/*
JMP Genomics 

Name: PLS_PCA
Author: Russ Wolfinger, SAS Institute Inc.
Support: Tzu-Ming Chu, SAS Institute Inc.
History:
   sasrdw 25Mar04 initial coding

Description: PLS_PCA contains SAS macro to utilize
Proc PLS.  
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/


%macro PLS_PCA(_outdata,_outdataname,_jslfile,_jslfilename,_forps=0,_annoflag=0);

%if &exiterror %then %goto exit;

data d;
	%if &_forps=0 %then %do;
		set InLib.&InData;
		%if %symexist(WhereStmt) and %length(&WhereStmt) %then %do;
			where &WhereStmt;
		%end;
	%end;
	%else %do;
		set &ds;
		%if %length(&WhereStmt) %then %do;
			where &WhereStmt;
		%end;
	%end;
run; 


%Nobsvars(d);
%if &syserr>4 or &nobs=0 %then %do;
	%put ERROR: Filter to Include Observations left no observations in the input data set, or is invalid.;
	%let exiterror=1;
	%goto exit;
%end;

/*---create variable lists and indicators for class variables---*/
%let NeedClassIndicators = Yes;
%if %index(&ListClassPredVars,:) %then %let ListClassColon = 1;
%else %let ListClassColon = 0;

/* %let PLSLimit = 32767000; */
%let PLSLimit = 2147483647;
%global ClassCheck ColumnLimit;
%let ClassCheck = Yes;
%NObsVars(d);
%if %upcase(&ProcessName)=PRINCIPALCOMPONENTSANALYSIS %then %let ColumnLimit = &nobs;
%else %let ColumnLimit = &PLSLimit;

%include "&MacroPath./ClassVars.sas";
%if &exiterror or &syserr>4 %then %goto exit;

/*---check for more than 32767 columns, set up transpose flag for PCA---*/
%global Transpose;
%let Transpose = No;
%global foundcn;
%let foundcn = 0; 
%if (%eval(&nix + &nip) > &ColumnLimit or &_forps=1) %then %do;
	%if %upcase(&ProcessName)=PARTIALLEASTSQUARES %then %do;
		%put ERROR: Partial Least Squares cannot handle more than &PLSLimit total predictor columns.  Please specify fewer predictors.;
		%let exiterror = 1;
		%goto exit;
	%end;
  %else %do;
  	%if (&nir > &PLSLimit) %then %do;
      	%put ERROR: PCA cannot handle matrices larger than &PLSLimit x &PLSLimit..  Please specify fewer variables.;
      	%let exiterror = 1;
      	%goto exit;
		%end;
    %else %do;
      	%put Transposing &nir x %eval(&nix + &nip) matrix for PCA efficiency.;
      	%let Transpose = Yes;
      	
	* use other variables as experimental design ds;
	%if ^%length(&ExpInData) %then %do;
		%let ExpInData = ed;
		%let ExpColorVars = &ColorVar;
		%let ExpLabelVar = &LabelVar;
		%let ColorVar = ;
		%let LabelVar = ;
		%let checkcn = 1;
		data ed;
			set d;
			%if &_forps=0 %then %do;
				drop 
					%if &PredictorVars = _SetToAllCol_ %then %do;
						%do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;
					%end;
					%else %do;
						&PredictorVars
					%end;
					&ClassPredVars
				;
			%end;
			%else %do;
				%if %length(&ExpColorVars &ExpLabelVar &TraitVar /*&ToKeepVars*/) %then %do;
					keep &ExpColorVars &ExpLabelVar &TraitVar /*&ToKeepVars*/;
				%end;
				%else	%do;
				/* Just create ColumnName here so have a variable to keep in ed */
					ColumnName = "i" || trim(left(_n_));
					keep ColumnName;
					%let checkcn = 0;
				%end;
			%end;
		run;

		* check if ColumnName already exists in order to save it;
		%if &checkcn = 1 %then %do;
			%let _ds_ = %sysfunc(open(ed));
			%let foundcn = %sysfunc(varnum(&_ds_,ColumnName));
			%let _rc_ = %sysfunc(close(&_ds_));
			%put foundcn = &foundcn;

			data ed;
				set ed;
				%if &foundcn %then %do;
					ColumnName0 = ColumnName;
				%end;
				ColumnName = "i" || trim(left(_n_));
			run;
		%end;

		proc sort data=ed;
			by ColumnName;
		run;
		%let SpecDesign = No;
	%end;
	%else %do;
		%let SpecDesign = Yes;
	%end;

      	* append transposed continuous and class variables;
      	%if (&nix > 0) %then %do;
      	
      		/* Have to do in blocks to prevent Out of Memory in Proc Transpose */
				%let BlockSize=%sysfunc(min(&nix,50000));
				%let first = 1;
				%let last = &BlockSize;

				data _null_;
					/* Make sure PredictorVars are in consecutive variables */
					set d(drop=
						%if &PredictorVars = _SetToAllCol_ %then %do;
							%do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;
						%end;
						%else %do;
							&PredictorVars
						%end;
					);
					set d(keep=
						%if &PredictorVars = _SetToAllCol_ %then %do;
							%do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;
						%end;
						%else %do;
							&PredictorVars
						%end;
					);
					array getname{*} 
						%if &PredictorVars = _SetToAllCol_ %then %do;
							%do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;
						%end;
						%else %do;
							&PredictorVars
						%end;
					;
					%let Block=1;
					%do %until (&first > &nix);	

						_f = vname(getname(&first));
						_l = vname(getname(&last));

						call symput("firstv&Block",_f);
						call symput("lastv&Block",_l);

						%let first = %eval(&last + 1);
						%let last = %eval(&first + &BlockSize - 1);
						%let Block = %eval(&Block + 1);
						%if &last > &nix %then %let last = &nix;
					%end;
				run;

				%let NBlocks = %eval(&Block-1);	

				data %do b=1 %to &NBlocks; sub_trans&b(keep=&&firstv&b--&&lastv&b) %end;;
					/* Make sure PredictorVars are in consecutive variables */
					set d(drop=
						%if &PredictorVars = _SetToAllCol_ %then %do;
							%do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;
						%end;
						%else %do;
							&PredictorVars
						%end;
					);
					set d(keep=
						%if &PredictorVars = _SetToAllCol_ %then %do;
							%do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;
						%end;
						%else %do;
							&PredictorVars
						%end;
					);
				run;

				%do b=1 %to &NBlocks;
					proc transpose data=sub_trans&b out=trans_ prefix=i;
						var &&firstv&b--&&lastv&b;
					run;

					%if &syserr>4 %then %let exiterror=1;

					proc datasets nolist;
						%if &b>1 %then %do;
							append base=d data=trans_ force;
						%end;
						%else %do;
							delete d;
							change trans_=d;
						%end;
						delete sub_trans&b;
					quit;

				%end;    	
      	
      	/************old call**
  					proc transpose data=d out=d prefix=i;
						var &PredictorVars;
					run;
			************/
					%if (&nip > 0) %then %do;
						proc append base=d data=opc force;
						run;
					%end;
      	%end;
      	%else %do;
      		proc datasets library=work nolist;
      			delete d;
      			change opc=d;
      		run;
      	%end;

				%let _rc_ = %sysfunc(close(&_dsix_));
				
				data dsix;
					set d(obs=0);
					keep i1-i&nir;
				run;
				
				%let _dsix_ = %sysfunc(open(dsix));	
				%do v=1 %to &nir;
					%if %sysfunc(vartype(&_dsix_,&v))=C %then %do;
						%let exiterror=1;
						%if &_forps=1 %then %do;
							%put ERROR: One or more of the Marker Variables are not numeric.;
							%goto psexit;
						%end;
						%else %do; 	
							%put ERROR: One or more of the Continuous Variables are not numeric.;
							%goto exit;
						%end;
					%end;
				%end;
					
				%let nix = &nir;
				%let nip = 0;

				%if &exiterror or &syserr>4 %then %goto exit;
     	
      	%let Temp = &CenterRows;
      	%let CenterRows = &CenterCols;
      	%let CenterCols = &Temp;
      	
      	%let Temp = &ScaleRows;
      	%let ScaleRows = &ScaleCols;
      	%let ScaleCols = &Temp;
      %end;
   %end;
   %if &exiterror or &syserr>4 %then %goto exit;
%end;
%else %let SpecDesign = No;
%let NeedPLS = Yes;

/*---centering and scaling for PCA---*/
%if &ProcessName=PrincipalComponentsAnalysis %then %do;

	%let ProcPLSOptions0 = &ProcPLSOptions;

	%if &CenterRows=No and &CenterMult=No and &ScaleRows=No %then %do;
		%if &CenterCols=No %then %let ProcPLSOptions = &ProcPLSOptions NOCENTER;
		%if &ScaleCols=No %then %let ProcPLSOptions = &ProcPLSOptions NOSCALE;
	%end;
	%else %do;
	
		%let ProcPLSOptions = &ProcPLSOptions NOCENTER NOSCALE;
		
		%if &CenterCols=Yes or &CenterMult=Yes %then %do;
			proc summary data=d noprint; 
				var 
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
				;
				output out=dmean(drop=_type_ _freq_)
					mean(
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
					) = 
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
				;
			run;
			data dmean;
				set dmean;
				%do i=1 %to &nix; 
					call symput("_colmean_x&i",trim(left(%sysfunc(varname(&_dsix_,&i)))));
				%end; 
				%do i=1 %to &nip; 
					call symput("_colmean_p&i",trim(left(%sysfunc(varname(&_dsip_,&i)))));
				%end; 
			run;
		%end;
		%else %do;
			%do i=1 %to &nix; %let _colmean_x&i = 0; %end; 
			%do i=1 %to &nip; %let _colmean_p&i = 0; %end; 
		%end;

		%if &ScaleCols=Yes %then %do;
			proc summary data=d noprint; 
				var 
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
				;
				output out=dstd(drop=_type_ _freq_)
					std(
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
					) = 
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
				;
			run;
			data dstd;
				set dstd;
				%do i=1 %to &nix; 
					call symput("_colstd_x&i",trim(left(%sysfunc(varname(&_dsix_,&i)))));
				%end; 
				%do i=1 %to &nip; 
					call symput("_colstd_p&i",trim(left(%sysfunc(varname(&_dsip_,&i)))));
				%end; 
			run;
		%end;
		%else %do;
			%do i=1 %to &nix; %let _colstd_x&i = 1; %end; 
			%do i=1 %to &nip; %let _colstd_p&i = 1; %end; 
		%end;

		data d;
			set d;
			
			%if &CenterRows=Yes or &CenterMult=Yes %then %do;
				_rowmean_ = mean(of 
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
				);
			%end;
			%else %do;
				_rowmean_ = 0;
			%end;
			_rown_ = n(of 
				%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
				%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
			);
			
			%if &ScaleRows=Yes %then %do;
				_rowstd_ = std(of 
					%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
					%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
				);
			%end;
			%else %if &_forps=1 %then %do;
				_rowstd_ = sqrt(_rowmean_/2 * (1 - _rowmean_/2));
			%end;
			%else %do;
				_rowstd_ = 1;
			%end;
		run;
		
		* compute overall mean for double centering;
		%if (&CenterCols=Yes and &CenterRows=Yes) or (&CenterMult=Yes) %then %do;
			proc summary data=d noprint; 
				var _rowmean_;
				weight _rown_;
				output out=dovm(drop=_type_ _freq_)
					mean(_rowmean_) = _rowmean_
				;
			run;
			data dovm;
				set dovm;
				%if (&CenterMult=Yes) %then %do;
					if abs(_rowmean_) < 1e-8 then _rowmean_ = 1;
				%end;
				call symputx("overall_mean",trim(left(_rowmean_)));
			run;
		%end;
		%else %do;
			%let overall_mean = 0;
		%end;	
		
		%put overall_mean = &overall_mean;

		* compute overall std for double scaling;
		%if (&ScaleCols=Yes and &ScaleRows=Yes) %then %do;
			proc summary data=d noprint; 
				var _rowstd_;
				weight _rown_;
				output out=dovm(drop=_type_ _freq_)
					mean(_rowstd_) = _rowstd_
				;
			run;
			data dovm;
				set dovm;
				if abs(_rowstd_) < 1e-8 then _rowstd_ = 1;
				call symputx("overall_std",trim(left(_rowstd_)));
			run;			
		%end;
		%else %do;
			%let overall_std = 1;
		%end;
		
		%put overall_std = &overall_std;
		
		data d;
			set d;

			%do i=1 %to &nix; 
				%if (&CenterMult=Yes) %then %do;	
					%sysfunc(varname(&_dsix_,&i)) = %sysfunc(varname(&_dsix_,&i)) - (_rowmean_ * &&_colmean_x&i / &overall_mean);
				%end;
				%else %do;
					%sysfunc(varname(&_dsix_,&i)) = %sysfunc(varname(&_dsix_,&i)) - _rowmean_ - &&_colmean_x&i + &overall_mean;
				%end;
				if _rowstd_ > 0 and &&_colstd_x&i > 0 then do;
					%sysfunc(varname(&_dsix_,&i)) = %sysfunc(varname(&_dsix_,&i)) / _rowstd_ / &&_colstd_x&i * &overall_std;
				end;
				/* Set missing to 0 for EIGENSTRAT */
				%if &_forps=1 %then %do;
					if missing(%sysfunc(varname(&_dsix_,&i))) then %sysfunc(varname(&_dsix_,&i)) = 0;
				%end;				
			%end; 			
			%do i=1 %to &nip; 
				%if (&CenterMult=Yes) %then %do;	
					%sysfunc(varname(&_dsip_,&i)) = %sysfunc(varname(&_dsip_,&i)) - (_rowmean_ * &&_colmean_p&i / &overall_mean);
				%end;
				%else %do;
					%sysfunc(varname(&_dsip_,&i)) = %sysfunc(varname(&_dsip_,&i)) - _rowmean_ - &&_colmean_p&i + &overall_mean;
				%end;
				if _rowstd_ and &&_colstd_p&i then do;
					%sysfunc(varname(&_dsip_,&i)) = %sysfunc(varname(&_dsip_,&i)) / _rowstd_ / &&_colstd_p&i * &overall_std;
				end;
			%end; 
		run;
		
	%end;	
	
	/*
	data outlib.d;
		set d;
	run;
	*/
	
	/*---use princomp if only need column scores---*/
	%if (&Transpose=No and %index(&Plot,Column)) or
		(&Transpose=Yes and %index(&Plot,Row)) %then %do;
		
		%if %index(%upcase(&ProcPLSOptions),NOCENTER) %then %do;
			%if %index(%upcase(&ProcPLSOptions),NOSCALE) %then %do;
				%let ProcPrincompOptions = NOINT COV;
				%let Type = UCOV;
				%let SType = USCORE;
			%end;
			%else %do;
				%let ProcPrincompOptions = NOINT;
				%let Type = UCORR;
				%let SType = USCORE;
			%end;
		%end;
		%else %if %index(%upcase(&ProcPLSOptions),NOSCALE) %then %do;
			%let ProcPrincompOptions = COV;
			%let Type = COV;
			%let SType = SCORE;
		%end;
		%else %do;
			%let ProcPrincompOptions = ;
			%let Type = CORR;
			%let SType = SCORE;
		%end;
		
		proc princomp data=d outstat=pc n=&NumberOfComponents &ProcPrincompOptions;
			var 
				%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
				%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
			;
			ods select eigenvalues eigenvectors nobsnvar;
			%global eigdata;
			%if &_forps=0 %then %do;
				ods output eigenvalues=OutLib.%substr(&OrigInData,1,23)_eigenval;
				%let EigData=%substr(&OrigInData,1,23)_eigenval;
			%end;
			%else %do;
				ods output eigenvalues=OutLib.%substr(&OutName,1,23)_eigenval;
				%let EigData=%substr(&OutName,1,23)_eigenval;
			%end;			
		run;
		
		%if &syserr>4 %then %goto exit;

		%if &_forps=0 %then %do;
			proc print data=pc;
			run;
		%end;
		
		data _null_;
			set OutLib.&EigData;
			currentyvariation = 100*proportion;
			if currentyvariation >= 9.95 then do;
				call symput("prinlab"||trim(left(_n_)),"&prefix" || trim(left(_n_)) || ' (' || put(currentyvariation,4.1) || '%)');
			end;
			else do;
				call symput("prinlab"||trim(left(_n_)),"&prefix" || trim(left(_n_)) || ' (' || put(currentyvariation,3.1) || '%)');		
			end;
		run;

		data xw;
			set pc;
			if _type_ = "&SType";
			drop _name_;
		run;
		
		%let NeedPLS = No;
	%end;
%end;
%else %do;
	%if %index(%upcase(&ProcPLSOptions),NOCENTER) %then %let CenterCols = No;
	%else %let CenterCols = Yes;
	%if %index(%upcase(&ProcPLSOptions),NOSCALE) %then %let ScaleCols = No;
	%else %let ScaleCols = Yes;
	%let CenterRows = No;
	%let CenterMult = No;
	%let ScaleRows = No;
%end;
		
%if &exiterror or &syserr>4 %then %goto exit;

/* Currently not translating these three -- appear to be in data table only. */
%let trRow = Row; /*%nrbquote(%LocaleGet(Row_Text));*/
%let trCol = Col; /*%nrbquote(%LocaleGet(Col_Text));*/
%let trColumn = Column; /*%nrbquote(%LocaleGet(Column_Text));*/

/* These two are shown in Tabbed Report results only. */
%let trRows = %nrbquote(%LocaleGet(Rows_Text));
%let trColumns = %nrbquote(%LocaleGet(Columns_Text));

/*---labels for JSL---*/
%if (&Transpose = No) %then %do;
	%let Lab1 = %nrbquote(&trRow);
	%let Lab2 = %nrbquote(&trCol);
	%let Lab3 = %nrbquote(&trRow);
	%let Lab3s = %nrbquote(&trRows);
	%let Lab4 = %nrbquote(&trColumn);
	%let Lab4s = %nrbquote(&trColumns);
%end;
%else %do;
	%let Lab1 = %nrbquote(&trCol);
	%let Lab2 = %nrbquote(&trRow);
	%let Lab3 = %nrbquote(&trColumn);
	%let Lab3s = %nrbquote(&trColumns);
	%let Lab4 = %nrbquote(&trRow);
	%let Lab4s = %nrbquote(&trRows);
%end;

%let Lab1Score = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(XScore_Text)), %nrbquote(&Lab1)));

/*---pls computations---*/
%if (&NeedPLS = Yes) %then %do;
	%put nix = &nix;
	%put nip = &nip;

	%if (&nix=0) and (&nip=0) %then %do;
		%put ERROR: The specified variables do not exist in the input data set.;
		%let exiterror= 1;
		%goto exit;
	%end;
		
	proc pls data=d nfac=&NumberOfComponents &ProcPLSOptions details missing=avg /* missing=em */;
		model 
			%if %upcase(&ProcessName)=PARTIALLEASTSQUARES %then %do;
				%do i=1 %to &nic; &&_c&i %end; 
				%do i=1 %to &nit; &&_t&i %end; 
			%end;
			%else %do;
				%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
				%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
			%end;
			=  
			%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
			%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
		;
		output out=&&&_outdata(drop=
				%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end;
				%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
			)
			%if %upcase(&ProcessName)=PARTIALLEASTSQUARES  %then %do; 
				predicted= 
					%do i=1 %to &nic; pred_&&_c&i %end;
					%do i=1 %to &nit; pred_&&_t&i %end;
			%end; 
			xscore=&prefix h=leverage tsquare=tsquare
		;
		ods exclude xloadings xweights yweights;
		ods output xweights=xw percentvariation=pv;
	run;
	
	%if ^%sysfunc(exist(xw)) %then %do;
		%put WARNING: No components selected.;
		%goto exit;
	%end;

	data _null_;
		set pv;
		if currentyvariation >= 9.95 then do;
			call symput("prinlab"||trim(left(_n_)),"&prefix" || trim(left(_n_)) || ' (' || put(currentyvariation,4.1) || '%)');
		end;
		else do;
			call symput("prinlab"||trim(left(_n_)),"&prefix" || trim(left(_n_)) || ' (' || put(currentyvariation,3.1) || '%)');		
		end;
	run;

	data &&&_outdata;
		length _ScoreType $ 10 _Label $ 32;
		set &&&_outdata;
		_ScoreType = "&Lab1";
		%if %length(&ColorVar) %then %do;
			_Color = trim(left(&ColorVar));
		%end;
		%else %do;
			_Color = "&Lab1";
		%end;
		%if %length(&LabelVar) %then %do;
			_Label = trim(left(&LabelVar));
		%end;
		%else %do;
			_Label = "&Lab1Score";
		%end;	
		label _Color=' ' _Label=' '
			%do i=1 %to &NumberOfComponents;
				&prefix.&i = "&&prinlab&i" 
			%end;
		;	
	run;
	%if &_AnnoFlag %then %do;
	   %MergeAnno(&&&_outdata);
	%end;
%end;


%NObsVars(xw);
%let NumberOfComponents = &nobs;
proc transpose data=xw out=xwt prefix=&prefix;
   var 
		%do i=1 %to &nip; %sysfunc(varname(&_dsip_,&i)) %end; 
		%do i=1 %to &nix; %sysfunc(varname(&_dsix_,&i)) %end; 
	;
run;
%let _rc_ = %sysfunc(close(&_dsix_));

data xwt;
   set xwt;
   _ScoreType = "&Lab2";
   ColumnName = _name_;
   drop _name_;
run;

* merge experimental design metadata with column scores;
%if %length(&ExpInData) %then %do;

	%if (&Transpose=Yes) and (&SpecDesign = Yes) %then %do;
		%let dsm = &&&_outdata;
		data &dsm;
			set &dsm;
			rename _name_ = ColumnName;
		run;
		%let foundcn = 0;
		data xwt;
			set xwt;
			_Color = "&Lab2";
			_Label = ColumnName;
			drop ColumnName;
		run;
	%end;
	%else %do;
		%let dsm = xwt;
	%end;

	proc sort data=&dsm;
		by ColumnName;
	run;
	
	%CheckLengthForMerging(ed,&dsm,ColumnName);
	
	data &dsm;
		merge ed &dsm;
		by ColumnName;
	run;

	%if &foundcn %then %do;
		data &dsm;
			set &dsm;
			%if &_forps=1 %then %do;
				OldColName = ColumnName;
			%end;
			ColumnName = ColumnName0;
			drop ColumnName0;
		run;
	%end;
	
	%if ^%length(&ExpColorVars) %then %let ExpColorVars = ColumnName;
	%if %length(&ExpLabelVar) %then %let labvar = &ExpLabelVar;
	%else %let labvar = ColumnName;
	%let i = 1;
	%do %while(%length(%scan(&ExpColorVars,&i)));
	   %let ExpColorVar&i = %scan(&ExpColorVars,&i);
		%let i = %eval(&i+1);
	%end;
	%let nExpColorVar = %eval(&i-1);
	
	data &dsm;
		set &dsm;
		_Color = trim(left(&ExpColorVar1))
					%do i = 2 %to &nExpColorVar;
					   ||"_"|| trim(left(&&ExpColorVar&i)) 
					%end; ;
		_Label = trim(left(&labvar));
		if _ScoreType = " " then delete;
	run;
%end;
%else %do;
	data xwt;
		set xwt;
		_Color = "&Lab2";
		_Label = ColumnName;
		drop ColumnName;
	run;
%end;

* standardize components for plotting;
%if %index(&Plot,Biplot) %then %do;
	proc stdize data=&&&_outdata out=&&&_outdata method=std;
		var &prefix.1-&prefix.&NumberOfComponents;
	run;

	proc stdize data=xwt out=xwt method=std;
		var &prefix.1-&prefix.&NumberOfComponents;
	run;
%end;

%let uprefix = %upcase(&prefix);
%if %index(&Plot,&Lab4) %then %do;
	data &&&_outdata;
		set xwt;
		label
			%do i=1 %to &NumberOfComponents;
				&prefix.&i = "&&prinlab&i" 
			%end;
		;	
	run;
	%let winname = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X2DXScores_Text)), &uprefix, &Lab4));
	%let win3name = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X3DXScores_Text)), &uprefix, &Lab4));
%end;
%else %if %index(&Plot,Both) %then %do;
	data &&&_outdata;
		set &&&_outdata xwt;
	run;
   %if %index(&Plot,Separate) %then %do;
		%let winname = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X2DXScores_Text)), &uprefix, %nrbquote(&trRow)));
		%let winname2 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X2DXScores_Text)), &uprefix, %nrbquote(&trColumn)));
		%let win3name = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X3DXScores_Text)), &uprefix, %nrbquote(&trRow)));
		%let win3name2 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X3DXScores_Text)), &uprefix, %nrbquote(&trColumn)));
	%end;
	%else %do;
		%let winname = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X2DBiplots_Text)), &uprefix));
		%let win3name = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X3DBiplot_Text)), &uprefix));
	%end;
%end;
%else %do;
	%let winname = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X2DXScores_Text)), &uprefix, %nrbquote(&Lab3)));
	%let win3name = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(X3DXScores_Text)), &uprefix, %nrbquote(&Lab3)));
%end;

%put syserr = &syserr;

%if &syserr > 4 %then %goto exit;

%if &_forps=1 %then %do;

	%if %length(&ToKeepVars) %then %do;
	
		/*data OutLib.PCA_Out0; set OutLib.&PCA_OutDataName; run;*/
		
		options label;
		data OutLib.&PCA_OutDataName;
			set OutLib.&PCA_OutDataName;
			%if &foundcn>0 %then %do;
				_ind=substr(OldColName,2)+0;
			%end;
			%else _ind=substr(ColumnName,2)+0;;
		run;
		
		options label;
		proc sort data=OutLib.&PCA_OutDataName out=OutLib.&PCA_OutDataName(drop=_ind);
			by _ind;
		run;		

		/*data OutLib.pca_ds_mg; set &ds_mg(keep= &ToKeepVars); run;*/
		/*data OutLib.PCA_Out1; set OutLib.&PCA_OutDataName; run;*/
		
		options label;
		data OutLib.&PCA_OutDataName;
			set &ds_mg(keep= &ToKeepVars);
			set OutLib.&PCA_OutDataName;
		run;
	%end;

 	%PackageInsertDataset(PCA data from PCAforPS, OutLib, &PCA_OutDataName);
%end;

%if ^%symexist(DisplayPlots) %then %let DisplayPlots = Yes;

%global PCAIncludeDS;
%let PCAIncludeDS =;

%if %upcase(&DisplayPlots)=NO %then %return;;

* change signs if transposed for consistency with non-transposed;
/*
%if (&Transpose=Yes) %then %do;
	data &&&_outdata;
		set &&&_outdata;
		%if &NumberOfComponents >= 2 %then %do;
			pca2 = -pca2;
		%end;
		%if &NumberOfComponents >= 5 %then %do;
			pca5 = -pca5;
		%end;
	run;	
%end;
*/

%let trNone = %nrbquote(%LocaleGet(None_Text));

%let ttext1 = %nrbquote(%LocaleGet(Centering_Text): );
%if (&CenterMult=Yes) %then 
	%let ttext2 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(BothXandXMultiplicatively_Text)), %nrbquote(&trRows), %nrbquote(&trColumns)));
%else %if (&CenterRows=Yes) and (&CenterCols=Yes) %then 
	%let ttext2 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(BothXandXAdditively_Text)), %nrbquote(&trRows), %nrbquote(&trColumns)));
%else %if (&CenterRows=Yes) %then
	%let ttext2 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(XOnly_Text)), %nrbquote(&Lab3s)));
%else %if (&CenterCols=Yes) %then
	%let ttext2 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(XOnly_Text)), %nrbquote(&Lab4s)));
%else 
	%let ttext2 = %nrbquote(&trNone);
	
%let ttext3 = %nrbquote(%LocaleGet(Scaling_Text): );

%if (&ScaleRows=Yes) and (&ScaleCols=Yes) %then 
	%let ttext4 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(BothXandX_Text)), %nrbquote(&trRows), %nrbquote(&trColumns)));
%else %if (&ScaleRows=Yes) %then
	%let ttext4 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(XOnly_Text)), %nrbquote(&Lab3s)));
%else %if (&ScaleCols=Yes) %then
	%let ttext4 = %nrbquote(%LocaleFill(%nrbquote(%LocaleGet(XOnly_Text)), %nrbquote(&Lab4s)));
%else 
	%let ttext4 = %nrbquote(&trNone);

	
%let ttext = &ttext1 &ttext2 \!N&ttext3 &ttext4;

%global _mainjslfile_ uID;
%let _mainjslfile_ = %quote(&&&_JSLFile);
%let uID = U&sysjobid;

* put JSL statements;
data _null_;
   file "&&&_JSLFile" encoding="utf-8";
   put "//!";
	put "Names Default to Here(1);";
	
	%if (%scan(&plot,1)=Both | %scan(&plot,1)=Row) %then %do;
   	%if &_AnnoFlag %then %do;
			put " ";
   	   %include "&MacroPath./WebSearchJSL.sas";
   	   %WebSearchJSL(Ds=,clientPath=,MakeWindow=0);
			put " ";
   		%if %length(&AnnoAccession) %then %do;
				put 'Here:ActionButtonDesc = {"Select Points in Graphs, then Click a Button"};';
   		   %WebButtons;
				%WebButtons_individual;
   		%end;
   	%end;
  	%end;
  	
  	put "jg:drill_down = 0;";
  	put "dNS&uid = New Namespace();";

   put "pcmv = Function({},{default local},";
   put "if(isEmpty(dNS&uid:pcadata),";
   put "%str(dNS&uid:pcadata=open(%"&ClientOutPath.&&&_outdataname...sas7bdat%", invisible);)";
	put "dNS&uid:pcadata<<Color By Column(:_Color);";
	%if %index(&Plot,Biplot) %then %do;
	 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trRow"");";
	 put "dNS&uid:pcadata << Markers(3);";
	 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trCol"");";
	 /* put "pcadata<<Marker By Column(:_ScoreType);"; */
	%end; 
	%else %do;
	   put "dNS&uid:pcadata << Select Where(1);";
	%end;
	put "dNS&uid:pcadata << Markers(8);";
	put "dNS&uid:pcadata << Clear Select;";
	put 'column("_Label") << label(1);';
	put ", //else";
	put "Current Data Table(dNS&uid:pcadata);";
	put ");"; 
   put " ";
   put "Vlistbox(";
   put "TextBox(""&ttext""),";
   put "Multivariate(Y( ";
   %do i=1 %to %eval(&NumberOfComponents-1); 
		put "   Column(""&&prinlab&i""),";
   %end;
	put "   Column(""&&prinlab&NumberofComponents"")),";
   %if %index(&Plot,Separate) %then %do;
   	put "Where(:_ScoreType==""&trRow""),";
   %end;
   put "Scatterplot Matrix(Ellipse Color(1), Horizontal(1)),";
   put "Parallel Coord Plot(1),"; 
   put "Outlier Analysis"; 
   /*
   put "SendToReport(";
   %do i=1 %to %eval(&NumberOfComponents * &NumberOfComponents);
      put "Dispatch({";
      put '"Scatterplot Matrix"';
      put "},";
      put '"Multiv Plot",';
      put "FrameBox(&i), {Marker Size(3), Marker Drawing Mode(Outlined)}),";
   %end;
   put "Dispatch({";
   put '"Outlier Analysis","Mahalanobis Distances"';
   put "},";
   put '"Multiv Outlier",';
   put "FrameBox, {Marker Size(3), Marker Drawing Mode(Outlined)})";
   put ");";
   */
   put ")));";
   put " ";
   
   /* change these 4 vars back to Here: */
	put "Here:DataList = {{""dNS&uid:pcadata""}};";
	put "Here:TabButtonList = {""&winname""};";
	put "Here:TabCodeList = {pcmv()};";
	put "Here:TabAfterCodeList = {"" ""};";
	put "Here:TabOpenList = {1};";
	
	%if &Include3D = Yes and &NumberOfComponents > 2 %then %do;
		put "pc3d = Function({},{default local},";
		put "if(isEmpty(dNS&uid:pcadata),";
		put "%str(dNS&uid:pcadata=open(%"&ClientOutPath.&&&_outdataname...sas7bdat%", invisible);)";
		put "dNS&uid:pcadata<<Color By Column(:_Color);";
		%if %index(&Plot,Biplot) %then %do;
		 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trRow"");";
		 put "dNS&uid:pcadata << Markers(3);";
		 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trCol"");";
		 /* put "pcadata<<Marker By Column(:_ScoreType);"; */
		%end; 
		%else %do;
			put "dNS&uid:pcadata << Select Where(1);";
		%end;
		put "dNS&uid:pcadata << Markers(8);";
		put "dNS&uid:pcadata << Clear Select;";
		put 'column("_Label") << label(1);';
		put ", //else";
		put "Current Data Table(dNS&uid:pcadata);";
		put ");"; 
		put "Vlistbox(";
	   put "TextBox(""&ttext""),";
		put "Scatterplot 3D(";
		put "   Y(Column(""&prinlab1""),Column(""&prinlab2""),Column(""&prinlab3"")),";
		%if %index(&Plot,Separate) %then %do;
		   put "   Where(:_ScoreType==""&trRow""),";
		%end;
		put "   Frame3D(";
		put "      Set Marker Transparency( 0.6 ),";
		put "      Set Marker Quality( 1 ),";
		put "      Set Marker Scale( 2.5 )";
		put "   )";
		put ")));";
		
		put "InsertInto(DataList,{{""dNS&uid:pcadata""}});";
		put "InsertInto(TabButtonList, ""&win3name"");";
		put "InsertInto(TabCodeList, expr(pc3d()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
		
	%end;
	%if &_forps=1 %then %do;
		put 'pcScree = Function({},{default local},';
		put "if(isEmpty(dNS&uid:eigdata),";
		put "dNS&uid:eigdata=open(""&ClientOutPath.&EigData..sas7bdat"", invisible);";
		put ", //else";
		put "Current Data Table(dNS&uid:eigdata);";
		put ");"; 
		put "	Column(""Number"") << Modeling Type(ordinal);"; 
		put "   Vlistbox(";
		/*
		put "   	Graph Builder(";
		put "   		Show Control Panel( 0 ),";
		put "   		Show Legend( 0 ),";
		put "   		Variables( X( :Number ), Y( :Eigenvalue ) ),";
		put "   		Elements( Line( X, Y, Legend( 4 ) ), Points( X, Y, Legend( 1 ) ) ),";
		put "   		SendToReport(";
		put "   			Dispatch(";
		put "   				{},";
		put "   				""Graph Builder"",";
		put "   				OutlineBox,";
		put "   				{Set Title( ""Scree Plot"" ), Image Export Display( Normal )}";
		put "   			),";
		put "   			Dispatch( {}, ""graph title"", TextEditBox, {Set Text( """" )} ),";
		put "   			Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""Component Number"" )} )";
		put "   		)";
		put "   	),";
		*/
		put "   	Graph Builder(";
		put "   		Show Control Panel( 0 ),";
		put "   		Legend Position( ""Bottom"" ),";
		put "   		Variables(";
		put "   			X( :Number ),";
		put "   			Y( :Cumulative, Side( ""Right"" ) ),";
		put "   			Y( :Proportion, Position( 1 ) )";
		put "   		),";
		put "   		Elements(";
		put "   			Line( X, Y( 2 ), Legend( 4 ) ),";
		put "   			Points( X, Y( 2 ), Legend( 1 ) ),";
		put "   			Line( X, Y( 1 ), Legend( 5 ) ),";
		put "   			Points( X, Y( 1 ), Legend( 6 ) )";
		put "   		),";
		put "   		SendToReport(";
		put "   			Dispatch(";
		put "   				{},";
		put "   				""Graph Builder"",";
		put "   				OutlineBox,";
		put "   				{Set Title( ""Scree Plot"" ), Image Export Display( Normal )}";
		put "   			),";
		put "   			Dispatch( {}, ""graph title"", TextEditBox, {Set Text( """" )} ),";
		put "   			Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""Component Number"" )} ),";
		put "   			Dispatch(";
		put "   				{},";
		put "   				""400"",";
		put "   				LegendBox,";
		put "   				{Orientation( ""Horizontal"" ), Sides( ""Left"" )}";
		put "   			)";
		put "   		)";
		put "   	);";
		put "   ),";
		put ");";		
		put "InsertInto(DataList,{{""dNS&uid:eigdata""}});";
		put "InsertInto(TabButtonList, ""Scree Plot"");";
		put "InsertInto(TabCodeList, expr(pcScree()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";		
	%end;	

   %if %index(&Plot,Separate) %then %do;
   	put "pcsepplot = Function({},{default local},";
		put "if(isEmpty(dNS&uid:pcadata),";
		put "%str(dNS&uid:pcadata=open(%"&ClientOutPath.&&&_outdataname...sas7bdat%", invisible);)";
		put "dNS&uid:pcadata<<Color By Column(:_Color);";
		%if %index(&Plot,Biplot) %then %do;
		 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trRow"");";
		 put "dNS&uid:pcadata << Markers(3);";
		 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trCol"");";
		 /* put "pcadata<<Marker By Column(:_ScoreType);"; */
		%end; 
		%else %do;
			put "dNS&uid:pcadata << Select Where(1);";
		%end;
		put "dNS&uid:pcadata << Markers(8);";
		put "dNS&uid:pcadata << Clear Select;";
		put 'column("_Label") << label(1);';
		put ", //else";
		put "Current Data Table(dNS&uid:pcadata);";
		put ");"; 	   
		put "Vlistbox(";   	
	   put "TextBox(""&ttext""),";
	   put "Multivariate(Y( ";
		%do i=1 %to %eval(&NumberOfComponents-1); 
			put "   Column(""&&prinlab&i""),";
		%end;
		put "   Column(""&&prinlab&NumberOfComponents"")),";

		put "Where(:_ScoreType==""&trCol""),";
		put "Scatterplot Matrix(Ellipse Color(1), Horizontal(1)),";
		put "Parallel Coord Plot(1),"; 
		put "Outlier Analysis"; 
		/*
		put "SendToReport(";
		%do i=1 %to %eval(&NumberOfComponents * &NumberOfComponents);
			put "Dispatch({";
			put '"Scatterplot Matrix"';
			put "},";
			put '"Multiv Plot",';
			put "FrameBox(&i), {Marker Size(3), Marker Drawing Mode(Outlined)}),";
		%end;
		put "Dispatch({";
		put '"Outlier Analysis","Mahalanobis Distances"';
		put "},";
		put '"Multiv Outlier",';
		put "FrameBox, {Marker Size(3), Marker Drawing Mode(Outlined)})";
		put ");";
		*/
		put ")));";
		put "InsertInto(DataList,{{""dNS&uid:pcadata""}});";
		put "InsertInto(TabButtonList, ""&winname2"");";
		put "InsertInto(TabCodeList, expr(pcsepplot()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";			
		
		put " "; 
		%if &Include3D = Yes and &NumberOfComponents > 2 %then %do;
	   	put "pc3dsepplot = Function({},{default local},";
			put "if(isEmpty(dNS&uid:pcadata),";
			put "%str(dNS&uid:pcadata=open(%"&ClientOutPath.&&&_outdataname...sas7bdat%", invisible);)";
			put "dNS&uid:pcadata<<Color By Column(:_Color);";
			%if %index(&Plot,Biplot) %then %do;
			 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trRow"");";
			 put "dNS&uid:pcadata << Markers(3);";
			 put "dNS&uid:pcadata << Select Where(:_ScoreType==""&trCol"");";
			 /* put "pcadata<<Marker By Column(:_ScoreType);"; */
			%end; 
			%else %do;
				put "dNS&uid:pcadata << Select Where(1);";
			%end;
			put "dNS&uid:pcadata << Markers(8);";
			put "dNS&uid:pcadata << Clear Select;";
			put 'column("_Label") << label(1);';
			put ", //else";
			put "Current Data Table(dNS&uid:pcadata);";
			put ");"; 
			put "Vlistbox(";	   	
		   put "TextBox(""&ttext""),";
		   put "Scatterplot 3D(";
			put "   Y(Column(""&prinlab1""),Column(""&prinlab2""),Column(""&prinlab3"")),";
			put "   Where(:_ScoreType==""&trCol""),";
			put "   Frame3D(";
			put "      Set Marker Transparency( 0.6 ),";
			put "      Set Marker Quality( 1 ),";
			put "      Set Marker Scale( 2.5 )";
			put "   )";
			put ")));";
			put "InsertInto(DataList,{{""dNS&uid:pcadata""}});";
			put "InsertInto(TabButtonList, ""&win3name2"");";
			put "InsertInto(TabCodeList, expr(pc3dsepplot()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);";			
		%end;
	%end;
   
   /*---extra code for creating separate bivariate plots of components---*/
   /*
   %do i=1 %to %eval(&NumberOfComponents-1); 
      %do j=%eval(&i+1) %to &NumberOfComponents; 
         put "New Window(";
         put "%str(%"Partial Least Squares &i Versus &j%",)";
         put "Bivariate(X(:&prefix.&i),Y(:&prefix.&j),";
		 put "SendToReport(Dispatch({},";
         put '"Bivar Plot",';
         put "FrameBox, {Marker Size(4), Marker Drawing Mode(Outlined)}))";
         put "));";
     %end;
   %end;
   */
run;

%let dopackage = 1;
%global PCAIncludeDS;
%let PCAIncludeDS = &&&_outdataname...sas7bdat;
%AddDatatoStudy(&&&_outdataname,_type=other);		

%if &_forps = 1 %then %do;
	%let PCAIncludeDS = &PCAIncludeDS, &EigData..sas7bdat;
	%AddDatatoStudy(&EigData,_type=other);			
	%if %length(&TraitVar) %then %let dopackage = 0;
%end;

%if &dopackage=1 %then %PackageInsertFile(jsl code from &ProcessName, &OutPath, &&&_JSLFileName,"&PCAIncludeDS", text/plain, yes);

%psexit:;

%if &_forps=1 %then %return;;

%exit:;

%if &_forps=1 %then %let ProcessName = PCAforPS;

%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend;
