/* SAS Macros to be used in Linkage Mapping and QTL Processes */
/* Support: Kelci Miclaus (kemicl)
	History:
		Initial Macros Written by Shengchu Wang
		kemicl 02Mar11 Add RecodeGenos Macro to do genotype recoding following new format of . 0 1 2 3 4
		kemicl 22Apr11 Add OldGenoCheck Macro
		kemicl 08Feb12 Add HaleyKnott Regression and Permutation Macro
		Luciano 2013 changed the grid of search to include markers
		Luciano 2013 changed estimates coefficients to match other softwares
		Luciano 2014 moved the code to estimate recombination frequency from LinkageGroups.sas to here, 
			because it is need to add dropped markers back in the Linkage Map Order AP. 
*/

%macro haleyknott(nx=,ng=,perm=);
	%if &ng=2 %then %do;
		%let geffect = p1;
		%let pestimate1 = p1 -1;
	%end;
	%else %if &ng=3 %then %do;
		%let geffect = p1 p2;
		%let pestimate1 = p1 -.5 p2 0;
		%*let pestimated = p1 -.333333333333333333333333 p2 .66666666666666666666666;
		%let pestimated = p1 -.5 p2 1; /*---Luciano---*/
	%end;
	%else %do; /*outcross*/
		%*let geffect = p1 p2 p3;
		%*let pestimate1 = p1 .25 p2  .25 p3 -.25;
		%*let pestimated = p1 .25 p2 -.25 p3 -.25;
		%*let pestimate2 = p1 .25 p2 -.25 p3  .25; 
		
		%let geffect = xa2_ xd_ xa1_;
		%let pestimate1 = xa2_ 1;
		%let pestimated = xd_ 1;	
		%let pestimate2 = xa1_ 1; 
	%end;
	
	%let xestimate = ;
	%do i = 1 %to &nx;
		%let xestimate = &xestimate x&i 0;
	%end;

	%if &perm %then %do;
		%let start = 1;
		%let PBsize = 100;
		%let nblocks = %sysevalf(&perm/&PBsize,ceil);
		%put nblocks = &nblocks;
		%let seed = 445322;
		/* create id for merging back in data for analysis with permuted traits*/
		data sma_p;
			set hkdata;
			*by trait chrom CM;
			by trait chrom CM CMID;/*---Luciano---*/
			retain newid;
			*if first.CM then newid =1;
			if first.CMID then newid =1;/*---Luciano---*/
			else newid+1;
			drop y;
		run;

		/*data OutLib.sma_p; set sma_p; run;*/
		/*data OutLib.TtFile; set &TtFile; run;*/

		/* generate permutation scheme for permuting Ys*/
		%NObsVars(&TtFile);
		%let ntobs = &nobs;
		proc plan %if %length(&Seed) %then seed=&Seed;;
			factors perm=&Perm ordered obs=&ntobs random/ noprint;
			output out=perms;
		run;

		/* traits data is referenced by TtFile macro */
		data  %do b=1 %to &nblocks; traits_p&b %end; ;
			set perms;
			by perm notsorted;
			set &TtFile point=obs;
			if first.perm then newid=1;
			else newid+1;
			%let i = 1;
			%do %while(%length(%scan(&TraitVar,&i)));
				y = %scan(&TraitVar,&i);
				trait = &i;
				if 1 <= perm <= &PBsize then output traits_p1;
				%do b=2 %to &nblocks;
					if %eval((&b-1)*&PBsize) < perm <= %eval(&b*&PBsize) then output traits_p&b;
				%end;
				%let i = %eval(&i+1);
			%end;
			drop &TraitVar;
		run;
	%end; /* end perm loop for setting up permuted y data sets*/
	%else %do;
		%let start = 0;
		%let nblocks = 0;
		data sma_tmp;
			set hkdata;
		run;
	%end;

	%do b = &start %to &nblocks;
		%if &perm %then %do;
			proc sql UNDO_POLICY=NONE; 
				create table sma_tmp(drop = tmp_t tmp_id newid) as
				select * 
				from traits_p&b as d1 full join sma_p(rename = (trait=tmp_t newid=tmp_id)) as d2
				on d1.trait = d2.tmp_t and d1.newid = d2.tmp_id
				order by perm, trait, chrom, CM, CMID, newid;/*---Luciano---*/
				*order by perm, trait, chrom, CM, newid;
			quit;
		%end;
		
		*ods html;/*luciano*/
		ods exclude all;
		ods noresults;
		proc mixed data=sma_tmp method=ml;
			*by %if &perm %then perm; trait chrom CM;
			by %if &perm %then perm; trait chrom CM CMID;/*---Luciano---*/
			%if &nx=0 %then %do;
				model y = &geffect ;
			%end;
			%else %if &nx=1 %then %do;
				model y = x1 &geffect ;
			%end;
			%else %do;
				model y = x1-x&nx &geffect ;
			%end;
			%if ^&perm %then %do;
				estimate "Additive2" &xestimate &pestimate1;
				%if &ng >= 3 %then %do;
					estimate "Dominance" &xestimate &pestimated;
				%end;
				%if &ng = 4 %then %do;
					estimate "Additive1" &xestimate &pestimate2;
				%end;
			%end;
			ods output fitstatistics=fs %if ^&perm %then estimates=ests;;
		run;

		proc mixed data=sma_tmp method=ml;
			/*by %if &perm %then perm; trait chrom CM;*/
			by %if &perm %then perm; trait chrom CM CMID;/*---Luciano---*/
			%if &nx=0 %then %do;
				model y = ;
			%end;
			%else %if &nx=1 %then %do;
				model y = x1;
			%end;
			%else %do;
				model y = x1-x&nx;
			%end;
			ods output fitstatistics=fs0;
		run;

		ods exclude none;
		ods results;

		data fs1;
			set fs;
			where descr = "-2 Log Likelihood";
		run;

		data fs01;
			set fs0;
			where descr = "-2 Log Likelihood";
			rename value=value0;
		run;
		
		data fsb;
			merge fs1 fs01;
			/*by %if &perm %then perm; trait chrom CM;*/
			by %if &perm %then perm; trait chrom CMID; /*---Luciano---*/
			/*kemicl calculate LOD like EM code does */
			lod = 0.217*(value0 - value);
			/*lod = log10(exp(-0.5*value)) - log10(exp(-0.5*value0));*/
			drop value value0 descr;
		run;
		
		/*
		proc print data=fsb(obs=10);
		run;
		*/
		proc sort data=fsb;
			*by trait chrom CM;
			by trait chrom CMID;/*---Luciano---*/
		run;

		%let fsdat = fsb;
		%if &perm %then %do;
			proc transpose data=fsb prefix=plod out=fsbt;
				*by trait chrom CM;
				by trait chrom CMID;/*---Luciano---*/
				var lod;
				id perm;
			run;
			%let fsdat = fsbt;
		%end;
		/*
		proc print data=fsbt(obs=10);
		run;
		*/
		data hk&b;
			set &fsdat;
		run;
	%end;

	%if &perm %then %do;
		/* compute 100-alpha */
		%let apct = %sysevalf(100*(1-&Alpha));
		/* merge all permutation LODs together and compute permutation p-values*/
		data hk_perm(drop =plod: p _name_ permprobl nperml) allprobs(keep = trait plod:);
			set hk0; %do b= 1 %to &nblocks; set hk&b; %end;
			array pps{*} plod1-plod&Perm;
			PermProbL = 0;
			NPermL = 0;
			if Lod=. then ProbLODPerm = .;
			else do;
				do p=1 to dim(pps);
					if LOD <= pps{p} then PermProbL+1;
					if pps{p} ne . then NPermL+1;
				end;
				ProbLODPerm = (PermProbL +1)/ (NPermL + 1);
			end;
			/* compute comparison-wise LOD threshold per  Churchill and Doerge 1994 */
			cw_LODthresh = PCTL(&apct,of plod1-plod&Perm); 
		run;
		
/* compute experiment-wise LOD threshold per Churchill and Doerge 1994 */
/* kemicl compute it by trait, should this be across all traits? */
		%let i = 1;
		%do %while(%length(%scan(&TraitVar,&i)));
		 	%global EWLODThresh&i;
		 	%let i = %eval(&i+1);
		%end;
		
		proc means data=allprobs max noprint;
			by trait;
			var plod:;
			output out=pLODmax max =/autoname;
		run;
		data _null_;
			set pLODMax;
			ew_LODthresh = PCTL(&apct,of plod:);
			ew_LODthresh = round(ew_LODthresh,0.0001);
			call symputx("EWLODThresh"||trim(left(trait)), ew_LODthresh);
		run;
		
/* split data for output format */
		data hkout; 
			%let i =1;
			%do %while(%length(%scan(&TraitVar,&i)));
				%let CurrT = %scan(&TraitVar,&i);
				%put Experimentwise threshold for &CurrT = &&EWLODThresh&i;
				set hk_perm(rename=(lod=LOD&i Additive2=Additive2&i 
							%if &ng >= 3 %then Dominance = Dominance&i; 
							%if &ng = 4 %then Additive1 = Additive1&i; 
							problodperm=PROBLODperm&i 
							cw_LODthresh = cw_LODthresh&i)); 
					where trait=&i; 
					label LOD&i = "LOD &CurrT";
					%if &ng < 4 %then %do;
						label Additive2&i = "Additive &CurrT";
					%end;
					%else %do;
						***assuming maximum ng = 4 ***;
						label Additive1&i = "Additive1 &CurrT";
						label Additive2&i = "Additive2 &CurrT";
					%end;
					%if &ng >= 3 %then %do;
						label Dominance&i = "Dominance &CurrT";
					%end;
					Nobs&i = &&Nobs&i;
					label Nobs&i = "NObs &CurrT";
					/* create Rsq */
					R2&i = 1 - 10**(-2*LOD&i / Nobs&i);
					label R2&i = "R2 &CurrT";					
					label PROBLODperm&i = "ProbLOD &CurrT";
					label cw_LODthresh&i = "CW LOD Threshold &CurrT";
				%let i = %eval(&i+1);
			%end;
			*drop trait;
			drop trait CMID; /*---Luciano---*/
		run;
	%end;
	%else %do;
		%if &ng >= 3 %then %do;
			*data ests (drop= Label); 
			data ests (drop= Label); /*---Luciano---*/
				set ests(rename = (Estimate = Additive2));
				where Label contains "Additive2";
				set ests(rename = (Estimate = Dominance));
				where Label contains "Dominance";
				%if &ng = 4 %then %do;
					set ests(rename = (Estimate = Additive1));
					where Label contains "Additive1";
				%end;
			run;
		%end;	
				
		data hk0; 
			*merge hk0 ests(keep = trait chrom CM %if &ng =3 %then Additive Dominance; %else estimate;);
			merge hk0 ests(keep = trait chrom CM CMID 
				%if &ng =4 %then Additive2 Dominance Additive1; 
				%else %if &ng =3 %then Additive2 Dominance; 
				%else estimate;);
		 	*by trait chrom CM;
		 	by trait chrom CMID;/*---Luciano---*/
			%if &ng = 2 %then %do;
		 		rename estimate = Additive2;
		 	%end;
		run;
		
		data hkout;	
			%let i = 1;
			%do %while(%length(%scan(&TraitVar,&i)));
				%let CurrT = %scan(&TraitVar,&i);
				set hk0(rename=(lod=LOD&i Additive2=Additive2&i 
						%if &ng >= 3 %then Dominance = Dominance&i;
						%if &ng = 4 %then Additive1 = Additive1&i;
						)
					); 
					where trait=&i; 
					label LOD&i = "LOD &CurrT";
					%if &ng < 4 %then %do;
						label Additive2&i = "Additive &CurrT";
					%end;
					%else %do;
						***assuming maximum ng = 4 ***;
						label Additive1&i = "Additive1 &CurrT";
						label Additive2&i = "Additive2 &CurrT";
					%end;
					%if &ng >= 3 %then %do;
						label Dominance&i = "Dominance &CurrT";
					%end;										
					Nobs&i = &&Nobs&i;
					label Nobs&i = "NObs &CurrT";
					/* create Rsq */
					R2&i = 1 - 10**(-2*LOD&i / Nobs&i);
					label R2&i = "R2 &CurrT";
				%let i = %eval(&i+1);
			%end;
			*drop trait;
			drop trait CMID; /*---Luciano---*/
		run;
	%end;
	
/* Compute Max LOD for plots */
	%global maxLOD;
	proc means data=hk0 max noprint;
		var lod;
		output out=Lodmax max=lmax;
	run;

	data _null_;
		set lodmax;
		call symputx("maxLOD",lmax+1);
	run;
	%*put maxLod = &maxLOD;
%mend;

%macro OldGenoCheck;
	%if %index(%Upcase(&MarkerFmt),NUMERIC) and %UPCASE(&OldNumCode) =  NO %then %do; 
		data tmpcheck; set gendat;
			minval = min( of &AlleleVars);
			keep minval;
		run;
		
		proc means data=tmpcheck min noprint;
			var minval;
			output out=tmpcheck min=minval;
		run;
		
		data _null_; 
			set tmpcheck;
			call symput("negvals",trim(left(minval)));
		run;
		
		%if &negvals = . %then %do;
			%put ERROR: No numeric data was found in the genotype data, numeric coding format for the data should consist of only the following values: .,0,1,2,3,4.;
			%put ERROR: Please check that your data matches the Format of Marker Variables chosen on the Options tab.;
			%let exiterror =1;
			%goto macroexit;
		%end;		
		%if &negvals < 0 %then %do;
			%put ERROR: Negative values were found in the genotype data, numeric coding format for the data should consist of only the following values: .,0,1,2,3,4.;
			%put ERROR: If your data follows a previous formatting for JMP Genomics, please use the check box option for this on the Options tab.;
			%let exiterror =1;
			%goto macroexit;
		%end;
	%end;
	%macroexit:;
%mend;

%macro RecodeGenos;

	/* recode to numeric */
	%if %index(%Upcase(&MarkerFmt),NUMERIC) = 0  %then %do;
		%if %length(%scan(&GenoMcode,2)) = 0 %then %do;
			%put ERROR: Please enter the missing required Allele Character that represents the P2 line.;
			%let exiterror =1;
			%goto macroexit;
		%end;
		
		data _mnames_; 
			set gendat;
			where 0;
		run;
	
		%let dsid = %sysfunc(open(_mnames_));
		%let MType = %sysfunc(vartype(&dsid,1));
		%let rc = %sysfunc(close(&dsid));
		%put Marker variable type is &MType;
		
		%nobsvars(gendat);
		%if %index(%Upcase(&MarkerFmt), GENOTYPES)>0 %then %do;
			%if %index(%Upcase(&MarkerFmt), NONDELIMITED)>0%then %do;
				%let code0 = %scan(&GenoMCode,1,' ')%scan(&GenoMCode,1,' ');
				%let code11 = %scan(&GenoMCode,1,' ')%scan(&GenoMCode,2,' ');
				%let code12 = %scan(&GenoMCode,2,' ')%scan(&GenoMCode,1,' ');
				%let code2 =  %scan(&GenoMCode,2,' ')%scan(&GenoMCode,2,' ');
				%put code0 = &code0, code11 = &code11, code12 = &code12, code2 = &code2;			
			%end;
			%else %if %index(%Upcase(&MarkerFmt), HOMOZYGOUS)>0 %then %do;
				%let code0 = %scan(&GenoMCode,1,' ');
				%let code11 = %scan(&GenoMCode,1,' ')%scan(&GenoMCode,2,' ');  /* AB coded */
				%let code12 = H;  /* H coded*/
				%let code2 =  %scan(&GenoMCode,2,' ');
				%put code0 = &code0, code11 = &code11, code12 = &code12, code2 = &code2;	
			%end;
			%else %do; /* delimited */
				%let code0 = %scan(&GenoMCode,1,' ')&GenoDelim%scan(&GenoMCode,1,' ');
				%let code11 = %scan(&GenoMCode,1,' ')&GenoDelim%scan(&GenoMCode,2,' ');
				%let code12 =  %scan(&GenoMCode,2,' ')&GenoDelim%scan(&GenoMCode,1,' ');
				%let code2 =  %scan(&GenoMCode,2,' ')&GenoDelim%scan(&GenoMCode,2,' ');
				%put code0 = &code0, code11 = &code11, code12 = &code12, code2 = &code2;
			%end;

			data gendat; set gendat;
				array d{*} &AlleleVars;
				array _geno_{&nvars};
				do i = 1 to &nvars;
					%if %Upcase(&MType) = C %then %do;
						if trim(left(d{i})) = "&code0" then _geno_{i} = 0;
						else if (trim(left(d{i})) = "&code11" or trim(left(d{i})) = "&code12") then _geno_{i} = 1;
						else if trim(left(d{i})) = "&code2" then _geno_{i} = 2;
					%end;
					%else %do;
						%if %sysfunc(anyalpha(&GenoMCode)) > 0 or %sysfunc(anyalpha(&GenoMCode)) >0 %then %do;
							%put ERROR:  The Allele Characters representing P1 and P2 line do not match the Marker Variable type.;
							%let exiterror =1;
							%goto macroexit;
						%end;
						if d{i} = &code0 then _geno_{i} = 0;
						else if d{i} = &code11 then _geno_{i} = 1;
						else if d{i} = &code2 then _geno_{i} = 2;
					%end;
				end;
				drop i &AlleleVars;
			run;
		%end;
		%else %if %index(%Upcase(&MarkerFmt), ALLELES) %then %do;
			%let nvars = %eval(&nvars/2);
			%let code0 = %scan(&GenoMCode,1,' ');
			%let code2 = %scan(&GenoMCode,2,' ');
			data gendat; set gendat;
				array d{*} &AlleleVars;
				array _geno_{&nvars};
				do i = 1 to &nvars;
					%if %Upcase(&MType) = C %then %do;
						if trim(left(d{2*i-1})) = "&code0"  and trim(left(d{2*i}))= "&code0" then _geno_{i} = 0;
						else if ( (trim(left(d{2*i-1})) = "&code0" and trim(left(d{2*i}))= "&code2") 
							  or ( (trim(left(d{2*i-1})) = "&code2" and trim(left(d{2*i}))= "&code0") then _geno_{i} = 1;
						else if trim(left(d{2*i-1})) = "&code2"  and trim(left(d{2*i}))= "&code2" then _geno_{i} = 2;
					%end;
					%else %do;
						%if %sysfunc(anyalpha(&GenoMCode)) > 0 or %sysfunc(anyalpha(&GenoMCode)) >0 %then %do;
							%put ERROR:  The Allele Characters representing P1 and P2 line do not match the Marker Variable type.;
							%let exiterror =1;
							%goto macroexit;
						%end;
						if d{2*i-1} = &code0  and d{2*i}= &code0 then _geno_{i} = 0;
							else if ( d{2*i-1} = &code0 and d{2*i}= &code2) 
								  or ( d{2*i-1}= &code2 and d{2*i}= &code0) then _geno_{i} = 1;
							else if d{2*i-1} = &code2  and d{2*i}= &code2 then _geno_{i} = 2;
					%end;
				end;
				drop i &AlleleVars;
			run;
		%end;
		data gfcheck; set gendat;
			sumcheck = sum(of _geno:);
			keep sumcheck;
			if sumcheck ne .;
		run;

		%nobsvars(gfcheck);
		
		%if &nobs = 0 %then %do;
			%put ERROR: Converting the Marker Variables to numeric format produced all missing data.;
			%put ERROR: Please check that your input marker data matches the Format of Marker Variables chosen on the Options tab.;
			%let exiterror =1;
			%goto macroexit;
		%end;		
	%end;
	
	%if %UPCASE(&OldNumCode) = YES %then %do;
		data gendat; set gendat;
			array d{*} &AlleleVars;
			do i = 1 to dim(d);
				if d{i} = 2 then d{i} = 0;
				else if d{i} = 0 then d{i} = 2;
				else if d{i} = -3 then d{i} = 3;
				else if d{i} = -2 then d{i} = 4;
				else if d{i} = -1 then d{i} = .;
			end;
			drop i;
		run;
	%end;
	
	%macroexit:;
%mend RecodeGenos;



/* Get variable number in a variable list  */
%macro getVarNumber(_vname);
	%global _vnum;
	data _null_;
		%let i = 1;
		%let var = %scan(&&&_vname,&i);
		%do %while(%length(&var));
			%let i = %eval(&i+1);
			%let var = %scan(&&&_vname,&i);
		%end;
		%let _vnum = %eval(&i-1);
	run;
%mend getVarNumber;

/* Get a varibale list from the contents of a SAS data varaible */
%macro getDataVarList(_data,_var);
	%global dataVarList;
	
	data _null_;
		set &&&_data end=last;
		call symputx("_tmp_"||strip(_n_),&_var);
		if last then call symputx("ntot",_n_);
	run;
	%let dataVarList=;
	%do i = 1 %to &ntot;
		%let dataVarList = &dataVarList %upcase(&&_tmp_&i);
	%end;
%mend getDataVarList;

/* Transfer to numeric if the factor variable is character */
%macro setFactorVarData();
/* Produce FtData data set */
	data FtData;
		set &CrossFile (Keep = &FactorVars);
	run;
	%if &debugflag = 1 %then %do;
		proc print data=&FtFile noobs; Title "&FtFile.-1 before change"; run;
	%end;
/* Change all character variables into numeric (all observations) */
	%let FtVarStr=;
	%let FtNum1 = 0;
	%let FtNum2 = 0;
/* FtNum2 = number of char var and FtNum1 = number of num var */
	proc contents data=&FtFile out=FtInfo1 (keep=name type) noprint; run;
/*proc print data=FtInfo1; Title "FtInfo-1";*/
	proc summary data = FtInfo1; by type; Output out = FtInfo2; run;
/*proc print data=FtInfo2; Title "FtInfo-2";*/
	data _null_;
		set FtInfo2;
		if type = 1 then call symput('FtNum1',_Freq_);
		else call symput('FtNum2',_Freq_);
	run;
/* FtData_Num */
%*put FtNum1 = &FtNum1 FtNum2 = &FtNum2;
	%if &FtNum2 > 0 %then %do;
/* FtData_Char*/
		data FtData_Char;
			set &FtFile (Keep = _CHAR_);
		run;
/*proc print data = FtData_Char; Title "FtData_Char"; run;*/
/* FtVarStr is the Char Var List */
		proc contents data=FtData_Char out=FtInfo (keep=name) noprint; run;
/* proc print data=FtInfo; Title "Factor Info"; run; */
		%let _ni_ = 0;
		data _null_;
			set FtInfo nobs=eof;
			call symput("_ni_",trim(left(eof)));
		run;
		%if (&_ni_ = ) %then %let _ni_ = 0;
		%do _i_ = 1 %to &_ni_;
			data _null_;
				set FtInfo (firstobs=&_i_ obs=&_i_);
				call symput('_tmp_',name);
			run;
			%let FtVarStr = &FtVarStr &_tmp_;
		%end;
/* %put FtVarStr = &FtVarStr;*/
/* For each char var, to transfer into numeric value */	
		%let i = 1;
		%let var = %scan(&FtVarStr,&i);
		%do %while(%length(&var));
			data _tmp_;
				set FtData_Char (keep = &var);
			run;
			proc sort data=_tmp_; by &var; run;
			proc summary data = _tmp_; by &var;	Output out = _ntmp_;
/*proc print data=_ntmp_; Title "each&i"; run; */
			%let _ni_ = 0;
			data _null_;
				set _ntmp_ nobs=eof;
				call symput("_ni_",trim(left(eof)));
			run;
			%do j = 1 %to &_ni_;
				data _null_;
					set _ntmp_ (firstobs=&j obs=&j);
					call symput("symb&j",&var);
					call symput("code&j",&j);
				run;
			%end;
			data _tmp_&i;
				ID = _n_;
				set FtData_Char (keep = &var);
				%do j = 1 %to &_ni_;
					if &var eq "&&symb&j" then &var =  %eval(&&code&j);
				%end;
			run;
/*proc print data = _tmp_&i noobs; title "_tmp_&i";	run;		*/
			%let i = %eval(&i+1);
			%let var = %scan(&FtVarStr,&i);
		%end;

		data nFtData;
			ID = _n_;
			set FtData;
		run;
/*proc print data=nFtData; Tilte "nFtData-1";run;*/
		%let i = 1;
		%let var = %scan(&FtVarStr,&i);
		%do %while(%length(&var));
			data nFtData;
				update nFtData _tmp_&i; by ID;
			run;	
			%let i = %eval(&i+1);
			%let var = %scan(&FtVarStr,&i);
		%end;
		%let FtFile = nFtData;
/* %put FactorVars = &FactorVars; */
	%end; /*&FtNum2 > 0*/
%mend setFactorVarData;

/* Make Marker genotype of 'MkFile' code as 2 1 0 -3 -2 -1 */
%macro setGenotypeCode();
	%global nMarkerVars;
	%if &MarkerFmt = Alleles %then %do;
		%let chrflag = 1;
	%*put The format is Alleles !;
		%getDataVarList(MapFile,&MarkerLab);
		%let nMarkerVars = &dataVarList;
	%*put nMarkerVars = &nMarkerVars;
	/* Two columns for one variable in nMarkerVars according to positions */
		%let rename_str =; %let varsav = ;
		%let i = 1;
		%let var = %scan(&MarkerVars,&i);
		%do %while(%length(&var));
			%let rename_str = &rename_str &var.=_V%eval(&i)_ ;
			%let varsav = &var; /*last variable*/
			%let i = %eval(&i+1);
			%let var = %scan(&MarkerVars,&i);
		%end;
		%let i1 = %eval(&i-1);
	/* Test var number in MarkerVars is odd or even */
		%if %sysfunc(MOD(&i1,2)) = 0 %then %do;
			%let _ni_ = %eval(&i1/2);
		%end;
		%else %do;
	/* delete last marker label of MarkerVars */
			%let MarkerVars = %substr(&MarkerVars,1,%length(&MarkerVars)-%length(&varsav)-1);
			%let _ni_ = %eval((&i1-1)/2);
		%end;
	/* nMarkerVars according to MarkerVars*0.5 */
		%let tVar = ;
		%do i=1 %to &_ni_;
			%let var = %scan(&nMarkerVars,&i);
			%let tVar = &tVar &var;
		%end;
		%let nMarkerVars = &tVar;
	%*put rename_str = &rename_str;
		data _t0_;
			set &CrossFile;
			rename &rename_str;
		run;
		data  _t1_;
			retain &nMarkerVars;
			set _t0_;
			%let i = 1;
			%let var1 = %scan(&MarkerVars,&i);
			%do %while(%length(&var1));
				%let j = %eval((&i+1)/2);
				%let var = %scan(&nMarkerVars,&j);
	%*put &i - &j and &var - &var1 or &var2;
				%let var1 = _V%eval(&i)_;
				%let var2 = _V%eval(&i+1)_;
				%if &chrflag = 1 %then %do;
					if      &var1 eq "&GM2" and &var2 eq "&GM2" then &var = 2;
					else if &var1 eq "&GM1" and &var2 eq "&GM2" then &var = 1;
					else if &var1 eq "&GM2" and &var2 eq "&GM1" then &var = 1;
					else if &var1 eq "&GM1" or  &var2 eq "&GM1" then &var = 0;
					else if &var1 eq "&GM2" or  &var2 eq "&GM2" then &var = 0;
					else &var = -1;
				%end;
				%else %do;
					if      &var1 eq &GM1 and &var2 eq &GM1 then &var = 2;
					else if &var1 eq &GM1 and &var2 eq &GM2 then &var = 1;
					else if &var1 eq &GM2 and &var2 eq &GM1 then &var = 1;
					else if &var1 eq &GM1 or  &var2 eq &GM1 then &var = 0;
					else if &var1 eq &GM2 or  &var2 eq &GM2 then &var = 0;
					else &var = -1;
				%end;
		/***************************************/
				%let i = %eval(&i+2);
				%let var1 = %scan(&MarkerVars,&i);
			%end;
		run;
		data nMkData;
			set _t1_ (keep = &nMarkerVars);
		run;
		%let MkFile = nMkData;
	%end;
	%else %if &MarkerFmt = Genotypes %then %do;
		%*put The format is Genotypes !;
		%let C11 = &GM1.&GenoDelim.&GM1;
		%let C12 = &GM1.&GenoDelim.&GM2;
		%let C21 = &GM2.&GenoDelim.&GM1;
		%let C22 = &GM2.&GenoDelim.&GM1;
		data _t1_;
			set &CrossFile;
			%let i = 1;
			%let var = %scan(&MarkerVars,&i);
			%do %while(%length(&var));
				if      &var = "&C11" then &var = 2;
				else if &var = "&C12" then &var = 1;
				else if &var = "&C21" then &var = 1;
				else if &var = "&C22" then &var = 0;
	/*for M- and m- */
				else if FIND(&var,"&GM1") then &var = -3;
				else if FIND(&var,"&GM2") then &var = -2;
				else 					       &var = -1;
				%let i = %eval(&i+1);
				%let var = %scan(&MarkerVars,&i);
			%end;
		run;
		data nMkData;
			set _t1_ (keep = &MarkerVars);
		run;
		%let MkFile = nMkData;
	%end;
	%else %if &MarkerFmt = Nondelimited Genotypes %then %do;
		%*put The format is Nondelimited Genotypes !;
		%let C11 = &GM1&GM1;
		%let C12 = &GM1&GM2;
		%let C21 = &GM2&GM1;
		%let C22 = &GM2&GM2;
	%*put &c11 - &c12 - &c21 - &c22;
	
		data _t1_;
			set &CrossFile;
			%let i = 1;
			%let var = %scan(&MarkerVars,&i);
			%do %while(%length(&var));
				%if &NonGNTypeFlag = 0 %then %do;
					if      strip(&var) = "&C11" then &var = 2;
					else if strip(&var) = "&C12" then &var = 1;
					else if strip(&var) = "&C21" then &var = 1;
					else if strip(&var) = "&C22" then &var = 0;
					/*for M- and m- */
					else if FIND(&var,"&GM1") then &var = -3;
					else if FIND(&var,"&GM2") then &var = -2;
					else 					       &var = -1;
				%end;
				%else %do;
					if      &var = &C11 then &var = 2;
					else if &var = &C12 then &var = 1;
					else if &var = &C21 then &var = 1;
					else if &var = &C22 then &var = 0;
					else if FIND(&var,"&GM1") then &var = -3;
					else if FIND(&var,"&GM2") then &var = -2;
					else 					         &var = -1;
				%end;
				%let i = %eval(&i+1);
				%let var = %scan(&MarkerVars,&i);
			%end;
		run;
		
		data nMkData;
			set _t1_ (keep = &MarkerVars);
		run;
		%let MkFile = nMkData;
	%end;
	%else %do;
		%*put The format is Numeric Genotypes !;
	%end;
%mend setGenotypeCode;

/* Using simple regression to sort markers (RR) */
%macro CMarkerSRegression(_markerVar);
	%let newData =; %let countData = 0;
	%let i = 1;
	%let vari = %scan(&TraitVar,&i);
	%do %while(%length(&vari));
	/***************************************************/
		%let j = 1;
		%let varj = %scan(&&&_markerVar,&j);
		%do %while(%length(&varj));
			proc reg data= &CrossFile outest = work.r EDF noprint;
			model &vari = &varj; run;
			data _null;
				set work.r;
				call symputx('RR',_RSQ_);
			run;
			%if &RR > 0.005 %then %do;
				%let newData = &newData &i &varj &RR;
				%let countData = %eval(&countData+1);
			%end;

			%let j = %eval(&j+1);
			%let varj = %scan(&&&_markerVar,&j);
		%end;
	/***************************************************/
		%let i = %eval(&i+1);
		%let vari = %scan(&TraitVar,&i);
	%end;
	proc iml;
		tmpN = J(&countData,2,0); *trait,chrom,rr;
		tmpC = J(&countData,1,'                          '); *marker;
		%let i = 1;
		%do i = 1 %to &countData;
			%let n1 = %eval((&i-1)*3+1);
			%let n2 = %eval((&i-1)*3+2);
			%let n3 = %eval((&i-1)*3+3);
			%let trait = %scan(&newData,&n1,' ');
			%let marker = %scan(&newData,&n2,' ');
			%let RR = %scan(&newData,&n3,' ');
			tmpN[&i,1] = &trait; tmpN[&i,2] = &RR; 
			tmpC[&i,1] = {"&marker"};
		%end;
		create work.r1 from tmpN [colname={'Trait' 'RR'}];
		append from tmpN; 
		create work.r2 from tmpC [colname={'Marker'}]; 
		append from tmpC;
	quit;
	data &CMFile;
		retain Trait Marker RR;
		set work.r1; set work.r2;
		output;
	run;
	proc sort data=&CMFile; by Trait descending RR; run;
	data &CMFile;
		set &CMFile (drop = RR);
	run;
%mend CMarkerSRegression;

/* Forward search of selecting control markers */
%macro CMarkerFSearch(_markerVar);

	/*Luciano*/
	proc datasets nolist;
		delete _all_ r;
	quit;
	/*Luciano*/

	%let i = 1;
	%let vari = %scan(&TraitVar,&i);
	%do %while(%length(&vari));	
	
		proc datasets nolist;
			delete r;
		quit;
	
		/***************************************************/
		proc reg data=&CrossFile;
		model &vari = &&&_markerVar 
		/ selection=forward SLE=&EntryProb details=summary;
		ods output SelectionSummary=r;
		run;
		/*
		data r;	set r; trait = &i; run;
		%if &i = 1 %then %do;
			data _all_;	set r; run;
		%end;
		%else %do;
			data _all_;	set _all_ r; run;
		%end;
		*/
		/*---begin: Luciano ---*/
		%if %sysfunc(exist(r)) %then %do;
			data r;	set r; trait = &i; run;
		%end;
		%else %do;
			data r;	set _null_; run;
		%end;
		%if ^%sysfunc(exist(_all_)) %then %do;
			data _all_; set r; run;
		%end;
		%else %do;
			data _all_; set _all_ r; run;
		%end;
		/*---end: Luciano ---*/
	/***************************************************/
		%let i = %eval(&i+1);
		%let vari = %scan(&TraitVar,&i);
	%end;
	
	%NObsVars(_all_);
	%if &nobs > 1 %then %do;
		data _all_;
			set _all_;
			rename VarEntered= Marker;
		run;
		data &CMFile;
			retain Trait Marker;
			set _all_ (keep= Trait Marker);
			output;
		run;
	%end;
	%else %do;
		data &CMFile;
			retain Trait . Marker '';
			if _n_ > 1 then output;
		run;
	%end;
	
%mend CMarkerFSearch;

/* Stepwise search of selecting control markers */
%macro CMarkerStepSearch(_markerVar);

	/*Luciano*/
	proc datasets nolist;
		delete _all_ r;
	quit;
	/*Luciano*/

	%let i = 1;
	%let vari = %scan(&TraitVar,&i);
	%do %while(%length(&vari));
	
		/*Luciano*/
		proc datasets nolist;
			delete r;
		quit;
		/*Luciano*/
	
		/***************************************************/
		proc reg data=&CrossFile;
		model &vari = &&&_markerVar 
		/ selection=stepwise SLE=&EntryProb SLS=&StayProb details=summary;
		ods output SelectionSummary=r;
		run;
		/*
		data r;	set r; trait = &i; run;
		%if &i = 1 %then %do;
			data _all_;	set r; run;
		%end;
		%else %do;
			data _all_;	set _all_ r; run;
		%end;
		*/
		/*---begin: Luciano ---*/
		%if %sysfunc(exist(r)) %then %do;
			data r;	set r; trait = &i; run;
		%end;
		%else %do;
			data r;	set _null_; run;
		%end;
		%if ^%sysfunc(exist(_all_)) %then %do;
			data _all_; set r; run;
		%end;
		%else %do;
			data _all_; set _all_ r; run;
		%end;
		/*---end: Luciano ---*/
	/***************************************************/
		%let i = %eval(&i+1);
		%let vari = %scan(&TraitVar,&i);
	%end;
	
	%NObsVars(_all_);
	%if &nobs > 1 %then %do;
		data _all_;
			set _all_;
			rename VarEntered= Marker;
		run;
		data _all_;
			set _all_;
			if missing(Marker) then delete;
		run;
		data &CMFile;
			retain Trait Marker;
			set _all_ (keep= Trait Marker);
			output;
		run;
	%end;
	%else %do;
		data &CMFile;
			retain Trait . Marker '';
			if _n_ > 1 then output;
		run;
	%end;
	
%mend CMarkerStepSearch;

/* Backward-search of selecting control markers */
%macro CMarkerBSearch(_markerVar);
	proc iml;
		tmpMat1 = J(1,1); tmpMat2 = J(1,1);
		%let i = 1;
		%let vari = %scan(&TraitVar,&i);
		%do %while(%length(&vari));
			%let j = 1;
			%let varj = %scan(&&&_markerVar,&j);
			%do %while(%length(&varj));
				tmpMat1 = &i; tmpMat2 = "&varj";
				flag = 1; 
				if flag=1 then do;
					addMat1 = addMat1//tmpMat1;
					addMat2 = addMat2//tmpMat2;
				end;
				%let j = %eval(&j+1);
				%let varj = %scan(&&&_markerVar,&j);
			%end;
			%let i = %eval(&i+1);
			%let vari = %scan(&TraitVar,&i);
		%end;
*print delMat1 delMat2;
*print addMat1 addMat2;
		create work.r1 from addMat1 [colname={'Trait'}];
		append from addMat1; 
		create work.r2 from addMat2 [colname={'Marker'}]; 
		append from addMat2;
	quit;
	data &CMFile;
		retain Trait Marker;
		set work.r1; set work.r2;
		output;
	run;
	data &CMFile;
		set &CMFile;
	run;
%mend CMarkerBSearch;

/*
%macro CMarkerBSearch(_markerVar);
	%let i = 1;
	%let vari = %scan(&TraitVar,&i);
	%do %while(%length(&vari));
		proc reg data=&CrossFile;
		model &vari = &&&_markerVar 
		/ selection=backward SLS=&StayProb details=summary;
		ods output SelectionSummary=r;
		run;
		data r;	set r; trait = &i; run;
		%if &i = 1 %then %do;
			data _all_;	set r; run;
		%end;
		%else %do;
			data _all_;	set _all_ r; run;
		%end;
		%let i = %eval(&i+1);
		%let vari = %scan(&TraitVar,&i);
	%end;
	data _tmp_;
		retain Trait VarRemoved;
		set _all_ (keep= Trait VarRemoved);
	run;
	proc iml;
		reset deflib=work;	use _tmp_;
		read all var {Trait} into delMat1;
		read all var {VarRemoved} into delMat2;
		tmpMat1 = J(1,1); tmpMat2 = J(1,1);
		%let i = 1;
		%let vari = %scan(&TraitVar,&i);
		%do %while(%length(&vari));
			%let j = 1;
			%let varj = %scan(&&&_markerVar,&j);
			%do %while(%length(&varj));
				tmpMat1 = &i; tmpMat2 = "&varj";
				flag = 1; 
				do row = 1 to nrow(delMat1);
					if delMat1[row]=tmpMat1 then do;
						if delMat2[row]=tmpMat2 then do;
							flag = 0; row = nrow(delMat1)+1;
						end;
					end;
				end;
				if flag=1 then do;
					addMat1 = addMat1//tmpMat1;
					addMat2 = addMat2//tmpMat2;
				end;
				%let j = %eval(&j+1);
				%let varj = %scan(&&&_markerVar,&j);
			%end;
			%let i = %eval(&i+1);
			%let vari = %scan(&TraitVar,&i);
		%end;
*print delMat1 delMat2;
*print addMat1 addMat2;
		create work.r1 from addMat1 [colname={'Trait'}];
		append from addMat1; 
		create work.r2 from addMat2 [colname={'Marker'}]; 
		append from addMat2;
	quit;
	data &CMFile;
		retain Trait Marker;
		set work.r1; set work.r2;
		output;
	run;
	data &CMFile;
		set &CMFile;
	run;
%mend CMarkerBSearch;
*/

/* _gendata_ is a data set with input genotypes coded 
		. missing,
		0 for AA, 
		1 for AB, 
		2 for BB 
	where A is from P1 parental line, B is P2
	for F2 cross, allow dominance coding:  
		3 is A- (dominant for P1: e.g. not BB so AA or AB)
		4 is B- (dominant for P2: not AA so AB or BB)
	cross is type of breeding: 1=BC, 2 = F2, 3=RILself, 4=RILsib, 5 = Double Haploid
	RILBiasCorr is 1 if use bias corrected rhat estimate from Martin and Hospital 2006
*/
%macro rfest(_gendata_=, _annodata_=, _RandLG_AP_=, _runchisq_=, _swindow_=, _wsize_=);
	%put RILBiasCorr= &RILBiasCorr;
	%put cross= &cross;
	
	options nonotes;
	proc iml; 
		/* For cross codes: 1=BC or 2=F2 or 3=RIL(self), 4=RIL(sib-mating) or 5=Double Haploid*/
		start estrf(r) global(dom1, dom2, cc);
			if  dom1=0 & dom2 = 0 then do;
				/* Likelihood of recombination for F2 from pg 53 of Wu, Ma, and Casella */
				LL = (cc[1,1])*log(0.25*(1-r)**2) 
					 +(cc[1,2])*log(0.5*r*(1-r)) 
					 +(cc[1,3])*log(0.25*(r)**2)
				     +(cc[1,4])*log(0.5*(1-r)**2 + 0.5*r**2);
			end;
			else if (dom1=0 & dom2>0)|(dom1>0 & dom2=0) then do;
				/* likelihood  if one marker dominant coded, one codominant */
				/*LL = cc[1,1]*log(1-r) 
					 + cc[1,2]*log(1+r)
					 + cc[1,3]*log(r) 
					 + cc[1,4]*log(2-r) 
					 + cc[1,5]*log(1-r+r**2);
					 */
				LL = cc[1,1]*log(r**2)
					+ cc[1,2]*log(1-r**2) 
					+ cc[1,3]*log(r*(1-r))  
					+ cc[1,4]*log(r*(2-r))  
					+ cc[1,5]*log(1-r+r**2)
					+ cc[1,6]*log((1-r)**2);
			end;
			else if (dom1=1 & dom2 = 1)|(dom1=2 & dom2=2) then do;  
				/* markers are both dominant coding for same parent*/
				LL = cc[1,1]*log(3-2*r+r**2)
					  +cc[1,2]*log(2*r - r**2) 
					  +2*cc[1,3]*log(1-r);
			end;
			else do;
				/* both dominant coding but one for P1 and one for P2 */
				LL = cc[1,1]*log(2+r**2)
					 +cc[1,2]*log(1-r**2)
					 +2*cc[1,3]*log(r);
			end;
			/* Likelihood of R to use in recombination for RILs from pg 109 of Wu, Ma, and Casella */
			/*else do;
				LL = (counts[3,1]+counts[1,3])*log(r)+(counts[1,1]+counts[3,3])*log(1-r);
			end;
			*/
			return(LL);
		finish estrf;
		/*begin: Luciano. Likelihood function of SFn population. 
		References:
		1- Bulmer MG (1985). The Mathematical Theory of Quantitative Genetics. Oxford University Press. (Page 33).
		2- Laura M Shannon, Brian S. Yandell and Karl Broman (January of 2013). Users Guide for New BCsFt Tools for R/qtl. 
		*/
		start estrf_sfn(r) global(dom1, dom2, cc, _t);
			/*From Bulmer (1985)*/
			B11= 1.000*((0.5/(1+2*r)) - (1/2)**_t - (0.5/(1+2*r))*(0.5 - r)**_t + 0.25*(0.5 - r*(1-r))**(_t-1));
			B14= 1.000*(  (r/(1+2*r)) - (1/2)**_t + (0.5/(1+2*r))*(0.5 - r)**_t + 0.25*(0.5 - r*(1-r))**(_t-1));
			B12= 0.500*(            0 + (1/2)**_t +                           0 - 0.50*(0.5 - r*(1-r))**(_t-1));
			B22= 0.500*(            0 +         0 +   (1/(1-2*r))*(0.5 - r)**_t + 0.50*(0.5 - r*(1-r))**(_t-1));
			B23= 0.500*(            0 +         0 -   (1/(1-2*r))*(0.5 - r)**_t + 0.50*(0.5 - r*(1-r))**(_t-1));
			*print B11 B14 B12 B22 B23;
					
			if  dom1=0 & dom2 = 0 then do;
				/* likelihood if both markers are codominant */
				LL = (cc[1,1])*log(2*B11) 
					 +(cc[1,2])*log(8*B12) 
					 +(cc[1,3])*log(2*B14)
				     +(cc[1,4])*log(2*(B22+B23));
			end;
			else if (dom1=0 & dom2>0)|(dom1>0 & dom2=0) then do;
				/* likelihood  if one marker dominant coded, one codominant */
				LL = cc[1,1]*log(B14)
					+ cc[1,2]*log(B11 + 2*B12) 
					+ cc[1,3]*log(2*B12)  
					+ cc[1,4]*log(B14 + 2*B12)  
					+ cc[1,5]*log(2*B12 + 2*(B22 + B23))
					+ cc[1,6]*log(B11); 
			end;
			else if (dom1=1 & dom2 = 1)|(dom1=2 & dom2=2) then do;  
				/* markers are both dominant coding for same parent*/
				LL = cc[1,1]*log(B11 + 2*(B22 + B23) + 4*B12)
					  +cc[1,2]*log(B14 + 2*B12) 
					  +cc[1,3]*log(B11);
			end;
			else do;
				/* both dominant coding but one for P1 and one for P2 */
				LL = cc[1,1]*log(4*B12 + B14 + 2*(B22 + B23))
					 +cc[1,2]*log(B11 + 2*B12)
					 +cc[1,3]*log(B14);
			end;
			return(LL);
		finish estrf_sfn;
		/*end: Luciano*/
		
		/* function for the derivatives of the likelihoods */
		start estrfd(r) global(dom1, dom2, cc);
			if  dom1=0 & dom2 = 0 then do;
				/* Likelihood of recombination for F2 from pg 53 of Wu, Ma, and Casella */
				LLD = -(2*cc[1,1]+cc[1,2])/(1-r)
					 +(cc[1,2]+2*cc[1,3])/r 
					 +(cc[1,4]*(4*r-2))/(2*r**2-2*r-1);
			end;
			else if (dom1=0 & dom2>0)|(dom1>0 & dom2=0) then do;
				
				/* likelihood derivative if one marker dominant coded, one codominant */
				LLD = -cc[1,1]/(1-r)
					 +cc[1,2]/(1+r)
					 +cc[1,3]/r
					 -cc[1,4]/(2-r)
					 +((2*r-1)*cc[1,5])/(1-r+r**2);
			end;
			else if (dom1=1 & dom2 = 1)|(dom1=2 & dom2=2) then do;  
				/* markers are both dominant coding for same parent*/
				LLD = -(2*(1-r)*cc[1,1])/(3-2*r+r**2)
					 +(2*(1-r)*cc[1,2])/(r*(2-r)) 
					 -(2*cc[1,3])/(1-r);
			end;
			else do;
				/* both dominant coding but one for P1 and one for P2 */
				LLD = (2*r*cc[1,1])/(2+r**2)
					 -(2*r*cc[1,2])/(1-r**2)
					 +(2*cc[1,3])/r;
			end;
			return(LLD);
		finish estrfd;
		/* module for finding IRIL (via selfing) recombination estimate */
		start iril_self_bisect(r,rlow,rup,tol,t);
				/*
				RF2:t=0; RF3:t=1; RF4:t=2; 
				Reference: Winkler et al. 2003 - Genetics 164: 741-745.
				*/
				delta = rup - rlow;
				c = 1;
				do while(delta>tol);
				 mid = (rup + rlow)/2;
				 *rmid = 0.5 * (1 - (1 - 2 * mid) / (1 + 2 * mid) * (1 - mid) ** t);
				 rmid = 0.5 * (1 - (1 - 2 * mid) / (1 + 2 * mid) * (1 - mid) ** (t-2));
				 if rmid > r then rup = mid;
				 else rlow = mid; 
				 delta = abs(r-rmid);
				 c = c+1;
				 if c > 50 then delta = tol/2;
				end;
				return(rmid);
		finish iril_self_bisect;
		/* module for finding IRIL (via random mating) recombination estimate */
		start iril_sib_bisect(r,rlow,rup,tol,t);
				/*
				RF2:t=0; RF3:t=1; RF4:t=2; 
				Reference: Winkler et al. 2003 - Genetics 164: 741-745.
				*/
				delta = rup - rlow;
				c = 1;
				do while(delta>tol);
				 mid = (rup + rlow)/2;
				 *rmid = 0.5 * (1 - (1 - 2 * mid) / (1 + 6 * mid) * (1 - mid) ** t);
				 rmid = 0.5 * (1 - (1 - 2 * mid) / (1 + 6 * mid) * (1 - mid) ** (t-2));
				 if rmid > r then rup = mid;
				 else rlow = mid; 
				 delta = abs(r-rmid);
				 c = c+1;
				 if c > 50 then delta = tol/2;
				end;
				return(rmid);
		finish iril_sib_bisect;		
		*use gendat;
		use &_gendata_;
		read all into geno;
		n_mar = ncol(Geno);
		n_ind = nrow(Geno);
		cross = &cross;
		no_st = &no_st;
		pl = &Parent;
		Sr = J(n_mar, 8,.); /* matrix to fill with geno counts, chisq, and pvalue */
		Rf = J(n_mar,n_mar,.);
		LOD = J(n_mar,n_mar,.);
		/* for genetic map distances 
		Gd = J(n_mar,n_mar,.);
		*/
		con = {1E-5 . .,
			  .99999 . .,
				. . .};
		x0 = j(1,1,.01);
		opt = {1 0 . 1};
		_t = &GenNum;
		%if &_swindow_=1 %then %do; 
			n_marorig= n_mar;
		%end;
		do j1=1 to n_mar;
			/* get dominance coding for marker j1 dom coding: 0 = codominant, 1 = dom P1 (M-), 2 = dom P2 (m-)*/
			if cross = 2|cross = 6|cross = 22 then do;  
				dom1 = 0;   
				if max(geno[1:n_ind,j1])>3 then dom1=2;
				else if max(geno[1:n_ind,j1])>2 then dom1=1;
			end;
			if &_runchisq_ then do;
				/* calculation segregation ratio and chisq */
				n_obs = J(1,5,0);
				n_tot = 0;
				do i=1 to n_ind;
					if Geno[i,j1] ^= . then do;
						n_tot = n_tot+1;
						n_obs[1,Geno[i,j1]+1] = n_obs[1,Geno[i,j1]+1]+1;  /* get observed counts for each genotype */
					end;
				end;
				
				if cross = 2|cross = 6|cross = 22 then do; /*f2 or AIL or SFn*/
					if dom1=1 then do; /* dominant for P1 line */
						if no_st = 0 then do;
							chisq = ((n_obs[1,3] - &BB_ef * n_tot)**2)/(&BB_ef * n_tot)
									+ ((n_obs[1,4] - ((&AB_ef+ &AA_ef)* n_tot))**2)/((&AB_ef+ &AA_ef)* n_tot);
							Sr[j1, 7] = 1-PROBCHI(chisq, 1);
						end;
						/* a frequency minus b frequency */
						Sr[j1,8] = ((2/3)*n_obs[1,4] - ((1/3)*n_obs[1,4]+n_obs[1,3]))/n_tot;
					end;
					else if dom1=2 then do; /* dominant for P2 line */
						if no_st = 0 then do;
							chisq = ((n_obs[1,1] - &AA_ef * n_tot)**2)/(&AA_ef * n_tot)
									+ ((n_obs[1,5] - ((&AB_ef+ &BB_ef)* n_tot))**2)/((&AB_ef+ &BB_ef)* n_tot);
							Sr[j1, 7] = 1-PROBCHI(chisq, 1);
						end;
						Sr[j1,8] = (((1/3)*n_obs[1,5]+n_obs[1,1]) - (2/3)*n_obs[1,5])/n_tot;
					end;
					else do; /* codominant marker */
						if no_st = 0 then do;
							chisq = ((n_obs[1,1]-&AA_ef * n_tot)**2)/(&AA_ef * n_tot)
									+ ((n_obs[1,2]-&AB_ef * n_tot)**2)/(&AB_ef * n_tot)
									+ ((n_obs[1,3]-&BB_ef * n_tot)**2)/(&BB_ef * n_tot);
							Sr[j1, 7] = 1-PROBCHI(chisq, 2);
						end;
						Sr[j1,8] = (((1/2)*n_obs[1,2]+n_obs[1,1]) - ((1/2)*n_obs[1,2]+ n_obs[1,3]))/n_tot;
					end;
				end;
				else if cross =1 then do;  /* seg ratio for bc A to AB 1 to 1 */
					if pl = 1 then do;
						chisq = ((n_obs[1,1] - &AA_ef*n_tot)**2)/(&AA_ef*n_tot)
								+ ((n_obs[1,2] - &AB_ef*n_tot)**2)/(&AB_ef*n_tot);
					end;
					else do;
						chisq = ((n_obs[1,3] - &BB_ef*n_tot)**2)/(&BB_ef*n_tot)
								+ ((n_obs[1,2] - &AB_ef*n_tot)**2)/(&AB_ef*n_tot);
					end;
					Sr[j1, 7] = 1-PROBCHI(chisq, 1);
					Sr[j1,8] = (n_obs[1,1] - n_obs[1,2])/n_tot;
				end;
				else do; /* for RIL and DH seg ratio is AA to BB 1 to 1 */
					if no_st = 0 then do;
						chisq = ((n_obs[1,1] - &AA_ef*n_tot)**2)/(&AA_ef*n_tot)
								 + ((n_obs[1,3] - &BB_ef*n_tot)**2)/(&BB_ef*n_tot);
						%if (%length(&SegRatios) & &AB_ef>0) %then %do;
							chisq = chisq + ((n_obs[1,1] - &AA_ef*n_tot)**2)/(&AA_ef*n_tot);
							Sr[j1, 7] = 1-PROBCHI(chisq, 2);
						%end;
						%else %do;
							Sr[j1, 7] = 1-PROBCHI(chisq, 1);
						%end;
					end;
					Sr[j1,8] = (((1/2)*n_obs[1,2]+n_obs[1,1]) - ((1/2)*n_obs[1,2]+ n_obs[1,3]))/n_tot;
				end;	
				Sr[j1, 1] = n_obs[1,1]/n_tot;  /* AA count freq */
				Sr[j1, 2] = n_obs[1,2]/n_tot;	 /* AB count freq */
				Sr[j1, 3] = n_obs[1,3]/n_tot;  /* BB count freq */
				Sr[j1, 4] = n_obs[1,4]/n_tot;  /* Not BB count freq */
				Sr[j1, 5] = n_obs[1,5]/n_tot;	 /* Not AA count freq*/ 	
				if no_st = 0 then do;  /* don't do seg test for AIL right now */
					Sr[j1, 6] = chisq;
					%if %length(&MultipleTestingMethod)=0 %then %do;
						/* Don't convert to NegLog10 yet if applying MTM */
						if Sr[j1,7] ^= . then do;
							if Sr[j1,7]>= 0.000000000000001 then Sr[j1, 7] = -log10(Sr[j1,7]);
							else Sr[j1,7] = 15;
						end;
					%end;
				end;
			end;
			/* begin for recombination fraction calculation */ 
			Rf[j1,j1] = 0;
			LOD[j1,j1] = 99;
			%if &_swindow_=1 %then %do; 
				n_mar= min(j1 + &_wsize_, n_marorig);
			%end;
			*print cross;
			do j2=j1+1 to n_mar;			
				if  cross = 2|cross = 6|cross = 22 then do;  /* f2 or AIL or SFn*/
					/* get dominance coding for marker j2 dom coding: 0 = codominant, 1 = dom P1 (M-), 2 = dom P2 (m-)*/ 
					dom2 = 0;   
					if max(geno[1:n_ind,j2])>3 then dom2=2;
					else if max(geno[1:n_ind,j2])>2 then dom2=1;
					xres = .;
					counts= J(5, 5,0);
					*cc = J(1,5,0);  /* vector for collapsed counts for likelihoods */
					cc = J(1,6,0);  /* vector for collapsed counts for likelihoods */
					do i=1 to n_ind;
						if Geno[i,j1] ^= . & Geno[i,j2] ^= . then do;
							/* use straight counts for likelihood in codominant markers */
							counts[Geno[i,j1]+1,Geno[i,j2]+1] = counts[Geno[i,j1]+1,Geno[i,j2]+1]+1;
							if dom1=0 & dom2=0 then do;
								 cc[1,1] = counts[1,1]+counts[3,3];
								 cc[1,2] = counts[2,3]+counts[3,2]+counts[1,2]+counts[2,1];
								 cc[1,3] = counts[1,3]+counts[3,1];
								 cc[1,4] = counts[2,2];
							end;
							else if dom1=0 & dom2>0 then do;
								/* collapse counts for LL when Marker j1 is codominant  Marker j2 is either Dom for P1 or P2 */
								/*
								cc[1,1] = counts[1,4] + counts[3,5] + counts[2,1]+ counts[2,3] + 2*(counts[1,1] + counts[3,3]);
								cc[1,2] = counts[1,4] + counts[3,5];
								cc[1,3] = counts[2,1]+ counts[2,3] + counts[3,4] + counts[1,5] + 2*(counts[1,3] + counts[3,1]);
								cc[1,4] = counts[3,4] + counts[1,5]; 
								cc[1,5] = counts[2,4] + counts[2,5];
								*/
								cc[1,1] = counts[1,3] + counts[3,1];
								cc[1,2] = counts[1,4] + counts[3,5];
								cc[1,3] = counts[2,3] + counts[2,1];
								cc[1,4] = counts[3,4] + counts[1,5]; 
								cc[1,5] = counts[2,4] + counts[2,5];
								cc[1,6] = counts[3,3] + counts[1,1];
							end;
							else if dom1>0 & dom2=0 then do;
								/* collapse counts for LL when Marker j1 is either Dom for P1 or P2 Marker j2 is codominant */
								/*
								cc[1,1] = counts[4,1] + counts[5,3] + counts[1,2]+ counts[2,3] + 2*(counts[1,1] + counts[3,3]);
								cc[1,2] = counts[4,1] + counts[5,3];
								cc[1,3] = counts[1,2]+ counts[3,2] + counts[4,3] + counts[5,1] + 2*(counts[3,1] + counts[1,3]);
								cc[1,4] = counts[4,3] + counts[5,1]; 
								cc[1,5] = counts[4,2] + counts[5,2];
								*/							
								cc[1,1] = counts[3,1] + counts[1,3];
								cc[1,2] = counts[4,1] + counts[5,3];
								cc[1,3] = counts[3,2] + counts[1,2];
								cc[1,4] = counts[4,3] + counts[5,1]; 
								cc[1,5] = counts[4,2] + counts[5,2];
								cc[1,6] = counts[3,3] + counts[1,1];
							end;
							else if (dom1=1 & dom2 = 1)|(dom1=2 & dom2=2) then do; 
							/* counts for LL when both markers are dominant coded p1p1 or p2p2 */
								cc[1,1] = counts[4,4] + counts[5,5];
								cc[1,2] = counts[4,3] + counts[5,1] + counts[3,4] + counts[1,5];
								cc[1,3] = counts[3,3] + counts[1,1];
							end;
							else do;
								/* counts for LL when dominant but switched parental dominance */
								/*cc[1,1] = counts[4,5] + counts[3,4];*/
								cc[1,1] = counts[4,5] + counts[5,4];
								cc[1,2] = counts[4,1] + counts[5,3] + counts[3,5] + counts[1,4];
								cc[1,3] = counts[3,1] + counts[1,3];
							end;
							/*
							if(i=300) then do;
								mk= j1 || j2;							
								ty=dom1 || dom2;
								print(mk);
								print(ty);
								print(counts);
								sumco=sum(counts);
								print(sumco);
								print(cc);
								sumcc=sum(cc);
								print(sumcc);
							end;
							*/
						end;
					end;
					if cross = 2|cross = 6 then do;
						call nlpqn(rc, xres,"estrf",x0,opt,con);
						Rf[j1,j2]= min(&MaxR,xres);
						Rf[j2,j1]= Rf[j1,j2];
						LL1_= estrf(Rf[j1,j2]);
						LL0_= estrf(0.5);						
						if cross = 6 then do;  /* correction for AIL recombination */
							/*Reference: Jiang and Zeng (1997) - Genetica 101:47-58*/
							Rf[j1,j2] = (1-(1-Rf[j1,j2])**(&GenNum - 2)*(1-2*Rf[j1,j2]))/2; 
							Rf[j2,j1] = Rf[j1,j2];
						end;						
					end;
					else if cross = 22 then do;
						call nlpqn(rc, xres,"estrf_sfn",x0,opt,con);
						Rf[j1,j2]= min(&MaxR,xres);
						Rf[j2,j1]= Rf[j1,j2];
						LL1_= estrf_sfn(Rf[j1,j2]);
						LL0_= estrf_sfn(0.499);
					end;
					
					LOD[j1,j2]= log10(exp(LL1_)/exp(LL0_));
					LOD[j2,j1]= LOD[j1,j2];
					LOD_= LOD[j1,j2];
					*print xres LL1_ LL0_ LOD_;
					*Rf[j2,j1]= log10(exp(LL1_)/exp(LL0_));
				end;
				else if cross = 1|cross=5 then do; /* BC or DH*/
					totlines=0;
					rc=0;
					do i=1 to n_ind;
						if Geno[i,j1] ^= . & Geno[i,j2] ^= . then do;
							totlines=totlines+1;
							/* if BC then recombinants are MM vs. Mm (BC to P1) or mm vs. Mm (BC to P2)
							   if DH then recombinants are MM vs. mm (data should contain only homozygotes for parental lines)
							*/
						  	if Geno[i,j1] ^= Geno[i,j2] then do;
						  		rc=rc+1;
						  	end;
						end;
					end;
					if totlines ^= 0 then do;   /* at least one informative meiosis */ 
						Rf[j1,j2] = min(&MaxR,rc/totlines);
						Rf[j2,j1] = min(&MaxR,rc/totlines);
						
						*compute LOD*;
						L1_ = (Rf[j1,j2]**rc)*(1-Rf[j1,j2])**(totlines-rc); 
						L0_ = (0.5**rc)*(1-0.5)**(totlines-rc); 
						LOD[j1,j2] = log10(L1_/L0_);
						LOD[j2,j1] = LOD[j1,j2];
						LOD_= LOD[j1,j2];
						*print L1_ L0_ LOD_;
						*Rf[j2,j1] = log10(L1_/L0_);
					end;
				end;
				else do;  /* RILself, RILsib */
					/* get totlines  and homozygous recombinant counts*/
				    	counts= J(5, 5,0);  /* if data is coded with 3's or 4's they are ignored for RIL */
					totlines = 0;
					rc = 0;
					do i=1 to n_ind;
						if Geno[i,j1] ^= . & Geno[i,j2] ^= . then do;
							totlines = totlines+1;
							counts[Geno[i,j1]+1,Geno[i,j2]+1] = counts[Geno[i,j1]+1,Geno[i,j2]+1]+1;
						end;
					end;
					*print totlines;
					if totlines ^=0 then do; /* at least one informative meiosis */
						rc = counts[3,1]+counts[1,3];  /* recombinant homozygous */ 
						if cross = 3|cross = 7 then do; /* RILself */
							if &RILBiasCorr then do;
								/*use bias corrected formula (Martin and Hospital 2006)*/
								Rf[j1,j2]= min(&MaxR,rc*(totlines - rc - 1)/(2*(totlines - rc)**2));
								Rf[j2,j1]= min(&MaxR,rc*(totlines - rc - 1)/(2*(totlines - rc)**2));
								
								*compute LOD*;
								R = (2*Rf[j1,j2])/(1 + 2*Rf[j1,j2]);
								*print R rc;
								if R > 0 then do;
									L1_= (R**rc)*(1 - R)**(totlines - rc); 
									L0_= (0.5**rc)*(1 - 0.5)**(totlines - rc);
									LOD[j1,j2] = log10(L1_/L0_);
									LOD[j2,j1] = LOD[j1,j2];
								end;
								else do;
									LOD[j1,j2] = 99;
									LOD[j2,j1] = 99;
								end;
								LOD_= LOD[j1,j2];
								*print L1_ L0_;
								*Rf[j2,j1]= log10(L1_/L0_);
							end;
							else do;
								/* use traditional formula, pg 109 in Wu, Ma, Casella */
								R = min(&MaxR,rc/totlines);
								if R = 1 then do;  /* trap for if all are recombinant */
									Rf[j1,j2] = 1;
									Rf[j2,j1] = 1;
									*Rf[j2,j1] = 0; /*if LOD is stored in lower triangular elements of Rf*/
								end;
								else if cross = 3 then do;
									Rf[j1,j2]= R/(2*(1-R));
									Rf[j2,j1]= R/(2*(1-R));

									*compute LOD*;	
									*print R rc;
									if R > 0 then do;
										L1_= (R**rc)*(1-R)**(totlines-rc); 
										L0_= (0.5**rc)*(1-0.5)**(totlines-rc);
										LOD[j1,j2] = log10(L1_/L0_);
										LOD[j2,j1] = LOD[j1,j2];
									end;
									else do;
										LOD[j1,j2] = 99;
										LOD[j2,j1] = 99;
									end;
									LOD_= LOD[j1,j2];
									*print L1_ L0_;
									*Rf[j2,j1]= log10(L1_/L0_);
								end;
								else if cross = 7 then do;
									maxval = &MaxR + .10;
									Rf[j1,j2] = iril_self_bisect(R,0,maxval,0.001,&GenNum);
									Rf[j2,j1] = Rf[j1,j2];
									
									R = (1/2)*(1 - ((1 - 2*Rf[j1,j2])/(1 + 2*Rf[j1,j2]))*(1 - Rf[j1,j2])**(&GenNum - 2));
									*print R rc;
									if R > 0 then do;
										L1_= (R**rc)*(1-R)**(totlines-rc); 
										L0_= (0.5**rc)*(1-0.5)**(totlines-rc);
										LOD[j1,j2] = log10(L1_/L0_);
										LOD[j2,j1] = LOD[j1,j2];
									end;
									else do;
										LOD[j1,j2] = 99;
										LOD[j2,j1] = 99;
									end;					
									LOD_= LOD[j1,j2];
									*print L1_ L0_ LOD_;
									*Rf[j2,j1] = log10(L1_/L0_);
								end;
							end;
							
						end;
						else do;  /* RILsib*/
							if &RILBiasCorr then do;
								/*use bias corrected formula (Martin and Hospital 2006)*/
								Rf[j1,j2]= min(&MaxR,(rc/(4*totlines - 6*rc))*(1-((6*(totlines - rc))/((2*totlines - 3*rc)**2))));
								Rf[j2,j1]= min(&MaxR,(rc/(4*totlines - 6*rc))*(1-((6*(totlines - rc))/((2*totlines - 3*rc)**2))));
								
								*compute LOD*;
								R = (4*Rf[j1,j2])/(1 + 6*Rf[j1,j2]);
								*print R rc;
								if R > 0 then do;
									L1_= (R**rc)*(1 - R)**(totlines - rc); 
									L0_= (0.5**rc)*(1 - 0.5)**(totlines - rc);
									LOD[j1,j2] = log10(L1_/L0_);
									LOD[j2,j1] = LOD[j1,j2];
								end;
								else do;
									LOD[j1,j2] = 99;
									LOD[j2,j1] = 99;
								end;
								LOD_= LOD[j1,j2];
								*print L1_ L0_ LOD_;
								*Rf[j2,j1]= log10(L1_/L0_);
							end;
							else do;
								/* use traditional formula (Martin and Hospital 2006)*/
								R = min(&MaxR,rc/totlines);
								if R = 1 then do;
									Rf[j1,j2] = 1;
									Rf[j2,j1] = 1;
									*Rf[j2,j1] = 0; /*if LOD is stored in the lower triangular elements of Rf*/
								end;
								else if cross = 4 then do;
									Rf[j1,j2]= R/(4-6*R);
									Rf[j2,j1]= R/(4-6*R);
									
									*compute LOD*;	
									*print R rc;
									if R > 0 then do;
										L1_= (R**rc)*(1-R)**(totlines-rc); 
										L0_= (0.5**rc)*(1-0.5)**(totlines-rc);
										LOD[j1,j2] = log10(L1_/L0_);
										LOD[j2,j1] = LOD[j1,j2];
									end;
									else do;
										LOD[j1,j2] = 99;
										LOD[j2,j1] = 99;
									end;									
									LOD_= LOD[j1,j2];
									*print L1_ L0_ LOD_;
									*Rf[j2,j1]= log10(L1_/L0_);
								end;
								else if cross = 8 then do;
									maxval = &MaxR + .10;
									Rf[j1,j2] = iril_sib_bisect(R,0,maxval,0.001,&GenNum);
									Rf[j2,j1] = Rf[j1,j2];
									
									R = (1/2)*(1 - ((1 - 2*Rf[j1,j2])/(1 + 6*Rf[j1,j2]))*(1 - Rf[j1,j2])**(&GenNum - 2));
									*print R rc;
									if R > 0 then do;
										L1_= (R**rc)*(1-R)**(totlines-rc); 
										L0_= (0.5**rc)*(1-0.5)**(totlines-rc);
										LOD[j1,j2] = log10(L1_/L0_);
										LOD[j2,j1] = LOD[j1,j2];
									end;
									else do;
										LOD[j1,j2] = 99;
										LOD[j2,j1] = 99;
									end;									
									LOD_= LOD[j1,j2];
									*print L1_ L0_ LOD_;
									*Rf[j2,j1] = log10(L1_/L0_);
								end;			
							end;
						end;
					end;
				end;
				/* S0742036:  replace missing with recomb of 0.5, assume no recombination */
				if Rf[j1,j2] = . then do;
					Rf[j1,j2]=0.5;
					Rf[j2,j1]=0.5;
					*Rf[j2,j1]=0.0; /*if LOD is stored in the lower triangular elements of Rf*/
					LOD[j1,j2]=0.0;
					LOD[j2,j1]=0.0;
				end;
				else if Rf[j1,j2] = 0 & LOD[j1,j2] = . then do;
					LOD[j1,j2]=99;
					LOD[j2,j1]=99;
				end;
				/* calc Haldane's
				rval = Rf[j1,j2];
				if rval = . then Gd[j1,j2] = .;
				else if rval = 0 then Gd[j1,j2] = 0;
				* cap cM at 250 so it doesn't go large OOP;
				else if rval >= 0.4966 then Gd[j1,j2] = 100*(-(1/2)*log(1-2*0.4966));	 
				else Gd[j1,j2] = 100*(-(1/2)*log(1-2*Rf[j1,j2]));
				Gd[j2,j1] = Gd[j1,j2];
				*/
			end;
		end;
		
		create rfmat from Rf;
		append from Rf;
		
		create lodmat from LOD;
		append from LOD;
				
		%if &_RandLG_AP_ = 1 %then %do;
				
			create srmat from Sr;
			append from Sr;
		
			%if %index(%upcase(&GroupingMethod),MINIMUM) %then %do;
				rstack = symsqr(Rf);
				
				create rstack from rstack;
				append from rstack;
			
				lodstack = symsqr(LOD);
				create lodstack from lodstack;
				append from lodstack;
				
			%end;
		%end;
		
		%if &_swindow_=1 %then %do;
			rfadj= 0;
			lodadj= 0;
			do i=1 to n_marorig;
				do j=i+1 to min(i+1, n_marorig);
					rfadj= rfadj // Rf[i,j];
					lodadj= lodadj // LOD[i,j];
				end;	
			end;			
			create rfadj from rfadj;
			append from rfadj;
			
			create lodadj from lodadj;
			append from lodadj;			
		%end;
		
		*close gendat;
		close _gendata_;
	quit;
	options notes;

	%if &syserr>4 %then %do;
		%put ERROR: Calculation of pairwise recombination could not be completed.;
		%put ERROR: Please be sure the appropriate columns were selected as Marker Variables and the Marker Variable Format coding for possible errors.;
		%let exiterror=1;
		%goto exitm;
	%end;
	
	/*data OutLib.rfmat; set rfmat; run;*/
	/*data OutLib.lodmat; set lodmat; run;*/
	
	%if &_RandLG_AP_ = 1 %then %do;
		%let CountList = AA_Pct AB_Pct BB_Pct NotAA_Pct NotBB_Pct;
		
		proc means data = srmat sum noprint;
			var Col1 - Col6;
			output out = csums sum = /autoname;
		run;
	
		data _null_; 
			set csums;
			%do i=1 %to 5;
				call symput("Cts&i",trim(left(Col&i._sum)));
			%end;
			call symput("chisum",trim(left(Col6_sum)));
		run;
	
		/* add to check if the ESR were possibly bad such as divide by 0 case */
		%global Col7Name;
		%let Col7Name = NegLog10_ProbChisq;
		
		%global notests;
		%if  &no_st = 0 %then %do;
			%if &chisum = . %then %do;
				%let notests = 1;
				%put ERROR: Statistical tests for segregation ratios could not be computed. This could be due to a division by 0. Please check that the specified segregation ratios match your experimental cross.;
				%put ; 
				%if &cross = 2 or %length(&SegRatios) %then %do;
					%put Segregation Ratios tested: AA = &AA_ef, AB = &AB_ef, BB = &BB_ef.;
				%end;
				%else %if &cross = 1 and &Parent =1 %then %do;
					%put Segregation Ratios tested: AA = &AA_ef, AB = &AB_ef, BB = 0;
				%end;
				%else %if &cross = 1 and &Parent =2 %then %do;
					%put Segregation Ratios tested: AA = 0, AB = &AB_ef, BB = &BB_ef.;
				%end;	
				%else %do;
					%put Segregation Ratios tested: AA = &AA_ef, AB = 0, BB = &BB_ef.;
				%end;	
			%end;	
			%else %do;
				%let notests = 0;
				%if %length(&MultipleTestingMethod) %then %do;

					%AdjPValueName(method=&MultipleTestingMethod);

					data srmat_raw;
						set srmat;
						rename col7 = raw_p;
					run;

					proc multtest pdata=srmat_raw &MultipleTestingMethod out=_op noprint;
					run;			

					data srmat_adj;
						set _op;
						if not missing(&mtpvar) then do;
							if (&mtpvar < 10**(-15)) then &mtpvar = 15;
							else &mtpvar = -log10(&mtpvar);
						end;
						keep &mtpvar;
					run;		

					data srmat;
						merge srmat(drop=col7) srmat_adj;
						rename &mtpvar = Col7;
					run;			
					%let mtm = %substr(&mtpvar,1,%length(&mtpvar)-2);
					%let Col7Name = NegLog10_ProbChisq_&mtm.Adj;
				%end;
			%end;
		%end;
		%else %let notests = 1;

		%global JMPClist;
		%let JMPCList = ;
		%do i=1 %to 5;
			%if &&Cts&i > 0 %then %do;
				%let JMPCList = &JMPCList %scan(&CountList,&i,' ');
			%end;
		%end;
		%put  ;
		%put  Genotypes with Counts > 0 = &JMPCList;
		
	%end;

	%if %index(%Upcase(&MarkerFmt), ALLELES) and %length(&AnnoInData) %then %do;
		data _null_;
			set &_annodata_ end = last; 
			call symputx("AMName"||trim(left(_n_)), &AnnoLabel);
			if last then call symputx("Totnames",_n_);
		run;
	%end;

	data mnames3; set &_gendata_;
		where 0;
	run;
	%let _tid3_ = %sysfunc(open(mnames3));
	%let _nVars3_ = %sysfunc(attrn(&_tid3_,NVARS));
	
	%*let maxvarlength = 27;

	options nonotes;
	proc datasets library=work nolist;
		modify rfmat;
		%if %index(%Upcase(&MarkerFmt), ALLELES) %then %do;
			%let _nm_ = %sysfunc(floor(%eval(&_nVars3_/2)));
			%put There are &_nm_ Markers;
			%if %length(&AnnoInData) %then %do;
				%if &Totnames >= &_nm_ %then %do;
					rename %do i=1 %to &_nm_; Col&i = &&AMName&i %end;;
				%end;
				%else %do;
					%put ERROR: The number of markers formed from the alleles in the Input Data Set do not match with the the number of marker names in the Annotation Data Set.;
					%let exiterror =1;
					%goto exitm;
				%end;
			%end;
			%else %do;
				rename %do i=1 %to &_nm_; Col&i = Marker_&i %end;;
			%end;
		%end;
		%else %do;
			rename 
				%do i=1 %to &_nVars3_; 
					Col&i = %sysfunc(varname(&_tid3_,&i))
					/*Col&i = %sysfunc(compress(%sysfunc(cats(_REC_,&i))))*/
				%end;
			;
		%end;
		%if &_RandLG_AP_ = 1 %then %do;
			modify srmat;
			rename %do i=1 %to 5; Col&i = %scan(&CountList,&i,' ') %end; Col6 = SegRatio_Chisq Col7 = &Col7Name Col8 = A_minus_B;
		%end;
		modify lodmat;
		%if %index(%Upcase(&MarkerFmt), ALLELES) %then %do;
			%let _nm_ = %sysfunc(floor(%eval(&_nVars3_/2)));
			%put There are &_nm_ Markers;
			%if %length(&AnnoInData) %then %do;
				%if &Totnames >= &_nm_ %then %do;
					rename 
						%do i=1 %to &_nm_; 
							Col&i = &&AMName&i
							/*Col&i = compress("_LOD_"||substr(&&AMName&i,1,&maxvarlength))*/ 
						%end;
					;
				%end;
				%else %do;
					%put ERROR: The number of markers formed from the alleles in the Input Data Set do not match with the the number of marker names in the Annotation Data Set.;
					%let exiterror =1;
					%goto exitm;
				%end;
			%end;
			%else %do;
				rename 
					%do i=1 %to &_nm_; 
						Col&i = Marker_&i 
					%end;
				;
			%end;
		%end;
		%else %do;
			rename 
				%do i=1 %to &_nVars3_; 
					Col&i = %sysfunc(varname(&_tid3_,&i))
					/*Col&i = %sysfunc(compress(%sysfunc(cats(_LOD_,%substr(%sysfunc(varname(&_tid3_,&i)),1,&maxvarlength)))))*/ 
				%end;
			;
		%end;		
	quit;
	%let _rc3_ = %sysfunc(close(&_tid3_)); *close opened file*; 
	
	options notes label;
		
	%if &_RandLG_AP_ = 1 %then %do;
		data srmat;
			set srmat;
			label A_minus_B="A_Pct minus B_Pct";
			drop %do i=1 %to 5; %if &&Cts&i = 0 or &&Cts&i = . %then %scan(&CountList,&i,' '); %end;;
			%if &no_st %then %do;
				drop  SegRatio_Chisq &Col7Name;
			%end;
			%else %do;
				label &Col7Name= "&Col7Name";
			%end;
		run;
	%end;
	
	%exitm:;
%mend;
