/*
JMP Genomics Analytical Process

Name: ConcatenateData
Author: Luciano Silva, SAS Institute Inc.
Support: Luciano Silva, SAS Institute Inc.
History:
   sbrlsi 03Aug20 initial coding 

Description:
      Concatenate Data (code name Concatenate Data) combine rows 
      of SAS data sets into a single data set containing rows of 
      all input files.
*/

/*
-----------------------------------------------------------
Copyright (C) 2020 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InPath FileType InFiles OutPath InData
KeepVars DropVars ServerOutPath MacroPath OutName 
CData_OutData CDataScript;

*ProcessBody;

%macro ConcatenateData;

	*initialization and argument checking; 
	%let ProcessName = ConcatenateData;
	%let FullAPName = %str(Concatenate Data);
	%include "&MacroPath./UtilityMacros.sas" / nosource;

	%CheckEndSlash(InPath);
	%let fLen = 0;
	%let FileList = %trim(%left(&InFiles)); 	
	%put Infiles=&FileList;
	libname DataLib "&InPath";
	
	* check name for OutName;
	%let maxnamelength = 28;
	%let OrigInData= %left(%scan(&FileList,1,'|'));
	%CheckDataName(_name=OutName,_default=&OrigInData);
	%if %length(&OutName)>&maxnamelength %then %let OutName=%substr(&OutName,1,&maxnamelength);
	
	%PrepOutDS(CData_OutData,CData_OutData_OutDataName,&OutName._ccd);
	
	data vtypes;
		length varname vartype $32;
		delete;
	run;	
	
	%let nfiles= 0;
	%let NewFileList= ;
	
	%if %length(&InData) %then %do;
		%PathName(&InData);
		%if &exiterror %then %do;
			%put ERROR: Input data set specification is invalid.;
	   		%goto exit;
		%end;
		%let InData = %trim(%left(&TmpName));
		libname InLib "&TmpPath";
		
		%let nfiles= %eval(&nfiles + 1);
		%let NewFileList= InLib.&InData %str('|');
		
		data vnames;
			set InLib.&InData;
			where 0;
			%if %length(&KeepVars) %then %do;
				keep &KeepVars;	
			%end;
			%if %length(&DropVars) %then %do;
				drop &DropVars;	
			%end;			
		run;
		
		%NOBsVars(vnames);
		%let nvvars= &nvars;
		%put nvvars= &nvvars;
		%if &nvvars %then %do;		
			proc contents data=vnames out=content noprint;
			run;
		%end;
		%else %do;			
			proc contents data=InLib.&InData out=content noprint;
			run;
		%end;
		
		/*data OutLib.content; set content; run;*/
		
		data content;
			set content;
			if type= 1 then vartype="N";
			else vartype="C";
			varfile= &nfiles;
			rename name=varname length=varleng;
			keep name vartype length varfile varnum;
		run;
		
		proc sort data=content out=content;
			by varnum;
		run;
		
		data vtypes;
			set vtypes content;
		run;		
	%end;

	%put NewFileList= &NewFileList;	
	
	%*goto exit;
	
	%let flagdupbase= 0;
	%let i=1;
	%let FileName = %left(%scan(&FileList,&i,'|'));
	%do %while(%length(&FileName));
	
		%put FileName = &FileName;
		%let fsas7bdat = %sysfunc(find(&FileName, .sas7bdat));
		%put fsas7bdat = &fsas7bdat;
		%let FileName = %substr(&FileName, 1, &fsas7bdat - 1);
		%put FileName = &FileName;
		
		%if %length(&InData) %then %do;
			proc compare data=InLib.&InData compare=DataLib.&FileName noprint;
			run;
			
			%if &sysinfo. = 0 %then %do;
				%put WARNING: "Base SAS Data Set &InData is identical to Concatenate SAS Data Set &FileName..";
				%let flagdupbase= 1;
			%end;
			%else %do;
				%let flagdupbase= 0;
			%end;			
		%end;
		
		%if &flagdupbase = 0 %then %do;
			%let nfiles= %eval(&nfiles + 1);
			%let NewFileList= &NewFileList DataLib.&FileName %str('|');

			data vnames;
				set DataLib.&FileName;
				where 0;
				%if %length(&KeepVars) %then %do;
					keep &KeepVars;	
				%end;
				%if %length(&DropVars) %then %do;
					drop &DropVars;	
				%end;			
			run;

			%NOBsVars(vnames);
			%let nvvars= &nvars;
			%put nvvars= &nvvars;
			%if &nvvars %then %do;		
				proc contents data=vnames out=content&i noprint;
				run;
			%end;
			%else %do;			
				proc contents data=DataLib.&FileName out=content&i noprint;
				run;
			%end;

			/*data OutLib.content&i; set content&i; run;*/

			data content&i;
				set content&i;
				if type= 1 then vartype="N";
				else vartype="C";
				varfile=&nfiles;
				rename name=varname length=varleng;
				keep name vartype length varfile varnum;
			run;

			proc sort data=content&i out=content&i;
				by varnum;
			run;

			data vtypes;
				set vtypes content&i;
			run;
		%end;
			
		%let file&i= &FileName;
		%let i=%eval(&i+1);
		%let FileName = %left(%scan(&FileList,&i,'|'));
		%put FileName = &FileName;
	%end;
	%put nfiles= &nfiles;
	%put NewFileList= &NewFileList;	

	%if &nfiles < 2 %then %do;
		%put ERROR: The &ProcessName needs two or more SAS Data Set that are not identical to run properly.;
		%let exiterror=1;
		%goto exit;
	%end;
		
	/*data OutLib.vtypes0; set vtypes; run;*/
	
	%let nvarkeep= 0;
	%NOBsVars(vtypes);
	%put nobs= &nobs;
	%if &nobs %then %do;
	
		proc sort data=vtypes out=vtypes;
			by varname varleng;
		run;
		
		data vtypes;
			set vtypes end=_eof_;
			by varname;
			if _n_= 1 then count=0;
			*if first.varname then len= varleng;
			*if last.varname and len ^= varleng then do;
			if last.varname then do;
				count= count + 1;
				call symputx("_varname"||strip(count),trim(left(varname)));	
				call symputx("_vartype"||strip(count),trim(left(vartype)));
				call symputx("_varleng"||trim(left(count)),trim(left(varleng)));	
			end;						
			retain count;
			*retain len;
			if _eof_ then call symputx("nvarkeep", count);
		run;
		
		/*data OutLib.vtypes1; set vtypes; run;*/
	%end;
	
	%put nvarkeep= &nvarkeep;
	%do j=1 %to &nvarkeep;
		%put _varname&j= &&_varname&j; 
		%put _vartype&j= &&_vartype&j; 
		%put _varleng&j= &&_varleng&j; 
	%end;
	
	data vbase;
		set %scan(&NewFileList,1,'|');
		where 0;
		%if %length(&KeepVars) %then %do;
			keep &KeepVars;	
		%end;
		%if %length(&DropVars) %then %do;
			drop &DropVars;	
		%end;			
	run;
	
	%NOBsVars(vbase);
	%let nbvars= &nvars;
	%put nbvars= &nbvars;
	%if &nbvars %then %do;	
		%let _tid_ = %sysfunc(open(vbase));		
	%end;
	%else %do;
		%let _tid_ = %sysfunc(open(%scan(&NewFileList,1,'|')));
	%end;
	%let _nVars_ = %sysfunc(attrn(&_tid_,NVARS));
	%put _nVars_= &_nVars_;	
	
	data OutLib.&CData_OutData_OutDataName;
		%if &_nVars_ %then %do;
			*maintaining the order of variables as is in the first data file*;
			retain %do j=1 %to &_nVars_; %sysfunc(varname(&_tid_,&j)) %end;;
		%end;
		length %do j=1 %to &nvarkeep; &&_varname&j %if "&&_vartype&j" = "C" %then $; &&_varleng&j %end;;
		set %scan(&NewFileList,1,'|');
		keep %do j=1 %to &nvarkeep; &&_varname&j %end;;
	run;
	%let rc = %sysfunc(close(&_tid_));
	
	%*goto exit;
	
	%do i=2 %to &nfiles;		
		data vbase;
			set %scan(&NewFileList,&i,'|');
			where 0;
			%if %length(&KeepVars) %then %do;
				keep &KeepVars;	
			%end;
			%if %length(&DropVars) %then %do;
				drop &DropVars;	
			%end;			
		run;
	
		%let _tid_ = %sysfunc(open(vbase));
		%let _nVars_ = %sysfunc(attrn(&_tid_,NVARS));
		%put _nVars_= &_nVars_;
		
		proc append 
			base=OutLib.&CData_OutData_OutDataName 
			data= %scan(&NewFileList,&i,'|')
			%if &_nVars_ %then %do;
				(keep= %do j=1 %to &_nVars_; %sysfunc(varname(&_tid_,&j)) %end;)
			%end;
			force;
		run;
		%let rc = %sysfunc(close(&_tid_));
	%end;	

	%PrepOutFile(CDataScript,CDataScriptName,&ProcessName Script for &OutName..jsl);
	
	%global _mainjslfile_;
	%let _mainjslfile_ = %quote(&CDataScript);
	
	**creating a JSL file without code just to force the final jsl file to contain the process name on it instead of sasmessage**;
	* create JSL code;
	data _null_;
		file "&CDataScript" lrecl=32767;
		put '//!';
		put "Names Default to Here(1);";
		put "Here:DataList = {};";
	run;
	%PackageInsertFile(jsl code from &ProcessName, &OutPath, &CDataScriptName, , text/plain, yes, yes);
	
	%PackageInsertDataset(Concatenated SAS Data Set data from &ProcessName,outlib,&CData_OutData_OutDataName,No);

	%exit:;
	%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend;
%ConcatenateData;