/*
JMP Genomics Analytical Process

Name: DataTransposeTalltoWide

History:
   tzuchu 23Jun03 initial coding 
   sasrdw 27Apr05 converted to tall input format
   sasrdw 13Sep05 added wide->tall
   saslii Sep2810 split the datatranspose into two different APs: TalltoWide and WidetoTall.

Description:
      Data Transpose transposes a SAS data set
      in tall format, along with an accompanying
      experimental design table, into a data
      set in wide format in which the rows
      and columns are transposed.  
      
      Conversely, you can transpose a wide data
      set into a tall one plus an accompanying
      experimental design data set.

      Be aware that wide data sets can 
      become unwieldy when there are thousands 
      of columns. In this case, it can be more 
      efficient to first use other analytical
      processes to reduce the number of
      rows in a tall input data set.  For example:
      
      	- Use Data Summary to reduce the 
      	  number of rows in the original
      	  data set by computing a summary statistic,
      	  
      	- Use K-Means Clustering to cluster the rows
      	  into similar groups.
      	  
      	- Use Data Merge or Data Step to extract
      	  a subset of the rows.
 */

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName ExpInData InData NamingVars NamingPrefix 
OutData OutPath ServerOutPath MacroPath;

*ProcessBody;

%macro DataTransposeTalltoWide;

*initialization and argument checking;
 
%let ProcessName = DataTransposeTalltoWide;
%let FullAPName = %str(Transpose Tall to Wide);
%include "&MacroPath./UtilityMacros.sas" / nosource;

%global OrigInData;

%GetOrigFileName(InData,OrigInData);
%CheckInData(_ExpFlag=1);
%if ^%length(&OutData) %then %AddSuffixToDataName(_wid,&OrigInData,OutData,_SDS=1);
%CheckDataName(_name=OutData,_default=&OrigInData);

%if &exiterror %then %goto exit;

libname OutLib "&OutPath"; 


* determine number of naming variables and construct id variable;
%let nv = 1;
%put NamingVars = &NamingVars NV = &nv;
%do %while(%length(%scan(&NamingVars,&nv,' ')));
	%let nv = %eval(&nv + 1);
%end;
%let nv = %eval(&nv - 1);
%if (&nv > 1) %then %do;
	data ind;
		set InLib.&InData;
		_id = trim(left(%scan(&NamingVars,1,' '))) || 
		%do v=2 %to &nv;
			"_" || trim(left(%scan(&NamingVars,&v,' '))) || 
		%end;
		"";
	run;
	%let ds = ind;
	%let ID_Vars = _id;
%end;
%else %do;
	%let ds = InLib.&InData;
	%let ID_Vars = &NamingVars;
%end;
%put ID_Vars = &ID_Vars. NV = &nv;


%let oriPrefix=;

%if %length(&NamingPrefix)=0 and &nv=0 %then %let NamingPrefix=Var;
	
%if %length(&NamingPrefix) %then %do;
	%let oriPrefix=&NamingPrefix;
	%Naming(_SingleVar=NamingPrefix);
%end;

/*
data tall;
	set &ds;
run;

%Nobsvars(tall);
*/
%Nobsvars(&ds);
%let nTallRows = &nobs;
%let BlockSize=%sysfunc(min(&nTallRows,50000));
%let first =1;

%* messages to the log of sizes for blocking
%put nTallRows = &nTallRows;
%put Blocksize = &BlockSize;

%if &nTallRows = &BlockSize %then %do;
	%let NBlocks = 1;
%end;
%else %do;
	%let NBlocks = %eval(&nTallRows/&BlockSize+1);
%end;
%let last =0;
* save the sas label and convert space and special characters to _;

/*
%put NamingPrefix is &NamingPrefix;
%if %length(&oriPrefix) AND %length(&ID_Vars) %then %do;
	data &ds;
		set &ds;
		*concate prefix to labels*;
		ID_label = cats("&oriPrefix",trim(left(&ID_Vars)));
	run;
%end;
%else %if %length(&oriPrefix) AND %length(&ID_Vars)=0 %then %do;
	data &ds;
		set &ds;
		*concate prefix to labels*;
		ID_label = cats("&oriPrefix","");
	run;
%end;
%else %do;
	data &ds;
		set &ds;
		*concate prefix to labels*;
		ID_label = trim(left(&ID_Vars));
	run;
%end;
*/
	
%do b=1 %to &NBlocks;
	%put Block &b out of &NBlocks;
	%let first =%eval(&last+1);
	%let last =%eval(&first + &BlockSize-1);

	%put NamingPrefix is &NamingPrefix;
	
	%if %length(&oriPrefix) AND %length(&ID_Vars) %then %do;
		*data &ds;
		data ds&b;
			set &ds;
			*concate prefix to labels*;
			ID_label = cats("&oriPrefix",trim(left(&ID_Vars)));
			if _n_<=&last & _n_>=&first;
		run;
	%end;
	%else %if %length(&oriPrefix) AND %length(&ID_Vars)=0 %then %do;
		*data &ds;
		data ds&b;
			set &ds;
			*concate prefix to labels*;
			ID_label = cats("&oriPrefix","");
			if _n_<=&last & _n_>=&first;
		run;
	%end;
	%else %do;
		*data &ds;
		data ds&b;
			set &ds;
			*concate prefix to labels*;
			ID_label = trim(left(&ID_Vars));
			if _n_<=&last & _n_>=&first;
		run;
	%end;

	/*
	data ds&b;
		set &ds;
		if _n_<=&last & _n_>=&first;
	run;
	*/

	proc transpose data=ds&b out=wide&b
		%if %length(&NamingPrefix) %then %do; prefix=&NamingPrefix %end; ;
		var %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
		%if (&nv ge 1) %then %do;
			id &ID_Vars;
			idlabel ID_label;
		%end;
	run;
	%if &syserr %then %do;
		%put ERROR: The naming variables likely do not have unique values for each row.;
	%end;
	data wide&b;
		set wide&b;
		ColumnName = _name_;
		drop _name_;
	run;
	
	proc sort data=wide&b;
		by ColumnName;
	run;
		
	data ed;
		set ExpInLib.&ExpInData;
		indCount = _n_;
	run;
	
	proc sort data=ed;
		by ColumnName;
	run;
	%CheckVarExist(Intensity,ed,,_ShowError=0);
	%if &exiterror %then %let DropVars = &FileAlias;
	%else %let DropVars = &FileAlias Intensity;
	%let exiterror = 0;
	
	%if &b =1 %then %do;
			
		data OutLib.&OutData;
			merge ed wide&b;
			by ColumnName;
			drop &DropVars;
		run;
	%end;
	%else %do;
		data OutLib.&OutData;
			set OutLib.&OutData; 
			set wide&b;
		run;

		%if &syserr > 4 %then %do;
			%put ERROR: Check the SAS log for more information.;
			%let exiterror=1;
			%goto exit;
		%end;
	%end;
		
proc datasets nolist;
	delete wide&b ds&b;
	quit;
%end;
proc sort data=OutLib.&OutData;
	by indCount;
run;
data OutLib.&OutData;
	set OutLib.&OutData;
	drop indCount;
run;	

%PackageInsertDataset(%nrbquote(%LocaleFill(%nrbquote(%LocaleGet(WidedatafromX_Text)),
	%nrbquote(%LocaleGet(&ProcessName._Label)))), Outlib, &OutData);


%exit:;

%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend DataTransposeTalltoWide;

* run the macro;
%DataTransposeTalltoWide;

