/*
JMP Life Sciences Analytical Process

Name: Local Control
Authors: Russ Wolfinger, SAS Institute Inc.

Description:
 
    	Local Control estimates treatment effects within
    	clusters of subjects as described in Lopiano, Obenchain and Young (2014).
    	It allows you to interactively select the number of clusters and cumulatively compare
    	results.
*/

/*
-----------------------------------------------------------
Copyright (C) 2013 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData ClassVar TrtVar PredictorVars ClassPredVars WeightVar 
ListContPredVars ListClassPredVars 
ScreenControlValue ClusterEffectType VariabilityAssumption Seed
LocalControl_OutData LocalControlScript
OutPath MacroPath;

*ProcessBody;


%macro LocalControl;

*initialization and argument checking;
%let ProcessName = LocalControl; 
%let FullAPName = %str(Local Control);

%include "&MacroPath./UtilityMacros.sas" / nosource;

%CheckInData;

%if %length(&LocalControl_OutData) %then %do;
	%let prefix = &LocalControl_OutData;
	%let LocalControl_OutData = ;
%end;
%else %let prefix = &InData;

%PrepOutDS(LocalControl_OutData,LocalControl_OutDataName,&prefix,_lc);

data all;
	set InLib.&InData;
	if missing(&ClassVar) or missing(&TrtVar) then delete;
run;

%NObsVars(all);
%if ^&nobs %then %do;
	%put ERROR: No valid observations due missing response or treatment variables.;
	%let exiterror = 1;
	%goto vexit;
%end;

%if %length(&WeightVar) %then %do;
	data all;
		set all;
		if missing(&WeightVar) or &WeightVar <= 0 then delete;
	run;
	%NObsVars(all);
	%if ^&nobs %then %do;
		%put ERROR: No valid observations due either to missing or nonpositive weight values.;
		%let exiterror = 1;
		%goto vexit;
	%end;
%end;

* get treatment levels;
proc sort data=all nodupkey out=tl;
	by &TrtVar;
run;

%NObsVars(tl);
%if &nobs ne 2 %then %do;
	%put ERROR: The treatment variable must have exactly two levels.  &TrtVar has &nobs levels..;
	%let exiterror = 1;
	%goto vexit;
%end;
data _null_;
	set tl;
	call symputx("tlev"||trim(left(_n_-1)),trim(left(&TrtVar)));
run;

%if %length(&ScreenControlValue) %then %do;
	%let trPlacebo = %AddQuotes(&ScreenControlValue);
	%if (%qupcase("&tlev0") ne %qupcase(&trPlacebo)) and (%qupcase("&tlev1") ne %qupcase(&trPlacebo)) %then %do;
		%put ERROR: The Treatment Control Level &ScreenControlValue does not match either of the treatment levels.;
		%put Please specify "&tlev0","&tlev1", or a blank value on the Analysis tab.;
		%let exiterror = 1;
		%goto exit;
	%end;
%end;
%else %do;
	%let trPlacebo = %str(%")%nrbquote(%LocaleGet(Placebo_Text))%str(%");
%end;

%if %qupcase("&tlev1") = %qupcase(&trPlacebo) %then %do;
	%let temp = &tlev1;
	%let tlev1 = &tlev0;
	%let tlev0 = &temp;
%end;
%put tlev0 = &tlev0;
%put tlev1 = &tlev1;
%let dsid = %sysfunc(open(tl));
%let tpos = %sysfunc(varnum(&dsid,&TrtVar));
%let cpos = %sysfunc(varnum(&dsid,&ClassVar));
%global ttype ctype;
%let ttype = %sysfunc(vartype(&dsid,&tpos)); 
%let ctype = %sysfunc(vartype(&dsid,&cpos)); 
%let dsid = %sysfunc(close(&dsid)); 
%put ttype = &ttype;
%let teffect = _Trt01_;
%let teffect2 = _Trt01_2;

data all;
	set all;
	_id_ = _n_;
	%if &ttype = C %then %do;
		if &TrtVar = "&tlev0" then &teffect = 0;
		else &teffect = 1;
	%end;
	%else %do;
		if &TrtVar = &tlev0 then &teffect = 0;
		else &teffect = 1;
	%end;
run;

* determine type of target variable;
%if &ctype = C %then %do;
	%let trait_type = qualitative;
%end;
%else %do;
	proc sort data=all nodupkey out=tl;
		by &ClassVar;
	run;
	%NObsVars(tl);
	%if &nobs < 2 %then %do;
		%put ERROR: Dependent variable &ClassVar is constant.  It must have at least two distinct values in order to model it.;
		%let exiterror = 1;
	%end;
	%else %if &nobs = 2 %then %do;
		%put Deeming &ClassVar to be binary since it only has two levels;
		%let trait_type = qualitative;
	%end;
	%else %do;
		%let trait_type = quantitative;
	%end;
%end;

%if &trait_type = qualitative %then %do;
	%if &ctype = C %then %do;
		proc sort data=all nodupkey out=tl;
			by &ClassVar;
		run;
		%NObsVars(tl);
	%end;
	data _null_;
		set tl;
		call symputx("clev"||trim(left(_n_)),trim(left(&ClassVar)));
	run;
	%let nclev = &nobs;
	%put nclev = &nclev;
%end;

data &LocalControl_Outdata;
	set all;
	keep &ClassVar &TrtVar &PredictorVars &ClassPredVars &WeightVar &ListContPredVars &ListClassPredVars;
run;

%let DepLabel = &ClassVar;
%CheckLabel(all, DepLabel);
%let TrtLabel = &TrtVar;
%CheckLabel(all, TrtLabel);
%let WeightLabel = &WeightVar;
%if %length(&WeightVar) %then %do;
	%CheckLabel(all, WeightLabel);
%end;
%if ^%length(&Seed) %then %let Seed = 12345;
data xv;
	set all;
	if (0);
	keep &PredictorVars &ClassPredVars &ListContPredVars &ListClassPredVars;
run;

%NObsVars(xv);

%let xvid = %sysfunc(open(xv));

* create JMP script;
%PrepOutFile(LocalControlScript,LocalControlScriptName,&ProcessName Script for &LocalControl_OutDataName..jsl);

%global _mainjslfile_;
%let _mainjslfile_ = &LocalControlScript;


data _null_;
   file "&LocalControlScript"  encoding="utf-8";
   put "//!";
   put "Names Default to Here(1);";
   put "include(""&CLIENT_LS_HOME.JSLFiles\Local Control Expressions.jsl"");";
	put "ds = ""&ClientOutPath.&LocalControl_OutDataName..sas7bdat"";";
   put "yv = {""&DepLabel""};";
   put "tv = {""&TrtLabel""};";
   put "wv = {""&WeightLabel""};";
   put "xv = {";
   %do i=1 %to &nvars;
    %let vlab = %sysfunc(varlabel(&xvid,&i));
    %*put vlab = &vlab;
   	%if %length(%trim(&vlab)) %then %do;
   		put " ""&vlab"" " @@;
   	%end;
   	%else %do;
   		%let vname = %sysfunc(varname(&xvid,&i));
   		%*put vname = &vname;
   		put " ""&vname"" " @@;
   	%end;
   	%if &i < &nvars %then %do;
   		put ",";
   	%end;
   %end;
   put "};";
   put "cetype = ""&ClusterEffectType"";";
   %if &VariabilityAssumption = Homoskedastic %then %do;
   	put "varas = 1;";
   %end;
   %else %do;
   	put "varas = 2;";
   %end;
   put "numy = 1;";
   put "tclev = &trPlacebo;";
   put "seed = &Seed;";
   put "Here:TabCodeList = {lcfunc()};";
	put "Here:TabButtonList = {locale:get(""Overview_Text"")};";
	put "Here:TabOpenList = {1};";	
	put "Here:TabAfterCodeList = {"" ""};";	
	put "Here:DataList= {{""dt"",""dtsum"",""dtmain""}};";	
run;

%let xvid = %sysfunc(close(&xvid));

%PackageInsertFile(jsl code from &ProcessName, &OutPath, &LocalControlScriptName, &LocalControl_OutDataName..sas7bdat, text/plain, yes);

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);


%mend LocalControl;

* run the macro;
%LocalControl;


