/*
JMP Genomics Analytical Process

Name: ArrayGroupCorrelation
Author: Tzu-Ming Chu, SAS Institute Inc.
Support: Tzu-Ming Chu, SAS Institute Inc.
History:
   tzuchu 28Aug02 initial coding  
	sasrdw 05Apr05 convert to tall input format                 

Description: 
Array Group Correlation computes 
correlations and scatterplot matrices
for expression measurements across 
groups of arrays. 
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData  WhereStmt ExpInData GroupVars HV_orientation PlotRate Pseed AnnoInData   
BaseMergeVars AnnoMergeVars AnnoOrganism AnnoAccession AnnoChrVar AnnoPosVar
AnnoOnly AnnoWhereStmt AnnoKeepVars MaxLength
AnnoLabel AnnoGeneSymbol AnnoDescription AnnoGeneId AnnoSnpId ArrayGroupCorr_OutData
OutPath ServerOutPath MacroPath KeepProcessInfo ArrayGroupCorr_Script;

*ProcessBody;

%macro ArrayGroupCorrelation;

*initialization and argument checking;
%let ProcessName = ArrayGroupCorrelation; 

%include "&MacroPath./UtilityMacros.sas" / nosource;

%GetOrigFileName(InData,OrigInData);

%CheckInData(_ExpFlag=1);
libname OutLib "&OutPath";

%if %length(&ArrayGroupCorr_OutData) %then %let ArrayGroupCorr_OutData2 = &ArrayGroupCorr_OutData.2;
%else %let ArrayGroupCorr_OutData2 = ;
%PrepOutDS(ArrayGroupCorr_OutData,ArrayGroupCorr_OutDataName,&OrigInData,_agc);
%PrepOutDS(ArrayGroupCorr_OutData2,ArrayGroupCorr_OutDataName2,&OrigInData,_ags);

%CheckAnno;
%if &exiterror %then %goto exit;
%if %length(&AnnoInData) and %length(&AnnoWhereStmt) %then %do;
	data OutLib.&AnnoData;
		set AnnoLib.&AnnoData;
		where &AnnoWhereStmt;
	run;
	libname AnnoLib "&OutPath";
%end;

%if %sysevalf(&PlotRate<=0 | &PlotRate>1) %then %do;
   %let exiterror = 1;
   %put ERROR: Plot rate is out of range.;
   %goto exit;
%end;
%if %sysevalf(&PlotRate < 1) %then %do;
	%if ^%length(&PSeed) %then %do;
		%let PSeed = 1234567;
	%end;
	%else %if %sysevalf(&PSeed <= 0) %then %do;
		%let exiterror = 1;
		%put ERROR: The PSeed is not positive integer.;
		%goto exit;
	%end;
	%else %do;
		%if %sysevalf(&PSeed-%sysevalf(&PSeed,integer) > 0) %then %do;
			%let PSeed = %sysevalf(&PSeed,integer);
			%put WARNING: The decimal portion of PSeed was truncated.;
		%end;
	%end; 
%end;

/*--drop unannotated data if necessary---*/
%if %length (&AnnoInData) and (&AnnoOnly=yes) %then %do;
	data InDataAll;
		set InLib.&InData;
	run;
	%OnlyAnno;
%end;
%else %do;
	%let indat = InLib.&InData;
%end;

data &ArrayGroupCorr_OutData;
	set &indat;
	%if %length(&WhereStmt) %then %do;
		where &WhereStmt;
	%end;
run;

%if &PlotRate < 1 %then %do;
   proc surveyselect data=&ArrayGroupCorr_OutData
      out=&ArrayGroupCorr_OutData seed=&Pseed rate=&PlotRate noprint;
   run;
%end;

%if %length (&AnnoInData) %then %do;
   %MergeAnno(&ArrayGroupCorr_OutData);
%end;

%NObsVars(&ArrayGroupCorr_OutData);
%if ^&nobs %then %do;
	%put ERROR: No observations in the plotting data set.;
	%if %length (&AnnoInData) and (&AnnoOnly=yes) %then %do;
		%put There may not be any matches with the annotation merge variables.;
	%end;
	%let exiterror = 1;
	%goto exit;
%end;

* determine groups;
data ed;
	set ExpInLib.&ExpInData;
run;

%if %length(&GroupVars) %then %do;
	proc sort data=ed;
		by &GroupVars;
	run;
	%let gvn = %scan(&GroupVars,-1," ");
	%let i = 1;
	%let gCon = ;
	%do %while(%length(%scan(&GroupVars,&i)));
	   %let gVar&i = %scan(&GroupVars,&i);
	   %if &i = 1 %then %let gCon = "Group = " || trim(left(&&gVar&i));
		%else %let gCon = &gCon ||","|| trim(left(&&gVar&i));
		%let i = %eval(&i+1);
	%end;
%end;
%else %let gID1 = Multivariate;

data ed;
	set ed end=_last;
	retain ng 0 nr 0 ngname "ng10000";
	nr = nr + 1;
	%if %length(&GroupVars) %then %do;
		by &GroupVars;
		if (first.&gvn) then do;
			ng = ng + 1;
			nr = 1;
			gID = "gID" || trim(left(ng));
			call symput(gID,&gCon);
	      ngname = "ng" || trim(left(ng));
		end;
		if (last.&gvn) then call symput(ngname,trim(left(nr)));
	%end;
	%else %do;
	   ng = 1;
	   ngname = "ng" || trim(left(ng));
		if (_last) then call symput(ngname,trim(left(nr)));
	%end;
	cname = "G" || trim(left(ng)) || "R" || trim(left(nr));
	call symput(cname,ColumnName);
	if (_last) then do;
		call symput("ng",trim(left(ng)));		
	end;
run;

%if %length(&GroupVars) %then %do;
	%do i = 1 %to &ng;
   	%*put gID&i = &&gID&i;
   	%*put i=&i, ngi=&&ng&i;
	
		%if &i = 1 %then %do;
			%let lgID = %length(&gID1);
		%end;
		%else %do;
			%if &lgID < %length(&&gID&i) %then %let lgID = %length(&&gID&i);
		%end;
	
	   proc corr data=&ArrayGroupCorr_OutData outp=outp&i noprint;
			var %do k = 1 %to &&ng&i; &&G&i.R&k %end;;
		run;
	
		data outp&i;
			set outp&i;
			length _Column1_ _Column2_ $32; *to avoid multiple lengths;
			array _Corr_{&&ng&i} %do k = 1 %to &&ng&i; &&G&i.R&k %end;;
			array _Var_{&&ng&i} $ (%do k = 1 %to %eval(&&ng&i-1); "&&G&i.R&k", %end; "&&G&i.R&k");
			retain _c_ 0;
			if _TYPE_ ne "CORR" then delete;
			else do;
				_c_+1;
				do j = 1 to dim(_Corr_);
					if j > _c_ then do;
						_Group_ = substr("&&gID&i",9);
						_Correlation_ = _Corr_(j);
						_Column1_ = _NAME_;
						_Column2_ = _Var_(j);
						output;
					end;
				end;
			end;
			keep _Group_ _Correlation_ _Column1_ _Column2_;
		run;
	%end;

	data &ArrayGroupCorr_OutData2;
		length _Group_ $&lgID;
		set %do i = 1 %to &ng; outp&i %end;;
	run;
%end;

%if &syserr %then %do;
   %put ERROR: Unable to complete calculations.;
   %goto exit;
%end;

* convert names to labels for JMP;
%do i = 1 %to &ng;
	%do k = 1 %to &&ng&i;
		%let mvar = &&G&i.R&k;
		 /* %put varname = &mvar; */
		%CheckLabel(&ArrayGroupCorr_OutData,mvar);
		/* %put varlabel = &mvar; */
		%let G&i.R&k = &mvar;
	%end;
%end;

* create JSL file;
%PrepOutFile(ArrayGroupCorr_Script,ArrayGroupCorr_ScriptName,&ProcessName Script for &ArrayGroupCorr_OutDataName..jsl);

* create JSL code;
%global _mainjslfile_ FullAPName;
%let _mainjslfile_ = &ArrayGroupCorr_Script;
%let FullAPName=%str(Correlation and Grouped Scatterplots);

data _null_;
   file "&ArrayGroupCorr_Script";
	put '//!';
	put "Names Default to Here(1);";
   %if %length(&AnnoInData) %then %do;
		put " ";
      %include "&MacroPath./WebSearchJSL.sas";
      %WebSearchJSL(Ds=,clientPath=,MakeWindow=0);
		put " ";
   %end;
   %if %length(&AnnoAccession) %then %do;
		put 'Here:ActionButtonDesc = {"Select Points in Graphs, then Click a Button"};';
      %WebButtons;
		%WebButtons_individual;
   %end;

	put "Here:DataList = {};";
	put "Here:TabButtonList = {};";
	put "Here:TabCodeList = {};";
	put "Here:TabAfterCodeList = {};";
	put "Here:TabOpenList = {};";
	put " ";
   put "Here:agccorr_TabCode=Function({},{default local},";
   put " If(IsEmpty(Here:agc_corrdata),";
	put "  Here:agc_corrdata=open(""&ClientOutPath.&ArrayGroupCorr_OutDataName..sas7bdat"",invisible);";
	%if %length(&AnnoInData) %then %do;
		%AnnoLabelJSL(Ds=&ArrayGroupCorr_OutData);
	%end;
	put ","; /*is empty else condition */
	put "   Current Data Table(Here:agc_corrdata);";
	put ");";
/*
   put "data=open("".\&ArrayGroupCorr_OutDataName..sas7bdat"");";
   put 'data<<Minimize Window;';
   put 'smatrix=New Window( "Correlation Scatterplots",';
*/
	%if %upcase(&HV_orientation) = HORIZONTALLY %then %do;
   	put '   HListBox( '; 
	%end;
	%else %do;
   	put '   VListBox( '; 
	%end;
   %do i = 1 %to &ng; 
      %if (&&ng&i > 0) %then %do;
		    put "      sm&i=Multivariate(Y( ";
          %do k = 1 %to %eval(&&ng&i - 1);
			   	%let mvar = &&G&i.R&k;
			  		put "         :Name(""&&mvar"") , " ;
          %end;
			 %let mvar = &&G&i.R&k;
          put "         :Name(""&&mvar"")),";
		    %if &i = &ng %then %let ending = ;
          %else %let ending = ,; 
          put "         Scatterplot Matrix(Density Ellipses(1), Show Correlations(1), Horizontal(1),";
          put '         Ellipse Color(3)), SendToReport(Dispatch({}, "Multivariate ", OutlineBox,';
          put '         Set Title("' @;
          put "&&gID&i" @;
          put '")))';
          put "   ) &ending";
      %end;
   %end;
	put '   )';
   put ');';
	put " ";
	put 'InsertInto(DataList,{{"Here:agc_corrdata"}});';
	put 'InsertInto(TabButtonList, "Grouped Correlation");';
	put 'InsertInto(TabCodeList,expr(Here:agccorr_TabCode()));';
	put 'InsertInto(TabAfterCodeList, " ");';
	put 'InsertInto(TabOpenList,1);';

	%if %length(&GroupVars) %then %do;
   	put "Here:agccorr_Tab2Code=Function({},{default local},";
   	put "   If(IsEmpty(Here:agc_corrdata2),";
		put "      Here:agc_corrdata2=open(""&ClientOutPath.&ArrayGroupCorr_OutDataName2..sas7bdat"",invisible);";
		put "   ,"; 
		put "      Current Data Table(Here:agc_corrdata2);";
		put "   );";
		put '   Column("_Column1_") << label(1);';
		put '   Column("_Column2_") << label(1);';
		put "   vListBox( ";
		put "      GSummary=Oneway(";
		put "         Y( :_Correlation_ ),";
		put "         X( :_Group_ ),";
		put "         With Best( 1 ),";
		put "         Quantiles( 1 ),";
		put '         Name( "Means/Anova" )(1),';
		put "         Box Plots( 1 ),";
		put "         Mean Diamonds( 1 ),";
		put "         Comparison Circles( 1 ),";
		put "         X Axis Proportional( 0 ),";
		put "         SendToReport(";
		put '            Dispatch( {}, "Quantiles", OutlineBox, {Close( 1 )} ),';
		put '            Dispatch( {}, "Oneway Anova", OutlineBox, {Close( 1 )} ),';
		put '            Dispatch( {}, "Means for Oneway Anova", OutlineBox, {Close( 1 )} )';
		put "         )";
		put "      )";
		put "   )";
  		put ');';
		put " ";
		put 'InsertInto(DataList,{{"Here:agc_corrdata2"}});';
		put 'InsertInto(TabButtonList, "Summary");';
		put 'InsertInto(TabCodeList,expr(Here:agccorr_Tab2Code()));';
		put 'InsertInto(TabAfterCodeList, " ");';
		put 'InsertInto(TabOpenList,1);';
	%end;
run;

%let inds = &ArrayGroupCorr_OutDataName..sas7bdat;
%if %length(&GroupVars) %then %do;
	%let inds = &inds,&ArrayGroupCorr_OutDataName2..sas7bdat;
%end;
	

%PackageInsertFile(jsl code from &ProcessName, &OutPath, &ArrayGroupCorr_ScriptName, "&inds", text/plain, yes);
%AddDatatoStudy(&ArrayGroupCorr_OutDataName,_type=tall);

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend ArrayGroupCorrelation;

* run the macro;
%ArrayGroupCorrelation;

