/*
JMP Genomics Analytical Process

Name: QTL Single Marker Analysis
Author: Shengchu Wang
Support: Kelci Miclaus

History:
	Luciano changed the grid of search to include markers.

Description:

	Interal Mapping (IM) is an extension of single marker analysis 
	and it can do analysis to any genomic location flanked by two 
	markers. IM uses complete marker linkage maps for genomic sacnning 
	(LR or LOD profile) of QTL. Composite Interval Mapping (CIM) is an 
	extension of IM analysis that uses a combination of IM with multiple 
	regression. CIM can improve the precision and efficiency of mapping 
	multiple QTL. 
*/

/*
-----------------------------------------------------------
Copyright (C) 2007 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData MarkerVars TraitVar FactorVars FactorClassVars ListMarkerVars ListTraitVars WhereStmt OutPath ServerOutPath MacroPath 
		AnnoInData ChromNum ChromLab MarkerLab MarkerPos MarkerPos_BP 
		GPInData GKNumber TestStep  MethodSelection MAlgorithm Perms Seed LODThresh Alpha 
		CMarkerNum TWinSize CMarkerSM CMarkerData EntryProb StayProb
		DisplayMarkers DisplayQTL MIMWinSize 
		MIMEntryCriterionMain MIMEntryLODThreshMain MIMStayLODThreshMain MIMEntryProbMain MIMStayProbMain MIMTestStepMain 
		MIMEntryCriterionEpi MIMEntryLODThreshEpi MIMStayLODThreshEpi MIMEntryProbEpi MIMStayProbEpi MIMTestStepEpi1D MIMTestStepEpi2D
		GDMetric MarkerFmt OldNumCode GenoDelim GenoMcode OutName CMData MappingResult_OutData MappingGraph_Script;

%global AnnoData tTraitNum tChromNum GPData VType ChLab_Type MkBP_Type NonGNTypeFlag;
%let tFactorNum = 0;
%let debugflag = 0; %let debugflagiml = 0;

*ProcessBody;

%macro SingleQTLModel;

%let ProcessName = SingleQTLModel;
%include "&MacroPath./UtilityMacros.sas" / nosource;
%include "&MacroPath./qtlUtilityMacros.sas" / nosource;
*ods html file="&OutPath./file.html";
*ods html close;

%GetOrigFileName(InData,OrigInData);

%PathName(&InData);
%if &exiterror %then %do;
	%put ERROR: Input data set (cross information) specification is invalid.;
	%let exiterror=1;
	%goto exit;
%end;
%let InData = &TmpName;
libname InLib "&TmpPath";
libname OutLib "&OutPath";

/* Annotation data set */
%if %length(&AnnoInData)=0 %then %do;
	%put ERROR: Annotation data set (map information) must be specified.;
	%let exiterror=1;
	%goto exit;
%end;
%PathName(&AnnoInData);
%if &exiterror %then %do;
	%put ERROR: Annotation data set (map information) specification is invalid.;
	%let exiterror=1;
	%goto exit;
%end;
%let AnnoData = &TmpName;
libname AnnoLib "&TmpPath";

/* Genotype data set */
%if %length(&GPInData)=0 %then %do;
	%put ERROR: GP data set (QTL Genotype information) must be specified.;
	%let exiterror=1;
	%goto exit;
%end;
%PathName(&GPInData);
%if &exiterror %then %do;
	%put ERROR: GP data set (QTL Genotype information) specification is invalid.;
	%let exiterror=1;
	%goto exit;
%end;
%let GPData = &TmpName;
libname GPLib "&TmpPath";

/* check name for OutName */
%let maxnamelength = 20;
%CheckDataName(_name=OutName,_default=&OrigInData);
%if %length(&OutName)>&maxnamelength %then %let OutName=%substr(&OutName,1,&maxnamelength);

/* Marker varaibles */
%if %length(&ListMarkerVars &MarkerVars)=0 %then %do;
	%put ERROR: Marker variables must be specified in either the Marker Variables or List-Style Specification of Marker Variables parameters.;
	%let exiterror=1;
	%goto exit;
%end;
%let MarkerVars = %upcase(&MarkerVars &ListMarkerVars);

/* Trait variables */
%if %length(&ListTraitVars &TraitVar)=0 %then %do;
	%put ERROR: One or more trait variables must be specified in either the Trait Variables or List-Style Specification of Trait Variables parameters.;
	%let exiterror=1;
	%goto exit;
%end;
%let TraitVar = &TraitVar &ListTraitVars;

%let oname = &OutName._cim;
%if %upcase(&MethodSelection) = %upcase(Interval Mapping) %then %do;
	%let oname = &OutName._im;
%end;
%else %if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
	*--- Begin: sbrlsi ---*;
	%if "&MAlgorithm" = "EM Algorithm" %then %do;
		%put ERROR: &MethodSelection of QTL can only be run with Haley-Knott Regression. Please, return to the Analysis tab and choose Haley-Knott Regression in the field QTL Mapping Model Algorithm.;
		
		%goto exit;
	%end;

	%let oname = &OutName._mim;
	%let Perms = 0;
	
	%let LODThresh = 0;
	*--- End: sbrlsi ---*;
%end;

%if %index(%UPCASE(&MAlgorithm),HALEY) %then %do;
	%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
		%PrepOutDS(MappingResult_OutData,MappingResult_OutDataName,&oname,_pr_hk);
	%end;
	%else %do;
		%PrepOutDS(MappingResult_OutData,MappingResult_OutDataName,&oname,_hk);
	%end;
%end;
%else %do;
	%PrepOutDS(MappingResult_OutData,MappingResult_OutDataName,&oname,_out);
	%let Perms = 0;
%end;
%*put MappingResult_OutData = &MappingResult_OutData; *--- sbrlsi checking code ---*;

%PrepOutFile(MappingGraph_Script,MappingGraph_ScriptName,&ProcessName Script for &oname..jsl);
%*put MappingGraph_ScriptName = &MappingGraph_ScriptName; *--- sbrlsi checking code ---*;

%global GPFile MapFile CrossFile TtFile FtFile MkFile;

%let GPFile = GPLib.&GPData;

/*** Get GP Data Set's distinct number ***/
data _tmp_;
	*set &GPFile (firstobs=2 obs=2 keep= col1 col2);
	set &GPFile (firstobs=3 obs=3 keep= col1 col2 col3); /*sbrlsi, outcross*/
run;

%let c1=; %let c2=; %let c3=; /*sbrlsi, outcross*/
data _null_;
	set _tmp_;
	if _n_ = 1 then do;
		call symputx('c1',col1);
		call symputx('c2',col2);
		call symputx('c3',col3);/*sbrlsi, outcross*/ 		
	end;	
run;

/*begin: sbrlsi, outcross*/
%if &c1 = &c2 and &c2 = &c3 %then %do;
	%let GPTypeNum = four;
	%let GPNum = 4;
%end;
%else %if &c1 = &c2 %then %do;
	%let GPTypeNum = three;
	%let GPNum = 3;
%end;
%else %do;
	%let GPTypeNum = two;
	%let GPNum = 2;
%end;

/*
%let GKNumber=;
%if %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(BC1n)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(BC2n)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(RI0)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(RI1)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(RI2)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(IRILself)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(IRILsib))
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(T(B1)SFn)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(T(B2)SFn)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(T(B1)RFn)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(T(B2)RFn)) 
	%then %let GKNumber = two;
%else %if %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(SFn)) 
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(RFn))  
	or %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(T(SFm)SFn)) 
	%then %let GKNumber = three; 
%else %if %nrbquote(%upcase(&CrossType)) = %nrbquote(%upcase(FSoutbred)) 
	%then %let GKNumber = four; 
%else %do;
	%put ERROR: Crosstype &CrossType is not supported.;
	%let exiterror=1;
	%goto exit;
%end;
*/

%if &GKNumber ^= &GPTypeNum %then %do;
	%put ERROR: You selected &GKNumber for Number of Distinct Genotypes when the value is &GPTypeNum in the Genotype Probability data set. ;
	%let exiterror=1;
	%goto exit;
%end;
/*end: sbrlsi, outcross*/

/*%if &MethodSelection = Interval Mapping %then %do;*/
%if %upcase(&MethodSelection) = %upcase(Interval Mapping) | %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do; /*--- sbrlsi ---*/
	%let CMarkerNum = 0;
%end;

/*** 1. Deal with map data set - AnnoLib.&AnnoData ***/
data nmapfile;
 	set AnnoLib.&AnnoData;
run;
%let MapFile = nmapfile;

/* Check ChromNum+MarkerPos is in ascending order or not */
%let _tmp_ = 1;
data _null_;
	set &MapFile;
	if (&ChromNum*10000+&MarkerPos/100)<(lag(&ChromNum)*10000+lag(&MarkerPos)/100) then do;
		call symput('_tmp_',"-1");
	end;
run;
%if &_tmp_ = -1 %then %do;
	%put ERROR: In the annotation data set (map information), &ChromNum + &MarkerPos must be in ascending order.;
	%let exiterror=1;
	%goto exit;
%end;

%let ListMapVars = &ChromNum &ChromLab &MarkerLab &MarkerPos &MarkerPos_BP;
%if &debugflag = 1 %then %do;
	proc print data=&MapFile; var &ListMapVars; Title "Map File"; run;
%end;
%if &ChromNum = &ChromLab %then %let ChromLab =;

/* Test the type of chromNum etc. */
%VarAttributes(_Ds=&MapFile,_Var=&ChromNum,_Len=,_Type=VType);
%if &VType ^= N %then %do;
	%put ERROR: Variable (&ChromNum) is not a numeric variable.;
	%let exiterror=1;
	%goto exit;
%end;

%if %length(&ChromLab)^=0 %then %do;
	%VarAttributes(_Ds=&MapFile,_Var=&ChromLab,_Len=,_Type=ChLab_Type);
%end;

%VarAttributes(_Ds=&MapFile,_Var=&MarkerLab,_Len=,_Type=VType);
%if &VType ^= C %then %do;
	%put ERROR: Variable (&MarkerLab) is not a character variable.;
	%let exiterror=1;
	%goto exit;
%end;

%VarAttributes(_Ds=&MapFile,_Var=&MarkerPos,_Len=,_Type=VType);
%if &VType ^= N %then %do;
	%put ERROR: Variable (&MarkerPos) is not a numeric variable.;
	%let exiterror=1;
	%goto exit;
%end;

%if %length(&MarkerPos_BP)^=0 %then %do;
	%VarAttributes(_Ds=&MapFile,_Var=&MarkerPos_BP,_Len=,_Type=MkBP_Type);
%end;

/*** 2. Deal with cross data set - InLib.&InData ***/
/* Apply WhereStmt clause */
data nCrossFile;
	set InLib.&InData; 
	%if %length(&wherestmt) %then where &wherestmt;;
run;
%let CrossFile = nCrossFile;

/* Produce TtData data set */
data TtData;
	set &CrossFile (keep = &TraitVar);
run;

%if %length(&ListTraitVars)^=0 %then %do;
/* Re-construct TraitVar */
	proc contents data=TtData out=_t0_(keep=name varnum) noprint; run;
    	proc sort data=_t0_; by varnum; run;
	%let _t1_ = _t0_;
	%getDataVarList(_t1_,name);
	%let TraitVar = &dataVarList;
	%put TraitVars = &TraitVar;
%end;

/* Test if the trait variable is numeric or not */
%let nTraitVars = 0; *--- sbrlsi ---*;
%let VType =;
%let i = 1;
%let var = %scan(&TraitVar,&i);
%do %while(%length(&var));
	%VarAttributes(_Ds=TtData,_Var=&var,_Len=,_Type=VType);
	%if &VType ^= N %then %do;
		%put ERROR: Trait variable (&var) is not a numeric variable.;
		%let exiterror=1;
		%goto exit;
	%end;
	%let i = %eval(&i+1);
	%let var = %scan(&TraitVar,&i);
	%let nTraitVars = %sysevalf(&nTraitVars + 1); *--- sbrlsi ---*;
%end;

*--- Begin: sbrlsi --- *;
%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;

	%if &nTraitVars > 1 %then %do;
		%put ERROR: &MethodSelection of QTL can only analyse one trait at the time. Please, return to the General tab and choose only one trait in the field Marker Variables, and leave the field List-Style Specification of Trait Variables blank.;
		
		%goto exit;
	
	%end;
%end;
*--- End: sbrlsi ---*;

/* Get total trait number */
%getVarNumber(TraitVar);
%let tTraitNum = &_vnum;
%*put tTraitNum = &tTraitNum;
/*************************/

%let TtFile = TtData;
data _null_ ;  
  if 0 then set &TtFile nobs=count;
  call symput('obsnum',left(put(count,8.))) ; 
  stop ;                        
run;
%if &debugflag = 1 %then %do;
	proc print data=&TtFile; Title "Trait File"; run;
%end;

%if &obsnum < 5 %then %do;
	%put ERROR: Input data set has too few observations left. ;
	%let exiterror=1;
	%goto exit;
%end;

/* Produce FtData data set that contains factor continuous variables*/
%let FtFile = FtData;
%let ftflag = 0;
%if %length(&FactorVars) %then %do;
	%let ftflag = 1;
	%setFactorVarData();
	%if &debugflag = 1 %then %do;
		proc print data=&FtFile noobs; Title "FtData-2 After change char. values into num. values";run;
	%end;
	
	*data OutLib.Ftfile0; *set &FtFile; *run;
%end;

%global FactorVarsNew;
%let FactorVarsNew= &FactorVars;
%if %length(&FactorClassVars) %then %do;

	/* Produce d data set that contains factor categorical variables*/
	* set up for ClassVars.sas;
	* The data set d must exist and the macro variables ClassVars, ContinuousVars,;
	* PredictorVars, and ClassPredVars must be defined prior to calling this code.;
	%global ClassVars ContinuousVars PredictorVars ClassPredVars ListClassColon;
	%let ClassVars=; 
	%let ContinuousVars=;
	%let PredictorVars=;
	%let ClassPredVars= &FactorClassVars;
	%let ListClassColon= 0;

	data d;
		set &CrossFile (keep = &ClassPredVars);
	run;

	%let NeedClassIndicators = Yes;
	%let RecoverClassPredVars = No;

	* save original class variables for later recovery;
	%if (&RecoverClassPredVars = Yes) %then %do;
		data orig_d_class;
			set d;
			keep &ClassPredVars;
		run;
	%end;

	* levelize class variables, create macro lists and indicator variables;
	* replace ClassPredVars accordingly;
	%put Beginning class levelization;
	%include "&MacroPath./ClassVars.sas";
	%if &syserr %then %goto exit;

	%put Ending class levelization:;
	%put nixl = &nixl;
	%put nipl = &nipl;
	%put nix = &nix;
	%put nip = &nip;

	*data OutLib.opn; *set opn; *run;
	*data OutLib.d; *set d; *run;

	**create list of class dummy variables, excluding dummy variable for missing values**;
	data d4;
		*set Outlib.opn;
		set opn;
		length orig_name vlevel $ 64;
		if name="Intercept" then delete;
		rename name = new_name;
		orig_name = scan(parameter,1);
		vlevel = scan(parameter,2);
		if missing(vlevel) or trim(left(vlevel)) = "_miss_" then delete; 
		drop parameter;
	run;

	proc sort data=d4;
		by orig_name vlevel;
	run;

	data _null_;
		set d4 end=_eof_;
		by orig_name vlevel;
		if _n_ = 1 then countv = 0;
		if first.orig_name then do;
			countl = 0;
			countv = countv + 1;
		end;		
		countl = countl + 1;
		call symputx("newcovar"||trim(left(countv))||"_"||trim(left(countl)),trim(left(new_name)));
		if last.orig_name then do;
			call symputx("norgcovar",trim(left(countv)));
			call symputx("orgcovar"||trim(left(countv)),trim(left(orig_name)));
			call symputx("nnewcovar"||trim(left(countv)),trim(left(countl)));
		end;
		retain countl countv;
	run;
	
	%let FactorDummyVars= ;
	%put norgcovar= &norgcovar;
	%do i=1 %to &norgcovar;
		%put &&orgcovar&i;
		%put &&nnewcovar&i;
		%do j=1 %to &&nnewcovar&i;
			%put &&newcovar&i._&j;
			%if %sysevalf(&j < &&nnewcovar&i) %then %do;
				%let FactorDummyVars= &FactorDummyVars &&newcovar&i._&j;
			%end;
		%end;
	%end;

	*exclude the last level of dummy variables*;
	data d;
		set d;
		keep &FactorDummyVars;
		;
	run;

	%if %length(&FactorVars) %then %do;
		**merging continuous and categorical dummy factor variables**;
		data &FtFile;
			merge &FtFile d;
		run;
	%end;
	%else %do;
		%let ftflag = 1;
		proc datasets nolist;
			change d= &FtFile;
		quit;	
	%end;
	
	*update FactorVarsNew*;
	%let FactorVarsNew= &FactorVarsNew &FactorDummyVars;
	
	*data OutLib.Ftfile1; *set &FtFile; *run;

%end;

%*goto exit;

/* Produce MkData data set */
%let MkFile = gendat;
data &MkFile;
	set &CrossFile (keep = &MarkerVars);
run;

data _nonmark; 
	set &CrossFile(drop = &MarkerVars);
run;

%if %length(&ListMarkerVars)^=0 %then %do;
/* Re-construct MarkerVars */
	proc contents data=&MkFile out=_t0_ (keep=name varnum) noprint; run;
	proc sort data=_t0_; by varnum; run;
	%let _t1_ = _t0_;
	%getDataVarList(_t1_,name);
	%let MarkerVars = &dataVarList;
	%put MarkerVars = &MarkerVars;
%end;
/*************************/
%if &debugflag = 1 %then %do;
	proc print data=&MkFile; Title "MkData1";run;
%end;

/****************************************************/
/* data set to use in iml to convert genotypes */
/* kemicl - changed to use new macro and new coding scheme */
%global AlleleVars;
%let AlleleVars = &MarkerVars;
data mnames; set &MkFile;
where 0;
run;

/*begin: sbrlsi, outcross*/
%if &GKNumber ^= four %then %do; 
	/* marker recoding macro to convert to numeric for IML code */
	%RecodeGenos;
	
	/* Add a check for numeric format for negative values and throw error if found */
	%OldGenoCheck;

	%if &exiterror %then %goto exit;
	%if &syserr>4 %then %do;
		%put ERROR: The genotype data could not be converted to numeric format for recombination computation.;
		%put Please check the Marker Format coding of your data for possible errors.;
	 	%let exiterror=1;
		%goto exit;
	%end;

	%nobsvars(gendat);
	%if %index(%Upcase(&MarkerFmt),NUMERIC) = 0  %then %do;
		%if %index(%Upcase(&MarkerFmt), ALLELES) %then %do;
			data _null_;
				set nmapfile end = last;
				call symputx("AMName"||trim(left(_n_)), &MarkerLab);
				if last then call symputx("Totnames",_n_);
			run;

			/* Kelci next version change to stop building large accumulated macros of variables */
			%let nMarkerVars = ;
			%do i=1 %to &nvars;
				%let nMarkerVars = &nMarkerVars &&AMName&i;
			%end;	
		%end;
		%else %do;
			%let _tid_ = %sysfunc(open(mnames));
			%let _nVars_ = %sysfunc(attrn(&_tid_,NVARS));
		%end;

		proc datasets library=work nolist;
			modify gendat;
			%if %index(%Upcase(&MarkerFmt), ALLELES) %then %do;
				%if &Totnames >= &nvars %then %do;
					rename %do i=1 %to &nvars; _geno_&i = &&AMName&i %end;;
				%end;
				%else %do;
					%put ERROR: The number of markers formed from the alleles in the Input Data Set do not match with the the number of marker names in the Annotation Data Set.;
					%let exiterror =1;
					%goto exit;
				%end;
			%end;
			%else %do;
				rename %do i=1 %to &_nVars_; _geno_&i = %sysfunc(varname(&_tid_,&i)) %end;;
			%end;
		quit;
	
		/* remove any extra markers that weren't renamed */
		data gendat; set gendat; drop _geno_:; run;
	%end;
	
	/* Kemicl - added to fix CIM with nonnumeric markers */
	data dataNum; 
		set &MkFile; set _nonmark;
	run;
	%let CrossFile = dataNum;	
%end;
%else %do;

	%if %index(%Upcase(&MarkerFmt), OUTCROSS) = 0 %then %do;
		%put ERROR: The Format of Marker Varibles for this Input data set must be Outcross Nondelimited Genotypes.;
 		%let exiterror=1;
		%goto exit;
	%end;
	
	%if %index(%upcase(&MAlgorithm), HALEY) = 0 %then %do;
		%put ERROR: The QTL Mapping Model Algorithm for this Input data set must be Haley-Knott Regression.;
	 	%let exiterror=1;
		%goto exit;
	%end;	

%end;
/*end: sbrlsi, outcross*/

/*************************/
%if &debugflag = 1 %then %do;
	data outlib.&MkFile; set &MkFile; run;
%end;
/****************************************************/

/**Check if number of markers in the GPData and Input Data (after manipulations) are the same**/
%NobsVars(&MkFile);
%let inmrks= &nvars;

data gpline3;
	set &GPFile(firstobs=3 obs=3);
	array a{*} _all_;
	count=0;
	do _i=1 to dim(a);
		if(int(a[_i])=a[_i] and a[_i]>0) then count= count + 1;
	end;
	put count;
	call symputx("gpmrks", trim(left(count)));
run;
%let gpmrks= %sysevalf(&gpmrks/(&GPNum-1));

%if %sysevalf(&inmrks ^= &gpmrks) %then %do;
	%put WARNING: Number of markers specified in the General tab (&inmrks markers) does not match markers in the Genotype Probability Data Set (&gpmrks markers).;
	%*goto exit;
%end;

/*** 3. Combine map (MapFile) and cross (&CrossFile = &TtFile,&FtFile,&MkFile) information - InLib.&InData ***/
 %let allmatch=1;
/* %put MarkerVars = &MarkerVars; */
/* Delete the value of MarkerVars that is not in MapFile.MarkerLab and orderly. */
/* Create nMarkerVars : contents of MarkerLab */
%if %index(%Upcase(&MarkerFmt), ALLELES)=0 %then %do;
	%getDataVarList(MapFile,&MarkerLab);
	%let nMarkerVars = &dataVarList;
%end;
/*%put nMarkerVars = &nMarkerVars;
 newMarkerVars is the nMarkerVars after delete */
%if %index(%Upcase(&MarkerFmt), ALLELES) %then %do;
	%let newMarkerVars = &nMarkerVars;
	%let MarkerVars = &nMarkerVars;
%end;
%else %do;
	%let newMarkerVars=;
	%let i = 1;
	%let vari = %scan(&nMarkerVars,&i);
	%let varj = %scan(&MarkerVars,&i);
	%do %while(%length(&varj));
		/*kemicl checks
		%put This is Vari = &Vari;
		%put this is Varj = &Varj;
		*/
	/****************************************************/
		%if %upcase(&vari) ne %upcase(&varj) %then %do;
			%let allmatch=0;
			%put For Variable &i, Marker Variable is &Varj and Marker Label in Annotation is &Vari.;
			%goto enddo;
		%end;
		%else %do;
			%let newMarkerVars = &newMarkerVars &Vari;
		%end;
	/***************************************************/
		%let i = %eval(&i+1);
		%let vari = %scan(&nMarkerVars,&i);
		%let varj = %scan(&MarkerVars,&i);
	%end;
	%if %length(&vari) %then %do;
		data &Mapfile;
			set &Mapfile;
			if _n_ < &i;
		run;
	%end;
%enddo:;
%end;
%put allmatch = &allmatch;

%if &allmatch=0 %then %do;  
	%put ERROR: The markers in the input data set do not match the annotation data set according to the variable selected as Marker Label Variable.;
	%put Please check that the Marker Label exactly matches the input data columns or run the Subset and Reorder Genetic Data AP to order these data sets correctly.;

		* create JSL file;
		%global SRButtonsScript;
		%PrepOutFile(SRButtonsScript,SRButtonsScriptName,launchAP_buttons_SR.jsl);
		filename jslbtns "&SRButtonsScript";

		options noquotelenmax;
		data _null_;
			file jslbtns lrecl=32767;
			put "//!";
			put "Names Default to Here(1);";
			put "wait(0.5);";
			put "if(host is(Windows),";
			put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, windows) ),";
			put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, POSIX) ));";
			put "	";
			put "srbuttonswindow=New Window(""Launch AP to Reorder Markers"",";
			put "		VListBox(";
			put "			TextBox(""Click on the button below to launch the Subset and Reorder Genetic Data process to reorder marker columns properly.""),";
			put "			TextBox(""Use the output data sets from that process as the new input and annotation data sets to .""),";

			put "			TextBox("" ""),";
			put "			ButtonBox(""Subset and Reorder Genetic Data"",";
			put "				if(host is(Windows),";
			put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/SubsetandReorderGeneticData.jsl"", absolute, windows)),";
			put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/SubsetandReorderGeneticData.jsl"", absolute, POSIX) ));";
			put "	";

			/* Set input data set and Output Folder */
			put "				jg:setDefaultParmValues(2, ""SubsetandReorderGeneticData"");";				
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""InData""),""%sysfunc(pathname(Inlib))\&InData..sas7bdat"");";
			put "				jg:SubsetandReorderGeneticData_updateDeps;";
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""ListMarkerVars""),""&MarkerVars"");";	
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""MarkerFmt""),""&MarkerFmt"");";
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""GenoDelim""),""&GenoDelim"");";			
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""MinPercent""),""0"");";		
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""OutPath""),""&ClientOutPath"");";					
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""AnnoInData""),""%sysfunc(pathname(AnnoLib))\&AnnoData..sas7bdat"");";					
			put "				jg:SubsetandReorderGeneticData_updateDeps;";				
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""AnnoMName""),""&MarkerLab"");";	
			put "			)";;

			put "		)";
			put ");";
		run;

	%PackageInsertFile(jsl code from &ProcessName, &OutPath, &SRButtonsScriptName, , text/plain, yes);		
	%let exiterror=1;
	%goto exit;
%end;

/* check marker var to make sure that each chromosome at least have two markers */
proc summary data = &MapFile; by &ChromNum; Output out = _tmp_; run;
%let _flag_ = 1;
data null;
	set _tmp_;
	if _FREQ_ = 1 then do;
		call symput('_flag_',"-1");
		call symputx('_chrom',&ChromNum);
	end;
run;
%if &_flag_ = -1 %then %do;
	%put ERROR: In the annotation data set (map information), each chromosome (&_chrom) must have more than one marker.;
	%let exiterror=1;
	%goto exit;
%end;

/*kemicl added to find unique values of chromosome numbers, and make into macros for building matrices */
proc sort data=&MapFile out=Numchrom nodupkey;
by &ChromNum;
run;

data _null_;
	set Numchrom end=last;
	call symputx("ChrVars"||strip(_n_),&ChromNum);  /*Use ChrVars to loop through chr. value so no longer needs to be consecutive */
	%if %length(&ChromLab) %then %do;
		call symputx("ChrLabel"||strip(_n_), &ChromLab);   /* kemicl added - use for JSL output */
	%end;
	if last then call symputx("tChromNum",_n_);  /* Total number of Chromosomes - use to build dimensions of matrices */
run;
*******************************************************************************;
%if &debugflag = 1 %then %do;
	proc print data=&MapFile; Title "Annotation File"; run;
	proc print data=&CrossFile; Title "Input File"; run;
%end;

%let TraitVar_nocofactor= ; /*sbrlsi*/
%if %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) %then %do; *CIM;

	%if &GKNumber = four %then %do;
		
		proc transpose data=GPlib.&GPData out=gptrans;
		run;

		data gptrans;
			set gptrans(rename=(col1= chrom col2= cm col3= MarkerIndicator));
			if MarkerIndicator > 0 then MarkerIndicator = 1;
			where MarkerIndicator > 0;
		run;

		proc transpose data=gptrans out=gpback prefix=col;
		run;

		data gpback(drop=_NAME_);
			set gpback;
			if _n_ > 3 then output;
		run;
		
		%let outMarkerVars=;
		%let renamestm=;
		%let i=1;
		%let mkvar = %scan(&newMarkerVars, &i);
		%do %while(%length(&mkvar));

			%let k= %sysevalf(&i * 3 - 2);
			%let renamestm= %str(&renamestm col&k= &mkvar._p1);
			%let k= %sysevalf(&i * 3 - 1);
			%let renamestm= %str(&renamestm col&k= &mkvar._p2);
			%let k= %sysevalf(&i * 3 - 0);
			%let renamestm= %str(&renamestm col&k= &mkvar._p3);
			
			%let outMarkerVars= &outMarkerVars &mkvar._p1 &mkvar._p2 &mkvar._p3;

			%let i= %sysevalf(&i + 1);
			%let mkvar = %scan(&newMarkerVars, &i);
		%end;

		proc datasets nolist;
			modify gpback;
			rename &renamestm;
		quit;

		%let MkFile = gpback;		
		data dataNum; 
			set &MkFile; set _nonmark;
		run;
		%let CrossFile = dataNum;
		
		%let newMarkerVarsOrig= &newMarkerVars;
		%let newMarkerVars= &outMarkerVars;
		%*put newMarkerVars= &newMarkerVars;
		
		data outMapFile;
			set &MapFile
				&MapFile
				&MapFile;
		run;

		proc sort data=outMapFile;
			by &ChromNum &MarkerPos;
		run;

		data outMapFile(drop= mklab_ p rename=(newmklab_= &MarkerLab));
			set outMapFile(rename=(&MarkerLab= mklab_));
			by &ChromNum &MarkerPos;
			if first.&MarkerPos then p=1;
			else p = p + 1;
			retain p;
			newmklab_= trim(left(mklab_))||trim(left('_p'))||trim(left(p));
		run;		
				
	%end;

	%*goto exitcheck;

	/*Kemicl - changed to Outlib to avoid write issues  S0602963*/
	%let CMData = &OutName._cml;  
	%let CMFile = Outlib.&CMData; *default control file name;
		
	%PrepOutDS(CMData,CMDataName,&OutName._cml);

	%*put CMFile= &CMFile; /*sbrlsi*/
	%*put CMDataName= &CMDataName; /*sbrlsi*/
	%*put CrossFile= &CrossFile; /*sbrlsi*/
	
	%if &CMarkerSM = Using SAS Data Set %then %do; *open data set;
		%PathName(&CMarkerData);
		%if &exiterror %then %do;
			%put ERROR: CIM control SAS data set specification is invalid.;
			%let exiterror=1;
			%goto exit;
		%end;
		%let CMDataName = &TmpName;
		libname CMLib "&TmpPath";
		%let CMData = CMLib.&CMDataName;
		
		%let CMDataOut= &CMData; /*sbrlsi, outcross*/
		
		%*put CMDataName= &CMDataName;
	%end;
	%else %do; *create and open data set;
	
		proc datasets nolist lib=outlib;
			delete &CMDataName;
		quit;
		
		%if &CMarkerSM = Simple Regression %then %do;
			%CMarkerSRegression(newMarkerVars);
		%end; /*Simple Regression*/
		%else %do;
			/* kemicl S0766766 - add check if too much missing data for PROC REG for CIM */ 
			data cimmiss; set &CrossFile;
				misscheck = nmiss(of &newMarkerVars);
				keep misscheck;
				if misscheck=0;
			run;
		
			%nobsvars(cimmiss);
			%if &nobs = 0 %then %do;
				%put ERROR: There is too much missing data for Composite Interval Mapping to be performed.;
				%put ERROR: Every row of the data contains at least one genotype column with a missing observation; 
				%put ERROR: Regression to select control markers cannot be performed.;
				%let exiterror =1;
				%goto exit;
			%end;	

			%if &CMarkerSM = Forward Regression %then %do;
				%if &EntryProb < 0 %then %do;
					%let EntryProb = 0.5;
				%end;
				%CMarkerFSearch(newMarkerVars);
			%end; /*Forward Regression*/
			%else %if &CMarkerSM = Backward Regression %then %do;
				%CMarkerBSearch(newMarkerVars);
			%end; /*Backward Regression*/
			%else %if &CMarkerSM = Stepwise Regression %then %do;
				%if &EntryProb < 0 %then %do;
					%let EntryProb = 0.5;
				%end;
				%if &StayProb < 0 %then %do;
					%let StayProb = 0.5;
				%end;			
				%CMarkerStepSearch(newMarkerVars);
			%end; /*Stepwise Regression*/
		%end;
		
		/*begin: sbrlsi*/		
		%let nTraitVar_nocofactor= 0;
		%let delimiter= ;
		%let u = 1;
		%let varu = %scan(&TraitVar,&u);
		%do %while(%length(&varu));
		
			data cmfile_tmp;
				set &CMFile;
				if Trait = &u then output;
			run;
		
			%NObsVars(cmfile_tmp);
			%if &nobs < 1 %then %do;
				%let TraitVar_nocofactor = &TraitVar_nocofactor.&delimiter &varu;
				%let delimiter =,;
				%let nTraitVar_nocofactor= %sysevalf(&nTraitVar_nocofactor + 1);
			%end;
			
			%let u= %sysevalf(&u + 1);
			%let varu = %scan(&TraitVar,&u);
		%end;
		%put TraitVar_nocofactor= &TraitVar_nocofactor;
		/*end: sbrlsi*/		
		
		%let CMDataOut= &CMFile; /*sbrlsi, outcross*/
		%*put CMDataOut= &CMDataout;

	%end; /*create and open data set*/
	%if &debugflag = 1 %then %do;
		proc print data = &CMData; title "&CMData"; run;
	%end;
	
	/*begin: sbrlsi, outcross*/
	%if &GKNumber = four %then %do;
	
		proc datasets nolist;
			delete CMDataOutcross;
		quit;
		
		%if &CMarkerSM ^= Using SAS Data Set %then %do;
	
			data CMDataOutcross;
				set &CMDataOut;
				by trait;
				if first.trait then 
					order =first.trait;
				else
					order = order + 1;
				retain order;
				markername= substr(trim(left(marker)), 1, length(trim(left(marker)))-3);
			run;	
	
			proc sort data=CMDataOutcross;
				by trait markername;
				run;
	
			data CMDataOutcross(rename=(markername=Marker));
				set CMDataOutcross(drop=Marker);
				by trait markername;
				if first.markername then output;
			run;
	
			proc sort data=CMDataOutcross out=CMDataOutcross(drop=order);
				by trait order;
			run;
			
			data outlib.&CMDataName;
				set CMDataOutcross;
			run;
	
			data CMDataOutcross(keep= Trait Marker);
				set CMDataOutcross(rename=(Marker=MarkerTmp));
				do i=1 to 3;
					Marker= compress(MarkerTmp||'_p'||i, '', 's');
					output;
				end;
			run;	
		%end;
		%else %do;
			
			data CMDataOutcross(keep= Trait Marker);
				set &CMDataOut(rename=(Marker=MarkerTmp));
				do i=1 to 3;
					Marker= compress(MarkerTmp||'_p'||i, '', 's');
					output;
				end;								
			run;
			
		%end;
			
		%let CMarkerNum = %sysevalf(&CMarkerNum * 3); 
	%end;	
	/*end: sbrlsi, outcross*/

%end; /*Composite Interval Mapping*/
*******************************************************************************;

%*goto exitcheck;

%let nx = %eval(&CMarkerNum+1);

/*sbrlsi
 Note: In the backward regression, why not use proc reg with model option selection= backward.
 I see that backward regression is not available in the QTL mapping AP.
 Note: I notice that IML function CheckVarSingular is not called anywhere.
sbrlsi*/

/*---begin: sbrlsi. Create grid of genomic positions ---*/
/* Assumptions: Markers in the Input Data Set = Markers in the Annotation SAS Data Set = Markers in the Genotype Probability Data Set.*/
%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
	%*let TestStep = &MIMTestStepMain;
	%let TestStep = 1;
%end;

/*
%if %length(&MIMTestStepEpi1D) %then %do;
	%if %sysevalf(&MIMTestStepEpi1D < &MIMTestStepMain) %then %do;
		%put ERROR: QTL Test Step in cM (1-D Genomewide Scan) from the MIM Options Tab cannot be smaller than Main QTL Test Step in cM from the Analysis Tab. Please, go to the specified Tab and change this option.;
		%let exiterror = 1;
		%goto exit;
	%end;
%end;

%if %length(&MIMTestStepEpi2D) %then %do;
	%if %sysevalf(&MIMTestStepEpi2D < &MIMTestStepMain) %then %do;
		%put ERROR: QTL Test Step in cM (2-D Genomewide Scan) from the MIM Options Tab cannot be smaller than Main QTL Test Step in cM from the Analysis Tab. Please, go to the specified Tab and change this option.;
		%let exiterror = 1;
		%goto exit;
	%end;
%end;
*/

data work.GridProb;
	set GPlib.&GPData (obs=3);
run;	

proc transpose data=work.GridProb out=work.GridProb;
run;

data work.GridProb(drop= _name_);
	set work.GridProb(rename=(col1=ChNum col2=Pos col3= MarkerIndicator));
	by ChNum;
	if MarkerIndicator > 0 then MarkerIndicator = 1;
	if first.ChNum then do;
		id= 0;
	end;
	id + 1;
run;

%if &GKNumber= four %then %let den= 3; /*sbrlsi, outcross*/
%else %if &GKNumber= three %then %let den= 2;
%else %let den= 1;

data work.GridProb(drop = id);
	set work.GridProb;
	by ChNum;
	where floor(id/&den) = ceil(id/&den);
	CMID = id/&den;
run;

data work.GridProb;
	retain ChNum Pos CMID MarkerIndicator;
	set work.GridProb;
run;

data work.TestStepGridTmp;
	set work.GridProb;
	by ChNum;
	retain _mkpos_;
	retain _Lmk_;
	retain _LmkCum_;
	if _n_ = 1 then _LmkCum_=0;
	if first.ChNum then _Lmk_=0;
	if MarkerIndicator = 1 then do;
		_mkpos_ = Pos;
		_Lmk_ = _Lmk_ + MarkerIndicator;
		_LmkCum_ = _LmKCum_ + MarkerIndicator;
		output;
	end;
	else do;
		if Pos >= _mkpos_ + &TestStep then do;
			_mkpos_ = Pos;
			_Lmk_ = _Lmk_;
			_LmkCum_ = _LmKCum_ + MarkerIndicator;
			output;
		end;
	end;
run; 

/*data OutLib.GridProb; set GridProb; run;*/
/*data OutLib.TestStepGridTmp; set TestStepGridTmp; run;*/

proc means data=work.TestStepGridTmp noprint;
	var MarkerIndicator;
	output out=TotalMkGrid sum=nMkGrid;
run;

data _null_;
	set TotalMkGrid;
	call symputx('nMkGrid',nMkGrid);
run;

/* Assumptions: If number of markers in the Annotation SAS Data Set > number of markers in the Genotype Probability Data Set, 
then trim the Annotation SAS Data Set to match number of markers in the Genotype Probability Data Set. */
data MapFileTmp(keep= MarkerName ChNum MrkPos _Lmk_ %if %length(&ChromLab)^=0 %then ChLab;);
	set &MapFile(rename=(&MarkerLab= MarkerName &ChromNum = ChNum &MarkerPos = MrkPos %if %length(&ChromLab)^=0 %then &ChromLab=ChLab;));
	by ChNum;
	retain _Lmk_;
	if first.ChNum then _Lmk_=1;
	else _Lmk_ = _Lmk_ + 1;
	if _n_ <= &nMkGrid then output;
run;
%*put MapFile = &MapFile;

data work.TestStepGrid;
	merge MapFileTmp TestStepGridTmp ;
	by ChNum _Lmk_;
run;

/* Assumptions: If number of markers in the Genotype Probability Data Set > number of selected markers, 
then trim Genotype Probability Data Set to match number of selected markers. */
data work.TestStepGrid;
	set work.TestStepGrid;
	if MarkerName = ' ' then delete;
run;

/*data OutLib.qtl_TestStepGrid; set TestStepGrid; run;*/

proc datasets nolist;
	delete TestStepGridTmp 
		MapFileTmp 
		TotalMkGrid;
quit;
/*---end: sbrlsi---*/

%*goto exitcheck;

/*************************************************************************************************/
proc iml symsize=10000;
	/* Check if x is inside of y or not */
	start CheckMkLab(x,y); 
		flag = 0;
		do i=1 to ncol(y) by 1;
			if UPCASE(x) = UPCASE(y[1,i]) then do;
				return (1);
			end;
		end;
		return (flag);
	finish CheckMkLab; 

	/* Check singular of factor variable */
	start CheckVarSingular(_var);
		_x = J(nrow(mFactorData),2,1.0);
		flag = 0;
		do j = 1 to ncol(mFactorData);
			do i=1 to nrow(mFactorData);
				_x[i,2] = mFactorData[i,j];
			end;
			xpx=T(_x)*_x;                 /* cross-products */ 
			detm = DET(xpx);
			if abs(detm)<1.0e-10 then do;
				return (1);
			end;
		end;
		return (flag);
	finish CheckVarSingular;

	/* Create of design matrix */
	start GetDMatrix(GKnum,qcomb);
		if GKnum = 1 then do; *BC;
			DMat = J(qcomb,1);
			DMat[1,1] = -0.5; DMat[2,1] = 0.5;
		end;
		else if GKnum = 2 then do; *SF;
			DMat = J(qcomb,2);
			DMat[1,1] = -1;	DMat[2,1] = 0; DMat[3,1] = 1;
			/*DMat[1,2] = -0.5; DMat[2,2] = 1; DMat[3,2] = -0.5;*/
			DMat[1,2] = -0.5; DMat[2,2] = .5; DMat[3,2] = -0.5;/*---sbrlsi---*/
		end;
		else do; *Outcross;
			DMat = J(qcomb,3);
			DMat[1,1] = 1;	DMat[2,1] = 1; DMat[3,1] = -1; DMat[4,1] = -1; 
			DMat[1,2] = 1;	DMat[2,2] = -1; DMat[3,2] = 1; DMat[4,2] = -1; 
			DMat[1,3] = 1;	DMat[2,3] = -1; DMat[3,3] = -1; DMat[4,3] = 1; 
		end;		
		return (DMat);
	finish GetDMatrix;

	start GetColNum(GKnum,cnum,mpos,GPMat);
		pos = -1; minval = 1.0e10;
		*kemicl check;
		*print GPMat;
		*print(cnum);
		*print(mpos);
		do j=1 to ncol(GPMat) by GKnum;
			if cnum = GPMat[1,j] then do;
				*print(GPMat[2,j]);
				val = abs(mpos-GPMat[2,j]);
				*print(val);
				if minval > val then do;
					minval = val; pos = j;
				end;
			end;
		end;
		if pos = -1 then do;
			pos = 1; 
			print "WRONG CHROMOSOME NUMBER!!!";
		end;
		return (pos);
	finish GetColNum;

	/* Create pi matrix from GP data set */
	start CreatePIMatrix(GKnum,qcomb,cnum,mpos,GPMat);
		skip= 3;/*--- because a row of marker indicator was added to the &GPdata. sbrlsi ---*/
		*inum = nrow(GPMat)-2; 
		inum = nrow(GPMat)-skip; /*--- sbrlsi ---*/
		pMat = J(inum,qcomb,0);
		colnum = GetColNum(GKnum,cnum,mpos,GPMat);
		do i=1 to inum;
			sum = 0;
			do j=1 to qcomb-1;
				*pMat[i,j] = GPMat[i+2,colnum+j-1];
				pMat[i,j] = GPMat[i+skip,colnum+j-1]; /*--- sbrlsi --- */
				sum = sum+pMat[i,j];
			end;
			pMat[i,qcomb] = 1-sum;
		end;
		return (pMat);
	finish CreatePIMatrix;

	/* Variance of YMat */
	start MatColVar(YMat,Mean);
		rnum = nrow(YMat); cnum = 1;
		MDiff = J(rnum,cnum,0);
		do i=1 to rnum;
			do j=1 to cnum;
				MDiff[i,j] = YMat[i,j]-Mean[j];
			end;
		end;
		Var = MDiff[##]/(rnum-1);
		return (Var);
	finish MatColVar;
	/* normal distribution function */
	start ZFunction(u,q);
		pi = 3.141592653589793;
		val = exp(-0.5*(u/q)*(u/q))/(sqrt(2*pi*q*q));
		return (val);
	finish ZFunction;
	/* LR when effect = 0 */
	start GetbaseLKVal(pMat,YMat,Mean,Var);
		rnum = nrow(YMat); cnum = ncol(pMat);
		PIMat = J(rnum,cnum);
		do i = 1 to rnum;
			yi = YMat[i];
			uu = Mean;
			q = sqrt(Var);
			do l = 1 to cnum;
				u = yi-uu;
				PIMat[i,l] = pMat[i,l]*ZFunction(u,q);
			end;
		end;
		LKVal = 0.0; sum = PIMat[,+];
		do i = 1 to rnum;
			if sum[i]>0 then LKVal = LKVal + log(sum[i]);
		end;
		return (LKVal);
	finish GetbaseLKVal;

	/* E-step */
	start UpdatePIMatrix(pMat,DMat,EMat,YMat,Var,XX) Global (LKVal,BB);
		rnum = nrow(YMat); cnum = ncol(pMat);
		PIMat = J(rnum,cnum);
		do i = 1 to rnum;
			yi = YMat[i];
			uu = XX[i,]*BB;
			q = sqrt(Var);
			do l = 1 to cnum;
				de = DMat[l,]*T(EMat);
				u = yi-uu-de;
				PIMat[i,l] = pMat[i,l]*ZFunction(u,q);
			end;
		end;
		LKVal = 0.0; sum = PIMat[,+];
		do i = 1 to rnum;
			if sum[i]>0 then LKVal = LKVal + log(sum[i]);
		end;
		*formalize PIMat by divided sum of row;
		do i = 1 to rnum;
			do l = 1 to cnum;
				if sum[i]>0 then PIMat[i,l] = PIMat[i,l]/sum[i];
			end;
		end;
		return (PIMat);
	finish UpdatePIMatrix;

	start getMkNum(chnum,mkpos,allmkpos);
		mknum = 1;
		do i = 1 to ncol(allmkpos);
			if mkpos < allmkpos[chnum,i] then do;
				mknum = i-1; 
				i =  ncol(allmkpos)+1;
			end;
		end;
		return (mknum); 
	finish getMkNum;

	start getAnnoPos(chnum,mkpos,mMapData);
		n = nrow(mMapData);
		do i = 1 to n;
			if (mMapData[i,1]=chnum) then do;
				*if (mMapData[i,2]>mkpos) then do;
					*return (i-1);
				*end;
				/*---begin: sbrlsi---*/
				if (mMapData[i,2] = mkpos) then do;
					return (i);
				end;
				else if (mMapData[i,2] > mkpos) then do; 
					return (i-1);
				end;
				/*---end: sbrlsi---*/
			end;
		end;
	finish getAnnoPos;

	start GetTheMarker(j,t,mCMData1,mCMData2);
		pos = -1;
		mk = J(1,1); n = nrow(mCMData1);
		do i=1 to n;
			if mCMData1[i] = t then do;
				pos = i; i = n+1; 
			end;
		end;
		i = pos+j-1;
		if pos = -1 | i > n then do;
			return ("NULL");
		end;
		if mCMData1[i] = t then do;
			mk = mCMData2[i];
		end;
		else do;
			mk = "NULL";
		end;
		return (mk);
	finish GetTheMarker;

	start GetFlankMarkersPos(chnum,mkpos,mMapData1) Global(m1,m2); *m1,m2 is position;
		n = nrow(mMapData1); pos = 2;
		do i = 1 to n by 1;
			if chnum = mMapData1[i,1] then do;
				if mMapData1[i,2] > mkpos then do;
					return (i);
				end;
			end;
		end;
		return (pos);
	finish GetFlankMarkersPos;

	start GetMarkerPos(mk,chnum,mMapData1,mMapData2);
		mm = 1000; n = nrow(mMapData1);
		do i = 1 to n;
			if chnum = mMapData1[i,1] then do;
				*if mMapData2[i,ncol(mMapData2)] = mk then do;
				if upcase(mMapData2[i,ncol(mMapData2)]) = upcase(mk) then do; /*sbrlsi*/
					mm = mMapData1[i,2];
					return (mm);
				end;
			end;
		end;
		return (mm);
	finish GetMarkerPos;

	start testCMMarker(mk,chnum,mkpos,tWSize,mMapData1,mMapData2);
		flag = 1;
		pos = GetFlankMarkersPos(chnum,mkpos,mMapData1);
		m1 = mMapData1[pos-1,2];
		m2 = mMapData1[pos,2];
		m1 = m1-tWSize; m2 = m2+tWSize;
		mm = GetMarkerPos(mk,chnum,mMapData1,mMapData2);
		if mm >= m1 & mm <= m2 then do;
			flag = 0; *not good;
		end;
		return (flag);
	finish testCMMarker;

	start GetCMarker(mk,mGenoData,mMapData1,mMapData2);
		n = nrow(mGenoData); XX0 = J(n,1); 
		do i=1 to nrow(mMapData1);
			if upcase(mk) = upcase(mMapData2[i,ncol(mMapData2)]) then do;
				j = i; i = nrow(mMapData1)+1;
			end;
		end;
		do i = 1 to n;
			XX0[i,1] = mGenoData[i,j];
		end;
		return (XX0);
	finish GetCMarker;

	start AddCMarkerToXXBB(t,chnum,mkpos,cMkNum,tWSize,mMapData1,mMapData2,mCMData1,mCMData2,mGenoData,XX) Global (BB);
		BBY = J(1,1,0); j = 1; k = 1;
		/*sbrlsi. Note: I think it could be 'do j=1 to ncol(loc(mCMData1[,1] = t))'*/
		do i=1 to 1000 by 1; 
			mk = GetTheMarker(j,t,mCMData1,mCMData2);
			*print(mk);
			if upcase(mk) = "NULL" then do;
				goto ex1;
			end;
			flag = testCMMarker(mk,chnum,mkpos,tWSize,mMapData1,mMapData2);
			if flag = 1 then do; *good;
				XXY = GetCMarker(mk,mGenoData,mMapData1,mMapData2);
				XX = XX||XXY; BB = BB//BBY;
				k = k+1;
			end;
			j = j+1; 
			if k > cMkNum then goto ex1;
		end;
		ex1:
	finish AddCMarkerToXXBB;

	start missingCount(y);
		count = 0;
		do i=1 to nrow(y);
			if y[i] = . then count = count+1;
		end;
		return (count);
	finish missingCount;

	start rmMissingTval(YMat,XX,pMat) global(_YMat,_XX,_pMat);
		ii = 1;
		do i=1 to nrow(YMat);
			if YMat[i] ^= . then do;
				_YMat[ii] = YMat[i];
				do j=1 to ncol(XX);
					_XX[ii,j] = XX[i,j];
				end;
				do j=1 to ncol(pMat);
					_pMat[ii,j] = pMat[i,j];
				end;
				ii = ii+1;
			end;
		end;
	finish rmMissingTval;

	start EM_Calculation(pMat,YMat,DMat,flag,XX) Global (LKVal,baseLKVal,LR,Loops,EMat,Var,BB,_YMat,_XX,_pMat,NTrtObs);
		if any(YMat = .) then do;
			nm = missingCount(YMat);
			n = nrow(YMat)-nm;	m = ncol(XX); m1 = ncol(pMat);
			_YMat = J(n,1,0); _XX = J(n,m,0); _pMat = J(n,m1,0);
			run rmMissingTval(YMat,XX,pMat);
			NTrtObs = Nrow(_YMat);
			run _EM_Calculation(_pMat,_YMat,DMat,flag,_XX);
		end;
		else do;
			NTrtObs = Nrow(YMat);
			run _EM_Calculation(pMat,YMat,DMat,flag,XX);                     /* execute module      */ 
		end;
	finish EM_Calculation;

	/* EM Calculation */
	/* 1:a<>0,d<>0,2:a=0,d<>0,3:a<>0,d=0,4:a=0,d=0 */
	start _EM_Calculation(pMat,YMat,DMat,flag,XX) Global (LKVal,baseLKVal,LR,Loops,EMat,Var,BB);
		*file LOG; 
		LKVal_Sav = 1.0e10; endloop = 300;
		do time = 1 to endloop;
			dcnum = ncol(DMat);
			*PI updated pij matrix;
			PIMat = UpdatePIMatrix(pMat,DMat,EMat,YMat,Var,XX);
			/*kemicl - add max to avoid divide by zero S0568537;
			 put "Time =" time;
			 put "LKVal =" LKVal;
			*/
			testval = ABS(LKVal-LKVal_Sav)/max(1e-4,ABS(LKVal)); 
			if  testval < 1.0e-6 then do; *end of loop;
				Loops = time-1;
				time = endloop+1; *break loop;
			end;
			else do; *continue loop;
				LKVal_Sav = LKVal;
				VMat = J(dcnum,dcnum);
				do i = 1 to dcnum;
					do j = 1 to dcnum;
						DD = DMat[,i]#DMat[,j];
						PDD = PIMat*DD;
						VMat[i,j] = SUM(PDD);
					end;
				end;
				V0Mat = J(dcnum,1);
				Y_1U = YMat-XX*BB;
				do i = 1 to dcnum;
					V0Mat[i] = T(Y_1U)*PIMat*DMat[,i];
				end;
				RMat = J(dcnum,1);
				/*kemicl - add max to avoid divide by zero S0568537;
			      put "VMat=" VMat;
			      put "Dcnum=" dcnum;
			    */
				do i = 1 to dcnum;
					RMat[i] = V0Mat[i]/max(1e-4,VMat[i,i]);
				end;
				MMat = J(dcnum,dcnum,0);
				do i = 1 to dcnum;
					do j = 1 to dcnum;
						if i ^= j then do;
							MMat[i,j] = VMat[i,j]/max(1e-4,VMat[i,i]);
						end;
					end;
				end;
				do j = 1 to ncol(EMat);
					if flag=1 then do;
						EMat[j] = RMat[j]-MMat[j,]*T(EMat);
					end;
					else if flag=2 then do;
						if j=1 then do;
							EMat[j] = 0.0;
						end;
						else do;
							EMat[j] = RMat[j]-MMat[j,]*T(EMat);
						end;
					end;
					else if flag=3 then do;
						if j=2 then do;
							EMat[j] = 0.0;
						end;
						else do;
							EMat[j] = RMat[j]-MMat[j,]*T(EMat);
						end;
					end;
					else if flag=4 then do;
						EMat[j] = 0.0;
					end;
				end;
				t1 = PIMat*DMat*T(EMat); t2 = YMat-t1;
				t3 = INV(T(XX)*XX)*T(XX); BB = t3*t2;
				Y_1U = YMat-XX*BB;
				t1 = T(Y_1U)*Y_1U;
				t2 = 2*T(Y_1U)*PIMat*DMat*T(EMat);
				t3 = EMat*VMat*T(EMat);
				Var = (t1-t2+t3)/nrow(YMat);
				var = max(Var,1e-4);
				*print var;
				/*  Kemicl - This goes negative for S0568537 - set max to avoid singularity */ 
			end; *continue loop;
		end; *time=;
	finish _EM_Calculation;

	/**************************************************************************************/
	/* Build matrices of mMapData1(num),mMapData2(char),mTraitData,mFactorData,mGenoData */
	/* column number of markerlab in matrix */
	%let mkcol = 2;
	/* IndexStr */
	IndexStr = {&newMarkerVars};
	%if &debugflagiml = 1 %then %do;
		print IndexStr;
	%end;
	/* mMapData */
	%*put MapFile= &MapFile; /*sbrlsi*/
	%*put MarkerPos= &MarkerPos; /*sbrlsi*/
	%*put ChromNum=&ChromNum;  /*sbrlsi*/
	%*put ChromLab=&ChromLab;
	%*put MarkerLab=&MarkerLab;
		
	reset deflib=Work;
	
	%if &GKNumber = four and %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) %then %do;

		IndexStrOrig = {&newMarkerVarsOrig};
		
		use &MapFile;
		read all var _NUM_ into xx;
		i1 = 1; i2 = nrow(xx);
		free xx;
		do i=i1 to i2 by 1;
			%if %length(&MarkerPos)^=0 %then %do;
				read point i var {&ChromNum &MarkerPos} into x0;
			%end;
			%else %do;
				read point i var {&ChromNum} into x0;
			%end;
			%if %length(&ChromLab)^=0 %then %do;
				read point i var {&ChromLab &MarkerLab} into x1;
			%end;
			%else %do;
				read point i var {&MarkerLab} into x1;
				%let mkcol = 1;
			%end;
			if CheckMkLab(x1[1,&mkcol],IndexStrOrig) then do;		
				mMapData1o = mMapData1o//x0;
				mMapData2o = mMapData2o//x1;
			end;
		end;	
		
		use outMapFile; 
	%end;
	%else %do;
		use &MapFile;
	%end;
		
	read all var _NUM_ into xx;
	i1 = 1; i2 = nrow(xx);
	free xx;
	do i=i1 to i2 by 1;
		%if %length(&MarkerPos)^=0 %then %do;
			read point i var {&ChromNum &MarkerPos} into x0;
		%end;
		%else %do;
			read point i var {&ChromNum} into x0;
		%end;
		%if %length(&ChromLab)^=0 %then %do;
			read point i var {&ChromLab &MarkerLab} into x1;
		%end;
		%else %do;
			read point i var {&MarkerLab} into x1;
			%let mkcol = 1;
		%end;
		if CheckMkLab(x1[1,&mkcol],IndexStr) then do;		
			mMapData1 = mMapData1//x0;
			mMapData2 = mMapData2//x1;
		end;
	end;
	*print(mMapData1); /*sbrlsi*/
	*print(mMapData2); /*sbrlsi*/
	*print(mMapData2o);/*sbrlsi*/
	
	%*goto exitcheck;

	/*  call symputx('tChromNum',mMapData1[nrow(mMapData1),1]); kemicl took and made 
	tChromNum from number of unique rows in mapdata set */
  	/* %put Tchromnum: &tChromNum; */
	%if &debugflagiml = 1 %then %do;
		print mMapData1 [colname={&ChromNum &MarkerPos}];
		print mMapData2 [colname={&ChromLab &MarkerLab}];
	%end;
	/* mTraitData */
	mTraitLab = {&TraitVar};
	use &TtFile;
	read all var mTraitLab into mTraitData;
	%if &debugflagiml = 1 %then %do;
		print mTraitData [colname={&TraitVar}];
	%end;
	/* Check missing value for each trait */
	%let tmPos = 0;
	ttCount = J(1,ncol(mTraitData),0.0);
	do j=1 to ncol(mTraitData);
		do i=1 to nrow(mTraitData);
			if mTraitData[i,j] ^= . then do; *non-missing;
				ttCount[j] = ttCount[j]+1.0;
			end;
		end;
		ttCount[j] = ttCount[j]*100.0/nrow(mTraitData);
		if ttCount[j] < &tFactorNum+5 then do;
			call symputx('tmPos',j);
			j = ncol(mTraitData)+1;
		end;
	end;
	%if &debugflagiml = 1 %then %do;
		print ttCount;
	%end;
	%if &tmPos > 0 %then %do;
		%put ERROR: Too many missing data in trait &tmPos.  Please de-select this trait.;
		%let exiterror=1;
		%goto exit;
	%end;
	/* mFactorData */
	%if &ftflag = 1 %then %do;
		use &FtFile;
		%let i = 1;
		%let var = %scan(&FactorVarsNew,&i);
		%do %while(%length(&var));
			read all var {&var} into x0;
			if TYPE(x0)='N' then do;
				mFactorData = mFactorData||(x0);
			end;
			else do;
				mFactorData = mFactorData||NUM(x0);
			end;
			%let i = %eval(&i+1);
			%let var = %scan(&FactorVarsNew,&i);
		%end;
		%if &debugflagiml = 1 %then %do;
			print mFactorData [colname={&FactorVarsNew}];
		%end;
		/* Treat missing as -9.99999 */
		*mFactorData[loc(mFactorData = .)] = -9.99999;
		do i = 1 to nrow(mFactorData);
			do j = 1 to ncol(mFactorData);
				if mFactorData[i,j] = . then do;
					mFactorData[i,j] = -999.999;
				end;
			end;
		end;
		%if &debugflagiml = 1 %then %do;
			print mFactorData [colname={&FactorVarsNew}];
		%end;
		/* Check singular of factor variable */
		_x = J(nrow(mFactorData),2,1.0);
		%let mfPos = 0;
		do j = 1 to ncol(mFactorData);
			do i=1 to nrow(mFactorData);
				_x[i,2] = mFactorData[i,j];
			end;
			xpx=T(_x)*_x;                 /* cross-products */ 
			detm = DET(xpx);
			if abs(detm)<1.0e-10 then do;
				call symputx('mfPos',j);
				j = ncol(mFactorData)+1;
			end;
		end;
		%if &mfPos > 0 %then %do;
			%put ERROR: Factor variable &mfPos causes a singular regression matrix.  Please de-select this variable.;
			%let exiterror=1;
			%goto exit;
		%end;
	%end; *if &ftflag;
	
	%if %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) %then %do;/*sbrlsi, outcross. The mGenoData matrix is used only in the CIM*/
		%*put MkFile= &MkFile; /*sbrlsi*/
		
		/* mGenoData */
		use &MkFile;
		%let i = 1;
		%let var = %scan(&newMarkerVars,&i);
		%do %while(%length(&var));
			read all var {&var} into x0;

			if TYPE(x0)='N' then do;
				mGenoData = mGenoData||x0;
			end;
			else do;
				mGenoData = mGenoData||NUM(x0);
			end;
			%let i = %eval(&i+1);
			%let var = %scan(&newMarkerVars,&i);
		%end;
		%if &debugflagiml = 1 %then %do;
			print mGenoData [colname={&newMarkerVars}];
		%end;
	%end; /*sbrlsi, outcross*/
	
	/* Treat missing as -1 for EM algorithm code in IML*/
	*	mGenoData[loc(mGenoData = .)] = -1;
	%if %index(%upcase(&MAlgorithm), HALEY) = 0 %then %do;
		do i = 1 to nrow(mGenoData);
			do j = 1 to ncol(mGenoData);
				if mGenoData[i,j] = . then do;
					mGenoData[i,j] = -1;
				end;
			end;
		end;
	%end;
	%if &debugflagiml = 1 %then %do;
		print mGenoData [colname={&newMarkerVars}];
	%end;
	*******************************************************************************;
	/* IM (MIM) module */
	/* mFactorData */
	%if &ftflag = 1 %then %do;
		call symputx('xNum',ncol(mFactorData)+1); *IM with cofactor;
	%end;
	%else %do;
		%let xNum = 1; *IM without cofactor;
	%end;

	*Whole pij matrix;
	reset deflib=GPLib;	use &GPData;
	read all into GPMat;
	*Trait matrix;
	traitNum = 1; *one trait at a time;
	ttLab = {&TraitVar}; 
	t_traitNum = ncol(ttLab);
	traitLab = ttLab[1];
	reset deflib=work;	use &CrossFile;
	read all var traitLab into YMat;
	indivNum = nrow(YMat);
	*print YMat;
	*Set parameters and E matrix;	
	GK = {"&GKNumber"};
	if GK = "four" then do; /*sbrlsi, outcross*/ 
	*Outcross;
	
		QCombNum = 4; GKnum = 3; 
		Dcname = {A1 A2 D}; Drname = {'' '' '' ''};
		EMat = J(traitNum,3,0); *effect matrix;
	end;
	else if GK = "three" then do; 
	*SF;
		QCombNum = 3; GKnum = 2; 
		Dcname = {A D}; Drname = {'2' '1' '0'};
		EMat = J(traitNum,2,0); *effect matrix;
	end;
	else do; 
	*BC;
		QCombNum = 2; GKnum = 1; 
		Dcname = {A}; Drname = {'1' '0'};
		EMat = J(traitNum,1,0); *effect matrix; /* makes a 1*1 matrix/element 0 */
	end;
	*print Emat;
	*Design matrix;
	
	DMat = GetDMatrix(GKnum,QCombNum);
			
	*print DMat [colname=Dcname][rowname=Drname];
	*Using XB instead of 1u;
	XX0 = J(indivNum,&xNum,1);
	%if &ftflag = 1 %then %do;
		do i = 1 to nrow(XX0);
			do j = 2 to ncol(XX0);
				XX0[i,j] = mFactorData[i,j-1];
			end;
		end;
	%end;
	BB0 = J(&xNum,1,0);
	XX = J(indivNum,&xNum,1); XX = XX0;
	BB = J(&xNum,1,0); BB = BB0;

	*print XX;
	******************************************************************************;
	/* mMapData1 -> allmknum & allmkpos */
	allmknum = J(1,&tChromNum,0);
	/*kemicl replace way to fill in allmknum with the number of markers per chromosome value 
	do i = 1 to nrow(mMapData1);
		j = mMapData1[i,1];
		allmknum[1,j] = allmknum[1,j]+1;
	end;
	*/
	%do i=1 %to &tChromNum;
		%*put Fill in number of markers for Chromosome: &&ChrVars&i;
		allmknum[1,&i] = nrow(mMapdata1[loc(mMapdata1[,1]=&&ChrVars&i),1]);
	%end;
	maxchnum = max(allmknum);
	allmkpos = J(&tChromNum,maxchnum,-1);
	/*Kemicl replace way to fill in allmkpos with marker positions to avoid need for consec. chromosome numbers 
	j = 1; prechnum = 1;
	do i = 1 to nrow(mMapData1);
		chnum = mMapData1[i,1];
		if chnum ^=prechnum then j = 1;
		allmkpos[chnum,j] = mMapData1[i,2];
		j = j+1; prechnum = chnum;
	end;
	*/
	%do i=1 %to &tChromNum;
		%*put Fill in marker positions for Chromosome: &&ChrVars&i;
		posfill = mMapdata1[loc(mMapdata1[,1]=&&ChrVars&i),2];
		do j=1 to nrow(posfill);
			allmkpos[&i,j] = posfill[j];
		end;
	%end;
	/* Kemicl check	
	print allmkpos;
	print allmknum;
	*/
	%if &debugflagiml = 1 %then %do;
		print allmknum, allmkpos;
	%end;
	
	*******************************************************************************;
	%if %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) %then %do; *CIM;
	/*proc print data = &CMFile; title "&CMFile"; run;*/
		
		%if &GKNumber ^= four %then %do;		
		
			%if &CMarkerSM = Using SAS Data Set %then %do; *open data set;
				reset deflib=CMLib; 
			%end;
			%else %do;
				reset deflib=OutLib;
			%end;
			use &CMDataName;
		
		%end;
		%else %do;
			reset deflib=Work;
			use CMDataOutcross;
		%end;
		
		read all var {Trait} into mCMData1;
		read all var {Marker} into mCMData2;
				
		cMkNum = &CMarkerNum; 
		tWSize = &TWinSize;		
		
		*print mMapData1 mMapData2, mCMData1 mCMData2;
		*print mGenoData [colname={&newMarkerVars}];
		*print cMkNum tWSize;
	%end;
	
	*******************************************************************************;
	_YMat = J(1,1); _XX = J(1,1); _pMat = J(1,1);
	*******************************************************************************;
	reset deflib=work;	use TestStepGrid;
	read all var {chnum pos _Lmk_ _LmkCum_ CMID} into Grid;
	reset deflib=work;	use &CrossFile;
	nx = &nx;
	/* add columns for cofactors */	
	nx = nx + ncol(mFactorData);
	%let pdstring = ;/*---necessary for HKMIM. sbrlsi---*/
	%let sepc =c; /*---necessary for HKMIM. sbrlsi---*/
	%let sepp =p; /*---necessary for HKMIM. sbrlsi---*/
	%*let id_TestStepGrid=%sysfunc(open(work.TestStepGrid));
	%*let NObs_TestStepGrid=%sysfunc(attrn(&id_TestStepGrid,NOBS));
	%*syscall set(id_TestStepGrid);	
	
	do t = 1 to t_traitNum;
		*print(t);
		if t > 1 then do;
			traitLab = ttLab[t];
			read all var traitLab into YMat;
		end;
		do j = 1 to ncol(EMat); 
			EMat[j] = 0.0;
		end;
		
		*pMat = CreatePIMatrix(GKnum,QCombNum,1,1,GPMat);
				
		Mean = YMat[:]; Var = MatColVar(YMat,Mean);
		BB[1] = Mean; BB0[1] = Mean;
		*print BB BB0;
		/*---begin: sbrlsi. Using the grid of genomic positions ---*/	
		do ob=1 to nrow(Grid);
		%*do ob= 1 %to &NObs_TestStepGrid;
		
			%*let b=%sysfunc(fetchobs(&id_TestStepGrid,&ob));
			%*put ### Processing chnum=&chnum pos=&pos MarkerName=&MarkerName ###;	
			chmacro= Grid[ob, 1]/*&chnum*/;
			chnum= Grid[ob, 1]/*&chnum*/;
			mkpos= Grid[ob, 2]/*&pos*/;
			mknum = Grid[ob, 3]/*&_Lmk_*/;
			j = Grid[ob, 4]/*&_LmkCum_*/;			
			CMID= Grid[ob, 5] /*&CMID*/;
			
			%if %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) %then %do; /*---CIM---*/
				* Modify XX, and BB;
				XX = J(indivNum,&xNum,1); XX = XX0;
				BB = J(&xNum,1,0); BB = BB0;
				call AddCMarkerToXXBB(t,chmacro,mkpos,cMkNum,tWSize,mMapData1,mMapData2,mCMData1,mCMData2,mGenoData,XX);
				*print t chnum mkpos XX BB;
			%end;

			*print(mknum);
			*print(j);
			*print(mMapData2);
			
			%if &GKNumber = four and %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) %then %do;
				%if %length(&ChromLab)^=0 %then %do;
					clab = propcase(mMapData2o[j,1]);
					cmlab = cmlab//(propcase(mTraitLab[1,t])||clab||propcase(mMapData2o[j,&mkcol]));
				%end;
				%else %do;
					cmlab = cmlab//(propcase(mTraitLab[1,t])||propcase(mMapData2o[j,&mkcol]));
				%end; 
			%end;
			%else %do;
				%if %length(&ChromLab)^=0 %then %do;
					clab = propcase(mMapData2[j,1]);
					cmlab = cmlab//(propcase(mTraitLab[1,t])||clab||propcase(mMapData2[j,&mkcol]));
				%end;
				%else %do;
					cmlab = cmlab//(propcase(mTraitLab[1,t])||propcase(mMapData2[j,&mkcol]));
				%end; 						
			%end;
			*print(chnum);		
			pMat = CreatePIMatrix(GKnum,QCombNum,chnum,mkpos,GPMat);
			mkpos_m = mkpos/100;
			%if %index(%UPCASE(&MAlgorithm),HALEY) %then %do;
			
				%if %upcase(&MethodSelection) ^= %upcase(Multiple Interval Mapping) 
					| &GKNumber = four %then %do;
					
					/*--- Prepare data to output from IML to run Haley-Knott macro ---*/
					traitMat = t * J(nrow(YMat),1,1);
					chrMat = chnum * J(nrow(YMat),1,1);
					posMat = mkpos * J(nrow(YMat),1,1);
					cmidMat = CMID * J(nrow(YMat),1,1); *---sbrlsi---*;
					ncx = ncol(XX);
					ncx1 = ncx-1;
					if ncx ^= nx then do;
						XXS = XX || J(nrow(XX),nx-ncx,0);
					end;
					else do;
						XXS = XX;
					end;

					SaveMat = traitMat || chrMat || posMat || cmidMat || YMat || XXS || pMat;/*--- sbrlsi ---*/					
					ncsm = ncol(SaveMat);
					if t=1 & /*chnum=1*/ ob=1 & CMID=1 then do; *--- sbrlsi ---*;
						SaveMatAll = SaveMat;
						ncsm0 = ncsm;
					end;
					else do;
						SaveMatAll = SaveMatAll // SaveMat;
					end;
				%end;
				
				%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
					/*---Begin: Creating matrix data for HKMIM---*/
					/*---Prepare data to output from IML to run Haley-Knott macro ---*/
					
					reset deflib=work;	
					use &CrossFile;
					read all var {&TraitVar} into YMat;
					
					traitMat2 = t * J(nrow(YMat),1,1);
					chrMat2 = chnum * J(nrow(YMat),1,1);
					posMat2 = mkpos * J(nrow(YMat),1,1);
					cmidMat2 = CMID * J(nrow(YMat),1,1); *---sbrlsi---*;
					obs_ = t(1:nrow(YMat)); *---sbrlsi---*;

					if t=1 then do;
						ncx = ncol(XX);
						ncx1 = ncx-1;
						if ncx ^= nx then do;
							XXS = XX || J(nrow(XX),nx-ncx,0);
						end;
						else do;
							XXS = XX;
						end;
				
						chrMat = chnum * J(1,1,1);
						posMat = mkpos * J(1,1,1);
						
						%if &GKNumber= four %then %do;
							a1 = pMat[,1] + pMat[,2] - pMat[,3] - pMat[,4];
							
							d = pMat[,1] - pMat[,2] - pMat[,3] + pMat[,4];
							
							a2 = pMat[,1] - pMat[,2] + pMat[,3] - pMat[,4];
							
							if /*chnum=1*/ ob=1 & CMID=1 then do;
								SaveMatAll2 = obs_ || traitMat2 || chrMat2 || posMat2 || cmidMat2 || YMat || XXS || a1 || d || a2; 
							end;
							else do;
								SaveMatAll2 = SaveMatAll2 // (obs_ || traitMat2 || chrMat2 || posMat2 || cmidMat2 || YMat || XXS || a1 || d || a2); 
							end;
							
						%end;						
						%else %if &GKNumber= three %then %do;
							a1 = pMat[,3] - pMat[,1];
						
							d = .5*pMat[,2] - .5*(pMat[,1] + pMat[,3]);
							
							if /*chnum=1*/ ob=1 & CMID=1 then do;
								SaveMatAll2 = obs_ || traitMat2 || chrMat2 || posMat2 || cmidMat2 || YMat || XXS || a1 || d; 
							end;
							else do;
								SaveMatAll2 = SaveMatAll2 // (obs_ || traitMat2 || chrMat2 || posMat2 || cmidMat2 || YMat || XXS || a1 || d); 
							end;
							
						%end;
						%else %do;
							a1 = pMat[,2] - pMat[,1];
							
							if /*chnum=1*/ ob=1 & CMID=1 then do;
								SaveMatAll2 = obs_ || traitMat2 || chrMat2 || posMat2 || cmidMat2 || YMat || XXS || a1 ; 
							end;
							else do;
								SaveMatAll2 = SaveMatAll2 // (obs_ || traitMat2 || chrMat2 || posMat2 || cmidMat2 || YMat || XXS || a1 ); 
							end;
							
						%end;											
					end; 
					
				%end;/*---end: creating matrix data for HKMIM---*/
				
			%end;/*---end: Haley-Knott---*/
			%else %do; 	
				/*--- Run EM Algorithm ---*/
				/************************************************************/
				/*--- 1:a<>0,d<>0,2:a=0,d<>0,3:a<>0,d=0,4:a=0,d=0 ---*/
				xpx = T(XX)*XX;	detm = DET(xpx);
				if abs(detm)<1.0e-10 then do;
					XX = J(indivNum,1,1); 
					BB = J(1,1,0);
				end;
				%if &GKNumber= three %then %do;
					Call EM_Calculation(pMat,YMat,DMat,1,XX); LKV1 = LKVal; EMat1 = EMat; Loops1 = Loops;
					Call EM_Calculation(pMat,YMat,DMat,2,XX); LKV2 = LKVal; EMat2 = EMat;
					Call EM_Calculation(pMat,YMat,DMat,3,XX); LKV3 = LKVal; EMat3 = EMat;
					Call EM_Calculation(pMat,YMat,DMat,4,XX); LKV4 = LKVal; EMat4 = EMat;
					LR1 = 2*(LKV1-LKV4); *test a=0,d=0;
					LR2 = 2*(LKV1-LKV2); *test a=0;
					LR3 = 2*(LKV1-LKV3); *test d=0;
					tResult = t||chmacro||mknum||mkpos_m||NTrtObs||LR1||LR2||LR3||EMat1||EMat2[2]||EMat3[1]||Loops1;
				%end;
				%else %do;
					Call EM_Calculation(pMat,YMat,DMat,1,XX); LKV1 = LKVal; EMat1 = EMat; Loops1 = Loops; 
					Call EM_Calculation(pMat,YMat,DMat,2,XX); LKV2 = LKVal; 
					LR = 2*(LKV1-LKV2);
					tResult = t||chmacro||mknum||mkpos_m||NTrtObs||LR||EMat1||Loops1;
				%end;
				*******************************************************************************;
				Result = Result//tResult;	
			%end;			
		%*end;
		end;
	end;/*---trait---*/ 
	%*let id_TestStepGrid = %sysfunc(close(&id_TestStepGrid)); /*---end: sbrlsi---*/
	
	%if %index(%UPCASE(&MAlgorithm),HALEY) %then %do;
		
		%if %upcase(&MethodSelection) ^= %upcase(Multiple Interval Mapping) 
			| &GKNumber = four %then %do;

			numxcols = ncol(XXS);
			call symputx('ncolx',numxcols-1);

			%let xstring = "x0";
			%do i=1 %to &ncolx;
				%let xstring = &xstring "x&i";
			%end;

			/*--- output prep for data to send to Haley Knott ---*/

			%if &GKNumber= four %then %do;
				%let ngprobs = 4;
				%let pdstring = "p1" "p2" "p3" "p4"; 
			%end;
			%else %if &GKNumber= three %then %do;
				%let ngprobs = 3;
				%let pdstring = "p1" "p2" "p3"; 
			%end;
			%else %do;
				%let ngprobs = 2;
				%let pdstring = "p1" "p2"; 
			%end;

			create hkdata from SaveMatAll [colname={"Trait" "Chrom" "CM" "CMID" "Y" &xstring &pdstring}]; 
			append from SaveMatAll;
		%end; 

		%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
			/*---begin: sbrlsi. Creating HKMIM data set---*/
			
			numxcols = ncol(XXS);
			call symputx('ncolx',numxcols-1);

			%let xstring = "x0";
			%do i=1 %to &ncolx;
				%let xstring = &xstring "x&i";
			%end;

			numYmatcols = ncol(Ymat);
			call symputx('ncolYmat',numYmatcols);
			
			%let Ystring=;
			%do i=1 %to &ncolYmat;
				%let Ystring= &Ystring "Y&i";
			%end;

			%global nt;
			call symputx("nt", t_traitNum);
			
			%if &GKNumber= four %then %do;
				%let ngprobs = 4;
				%let pdstring2 = "xa2_" "xd_" "xa1_"; 
			%end;
			%else %if &GKNumber= three %then %do;
				%let ngprobs = 3;
				%let pdstring2 = "xa2_" "xd_"; 
			%end;
			%else %do;
				%let ngprobs = 2;
				%let pdstring2 = "xa2_"; 
			%end;		
			
			create hkmimdata from SaveMatAll2 [colname={"obs_" "Trait" "Chrom" "CM" "CMID" &Ystring &xstring &pdstring2}];
			append from SaveMatAll2; 
			
		%end; /*---end: sbrlsi. creating HKMIM data set---*/
	%end;
	%else %do;
		%if &GKNumber= four %then %do;
			%let ngprobs = 4;
		%end;
		%else %if &GKNumber= three %then %do;
			%let ngprobs = 3;
		%end;
		%else %do;
			%let ngprobs = 2;
		%end;
			
		/*output data prep for EM Alg results */
		%if &GKNumber = two %then %do;
			cname = {Trait Chrom MarkerNum Morgans NObs LR Additive2 Iteration};
		%end;
		%if &GKNumber = three %then %do;
			cname = {Trait Chrom MarkerNum Morgans NObs LR LR2 LR3 Additive01 Dominance01 Dominance02 Additive03 Iteration};
		%end;
		*print Result [colname = cname];
		/* put result into SAS data set */
		create work.r1 from Result [colname = cname]; 
		append from Result;
	%end;
	%if %length(&ChromLab)^=0 %then %do;
		create work.r2 from cmlab [colname={'ttLab' "&ChromLab" "&MarkerLab"}]; 
	%end;
	%else %do;
		create work.r2 from cmlab [colname={'ttLab' "&MarkerLab"}]; 
	%end;
	append from cmlab;
	*print cmlab;
	*ods html close;
	
quit;/*End of proc IML*/

%*goto exitcheck;

%if &debugflag = 1 %then %do;
	data outlib.hkdata; set hkdata; run;
%end;

/*create macro var lists for output variables */
%let i = 1;
%do %while(%length(%scan(&TraitVar,&i)));
	%let var = %scan(&TraitVar,&i);
	%let tlab_lod&i = LOD &var;
	/*
	%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
		%let tlab_a2&i = Additive2 &var;
		%let tlab_d&i = Dominance &var;
		%let tlab_a1&i = Additive1 &var;
	%end;
	%else %do;
		%let tlab_a2&i = Additive &var;
		%let tlab_d&i = Dominance &var;
	%end;
	*/
	%if &GKNumber = four %then %do;
		%let tlab_a2&i = Additive2 &var;
		%let tlab_d&i = Dominance &var;
		%let tlab_a1&i = Additive1 &var;	
	%end;
	%else %do;
		%let tlab_a2&i = Additive &var;
		%let tlab_d&i = Dominance &var;
	%end;
	
	%let tlab_r2&i = R2 &var;
	%let tlab_n&i = NObs &var;
	/* use for permutations*/
	%let tlab_plod&i = ProbLOD &var;
	%let tlab_cwt&i = CW LOD Threshold &var;
	
	%let i = %eval(&i+1);
%end;

%if %index(%UPCASE(&MAlgorithm), HALEY) %then %do;

	/*begin: sbrlsi, outcross*/
	%if &GKNumber = four %then %do;
	
		data hkdata_t1;
			set hkdata;
			%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
				where trait = 1;
			%end;
		run;
		
		ods exclude all;
		ods noresults;
		proc mixed data=hkdata_t1;
			%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
				by chrom CMID;
			%end;
			%else %do;
			 	by trait chrom CMID;
			%end;
			model y= p1 p2 p3/ solution;
			estimate 'Additive2' p1 .25 p2  .25 p3 -.25;
			estimate 'Dominance' p1 .25 p2 -.25 p3 -.25;  
			estimate 'Additive1' p1 .25 p2 -.25 p3  .25;
			ods output estimates=ests_t1;
		run;
	
		data ests_t1 (drop= Label StdErr DF tValue Probt); 
			set ests_t1(rename = (Estimate = Additive2));
			where Label contains "Additive2";
			set ests_t1(rename = (Estimate = Dominance));
			where Label contains "Dominance";
			set ests_t1(rename = (Estimate = Additive1));
			where Label contains "Additive1";
		run;
		
		proc datasets nolist;
			delete hkdata_t1;
		quit;

	%end;
	/*end: sbrlsi, outcross*/
	
	%if %upcase(&MethodSelection) = %upcase(Interval Mapping) | %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) %then %do; *--- sbrlsi---*;

		/* get nobs as macros for Rsqr calculation */
		proc means data=&TtFile n noprint;
			var &TraitVar;
			output out = ntraits n=/autoname;
		run;

		data _null_;
			set ntraits;
			%let i = 1;
			%do %while(%length(%scan(&TraitVar,&i)));
				%let CurrT = %scan(&TraitVar,&i);
				call symputx("Nobs&i", trim(left(&CurrT._N)));
				%*put &&Nobs&i;
				%let i = %eval(&i+1);
			%end;
		run;
		
		%if &GKNumber = four %then %do; /*sbrlsi, outcross*/
			
			data hkdata(drop= p1 p2 p3 p4 Additive2 Dominance Additive1);
				merge hkdata ests_t1;
				by trait chrom CMID;
				if Additive2 = . and Dominance = . and Additive1 =. then do;
					xa2_ = 0;
					xd_  = 0;
					xa1_ = 0;
					*case1 = all effects are non-estimable*;
				end;
				else if Additive2 = . and Dominance = . and Additive1 ^=. then do;
					xa2_ = 0;
					xd_  = 0;
					xa1_ = p1 - p2 + p3 - p4;
					*case2= a1 is estimable*;
				end;
				else if Additive2 = . and Dominance ^= . and Additive1 = . then do;
					xa2_ = p1 + p2 - p3 - p4;
					xd_  = p1 - p2 - p3 + p4;
					xa1_ = 0;
					*case3= a2 is estimable*;
				end;
				else if Additive2 = . and Dominance ^= . and Additive1 ^= . then do;
					xa2_ = 0;
					xd_  = p1 - p2 - p3 + p4;
					xa1_ = p1 - p2 + p3 - p4;
					*case4= d and a1 are estimable*;
				end;
				else if Additive2 ^= . and Dominance = . and Additive1 = . then do;
					xa2_ = p1 + p2 - p3 - p4;
					xd_  = 0;
					xa1_ = 0;
					*case5= a2 is estimable*;
				end;
				else if Additive2 ^= . and Dominance = . and Additive1 ^= . then do;
					xa2_ = p1 + p2 - p3 - p4;
					xd_  = 0;
					xa1_ = p1 - p2 + p3 - p4;
					*case6= a1 and a2 are estimable*;
				end;
				else if Additive2 ^= . and Dominance ^= . and Additive1 = . then do;
					xa2_ = p1 + p2 - p3 - p4;
					xd_  = p1 - p2 - p3 + p4;
					xa1_ = 0;
					*case7= a2 and d are estimable*;
				end;
				else if Additive2 ^= . and Dominance ^= . and Additive1 ^= . then do;
					xa2_ = p1 + p2 - p3 - p4;
					xd_  = p1 - p2 - p3 + p4;
					xa1_ = p1 - p2 + p3 - p4;
					*case8= all effects are estimable*;
				end;
			run;	
	
			proc datasets nolist;
				delete ests_t1;
			quit;
		
		%end;
	
		%if &Perms %then %do;
			/* run first to get actual LOD*/
			%haleyknott(nx=&ncolx,ng=&ngprobs,perm=0);	
		%end;

		/* run to get permutations or if perms is 0 then get actual LOD */
		%haleyknott(nx=&ncolx,ng=&ngprobs,perm=&Perms);

		/* check tall format for all missing lod if so throw error*/
		data checkresults; set hk0;
			where LOD > 0;
			if LOD ne .;
		run;

		%nobsvars(checkresults);
		%if &nobs = 0 %then %do;
			%put ERROR:  The specified model returned an infinite likelihood or was not able to calculate LOD scores. Please check your model for possible errors.;
			%let exiterror = 1;
			%goto exit;
		%end;

		%if &debugflag = 1 %then %do;
			data outlib.hkout; set hkout; run;
		%end;
		%let ftrait = %upcase(%scan(&TraitVar,1));
		
		/*data OutLib.qtl_r2; set r2; run;*/
		
		data r2; set r2;
			where trim(left(upcase(ttLab))) = "&ftrait";
			drop ttLab;
		run;
		
		data &MappingResult_OutData;
			set r2;
			length Chrom 5 Morgans 8;
			set hkout; 
			rename Chrom = &ChromNum;
			Morgans = cm/100;
		run;
		
		%let outsasdata = &MappingResult_OutData;
	%end;
	%else %if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
		
		/*---Log Window and Output Window options ---*/
		*DM "log; *clear; *";
		ods exclude all; /*--- prevents all output objects from appearing in the output window ---*/
		*ods select listing;/*--- enables output objects to appear in the output window --- */
		ods noresults; /*--- prevents output objects from appearing in the results window ---*/
		*ods results on; /*--- enables output objects to appearing in the results window ---*/
		options nosource nonotes;/*--- prevents NOTE from being written down in the LOG window---*/
		*ods trace off; /*--- prevents display of information on output created on procs from appearing in the LOG window ---*/			
		
		/*begin: sbrlsi*/
		%if &GKNumber = four %then %do;
			
			proc datasets nolist;
				delete hkdata;
			quit;
		
			data hkmimdata(drop= za1_ zd_ za2_ Additive2 Dominance Additive1);
				merge hkmimdata(rename=(xa2_= za1_ xd_ = zd_ xa1_= za2_)) ests_t1;
				by chrom CMID;
				if Additive2 = . and Dominance = . and Additive1 =. then do;
					xa2_ = 0;
					xd_  = 0;
					xa1_ = 0;
					contrast_= '1';
					*case1= all effects are non-estimable*;	
				end;
				else if Additive2 = . and Dominance = . and Additive1 ^=. then do;
					xa2_ = 0;
					xd_  = 0;
					xa1_ = za2_;
					contrast_= '2';
					*case2= a2 is estimable*;
				end;
				else if Additive2 = . and Dominance ^= . and Additive1 = . then do;
					xa2_ = za1_;
					xd_  = zd_;
					xa1_ = 0;
					contrast_= '3';
					*case3= a1 is estimable*;
				end;
				else if Additive2 = . and Dominance ^= . and Additive1 ^= . then do;
					xa2_ = 0;
					xd_  = zd_;
					xa1_ = za2_;
					contrast_= '4';
					*case4= d and a2 are estimable*;
				end;
				else if Additive2 ^= . and Dominance = . and Additive1 = . then do;
					xa2_ = za1_;
					xd_  = 0;
					xa1_ = 0;
					contrast_= '5';
					*case5= a1 is estimable*;
				end;
				else if Additive2 ^= . and Dominance = . and Additive1 ^= . then do;
					xa2_ = za1_;
					xd_  = 0;
					xa1_ = za2_;
					contrast_= '6';
					*case6= a1 and a2 are estimable*;
				end;
				else if Additive2 ^= . and Dominance ^= . and Additive1 = . then do;
					xa2_ = za1_;
					xd_  = zd_;
					xa1_ = 0;
					contrast_= '7';
					*case7= a1 and d are estimable*;
				end;
				else if Additive2 ^= . and Dominance ^= . and Additive1 ^= . then do;
					xa2_ = za1_;
					xd_  = zd_;
					xa1_ = za2_;
					contrast_= '8';
					*case8= all effects are estimable*;
				end;
			run;
			
			data hkmimdata_tmp;
				set hkmimdata;
				by chrom CMID;
				if first.CMID then output;
			run;
			
			data TestStepGrid;
				merge TestStepGrid hkmimdata_tmp(rename=(chrom= chnum) keep=chrom CMID contrast_);
				by chnum CMID;
			run;
			
			proc datasets nolist;
				delete ests_t1
					hkmimdata_tmp;
			quit;
		%end;
		/*end: sbrlsi, outcross*/
		
		%global HKMIMData_OutData HKMIMTestGrid_OutData 
			HKMIMModelRaw_OutData HKMIMModelSum_OutData
			HKMIMFitTable_OutData HKMIMQtlTable_OutData
			HKMIMEffTable_OutData HKMIMCofTable_OutData
			HKMIMSetTable_OutData;
			
		%PrepOutDS(HKMIMData_OutData,HKMIMData_OutDataName,&oname,_rd_hk);
		%PrepOutDS(HKMIMTestGrid_OutData,HKMIMTestGrid_OutDataName,&oname,_ts_hk);
		%PrepOutDS(HKMIMModelRaw_OutData,HKMIMModelRaw_OutDataName,&oname,_mr_hk);
		%PrepOutDS(HKMIMModelSum_OutData,HKMIMModelSum_OutDataName,&oname,_ms_hk);
		
		%PrepOutDS(HKMIMFitTable_OutData,HKMIMFitTable_OutDataName,&oname,_ft_hk);
		%PrepOutDS(HKMIMQtlTable_OutData,HKMIMQtlTable_OutDataName,&oname,_qt_hk);
		%PrepOutDS(HKMIMEffTable_OutData,HKMIMEffTable_OutDataName,&oname,_et_hk);
		%PrepOutDS(HKMIMCofTable_OutData,HKMIMCofTable_OutDataName,&oname,_ct_hk);
		%PrepOutDS(HKMIMSetTable_OutData,HKMIMSetTable_OutDataName,&oname,_st_hk);
		
		*--- save SAS data sets to OutLib ---*; 
		%*put HKMIMData_OutData = &HKMIMData_OutData;
		%*put HKMIMData_OutDataName = &HKMIMData_OutDataName;
		data &HKMIMData_OutData(drop = X0);
			set work.hkmimdata;
			label Y1= "&TraitVar";
			%do i=1 %to &ncolx;
				label x&i = "%scan(&FactorVarsNew, &i)";
			%end;
		run;

		%*put HKMIMTestGrid_OutData = &HKMIMTestGrid_OutData;		
		data &HKMIMTestGrid_OutData(rename=(%if %length(&ChromLab)^=0 %then ChLab= &ChromLab; 
							MrkPos= &MarkerPos MarkerName= &MarkerLab 
							Pos= CM ChNum = &ChromNum) 
					   drop= _Lmk_ _LmkCum_ _mkpos_); /*Must drop mkpos here to avoid error in the MultiQTLModel_Macro.sas*/
			set work.TestStepGrid;
		run;
						
		/*--- output library to store results from MIM analyses ---*/
		%let g_OutLib = OutLib; 
					
		/* MIM macro variables not from MIM Options tab but that are already created before calling MIM macro fuctions */
		%let g_mapgrid= &HKMIMTestGrid_OutData;
		%let g_hkmimdata= &HKMIMData_OutData;
		
		%let g_fittable= &HKMIMFitTable_OutData;
		%let g_qtltable= &HKMIMQtlTable_OutData;
		%let g_efftable= &HKMIMEffTable_OutData;
		%let g_coftable= &HKMIMCofTable_OutData;
		%let g_settable= &HKMIMSetTable_OutData;		
		
		/* MIM macro variables neither from MIM Options tab nor created before calling MIM macro functions */
		%let g_hkmimmodel= hkmimmodel;
		
		/* Passing values of macro variables not from MIM Options tab to my old macro variables */
		%let g_trt = 1;
		%let g_ng= &ngprobs;
		%let g_nx= &ncolx;
		
		/*
		%let g_ChromLab= &ChromLab;
		%let g_ChromNum= &ChromNum; 
		%let g_MarkerLab= &MarkerLab; 
		%let g_MarkerPos= &MarkerPos; 
		%let g_TraitVar= &TraitVar;
		%let g_FactorVars= &FactorVarsNew;	
		%let g_GDMetric= &GDMetric;
		%let g_LODThresh = &LODThresh;
		*/
		
		/* Passing values of macro variables from MIM Options tab to my old macro variables */
		%let g_teststep= &TestStep;
		%let g_wsize= &MIMWinSize;
		%let g_criterion_ref= &MIMEntryCriterionMain;	
		
		%let g_criterion_main= &MIMEntryCriterionMain;
		%let g_lod_threshold_main= &MIMEntryLODThreshMain; /*=.4*/
		%let g_probf_entry_main= &MIMEntryProbMain; 

		%let g_criterion_dropmqtlonly= &MIMEntryCriterionMain;
		%let g_lod_threshold_dropmqtlonly= &MIMStayLODThreshMain;
		%let g_probf_stay_dropmqtlonly= &MIMStayProbMain; 
		%let g_auto_dropmqtlonly = TRUE;
		%let g_teststep_main= &MIMTestStepMain;
		
		%let g_criterion_epi0D= &MIMEntryCriterionEpi;
		%let g_lod_threshold_epi0D= &MIMEntryLODThreshEpi; /*=.4*/
		%let g_probf_entry_epi0D= &MIMEntryProbEpi; 
		
		%let g_criterion_epi1D= &MIMEntryCriterionEpi;
		%let g_lod_threshold_epi1D= &MIMEntryLODThreshEpi; /*=.4*/
		%let g_probf_entry_epi1D= &MIMEntryProbEpi; 
		%let g_teststep_epi1D= &MIMTestStepEpi1D; *--- g_teststep_epi1D must be > TestStep ---;
		
		%let g_criterion_epi2D= &MIMEntryCriterionEpi;
		%let g_lod_threshold_epi2D= &MIMEntryLODThreshEpi; /*=.18*/
		%let g_probf_entry_epi2D= &MIMEntryProbEpi; 
		%let g_teststep_epi2D= &MIMTestStepEpi2D; *--- g_teststep_epi2D must be > TestStep ---;
		
		%let g_criterion_dropepieff= &MIMEntryCriterionEpi;
		%let g_lod_threshold_dropepieff= &MIMStayLODThreshEpi;
		%let g_probf_stay_dropepieff= &MIMStayProbEpi; 
		%let g_auto_dropepieff = TRUE;
		
		/* Temporary macro variables for testing purpose only */
		%global g_firstobs;
		%let g_firstobs=;
		%*let g_firstobs=1;
		%global g_obs;
		%let g_obs=;
		%*let g_obs= 70;

		%include "&MacroPath./MultiQTLModel_Macro.sas" / nosource;
		
		%*goto exitcheck;
			
		%nullmodel(nx=&g_nx, trt=&g_trt, hkmimmodel= &g_hkmimmodel);

		%fitmodel(trt=&g_trt, ng= &g_ng, hkmimdata= &g_hkmimdata, hkmimmodel=&g_hkmimmodel);			

		%PrepTablesToAdd(fittedstat= fitmodel_st, fittedmodel= fitmodel_sum, modelnumb=0); /*fitmodel_st and fitmodel_sum come from fitmodel()*/
	        		
		%profile(trt=&g_trt, ng= &g_ng, hkmimdata=&g_hkmimdata, mapgrid=&g_mapgrid, 
			hkmimmodel=&g_hkmimmodel, wsize=&g_wsize);
							
		data &MappingResult_OutData.0;
			set profile; /*output data set from profile()*/
			
		run;
		
		/*
		proc means data= &MappingResult_OutData.0 max noprint;
			var LOD&g_trt ;
			output out=LODmaxdata max=maxLODvar;
		run;
		*/		
		
		proc means data=&MappingResult_OutData.0 max min noprint;
			var LOD&g_trt Additive2&g_trt 
				%if &ngprobs >= 3 %then Dominance&g_trt;
				%if &ngprobs = 4 %then Additive1&g_trt;
				;
			output out=LODmaxdata max=maxLODvar maxadd2Var 
						%if &ngprobs >= 3 %then maxDomVar; 
						%if &ngprobs = 4 %then maxadd1Var;
						min=minLODvar minadd2Var 
						%if &ngprobs >= 3 %then minDomVar;
						%if &ngprobs = 4 %then minadd1Var;;
		run;

		%global maxLOD;
		%let amaxadd2 = 0;
		%let aminadd2 = 0;
		%let amaxadd1 = 0;
		%let aminadd1 = 0;
		%let amaxDom = 0;
		%let aminDom = 0;
		data _null_;
			set LODmaxdata;
			call symputx('maxLOD', round(maxLODvar, .01));
			call symputx('amaxadd2', abs(round(maxadd2var, .01)), 'L');
			call symputx('aminadd2', abs(round(minadd2var, .01)), 'L');
			%if &ngprobs >= 3 %then %do;
				call symputx('amaxDom', abs(round(maxDomvar, .01)), 'L');
				call symputx('aminDom', abs(round(minDomvar, .01)), 'L');			
			%end;
			%if &ngprobs = 4 %then %do;
				call symputx('amaxadd1', abs(round(maxadd1var, .01)), 'L');
				call symputx('aminadd1', abs(round(minadd1var, .01)), 'L');			
			%end;
		run;
		
		%global maxEff;
		%let maxEff = %sysfunc(max(&amaxadd2, &aminadd2, &amaxDom, &aminDom, &amaxadd1, &aminadd1));
		*--- to improve graphical resolution, add 0.1 to the maxEff ---*;
		%if %sysevalf(&maxEff > 10000) %then 
			%let maxEff = 10000;
		%else %if %sysevalf(&maxEff < -10000) %then 
			%let maxEff = -10000;
		%else 
			%let maxEff = %sysevalf(&maxEff + 0.15); 		
		
		*--- to improve graphical resolution, add 0.5 to the maxLOD ---*;
		%if %sysevalf(&maxLOD > 10000) %then 
			%let maxLOD = 10000;
		%else %if %sysevalf(&maxLOD < -10000) %then 
			%let maxLOD = -10000;
		%else 
			%let maxLOD = %sysevalf(&maxLOD + 0.5); 
		%*put maxLOD = &maxLOD;
		
		*--- delete temporary data set ---*;
		proc datasets nolist;
			delete LODmaxdata;
		run;		
		
		%let outsasdata = &MappingResult_OutData.0;

		/*send raw model data set to outlib*/
		data &HKMIMModelRaw_OutData.0;
			set &g_hkmimmodel;
		run;		
		
		/*send setting data set to outlib*/
		data &HKMIMSetTable_OutData.0;
			*TestStep= &MIMTestStepMain;
			TestStep= &TestStep; /*The IM uses TestStep instead of MIMTestStepMain*/
			label TestStep= "QTL Test Step in cM";
		run;
		
		%*goto exitcheck; *--- to be used when running from SAS ---*;
		%*goto exit; *--- to be used when running from JMPG ---*;
	%end;
%end;
%else %do; /*EM alg.*/
	%let rstr = Trait ttLab Chrom;
	%if %length(&ChromLab)^=0 %then %do;
		%let rstr = &rstr &ChromLab;
	%end;
	%if &GKNumber = two %then %do;
		%let rstr = &rstr MarkerNum &MarkerLab Morgans NObs LR Additive2;
		%let rstr = &rstr Iteration;
	%end;
	%else %do;
		%let rstr = &rstr MarkerNum &MarkerLab Morgans NObs LR LR2 LR3 Additive01 Dominance01 Dominance02 Additive03 ;
		%let rstr = &rstr Iteration;
	%end;

	data &MappingResult_OutData;
		retain &rstr; 
		set work.r2; set work.r1;
		sindex=_n_;
		Output;
	run;

	proc sort data=&MappingResult_OutData;
		by &MarkerLab;
	run;

	data mapsort;
		set &MapFile; 
		&MarkerLab = propcase(&MarkerLab);
	run;

	proc sort data=mapsort;
		by &MarkerLab;
	run;
	
	data &MappingResult_OutData;
		merge mapsort(in=a) &MappingResult_OutData(in=b);
		by &MarkerLab;
		if b;
	run;
	
	proc sort data=&MappingResult_OutData out=&MappingResult_OutData(drop=sindex);
		by sindex;
	run;

	%if &debugflag = 1 %then %do;
		proc print data=&MappingResult_OutData; Title "Results of QTL Mapping"; Title2 "&MethodSelection Analysis";run;		
	%end;

	%let maxLOD = 5.00;
	proc means data = &MappingResult_OutData noprint; Output out = _ttt_; var LR; run;
	data _null_;
		set _ttt_;
		if _n_ = 3 then do;
			call symputx('maxLOD',round(LR*0.217, .01) + 2);
		end;
	run;
	%*put maxLOD = &maxLOD;

	%*put MappingResult_OutData = &MappingResult_OutData;
	%*put MappingResult_OutDataName = &MappingResult_OutDataName;
	%let outhtml = &OutPath.&MappingResult_OutDataName..html;
	%*put outhtml = %str(&outhtml);

	/* kemicl added to have cM as well as morgans, flip additive sign to account for new coding*/
	data &MappingResult_OutData;
		set &MappingResult_OutData;
		cm = 100*Morgans;
		%if &GKNumber = two %then %do;
			Additive2 = Additive2;
		%end;
		%else %do;
			Additive01 = Additive01;
			Additive03 = Additive03;
		%end;
		label cm="cM";
	run;

	ods html file="&outhtml";
	*ods listing close;

	*options nodate pageno=1 linesize=150 pagesize=100;
	options date pageno=1 linesize=200 pagesize=100;
	Title "   "; Title2 "Results of QTL Mapping - &MethodSelection Analysis";
	%if &GKNumber = two %then %do;
		%if %length(&ChromLab)^=0 %then %do;
			%let cname = Trait ttLab Chrom &ChromLab MarkerNum &MarkerLab Morgans NObs LR Additive2;
		%end;
		%else %do;
			%let cname = Trait ttLab Chrom MarkerNum &MarkerLab Morgans NObs LR Additive2;
		%end;
		%let cname = &cname Iteration;
	%end;
	%else %do;
		%if %length(&ChromLab)^=0 %then %do;
			%let cname = Trait ttLab Chrom &ChromLab MarkerNum &MarkerLab Morgans cM NObs LR LR2 LR3 Additive01 Dominance01 Dominance02 Additive03 ;
		%end;
		%else %do;
			%let cname = Trait ttLab Chrom MarkerNum &MarkerLab Morgans cM NObs LR LR2 LR3 Additive01 Dominance01 Dominance02 Additive03 ;
		%end;
		%let cname = &cname Iteration;
	%end;

	proc report data=&MappingResult_OutData nowd;
		column &cname;
		define Trait / display "Trait" width=5;
		define ttLab / display "t-Label" width = 15;
		define Chrom / display "Chrom" width=5;
		%if %length(&ChromLab)^=0 %then %do;	
			define &ChromLab / display "c-Label" width = 15;
		%end;
		define MarkerNum / display "Marker" width=6;
		%if %length(&ChromLab)^=0 %then %do;
			define &MarkerLab / display "m-Label" width = 15;
		%end;
		define Morgans / display "Morgans" width=7;
		%if &GKNumber ne two %then %do;
			define cM / display "cM" width=6;
		%end;
		define NObs / display "NObs" width=6;	
		%if &GKNumber = three %then %do;
			define LR2 / display "LR(a=0)";
			define LR3 / display "LR(d=0)";
			define Additive01 / display "Additive";
			define Additive03 / display "A(d=0)";
			define Dominance01 / display "Dominance";
			define Dominance02 / display "D(a=0)";
		%end;
	run;

	ods html close;
	*ods listing;

	* put JSL statements;
	/* Produce new data set for graph */
	data _tmp0;
		set &MappingResult_OutData.(drop=ttlab iteration);
		where trait = 1;
		%if &tChromNum > 1 %then %do;
			ChMkPos = put(chrom,z2.0)||"_"||put(markernum,z3.0);
		%end;
		%else %do;
			ChMkPos = put(markernum,z3.0);
		%end;
	run;
	/*proc print data = _tmp0; title "_tmp0";*/

	%let dstr = ttlab chrom;
	%if %length(&ChromLab)^=0 %then %do;
		%let dstr = &dstr &ChromLab;
	%end;
	%let dstr = &dstr markernum &MarkerLab;
	%if %length(&MarkerPos)^=0 %then %do;
		%let dstr = &dstr Morgans;
	%end;
	%let dstr = &dstr iteration;
	%*put dropstr = &dstr;
	%if &GKNumber = two %then %do;
		%let str = _tmp0;
		%let i = 1;
		%let var = %scan(&TraitVar,&i);
		%do %while(%length(&var));
			data _t0;
				set &MappingResult_OutData.(drop=&dstr);
				where trait = &i; 
				LOD&i = LR*0.217;
				Additive2&i = Additive2;
				R2&i = 1 - 10**(-2*LOD&i / NObs);
				NObs&i = Nobs;
				label LOD&i="&&tlab_lod&i" Additive2&i="&&tlab_a2&i" R2&i="&&tlab_r2&i" NObs&i="&&tlab_n&i";
				drop nobs;
			run;
			data _tmp&i;
				set _t0(drop=LR Additive2);
				trait = 1;
			run;
			
			/*proc print data = _tmp&i; title "_tmp&i"; run;*/
			
			%let str = &str _tmp&i;
			%let i = %eval(&i+1);
			%let var = %scan(&TraitVar,&i);
		%end;

		%let outsasdata = &MappingResult_OutData.1;
		data _tmp0;
			merge &str; by trait;
			output;
		run;
		data &outsasdata;
			set _tmp0 (drop=trait LR Additive2 Nobs);
		run;
	%end;
	%else %do;
		%let str = _tmp0;
		%let i = 1;
		%let var = %scan(&TraitVar,&i);
		%do %while(%length(&var));
			data _t0;
				set &MappingResult_OutData.(drop=&dstr);
				where trait = &i; 
				LOD&i = LR*0.217;
				Additive2&i = Additive01;
				Dominance&i = Dominance01;
				R2&i = 1 - 10**(-2*LOD&i / NObs);			
				NObs&i = Nobs;
				label LOD&i="&&tlab_lod&i" Additive2&i="&&tlab_a2&i" Dominance&i="&&tlab_d&i" R2&i="&&tlab_r2&i" NObs&i="&&tlab_n&i";
				drop nobs;
			run;
			data _tmp&i;
				set _t0(drop=LR Additive01 Dominance01);
				trait = 1;
			run;
			/*proc print data = _tmp&i; title "_tmp&i";*/
			%let str = &str _tmp&i;
			%let i = %eval(&i+1);
			%let var = %scan(&TraitVar,&i);
		%end;
		%let outsasdata = &MappingResult_OutData.1;
		data _tmp0;
			merge &str; by trait;
			output;
		run;
		data &outsasdata;
			set _tmp0 (drop=trait LR LR2 LR3 Additive01 Dominance01 Dominance02 Additive03 NObs);
		run;
	%end;	
%end;  /* close if condition for EM algorithm data output prep */

/*sbrlsi added marker indicator for the graphic results*/	
data &outsasdata;
	set &outsasdata;
	_sindex_ = _n_;
run;

proc sort data=&outsasdata;
	by &MarkerLab;
run;

data &outsasdata;
	set &outsasdata;
	by &MarkerLab;
	MarkerIndicator= 0;
	if first.&MarkerLab then MarkerIndicator= 1;
	label MarkerIndicator= "Marker Indicator";
run;

proc sort data=&outsasdata out=&outsasdata(drop= _sindex_);
	by _sindex_;
run;

proc means data=&outsasdata noprint;
	var MarkerIndicator %if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then QTLIndicator1; ;
	by &ChromNum;
	output out=nmbych sum=nm %if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then nQTL;;
run;

/*data outlib.qtl_nmbych; set nmbych; run;*/

data _null_;
	set nmbych;
	call symputx("mbych"||trim(left(_n_)),nm);
	%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
		call symputx("qtlbych"||trim(left(_n_)),nQTL);
	%end;
run;	
/*end adding marker indicator*/

**Creating y-axis limits for LOD and Effect plots**;
proc means data=&outsasdata max min noprint;
	var 
		%do t=1 %to &nTraitVars;
			LOD&t Additive2&t 
			%if &ngprobs >= 3 %then Dominance&t;
			%if &ngprobs = 4 %then Additive1&t;
		%end;
	;
	output out=LODmaxdata 
			max=
			%do t=1 %to &nTraitVars; 
				maxLODvar&t maxadd2Var&t 
				%if &ngprobs >= 3 %then maxDomVar&t; 
				%if &ngprobs = 4 %then maxadd1Var&t;
			%end;
			min=
			%do t=1 %to &nTraitVars; 
				minLODvar&t minadd2Var&t 
				%if &ngprobs >= 3 %then minDomVar&t;
				%if &ngprobs = 4 %then minadd1Var&t;
			%end;
			;
run;

data _null_;
	set LODMaxData;
	%do t=1 %to &nTraitVars; 
		call symputx("maxLOD&t", round(maxLODvar&t,.001));
		call symputx("minLOD&t", round(minLODvar&t,.001));
		call symputx("maxadd2_&t", abs(round(maxadd2Var&t,.001)));
		call symputx("minadd2_&t", abs(round(minadd2Var&t,.001)));  
		%if &ngprobs >= 3 %then %do;
			call symputx("maxdom&t", abs(round(maxDomVar&t,.001)));
			call symputx("mindom&t", abs(round(minDomVar&t,.001)));
		%end; 
		%if &ngprobs = 4 %then %do;
			call symputx("maxadd1_&t", abs(round(maxadd1Var&t,.001)));
			call symputx("minadd1_&t", abs(round(minadd1Var&t,.001)));
		%end;
	%end;
run;

%let maxLOD= -10000;
%if %length(&MIMEntryLODThreshMain) %then %do;
	%put maxLOD= &maxLOD MIMEntryLODThreshMain= &MIMEntryLODThreshMain;
	%let maxLOD = %sysfunc(max(&maxLOD, &MIMEntryLODThreshMain));
%end;
%else %if %length(&LODThresh) %then %do;
	%let maxLOD = %sysfunc(max(&maxLOD, &LODThresh));
%end;

%put maxLOD= &maxLOD;

%let minLOD= +10000;
%let maxEFF= -10000;
%do t=1 %to &nTraitVars; 
		%if %index(%Upcase(&MAlgorithm),HALEY) and &Perms %then %do;
			%let maxLOD = %sysfunc(max(&maxLOD, &&EWLODThresh&t));
		%end;

		%put maxLOD&t= &&maxLOD&t;
		%let maxLOD = %sysfunc(max(&maxLOD, &&maxLOD&t));
		%put minLOD&t= &&minLOD&t;
		%let minLOD = %sysfunc(min(&minLOD, &&minLOD&t));
		%put maxadd2_&t= &&maxadd2_&t;
		%put minadd2_&t= &&minadd2_&t;
		%let maxEFF = %sysfunc(max(&maxEFF, &&maxadd2_&t, &&minadd2_&t));
		%if &ngprobs >= 3 %then %do;
			%put maxdom&t= &&maxdom&t;
			%put mindom&t= &&mindom&t;
			%let maxEFF = %sysfunc(max(&maxEFF, &&maxdom&t, &&mindom&t));
		%end; 
		%if &ngprobs = 4 %then %do;
			%put maxadd1_&t= &&maxadd1_&t;
			%put minadd1_&t= &&minadd1_&t;
			%let maxEFF = %sysfunc(max(&maxEFF, &&maxadd1_&t, &&minadd1_&t));
		%end;
%end;
%let maxLOD= %sysevalf(&maxLOD + 0.5);
%put maxLOD= &maxLOD;
%put minLOD= &minLOD;
%put maxEFF= &maxEFF;

%if %length(&ChromLab) %then %CheckLabel(&outsasdata,ChromLab);
%CheckLabel(&outsasdata,ChromNum);
%CheckLabel(&outsasdata,MarkerLab);

%global _mainjslfile_ FullAPName;
%let _mainjslfile_ = %quote(&MappingGraph_Script);
%let FullAPName = %str(QTL IM, CIM and MIM Analysis);
/*********************************/
data _null_;
	file "&MappingGraph_Script" lrecl=32767;
	put '//!';
	put "Names Default to Here(1);";
	put "//Begin Parameters for MIM follow-up runs";
	put "Here:MacroPath = ""&MacroPath"";";
	put "Here:OutPath= ""&OutPath"";";
	put "Here:ProcessName= ""&ProcessName"";";
	put "Here:ClientOutPath = ""&ClientOutPath""; ";
	put "//End parameters for MIM follow-up runs";
	put "Here:DataList = {};";
	put "Here:TabButtonList = {};";
	put "Here:TabCodeList = {};";
	put "Here:TabAfterCodeList = {};";
	put "Here:TabOpenList = {};";	
	put 'wait(0.5);';
	put "Here:CofactorsList = {"@;
	%let i = 1;
	%let fvar = %scan(&FactorVarsNew,&i);
	%do %while(%length(&fvar));
		%if &i = 1 %then %do;
			put " ""&fvar"" "@;
		%end;
		%else %do;
			put ", ""&fvar"" "@;
		%end;
		%let i = %eval(&i+1);
		%let fvar = %scan(&FactorVarsNew,&i);
	%end;
	put "};"@;	
	put ;
	put "Here:maxLOD= &maxLOD;";
	put "Here:ManageQTLFlag= 0;";
	%if %index(%Upcase(&MAlgorithm),HALEY) %then %do;
		%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
			put "Here:maxEff= &maxEff; "; /*sbrlsi*/
		%end;
	%end;		
	put "Here:NumChrom = &tChromNum;";	
	put "Here:Nmodels = 0; ";
	put "Here:Ntables = 0; ";
	
	%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
		/*hard coding the maximum number of models to build within a single MIM framework*/
		%let maxnmodels= 50;
	%end;
	%else %do;
		%let maxnmodels= 1;
	%end;
		
	%macro QTLFnMacro(FunctionName=, VLBName=, OLBName=,);
		put "&FunctionName = Function({},{default local}," ;
		put "If(IsEmpty(Here:qtldata0),";
		%if %index(%Upcase(&MAlgorithm),HALEY) %then %do;
			%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
				put "  Here:qtldata0=open(""&ClientOutpath.&MappingResult_OutDataName.0.sas7bdat"", invisible);";
			%end;
			%else %do;
				put "  Here:qtldata0=open(""&ClientOutpath.&MappingResult_OutDataName..sas7bdat"", invisible);";
			%end;
		%end;
		%else %do;
			/* EM data format */
			put "  Here:qtldata0=open(""&ClientOutpath.&MappingResult_OutDataName.1.sas7bdat"", invisible);";
		%end;
		put "  column(""&ChromNum"")<<Set Modeling Type(Ordinal);";
		put "  column(""&MarkerLab"")<<Label(1);";
		put "  column(""Morgans"")<<data type(numeric);";
		%if %length(&ChromLab) %then %do;
			put "  column(""&ChromLab"")<<data type(character);";
		%end;
		put ","; /*else for Is empty check*/
		put "	 Current Data Table(Here:qtldata0);";
		put ");";
		/*sbrlsi getting marker indicator*/
		%if %index(%upcase(&DisplayMarkers), YES) %then %do;
			put "	mi = :Name(""Marker Indicator"") << Get Values; ";
			put "	nm = sum(mi); ";
			put "	Here:ym = J(nm,1,-.1);";
			put "	Here:xm = J(nm,1,0);";
			put "	Here:nr= N Row(mi);";
			put "	i= 1;";
			put "	for (j =1, j <= Here:nr, j++,";
			put "		if (mi[j]==1,";
			put "			Here:xm[i]= :Name(""&GDMetric"")[j];";
			put "			i= i + 1;";
			put "		)";
			put "	);";		
		%end;
		/*sbrlsi getting QTL indicator*/
		%if %index(%upcase(&DisplayQTL), YES) %then %do;
			put "	Here:nqtl = 0; ";		
		%end;
		/*end getting QTL indicator*/	

		%if ^(%index(%Upcase(&MAlgorithm),HALEY) and &Perms) %then %do;
			%do i=1 %to &tChromNum %by 1;	
				put "	Here:opname&i= ""Here:op""||char(Here:Nmodels)||char(""_"")||char(&i);";
				put "	Here:ropname&i= ""Here:rop""||char(Here:Nmodels)||char(""_"")||char(&i);";
				put "	Here:opename&i= ""Here:ope""||char(Here:Nmodels)||char(""_"")||char(&i);";
				put "	Here:ropename&i= ""Here:rope""||char(Here:Nmodels)||char(""_"")||char(&i);";			
			%end;
		%end;
		
		%if &FunctionName = Here:QTLFn %then %do;
			put "	 Here:qt_uf0="""";";
			put '	 qt_uf0hv=EvalInsert("\[';
			put "	     Here:qt_hfsize0=400;";
			put "	     Here:qt_vfsize0=250;";
			put '	 ]\");';
			put "	 Eval(Parse(qt_uf0hv));";   

			%if %index(%Upcase(&MAlgorithm),HALEY) and &Perms %then %do;
				%do j=1 %to &tTraitNum;
					put '	   Here:qt_uf0||=EvalInsert("\[';
					put "	     for(k=1,k<=NItems(Here:rop&j),k++,";
					put "	      try(";
					put "	        Here:rop&j[k][framebox(1)] << FrameSize(qt_hfsize0,qt_vfsize0);";
					put "	   ));";
					put '	   ]\");';
				%end;
			%end;
			%else %do;
				%do i=1 %to &tChromNum %by 1;
					put '	   Here:qt_uf0||=EvalInsert("\[';			
					%do jj=0 %to %sysevalf(&maxnmodels-1);
						put "	     try(";    
						put "	       	for(k=1,k<=NItems(Here:rop&jj._&i),k++,";     
						put "	          Here:rop&jj._&i[k][framebox(1)] << FrameSize(qt_hfsize0,qt_vfsize0);";
						put "	     	);";
						put "	     );";
					%end;
					put '	   ]\");';
	
					put '	   Here:qt_uf0||=EvalInsert("\[';		
					%do jj=0 %to %sysevalf(&maxnmodels-1);
						put "        try(";
						put "       	for(k=1,k<=NItems(Here:rope&jj._&i),k++,";
						put "          		Here:rope&jj._&i[k][framebox(1)] << FrameSize(qt_hfsize0,qt_vfsize0);";
						put "	     	);";
						put "	     );";
					%end;
					put '	]\");';
				%end;
				/*			
				   put "       for(u=1, u<= &tChromNum, u++,";
				   put "       for(j=0, j<= Here:Nmodels, j++,";
				   put '	Here:qt_uf0||=EvalInsert("\[';
				   put "          		//show(Here:rop^j^_^u^);";
				   put "          		for(k=1,k<=NItems(Here:rop^j^_^u^),k++,";
				   put "          		    try(";
				   put "          		    	Here:rop^j^_^u^[k][framebox(1)] << FrameSize(qtlast_hfsize0,qtlast_vfsize0);";
				   put "          		    );";
				   put "          		);";

				   put "          		for(k=1,k<=NItems(Here:rope^j^_^u^),k++,";
				   put "          		    try(";
				   put "          		    	Here:rope^j^_^u^[k][framebox(1)] << FrameSize(qtlast_hfsize0,qtlast_vfsize0);";
				   put "          		    );";
				   put "          		);";
				   put '  	]\");';
				  put "        );";
				  put "        );";
				  */
			%end;

			put '	 upf_expr= EvalInsert("\[';
			put "	   //show(""Debugging"");";
			put "	   Here:qt_update_Frames0 = Expr(Eval(Parse(Here:qt_uf0)));";
			put '	 ]\");';	
			put "	   Eval(Parse(upf_expr));";

			put '	   qt_uf0_2 = EvalInsert("\[';
			put '	   qt_update_HFS_TextEditBox0 = Expr(';
			put '	      value = Round(Num(qt_hfsize0),0);';
			put '	      qt_HFS_TEB0 << SetText(char(num(value)));';
			put '	      qt_HFS_TEB0 << Reshow;';
			put '	      qt_hfsize0 = value;';
			put '	      qt_update_Frames0;';
			put '	   );';
			put '	   qt_update_HFS_SliderBox0 = Expr(';
			put '	      qt_hfsize0 = num(qt_HFS_TEB0<<GetText);';
			put '	      qt_HFS_SB0 << Reshow;';
			put '	      qt_update_HFS_TextEditBox0;';
			put '	   );';
			put '	   qt_update_VFS_TextEditBox0 = Expr(';
			put '	      value = Round(Num(qt_vfsize0),0);';
			put '	      qt_VFS_TEB0 << SetText(char(num(value)));';
			put '	      qt_VFS_TEB0 << Reshow;';
			put '	      qt_vfsize0 = value;';
			put '	      qt_update_Frames0;';
			put '	   );';
			put '	   qt_update_VFS_SliderBox0 = Expr(';
			put '	      qt_vfsize0 = num(qt_VFS_TEB0<<GetText);';
			put '	      qt_VFS_SB0 << Reshow;';
			put '	      qt_update_VFS_TextEditBox0;';
			put '	   );';
			put '	   ]\");';
			put '	   Eval(Parse(qt_uf0_2));';
			put '	   qt_vfsteb0 = EvalInsert("\[';
			put '	      qt_VFS_TEB0 = TextEditBox(char(num(qt_vfsize0)),';
			put '	         <<Script(qt_update_VFS_SliderBox0)';
			put '	      )';
			put '	   ]\");';
			put '	   qt_vfssb0 = EvalInsert("\[';
			put "	      qt_VFS_SB0 = SliderBox(50,2000,qt_vfsize0,";
			put '	         qt_update_VFS_TextEditBox0';
			put '	      )';
			put '	   ]\");';
			put '	   qt_vfssb0_2 = EvalInsert("\[';
			put '	      qt_VFS_SB0 << SetValue(num(qt_vfsize0));';
			put '	      qt_VFS_SB0 << SetWidth(120);';
			put '	   ]\");';
			put '	   qt_hfsteb0 = EvalInsert("\[';
			put '	      qt_HFS_TEB0 = TextEditBox(char(num(qt_hfsize0)),';
			put '	         <<Script(qt_update_HFS_SliderBox0)';
			put '	      )';
			put '	   ]\");';
			put '	   qt_hfssb0 = EvalInsert("\[';
			put '	      qt_HFS_SB0 = SliderBox(50,2000,qt_hfsize0,';
			put '	         qt_update_HFS_TextEditBox0';
			put '	      )';
			put '	   ]\");';
			put '	   qt_hfssb0_2 = EvalInsert("\[';
			put '	      qt_HFS_SB0 << SetValue(num(qt_hfsize0));';
			put '	      qt_HFS_SB0 << SetWidth(120);';
			put '	   ]\");	';
		%end;

		/*put	'ScrollBox(size(8000,1200),';*/	
		%if %index(%Upcase(&MAlgorithm),HALEY) and &Perms %then %do;
			put "VListBox(";
			%if &FunctionName = Here:QTLFn %then %do;
				put	'   HListBox(';
				put	'      Text Box( " Vertical Frame Size " ),';
				put	'      HListBox(';
				put	'         eval(parse(qt_vfsteb0)),';
				put	'         eval(parse(qt_vfssb0)),';
				put	'         eval(parse(qt_vfssb0_2))';
				put	'      ),';
				put	'      Text Box( " Horizontal Frame Size " ),';
				put	'      HListBox(';
				put	'         eval(parse(qt_hfsteb0)),';
				put	'         eval(parse(qt_hfssb0)),';
				put	'         eval(parse(qt_hfssb0_2))';
				put	'      )';
				put	'   ), ';
			%end;
			put "	  TextBox(""     The horizontal red dashed reference line represents the empirical experiment-wise LOD threshold at Alpha = &Alpha."", <<Set Wrap(800),<<Set Font Size(9)),"; 
			put "	  TextBox(""     The grey dotted line represents the empirical comparison-wise LOD threshold at Alpha = &Alpha."", <<Set Wrap(800),<<Set Font Size(9)),"; 
			%do j=1 %to &tTraitNum;
				%let CurrT = %scan(&TraitVar,&j);
				put "Here:o=OutlineBox(""&MethodSelection Results for &CurrT"",";
				put "TextBox(""    Experiment-wise LOD threshold line drawn at &&EWLODThresh&j."", <<Set Wrap(800)),";
				put "HListBox(";
				put "Here:op&j=Graph Builder(";
				put "	Size(500,300),";
				put "	Show Control Panel( 0 ),";
				put "	Legend Position( ""Bottom"" ),";
				put "	Show Title( 0 ),";
				put "	Variables(";
				put "		X(:Name(""&GDMetric"")),";
				put "		Y(:Name(""&&tlab_lod&j"")),";
				put "		Y(:Name(""CW LOD Threshold &CurrT""), Position(1))";
				put "	),";
				put "	Elements(";
				put "		Points( X, Y( 1 ), ";
				put "		Y( 2 ), ";
				put "		Legend( 1 ) ),";
				put "		Line( X, Y( 1 ), ";
				put "		Y( 2 ), ";
				put "		Legend( 3 ) ),";
				put "	),";
				%if %length(&ChromLab) %then %do;
					put "	By( :Name(""&ChromLab"")),";
				%end;
				%else %do;
					put "	By( :Name(""&ChromNum"")),";
				%end;
				put "	SendToReport(";
				put "		Dispatch({}, ""Y title"", TextEditBox, {Set Text( ""LOD"" )} ),";
				put "		Dispatch({},""&&tlab_lod&j"",ScaleBox,";
				put "			{Min( -0.20 ), Max( Here:maxLOD ), Inc(2), Minor Ticks( 0 )";
				/*put "			,Add Ref Line(&&EWLODThresh&j,Dashed, ""Red"")";*/
				put "			}),";
				put "		Dispatch(";
				put "			{},";
				put "			""400"",";
				put "			ScaleBox,";
				put "			{Legend Model(";
				put "				3,";
				put "				Properties(";
				put "					1,";
				put "					{Line Color( 1 ), Line Style( ""Dotted"" ), Fill Color( 0 )},";
				put "					Item ID( ""CW LOD Threshold &CurrT"", 1 )";
				put "				)";
				put "			)}";
				put "		),";
				/*
				put "		Dispatch(";
				put "			{},";
				put "			""Graph Builder"",";
				put "			OutlineBox,";
				put "			{Set Title( ""LOD Plot"" ), Image Export Display( Normal )}";
				put "		),";
				*/				
				/*put "    	Dispatch( {}, ""Graph Builder"", FrameBox, {Frame Size( 500, 300 )} )";*/
				put "	),";
				put "),";
				put ")),";
				put "if(type(Here:op&j)==""Scriptable"" | type(Here:op&j)==""List"",";
				put "Here:rop&j = op&j << report;";
				/*put "try(o[TextBox(1)] << Set Text("""") << Set Font Size(1));";*/		
				%let bem=1;
				%let enm=0;
				%if &tChromNum > 1 %then %do;
					%do i=1 %to &tChromNum %by 1;
						%let bem= %sysevalf(&enm + 1);
						%let enm= %sysevalf(&bem + &&&mbych&i - 1);
						put "try(";
						/*sbrlsi including marker indicator*/
						%if %index(%upcase(&DisplayMarkers), YES) %then %do;
							put "	rop&j[&i][framebox(1)] << add graphics script(Marker(16,ColorState(""black""),Here:xm[&bem::&enm],Here:ym[&bem::&enm]));";
						%end;
						/*end including marker indicator*/
						put "   rop&j[&i] << Dispatch( {}, ""graph title"", TextEditBox, {Set Text( """" )} );";
		      				put "   //show(rop&j[&i][OutLineBox(1)]);";
						put "   gtitle= rop&j[&i][OutLineBox(1)]<< Get Text;";
      						put "   //show(gtitle);";
      						put "   gtitle = Trim(Munger(Trim(gtitle), 1, ""Graph Builder"", """"));";
      						put "   //show(gtitle);";
						put "   pos= Munger(gtitle, 1, ""Where("");";
						put "   //show(pos);";
						put "   if(pos, ";
      						put "        	del= Trim(Substr(gtitle, pos - 6));";
      						put "        	rop&j[&i][OutLineBox(1)] << Dispatch( {}, del, TextBox, {Hide( 1 )} );"; 					
						put "    	gtitle= Trim(Substr(gtitle, 1, pos - 6));";						
						put "   );";      			
      						put "   rop&j[&i][OutLineBox(1)] << {Set Title(""LOD Plot for ""||gtitle), Image Export Display( Normal )};";					
						put "	rop&j[&i][framebox(1)] << add graphics script(pencolor(""red"");linestyle(""dashed"");hline(&&EWLODThresh&j));";					
						put "	rop&j[&i][framebox(1)] << MarkerSize(1);";
						put "	rop&j[&i][framebox(1)] << MarkerDrawingMode(Normal);";		
						put "	rop&j[&i][Axisbox(2)] << RemoveAxisLabel;";
						put "	rop&j[&i][Axisbox(2)] << Min(-0.5);";
						put "	rop&j[&i][Axisbox(2)] << Max(Here:maxLOD);";
						put "	rop&j[&i][Axisbox(2)] << Inc(2);";
						put "	rop&j[&i][Axisbox(2)] << Minor Ticks(0);";
						put "	//^rop&j[&i][Axisbox(2)] << AddAxisLabel(""LOD"");";
						%if %Upcase(&GDMetric) = CM %then %do;
							put "	rop&j[&i][Axisbox(1)] << Min(-.5);";
						%end;						
						put ");";
					%end;
				%end;
				%else %do;
					%let i=1;
					%let bem= %sysevalf(&enm + 1);
					%let enm= %sysevalf(&bem + &&&mbych&i - 1);
					put "try(";
					/*sbrlsi including marker indicator*/
					%if %index(%upcase(&DisplayMarkers), YES) %then %do;
						put "	rop&j[framebox(1)] << add graphics script(Marker(16,ColorState(""black""),Here:xm[&bem::&enm],Here:ym[&bem::&enm]));";
					%end;
					/*end including marker indicator*/
					put "   rop&j << Dispatch( {}, ""graph title"", TextEditBox, {Set Text( """" )} );";
		      			put "   //show(rop&j[OutLineBox(1)]);";
					put "   gtitle= rop&j[OutLineBox(1)]<< Get Text;";
      					put "   //show(gtitle);";
      					put "   gtitle = Trim(Munger(Trim(gtitle), 1, ""Graph Builder"", """"));";
      					put "   //show(gtitle);";
					put "   pos= Munger(gtitle, 1, ""Where("");";
					put "   //show(pos);";
					put "   if(pos, ";
      					put "        	del= Trim(Substr(gtitle, pos - 6));";
      					put "        	rop&j[OutLineBox(1)] << Dispatch( {}, del, TextBox, {Hide( 1 )} );"; 					
					put "    	gtitle= Trim(Substr(gtitle, 1, pos - 6));";						
					put "   );";      			
      					put "   rop&j[OutLineBox(1)] << {Set Title(""LOD Plot for ""||gtitle), Image Export Display( Normal )};";								
					put "	rop&j[framebox(1)] << add graphics script(pencolor(""red"");linestyle(""dashed"");hline(&&EWLODThresh&j));";					
					put "	rop&j[framebox(1)] << MarkerSize(1);";
					put "	rop&j[framebox(1)] << MarkerDrawingMode(Normal);";		
					put "	rop&j[Axisbox(2)] << RemoveAxisLabel;";
					put "	rop&j[Axisbox(2)] << Min(-0.5);";
					put "	rop&j[Axisbox(2)] << Max(Here:maxLOD);";
					put "	rop&j[Axisbox(2)] << Inc(2);";
					put "	rop&j[Axisbox(2)] << Minor Ticks(0);";
					put "	//^rop&j[Axisbox(2)] << AddAxisLabel(""LOD"");";
					%if %Upcase(&GDMetric) = CM %then %do;
						put "	rop&j[Axisbox(1)] << Min(-.5);";
					%end;						
					put ");";				
				
				%end;
				put ");";
			%end;
			put ");";
		%end;
		%else %do;
			put "  &VLBName = VListBox(";

			%if &FunctionName = Here:QTLFn %then %do;		
				put	'   HListBox(';
				put	'      Text Box( " Vertical Frame Size " ),';
				put	'      HListBox(';
				put	'         eval(parse(qt_vfsteb0)),';
				put	'         eval(parse(qt_vfssb0)),';
				put	'         eval(parse(qt_vfssb0_2))';
				put	'      ),';
				put	'      Text Box( " Horizontal Frame Size " ),';
				put	'      HListBox(';
				put	'         eval(parse(qt_hfsteb0)),';
				put	'         eval(parse(qt_hfssb0)),';
				put	'         eval(parse(qt_hfssb0_2))';
				put	'      )';
				put	'   ), ';		
			%end;

			%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;	
				put "  &OLBName = outlineBox(""Interval Mapping Model"", ";
			%end;

			%macro qtlplots(trackcall=);	
				
				put "Here:chrom = :Name(""&ChromNum"") << Get Values;";
				%if %index(%upcase(&DisplayQTL), YES) %then %do;
					put "Here:qtli = :Name(""QTL Position Indicator &TraitVar"") << Get Values;";
					put "Here:nr= N Row(Here:qtli);";
				%end;

				%do i=1 %to &tChromNum %by 1;
					put "if(!Here:ManageQTLFlag,"; 
					put "   //for code that does not run in the Manage QTL Model";
					put "	Here:opname&i= ""Here:op""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put "	Here:ropname&i= ""Here:rop""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put "	,";
					put "   //for code that does run in the Manage QTL Model";
					put "	Here:mopname&i= ""Here:mop""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put "	Here:rmopname&i= ""Here:rmop""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put ");";						

					%if %index(%upcase(&DisplayQTL), YES) %then %do;
						put "Here:xqtl&i=.;";
						put "Here:yqtl&i=.;";
						put "Here:nqtl&i= 0;";
						put "k=1;";
						put "for(j=1 , j <= Here:nr, j++,";
						put "	if (Here:chrom[j]==&&ChrVars&i & Here:qtli[j]==1,";
						put "		if(k==1, ";
						put "			Here:xqtl&i= :Name(""&GDMetric"")[j];";
						put "			Here:yqtl&i= 0.5;";
						put "		,";
						put "			Here:xqtl&i= Here:xqtl&i || :Name(""&GDMetric"")[j];";
						put "			Here:yqtl&i= Here:yqtl&i || 0.5;";			
						put "		);";
						put "		Here:nqtl&i= Here:nqtl&i + 1;";
						put "		k= k + 1;";
						put "	);";
						put ");	";
					%end;
				%end;
				
				put "if(!Here:ManageQTLFlag,"; 
				put "  Here:olbq = outlineBox(""LOD Profile"",";
				put "  Here:hlbq = hlistbox(";
				%let bem=1;
				%let enm=0;
				%do i=1 %to &tChromNum %by 1;
					put "Here:o=OutlineBox("@;
					%if %length(&ChromLab) %then %do;
						put " ""&ChromLab &&ChrLabel&i"",";
					%end;
					%else %do;
						put " ""&ChromNum &&ChrVars&i"",";
					%end;
					put "Vlistbox(";
					put 'Eval(Parse(EvalInsert("\[';
					put "//show(""&trackcall"");";
					put "//show(Here:opname&i);";
					put "^Here:opname&i^ = Graph Builder(";
					put "	Size(500,300),";
					put "	Show Control Panel( 0 ),";
					put "	Legend Position( ""Bottom"" ),";
					put "	Show Title( 0 ),";
					put "	Variables(";
					put "		X(:&GDMetric),";
					put "		Y(:(&tlab_lod1)),";
					%do j=2 %to &tTraitNum %by 1;
						put "		 Y(:(&&tlab_lod&j), Position(1)),";
					%end;
					put "	),";
					put "	Elements(";
					put "		Points( X, Y( 1 ), ";
					%do j=2 %to &tTraitNum %by 1;
						put "		Y( &j ), ";
					%end;
					put "		Legend( 1 ) ),";
					put "		Line( X, Y( 1 ), ";
					%do j=2 %to &tTraitNum %by 1;
						put "		Y( &j ), ";
					%end;
					put "		Legend( 3 ) ),";
					put "	),";
					%if &tChromNum > 1 %then %do;
						%if %length(&ChromLab) %then %do;
							put ", Where(uppercase(:&ChromLab)==uppercase(""&&ChrLabel&i"")),";
						%end;
						%else %do;
							put ", Where(:Name(""&ChromNum"")==&&ChrVars&i),";
						%end;
					%end;
					put "	SendToReport(";
					put "		Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""LOD"" )} ),";
					%if &tTraitNum < 2 %then %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Legend Position( {1, [-1], 3, [0]} ), Position( {-1, 0} )}";
						put "		),";				
					%end;	
					%else %do;
						put "		Dispatch( {}, ""400"", LegendBox, {Set Wrap( 2 )} ),";
					%end;
					put "		Dispatch(";
					put "			{},";
					put "			""Graph Builder"",";
					put "			OutlineBox,";
					put "			{Set Title( ""LOD Plot"" ), Image Export Display( Normal )}";
					put "		),";					
					/*put "    	Dispatch( {}, ""Graph Builder"", FrameBox, {Frame Size( 500, 300 )} ),";*/
					put "	),";
					put ");";
					put "if(type(^Here:opname&i^)==""Scriptable"" | type(^Here:opname&i^)==""List"",";
					put "^Here:ropname&i^ = ^Here:opname&i^ << report;";
					%if &tChromNum > 1 %then %do;
						put 'try(o[TextBox(1)] << Set Text("") << Set Font Size(1));';
					%end;	
					put "for(i=1, i<=1, i++,";
					put "try(";
					%if &LODThresh > 0 %then %do;
						put "^Here:ropname&i^[i][framebox(1)] << add graphics script(pencolor(""red"");linestyle(""dashed"");hline(&LODThresh));";
					%end;
					/*sbrlsi including marker indicator*/
					%if %index(%upcase(&DisplayMarkers), YES) %then %do;
						%let bem= %sysevalf(&enm + 1);
						%let enm= %sysevalf(&bem + &&&mbych&i - 1);
						put "^Here:ropname&i^[i][framebox(1)] << add graphics script(";
						put "			Marker(16,";
						put "				ColorState(""black""),";
						put "				Here:xm[&bem::&enm],";
						put "				Here:ym[&bem::&enm] ";
						put "			)";
						put "		);";
					%end;
					/*end including marker indicator*/
					/*sbrlsi including qtl indicator*/
					%if %index(%upcase(&DisplayQTL), YES) %then %do;
						put "if(Here:nqtl&i > 0, ";
						put "		^Here:ropname&i^[i][framebox(1)] << add graphics script(";
						put "			Marker(21,";
						put "				ColorState(""red""),";
						put "				[^Here:xqtl&i^],";
						put "				[^Here:yqtl&i^] ";					
						put "				)";
						put "		)";
						put ");";
					%end;
					/*end including qtl indicator*/
					put "^Here:ropname&i^[i][framebox(1)] << MarkerSize(1);";
					put "^Here:ropname&i^[i][framebox(1)] << MarkerDrawingMode(Normal);";		
					put "^Here:ropname&i^[i][Axisbox(2)] << RemoveAxisLabel;";
					put "^Here:ropname&i^[i][Axisbox(2)] << Min(-0.5);";
					put "^Here:ropname&i^[i][Axisbox(2)] << Max(Here:maxLOD);";
					put "^Here:ropname&i^[i][Axisbox(2)] << Inc(2);";
					put "^Here:ropname&i^[i][Axisbox(2)] << Minor Ticks(0);";
					put "//^Here:ropname&i^[i][Axisbox(2)] << AddAxisLabel(""LOD"");";
					%if %Upcase(&GDMetric) = CM %then %do;
						put "^Here:ropname&i^[i][Axisbox(1)] << Min(-.5);";
					%end;
					/*put "^Here:ropname&i^[i][framebox(1)] << FrameSize(275,175);";*/
					put ")));";
					put '	]\")));';
					put ")),";
					put "try(o[textBox(1)] << Delete);";	
					put " ";
				%end;
				put "	);";	
				put "	);";
				put "	,";
				
				put "  Here:olbq = outlineBox(""LOD Profile"",";
				put "  Here:hlbq = hlistbox(";
				%let bem=1;
				%let enm=0;
				%do i=1 %to &tChromNum %by 1;
					put "Here:o=OutlineBox("@;
					%if %length(&ChromLab) %then %do;
						put " ""&ChromLab &&ChrLabel&i"",";
					%end;
					%else %do;
						put " ""&ChromNum &&ChrVars&i"",";
					%end;
					put "Vlistbox(";
					put 'Eval(Parse(EvalInsert("\[';
					put "//show(""&trackcall"");";
					put "//show(Here:mopname&i);";
					put "^Here:mopname&i^ = Graph Builder(";
					put "	Size(500,300),";
					put "	Show Control Panel( 0 ),";
					put "	Legend Position( ""Bottom"" ),";
					put "	Show Title( 0 ),";
					put "	Variables(";
					put "		X(:&GDMetric),";
					put "		Y(:(&tlab_lod1)),";
					%do j=2 %to &tTraitNum %by 1;
						put "		 Y(:(&&tlab_lod&j), Position(1)),";
					%end;
					put "	),";
					put "	Elements(";
					put "		Points( X, Y( 1 ), ";
					%do j=2 %to &tTraitNum %by 1;
						put "		Y( &j ), ";
					%end;
					put "		Legend( 1 ) ),";
					put "		Line( X, Y( 1 ), ";
					%do j=2 %to &tTraitNum %by 1;
						put "		Y( &j ), ";
					%end;
					put "		Legend( 3 ) ),";
					put "	),";
					%if &tChromNum > 1 %then %do;
						%if %length(&ChromLab) %then %do;
							put ", Where(uppercase(:&ChromLab)==uppercase(""&&ChrLabel&i"")),";
						%end;
						%else %do;
							put ", Where(:Name(""&ChromNum"")==&&ChrVars&i),";
						%end;
					%end;
					put "	SendToReport(";
					put "		Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""LOD"" )} ),";
					%if &tTraitNum < 2 %then %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Legend Position( {1, [-1], 3, [0]} ), Position( {-1, 0} )}";
						put "		),";				
					%end;	
					%else %do;
						put "		Dispatch( {}, ""400"", LegendBox, {Set Wrap( 2 )} ),";
					%end;
					put "		Dispatch(";
					put "			{},";
					put "			""Graph Builder"",";
					put "			OutlineBox,";
					put "			{Set Title( ""LOD Plot"" ), Image Export Display( Normal )}";
					put "		),";					
					/*put "    	Dispatch( {}, ""Graph Builder"", FrameBox, {Frame Size( 500, 300 )} ),";*/
					put "	),";
					put ");";
					put "if(type(^Here:mopname&i^)==""Scriptable"" | type(^Here:mopname&i^)==""List"",";
					put "^Here:rmopname&i^ = ^Here:mopname&i^ << report;";
					%if &tChromNum > 1 %then %do;
						put 'try(o[TextBox(1)] << Set Text("") << Set Font Size(1));';
					%end;	
					put "for(i=1, i<=1, i++,";
					put "try(";
					%if &LODThresh > 0 %then %do;
						put "^Here:rmopname&i^[i][framebox(1)] << add graphics script(pencolor(""red"");linestyle(""dashed"");hline(&LODThresh));";
					%end;
					/*sbrlsi including marker indicator*/
					%if %index(%upcase(&DisplayMarkers), YES) %then %do;
						%let bem= %sysevalf(&enm + 1);
						%let enm= %sysevalf(&bem + &&&mbych&i - 1);
						put "^Here:rmopname&i^[i][framebox(1)] << add graphics script(";
						put "		Marker(16,";
						put "		ColorState(""black""),";
						put "		Here:xm[&bem::&enm],";
						put "		Here:ym[&bem::&enm] ";
						put "		)";
						put ");";
					%end;
					/*end including marker indicator*/
					/*sbrlsi including qtl indicator*/
					%if %index(%upcase(&DisplayQTL), YES) %then %do;
						put "if(Here:nqtl&i > 0, ";
						put "		^Here:rmopname&i^[i][framebox(1)] << add graphics script(";
						put "		Marker(21,";
						put "		ColorState(""red""),";
						put "		[^Here:xqtl&i^],";
						put "		[^Here:yqtl&i^] ";					
						put "		)";
						put "	)";
						put ");";
					%end;
					/*end including qtl indicator*/
					put "^Here:rmopname&i^[i][framebox(1)] << MarkerSize(1);";
					put "^Here:rmopname&i^[i][framebox(1)] << MarkerDrawingMode(Normal);";		
					put "^Here:rmopname&i^[i][Axisbox(2)] << RemoveAxisLabel;";
					put "^Here:rmopname&i^[i][Axisbox(2)] << Min(-0.5);";
					put "^Here:rmopname&i^[i][Axisbox(2)] << Max(Here:maxLOD);";
					put "^Here:rmopname&i^[i][Axisbox(2)] << Inc(2);";
					put "^Here:rmopname&i^[i][Axisbox(2)] << Minor Ticks(0);";
					put "//^Here:rmopname&i^[i][Axisbox(2)] << AddAxisLabel(""LOD"");";
					%if %Upcase(&GDMetric) = CM %then %do;
						put "^Here:rmopname&i^[i][Axisbox(1)] << Min(-.5);";
					%end;
					/*put "^Here:rmopname&i^[i][framebox(1)] << FrameSize(275,175);";*/
					put ")));";
					put '	]\")));';
					put ")),";
					put "try(o[textBox(1)] << Delete);";	
					put " ";
				%end;
				put "	);";	
				put "	);";
				put ");";
			%mend qtlplots; 
			%qtlplots(trackcall= &FunctionName);

			%if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;	
				put "	);";					
			%end;
			put ");";
		%end;
		/*put ");";*/
		put ");";
	%mend QTLFnMacro;
	%QTLFnMacro(FunctionName= Here:QTLFn, VLBName= Here:vlb, OLBName= Here:olb0);

	put "InsertInto(DataList,{{""Here:qtldata0""}});";
	put "InsertInto(TabButtonList, ""Results"");";
	put "InsertInto(TabCodeList,expr(Here:QTLFn()));";
	
	%if %upcase(&MethodSelection) ^= %upcase(Multiple Interval Mapping) %then %do;	
	
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
	
		put "Here:VFn = Function({},{default local}," ;
		put "If(IsEmpty(Here:qtldata0),";
		%if %index(%Upcase(&MAlgorithm),HALEY) %then %do;
			put "	Here:qtldata0=open(""&ClientOutpath.&MappingResult_OutDataName..sas7bdat"", invisible);";
		%end;
		%else %do;
		/* EM data format */
			put "  Here:qtldata0=open(""&ClientOutpath.&MappingResult_OutDataName.1.sas7bdat"", invisible);";
		%end;
		put "  column(""&ChromNum"")<<Set Modeling Type(Ordinal);";
		put "  column(""Morgans"")<<data type(numeric);";
		%if %length(&ChromLab) %then %do;
			put "  column(""&ChromLab"")<<data type(character);";
		%end;
		put ","; /*else for Is empty check*/
		put "	 Current Data Table(Here:qtldata0);";
		put ");";
		put "HListBox(";
		put "VListBox(";
		%do j=1 %to &tTraitNum;
			%let CurrT = %scan(&TraitVar,&j);			
			%if &GKNumber = three or &GKNumber = four %then %do;
				put "HListBox(";
			%end;
			put "Here:vp=Graph Builder(Size( 500, 300 ),Show Control Panel( 0 ), Auto Stretching(0), Lock Scales( 1 ), ";
			%if &GKNumber ^= four %then %do;
				put "	Variables(X(:Name(""Additive &CurrT"")), Y(:Name(""LOD &CurrT"")),";
			%end;
			%else %do;
				put "	Variables(X(:Name(""Additive2 &CurrT"")), Y(:Name(""LOD &CurrT"")),";
			%end;
			%if %length(&ChromLab) %then %do;
				put "		Color(:Name(""&ChromLab""))";
			%end;
			%else %do;
				put "		Color(:Name(""&ChromNum""))";
			%end;
			put "	),";  /*close variables*/
			put " Elements(Points(X,Y,Color, Legend(1))),";
			put "	SendToReport(";
			put "		Dispatch({},""Graph Builder"", OutlineBox, {Set Title(""Estimate Plot for &CurrT"")}),";
			%if &GKNumber ^= four %then %do;
				put "		Dispatch({},""graph title"", TextEditBox,{Set Text(""Additive Effect vs. LOD Score"")}),";
			%end;
			%else %do;
				put "		Dispatch({},""graph title"", TextEditBox,{Set Text(""Additive2 Effect vs. LOD Score"")}),";
			%end;
			put "		Dispatch({},""X title"", TextEditBox,{Set Text(""Estimate of P2 Allele Additive Effect"")}),";
			put "		Dispatch({},""Graph Builder"", FrameBox, {Marker Size(2)}),";
			put "		Dispatch({},""LOD &CurrT"",ScaleBox,";
			put "	    {Format( ""Fixed Dec"", 9, 2 ),Min( -0.5), Max( Here:maxLOD ), Inc(2), Minor Ticks( 0 ),";
			%if &perms %then %do;
				put "			Add Ref Line( &&EWLODThresh&j,Dashed, ""Red"" ),";
			%end;
			%else %do;
				put "			Add Ref Line( &LODThresh,Dashed, ""Red"" ),";
			%end;		
			put "	    	Rotated Labels( ""Automatic"" )}";
			put "    )";
			put "	)";
			put "),";
	
			%if &GKNumber = three or &GKNumber = four %then %do;
				put "Here:vp=Graph Builder(Size( 500, 300 ),Show Control Panel( 0 ), Auto Stretching(0), Lock Scales( 1 ), ";
				put "	Variables(X(:Name(""Dominance &CurrT"")), Y(:Name(""LOD &CurrT"")),";
				%if %length(&ChromLab) %then %do;
					put "		Color(:Name(""&ChromLab""))";
				%end;
				%else %do;
					put "		Color(:Name(""&ChromNum""))";
				%end;
				put "	),";  /*close variables*/
				put " Elements(Points(X,Y,Color, Legend(1))),";
				put "	SendToReport(";
				put "		Dispatch({},""Graph Builder"", OutlineBox, {Set Title(""Estimate Plot for &CurrT"")}),";
				put "		Dispatch({},""graph title"", TextEditBox,{Set Text(""Dominance Effect vs. LOD Score"")}),";
				put "		Dispatch({},""X title"", TextEditBox,{Set Text(""Estimate of Dominance Effect"")}),";
				put "		Dispatch({},""Graph Builder"", FrameBox, {Marker Size(2)}),";
				put "		Dispatch({},""LOD &CurrT"",ScaleBox,";
				put "	    {Format( ""Fixed Dec"", 9, 2 ),Min( -0.5), Max( Here:maxLOD ), Inc(2), Minor Ticks( 0 ),";
				%if &perms %then %do;
					put "			Add Ref Line( &&EWLODThresh&j,Dashed, ""Red"" ),";
				%end;
				%else %do;
					put "			Add Ref Line( &LODThresh,Dashed, ""Red"" ),";
				%end;		
				put "	    	Rotated Labels( ""Automatic"" )}";
				put "    )";
				put "	)";
				put "),";
			%end;
			%if &GKNumber = four %then %do;
				put "Here:vp=Graph Builder(Size( 500, 300 ),Show Control Panel( 0 ), Auto Stretching(0), Lock Scales( 1 ), ";
				put "	Variables(X(:Name(""Additive1 &CurrT"")), Y(:Name(""LOD &CurrT"")),";
				%if %length(&ChromLab) %then %do;
					put "		Color(:Name(""&ChromLab""))";
				%end;
				%else %do;
					put "		Color(:Name(""&ChromNum""))";
				%end;
				put "	),";  /*close variables*/
				put " Elements(Points(X,Y,Color, Legend(1))),";
				put "	SendToReport(";
				put "		Dispatch({},""Graph Builder"", OutlineBox, {Set Title(""Estimate Plot for &CurrT"")}),";
				put "		Dispatch({},""graph title"", TextEditBox,{Set Text(""Additive1 Effect vs. LOD Score"")}),";
				put "		Dispatch({},""X title"", TextEditBox,{Set Text(""Estimate of P1 Allele Additive Effect"")}),";
				put "		Dispatch({},""Graph Builder"", FrameBox, {Marker Size(2)}),";
				put "		Dispatch({},""LOD &CurrT"",ScaleBox,";
				put "	    {Format( ""Fixed Dec"", 9, 2 ),Min( -0.5), Max( Here:maxLOD ), Inc(2), Minor Ticks( 0 ),";
				%if &perms %then %do;
					put "			Add Ref Line( &&EWLODThresh&j,Dashed, ""Red"" ),";
				%end;
				%else %do;
					put "			Add Ref Line( &LODThresh,Dashed, ""Red"" ),";
				%end;		
				put "	    	Rotated Labels( ""Automatic"" )}";
				put "    )";
				put "	)";
				put "),";
			%end;				
			%if &GKNumber = three or &GKNumber = four %then %do;
				put "),";
			%end;
		%end;
		put "),"; /*close vlist*/
		put "Here:qtldata0<<Data Filter(<<Conditional(1),";
		%if %length(&ChromLab) %then %do;
			put "		Add Filter(Columns(""&ChromLab""), Display(:Name(""&ChromLab"")), Size(104, 55), List Display), Mode(Show(1), Include(0))";
		%end;
		%else %do;
			put "		Add Filter(Columns(""&ChromNum""), Display(:Name(""&ChromNum"")), Size(104, 55), List Display), Mode(Show(1), Include(0))";
		%end;
		put ");";
		put "));"; /* close hlist and function*/
	
		put "InsertInto(DataList,{{""Here:qtldata0""}});";
		put "InsertInto(TabButtonList, ""Estimate Plots"");";
		put "InsertInto(TabCodeList,expr(Here:VFn()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";

		put "Here:DFn = Function({},{default local}," ;
		put "If(IsEmpty(Here:qtldata0),";
		%if %index(%Upcase(&MAlgorithm),HALEY) %then %do;
			put "		Here:qtldata0=open(""&ClientOutpath.&MappingResult_OutDataName..sas7bdat"", invisible);";
		%end;
		%else %do;
			/* EM data format */
			put "  Here:qtldata0=open(""&ClientOutpath.&MappingResult_OutDataName.1.sas7bdat"", invisible);";
		%end;
		put "  column(""&ChromNum"")<<Set Modeling Type(Ordinal);";
		put "  column(""Morgans"")<<data type(numeric);";
		%if %length(&ChromLab) %then %do;
			put "  column(""&ChromLab"")<<data type(character);";
		%end;
		put ","; /*else for Is empty check*/
		put "	 Current Data Table(Here:qtldata0);";
		put ");";
	
		put "VListBox(";
		put "	HListBox(";
		put "		Here:dp = Distribution(";
		%do j=1 %to &tTraitNum;
			%let CurrT = %scan(&TraitVar,&j);
			put "			Continuous Distribution(Column(""R2 &CurrT"")),";
		%end;
		put "	)),";
		%if &tTraitNum > 1 %then %do;
			put "	Here:mv = Multivariate(";
			put "		Y(";
			%do j=1 %to &tTraitNum;
				%let CurrT = %scan(&TraitVar,&j);
				put "		Column(""R2 &CurrT""),";
			%end;
			put "		),";
			put "		Estimation Method( ""Row-wise"" ),";
			put "		Scatterplot Matrix(Density Ellipses( 0 ),Shaded Ellipses( 0 ))";
			put "	)";
		%end;	
		put ")"; /*close vlist*/
		put ");"; /* close hlist and function*/		
		
		put "	InsertInto(DataList,{{""Here:qtldata0""}});";
		put "	InsertInto(TabButtonList, ""R2 Plots"");";
		put "	InsertInto(TabCodeList,expr(Here:DFn()));";
		put "	InsertInto(TabAfterCodeList, "" "");";
		put "	InsertInto(TabOpenList,1);";		
	%end; 
 	%else %if %upcase(&MethodSelection) = %upcase(Multiple Interval Mapping) %then %do;
 	
 		put ;
		put "//Auxiliary function for IM and MIM follow-up runs";
		put ;
		
		%QTLFnMacro(FunctionName= Here:AuxiliarQTLFnToUpdateIMModel, VLBName= Here:vlbtmp, OLBName= Here:olbtmp);
		
		put ;
		put "//Here:Nmodels = 0; ";
		put "//Here:Ntables = 0; ";
		/*begin: insert code to add table into the jsl file for posterior runs of the AP*/
		%CodeToAddDataTableToJsl(iswithindatastep= TRUE, fname=);
		/*end: insert code to add table into the jsl file for posterior runs of the AP*/
				 	
 		put ;
 		put "//append effects plot and tables from IM";
 		
 		%macro effplots(FunctionName=);
 			%if &FunctionName = Here:AppendEffAndTabToOlbFn %then %do;
				put " Here:AppendEffAndTabToOlbFn = Function({AppendToBox},{default local}," ;
				put " Here:olbe = outlineBox(""Effect Size Profile"", ";
				/*begin: insert effect profile into the jsl from the first run*/
				put " if(!Here:ManageQTLFlag,";
				put "   //insert effect profile code that does not run in the Manage QTL Model";
				put " Here:hlbe = hlistbox(";
				%do i=1 %to &tChromNum %by 1;
			
					put "Here:opename&i= ""Here:ope""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put "Here:ropename&i= ""Here:rope""||char(Here:Nmodels)||char(""_"")||char(&i);";			
				
					put "Here:o=OutlineBox("@;
					%if %length(&ChromLab) %then %do;
						put " ""&ChromLab &&ChrLabel&i"",";
					%end;
					%else %do;
						put " ""&ChromNum &&ChrVars&i"",";
					%end;
					put "Vlistbox(";
					put 'Eval(Parse(EvalInsert("\[';
					put " //show(""Here:AppendEffAndTabToOlbFn"");";
					put " //show(Here:opename&i);";
					put " ^Here:opename&i^ = Graph Builder(";
					put "	Size(500,300),";
					put "	Show Control Panel( 0 ),";
					put "	Legend Position( ""Bottom"" ),";
					put "	Show Title( 0 ),";				
					put " 	Variables(";
					put " 		X( :&GDMetric),"; 
					%if &ngprobs = 2 %then %do;
						put "	Y(:(&tlab_a21)),";
					%end;
					%else %if &ngprobs = 3 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)),";
					%end;
					%else %if &ngprobs = 4 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)), Y(:(&tlab_a11),Position(1)),";
					%end;				
					%do j=2 %to &tTraitNum %by 1;
						%if &ngprobs = 2 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)),";
						%end;
						%else %if &ngprobs = 3 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)), Y(:(&&tlab_d&j),Position(1)),";
						%end;
						%else %if &ngprobs = 4 %then %do;
							put "	Y(:(&&tlab_a2&j),Position(1)), Y(:(&&tlab_d&j),Position(1)), Y(:(&&tlab_a1&j),Position(1)),";
						%end;
					%end;
					put "),";
					put "	Elements(";
					put "		Points( X,";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;
					put "		Legend( 1 ) ),";
					put "		Line( X, ";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;
					put "		Legend( 3 ) ),";
					put "	),";
					%if &tChromNum > 1 %then %do;
						%if %length(&ChromLab) %then %do;
							put ", Where(uppercase(:&ChromLab)==uppercase(""&&ChrLabel&i"")),";
						%end;
						%else %do;
							put ", Where(:Name(""&ChromNum"")==&&ChrVars&i),";
						%end;
					%end;
					put "	SendToReport(";
					put "		Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Effect Size"" )} ),";
					%if &tTraitNum < 2 and &ngprobs <= 2 %then %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Legend Position( {1, [-1], 3, [0]} ), Position( {-1, 0} )}";
						put "		),";				
					%end;	
					%else %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Set Wrap( 2 )} ";
						put "		),";					
					%end;
					put "		Dispatch(";
					put "			{},";
					put "			""Graph Builder"",";
					put "			OutlineBox,";
					put "			{Set Title( ""Effect Size Plot"" ), Image Export Display( Normal )}";
					put "		),";					
					put "	),";
					put ");";
					put "if(type(^Here:opename&i^)==""Scriptable"" | type(^Here:opename&i^)==""List"",";
					put "^Here:ropename&i^ = ^Here:opename&i^ << report;";
					%if &tChromNum > 1 %then %do;
						put 'try(o[TextBox(1)] << Set Text("") << Set Font Size(1));';
					%end;	
					put "for(i=1, i<=1, i++,";
					put "try(";
					put "^Here:ropename&i^[i][framebox(1)] << add graphics script(hline(0));";		
					put "^Here:ropename&i^[i][framebox(1)] << MarkerSize(1);";
					put "^Here:ropename&i^[i][framebox(1)] << MarkerDrawingMode(Normal);";		
					put "^Here:ropename&i^[i][Axisbox(2)] << RemoveAxisLabel;";
					put "^Here:ropename&i^[i][Axisbox(2)] << Min(-Here:maxEff);";
					put "^Here:ropename&i^[i][Axisbox(2)] << Max(Here:maxEff);";
					put "^Here:ropename&i^[i][Axisbox(2)] << Inc(.1);";
					put "^Here:ropename&i^[i][Axisbox(2)] << Minor Ticks(0);";
					put "//^Here:ropename&i^[i][Axisbox(2)] << AddAxisLabel(""Effect Size"");";
					%if %Upcase(&GDMetric) = CM %then %do;
						put "^Here:ropename&i^[i][Axisbox(1)] << Min(-.5);";
					%end;
					/*put "^Here:ropename&i^[i][framebox(1)] << FrameSize(275,175);";*/
					put ")));";
					put '	]\")));';
					put ")),";
					put "try(o[textBox(1)] << Delete);";
				%end;
				put " ";
				put "	);";
				
				put "   ,";
				
				put "   //insert effect profile code that does not run in the Manage QTL Model";
				put " Here:hlbe = hlistbox(";
				%do i=1 %to &tChromNum %by 1;
			
					put "Here:mopename&i= ""Here:mope""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put "Here:rmopename&i= ""Here:rmope""||char(Here:Nmodels)||char(""_"")||char(&i);";			
				
					put "Here:o=OutlineBox("@;
					%if %length(&ChromLab) %then %do;
						put " ""&ChromLab &&ChrLabel&i"",";
					%end;
					%else %do;
						put " ""&ChromNum &&ChrVars&i"",";
					%end;
					put "Vlistbox(";
					put 'Eval(Parse(EvalInsert("\[';
					put " //show(""Here:AppendEffAndTabToOlbFn"");";
					put " //show(Here:mopename&i);";
					put " ^Here:mopename&i^ = Graph Builder(";
					put "	Size(500,300),";
					put "	Show Control Panel( 0 ),";
					put "	Legend Position( ""Bottom"" ),";
					put "	Show Title( 0 ),";				
					put " 	Variables(";
					put " 		X( :&GDMetric),"; 
					%if &ngprobs = 2 %then %do;
						put "	Y(:(&tlab_a21)),";
					%end;
					%else %if &ngprobs = 3 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)),";
					%end;
					%else %if &ngprobs = 4 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)), Y(:(&tlab_a11),Position(1)),";
					%end;				
					%do j=2 %to &tTraitNum %by 1;
						%if &ngprobs = 2 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)),";
						%end;
						%else %if &ngprobs = 3 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)), Y(:(&&tlab_d&j),Position(1)),";
						%end;
						%else %if &ngprobs = 4 %then %do;
							put "	Y(:(&&tlab_a2&j),Position(1)), Y(:(&&tlab_d&j),Position(1)), Y(:(&&tlab_a1&j),Position(1)),";
						%end;
					%end;
					put "),";
					put "	Elements(";
					put "		Points( X,";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;
					put "		Legend( 1 ) ),";
					put "		Line( X, ";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;
					put "		Legend( 3 ) ),";
					put "	),";
					%if &tChromNum > 1 %then %do;
						%if %length(&ChromLab) %then %do;
							put ", Where(uppercase(:&ChromLab)==uppercase(""&&ChrLabel&i"")),";
						%end;
						%else %do;
							put ", Where(:Name(""&ChromNum"")==&&ChrVars&i),";
						%end;
					%end;
					put "	SendToReport(";
					put "		Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Effect Size"" )} ),";
					%if &tTraitNum < 2 and &ngprobs <= 2 %then %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Legend Position( {1, [-1], 3, [0]} ), Position( {-1, 0} )}";
						put "		),";				
					%end;	
					%else %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Set Wrap( 2 )} ";
						put "		),";					
					%end;					
					put "		Dispatch(";
					put "			{},";
					put "			""Graph Builder"",";
					put "			OutlineBox,";
					put "			{Set Title( ""Effect Size Plot"" ), Image Export Display( Normal )}";
					put "		),";					
					put "	),";
					put ");";
					put "if(type(^Here:mopename&i^)==""Scriptable"" | type(^Here:mopename&i^)==""List"",";
					put "^Here:rmopename&i^ = ^Here:mopename&i^ << report;";
					%if &tChromNum > 1 %then %do;
						put 'try(o[TextBox(1)] << Set Text("") << Set Font Size(1));';
					%end;	
					put "for(i=1, i<=1, i++,";
					put "try(";
					put "^Here:rmopename&i^[i][framebox(1)] << add graphics script(hline(0));";		
					put "^Here:rmopename&i^[i][framebox(1)] << MarkerSize(1);";
					put "^Here:rmopename&i^[i][framebox(1)] << MarkerDrawingMode(Normal);";		
					put "^Here:rmopename&i^[i][Axisbox(2)] << RemoveAxisLabel;";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Min(-Here:maxEff);";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Max(Here:maxEff);";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Inc(.1);";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Minor Ticks(0);";
					put "//^Here:rmopename&i^[i][Axisbox(2)] << AddAxisLabel(""Effect Size"");";
					%if %Upcase(&GDMetric) = CM %then %do;
						put "^Here:rmopename&i^[i][Axisbox(1)] << Min(-.5);";
					%end;
					/*put "^Here:rmopename&i^[i][framebox(1)] << FrameSize(275,175);";*/
					put ")));";
					put '	]\")));';
					put ")),";
					put "try(o[textBox(1)] << Delete);";
				%end;
				put " ";
				put "	);";
				
				put " );";
				put "	);";
				put "	//Here:olb0 << Append(Here:olbe);" ; 
				put "	AppendToBox << Append(Here:olbe);" ; 
				/*end: insert effect profile into the jsl  from the first run*/

				/*begin: insert table values into the jsl  from the first run*/
				%PrepModelTables(iswithindatastep= TRUE, fname=, modelnumb=0);
				
				put " Here:AppendHlbToOlbfn(AppendToBox= AppendToBox);"; /*Here:AppendHlbToOlbFn() comes from PrepModelTables()*/
				/*end: insert table values into the jsl  from the first run*/
				put " );";

			%end;
			%else %do;
				/*Here:AppendEffPlotExpr*/
				/*put " Here:AppendEffPlotExpr = Expr(";*/
				put " if(!Here:ManageQTLFlag,";
				put "   //insert effect profile code that does not run in the Manage QTL Model";
				put "  Here:hlbe = hlistbox(";
				%do i=1 %to &tChromNum %by 1;
			
					put "Here:opename&i= ""Here:ope""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put "Here:ropename&i= ""Here:rope""||char(Here:Nmodels)||char(""_"")||char(&i);";			
				
					put "Here:o=OutlineBox("@;
					%if %length(&ChromLab) %then %do;
						put " ""&ChromLab &&ChrLabel&i"",";
					%end;
					%else %do;
						put " ""&ChromNum &&ChrVars&i"",";
					%end;
					put "Vlistbox(";
					put 'Eval(Parse(EvalInsert("\[';
					put " //show(""Here:AppendEffAndTabToOlbFn"");";
					put " //show(Here:opename&i);";
					put " ^Here:opename&i^ = Graph Builder(";
					put "	Size(500,300),";
					put "	Show Control Panel( 0 ),";
					put "	Legend Position( ""Bottom"" ),";
					put "	Show Title( 0 ),";				
					put " 	Variables(";
					put " 		X( :&GDMetric),"; 
					%if &ngprobs = 2 %then %do;
						put "	Y(:(&tlab_a21)),";
					%end;
					%else %if &ngprobs = 3 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)),";
					%end;
					%else %if &ngprobs = 4 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)), Y(:(&tlab_a11),Position(1)),";
					%end;				
					%do j=2 %to &tTraitNum %by 1;
						%if &ngprobs = 2 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)),";
						%end;
						%else %if &ngprobs = 3 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)), Y(:(&&tlab_d&j),Position(1)),";
						%end;
						%else %if &ngprobs = 4 %then %do;
							put "	Y(:(&&tlab_a2&j),Position(1)), Y(:(&&tlab_d&j),Position(1)), Y(:(&&tlab_a1&j),Position(1)),";
						%end;
					%end;
					put "),";

					put "	Elements(";
					put "		Points( X,";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;	
					put "		Legend( 1 ) ),";
					put "		Line( X, ";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;
					put "		Legend( 3 ) ),";
					put "	),";
					%if &tChromNum > 1 %then %do;
						%if %length(&ChromLab) %then %do;
							put ", Where(uppercase(:&ChromLab)==uppercase(""&&ChrLabel&i"")),";
						%end;
						%else %do;
							put ", Where(:Name(""&ChromNum"")==&&ChrVars&i),";
						%end;
					%end;
					put "	SendToReport(";
					put "		Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Effect Size"" )} ),";
					%if &tTraitNum < 2 and &ngprobs <= 2 %then %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Legend Position( {1, [-1], 3, [0]} ), Position( {-1, 0} )}";
						put "		),";				
					%end;	
					%else %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Set Wrap( 2 )} ";
						put "		),";					
					%end;					
					put "		Dispatch(";
					put "			{},";
					put "			""Graph Builder"",";
					put "			OutlineBox,";
					put "			{Set Title( ""Effect Size Plot"" ), Image Export Display( Normal )}";
					put "		),";					
					put "	),";
					put ");";
					put "if(type(^Here:opename&i^)==""Scriptable"" | type(^Here:opename&i^)==""List"",";
					put "^Here:ropename&i^ = ^Here:opename&i^ << report;";
					%if &tChromNum > 1 %then %do;
						put 'try(o[TextBox(1)] << Set Text("") << Set Font Size(1));';
					%end;	
					put "for(i=1, i<=1, i++,";
					put "try(";
					put "^Here:ropename&i^[i][framebox(1)] << add graphics script(hline(0));";		
					put "^Here:ropename&i^[i][framebox(1)] << MarkerSize(1);";
					put "^Here:ropename&i^[i][framebox(1)] << MarkerDrawingMode(Normal);";		
					put "^Here:ropename&i^[i][Axisbox(2)] << RemoveAxisLabel;";
					put "^Here:ropename&i^[i][Axisbox(2)] << Min(-Here:maxEff);";
					put "^Here:ropename&i^[i][Axisbox(2)] << Max(Here:maxEff);";
					put "^Here:ropename&i^[i][Axisbox(2)] << Inc(.1);";
					put "^Here:ropename&i^[i][Axisbox(2)] << Minor Ticks(0);";
					put "//^Here:ropename&i^[i][Axisbox(2)] << AddAxisLabel(""Effect Size"");";
					%if %Upcase(&GDMetric) = CM %then %do;
						put "^Here:ropename&i^[i][Axisbox(1)] << Min(-.5);";
					%end;
					/*put "^Here:ropename&i^[i][framebox(1)] << FrameSize(275,175);";*/
					put ")));";
					put '	]\")));';
					put ")),";
					put "try(o[textBox(1)] << Delete);";				
					put " ";
				%end;
				put "	);";	
			        put "	,";	
				put "   //insert effect profile code that does run in the Manage QTL Model";
				put "  Here:hlbe = hlistbox(";
				%do i=1 %to &tChromNum %by 1;
				
					put "Here:mopename&i= ""Here:mope""||char(Here:Nmodels)||char(""_"")||char(&i);";
					put "Here:rmopename&i= ""Here:rmope""||char(Here:Nmodels)||char(""_"")||char(&i);";			
			
					put "Here:o=OutlineBox("@;
					%if %length(&ChromLab) %then %do;
						put " ""&ChromLab &&ChrLabel&i"",";
					%end;
					%else %do;
						put " ""&ChromNum &&ChrVars&i"",";
					%end;
					put "Vlistbox(";
					put 'Eval(Parse(EvalInsert("\[';
					put " //show(""Here:AppendEffAndTabToOlbFn"");";
					put " //show(Here:mopename&i);";
					put " ^Here:mopename&i^ = Graph Builder(";
					put "	Size(500,300),";
					put "	Show Control Panel( 0 ),";
					put "	Legend Position( ""Bottom"" ),";
					put "	Show Title( 0 ),";				
					put " 	Variables(";
					put " 		X( :&GDMetric),"; 
					%if &ngprobs = 2 %then %do;
						put "	Y(:(&tlab_a21)),";
					%end;
					%else %if &ngprobs = 3 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)),";
					%end;
					%else %if &ngprobs = 4 %then %do;
						put "	Y(:(&tlab_a21)), Y(:(&tlab_d1),Position(1)), Y(:(&tlab_a11),Position(1)),";
					%end;				
					%do j=2 %to &tTraitNum %by 1;
						%if &ngprobs = 2 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)),";
						%end;
						%else %if &ngprobs = 3 %then %do;
							put "	Y(:(&&tlab_a2&j), Position(1)), Y(:(&&tlab_d&j),Position(1)),";
						%end;
						%else %if &ngprobs = 4 %then %do;
							put "	Y(:(&&tlab_a2&j),Position(1)), Y(:(&&tlab_d&j),Position(1)), Y(:(&&tlab_a1&j),Position(1)),";
						%end;
					%end;
					put "),";
	
					put "	Elements(";
					put "		Points( X,";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;
					put "		Legend( 1 ) ),";
					put "		Line( X, ";
					%let l=1;
					%do j=1 %to &tTraitNum %by 1;
						%do k= 1 %to %eval(&ngprobs-1);
							put "		Y( &l ), ";
							%let l= %eval(&l + 1);
						%end;
					%end;
					put "		Legend( 3 ) ),";
					put "	),";
					%if &tChromNum > 1 %then %do;
						%if %length(&ChromLab) %then %do;
							put ", Where(uppercase(:&ChromLab)==uppercase(""&&ChrLabel&i"")),";
						%end;
						%else %do;
							put ", Where(:Name(""&ChromNum"")==&&ChrVars&i),";
						%end;
					%end;
					put "	SendToReport(";
					put "		Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Effect Size"" )} ),";
					%if &tTraitNum < 2 and &ngprobs <= 2 %then %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Legend Position( {1, [-1], 3, [0]} ), Position( {-1, 0} )}";
						put "		),";				
					%end;	
					%else %do;
						put "		Dispatch(";
						put "			{},";
						put "			""400"",";
						put "			LegendBox,";
						put "			{Set Wrap( 2 )} ";
						put "		),";					
					%end;					
					put "		Dispatch(";
					put "			{},";
					put "			""Graph Builder"",";
					put "			OutlineBox,";
					put "			{Set Title( ""Effect Size Plot"" ), Image Export Display( Normal )}";
					put "		),";					
					put "	),";
					put ");";
					put "if(type(^Here:mopename&i^)==""Scriptable"" | type(^Here:mopename&i^)==""List"",";
					put "^Here:rmopename&i^ = ^Here:mopename&i^ << report;";
					%if &tChromNum > 1 %then %do;
						put 'try(o[TextBox(1)] << Set Text("") << Set Font Size(1));';
					%end;	
					put "for(i=1, i<=1, i++,";
					put "try(";
					put "^Here:rmopename&i^[i][framebox(1)] << add graphics script(hline(0));";		
					put "^Here:rmopename&i^[i][framebox(1)] << MarkerSize(1);";
					put "^Here:rmopename&i^[i][framebox(1)] << MarkerDrawingMode(Normal);";		
					put "^Here:rmopename&i^[i][Axisbox(2)] << RemoveAxisLabel;";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Min(-Here:maxEff);";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Max(Here:maxEff);";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Inc(.1);";
					put "^Here:rmopename&i^[i][Axisbox(2)] << Minor Ticks(0);";
					put "//^Here:rmopename&i^[i][Axisbox(2)] << AddAxisLabel(""Effect Size"");";
					%if %Upcase(&GDMetric) = CM %then %do;
						put "^Here:rmopename&i^[i][Axisbox(1)] << Min(-.5);";
					%end;
					/*put "^Here:rmopename&i^[i][framebox(1)] << FrameSize(275,175);";*/
					put ")));";
					put '	]\")));';
					put ")),";
					put "try(o[textBox(1)] << Delete);";				
					put " ";
						
				%end;
				put "	);";	
				
				put " );";
				/*put " );";*/	
			%end;


		%mend effplots; 

		%effplots(FunctionName = Here:AppendEffAndTabToOlbFn);
	
		put "InsertInto(TabAfterCodeList, expr(Here:AppendEffAndTabToOlbFn(AppendToBox= Here:olb0))); ";	
		put "InsertInto(TabOpenList,1); ";
		
		
 		/*Outpath and ProcessName for use in the follow-up MIM analysis*/
 		put ;
 		put "//Parameters for MIM follow-up runs";
 		put "MacroPath = ""&MacroPath"";";
 		put "OutPath= ""&OutPath"";";
 		put "ProcessName= ""&ProcessName"";";
 		put "ClientOutPath = ""&ClientOutPath""; ";
 		put "MappingGraph_Script= ""&MappingGraph_Script"";";
 		
		/*output data set for the results of sas macro fitmodel() */
		put "HKMIMTestGrid_OutData = ""&HKMIMTestGrid_OutData""; ";
		put "HKMIMTestGrid_OutDataName = ""&HKMIMTestGrid_OutDataName""; ";
		put "HKMIMData_OutData = ""&HKMIMData_OutData"";";
		put "HKMIMData_OutDataName = ""&HKMIMData_OutDataName"";";
 		put "HKMIMModelRaw_OutData = ""&HKMIMModelRaw_OutData"";";
 		put "HKMIMModelRaw_OutDataName = ""&HKMIMModelRaw_OutDataName"";";		
 		put "HKMIMModelSum_OutData = ""&HKMIMModelSum_OutData"";";
 		put "HKMIMModelSum_OutDataName = ""&HKMIMModelSum_OutDataName"";";
 		
 		/*output data set for the results of sas macro profile() */	
		put "MappingResult_OutData = ""&MappingResult_OutData"";";
		put "MappingResult_OutDataName = ""&MappingResult_OutDataName"";";
		
		*--- output data set from sas macro PrepTablesToAdd()---*;
		put "HKMIMFitTable_OutData= ""&HKMIMFitTable_OutData"";";
		put "HKMIMFitTable_OutDataName= ""&HKMIMFitTable_OutDataName"";";
		put "HKMIMQtlTable_OutData= ""&HKMIMQtlTable_OutData"";";
		put "HKMIMQtlTable_OutDataName= ""&HKMIMQtlTable_OutDataName"";";
		put "HKMIMEffTable_OutData= ""&HKMIMEffTable_OutData"";";
		put "HKMIMEffTable_OutDataName= ""&HKMIMEffTable_OutDataName"";";
		put "HKMIMCofTable_OutData= ""&HKMIMCofTable_OutData"";";
		put "HKMIMCofTable_OutDataName= ""&HKMIMCofTable_OutDataName"";";
		
		*--- output setting data table***;	
		put "HKMIMSetTable_OutData= ""&HKMIMSetTable_OutData"";";		
		put "HKMIMSetTable_OutDataName= ""&HKMIMSetTable_OutDataName"";";		
			 
 		/*output library to store results from MIM analyses */
 		put "g_OutLib = ""&g_OutLib"";"; 
 					
 		/* MIM macro variables not from MIM Options tab but that are already created before calling MIM macro fuctions */
  		put "g_mapgrid= HKMIMTestGrid_OutData;";
 		put "g_hkmimdata= HKMIMData_OutData;";	
		put "g_fittable= HKMIMFitTable_OutData;";
		put "g_qtltable= HKMIMQtlTable_OutData;";
		put "g_efftable= HKMIMEffTable_OutData;";
		put "g_coftable= HKMIMCofTable_OutData;";
		put "g_settable= HKMIMSetTable_OutData;"; 		
 		
 		/* MIM macro variables neither from MIM Options tab nor created before calling MIM macro functions */
 		put "g_hkmimmodel= HKMIMModelRaw_OutDataName;";
 		
 		/* Passing values of macro variables not from MIM Options tab to my old macro variables */
 		put "g_trt = ""&g_trt"";";
 		put "g_ng= ""&g_ng"";";
 		put "g_nx= ""&g_nx"";";	

 		put "g_ChromLab= ""&ChromLab"";";
 		put "g_ChromNum= ""&ChromNum"";"; 
 		put "g_NumChrom= ""&tChromNum"";";  		
		put "g_MarkerLab= ""&MarkerLab"";";
		put "g_MarkerPos= ""&MarkerPos"";";
		put "g_TraitVar= ""&TraitVar"";";
		put "g_FactorVars= ""&FactorVarsNew"";";
		put "g_GDMetric= ""&GDMetric"";";
		put "g_LODThresh= ""&LODThresh"";";
		put "g_tlab_lod1= ""&tlab_lod1"";";

 		/* Passing values of macro variables from MIM Options tab to my old macro variables */
 		put "g_teststep= ""&g_teststep"";";
 		put "g_wsize= ""&g_wsize"";";
 		put "g_criterion_ref= ""&g_criterion_ref"";";	
 		
 		put "g_criterion_main= ""&g_criterion_main"";";
 		put "g_lod_threshold_main= ""&g_lod_threshold_main"";";
 		put "g_probf_entry_main= ""&g_probf_entry_main"";"; 
 		put "g_teststep_main= ""&g_teststep_main"";";
 		
 		put "g_criterion_epi0D= ""&g_criterion_epi0D"";";
 		put "g_lod_threshold_epi0D= ""&g_lod_threshold_epi0D"";"; 
 		put "g_probf_entry_epi0D= ""&g_probf_entry_epi0D"";"; 
 		
 		put "g_criterion_epi1D= ""&g_criterion_epi1D"";";
 		put "g_lod_threshold_epi1D= ""&g_lod_threshold_epi1D"";";
 		put "g_probf_entry_epi1D= ""&g_probf_entry_epi1D"";"; 
 		put "g_teststep_epi1D= ""&g_teststep_epi1D"";";
 		
 		put "g_criterion_epi2D= ""&g_criterion_epi2D"";";
 		put "g_lod_threshold_epi2D= ""&g_lod_threshold_epi2D"";";
 		put "g_probf_entry_epi2D= ""&g_probf_entry_epi2D"";"; 
 		put "g_teststep_epi2D= ""&g_teststep_epi2D"";";
 
 		put "g_criterion_dropmqtlonly= ""&g_criterion_dropmqtlonly"";";
 		put "g_lod_threshold_dropmqtlonly= ""&g_lod_threshold_dropmqtlonly"";";
 		put "g_probf_stay_dropmqtlonly= ""&g_probf_stay_dropmqtlonly"";"; 
 		put "g_auto_dropmqtlonly = ""&g_auto_dropmqtlonly"";";
 		
 		put "g_criterion_dropepieff= ""&g_criterion_dropepieff"";";
 		put "g_lod_threshold_dropepieff= ""&g_lod_threshold_dropepieff"";";
 		put "g_probf_stay_dropepieff= ""&g_probf_stay_dropepieff"";"; 
 		put "g_auto_dropepieff = ""&g_auto_dropepieff"";";
 		
 		/* Temporary macro variables for testing purpose only */
 		put "g_firstobs= ""&g_firstobs"";";
		put "g_obs= ""&g_obs"";";
		
		/* Instructions to run AP*/
		put "g_Server=""&Server"";";
		put "g_LS_HOME=""&LS_HOME"";";
		put "g_CLIENT_LS_HOME=""&CLIENT_LS_HOME"";";
		put "g_GENOMICS_HOME=""&GENOMICS_HOME"";";
		put "g_USER_HOME=""&USER_HOME"";";
		put "g_LSUSER_HOME=""&LSUSER_HOME"";";
		put "g_CLIENT_USER_HOME=""&CLIENT_USER_HOME"";";
		put "g_CLIENT_LSUSER_HOME=""&CLIENT_LSUSER_HOME"";";
		put "g_Metapath=""&Metapath"";";
		put "g_Starter_Index=""&Starter_Index"";";
		put "g_LS_Product=""&LS_Product"";";
		put "g_CLIENT_HOST=""&CLIENT_HOST"";";
		
		put ;
		put "//Expressions for MIM follow-up runs";
		/*begin: insert code to plot qtl profile into the jsl file for posterior runs of the AP*/
		put "Here:AppendQtlPlotExpr = Expr( " ;
		
		%qtlplots(trackcall= Here:AppendQtlPlotExpr);
		
		put "); " ;
		/*end: insert code to plot qtl profile into the jsl file for posterior runs of the AP*/
		
		/*begin: insert code to plot effect profile into the jsl file for posterior runs of the AP*/
		put ;
		put "Here:AppendEffPlotExpr = Expr( " ;
		put "  Here:olbe = outlineBox(""Effect Size Profile"", ";
		
		%effplots(FunctionName = Here:AppendEffPlotExpr);
		
		put "); " ;
		put "); " ;
		/*end: insert code to plot effect profile into the jsl file for posterior runs of the AP*/
		
		/* File convertion */
		put ;
		put "if(host is(Windows),";
		put "   include( ConvertFilePath(""$GENOMICS_HOME/JSLFiles/MultiQTlModel_DrillDowns.jsl"", absolute, windows));";
		put ",";
		put "   include( ConvertFilePath(""$GENOMICS_HOME/JSLFiles/MultiQTlModel_DrillDowns.jsl"", absolute, POSIX));";
		put ");";
		put ;
		
		put ;
		put "//MIM Action Buttons";
		put "Here:RunFwdMainExpr= Expr( ";
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:FwdMainExpr;" ;
		put "); ";
		put ;
		put "Here:RunBackDropMainOnlyExpr= Expr( ";
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:BackDropMainOnlyExpr; ";
		put "); ";
		put ;		
		put "Here:RunFwdEpi0DExpr= Expr( ";
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:FwdEpi0DExpr; ";
		put "); ";
		put ;
		put "Here:RunFwdEpi1DExpr= Expr( ";
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:FwdEpi1DExpr; ";	
		put "); ";	
		put ;
		put "Here:RunFwdEpi2DExpr= Expr( ";		
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:FwdEpi2DExpr; ";		
		put "); ";	
		put ;	
		put "Here:RunBackDropEpiExpr= Expr( ";		
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:BackDropEpiExpr; ";
		put "); ";		
		put ;
		put "Here:RunGFwdQTLxCofactorExpr= Expr( ";
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:GFwdQTLxCofactorExpr;" ;
		put "); ";
		put ;
		put "Here:RunMFwdQTLxCofactorExpr= Expr( ";
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:MFwdQTLxCofactorExpr;" ;
		put "); ";
		put ;
		put "Here:RunBackDropQTLxCofactorExpr= Expr( ";		
		put "	Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:BackDropQTLxCofactorExpr; ";
		put "); ";		
		put ;		
		put "Here:NChanges = 0; ";
		put "Here:RunChangeMIMOptionsExpr= Expr( ";		
		put "	Here:NChanges= Here:NChanges + 1; ";
		put "	Here:ChangeMIMOptionsExpr; ";
		put "); ";		
		put ;
		put "Here:RunInteractivelyManageModelExpr= Expr( ";
		put "	//Here:Nmodels= Here:Nmodels + 1; ";
		put "	Here:InteractivelyManageModelExpr; ";
		put "); ";
		put ;
		
		put "Here:ActionButtonDesc = {{}};";
		put "Here:ActionButtonList = {{}};";
		put "Here:ActionCodeList = {{}};";		
	
 		put "Here:ActionButtonDesc = {""Genomic Search for QTL Main Effects""};";
		put "Here:ActionButtonList= {{""Search for QTL Main Effects""}};";
		put "Here:ActionCodeList= {{Here:RunFwdMainExpr}};";	

		put "InsertInto(Here:ActionButtonDesc, {""Backward Elimination of QTL Main Effects""});";
		put "InsertInto(Here:ActionButtonList, {{""Elimination of QTL Main Effects""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunBackDropMainOnlyExpr}});";
			
		put "InsertInto(Here:ActionButtonDesc, {""Model Search for QTL x QTL Effects""});";
		put "InsertInto(Here:ActionButtonList, {{""QTL in the Model x QTL in the Model""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunFwdEpi0DExpr}});";	
 			
		put "InsertInto(Here:ActionButtonDesc,{""1-D Genomewide Scan for QTL x QTL Effects""});";		
		put "InsertInto(Here:ActionButtonList,{{""QTL in the Model x New QTL""}});";
		put "InsertInto(Here:ActionCodeList ,{{Here:RunFwdEpi1DExpr}});";
			
		put "InsertInto(Here:ActionButtonDesc,{""2-D Genomewide Scan for QTL x QTL Effects""});";
		put "InsertInto(Here:ActionButtonList,{{""New QTL x New QTL""}});";
		put "InsertInto(Here:ActionCodeList ,{{Here:RunFwdEpi2DExpr}});"; 
			
		put "InsertInto(Here:ActionButtonDesc, {""Backward Elimination of QTL x QTL Effects""});";
		put "InsertInto(Here:ActionButtonList, {{""Elimination of QTL x QTL""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunBackDropEpiExpr}});";

		put "InsertInto(Here:ActionButtonDesc, {""Genomic Search for Cofactor x QTL""});";
		put "InsertInto(Here:ActionButtonList, {{""Cofactor x New QTL""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunGFwdQTLxCofactorExpr}});";

		put "InsertInto(Here:ActionButtonDesc, {""Model Search for Cofactor x QTL""});";
		put "InsertInto(Here:ActionButtonList, {{""Cofactor x QTL in the Model""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunMFwdQTLxCofactorExpr}});";
		
		put "InsertInto(Here:ActionButtonDesc, {""Backward Elimination of Cofactor x QTL""});";
		put "InsertInto(Here:ActionButtonList, {{""Elimination of Cofactor x QTL""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunBackDropQTLxCofactorExpr}});";		
		
 		/*put "InsertInto(Here:ActionButtonDesc, {""Summary of the MIM Model""});";*/
		/*put "InsertInto(Here:ActionButtonList, {{""Summary of the MIM Model""}});";*/
		/*put "InsertInto(Here:ActionCodeList, {{Here:FitModelExpr}});";*/
		
		put "InsertInto(Here:ActionButtonDesc, {""Interactively Add or Delete Effects""});";
		put "InsertInto(Here:ActionButtonList, {{""Manage QTL Model""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunInteractivelyManageModelExpr}});";
		
		put "InsertInto(Here:ActionButtonDesc, {""Change MIM Options""});";
		put "InsertInto(Here:ActionButtonList, {{""Change MIM Options""}});";
		put "InsertInto(Here:ActionCodeList, {{Here:RunChangeMIMOptionsExpr}});";
	
	%end;	
	
	%if %upcase(&MethodSelection) = %upcase(Composite Interval Mapping) 
		and &CMarkerSM ^= Using SAS Data Set 
		and %length(&TraitVar_nocofactor) > 0 %then %do;
		
		%*put TraitVar_nocofactor= &TraitVar_nocofactor;
		%if &nTraitVar_nocofactor > 1 %then %do;
			%let word1 = traits;
			%let word2 = these;
		%end;
		%else %do;
			%let word1 = trait;
			%let word2 = this;
		%end;

		put "ActionButtonMessage=NewWindow(""&ProcessName Warning Message"", showmenu(0), showtoolbars(0),";
		put "	VListBox( ";
		put "		tb=TextBox(""Warning message: For the given Control Marker Selection Method, no cofactor marker was found for &word1 &TraitVar_nocofactor.."",";
		put "			<<Font Color(5),<<Set Font Size(10),<<Set Wrap(500)";
		put "		),";
		put "		TextBox(" "),";
		put "		ButtonBox(""OK"",ActionButtonMessage<<CloseWindow);";
		put "	);";
		put ");";
	
	%end;	
run;

%if %index(%Upcase(&MAlgorithm),HALEY) %then %do;
	%PackageInsertFile(jsl code from &ProcessName, &OutPath, &MappingGraph_ScriptName, &MappingResult_OutDataName..sas7bdat, text/plain, yes);
%end;
%else %do;
	%PackageInsertFile(jsl code from &ProcessName, &OutPath, &MappingGraph_ScriptName, &MappingResult_OutDataName.1.sas7bdat, text/plain, yes);
%end;
/*Kemicl - surface data table created by CIM if no control marker data set was specified */
%if %quote(&MethodSelection) = %str(Composite Interval Mapping) and %quote(&CMarkerSM) ne %str(Using SAS Data Set) %then %do;
    %PackageInsertDataset(Control Marker Data Set from &ProcessName, OutLib, &CMDataName);
%end;
 
%exit:;
%*goto exitcheck;
 
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%exitcheck:; *---sbrlsi used it while testing the MIM model ---*;

%mend SingleQTLModel;

* run the macro;
%SingleQTLModel;