/*
JMP Genomics Analytical Process

Name: Tassel-GBS Import Engine
Author: Luciano Silva
History:
	sbrlsi 07/15/14 Initial Coding
Description:
		Tassel-GBS Import Engine transcodes output data from the TASSEL 3 Genotyping by Sequencing (GBS) pipeline to 
		into Input and Annotation Data Sets ready for use in downstream JMP Genomics. The Annotation Data Set contains
		columns of data summary and statistics relevant to downstream quality control analyses.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData IndVars ListIndVars IDVar Parent1Var Parent2Var KeepVars WhereStmt OutPath
AnnoInData AnnoMName AnnoWhereStmt PloidyLevel PloidyType CrossType GenNum SegRatios MultipleTestingMethod Alpha 
SegRemove MaxGenoError MaxMissInd SetHetToMiss DoCellPlots
ServerOutPath OutPath OutName MacroPath TasselGBSImport_OutData TasselGBSImportGen_OutData TasselGBSImportAnn_Outdata 
TasselGBSImportSeg_OutData TasselGBSImportScript 
;

%macro TasselGBSImportEngine;
	
	**ProcessName must be declared before including UtilityMacros*;
	%global FullAPName ProcessName;
	%let FullAPName = %str(Tassel-GBS Import Engine);		
	%let ProcessName = TasselGBSImportEngine;
	
	***begin: Including macros*;
	**The macro variable ClientOutpath is created upon including the UtilityMacros*;
	%include "&MacroPath./UtilityMacros.sas" / nosource;
	%*include "&MacroPath./MultipleTestingAdjustment.sas" / nosource;
	***end: Including macros*;
	
	**Setting ods output*;
	ods html close;
	*options nolabel;
	
	**processing input parameters from the xml file*;	
	%GetOrigFileName(InData,OrigInData);
	%CheckDataName(_name=OutName,_default=&OrigInData);
	
	%PathName(&InData);
	%if &exiterror %then %do;
		%put ERROR: The Input Data Set specification is invalid.;
		%goto exit;
	%end;
	%let InData = &TmpName;
	libname InLib "&TmpPath";
	libname OutLib "&OutPath";
	
	%if %length(&AnnoInData) %then %do;
		%PathName(&AnnoInData);
			%if &exiterror %then %do;
			%put ERROR: Annotation data set (map information) specification is invalid.;
			%let exiterror=1;
			%goto exit;
		%end;
		%let AnnoData = &TmpName;
		libname AnnoLib "&TmpPath";
	%end;
	
	%if %length(&PloidyLevel) = 0 %then %do;
		%let PloidyLevel = 2;
	%end;
		
	**xml variables error checking*;
	%put IndVars= &IndVars &ListIndVars;
	%if %length(&ListIndVars &IndVars)=0 %then %do;
		%put ERROR: Individual variables must be specified in either the Individual Variables or List-Style Specification of Individual Variables parameters.;
		%let exiterror=1;
		%goto exit;
	%end;

	**S1121578: check if cross type is either panel or bc, all others require parent ID variables*;
	%if ^%index(%upcase(&CrossType),PANEL) and ^%index(%upcase(&CrossType),BC) and (%length(&Parent1Var) = 0 and %length(&Parent2Var) = 0) %then %do;
		%put ERROR: Parent 1 ID Variable or Parent 2 ID Variable must be specified when Cross Type is &CrossType.. Please, return to the General Tab and specify one or both Parent ID Variables.;
		%let exiterror=1;
		%goto exit;
	%end;

	%if %length(&Parent1Var) > 0 and %length(&Parent2Var) > 0 %then %do;
		%if (%upcase(&Parent1Var) = %upcase(&Parent2Var)) %then %do;
			%put ERROR: Parent 1 ID Variable and Parent 2 ID Variable cannot be equal to each other. Please, return to the General Tab and change these two parameters.;
			%let exiterror=1;
			%goto exit;
		%end;
	%end;
		
	**S1123676: check if IDVar is a character variable to avoid errors.*;
	**Checking IDVar variable type*;
	%let IDVarType=;
	%VarAttributes(_Ds=InLib.&InData,_Var=&IDVar,_Len=,_Type=IDVarType);
	/*
	**Commented out when pushing a fix to defect S1144206.*;
	%if &IDVarType ^= C %then %do;
		%put ERROR: Marker ID Variable (&IDVar) must be a character variable. Please, return to the General Tab and specify another character variable.;
		%let exiterror=1;
		%goto exit;
	%end;	
	*/

	**S1144206: making sure that &IDVar and &AnnoMName variables are of same type.*;
	%if %length(&AnnoInData) %then %do;
		%let AnnoMNameType=;
		%VarAttributes(_Ds=AnnoLib.&AnnoData,_Var=&AnnoMName,_Len=,_Type=AnnoMNameType);
		%put AnnoMNameType= &AnnoMNameType;
		%if &IDVarType ^= &AnnoMNameType %then %do;
			%put ERROR: Marker ID Variable (&IDVar) from the Input SAS Data Set and Marker Name Variable (&AnnoMName) from the Annotation SAS Data Set must be off same type, i.e., both character or both numerical variables.;
			%let exiterror=1;
			%goto exit;
		%end;				
	%end;
		
	**applying where statement to InData*;		
	%let SubIndata= SubInData;
	data &subIndata;
		set InLib.&InData;
		OrigMarkerOrder= _n_;
	run;
	
	%if %length(&wherestmt) %then %do;
		data &SubIndata;
			set &SubIndata;
			where &wherestmt;
		run;
		
		%NObsVars(&SubIndata);
		%if &nobs < 1 %then %do;
			%put ERROR: &nobs observations left in the Input Data Set after applying the Where Statement specified in the General Tab. At least 1 observations must remain in the Input Data Set after applying the Where Statement.; 
			%let exiterror=1;
			%goto exit;
		%end;
	%end;
	
	%if %length(&AnnoInData) and %length(&AnnoWhereStmt) %then %do;
	
		%let SubAnnodata= SubAnnoData;
		data &SubAnnodata;
			set AnnoLib.&AnnoData;
			where &AnnoWhereStmt;
		run;
		/*data outlib.SubAnnoData; set &SubAnnoData; run;*/
		
		proc sql UNDO_POLICY= NONE;
			create table sub2Indat(drop= _AnnoMName) as
			select a1.*, a2.*
			from 
			&SubIndata as a1
			right join 
			&SubAnnodata(keep=&AnnoMName rename=(&AnnoMName=_AnnoMName)) as a2
			%if &IDVarType = C %then %do;
				on upcase(a1.&IDVar)= upcase(a2._AnnoMName)
			%end;
			%else %do;
				on a1.&IDVar = a2._AnnoMName
			%end;
			order by OrigMarkerOrder;
		quit;
		/*data outlib.sub2InDat; set sub2Indat; run;*/
		
		%NObsVars(&SubIndata);
		%let nobs_in= &nobs;
		%NObsVars(sub2Indat);
		%let nobs_out= &nobs;

		%if &nobs_in = &nobs_out %then %do;
			%put ERROR: No observation was removed from the Input Data Set when the Where Statement specified in the Annotation Tab was applyied to this data set. Please, return to the Annotation Tab and change the Where Statement that is being specified.; 
			%let exiterror=1;
			%goto exit;
		%end;

		%if &nobs_out < 1 %then %do;
			%put ERROR: There were &nobs observations left in the Input Data Set after applying the Where Statement specified in the Annotation Tab. At least 1 observations must remain in the Input Data Set after applying the Where Statement.; 
			%let exiterror=1;
			%goto exit;
		%end;
				
		%let SubIndata= sub2Indat;
	%end;
	/*data outlib.SubInData; set &SubIndata; run;*/
	
	**get number of individuals*;
	data indvnames;
		set &subIndata;
		where 0;
		keep &IndVars &ListIndVars;
	run;
	
	%let _did_= %sysfunc(open(indvnames));
	%let _nind_= %sysfunc(attrn(&_did_, nvars));
	%let fmvar = %sysfunc(varname(&_did_,1));
	%let lmvar = %sysfunc(varname(&_did_,&_nind_));
	%put _nind_= &_nind_;
	
	%if %length(&ListIndVars) %then %do;
		%let OrigListIndVars= &ListIndVars;
		%let OrigIndVars= &IndVars;
		%let ListIndVars= &fmvar.--&lmvar;	
		%let IndVars= &fmvar.--&lmvar;
		%put IndVars= &IndVars;
	%end;
		
	**creating lists of names of individuals*;
	%let nIndBlock= 0;
	%let TmpVars= ;
	%do k=1 %to &_nind_;
		%if %length(&TmpVars) le 32700 %then %do;
			%let TmpVars = &TmpVars %sysfunc(varname(&_did_,&k));
		%end;
		%else %do;
			%let nIndBlock= %sysevalf(&nIndBlock + 1);
			%let TmpIndVars&nIndBlock = &TmpVars %sysfunc(varname(&_did_,&k));
			%let TmpVars = ;
		%end;
	%end;
	%if (&nIndBlock = 0) or (&nIndBlock > 0 and %length(&TmpVars) > 0) %then %do;
		%let nIndBlock= %sysevalf(&nIndBlock + 1);
		%let TmpIndVars&nIndBlock = &TmpVars;
		%let TmpVars = ;
	%end;
	%let _rc_= %sysfunc(close(&_did_));
	%put nIndBlock= &nIndBlock;
	%*put TmpIndVars&nIndBlock= &&TmpIndVars&nIndBlock;
	
	/*S1364776: lists of names of individuals to drop when keepvars is not empty*/
	%let DropIndBlock= 0;
	%let DropTmpVars= ;
	%if %length(&KeepVars) %then %do;
		proc transpose data=indvnames out=indvnamest(rename=(_name_=_indvars_));
			var &IndVars &ListIndVars;
		run;

		data keepvarsdat;
			set &subIndata;
			where 0;
			keep &KeepVars;
		run;

		proc transpose data=keepvarsdat out=keepvarst(rename=(_name_= _keepvars_));
			var &KeepVars;
		run;

		proc sql Undo_Policy= NONE;
			create table leftind as
			select a.*, b.*
			from indvnamest as a left join keepvarst as b
			on a._indvars_=b._keepvars_;
		quit;

		data dropind;
			set leftind;
			where missing(_keepvars_);
			label ;
		run;

		%let _id_= %sysfunc(open(dropind));
		%let _nid_= %sysfunc(attrn(&_id_, nobs));
		%syscall set(_id_);
		%do k=1 %to &_nid_;
			%let b= %sysfunc(fetchobs(&_id_,&k));
			%*put %length(&DropTmpVars);
			%if %length(&DropTmpVars) le 32700 %then %do;
				%*put k1= &k;
				%let DropTmpVars = &DropTmpVars &_indvars_;
			%end;
			%else %do;
				%*put k2= &k;
				%let DropIndBlock= %sysevalf(&DropIndBlock + 1);
				%let DropTmpIndVars&DropIndBlock = &DropTmpVars &_indvars_;
				%let DropTmpVars = ;
			%end;						
		%end;
		%if (&DropIndBlock = 0) or (&DropIndBlock > 0 and %length(&DropTmpVars) > 0) %then %do;
			%*put here;
			%let DropIndBlock= %sysevalf(&DropIndBlock + 1);
			%let DropTmpIndVars&DropIndBlock = &DropTmpVars;
			%let DropTmpVars = ;
		%end;
		%let _ci_= %sysfunc(close(&_id_));
		%put DropIndBlock= &DropIndBlock;
		%put DropTmpIndVars&DropIndBlock= &&DropTmpIndVars&DropIndBlock;
	%end;	
	
	*creating a data set with individual variables (will be rows) to join with genotype data set*;
	proc transpose data=indvnames out=indvnamest(keep=_name_);
		var &IndVars;
	run;

	**creating lists of names of recoded individuals*;
	%let MaxLen= 30;	
	%let Flag_IndLen= 0;
	data indvnamest;
		set indvnamest;
		if length(trim(left(_name_))) > &MaxLen then do; 
			_newname_= substr(trim(left(_name_)), 1, &MaxLen);
			call symputx("Flag_IndLen", 1);
		end;
		else do;
			_newname_= trim(left(_name_));
		end;
		_order_= _n_;
	run;
	
	%if &Flag_IndLen %then %do;
		proc sort data=indvnamest out=indvnamest;
			by _name_;
		run;
		
		data indvnamest;
			set indvnamest;
			if _newname_= lag(_newname_) then do;
				_newname_= substr(_newname_, 1, &MaxLen - length(trim(left(_order_))) - 1)||"_"||trim(left(_order_));
				_dup_= 1;
			end;
		run;
	
		/*data OutLib.indvnamest; set indvnamest; run;*/
		
		proc sort data=indvnamest out= indvnamest(drop= _order_ _dup_);
			by _order_;
		run;
			
		proc transpose data= indvnamest(drop=_name_) out= indvnames(drop=_name_ where= (0));
			var _newname_;
			id _newname_;
		run;
	%end;
	
	/*data OutLib.indvnames2; set indvnames; run;*/
	
	%let _did_= %sysfunc(open(indvnames));
	%let _nind_= %sysfunc(attrn(&_did_, nvars));
		
	%let nRecIndBlock= 0;
	%let TmpVars= ;
	%do k=1 %to &_nind_;
		%if %length(&TmpVars) le 32700 %then %do;
			%let TmpVars = &TmpVars r_%sysfunc(varname(&_did_,&k));
		%end;
		%else %do;
			%let nRecIndBlock= %sysevalf(&nRecIndBlock + 1);
			%let RecIndVars&nRecIndBlock = &TmpVars r_%sysfunc(varname(&_did_,&k));
			%let TmpVars = ;
		%end;
	%end;
	%if (&nRecIndBlock = 0) or (&nRecIndBlock > 0 and %length(&TmpVars) > 0) %then %do;
		%let nRecIndBlock= %sysevalf(&nRecIndBlock + 1);
		%let RecIndVars&nRecIndBlock = &TmpVars;
		%let TmpVars = ;
	%end;
	%let _rc_= %sysfunc(close(&_did_));
	%put nRecIndBlock= &nRecIndBlock;
	%*put RecIndVars&nRecIndBlock= &&RecIndVars&nRecIndBlock;
	
	%let k= 1;
	%do i=1 %to &nRecIndBlock;
		%let recodedIndVar&i = ;
		%let j= 1;
		%let recVar= %scan(&&RecIndVars&i, &j);
		%do %while(%length(&recVar));	
			%let recodedIndVar&i = &&recodedIndVar&i _nf_&k= &recVar;
			%let j= %sysevalf(&j + 1);
			%let recVar= %scan(&&RecIndVars&i, &j);
			%let k= %sysevalf(&k + 1);
		%end;
	%end;
	%*put recodedIndVar&nRecIndBlock= &&recodedIndVar&nRecIndBlock;
	
	%*let exiterror= 1;
	%*goto exit;
		
	**preparing some output variables*;
	%if %length(&Parent1Var) %then %do;
		%let newP1Var= new_&Parent1Var;
		%let recP1Var= rec_&Parent1Var; 
		%let newP1Lab= New &Parent1Var;
		%let recP1Lab= Recoded &Parent1Var; 
	%end;
	%else %do;
		%let newP1Var= _newParent1_;
		%let recP1Var= _recParent1_; 
		%let newP1Lab= New Parent 1;
		%let recP1Lab= Recoded Parent 1; 
	%end;
	%if %length(&Parent2Var) %then %do;
		%let newP2Var= new_&Parent2Var;
		%let recP2Var= rec_&Parent2Var; 
		%let newP2Lab= New &Parent2Var;
		%let recP2Lab= Recoded &Parent2Var; 
	%end;
	%else %do;
		%let newP2Var= _newParent2_;
		%let recP2Var= _recParent2_;
		%let newP2Lab= New Parent 2;
		%let recP2Lab= Recoded Parent 2;
	%end;
	
	**maxrf is used when testing quality of unphased markers in OUTCROSS, SFN, RFN, RI0, RI1, RI2, and IRIL;
	%let maxrf= 0.35; 
	**minrf is used when testing quality of unphased markers in OUTCROSS, SFN, RFN, RI0, RI1, RI2, and IRIL;
	%let minrf= 0.00; 
	**rfdif is used to untie D1 and D2 markers in OUTCROSS;
	%global rfdif;
	%let rfdif= 0.10;
	
	**p_alpha is used to selected markers in outcross data;
	%let p_alpha= &Alpha;
	**end: processing Input Data Set and Annotation Data Set*;
	
	**Prep output data set names*;
	%GLOBAL TasselGBSImport_OutData TasselGBSImportGen_Outdata TasselGBSImportAnn_Outdata TasselGBSImportSeg_OutData ;
	%let TasselGBSImport_OutData = ;
	%let TasselGBSImportGen_OutData = ;
	%let TasselGBSImportSeg_OutData = ;
	%let TasselGBSImportAnn_Outdata = ;
	%PrepOutDS(TasselGBSImport_OutData,TasselGBSImport_OutDataName,&Outname,_ori);
	%PrepOutDS(TasselGBSImportGen_OutData,TasselGBSImportGen_OutDataName,&Outname,_gen);
	%PrepOutDS(TasselGBSImportAnn_OutData,TasselGBSImportAnn_OutDataName,&Outname,_ann);
	%PrepOutDS(TasselGBSImportSeg_OutData,TasselGBSImportSeg_OutDataName,&Outname,_seg);
	
	**MissCodedMarker indicates if there is any misscoced parental marker genotype in the input data*;
	%let MissCodedMarker=;
	%let MissCodedP1=;
	%let MissCodedP2=;
	
	%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
	
		proc iml;
			start autogametes(ploid,_dom);
				m= ploid/2;
				r= _dom;
				ub_dom= min(m, _dom);
				lb_dom= max(0, _dom - m);
				ngam= ub_dom - lb_dom + 1;
				gam= J(ngam, 3, 0);
				k=0;
				do i=1 to ngam;
					s= ub_dom - k;
					gam[i,1]= s;
					gam[i,2]= m - s;
					gam[i,3]= (fact(m)**2/fact(2*m)) * ( fact(2*m-r)*fact(r) / (fact(m-s)*fact(m-r+s)*fact(s)*fact(r-s)) );
					k=k+1;
				end;
				create outdata from gam;
				append from gam;
				*print(ngam);
				*print(gam);
			finish autogametes;
			store module=(autogametes);
		quit;
	
		**beging: macro to generate gametes of autopolyploid with even ploid level (2m= 2, 4, 6, ...)*;
		%macro AutoGametes(_dom=, _rec=);
			%local ploidlevel;
			%let ploidlevel= %sysevalf(&_dom + &_rec);

			proc datasets nolist;
				delete autogamdata;
			quit;

			proc iml;
				load module=(autogametes);
				run autogametes(&ploidlevel,&_dom);
				use outdata;
				read all into outdata_tmp;
				create autogamdata from outdata_tmp[colname= {"_dom_A" "_rec_a" "prob"}];
				append from outdata_tmp;
				close outdata;
			quit;

			proc datasets nolist;
				delete outdata;
			quit;

			proc datasets nolist;
				delete autogamdata&ploidlevel._&_dom;
				change autogamdata= autogamdata&ploidlevel._&_dom;
			quit;

			%exitAutoGametes:;
		%mend AutoGametes;

		**beging: macro to get segregation ratios of GBS markers on autopolyploid with even ploid level (2m= 2, 4, 6, ...)*;
		%macro AutoPolySegrRatio(_dom1=, _rec1=, _dom2=, _rec2=, gbs=1);
			%local ploidlevel1 ploidlevel2;
			%let ploidlevel1= %sysevalf(&_dom1 + &_rec1);
			%let ploidlevel2= %sysevalf(&_dom2 + &_rec2);

			%if &ploidlevel1 ^= &ploidlevel2 %then %do;
				%put ERROR: Ploid level for parent 1 must be equal to parent 2.;
				%let exiterror = 1;
				%goto exitAuto;
			%end;

			%AutoGametes(_dom=&_dom1, _rec=&_rec1);
			proc datasets nolist;
				delete gametesP1;
				change autogamdata&ploidlevel1._&_dom1= gametesP1;
			quit;		

			%AutoGametes(_dom=&_dom2, _rec=&_rec2);
			proc datasets nolist;
				delete gametesP2;
				change autogamdata&ploidlevel2._&_dom2= gametesP2;
			quit;

			proc sql UNDO_POLICY= NONE;
				create table genoratios_auto as
				select f1.*, f2.*,
				sum(f1._dom_A1, f2._dom_A2) as _dom_A, 
				sum(f1._rec_A1, f2._rec_A2) as _rec_a, 
				(f1.prob1 * f2.prob2) as prob
				from 
				gametesP1(rename = (_dom_A=_dom_A1 _rec_a=_rec_a1 prob=prob1)) as f1 
				cross join 
				gametesP2(rename = (_dom_A=_dom_A2 _rec_a=_rec_a2 prob=prob2)) as f2
				order by _dom_A descending
			;
			quit;

			data genoratios_auto;
				set genoratios_auto(rename=(prob=prob_));
				by descending _dom_A;
				if first._dom_A then do;
					prob=prob_;
					first_=1;
				end;
				else do;
					prob= prob + prob_;
					first_=0;
				end;
				%if &gbs %then %do;
					if _dom_A = 0 or _rec_a = 0 then do;
						*probcum= 0;
						gbsclass= _n_;
					end;
					else do;
						*probcum= probcum + prob_;
						gbsclass= 0;
					end;
					*retain prob probcum;
				%end;
				retain prob;
				if last._dom_A then output;
				keep _dom_A _rec_a prob %if &gbs %then gbsclass;;
			run;

			%if &gbs %then %do;
				proc sort data=genoratios_auto;
					by gbsclass;
				run;

				data genoratios_auto;
					set genoratios_auto(rename=(prob=prob_));
					by gbsclass;
					if first.gbsclass then do;
						prob=prob_;
						first_=1;
					end;
					else do;
						prob= prob + prob_;
						first_=0;
					end;
					retain prob;
					if last.gbsclass then output;
					keep _dom_A _rec_a prob;
				run;

				proc sort data=genoratios_auto;
					by descending _dom_A;
				run;
			%end; 

			proc datasets nolist;
				delete genoratios_auto&ploidlevel1._&_dom1._&_dom2;
				change genoratios_auto= genoratios_auto&ploidlevel1._&_dom1._&_dom2;
			quit;

			%exitAuto:;
		%mend AutoPolySegrRatio;

		**beging: macro to generate gametes of allopolyploid with even ploid level (2m= 2, 4, 6, ...)*;
		**This code does not produce the full spectrum of all gametic classes, as multiple classes of**; 
		**heterozygous gametes will be collapsed into a single class reflecting the inability of GBS **;
		**markers to distinct allele dosages. **;
		%macro AlloGametes(_dom=, _rec=);
			%global exiterror;
			%let exiterror = 0;
		
			%local ploidlevel;
			%let ploidlevel= %sysevalf(&_dom + &_rec);

			proc datasets nolist;
				delete allogamdata;
			quit;
			
			%if &_dom <= %sysevalf(&ploidlevel/2) %then %do;
				%if &_dom= 0 %then %do;
					%let p1= 0.0; 
					%let p2= 0.0;
					%let p3= 1.0;			
				%end;
				%else %if &_dom= 1 %then %do;
					%let p1= 0.0; 
					%let p2= 0.5;
					%let p3= 0.5;			
				%end;
				%else %if &_dom= 2 %then %do;
					%let p1= %sysevalf((1/2)**2); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;	
				%end;
				%else %if &_dom= 3 %then %do;
					%let p1= %sysevalf((1/2)**4); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;	
				%else %if &_dom= 4 %then %do;
					%let p1= %sysevalf((1/2)**6); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 5 %then %do;
					%let p1= %sysevalf((1/2)**8); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;
				%else %if &_dom= 6 %then %do;
					%let p1= %sysevalf((1/2)**10); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;
				%else %if &_dom= 7 %then %do;
					%let p1= %sysevalf((1/2)**12); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 8 %then %do;
					%let p1= %sysevalf((1/2)**14); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 9 %then %do;
					%let p1= %sysevalf((1/2)**16); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;
				%else %if &_dom= 10 %then %do;
					%let p1= %sysevalf((1/2)**18); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 11 %then %do;
					%let p1= %sysevalf((1/2)**20); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 12 %then %do;
					%let p1= %sysevalf((1/2)**22); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 13 %then %do;
					%let p1= %sysevalf((1/2)**24); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 14 %then %do;
					%let p1= %sysevalf((1/2)**26); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;
				%else %if &_dom= 15 %then %do;
					%let p1= %sysevalf((1/2)**28); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 16 %then %do;
					%let p1= %sysevalf((1/2)**30); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;
				%else %if &_dom= 17 %then %do;
					%let p1= %sysevalf((1/2)**32); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 18 %then %do;
					%let p1= %sysevalf((1/2)**34); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;			
				%else %if &_dom= 19 %then %do;
					%let p1= %sysevalf((1/2)**36); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;
				%else %if &_dom= 20 %then %do;
					%let p1= %sysevalf((1/2)**38); 
					%let p2= %sysevalf(1 - 2 * &p1);
					%let p3= &p1;			
				%end;
				%*put p1 p2 p3 &p1 &p2 &p3;
			%end;
			%else %do;
				%put ERROR: GBS marker not usefull.;
				%let exiterror= 1;
				%goto exitAlloGametes;
			%end;
			
			proc sql;
				create table allogamdata
				(_dom_A Num(8), _rec_a Num(8), prob Num(8));
			quit;
			
			proc sql;
				insert into allogamdata
				(_dom_A, _rec_a, prob)
				%if &_dom > 0 %then %do;
					%if &_dom < %sysevalf(&ploidlevel/2) %then %do;
						values(1, %sysevalf(&ploidlevel1/2 - 1), %sysevalf(&p1 + &p2))
						values(0, %sysevalf(&ploidlevel1/2), &p3)
					%end;
					%else %if &_dom = %sysevalf(&ploidlevel/2) %then %do;
						values(%sysevalf(&ploidlevel1/2), 0, &p1)
						values(1, %sysevalf(&ploidlevel1/2 - 1), &p2)
						values(0, %sysevalf(&ploidlevel1/2), &p3)
					%end;
				%end;
				%else %if &_dom = 0 %then %do;
					values(0, %sysevalf(&ploidlevel1/2), &p3)			
				%end;
				;
			quit;

			proc datasets nolist;
				delete allogamdata&ploidlevel._&_dom;
				change allogamdata= allogamdata&ploidlevel._&_dom;
			quit;

			%exitAlloGametes:;
		%mend AlloGametes;
		
		**beging: macro to get segregation ratios of GBS markers on allopolyploid with even ploid level (2m= 2, 4, 6, ...)*;
		%macro AlloPolySegrRatio(_dom1=,_rec1=,_dom2=,_rec2=, gbs=1);
			%global exiterror;
			%let exiterror = 0;
		
			%local ploidlevel1 ploidlevel2;
			%let ploidlevel1= %sysevalf(&_dom1 + &_rec1);
			%let ploidlevel2= %sysevalf(&_dom2 + &_rec2);

			%if &ploidlevel1 ^= &ploidlevel2 %then %do;
				%put ERROR: Ploid level for parent 1 must be equal to parent 2.;
				%let exiterror = 1;
				%goto exitAllo;
			%end;

			proc datasets nolist;
				delete genoratios_allo;
			quit;
			
			%AlloGametes(_dom=&_dom1, _rec=&_rec1);
			%if &exiterror %then %do;
				%goto exitAllo;
			%end;
			proc datasets nolist;
				delete gametesP1;
				change allogamdata&ploidlevel1._&_dom1= gametesP1;
			quit;		

			%AlloGametes(_dom=&_dom2, _rec=&_rec2);
			%if &exiterror %then %do;
				%goto exitAllo;
			%end;
			proc datasets nolist;
				delete gametesP2;
				change allogamdata&ploidlevel2._&_dom2= gametesP2;
			quit;

			proc sql UNDO_POLICY= NONE;
				create table genoratios_allo as
				select f1.*, f2.*,
				sum(f1._dom_A1, f2._dom_A2) as _dom_A, 
				sum(f1._rec_A1, f2._rec_A2) as _rec_a, 
				(f1.prob1 * f2.prob2) as prob
				from 
				gametesP1(rename = (_dom_A=_dom_A1 _rec_a=_rec_a1 prob=prob1)) as f1 
				cross join 
				gametesP2(rename = (_dom_A=_dom_A2 _rec_a=_rec_a2 prob=prob2)) as f2
				order by _dom_A descending
				;
			quit;

			data genoratios_allo;
				set genoratios_allo(rename=(prob=prob_));
				by descending _dom_A;
				if first._dom_A then do;
					prob=prob_;
					first_=1;
				end;
				else do;
					prob= prob + prob_;
					first_=0;
				end;
				%if &gbs %then %do;
					if _dom_A = 0 or _rec_a = 0 then do;
						*probcum= 0;
						gbsclass= _n_;
					end;
					else do;
						*probcum= probcum + prob_;
						gbsclass= 0;
					end;
					*retain prob probcum;
				%end;
				retain prob;
				if last._dom_A then output;
				keep _dom_A _rec_a prob %if &gbs %then gbsclass;;
			run;

			%if &gbs %then %do;
				proc sort data=genoratios_allo;
					by gbsclass;
				run;

				data genoratios_allo;
					set genoratios_allo(rename=(prob=prob_));
					by gbsclass;
					if first.gbsclass then do;
						prob=prob_;
						first_=1;
					end;
					else do;
						prob= prob + prob_;
						first_=0;
					end;
					retain prob;
					if last.gbsclass then output;
					keep _dom_A _rec_a prob;
				run;

				proc sort data=genoratios_allo;
					by descending _dom_A;
				run;
			%end; 

			proc datasets nolist;
				delete genoratios_allo&ploidlevel1._&_dom1._&_dom2;
				change genoratios_allo= genoratios_allo&ploidlevel1._&_dom1._&_dom2;
			quit;

			%exitAllo:;
		%mend AlloPolySegrRatio;

		**Creates a data set with segregation ratios for all GBS marker dosages*;
		%macro AppendSegrRatio(PloidList=, _PloidyType=);  

			%local suff_ i ploid_ id m j u k w;

			%if %upcase(&_PloidyType) = AUTOPOLYPLOID %then 
				%let suff_ = _auto;
			%else 
				%let suff_ = _allo;

			data GenoRatios_All;
				set _null_;
			run;

			%let i= 1;
			%let ploid_= %scan(&PloidList, &i);
			%do %while(%length(&ploid_));
				%if &ploid_ > 40 %then %do;
					%put ERROR: The maximum ploid level allowed is 40.;
				%end;

				%let id= 1;
				%let m= %sysevalf(&ploid_ / 2);
				%do j= 0 %to &m;
					%let u= %sysevalf(&ploid_ - &j);
					%do k=%sysfunc(max(&j, 1)) %to &m; 
						%let w= %sysevalf(&ploid_ - &k);

						%if %upcase(&_PloidyType) = AUTOPOLYPLOID %then %do;
							%AutoPolySegrRatio(_dom1=&j, _rec1=&u, _dom2=&k, _rec2=&w, gbs=1);
						%end;
						%else %do;
							%AlloPolySegrRatio(_dom1=&j, _rec1=&u, _dom2=&k, _rec2=&w, gbs=1);
						%end;

						data GenoRatios_All;
							retain ploid cross bigA_L1 bigA_L2 Dosage_L1 Dosage_L2;
							set GenoRatios_All genoratios&suff_&ploid_._&j._&k(in=_a_);
							if _a_ then do;
								ploid= &ploid_;
								cross= &id;
								bigA_L1= &j;
								bigA_L2= &k;
								Dosage_L1= 'D'||trim(left(bigA_L1));
								Dosage_L2= 'D'||trim(left(bigA_L2));
							end;
						run;
						%let id= %sysevalf(&id + 1);
					%end;
				%end;
				%let i= %sysevalf(&i + 1);
				%let ploid_= %scan(&PloidList, &i);
			%end;

			%exitAppend:;
		%mend AppendSegrRatio;

		**Creates macro variables holding segregation ratios for all GBS marker dosages*;
		%macro CreateSegrMacroVars;

			data _null_;
				set  GenoRatios_All end=_eof_;
				call symputx("cross_"||trim(left(_n_)), trim(left(cross)));
				call symputx("DosL1_"||trim(left(_n_)), trim(left(Dosage_L1)));
				call symputx("DosL2_"||trim(left(_n_)), trim(left(Dosage_L2)));
				call symputx("prob_"||trim(left(_n_)), trim(left(prob)));
				if _eof_ then do;
					call symputx("ncross", trim(left(cross)));
					call symputx("nobs", trim(left(_n_)));
				end;
			run;

			%local cross_ cross_f probs_tmp k j i ii;
			%let cross_f = &cross_1;
			%let probs_tmp= ;
			%let k=1;
			%let j=0;
			%do i=1 %to &nobs;
				%let cross_s= &&cross_&i;
				%if &cross_s = &cross_f %then %do;
					%let probs_tmp= &probs_tmp &&prob_&i;
					%let j= %sysevalf(&j + 1);
				%end;
				%else %do;
					%global probs&k dosage&k;

					%if &j = 2 %then %do;
						%let probs&k= 0 &probs_tmp;
					%end;
					%else %do;
						%let probs&k= &probs_tmp;
					%end;
					%let ii= %sysevalf(&i-1);
					%let dosage&k= &&DosL1_&ii..x&&DosL2_&ii;
					%let j=1;
					%let probs_tmp= &&prob_&i;
					%let k= %sysevalf(&k + 1);
				%end;
				%if &i = &nobs %then %do;
					%global probs&k dosage&k;

					%let dosage&k= &&DosL1_&i..x&&DosL2_&i;

					%if &j = 2 %then %do;
						%let probs&k= 0 &probs_tmp;
					%end;
					%else %do;
						%let probs&k= &probs_tmp;
					%end;
				%end;
				%let cross_f= &cross_s;
			%end;

			%do i=1 %to &k;
				%put dosage&i= &&dosage&i;
				%put probs&i= &&probs&i;
			%end;

			%exitCreate:;
		%mend CreateSegrMacroVars;
	
		%let m_= %sysevalf(&PloidyLevel/2);

		%put PloidyType= &PloidyType; 
		%AppendSegrRatio(PloidList= &PloidyLevel, _PloidyType= &PloidyType);
		%if &syserr>4 %then %do;
			%put ERROR: Error in the segregation ratio macro.;
			%let exiterror= 1;
			%goto exit;
		%end;

		%CreateSegrMacroVars;
		%if &syserr>4 %then %do;
			%put ERROR: Error in the create segregation ratio variables macro.;
			%let exiterror= 1;
			%goto exit;
		%end;
		
		%*goto exit;
		
		**start code to transcode Tassel-GBS*;
		data _recodedData;
			set &subIndata;
			*defining temporary arrays*;
			array _p_{2} $5;
			%if %length(&Parent1Var) %then %do;
				_p_[1]= upcase(trim(left(&Parent1Var)));
			%end;
			%else %do;
				_p_[1]= 'N';
			%end;
			%if %length(&Parent2Var) %then %do;
				_p_[2]= upcase(trim(left(&Parent2Var)));
			%end;
			%else %do;
				_p_[2]= 'N';
			%end;
			*put _n_ _p_[1] _p_[2];
			if missing(_p_[1]) then _p_[1]= 'N';
			if missing(_p_[2]) then _p_[2]= 'N';
			if ^indexc(trim(left(_p_[1])), 'A','C','G','T','M','R','W','S','Y','K','N') or ^indexc(trim(left(_p_[2])), 'A','C','G','T','M','R','W','S','Y','K','N') then do;
				call symputx("MissCodedMarker", trim(left(&IDVar)));
				if ^indexc(trim(left(_p_[1])), 'A','C','G','T','M','R','W','S','Y','K','N') then do;
					call symputx("MissCodedP1", trim(left(_p_[1])));
				end;
				if ^indexc(trim(left(_p_[2])), 'A','C','G','T','M','R','W','S','Y','K','N') then do;
					call symputx("MissCodedP2", trim(left(_p_[2])));
				end;
			end;
			array _o_{*} &IndVars;
			array _ne_(&_nind_) $ 2;
			array _nf_(&_nind_) $ 2;
			**defining new variables*;
			_a=0;_c=0;_g=0;_t=0;_m=0;_r=0;_w=0;_s=0;_y=0;_k=0;_n=0;
			_offsA='   ';
			_offsC='   ';
			_offsG='   ';
			_offsT='   ';
			_offsM='   ';
			_offsR='   ';
			_offsW='   ';
			_offsS='   ';
			_offsY='   ';
			_offsK='   ';
			_offsexp='                                   ';
			_offsobs='                                   ';
			_predGenoP1='    ';
			_predGenoP2='    ';
			_newGenoP1='    ';
			_newGenoP2='    ';
			_tot_ind=0;
			**marker type call based on parents genotypes*;
			_tot_testI=0;
			_AA_testI=0;_AB_testI=0;_BB_testI=0;_mis_testI=0;_genotype_error_testI=0;_p_genotype_error_testI= 0;
			%do j=1 %to %sysevalf(&ncross);
				_chisq_testI_&&Dosage&j= .;
				_pchisq_testI_&&Dosage&j= .;
				_indicator_testI_&&Dosage&j= 0;
				_mktype_testI_&&Dosage&j= '                 ';
			%end;
			_mktype_testI= '                 ';
			_chisq_testI= .;
			_pchisq_testI= .;
			_sum_indicator_testI= 0;
			_recP1exp= '     ';
			_recP2exp= '     ';
			*****;
			**marker type call based on offspring genotypes*;
			_tot_testII=0;
			_AA_testII=0;_AB_testII=0;_BB_testII=0;_mis_testII=0;
			%do j=1 %to %sysevalf(&ncross);
				_chisq_testII_&&Dosage&j= .;
				_pchisq_testII_&&Dosage&j= .;
				_indicator_testII_&&Dosage&j=0 ;
				_mktype_testII_&&Dosage&j= '                 ';
			%end;
			_deliberate = 0;
			_mktype_testII= '                 ';
			_chisq_testII= .;
			_pchisq_testII= .;
			_sum_indicator_testII= 0;
			_recP1obs= '     ';
			_recP2obs= '     ';
			*****;
			**final calls*;
			_newP1='   ';
			_newP2='   ';
			_recP1='   ';
			_recP2='   ';
			_testI_OR_testII='  ';
			_uniquedosage= 0;
			_singledosage= 0;
			_mktype='                 ';
			_selected= 0;
			_n_ind=.;
			_n_mis=.;
			_n_err=.;
			_p_err=.;
			_n_dom=.;
			_n_het=.;
			_n_rec=.;		
			_chisq= .;
			_pchisq= .;		
			*****;
			do i=1 to dim(_o_);
				_o_[i]= upcase(trim(left(_o_[i])));
				_ne_[i]= _o_[i];
				if _o_[i]= 'A' then do;
					_a=_a + 1;
					_offsA= 'A';
				end;
				else if _o_[i]= 'C' then do;
					_c=_c + 1;
					_offsC= 'C';
				end;
				else if _o_[i]= 'G' then do;
					_g=_g + 1;
					_offsG= 'G';
				end;
				else if _o_[i]= 'T' then do;
					_t=_t + 1;
					_offsT= 'T';
				end;
				else if _o_[i]= 'M' then do;
					_m=_m + 1;
					_offsM= 'M';
				end;
				else if _o_[i]= 'R' then do;
					_r=_r + 1;
					_offsR= 'R';
				end;
				else if _o_[i]= 'W' then do;
					_w=_w + 1;
					_offsW= 'W';
				end;
				else if _o_[i]= 'S' then do;
					_s=_s + 1;
					_offsS= 'S';
				end;
				else if _o_[i]= 'Y' then do;
					_y=_y + 1;
					_offsY= 'Y';
				end;
				else if _o_[i]= 'K' then do;
					_k=_k + 1;
					_offsK= 'K';
				end;
				else _n=_n + 1;
			end;
			*_tot_testII= _a +_c +_g + _t + _m + _r + _w + _s + _y + _k;
			_tot_ind= _a +_c +_g + _t + _m + _r + _w + _s + _y + _k + _n;
			_offsobs=_offsA||_offsC||_offsG||_offsT||_offsM||_offsR||_offsW||_offsS||_offsY||_offsK;
			_offsobs=upcase(COMPRESS(_offsobs));
			*****;
			**begin predict parents genotypes based on offspring genotypes*;
			select;
				when(_offsobs= 'A') do;
					*_predGenoP1='A';
					_predGenoP2='A';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'C') do;
					*_predGenoP1='C';
					_predGenoP2='C';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'G') do;
					*_predGenoP1='G';
					_predGenoP2='G';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'T') do;
					*_predGenoP1='T';
					_predGenoP2='T';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'M') do;
					*_predGenoP1='A';
					_predGenoP2='C';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'R') do;
					*_predGenoP1='A';
					_predGenoP2='G';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'W') do;
					*_predGenoP1='A';
					_predGenoP2='T';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'S') do;
					*_predGenoP1='C';
					_predGenoP2='G';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'Y') do;
					*_predGenoP1='C';
					_predGenoP2='T';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'K') do;
					*_predGenoP1='G';
					_predGenoP2='T';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'N') do;
					*_predGenoP1='';
					_predGenoP2='';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'AM') do;
					*_predGenoP1='A';
					_predGenoP2='M';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'AR') do;
					*_predGenoP1='A';
					_predGenoP2='R';	
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'AW') do;
					*_predGenoP1='A';
					_predGenoP2='W';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'CM') do;
					*_predGenoP1='C';
					_predGenoP2='M';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end; 
				when(_offsobs= 'CS') do;
					*_predGenoP1='C';
					_predGenoP2='S';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end; 
				when(_offsobs= 'CY') do; 
					*_predGenoP1='C';
					_predGenoP2='Y';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'GR') do; 
					*_predGenoP1='G';
					_predGenoP2='R';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'GS')do; 
					*_predGenoP1='G';
					_predGenoP2='S';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end; 
				when(_offsobs= 'GK') do; 
					*_predGenoP1='G';
					_predGenoP2='K';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'TW') do;
					*_predGenoP1='T';
					_predGenoP2='W';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end; 
				when(_offsobs= 'TY') do; 
					*_predGenoP1='T';
					_predGenoP2='Y';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'TK') do;
					*_predGenoP1='T';
					_predGenoP2='K';
					/*%if &PloidyLevel > 2 %then*/
						_predGenoP1= _predGenoP2;
				end;
				when(_offsobs= 'ACM') do;
					_predGenoP1='M';
					_predGenoP2='M';
				end;
				when(_offsobs= 'AGR') do;
					_predGenoP1='R';
					_predGenoP2='R';
				end;
				when(_offsobs= 'ATW') do;
					_predGenoP1='W';
					_predGenoP2='W';
				end;
				when(_offsobs= 'CGS') do;
					_predGenoP1='S';
					_predGenoP2='S';
				end;
				when(_offsobs= 'CTY') do;
					_predGenoP1='Y';
					_predGenoP2='Y';
				end;
				when(_offsobs= 'GTK') do;
					_predGenoP1='K';
					_predGenoP2='K';
				end;
				**The following cases are very unlikely to be observed, and when observed, they might be due to genotyping calling error*;				
				when(_offsobs= 'AC') do;
					_predGenoP1='M';
					_predGenoP2='M';
				end;
				when(_offsobs= 'AG') do;
					_predGenoP1='R';
					_predGenoP2='R';
				end;
				when(_offsobs= 'AT') do;
					_predGenoP1='W';
					_predGenoP2='W';
				end;
				when(_offsobs= 'CG') do;
					_predGenoP1='S';
					_predGenoP2='S';
				end;
				when(_offsobs= 'CT') do;
					_predGenoP1='Y';
					_predGenoP2='Y';
				end;
				when(_offsobs= 'GT') do;
					_predGenoP1='K';
					_predGenoP2='K';
				end;
				otherwise do;
					put 'Impossible Cross for Loci=' &IDVar 'P1=' _p_[1] 'P2=' _p_[2] 'Offspring=' _offsobs;
					_predGenoP1='';
					_predGenoP2='';
				end;
			end;
			*****;
			**begin assiging new genotypes to parents based on observed offspring genotypes*;
			if _p_[1]= 'N' and _p_[2]= 'N' then do;
				**both parent have missing genotype. Deliberately assign genotype to parents*;
				_newGenoP1= _predGenoP1; 
				_newGenoP2= _predGenoP2;
			end;
			else if _p_[1]= 'N' and _p_[2] ^= 'N' then do;
				**parent 1 has missing genotype*;
				if _p_[2]= _predGenoP1 then do;
					_newGenoP1= _predGenoP2;
					_newGenoP2= _p_[2];
				end;
				else if _p_[2]= _predGenoP2 then do;
					_newGenoP1= _predGenoP1;
					_newGenoP2= _p_[2];
				end; 
				else do;
					**Deliberately assign genotype to parents*;
					_newGenoP1= _predGenoP1;
					_newGenoP2= _predGenoP2;
				end;
			end;
			else if _p_[1] ^= 'N' and _p_[2]= 'N' then do;
				**parent 2 has missing genotype. *;
				if _p_[1]= _predGenoP1 then do;
					_newGenoP1= _p_[1];
					_newGenoP2= _predGenoP2;
				end;
				else if _p_[1]= _predGenoP2 then do;
					_newGenoP1= _p_[1];
					_newGenoP2= _predGenoP1;
				end; 
				else do;
					**Deliberately assign genotype to parents*;
					_newGenoP1= _predGenoP1;
					_newGenoP2= _predGenoP2;
				end;
			end;
			else do;
				**neither parent has missing genotype. *;
				if _p_[1] ^= _p_[2] then do;
					if _p_[1]= _predGenoP1 then do;
						_newGenoP1= _p_[1];
						_newGenoP2= _predGenoP2;
					end;
					else if _p_[1]= _predGenoP2 then do;
						_newGenoP1= _p_[1];
						_newGenoP2= _predGenoP1;
					end; 
					else if _p_[2]= _predGenoP1 then do;
						_newGenoP1= _predGenoP2;
						_newGenoP2= _p_[2];
					end;
					else if _p_[2]= _predGenoP2 then do;
						_newGenoP1= _predGenoP1;
						_newGenoP2= _p_[2];
					end;
					else do;
						**Deliberately assign genotype to parents*;
						_newGenoP1= _predGenoP1;
						_newGenoP2= _predGenoP2;
					end;
				end;
				else do;
					**Deliberately assign genotype to parents*;
					_newGenoP1= _predGenoP1;
					_newGenoP2= _predGenoP2;
				end;
			end;
			*****;
			**begin genotypic counting based on predicted parental genotypes*;
			array _nptmp_{*} _newGenoP1 _newGenoP2;
			_nptmp_[1]= upcase(trim(left(_nptmp_[1])));
			_nptmp_[2]= upcase(trim(left(_nptmp_[2])));
			*put 'nptmp ' _n_ _nptmp_[1] _nptmp_[2];
			if _nptmp_[1]= 'N' or _nptmp_[2]= 'N' then do;
				**either parent has missing genotype*;
				_AA_testII=.;_AB_testII=.;_BB_testII=.; _mis_testII=.;
			end;
			else if _nptmp_[1] = _nptmp_[2] then do;
				**parents have same genotypes*;
				if _nptmp_[1]= 'A' or _nptmp_[1]= 'C' or _nptmp_[1]= 'G' or _nptmp_[1]= 'T' then do;
					*both parents are homozygous with same genotypes*;
					_AA_testII=.;_AB_testII=.;_BB_testII=.; _mis_testII=.;
				end;
				else do;
					**both parents are heterozygous with same genotypes*;
					if _nptmp_[1]= 'M' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'M' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else if _o_[i]= 'C' then do;	
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='b';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'R' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'R' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else if _o_[i]= 'G' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='b';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'W' then do;
						do i=1 to dim(_o_);		
							if _o_[i]= 'A' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'W' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else if _o_[i]= 'T' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='b';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'S' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'S' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else if _o_[i]= 'G' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='b';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'Y' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'Y' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else if _o_[i]= 'T' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='b';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'K' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'G' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'K' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else if _o_[i]= 'T' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='b';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
				end;
			end;
			else do;
				**parents have different genotypes*;
				if (_nptmp_[1]= 'A' or _nptmp_[1]= 'C' or _nptmp_[1]= 'G' or _nptmp_[1]= 'T') 
					and (_nptmp_[2]^= 'A' or _nptmp_[2]^= 'C' or _nptmp_[2]^= 'G' or _nptmp_[2]^= 'T')
				then do;
					**p[1] is homozygous and p[2] is heretozygous*;
					if _nptmp_[1]= 'A' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'M' or  _o_[i]= 'R' or  _o_[i]= 'W' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'C' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'M' or  _o_[i]= 'S' or  _o_[i]= 'Y' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'G' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'G' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'R' or  _o_[i]= 'S' or  _o_[i]= 'K' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[1]= 'T' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'T' then do;
								_AA_testII=_AA_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'W' or  _o_[i]= 'Y' or  _o_[i]= 'K' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
				end;
				else if (_nptmp_[2]= 'A' or _nptmp_[2]= 'C' or _nptmp_[2]= 'G' or _nptmp_[2]= 'T') 
					and (_nptmp_[1]^= 'A' or _nptmp_[1]^= 'C' or _nptmp_[1]^= 'G' or _nptmp_[1]^= 'T')
				then do;
					**p[2] is homozygous and p[1] is heretozygous*;
					if _nptmp_[2]= 'A' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'M' or  _o_[i]= 'R' or  _o_[i]= 'W' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[2]= 'C' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'M' or  _o_[i]= 'S' or  _o_[i]= 'Y' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[2]= 'G' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'G' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'R' or  _o_[i]= 'S' or  _o_[i]= 'K' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
					else if _nptmp_[2]= 'T' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'T' then do;
								_BB_testII=_BB_testII + 1;
								*n_o_[i]='a';
							end;
							else if _o_[i]= 'W' or  _o_[i]= 'Y' or  _o_[i]= 'K' then do;
								_AB_testII=_AB_testII + 1;
								*n_o_[i]='ab';
							end;
							else do;
								_mis_testII= _mis_testII + 1;
								*n_o_[i]='';
							end;
						end;
					end;
				end;
				else do;
					**p[1] and p[2] are homozygous with different genotypes*;
					do i=1 to dim(_o_);
						if _o_[i]= 'M' or _o_[i]= 'R' or _o_[i]= 'W' 
						or _o_[i]= 'S' or _o_[i]= 'Y' or  _o_[i]= 'K' 
						then do;
							_AB_testII=_AB_testII + 1;
							*n_o_[i]='ab';
						end;
						else do;
							_mis_testII= _mis_testII + 1;
							*n_o_[i]='';
						end;
					end;
				end;
			end;
			*****;
			**begin chi-square based solely on offsprings*;
			_tot_testII= _AA_testII + _AB_testII + _BB_testII;
			if ((_AA_testII = 0 and _AB_testII > 0 and _BB_testII > 0) or 
				(_AA_testII > 0 and _AB_testII > 0 and _BB_testII = 0))
			then do;
				Hom_o= _AA_testII + _BB_testII; 
				*put _AA_testII _BB_testII Hom_o;
				%put ncross= &ncross;
				%do j=1 %to %sysevalf(&ncross - 1);
					%let pAA= %scan(&&probs&j, 1, ' ');
					%let pAB= %scan(&&probs&j, 2, ' ');
					%let pBB= %scan(&&probs&j, 3, ' ');
					%put Probs= &pAA &pAB &pBB;
					_chisq_testII_&&Dosage&j= (_AB_testII - &pAB*_tot_testII)**2 / (&pAB*_tot_testII)
								+ (Hom_o - &pBB*_tot_testII)**2 / (&pBB*_tot_testII);
					_pchisq_testII_&&Dosage&j= 1- PROBCHI(_chisq_testII_&&Dosage&j, 1);

					if _pchisq_testII_&&Dosage&j > &p_alpha then do;
						_indicator_testII_&&Dosage&j= 1; 
					end;
					else _indicator_testII_&&Dosage&j= 0;
				%end;
				if _indicator_testII_D0xD1 = 1 then do;
					if  ^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G', 'N') and indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G') then do;
						**p[2] is homozygous and p[1] is heterozygous*;
						_mktype_testII_D0xD1= 'D1.10';
						_newGenoP1= _p_[1];
						_newGenoP2= _p_[2];
					end;
					else if indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G', 'N') then do;
						**_p_[1] is homozygous and _p_[2] is heterozygous*;
						_mktype_testII_D0xD1= 'D2.15';
						_newGenoP1= _p_[1];
						_newGenoP2= _p_[2];
					end;
					else if indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G') then do;
						**_p_[1] and _p_[2] are homozygous; 
						if _p_[1] = _p_[2] then do;
							**Deliberately assign genotype to parents*;
							_deliberate = 1;
							_mktype_testII_D0xD1= 'D1.10';
							if _m > 0 then _newGenoP1= 'M';
							else if _r > 0 then _newGenoP1= 'R';
							else if _w > 0 then _newGenoP1= 'W';
							else if _s > 0 then _newGenoP1= 'S';
							else if _y > 0 then _newGenoP1= 'Y';
							else if _k > 0 then _newGenoP1= 'K';
	
							if _a > 0 then _newGenoP2= 'A';
							else if _c > 0 then _newGenoP2= 'C';
							else if _g > 0 then _newGenoP2= 'G';
							else if _t > 0 then _newGenoP2= 'T';
						end;
						else do;
							if indexc(trim(left(_p_[1])), trim(left(_offsobs))) then do;
								**_p_[1] is homozygous;
								_mktype_testII_D0xD1= 'D2.15';
								_newGenoP1= _p_[1];
								if _m > 0 then _newGenoP2= 'M';
								else if _r > 0 then _newGenoP2= 'R';
								else if _w > 0 then _newGenoP2= 'W';
								else if _s > 0 then _newGenoP2= 'S';
								else if _y > 0 then _newGenoP2= 'Y';
								else if _k > 0 then _newGenoP2= 'K';
								*put &IDVar _p_[1] _offsobs;
							end;
							else if indexc(trim(left(_p_[2])), trim(left(_offsobs))) then do;
								**_p_[2] is homozygous;
								_mktype_testII_D0xD1= 'D1.10';
								_newGenoP2= _p_[2];
								if _m > 0 then _newGenoP1= 'M';
								else if _r > 0 then _newGenoP1= 'R';
								else if _w > 0 then _newGenoP1= 'W';
								else if _s > 0 then _newGenoP1= 'S';
								else if _y > 0 then _newGenoP1= 'Y';
								else if _k > 0 then _newGenoP1= 'K';
							end;
						end;
					end;
					else if indexc(trim(left(_p_[1])), 'N') and indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G') and indexc(trim(left(_p_[2])), trim(left(_offsobs))) then do;
					*else if indexc(trim(left(_p_[1])), 'N') and indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G') then do;
						**_p_[2] is homozygous and p1 is missing*;
						_mktype_testII_D0xD1= 'D1.10';
						_newGenoP2= _p_[2];
						if _m > 0 then _newGenoP1= 'M';
						else if _r > 0 then _newGenoP1= 'R';
						else if _w > 0 then _newGenoP1= 'W';
						else if _s > 0 then _newGenoP1= 'S';
						else if _y > 0 then _newGenoP1= 'Y';
						else if _k > 0 then _newGenoP1= 'K';
					end;
					else if indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and indexc(trim(left(_p_[2])), 'N') and indexc(trim(left(_p_[1])), trim(left(_offsobs))) then do;
					*else if indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and indexc(trim(left(_p_[2])), 'N') then do;
						**_p_[1] is homozygous and _p_[2] is missing*;
						_mktype_testII_D0xD1= 'D2.15';
						_newGenoP1= _p_[1];
						if _m > 0 then _newGenoP2= 'M';
						else if _r > 0 then _newGenoP2= 'R';
						else if _w > 0 then _newGenoP2= 'W';
						else if _s > 0 then _newGenoP2= 'S';
						else if _y > 0 then _newGenoP2= 'Y';
						else if _k > 0 then _newGenoP2= 'K';
					end;
					else if ^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G', 'N') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G', 'N') then do;
						**_p_[1] and _p_[2] are heterozygous. Deliberately assign genotype to parents*;
						_deliberate = 1;
						_mktype_testII_D0xD1= 'D1.10';
						_newGenoP1= _p_[1];
						if _a > 0 then _newGenoP2= 'A';
						else if _c > 0 then _newGenoP2= 'C';
						else if _g > 0 then _newGenoP2= 'G';
						else if _t > 0 then _newGenoP2= 'T';
					end;
					else if indexc(trim(left(_p_[1])), 'N') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G', 'N') then do;
						**_p_[2] is heterozygous and _p_[1] is missing*;
						_mktype_testII_D0xD1= 'D2.15';
						_newGenoP2= _p_[2];
						if _a > 0 then _newGenoP1= 'A';
						else if _c > 0 then _newGenoP1= 'C';
						else if _g > 0 then _newGenoP1= 'G';
						else if _t > 0 then _newGenoP1= 'T';
					end;
					else if ^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G', 'N') and indexc(trim(left(_p_[2])), 'N') then do;
						**_p_[1] is heterozygous and _p_[2] is missing*;
						_mktype_testII_D0xD1= 'D1.10';
						_newGenoP1= _p_[1];
						if _a > 0 then _newGenoP2= 'A';
						else if _c > 0 then _newGenoP2= 'C';
						else if _g > 0 then _newGenoP2= 'G';
						else if _t > 0 then _newGenoP2= 'T';
					end;				
					else if indexc(trim(left(_p_[1])), 'N') and indexc(trim(left(_p_[2])), 'N') then do;
						**_p_[1] and _p_[2] are missing. Deliberately assign genotype to parents*;
						_deliberate = 1;
						_mktype_testII_D0xD1= 'D1.10';
						if _m > 0 then _newGenoP1= 'M';
						else if _r > 0 then _newGenoP1= 'R';
						else if _w > 0 then _newGenoP1= 'W';
						else if _s > 0 then _newGenoP1= 'S';
						else if _y > 0 then _newGenoP1= 'Y';
						else if _k > 0 then _newGenoP1= 'K';

						if _a > 0 then _newGenoP2= 'A';
						else if _c > 0 then _newGenoP2= 'C';
						else if _g > 0 then _newGenoP2= 'G';
						else if _t > 0 then _newGenoP2= 'T';
					end;
					else _indicator_testII_D0xD1 = 0;
				end;
				if _indicator_testII_D1xD1 = 1 then do;
					_mktype_testII_D1xD1= 'C.8';
				end;	
			end;
			else if (_AA_testII > 0 and _AB_testII > 0 and _BB_testII > 0) then do;
					%let pAA= %scan(&&probs&ncross, 1, ' ');
					%let pAB= %scan(&&probs&ncross, 2, ' ');
					%let pBB= %scan(&&probs&ncross, 3, ' ');
					%put Probs= &pAA &pAB &pBB;
					_chisq_testII_&&Dosage&ncross= (_AA_testII - &pAA*_tot_testII)**2 / (&pAA*_tot_testII) 
								+ (_AB_testII - &pAB*_tot_testII)**2 / (&pAB*_tot_testII) 
								+ (_BB_testII - &pBB*_tot_testII)**2 / (&pBB*_tot_testII);
					_pchisq_testII_&&Dosage&ncross= 1 - PROBCHI(_chisq_testII_&&Dosage&ncross, 2);

					if _pchisq_testII_&&Dosage&ncross > &p_alpha then do;
						_indicator_testII_&&Dosage&ncross= 1;
					end;
					else _indicator_testII_&&Dosage&ncross= 0;
					if _indicator_testII_D1xD1 = 1 then do;
						_mktype_testII_D1xD1= 'B3.7';
					end;
			end;	
			array _np_{*} _newGenoP1 _newGenoP2;
			_np_[1]= upcase(trim(left(_np_[1])));
			_np_[2]= upcase(trim(left(_np_[2])));
			**begin recoding parents*;
			if trim(left(_mktype_testII_D0xD1))= 'D1.10' then do;
				_recP1obs= 'ab';
				_recP2obs= 'aa';
			end;
			else if trim(left(_mktype_testII_D0xD1))= 'D2.15' then do;
				_recP1obs= 'aa';
				_recP2obs= 'ab';
			end;
			else if trim(left(_mktype_testII_D1xD1))= 'C.8' then do;
				_recP1obs= 'ao';
				_recP2obs= 'ao';
			end;
			else if trim(left(_mktype_testII_D1xD1))= 'B3.7' then do;
				_recP1obs= 'ab';
				_recP2obs= 'ab';
			end;
			*****;	
			****************************************************************;
			**begin analysis based on genotype of parents*;
			select;
				**one parent has genotype A and the other parent varies*;
				when((_p_[1]='A' and _p_[2]='A')) do;
					_offsexp= 'A';
					*_tot_testI= _a;
				end;
				when((_p_[1]='A' and _p_[2]='C') or (_p_[2]='A' and _p_[1]='C')) do;
					_offsexp= 'M';
					*_tot_testI= _m;
				end;
				when((_p_[1]='A' and _p_[2]='G') or (_p_[2]='A' and _p_[1]='G')) do;
					_offsexp= 'R';
					*_tot_testI= _r;
				end;
				when((_p_[1]='A' and _p_[2]='T') or (_p_[2]='A' and _p_[1]='T')) do;
					_offsexp= 'W';
					*_tot_testI= _w;
				end;
				when((_p_[1]='A' and _p_[2]='M') or (_p_[2]='A' and _p_[1]='M')) do;
					_offsexp= 'A M';
					*_tot_testI= _a + _m;
				end;
				when((_p_[1]='A' and _p_[2]='R') or (_p_[2]='A' and _p_[1]='R')) do;
					_offsexp= 'A R';
					*_tot_testI= _a + _r;
				end;
				when((_p_[1]='A' and _p_[2]='W') or (_p_[2]='A' and _p_[1]='W')) do;
					_offsexp= 'A W';
					*_tot_testI= _a + _w;
				end;
				when((_p_[1]='A' and _p_[2]='S') or (_p_[2]='A' and _p_[1]='S')) do;
					_offsexp= 'M R';
					*_tot_testI= _m + _r;
				end;
				when((_p_[1]='A' and _p_[2]='Y') or (_p_[2]='A' and _p_[1]='Y')) do;
					_offsexp= 'M W';
					*_tot_testI= _m + _w;
				end;
				when((_p_[1]='A' and _p_[2]='K') or (_p_[2]='A' and _p_[1]='K')) do;
					_offsexp= 'R W';
					*_tot_testI= _r + _w;
				end;
				*one parent has genotype C and the other parent varies*;
				when((_p_[1]='C' and _p_[2]='C')) do;
					_offsexp= 'C';
				end;
				when((_p_[1]='C' and _p_[2]='G') or (_p_[2]='C' and _p_[1]='G')) do;
					_offsexp= 'S';
				end;
				when((_p_[1]='C' and _p_[2]='T') or (_p_[2]='C' and _p_[1]='T')) do;
					_offsexp= 'Y';
				end;
				when((_p_[1]='C' and _p_[2]='M') or (_p_[2]='C' and _p_[1]='M')) do;
					_offsexp= 'C M';
					*_tot_testI= _c + _m;
				end;
				when((_p_[1]='C' and _p_[2]='R') or (_p_[2]='C' and _p_[1]='R')) do;
					_offsexp= 'M S';
					*_tot_testI= _m + _s;
				end;
				when((_p_[1]='C' and _p_[2]='W') or (_p_[2]='C' and _p_[1]='W')) do;
					_offsexp= 'M Y';
					*_tot_testI= _m + _y;
				end;
				when((_p_[1]='C' and _p_[2]='S') or (_p_[2]='C' and _p_[1]='S')) do;
					_offsexp= 'C S';
					*_tot_testI= _c + _s;
				end;
				when((_p_[1]='C' and _p_[2]='Y') or (_p_[2]='C' and _p_[1]='Y')) do;
					_offsexp= 'C Y';
					*_tot_testI= _c + _y;
				end;
				when((_p_[1]='C' and _p_[2]='K') or (_p_[2]='C' and _p_[1]='K')) do;
					_offsexp= 'S Y';
					*_tot_testI= _s + _y;
				end;
				*one parent has genotype G and the other parent varies*;
				when((_p_[1]='G' and _p_[2]='G')) do;
					_offsexp= 'G';
				end;
				when((_p_[1]='G' and _p_[2]='T') or (_p_[2]='G' and _p_[1]='T')) do;
					_offsexp= 'K';
				end;
				when((_p_[1]='G' and _p_[2]='M') or (_p_[2]='G' and _p_[1]='M')) do;
					_offsexp= 'R S';
					*_tot_testI= _r + _s;
				end;
				when((_p_[1]='G' and _p_[2]='R') or (_p_[2]='G' and _p_[1]='R')) do;
					_offsexp= 'G R';
					*_tot_testI= _g + _r;
				end;
				when((_p_[1]='G' and _p_[2]='W') or (_p_[2]='G' and _p_[1]='W')) do;
					_offsexp= 'R K';
					*_tot_testI= _r + _k;
				end;
				when((_p_[1]='G' and _p_[2]='S') or (_p_[2]='G' and _p_[1]='S')) do;
					_offsexp= 'G S';
					*_tot_testI= _g + _s;
				end;
				when((_p_[1]='G' and _p_[2]='Y') or (_p_[2]='G' and _p_[1]='Y')) do;
					_offsexp= 'S K';
					*_tot_testI= _s + _k;
				end;
				when((_p_[1]='G' and _p_[2]='K') or (_p_[2]='G' and _p_[1]='K')) do;
					_offsexp= 'G K';
					*_tot_testI= _g + _k;
				end;
				*one parent has genotype T and the other parent varies*;
				when((_p_[1]='T' and _p_[2]='T')) do;
					_offsexp= 'T';
				end;
				when((_p_[1]='T' and _p_[2]='M') or (_p_[2]='T' and _p_[1]='M')) do;
					_offsexp= 'W Y';
					*_tot_testI= _w + _y;
				end;
				when((_p_[1]='T' and _p_[2]='R') or (_p_[2]='T' and _p_[1]='R')) do;
					_offsexp= 'W K';
					*_tot_testI= _w + _k;
				end;
				when((_p_[1]='T' and _p_[2]='W') or (_p_[2]='T' and _p_[1]='W')) do;
					_offsexp= 'T W';
					*_tot_testI= _t + _w;
				end;
				when((_p_[1]='T' and _p_[2]='S') or (_p_[2]='T' and _p_[1]='S')) do;
					_offsexp= 'Y K';
					*_tot_testI= _y + _k;
				end;
				when((_p_[1]='T' and _p_[2]='Y') or (_p_[2]='T' and _p_[1]='Y')) do;
					_offsexp= 'T Y';
					*_tot_testI= _t + _y;
				end;
				when((_p_[1]='T' and _p_[2]='K') or (_p_[2]='T' and _p_[1]='K')) do;
					_offsexp= 'T K';
					*_tot_testI= _t + _k;
				end;
				*parent has genotype M and the other parent varies*;
				when((_p_[1]='M' and _p_[2]='M')) do;
					_offsexp= 'A C M';
					*_tot_testI= _a + _c + _m;
				end;
				when((_p_[1]='M' and _p_[2]='R') or (_p_[2]='M' and _p_[1]='R')) do;
					_offsexp= 'A M R S';
				end;
				when((_p_[1]='M' and _p_[2]='W') or (_p_[2]='M' and _p_[1]='W')) do;
					_offsexp= 'A M W Y';
				end;
				when((_p_[1]='M' and _p_[2]='S') or (_p_[2]='M' and _p_[1]='S')) do;
					_offsexp= 'C M R S';
				end;
				when((_p_[1]='M' and _p_[2]='Y') or (_p_[2]='M' and _p_[1]='Y')) do;
					_offsexp= 'C M W Y';
				end;
				when((_p_[1]='M' and _p_[2]='K') or (_p_[2]='M' and _p_[1]='K')) do;
					_offsexp= 'R W S Y';
				end;
				*one parent has genotype R and the other parent varies*;
				when((_p_[1]='R' and _p_[2]='R')) do;
					_offsexp= 'A G R';
					*_tot_testI= _a + _g + _r;
				end;
				when((_p_[1]='R' and _p_[2]='W') or (_p_[2]='R' and _p_[1]='W')) do;
					_offsexp= 'A R W K';
				end;
				when((_p_[1]='R' and _p_[2]='S') or (_p_[2]='R' and _p_[1]='S')) do;
					_offsexp= 'G M R S';
				end;
				when((_p_[1]='R' and _p_[2]='Y') or (_p_[2]='R' and _p_[1]='Y')) do;
					_offsexp= 'M W S K';
				end;
				when((_p_[1]='R' and _p_[2]='K') or (_p_[2]='R' and _p_[1]='K')) do;
					_offsexp= 'G R W K';
				end;
				*one parent has genotype W and the other parent varies*;
				when((_p_[1]='W' and _p_[2]='W')) do;
					_offsexp= 'A T W';
					*_tot_testI= _a + _t + _w;
				end;
				when((_p_[1]='W' and _p_[2]='S') or (_p_[2]='W' and _p_[1]='S')) do;
					_offsexp= 'M R Y K';
				end;
				when((_p_[1]='W' and _p_[2]='Y') or (_p_[2]='W' and _p_[1]='Y')) do;
					_offsexp= 'T M W Y';
				end;
				when((_p_[1]='W' and _p_[2]='K') or (_p_[2]='W' and _p_[1]='K')) do;
					_offsexp= 'T R W K';
				end;
				*one parent has genotype S and the other parent varies*;
				when((_p_[1]='S' and _p_[2]='S')) do;
					_offsexp= 'C G S';
					*_tot_testI= _c + _g + _s;
				end;
				when((_p_[1]='S' and _p_[2]='Y') or (_p_[2]='S' and _p_[1]='Y')) do;
					_offsexp= 'C Y S K';
				end;
				when((_p_[1]='S' and _p_[2]='K') or (_p_[2]='S' and _p_[1]='K')) do;
					_offsexp= 'G S Y K';
				end;
				*one parent has genotype Y and the other parent varies*;
				when((_p_[1]='Y' and _p_[2]='Y')) do;
					_offsexp= 'C T Y';
					*_tot_testI= _c + _t + _y;
				end;
				when((_p_[1]='Y' and _p_[2]='K') or (_p_[2]='Y' and _p_[1]='K')) do;
					_offsexp= 'T S Y K';
				end;
				*one parent has genotype K and other parent varies*;
				when((_p_[1]='K' and _p_[2]='K')) do;
					_offsexp= 'G T K';
					*_tot_testI= _g + _t + _k;
				end;
				otherwise;
			end;
			_offsexp=COMPRESS(_offsexp); 
			*****;
			**assigning marker type based on observed parental genotypes*;
			*put _n_ _p_[1] _p_[2];
			if _p_[1]= 'N' or _p_[2]= 'N' then do;
				**either parent has missing genotype*;
				_AA_testI=.;_AB_testI=.;_BB_testI=.;_mis_testI=.;_genotype_error_testI=.;
				*put 'p1=N or p2=N ' _n_ _p_[1] _p_[2];
			end;
			else if _p_[1] = _p_[2] then do;
				**parents have same genotypes*;
				if _p_[1]= 'A' or _p_[1]= 'C' or _p_[1]= 'G' or _p_[1]= 'T' then do;
					**both parents are homozygous with same genotypes*;
					_AA_testI=.;_AB_testI=.;_BB_testI=.;_mis_testI=.;_genotype_error_testI=.;
				end;
				else do;
					**both parents are heterozygous with same genotypes*;
					if _p_[1]= 'M' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'M' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'C' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'b';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'R' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'R' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'G' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'b';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'W' then do;
						do i=1 to dim(_o_);		
							if _o_[i]= 'A' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'W' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'T' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'b';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'S' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'S' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'G' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'b';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'Y' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'Y' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'T' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'b';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'K' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'G' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'K' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'T' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'b';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
				end;
			end;
			else do;
				**parents have different genotypes*;
				if (_p_[1]= 'A' or _p_[1]= 'C' or _p_[1]= 'G' or _p_[1]= 'T') 
					and (_p_[2]^= 'A' and _p_[2]^= 'C' and _p_[2]^= 'G' and _p_[2]^= 'T')
				then do;
					*parent 1 is homozygous and parent 2 is heretozygous*;
					if _p_[1]= 'A' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'M' or _o_[i]= 'R' or _o_[i]= 'W' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'C' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'M' or  _o_[i]= 'S' or  _o_[i]= 'Y' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'G' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'G' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'R' or  _o_[i]= 'S' or  _o_[i]= 'K' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[1]= 'T' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'T' then do;
								_AA_testI=_AA_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'W' or  _o_[i]= 'Y' or  _o_[i]= 'K' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
				end;
				else if (_p_[2]= 'A' or _p_[2]= 'C' or _p_[2]= 'G' or _p_[2]= 'T') 
					and (_p_[1]^= 'A' and _p_[1]^= 'C' and _p_[1]^= 'G' and _p_[1]^= 'T')
				then do;
					**parent 2 is homozygous and parent 1 is heretozygous*;
					if _p_[2]= 'A' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'A' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'M' or  _o_[i]= 'R' or  _o_[i]= 'W' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[2]= 'C' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'C' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'M' or  _o_[i]= 'S' or  _o_[i]= 'Y' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[2]= 'G' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'G' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'R' or  _o_[i]= 'S' or  _o_[i]= 'K' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
					else if _p_[2]= 'T' then do;
						do i=1 to dim(_o_);
							if _o_[i]= 'T' then do;
								_BB_testI=_BB_testI + 1;
								*_ne_[i]= 'a';
							end;
							else if _o_[i]= 'W' or  _o_[i]= 'Y' or  _o_[i]= 'K' then do;
								_AB_testI=_AB_testI + 1;
								*_ne_[i]= 'ab';
							end;
							else if _o_[i]= 'N' then do;
								_mis_testI= _mis_testI + 1;
								*_ne_[i]= '';
							end;
							else do;
								_genotype_error_testI= _genotype_error_testI + 1;
								_ne_[i]= 'N';
							end;
						end;
					end;
				end;
				else do;
					**parent 1 and parent 2 are homozygous with different genotypes*;
					do i=1 to dim(_o_);
						if _o_[i]= 'M' or _o_[i]= 'R' or _o_[i]= 'W' 
						or _o_[i]= 'S' or _o_[i]= 'Y' or _o_[i]= 'K'
						then do;
							_AB_testI=_AB_testI + 1;
							*_ne_[i]= 'ab';
						end;
						else if _o_[i]= 'N' then do;
							_mis_testI= _mis_testI + 1;
							*_ne_[i]= '';
						end;
						else do;
							_genotype_error_testI= _genotype_error_testI + 1;
							_ne_[i]= 'N';
						end;
					end;
				end;
			end;
			*****;
			**begin chi-square based on parents genotypes*;
			_tot_testI= _AA_testI + _AB_testI + _BB_testI;
			if ((_AA_testI = 0 and _AB_testI > 0 and _BB_testI > 0) or (_AA_testI > 0 and _AB_testI > 0 and _BB_testI = 0))
				then do;
				if (^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G')) or
					(indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G')) then do;
					
					Hom_e= _AA_testI + _BB_testI; 
					%put ncross= &ncross;
					%do j=1 %to &m_;
						%let pAA= %scan(&&probs&j, 1, ' ');
						%let pAB= %scan(&&probs&j, 2, ' ');
						%let pBB= %scan(&&probs&j, 3, ' ');
						%put Probs= &pAA &pAB &pBB;
						_chisq_testI_&&Dosage&j= (_AB_testI - &pAB*_tot_testI)**2 / (&pAB*_tot_testI)
									+ (Hom_e - &pBB*_tot_testI)**2 / (&pBB*_tot_testI);
						_pchisq_testI_&&Dosage&j= 1- PROBCHI(_chisq_testI_&&Dosage&j, 1);

						if _pchisq_testI_&&Dosage&j > &p_alpha then do;
							_indicator_testI_&&Dosage&j= 1; 
						end;
						else _indicator_testI_&&Dosage&j= 0;
					%end;
					*put _p_[1] _p_[2] _indicator_testI_D0xD1;
					if _indicator_testI_D0xD1 = 1 then do;
						if ^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G') then do;
							_mktype_testI_D0xD1= 'D1.10';
						end;
						else if indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G') then do;
							_mktype_testI_D0xD1= 'D2.15';
						end;
						else _indicator_testI_D0xD1= 0;
					end;
				end;
				else if (^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G', 'N') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G', 'N')) then do;
					Hom_e= _AA_testI + _BB_testI; 
					%put ncross= &ncross;
					%do j=%sysevalf(&m_ + 1) %to %sysevalf(&ncross - 1);
						%let pAA= %scan(&&probs&j, 1, ' ');
						%let pAB= %scan(&&probs&j, 2, ' ');
						%let pBB= %scan(&&probs&j, 3, ' ');
						%put Probs= &pAA &pAB &pBB;
						_chisq_testI_&&Dosage&j= (_AB_testI - &pAB*_tot_testI)**2 / (&pAB*_tot_testI)
									+ (Hom_e - &pBB*_tot_testI)**2 / (&pBB*_tot_testI);
						_pchisq_testI_&&Dosage&j= 1- PROBCHI(_chisq_testI_&&Dosage&j, 1);

						if _pchisq_testI_&&Dosage&j > &p_alpha then do;
							_indicator_testI_&&Dosage&j= 1; 
						end;
						else _indicator_testI_&&Dosage&j= 0;
					%end;
					if _indicator_testI_D1xD1 = 1 then do;
						if ^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G', 'N') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G', 'N') then do;
							_mktype_testI_D1xD1= 'C.8';
						end;
						else _indicator_testI_D1xD1 = 0;
					end;
					*put _p_[1] _p_[2] _indicator_testI_D0xD1;
				end;
			end;
			else if (_AA_testI > 0 and _AB_testI > 0 and _BB_testI > 0) then do;
				if ^indexc(trim(left(_p_[1])), 'A', 'C', 'T', 'G', 'N') and ^indexc(trim(left(_p_[2])), 'A', 'C', 'T', 'G', 'N') then do;
					%let pAA= %scan(&&probs&ncross, 1, ' ');
					%let pAB= %scan(&&probs&ncross, 2, ' ');
					%let pBB= %scan(&&probs&ncross, 3, ' ');
					%put Probs= &pAA &pAB &pBB;
					_chisq_testI_&&Dosage&ncross= (_AA_testI - &pAA*_tot_testI)**2 / (&pAA*_tot_testI) 
								+ (_AB_testI - &pAB*_tot_testI)**2 / (&pAB*_tot_testI) 
								+ (_BB_testI - &pBB*_tot_testI)**2 / (&pBB*_tot_testI);
					_pchisq_testI_&&Dosage&ncross= 1 - PROBCHI(_chisq_testI_&&Dosage&ncross, 2);

					if _pchisq_testI_&&Dosage&ncross > &p_alpha then do;
						_indicator_testI_&&Dosage&ncross= 1;
					end;
					else _indicator_testI_&&Dosage&ncross= 0;	
					if _indicator_testI_D1xD1 = 1 then do;
						_mktype_testI_D1xD1= 'B3.7';
					end;
				end;
			end;	
			**begin recoding parents*;
			if trim(left(_mktype_testI_D0xD1))= 'D1.10' then do;
				_recP1exp= 'ab';
				_recP2exp= 'aa';
			end;
			else if trim(left(_mktype_testI_D0xD1))= 'D2.15' then do;
				_recP1exp= 'aa';
				_recP2exp= 'ab';
			end;
			else if trim(left(_mktype_testI_D1xD1))= 'C.8' then do;
				_recP1exp= 'ao';
				_recP2exp= 'ao';
			end;
			else if trim(left(_mktype_testI_D1xD1))= 'B3.7' then do;
				_recP1exp= 'ab';
				_recP2exp= 'ab';
			end;
			*****;			
			*************************************************************************************;
			**begin selecting SD markers*;
			_p_genotype_error_testI= _genotype_error_testI/&_nind_;
			_sum_indicator_testI= 
				%do j=1 %to %sysevalf(&ncross);
					_sum_indicator_testI + _indicator_testI_&&Dosage&j;
				%end;
			_sum_indicator_testII= 
				%do j=1 %to %sysevalf(&ncross);
					_sum_indicator_testII + _indicator_testII_&&Dosage&j;
				%end;
			if _sum_indicator_testI = 1 or _sum_indicator_testII = 1 then do;
				_uniquedosage= 1;
				if (_sum_indicator_testI = 1 and (_indicator_testI_D0xD1 = 1 or _indicator_testI_D1xD1 = 1)) then do;
					_singledosage = 1;
				end;
				else if (_sum_indicator_testII = 1 and (_indicator_testII_D0xD1 = 1 or _indicator_testII_D1xD1 = 1)) then do; 
					_singledosage = 1;
				end;
					
				if (_p_genotype_error_testI < &MaxGenoError and ((_mis_testI + _genotype_error_testI)/&_nind_ < &MaxMissInd) and _sum_indicator_testI= 1 and sum(_indicator_testI_D0xD1, _indicator_testI_D1xD1)=1) 
					or (_mis_testII/&_nind_ < &MaxMissInd and _sum_indicator_testII= 1 and sum(_indicator_testII_D0xD1, _indicator_testII_D1xD1)=1) 
					then do;
					if (_p_genotype_error_testI < &MaxGenoError and (_mis_testI + _genotype_error_testI)/&_nind_ < &MaxMissInd and _sum_indicator_testI= 1 and sum(_indicator_testI_D0xD1, _indicator_testI_D1xD1)=1) then do;
						if _indicator_testI_D0xD1= 1 then do;
							**Null x SD= 1:1 ratio. Works for any ploid level as well as for auto and poly*;
							_mktype_testI= _mktype_testI_D0xD1;
							_chisq_testI= _chisq_testI_D0xD1;
							_pchisq_testI= _pchisq_testI_D0xD1;
						end;
						else if _indicator_testI_D1xD1= 1 then do;
							**SD x SD= 1:2:1 ratio for ploid= 2*;
							_mktype_testI= _mktype_testI_D1xD1;
							_chisq_testI= _chisq_testI_D1xD1;
							_pchisq_testI= _pchisq_testI_D1xD1;
						end;
						_newP1=_p_[1];
						_newP2=_p_[2];
						_recP1= _recP1exp;
						_recP2= _recP2exp;
						_testI_OR_testII= 'I';
						_mktype= upcase(trim(left(_mktype_testI))); 
						_selected= 1;
						_n_ind= _tot_ind;
						_n_mis=_mis_testI + _genotype_error_testI;
						_n_err=_genotype_error_testI;
						_p_err=_p_genotype_error_testI;
						_n_dom= _AA_testI;
						_n_het= _AB_testI;
						_n_rec= _BB_testI;
						_chisq=_chisq_testI;
						_pchisq=_pchisq_testI;
						if _mktype = 'D1.10' or _mktype = 'D2.15' then do;
							**_p_[1] is homozygous and _p_[2] is heretozygous or vice-versa*;
							if _p_[1]= 'A' or _p_[2]= 'A' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'A' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'M' or _ne_[i]= 'R' or _ne_[i]= 'W' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'C' or _p_[2]= 'C' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'C' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'M' or  _ne_[i]= 'S' or  _ne_[i]= 'Y' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'G' or _p_[2]= 'G' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'G' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'R' or  _ne_[i]= 'S' or  _ne_[i]= 'K' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'T' or _p_[2]= 'T' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'T' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'W' or  _ne_[i]= 'Y' or  _ne_[i]= 'K' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
						end;
						else if _mktype = 'B3.7' then do;
							**both parents are heterozygous with same genotypes*;
							if _p_[1]= 'M' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'A' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'M' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'C' then do;
										_nf_[i]= 'b';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'R' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'A' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'R' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'G' then do;
										_nf_[i]= 'b';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'W' then do;
								do i=1 to dim(_ne_);	
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'A' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'W' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'T' then do;
										_nf_[i]= 'b';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'S' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'C' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'S' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'G' then do;
										_nf_[i]= 'b';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'Y' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'C' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'Y' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'T' then do;
										_nf_[i]= 'b';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
							else if _p_[1]= 'K' then do;
								do i=1 to dim(_ne_);
									_ne_[i]= upcase(trim(left(_ne_[i])));
									if _ne_[i]= 'G' then do;
										_nf_[i]= 'a';
									end;
									else if _ne_[i]= 'K' then do;
										_nf_[i]= 'ab';
									end;
									else if _ne_[i]= 'T' then do;
										_nf_[i]= 'b';
									end;
									else if _ne_[i]= 'N' then do;
										_nf_[i]= '';
									end;
									else do;
										_nf_[i]= '';
									end;
								end;
							end;
						end;
						%if &PloidyLevel > 2 %then %do;
							else if _mktype = 'C.8' then do;
								**both parents are heterozygous with same genotypes*;
								if _p_[1]= 'M' then do;
									do i=1 to dim(_ne_);
										_ne_[i]= upcase(trim(left(_ne_[i])));
										if _ne_[i]= 'A' or _ne_[i]= 'C' then do;
											_nf_[i]= 'o';
										end;
										else if _ne_[i]= 'M' then do;
											_nf_[i]= 'a';
										end;
										else if _ne_[i]= 'N' then do;
											_nf_[i]= '';
										end;
										else do;
											_nf_[i]= '';
										end;
									end;
								end;
								else if _p_[1]= 'R' then do;
									do i=1 to dim(_ne_);
										_ne_[i]= upcase(trim(left(_ne_[i])));
										if _ne_[i]= 'A' or _ne_[i]= 'G' then do;
											_nf_[i]= 'o';
										end;
										else if _ne_[i]= 'R' then do;
											_nf_[i]= 'a';
										end;
										else if _ne_[i]= 'N' then do;
											_nf_[i]= '';
										end;
										else do;
											_nf_[i]= '';
										end;
									end;
								end;
								else if _p_[1]= 'W' then do;
									do i=1 to dim(_ne_);
										_ne_[i]= upcase(trim(left(_ne_[i])));
										if _ne_[i]= 'A' or _ne_[i]= 'T' then do;
											_nf_[i]= 'o';
										end;
										else if _ne_[i]= 'W' then do;
											_nf_[i]= 'a';
										end;
										else if _ne_[i]= 'N' then do;
											_nf_[i]= '';
										end;
										else do;
											_nf_[i]= '';
										end;
									end;
								end;
								else if _p_[1]= 'S' then do;
									do i=1 to dim(_ne_);
										_ne_[i]= upcase(trim(left(_ne_[i])));
										if _ne_[i]= 'C' or _ne_[i]= 'G' then do;
											_nf_[i]= 'o';
										end;
										else if _ne_[i]= 'S' then do;
											_nf_[i]= 'a';
										end;
										else if _ne_[i]= 'N' then do;
											_nf_[i]= '';
										end;
										else do;
											_nf_[i]= '';
										end;
									end;
								end;
								else if _p_[1]= 'Y' then do;
									do i=1 to dim(_ne_);
										_ne_[i]= upcase(trim(left(_ne_[i])));
										if _ne_[i]= 'C' or _ne_[i]= 'T' then do;
											_nf_[i]= 'o';
										end;
										else if _ne_[i]= 'Y' then do;
											_nf_[i]= 'a';
										end;
										else if _ne_[i]= 'N' then do;
											_nf_[i]= '';
										end;
										else do;
											_nf_[i]= '';
										end;
									end;
								end;
								else if _p_[1]= 'K' then do;
									do i=1 to dim(_ne_);
										_ne_[i]= upcase(trim(left(_ne_[i])));
										if _ne_[i]= 'G' or _ne_[i]= 'T' then do;
											_nf_[i]= 'o';
										end;
										else if _ne_[i]= 'K' then do;
											_nf_[i]= 'a';
										end;
										else if _ne_[i]= 'N' then do;
											_nf_[i]= '';
										end;
										else do;
											_nf_[i]= '';
										end;
									end;
								end;
							end;
						%end;
					end;
					else do;
						if _indicator_testII_D0xD1= 1 then do;
							**Null x SD= 1:1 ratio. Works for any ploid level as well as for auto and poly*;
							_mktype_testII= _mktype_testII_D0xD1; 
							_chisq_testII= _chisq_testII_D0xD1;
							_pchisq_testII= _pchisq_testII_D0xD1;
						end;
						else if _indicator_testII_D1xD1= 1 then do;
							**SD x SD= 1:2:1 ratio for ploid= 2*;
							_mktype_testII= _mktype_testII_D1xD1;
							_chisq_testII= _chisq_testII_D1xD1;
							_pchisq_testII= _pchisq_testII_D1xD1;
						end;
						_newP1=_newGenoP1;
						_newP2=_newGenoP2;
						_recP1= _recP1obs;
						_recP2= _recP2obs;
						_testI_OR_testII='II';
						_mktype= upcase(trim(left(_mktype_testII)));
						_selected= 1;
						_n_ind= _tot_ind;
						_n_mis=_mis_testII;
						_n_err= 0;
						_p_err= 0;
						_n_dom= _AA_testII;
						_n_het= _AB_testII;
						_n_rec= _BB_testII;
						_chisq=_chisq_testII;
						_pchisq=_pchisq_testII;
						if _mktype = 'D1.10' or _mktype = 'D2.15' then do;
							**parent 1 is homozygous and parent 2 is heretozygous*;
							if _np_[1]= 'A' or _np_[2]= 'A' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'A' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'M' or  _o_[i]= 'R' or  _o_[i]= 'W' then do;
										_nf_[i]='ab';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'C' or _np_[2]= 'C' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'C' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'M' or  _o_[i]= 'S' or  _o_[i]= 'Y' then do;
										_nf_[i]='ab';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'G' or _np_[2]= 'G' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'G' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'R' or  _o_[i]= 'S' or  _o_[i]= 'K' then do;
										_nf_[i]='ab';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'T' or _np_[2]= 'T' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'T' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'W' or  _o_[i]= 'Y' or  _o_[i]= 'K' then do;
										_nf_[i]='ab';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
						end;
						else if _mktype = 'B3.7' then do;
							**both parents are heterozygous with same genotypes*;
							if _np_[1]= 'M' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'A' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'M' then do;
										_nf_[i]='ab';
									end;
									else if _o_[i]= 'C' then do;	
										_nf_[i]='b';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'R' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'A' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'R' then do;
										_nf_[i]='ab';
									end;
									else if _o_[i]= 'G' then do;
										_nf_[i]='b';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'W' then do;
								do i=1 to dim(_o_);		
									if _o_[i]= 'A' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'W' then do;
										_nf_[i]='ab';
									end;
									else if _o_[i]= 'T' then do;
										_nf_[i]='b';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'S' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'C' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'S' then do;
										_nf_[i]='ab';
									end;
									else if _o_[i]= 'G' then do;
										_nf_[i]='b';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'Y' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'C' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'Y' then do;
										_nf_[i]='ab';
									end;
									else if _o_[i]= 'T' then do;
										_nf_[i]='b';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
							else if _np_[1]= 'K' then do;
								do i=1 to dim(_o_);
									if _o_[i]= 'G' then do;
										_nf_[i]='a';
									end;
									else if _o_[i]= 'K' then do;
										_nf_[i]='ab';
									end;
									else if _o_[i]= 'T' then do;
										_nf_[i]='b';
									end;
									else do;
										_nf_[i]='';
									end;
								end;
							end;
						end;
						%if &PloidyLevel > 2 %then %do;
							else if _mktype = 'C.8' then do;
								**both parents are heterozygous with same genotypes*;
								if _np_[1]= 'M' then do;
									do i=1 to dim(_o_);
										if _o_[i]= 'A' or _o_[i]= 'C' then do;
											_nf_[i]='o';
										end;
										else if _o_[i]= 'M' then do;
											_nf_[i]='a';
										end;
										else do;
											_nf_[i]='';
										end;
									end;
								end;
								else if _np_[1]= 'R' then do;
									do i=1 to dim(_o_);
										if _o_[i]= 'A' or _o_[i]= 'G' then do;
											_nf_[i]='o';
										end;
										else if _o_[i]= 'R' then do;
											_nf_[i]='a';
										end;
										else do;
											_nf_[i]='';
										end;
									end;
								end;
								else if _np_[1]= 'W' then do;
									do i=1 to dim(_o_);		
										if _o_[i]= 'A' or _o_[i]= 'T' then do;
											_nf_[i]='o';
										end;
										else if _o_[i]= 'W' then do;
											_nf_[i]='a';
										end;
										else do;
											_nf_[i]='';
										end;
									end;
								end;
								else if _np_[1]= 'S' then do;
									do i=1 to dim(_o_);
										if _o_[i]= 'C' or _o_[i]= 'G' then do;
											_nf_[i]='o';
										end;
										else if _o_[i]= 'S' then do;
											_nf_[i]='a';
										end;
										else do;
											_nf_[i]='';
										end;
									end;
								end;
								else if _np_[1]= 'Y' then do;
									do i=1 to dim(_o_);
										if _o_[i]= 'C' or _o_[i]= 'T' then do;
											_nf_[i]='o';
										end;
										else if _o_[i]= 'Y' then do;
											_nf_[i]='a';
										end;
										else do;
											_nf_[i]='';
										end;
									end;
								end;
								else if _np_[1]= 'K' then do;
									do i=1 to dim(_o_);
										if _o_[i]= 'G' or _o_[i]= 'T' then do;
											_nf_[i]='o';
										end;
										else if _o_[i]= 'K' then do;
											_nf_[i]='a';
										end;
										else do;
											_nf_[i]='';
										end;
									end;
								end;

							end;
						%end;
					end; 
					*output Sel_recodedData;
				end;
			end;
			else do;
				_mktype= 'Undefined';
			end;
			output _recodedData;
			/*S1364776: lists of names of individuals to drop when keepvars is not empty*/
			/*drop &IndVars i %if &PloidyLevel > 2 %then _offsexp;;*/
			%if &DropIndBlock or &PloidyLevel > 2 %then %do;
				drop 
				%do i=1 %to &DropIndBlock;
					%let j= 1;
					%let DropInd= %scan(&&DropTmpIndVars&i, &j);
					%do %while(%length(%scan(&&DropTmpIndVars&i, &j)));	
						&DropInd
						%let j= %sysevalf(&j + 1);
						%let DropInd= %scan(&&DropTmpIndVars&i, &j);
					%end;
				%end;
				_offsexp
				;
			%end;
			
		run;	
		
		/*data outlib.gbs_recoded&OutName; set _recodedData; run;*/
		
		%if %length(&MissCodedMarker) %then %do;
			%if %length(&MissCodedP1) and %length(&MissCodedP2) %then %do;
				%let string1= Invalid codes &MissCodedP1 and &MissCodedP2 were found in both parents.;
			%end;
			%else %do;
				%if %length(&MissCodedP1) %then 
					%let MissCode= &MissCodedP1; 
				%else 
					%let MissCode= &MissCodedP2; 
					
				%let string1= An invalid code &MissCode was found in one of the parents.;
			%end;
			%put ERROR: An error ocurred while recoding marker &MissCodedMarker.. &string1. The Valid codes are: A, C, G, T, M, R, W, S, Y, K, N. Please, check the parents coding for this marker.;
			%let exiterror=1;
			%goto exit;
		%end;

		%let ChangeDType= 1;
		%put ChangeDType= &ChangeDType;
		%if &ChangeDType %then %do;
		
			data Sel_recodedData;
				set _recodedData;
				where _selected= 1;
			run;
		
			%let runKMC = 0;
			%if &runKMC %then %do;
				data Sel_recodedData_Num;
					set Sel_recodedData; 
					where (_mktype= 'D1.10' or _mktype= 'D2.15'); 
					array _nu_(&_nind_);
					array _nf_{*} _nf_:;
					do i=1 to dim(_nf_);
						if trim(left(_nf_[i]))= 'a' then _nu_[i]= 0;
						else if trim(left(_nf_[i]))= 'ab' then _nu_[i]= 1;
						else _nu_[i]= .;
					end;
					keep &IDVar 
						%if %length(&Parent1Var) %then &Parent1Var;
						%if %length(&Parent2Var) %then &Parent2Var;
						_newP1 _newP2 _recP1 _recP2 _mktype _testI_OR_testII nu:;
				run;
				%let fclusdat= Sel_recodedData_Num;
				/*data outlib.Sel_recodedData_Num; set Sel_recodedData_Num; run;*/
				
				%*goto exit;
	
				/*Markers are in rows. Centering and/or scaling markers. It is a good 
				practice to center and scale rows to avoid differences in clustering outcomes. If, for 
				instance, markers in a backcross data were coded as 1 and 2, or 0 and 1, these would
				produce different results if they were not centered and scaled.*/
				%let CenterRows = Yes; 
				%let ScaleRows = Yes; 
				%let ClusterMethod= Automated Radius K Means;
				%let Radius = .6;
				%*let ClusterMethod= Automated K Means;
				%*let NumberOfClusters= 2;
				%let ListVars= nu:;
	
				**center and/or scale rows;
				%if &CenterRows=Yes or &ScaleRows=Yes %then %do;
					data ind;
						set &fclusdat;
	
						%if &CenterRows=Yes %then %do;
							_rowmean_ = mean(of &ListVars);
						%end;
						%else %do;
							_rowmean_ = 0;
						%end;
	
						%if &ScaleRows=Yes %then %do;
							_rowstd_ = std(of &ListVars);
						%end;
						%else %do;
							_rowstd_ = 1;
						%end;
	
						array _v_{*} &ListVars;
						do i=1 to dim(_v_);
							if _rowstd_ > 0 then _v_[i] = (_v_[i] - _rowmean_)/_rowstd_;
							else _v_[i] = .;
						end;
						drop i;
					run;	
					%let fclusdat = ind;		
				%end;	
	
				**determine radius as a function of a correlation;
				%if %index(&MarkerReductionMethod,Radius) %then %do;
					data std;
						set &fclusdat;
						_rowuss_ = uss(of &ListVars);
						keep _rowuss_;
					run;
					proc means data=std noprint;
						var _rowuss_;
						output out=stats(drop=_TYPE_);
					run;
					data stats;
						set stats;
						if _STAT_ in ("MEAN");
						call symputx('mean_uss',_rowuss_);
					run;
	
					data _null_;
						call symputx('Radius',sqrt(2 * &mean_uss * (1 - &CorrRadius)));
					run;
	
					%put mean_uss = &mean_uss;
				%end;
	
				ods exclude all;
				ods noresults;
				proc fastclus data=&fclusdat out=kmd(rename=(cluster=_KMCluster_ distance=_KMClusterDistance_) drop= nu:) outseed=kmc(rename=(cluster=_KMCluster_) drop= nu:)
					%if %index(&ClusterMethod,Radius) %then %do; 
						%put radius=&Radius;
						radius=&Radius 
						maxclusters=100000
					%end;
					%else %do; 
						%put maxclusters=&NumberOfClusters;
						maxclusters=&NumberOfClusters 
					%end; 
					;
				   var &ListVars; 
				run;
				ods exclude none;
				ods results;
	
				/*data outlib.kmd; set kmd; run;*/
				/*data outlib.kmc; set kmc; run;*/
			%end;
	
			%macro ChangeDTypeMarkers(_TestData=, _TestType=, _TestTypeOorE=, _nTestObs=, _RefData=, _RefType=, _RefTypeOorE=, _maxrf=, _minrf=);
				%let tsuf1= %trim(&_TestType);
				%let tsuf2= %trim(&_TestTypeOorE);
				%let rsuf1= %trim(&_RefType);
				%let rsuf2= %trim(&_RefTypeOorE);
				%let tsuf1= %sysfunc(substr(&tsuf1, 1, 2));
				%let tsuf2= %sysfunc(substr(&tsuf2, 1, 1));
				%let rsuf1= %sysfunc(substr(&rsuf1, 1, 2));
				%let rsuf2= %sysfunc(substr(&rsuf2, 1, 1));
				%put tsuf1= &tsuf1;
				%put tsuf2= &tsuf2;
				%put _TestTypeOorE= &_TestTypeOorE;
				
				%if %index(%upcase(&_RefData), D1) %then %let _newmktype= D1.10; 
				%else %if %index(%upcase(&_RefData), D2) %then %let _newmktype= D2.15; 
				%else %do;
					%put ERROR: Invalid _RefData;
					%let exiterror= 1;
					%goto exitChange;
				%end;

				data testmk_all;
					set _null_;
				run;

				%global appendtestmk_all;
				%let appendtestmk_all= 0;
				%do i=1 %to &_nTestObs;

					data testmk(keep= _newP1 _newP2 _recP1 _recP2 _by _tmpIDVar _mo:);
						set &_TestData(keep= _newP1 _newP2 _recP1 _recP2 _by &IDVar /*&IndVars*/ _nf_: _n);
						where _n= &i;
						array _mo_(&_nind_) $3;
						array _mk_{*} _nf_:;
						do i= 1 to dim(_mk_);
							_mo_[i]= trim(left(_mk_[i]));
						end;
						call symputx("mkname_", trim(left(&IDVar)));
						rename &IDVar= _tmpIDVar;
					run;
					%put mkname_&i= &mkname_;

					/*data outlib.testmk&i; set testmk; run;*/

					data testmkm;
						merge &_RefData(keep= _by &IDVar _nf_:) testmk;
						by _by;
						if _stop = 1 then do;
							stop;
						end;
						array _mo_{*} _mo_:;
						array _mk_{*} _nf_:;
						_same= 0;
						_diff= 0;
						do i=1 to dim(_mk_);
							if (trim(left(_mk_[i]))= 'a' and trim(left(_mo_[i]))= 'a') 
								or (trim(left(_mk_[i]))= 'ab' and trim(left(_mo_[i]))= 'ab') 
							then _same= _same + 1; 
							else if (trim(left(_mk_[i]))= 'a' and trim(left(_mo_[i]))= 'ab') 
								or (trim(left(_mk_[i]))= 'ab' and trim(left(_mo_[i]))= 'a') 
							then _diff= _diff + 1;
						end;
						if (_same + _diff) > 0 then do;
							if _same <= _diff then do;
								_rf= _same / (_same + _diff);
							end;
							else do;
								_rf= _diff / (_same + _diff);
							end;
						end;
						else _rf = 1;
						if _rf < &_maxrf then do;
							*_tmpmktype= "&_RefType";
							_tmpmktype= "&_newmktype";
							%if &tsuf1 ^= &rsuf1 %then %do;
								_tmpnewP1= _newP2;
								_tmpnewP2= _newP1;
								_tmprecP1= _recP2;
								_tmprecP2= _recP1;
							%end;
							%else %do;
								_tmpnewP1= _newP1;
								_tmpnewP2= _newP2;
								_tmprecP1= _recP1;
								_tmprecP2= _recP2;
							%end;
							output;
						end;
						if _rf < &_minrf then do;
							_stop = 1;
						end;
						retain _stop;
					run;

					/*data outlib.testmkm&tsuf1.&tsuf2.&rsuf1.&rsuf2._&i; set testmkm; run;*/

					%NObsVars(testmkm);
					%if &nobs > 1 %then %do;
						proc sort data=testmkm; 
							by _rf;
						run;

						data testmkm;
							set testmkm(obs=1);
						run;
					%end;

					/*data outlib.testmkm&i; set testmkm; run;*/

					%if &appendtestmk_all %then %do;
						proc append base=testmk_all data= testmkm(keep= _tmpIDVar _tmpnewP1 _tmpnewP2 _tmprecP1 _tmprecP2  _tmpmktype _rf); 
					%end;
					%else %do;
						 data testmk_all;
							set testmkm(keep= _tmpIDVar _tmpnewP1 _tmpnewP2 _tmprecP1 _tmprecP2 _tmpmktype _rf);
						 run;

						 %let appendtestmk_all= 1;
					%end;
				%end;

				/*data outlib.testmk_all&tsuf1.&tsuf2.&rsuf1.&rsuf2; set testmk_all; run;*/

				%exitChange:;
				
			%mend ChangeDTypeMarkers;

			%let nD1obs= 0;
			data D1obs;
				set Sel_recodedData end= _eof_;
				where _mktype = "D1.10" and upcase(trim(left(_testI_OR_testII))) = "II"
					and _deliberate = 1;
				_n= _n_;
				_by= 1;
				if _eof_ then call symputx("nD1obs", _n_);
			run;

			%let nD2obs= 0;
			data D2obs;
				set Sel_recodedData end= _eof_;
				where _mktype = "D2.15" and upcase(trim(left(_testI_OR_testII))) = "II"
					and _deliberate = 1;
				_n= _n_;
				_by= 1;
				if _eof_ then call symputx("nD2obs", _n_);
			run;

			%let nD1exp= 0;
			data D1exp;
				set Sel_recodedData end= _eof_;
				where _mktype = "D1.10" and upcase(trim(left(_testI_OR_testII))) = "I";
				_n= _n_;
				_by= 1;
				if _eof_ then call symputx("nD1exp", _n_);
			run;

			%let nD2exp= 0;
			data D2exp;
				set Sel_recodedData end= _eof_;
				where _mktype = "D2.15" and upcase(trim(left(_testI_OR_testII))) = "I";
				_n= _n_;
				_by= 1;
				if _eof_ then call symputx("nD2exp", _n_);
			run;
			
			/*data OutLib.gbs_Sel_recodedData0; set Sel_recodedData; run;*/

			**evaluating the quality of D2.15 assigned from offsprings*;
			%put nD2obs nD1exp nD2exp= &nD2obs &nD1exp &nD2exp;
			%let D2obsResultExist= 0;
			%if &nD2obs > 0 and (&nD1exp > 0 or &nD2exp > 0) %then %do;
				%if &nD2obs > 0 and &nD1exp > 0 %then %do;
					%ChangeDTypeMarkers(_TestData=D2obs, _TestType= D2.15, _TestTypeOorE=o, _nTestObs=&nD2obs, _RefData=D1exp, _RefType= D1.10, _RefTypeOorE=e, _maxrf= &maxrf, _minrf= &minrf);
					%if &syserr>4 %then %do;
						%put ERROR: Error in the segregation ratio macro.;
						%let exiterror= 1;
						%goto exit;
					%end;

					%put D2D1_0;
					%if &appendtestmk_all %then %do;
						%put D2D1_1;
						data D2obsResult;
							set testmk_all;
						run;
						%let D2obsResultExist= 1;
						
						/*data outlib.testmk_allD2D1; set testmk_all; run;*/
					%end;
				%end;
				%if &nD2obs > 0 and &nD2exp > 0 %then %do;
					%ChangeDTypeMarkers(_TestData=D2obs, _TestType= D2.15, _TestTypeOorE=o, _nTestObs=&nD2obs, _RefData=D2exp, _RefType= D2.15, _RefTypeOorE=e, _maxrf= &maxrf, _minrf= &minrf);
					%if &syserr>4 %then %do;
						%put ERROR: Error in the segregation ratio macro.;
						%let exiterror= 1;
						%goto exit;
					%end;
					
					%if &appendtestmk_all and &nD2obs > 0  and &nD1exp > 0 %then %do;
						/*data outlib.testmk_allD2D2; set testmk_all; run;*/
					
						proc append base=D2ObsResult data=testmk_all;
						
						**filtering out ties between D2-D2 and D1-D2 recombination*;						
						proc sort data= D2ObsResult;
							by _tmpIDVar descending _rf;
						run;
						
						data D2ObsResult;
							set D2ObsResult;
							by _tmpIDVar;
							if first._tmpIDVar then do;
								temp_rf= _rf;
								temp_mk= _tmpIDVar;
								retain temp_rf temp_mk;
							end;
							if temp_mk = _tmpIDVar and abs(temp_rf - _rf) > &rfdif then output;
							else if temp_mk ^= _tmpIDVar then output;
						run;
						/*data outlib.D2ObsResult; set D2ObsResult; run;*/
												
						%let D2obsResultExist= 1;
					%end;
					%else %if &appendtestmk_all %then %do;
						data D2obsResult;
							set testmk_all;
						run;
						%let D2obsResultExist= 1;
					%end;
				%end;

				%if &D2obsResultExist %then %do;
					proc sql UNDO_POLICY= NONE;
						create table Sel_recodedData as
						select a1.*, a2.*
						from 
						Sel_recodedData as a1 
						left join 
						D2obsResult(rename= (_tmpIDVar=_tmpIDVarD2 
									_tmpnewP1=_tmpnewP1D2 
									_tmpnewP2=_tmpnewP2D2 
									_tmprecP1=_tmprecP1D2 
									_tmprecP2=_tmprecP2D2
									_tmpmktype=_tmpmktypeD2 
									_rf=_rfD2)) as a2
						%if &IDVarType = C %then %do;
							on upcase(a1.&IDVar) = upcase(a2._tmpIDVarD2)
						%end;
						%else %do;
							on a1.&IDVar = a2._tmpIDVarD2
						%end;
						order by a1.OrigMarkerOrder;
					quit;
				%end;
			%end;

			**evaluating the quality of D1.10 assigned from offsprings*;
			%put nD1obs nD1exp nD2exp= &nD1obs &nD1exp &nD2exp;
			%let D1obsResultExist= 0;
			%if &nD1obs > 0 and (&nD1exp > 0 or &nD2exp > 0) %then %do;
				%if &nD1obs > 0 and &nD2exp > 0 %then %do;
					%ChangeDTypeMarkers(_TestData=D1obs, _TestType= D1.10, _TestTypeOorE=o, _nTestObs=&nD1obs, _RefData=D2exp, _RefType= D2.15, _RefTypeOorE=e, _maxrf= &maxrf, _minrf= &minrf);
					%if &syserr>4 %then %do;
						%put ERROR: Error in the segregation ratio macro.;
						%let exiterror= 1;
						%goto exit;
					%end;
					
					%if &appendtestmk_all %then %do;
						data D1obsResult;
							set testmk_all;
						run;
						%let D1obsResultExist= 1;
						
						/*data outlib.testmk_allD1D2; set testmk_all; run;*/
					%end;
				%end;
				%if &nD1obs > 0 and &nD1exp > 0 %then %do;
					%ChangeDTypeMarkers(_TestData=D1obs, _TestType= D1.10, _TestTypeOorE=o, _nTestObs=&nD1obs, _RefData=D1exp, _RefType= D1.10, _RefTypeOorE=e, _maxrf= &maxrf, _minrf= &minrf);
					%if &syserr>4 %then %do;
						%put ERROR: Error in the segregation ratio macro.;
						%let exiterror= 1;
						%goto exit;
					%end;
					
					%if &appendtestmk_all and &nD1obs > 0 and &nD2exp > 0 %then %do;
						/*data outlib.testmk_allD1xD1; set testmk_all; run;*/
					
						proc append base=D1ObsResult data=testmk_all;

						**filtering out ties between D1-D1 and D1-D2 recombination*;
						proc sort data= D1ObsResult;
							by _tmpIDVar descending _rf;
						run;
						
						data D1ObsResult;
							set D1ObsResult;
							by _tmpIDVar;
							if first._tmpIDVar then do;
								temp_rf= _rf;
								temp_mk= _tmpIDVar;
								retain temp_rf temp_mk;
							end;
							if temp_mk = _tmpIDVar and abs(temp_rf - _rf) > &rfdif then output;
							else if temp_mk ^= _tmpIDVar then output;
						run;
						
						%let D1obsResultExist= 1;
					%end;
					%else %if &appendtestmk_all %then %do;
						data D1obsResult;
							set testmk_all;
						run;
						%let D1obsResultExist= 1;
					%end;
				%end;

				%if &D1obsResultExist %then %do;
					proc sql UNDO_POLICY= NONE;
						create table Sel_recodedData as
						select a1.*, a2.*
						from 
						Sel_recodedData as a1 
						left join 
						D1obsResult(rename= (_tmpIDVar=
									_tmpIDVarD1 
									_tmpnewP1=_tmpnewP1D1 
									_tmpnewP2=_tmpnewP2D1 
									_tmprecP1=_tmprecP1D1 
									_tmprecP2=_tmprecP2D1 
									_tmpmktype=_tmpmktypeD1 
									_rf=_rfD1)) as a2
						%if &IDVarType = C %then %do;
							on upcase(a1.&IDVar) = upcase(a2._tmpIDVarD1)
						%end;
						%else %do;
							on a1.&IDVar = a2._tmpIDVarD1
						%end;
						order by a1.OrigMarkerOrder;
					quit;
				%end;
			%end;

			%put D2obsResultExist D1obsResultExist= &D2obsResultExist &D1obsResultExist;
			%if &D2obsResultExist or &D1obsResultExist %then %do;
				data Sel_recodedData;
					set Sel_recodedData(rename= (_mktype= _origmktype));
					%if &D2obsResultExist %then %do;
						if upcase(trim(left(_origmktype))) = "D2.15" and upcase(trim(left(_testI_OR_testII))) = 'II' then do;
							if not missing(_tmpIDVarD2) then do;
								_newP1= _tmpnewP1D2;
								_newP2= _tmpnewP2D2;
								_mktype= _tmpmktypeD2;
								_recP1= _tmprecP1D2;
								_recP2= _tmprecP2D2;
								_selected= 1;
							end;
							else do;
								*delete;
								_mktype= _origmktype;
								_selected= 0;
							end;
						end;
						else do;
							_mktype= _origmktype;
						end;
					%end;
					%if &D1obsResultExist %then %do;
						if upcase(trim(left(_origmktype))) = "D1.10" and upcase(trim(left(_testI_OR_testII))) = 'II' then do;
							if not missing(_tmpIDVarD1) then do;
								_newP1= _tmpnewP1D1;
								_newP2= _tmpnewP2D1;
								_mktype= _tmpmktypeD1;
								_recP1= _tmprecP1D1;
								_recP2= _tmprecP2D1;
								_selected= 1;
								_deliberate= 0;
							end;
							else do;
								*delete;
								_mktype= _origmktype;
								_selected= 0;
								_deliberate= 1;
							end;
						end;
						else do;
							_mktype= _origmktype;
						end;
					%end;
				run;
			%end;
			%else %do;
				data Sel_recodedData;
					set Sel_recodedData;
					if ((_mktype = "D1.10" and upcase(trim(left(_testI_OR_testII))) = 'II') 
						or (_mktype = "D2.15" and upcase(trim(left(_testI_OR_testII))) = 'II')) 
						and _deliberate = 1 then do;
						_selected= 0;
					end;
				run;
			%end;
			
			/*data OutLib.gbs_Sel_recodedData1; set Sel_recodedData; run;*/
			
			data Sel_recodedData;
				set Sel_recodedData;
				*where _selected= 1;
				where _selected= 1 and _deliberate = 0;
			run;
		
		%end; /*end if ChangeDType*/
		%else %do;
			/*data OutLib.gbs_Sel_recodedData1; set Sel_recodedData; run;*/
		
			data Sel_recodedData;
				set Sel_recodedData;
				*where _selected= 1;
				where _selected= 1 and _deliberate = 0;
			run;		
		%end;
			
		%NObsVars(Sel_recodedData);
		%let nobsr= &nobs;
			
		%if &nobsr = 0 %then %do;
			%put ERROR: Aborting process because none of the markers could be classified correctly.;
			%let exiterror=1;
			%goto exit;
		%end;
			
		data _nf_tmp;
			set Sel_recodedData(keep= _nf_:);
			where 0;
		run;
		%NObsVars(_nf_tmp);
		%let nvarsr= &nvars;
		
		%let BlockSize= %sysfunc(min(%sysevalf(3000000/&nobsr,floor),&nvarsr));	
		%put BlockSize= &BlockSize;
		
		%let first1 = 1;
		%let last1 = &BlockSize;
		%let Block=1;
		%do %until (&&first&Block > &nvars );
			%let first%eval(&Block + 1) = %eval(&&last&Block + 1);
			%let last%eval(&Block + 1) = %eval(&&last&Block + &BlockSize);
			%let Block = %eval(&Block + 1);
			%*put first&Block &&first&Block lastb &&last&Block;
			%if &&last&Block > &nvars %then %let last&Block = &nvars;
			%*put first&Block &&first&Block lastb &&last&Block;
		%end;
		%let NBlocks= %sysevalf(&Block - 1);
		%put NBlocks= &NBlocks;
		
		%do i=1 %to &NBlocks;
			proc transpose data= Sel_recodedData(keep= _nf_&&first&i - _nf_&&last&i /*%if &i=1 %then*/ &IDVar/*;*/) out=Sel_recodedDatat(drop=_name_);
				var _nf_&&first&i - _nf_&&last&i;
				id &IDVar;
			run;
			
			proc datasets nolist;
				%if &i=1 %then %do;
					delete SelAnd_recoded_Geno;
					change Sel_recodedDatat = SelAnd_recoded_Geno;
				%end;
				%else %do;
					append base=SelAnd_recoded_Geno data=Sel_recodedDatat force;
					delete Sel_recodedDatat;
				%end;
			quit;
		%end;
	
		/*
		proc transpose data= Sel_recodedData(keep= _nf_: &IDVar) out=SelAnd_recoded_Geno(drop=_name_);
			var _nf_:;
			id &IDVar;
		run;
		*/
		
		%let oldnames= _a _c _g _t _m _r _w _S _y _k _n _tot_ind _AA_testI _AB_testI _BB_testI 
			_mis_testI _genotype_error_testI _chisq_testI _pchisq_testI _mktype_testI
			_offsobs _AA_testII _AB_testII _BB_testII _mis_testII _chisq_testII _pchisq_testII _mktype_testII 
			_sum_indicator_testI _sum_indicator_testII _uniquedosage _singledosage _testI_OR_testII _selected 
			_mktype  _chisq _pchisq 
			_n_ind _n_dom _n_het _n_rec _n_mis _n_err _p_err _newP1 _newP2 _recP1 _recP2;
		%let newnames= n_A n_C n_G n_T n_M n_R n_W n_S n_Y n_K n_N Total_Ind AA_testI AB_testI BB_testI
			mis_testI genotype_error_testI chisq_testI pchisq_testI mktype_testI
			offsobs AA_testII AB_testII BB_testII mis_testII chisq_testII pchisq_testII mktype_testII
			sum_indicator_testI sum_indicator_testII uniquedosage singledosage testI_OR_testII selected
			mktype chisq pchisq 
			n_ind n_A1A1 n_A1A2 n_A2A2 n_mis n_err p_err &newP1Var &newP2Var &recP1Var &recP2Var;
		%let labels= A obs @$@C obs @$@G obs @$@T obs @$@M obs @$@R obs @$@W obs @$@S obs @$@Y obs @$@K obs @$@N obs @$@Total Obs @$@AA Obs for Test I @$@AB Obs for Test I @$@BB Obs for Test I
			@$@Miss Obs for Test I @$@Genotyping Error for Test I @$@Chi-square of Test I @$@p Value of Test I @$@Marker Type for Test I
			@$@Offspring Obs @$@AA Obs for Test II @$@AB Obs for Test II @$@BB Obs for Test II @$@Miss Obs for Test II @$@Chi-square of Test II @$@p Value of Test II @$@Marker Type for Test II
			@$@Sum of Indicators for Test I @$@Sum of Indicators for Test II @$@Unique Dosage @$@Single Dosage @$@Test I or II @$@Selected Markers
			@$@Marker Type @$@Chi-square @$@p Value 
			@$@Total Individuals @$@AA Obs @$@AB Obs @$@BB Obs @$@Miss Obs @$@Genotyping Error @$@Ratio Of Genotyping Error @$@&newP1Lab @$@&newP2Lab @$@&recP1Lab @$@&recP2Lab;
		
		%let k= 1;
		%let _oldname= %scan(&oldnames, &k);
		%let _newname= %scan(&newnames, &k);
		%let _label= %scan(&labels, &k, @$@);
		%let OldNamesToNewNames=;
		%let LabelsStmt=;
		%do %while(%length(&_newname));
			%let OldNamesToNewNames= &OldNamesToNewNames &_oldname= &_newname;
			%let LabelsStmt= &LabelsStmt &_oldname= "&_label";
			%let k= %sysevalf(&k + 1);
			%let _oldname= %scan(&oldnames, &k);
			%let _newname= %scan(&newnames, &k);	
			%let _label= %scan(&labels, &k, @$@);
		%end;
		%put OldNamesToNewNames= &OldNamesToNewNames;
		%put LabelsStmt= &LabelsStmt;

		%let Miss_mktype = 0;
		data SelAnd_recoded_Anno(keep= &IDVar 
						%if %length(&Parent1Var) %then &Parent1Var; 
						%if %length(&Parent2Var) %then &Parent2Var;
						&KeepVars &newnames OrigMarkerOrder);
			retain &IDVar 
				%if %length(&Parent1Var) %then &Parent1Var; 
				%if %length(&Parent2Var) %then &Parent2Var;
				&KeepVars &oldnames OrigMarkerOrder;
			set Sel_recodedData;
			if missing(_mktype) then 
				call symputx("Miss_mktype", 1);
			rename &OldNamesToNewNames;
			label &LabelsStmt;
		run;

		%if &Miss_mktype = 1 %then %do;
			%put ERROR: Markers with missing type were detected.;
		%end;
		
		/*data Outlib.Sel_recodedData; set Sel_recodedData; run;*/
		
		%let oldDosageVarsexp=;
		%let oldDosageVarsobs=;
		%let newDosageVarsexp=;
		%let newDosageVarsobs=;
		%let DosageLabelsexp=;
		%let DosageLabelsobs=;
		%do j=1 %to %sysevalf(&ncross);
			%let oldDosageVarsexp= &oldDosageVarsexp _chisq_testI_&&Dosage&j _pchisq_testI_&&Dosage&j _indicator_testI_&&Dosage&j;
			%let oldDosageVarsobs= &oldDosageVarsobs _chisq_testII_&&Dosage&j _pchisq_testII_&&Dosage&j _indicator_testII_&&Dosage&j;
			%let newDosageVarsexp= &newDosageVarsexp chisq_testI_&&Dosage&j pchisq_testI_&&Dosage&j indicator_testI_&&Dosage&j;
			%let newDosageVarsobs= &newDosageVarsobs chisq_testII_&&Dosage&j pchisq_testII_&&Dosage&j indicator_testII_&&Dosage&j;
			%let DosageLabelsexp= &DosageLabelsexp @$@Chi-square Test I for Dosages &&Dosage&j @$@p Value of Test I for Dosages &&Dosage&j @$@Indicator of Test I for Dosages &&Dosage&j;
			%let DosageLabelsobs= &DosageLabelsobs @$@Chi-square Test II for Dosages &&Dosage&j @$@p Value of Test II for Dosages &&Dosage&j @$@Indicator of Test II for Dosages &&Dosage&j;
		%end; 
			
		%let keepFrom_recoded= _a _c _g _t _m _r _w _S _y _k _n _tot_ind 
			_AA_testI _AB_testI _BB_testI _mis_testI _genotype_error_testI 
			&oldDosageVarsexp			
			_chisq_testI _pchisq_testI _mktype_testI
			_offsobs _AA_testII _AB_testII _BB_testII _mis_testII 
			&oldDosageVarsobs 		
			_chisq_testII _pchisq_testII _mktype_testII
			_sum_indicator_testI _sum_indicator_testII _uniquedosage _singledosage _testI_OR_testII 
			_n_ind _n_dom _n_het _n_rec _n_mis _n_err _p_err _chisq _pchisq;		
		%let keepFromSel_recoded= _selected _mktype _newP1 _newP2 _recP1 _recP2;
		%let oldnames2= &keepFrom_recoded &keepFromSel_recoded;
			
		%let keepFrom_recoded_New= n_A n_C n_G n_T n_M n_R n_W n_S n_Y n_K n_N Total_Ind
			AA_testI AB_testI BB_testI mis_testI genotype_error_testI 
			&newDosageVarsexp
			chisq_testI pchisq_testI mktype_testI		
			offsobs AA_testII AB_testII BB_testII mis_testII 
			&newDosageVarsobs 			
			chisq_testII pchisq_testII mktype_testII
			sum_indicator_testI sum_indicator_testII uniquedosage singledosage testI_OR_testII 
			n_ind n_A1A1 n_A1A2 n_A2A2 n_mis n_err p_err chisq pchisq;
		%let keepFromSel_recoded_New= selected mktype &newP1Var &newP2Var &recP1Var &recP2Var;	
		%let newnames2= &keepFrom_recoded_New &keepFromSel_recoded_New;
		
		%let keepFrom_recoded_New_Labels= A Obs @$@C Obs @$@G Obs @$@T Obs @$@M Obs @$@R Obs @$@W Obs @$@S Obs @$@Y Obs @$@K Obs @$@N Obs @$@Total Obs 
			@$@AA Obs for Test I @$@AB Obs for Test I @$@BB Obs for Test I @$@Miss Obs for Test I @$@Genotyping Error for Test I 
			&DosageLabelsexp
			@$@Chi-square of Test I @$@p Value of Test I @$@Marker Type for Test I		
			@$@Offspring Obs @$@AA Obs for Test II @$@AB Obs for Test II @$@BB Obs for Test II @$@Miss Obs for Test II 
			&DosageLabelsobs			
			@$@Chi-square of Test II @$@p Value of Test II @$@Marker Type for Test II
			@$@Sum of Indicators for Test I @$@Sum of Indicators for Test II @$@Unique Dosage @$@Single Dosage @$@Test I or II 
			@$@Total Individuals @$@AA Obs @$@AB Obs @$@BB Obs @$@Miss Obs @$@Genotyping Error @$@Rate of Genotyping Error @$@Chi-square @$@p Value;
		%let keepFromSel_recoded_New_Labels= @$@Selected Markers @$@Marker Type @$@&newP1Lab @$@&newP2Lab @$@&recP1Lab @$@&recP2Lab;	
		%let labels2= &keepFrom_recoded_New_Labels &keepFromSel_recoded_New_Labels;
		
		**creating lists of variables*;
		%let k= 1;
		%let _oldname2= %scan(&oldnames2, &k);
		%let _newname2= %scan(&newnames2, &k);
		%let nVarBlock= 0;
		%let TmpVars=;
		%do %while(%length(&_newname2));
			%if %length(&TmpVars) le 32700 %then %do;
				%let TmpVars= &TmpVars &_oldname2= &_newname2;
			%end;
			%else %do;
				%let nVarBlock= %sysevalf(&nVarBlock + 1);
				%let OldNames2ToNewNames2&nVarBlock = &TmpVars &_oldname2= &_newname2;
				%let TmpVars = ;
			%end;
			%let k= %sysevalf(&k + 1);
			%let _oldname2= %scan(&oldnames2, &k);
			%let _newname2= %scan(&newnames2, &k);	
		%end;
		%if (&nVarBlock = 0) or (&nVarBlock > 0 and %length(&TmpVars) > 0) %then %do;
			%let nVarBlock= %sysevalf(&nVarBlock + 1);
			%let OldNames2ToNewNames2&nVarBlock = &TmpVars;
			%let TmpVars = ;
		%end;
		%put nVarBlock= &nVarBlock;
		%put OldNames2ToNewNames2&nVarBlock= &&OldNames2ToNewNames2&nVarBlock;

		**creating lists of labels*;
		%let k= 1;
		%let _newname2= %scan(&newnames2, &k);
		%let _label2= %scan(&labels2, &k, @$@);
		%let nLabBlock= 0;
		%let TmpVars=;
		%do %while(%length(&_newname2));
			%if %length(&TmpVars) le 32700 %then %do;
				%let TmpVars= &TmpVars &_newname2= "&_label2";
			%end;
			%else %do;
				%let nLabBlock= %sysevalf(&nLabBlock + 1);
				%let LabelsStmt2&nLabBlock= &TmpVars &_newname2= "&_label2";
				%let TmpVars = ;
			%end;
			%let k= %sysevalf(&k + 1);
			%let _newname2= %scan(&newnames2, &k);	
			%let _label2= %scan(&labels2, &k, @$@);
		%end;
		%if (&nLabBlock = 0) or (&nLabBlock > 0 and %length(&TmpVars) > 0) %then %do;
			%let nLabBlock= %sysevalf(&nLabBlock + 1);
			%let LabelsStmt2&nLabBlock = &TmpVars;
			%let TmpVars = ;
		%end;
		%put nLabBlock= &nLabBlock;
		%put LabelsStmt2&nLabBlock= &&LabelsStmt2&nLabBlock;
		
		/*data outlib._recodedData; set _recodedData; run;*/
		
		%*goto exit;
			
		proc sql UNDO_POLICY= NONE;
			create table All_recodedData(rename=(
								%do i=1 %to &nRecIndBlock;
									&&recodedIndVar&i
								%end;
								
								%do i=1 %to &nVarBlock;
									&&OldNames2ToNewNames2&i
								%end;
								
								)) as
			select a1.*, a2.*
			from _recodedData(keep= &IDVar 
						%if %length(&Parent1Var) %then &Parent1Var;
						%if %length(&Parent2Var) %then &Parent2Var;
						/*&IndVars*/ &KeepVars &keepFrom_recoded OrigMarkerOrder) as a1 
			left join 
			Sel_recodedData(keep= &keepFromSel_recoded _nf_: OrigMarkerOrder rename=(OrigMarkerOrder= _OrigMarkerOrder_)) as a2
			on a1.OrigMarkerOrder= a2._OrigMarkerOrder_
			order by OrigMarkerOrder;
		quit;
		/*data outlib.All_recodedDataL; set All_recodedData; run;*/
					
		data All_recodedData;
			retain &IDVar
				%if %length(&Parent1Var) %then &Parent1Var;
				%if %length(&Parent2Var) %then &Parent2Var;	
				%do i=1 %to &nIndBlock;
					%let j= 1;
					%let InVar= %scan(&&TmpIndVars&i, &j);
					%do %while(%length(&InVar));	
						&InVar
						%let j= %sysevalf(&j + 1);
						%let InVar= %scan(&&TmpIndVars&i, &j);
					%end;
				%end;
				&KeepVars &newnames2;
			length mktype $12;
			set All_recodedData;
			*if missing(mktype) then do; /*needed because of the proc sql left join done previously*/
			if missing(selected) then do; /*needed because of the proc sql left join done previously*/
				mktype= 'Undefined';
				selected = 0;
			end;
			drop _OrigMarkerOrder_;
			label 
				%do i=1 %to &nLabBlock;
					&&LabelsStmt2&i
				%end;
				;
		run;
		
		%let PlotVar= mktype;
		%let PlotLabel= Marker Type;
		
		%*goto exit;

	%end; /*end if crosstype is Outcross*/
	%else %do;
		**SetHetToMiss = Yes will set all heterozygous genotypes from RI0, RI1, RI2, RILSelf and RILSib to missing*;
		%*let SetHetToMiss = No;
	
		%put ParentVars0 Parent1Var= &Parent1Var Parent2Var= &Parent2Var;	

		%if %index(%upcase(&CrossType),BC) %then %do;
			** This code checks if Parent1Var is the backcrossed parent for BC1*;
			** This code checks if Parent2Var is the backcrossed parent for BC2*;
			%if %length(&Parent1Var &Parent2Var) %then %do;
				%let bc_ok= 0;
				data checkBC;
					*set &subIndata(firstobs=1 obs=20);
					set &subIndata;
					**defining temporary arrays*;
					if _stop = 1 then do;
						stop;
					end;
					array _o_{*} &IndVars;
					_offsA='   ';
					_offsC='   ';
					_offsG='   ';
					_offsT='   ';
					_offsM='   ';
					_offsR='   ';
					_offsW='   ';
					_offsS='   ';
					_offsY='   ';
					_offsK='   ';
					_offsobs='                                   ';
					_bc_ok= 0;
					do i=1 to dim(_o_);
						_o_[i]= upcase(trim(left(_o_[i])));
						if _o_[i]= 'A' then do;
							_offsA= 'A';
						end;
						else if _o_[i]= 'C' then do;
							_offsC= 'C';
						end;
						else if _o_[i]= 'G' then do;
							_offsG= 'G';
						end;
						else if _o_[i]= 'T' then do;
							_offsT= 'T';
						end;
						else if _o_[i]= 'M' then do;
							_offsM= 'M';
						end;
						else if _o_[i]= 'R' then do;
							_offsR= 'R';
						end;
						else if _o_[i]= 'W' then do;
							_offsW= 'W';
						end;
						else if _o_[i]= 'S' then do;
							_offsS= 'S';
						end;
						else if _o_[i]= 'Y' then do;
							_offsY= 'Y';
						end;
						else if _o_[i]= 'K' then do;
							_offsK= 'K';
						end;
					end;
					_offsobs=_offsA||_offsC||_offsG||_offsT||_offsM||_offsR||_offsW||_offsS||_offsY||_offsK;
					_offsobs=COMPRESS(_offsobs);
					%if %index(%upcase(&CrossType),BC1) and %length(&Parent1Var) %then %do;
						if upcase(trim(left(&Parent1Var))) ^= 'N' then do;
							if indexc(trim(left(_offsobs)), upcase(trim(left(&Parent1Var)))) then _bc_ok= 1;
							else _bc_ok= 0;
						 	_stop= 1;
						 	call symputx("bc_ok", _bc_ok);
						 	output;
						end;
					%end;
					%else %if %index(%upcase(&CrossType),BC1) and %length(&Parent2Var) %then %do;
						if upcase(trim(left(&Parent2Var))) ^= 'N' then do;
							if indexc(trim(left(_offsobs)), upcase(trim(left(&Parent2Var)))) then _bc_ok= 0;
							else _bc_ok= 1;
						 	_stop= 1;
						 	call symputx("bc_ok", _bc_ok);
						 	output;
						end;
					%end;
					%else %if %index(%upcase(&CrossType),BC2) and %length(&Parent2Var) %then %do;
						if  upcase(trim(left(&Parent2Var))) ^= 'N' then do;
							if indexc(trim(left(_offsobs)), upcase(trim(left(&Parent2Var)))) then _bc_ok= 1;
							else _bc_ok= 0;
							_stop= 1;
							call symputx("bc_ok", _bc_ok);
							output;
						end;
					%end;
					%else %if %index(%upcase(&CrossType),BC2) and %length(&Parent1Var) %then %do;
						if  upcase(trim(left(&Parent1Var))) ^= 'N' then do;
							if indexc(trim(left(_offsobs)), upcase(trim(left(&Parent1Var)))) then _bc_ok= 0;
							else _bc_ok= 1;
							_stop= 1;
							call symputx("bc_ok", _bc_ok);
							output;
						end;
					%end;
					retain _stop;
				run;
				/*data outlib.CHECKBC; set CHECKBC;run;*/
				
				%NObsVars(checkBC);
				%if &nobs > 0 %then %do;
					%put bc_ok= &bc_ok;
					%if &bc_ok = 0 %then %do;
						%if %index(%upcase(&CrossType),BC1) and %length(&Parent1Var) %then %do;
							%put WARNING: Cross Type is BC1, however, Parent 1 ID Variable is not consistent with genotypes observed in the offsprings.;
						%end;
						%else %if %index(%upcase(&CrossType),BC1) and %length(&Parent2Var) %then %do;
							%put WARNING: Cross Type is BC1, however, Parent 2 ID Variable is not consistent with genotypes observed in the offsprings.;
						%end;
						%else %if %index(%upcase(&CrossType),BC2) and %length(&Parent2Var) %then %do;
							%put WARNING: Cross Type is BC2, however, Parent 2 ID Variable is not consistent with genotypes observed in the offsprings.;
						%end;
						%else %if %index(%upcase(&CrossType),BC2) and %length(&Parent1Var) %then %do;
							%put WARNING: Cross Type is BC2, however, Parent 1 ID Variable is not consistent with genotypes observed in the offsprings.;
						%end;
						%let OrigParent1Var= &Parent1Var;
						%let OrigParent2Var= &Parent2Var;
						%let Parent1Var= &OrigParent2Var;
						%let Parent2Var= &OrigParent1Var;
					%end;
				%end;
				%put ParentVars1 Parent1Var= &Parent1Var Parent2Var= &Parent2Var;
			%end;	
			
			**Switch parents for BC2, such that the rest of the code works for both BC1 and BC2*;
			**Parent1Var must always the backcrossed parent when cross is BC*;
			%if %index(%upcase(&CrossType),BC2) %then %do;
				%let OrigParent1Var= &Parent1Var;
				%let OrigParent2Var= &Parent2Var;
				%let Parent1Var= &OrigParent2Var;
				%let Parent2Var= &OrigParent1Var;
				
				%let OrignewP1Var= &newP1Var;
				%let OrignewP2Var= &newP2Var;
				%let newP1Var= &OrignewP2Var;
				%let newP2Var= &OrignewP1Var;
				
				%*let OrigrecP1Var= &recP1Var; 
				%*let OrigrecP2Var= &recP2Var; 		
				%*let recP1Var= &OrigrecP2Var; 
				%*let recP2Var= &OrigrecP1Var; 
				
				%let codeAA= 2;
				%let codeAB= 1;
			%end;
			%else %if %index(%upcase(&CrossType),BC1) %then %do;
				%let codeAA= 0;
				%let codeAB= 1;
			%end;
			%put ParentVars2 Parent1Var= &Parent1Var Parent2Var= &Parent2Var;
		%end;		
		
		data _recodedData;
			set &subIndata;
			**defining temporary arrays*;
			array _p_{2} $5;
			%if %length(&Parent1Var) %then %do;
				_p_[1]= upcase(trim(left(&Parent1Var)));
			%end;
			%else %do;
				_p_[1]= 'N';
			%end;
			%if %length(&Parent2Var) %then %do;
				_p_[2]= upcase(trim(left(&Parent2Var)));
			%end;
			%else %do;
				_p_[2]= 'N';
			%end;
			if missing(_p_[1]) then _p_[1]= 'N';
			if missing(_p_[2]) then _p_[2]= 'N';
			if ^indexc(trim(left(_p_[1])), 'A','C','G','T','M','R','W','S','Y','K','N') or ^indexc(trim(left(_p_[2])), 'A','C','G','T','M','R','W','S','Y','K','N') then do;
				call symputx("MissCodedMarker", trim(left(&IDVar)));
				if ^indexc(trim(left(_p_[1])), 'A','C','G','T','M','R','W','S','Y','K','N') then do;
					call symputx("MissCodedP1", trim(left(_p_[1])));
				end;
				if ^indexc(trim(left(_p_[2])), 'A','C','G','T','M','R','W','S','Y','K','N') then do;
					call symputx("MissCodedP2", trim(left(_p_[2])));
				end;
			end;
			array _o_{*} &IndVars;
			array _nf_(&_nind_) 3;
			**defining new variables*;
			_a=0;_c=0;_g=0;_t=0;_m=0;_r=0;_w=0;_s=0;_y=0;_k=0;_n=0;
			_offsA='   ';
			_offsC='   ';
			_offsG='   ';
			_offsT='   ';
			_offsM='   ';
			_offsR='   ';
			_offsW='   ';
			_offsS='   ';
			_offsY='   ';
			_offsK='   ';
			*_offsexp='                                   ';
			_offsobs='                                   ';
			_tot_ind=0;
			**markers based on parents genotypes*;
			_AA_testI=0;_AB_testI=0;_BB_testI=0;_mis_testI=0;_genotype_error_testI=0; _p_genotype_error_testI= 0;
			*****;
			**markers based on offspring genotypes*;
			_tot_testII=0;
			*****;
			**final calls**;
			_newP1='   ';
			_newP2='   ';
			_recP1=.;
			_recP2=.;
			_selected= 1;
			_phased= 1;				
			*****;
			do i=1 to dim(_o_);
				_o_[i]= upcase(trim(left(_o_[i])));
				if _o_[i]= 'A' then do;
					_a=_a + 1;
					_offsA= 'A';
				end;
				else if _o_[i]= 'C' then do;
					_c=_c + 1;
					_offsC= 'C';
				end;
				else if _o_[i]= 'G' then do;
					_g=_g + 1;
					_offsG= 'G';
				end;
				else if _o_[i]= 'T' then do;
					_t=_t + 1;
					_offsT= 'T';
				end;
				else if _o_[i]= 'M' then do;
					_m=_m + 1;
					_offsM= 'M';
				end;
				else if _o_[i]= 'R' then do;
					_r=_r + 1;
					_offsR= 'R';
				end;
				else if _o_[i]= 'W' then do;
					_w=_w + 1;
					_offsW= 'W';
				end;
				else if _o_[i]= 'S' then do;
					_s=_s + 1;
					_offsS= 'S';
				end;
				else if _o_[i]= 'Y' then do;
					_y=_y + 1;
					_offsY= 'Y';
				end;
				else if _o_[i]= 'K' then do;
					_k=_k + 1;
					_offsK= 'K';
				end;
				else _n=_n + 1;
			end;
			_tot_ind= _a +_c +_g + _t + _m + _r + _w + _s + _y + _k + _n;
			_offsobs=_offsA||_offsC||_offsG||_offsT||_offsM||_offsR||_offsW||_offsS||_offsY||_offsK;
			_offsobs=upcase(COMPRESS(_offsobs));
			*****;
			%if %index(%upcase(&CrossType),BC) %then %do;
				_newP1= upcase(trim(left(_p_[1])));
				_newP2= upcase(trim(left(_p_[2])));
				_recP1= 0;
				_recP2= 2;
				if indexc(trim(left(_p_[1])), 'A','C','G','T') and indexc(trim(left(_p_[2])), 'A','C','G','T') then do;
					**both parents are homozygous;
					if _p_[1] = _p_[2] then do;
						**parents are homozygous with same genotype;
						if length(_offsobs) = 2 then do;
							if _offsobs = 'AM' then do;
								_newP1= 'A';
								_newP2= 'C';
							end;
							else if _offsobs = 'CM' then do;
								_newP1= 'C';
								_newP2= 'A';
							end;							
							else if _offsobs = 'AR' then do;
								_newP1= 'A';
								_newP2= 'G';
							end;
							else if _offsobs = 'GR' then do;
								_newP1= 'G';
								_newP2= 'A';
							end;							
							else if _offsobs = 'AW' then do;
								_newP1= 'A';
								_newP2= 'T';
							end;
							else if _offsobs = 'TW' then do;
								_newP1= 'T';
								_newP2= 'A';
							end;							
							else if _offsobs = 'CS' then do;
								_newP1= 'C';
								_newP2= 'G';
							end;
							else if _offsobs = 'GS' then do;
								_newP1= 'G';
								_newP2= 'C';
							end;
							else if _offsobs = 'CY' then do;
								_newP1= 'C';
								_newP2= 'T';
							end;
							else if _offsobs = 'TY' then do;
								_newP1= 'T';
								_newP2= 'C';
							end;
							else if _offsobs = 'GK' then do;
								_newP1= 'G';
								_newP2= 'T';
							end;
							else if _offsobs = 'TK' then do;
								_newP1= 'T';
								_newP2= 'G';
							end;
							_newP1= upcase(trim(left(_newP1)));
							_newP2= upcase(trim(left(_newP2)));
							do i=1 to dim(_o_);
								if _o_[i] = _newP1 then do;
									_nf_[i]= &codeAA;
									_AA_testI= _AA_testI + 1;
								end;
								else if (indexc(trim(left(_o_[i])), 'M', 'R', 'W') and indexc(trim(left(_newP1)), 'A')) 
									or (indexc(trim(left(_o_[i])), 'M', 'S', 'Y') and indexc(trim(left(_newP1)), 'C')) 
									or (indexc(trim(left(_o_[i])), 'R', 'S', 'K') and indexc(trim(left(_newP1)), 'G'))
									or (indexc(trim(left(_o_[i])), 'W', 'Y', 'K') and indexc(trim(left(_newP1)), 'T')) then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							_newP1= ' ';
							_newP2= ' ';
							_recP1=.;
							_recP2=.;
							do i=1 to dim(_o_);
								if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
								else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
								else _AB_testI= _AB_testI + 1;
								_nf_[i]= .;
							end;
							_selected= 0;
							_phased= 0;
						end;						
					end;
					else if (indexc(trim(left(_p_[1])), 'A') and indexc(trim(left(_p_[2])), 'C')) 
						or (indexc(trim(left(_p_[1])), 'C') and indexc(trim(left(_p_[2])), 'A')) then do;
						**parents are homozygous with distinct genotype;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'M') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;	
					end;
					else if (indexc(trim(left(_p_[1])), 'A') and indexc(trim(left(_p_[2])), 'G'))
						or (indexc(trim(left(_p_[1])), 'G') and indexc(trim(left(_p_[2])), 'A')) then do;
						**parents are homozygous with distinct genotype;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'R') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;	
					end;
					else if (indexc(trim(left(_p_[1])), 'A') and indexc(trim(left(_p_[2])), 'T')) 
						or (indexc(trim(left(_p_[1])), 'T') and indexc(trim(left(_p_[2])), 'A')) then do;
						**parents are homozygous with distinct genotype;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'W') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;	
					end;
					else if (indexc(trim(left(_p_[1])), 'C') and indexc(trim(left(_p_[2])), 'G')) 
						or (indexc(trim(left(_p_[1])), 'G') and indexc(trim(left(_p_[2])), 'C')) then do;
						**parents are homozygous with distinct genotype;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'S') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;	
					end;
					else if (indexc(trim(left(_p_[1])), 'C') and indexc(trim(left(_p_[2])), 'T')) 
						or (indexc(trim(left(_p_[1])), 'T') and indexc(trim(left(_p_[2])), 'C')) then do;
						**parents are homozygous with distinct genotype;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'Y') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;	
					end;
					else if (indexc(trim(left(_p_[1])), 'G') and indexc(trim(left(_p_[2])), 'T')) 
						or (indexc(trim(left(_p_[1])), 'T') and indexc(trim(left(_p_[2])), 'G')) then do;
						**parents are homozygous with distinct genotype;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'K') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;	
					end;
				end;
				else if (indexc(trim(left(_p_[1])), 'A','C','G','T') and indexc(trim(left(_p_[2])), 'N')) 
					or (indexc(trim(left(_p_[1])), 'N') and indexc(trim(left(_p_[2])), 'A','C','G','T')) then do;
					**one parent is homozygous and the other is missing;
					if length(_offsobs) = 2 then do;
						if _offsobs = 'AM' then do;
							_newP1= 'A';
							_newP2= 'C';
						end;
						else if _offsobs = 'CM' then do;
							_newP1= 'C';
							_newP2= 'A';
						end;							
						else if _offsobs = 'AR' then do;
							_newP1= 'A';
							_newP2= 'G';
						end;
						else if _offsobs = 'GR' then do;
							_newP1= 'G';
							_newP2= 'A';
						end;							
						else if _offsobs = 'AW' then do;
							_newP1= 'A';
							_newP2= 'T';
						end;
						else if _offsobs = 'TW' then do;
							_newP1= 'T';
							_newP2= 'A';
						end;							
						else if _offsobs = 'CS' then do;
							_newP1= 'C';
							_newP2= 'G';
						end;
						else if _offsobs = 'GS' then do;
							_newP1= 'G';
							_newP2= 'C';
						end;
						else if _offsobs = 'CY' then do;
							_newP1= 'C';
							_newP2= 'T';
						end;
						else if _offsobs = 'TY' then do;
							_newP1= 'T';
							_newP2= 'C';
						end;
						else if _offsobs = 'GK' then do;
							_newP1= 'G';
							_newP2= 'T';
						end;
						else if _offsobs = 'TK' then do;
							_newP1= 'T';
							_newP2= 'G';
						end;
						_newP1= upcase(trim(left(_newP1)));
						_newP2= upcase(trim(left(_newP2)));
						if indexc(trim(left(_p_[1])), 'A','C','G','T') then do;
							**parent 1 is homozygous;
							if _p_[1] ^= _newP1 then do;
								_newP1_1= _newP1;
								_newP1= _newP2;
								_newP2= _newP1_1;
							end;
						end;
						else do;	
							**parent 2 is homozygous;
							if trim(left(_p_[2])) ^= _newP2 then do;
								_newP2_2= _newP2;
								_newP2= _newP1;
								_newP1= _newP2_2;
							end;						
						end;
						_newP1= upcase(trim(left(_newP1)));
						_newP2= upcase(trim(left(_newP2)));
						do i=1 to dim(_o_);
							if _o_[i] = _newP1 then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if (indexc(trim(left(_o_[i])), 'M', 'R', 'W') and indexc(trim(left(_newP1)), 'A')) 
								or (indexc(trim(left(_o_[i])), 'M', 'S', 'Y') and indexc(trim(left(_newP1)), 'C')) 
								or (indexc(trim(left(_o_[i])), 'R', 'S', 'K') and indexc(trim(left(_newP1)), 'G'))
								or (indexc(trim(left(_o_[i])), 'W', 'Y', 'K') and indexc(trim(left(_newP1)), 'T')) then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					else do;
						_newP1= ' ';
						_newP2= ' ';
						_recP1=.;
						_recP2=.;
						do i=1 to dim(_o_);
							if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
							else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
							else _AB_testI= _AB_testI + 1;
							_nf_[i]= .;
						end;
						_selected= 0;
						_phased= 0;
					end;					
				end;
				else if (indexc(trim(left(_p_[1])), 'N') and indexc(trim(left(_p_[2])), 'N')) then do;
					**both parents are undefined missing;
					if length(_offsobs) > 1 and length(_offsobs) < 4 then do;
						if _offsobs = 'AM' then do;
							_newP1= 'A';
							_newP2= 'C';
						end;
						else if _offsobs = 'CM' then do;
							_newP1= 'C';
							_newP2= 'A';
						end;							
						else if _offsobs = 'AR' then do;
							_newP1= 'A';
							_newP2= 'G';
						end;
						else if _offsobs = 'GR' then do;
							_newP1= 'G';
							_newP2= 'A';
						end;							
						else if _offsobs = 'AW' then do;
							_newP1= 'A';
							_newP2= 'T';
						end;
						else if _offsobs = 'TW' then do;
							_newP1= 'T';
							_newP2= 'A';
						end;							
						else if _offsobs = 'CS' then do;
							_newP1= 'C';
							_newP2= 'G';
						end;
						else if _offsobs = 'GS' then do;
							_newP1= 'G';
							_newP2= 'C';
						end;
						else if _offsobs = 'CY' then do;
							_newP1= 'C';
							_newP2= 'T';
						end;
						else if _offsobs = 'TY' then do;
							_newP1= 'T';
							_newP2= 'C';
						end;
						else if _offsobs = 'GK' then do;
							_newP1= 'G';
							_newP2= 'T';
						end;
						else if _offsobs = 'TK' then do;
							_newP1= 'T';
							_newP2= 'G';
						end;
						_newP1= upcase(trim(left(_newP1)));
						_newP2= upcase(trim(left(_newP2)));
						do i=1 to dim(_o_);
							if _o_[i] = _newP1 then do;
								_nf_[i]= &codeAA;
								_AA_testI= _AA_testI + 1;
							end;
							else if (indexc(trim(left(_o_[i])), 'M', 'R', 'W') and indexc(trim(left(_newP1)), 'A')) 
								or (indexc(trim(left(_o_[i])), 'M', 'S', 'Y') and indexc(trim(left(_newP1)), 'C')) 
								or (indexc(trim(left(_o_[i])), 'R', 'S', 'K') and indexc(trim(left(_newP1)), 'G'))
								or (indexc(trim(left(_o_[i])), 'W', 'Y', 'K') and indexc(trim(left(_newP1)), 'T')) then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					else do;
						_newP1= ' ';
						_newP2= ' ';
						_recP1= .;
						_recP2= .;
						do i=1 to dim(_o_);
							if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
							else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
							else _AB_testI= _AB_testI + 1;
							_nf_[i]= .;
						end;
						_selected= 0;
						_phased= 0;
					end;
				end;
				else if (indexc(trim(left(_p_[1])), 'M','R','W','S','Y','K') or indexc(trim(left(_p_[2])), 'M','R','W','S','Y','K')) then do;
					**at least one parent is heterozygous*;
					_newP1= ' ';
					_newP2= ' ';
					_recP1= .;
					_recP2= .;
					do i=1 to dim(_o_);
						if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
						else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
						else _AB_testI= _AB_testI + 1;
						_nf_[i]= .;
					end;
					_selected= 0;	
					_phased= 0;
				end;
			%end;
			%else %if (%index(%upcase(&CrossType),SFN) or %index(%upcase(&CrossType),RFN) 
				or %index(%upcase(&CrossType),RI0) or %index(%upcase(&CrossType),RI1)
				or %index(%upcase(&CrossType),RI2) or %index(%upcase(&CrossType),IRILSELF)
				or %index(%upcase(&CrossType),IRILSIB))	%then %do;
				_newP1= upcase(trim(left(_p_[1])));
				_newP2= upcase(trim(left(_p_[2])));
				_recP1= 0;
				_recP2= 2;
				if indexc(trim(left(_p_[1])), 'A','C','G','T') and indexc(trim(left(_p_[2])), 'A','C','G','T') then do;
					**both parents are homozygous;
					if _p_[1] = _p_[2] then do;
						**parents are homozygous with same genotypes;
						if length(_offsobs) > 1 and length(_offsobs) < 4 then do;
							if _offsobs = 'ACM' or _offsobs = 'AC' or _offsobs = 'AM' or _offsobs = 'CM' then do;
								_newP1= 'A';
								_newP2= 'C';
							end;
							else if _offsobs = 'AGR' or _offsobs = 'AG' or _offsobs = 'AR' or _offsobs = 'GR' then do;
								_newP1= 'A';
								_newP2= 'G';
							end;
							else if _offsobs = 'ATW' or _offsobs = 'AT' or _offsobs = 'AW' or _offsobs = 'TW' then do;
								_newP1= 'A';
								_newP2= 'T';
							end;
							else if _offsobs = 'CGS' or _offsobs = 'CG' or _offsobs = 'CS' or _offsobs = 'GS' then do;
								_newP1= 'C';
								_newP2= 'G';
							end;
							else if _offsobs = 'CTY' or _offsobs = 'CT' or _offsobs = 'CY' or _offsobs = 'TY' then do;
								_newP1= 'C';
								_newP2= 'T';
							end;
							else if _offsobs = 'GTK' or _offsobs = 'GT' or _offsobs = 'GK' or _offsobs = 'TK' then do;
								_newP1= 'G';
								_newP2= 'T';
							end;	
							_newP1= upcase(trim(left(_newP1)));
							_newP2= upcase(trim(left(_newP2)));
							do i=1 to dim(_o_);
								if _o_[i] = _newP1 then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = _newP2 then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if (indexc(trim(left(_o_[i])), 'M', 'R', 'W') and indexc(trim(left(_newP1)), 'A')) 
									or (indexc(trim(left(_o_[i])), 'M', 'S', 'Y') and indexc(trim(left(_newP1)), 'C')) 
									or (indexc(trim(left(_o_[i])), 'R', 'S', 'K') and indexc(trim(left(_newP1)), 'G'))
									or (indexc(trim(left(_o_[i])), 'W', 'Y', 'K') and indexc(trim(left(_newP1)), 'T')) then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
							_phased= 0;
						end;
						else do;
							_newP1= ' ';
							_newP2= ' ';
							_recP1=.;
							_recP2=.;
							do i=1 to dim(_o_);
								if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
								else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
								else _AB_testI= _AB_testI + 1;
								_nf_[i]= .;
							end;
							_selected= 0;
							_phased= 0;
						end;
					end;
					else if (indexc(trim(left(_p_[1])), 'A') and indexc(trim(left(_p_[2])), 'C')) 
						or (indexc(trim(left(_p_[1])), 'C') and indexc(trim(left(_p_[2])), 'A')) then do;
						**parents are homozygous with distinct genotypes;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if _o_[i] = _p_[2] then do;
								_nf_[i]= 2;
								_BB_testI= _BB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'M') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					else if (indexc(trim(left(_p_[1])), 'A') and indexc(trim(left(_p_[2])), 'G')) 
						or (indexc(trim(left(_p_[1])), 'G') and indexc(trim(left(_p_[2])), 'A')) then do;
						**parents are homozygous with distinct genotypes;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if _o_[i] = _p_[2] then do;
								_nf_[i]= 2;
								_BB_testI= _BB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'R') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					else if (indexc(trim(left(_p_[1])), 'A') and indexc(trim(left(_p_[2])), 'T')) 
						or (indexc(trim(left(_p_[1])), 'T') and indexc(trim(left(_p_[2])), 'A')) then do;
						**parents are homozygous with distinct genotypes;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if _o_[i] = _p_[2] then do;
								_nf_[i]= 2;
								_BB_testI= _BB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'W') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					else if (indexc(trim(left(_p_[1])), 'C') and indexc(trim(left(_p_[2])), 'G')) 
						or (indexc(trim(left(_p_[1])), 'G') and indexc(trim(left(_p_[2])), 'C')) then do;
						**parents are homozygous with distinct genotypes;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if _o_[i] = _p_[2] then do;
								_nf_[i]= 2;
								_BB_testI= _BB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'S') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					else if (indexc(trim(left(_p_[1])), 'C') and indexc(trim(left(_p_[2])), 'T')) 
						or (indexc(trim(left(_p_[1])), 'T') and indexc(trim(left(_p_[2])), 'C')) then do;
						**parents are homozygous with distinct genotypes;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if _o_[i] = _p_[2] then do;
								_nf_[i]= 2;
								_BB_testI= _BB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'Y') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					else if (indexc(trim(left(_p_[1])), 'G') and indexc(trim(left(_p_[2])), 'T')) 
						or (indexc(trim(left(_p_[1])), 'T') and indexc(trim(left(_p_[2])), 'G')) then do;
						**parents are homozygous with distinct genotypes;
						do i=1 to dim(_o_);
							if _o_[i] = _p_[1] then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if _o_[i] = _p_[2] then do;
								_nf_[i]= 2;
								_BB_testI= _BB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'K') then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
				end;
				else if (indexc(trim(left(_p_[1])), 'A','C','G','T') and indexc(trim(left(_p_[2])), 'N'))
					or (indexc(trim(left(_p_[1])), 'N') and indexc(trim(left(_p_[2])), 'A','C','G','T')) then do;
					**one parent is homozygous and the other is missing;		
					if length(_offsobs) > 1 and length(_offsobs) < 4 then do;
						if _offsobs = 'ACM' or _offsobs = 'AC' or _offsobs = 'AM' or _offsobs = 'CM' then do;
							_newP1= 'A';
							_newP2= 'C';
						end;
						else if _offsobs = 'AGR' or _offsobs = 'AG' or _offsobs = 'AR' or _offsobs = 'GR' then do;
							_newP1= 'A';
							_newP2= 'G';
						end;
						else if _offsobs = 'ATW' or _offsobs = 'AT' or _offsobs = 'AW' or _offsobs = 'TW' then do;
							_newP1= 'A';
							_newP2= 'T';
						end;
						else if _offsobs = 'CGS' or _offsobs = 'CG' or _offsobs = 'CS' or _offsobs = 'GS' then do;
							_newP1= 'C';
							_newP2= 'G';
						end;
						else if _offsobs = 'CTY' or _offsobs = 'CT' or _offsobs = 'CY' or _offsobs = 'TY' then do;
							_newP1= 'C';
							_newP2= 'T';
						end;
						else if _offsobs = 'GTK' or _offsobs = 'GT' or _offsobs = 'GK' or _offsobs = 'TK' then do;
							_newP1= 'G';
							_newP2= 'T';
						end;
						_newP1= upcase(trim(left(_newP1)));
						_newP2= upcase(trim(left(_newP2)));
						if indexc(trim(left(_p_[1])), 'A','C','G','T') then do;
							**parent 1 is homozygous;
							if _p_[1] ^= _newP1 then do;
								_newP1_1= _newP1;
								_newP1= _newP2;
								_newP2= _newP1_1;
							end;
							_newP1= upcase(trim(left(_newP1)));
							_newP2= upcase(trim(left(_newP2)));
							do i=1 to dim(_o_);
								if _o_[i] = _newP1 then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = _newP2 then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if (indexc(trim(left(_o_[i])), 'M', 'R', 'W') and indexc(trim(left(_newP1)), 'A')) 
									or (indexc(trim(left(_o_[i])), 'M', 'S', 'Y') and indexc(trim(left(_newP1)), 'C')) 
									or (indexc(trim(left(_o_[i])), 'R', 'S', 'K') and indexc(trim(left(_newP1)), 'G'))
									or (indexc(trim(left(_o_[i])), 'W', 'Y', 'K') and indexc(trim(left(_newP1)), 'T')) then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							**parent 2 is homozygous;
							if _p_[2] ^= _newP2 then do;
								_newP2_2= _newP2;
								_newP2= _newP1;
								_newP1= _newP2_2;
							end;							
							_newP1= upcase(trim(left(_newP1)));
							_newP2= upcase(trim(left(_newP2)));
							do i=1 to dim(_o_);
								if _o_[i] = _newP2 then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = _newP1 then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if (indexc(trim(left(_o_[i])), 'M', 'R', 'W') and indexc(trim(left(_newP2)), 'A')) 
									or (indexc(trim(left(_o_[i])), 'M', 'S', 'Y') and indexc(trim(left(_newP2)), 'C')) 
									or (indexc(trim(left(_o_[i])), 'R', 'S', 'K') and indexc(trim(left(_newP2)), 'G'))
									or (indexc(trim(left(_o_[i])), 'W', 'Y', 'K') and indexc(trim(left(_newP2)), 'T')) then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
					end;
					else do;
						_newP1= ' ';
						_newP2= ' ';
						_recP1=.;
						_recP2=.;
						do i=1 to dim(_o_);
							if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
							else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
							else _AB_testI= _AB_testI + 1;
							_nf_[i]= .;
						end;
						_selected= 0;
						_phased= 0;
					end;
				end;
				else if (indexc(trim(left(_p_[1])), 'N') and indexc(trim(left(_p_[2])), 'N')) then do;
					**both parents are missing;
					if length(_offsobs) > 1 and length(_offsobs) < 4 then do;
						if _offsobs = 'ACM' or _offsobs = 'AC' or _offsobs = 'AM' or _offsobs = 'CM' then do;
							_newP1= 'A';
							_newP2= 'C';
						end;
						else if _offsobs = 'AGR' or _offsobs = 'AG' or _offsobs = 'AR' or _offsobs = 'GR' then do;
							_newP1= 'A';
							_newP2= 'G';
						end;
						else if _offsobs = 'ATW' or _offsobs = 'AT' or _offsobs = 'AW' or _offsobs = 'TW' then do;
							_newP1= 'A';
							_newP2= 'T';
						end;
						else if _offsobs = 'CGS' or _offsobs = 'CG' or _offsobs = 'CS' or _offsobs = 'GS' then do;
							_newP1= 'C';
							_newP2= 'G';
						end;
						else if _offsobs = 'CTY' or _offsobs = 'CT' or _offsobs = 'CY' or _offsobs = 'TY' then do;
							_newP1= 'C';
							_newP2= 'T';
						end;
						else if _offsobs = 'GTK' or _offsobs = 'GT' or _offsobs = 'GK' or _offsobs = 'TK' then do;
							_newP1= 'G';
							_newP2= 'T';
						end;
						_newP1= upcase(trim(left(_newP1)));
						_newP2= upcase(trim(left(_newP2)));
						do i=1 to dim(_o_);
							if _o_[i] = _newP1 then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if _o_[i] = _newP2 then do;
								_nf_[i]= 2;
								_BB_testI= _BB_testI + 1;
							end;
							else if (indexc(trim(left(_o_[i])), 'M', 'R', 'W') and indexc(trim(left(_newP1)), 'A')) 
								or (indexc(trim(left(_o_[i])), 'M', 'S', 'Y') and indexc(trim(left(_newP1)), 'C')) 
								or (indexc(trim(left(_o_[i])), 'R', 'S', 'K') and indexc(trim(left(_newP1)), 'G'))
								or (indexc(trim(left(_o_[i])), 'W', 'Y', 'K') and indexc(trim(left(_newP1)), 'T')) then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
						_phased= 0;	
					end;
					else do;
						_newP1= ' ';
						_newP2= ' ';
						_recP1= .;
						_recP2= .;
						do i=1 to dim(_o_);
							if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
							else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
							else _AB_testI= _AB_testI + 1;
							_nf_[i]= .;
						end;
						_selected= 0;
						_phased= 0;
					end;
				end;
				else if (indexc(trim(left(_p_[1])), 'M','R','W','S','Y','K') or indexc(trim(left(_p_[2])), 'M','R','W','S','Y','K')) then do;
					**at least one parent is heterozygous*;
						_newP1= ' ';
						_newP2= ' ';
						_recP1= .;
						_recP2= .;
						do i=1 to dim(_o_);
							if indexc(trim(left(_o_[i])), 'N') then _mis_testI= _mis_testI + 1;
							else if indexc(trim(left(_o_[i])), 'A', 'C', 'G', 'T') then _AA_testI= _AA_testI + 1;
							else _AB_testI= _AB_testI + 1;
							_nf_[i]= .;
						end;
						_selected= 0;
						_phased= 0;				
				end;
				%if (%index(%upcase(&CrossType),RI0) or %index(%upcase(&CrossType),RI1)
					or %index(%upcase(&CrossType),RI2) or %index(%upcase(&CrossType),IRILSELF)
					or %index(%upcase(&CrossType),IRILSIB))	
					and %upcase(&SetHetToMiss)= YES %then %do;
					do i=1 to dim(_nf_);
						if _nf_[i] = 1 then _nf_[i]= .;
					end;	
				%end;	
			%end;
			%else %if %index(%upcase(&CrossType),PANEL) %then %do;
				_phased= 0;
				select;
					when(_offsobs= 'A') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'A' then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'C') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'C' then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'G') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'G' then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'T') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'T' then do;
								_nf_[i]= 0;
								_AA_testI= _AA_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'M') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'M' then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'R') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'R' then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'W') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'W' then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'S') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'S' then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'Y') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'Y' then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'K') do;
						do i=1 to dim(_o_);
							if _o_[i] = 'K' then do;
								_nf_[i]= 1;
								_AB_testI= _AB_testI + 1;
							end;
							else if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'N') do;
						do i=1 to dim(_o_);
							if indexc(trim(left(_o_[i])), 'N') then do;
								_nf_[i]= .;
								_mis_testI= _mis_testI + 1;
							end;
							else do;
								_nf_[i]= .;
								_genotype_error_testI= _genotype_error_testI + 1;
							end;
						end;
					end;
					when(_offsobs= 'AM') do;
						if ((2*_a + _m) / (2*(_a + _m))) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'M' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'M' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;							
						end;
					end;
					when(_offsobs= 'AR') do;
						if ((2*_a + _r)/ (2*(_a + _r))) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'R' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'R' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;	
					end;
					when(_offsobs= 'AW') do;
						if ((2*_a + _w)/ (2*(_a + _w))) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'W' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'W' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'CM') do;
						if ((2*_c + _m)/ (2*(_c + _m))) >= 0.50 then do;
							*C is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'M' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*C is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'M' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end; 
					when(_offsobs= 'CS') do;
						if ((2*_c + _s)/ (2*(_c + _s))) >= 0.50 then do;
							*C is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'S' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*C is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'S' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end; 
					when(_offsobs= 'CY') do; 
						if ((2*_c + _y) / (2*(_c + _y))) >= 0.50 then do;
							*C is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'Y' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*C is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'Y' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
					end;
					when(_offsobs= 'GR') do; 
						if ((2*_g + _r) / (2*(_g + _r))) >= 0.50 then do;
							*G is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'R' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*G is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'R' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'GS')do; 
						if ((2*_g + _s) / (2*(_g + _s))) >= 0.50 then do;
							*G is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'S' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*G is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'S' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end; 
					when(_offsobs= 'GK') do; 
						if ((2*_g + _k) / (2*(_g + _k))) >= 0.50 then do;
							*G is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'K' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*G is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'K' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'TW') do;
						if ((2*_t + _w)/ (2*(_t + _w))) >= 0.50 then do;
							*T is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'W' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*T is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'W' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end; 
					when(_offsobs= 'TY') do; 
						if ((2*_t + _y)/ (2*(_t + _y))) >= 0.50 then do;
							*T is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'Y' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*T is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'Y' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
					end;
					when(_offsobs= 'TK') do;
						if ((2*_t + _k)/ (2*(_t + _k))) >= 0.50 then do;
							*T is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'K' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*T is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'K' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'ACM') do;
						if ((2*_a + _m)/ (2*(_a + _c + _m))) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'M' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'M' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'AGR') do;
						if ((2*_a + _r)/ (2*(_a + _g + _r))) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'R' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'R' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'ATW') do;
						if ((2*_a + _w)/ (2*(_a + _t + _w))) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'W' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'W' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'CGS') do;
						if ((2*_c + _s)/ (2*(_c + _g + _s))) >= 0.50 then do;
							*C is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'S' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*C is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'S' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'CTY') do;
						if ((2*_c + _y)/ (2*(_c + _t + _y))) >= 0.50 then do;
							*C is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'Y' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*C is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'Y' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'GTK') do;
						if ((2*_g + _k)/ (2*(_g + _t + _k))) >= 0.50 then do;
							*G is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'K' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*G is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if _o_[i] = 'K' then do;
									_nf_[i]= 1;
									_AB_testI= _AB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					**The following cases are very unlikely to be observed, and when observed, they might be due to genotyping calling error*;
					when(_offsobs= 'AC') do;
						if (_a / (_a + _c)) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'AG') do;
						if (_a / (_a + _g)) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
					end;
					when(_offsobs= 'AT') do;
						if (_a / (_a + _t)) >= 0.50 then do;
							*A is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*A is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'A' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
					end;
					when(_offsobs= 'CG') do;
						if (_c / (_c + _g)) >= 0.50 then do;
							*C is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*C is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
					end;
					when(_offsobs= 'CT') do;
						if (_c / (_c + _t)) >= 0.50 then do;
							*C is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*C is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'C' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					when(_offsobs= 'GT') do;
						if (_g / (_g + _t)) >= 0.50 then do;
							*G is the major allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 0;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 2;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;
						end;
						else do;
							*G is the minor allele, assuming diploid*;
							do i=1 to dim(_o_);
								if _o_[i] = 'G' then do;
									_nf_[i]= 2;
									_AA_testI= _AA_testI + 1;
								end;
								else if _o_[i] = 'T' then do;
									_nf_[i]= 0;
									_BB_testI= _BB_testI + 1;
								end;
								else if indexc(trim(left(_o_[i])), 'N') then do;
									_nf_[i]= .;
									_mis_testI= _mis_testI + 1;
								end;
								else do;
									_nf_[i]= .;
									_genotype_error_testI= _genotype_error_testI + 1;
								end;
							end;						
						end;
					end;
					otherwise do;
						put 'Impossible Cross for Loci=' &IDVar 'P1=' _p_[1] 'P2=' _p_[2] 'Offspring=' _offsobs;
						_selected= 0;					
					end;
				end;
			%end;
			%else %do;
				%put ERROR: Unknown Cross Type.;
				%let exiterror=1;
				%goto exit;
			%end;
			_p_genotype_error_testI= _genotype_error_testI/&_nind_;
			/*S1364776: lists of names of individuals to drop when keepvars is not empty*/
			/*drop &IndVars i;*/
			drop
			%if &DropIndBlock %then %do;
				%do i=1 %to &DropIndBlock;
					%let j= 1;
					%let DropInd= %scan(&&DropTmpIndVars&i, &j);
					%do %while(%length(%scan(&&DropTmpIndVars&i, &j)));	
						&DropInd
						%let j= %sysevalf(&j + 1);
						%let DropInd= %scan(&&DropTmpIndVars&i, &j);
					%end;
				%end;
			%end;
			i;			
		run;
		/*data outlib._recodedData&OutName; set _recodedData; run;*/
		
		%if %length(&MissCodedMarker) %then %do;
			%if %length(&MissCodedP1) and %length(&MissCodedP2) %then %do;
				%let string1= Invalid codes &MissCodedP1 and &MissCodedP2 were found in both parents.;
			%end;
			%else %do;
				%if %length(&MissCodedP1) %then 
					%let MissCode= &MissCodedP1; 
				%else 
					%let MissCode= &MissCodedP2; 
					
				%let string1= An invalid code &MissCode was found in one of the parents.;
			%end;
			%put ERROR: An error ocurred while recoding marker &MissCodedMarker.. &string1. The Valid codes are: A, C, G, T, M, R, W, S, Y, K, N. Please, check the parents coding for this marker.;
			%let exiterror=1;
			%goto exit;
		%end;
				
		%if (%index(%upcase(&CrossType),SFN) or %index(%upcase(&CrossType),RFN) 
			or %index(%upcase(&CrossType),RI0) or %index(%upcase(&CrossType),RI1)
			or %index(%upcase(&CrossType),RI2) or %index(%upcase(&CrossType),IRILSELF)
			or %index(%upcase(&CrossType),IRILSIB))	%then %do;
			%let RecoverUnphased= 1;
		%end;
		%else %do;
			%let RecoverUnphased= 0;
		%end;
						
		%put RecoverUnphased= &RecoverUnphased;
		%if &RecoverUnphased %then %do;
		
			data Sel_recodedData;
				set _recodedData;
				where _selected= 1;
			run;
			/*data outlib.Sel_recodedData0; set Sel_recodedData; run;*/

			%macro RecoverUnPhasedMarkers(_TestData=, _nTestObs=, _RefData=, _maxrf=, _minrf=);
				
				data testmk_all;
					set _null_;
				run;

				%global appendtestmk_all;
				%let appendtestmk_all= 0;
				%do i=1 %to &_nTestObs;

					data testmk(keep= _newP1 _newP2 _recP1 _recP2 _by _tmpIDVar _mo:);
						set &_TestData(keep= _newP1 _newP2 _recP1 _recP2 _by &IDVar /*&IndVars*/ _nf_: _n);
						where _n= &i;
						array _mo_(&_nind_);
						array _mk_{*} _nf_:;
						do i= 1 to dim(_mk_);
							_mo_[i]= _mk_[i];
						end;
						call symputx("mkname_", trim(left(&IDVar)));
						rename &IDVar= _tmpIDVar;
					run;
					%put mkname_&i= &mkname_;

					/*data outlib.testmk&i; set testmk; run;*/

					data testmkm;
						merge &_RefData(keep= _by &IDVar _nf_:) testmk;
						by _by;
						if _stop = 1 then do;
							stop;
						end;
						array _mo_{*} _mo_:;
						array _mk_{*} _nf_:;
						_n22= 0;_n21= 0;_n20= 0;_n12= 0;_n11= 0;_n10= 0;_n02= 0;_n01= 0;_n00= 0;
						%if (%index(%upcase(&CrossType),RI0) or %index(%upcase(&CrossType),RI1) 
							or %index(%upcase(&CrossType),RI2) or %index(%upcase(&CrossType),IRILSELF)
							or %index(%upcase(&CrossType),IRILSIB)) %then %do;
							do i=1 to dim(_mk_);								
								if (_mk_[i]= 2 and _mo_[i]= 2) then _n22= _n22 + 1;
								else if (_mk_[i]= 2 and _mo_[i]= 0) then _n20= _n20 + 1;
								else if (_mk_[i]= 0 and _mo_[i]= 2) then _n02= _n02 + 1;
								else if (_mk_[i]= 0 and _mo_[i]= 0) then _n00= _n00 + 1;
							end;
							
							if (_n22 + _n20 + _n02 + _n00) > 0 then do;
								_rc= (_n20 + _n02) / (_n22 + _n20 + _n02 + _n00);
								_rr= (_n22 + _n00) / (_n22 + _n20 + _n02 + _n00);
							end;
							else do;
								_rc = .99;
								_rr = .99;
							end;
						%end;
						%else %if (%index(%upcase(&CrossType),SFN) or %index(%upcase(&CrossType),RFN)) %then %do;
							%*put SFN RFN;
							do i=1 to dim(_mk_);
								*put i _mk_[i] _mo_[i];
								if (_mk_[i]= 2 and _mo_[i]= 2) then _n22= _n22 + 1;
								else if (_mk_[i]= 2 and _mo_[i]= 1) then _n21= _n21 + 1;
								else if (_mk_[i]= 2 and _mo_[i]= 0) then _n20= _n20 + 1;
								else if (_mk_[i]= 1 and _mo_[i]= 2) then _n12= _n12 + 1;
								else if (_mk_[i]= 1 and _mo_[i]= 1) then _n11= _n11 + 1;
								else if (_mk_[i]= 1 and _mo_[i]= 0) then _n10= _n10 + 1;
								else if (_mk_[i]= 0 and _mo_[i]= 2) then _n02= _n02 + 1;
								else if (_mk_[i]= 0 and _mo_[i]= 1) then _n01= _n01 + 1;
								else if (_mk_[i]= 0 and _mo_[i]= 0) then _n00= _n00 + 1;
							end;	
							
							if (_n22+_n21+_n20+_n12+_n11+_n10+_n02+_n01+_n00) > 0 then do;
								_r0=0.25;
								_alpha= _r0**2/((1-_r0)**2 + _r0**2);
								_iter= 1;
								_tol = 10;
								do while(_tol > 0.00001 and _r0 <= 1 and _iter <= 20);
									_rc= ((_n22 + _n00)*0 + (_n21 + _n12 + _n10 + _n01)*0.5 + (_n20  + _n02)*1 + _n11*_alpha)/(_n22+_n21+_n20+_n12+_n11+_n10+_n02+_n01+_n00); 
									_alpha= _rc**2/((1-_rc)**2 + _rc**2);
									_tol= abs(_rc-_r0);
									put ' iter ' ' r0 ' ' rc ' ' tol ' _iter _r0 _rc _tol;
									_r0= _rc;
									_iter= _iter + 1;
								end;
								_r0=0.25;
								_alpha= _r0**2/((1-_r0)**2 + _r0**2);
								_iter=1;
								_tol= 10;
								do while(_tol > 0.00001 and _r0 <= 1 and _iter <= 20);
									_rr= ((_n22 + _n00)*1 + (_n21 + _n12 + _n10 + _n01)*0.5 + (_n20  + _n02)*0 + _n11*_alpha)/(_n22+_n21+_n20+_n12+_n11+_n10+_n02+_n01+_n00); 
									_alpha= _rr**2/((1-_rr)**2 + _rr**2);
									_tol= abs(_rr-_r0);
									put ' iter ' ' r0 ' ' rr ' ' tol ' _iter _r0 _rr _tol;
									_r0= _rr;
									_iter= _iter + 1;
								end;
							end;
							else do;
								_rc = .99;
								_rr = .99;
							end;
						%end;
						if _rc <= _rr then do;
							**In the initial calls, _recP1=0 and _recP2=2 for all markers (coupling).;
							**Therefore, it is NOT needed to switch marker code;
							_rf= _rc;
							_tmpphase= 'coupl';
							_tmpnewP1=_newP1;
							_tmpnewP2=_newP2;
							_tmprecP1=_recP1;
							_tmprecP2=_recP2;
						end;
						else do;
							**In the initial calls, _recP1=0 and _recP2=2 for all markers (coupling).; 
							**Therefore, it is needed to switch marker code for newP1 <=> newP2, and in the offsprings: 2<=>0.;
							_rf= _rr;
							_tmpphase= 'repul';
							_tmpnewP1= _newP2;
							_tmpnewP2= _newP1;
							_tmprecP1= _recP1; /*no need to switch*/
							_tmprecP2= _recP2; /*no need to switch*/		
						end;
						%if %index(%upcase(&CrossType),RI1) %then %do;
							_rf= _rf/(2*(1-_rf));
						%end;
						%else %if %index(%upcase(&CrossType),RI2) %then %do;
							_rf= _rf/(4-6*_rf);
						%end;	
						if _rf < &_maxrf then do;
							output;
						end;
						if _rf < &_minrf then do;
							_stop = 1;
						end;
						retain _stop;
					run;

					/*data outlib.testmkm_&i; set testmkm; run;*/

					%NObsVars(testmkm);
					%if &nobs > 1 %then %do;
						proc sort data=testmkm; 
							by _rf;
						run;

						data testmkm;
							set testmkm(obs=1);
						run;
					%end;

					/*data outlib.testmkm&i; set testmkm; run;*/

					%if &appendtestmk_all %then %do;
						proc append base=testmk_all data= testmkm(keep= _tmpIDVar _tmpnewP1 _tmpnewP2 _tmprecP1 _tmprecP2  _tmpphase _rf); 
					%end;
					%else %do;
						 data testmk_all;
							set testmkm(keep= _tmpIDVar _tmpnewP1 _tmpnewP2 _tmprecP1 _tmprecP2 _tmpphase _rf);
						 run;

						 %let appendtestmk_all= 1;
					%end;
				%end;
				
				/*data outlib.testmk_all; set testmk_all; run;*/

				%exitRecover:;
				
			%mend RecoverUnPhasedMarkers;	
			
			%let nUnPhased= 0;
			data UnPhased;
				set Sel_recodedData end= _eof_;
				where _selected = 1 and _phased = 0;
				_n= _n_;
				_by= 1;
				if _eof_ then call symputx("nUnphased", _n_);
			run;	
			
			%let nPhased= 0;
			data Phased;
				set Sel_recodedData end= _eof_;
				where _selected = 1 and _phased = 1;
				_n= _n_;
				_by= 1;
				if _eof_ then call symputx("nPhased", _n_);
			run;
			
			**evaluating the quality of Unphased markers*;
			%put nUnPhased nPhased= &nUnPhased &nPhased;
			%if &nUnPhased > 0 and &nPhased > 0 %then %do;
				%RecoverUnPhasedMarkers(_TestData=UnPhased, _nTestObs=&nUnPhased, _RefData=Phased, _maxrf= &maxrf, _minrf= &minrf);
				%if &syserr>4 %then %do;
					%put ERROR: Error in the segregation ratio macro.;
					%let exiterror= 1;
					%goto exit;
				%end;				
			
				%put appendtestmk_all= &appendtestmk_all;
				%if &appendtestmk_all %then %do;
				
					/*data outlib.Sel_recodedData;set Sel_recodedData; run;*/
					/*data outlib.testmk_all;set testmk_all; run;*/
				
					proc sql UNDO_POLICY= NONE;
						create table Sel_recodedData as
						select a1.*, a2.*
						from 
						Sel_recodedData as a1 
						left join 
						testmk_all as a2
						%if &IDVarType = C %then %do;
							on upcase(a1.&IDVar) = upcase(a2._tmpIDVar)
						%end;
						%else %do;
							on a1.&IDVar = a2._tmpIDVar
						%end;
						order by a1.OrigMarkerOrder;
					quit;

					/*data outlib.Sel_recodedData1; set Sel_recodedData; run;*/
					
					data Sel_recodedData;
						set Sel_recodedData;
						if _phased = 0 then do;
							if not missing(_tmpIDVar) then do;
								_newP1= _tmpnewP1;
								_newP2= _tmpnewP2;
								_recP1= _tmprecP1;
								_recP2= _tmprecP2;
								_selected= 1;
								_phased= 1;
								array _mk_{*} _nf_:;
								do i = 1 to dim(_mk_);
									if (_mk_[i] = 2) then _mk_[i]= 0;
									else if (_mk_[i] = 0) then _mk_[i]= 2;
								end;
							end;
							else do;
								*delete;
								_selected= 0;
								_phased= 0;
							end;
						end;
					run;
				%end;
				%else %do;
					data Sel_recodedData;
						set Sel_recodedData;
						if _phased = 0 then do;
							_selected= 0;
						end;
					run;				
				%end;
			%end;
			
			/*data outlib.Sel_recodedData2; set Sel_recodedData; run;*/
			
			data Sel_recodedData;
				set Sel_recodedData;
				where _selected= 1 and _phased= 1;
			run;
			
		%end; /*end if RecoverUnPhased*/
		%else %do;
			data Sel_recodedData;
				set _recodedData;
				where _selected= 1 %if ^%index(%upcase(&CrossType),PANEL) %then and _phased= 1;;
			run;
			/*data outlib.Sel_recodedData0; set Sel_recodedData; run;*/		
		%end;
		/*data outlib.Sel_recodedData0&OutName; set Sel_recodedData; run;*/
		
		%*goto exit;
		
		%NObsVars(Sel_recodedData);
		%let nobsr= &nobs;
			
		data _nf_tmp;
			set Sel_recodedData(keep= _nf_:);
			where 0;
		run;
		%NObsVars(_nf_tmp);
		%let nvarsr= &nvars;
		
		%let BlockSize= %sysfunc(min(%sysevalf(3000000/&nobsr,floor),&nvarsr));
		%*let BlockSize = 20;
		%put BlockSize= &BlockSize;
		
		%let first1 = 1;
		%let last1 = &BlockSize;
		%let Block=1;
		%do %until (&&first&Block > &nvars );
			%let first%eval(&Block + 1) = %eval(&&last&Block + 1);
			%let last%eval(&Block + 1) = %eval(&&last&Block + &BlockSize);
			%let Block = %eval(&Block + 1);
			%put first&Block &&first&Block lastb &&last&Block;
			%if &&last&Block > &nvars %then %let last&Block = &nvars;
			%put first&Block &&first&Block lastb &&last&Block;
		%end;
		%let NBlocks= %sysevalf(&Block - 1);
		%put NBlocks= &NBlocks;
		
		%do i=1 %to &NBlocks;
			proc transpose data= Sel_recodedData(keep= _nf_&&first&i - _nf_&&last&i /*%if &i=1 %then*/ &IDVar/*;*/) out=Sel_recodedDatat(drop=_name_);
				var _nf_&&first&i - _nf_&&last&i;
				id &IDVar;
			run;
			
			/*data OutLib.Sel_recodedDatat&i; set Sel_recodedDatat; run;*/
			
			proc datasets nolist;
				%if &i=1 %then %do;
					delete SelAnd_recoded_Geno;
					change Sel_recodedDatat = SelAnd_recoded_Geno;
				%end;
				%else %do;
					append base=SelAnd_recoded_Geno data=Sel_recodedDatat force;
					delete Sel_recodedDatat;
				%end;
			quit;
		%end;		
		
		/*
		proc transpose data= Sel_recodedData(keep= _nf_: &IDVar) out=SelAnd_recoded_Geno(drop=_name_);
			var _nf_:;
			id &IDVar;
		run;
		*/

		%if %index(%upcase(&CrossType),PANEL) %then %do;
			%let oldnames= _a _c _g _t _m _r _w _S _y _k _n _offsobs _tot_ind _AA_testI _AB_testI _BB_testI _mis_testI 
					_selected;
			%let newnames= n_A n_C n_G n_T n_M n_R n_W n_S n_Y n_K n_N offs_obs n_ind n_A1A1 n_A1A2 n_A2A2 n_mis 
					selected;	
			%let labels= A Obs @$@C Obs @$@G Obs @$@T Obs @$@M Obs @$@R Obs @$@W Obs @$@S Obs @$@Y Obs @$@K Obs @$@N Obs @$@Offspring Obs @$@Total Individuals @$@AA Obs @$@AB Obs @$@BB Obs @$@Miss Obs 
					@$@Selected Markers;	
		%end;
		%else %do;
			%let oldnames= _a _c _g _t _m _r _w _S _y _k _n _offsobs _tot_ind _AA_testI _AB_testI _BB_testI _mis_testI 
					_genotype_error_testI _p_genotype_error_testI _phased _selected _newP1 _newP2 _recP1 _recP2;
			%let newnames= n_A n_C n_G n_T n_M n_R n_W n_S n_Y n_K n_N offs_obs n_ind n_A1A1 n_A1A2 n_A2A2 n_mis 
					n_err p_err phased selected &newP1Var &newP2Var &recP1Var &recP2Var;	
			%let labels= A Obs @$@C Obs @$@G Obs @$@T Obs @$@M Obs @$@R Obs @$@W Obs @$@S Obs @$@Y Obs @$@K Obs @$@N Obs @$@Offspring Obs @$@Total Individuals @$@AA Obs @$@AB Obs @$@BB Obs @$@Miss Obs 
					@$@Genotyping Error @$@Ratio of Genotyping Error @$@Phased Markers @$@Selected Markers @$@&newP1Lab @$@&newP2Lab @$@&recP1Lab @$@&recP2Lab;	
		%end;
		%put newnames= &newnames;
			
		%let k= 1;
		%let _oldname= %scan(&oldnames, &k);
		%let _newname= %scan(&newnames, &k);
		%let _label= %scan(&labels, &k, @$@);
		%let OldNamesToNewNames=;
		%let LabelsStmt= ;
		%do %while(%length(&_newname));
			%let OldNamesToNewNames= &OldNamesToNewNames &_oldname= &_newname;
			%let LabelsStmt= &LabelsStmt &_oldname= "&_label";
			%let k= %sysevalf(&k + 1);
			%let _oldname= %scan(&oldnames, &k);
			%let _newname= %scan(&newnames, &k);
			%let _label= %scan(&labels, &k, @$@);
		%end;
		%put OldNamesToNewNames= &OldNamesToNewNames;
		%put LabelsStmt= &LabelsStmt;
		
		data SelAnd_recoded_Anno(keep= &IDVar 
						%if %length(&Parent1Var) %then &Parent1Var;
						%if %length(&Parent2Var) %then &Parent2Var;
						&KeepVars &newnames OrigMarkerOrder);
			retain &IDVar 
				%if %length(&Parent1Var) %then &Parent1Var;
				%if %length(&Parent2Var) %then &Parent2Var;
				&KeepVars &oldnames OrigMarkerOrder;
			set Sel_recodedData;
			rename &OldNamesToNewNames;
			label &LabelsStmt;
		run;
				
		%*goto exit;
		
		%if %index(%upcase(&CrossType),PANEL) %then %do;
			%let keepFrom_recoded= _a _c _g _t _m _r _w _S _y _k _n _offsobs 
				_tot_ind _AA_testI _AB_testI _BB_testI _mis_testI; 	
			%let keepFromSel_recoded= _selected;
			%let oldnames2= &keepFrom_recoded &keepFromSel_recoded;
				
			%let keepFrom_recoded_New= n_A n_C n_G n_T n_M n_R n_W n_S n_Y n_K n_N offs_obs 
				n_ind n_A1A1 n_A1A2 n_A2A2 n_mis;
			%let keepFromSel_recoded_New= selected;
			%let newnames2= &keepFrom_recoded_New &keepFromSel_recoded_New;
			
			%let keepFrom_recoded_New_Labels= A Obs @$@C Obs @$@G Obs @$@T Obs @$@M Obs @$@R Obs @$@W Obs @$@S Obs @$@Y Obs @$@K Obs @$@N Obs @$@Offspring Obs 
				@$@Total Individuals @$@AA Obs @$@AB Obs @$@BB Obs @$@Miss Obs;
			%let keepFromSel_recoded_New_Labels= @$@Selected Markers;
			%let labels2= &keepFrom_recoded_New_Labels &keepFromSel_recoded_New_Labels;	
		%end;
		%else %do;			
			%let keepFrom_recoded= _a _c _g _t _m _r _w _S _y _k _n _offsobs 
				_tot_ind _AA_testI _AB_testI _BB_testI _mis_testI _genotype_error_testI _p_genotype_error_testI; 	
			%let keepFromSel_recoded= _phased _selected  _newP1 _newP2 _recP1 _recP2;
			%let oldnames2= &keepFrom_recoded &keepFromSel_recoded;
				
			%let keepFrom_recoded_New= n_A n_C n_G n_T n_M n_R n_W n_S n_Y n_K n_N offs_obs 
				n_ind n_A1A1 n_A1A2 n_A2A2 n_mis n_err p_err;
			%let keepFromSel_recoded_New= phased selected &newP1Var &newP2Var &recP1Var &recP2Var;
			%let newnames2= &keepFrom_recoded_New &keepFromSel_recoded_New;
			
			%let keepFrom_recoded_New_Labels= A Obs @$@C Obs @$@G Obs @$@T Obs @$@M Obs @$@R Obs @$@W Obs @$@S Obs @$@Y Obs @$@K Obs @$@N Obs @$@Offspring Obs 
				@$@Total Individuals @$@AA Obs @$@AB Obs @$@BB Obs @$@Miss Obs @$@Genotyping Error Obs @$@Ratio of Genotyping Error;
			%let keepFromSel_recoded_New_Labels= @$@Phased Markers @$@Selected Markers @$@&newP1Lab @$@&newP2Lab @$@&recP1Lab @$@&recP2Lab;
			%let labels2= &keepFrom_recoded_New_Labels &keepFromSel_recoded_New_Labels;
		%end;
				
		%let k= 1;
		%let _oldname2= %scan(&oldnames2, &k);
		%let _newname2= %scan(&newnames2, &k);
		%let _labels2= %scan(&labels2, &k, @$@);
		%let OldNames2ToNewNames2=;
		%let LabelsStmt2= ;
		%do %while(%length(&_newname2));
			%let OldNames2ToNewNames2= &OldNames2ToNewNames2 &_oldname2= &_newname2;
			%let LabelsStmt2= &LabelsStmt2 &_newname2= "&_labels2";
			%let k= %sysevalf(&k + 1);
			%let _oldname2= %scan(&oldnames2, &k);
			%let _newname2= %scan(&newnames2, &k);	
			%let _labels2= %scan(&labels2, &k, @$@);
		%end;
		%put OldNames2ToNewNames2= &OldNames2ToNewNames2;
		
		/*data outlib._recodedData; set _recodedData; run;*/
			
		proc sql UNDO_POLICY= NONE;
			create table All_recodedData(rename=(
							%do i=1 %to &nRecIndBlock;
								&&recodedIndVar&i
							%end;
							&OldNames2ToNewNames2)) as
			select a1.*, a2.*
			from _recodedData(keep= &IDVar 
						%if %length(&Parent1Var) %then &Parent1Var;
						%if %length(&Parent2Var) %then &Parent2Var;
						/*&IndVars*/ &KeepVars &keepFrom_recoded OrigMarkerOrder) as a1 
			left join 
			Sel_recodedData(keep= &keepFromSel_recoded _nf_: OrigMarkerOrder rename=(OrigMarkerOrder= _OrigMarkerOrder_)) as a2
			on a1.OrigMarkerOrder= a2._OrigMarkerOrder_
			order by OrigMarkerOrder;
		quit;
		
		%put IndVars= &IndVars;
		data All_recodedData;
			retain &IDVar
				%if %length(&Parent1Var) %then &Parent1Var;
				%if %length(&Parent2Var) %then &Parent2Var;
				%do i=1 %to &nIndBlock;
					%let j= 1;
					%let InVar= %scan(&&TmpIndVars&i, &j);
					%do %while(%length(&InVar));	
						&InVar
						%let j= %sysevalf(&j + 1);
						%let InVar= %scan(&&TmpIndVars&i, &j);
					%end;
				%end;
				&KeepVars &newnames2;
			set All_recodedData;
			if missing(selected) then do; /*needed because of the proc sql left join done previously*/
				%if ^%index(%upcase(&CrossType),PANEL) %then %do; 
					phased= 0;
				%end;
				selected= 0; 
			end;
			drop _OrigMarkerOrder_;
			label &LabelsStmt2;
		run;
		
		%let PlotVar= selected;
		%let PlotLabel= Selected Markers;
	%end;
	
	%*goto exit;
	
	%NObsVars(SelAnd_recoded_Geno);
	%let nSelMrks= &nvars;
	%if &nSelMrks > 0 %then %do;
		**begin output selected markers into genotype and annotation data set*;
		
		**begin S1144206: creating a new column of marker name for the output annotation data set.*;
		proc transpose data= Sel_recodedData(keep= &IDVar) out=SelAnd_recoded_MarkerName(drop=_name_);
			var &IDVar;
			id &IDVar;
		run;
		
		data SelAnd_recoded_MarkerName;
			set SelAnd_recoded_MarkerName;
			where 0;
		run;
		
		%let _dsid_ = %sysfunc(open(SelAnd_recoded_MarkerName));
  		%let _name_exist= %sysfunc(varnum(&_dsid_,_label_));
  		%let _rc_ = %sysfunc(close(&_dsid_));		
		
		proc transpose data= SelAnd_recoded_MarkerName %if &_name_exist > 0 %then %do; (drop=_label_) %end; out= SelAnd_recoded_MarkerName;
			var _all_;
		run;	

		data SelAnd_recoded_Anno;
			merge SelAnd_recoded_MarkerName(rename=(_name_= _&IDVar)) SelAnd_recoded_Anno;
			label _&IDVar = "_&IDVar";
		run;
		**end*;
		
		%if %length(&AnnoInData) %then %do;
			proc sql UNDO_POLICY= NONE;
				create table OutLib.&TasselGBSImportAnn_OutDataName(drop=_AnnoMName) as
				select a1.*, a2.* 
				from 
				SelAnd_recoded_Anno as a1
				left join
				AnnoLib.&AnnoData(rename=(&AnnoMName=_AnnoMName)) as a2
				%if &IDVarType = C %then %do;
					on upcase(a1.&IDVar) = upcase(a2._AnnoMName)
				%end;
				%else %do;
					on a1.&IDVar = a2._AnnoMName
				%end;
				order by OrigMarkerOrder
				;
			quit;
		%end;
		%else %do;
			data OutLib.&TasselGBSImportAnn_OutDataName;
				set SelAnd_recoded_Anno;
			run;
		%end;
		%PackageInsertDataset(Annotation Data Set of Selected Markers from &ProcessName, OutLib, &TasselGBSImportAnn_OutDataName,no);
		
		data OutLib.&TasselGBSImportGen_OutDataName;
			merge indvnamest(rename=(_name_=IndID %if &Flag_IndLen %then _newname_= IndID_New;)) SelAnd_recoded_Geno;
			label IndID= "IndID";
		run;
		%PackageInsertDataset(Genotypes Data Set of Selected Markers from &ProcessName, OutLib, &TasselGBSImportGen_OutDataName,no);
	
		**collecting names of selected markers*;
		data SelMkrNames;
			set SelAnd_recoded_Geno;
			where 0;
		run;

		%let _did_= %sysfunc(open(SelMkrNames));
		%let _nmrk_= %sysfunc(attrn(&_did_, nvars));
		%let _fmrk_= %sysfunc(varname(&_did_,1));
		%let _lmrk_= %sysfunc(varname(&_did_,&_nmrk_));
		%put _fmrk_= &_fmrk_;
		%put _lmrk_= &_lmrk_;

		**code values for cell plot*;
		%if %upcase(&DoCellPlots) = YES %then %do;

			%let _maxm_= 50;
			%let _div_= %sysevalf(&_nmrk_/&_maxm_);
			%let _int_= %sysfunc(int(&_div_));
			%let _dec_= %sysfunc(mod(&_div_, 1));
			%if &_dec_ ge 0.5 %then 
				%let _den_= %sysevalf(&_int_ + 1);
			%else 
				%let _den_= %sysfunc(max(&_int_, 1));
			%let _maxm_= %sysfunc(ceil(&_nmrk_/&_den_));

			%let nMrkBlock= 0;
			%let TmpVars= ;
			%let l= 1;
			%do k=1 %to &_nmrk_;
				%*if %length(&TmpVars) le 32700 %then %do;
				%if %length(&TmpVars) le 500 and &l < &_maxm_ %then %do;
					%let TmpVars = &TmpVars %sysfunc(varname(&_did_,&k));
					%let l= %sysevalf(&l + 1);
					%*put k= &k;
				%end;
				%else %do;
					%*put l= &l;
					%let nMrkBlock= %sysevalf(&nMrkBlock + 1);
					%let SelMrks&nMrkBlock = &TmpVars %sysfunc(varname(&_did_,&k));
					%let TmpVars = ;
					%let l= 0;
				%end;
			%end;		
			%if (&nMrkBlock = 0) or (&nMrkBlock > 0 and %length(&TmpVars) > 0) %then %do;
				%let nMrkBlock= %sysevalf(&nMrkBlock + 1);
				%let SelMrks&nMrkBlock = &TmpVars;
				%let TmpVars = ;
			%end;
			%put nMrkBlock= &nMrkBlock;
			%put SelMrks&nMrkBlock= &&SelMrks&nMrkBlock;	

			%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
				%let code0 = ; 
				%let code11 = ;
				%let code12 = ;
				%let code2 = ;
			%end;
			%else %do;
				%let code0 = 0; 
				%let code11 = 1;
				%let code12 = 1;
				%let code2 = 2;
			%end;
		%end;
		%let _rc_= %sysfunc(close(&_did_));
	%end;
	*****;
	
	*begin output original data set with some extra columns*;
	data OutLib.&TasselGBSImport_OutDataName;
		set All_recodedData;
	run;
	*****;
		
	%*goto exit;
	
	**create JSL file*;
	%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
		%let stringTI_1= In the Chi-square Test I, the number of classes in the chi-square statistic is equal to the number of offspring classes as expected according to the genotypes of the parents.; 
		%let stringTI_2= For instance, if the genotype of parents are A and M, the expected offspring classes are A and M, and the chi-square statistic has two classes.;
		%let stringTI_3= Now, if the genotype of parents are M and M, the expected offspring classes are A, C, and M, and the chi-square statistic has three classes.;
		%let stringTI_4= The Chi-square Test I will NOT be carried out when both parents have missing genotypes.; 
		%let stringTI_5= In the output data sets, variables that have '_testI' in their names come from the Chi-square Test I.;
	
		%let stringTII_1= In the Chi-square Test II, the number of classes in the chi-square statistic is equal to the number of classes as observed in the offsprings.; 
		%let stringTII_2= For instance, if the genotype of parents are A and M, the expected offspring classes are A and M, but the observed classes are A, C, and M, then the chi-square statistic has three classes.; 
		%let stringTII_3= Now, if the genotype of parents are M and M, the expected offspring classes are A, C, and M, but the observed classes are A and M, then the chi-square statistic has two classes.; 
		%let stringTII_4= The Chi-square Test II will be carried out even when both parents have missing genotypes. In the output data sets, variables that have '_testII' in their names come from the Chi-square Test II.;
		
		%*put stringsTI= &stringTI_1 &stringTI_2 &stringTI_3 &stringTI_4 &stringTI_5;
	%end; 
	
	%PrepOutFile(TasselGBSImportScript,TasselGBSImportScriptName,&ProcessName Script for &OutName..jsl);
	%global _mainjslfile_;
	%let _mainjslfile_ = &TasselGBSImportScript;
		
	data _null_;
	   file "&TasselGBSImportScript" lrecl=32767;
		put "//!";
		put "Names Default to Here(1);";
		put "Here:DataList = {};";
		put "Here:TabButtonList = {};";
		put "Here:TabCodeList = {};";
		put "Here:TabAfterCodeList = {};";
		put "Here:TabOpenList = {};";
		put "jg:clientoutpath = ""&ClientOutPath"";";
		put "jg:serveroutpath = ""&ServerOutPath"";";
		put "OutDataName = ""&TasselGBSImport_OutDataName"";";	
		
		put "Here:mrktypechart=Function({},{default local},";
		put "	if(isEmpty(Here:alldat),";
		put " 		Here:alldat = open(""&ClientOutPath.&TasselGBSImport_OutDataName..sas7bdat"",invisible);";
		put "		Column(""Selected Markers"")<<Set Modeling Type(Nominal);";
		%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
			put "		Column(""Sum of Indicators for Test I"")<<Set Modeling Type(Nominal);";
			put "		Column(""Sum of Indicators for Test II"")<<Set Modeling Type(Nominal);";
			put "		Column(""Unique Dosage"")<<Set Modeling Type(Nominal);";
			put "		Column(""Single Dosage"")<<Set Modeling Type(Nominal);";
			put "		Column(""Test I or II"")<<Set Modeling Type(Nominal);";
			%do j=1 %to %sysevalf(&ncross);
				put "		Column(""Indicator of Test I for Dosages &&Dosage&j"")<<Set Modeling Type(Nominal);";
				put "		Column(""Indicator of Test II for Dosages &&Dosage&j"")<<Set Modeling Type(Nominal);";
			%end;
		%end;
		put "		Here:alldat << Color Or Mark by Column(:Name(""&PlotLabel""), Marker Theme(standard));";
		put "		,//else";
		put "		CurrentDataTable(Here:alldat);";
		put "	);";
		put "	VListBox(";
		put "		Here:gb= Graph Builder( ";
		put "			Size( 570, 537),";
		put "			Show Control Panel( 0 ),";
		put "			Variables(";
		put "				X( :&PlotLabel),";
		put "				Color( :&PlotLabel),";
		put "			),";
		put "			Elements(";
		put "				Bar(";
		put "					X( 1 ),";
		put "					Legend( 20 ),";
		put "					Bar Style( ""Side by side"" ),";
		put "					Summary Statistic( ""N"" ),";
		put "				)";
		put "			),";
		put "			SendToReport(";
		put "				Dispatch( {}, ""&PlotLabel"", ScaleBox, {Rotated Labels( ""Horizontal"" )} ),";
		put "				Dispatch( {}, ""400"", LegendBox, {Set Title( ""&PlotLabel"" )} ),";
		%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
			put "				Dispatch( {}, ""Graph Builder"",OutlineBox,{Set Title( ""Marker Type Counts"" )}),";
		%end;
		%else %do;
			put "				Dispatch( {}, ""Graph Builder"",OutlineBox,{Set Title( ""Selected Marker Counts"" )}),";
		%end;
		put "				Dispatch( {}, ""graph title"", TextEditBox, {Set Text( """" )} ),";
		put "				Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""&PlotLabel"" )} )";
		put "				),";
		put "			Local Data Filter(""on right"",";
		put "				Location( {0, 15} ),";
		put "				Mode( Show( 1 ), Include( 1 ) ),";
		put "				Add Filter(";
		put "					columns( :&PlotLabel, :testI_OR_testII), ";
		put "				),";
		put "				Display( :&PlotLabel, Size( 150, 100), List Display ),";
		put "				Mode(  Show(1), Include(1)),";
		put "				Animation( Animate Column( :&PlotLabel) )";
		put "			),";
		put "		),";
		put "	);";
		put ");";
		put "InsertInto(DataList,{{""Here:alldat""}});";
		%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
			put "InsertInto(TabButtonList, ""Marker Type Counts Plot"");";
		%end;
		%else %do;
			put "InsertInto(TabButtonList, ""Selected Marker Counts Plot"");";
		%end;
		put "InsertInto(TabCodelist,expr(Here:mrktypechart()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
		
		%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
			put "Here:PValParentalPlots=Function({},{default local},";
			put "	if(isEmpty(Here:alldat),";
			put " 		Here:alldat = open(""&ClientOutPath.&TasselGBSImport_OutDataName..sas7bdat"",invisible);";
			put "		Column(""Selected Markers"")<<Set Modeling Type(Nominal);";
			put "		Column(""Sum of Indicators for Test I"")<<Set Modeling Type(Nominal);";
			put "		Column(""Sum of Indicators for Test II"")<<Set Modeling Type(Nominal);";
			put "		Column(""Unique Dosage"")<<Set Modeling Type(Nominal);";
			put "		Column(""Single Dosage"")<<Set Modeling Type(Nominal);";
			put "		Column(""Test I or Test II"")<<Set Modeling Type(Nominal);";
			%do j=1 %to %sysevalf(&ncross);
				put "		Column(""Indicator of Test I for Dosages &&Dosage&j"")<<Set Modeling Type(Nominal);";
				put "		Column(""Indicator of Test II for Dosages &&Dosage&j"")<<Set Modeling Type(Nominal);";
			%end;
			put "		Here:alldat << Color Or Mark by Column(:Name(""&PlotLabel""), Marker Theme(standard));";
			put "		,//else";
			put "		CurrentDataTable(Here:alldat);";
			put "	);";
			put "	VListBox(";
			put "		Here:olb= OutLineBox(""Chi-square Test I"",";
			put "			TextBox("" &stringTI_1 "@;
			put "&stringTI_2 "@;
			put "&stringTI_3 "@;
			put "&stringTI_4 "@;
			put "&stringTI_5"", ";
			put "				,<< Justify Text()";
			put "				,SetWrap(800))";
			put "				,<<Close(1)";
			put "				,";
			put "		),";

			%do j=1 %to %sysevalf(&ncross);
				put "		Here:gb&j= Graph Builder(";
				put "			Size( 968, 572 ),";
				put "			Show Control Panel( 0 ),";
				put "				Show Legend( 0 ),";
				put "			Variables( X( :&IDVar ), Y( :p Value of Test I for Dosages &&Dosage&j ), Group X( :&PlotLabel, Show Title(0))),";
				put "			Elements( Points( X, Y, Legend( 14 ), Jitter( 1 ) ) ),";
				put "			SendToReport(";
				put "				Dispatch( {}, ""400"", LegendBox, {Set Title( """" ), Position( {0} )} ),";
				put "				Dispatch( {}, ""Graph Builder"", OutlineBox, {Set Title( ""P-Values for Dosages &&Dosage&j"" )}),";
				put "				Dispatch( {}, ""graph title"", TextEditBox, {Set Text( ""&PlotLabel"" )} ),";
				put "				Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""raw p-value of test I for dosages &&Dosage&j"" )} )";
				put "			),";
				put "			Local Data Filter(""on right"",";
				put "				Location( {0, 15} ),";
				put "				Mode( Show( 1 ), Include( 1 ) ),";
				put "				Add Filter(";
				put "					columns(:Indicator of Test I for Dosages &&Dosage&j, :Unique Dosage, :Single Dosage, :Selected Markers, :Test I or Test II), ";			
				put "				),";
				/*put "				Display( :uniquedosage, Size( 150, 100), List Display ),";*/
				put "				Mode(  Show(1), Include(1)),";
				put "				Animation( Animate Column( :singledosage) )";
				put "			),";
				put "		);";
			%end;
			put "	);";
			put ");";
			
			put "InsertInto(DataList,{{""Here:alldat""}});";
			put "InsertInto(TabButtonList, ""P-Value Plots of Test I"");";
			put "InsertInto(TabCodeList, expr(Here:PValParentalPlots()), Loc(TabButtonList,""Marker Type Counts Plot"")+1+1);";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,0);";
			
			put "Here:PValOffsPlots=Function({},{default local},";
			put "	if(isEmpty(Here:alldat),";
			put " 		Here:alldat = open(""&ClientOutPath.&TasselGBSImport_OutDataName..sas7bdat"",invisible);";
			put "		Column(""Selected Markers"")<<Set Modeling Type(Nominal);";
			put "		Column(""Sum of Indicators for Test I"")<<Set Modeling Type(Nominal);";
			put "		Column(""Sum of Indicators for Test II"")<<Set Modeling Type(Nominal);";
			put "		Column(""Unique Dosage"")<<Set Modeling Type(Nominal);";
			put "		Column(""single Dosage"")<<Set Modeling Type(Nominal);";
			put "		Column(""Test I or Test II"")<<Set Modeling Type(Nominal);";
			%do j=1 %to %sysevalf(&ncross);
				put "		Column(""indicator_testI_&&Dosage&j"")<<Set Modeling Type(Nominal);";
				put "		Column(""indicator_testII_&&Dosage&j"")<<Set Modeling Type(Nominal);";
			%end;
			put "		Here:alldat << Color Or Mark by Column(:Name(""&PlotLabel""), Marker Theme(standard));";
			put "		,//else";
			put "		CurrentDataTable(Here:alldat);";
			put "	);";
			put "	VListBox(";
			put "		Here:olb= OutLineBox(""Chi-square Test II"",";
			put "			TextBox("" &stringTII_1 "@;
			put "&stringTII_2 "@;
			put "&stringTII_3 "@;
			put "&stringTII_4"", ";
			put "				,<< Justify Text()";
			put "				,SetWrap(800))";
			put "				,<<Close(1)";
			put "				,";
			put "		),";
			%do j=1 %to %sysevalf(&ncross);
				put "		Here:gb&j= Graph Builder(";
				put "			Size( 968, 572 ),";
				put "			Show Control Panel( 0 ),";
				put "				Show Legend( 0 ),";
				put "			Variables( X( :&IDVar ), Y( :p Value of Test II for Dosages &&Dosage&j ), Group X( :&PlotLabel, Show Title(0))),";
				put "			Elements( Points( X, Y, Legend( 14 ), Jitter( 1 ) ) ),";
				put "			SendToReport(";
				put "				Dispatch( {}, ""400"", LegendBox, {Set Title( """" ), Position( {0} )} ),";
				put "				Dispatch( {}, ""Graph Builder"", OutlineBox, {Set Title( ""P-Values for Dosages &&Dosage&j"" )}),";
				put "				Dispatch( {}, ""graph title"", TextEditBox, {Set Text( ""&PlotLabel"" )} ),";
				put "				Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""raw p-value of test II for dosages &&Dosage&j"" )} )";
				put "			),";
				put "			Local Data Filter(""on right"",";
				put "				Location( {0, 15} ),";
				put "				Mode( Show( 1 ), Include( 1 ) ),";
				put "				Add Filter(";
				put "					columns(:Indicator of Test II for Dosages &&Dosage&j, :Unique Dosage, :Single Dosage, :Selected Markers, :Type I or II), ";
				put "				),";
				/*put "				Display( :uniquedosage, Size( 150, 100), List Display ),";*/
				put "				Mode(  Show(1), Include(1)),";
				put "				Animation( Animate Column( :singledosage) )";
				put "			),";
				put "		);";
			%end;
			put "	);";
			put ");";
			
			put "InsertInto(DataList,{{""Here:alldat""}});";
			put "InsertInto(TabButtonList, ""P-Value Plots of Test II"");";
			put "InsertInto(TabCodeList, expr(Here:PValOffsPlots()), Loc(TabButtonList,""Marker Type Counts Plot"")+1+2);";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,0);";
		%end;
		
		%if %upcase(&DoCellPlots) = YES and &nSelMrks %then %do;
			put "Here:CPfunc = Function({},{default local},";
			put "	If(IsEmpty(Here:gendat),";
			put " 	Here:gendat = open(""&ClientOutPath.&TasselGBSImportGen_OutDataName..sas7bdat"",invisible);";
			**Force to use SAS names so markers match*;
			put "		cols = Here:gendat  << GetColumnNames();";
			put "		for(i=1,i<=NItems(cols),i++,";
			put "			ci = Column(cols[i]);";
			put "			sasnm = char(ci << GetProperty(""SAS Name""));";
			put "			if(length(sasnm),ci << SetName(sasnm))";
			put "		);";
			%let k= 1;
			%do i=1 %to &nMrkBlock;
				%let j= 1;
				%let mrkVar= %scan(&&SelMrks&i, &j);
				%do %while(%length(&mrkVar));	
					put "  Column(""&mrkVar"")<<Set Modeling Type(""Ordinal"");";		 
					put "  Column(""&mrkVar"")<<Set Property(""Value Colors"",{""&code2""=21,""&code11""=1, ""&code12"" = 1, ""&code0""=19});"; 			
					%let j= %sysevalf(&j + 1);
					%let mrkVar= %scan(&&SelMrks&i, &j);
					%let k= %sysevalf(&k + 1);		
				%end;
			%end;
			/*
			%if %length(&LabelVar) %then %do;
				put "		Column(""&LabelVar"")<<Label(1);";
			%end;
			*/
			put " ,"; 
			put "		Current Data Table(Here:gendat);";
			put "	);";
			put "HListBox(";
			put "	VListBox(";
			put "		LineUpBox(NCol(2),";
			put "		Text Box( "" Vertical Frame Size "" ),";
			put "		vframeSizeSlider = SliderBox(num(200),num(1500),vfsize,";
			%do i=1 %to &nMrkBlock;
				put "	try((Here:cp&i<<report)[CellPlotBox(1)] << FrameSize(hfsize,vfsize);); ";
			%end;
			put "					  ),";
			put "           Text Box( "" Horizontal Frame Size "" ),";
			put "			hframeSizeSlider = SliderBox(num(100),num(800),hfsize,";
			%do i=1 %to &nMrkBlock;
					put "								try((Here:cp&i<<report)[CellPlotBox(1)] << FrameSize(hfsize,vfsize); );";
			%end;
			put "					  ),";
			put "            ),";	
			%if ^%index(%upcase(&CrossType),OUTCROSS) and ^%index(%upcase(&CrossType),PANEL) %then %do;
				put "panelbox(""Legend"",";
				put "   t1=Textbox(""Homozygous for Parental Line 1""),";
				put "   t1<<Font Color(3),";
				put "   t1<<Bullet Point(1);";
				put "   t2=Textbox(""Heterozygous""),";
				put "   t2<<Font Color(1),";
				put "   t2<<Bullet Point(1);";
				put "   t3=Textbox(""Homozygous for Parental Line 2""),";
				put "   t3<<Font Color(5),";
				put "   t3<<Bullet Point(1);";
				put "   t4=Textbox(""x  Missing value""),";
				put ")";
			%end;
			%else %if %index(%upcase(&CrossType),PANEL) %then %do;
				put "panelbox(""Legend"",";
				put "   t1=Textbox(""Homozygous AA""),";
				put "   t1<<Font Color(3),";
				put "   t1<<Bullet Point(1);";
				put "   t2=Textbox(""Heterozygous AB""),";
				put "   t2<<Font Color(1),";
				put "   t2<<Bullet Point(1);";
				put "   t3=Textbox(""Homozygous BB""),";
				put "   t3<<Font Color(5),";
				put "   t3<<Bullet Point(1);";
				put "   t4=Textbox(""x  Missing value""),";
				put ")";
			%end;
			put "),";
			%let k= 0;
			%do i=1 %to &nMrkBlock;
				put "		Here:cp&i =  Cell Plot( ScaleUniformly(1), Center at Zero(0),";
				put "			Y(";
				%let j= 1;
				%let mrkVar= %scan(&&SelMrks&i, &j);
				%do %while(%length(&mrkVar));					
					put "			:Name(""&mrkVar""),";
					%let j= %sysevalf(&j + 1);
					%let mrkVar= %scan(&&SelMrks&i, &j);
					%let k= %sysevalf(&k + 1);
				%end;
				put "			),";
				/*
				%if %length(&LabelVar) %then %do;
					put "			Label(:Name(""&LabelVar"")),";
				%end;
				*/
				put "			SendToReport(";
				put "				Dispatch({},""Cell Plot"",OutlineBox,";
				put "					{Set Title(""Genotype Color Plot - Block &i"")}";
				put "				),";
				%if &k > 50 %then %do;
					put "				Dispatch( {}, """", CellPlotBox, {Frame Size(400, 650), No separator lines(1)})";
				%end;
				%else %do;
					put "				Dispatch( {}, """", CellPlotBox, {Frame Size(400, 650)})";
				%end;
				put "			)";
				put "		);";
				%let k= 0;
			%end;
			put " ));";
			
			put "InsertInto(DataList,{{""Here:gendat""}});";
			put "InsertInto(TabButtonList, ""Genotype Color Plots"");";
			%if %index(%upcase(&CrossType),OUTCROSS) %then %do;
				put "InsertInto(TabCodeList, expr(Here:CPfunc()), Loc(TabButtonList,""Marker Type Counts Plot"")+1+3);";
			%end;
			%else %do;
				put "InsertInto(TabCodeList, expr(Here:CPfunc()), Loc(TabButtonList,""Selected Marker Counts Plot"")+1+1);";
			%end;
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,0);";
		%end;
		
		put "if(host is(Windows),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, windows) ),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, POSIX) ));";
		put "	";

		%if ^%index(%upcase(&CrossType),OUTCROSS) and &nSelMrks %then %do;
			put "lobuttonwindow=Function({},{default local},";
			put "		VListBox(";
			put "			TextBox(""Click the button below to launch the Marker Properties process.""),";
			put "			ButtonBox(""Marker Properties"",";
			put "				if(host is(Windows),";
			put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/MarkerProperties.jsl"", absolute, windows)),";
			put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/MarkerProperties.jsl"", absolute, POSIX) ));";
			put "	";
			** Set input data set and Output Folder *;
			put "				jg:setDefaultParmValues(2, ""MarkerProperties"");";				
			put "				jg:MarkerProperties_updateDeps;";
			put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""InData""),""&ClientOutPath.&TasselGBSImportGen_OutDataName..sas7bdat"");";		
			put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""OutPath""),""&ClientOutPath"");";										
			put "				jg:MarkerProperties_updateDeps;";				
			put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""ListMarkerVars""),""&_fmrk_.--&_lmrk_"");";
			%if %length(&AnnoInData) %then %do;
				put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoInData""),""&ClientOutPath.&TasselGBSImportAnn_OutDataName..sas7bdat"");";
			%end;
			put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""PloidyNum""),""&PloidyLevel"");";				
			put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoLabel""),""&IDVar"");";
			put "				jg:MarkerProperties_updateDeps;";	
			put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""MarkerFmt""),""Numeric Genotypes"");";
			put "				jg:MarkerProperties_updateDeps;";			
			put "			)";
			put "		)";
			put ");";

			put "Here:LaunchCode = lobuttonwindow();";
		%end;
		%else %if %index(%upcase(&CrossType),OUTCROSS) and &nSelMrks %then %do;
			put "lobuttonwindow=Function({},{default local},";
			put "		VListBox(";
			put "			TextBox(""Click the button below to launch the Linkage Group and Recombination for Outcross process.""),";
			put "			ButtonBox(""Linkage Group and Recombination for Outcross"",";
			put "				if(host is(Windows),";
			put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/LinkageGroupsOutcross.jsl"", absolute, windows)),";
			put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/LinkageGroupsOutcross.jsl"", absolute, POSIX) ));";
			put "	";
			** Set input data set and Output Folder *;
			put "				jg:setDefaultParmValues(2, ""LinkageGroupsOutcross"");";				
			put "				jg:LinkageGroupsOutcross_updateDeps;";	
			put "				jg:setCurrentParmValue(2, ""LinkageGroupsOutcross"", jg:getParmNum(""LinkageGroupsOutcross"",""InData""),""&ClientOutPath.&TasselGBSImportGen_OutDataName..sas7bdat"");";		
			put "				jg:setCurrentParmValue(2, ""LinkageGroupsOutcross"", jg:getParmNum(""LinkageGroupsOutcross"",""OutPath""),""&ClientOutPath"");";										
			put "				jg:LinkageGroupsOutcross_updateDeps;";
			put "				jg:setCurrentParmValue(2, ""LinkageGroupsOutcross"", jg:getParmNum(""LinkageGroupsOutcross"",""ListMarkerVars""),""&_fmrk_.--&_lmrk_"");";
			** Set annotation data set*;
			put "				jg:setCurrentParmValue(2, ""LinkageGroupsOutcross"", jg:getParmNum(""LinkageGroupsOutcross"",""AnnoInData""),""&ClientOutPath.&TasselGBSImportAnn_OutDataName..sas7bdat"");";
			put "				jg:setCurrentParmValue(2, ""LinkageGroupsOutcross"", jg:getParmNum(""LinkageGroupsOutcross"",""AnnoMName""),""&IDVar"");";
			put "				jg:setCurrentParmValue(2, ""LinkageGroupsOutcross"", jg:getParmNum(""LinkageGroupsOutcross"",""MarkerType""),""mktype"");";
			put "				jg:LinkageGroupsOutcross_updateDeps;";			
			put "			)";
			put "		)";
			put ");";

			put "Here:LaunchCode = lobuttonwindow();";		
		%end;
	run;
	
	%PackageInsertFile(JSL Code from &ProcessName, &OutPath, &TasselGBSImportScriptName, "&TasselGBSImport_OutDataName..sas7bdat", yes);

	%exit:;
	%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);
	
%mend TasselGBSImportEngine;
%TasselGBSImportEngine;
