/*********************************************/
/**Macro functions for the MIM QTL procedure**/
/*********************************************/

/* nullmodel() creates a data set 'hkmimmodel' framework that stores model for each trait. This function should always be the first to call 
when building a multi-QTL model */
%macro nullmodel(nx=, trt=, hkmimmodel= );

	*--- NEVER uses same prefix anywhere throughout the code ---*;
	
	%local i;
	%local xcovar;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	
	*---create a data set that holds QTL model for all traits ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		data &hkmimmodel;
			set _null_;
		run;
	%end;
	%else  %do;
		data &hkmimmodel;
			set &hkmimmodel;
			where trait ^= &trt;
		run;
	%end;
	
	data hkmodel_tmp;
		length trait 8 
			cofactors $200 
			ncofactors
			inter_cofactor_id
			nqtl 
			neffect
			indic_epi_eff indic_qtl_without_inter 
			qtl1 qtl2 ch1 ch2 pos1 pos2 posid1 posid2 efftype1 efftype2 
			contrast InfinityLike 8;
		retain trait &trt
			cofactors "&xcovar"
			ncofactors &nx
			inter_cofactor_id .
			nqtl 0 
			neffect 0
			indic_epi_eff indic_qtl_without_inter 
			qtl1 qtl2 ch1 ch2 pos1 pos2 posid1 posid2 efftype1 efftype2 
			contrast . InfinityLike 0;
	run;
	
	data &hkmimmodel;
		set &hkmimmodel hkmodel_tmp;
	run;
	
	proc datasets nolist;
		delete hkmodel_tmp;
	quit;
	
	proc sort data=&hkmimmodel;
		by trait;
	run;
	
	%exitnm:;
	
%mend nullmodel;

/* fitmodel() fits a model for a given set of variables */ 
%macro fitmodel(trt=, ng=, hkmimdata=, hkmimmodel=);
	
	ods exclude all;
	ods noresults;	
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmpf;
	%local NOBS_hkmodel_tmpf;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local indicator_epi;
	%local indicator_qxc;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local rc;
	%local currNeg2LogLike;
	%local geffect_main_red;
	%local geffect_epi_red;
	%local geffect_qxc_red;	
	%local id_fit_ind_eff;
	%local NOBS_id_fit_ind_eff;
	%local pastqtl;
	%local contrasts;
	%local nullNeg2LogLike;
	%local LODmodel;
	%local nobsread;
	%local R2;
	%local factorvar_tmp;

	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitfit;
	%end;
	
	*--- create data set that holds the model from the current trait ---*;	
	data hkmodel_tmpf;
		set &hkmimmodel;
		where trait = &trt;
		length qtleffect qtleffect_rec $70;
		if qtl1 ^= . and qtl1 = qtl2 and inter_cofactor_id = . then do; 
			qtleffect = 'e'||trim(left(efftype1))||'c'||trim(left(ch1))||'p'||trim(left(posid1));
			qtleffect_rec = qtleffect;
		end;
		else if qtl1 ^= . and qtl1 ^= qtl2 and inter_cofactor_id = . then do;
			qtleffect = 'e'||trim(left(efftype1))||'c'||trim(left(ch1))||'p'||trim(left(posid1))||'*'||
				    'e'||trim(left(efftype2))||'c'||trim(left(ch2))||'p'||trim(left(posid2));
			qtleffect_rec = 'e'||trim(left(efftype2))||'c'||trim(left(ch2))||'p'||trim(left(posid2))||'*'||
				    	'e'||trim(left(efftype1))||'c'||trim(left(ch1))||'p'||trim(left(posid1));				    
		end;	
		else if qtl1 ^= . and qtl1 = qtl2 and inter_cofactor_id ^= . then do;
			qtleffect = 'x'||trim(left(inter_cofactor_id))||'*'||
				    'e'||trim(left(efftype1))||'c'||trim(left(ch1))||'p'||trim(left(posid1));
			qtleffect_rec = 'e'||trim(left(efftype1))||'c'||trim(left(ch1))||'p'||trim(left(posid1))||'*'||
				    	'x'||trim(left(inter_cofactor_id));				    
		end;			
		effectcount = (_n_ - 1);
	run;
	
	*data OutLib.hkmodel_tmpf; *set hkmodel_tmpf; *run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmpf = %sysfunc(open(hkmodel_tmpf));
	%let NOBS_hkmodel_tmpf =%sysfunc(attrn(&id_hkmodel_tmpf, NOBS));
	*--- very important to close ---*;
	%let id_hkmodel_tmpf = %sysfunc(close(&id_hkmodel_tmpf));
	%if &NOBS_hkmodel_tmpf < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;
		%let exiterror=1;
	%end;
	%else %do;
	
		*---get covariates in the input model---*;
		%let xcovar = ;
		data _null_;
			set hkmodel_tmpf(obs=1);
			call symputx("xcovar", trim(left(cofactors)), 'L');
		run;	
		%put xcovar= &xcovar;
		
		%let u = 1;
		%let cofvar= %scan(&xcovar,&u);
		%do %while(%length(&cofvar));		
			%let u= %sysevalf(&u + 1);		
			%let cofvar= %scan(&xcovar,&u);
		%end;
		%let nx= %sysevalf(&u - 1);
	
		*--- get the number of QTL and effects in the input model ---*;
		proc means data=hkmodel_tmpf noprint;
			var nqtl neffect indic_epi_eff inter_cofactor_id InfinityLike;
			output out=nqtlin_tmp max=numbqtl numbeff indic_epi indic_qxc InfinityLike;
		run;

		data _null_;
			set nqtlin_tmp;
			call symputx("numbqtlin", numbqtl, 'L');
			call symputx("numbeffin", numbeff, 'L');
			call symputx("indicator_epi", indic_epi, 'L');
			call symputx("indicator_qxc", indic_qxc, 'L');
			call symputx("InfinityLikein", InfinityLike, 'L');
		run;	
		
		%if %trim(%left(&indicator_qxc)) = %str(.) %then %do;
			%put indicator_qxc= &indicator_qxc;
			%let indicator_qxc= 0;
		%end;
		%put indicator_qxc= &indicator_qxc;

		proc datasets nolist;
			delete nqtlin_tmp;
		quit;

		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		%*put # # # Number of QTL = &numbqtlin # # # # # #;
		%*put # # # Number of Effects = &numbeffin # # # #;
		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		
		%if %sysevalf(&numbqtlin > 0) %then %do;
		
			*--- begin Individual effects tests ---*;
			%let geffect_main_in= ;
			%let geffect_epi_in= ;
			%let geffect_qxc_in= ;
			
			proc sql;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_in separated by ' '
				from hkmodel_tmpf
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
				
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
					'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_in separated by ' '
				from hkmodel_tmpf
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
								
				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_in separated by ' '
				from hkmodel_tmpf
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;				
			quit;
			%put geffect_main_in= &geffect_main_in;
			%put geffect_epi_in= &geffect_epi_in;
			%put geffect_qxc_in= &geffect_qxc_in;

			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);
			
			proc datasets nolist;
				delete hkdata3f;
				change hkdata3=hkdata3f;
			quit;
			
			*data OutLib.ft_hkdata3f; *set hkdata3f; *run;

			*--- get estimates and ProbT for each effect in the model (individual effect tests - ProbT) ---;
			*--- fit input model ---*;
			proc mixed data=hkdata3f method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in / solution;
				ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
			run;
			
			data fitmodel_st;
				set st_tmp1;
				if upcase(descr) = "-2 LOG LIKELIHOOD" then 
					call symputx("currNeg2LogLike", trim(left(value)), 'L');
				label value = "Value";
				format value 8.2;
			run;
			
			*--- include QTL number and effect type in the output data set ---*;			
			data interc_cov_tmp;
				set es_tmp1(rename=(Effect= Effect_var));
				where (substr(trim(left(Effect_var)),1,1) = 'I' or 
					(substr(trim(left(Effect_var)),1,1) = 'x') and find(Effect_var,'*')=0);
			run;
			
			*data OutLib.es_tmp1; *set es_tmp1; *run;

			proc sql;
				create table ind_eff_tmp as
				select b.qtl1 as QTL1, b.qtl2 as QTL2, b.ch1 as Ch1, b.ch2 as Ch2, b.pos1 as Pos1, b.pos2 as Pos2, 
				b.inter_cofactor_id, b.indic_epi_eff, b.indic_qtl_without_inter, b.effectcount,
				a.Effect as Effect_var, a.Estimate, a.StdErr, a.DF, a.tValue, a.Probt  
				from es_tmp1 as a, hkmodel_tmpf as b
				where a.effect= b.qtleffect or a.effect= b.qtleffect_rec;
			quit;	
			
			*data OutLib.ind_eff_tmp; *set ind_eff_tmp; *run;

			data ind_eff_tmp(drop = eff1_tmp eff2_tmp effect1_tmp effect2_tmp);
				retain inter_cofactor_id QTL1 QTL2 Ch1 Ch2 Pos1 Pos2 Effect_type Effect Estimate StdErr DF tValue Probt effectcount;
				length Effect_type $20;
				set ind_eff_tmp interc_cov_tmp;
				length Effect $30 Effect1_tmp $20 Effect2_tmp $20;
				if QTL1 ^= . and QTL1 = QTL2 and inter_cofactor_id =. then do;
					if substr(trim(left(effect_var)),1,2) = 'e1' then Effect = 'Additive2';
					else if substr(trim(left(effect_var)),1,2) = 'e2' then Effect = 'Dominance';
					else if substr(trim(left(effect_var)),1,2) = 'e3' then Effect = 'Additive1';
				end;
				else if QTL1 ^=. and QTL1 ^= QTL2 and inter_cofactor_id = . then do;
					eff1_tmp = scan(trim(left(effect_var)), 1, '*');
					eff2_tmp = scan(trim(left(effect_var)), 2, '*');
					if substr(trim(left(eff1_tmp)),1,2) = 'e1' then Effect1_tmp = 'Additive2';
					else if substr(trim(left(eff1_tmp)),1,2) = 'e2' then Effect1_tmp = 'Dominance';	
					else if substr(trim(left(eff1_tmp)),1,2) = 'e3' then Effect1_tmp = 'Additive1';	
					if substr(trim(left(eff2_tmp)),1,2) = 'e1' then Effect2_tmp = 'Additive2';
					else if substr(trim(left(eff2_tmp)),1,2) = 'e2' then Effect2_tmp = 'Dominance';	
					else if substr(trim(left(eff2_tmp)),1,2) = 'e3' then Effect2_tmp = 'Additive1';	
					Effect = trim(left(Effect1_tmp))||' * '||trim(left(Effect2_tmp));
				end;
				else if QTL1 ^=. and QTL1 = QTL2 and inter_cofactor_id ^= . then do;
					eff2_tmp = scan(trim(left(effect_var)), 2, '*');	
					if substr(trim(left(eff2_tmp)),1,2) = 'e1' then Effect2_tmp = 'Additive2';
					else if substr(trim(left(eff2_tmp)),1,2) = 'e2' then Effect2_tmp = 'Dominance';	
					else if substr(trim(left(eff2_tmp)),1,2) = 'e3' then Effect2_tmp = 'Additive1';	
					Effect = "x"||trim(left(inter_cofactor_id))||' * '||trim(left(Effect2_tmp));
				end;				
				else Effect = Effect_var;
				if indic_epi_eff = 0 and inter_cofactor_id = . then Effect_type = 'Main';
				else if indic_epi_eff = 1 and inter_cofactor_id = . then Effect_type = 'Interaction';
				else if indic_epi_eff = 0 and inter_cofactor_id ^= . then Effect_type = 'Interaction';
				else Effect_type = '';
			run;
			
			*data OutLib.ind_eff_tmp; *set ind_eff_tmp; *run;
			
			*--- detele temporary datasets ---*;
			proc datasets nolist;
				delete	st_tmp1
					es_tmp1;
			quit;			
			
			*--- This step makes sure that possibly existing previous files will be deleted to avoid pilling up unwanted results together ---*;
			data st_tmp1_all;
				set _null_;
			run;
			
			*--- get estimates of AIC, AICc and BIC for each effect in the model (Individual effect tests - AIC, AICc, BIC) ---;
			%put # # # Individual test of effects of QTL # # #;
			%do j=1 %to &numbeffin;
			
				%put # # # Fitmodel: Trait = &trt, Effect= &j # # #;
			
				data hkmodel_red_tmp;
					set hkmodel_tmpf;
					*--- removes each effect from the complete model. ---*;
					*--- (j + 1) because the first line has no effect ---*;					
					if effectcount ^= &j then output; 
				run;
			
				*--- get effects of QTL in the reduced model ---*;
				%let geffect_main_red=;
				%let geffect_epi_red =;
				%let geffect_qxc_red =;
			
				proc sql;
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
					
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
					
					select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
						'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;					
					
				quit;
				%put geffect_main_red= &geffect_main_red;
				%put geffect_epi_red= &geffect_epi_red;
				%put geffect_qxc_red= &geffect_qxc_red;
			
				proc mixed data=hkdata3f method=ml;
					model y&trt = &xcovar &geffect_main_red &geffect_epi_red &geffect_qxc_red/ solution;
					ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
				run;
			
				proc transpose data=st_tmp1 out=st_tmp1;
					id Descr;
					var value;
				run;
			
				data st_tmp1;
					set st_tmp1;
					effectcount = &j;
				run;
			
				data st_tmp1_all (drop= _name_);
					set %if %sysfunc(exist(st_tmp1_all)) %then st_tmp1_all; st_tmp1;
				run;
				
				proc datasets nolist;
					delete	hkmodel_red_tmp
						st_tmp1
						es_tmp1;
				quit;
			
			%end; *--- end do loop over &numbeffin ---*;	
			
			*data OutLib.ind_eff_tmp; *set ind_eff_tmp; *run;
			*data OutLib.st_tmp1_all; *set st_tmp1_all; *run;
			
			*--- Merge Log Likelihood, AIC, AICC, and BIC to ProbT for each effect (Individual effects tests - all statistics)---*;
			proc sql;
				create table ind_eff as
				select a.*, b._2_Log_Likelihood, b.AIC__smaller_is_better_, b.AICC__smaller_is_better_, b.BIC__smaller_is_better_  
				from ind_eff_tmp as a, st_tmp1_all as b
				where a.effectcount = b.effectcount;
			quit;
			
			*data OutLib.ind_eff; *set ind_eff; *run;
			
			*--- Intercept and covariates ---*;
			data interc_cov;
				length Predictors $40;
				set interc_cov_tmp(rename= (Effect_var= Predictors
									DF= DF_IndTest
									tvalue= tvalue_IndTest
									Probt= Probt_IndTest));
				label Predictors = "Predictors";
				label Estimate = "Estimate";
				label StdErr = "Standard Error";
				label DF_IndTest = "Individual Effect Test DF";
				label tvalue_IndTest = "Individual Effect Test t Value";
				label Probt_IndTest = "Individual Effect Test Pr>|t|";
			run;
			
			*data OutLib.interc_cov0; *set interc_cov; *run;

			%put FACTORVARSNEW= &FACTORVARSNEW;
			%put xcovar= &xcovar;
			%let u = 1;
			%let cofvar= %scan(&xcovar,&u);
			data interc_cov; 
				set interc_cov;
				%do %while(%length(&cofvar));
					%let cofid= %substr(&cofvar, 2);	
					%let factorvar_tmp = %scan(&FACTORVARSNEW,&cofid);
					%put factorvar_tmp = &factorvar_tmp;
					if trim(left(Predictors)) = "&cofvar" then Predictors = "&factorvar_tmp";
					%let u= %sysevalf(&u + 1);		
					%let cofvar= %scan(&xcovar,&u);
				%end;
			run;			
			
			*data OutLib.interc_cov1; *set interc_cov; *run;
			
			*--- delete temporary data sets ---*;
			proc datasets nolist;
				delete interc_cov_tmp
					ind_eff_tmp
					st_tmp1_all;
			quit;
			*--- end Individual effects tests ---*;
						
			*--- start Joint main test ---*;
			%put # # # Joint test of main effects of each QTL # # #;	
			*--- AIC, AICc and BIC for the joint test of main effects of each QTL (Joint main tests) ---;									
			%do j=1 %to &numbqtlin;
			
				%put # # # Fitmodel: Trait = &trt, QTL= &j # # #;
			
				data hkmodel_red_tmp;
					set hkmodel_tmpf;
					*--- main effects of each QTL are removed from the complete model ---*; 
					if  qtl1 = &j and qtl2 = &j and inter_cofactor_id = . then delete;
				run;
			
				*--- get effects of QTL in the reduced model ---*;
				%let geffect_main_red=;
				%let geffect_epi_red =;
				%let geffect_qxc_red =;
			
				proc sql;
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
					
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
					
					select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
						'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;					
				quit;
				%put geffect_main_red= &geffect_main_red;
				%put geffect_epi_red= &geffect_epi_red;
				%put geffect_qxc_red= &geffect_qxc_red;
			
				proc mixed data=hkdata3f method=ml;
					model y&trt = &xcovar &geffect_main_red &geffect_epi_red &geffect_qxc_red/ solution;
					ods output fitstatistics=st_tmp2 solutionf=es_tmp2;
				run;
			
				proc transpose data=st_tmp2 out=st_tmp2;
					id Descr;
					var value;
				run;
			
				data st_tmp2;
					set st_tmp2;
					QTL = &j;
				run;
			
				data st_tmp2_all (drop= _name_);
					set %if %sysfunc(exist(st_tmp2_all)) %then st_tmp2_all; st_tmp2;
				run;	
				
				proc datasets nolist;
					delete	hkmodel_red_tmp
						st_tmp2
						es_tmp2;
				quit;
			
			%end; *--- end do loop over &numbqtlin (j) ---*;	

			*--- contrasts to test joint main effects of each QTL (Joint main tests - ProbF) ---*;
			%let id_fit_ind_eff = %sysfunc(open(ind_eff));
			%let NOBS_id_fit_ind_eff = %sysfunc(attrn(&id_fit_ind_eff, NOBS));
					
			%let pastqtl = -99;
			%let contrasts = ;			
			%syscall set(id_fit_ind_eff);
			%do i=1 %to &NOBS_id_fit_ind_eff;
				%let rc = %sysfunc(fetchobs(&id_fit_ind_eff,&i));
				
				%if &DF ^= . and &qtl1 ^= . and &qtl1 = &qtl2 and &inter_cofactor_id = . %then %do;
				
					%put # # # Fitmodel: QTL = &qtl1 # # #;
		
					%if (%sysevalf(&pastqtl ^= &qtl1)) %then %do;
						%if &pastqtl = -99 %then 
							%let contrasts = &contrasts %str(contrast "&qtl1" %trim(%left(&effect_var)) 1);
						%else 
							%let contrasts = &contrasts %str(; contrast "&qtl1" %trim(%left(&effect_var)) 1);
					%end;
					%else %do;
						%let contrasts = &contrasts %str(, %trim(%left(&effect_var)) 1);	
					%end;
					
				%end;
				%let pastqtl = &qtl1;
			%end;
			%let id_fit_ind_eff = %sysfunc(close(&id_fit_ind_eff)); *--- very important to close ---*;

			%if %length(contrasts) %then %do;
				%let contrasts = %str(&contrasts ;);
			%end;
			%put contrasts = &contrasts;
		
			*--- probabilities of joint test of main effects of each QTL using contrast statement (Joint main tests - ProbF ) ---*;
			proc mixed data=hkdata3f method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
				ods output fitstatistics=st_tmp3 solutionf=es_tmp3 contrasts=ct_tmp3;
				&contrasts;
			run;
			
			proc sql;
				create table joint_main as
				select distinct input(a.label,7.) as QTL, a.NumDF, a.DenDF, a.FValue, a.Probf, 
				b._2_Log_likelihood, b.AIC__smaller_is_better_, b.AICC__smaller_is_better_, b.BIC__smaller_is_better_, 
				c.indic_qtl_without_inter, c.contrast, c.inter_cofactor_id
				from ct_tmp3 as a, st_tmp2_all as b, hkmodel_tmpf as c
				where input(a.label,7.) = b.QTL and b.QTL = c.qtl1 and c.qtl1 = c.qtl2 and c.inter_cofactor_id = . 
				order by QTL;
			quit;
			
			*data OutLib.joint_main; *set joint_main; *run;
			
			%*goto exitfit;
			
			data joint_main(drop = indic_qtl_without_inter inter_cofactor_id _2_Log_likelihood contrast);
				retain QTL H0_JointMain NumDF_JointMain DenDF_JointMain fValue_JointMain Probf_JointMain
					LOD_JointMain AIC_JointMain AICC_JointMain BIC_JointMain;
				length H0_JointMain $60;
				set joint_main(rename=(NumDF= NumDF_JointMain 
							DenDF= DenDF_JointMain 
							fValue= fValue_JointMain 
							Probf= Probf_JointMain
							AIC__smaller_is_better_ = AIC_JointMain
							AICC__smaller_is_better_= AICC_JointMain
							BIC__smaller_is_better_ = BIC_JointMain
						));
				LOD_JointMain = (_2_Log_Likelihood - &currNeg2LogLike) / (2 * %sysfunc(log(10)));											
				if indic_qtl_without_inter = 1 then do; 
					/*
					if &ng = 2 then H0_JointMain = "Additive2=0";
					else if &ng = 3  then H0_JointMain = "Additive2=Dominance=0";
					else if &ng = 4  then do;
						if contrast=2 then H0_JointMain = "Additive1=0";
						else if contrast=3 then H0_JointMain = "Additive2=Dominance=0";
						else if contrast=4 then H0_JointMain = "Additive1=Dominance=0";
						else if contrast=5 then H0_JointMain = "Additive2=0";
						else if contrast=6 then H0_JointMain = "Additive2=Additive1=0";
						else if contrast=7 then H0_JointMain = "Additive2=Dominance=0";
						else if contrast=8 then H0_JointMain = "Additive2=Additive1=Dominance=0";
					end;
					*/
					if &ng = 2 then do;
						H0_JointMain = "Additive=0";
					end;
					else if &ng = 3  then do;
						H0_JointMain = "Additive=Dominance=0";
					end;
					else if &ng = 4  then do;
						if contrast=2 then H0_JointMain = "Additive1=0";
						else if contrast=3 then H0_JointMain = "Additive2=Dominance=0";
						else if contrast=4 then H0_JointMain = "Additive1=Dominance=0";
						else if contrast=5 then H0_JointMain = "Additive2=0";
						else if contrast=6 then H0_JointMain = "Additive2=Additive1=0";
						else if contrast=7 then H0_JointMain = "Additive2=Dominance=0";
						else if contrast=8 then H0_JointMain = "Additive2=Additive1=Dominance=0";
					end;					
				end;
				else if indic_qtl_without_inter = 0 then do;
					/*
					if &ng = 2 then H0_JointMain = "Additive2=0 & Interaction not 0";
					else if &ng = 3  then H0_JointMain = "Additive2=Dominance=0 & Interaction not 0";
					else if &ng = 4  then do;
						if contrast= 2 then H0_JointMain = "Additive1=0 & Interaction not 0";
						else if contrast= 3 then H0_JointMain = "Additive2=Dominance=0 & Interaction not 0";
						else if contrast= 4 then H0_JointMain = "Additive1=Dominance=0 & Interaction not 0";
						else if contrast= 5 then H0_JointMain = "Additive2=0 & Interaction not 0";
						else if contrast= 6 then H0_JointMain = "Additive2=Additive1=0 & Interaction not 0";
						else if contrast= 7 then H0_JointMain = "Additive2=Dominance=0 & Interaction not 0";
						else if contrast= 8 then H0_JointMain = "Additive2=Additive1=Dominance=0 & Interaction not 0";
					end;
					*/
					if &ng = 2 then do;
						H0_JointMain = "Additive=0 & Interaction not 0";
					end;
					else if &ng = 3 then do;
						H0_JointMain = "Additive=Dominance=0 & Interaction not 0";
					end;
					else if &ng = 4 then do;
						if contrast= 2 then H0_JointMain = "Additive1=0 & Interaction not 0";
						else if contrast= 3 then H0_JointMain = "Additive2=Dominance=0 & Interaction not 0";
						else if contrast= 4 then H0_JointMain = "Additive1=Dominance=0 & Interaction not 0";
						else if contrast= 5 then H0_JointMain = "Additive2=0 & Interaction not 0";
						else if contrast= 6 then H0_JointMain = "Additive2=Additive1=0 & Interaction not 0";
						else if contrast= 7 then H0_JointMain = "Additive2=Dominance=0 & Interaction not 0";
						else if contrast= 8 then H0_JointMain = "Additive2=Additive1=Dominance=0 & Interaction not 0";
					end;					
				end;
				label H0_JointMain = "Joint Main Effect Test";
				label NumDF_JointMain = "Joint Main Effect Test NumDF";
				label DenDF_JointMain = "Joint Main Effect Test NenDF";
				label fValue_JointMain = "Joint Main Effect Test F Value";
				label Probf_JointMain = "Joint Main Effect Test Pr>F";
				label LOD_JointMain = "Joint Main Effect Test LOD";
				label AIC_JointMain = "Joint Main Effect Test AIC";
				label AICC_JointMain = "Joint Main Effect Test AICC";
				label BIC_JointMain = "Joint Main Effect Test BIC";
			run;
			
			%*goto exitfit;
			
			*--- delete temporaty data sets ---*;
			proc datasets nolist;
				delete 	st_tmp2_all
					st_tmp3
					es_tmp3
					ct_tmp3;
			quit;
			*--- end Joint main tests ---*;

			*--- start Joint main and interaction tests ---*;
			%if %sysevalf(&indicator_epi > 0 or &indicator_qxc > 0) %then %do;
				%put # # # Joint test of main and interaction effects of each QTL # # #;

				*--- This step makes sure that possibly existing previous files will be deleted to avoid pilling up unwanted results together ---*;
				data st_tmp2_all;
					set _null_;
				run;

				*--- AIC, AICc and BIC for the joint test of main and interaction effects of each QTL (Joint Main and Interaction tests) ---;									
				%do j=1 %to &numbqtlin;

					%put # # # Fitmodel: Trait = &trt, QTL= &j # # #;

					data hkmodel_red_tmp;
						set hkmodel_tmpf;
						*--- main and interation effects of each QTL are removed from the complete model ---*; 
						if  qtl1 ^= &j and qtl2 ^= &j then output;
					run;

					*--- get effects of QTL in the reduced model ---*;
					%let geffect_main_red=;
					%let geffect_epi_red =;
					%let geffect_qxc_red =;

					proc sql;
						select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_red separated by ' '
						from hkmodel_red_tmp
						where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
						
						select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
							'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_red separated by ' '
						from hkmodel_red_tmp
						where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
						
						select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
							'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_red separated by ' '
						from hkmodel_red_tmp
						where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;						
					quit;
					%put geffect_main_red= &geffect_main_red;
					%put geffect_epi_red= &geffect_epi_red;
					%put geffect_qxc_red= &geffect_qxc_red;

					proc mixed data=hkdata3f method=ml;
						model y&trt = &xcovar &geffect_main_red &geffect_epi_red &geffect_qxc_red/ solution;
						ods output fitstatistics=st_tmp2 solutionf=es_tmp2;
					run;

					proc transpose data=st_tmp2 out=st_tmp2;
						id Descr;
						var value;
					run;

					data st_tmp2;
						set st_tmp2;
						QTL = &j;
					run;

					data st_tmp2_all (drop= _name_);
						set %if %sysfunc(exist(st_tmp2_all)) %then st_tmp2_all; st_tmp2;
					run;	

					proc datasets nolist;
						delete	hkmodel_red_tmp
							st_tmp2
							es_tmp2;
					quit;

				%end; *--- end do loop over &numbqtlin (j) ---*;	

				/*
				proc sql;
					create table joint_main_inter as
					select distinct a.*, b.indic_qtl_without_inter
					from st_tmp2_all as a, hkmodel_tmpf as b
					where a.QTL = b.qtl1 and b.qtl1 = b.qtl2 and inter_cofactor_id = .;
					order by QTL;
				quit;
				*/
				
				%put # # # Contrasts to test joint main and interaction effects of each QTL # # #;
				*--- contrasts to test joint main and interaction effects of each QTL (Joint main and interaction tests - ProbF) ---*;
				%let id_fit_ind_eff = %sysfunc(open(ind_eff));
				%let NOBS_id_fit_ind_eff = %sysfunc(attrn(&id_fit_ind_eff, NOBS));

				%let contrasts = ;			
				%syscall set(id_fit_ind_eff);

				%do k=1 %to &numbqtlin;
					%let pastqtl = -99;
					%do i=1 %to &NOBS_id_fit_ind_eff;
						%let rc = %sysfunc(fetchobs(&id_fit_ind_eff,&i));

						%put # # # Fitmodel: QTL = &qtl1 # # #;

						%if %sysevalf(&DF ^= .) %then %do;

							%if (%sysevalf(&qtl1 = &k and &qtl2 = &k)) %then %do;

								%if %sysevalf(&pastqtl = -99) %then %do; 
									%let contrasts = &contrasts %str(contrast "&qtl1" %trim(%left(&effect_var)) 1);
									%let pastqtl = &k;
								%end;
								%else %let contrasts = &contrasts %str(, %trim(%left(&effect_var)) 1);	
							%end;
							%else %if (%sysevalf(&qtl1 = &k and &qtl2 ^= &k or &qtl1 ^= &k and &qtl2 = &k)) %then %do;

								%if %sysevalf(&pastqtl = -99) %then %do;
									%let contrasts = &contrasts %str(contrast "&qtl1" %trim(%left(&effect_var)) 1);
									%let pastqtl = &k;
								%end;
								%else %let contrasts = &contrasts %str(, %trim(%left(&effect_var)) 1);	
							%end;	
						%end;
					%end;
					%let contrasts = %str(&contrasts ;);
				%end;
				*--- very important to close ---*;
				%let id_fit_ind_eff = %sysfunc(close(&id_fit_ind_eff)); 
				%put contrasts = &contrasts;

				*--- probabilities of joint test of main effects of each QTL using contrast statement (Joint main tests - ProbF ) ---*;
				proc mixed data=hkdata3f method=ml;
					model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
					ods output fitstatistics=st_tmp3 solutionf=es_tmp3 contrasts=ct_tmp3;
					&contrasts;
				run;
				
				proc sql;
					create table joint_main_inter as
					select distinct input(a.label,7.) as QTL, a.NumDF, a.DenDF, a.FValue, a.Probf, 
					b._2_Log_likelihood, b.AIC__smaller_is_better_, b.AICC__smaller_is_better_, b.BIC__smaller_is_better_, 
					c.indic_qtl_without_inter, c.inter_cofactor_id
					from ct_tmp3 as a, st_tmp2_all as b, hkmodel_tmpf as c
					where input(a.label, 7.) = b.QTL and b.QTL = c.qtl1 and c.qtl1 = c.qtl2 and c.inter_cofactor_id = . 
					order by QTL;
				quit;
				
				*data Outlib.joint_main_inter; *set joint_main_inter; *run;
				
				%*goto exitfit;
				
				data joint_main_inter(drop = _2_Log_Likelihood indic_qtl_without_inter inter_cofactor_id);
					retain  QTL H0_JointMainInter NumDF_JointMainInter DenDF_JointMainInter fvalue_JointMainInter ProbF_JointMainInter 
						LOD_JointMainInter AIC_JointMainInter AICC_JointMainInter BIC_JointMainInter; 
					length H0_JointMainInter $40;
					set joint_main_inter(rename=(NumDF= NumDF_JointMainInter
								DenDF= DenDF_JointMainInter
								fvalue= fvalue_JointMainInter
								ProbF= ProbF_JointMainInter
								AIC__smaller_is_better_= AIC_JointMainInter
								AICC__smaller_is_better_=AICC_JointMainInter
								BIC__smaller_is_better_=BIC_JointMainInter));
					LOD_JointMainInter = (_2_Log_Likelihood - &currNeg2LogLike) / (2 * %sysfunc(log(10)));
					if indic_qtl_without_inter = 0 then H0_JointMainInter = 'All Effects (Main & Interaction) = 0';
					else if indic_qtl_without_inter = 1 then H0_JointMainInter = 'All Effects (Main) = 0';
					label H0_JointMainInter ="Joint Main and Interaction Effect Test";
					label NumDF_JointMainInter ="Joint Main and Interaction Effect Test NumDF";
					label DenDF_JointMainInter ="Joint Main and Interaction Effect Test DenDF";
					label fvalue_JointMainInter ="Joint Main and Interaction Effect Test F Value";
					label ProbF_JointMainInter ="Joint Main and Interaction Effect Test Pr>F";
					label LOD_JointMainInter ="Joint Main and Interaction Effect Test LOD";
					label AIC_JointMainInter ="Joint Main and Interaction Effect Test AIC";
					label AICC_JointMainInter ="Joint Main and Interaction Effect Test AICC";
					label BIC_JointMainInter ="Joint Main and Interaction Effect Test BIC";
				run;

				%*goto exitfit;

				*--- delete temporary data sets ---*;
				proc datasets nolist;
					delete st_tmp2_all
						st_tmp3 
						es_tmp3 
						ct_tmp3;
				quit;
	 		%end;
	 		*--- end Joint main and interaction tests ---*;
	 
			*--- clean up the rest of unwanted variables in the output data set (Individual effects tests - all statistics) ---*;
			data ind_eff(drop = Effect_var indic_epi_eff 
					 indic_qtl_without_inter
					effectcount _2_Log_Likelihood);
				retain Predictors QTL1 QTL2 Ch1 Ch2 Pos1 Pos2 Effect_type Effect Estimate StdErr 
					H0_IndTest DF_IndTest tValue_IndTest Probt_IndTest LOD_IndTest AIC_IndTest 
					AICC_IndTest BIC_IndTest;
				length Predictors $40 H0_IndTest $15;
				set ind_eff(rename=(DF=DF_IndTest 
							tValue= tValue_IndTest 
							Probt= Probt_IndTest 
							AIC__smaller_is_better_= AIC_IndTest 
							AICC__smaller_is_better_=AICC_IndTest
							BIC__smaller_is_better_= BIC_IndTest
							)
						);
				Predictors = 'QTL Effect';
				LOD_IndTest = (_2_Log_Likelihood - &currNeg2LogLike) / (2 * %sysfunc(log(10)));
				%if &ng < 4 %then %do;
					if Effect_type = 'Main' and Effect = 'Additive2' then do;
						Effect = 'Additive';
						H0_IndTest = "Additive=0";
					end;
					else if Effect_type = 'Main' and Effect = 'Dominance' then do;
						H0_IndTest = "Dominance=0";
					end;
					else if Effect_type = 'Interaction' then do;
						Effect = tranwrd(Effect, 'Additive2','Additive');
						H0_IndTest = "Interaction=0";
					end;
				%end;
				%else %do;
					***assuming maximum ng = 4 ***;
					if Effect_type = 'Main' and Effect = 'Additive2' then do;
						H0_IndTest = "Additive2=0";
					end;
					else if Effect_type = 'Main' and Effect = 'Dominance' then do;
						H0_IndTest = "Dominance=0";
					end;
					else if Effect_type = 'Main' and Effect = 'Additive1' then do;
						H0_IndTest = "Additive1=0";
					end;
					else if Effect_type = 'Interaction' then do;
						H0_IndTest = "Interaction=0";
					end;
				%end;
				label Predictors = "Predictors";
				label QTL1 = "QTL1";
				label QTL2 = "QTL2";
				label Ch1 = "Chrom1";
				label Ch2 = "Chrom2";
				label Pos1 = "Position1";
				label Pos2 = "Position2";
				label Effect_type = "Effect Type";
				label Effect = "Effect";
				label Estimate = "Estimate";
				label StdErr = "Standard Error";
				label H0_IndTest = "Individual Effect Test";
				label DF_IndTest = "Individual Effect Test DF";
				label tValue_IndTest="Individual Effect Test t Value";
				label Probt_IndTest="Individual Effect Test Pr > |t|";
				label LOD_IndTest = "Individual Effect Test LOD";
				label AIC_IndTest ="Individual Effect Test AIC";
				label AICC_IndTest ="Individual Effect Test AICC";
				label BIC_IndTest ="Individual Effect Test BIC";
			run;
			
			*--- final output dataset: merge individual, joint main, and joint main and interaction effect tests results ---*;
			%put indicator_epi= &indicator_epi;
			%put indicator_qxc= &indicator_qxc;
			%if %sysevalf(&indicator_epi > 0 or &indicator_qxc > 0) %then %do;

				data fitmodel_sum;
					set _null_;
				run;
							
				proc sql;
					create table fitmodel_sum(drop= QTL) as
					select a.*, b.*, c.*
					from ind_eff as a, joint_main as b, joint_main_inter as c
					where a.QTL1 = b.QTL and b.QTL = c.QTL;
				quit;

				proc datasets nolist;
					delete ind_eff
						joint_main
						joint_main_inter;
				quit;
				
				*data OutLib.fitmodel_sum0; *set fitmodel_sum; *run;

				data fitmodel_sum;
					set fitmodel_sum;
					if DF_IndTest = . then do;
						H0_IndTest= '';
						LOD_IndTest = .;
						AIC_IndTest = .;
						AICC_IndTest = .;
						BIC_IndTest = .;
						H0_JointMain='';
						NumDF_JointMain=.;
						DenDF_JointMain=.;
						fValue_JointMain=.;
						Probf_JointMain=.;
						LOD_JointMain=.;
						AIC_JointMain=.;
						AICC_JointMain=.;
						BIC_JointMain=.;
						H0_JointMainInter='';
						NumDF_JointMainInter=.;
						DenDF_JointMainInter=.;
						fValue_JointMainInter=.;
						Probf_JointMainInter=.;
						LOD_JointMainInter=.;
						AIC_JointMainInter=.;
						AICC_JointMainInter=.;
						BIC_JointMainInter=.;	
					end;
					if QTL1 ^= QTL2 then do;
						H0_JointMain='';
						NumDF_JointMain=.;
						DenDF_JointMain=.;
						fValue_JointMain=.;
						Probf_JointMain=.;
						LOD_JointMain=.;
						AIC_JointMain=.;
						AICC_JointMain=.;
						BIC_JointMain=.;						
						H0_JointMainInter='';
						NumDF_JointMainInter=.;
						DenDF_JointMainInter=.;
						fValue_JointMainInter=.;
						Probf_JointMainInter=.;						
						LOD_JointMainInter=.;
						AIC_JointMainInter=.;
						AICC_JointMainInter=.;
						BIC_JointMainInter=.;				
					end;
				run;
			%end;
			%else %do;
				proc sql;		
					create table fitmodel_sum(drop= QTL) as
					select a.*, b.*
					from ind_eff as a, joint_main as b
					where a.QTL1 = b.QTL;
				quit;

				data fitmodel_sum;
					set fitmodel_sum;					
					if DF_IndTest = . then do;
						H0_IndTest= '';
						LOD_IndTest = .;
						AIC_IndTest = .;
						AICC_IndTest = .;
						BIC_IndTest = .;
						H0_JointMain='';
						NumDF_JointMain=.;
						DenDF_JointMain=.;
						fValue_JointMain=.;
						Probf_JointMain=.;
						LOD_JointMain=.;
						AIC_JointMain=.;
						AICC_JointMain=.;
						BIC_JointMain=.;						
					end;
					if QTL1 ^= QTL2 then do;
						H0_JointMain='';
						NumDF_JointMain=.;
						DenDF_JointMain=.;
						fValue_JointMain=.;
						Probf_JointMain=.;
						LOD_JointMain=.;
						AIC_JointMain=.;
						AICC_JointMain=.;
						BIC_JointMain=.;						
					end;
				run;
				
				proc datasets nolist;
					delete ind_eff
						joint_main;
				quit;
				
			*end if indicator_epi < 1 or indicator_qxc < 1*;	
			%end; 
			
			proc datasets nolist nowarn force;
				append 
				base= fitmodel_sum
				data = interc_cov;
			quit;
		
		*--- end if numbqtlin > 0 ---*;
		%end; 
		%else %do;
		
			%prephkdata3(ng=&ng, qtlvars=, talldsn=&hkmimdata);
						
			proc datasets nolist;
				delete hkdata3f;
				change hkdata3=hkdata3f;
			quit;
			
			*data OutLib.ft_hkdata3f; *set hkdata3f; *run;

			*--- fit input model ---*;
			proc mixed data=hkdata3f method=ml;
				model y&trt = &xcovar / solution;
				ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
			run;
			
			data fitmodel_st;
				set st_tmp1;
				if upcase(descr) = "-2 LOG LIKELIHOOD" then 
					call symputx("currNeg2LogLike", trim(left(value)), 'L');
				label value = "Value";
				format value 8.2;
			run;
			
			data fitmodel_sum;
				length Predictors $40;
				set es_tmp1 (rename = (	Effect= Predictors
							DF= DF_IndTest
							tvalue= tvalue_IndTest
							Probt= Probt_IndTest
						));
				label Predictors = "Predictors";
				label Estimate = "Estimate";
				label StdErr = "Standard Error";
				label DF_Indtest = "Individual Test Effect DF";
				label tvalue_Indtest = "Individual Test Effect t Value";
				label Probt_Indtest = "Individual Test Effect Pr>|t|";
			run;
			
			
			%put FACTORVARSNEW= &FACTORVARSNEW;
			%put xcovar= &xcovar;
			%let u = 1;
			%let cofvar= %scan(&xcovar,&u);
			data fitmodel_sum;
				set fitmodel_sum;
				%do %while(%length(&cofvar));
					%let cofid= %substr(&cofvar, 2);	
					%let factorvar_tmp = %scan(&FACTORVARSNEW,&cofid);
					%put factorvar_tmp = &factorvar_tmp;
					if trim(left(Predictors)) = "&cofvar" then Predictors = "&factorvar_tmp";
					%let u= %sysevalf(&u + 1);		
					%let cofvar= %scan(&xcovar,&u);
				%end;
			run;
			
			*--- delete temporary data sets ---*;
			proc datasets nolist;
				delete 	st_tmp1
					es_tmp1;
			quit;
		%end;
				
	%end; *--- end if &NOBS_hkmodel_tmpf >= 1 ---*;
	
	*--- Estimate R2 ---*;
	proc mixed data=hkdata3f method=ml;
		model y&trt = / solution;
		ods output fitstatistics=st_tmp solutionf=es_tmp nobs=nobm;
	run;
	
	data _null_;
		set st_tmp;
		if upcase(descr) = "-2 LOG LIKELIHOOD" then 
			call symputx("nullNeg2LogLike", trim(left(value)), 'L');
	run;
	
	data _null_;
		set nobm;
		if label = "Number of Observations Read" then
			call symputx("nobsread", trim(left(NObsRead)), 'L');		
	run;	
	
	%let LODmodel = %sysevalf((1/(2 * 2.30258509299404))*(&nullneg2loglike - &currneg2loglike));
	
	%let R2 = %sysevalf(1 - 10**(-2*&LODmodel / &nobsread));
	
	data Rsq;
		length descr $25;
		retain descr value;
		descr = "R-squared";
		value = &R2;
		format value 8.2;
	run;
	
	proc datasets nolist;
		append
		base=fitmodel_st
		data=Rsq;;
	quit;
	
	*--- delete temporary data sets ---*;
	proc datasets;
		delete hkmodel_tmpf
			hkdata3f
			st_tmp
			es_tmp;
	quit;
	
	%exitfit:;
	
	ods exclude none;
	ods results;
	
%mend fitmodel;

/* sortqtl() re-numbers QTL after sorting by chromosome and position */
%macro sortqtl(trt=, hkmimmodel=);

	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local id_allqtl_tmp;
	%local NOBS_id_allqtl_tmp;
	%local rc;
	%local id_epiqtl_tmp;
	%local NOBS_id_epiqtl_tmp;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitsortq;
	%end;

	*--- create data set that holds the model from the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitsortq;
	%end;
	
	*--- get the number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	%if %sysevalf(&numbqtlin < 1) %then %do;
		%*put WARNING: The current model has no QTL. Therefore, the sortqtl procedure was aborted.;
	%end;
	%else %do;
		
		*--- create a temporary data set that holds QTL in the input model ---*;
		data qtl1_tmp(keep= trait qtl1 ch1 pos1 posid1) qtl2_tmp(keep= trait qtl2 ch2 pos2 posid2);
			set hkmodel_tmp;
			where trait=&trt and nqtl > 0;				
		run;

		proc sort data= qtl1_tmp;
			by qtl1;
		run;

		proc sort data= qtl2_tmp;
			by qtl2;
		run;

		data qtl_tmp1;
			merge qtl1_tmp qtl2_tmp(rename= (qtl2= qtl1 ch2= ch1 pos2= pos1 posid2= posid1));
			by qtl1;
		run;

		data allqtl_tmp;
			set qtl_tmp1;
			by qtl1;
			if first.qtl1 then output;
		run;
		
		proc datasets nolist;
			delete qtl1_tmp qtl2_tmp qtl_tmp1;
		quit;

		*--- sort by chromosome and positon, and create new numbering for QTL ---*;
		proc sort data=allqtl_tmp;
			by ch1 posid1;
		run;

		data allqtl_tmp (rename = (qtl1= QTL));
			set allqtl_tmp;
			updatedqtl = _n_;
		run;
	
		*--- renumber qtl1 and qtl2 in the input data set after sorting by chromosome and positon ---*;
		%let id_allqtl_tmp = %sysfunc(open(allqtl_tmp));
		%let NOBS_id_allqtl_tmp = %sysfunc(attrn(&id_allqtl_tmp, NOBS));
		
		%syscall set(id_allqtl_tmp);
		%do i=1 %to &NOBS_id_allqtl_tmp;
			%let rc = %sysfunc(fetchobs(&id_allqtl_tmp, &i));
			
			data hkmodel_tmp;
				set hkmodel_tmp;
				if qtl1 = &QTL then updatedqtl1 = &updatedqtl;
				if qtl2 = &QTL then updatedqtl2 = &updatedqtl;
			run;
		
		%end;
		*--- very important to close ---*;
		%let id_allqtl_tmp = %sysfunc(close(&id_allqtl_tmp)); 
		
		data hkmodel_tmp(drop= updatedqtl1 updatedqtl2);
			set hkmodel_tmp;
			qtl1 = updatedqtl1;
			qtl2 = updatedqtl2;
		run;
		
		*--- rearrange order of epistatic QTL.  Qtl1 < Qtl2 for all pairs of epistatic QTL. ---*;
		data hkmodel_tmp(drop=qtl1_tmp ch1_tmp pos1_tmp posid1_tmp efftype1_tmp);
			set hkmodel_tmp;
			if qtl1 > qtl2 then do;
				qtl1_tmp = qtl1;
				qtl1 = qtl2;
				qtl2 = qtl1_tmp; 
				
				ch1_tmp = ch1;
				ch1 = ch2;
				ch2 = ch1_tmp;
				
				pos1_tmp = pos1;
				pos1 = pos2;
				pos2 = pos1_tmp;
				
				posid1_tmp = posid1;
				posid1 = posid2;
				posid2 = posid1_tmp;
				
				efftype1_tmp = efftype1;
				efftype1 = efftype2;
				efftype2 = efftype1_tmp;
			end;
		run;

		proc sort data=hkmodel_tmp;
			by indic_epi_eff qtl1 qtl2;
		run;
		
		*--- Update indicator for QTL with main effects only. ---*;
		*--- Needed because interaction effects can be dropped off ---*;
		data epiqtl1_tmp(keep= trait qtl1 ch1 pos1 posid1) epiqtl2_tmp(keep= trait qtl2 ch2 pos2 posid2);
			set hkmodel_tmp;
			where (trait=&trt and nqtl > 0) and (qtl1 ^= qtl2 or inter_cofactor_id ^= .);					
		run;
		
		proc sort data= epiqtl1_tmp;
			by qtl1;
		run;
		
		proc sort data= epiqtl2_tmp;
			by qtl2;
		run;
		
		data epiqtl_tmp1;
			merge epiqtl1_tmp epiqtl2_tmp(rename= (qtl2= qtl1 ch2= ch1 pos2= pos1 posid2= posid1));
			by qtl1;
		run;
		
		data epiqtl_tmp (rename= (qtl1 = QTL));
			set epiqtl_tmp1;
			by qtl1;
			if first.qtl1 then output;
		run;
		
		proc datasets nolist;
			delete epiqtl1_tmp epiqtl2_tmp epiqtl_tmp1;
		quit;
		
		data hkmodel_tmp;
			set hkmodel_tmp;
			length donealready $6;
			donealready = 'no';
			indic_qtl_without_inter = .;
		run;
		
		%*goto exitsortq;
		
		%let id_epiqtl_tmp = %sysfunc(open(epiqtl_tmp));
		%let NOBS_id_epiqtl_tmp = %sysfunc(attrn(&id_epiqtl_tmp, NOBS));
		%put NOBS_id_epiqtl_tmp= &NOBS_id_epiqtl_tmp;
		
		%if (&NOBS_id_epiqtl_tmp > 0) %then %do;
			%syscall set(id_epiqtl_tmp);
			%do i=1 %to &NOBS_id_epiqtl_tmp;
				%let rc = %sysfunc(fetchobs(&id_epiqtl_tmp, &i));
				
				data hkmodel_tmp;
					set hkmodel_tmp;
					if donealready = 'no' then do;
						if (qtl1 ^= . and qtl1 = qtl2) and (qtl1 = &QTL or inter_cofactor_id ^= .) then do;
							indic_qtl_without_inter = 0;
							donealready = 'yes';
						end;
						else if (qtl1 ^= . and qtl1 = qtl2) and (qtl1 ^= &QTL or inter_cofactor_id = .) then do;
							indic_qtl_without_inter = 1;
						end;					
					end;
				run;
			%end;
		%end;
		%else %do;
			data hkmodel_tmp;
				set hkmodel_tmp;
				if qtl1 ^= . then indic_qtl_without_inter = 1;
			run;
		%end;		
		%let id_epiqtl_tmp = %sysfunc(close(&id_epiqtl_tmp)); 
		*--- very important to close ---*;
		
		data hkmodel_tmp;
			set hkmodel_tmp;
			if (qtl1 ^= .) and (qtl1 ^= qtl2 or inter_cofactor_id ^= .) then do;
				indic_qtl_without_inter = .;
			end;
		run;

		*--- update the input model ---*;
		data &hkmimmodel;
			set &hkmimmodel;
			where trait ^= &trt;
		run;
		
		data &hkmimmodel;
			set &hkmimmodel hkmodel_tmp(drop= donealready);
		run;
		
		proc sort data=&hkmimmodel;
			by trait;
		run;
		
		*--- delete temporary data sets ---*;
		proc datasets nolist;
			delete allqtl_tmp
				epiqtl_tmp;
		quit;

	%end;
	
	*--- delete temporary data sets ---*;
	proc datasets nolist;
		delete hkmodel_tmp;
	quit;
	
	%exitsortq:;
	
%mend sortqtl;

%macro prephkdata1(ng=, qtlvars=, talldsn=, mappedgr=, chrom_test=, cmid_test=);
	
	
	%local u;
	%local qtlvar1;
	%local qtle;
	%local qtlc;
	%local qtlp;
	%local qtlstr;
	%local which_n;
	
	*create hkdata1 without positions of qtl and their window sizes*;
	data hkdata1;
		set _null_;
	run;
	
	proc sql;
		create table hkdata1 as
		select a.*
		from &talldsn as a, &mappedgr as b
		where a.chrom = b.chrom and a.cmid=b.cmid;
	quit;		
	
	*loop to add columns of mapped qtl into the hkdata1*;
	%put qtlvars = &qtlvars;
	%let u= 1;
	*I can do qtlvar1= %scan(&qtlvars,&u) because all epistatic QTL also have main effect in the model*;
	%let qtlvar1= %scan(&qtlvars,&u); 
	%do %while(%length(&qtlvar1));
		%put prephkdata1;
		%put u=&u;
				
		%let qtlstr = %sysfunc(tranwrd(&qtlvar1, e, ));
		%let qtlstr = %sysfunc(tranwrd(&qtlstr, c, ));
		%let qtlstr = %sysfunc(tranwrd(&qtlstr, p, ));
		
		%let qtle= %scan(&qtlstr, 1);
		%let qtlc= %scan(&qtlstr, 2);
		%let qtlp= %scan(&qtlstr, 3);
		
		%put qtle=&qtle;
		%put qtlc=&qtlc;
		%put qtlp=&qtlp;
				
		*get variables of each mapped qtl*;
		data qtldata;
			set &talldsn;
			where chrom = &qtlc and cmid = &qtlp;
			keep obs_ %if &qtle = 1 %then xa2_;
				  %else %if &qtle = 2 %then xd_;
				  %else xa1_;
				  ;
			rename obs_=obs_q 
				%if &qtle = 1 %then xa2_= &qtlvar1;
				%else %if &qtle = 2 %then xd_= &qtlvar1; 
				%else xa1_= &qtlvar1;
				;
		run;		
				
		*merge variables of each mappend qtl into the hkdata1*;
		proc sql; 
			create table hkdata_tmp(drop=obs_q) as 
			select a.*, b.* 
			from hkdata1 as a, qtldata as b
			where a.obs_=b.obs_q;
		quit;

		proc sort data=hkdata_tmp out=hkdata1;
			by chrom cm cmid obs_;
		run;
						
		%let u = %sysevalf(&u + 1);
					
		%let qtlvar1= %scan(&qtlvars,&u);
			
	%end;

	%if(%length(&chrom_test) & %length(&cmid_test)) %then %do;
		*get variables of tested position*;
		data testdata;
			set &talldsn;
			where chrom = &chrom_test and cmid = &cmid_test;
			keep obs_ %if &ng = 2 %then xa2_;
				  %else %if &ng = 3 %then xa2_ xd_; 
				  %else xa2_ xd_ xa1_;
			;
			rename obs_=obs_z 
				%if &ng=2 %then xa2_= za2_;
				%else %if &ng = 3 %then xa2_= za2_ xd_= zd_; 
				%else xa2_= za2_ xd_= zd_ xa1_= za1_;
			;
		run;

		*merge variables of tested position into the hkdata1*;
		proc sql; 
			create table hkdata_tmp(drop=obs_z) as 
			select a.*, b.* 
			from hkdata1 as a, testdata as b
			where a.obs_=b.obs_z;
		quit;

		proc sort data=hkdata_tmp;
			by chrom cm cmid obs_;
		run;
		
 		data _null_;
			set hkdata_tmp;
			by chrom cm cmid;
			if first.cmid and chrom = &chrom_test and cmid= &cmid_test then 
				call symputx("which_n", _n_, 'L');
		run;
		%*put which_n= &which_n;

		data hkdata1;
			set hkdata_tmp(firstobs=&which_n);
		run;		
		
		proc datasets nolist;
			delete testdata;
		quit;
	
	%end;

	proc datasets nolist;
		delete hkdata_tmp qtldata;
	quit;
	
	%exitpdata1:;
	
%mend prephkdata1;

%macro prephkdata3(ng=, qtlvars=, talldsn=);

	%local u;
	%local qtlvar1;
	%local qtl2;
	%local qtlc;
	%local qtlp;
	%local qtlstr;
	
	*create hkdata3 with trait and covariates*;
	data hkdata3;
		set _null_;
	run;
	
	data hkdata3;
		set &talldsn;
		if _n_ <= obs_ then output;
		drop chrom cm cmid %if &ng = 2 %then xa2_;
				   %else %if &ng = 3 %then xa2_ xd_; 
				   %else xa2_ xd_ xa1_;
				   ;
	run; 		
	
	%*goto exitpdata3;
	
	*loop to add columns of mapped qtl into the hkdata3*;
	%let u= 1;
	/*I can do qtlvar1= %scan(&qtlvars,&u) because all epistatic QTL also have main effect in the model*/
	%let qtlvar1= %scan(&qtlvars,&u); 
	%do %while(%length(&qtlvar1));
		%put prephkdata3;
		%put u=&u;
	
		%let qtlstr = %sysfunc(tranwrd(&qtlvar1, e, ));
		%let qtlstr = %sysfunc(tranwrd(&qtlstr, c, ));
		%let qtlstr = %sysfunc(tranwrd(&qtlstr, p, ));
		
		%let qtle= %scan(&qtlstr, 1);		
		%let qtlc= %scan(&qtlstr, 2);
		%let qtlp= %scan(&qtlstr, 3);
		
		%put qtle=&qtle;		
		%put qtlc=&qtlc;
		%put qtlp=&qtlp;
				
		*get variables of each mapped qtl*;
		data qtldata;
			set &talldsn;
			where chrom = &qtlc and cmid = &qtlp;
			keep obs_ %if &qtle = 1 %then xa2_;
				  %else %if &qtle = 2 %then xd_; 
				  %else xa1_;
			;
			rename obs_=obs_q 
				%if &qtle = 1 %then xa2_= &qtlvar1;
				%else %if &qtle = 2 %then xd_= &qtlvar1; 
				%else xa1_= &qtlvar1;
			;
		run;		

		*merge variables of each mappend qtl into the hkdata3*;
		proc sql; 
			create table hkdata_tmp(drop=obs_q) as 
			select a.*, b.* 
			from hkdata3 as a, qtldata as b
			where a.obs_=b.obs_q;
		quit;

		proc sort data=hkdata_tmp out=hkdata3;
			by obs_;
		run;
		
		%let u = %sysevalf(&u + 1);

		%let qtlvar1= %scan(&qtlvars,&u);
			
	%end;

	%*goto exitpdata3;

	proc datasets nolist;
		delete hkdata_tmp
			qtldata;
	quit;
	
	%exitpdata3:;
	
%mend prephkdata3;

/* scanmain() scans through the genome grid and adds the best putative QTL into model */
%macro scanmain(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_main=);

	*ods html;
	ods exclude all;
	ods noresults;

	%global flag_warn_sm1 warn_sm1_message;
	%let flag_warn_sm1 =0;
	%let warn_sm1_message=;

	*--- NEVER uses same prefix anywhere throughout the code ---*;	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local sepc;
	%local sepp;	
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local maing1;
	%local maing2;
	%local L_window1; 
	%local R_window1;
	%local L_window2; 
	%local R_window2;
	%local id_mappedgrid_tmp;
	%local NObs_mappedgrid_tmp;	
	%local bestchrom;
	%local bestpos;
	%local bestposid;
	%local rc;
	%local threshold;

	%local teststmt;
	%local id_bestmain;
	%local NObs_bestmain;
	
	%local totalqtl;	
	%local totaleff;
	
	%local orderdirection;
	%local bool;
	%local neg2loglike0;
	%local AIC0;
	%local AICC0;
	%local BIC0;
	%local bestvalue;
	%local bestInfinityLike;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitscm;
	%end;
	
	*---get criterion to select best genomic position---*;
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then %do;
		%let threshold = &lod_threshold;
	%end;
	%else %if (&criterion=AIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=AICC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=BIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=ProbF) %then %do;
		%let threshold = &probf_entry;
	%end;
	
	*--- created temporary data set that holds the model for the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;	

	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	*--- very important to close ---*;
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); 

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitscm;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;	

	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);

	*--- get the number of QTL in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%put # # # # # # # # # # # # # # # # # # # # # # ;
	%put # # # Number of QTL = &numbqtlin # # # # # #;
	%put # # # Number of Effects = &numbeffin # # # #;
	%put # # # # # # # # # # # # # # # # # # # # # # ;
	
	%*put teststep_main= &teststep_main;
	*--- create data sets for grid search and profiling results ---*;
	data MappedGrid_tmp;
		*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;			
		set &mapgrid(rename= (&CHROMNUM= chrom) 
					%if %length(&g_firstobs) %then firstobs= &g_firstobs;
					%if %length(&g_obs) %then obs=&g_obs;
					);
		retain _mkpos_;
		if MarkerIndicator = 1 then do;
			_mkpos_ = cm;
			output;
		end;
		else if (MarkerIndicator = 0 and (cm - _mkpos_ >= &teststep_main)) then do;
			_mkpos_ = cm;
			output;
		end;
	run;
	
	*data OutLib.MappedGrid_tmpMain; *set MappedGrid_tmp; *run;
	
	%*goto exitscm;
	
	*--- get effects of QTL already in the model ---*;
	%let sepc=c;
	%let sepp=p;	
	%let geffect_main_in=;
	%let geffect_epi_in=;
	%let geffect_qxc_in=;
	%if %sysevalf(&numbqtlin > 0) %then %do;
		%let id_hkmodel_tmp=%sysfunc(open(hkmodel_tmp));
		%let NObs_hkmodel_tmp=%sysfunc(attrn(&id_hkmodel_tmp,NOBS));
		
		%syscall set(id_hkmodel_tmp);
		*--- first row of id_hkmodel_tmp is not used ---**;
		%do i=2 %to &NObs_hkmodel_tmp;
			%let rc=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));

			%if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id = .) %then %do;
				*---get main effects of QTL already in the input model ---*;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let geffect_main_in= &geffect_main_in &maing1;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 ^= &qtl2 and &ch1 ^= &ch2 and &posid1 ^= &posid2 and &inter_cofactor_id = .) %then %do; 
				*---get two-way interaction (epistasis) already in the model ---*;  
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1;
				%let maing2= e&efftype2&sepc&ch2&sepp&posid2;
				%let geffect_epi_in= &geffect_epi_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%let L_window2 = %sysevalf(&pos2 - &wsize); 
				%let R_window2 = %sysevalf(&pos2 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				%*put L2=&L_window2;
				%*put R2=&R_window2;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;					
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 & chrom ~= &ch2 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) |
						chrom = &ch2 and (cm <= &L_window2 or cm >= &R_window2) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id ^= .) %then %do;
				*---get two-way interaction of Cofactor x QTL already in the input model ---*;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let maing2= x&inter_cofactor_id;
				%let geffect_qxc_in= &geffect_qxc_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;
		%end;
		*--- very important to close ---*;
		%let id_hkmodel_tmp=%sysfunc(close(&id_hkmodel_tmp)); 
	%end;
	%put geffect_main_in= &geffect_main_in;
	%put geffect_epi_in= &geffect_epi_in;
	%put geffect_qxc_in= &geffect_qxc_in;
	
	%let id_mappedgrid_tmp=%sysfunc(open(MappedGrid_tmp));
	%let NObs_mappedgrid_tmp=%sysfunc(attrn(&id_mappedgrid_tmp,NOBS));
	*--- very important to close ---*;
	%let id_mappedgrid_tmp=%sysfunc(close(&id_mappedgrid_tmp)); 
		
	%if %sysevalf(&NObs_mappedgrid_tmp < 1) %then %do;
	
		%let flag_warn_sm1 = 1;
		%let warn_sm1_message = The search for QTL main effects stopped because all positions in the genome grid are occuppied.;
		%put WARNING: &warn_sm1_message;	

		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp MappedGrid_tmp;
		quit;

		%goto exitscm;
	%end;
	%else %do;
		%prephkdata1(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata, mappedgr=MappedGrid_tmp, chrom_test=, cmid_test=);
								
		%*goto exitscm;

		%if &ng= 2 %then %do;
			%let teststmt= 'contq_' xa2_ 1;
		%end;
		%else %if &ng = 3 %then %do;
			%let teststmt= 'contq_' xa2_ 1, xd_ 1;
		%end;
		%else %if &ng = 4 %then %do;
			%*let teststmt= %str(&teststmt, xa1_ 1); 

			*case2*;
			%let teststmt2= '2' xa1_ 1;
			*case3;
			%let teststmt3= '3' xa2_ 1, xd_  1;
			*case4*;
			%let teststmt4= '4' xd_  1, xa1_ 1; 
			*case3;
			%let teststmt5= '5' xa2_ 1;
			*case6*;
			%let teststmt6= '6' xa2_ 1, xa1_ 1; 
			*case7*;
			%let teststmt7= '7' xa2_ 1, xd_  1; 
			*case8*;
			%let teststmt8= '8' xa2_ 1, xd_  1, xa1_ 1;		
		%end;
		
		%*goto exitscm;
		
		*data OutLib.hkdata1main; *set hkdata1; *run;
		
		ods output contrasts= contrq CovParms=CovP;
		proc mixed data=hkdata1 method=ml;
			by chrom cm cmid;
			model y&trt = &xcovar xa2_ %if &ng >= 3 %then xd_; %if &ng = 4 %then xa1_; &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
			ods output fitstatistics=fs solutionf=sf nobs=nobm;
			%if &ng ^= 4 %then %do;
				contrast &teststmt / chisq;
			%end;
			%else %do;
				contrast &teststmt2 / chisq;
				contrast &teststmt3 / chisq;
				contrast &teststmt4 / chisq;
				contrast &teststmt5 / chisq;
				contrast &teststmt6 / chisq;
				contrast &teststmt7 / chisq;
				contrast &teststmt8 / chisq;
			%end;
		run;
		
		data CovP;
			set CovP;
			where  CovParm = 'Residual';
			if Estimate=0 then InfinityLike= 1;
			else InfinityLike= 0;
		run; 
		
		proc transpose data=fs out=fs1(rename= (
							_2_Log_Likelihood=neg2loglike
							AIC__smaller_is_better_ =AIC
							AICC__smaller_is_better_= AICC
							BIC__smaller_is_better_=BIC
							));
			var value;
			id descr;
			by chrom cm cmid;
		run;		
		
		%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);
		
		*data OutLib.hkdata3main; *set hkdata3; *run;
		
		*ods output CovParms=CovP0;
		proc mixed data=hkdata3 method=ml;
			model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
			ods output fitstatistics=fs0 solutionf=sf0;
		run;	
		
		/*
		data CovP0;
			set CovP0;
			where  CovParm = 'Residual';
			if Estimate=0 then InfinityLike= 1;
			else InfinityLike= 0;
		run; 
		*/

		proc transpose data=fs0 out=fs01(rename= (
							_2_Log_Likelihood=neg2loglike0
							AIC__smaller_is_better_ =AIC0
							AICC__smaller_is_better_= AICC0
							BIC__smaller_is_better_=BIC0
							));
			var value;
			id descr;
		run;
		
		data _null_;
			set fs01;
			call symputx("neg2loglike0", trim(left(neg2loglike0)), 'L');
			call symputx("AIC0", trim(left(AIC0)), 'L');
			call symputx("AICC0", trim(left(AICC0)), 'L');
			call symputx("BIC0", trim(left(BIC0)), 'L');
		run;

		%if %sysevalf(&criterion=AIC or &criterion=AICC or &criterion=BIC) %then %do; 
			%let threshold = &&&criterion.0;		
		%end;
		
		data nobm(keep=chrom cm cmid NObsRead NObsUsed);
			set nobm;
			where label = "Number of Observations Read";
		run;		

		data sf(drop= Effect stderr df tvalue probt);
			set sf(rename = (Estimate = Additive2));
			where Effect = "xa2_";
			%if &ng >= 3 %then %do;
				set sf(rename = (Estimate = Dominance));
				where Effect = "xd_";
			%end;
			%if &ng = 4 %then %do;
				set sf(rename = (Estimate = Additive1));
				where Effect = "xa1_";
			%end;
		run;
		
		%if &ng = 4 %then %do;
			data hkdata1_sub;
				set hkdata1;
				by chrom cmid;
				if first.cmid then output;
			run;
				
			proc sql;
				create table contrq1 as
				select a.*
				from contrq as a, hkdata1_sub as b
				where a.chrom= b.chrom and a.cmid= b.cmid and a.label= b.contrast_;
			quit;
						
			proc datasets nolist;
				delete contrq;
				change contrq1= contrq;
				modify contrq;
				rename label = contrast_;
				delete hkdata1_sub;
			quit;
		%end;		

		data scmain;
			merge sf CovP fs1 nobm contrq;
			*by chrom cm cmid;
			by chrom cmid;
			LOD = (1/(2 * 2.30258509299404))*(&neg2loglike0 - neg2loglike);
			%if "&criterion" = "LOD" %then %do; 
				valuet = LOD;
			%end;
			%else %if "&criterion" = "ProbF" %then %do;
				valuet = probf;
			%end;
			%else %if "&criterion" = "AIC" %then %do;
				valuet = AIC;
			%end;	
			%else %if "&criterion" = "AICC" %then %do;
				valuet = AICC;
			%end;
			%else %if "&criterion" = "BIC" %then %do;
				valuet = BIC;
			%end;
		run;
		
		%*goto exitscm;
		
		*data OutLib.scmain1; *set scmain; *run;
		
		*--- CHROMNUM, MARKERLAB are global variable passed from SingleQTLModel.sas ---*;	
		data scmain(keep = chrom cm cmid 
				AIC BIC AICC LOD NumDF DenDF ProbF
				%if &ng =2 %then Additive2;
				%else %if &ng =3 %then Additive2 Dominance; 
				%else Additive2 Dominance Additive1 contrast_; 
				MarkerIndicator &MARKERLAB NObsRead NObsUsed
				valuet CovParm Estimate InfinityLike); 
			merge &mapgrid(rename=(&CHROMNUM= chrom)) scmain;
			*by chrom cm cmid;
			by chrom cmid;
		run;
		
		*data OutLib.scmain2; *set scmain; *run;
				
		%let orderdirection = ;
		%let bool = < ;
		%if "&criterion" = "LOD" %then %do; 
			%let orderdirection = descending;
			%let bool = > ;
		%end;
		
		data bestmain;
			set scmain;
			where valuet ^=. ;
		run;
		
		proc sort data=bestmain;
			by &orderdirection valuet;
		run;
		
		data _null_;
			set bestmain(obs=1);
			call symputx("bestchrom", chrom, "L");
			call symputx("bestpos", cm, "L");
			call symputx("bestposid", cmid, "L");
			call symputx("bestvalue", valuet, "L");
			%if &ng=4 %then %do;
				call symputx("case_", contrast_, "L");
			%end;
			call symputx("bestInfinityLike", InfinityLike, "L");
		run;
		
		%put ## Scan Main Effects ##;
		%put criterion= &criterion;
		%put threshold= &threshold;
		%put bestvalue= &bestvalue;
		
		data bestmain;
			set bestmain;
			where valuet &bool &threshold;
		run;
			
		%*goto exitscm;

		*---saving best model---*;
		%let id_bestmain=%sysfunc(open(bestmain));
		%let NObs_bestmain=%sysfunc(attrn(&id_bestmain,NOBS));
		%let id_bestmain=%sysfunc(close(&id_bestmain)); *--- very important to close ---*;			
		%if (&NObs_bestmain > 0) %then %do;
			
			%*put bestchrom= &bestchrom;
			%*put bestpos= &bestchrom;
			%*put bestposid= &bestposid;
			
			%let totalqtl= %sysevalf(&numbqtlin + 1);
			
			%if &ng=2 %then %do;
				%let totaleff = %sysevalf(&numbeffin + 1);
				%let case_= .;
			%end;
			%else %if &ng=3 %then %do;
				%let totaleff = %sysevalf(&numbeffin + 2);
				%let case_= .;
			%end;
			%else %if &ng=4 %then %do;
				%if &case_ = 2 %then %let totaleff = %sysevalf(&numbeffin + 1);
				%else %if &case_ = 3 %then %let totaleff = %sysevalf(&numbeffin + 2);
				%else %if &case_ = 4 %then %let totaleff = %sysevalf(&numbeffin + 2);
				%else %if &case_ = 5 %then %let totaleff = %sysevalf(&numbeffin + 1);
				%else %if &case_ = 6 %then %let totaleff = %sysevalf(&numbeffin + 2);
				%else %if &case_ = 7 %then %let totaleff = %sysevalf(&numbeffin + 2);
				%else %if &case_ = 8 %then %let totaleff = %sysevalf(&numbeffin + 3);				
			%end;		
			
			data add2_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl 
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .
				qtl1 &totalqtl 
				qtl2 &totalqtl 
				ch1 &bestchrom
				ch2 &bestchrom
				pos1 &bestpos
				pos2 &bestpos
				posid1 &bestposid
				posid2 &bestposid
				efftype1 1 
				efftype2 1
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;
			
			data dom_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl 
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .			
				qtl1 &totalqtl 
				qtl2 &totalqtl 
				ch1 &bestchrom
				ch2 &bestchrom
				pos1 &bestpos
				pos2 &bestpos
				posid1 &bestposid
				posid2 &bestposid
				efftype1 2 
				efftype2 2
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;
			
			data add1_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl 
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .			
				qtl1 &totalqtl 
				qtl2 &totalqtl 
				ch1 &bestchrom
				ch2 &bestchrom
				pos1 &bestpos
				pos2 &bestpos
				posid1 &bestposid
				posid2 &bestposid
				efftype1 3 
				efftype2 3
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;			
	
			data &hkmimmodel;
				set &hkmimmodel 
					%if %sysevalf(&ng=2) %then add2_tmp; 
					%if %sysevalf(&ng=3) %then add2_tmp dom_tmp; 
					%if %sysevalf(&ng=4) %then %do;
						%if &case_=2 %then add1_tmp;
						%else %if &case_=3 %then add2_tmp dom_tmp;
						%else %if &case_=4 %then add1_tmp dom_tmp;
						%else %if &case_=5 %then add2_tmp;
						%else %if &case_=6 %then add2_tmp add1_tmp;
						%else %if &case_=7 %then add2_tmp dom_tmp;
						%else %if &case_=8 %then add2_tmp dom_tmp add1_tmp;
					%end;
					;
				if trait= &trt and qtl1 ^=. then do;
					nqtl = &totalqtl;
					neffect = &totaleff;
				end;
			run;
			
		%end;	
  
	%end;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp
			MappedGrid_tmp
			hkdata1
			hkdata3
			add2_tmp
			dom_tmp
			add1_tmp;
	quit;	

	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);
	
	%exitscm:;

	ods exclude none;
	ods results;	
	
%mend scanmain;

/* fwdmain() calls function scanmain() consecutively until no more putative QTL 
in the genome grid cannot be added into the MIM model */
%macro fwdmain(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_main= );

	**need to create flag variables for scanmain() here because it is needed in the MultiQTLModel_DrillDowns.jsl, even when scanmain is not used*;
	%global flag_warn_sm1 warn_sm1_message;
	%let flag_warn_sm1 = 0;
	%let warn_sm1_message=;

	%global flag_warn_fm1 warn_fm1_message;
	%let flag_warn_fm1 = 0;
	%let warn_fm1_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbqtlout;
	%local numbeffout;
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitfwdm;
	%end;
				
	*--- get number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;

	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	%put InfinityLikein= &InfinityLikein;
		
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitfwdm;
	%end;
	
	%if &InfinityLikein %then %do;
		%let flag_warn_fm1 = 1;
		%let warn_fm1_message = No more QTL effects can be added into the model because the residual variance is zero.;
		%put WARNING: &warn_fm1_message;
		%goto exitfwdm;
	%end;
	
	*--- search for new putative QTL ---*;
	%scanmain(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, mapgrid= &mapgrid, hkmimmodel= &hkmimmodel, 
		wsize= &wsize, criterion= &criterion, lod_threshold= &lod_threshold, probf_entry= &probf_entry, teststep_main= &teststep_main); 
	
	*data OutLib.hkmimmodelfwdmain; *set &hkmimmodel; *run;
	
	*--- get number of QTL and effects after the scan search ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbqtlout", numbqtl, 'L');
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%put numbeffin = &numbeffin;
	%put numbeffout = &numbeffout;
	%put InfinityLikeout= &InfinityLikeout;
	
	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn_fm1 = 1;
		%let warn_fm1_message = No significant QTL effects were found in the search for main effects.;
		%put WARNING: &warn_fm1_message;
	%end;
	%else %do %while(%sysevalf(&numbeffout > &numbeffin) and &InfinityLikeout=0);
		*--- search for new putative QTL until no more QTL can be found ---*;
		%*put numbeffin = &numbeffin;
		%*put numbeffout = &numbeffout;
		
		%let numbeffin = &numbeffout;
	
		*--- search for new putative QTL ---*;
		%scanmain(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, mapgrid= &mapgrid, hkmimmodel= &hkmimmodel, 
			wsize= &wsize, criterion= &criterion, lod_threshold= &lod_threshold, probf_entry= &probf_entry, teststep_main= &teststep_main); 
				
		*--- get number of QTL and effects after the scan search ---*;
		proc means data=&hkmimmodel noprint;
			where trait=&trt;
			var nqtl neffect InfinityLike;
			output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
		run;
		
		data _null_;
			set nqtlout_tmp;
			call symputx("numbqtlout", numbqtl, 'L');
			call symputx("numbeffout", numbeff, 'L');
			call symputx("InfinityLikeout", InfinityLike, 'L');
		run;
		
		*--- refine model before exiting the while loop ---*;
		%if &InfinityLikeout = 1 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);	
			%let IsRefined= 1;
		%end;
		%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);
			*setting IsRefined to 1 to stop refining QTL position*;
			%let IsRefined= 1;
			*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
			%let numbeffin= %sysevalf(&numbeffin - 1);
		%end;
		%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
			*setting IsRefined to 0 to allow refining QTL position*;
			%let IsRefined= 0;
		%end;
		
	%end; *--- end while loop over new QTL ---*; 

	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;
	
	*--- deleting temporary data sets ---*;
	proc datasets nolist;
		delete nqtlin_tmp nqtlout_tmp;
	quit;	
	
	%exitfwdm:;
	
%mend fwdmain;

/* scanepi0D() scans for two-way interaction among QTL already in the model */
%macro scanepi0D(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_entry=);

	ods exclude all;
	ods noresults;

	%global flag_warn1_s0D warn1_s0D_message;
	%let flag_warn1_s0D=0;
	%let warn1_s0D_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local criterionname;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local sepc;
	%local sepp;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local rc;
	%local maing1;
	%local maing2;
	%local currNeg2LogLike;
	%local currAIC;
	%local currAICC;
	%local currBIC;
	%local currvalue;
	%local bestvalue;
	%local bestqtl1;
	%local bestchrom1;
	%local bestpos1;
	%local bestposid1;
	%local bestepitype1;
	%local bestqtl2;
	%local bestchrom2;
	%local bestpos2;
	%local bestposid2;
	%local bestepitype2;
	%local Neg2LogLikeatbest;
	%local AICatbest;
	%local AICCatbest;
	%local BICatbest;
	%local LODatbest;
	%local r_i;
	%local qtl_1;
	%local ch_1;
	%local pos_1;
	%local posid_1;
	%local efftype_1;
	%local maing_1;
	%local r_j;
	%local qtl_2;
	%local ch_2;
	%local pos_2;
	%local posid_2;
	%local efftype_2;
	%local maing_2;
	%local flag_epi_in_model;
	%local i;
	%local j;
	%local k; 
	%local r_k;
	%local geffect_epi_test;
	%local tmpNeg2LogLike;
	%local tmpAIC;
	%local tmpAICC;
	%local tmpBIC;
	%local tmpvalue;
	%local tmpLOD; 	
	%local threshold;
	%local totalqtl;
	%local totaleff;
	
	%local ProbFatbest;
	%local contrasts;
	%local tmpProbF;	
			
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitsce0D;
	%end;
	
	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/

	*---get criterion to select best genomic position---*;
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then
		%let criterionname = -2 LOG LIKELIHOOD;
	%else %if (&criterion=AIC) %then
		%let criterionname = AIC (SMALLER IS BETTER);
	%else %if (&criterion=AICC) %then
		%let criterionname = AICC (SMALLER IS BETTER);
	%else %if (&criterion=BIC) %then
		%let criterionname = BIC (SMALLER IS BETTER);
	%else %if (&criterion=ProbF) %then
		%let criterionname = PROBF;
	
	*---crete temporary data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitsce0D;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);	
	
	*--- get number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLikein;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%if %sysevalf(&numbqtlin < 2) %then %do;

		%let flag_warn1_s0D = 1;
		%let warn1_s0D_message = The 0-D search procedure for QTL x QTL needs a model with at least 2 QTL.;
		%put WARNING: &warn1_s0D_message;  
			
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitsce0D;	
	%end;

	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	*--- get effects of QTL already in the model ---*;
	%let sepc=c;
	%let sepp=p;
	%let geffect_main_in=;
	%let geffect_epi_in=;
	%let geffect_qxc_in=;
	
	%let id_hkmodel_tmp=%sysfunc(open(hkmodel_tmp));
	%let NObs_hkmodel_tmp=%sysfunc(attrn(&id_hkmodel_tmp,NOBS));
	
	%syscall set(id_hkmodel_tmp);
	%do i=2 %to &NObs_hkmodel_tmp;
		%let rc=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));

		*---get main effect of QTL already in the model ---*;
		%if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id = .) %then %do;
			%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
			%let geffect_main_in= &geffect_main_in &maing1;
		%end;
		%else %if %sysevalf(&qtl1 ^= &qtl2 and &ch1 ^= &ch2 and &posid1 ^= &posid2 and &inter_cofactor_id = .) %then %do;
			*---get two-way interaction (epistasis) already in the model ---*; 
			%let maing1= e&efftype1&sepc&ch1&sepp&posid1;
			%let maing2= e&efftype2&sepc&ch2&sepp&posid2;
			%let geffect_epi_in= &geffect_epi_in &maing1 * &maing2;
		%end;
		%else %if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id ^= .) %then %do;
			*---get two-way interaction of Cofactor x QTL already in the input model ---*;
			%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
			%let maing2= x&inter_cofactor_id;
			%let geffect_qxc_in= &geffect_qxc_in &maing1 * &maing2;
		%end;
	%end;
	%*put geffect_main_in= &geffect_main_in;
	%*put geffect_epi_in= &geffect_epi_in;
	%*put geffect_qxc_in= &geffect_qxc_in;
	
	%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);

	*--- get value of criterion for input model ---*;		
	proc mixed data=hkdata3 method=ml;
		model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
		ods output fitstatistics=curr_st_tmp;
	run;
	
	data _null_;
		set curr_st_tmp;
		if upcase(descr) = "-2 LOG LIKELIHOOD" then 
			call symputx("currNeg2LogLike", trim(left(value)), 'L');
		else if upcase(descr) = "AIC (SMALLER IS BETTER)" then 
			call symputx("currAIC", trim(left(value)), 'L');
		else if upcase(descr) = "AICC (SMALLER IS BETTER)" then 
			call symputx("currAICC", trim(left(value)), 'L');
		else if upcase(descr) = "BIC (SMALLER IS BETTER)" then 
			call symputx("currBIC", trim(left(value)), 'L');
		if upcase(descr) = "&criterionname" then
			call symputx("currvalue", trim(left(value)), 'L');
	run;
	%*put currvalue= &currvalue;
	
	proc datasets nolist;
		delete curr_st_tmp;
	quit;
	
	*--- create data set for profiling results ---*;
	data sce0d sce0d_tmp;
		retain qtl1 qtl2 ch1 ch2 pos1 pos2 posid1 posid2 efftype1 efftype2 LOD AIC AICC BIC ProbF .;
		stop;
	run;

	*--- begin search for two-way interaction among QTL in the model ---*;	
	%let bestvalue = 10000000000;
	%let bestqtl1 = ;
	%let bestchrom1 = ;
	%let bestpos1 = ;
	%let bestposid1= ;
	%let bestepitype1= ;
	%let bestqtl2 = ;
	%let bestchrom2 = ;
	%let bestpos2 = ;
	%let bestposid2= ;
	%let bestepitype2= ;
	%let Neg2LogLikeatbest = ;
	%let AICatbest = ;
	%let AICCatbest = ;
	%let BICatbest = ;
	%let LODatbest = ;
	%let ProbFatbest = ;
			
	*--- qtl1 ---*;
	%do i=2 %to &NObs_hkmodel_tmp; 
		*--- i starts at 2 because the first obsearvation for the null model is nqll=0 --- *;	
		%*put i = &i;
		%let r_i=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));
		*I use this if statement because I assume all epistatic QTL in the model have main effects*;
		%if %sysevalf(&qtl1 = &qtl2 and &inter_cofactor_id = .) %then %do;
		%let qtl_1 = &qtl1;
		%let ch_1 = &ch1;
		%let pos_1 = &pos1;
		%let posid_1 = &posid1;
		%let efftype_1 = &efftype1;
		%let maing_1= e&efftype_1&sepc&ch_1&sepp&posid_1;
		*--- qtl2 ---*;
		%do j=&i %to &NObs_hkmodel_tmp; 
			*--- j starts at 2 because the first obsearvation for the null model is nqll=0 --- *;
			%*put j=&j;
			%let r_j=%sysfunc(fetchobs(&id_hkmodel_tmp,&j));
			*I use this if statement because I assume all epistatic QTL in the model have main effects too*;
			%if %sysevalf(&qtl1 = &qtl2 and &inter_cofactor_id = .) %then %do;

			%let qtl_2 = &qtl2;
			%if %sysevalf(&qtl_1 ^= &qtl_2) %then %do;
				*--- build two-way interaction (epistasis) ---*;
				%let ch_2 = &ch2;
				%let pos_2 = &pos2;
				%let posid_2 = &posid2;
				%let efftype_2= &efftype2;
				%let maing_2= e&efftype_2&sepc&ch_2&sepp&posid_2;
									
				*--- checking if the epistasis of qtl1 and qtl2 is already in the model ---*;
				%let flag_epi_in_model = 0;
				%let k = 2;  
				*--- k starts at 2 because the first obsearvation for the null model is nqll=0 --- *;
				%do %while (%sysevalf(&k <= &NObs_hkmodel_tmp and &flag_epi_in_model = 0));
					%let r_k=%sysfunc(fetchobs(&id_hkmodel_tmp,&k));
					%*put k=&k;
					%if (%sysevalf(&qtl1 = &qtl_1 and &qtl2 = &qtl_2 and &efftype1 = &efftype_1 and &efftype2 = &efftype_2)) %then %do;	
						%let flag_epi_in_model = 1;
					%end; 
					*--- end if epistasis is already in the model ---*;
					%let k = %sysevalf(&k + 1);
				%end; 
				*--- end loop for checking if the epistasis is already in the model ---*;
				
				*--- fitting model with epistasis that is not yet in the model ---*;
				%if %sysevalf(&flag_epi_in_model = 0) %then %do;
				
					%put # # # Scanepi0D: Trait= &trt, QTL1= &qtl_1 type1= &efftype_1, QTL2= &qtl_2 type2= &efftype_2 # # #;
					
					%let geffect_epi_test = &maing_1*&maing_2;
					%put geffect_epi_test = &geffect_epi_test;
					
					proc mixed data=hkdata3 method=ml;
						model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &geffect_epi_test/ solution;
						ods output fitstatistics=test_st_tmp solutionf=test_es_tmp;
					run;
					
					*--- collect estimable epistatic effect entering the growing model ---*;
					data test_es_tmp2;
						set test_es_tmp;
						where (Effect = "&maing_1*&maing_2" or Effect = "&maing_2*&maing_1") and DF ^= . ;
					run;					
					
					%let contrasts = ;
					proc sql;
						select Effect into : contrasts separated by ' 1, '
						from test_es_tmp2;
					quit;
					
					%let tmpvalue = .;	
					%let tmpNeg2LogLike= .;
					%let tmpLOD =.;
					%let tmpAIC =.;
					%let tmpAICC =.;
					%let tmpBIC = .;
					%let tmpProbF= .;

					*--- make sure to test a contrast only if it has at least one epistasis ---*;						
					%if %length(&contrasts) %then %do;
						%let contrasts = %str(contrast 'test_qtl_epi' &contrasts 1;);
						%*put contrasts = &contrasts;
					
						*--- get values of -2 Log Likehihood, AIC, AICC, BIC for the growing model ---*;
						data _null_;
							set test_st_tmp;
							if upcase(descr) = "-2 LOG LIKELIHOOD" then 
								call symputx("tmpNeg2LogLike", trim(left(value)), 'L');
							else if upcase(descr) = "AIC (SMALLER IS BETTER)" then 
								call symputx("tmpAIC", trim(left(value)), 'L');
							else if upcase(descr) = "AICC (SMALLER IS BETTER)" then 
								call symputx("tmpAICC", trim(left(value)), 'L');
							else if upcase(descr) = "BIC (SMALLER IS BETTER)" then 
								call symputx("tmpBIC", trim(left(value)), 'L');
							if upcase(descr) = "&criterionname" then
								call symputx("tmpvalue", trim(left(value)), 'L');
						run;
						%*put tmpvalue= &tmpvalue;
						
						%let tmpLOD = %sysevalf((&currNeg2LogLike - &tmpNeg2LogLike)/(2 * 2.30258509299404));						

						*--- get value of ProbF of the epistatic effect entering the growing model ---*;
						proc mixed data=hkdata3 method=ml;
							model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &geffect_epi_test/ solution;
							ods output contrasts=test_ct_tmp;
							&contrasts;
						run;

						data _null_;
							set test_ct_tmp;
							call symputx("tmpProbF", trim(left(ProbF)), 'L');
							if "&criterionname" = "PROBF" then
								call symputx("tmpvalue", trim(left(ProbF)), 'L');
						run;
						%*put tmpProbF = &tmpProbF;

						*--- delete temporary data sets ---*;
						proc datasets nolist;
							delete test_ct_tmp;
						quit;
	
					%end;
					
					*--- delete temporary data sets ---*;
					proc datasets nolist;
						delete test_st_tmp
							test_es_tmp
							test_es_tmp2;
					quit;
					
					*---saving profiling to data set---*;
					data sce0d_tmp; 
						retain	
						qtl1 &qtl_1 
						qtl2 &qtl_2
						ch1 &ch_1
						ch2 &ch_2
						pos1 &pos_1
						pos2 &pos_2
						posid1 &posid_1
						posid2 &posid_2
						efftype1 &efftype_1
						efftype2 &efftype_2 
						LOD &tmpLOD
						AIC &tmpAIC
						AICC &tmpAICC
						BIC &tmpBIC
						ProbF &tmpProbF;
					run;
					
					data sce0d;
					 	set sce0d sce0d_tmp;
					run;
					
					*---selecting best model---*;
					%*put criterionname= &criterionname;
					%put bestvalue= &bestvalue;
					%put tmpvalue= &tmpvalue;
					%if (%sysevalf(&tmpvalue ^=. and &tmpvalue < &bestvalue)) %then %do;
						%let bestvalue = &tmpvalue;
						%let bestqtl1= &qtl_1;
						%let bestchrom1 = &ch_1;
						%let bestpos1 = &pos_1;
						%let bestposid1 = &posid_1;
						%let bestepitype1 = &efftype_1;
						%let bestqtl2 = &qtl_2;
						%let bestchrom2 = &ch_2;
						%let bestpos2 = &pos_2;
						%let bestposid2 = &posid_2;
						%let bestepitype2 = &efftype_2;
						%let Neg2LogLikeatbest = &tmpNeg2LogLike;
						%let AICatbest = &tmpAIC;
						%let AICCatbest = &tmpAICC;
						%let BICatbest = &tmpBIC;
						%let LODatbest = &tmpLOD;
						%let ProbFatbest = &tmpProbF;
					%end;
				%end; *--- and if flat_epi_in_model ---*;
			%end; *--- end if qtl1 ^= qtl2 ---*;
			%end;
		%end; *--- end do loop for qtl2 ---*;
		%end; 
	%end; *--- end do loop for qtl1 ---*;
	%let id_hkmodel_tmp=%sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;
			
	*---saving best model---*;
	%if %length(&bestpos1) %then %do;
		%if "&criterion" = "LOD" %then %do;
			%let threshold = &lod_threshold;
			%let bestvalue = &LODatbest;
		%end;
		%else %if "&criterion" = "ProbF" %then %do;
			%let threshold = &probf_entry;
		%end;
		%else %do;
			%let threshold = &currvalue;
		%end;
		%*put criterion = &criterion;
		%put ## Scan 0-D Epistasis ##;
		%put criterion = &criterion;
		%put threshold = &threshold;			
		%put bestvalue = &bestvalue;
			
		%if ("&criterion" ^= "LOD" and %sysevalf(&bestvalue < &threshold)) or 
			("&criterion" = "LOD" and %sysevalf(&bestvalue > &threshold)) %then %do;
				
			%let totalqtl= &numbqtlin;
			%let totaleff= %sysevalf(&numbeffin + 1);
			
			data neweff_tmp;
				retain 
				trait &trt
				cofactors "&xcovar"				
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 1
				indic_qtl_without_inter . 				
				qtl1 &bestqtl1 
				qtl2 &bestqtl2
				ch1 &bestchrom1
				ch2 &bestchrom2
				pos1 &bestpos1
				pos2 &bestpos2
				posid1 &bestposid1
				posid2 &bestposid2
				efftype1 &bestepitype1
				efftype2 &bestepitype2
				contrast .;
			run;

			data &hkmimmodel;
				set &hkmimmodel neweff_tmp;
				if trait= &trt and neffect > 0 then do;
					nqtl = &totalqtl;
					neffect = &totaleff;
				end;
			run;

			proc datasets nolist;
				delete neweff_tmp;
			quit;
		%end;
	%end; *---end if for saving best model---*;

	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp 
			sce0d_tmp
			hkdata3;
	quit;
	
	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);
				
	%exitsce0D:;
	
	ods exclude none;
	ods results;
	
%mend scanepi0D;

/* fwdepi0D() calls function scanepi0D() consecutively until no more two-way intaraction between QTL 
alrealdy in the model cannot be added into the MIM model */
%macro fwdepi0D(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_entry=, mapgrid=, wsize= );

	%global flag_warn1_s0D warn1_s0D_message;
	%let flag_warn1_s0D=0;
	%let warn1_s0D_message=;

	%global flag_warn1_f0D warn1_f0D_message;
	%let flag_warn1_f0D=0;
	%let warn1_f0D_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbeffout;
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;

	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitfwd0D;
	%end;
	
	*--- get the number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitfwd0D;
	%end;
	
	%if %sysevalf(&numbqtlin < 2) %then %do;
		%*let flag_warn1_s0D = 1;
		%let flag_warn1_f0D = 1;
		%let warn1_f0D_message= The 0-D search procedure for QTL x QTL needs a model with at least 2 QTL.;
		%put WARNING: &warn1_f0D_message; 
		%goto exitfwd0D;
	%end;
	
	%if &InfinityLikein %then %do;
		%let flag_warn1_f0D = 1;
		%let warn1_f0D_message= No more QTL effects can be added into the model because the residual variance is zero.;
		%put WARNING: &warn1_f0D_message;
		%goto exitfwd0D;
	%end;	
	
	*--- search for two-way interacton among QTL in the model ---*;
	%scanepi0D(trt=&trt, ng=&ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, criterion= &criterion, 
		lod_threshold= &lod_threshold, probf_entry= &probf_entry);
	
	*--- get the number of QTL and effects after the scan search ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%*put numbeffout = &numbeffout;
	%*put numbeffin = &numbeffin;
	
	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn1_f0D = 1;
		%let warn1_f0D_message= No significant QTL x QTL effects were found in the 0-D search for QTL x QTL.;
		%put WARNING: &warn1_f0D_message;
	
	%end;
	%else %do %while(%sysevalf(&numbeffout > &numbeffin) and &InfinityLikeout=0);
		*--- search for new effects until no more effects can be found ---*;
		%let numbeffin = &numbeffout;
	
		*--- search for new effect ---*;
		%scanepi0D(trt=&trt, ng=&ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, criterion= &criterion, 
			lod_threshold= &lod_threshold, probf_entry= &probf_entry);
				
		*--- get the number of QTL and effects after the scan search ---*;
		proc means data=&hkmimmodel noprint;
			where trait=&trt;
			var nqtl neffect InfinityLike;
			output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
		run;
		
		data _null_;
			set nqtlout_tmp;
			call symputx("numbeffout", numbeff, 'L');
			call symputx("InfinityLikeout", InfinityLike, 'L');
		run;

		*--- refine model before exiting the while loop ---*;
		%if &InfinityLikeout = 1 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);	
			%let IsRefined= 1;
		%end;
		%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);
			*setting IsRefined to 1 to stop refining QTL position*;
			%let IsRefined= 1;
			*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
			%let numbeffin= %sysevalf(&numbeffin - 1);
		%end;
		%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
			*setting IsRefined to 0 to allow refining QTL position*;
			%let IsRefined= 0;
		%end;
		
	%end; *--- end while loop over search for new effect ---*;
	
	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;
	
	*--- deleting temporary data sets ---*;
	proc datasets nolist;
		delete nqtlin_tmp 
			nqtlout_tmp;
	quit;	

	%exitfwd0D:;
	
%mend fwdepi0D;

/* scanepi1D() scans for two-way interaction among QTL already in the model and genomic positions */ 
%macro scanepi1D(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_epi1D=);

	ods exclude all;
	ods noresults;	
	*options nosource nonotes;/*--- prevents NOTE from being written down in the LOG window---*/

	%global flag_warn1_s1D warn1_s1D_message;
	%global flag_warn2_s1D warn2_s1D_message;
	%let flag_warn1_s1D=0;
	%let flag_warn2_s1D=0;
	%let warn1_s1D_message=;
	%let warn2_s1D_message=;
		
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;	
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local sepc;
	%local sepp;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local rc;
	%local maing1;
	%local maing2;
	%local L_window1;
	%local R_window1;
	%local L_window2;
	%local R_window2;
	%local id_mappedgrid_tmp;
	%local NObs_mappedgrid_tmp;
	%local bestqtl1;
	%local bestchrom1;
	%local bestpos1;
	%local bestposid1;
	%local bestepitype1;
	%local bestqtl2;
	%local bestchrom2;
	%local bestpos2;
	%local bestposid2;
	%local bestepitype2;					
	%local threshold;
	%local totalqtl;
	%local totaleff;
	
	%local teststmt;
	%local u;
	%local qtlvar;
	%local newqtl1var;
	%local newqtl1;
	%local z;
	%local newepi;
	%local newepi_rec;
	%local start_e;
	%local start_c;
	%local start_p;
	%local start_x;			
	%local bestepitype2_tmp;
	%local id_bestepi1D;
	%local NObs_bestepi1D;
	%local besteff_epi;
	
	%local orderdirection;
	%local bool;
	%local neg2loglike0;
	%local AIC0;
	%local AICC0;
	%local BIC0;
	%local bestvalue;
	%local case_;
	
	%local bestInfinityLike;
		
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitsce1D;
	%end;
	
	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/

	*---get criterion to select best genomic position---*;
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then %do;
		%let threshold = &lod_threshold;
	%end;
	%else %if (&criterion=AIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=AICC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=BIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=ProbF) %then %do;
		%let threshold = &probf_entry;
	%end;
		
	*--- create data set that holds the model from the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitsce1D;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);
	
	*--- get the number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%if %sysevalf(&numbqtlin < 1) %then %do;
	
		%let flag_warn1_s1D = 1;
		%let warn1_s1D_message = The 1-D search procecure for QTL x QTL needs a model with at least 1 QTL.;
		%put WARNING: &warn1_s1D_message; 
			
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitsce1D;	
	%end;
	
	%*goto exitsce1D;
	
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	%*put teststep_epi1D= &teststep_epi1D;
	*--- create data sets for grid search and profiling results ---*;
	data MappedGrid_tmp;
		*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;	
		set &mapgrid(rename= (&CHROMNUM= chrom) 
					%if %length(&g_firstobs) %then firstobs= &g_firstobs;
					%if %length(&g_obs) %then obs=&g_obs;
					);
		retain _mkpos_;
		if MarkerIndicator = 1 then do;
			_mkpos_ = cm;
			output;
		end;
		else if (MarkerIndicator = 0 and (cm - _mkpos_ >= &teststep_epi1D)) then do;
			_mkpos_ = cm;
			output;
		end;
	run;
	
	*data OutLib.MappedGrid_tmp1D; *set MappedGrid_tmp; *run;
	
	%*goto exitsce1D;
	
	*--- get effects already in the model ---*;
	%let sepc=c;
	%let sepp=p;
	%let geffect_main_in=;
	%let geffect_epi_in =;
	%let geffect_qxc_in=;
	%if %sysevalf(&numbqtlin > 0) %then %do;
	
		%let id_hkmodel_tmp=%sysfunc(open(hkmodel_tmp));
		%let NObs_hkmodel_tmp=%sysfunc(attrn(&id_hkmodel_tmp,NOBS));

		%syscall set(id_hkmodel_tmp);
		%do i=2 %to &NObs_hkmodel_tmp;
			%let rc=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));

			*---get main effects of QTL already in the input model ---*;
			%if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id = .) %then %do;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let geffect_main_in= &geffect_main_in &maing1;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 ^= &qtl2 and &ch1 ^= &ch2 and &posid1 ^= &posid2 and &inter_cofactor_id = .) %then %do; 
				*---get two-way interaction (epistasis) already in the model ---*;  	
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1;
				%let maing2= e&efftype2&sepc&ch2&sepp&posid2;
				%let geffect_epi_in= &geffect_epi_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%let L_window2 = %sysevalf(&pos2 - &wsize); 
				%let R_window2 = %sysevalf(&pos2 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				%*put L2=&L_window2;
				%*put R2=&R_window2;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;					
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 & chrom ~= &ch2 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) |
						chrom = &ch2 and (cm <= &L_window2 or cm >= &R_window2) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id ^= .) %then %do;
				*---get two-way interaction of Cofactor x QTL already in the input model ---*;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let maing2= x&inter_cofactor_id;
				%let geffect_qxc_in= &geffect_qxc_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;
		%end;
		%let id_hkmodel_tmp=%sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;
	%end;
	%put geffect_main_in= &geffect_main_in;
	%put geffect_epi_in= &geffect_epi_in;
	%put geffect_qxc_in= &geffect_qxc_in;
	
	%*goto exitsce1D;
	
	%let id_mappedgrid_tmp=%sysfunc(open(MappedGrid_tmp));
	%let NObs_mappedgrid_tmp=%sysfunc(attrn(&id_mappedgrid_tmp,NOBS));
	%let id_mappedgrid_tmp=%sysfunc(close(&id_mappedgrid_tmp)); *--- very important to close ---*;
	%if %sysevalf(&NObs_mappedgrid_tmp < 1) %then %do;
	
		%let flag_warn2_s1D = 1;
		%let warn2_s1D_message = The 1-D search for QTL x QTL stopped because all positions in the genome grid are occuppied.;
		%put WARNING: &warn2_s1D_message;
		
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp MappedGrid_tmp;
		quit;
		
		%goto exitsce1D;
	%end;
	%else %do;
		%put geffect_main_in= &geffect_main_in;
		
		%prephkdata1(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata, mappedgr=MappedGrid_tmp, chrom_test=, cmid_test=);

		*data OutLib.hkdata1epi1D; *set hkdata1; *run;
		
		%*goto exitsce1D;
		
		%if &ng = 2 %then %do;
			%let newqtl1 = xa2_;
		%end;
		%else %if &ng = 3 %then %do;
			%let newqtl1 = xa2_ xd_;	
		%end;
		%else %do; 
			%let newqtl1 = xa2_ xd_ xa1_;	
		%end;
		%put newqtl1= &newqtl1;		
		
		%if %sysevalf(&criterion=AIC or &criterion=AICC or &criterion=BIC) %then %do; 
			%let threshold = &&&criterion.0;		
		%end;
		
		%let orderdirection = ;
		%let bool = < ;
	
		%if "&criterion" = "LOD" %then %do; 
		
			%let orderdirection = descending;
			%let bool = > ;
		
			/*
			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);
	
			*data outlib.hkdata3; *set hkdata3; *run;
	
			proc mixed data=hkdata3 method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
				ods output fitstatistics=fs0 solutionf=sf0;
			run;
			*/
			
			proc mixed data=hkdata1 method=ml;
				by chrom cm cmid;					
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newqtl1 / solution;
				ods output fitstatistics=fs0 solutionf=sf0;
			run;	
	
			proc transpose data=fs0 out=fs01(rename= (
								_2_Log_Likelihood=neg2loglike0
								AIC__smaller_is_better_ =AIC0
								AICC__smaller_is_better_= AICC0
								BIC__smaller_is_better_=BIC0
								));
				var value;
				id descr;
				by chrom cm cmid;
			run;
	
			/*
			data _null_;
				set fs01;
				call symputx("neg2loglike0", trim(left(neg2loglike0)), 'L');
				call symputx("AIC0", trim(left(AIC0)), 'L');
				call symputx("AICC0", trim(left(AICC0)), 'L');
				call symputx("BIC0", trim(left(BIC0)), 'L');
			run;
			*/
					
			*data OutLib.fs01; *set fs01; *run;
		%end; 
								
		%*goto exitsce1D;
				
		data bestepi1D_all;
			set _null_;
		run;
				
		%let u = 1;
		%let qtlvar= %scan(&geffect_main_in,&u); /*I can do this because all epistatic QTL also have main effect in the model*/
		%do %while(%length(&qtlvar));
			%*put u=&u;
				
			%let z = 1;
			%let newqtl1var= %scan(&newqtl1,&z);	
			%do %while(%length(&newqtl1var));
				%*put z=&z;
			
				%let newepi = &newqtl1var.*&qtlvar;
				%let newepi_rec = &qtlvar.*&newqtl1var;
				
				/*joint testing of new main and new epistasis*/
				/*
				%let teststmt= 'contq_' xa2_ 1;
				%if &ng >= 3 %then 
					%let teststmt= %str(&teststmt, xd_ 1); 
				%if &ng = 4 %then 
					%let teststmt= %str(&teststmt, xa1_ 1); 
					
				%let teststmt= %str(&teststmt, &newepi 1); 
				%put teststmt = &teststmt;
				*/				

				/*test only new epistasis*/
				%let teststmt= 'contq_' &newepi 1;
				%put teststmt = &teststmt;
				
				ods output contrasts= contrq CovParms=CovP;
				proc mixed data=hkdata1 method=ml;
					by chrom cm cmid;		
					model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newqtl1 &newepi/ solution; /*main and epistasis*/
					*model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newepi/ solution; /*new epistasis only*/
					ods output fitstatistics=fs solutionf=sf nobs=nobm;
					contrast &teststmt / chisq;
				run;
				
				data CovP;
					set CovP;
					where  CovParm = 'Residual';
					if Estimate=0 then InfinityLike= 1;
					else InfinityLike= 0;
				run; 				
		
				proc transpose data=fs out=fs1(rename= (
									_2_Log_Likelihood=neg2loglike
									AIC__smaller_is_better_ =AIC
									AICC__smaller_is_better_= AICC
									BIC__smaller_is_better_=BIC
									));
					var value;
					id descr;
					by chrom cm cmid;
				run;	
				
				*data OutLib.fs1; *set fs1; *run;
				
				data nobm(keep=chrom cm cmid NObsRead NObsUsed);
					set nobm;
					where label = "Number of Observations Read";
				run;	

				%*goto exitsce1D;
				
				data sf(drop= stderr df tvalue probt);
					length Eff_epi $20;
					set sf(rename=(Estimate = Est_epi Effect= Eff_epi));
					where Eff_epi = "&newepi" or Eff_epi = "&newepi_rec";
				run;
				
				data scepi1D_tmp;
					merge sf CovP fs1 nobm contrq %if "&criterion" = "LOD" %then fs01;;
					*by chrom cm cmid;
					by chrom cmid;
					*LOD = (1/(2 * 2.30258509299404))*(&neg2loglike0 - neg2loglike);
					LOD = .;
					%if "&criterion" = "LOD" %then %do; 
						LOD = (1/(2 * 2.30258509299404))*(neg2loglike0 - neg2loglike);
						valuet = LOD;
					%end;
					%else %if "&criterion" = "ProbF" %then %do;
						valuet = probf;
					%end;
					%else %if "&criterion" = "AIC" %then %do;
						valuet = AIC;
					%end;
					%else %if "&criterion" = "AICC" %then %do;
						valuet = AICC;
					%end;
					%else %if "&criterion" = "BIC" %then %do;
						valuet = BIC;
					%end;	
				run;
					
				*--- CHROMNUM, MARKERLAB are global variable passed from SingleQTLModel.sas ---*;	
				data scepi1D_tmp(keep = chrom cm cmid 
						neg2loglike AIC BIC AICC LOD NumDF DenDF ProbF
						Eff_epi Est_epi
						MarkerIndicator &MARKERLAB NObsRead NObsUsed
						valuet 
						%if &ng = 4 %then contrast_;
						%if "&criterion" = "LOD" %then neg2loglike0 AIC0 BIC0 AICC0;
						); 
					merge &mapgrid(rename=(&CHROMNUM= chrom)) scepi1D_tmp;
					*by chrom cm cmid;
					by chrom cmid;
				run;	
				
				data bestepi1D_ea;
					set scepi1D_tmp;
					where valuet ^=. ;
				run;

				proc sort data=bestepi1D_ea;
					by &orderdirection valuet;
				run;
				
				data bestepi1D_all;
					set bestepi1D_all bestepi1D_ea(obs=1);
				run;

				%let z = %sysevalf(&z + 1);
				%let newqtl1var= %scan(&newqtl1,&z);					
				
				%*goto exitsce1D;
			%end;
			
			%let u= %sysevalf(&u + 1);		
			%let qtlvar= %scan(&&geffect_main_in,&u);		
		%end;
		
		*data OutLib.bestepi1D_all; *set bestepi1D_all; *run;
		
		data bestepi1D;
			set bestepi1D_all;
			where valuet ^=. ;
		run;
		
		proc sort data= bestepi1D;
			by &orderdirection valuet;
		run;
		
		data _null_;
			set bestepi1D(obs=1);
			call symputx("besteff_epi", trim(left(Eff_epi)), "L");
			call symputx("bestchrom2", trim(left(chrom)), "L");
			call symputx("bestpos2", trim(left(cm)), "L");
			call symputx("bestposid2", trim(left(cmid)), "L");
			call symputx("bestvalue", trim(left(valuet)), "L");
			%if &ng=4 %then %do;
				call symputx("case_", contrast_, "L");
			%end;
			call symputx("bestInfinityLike", InfinityLike, "L");
		run;	
		
		%put ## Scan 1-D Epistasis ##;
		%put criterion= &criterion;
		%put threshold= &threshold;
		%put bestvalue= &bestvalue;
		
		%*goto exitsce1D;
		
		data bestepi1D;
			set bestepi1D;
			where valuet &bool &threshold;
		run;
		
		*---saving best model---*;
		%let id_bestepi1D=%sysfunc(open(bestepi1D));
		%let NObs_bestepi1D=%sysfunc(attrn(&id_bestepi1D,NOBS));
		%let id_bestepi1D=%sysfunc(close(&id_bestepi1D)); *--- very important to close ---*;			
		%if (&NObs_bestepi1D > 0) %then %do;

			%let start_e = %sysfunc(find(&besteff_epi, e));
			%let start_c = %sysfunc(find(&besteff_epi, c));
			%let start_p = %sysfunc(find(&besteff_epi, p));
			%let start_x = %sysfunc(find(&besteff_epi, x));
			%*put start_e=&start_e;
			%*put start_c=&start_c;
			%*put start_p=&start_p;
			%*put start_x=&start_x;
			
			%let bestepitype2_tmp = %substr(&besteff_epi, &start_x + 1, 2);
			%if &bestepitype2_tmp = a1 %then 
				%let bestepitype2 = 3;
			%else %if &bestepitype2_tmp = d_ %then 
				%let bestepitype2 = 2;
			%else %if &bestepitype2_tmp = a2 %then 
				%let bestepitype2 = 1;				
			
			%let bestepitype1 = %substr(&besteff_epi, &start_e + 1, &start_c - &start_e - 1);
			%let bestchrom1 = %substr(&besteff_epi, &start_c + 1, &start_p - &start_c - 1);
			%if &start_e ^= 1 %then
				%let bestposid1 = %substr(&besteff_epi, &start_p + 1, %length(&besteff_epi) - &start_p);
			%else 
				%let bestposid1 = %substr(&besteff_epi, &start_p + 1, &start_x - &start_p - 2);
			
			%*put bestepitype1=&bestepitype1;
			%*put bestchrom1=&bestchrom1;
			%*put bestposid1=&bestposid1;
			
			%*goto exitsce1D;
			
			data _null_;
				set &hkmimdata;
				where chrom = &bestchrom1 and cmid = &bestposid1;
				call symputx("bestpos1", cm, "L");
			run;

			%let bestqtl2 = %sysevalf(&numbqtlin + 1);
			
			data _null_;
				set &hkmimmodel;
				where qtl1 = qtl2 and ch1 = &bestchrom1 and posid1 = &bestposid1;
				call symputx("bestqtl1", qtl1, "L");
			run;
		
			*--- saving main effects of the new QTL ---*;
			%let totalqtl= %sysevalf(&numbqtlin + 1);
			
			%if &ng= 2 %then %do;
				%let totaleff = %sysevalf(&numbeffin + 2);
				%let case_ = .;
			%end;
			%else %if &ng=3 %then %do;
				%let totaleff = %sysevalf(&numbeffin + 3);
				%let case_ = .;
			%end;
			%else %if &ng=4 %then %do;
				%if &case_ = 2 %then %let totaleff = %sysevalf(&numbeffin + 2);
				%else %if &case_ = 3 %then %let totaleff = %sysevalf(&numbeffin + 3);
				%else %if &case_ = 4 %then %let totaleff = %sysevalf(&numbeffin + 3);
				%else %if &case_ = 5 %then %let totaleff = %sysevalf(&numbeffin + 2);
				%else %if &case_ = 6 %then %let totaleff = %sysevalf(&numbeffin + 3);
				%else %if &case_ = 7 %then %let totaleff = %sysevalf(&numbeffin + 3);
				%else %if &case_ = 8 %then %let totaleff = %sysevalf(&numbeffin + 4);				
			%end;						
						
			data add2_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl 
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter . 			
				qtl1 &bestqtl2 
				qtl2 &bestqtl2 
				ch1 &bestchrom2
				ch2 &bestchrom2
				pos1 &bestpos2
				pos2 &bestpos2
				posid1 &bestposid2
				posid2 &bestposid2
				efftype1 1 
				efftype2 1
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;
				
			data dom_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl 
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter . 				
				qtl1 &bestqtl2 
				qtl2 &bestqtl2 
				ch1 &bestchrom2
				ch2 &bestchrom2
				pos1 &bestpos2
				pos2 &bestpos2
				posid1 &bestposid2
				posid2 &bestposid2
				efftype1 2 
				efftype2 2
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;
	
			data add1_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl 
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .				
				qtl1 &bestqtl2 
				qtl2 &bestqtl2 
				ch1 &bestchrom2
				ch2 &bestchrom2
				pos1 &bestpos2
				pos2 &bestpos2
				posid1 &bestposid2
				posid2 &bestposid2
				efftype1 3 
				efftype2 3
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;
	
			*--- saving new two-way interaction effect ---*;
			data epi_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl 
				neffect &totaleff
				indic_epi_eff 1
				indic_qtl_without_inter .				
				qtl1 &bestqtl1 
				qtl2 &bestqtl2 
				ch1 &bestchrom1
				ch2 &bestchrom2
				pos1 &bestpos1
				pos2 &bestpos2
				posid1 &bestposid1
				posid2 &bestposid2
				efftype1 &bestepitype1
				efftype2 &bestepitype2
				contrast .
				InfinityLike &bestInfinityLike;
			run;				
	
			data &hkmimmodel;
				set &hkmimmodel 
					%if %sysevalf(&ng=2) %then add2_tmp; 
					%if %sysevalf(&ng=3) %then add2_tmp dom_tmp; 
					%if %sysevalf(&ng=4) %then %do;
						%if &case_ = 2 %then add1_tmp; 
						%else %if &case_ = 3 %then add2_tmp dom_tmp;  
						%else %if &case_ = 4 %then add1_tmp dom_tmp;  
						%else %if &case_ = 5 %then add2_tmp;  
						%else %if &case_ = 6 %then add2_tmp add1_tmp;   
						%else %if &case_ = 7 %then add2_tmp dom_tmp;   
						%else %if &case_ = 8 %then add2_tmp dom_tmp add1_tmp;   
					%end;
					epi_tmp;
				if trait= &trt and neffect > 0 then do;
					nqtl = &totalqtl;
					neffect = &totaleff;
				end;
			run;

			proc datasets nolist;
				delete add2_tmp 
					dom_tmp
					add1_tmp
					epi_tmp; 
			quit;
			
		%end; *---end save best model---*;	
		
	%end;
	
	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp
			MappedGrid_tmp
			hkdata1
			hkdata3;
	quit;
	
	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel); 

	%exitsce1D:;

	ods exclude none;
	ods results;
	*options source notes;

%mend scanepi1D;

/* fwdepi1D() calls function scanepi1D() consecutively until no more two-way intaraction between QTL 
alrealdy in the model and putative QTL in the genome grid cannot be added into the MIM model */
%macro fwdepi1D(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_epi1D=);
	
	%global flag_warn1_s1D warn1_s1D_message;
	%let flag_warn1_s1D=0;
	%let warn1_s1D_message=;
	
	%global flag_warn2_s1D warn2_s1D_message;
	%let flag_warn2_s1D=0;
	%let warn2_s1D_message=;
	
	%global flag_warn1_f1D warn1_f1D_message;	
	%let flag_warn1_f1D=0;	
	%let warn1_f1D_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbeffout;
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitfwd1D;
	%end;
	
	*--- get number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitfwd1D;
	%end;
	
	%if %sysevalf(&numbqtlin < 1) %then %do;
		%let flag_warn1_f1D = 1;
		%let warn1_f1D_message = The 1-D search procecure for QTL x QTL needs a model with at least 1 QTL.;
		%put WARNING: &warn1_f1D_message;
		%goto exitfwd1D;
	%end;
	
	%if &InfinityLikein %then %do;
		%let flag_warn1_f1D = 1;
		%let warn1_f1D_message = No more QTL effects can be added into the model because the residual variance is zero.;
		%put WARNING: &warn1_f1D_message;
		%goto exitfwd1D;
	%end;
	
	*--- search for two-way interaction among QTL already in the model and putative QTL in the genome grid---*;
	%scanepi1D(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, mapgrid= &mapgrid, hkmimmodel= &hkmimmodel, 
		wsize= &wsize, criterion= &criterion, lod_threshold= &lod_threshold, probf_entry= &probf_entry,
		teststep_epi1D= &teststep_epi1D);

	*--- get number of QTL and effects after the scan search ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%*put numbeffout = &numbeffout;
	%*put numbeffin = &numbeffin;
	
	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn1_f1D = 1;
		%let warn1_f1D_message = No significant QTL x QTL effects were found in the 1-D search for QTL x QTL.;
		%put WARNING: &warn1_f1D_message;
	%end;
	%else %do %while(%sysevalf(&numbeffout > &numbeffin) and &InfinityLikeout=0);
		*--- search for new putative QTL until no more QTL can be found ---*;
		%let numbeffin = &numbeffout;
		
		*--- search for another putative QTL ---*;
		%scanepi1D(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, mapgrid= &mapgrid, hkmimmodel= &hkmimmodel, 
			wsize= &wsize, criterion= &criterion, lod_threshold= &lod_threshold, probf_entry= &probf_entry,
			teststep_epi1D= &teststep_epi1D); 
				
		*--- get number of QTL after the scan search ---*;
		proc means data=&hkmimmodel noprint;
			where trait=&trt;
			var nqtl neffect InfinityLike;
			output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
		run;
		
		data _null_;
			set nqtlout_tmp;
			call symputx("numbeffout", numbeff, 'L');
			call symputx("InfinityLikeout", InfinityLike, 'L');
		run;
		
		*--- refine model before exiting the while loop ---*;
		%if &InfinityLikeout = 1 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);	
			%let IsRefined= 1;
		%end;
		%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);
			*setting IsRefined to 1 to stop refining QTL position*;
			%let IsRefined= 1;
			*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
			%let numbeffin= %sysevalf(&numbeffin - 1);
		%end;
		%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
			*setting IsRefined to 0 to allow refining QTL position*;
			%let IsRefined= 0;
		%end;
		
	%end; *--- end while loop over search for new putative QTL ---*;
	
	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;
	
	*--- deleting temporary data sets ---*;
	proc datasets nolist;
		delete nqtlin_tmp nqtlout_tmp;
	quit;	
	
	*scanepi1D() includes in the model only the best interaction effect (e.g., a*a, or a*d, etc.) for a pair of qtls**;
	*To search for other interaction effects among pairs of qtl, we need to call fwdepi0D()*;	
	%if &flag_warn1_f1D = 0 %then %do;
		%fwdepi0D(trt=&trt, ng=&ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
			criterion= &criterion, lod_threshold= &lod_threshold, 
			probf_entry= &probf_entry, mapgrid= &mapgrid, wsize= &wsize);		
	%end;
	%exitfwd1D:;
	
%mend fwdepi1D;

/* scanepi2D() scans for two-way interaction between genomic positions */ 
%macro scanepi2D(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_epi2D=);

	ods exclude all;
	ods noresults;
	*options nosource nonotes;/*--- prevents NOTE from being written down in the LOG window---*/

	%global flag_warn1_s2D warn1_s2D_message;
	%let flag_warn1_s2D=0;	
	%let warn1_s2D_message=;
		
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local sepc;
	%local sepp;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local rc;
	%local maing1;
	%local maing2;
	%local L_window1;
	%local R_window1;
	%local L_window2;
	%local R_window2;
	%local id_mappedgrid_tmp;
	%local NObs_mappedgrid_tmp;
	%local bestqtl1;
	%local bestchrom1;
	%local bestpos1;
	%local bestposid1;
	%local bestepitype1;
	%local bestqtl2;
	%local bestchrom2;
	%local bestpos2;
	%local bestposid2;
	%local bestepitype2;
	%local threshold;	
	%local newqtl1;
	%local newqtl2;
	%local z;
	%local j;
	%local Cmidj;
	%local Chroj;
	%local Cmj;
	%local u;
	%local newqtl1var;
	%local newqtl2var;
	%local newepi;
	%local newepi_rec;
	%local teststmt;
	%local id_bestepi2D;
	%local NObs_bestepi2D;
	%local totalqtl;
	%local totaleff;	
	%local start_z;
	%local start_x;
	%local bestepitype1_tmp;
	%local bestepitype2_tmp;
	%local besteff_epi;
	
	%local ChroTest2D;
	%local CmidTest2D;
	%local CmTest2D;

	%local orderdirection;
	%local bool;
	%local neg2loglike0;
	%local AIC0;
	%local AICC0;
	%local BIC0;

	%local id_hkdata1;
	%local NObs_hkdata1;
	%local bestvalue;
	%local case_;
	%local case_1;
	%local ContrastTest2D;
	%local Contrastj;
	%local totaleff_q1;
	%local totaleff_q2;
	
	%local bestInfinityLike;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitsce2D;
	%end;
	
	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/

	*---get criterion to select best genomic position---*;
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then %do;
		%let threshold = &lod_threshold;
	%end;
	%else %if (&criterion=AIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=AICC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=BIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=ProbF) %then %do;
		%let threshold = &probf_entry;
	%end;

	*--- create data set that holds the model from the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitsce2D;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;	
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);
	
	*--- get the number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	%*put teststep_epi2D= &teststep_epi2D;
	*--- create data sets for grid search and profiling results ---*;
	data MappedGrid_tmp;
		*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;	
		set &mapgrid(rename= (&CHROMNUM= chrom) 
					%if %length(&g_firstobs) %then firstobs= &g_firstobs;
					%if %length(&g_obs) %then obs=&g_obs;
					);
		retain _mkpos_;
		if MarkerIndicator = 1 then do;
			_mkpos_ = cm;
			output;
		end;
		else if (MarkerIndicator = 0 and (cm - _mkpos_ >= &teststep_epi2D)) then do;
			_mkpos_ = cm;
			output;
		end;
	run;
	
	%*goto exitsce2D;

	*--- get effects already in the model ---*;
	%let sepc=c;
	%let sepp=p;
	%let geffect_main_in=;
	%let geffect_epi_in =;
	%let geffect_qxc_in=;
	%if %sysevalf(&numbqtlin > 0) %then %do;
		
		%let id_hkmodel_tmp=%sysfunc(open(hkmodel_tmp));
		%let NObs_hkmodel_tmp=%sysfunc(attrn(&id_hkmodel_tmp,NOBS));
		
		%syscall set(id_hkmodel_tmp);
		%do i=2 %to &NObs_hkmodel_tmp;
			%let rc=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));

			*---get main effects of QTL already in the input model ---*;
			%if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id = .) %then %do;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let geffect_main_in= &geffect_main_in &maing1;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 ^= &qtl2 and &ch1 ^= &ch2 and &posid1 ^= &posid2 and &inter_cofactor_id = .) %then %do; 
				*---get two-way interaction (epistasis) already in the model ---*;  	
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1;
				%let maing2= e&efftype2&sepc&ch2&sepp&posid2;
				%let geffect_epi_in= &geffect_epi_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%let L_window2 = %sysevalf(&pos2 - &wsize); 
				%let R_window2 = %sysevalf(&pos2 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				%*put L2=&L_window2;
				%*put R2=&R_window2;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;					
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 & chrom ~= &ch2 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) |
						chrom = &ch2 and (cm <= &L_window2 or cm >= &R_window2) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id ^= .) %then %do;
				*---get two-way interaction of Cofactor x QTL already in the input model ---*;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let maing2= x&inter_cofactor_id;
				%let geffect_qxc_in= &geffect_qxc_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;			
		%end;
		%let id_hkmodel_tmp=%sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;
	%end;
	%*put geffect_main_in= &geffect_main_in;
	%*put geffect_epi_in= &geffect_epi_in;
	%*put geffect_qxc_in= &geffect_qxc_in;
	
	%let id_mappedgrid_tmp=%sysfunc(open(MappedGrid_tmp));
	%let NObs_mappedgrid_tmp=%sysfunc(attrn(&id_mappedgrid_tmp,NOBS));
	%let id_mappedgrid_tmp=%sysfunc(close(&id_mappedgrid_tmp)); *--- very important to close ---*;

	%if %sysevalf(&NObs_mappedgrid_tmp < 2) %then %do;
	
		%let flag_warn1_s2D = 1;
		%let warn1_s2D_message = The 2-D search for QTL x QTL stopped because all positions in the genome grid are occuppied.;
		%put WARNING: &warn1_s2D_message;
		
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp MappedGrid_tmp;
		quit;
		
		%goto exitsce2D;
	%end;
	%else %do;		
		%put geffect_main_in=&geffect_main_in;
		
		%*goto exitsce2D;
		
		%if &ng = 2 %then %do;
			%let newqtl1 = xa2_;
			%let newqtl2 = za2_;
		%end;
		%else %if &ng = 3 %then %do;
			%let newqtl1 = xa2_ xd_;
			%let newqtl2 = za2_ zd_;
		%end;
		%else %do;
			%let newqtl1 = xa2_ xd_ xa1_;
			%let newqtl2 = za2_ zd_ za1_;
		%end;		
		%put newqtl1= &newqtl1;
		%put newqtl2= &newqtl2;		
		
		%if %sysevalf(&criterion=AIC or &criterion=AICC or &criterion=BIC) %then %do; 
			%let threshold = &&&criterion.0;		
		%end;		
		
		%let orderdirection = ;
		%let bool = < ;
		%if "&criterion" = "LOD" %then %do; 
		
			%let orderdirection = descending;
			%let bool = > ;
		
			
			/*
			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);
	
			proc mixed data=hkdata3 method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
				ods output fitstatistics=fs0 solutionf=sf0;
			run;
			
			proc transpose data=fs0 out=fs01(rename= (
								_2_Log_Likelihood=neg2loglike0
								AIC__smaller_is_better_ =AIC0
								AICC__smaller_is_better_= AICC0
								BIC__smaller_is_better_=BIC0
								));
				var value;
				id descr;
			run;
			
			*data OutLib.fs01_2d; *set fs01; *run;
	
			data _null_;
				set fs01;
				call symputx("neg2loglike0", trim(left(neg2loglike0)), 'L');
				call symputx("AIC0", trim(left(AIC0)), 'L');
				call symputx("AICC0", trim(left(AICC0)), 'L');
				call symputx("BIC0", trim(left(BIC0)), 'L');
			run;
			*/
			

		%end;
		
		proc sql;
			select chrom into: ChroTest2D separated by ' '
			from MappedGrid_tmp;
			select cmid into: CmidTest2D separated by ' '
			from MappedGrid_tmp;
			select cm into: CmTest2D separated by ' '
			from MappedGrid_tmp;	
			%if &ng= 4 %then %do;
				select contrast_ into: ContrastTest2D separated by ' '
				from MappedGrid_tmp;
			%end;
		quit;
				
		%*put ChroTest2D=&ChroTest2D;
		%*put CmidTest2D=&CmidTest2D;
		%*put CmTest2D= &CmTest2D;
		
		data bestepi2D_all;
			set _null_;
		run;
		
		%let j = 1;
		%let Cmidj= %scan(&CmidTest2D,&j, ' ');
		%let Chroj= %scan(&ChroTest2D,&j, ' ');
		%let Cmj= %scan(&CmTest2D,&j, ' ');
		%if &ng = 4 %then %do;
			%let Contrastj= %scan(&ContrastTest2D,&j, ' ');
		%end;
		%do %while(%length(&Cmidj));
		
			%*put Chroj= &Chroj;
			%*put Cmidj= &Cmidj;
			%*put Cmj= &Cmj;

			%prephkdata1(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata, mappedgr=MappedGrid_tmp, chrom_test=&Chroj, cmid_test=&Cmidj);

			*apply window size*;
			data hkdata1;
				set hkdata1;
				if chrom= &Chroj and cm <= &Cmj + &wsize then delete;
			run;
								
			%*goto exitsce2D;
			
			%let id_hkdata1=%sysfunc(open(hkdata1));
			%let NObs_hkdata1=%sysfunc(attrn(&id_hkdata1,NOBS));
			%let id_hkdata1=%sysfunc(close(&id_hkdata1)); *--- very important to close ---*;

			%*goto exitsce2D;

			%if %sysevalf(&NObs_hkdata1 > 1) %then %do;
			
				%if "&criterion" = "LOD" %then %do; 
					proc mixed data=hkdata1 method=ml;
						by chrom cm cmid;
						model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newqtl1 &newqtl2 / solution;
						ods output fitstatistics=fs0 solutionf=sf0;
					run;
					
					proc transpose data=fs0 out=fs01(rename= (
									_2_Log_Likelihood=neg2loglike0
									AIC__smaller_is_better_ =AIC0
									AICC__smaller_is_better_= AICC0
									BIC__smaller_is_better_=BIC0
									));
						var value;
						id descr;
						by chrom cm cmid;
					run;						
				%end;			

				%let u = 1;
				%let newqtl1var= %scan(&newqtl1,&u);
				%do %while(%length(&newqtl1var));
					%*put u=&u;

					%let z = 1;
					%let newqtl2var= %scan(&newqtl2,&z);
					%do %while(%length(&newqtl2var));
						%*put z=&z;

						%let newepi = &newqtl1var.*&newqtl2var;
						%let newepi_rec = &newqtl2var.*&newqtl1var;

						/*joint testing of new main and new epistasis*/
						/*
						%let teststmt= 'contq_' xa2_ 1, za1_ 1;
						%if &ng >= 3 %then 
							%let teststmt= %str(&teststmt, xd_ 1, zd_ 1);
						%if &ng = 4 %then 
							%let teststmt= %str(&teststmt, xa1_ 1, za2_ 1);

						%let teststmt= %str(&teststmt, &newepi 1); 
						%put teststmt = &teststmt;
						*/				

						/*test only new epistasis*/
						%let teststmt= 'contq_' &newepi 1;
						%put teststmt = &teststmt;

						ods output contrasts= contrq CovParms=CovP;
						proc mixed data=hkdata1 method=ml;
							by chrom cm cmid;
							model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newqtl1 &newqtl2 &newepi/ solution;
							*model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newepi/ solution; /*new epistasis only*/
							ods output fitstatistics=fs solutionf=sf nobs=nobm;
							contrast &teststmt / chisq;
						run;
						
						data CovP;
							set CovP;
							where  CovParm = 'Residual';
							if Estimate=0 then InfinityLike= 1;
							else InfinityLike= 0;
						run; 						

						proc transpose data=fs out=fs1(rename= (
											_2_Log_Likelihood=neg2loglike
											AIC__smaller_is_better_ =AIC
											AICC__smaller_is_better_= AICC
											BIC__smaller_is_better_=BIC
											));
							var value;
							id descr;
							by chrom cm cmid;
						run;				

						data nobm(keep=chrom cm cmid NObsRead NObsUsed);
							set nobm;
							where label = "Number of Observations Read";
						run;	

						%*goto exitsce2D;

						data sf(drop= stderr df tvalue probt);
							length Eff_epi $20;
							set sf(rename=(Estimate = Est_epi Effect= Eff_epi));
							where Eff_epi = "&newepi" or Eff_epi = "&newepi_rec";
						run;

						%put neg2loglike0= &neg2loglike0;
						data scepi2D_tmp;
							merge sf CovP fs1 nobm contrq %if "&criterion" = "LOD" %then fs01;;
							*by chrom cm cmid;
							by chrom cmid;
							*LOD = (1/(2 * 2.30258509299404))*(&neg2loglike0 - neg2loglike);
							LOD = .;
							%if "&criterion" = "LOD" %then %do; 
								LOD = (1/(2 * 2.30258509299404))*(neg2loglike0 - neg2loglike);
								valuet = LOD;
							%end;
							%else %if "&criterion" = "ProbF" %then %do;
								valuet = probf;
							%end;	
							%else %if "&criterion" = "AIC" %then %do;
								valuet = AIC;
							%end;	
							%else %if "&criterion" = "AICC" %then %do;
								valuet = AICC;
							%end;	
							%else %if "&criterion" = "BIC" %then %do;
								valuet = BIC;
							%end;	
							retain chrom1 cm1 cmid1 %if &ng= 4 %then contrast_1;;
							chrom1=&Chroj;
							cm1=&Cmj;
							cmid1=&Cmidj;
							%if &ng= 4 %then contrast_1= &Contrastj;;
							
						run;	

						*--- CHROMNUM, MARKERLAB are global variable passed from SingleQTLModel.sas ---*;	
						data scepi2D_tmp(keep = chrom cm cmid 
								neg2loglike AIC BIC AICC LOD NumDF DenDF ProbF
								Eff_epi Est_epi
								MarkerIndicator &MARKERLAB NObsRead NObsUsed
								valuet chrom1 cm1 cmid1 
								%if &ng = 4 %then contrast_ contrast_1;
								%if "&criterion" = "LOD" %then neg2loglike0 AIC0 BIC0 AICC0;
								); 
							merge &mapgrid(rename=(&CHROMNUM= chrom)) scepi2D_tmp;
							*by chrom cm cmid;
							by chrom cmid;
						run;		

						data bestepi2D_ea;
							set scepi2D_tmp;
							where valuet ^=. ;
						run;

						proc sort data=bestepi2D_ea;
							by &orderdirection valuet;
						run;

						data bestepi2D_all;
							set bestepi2D_all bestepi2D_ea(obs=1);
						run;

						%let z = %sysevalf(&z + 1);
						%let newqtl2var = %scan(&newqtl2,&z);

						%*goto exitsce2D;

					%end;

					%let u= %sysevalf(&u + 1);		
					%let newqtl1var= %scan(&newqtl1,&u);
				%end;
			%end;
			
			%let j= %sysevalf(&j + 1);		
			%let Cmidj= %scan(&CmidTest2D,&j, ' ');
			%let Chroj= %scan(&ChroTest2D,&j, ' ');
			%let Cmj= %scan(&CmTest2D,&j, ' ');
			%if &ng = 4 %then
				%let Contrastj= %scan(&ContrastTest2D,&j, ' ');
			
		%end;
		
		*data OutLib.bestepi2D_all; *set bestepi2D_all; *run;
		
		data bestepi2D;
			set bestepi2D_all;
			where valuet ^=. ;
		run;
		
		proc sort data=bestepi2D;
			by &orderdirection valuet;
		run;	
		
		data _null_;
			set bestepi2D(obs=1);
			call symputx("besteff_epi", trim(left(Eff_epi)), "L");
			call symputx("bestchrom2", trim(left(chrom)), "L");
			call symputx("bestpos2", trim(left(cm)), "L");
			call symputx("bestposid2", trim(left(cmid)), "L");
			call symputx("bestchrom1", trim(left(chrom1)), "L");
			call symputx("bestpos1", trim(left(cm1)), "L");
			call symputx("bestposid1", trim(left(cmid1)), "L");
			call symputx("bestvalue", trim(left(valuet)), "L");
			%if &ng=4 %then %do;
				call symputx("case_", contrast_, "L");
				call symputx("case_1", contrast_1, "L");
			%end;
			call symputx("bestInfinityLike", InfinityLike, "L");
		run;
		
		%put ## Scan 2-D Epistasis ##;
		%put criterion= &criterion;
		%put threshold= &threshold;
		%put bestvalue= &bestvalue;		

		data bestepi2D;
			set bestepi2D;
			where valuet &bool &threshold;
		run;	
		
		%*goto exitsce2D;

		*---saving best model---*;
		%let id_bestepi2D=%sysfunc(open(bestepi2D));
		%let NObs_bestepi2D=%sysfunc(attrn(&id_bestepi2D,NOBS));
		%let id_bestepi2D=%sysfunc(close(&id_bestepi2D)); *--- very important to close ---*;			
		%if (&NObs_bestepi2D > 0) %then %do;
			
			%let start_z = %sysfunc(find(&besteff_epi, z));
			%let start_x = %sysfunc(find(&besteff_epi, x));
			%*put start_z=&start_z;
			%*put start_x=&start_x;
			
			%let bestepitype1_tmp = %substr(&besteff_epi, &start_z + 1, 2);
			%if &bestepitype1_tmp = a1 %then 
				%let bestepitype1= 3;
			%else %if &bestepitype1_tmp = d_ %then 
				%let bestepitype1= 2;
			%else %if &bestepitype1_tmp = a2 %then 
				%let bestepitype1= 1;

			%*put bestepitype1=&bestepitype1;
			%*put bestchrom1=&bestchrom1;
			%*put bestposid1=&bestposid1;
			
			%let bestepitype2_tmp = %substr(&besteff_epi, &start_x + 1, 2);
			%if &bestepitype2_tmp = a1 %then 
				%let bestepitype2 = 3;
			%else %if &bestepitype2_tmp = d_ %then 
				%let bestepitype2 = 2;
			%else %if &bestepitype2_tmp = a2 %then 
				%let bestepitype2 = 1;

			%*put bestepitype2=&bestepitype2;
			%*put bestchrom2=&bestchrom2;
			%*put bestposid2=&bestposid2;
			
			%*goto exitsce1D;			
			
			*--- saving main effects of the new QTL ---*;
			%let totalqtl = %sysevalf(&numbqtlin + 2);
			
			%if &ng= 2 %then %do;
				%let totaleff = %sysevalf(&numbeffin + 3);
				%let case_ = .;
				%let case_1= .;
			%end;
			%else %if &ng=3 %then %do;
				%let totaleff = %sysevalf(&numbeffin + 5);
				%let case_ = .;
				%let case_1 = .;
			%end;
			%else %if &ng=4 %then %do;
				%if &case_1 = 2 %then %let totaleff_q1 = 1;
				%else %if &case_1 = 3 %then %let totaleff_q1 = 2;
				%else %if &case_1 = 4 %then %let totaleff_q1 = 2;
				%else %if &case_1 = 5 %then %let totaleff_q1 = 1;
				%else %if &case_1 = 6 %then %let totaleff_q1 = 2;
				%else %if &case_1 = 7 %then %let totaleff_q1 = 2;
				%else %if &case_1 = 8 %then %let totaleff_q1 = 3;
			
				%if &case_ = 2 %then %let totaleff_q2 = 1;
				%else %if &case_ = 3 %then %let totaleff_q2 = 2;
				%else %if &case_ = 4 %then %let totaleff_q2 = 2;
				%else %if &case_ = 5 %then %let totaleff_q2 = 1;
				%else %if &case_ = 6 %then %let totaleff_q2 = 2;
				%else %if &case_ = 7 %then %let totaleff_q2 = 2;
				%else %if &case_ = 8 %then %let totaleff_q2 = 3;
				
				%let totaleff = %sysevalf(&numbeffin + &totaleff_q1 + &totaleff_q2 + 1);
			%end;
			
			%let bestqtl1= %sysevalf(&numbqtlin + 1);
			%let bestqtl2= %sysevalf(&numbqtlin + 2);			
			
			data add2_q1;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .				
				qtl1 &bestqtl1 
				qtl2 &bestqtl1 
				ch1 &bestchrom1
				ch2 &bestchrom1
				pos1 &bestpos1
				pos2 &bestpos1
				posid1 &bestposid1
				posid2 &bestposid1
				efftype1 1 
				efftype2 1
				contrast &case_1
				InfinityLike &bestInfinityLike;
			run;
			
			data dom_q1;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter . 				
				qtl1 &bestqtl1 
				qtl2 &bestqtl1 
				ch1 &bestchrom1
				ch2 &bestchrom1
				pos1 &bestpos1
				pos2 &bestpos1
				posid1 &bestposid1
				posid2 &bestposid1
				efftype1 2 
				efftype2 2
				contrast &case_1
				InfinityLike &bestInfinityLike;
			run;
			
			data add1_q1;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .				
				qtl1 &bestqtl1 
				qtl2 &bestqtl1 
				ch1 &bestchrom1
				ch2 &bestchrom1
				pos1 &bestpos1
				pos2 &bestpos1
				posid1 &bestposid1
				posid2 &bestposid1
				efftype1 3 
				efftype2 3
				contrast &case_1
				InfinityLike &bestInfinityLike;
			run;
			
			data add2_q2;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .				
				qtl1 &bestqtl2 
				qtl2 &bestqtl2 
				ch1 &bestchrom2
				ch2 &bestchrom2
				pos1 &bestpos2
				pos2 &bestpos2
				posid1 &bestposid2
				posid2 &bestposid2
				efftype1 1 
				efftype2 1
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;
			
			data dom_q2;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .				
				qtl1 &bestqtl2 
				qtl2 &bestqtl2 
				ch1 &bestchrom2
				ch2 &bestchrom2
				pos1 &bestpos2
				pos2 &bestpos2
				posid1 &bestposid2
				posid2 &bestposid2
				efftype1 2 
				efftype2 2
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;

			data add1_q2;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter .			
				qtl1 &bestqtl2 
				qtl2 &bestqtl2 
				ch1 &bestchrom2
				ch2 &bestchrom2
				pos1 &bestpos2
				pos2 &bestpos2
				posid1 &bestposid2
				posid2 &bestposid2
				efftype1 3 
				efftype2 3
				contrast &case_
				InfinityLike &bestInfinityLike;
			run;
			
			*--- saving new two-way interaction effect ---*;
			data epi_tmp;
				retain 
				trait &trt 
				cofactors "&xcovar"
				ncofactors &nx
				inter_cofactor_id .
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 1
				indic_qtl_without_inter .			
				qtl1 &bestqtl1 
				qtl2 &bestqtl2 
				ch1 &bestchrom1
				ch2 &bestchrom2
				pos1 &bestpos1
				pos2 &bestpos2
				posid1 &bestposid1
				posid2 &bestposid2
				efftype1 &bestepitype1
				efftype2 &bestepitype2
				contrast .
				InfinityLike &bestInfinityLike;
			run;				
	
			data &hkmimmodel;
				set &hkmimmodel 
					%if %sysevalf(&ng=2) %then add2_q1 add2_q2;
					%if %sysevalf(&ng=3) %then add2_q1 dom_q1 add2_q2 dom_q2; 
					%if %sysevalf(&ng=4) %then %do;
						%if &case_1 = 2 %then add1_q1; 
						%else %if &case_1 = 3 %then add2_q1 dom_q1;  
						%else %if &case_1 = 4 %then add1_q1 dom_q1;  
						%else %if &case_1 = 5 %then add2_q1;  
						%else %if &case_1 = 6 %then add2_q1 add1_q1;   
						%else %if &case_1 = 7 %then add2_q1 dom_q1;   
						%else %if &case_1 = 8 %then add2_q1 dom_q1 add1_q1; 
						
						%if &case_ = 2 %then add1_q2; 
						%else %if &case_ = 3 %then add2_q2 dom_q2;  
						%else %if &case_ = 4 %then add1_q2 dom_q2;  
						%else %if &case_ = 5 %then add2_q2;  
						%else %if &case_ = 6 %then add2_q2 add1_q2;   
						%else %if &case_ = 7 %then add2_q2 dom_q2;   
						%else %if &case_ = 8 %then add2_q2 dom_q2 add1_q2; 
					%end;
					epi_tmp;
				if trait= &trt and neffect > 0 then do;
					nqtl = &totalqtl;
					neffect = &totaleff;
				end;
			run;

			proc datasets nolist;
				delete add2_q1 dom_q1 add1_q1 add2_q2 dom_q2 add1_q2 epi_tmp; 
			quit;

		%end;
	%end;

	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	
	proc datasets nolist;
		delete hkmodel_tmp
			MappedGrid_tmp
			MappedGrid_tmpNoDup
			hkdata1
			hkdata3;
	quit;
	
	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel); 

	%exitsce2D:;

	ods exclude none;
	ods results;
	*options source notes;

%mend scanepi2D;

/* fwdepi2D() calls function scanepi2D() consecutively until no more two-way intaraction between 
genomic poisitios cannot be added into the MIM model */
%macro fwdepi2D(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_epi2D=);

	%global flag_warn1_s2D warn1_s2D_message;
	%let flag_warn1_s2D=0;
	%let warn1_s2D_message=;

	%global flag_warn1_f2D warn1_f2D_message;	
	%let flag_warn1_f2D = 0;
	%let warn1_f2D_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbeffout;
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;
		
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitfwd2D;
	%end;
	
	*--- get the number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitfwd2D;
	%end;
	
	%if &InfinityLikein %then %do;
		%let flag_warn1_f2D = 1;
		%let warn1_f2D_message = No more QTL effects can be added into the model because the residual variance is zero.;
		%put WARNING: &warn1_f2D_message;
		%goto exitfwd2D;
	%end;	
	
	*--- search for two-way interacton among QTL in the model ---*;
	%scanepi2D(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, wsize=&wsize, 
		criterion=&criterion, lod_threshold= &lod_threshold, probf_entry= &probf_entry, teststep_epi2d= &teststep_epi2d); 
	
	*--- get the number of QTL and effects after the scan search ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%*put numbeffout = &numbeffout;
	%*put numbeffin = &numbeffin;
	
	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn1_f2D = 1;
		%let warn1_f2D_message = No significant QTL x QTL effects were found in the 2-D search for QTL x QTL.;
		%put WARNING: &warn1_f2D_message;
	%end;
	%else %do %while(%sysevalf(&numbeffout > &numbeffin) and &InfinityLikeout=0);
		*--- search for new effects until no more effects can be found ---*;
		%let numbeffin = &numbeffout;
	
		*--- search for new effect ---*;
		%scanepi2D(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, wsize=&wsize, 
			criterion=&criterion, lod_threshold= &lod_threshold, probf_entry= &probf_entry, teststep_epi2d= &teststep_epi2d); 
				
		*--- get the number of QTL and effects after the scan search ---*;
		proc means data=&hkmimmodel noprint;
			where trait=&trt;
			var nqtl neffect InfinityLike;
			output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
		run;
		
		data _null_;
			set nqtlout_tmp;
			call symputx("numbeffout", numbeff, 'L');
			call symputx("InfinityLikeout", InfinityLike, 'L');
		run;

		*--- refine model before exiting the while loop ---*;
		%if &InfinityLikeout = 1 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);	
			%let IsRefined= 1;
		%end;
		%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);
			*setting IsRefined to 1 to stop refining QTL position*;
			%let IsRefined= 1;
			*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
			%let numbeffin= %sysevalf(&numbeffin - 1);
		%end;
		%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
			*setting IsRefined to 0 to allow refining QTL position*;
			%let IsRefined= 0;
		%end;		
		
	%end; *--- end while loop over search for new effect ---*;
	
	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;
	
	*--- deleting temporary data sets ---*;
	proc datasets nolist;
		delete nqtlin_tmp nqtlout_tmp;
	quit;	
	
	*scanepi2D() includes in the model only the best interaction effect (e.g., a*a, or a*d, etc.) for a pair of qtls**;
	*To search for other interaction effects among pairs of qtl, we need to call fwdepi0D*;
	%if &flag_warn1_f2D = 0 %then %do;
		%fwdepi0D(trt=&trt, ng=&ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
			criterion= &criterion, lod_threshold= &lod_threshold, 
			probf_entry= &probf_entry, mapgrid= &mapgrid, wsize= &wsize);			
	%end;	

	%exitfwd2D:;
	
%mend fwdepi2D;

/* mscanqtlxcofactor() scans for interaction between cofactor and QTL already in the model */
%macro mscanqtlxcofactor(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_entry=);
	***this function uses global variable selcofactorsIDList that cames from Here:GFwdQTLxCofactorExpr***;
	
	ods exclude all;
	ods noresults;

	%global flag_warn_msqxc warn_msqxc_message;
	%let flag_warn_msqxc=0;
	%let warn_msqxc_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local cofvarid;
	%local u;
	%local criterionname;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local sepc;
	%local sepp;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local rc;
	%local maing1;
	%local maing2;
	%local currNeg2LogLike;
	%local currAIC;
	%local currAICC;
	%local currBIC;
	%local currvalue;
	%local bestvalue;
	%local bestqtl1;
	%local bestchrom1;
	%local bestpos1;
	%local bestposid1;
	%local bestepitype1;
	%local Neg2LogLikeatbest;
	%local AICatbest;
	%local AICCatbest;
	%local BICatbest;
	%local LODatbest;
	%local r_i;
	%local qtl_1;
	%local ch_1;
	%local pos_1;
	%local posid_1;
	%local efftype_1;
	%local maing_1;
	%local r_j;
	%local qtl_2;
	%local ch_2;
	%local pos_2;
	%local posid_2;
	%local efftype_2;
	%local maing_2;
	%local flag_qxc_in_model;
	%local i;
	%local j;
	%local k; 
	%local z;
	%local r_k;
	%local geffect_qxc_test;
	%local tmpNeg2LogLike;
	%local tmpAIC;
	%local tmpAICC;
	%local tmpBIC;
	%local tmpvalue;
	%local tmpLOD; 	
	%local threshold;
	%local totalqtl;
	%local totaleff;
	
	%local ProbFatbest;
	%local contrasts;
	%local tmpProbF;	
			
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitmsqxc;
	%end;
	
	%if %symexist(selcofactorsIDList) = 0 %then %let selcofactorsIDList= ;
	
	%if %length(&selcofactorsIDList) = 0 %then %do;
		%put ERROR: List of selected cofactors does is empty.;
		%let exiterror=1;
		%goto exitmsqxc;	
	%end;		
	
	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/

	*---get criterion to select best genomic position---*;
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then
		%let criterionname = -2 LOG LIKELIHOOD;
	%else %if (&criterion=AIC) %then
		%let criterionname = AIC (SMALLER IS BETTER);
	%else %if (&criterion=AICC) %then
		%let criterionname = AICC (SMALLER IS BETTER);
	%else %if (&criterion=BIC) %then
		%let criterionname = BIC (SMALLER IS BETTER);
	%else %if (&criterion=ProbF) %then
		%let criterionname = PROBF;
	
	*---crete temporary data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); 
	*--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitmsqxc;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);	
	
	*--- get number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLikein;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%if %sysevalf(&numbqtlin < 1 or &nx < 1) %then %do;
		%let flag_warn_msqxc = 1;
		
		%let warn_msqxc_message= ;
		%if &numbqtlin < 1 %then %do;
			%let warn_msqxc_message = &warn_msqxc_message The search procedure Cofactor x QTL in the model needs a model with at least 1 QTL.;
		%end;
		%if &nx < 1 %then %do;
			%let warn_msqxc_message = &warn_msqxc_message The search procedure Cofactor x QTL in the model needs a model with at least 1 cofactor.;		
		%end;		
		%put WARNING: &warn_msqxc_message;  
			
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitmsqxc;	
	%end;

	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	*--- get effects of QTL already in the model ---*;
	%let sepc=c;
	%let sepp=p;
	%let geffect_main_in=;
	%let geffect_epi_in=;
	%let geffect_qxc_in=;
	
	%let id_hkmodel_tmp=%sysfunc(open(hkmodel_tmp));
	%let NObs_hkmodel_tmp=%sysfunc(attrn(&id_hkmodel_tmp,NOBS));
	
	%syscall set(id_hkmodel_tmp);
	%do i=2 %to &NObs_hkmodel_tmp;
		%let rc=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));

		*---get main effect of QTL already in the model ---*;
		%if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id = .) %then %do;
			%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
			%let geffect_main_in= &geffect_main_in &maing1;
		%end;
		%else %if %sysevalf(&qtl1 ^= &qtl2 and &ch1 ^= &ch2 and &posid1 ^= &posid2 and &inter_cofactor_id = .) %then %do;
			*---get two-way interaction (epistasis) already in the model ---*; 
			%let maing1= e&efftype1&sepc&ch1&sepp&posid1;
			%let maing2= e&efftype2&sepc&ch2&sepp&posid2;
			%let geffect_epi_in= &geffect_epi_in &maing1 * &maing2;
		%end;
		%else %if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id ^= .) %then %do;
			*---get two-way interaction of Cofactor x QTL already in the input model ---*;
			%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
			%let maing2= x&inter_cofactor_id;
			%let geffect_qxc_in= &geffect_qxc_in &maing1 * &maing2;
		%end;
	%end;
	%*put geffect_main_in= &geffect_main_in;
	%*put geffect_epi_in= &geffect_epi_in;
	%*put geffect_qxc_in= &geffect_qxc_in;
	
	%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);

	*--- get value of criterion for input model ---*;		
	proc mixed data=hkdata3 method=ml;
		model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
		ods output fitstatistics=curr_st_tmp;
	run;
	
	data _null_;
		set curr_st_tmp;
		if upcase(descr) = "-2 LOG LIKELIHOOD" then 
			call symputx("currNeg2LogLike", trim(left(value)), 'L');
		else if upcase(descr) = "AIC (SMALLER IS BETTER)" then 
			call symputx("currAIC", trim(left(value)), 'L');
		else if upcase(descr) = "AICC (SMALLER IS BETTER)" then 
			call symputx("currAICC", trim(left(value)), 'L');
		else if upcase(descr) = "BIC (SMALLER IS BETTER)" then 
			call symputx("currBIC", trim(left(value)), 'L');
		if upcase(descr) = "&criterionname" then
			call symputx("currvalue", trim(left(value)), 'L');
	run;
	%*put currvalue= &currvalue;
	
	proc datasets nolist;
		delete curr_st_tmp;
	quit;
	
	*--- create data set for profiling results ---*;
	data msqxc msqxc_tmp;
		retain qtl1 qtl2 ch1 ch2 pos1 pos2 posid1 posid2 efftype1 efftype2 LOD AIC AICC BIC ProbF .;
		stop;
	run;

	*--- begin search for interaction between cofactor and QTL in the model ---*;	
	%let bestvalue = 10000000000;
	%let bestqtl1 = ;
	%let bestchrom1 = ;
	%let bestpos1 = ;
	%let bestposid1= ;
	%let bestepitype1= ;
	%let Neg2LogLikeatbest = ;
	%let AICatbest = ;
	%let AICCatbest = ;
	%let BICatbest = ;
	%let LODatbest = ;
	%let ProbFatbest = ;
			
	*--- qtl1 ---*;
	%do i=2 %to &NObs_hkmodel_tmp; 
		*--- i starts at 2 because the first obsearvation for the null model is nqll=0 --- *;	
		%*put i = &i;
		%let r_i=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));
		*I use this if statement because I assume all QTL in the model have main effects*;
		%if %sysevalf(&qtl1 = &qtl2 and &inter_cofactor_id = .) %then %do;
			%let qtl_1 = &qtl1;
			%let ch_1 = &ch1;
			%let pos_1 = &pos1;
			%let posid_1 = &posid1;
			%let efftype_1 = &efftype1;
			%let maing_1= e&efftype_1&sepc&ch_1&sepp&posid_1;
						
			%let z = 1;
			%*let cofvar= %scan(&xcovar,&z);
			%*do %while(%length(&cofvar));
			%let cofvarid= %scan(%superQ(selcofactorsIDList), &z, %str(,));
			%do %while(%length(&cofvarid));
				%put z= &z;
				%let cofvar= %str(x)&cofvarid;
				%put cofvar= &cofvar;

				*--- checking if the interaction between qtl1 and cofactor is already in the model ---*;
				%let flag_qxc_in_model = 0;
				%let k = 2;  
				*--- k starts at 2 because the first obsearvation for the null model is nqll=0 --- *;
				%do %while (%sysevalf(&k <= &NObs_hkmodel_tmp and &flag_qxc_in_model = 0));
					%let r_k=%sysfunc(fetchobs(&id_hkmodel_tmp,&k));
					%*put k=&k;
					%if (%sysevalf(&inter_cofactor_id ^= . and &inter_cofactor_id = &cofvarid and &qtl1 = &qtl_1 and &efftype1 = &efftype_1)) %then %do;	
						%let flag_qxc_in_model = 1;
					%end; 
					*--- end if Cofactor x QTL is already in the model ---*;
					%let k = %sysevalf(&k + 1);
				%end;
				*--- end loop for checking if the Cofactor x QTL interaction is already in the model ---*;
				
				*--- fitting model with Cofactor x QTL interation that is not yet in the model ---*;
				%if %sysevalf(&flag_qxc_in_model = 0) %then %do;

					%put ### mscanqtlxcofactor: Trait= &trt, QTL1= &qtl_1 type1= &efftype_1, Cofactor= &cofvar ###;

					%let geffect_qxc_test = &maing_1.*&cofvar;
					%put geffect_qxc_test = &geffect_qxc_test;

					proc mixed data=hkdata3 method=ml;
						model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &geffect_qxc_test/ solution;
						ods output fitstatistics=test_st_tmp solutionf=test_es_tmp;
					run;

					*--- collect estimable interaction effect entering the growing model ---*;
					data test_es_tmp2;
						set test_es_tmp;
						where (Effect = "&maing_1.*&cofvar" or Effect = "&cofvar.*&maing_1") and DF ^= . ;
					run;					

					%let contrasts = ;
					proc sql;
						select Effect into : contrasts separated by ' 1, '
						from test_es_tmp2;
					quit;

					%let tmpvalue = .;	
					%let tmpNeg2LogLike= .;
					%let tmpLOD =.;
					%let tmpAIC =.;
					%let tmpAICC =.;
					%let tmpBIC = .;
					%let tmpProbF= .;

					*--- make sure to test a contrast only if it has at least one interaction ---*;						
					%if %length(&contrasts) %then %do;
						%let contrasts = %str(contrast 'test_qtl_cofactor' &contrasts 1;);
						%*put contrasts = &contrasts;

						*--- get values of -2 Log Likehihood, AIC, AICC, BIC for the growing model ---*;
						data _null_;
							set test_st_tmp;
							if upcase(descr) = "-2 LOG LIKELIHOOD" then 
								call symputx("tmpNeg2LogLike", trim(left(value)), 'L');
							else if upcase(descr) = "AIC (SMALLER IS BETTER)" then 
								call symputx("tmpAIC", trim(left(value)), 'L');
							else if upcase(descr) = "AICC (SMALLER IS BETTER)" then 
								call symputx("tmpAICC", trim(left(value)), 'L');
							else if upcase(descr) = "BIC (SMALLER IS BETTER)" then 
								call symputx("tmpBIC", trim(left(value)), 'L');
							if upcase(descr) = "&criterionname" then
								call symputx("tmpvalue", trim(left(value)), 'L');
						run;
						%*put tmpvalue= &tmpvalue;

						%let tmpLOD = %sysevalf((&currNeg2LogLike - &tmpNeg2LogLike)/(2 * 2.30258509299404));						

						*--- get value of ProbF of the interaction effect entering the growing model ---*;
						proc mixed data=hkdata3 method=ml;
							model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &geffect_qxc_test/ solution;
							ods output contrasts=test_ct_tmp;
							&contrasts;
						run;

						data _null_;
							set test_ct_tmp;
							call symputx("tmpProbF", trim(left(ProbF)), 'L');
							if "&criterionname" = "PROBF" then
								call symputx("tmpvalue", trim(left(ProbF)), 'L');
						run;
						%*put tmpProbF = &tmpProbF;

						*--- delete temporary data sets ---*;
						proc datasets nolist;
							delete test_ct_tmp;
						quit;

					%end;

					*--- delete temporary data sets ---*;
					proc datasets nolist;
						delete test_st_tmp
							test_es_tmp
							test_es_tmp2;
					quit;

					*---saving profiling to data set---*;
					data msqxc_tmp; 
						retain	
						qtl1 &qtl_1 
						qtl2 &qtl_1
						ch1 &ch_1
						ch2 &ch_2
						pos1 &pos_1
						pos2 &pos_1
						posid1 &posid_1
						posid2 &posid_1
						efftype1 &efftype_1
						efftype2 &efftype_1 
						LOD &tmpLOD
						AIC &tmpAIC
						AICC &tmpAICC
						BIC &tmpBIC
						ProbF &tmpProbF;
					run;

					data msqxc;
					 	set msqxc msqxc_tmp;
					run;

					*---selecting best model---*;
					%put criterionname= &criterionname;
					%put bestvalue= &bestvalue;
					%put tmpvalue= &tmpvalue;
					%if (%sysevalf(&tmpvalue ^=. and &tmpvalue < &bestvalue)) %then %do;
						%let bestcofactor_id= %substr(&cofvar, 2);
					
						%let bestvalue = &tmpvalue;
						%let bestqtl1= &qtl_1;
						%let bestchrom1 = &ch_1;
						%let bestpos1 = &pos_1;
						%let bestposid1 = &posid_1;
						%let bestepitype1 = &efftype_1;

						%let Neg2LogLikeatbest = &tmpNeg2LogLike;
						%let AICatbest = &tmpAIC;
						%let AICCatbest = &tmpAICC;
						%let BICatbest = &tmpBIC;
						%let LODatbest = &tmpLOD;
						%let ProbFatbest = &tmpProbF;
					%end;
				%end; 
				*--- and if flat_qxc_in_model ---*;
				
				%let z = %sysevalf(&z + 1);
				%*let cofvar = %scan(&xcovar,&z);
				%let cofvarid= %scan(%superQ(selcofactorsIDList), &z, %str(,));
				
			%end; 
			*--- end while loop ---*;
		%end; 
	%end; 
	*--- end do loop for qtl1 ---*;
	%let id_hkmodel_tmp=%sysfunc(close(&id_hkmodel_tmp)); 
	*--- very important to close ---*;
			
	*---saving best model---*;
	%if %length(&bestpos1) %then %do;
		%put ## Scan for Cofactor x QTL interaction among effects in the model ##;
		
		%if "&criterion" = "LOD" %then %do;
			%let threshold = &lod_threshold;
			%let bestvalue = &LODatbest;
		%end;
		%else %if "&criterion" = "ProbF" %then %do;
			%let threshold = &probf_entry;
		%end;
		%else %do;
			%let threshold = &currvalue;
		%end;
		
		%put criterion = &criterion;
		%put threshold = &threshold;			
		%put bestvalue = &bestvalue;
			
		%if ("&criterion" ^= "LOD" and %sysevalf(&bestvalue < &threshold)) or 
			("&criterion" = "LOD" and %sysevalf(&bestvalue > &threshold)) %then %do;
				
			%let totalqtl= &numbqtlin;
			%let totaleff= %sysevalf(&numbeffin + 1);
			
			data neweff_tmp;
				retain 
				trait &trt
				cofactors "&xcovar"				
				ncofactors &nx
				inter_cofactor_id &bestcofactor_id
				nqtl &totalqtl
				neffect &totaleff
				indic_epi_eff 0
				indic_qtl_without_inter . 				
				qtl1 &bestqtl1 
				qtl2 &bestqtl1
				ch1 &bestchrom1
				ch2 &bestchrom1
				pos1 &bestpos1
				pos2 &bestpos1
				posid1 &bestposid1
				posid2 &bestposid1
				efftype1 &bestepitype1
				efftype2 &bestepitype1
				contrast .;
			run;

			data &hkmimmodel;
				set &hkmimmodel neweff_tmp;
				if trait= &trt and neffect > 0 then do;
					nqtl = &totalqtl;
					neffect = &totaleff;
				end;
			run;

			proc datasets nolist;
				delete neweff_tmp;
			quit;
		%end;
	%end; *---end if for saving best model---*;

	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp 
			msqxc_tmp
			hkdata3;
	quit;
	
	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);
				
	%exitmsqxc:;
	
	ods exclude none;
	ods results;
	
%mend mscanqtlxcofactor;

/* mfwdqtlxcofactor() calls function mscanqtlxcofactor() consecutively until no more two-way intaraction between QTL 
alrealdy in the model cannot be added into the MIM model */
%macro mfwdqtlxcofactor(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_entry=, mapgrid=, wsize= );
	***this function uses global variable selcofactorsIDList that cames from Here:MFwdQTLxCofactorExpr***;

	%global flag_warn_msqxc warn_msqxc_message;
	%let flag_warn_msqxc=0;
	%let warn_msqxc_message=;

	%global flag_warn_mfqxc warn_mfqxc_message;
	%let flag_warn_mfqxc=0;
	%let warn_mfqxc_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbeffout;
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;

	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitmfqxc;
	%end;
	
	%if %symexist(selcofactorsIDList) = 0 %then %let selcofactorsIDList= ;
	
	%if %length(&selcofactorsIDList) = 0 %then %do;
		%put ERROR: List of selected cofactors does is empty.;
		%let exiterror=1;
		%goto exitmfqxc;	
	%end;		
		
	*--- get the number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); 
	*--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitmfqxc;
	%end;
	
	%if %sysevalf(&numbqtlin < 1) %then %do;
		%let flag_warn_mfqxc = 1;
		%let warn_mfqxc_message= The search procedure for Cofactor x QTL needs a model with at least 1 QTL.;
		%put WARNING: &warn_mfqxc_message; 
		%goto exitmfqxc;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set &hkmimmodel(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	%put xcovar= &xcovar;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);
	
	%if %sysevalf(&numbqtlin < 1 or &nx < 1) %then %do;
		%let flag_warn_fsqxc = 1;
		
		%let warn_fsqxc_message= ;
		%if &numbqtlin < 1 %then %do;
			%let warn_fsqxc_message = &warn_fsqxc_message The search procedure Cofactor x QTL in the model needs a model with at least 1 QTL.;
		%end;
		%if &nx < 1 %then %do;
			%let warn_fsqxc_message = &warn_fsqxc_message The search procedure Cofactor x QTL in the model needs a model with at least 1 cofactor.;		
		%end;		
		%put WARNING: &warn_fsqxc_message;  
			
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitmfqxc;
	%end;	
	
	%if &InfinityLikein %then %do;
		%let flag_warn_mfqxc = 1;
		%let warn_mfqxc_message= No more QTL effects can be added into the model because the residual variance is zero in the forward search procedure.;
		%put WARNING: &warn_mfqxc_message;
		%goto exitmfqxc;
	%end;	
	
	*--- search for interacton between cofactor and QTL in the model ---*;
	%mscanqtlxcofactor(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, hkmimmodel=&hkmimmodel, 
		criterion=&criterion, lod_threshold=&lod_threshold, probf_entry=&probf_entry);		
	
	*--- get the number of QTL and effects after the scan search ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%*put numbeffout = &numbeffout;
	%*put numbeffin = &numbeffin;
	
	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn_mfqxc = 1;
		%let warn_mfqxc_message= No significant Cofactor x QTL effects were found in the search procedure.;
		%put WARNING: &warn_mfqxc_message;
	
	%end;
	%else %do %while(%sysevalf(&numbeffout > &numbeffin) and &InfinityLikeout=0);
		*--- search for new effects until no more effects can be found ---*;
		%let numbeffin = &numbeffout;
	
		*--- search for new effect ---*;
		%mscanqtlxcofactor(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, hkmimmodel=&hkmimmodel, 
			criterion=&criterion, lod_threshold=&lod_threshold, probf_entry=&probf_entry);	
				
		*--- get the number of QTL and effects after the scan search ---*;
		proc means data=&hkmimmodel noprint;
			where trait=&trt;
			var nqtl neffect InfinityLike;
			output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
		run;
		
		data _null_;
			set nqtlout_tmp;
			call symputx("numbeffout", numbeff, 'L');
			call symputx("InfinityLikeout", InfinityLike, 'L');
		run;

		*--- refine model before exiting the while loop ---*;
		%if &InfinityLikeout = 1 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);	
			%let IsRefined= 1;
		%end;
		%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);
			*setting IsRefined to 1 to stop refining QTL position*;
			%let IsRefined= 1;
			*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
			%let numbeffin= %sysevalf(&numbeffin - 1);
		%end;
		%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
			*setting IsRefined to 0 to allow refining QTL position*;
			%let IsRefined= 0;
		%end;
		
	%end; *--- end while loop over search for new effect ---*;
	
	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;
	
	*--- deleting temporary data sets ---*;
	proc datasets nolist;
		delete nqtlin_tmp 
			nqtlout_tmp;
	quit;	

	%exitmfqxc:;
	
%mend mfwdqtlxcofactor;

/* gscanqtlxcofactor() scans through the genome grid and adds the best putative QTL x Cofactor effect into model */
%macro gscanqtlxcofactor(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_qxc=);
	***this function uses global variable selcofactorsIDList that cames from Here:GFwdQTLxCofactorExpr***;

	ods exclude all;
	ods noresults;
	*options nosource nonotes;/*--- prevents NOTE from being written down in the LOG window---*/

	%global flag_warn1_gsqxc warn1_gsqxc_message;
	%let flag_warn1_gsqxc=0;	
	%let warn1_gsqxc_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local cofvarid;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local sepc;
	%local sepp;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local rc;
	%local maing1;
	%local maing2;
	%local L_window1;
	%local R_window1;
	%local L_window2;
	%local R_window2;
	%local id_mappedgrid_tmp;
	%local NObs_mappedgrid_tmp;
	%local bestqtl1;
	%local bestchrom1;
	%local bestpos1;
	%local bestposid1;
	%local bestqxctype1;
	%local threshold;	
	%local newqtl1;
	%local z;
	%local j;
	%local u;
	%local newqtl1var;
	%local cofvar;
	%local newqxc;
	%local newqxc_rec;
	%local teststmt;
	%local id_bestqxc;
	%local NObs_bestqxc;
	%local totalqtl;
	%local totaleff;	
	%local start_z;
	%local start_x;
	%local bestqxctype1_tmp;
	%local besteff_qxc;

	%local orderdirection;
	%local bool;
	%local neg2loglike0;
	%local AIC0;
	%local AICC0;
	%local BIC0;

	%local id_hkdata1;
	%local NObs_hkdata1;
	%local bestvalue;
	%local case_;
	%local case_1;
	%local ContrastTestqxc;
	%local Contrastj;
	%local totaleff_q1;
	%local bestcofactor_id;
	%local bestInfinityLike;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitgsqxc;
	%end;
	
	%if %symexist(selcofactorsIDList) = 0 %then %let selcofactorsIDList= ;
	
	%put selcofactorsIDList= &selcofactorsIDList;
	%if %length(&selcofactorsIDList) = 0 %then %do;
		%put ERROR: List of selected cofactors does is empty.;
		%let exiterror=1;
		%goto exitgsqxc;	
	%end;
	
	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/

	*---get criterion to select best genomic position---*;
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then %do;
		%let threshold = &lod_threshold;
	%end;
	%else %if (&criterion=AIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=AICC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=BIC) %then %do;
		%let threshold = ;
	%end;
	%else %if (&criterion=ProbF) %then %do;
		%let threshold = &probf_entry;
	%end;

	*--- create data set that holds the model from the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); 
	*--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitgsqxc;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	%put xcovar= &xcovar;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);
	
	%if &nx < 1 %then %do;
	
		%let flag_warn1_gsqxc = 1;
		%let warn1_gsqxc_message = The model does not have cofactors.;
		%put WARNING: &warn1_gsqxc_message;
	
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitgsqxc;
	%end;
	
	*--- get the number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	%*put teststep_qxc= &teststep_qxc;
	*--- create data sets for grid search and profiling results ---*;
	data MappedGrid_tmp;
		*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;	
		set &mapgrid(rename= (&CHROMNUM= chrom) 
					%if %length(&g_firstobs) %then firstobs= &g_firstobs;
					%if %length(&g_obs) %then obs=&g_obs;
					);
		retain _mkpos_;
		if MarkerIndicator = 1 then do;
			_mkpos_ = cm;
			output;
		end;
		else if (MarkerIndicator = 0 and (cm - _mkpos_ >= &teststep_qxc)) then do;
			_mkpos_ = cm;
			output;
		end;
	run;
	
	%*goto exitgsqxc;

	*--- get effects already in the model ---*;
	%let sepc=c;
	%let sepp=p;
	%let geffect_main_in=;
	%let geffect_epi_in=;
	%let geffect_qxc_in=;
	%if %sysevalf(&numbqtlin > 0) %then %do;
		
		%let id_hkmodel_tmp=%sysfunc(open(hkmodel_tmp));
		%let NObs_hkmodel_tmp=%sysfunc(attrn(&id_hkmodel_tmp,NOBS));
		
		%syscall set(id_hkmodel_tmp);
		%do i=2 %to &NObs_hkmodel_tmp;
			%let rc=%sysfunc(fetchobs(&id_hkmodel_tmp,&i));

			%if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id = .) %then %do;
				*---get main effects of QTL already in the input model ---*;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let geffect_main_in= &geffect_main_in &maing1;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 ^= &qtl2 and &ch1 ^= &ch2 and &posid1 ^= &posid2 and &inter_cofactor_id = .) %then %do; 
				*---get two-way interaction (epistasis) already in the model ---*;  	
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1;
				%let maing2= e&efftype2&sepc&ch2&sepp&posid2;
				%let geffect_epi_in= &geffect_epi_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%let L_window2 = %sysevalf(&pos2 - &wsize); 
				%let R_window2 = %sysevalf(&pos2 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				%*put L2=&L_window2;
				%*put R2=&R_window2;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;					
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 & chrom ~= &ch2 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) |
						chrom = &ch2 and (cm <= &L_window2 or cm >= &R_window2) then do; 
						output;
					end;
				run;
			%end;
			%else %if %sysevalf(&qtl1 = &qtl2 and &ch1 = &ch2 and &posid1 = &posid2 and &inter_cofactor_id ^= .) %then %do;
				*---get two-way interaction of Cofactor x QTL already in the input model ---*;
				%let maing1= e&efftype1&sepc&ch1&sepp&posid1; 
				%let maing2= x&inter_cofactor_id;
				%let geffect_qxc_in= &geffect_qxc_in &maing1 * &maing2;
				*---get window size for each QTL already in the model ---*;
				%let L_window1 = %sysevalf(&pos1 - &wsize); 
				%let R_window1 = %sysevalf(&pos1 + &wsize);
				%*put L1=&L_window1;
				%*put R1=&R_window1;
				*---get genome grid if the input model has QTL. Positions within window size of QTL are deleted---*;
				data MappedGrid_tmp;
					set MappedGrid_tmp;
					*---chrom without QTL---*;
					if chrom ~= &ch1 then do; 
						output;
					end;
					*---chrom with QTL---*;
					else if chrom = &ch1 and (cm <= &L_window1 or cm >= &R_window1) then do; 
						output;
					end;
				run;
			%end;			
		%end;
		%let id_hkmodel_tmp=%sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;
	%end;
	%put geffect_main_in= &geffect_main_in;
	%put geffect_epi_in= &geffect_epi_in;
	%put geffect_qxc_in= &geffect_qxc_in;
	
	%let id_mappedgrid_tmp=%sysfunc(open(MappedGrid_tmp));
	%let NObs_mappedgrid_tmp=%sysfunc(attrn(&id_mappedgrid_tmp,NOBS));
	%let id_mappedgrid_tmp=%sysfunc(close(&id_mappedgrid_tmp)); *--- very important to close ---*;
	
	%if %sysevalf(&NObs_mappedgrid_tmp < 2) %then %do;
	
		%let flag_warn1_gsqxc = 1;
		%let warn1_gsqxc_message = The genomic scan for Cofactor x QTL stopped because all positions in the genome grid are occuppied.;
		%put WARNING: &warn1_gsqxc_message;
		
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp MappedGrid_tmp;
		quit;
		
		%goto exitgsqxc;
	%end;
	%else %do;		
		%put geffect_main_in=&geffect_main_in;
		
		%prephkdata1(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata, mappedgr=MappedGrid_tmp, chrom_test=, cmid_test=);
		
		*data OutLib.hkdata1; *set hkdata1; *run;
		
		%*goto exitgsqxc;
		
		%if &ng = 2 %then %do;
			%let newqtl1 = xa2_;
		%end;
		%else %if &ng = 3 %then %do;
			%let newqtl1 = xa2_ xd_;
		%end;
		%else %do;
			%let newqtl1 = xa2_ xd_ xa1_;
		%end; 
		%put newqtl1= &newqtl1;		

		%if %sysevalf(&criterion=AIC or &criterion=AICC or &criterion=BIC) %then %do; 
			%let threshold = &&&criterion.0;		
		%end;
		
		%let orderdirection = ;
		%let bool = < ;
		%if "&criterion" = "LOD" %then %do; 
			%let orderdirection = descending;
			%let bool = > ;
			
			/*
			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);
			
			*data OutLib.hkdata3; *set hkdata3; *run;
	
			proc mixed data=hkdata3 method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
				ods output fitstatistics=fs0 solutionf=sf0;
			run;
	
			proc transpose data=fs0 out=fs01(rename= (
								_2_Log_Likelihood=neg2loglike0
								AIC__smaller_is_better_ =AIC0
								AICC__smaller_is_better_= AICC0
								BIC__smaller_is_better_=BIC0
								));
				var value;
				id descr;
			run;
	
			data _null_;
				set fs01;
				call symputx("neg2loglike0", trim(left(neg2loglike0)), 'L');
				call symputx("AIC0", trim(left(AIC0)), 'L');
				call symputx("AICC0", trim(left(AICC0)), 'L');
				call symputx("BIC0", trim(left(BIC0)), 'L');
			run;
			*/
			
			proc mixed data=hkdata1 method=ml;
				by chrom cm cmid;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newqtl1/ solution; 					
				ods output fitstatistics=fs0 solutionf=sf0;
			run;			
			
			proc transpose data=fs0 out=fs01(rename= (
								_2_Log_Likelihood=neg2loglike0
								AIC__smaller_is_better_ =AIC0
								AICC__smaller_is_better_= AICC0
								BIC__smaller_is_better_=BIC0
								));
				var value;
				id descr;
				by chrom cm cmid;
			run;			
			
		%end;
				
		data bestqxc_all;
			set _null_;
		run;
		
		data bestqxc_list;
			set _null_;
		run;

		%let u = 1;
		%let newqtl1var= %scan(&newqtl1,&u);
		%do %while(%length(&newqtl1var));
			%put u=&u;

			%let z = 1;
			%*let cofvar= %scan(&xcovar,&z);
			%*do %while(%length(&cofvar));
			%let cofvarid= %scan(%superQ(selcofactorsIDList), &z, %str(,));
			%do %while(%length(&cofvarid));
				%put z= &z;
				%let cofvar= %str(x)&cofvarid;
				%let newqxc= &newqtl1var.*&cofvar;
				%let newqxc_rec= &cofvar.*&newqtl1var;
				%put cofvar= &cofvar;

				/*joint testing of new main and cofactor*/
				/*
				%let teststmt= 'contq_' &cofvar 1;
				%if &ng >= 3 %then 
					%let teststmt= %str(&teststmt, xa2_ 1, xd_ 1);
				%if &ng = 4 %then 
					%let teststmt= %str(&teststmt, xa2_ 1, xd_ 1, xa1_ 1);

				%let teststmt= %str(&teststmt, &newqxc 1); 
				%put teststmt = &teststmt;
				*/				

				*test only new Cofactor x QTL interaction*;
				%let teststmt= 'contq_' &newqxc 1;
				%put teststmt = &teststmt;

				ods output contrasts= contrq CovParms=CovP;
				proc mixed data=hkdata1 method=ml;
					by chrom cm cmid;
					model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newqtl1 &newqxc/ solution; /*main and new cofactor x qtl only*/					
					*model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in &newqxc/ solution; /*new cofactor x qtl only*/
					ods output fitstatistics=fs solutionf=sf nobs=nobm;
					contrast &teststmt / chisq;
				run;
				
				data CovP;
					set CovP;
					where  CovParm = 'Residual';
					if Estimate=0 then InfinityLike= 1;
					else InfinityLike= 0;
				run; 						

				proc transpose data=fs out=fs1(rename= (
									_2_Log_Likelihood=neg2loglike
									AIC__smaller_is_better_ =AIC
									AICC__smaller_is_better_= AICC
									BIC__smaller_is_better_=BIC
									));
					var value;
					id descr;
					by chrom cm cmid;
				run;				

				data nobm(keep=chrom cm cmid NObsRead NObsUsed);
					set nobm;
					where label = "Number of Observations Read";
				run;	

				%*goto exitgsqxc;

				data sf(drop= stderr df tvalue probt);
					length Eff_qxc $20;
					set sf(rename=(Estimate = Est_qxc Effect= Eff_qxc));
					where Eff_qxc = "&newqxc" or Eff_qxc = "&newqxc_rec";
				run;
				
				*data OutLib.contrq; *set contrq; *run;

				data gsqxc_tmp;
					merge sf CovP fs1 nobm contrq %if "&criterion" = "LOD" %then fs01;;
					*by chrom cm cmid;
					by chrom cmid;
					*LOD = (1/(2 * 2.30258509299404))*(&neg2loglike0 - neg2loglike);
					LOD = .;
					%if "&criterion" = "LOD" %then %do; 
						LOD = (1/(2 * 2.30258509299404))*(neg2loglike0 - neg2loglike);
						valuet = LOD;
					%end;
					%else %if "&criterion" = "ProbF" %then %do;
						valuet = probf;
					%end;	
					%else %if "&criterion" = "AIC" %then %do;
						valuet = AIC;
					%end;	
					%else %if "&criterion" = "AICC" %then %do;
						valuet = AICC;
					%end;	
					%else %if "&criterion" = "BIC" %then %do;
						valuet = BIC;
					%end;								
				run;	

				*--- CHROMNUM, MARKERLAB are global variable passed from SingleQTLModel.sas ---*;	
				data gsqxc_tmp(keep = chrom cm cmid 
						neg2loglike AIC BIC AICC LOD NumDF DenDF ProbF
						Eff_qxc Est_qxc
						MarkerIndicator &MARKERLAB NObsRead NObsUsed
						valuet 
						%if "&criterion" = "LOD" %then neg2loglike0 AIC0 BIC0 AICC0;
						); 
					merge &mapgrid(rename=(&CHROMNUM= chrom)) gsqxc_tmp;
					*by chrom cm cmid;
					by chrom cmid;
				run;		

				data bestqxc_ea;
					set gsqxc_tmp;
					where valuet ^=. ;
				run;

				proc sort data=bestqxc_ea;
					by &orderdirection valuet;
				run;
				
				data bestqxc_ea;
					length cofactorvar $20;
					set bestqxc_ea(obs=1);
					cofactorvar= "&cofvar";
				run;

				data bestqxc_all;
					set bestqxc_all bestqxc_ea;
				run;
				
				*data OutLib.bestqxc_all; *set bestqxc_all; *run;
				
				%let z = %sysevalf(&z + 1);
				%*let cofvar = %scan(&xcovar,&z);
				%let cofvarid= %scan(%superQ(selcofactorsIDList), &z, %str(,));

				%*goto exitgsqxc;
			%end;

			%let u= %sysevalf(&u + 1);		
			%let newqtl1var= %scan(&newqtl1,&u);
		%end;
		
		%if 0 %then %do;
			data OutLib.bestqxc_all; set bestqxc_all; run;
		%end;
		
		%*goto exitgsqxc;

		%let z = 1;
		%*let cofvar= %scan(&xcovar,&z);
		%*do %while(%length(&cofvar));
		%let cofvarid= %scan(%superQ(selcofactorsIDList), &z, %str(,));
		%do %while(%length(&cofvarid));
			%put ## Scan Cofactor x QTL interaction ##;
			%put z= &z;
			%let cofvar= %str(x)&cofvarid;
			%put cofvar= &cofvar;

			data bestqxc;
				set bestqxc_all;
				where valuet ^=. and cofactorvar= "&cofvar";
			run;

			proc sort data=bestqxc;
				by &orderdirection valuet;
			run;	

			data _null_;
				set bestqxc(obs=1);
				call symputx("besteff_qxc", trim(left(Eff_qxc)), "L");
				call symputx("bestchrom1", trim(left(chrom)), "L");
				call symputx("bestpos1", trim(left(cm)), "L");
				call symputx("bestposid1", trim(left(cmid)), "L");
				call symputx("bestvalue", trim(left(valuet)), "L");
				%if &ng=4 %then %do;
					call symputx("case_", contrast_, "L");
				%end;
				call symputx("bestInfinityLike", InfinityLike, "L");
			run;

			%put criterion= &criterion;
			%put threshold= &threshold;
			%put bestvalue= &bestvalue;		

			data bestqxc;
				set bestqxc;
				where valuet &bool &threshold;
			run;	

			%*goto exitgsqxc;

			*---saving best model---*;
			%let id_bestqxc=%sysfunc(open(bestqxc));
			%let NObs_bestqxc=%sysfunc(attrn(&id_bestqxc,NOBS));
			%let id_bestqxc=%sysfunc(close(&id_bestqxc));
			*--- very important to close ---*;			
			%if (&NObs_bestqxc > 0) %then %do;
			
				%let sameQTLflag= 0;
			
				*Collect all best qtl to check if the same qtl interacts with disctinct cofactors.*; 
				*This is important to count the number of qtl in the model*;
				data bestqxc_list;
					set bestqxc_list bestqxc(obs=1);
				run;
									
				%if &z > 1 %then %do;
					%NObsVars(bestqxc_list);
					%let nbest_in= &nobs;
				
					proc sort data= bestqxc_list nodupkey;
						by chrom cmid;
					run;
					
					%NObsVars(bestqxc_list);
					%let nbest_out= &nobs;	
					
					%if &nbest_in > &nbest_out %then %do;
						%let sameQTLflag= 1;
						
						*data OutLib.hkmodel_tmpz&z; *set hkmodel_tmp; *run; 
						
						data _null_;
							set hkmodel_tmp;
							if qtl1 = qtl2 and ch1= &bestchrom1 and posid1= &bestposid1 then do;
								call symputx("IDsameqtl", qtl1, "L");
							end;
						run;
					%end;
				%end;
			
				*get cofactor id*;
				%let bestcofactor_id= %substr(&cofvar, 2);
				%put bestcofactor_id= &bestcofactor_id;

				*get qtl effect type*;
				%let besteff_qxctmp = %sysfunc(tranwrd(&besteff_qxc, &cofvar, ));
				%let start_x = %sysfunc(find(&besteff_qxctmp, x));
				%let bestqxctype1_tmp = %substr(&besteff_qxctmp, &start_x + 1, 2);
				
				%if &bestqxctype1_tmp = a1 %then 
					%let bestqxctype1 = 3;
				%else %if &bestqxctype1_tmp = d_ %then 
					%let bestqxctype1 = 2;
				%else %if &bestqxctype1_tmp = a2 %then 
					%let bestqxctype1 = 1;
					
				%put bestqxctype1=&bestqxctype1;
				%put bestchrom1=&bestchrom1;
				%put bestposid1=&bestposid1;
					
				*sameQTLflag is always 0 for the first cofactor*;
				*sameQTLflag is 0 or 1 for second and greater cofactors*;
				%if &sameQTLflag= 0 %then %do;

					%*goto exitgsqxc;	
					
					*updating number of qtl effects (add main plus interaction)*;
					%if &ng= 2 %then %do;
						%let numbeffin = %sysevalf(&numbeffin + 2);
						%let case_ = .;
					%end;
					%else %if &ng=3 %then %do;
						%let numbeffin = %sysevalf(&numbeffin + 3);
						%let case_ = .;
					%end;
					%else %if &ng=4 %then %do;

						%if &case_ = 1 %then %let totaleff_q1 = 1;
						%else %if &case_ = 3 %then %let totaleff_q1 = 1;
						%else %if &case_ = 4 %then %let totaleff_q1 = 1;
						%else %if &case_ = 5 %then %let totaleff_q1 = 1;
						%else %if &case_ = 6 %then %let totaleff_q1 = 1;
						%else %if &case_ = 7 %then %let totaleff_q1 = 1;
						%else %if &case_ = 8 %then %let totaleff_q1 = 3;

						%let numbeffin = %sysevalf(&numbeffin + &totaleff_q1 + 1);
					%end;	
					
					*updating number of qtl*;
					%let numbqtlin = %sysevalf(&numbqtlin + 1);					

					*updating ID for new qtl*;
					%let bestqtl1= %sysevalf(&numbqtlin);
				
					*--- saving main effects of the new QTL ---*;
					data add2_q1;
						retain 
						trait &trt 
						cofactors "&xcovar"
						ncofactors &nx
						nqtl &numbqtlin
						neffect &numbeffin
						indic_epi_eff 0
						indic_qtl_without_inter .
						inter_cofactor_id .				
						qtl1 &bestqtl1 
						qtl2 &bestqtl1 
						ch1 &bestchrom1
						ch2 &bestchrom1
						pos1 &bestpos1
						pos2 &bestpos1
						posid1 &bestposid1
						posid2 &bestposid1
						efftype1 1 
						efftype2 1
						contrast &case_
						InfinityLike &bestInfinityLike;
					run;

					data dom_q1;
						retain 
						trait &trt 
						cofactors "&xcovar"
						ncofactors &nx
						nqtl &numbqtlin
						neffect &numbeffin
						indic_epi_eff 0
						indic_qtl_without_inter .
						inter_cofactor_id .				
						qtl1 &bestqtl1 
						qtl2 &bestqtl1 
						ch1 &bestchrom1
						ch2 &bestchrom1
						pos1 &bestpos1
						pos2 &bestpos1
						posid1 &bestposid1
						posid2 &bestposid1
						efftype1 2 
						efftype2 2
						contrast &case_
						InfinityLike &bestInfinityLike;
					run;

					data add1_q1;
						retain 
						trait &trt 
						cofactors "&xcovar"
						ncofactors &nx
						nqtl &numbqtlin
						neffect &numbeffin
						indic_epi_eff 0
						indic_qtl_without_inter .
						inter_cofactor_id .				
						qtl1 &bestqtl1 
						qtl2 &bestqtl1 
						ch1 &bestchrom1
						ch2 &bestchrom1
						pos1 &bestpos1
						pos2 &bestpos1
						posid1 &bestposid1
						posid2 &bestposid1
						efftype1 3 
						efftype2 3
						contrast &case_
						InfinityLike &bestInfinityLike;
					run;
				
				%end;
				%else %do;
					*number of qtl is kept the same*;
					%let numbqtlin= &numbqtlin;
					
					*updating number of qtl effects (add only the interaction)*;
					%let numbeffin = %sysevalf(&numbeffin + 1);
					
					%let bestqtl1= &IDsameqtl;					
				%end;

				*--- saving new two-way interaction effect ---*;
				data qxc_tmp;
					retain 
					trait &trt 
					cofactors "&xcovar"
					ncofactors &nx
					nqtl &numbqtlin
					neffect &numbeffin
					indic_epi_eff 0
					indic_qtl_without_inter .
					inter_cofactor_id &bestcofactor_id				
					qtl1 &bestqtl1 
					qtl2 &bestqtl1 
					ch1 &bestchrom1
					ch2 &bestchrom1
					pos1 &bestpos1
					pos2 &bestpos1
					posid1 &bestposid1
					posid2 &bestposid1
					efftype1 &bestqxctype1
					efftype2 &bestqxctype1
					contrast .
					InfinityLike &bestInfinityLike;
				run;				
				
				data &hkmimmodel;
					set &hkmimmodel 
					%if &sameQTLflag= 0 %then %do;
						%if %sysevalf(&ng=2) %then add2_q1;
						%if %sysevalf(&ng=3) %then add2_q1 dom_q1; 
						%if %sysevalf(&ng=4) %then %do;
							%if &case_ = 2 %then add1_q1; 
							%else %if &case_ = 3 %then add2_q1 dom_q1;  
							%else %if &case_ = 4 %then add1_q1 dom_q1;  
							%else %if &case_ = 5 %then add2_q1;  
							%else %if &case_ = 6 %then add2_q1 add1_q1;   
							%else %if &case_ = 7 %then add2_q1 dom_q1;   
							%else %if &case_ = 8 %then add2_q1 dom_q1 add1_q1; 	
						%end;
					%end;
					qxc_tmp
					;
					if trait= &trt and neffect > 0 then do;
						nqtl = &numbqtlin;
						neffect = &numbeffin;
					end;
				run;
				
				data hkmodel_tmp;
					set &hkmimmodel;
					where trait = &trt;
				run;

				proc datasets nolist;
					delete %if &sameQTLflag= 0 %then %do; add2_q1 dom_q1 add1_q1 %end; qxc_tmp; 
				quit;
				
			*end saving best model for one of the cofactors*;
			%end;

			%let z = %sysevalf(&z + 1);
			%*let cofvar = %scan(&xcovar,&z);
			%let cofvarid= %scan(%superQ(selcofactorsIDList), &z, %str(,));

			%*goto exitgsqxc;
			
		*end when loop for cofactors*;
		%end;
	%end;
	
	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp
			MappedGrid_tmp
			MappedGrid_tmpNoDup
			hkdata1
			hkdata3;
	quit;

	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel); 
	
	%exitgsqxc:;

	ods exclude none;
	ods results;
	*options source notes;		
	
%mend gscanqtlxcofactor;

/* gfwdqtlxcofactor() calls function gscanqtlxcofactor() consecutively until no more two-way intaraction between cofactor 
and putative QTL in the genome grid cannot be added into the MIM model */
%macro gfwdqtlxcofactor(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=, lod_threshold=, probf_entry=, teststep_qxc=);
	***this function uses global variable selcofactorsIDList that cames from Here:GFwdQTLxCofactorExpr***;
	
	%global flag_warn1_gsqxc warn1_gsqxc_message;
	%let flag_warn1_gsqxc=0;
	%let warn1_gsqxc_message=;
	
	%global flag_warn1_gfqxc warn1_gfqxc_message;	
	%let flag_warn1_gfqxc=0;	
	%let warn1_gfqxc_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbeffout;
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitgfwdqxc;
	%end;
	
	%if %symexist(selcofactorsIDList) = 0 %then %let selcofactorsIDList= ;
	
	%if %length(&selcofactorsIDList) = 0 %then %do;
		%put ERROR: List of selected cofactors does is empty.;
		%let exiterror=1;
		%goto exitgfwdqxc;	
	%end;	
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set &hkmimmodel(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	%put xcovar= &xcovar;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);
	
	%if &nx < 1 %then %do;
	
		%let flag_warn1_gfqxc = 1;
		%let warn1_gfqxc_message = The model does not have cofactors.;
		%put WARNING: &warn1_gfqxc_message;
			
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitgfwdqxc;
	%end;	
	
	*--- get number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitgfwdqxc;
	%end;
	
	%if &InfinityLikein %then %do;
		%let flag_warn1_gfqxc = 1;
		%let warn1_gfqxc_message = No more QTL effects can be added into the model because the residual variance is zero in the genomic forward search.;
		%put WARNING: &warn1_gfqxc_message;
		%goto exitgfwdqxc;
	%end;
	
	*--- search for two-way interaction between cofactor and putative QTL in the genome grid---*;
	%gscanqtlxcofactor(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, 
		hkmimmodel=&hkmimmodel, wsize=&wsize, criterion=&criterion, 
		lod_threshold=&lod_threshold, probf_entry=&probf_entry, 
		teststep_qxc=&teststep_qxc);		

	*--- get number of QTL and effects after the scan search ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%put numbeffout = &numbeffout;
	%put numbeffin = &numbeffin;

	%local b;
	%let b=0;
	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn1_gfqxc = 1;
		%let warn1_gfqxc_message = No significant Cofactor x QTL were found in the genomic search.;
		%put WARNING: &warn1_gfqxc_message;
	%end;
	%else %do %while(%sysevalf(&numbeffout > &numbeffin) and &InfinityLikeout=0);
		%let b= %sysevalf(&b + 1);
		%put b= &b;
		%put numbeffout = &numbeffout;
		%put numbeffin = &numbeffin;

		*--- search for new putative QTL until no more QTL can be found ---*;
		%let numbeffin = &numbeffout;
			
		*--- search for two-way interaction between cofactor and putative QTL in the genome grid---*;
		%gscanqtlxcofactor(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, 
			hkmimmodel=&hkmimmodel, wsize=&wsize, criterion=&criterion, 
			lod_threshold=&lod_threshold, probf_entry=&probf_entry, 
			teststep_qxc=&teststep_qxc);				
				
		*--- get number of QTL after the scan search ---*;
		proc means data=&hkmimmodel noprint;
			where trait=&trt;
			var nqtl neffect InfinityLike;
			output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
		run;
		
		data _null_;
			set nqtlout_tmp;
			call symputx("numbeffout", numbeff, 'L');
			call symputx("InfinityLikeout", InfinityLike, 'L');
		run;
		
		*--- refine model before exiting the while loop ---*;
		%if &InfinityLikeout = 1 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);	
			%let IsRefined= 1;
		%end;
		%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
			%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
				wsize=&wsize, criterion=&criterion);
			*setting IsRefined to 1 to stop refining QTL position*;
			%let IsRefined= 1;
			*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
			%let numbeffin= %sysevalf(&numbeffin - 1);
		%end;
		%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
			*setting IsRefined to 0 to allow refining QTL position*;
			%let IsRefined= 0;
		%end;
		
	%end; *--- end while loop over search for new putative QTL ---*;
	
	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;
	
	*--- deleting temporary data sets ---*;
	proc datasets nolist;
		delete nqtlin_tmp nqtlout_tmp;
	quit;	
	
	*gscanqtlxcofactor() includes in the model only the best interaction effect (e.g., a*x1, or d*x1, etc.) between a qtl and cofactor**;
	*To search for other interaction effects between qtl and cofactor, we need to call mfwdqtlxcofactor()*;	
	%if &flag_warn1_gfqxc = 0 %then %do;			
		%mfwdqtlxcofactor(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, hkmimmodel=&hkmimmodel, 
			criterion=&criterion, lod_threshold=&lod_threshold, probf_entry=&probf_entry, 
			mapgrid=&mapgrid, wsize=&wsize);			
	%end;
	
	%exitgfwdqxc:;
	
%mend gfwdqtlxcofactor;

/* dropqxceff() tests each Cofactor x QTL effect (one degree of freedom). The argument AUTO gives the option to remove or 
not the lesser significant interaction effect from model if the test does not exceed a threshold. Its default value is TRUE. */
%macro dropqxceff(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_stay=, auto= TRUE);

	ods exclude all;
	ods noresults;

	%global flag_warn1_dqxc warn1_dqxc_message;
	%global flag_warn2_dqxc warn2_dqxc_message;
	%let flag_warn1_dqxc=0;
	%let warn1_dqxc_message=;
	%let flag_warn2_dqxc=0;
	%let warn2_dqxc_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local criterionname;
	%local criterionvar;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp; 
	%local numbqtlin;
	%local numbeffin;
	%local geffect_main_in;
	%local geffect_epi_in;	
	%local geffect_qxc_in;	
	%local currvalue;
	%local droppedeff;
	%local droppedtype;
	%local droppedtype2;
	%local qxc_qtl;
	%local qxc_cof;
	%local worstvalue;
	%local threshold;
	%local id_getqxc;
	%local NOBS_id_getqxc;
	%local orderdirection;
		
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitdqxc;
	%end;

	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/
	
	*---get criterion to select best genomic position---*;	
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then %do;
		%let criterionvar = LOD_IndTest;
		%let criterionname = LOD;
	%end;
	%else %if (&criterion=AIC) %then %do;
		%let criterionvar = AIC_IndTest;
		%let criterionname = AIC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=AICC) %then %do;
		%let criterionvar = AICC_IndTest;
		%let criterionname = AICC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=BIC) %then %do;
		%let criterionvar = BIC_IndTest;
		%let criterionname = BIC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=ProbF) %then %do;
		%let criterionvar = Probt_IndTest;
		%let criterionname = PROBT;
	%end;

	*--- create data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;
	
	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;
		%let exiterror=1;
	%end;
	%else %do;
	
		*---get covariates in the input model---*;
		%let xcovar = ;
		data _null_;
			set hkmodel_tmp(obs=1);
			call symputx("xcovar", trim(left(cofactors)), 'L');
		run;	
		
		%let u = 1;
		%let cofvar= %scan(&xcovar,&u);
		%do %while(%length(&cofvar));		
			%let u= %sysevalf(&u + 1);		
			%let cofvar= %scan(&xcovar,&u);
		%end;
		%let nx= %sysevalf(&u - 1);	
	
		*--- get the number of QTL and effects in the input model ---*;
		proc means data=hkmodel_tmp noprint;
			var nqtl neffect InfinityLike;
			output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
		run;

		data _null_;
			set nqtlin_tmp;
			call symputx("numbqtlin", numbqtl, 'L');
			call symputx("numbeffin", numbeff, 'L');
			call symputx("InfinityLikein", InfinityLike, 'L');
		run;	

		proc datasets nolist;
			delete nqtlin_tmp;
		quit;

		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		%*put # # # Number of QTL = &numbqtlin # # # # # #;
		%*put # # # Number of Effects = &numbeffin # # # #;
		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		
		data qxc_del;
			set _null_;
		run;
		
		%if %sysevalf(&numbqtlin < 1) %then %do;
			
			%let flag_warn1_dqxc = 1;
			%let warn1_dqxc_message= The procecure for the elimination of Cofactor x QTL interaction effects needs a model with some Cofactor x QTL interaction.;
			%put WARNING: &warn1_dqxc_message;  
							
		%end;
		%else %do;
			*--- get effects already in the model ---*;
			%let geffect_main_in=;
			%let geffect_epi_in =;
			%let geffect_qxc_in =;

			proc sql;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
				
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
				
				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;				
			quit;
			%*put geffect_main_in= &geffect_main_in;
			%*put geffect_epi_in= &geffect_epi_in;
			%*put geffect_qxc_in= &geffect_qxc_in;
			
			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);

			*--- fit input model ---*;
			proc mixed data=hkdata3 method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
				ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
			run;
			
			%let currvalue = ;
			data _null_;
				set st_tmp1;
				if upcase(descr) = "&criterionname" then 
					call symputx("currvalue", trim(left(value)), 'L');
			run;
			
			*--- delete temporary data sets ---;
			proc datasets nolist;
				delete st_tmp1 es_tmp1;
			quit;
	
			*--- get parameter estimates, probability and fitting measures (AIC, AICC, BIC, LOD) for each QTL ---*;
			%fitmodel(trt=&trt, ng= &ng, hkmimdata=&hkmimdata, hkmimmodel=&hkmimmodel);
	
			*--- get cofactor x qtl effects (if any) ---*;
			%let orderdirection = ;
			%if "&criterionname" = "PROBT" %then 
				%let orderdirection = descending;
				
			%put criterionvar= &criterionvar;
			%put orderdirection= &orderdirection;
			*fitmodel_sum comes from fitmodel()*;
			proc sql;
				create table getqxc as
				select a.* 
				from fitmodel_sum as a
				where trim(left(Effect_type)) = 'Interaction' and trim(left(Predictors)) = "QTL Effect" and inter_cofactor_id ^= .
				order by &criterionvar &orderdirection;
			quit;
			
			*data OutLib.qxcfitmodel_sum; *set fitmodel_sum; *run;
			
			%*goto exitdqxc;
			
			%let id_getqxc= %sysfunc(open(getqxc));
			%let NOBS_id_getqxc = %sysfunc(attrn(&id_getqxc,NOBS));		
			%let id_getqxc= %sysfunc(close(&id_getqxc)); *--- very important to close ---*;
			%if &NOBS_id_getqxc < 1 %then %do;
			
				%let flag_warn2_dqxc = 1;	
				%let warn2_dqxc_message= The elimination procedure of Cofactor x QTL interaction effects needs a model with least 1 Cofactor x QTL interaction.;
				%put WARNING: &warn2_dqxc_message;  
									
			%end;
			%else %do;
				
				%if &criterion = LOD %then %do;
					%let threshold = &lod_threshold;
				%end;
				%else %if &criterion = ProbF %then %do;
					%let threshold = &probf_stay;
				%end;
				%else %do;
					%let threshold = &currvalue;
				%end;
				%*put criterion = &criterion;
				%*put criterionname = &criterionname;
				%*put threshold = &threshold;

				%let droppedeff = ;
				%let qxc_qtl =;
				%let qxc_cof =;
				data qxc_less_sig;
					set getqxc(obs=1);					
					if "&criterionname" ^= "PROBT" and &criterionvar < &threshold then do;
						output;
						call symputx('droppedeff', trim(left(Effect)), 'L');
						call symputx('qxc_qtl', trim(left(qtl1)), 'L');
						call symputx('qxc_cof', trim(left(inter_cofactor_id)), 'L');
					end;
					else if "&criterionname" = "PROBT" and &criterionvar > &threshold then do;
						output;
						call symputx('droppedeff', trim(left(Effect)), 'L');						
						call symputx('qxc_qtl', trim(left(qtl1)), 'L');
						call symputx('qxc_cof', trim(left(inter_cofactor_id)), 'L');
					end;
					call symputx('worstvalue', &criterionvar, 'L');
				run;	
				%put ## Drop Cofactor x QTL Effect ##;
				%put criterionname = &criterionname;
				%put criterionvar = &criterionvar;
				%put criterion = &criterion;
				%put threshold = &threshold;
				%put worstvalue = &worstvalue;
				%put droppedeff = &droppedeff;
				%put qxc_qtl = &qxc_qtl; 
				%put qxc_cof = &qxc_cof;
				
				%if &ng < 4 %then %do;
					%if %index(&droppedeff, Additive) %then %do;
						%let droppedtype = 1;
					%end;
					%else %if %index(&droppedeff, Dominance) %then %do;
						%let droppedtype = 2;
					%end; 
				%end;
				%else %do;
					***assuming maximum ng = 4 ***;
					%if %index(&droppedeff, Additive2) %then %do;
						%let droppedtype = 1;
					%end;
					%else %if %index(&droppedeff, Dominance) %then %do;
						%let droppedtype = 2;
					%end; 
					%else %if %index(&droppedeff, Additive1) %then %do;
						%let droppedtype = 3;
					%end; 
				%end;				
				%put droppedtype = &droppedtype;
				
				*--- remove lesser significant cofactor x qtl effect, conditional on a threshold ---*;				
				%if %UPCASE(&auto) = %UPCASE(TRUE) %then %do;
					
					%if(%length(&droppedeff)) %then %do;
					
						data qxc_del;
							set qxc_less_sig;
						run;
					
						data &hkmimmodel;
							set &hkmimmodel;
							if trait = &trt and qtl1 = &qxc_qtl and qtl2 = &qxc_qtl 
								and efftype1 = &droppedtype and efftype2 = &droppedtype 
								and inter_cofactor_id = &qxc_cof
							then delete;
							else if neffect > 0 then do;
								neffect = neffect - 1;
							end;							
						run;

					%end; *--- end if droppedeff1 is true ---*;	
				%end; *--- end if auto is true ---*;
				
				proc datasets nolist;
					delete qxc_less_sig;
				quit;
				
			%end; *--- end if NOBS_id_getqxc >= 1 ---*;
			
			*--- delete temporary data sets ---*;
			proc datasets nolist;
				delete fitmodel_sum
					fitmodel_st
					interc_cov
					getqxc;			
			quit;
			
			*--- re-number QTL after sorting by chromosome and position ---*;
			%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);				
						
		%end; *--- end if numbqtlin > 0 ---*;
	%end; *--- end if NOBS_hkmodel_tmp >=1 ---*;
	
	proc datasets nolist;
	        delete hkmodel_tmp
	        	hkdata3;
	quit;

	ods exclude none;
	ods results;

	%exitdqxc:;

%mend dropqxceff;

/* backdropqxceff() tests each Cofactor x QTL effect. The argument AUTO gives the option to remove or not the lesser significant 
interaction effect from the model if the test does not exceed a threshold. Its default value is TRUE. If AUTO= TRUE, this function 
keeps removing interaction effects if they don't exceed a threshold*/
%macro backdropqxceff(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_stay=, auto=, mapgrid=, wsize=);

	%global flag_warn1_bdqxc warn1_bdqxc_message;
	%let flag_warn1_bdqxc = 0;	
	%let warn1_bdqxc_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbqtlout;
	%local numbeffout; 
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;

	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitbdqxc;
	%end;
				
	*--- get number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;

	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
		
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;
		%let exiterror=1;
		%goto exitbdqxc;
	%end;
	
	*--- call the procedure to drop a non-significant cofactor x QTL effect ---*; 
	%dropqxceff(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
		criterion= &criterion, lod_threshold= &lod_threshold, probf_stay= &probf_stay, 
		auto= &auto);

	%if %sysfunc(exist(dqxc_qxc_del)) %then %do;
		data qxc_del_all;
			set dqxc_qxc_del;
		run;
	%end;
	
	*--- get number of QTL and effects after trying to remove cofactor x QTL effect ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbqtlout", numbqtl, 'L');
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%put numbqtlin = &numbqtlin;
	%put numbqtlout = &numbqtlout;
	%put numbeffin = &numbeffin;
	%put numbeffout = &numbeffout;	

	%local b;
	%let b=0;
	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn1_bdqxc = 1;
		%let warn1_bdqxc_message= No Cofactor x QTL interaction effects were removed from the model.;
		%put WARNING: &warn1_bdqxc_message;
	%end;
	%else %do;
		%do %while(%sysevalf(&numbeffout < &numbeffin) and &InfinityLikeout=0);
			%let b= %sysevalf(&b + 1);
			%put b= &b;
			%put numbeffout = &numbeffout;
			%put numbeffin = &numbeffin;			

			%let numbeffin = &numbeffout;

			*--- detele data set to avoid pilling results from previous runs ---*;
			proc datasets nolist;
				delete dqxc_qxc_del;
			quit;

			*--- call the procedure to drop a non-significant QTL that has main effects only ---*; 
			%dropqxceff(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
				criterion= &criterion, lod_threshold= &lod_threshold, probf_stay= &probf_stay, 
				auto= &auto);

			%if %sysfunc(exist(dqxc_qxc_del)) %then %do;
				data qxc_del_all;
					set qxc_del_all dqxc_qxc_del;
				run;
			%end;	

			*--- get number of QTL and effects after trying to remove cofactor x QTL effect ---*;
			proc means data=&hkmimmodel noprint;
				where trait=&trt;
				var nqtl neffect InfinityLike;
				output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
			run;

			data _null_;
				set nqtlout_tmp;
				call symputx("numbqtlout", numbqtl, 'L');
				call symputx("numbeffout", numbeff, 'L');
				call symputx("InfinityLikeout", InfinityLike, 'L');
			run;

			%put numbeffout = &numbeffout;
			%put numbeffin = &numbeffin;
			
			*--- refine model before exiting the while loop ---*;
			%if &InfinityLikeout = 1 %then %do;
				%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
					wsize=&wsize, criterion=&criterion);	
				%let IsRefined= 1;
			%end;
			%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
				%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
					wsize=&wsize, criterion=&criterion);
				*setting IsRefined to 1 to stop refining QTL position*;
				%let IsRefined= 1;
				*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
				%let numbeffin= %sysevalf(&numbeffin + 1);
			%end;
			%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
				*setting IsRefined to 0 to allow refining QTL position*;
				%let IsRefined= 0;
			%end;			

		%end; *--- end while loop ---*;
		
		%let flag_warn1_dqxc = 0; /*comes from macro dropqxceff()*/
		%let flag_warn2_dqxc = 0; /*comes from macro dropqxceff()*/
	%end;
	
	*--- delete temporary data sets ---*;
	proc datasets nolist;
		delete 	nqtlin_tmp
			nqtlout_tmp;
	quit;
	
	*--- delete temporary data sets ---*;
	proc datasets nolist;	
		delete dqxc_qxc_del;
	quit;
	
	%exitbdqxc:;
	
%mend backdropqxceff;

/* dropmainqtlonly() jointly tests main effects of each of those QTL that have main effects only. 
If genotypes have two classes, then the test has 1 degree of freedom, if three classes, then 2 degrees of freemom. 
The argument AUTO gives the option to remove or not the lesser significant QTL from the model if the test does not
exceed a threshold. Its default value is TRUE. */
%macro dropmainqtlonly(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_stay=, auto= TRUE);

	ods exclude all;
	ods noresults;

	%global flag_warn1_dmo warn1_dmo_message;
	%global flag_warn2_dmo warn2_dmo_message;
	%let flag_warn1_dmo=0;
	%let warn1_dmo_message=;
	%let flag_warn2_dmo=0;
	%let warn2_dmo_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local criterionname;
	%local criterionvar;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp; 
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local currvalue;
	%local orderdirection;
	%local droppedqtl;
	%local case_;
	%local worstvalue;
	%local threshold;
	%local id_getmqtlo;
	%local NObs_id_getmqtlo;
		
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitdmqo;
	%end;

	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/
	
	*---get criterion to select best genomic position---*;	
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then %do;
		%let criterionvar = LOD_JointMain;
		%let criterionname = LOD;
	%end;
	%else %if (&criterion=AIC) %then %do;
		%let criterionvar = AIC_JointMain;
		%let criterionname = AIC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=AICC) %then %do;
		%let criterionvar = AICC_JointMain;
		%let criterionname = AICC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=BIC) %then %do;
		%let criterionvar = BIC_JointMain;
		%let criterionname = BIC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=ProbF) %then %do;
		%let criterionvar = ProbF_JointMain;
		%let criterionname = PROBF;
	%end;
	
	*--- create data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;	
		%let exiterror=1;
	%end;
	%else %do;
	
		*---get covariates in the input model---*;
		%let xcovar = ;
		data _null_;
			set hkmodel_tmp(obs=1);
			call symputx("xcovar", trim(left(cofactors)), 'L');
		run;
		
		%let u = 1;
		%let cofvar= %scan(&xcovar,&u);
		%do %while(%length(&cofvar));		
			%let u= %sysevalf(&u + 1);		
			%let cofvar= %scan(&xcovar,&u);
		%end;
		%let nx= %sysevalf(&u - 1);	
	
		*--- get the number of QTL and effects in the input model ---*;
		proc means data=hkmodel_tmp noprint;
			var nqtl neffect InfinityLike;
			output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
		run;

		data _null_;
			set nqtlin_tmp;
			call symputx("numbqtlin", numbqtl, 'L');
			call symputx("numbeffin", numbeff, 'L');
			call symputx("InfinityLikein", InfinityLike, 'L');
		run;	

		proc datasets nolist;
			delete nqtlin_tmp;
		quit;

		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		%*put # # # Number of QTL = &numbqtlin # # # # # #;
		%*put # # # Number of Effects = &numbeffin # # # #;
		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		
		data mqtlo_del;
			set _null_;
		run;
		
		%if %sysevalf(&numbqtlin < 1) %then %do;
		
			%let flag_warn1_dmo = 1;
			%let warn1_dmo_message= The elimination procedure of QTL main effects needs a model with some QTL.;
			%put WARNING: &warn1_dmo_message;  
							
		%end;
		%else %do;
			*--- get effects already in the model ---*;
			%let geffect_main_in=;
			%let geffect_epi_in =;
			%let geffect_qxc_in=;

			proc sql;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
				
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
				
				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;				
			quit;
			%*put geffect_main_in= &geffect_main_in;
			%*put geffect_epi_in= &geffect_epi_in;
			%*put geffect_qxc_in= &geffect_qxc_in;
			
			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);

			*--- fit input model ---*;
			proc mixed data=hkdata3 method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in / solution;
				ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
			run;
			
			%let currvalue = ;
			data _null_;
				set st_tmp1;
				if upcase(descr) = "&criterionname" then 
					call symputx("currvalue", trim(left(value)), 'L');
			run;
			%put currvalue= &currvalue;
			
			*data outlib.st_tmp1; *set st_tmp1; *run;
			
			*--- delete temporary data sets ---*;
			proc datasets nolist;
				delete st_tmp1 es_tmp1;
			quit;
	
			*--- get parameter estimates, probability and fitting measures (AIC, AICC, BIC, LOD) for each QTL ---*;
			%fitmodel(trt=&trt, ng= &ng, hkmimdata=&hkmimdata, hkmimmodel=&hkmimmodel);
	
			*--- get those QTL that have main effects only ---*;
			%let orderdirection = ;
			%if "&criterionname" = "PROBF" %then 
				%let orderdirection = descending;
			
			*fitmodel_sum comes from fitmodel()*;			
			proc sql;
				create table getmqtlo as
				select distinct a.*
				from fitmodel_sum as a, hkmodel_tmp as b
				where a.qtl1 = b.qtl1 and a.qtl2= b.qtl2 and b.indic_qtl_without_inter = 1 and &criterionvar ^= . 
				order by &criterionvar &orderdirection;
			quit;			
			
			%*goto exitdmqo;
						
			%let id_getmqtlo= %sysfunc(open(getmqtlo));
			%let NOBS_id_getmqtlo = %sysfunc(attrn(&id_getmqtlo,NOBS));		
			%let id_getmqtlo= %sysfunc(close(&id_getmqtlo)); *--- very important to close ---*;
			%if &NOBS_id_getmqtlo < 1 %then %do;
			
				%let flag_warn2_dmo = 1;
				%let warn2_dmo_message = The elimination procedure of QTL main effects needs a model with at least 1 QTL with main effects only.;
				%put WARNING: &warn2_dmo_message;  
									
			%end;
			%else %do;
				
				%if &criterion = LOD %then %do;
					%let threshold = &lod_threshold;
				%end;
				%else %if &criterion = ProbF %then %do;
					%let threshold = &probf_stay;
				%end;
				%else %do;
					%let threshold = &currvalue;
				%end;
				%*put criterion = &criterion;
				%*put criterionname = &criterionname;
				%*put threshold = &threshold;
 				
 				%let case_ = ;
				%let droppedqtl = ;
				%let worstvalue = ;
				
				*data outlib.getmqtlo; *set getmqtlo; *run;
				
				data mqtlo_less_sig;
					set getmqtlo(obs=1);					
					if "&criterionname" ^= "PROBF" and &criterionvar < &threshold then do;
						call symputx('droppedqtl', trim(left(qtl1)), 'L');
						call symputx('case_', trim(left(contrast_)), 'L');
						output;
					end;
					else if "&criterionname" = "PROBF" and &criterionvar > &threshold then do;
						call symputx('droppedqtl', trim(left(qtl1)), 'L');
						call symputx('case_', trim(left(contrast_)), 'L');
						output;
					end;	
					call symputx('worstvalue', &criterionvar, 'L');
				run;	
				%put ## Drop Main Effect Only ##;
				%put criterionname = &criterionname;
				%put criterionvar = &criterionvar;
				%put criterion = &criterion;
				%put threshold = &threshold;
				%put worstvalue = &worstvalue;
				%put droppedqtl = &droppedqtl;
				%*put case_ = &case_;
				
				%*goto exitdmqo;
			
				*--- remove lesser significant QTL that has main effects only, conditional on a threshold ---*;
				%if %UPCASE(&auto) = %UPCASE(TRUE) %then %do;
					
					%if(%length(&droppedqtl)) %then %do;
						
						data mqtlo_del;
							set mqtlo_less_sig;
						run;

						
						data &hkmimmodel;
							set &hkmimmodel;
							if trait = &trt and qtl1 = &droppedqtl and qtl2 = &droppedqtl then delete;
							else if neffect > 0 then do;
								nqtl = nqtl - 1;
								%if &ng = 2 %then %do;
									neffect = neffect - 1;
								%end;
								%else %if &ng = 3 %then %do;
									neffect = neffect - 2;
								%end;
								%else %do;
									%if &case_ = 2 %then neffect = neffect - 1;
									%else %if &case_ = 3 %then neffect = neffect - 2;
									%else %if &case_ = 4 %then neffect = neffect - 2;
									%else %if &case_ = 5 %then neffect = neffect - 1;
									%else %if &case_ = 6 %then neffect = neffect - 2;
									%else %if &case_ = 7 %then neffect = neffect - 2;
									%else %if &case_ = 8 %then neffect = neffect - 3;
									;
								%end;
							end;
						run;
						
					%end; *--- end if droppedqtl is true ---*;	
				%end; *--- end if auto is true ---*;
			
				proc datasets nolist;
					delete mqtlo_less_sig;
				quit;
		
			%end; *--- end if NOBS_id_getmqtlo >= 1 ---*;
			
			*--- delete temporary data sets ---*;	
			proc datasets nolist;
				delete fitmodel_sum
					fitmodel_st
					interc_cov
					getmqtlo;
			quit;
			
			
		%end; *--- end if numbqtlin > 0 ---*;
	%end; *--- end if NOBS_hkmodel_tmp >=1 ---*;

	proc datasets nolist;
		delete hkmodel_tmp
			hkdata3;
	quit;
	
	%*goto exitdmqo;

	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel); 
	
	ods exclude none;
	ods results;	
	
	%exitdmqo:;
	
%mend dropmainqtlonly;

/* backdropmainqtlonly() jointly tests main effects of each of those QTL that have main effects only. If genotypes have 
two classes, then the test has 1 degree of freedom, if three classes, then 2 degrees of freemom. The argument AUTO gives 
the option to remove or not the lesser significant QTL from the model if the test does not exceed a threshold. Its default 
value is TRUE. This function keeps removing QTL with effects only if they don't exceed a threshold.*/
%macro backdropmainqtlonly(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_stay=, auto=, mapgrid=, wsize=);

	%global flag_warn1_bdmo warn1_bdmo_message;
	%let flag_warn1_bdmo=0;
	%let warn1_bdmo_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbqtlout;
	%local numbeffout; 
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;

	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitbdmqo;
	%end;
				
	*--- get number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;

	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
		
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;
	
	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;
		%let exiterror=1;
		%goto exitbdmqo;
	%end;	
	
	*--- call the procedure to drop a non-significant QTL that has main effects only ---*; 
	%dropmainqtlonly(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
		criterion= &criterion, lod_threshold= &lod_threshold, probf_stay= &probf_stay, auto= &auto);
	
	%if %sysfunc(exist(dmqo_mqtlo_del)) %then %do;
		data qtlmo_del_all;
			set dmqo_mqtlo_del;
		run;
	%end;	
	
	*--- get number of QTL and effects after trying to remove a QTL that has main effects only ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbqtlout", numbqtl, 'L');
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%*put numbqtlin = &numbqtlin;
	%*put numbqtlout = &numbqtlout;

	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn1_bdmo = 1;
		%let warn1_bdmo_message= No QTL main effects were removed from the model.;
		%put WARNING: &warn1_bdmo_message;
	%end;
	%else %do;
		%do %while(%sysevalf(&numbeffout < &numbeffin) and &InfinityLikeout=0);

			%let numbeffin = &numbeffout;

			*--- detele data set to avoid pilling results from previous runs ---*;
			proc datasets nolist;
				delete dmqo_mqtlo_del;
			quit;

			*--- call the procedure to drop a non-significant QTL that has main effects only ---*; 
			%dropmainqtlonly(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
				criterion= &criterion, lod_threshold= &lod_threshold, probf_stay= &probf_stay, auto= &auto);

			%if %sysfunc(exist(dmqo_mqtlo_del)) %then %do;
				data qtlmo_del_all;
					set qtlmo_del_all dmqo_mqtlo_del;
				run;
			%end;		

			*--- get number of QTL and effects after trying to remove a QTL that has main effects only ---*;
			proc means data=&hkmimmodel noprint;
				where trait=&trt;
				var nqtl neffect InfinityLike;
				output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
			run;

			data _null_;
				set nqtlout_tmp;
				call symputx("numbqtlout", numbqtl, 'L');
				call symputx("numbeffout", numbeff, 'L');
				call symputx("InfinityLikeout", InfinityLike, 'L');
			run;
			
			*--- refine model before exiting the while loop ---*;
			%if &InfinityLikeout = 1 %then %do;
				%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
					wsize=&wsize, criterion=&criterion);	
				%let IsRefined= 1;
			%end;
			%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
				%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
					wsize=&wsize, criterion=&criterion);
				*setting IsRefined to 1 to stop refining QTL position*;
				%let IsRefined= 1;
				*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
				%let numbeffin= %sysevalf(&numbeffin + 1);
			%end;
			%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
				*setting IsRefined to 0 to allow refining QTL position*;
				%let IsRefined= 0;
			%end;

		%end; *--- end while loop ---*;
		
		%let flag_warn1_dmo = 0; /*created within macro dropmainqtlonly()*/
		%let flag_warn2_dmo = 0; /*created within macro dropmainqtlonly()*/
	%end;

	*--- delete temporary data sets ---*;
	proc datasets nolist;
		delete nqtlin_tmp
			nqtlout_tmp;
	quit;
	
	*--- delete temporary data sets ---*;
	proc datasets nolist;
		delete dmqo_mqtlo_del;
	quit;
	
	%exitbdmqo:;

%mend backdropmainqtlonly;

/* dropepieff() tests each epistatic effect (one degree of freedom). The argument AUTO gives the option to remove or 
not the lesser significant epistatic effect from model if the test does not exceed a threshold. Its default value is TRUE. */
%macro dropepieff(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_stay=, auto= TRUE);

	ods exclude all;
	ods noresults;

	%global flag_warn1_de warn1_de_message;
	%global flag_warn2_de warn2_de_message;
	%let flag_warn1_de=0;
	%let flag_warn2_de=0;
	%let warn1_de_message=;
	%let warn2_de_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local criterionname;
	%local criterionvar;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp; 
	%local numbqtlin;
	%local numbeffin;
	%local geffect_main_in;
	%local geffect_epi_in;	
	%local geffect_qxc_in;	
	%local currvalue;
	%local droppedeff;
	%local droppedtype1;
	%local droppedtype2;
	%local epi_qtl1;
	%local epi_qtl2;
	%local worstvalue;
	%local threshold;
	%local id_getepi;
	%local NOBS_id_getepi;
	%local orderdirection;
		
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitdepi;
	%end;

	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/
	
	*---get criterion to select best genomic position---*;	
	%let criterion = %sysfunc(trim(%left(&criterion)));
	%if (&criterion=LOD) %then %do;
		%let criterionvar = LOD_IndTest;
		%let criterionname = LOD;
	%end;
	%else %if (&criterion=AIC) %then %do;
		%let criterionvar = AIC_IndTest;
		%let criterionname = AIC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=AICC) %then %do;
		%let criterionvar = AICC_IndTest;
		%let criterionname = AICC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=BIC) %then %do;
		%let criterionvar = BIC_IndTest;
		%let criterionname = BIC (SMALLER IS BETTER);
	%end;
	%else %if (&criterion=ProbF) %then %do;
		%let criterionvar = Probt_IndTest;
		%let criterionname = PROBT;
	%end;

	*--- create data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;
	
	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;
		%let exiterror=1;
	%end;
	%else %do;
	
		*---get covariates in the input model---*;
		%let xcovar = ;
		data _null_;
			set hkmodel_tmp(obs=1);
			call symputx("xcovar", trim(left(cofactors)), 'L');
		run;	
		
		%let u = 1;
		%let cofvar= %scan(&xcovar,&u);
		%do %while(%length(&cofvar));		
			%let u= %sysevalf(&u + 1);		
			%let cofvar= %scan(&xcovar,&u);
		%end;
		%let nx= %sysevalf(&u - 1);	
	
		*--- get the number of QTL and effects in the input model ---*;
		proc means data=hkmodel_tmp noprint;
			var nqtl neffect InfinityLike;
			output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
		run;

		data _null_;
			set nqtlin_tmp;
			call symputx("numbqtlin", numbqtl, 'L');
			call symputx("numbeffin", numbeff, 'L');
			call symputx("InfinityLikein", InfinityLike, 'L');
		run;	

		proc datasets nolist;
			delete nqtlin_tmp;
		quit;

		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		%*put # # # Number of QTL = &numbqtlin # # # # # #;
		%*put # # # Number of Effects = &numbeffin # # # #;
		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		
		data epe_del;
			set _null_;
		run;
		
		%if %sysevalf(&numbqtlin < 1) %then %do;
			
			%let flag_warn1_de = 1;	
			%let warn1_de_message = The procecure for the elimination of QTL x QTL effects needs a model with some QTL.;
			%put WARNING: &warn1_de_message;  
							
		%end;
		%else %do;
			*--- get effects already in the model ---*;
			%let geffect_main_in=;
			%let geffect_epi_in =;
			%let geffect_qxc_in =;

			proc sql;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
				
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
				
				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;				
			quit;
			%*put geffect_main_in= &geffect_main_in;
			%*put geffect_epi_in= &geffect_epi_in;
			%*put geffect_qxc_in= &geffect_qxc_in;
			
			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);

			*--- fit input model ---*;
			proc mixed data=hkdata3 method=ml;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
				ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
			run;
			
			%let currvalue = ;
			data _null_;
				set st_tmp1;
				if upcase(descr) = "&criterionname" then 
					call symputx("currvalue", trim(left(value)), 'L');
			run;
			
			*--- delete temporary data sets ---;
			proc datasets nolist;
				delete st_tmp1 es_tmp1;
			quit;
	
			*--- get parameter estimates, probability and fitting measures (AIC, AICC, BIC, LOD) for each QTL ---*;
			%fitmodel(trt=&trt, ng= &ng, hkmimdata=&hkmimdata, hkmimmodel=&hkmimmodel);
	
			*--- get epistatic effects (if any) ---*;
			%let orderdirection = ;
			%if "&criterionname" = "PROBT" %then 
				%let orderdirection = descending;
				
			%*goto exitdepi;
			
			*fitmodel_sum comes from fitmodel()*;
			proc sql;
				create table getepi as
				select a.* 
				from fitmodel_sum as a
				where trim(left(Effect_type)) = 'Interaction' and trim(left(Predictors)) = "QTL Effect" and inter_cofactor_id = .
				order by &criterionvar &orderdirection;
			quit;
			
			%let id_getepi= %sysfunc(open(getepi));
			%let NOBS_id_getepi = %sysfunc(attrn(&id_getepi,NOBS));		
			%let id_getepi= %sysfunc(close(&id_getepi)); *--- very important to close ---*;
			%if &NOBS_id_getepi < 1 %then %do;
			
				%let flag_warn2_de = 1;
				%let warn2_de_message= The elimination procedure of QTL x QTL effects needs a model with least 1 interaction effect.;
				%put WARNING: &warn2_de_message;  
									
			%end;
			%else %do;
				
				%if &criterion = LOD %then %do;
					%let threshold = &lod_threshold;
				%end;
				%else %if &criterion = ProbF %then %do;
					%let threshold = &probf_stay;
				%end;
				%else %do;
					%let threshold = &currvalue;
				%end;
				%*put criterion = &criterion;
				%*put criterionname = &criterionname;
				%*put threshold = &threshold;

				%let droppedeff = ;
				%let epi_qtl1 =;
				%let epi_qtl2 =;
				data epe_less_sig;
					set getepi(obs=1);					
					if "&criterionname" ^= "PROBT" and &criterionvar < &threshold then do;
						output;
						call symputx('droppedeff', trim(left(Effect)), 'L');
						call symputx('epi_qtl1', trim(left(qtl1)), 'L');
						call symputx('epi_qtl2', trim(left(qtl2)), 'L');
					end;
					else if "&criterionname" = "PROBT" and &criterionvar > &threshold then do;
						output;
						call symputx('droppedeff', trim(left(Effect)), 'L');						
						call symputx('epi_qtl1', trim(left(qtl1)), 'L');
						call symputx('epi_qtl2', trim(left(qtl2)), 'L');
					end;
					call symputx('worstvalue', &criterionvar, 'L');
				run;	
				%put ## Drop Epistasis Effect ##;
				%*put criterionname = &criterionname;
				%*put criterionvar = &criterionvar;
				%put criterion = &criterion;
				%put threshold = &threshold;
				%put worstvalue = &worstvalue;
				%put droppedeff = &droppedeff;
				%*put epi_qtl1 = &epi_qtl1; 
				%*put epi_qtl2 = &epi_qtl2;
				
				%if &ng < 4 %then %do;
					%if "&droppedeff" = "Additive * Additive" %then %do;
						%let droppedtype1 = 1;
						%let droppedtype2 = 1;
					%end;
					%else %if "&droppedeff" = "Additive * Dominance" %then %do;
						%let droppedtype1 = 1;
						%let droppedtype2 = 2;
					%end; 
					%else %if "&droppedeff" = "Dominance * Additive" %then %do;
						%let droppedtype1 = 2;
						%let droppedtype2 = 1;
					%end; 
					%else %if "&droppedeff" = "Dominance * Dominance" %then %do;
						%let droppedtype1 = 2;
						%let droppedtype2 = 2;
					%end; 			
				%end;
				%else %do;
					***assuming maximum ng = 4 ***;
					%if "&droppedeff" = "Additive2 * Additive2" %then %do;
						%let droppedtype1 = 1;
						%let droppedtype2 = 1;
					%end;
					%else %if "&droppedeff" = "Additive2 * Dominance" %then %do;
						%let droppedtype1 = 1;
						%let droppedtype2 = 2;
					%end; 
					%else %if "&droppedeff" = "Dominance * Additive2" %then %do;
						%let droppedtype1 = 2;
						%let droppedtype2 = 1;
					%end; 
					%else %if "&droppedeff" = "Dominance * Dominance" %then %do;
						%let droppedtype1 = 2;
						%let droppedtype2 = 2;
					%end; 
					%else %if "&droppedeff" = "Additive1 * Additive1" %then %do;
						%let droppedtype1 = 3;
						%let droppedtype2 = 3;
					%end; 
					%else %if "&droppedeff" = "Additive1 * Additive2" %then %do;
						%let droppedtype1 = 3;
						%let droppedtype2 = 1;
					%end; 	
					%else %if "&droppedeff" = "Additive1 * Dominance" %then %do;
						%let droppedtype1 = 3;
						%let droppedtype2 = 2;
					%end; 	
					%else %if "&droppedeff" = "Additive2 * Additive1" %then %do;
						%let droppedtype1 = 1;
						%let droppedtype2 = 3;
					%end;
					%else %if "&droppedeff" = "Dominance * Additive1" %then %do;
						%let droppedtype1 = 2;
						%let droppedtype2 = 3;
					%end;
				%end;
				
				%*put droppedtype1 = &droppedtype1;
				%*put droppedtype2 = &droppedtype1;
				
				*--- remove lesser significant epistasis, conditional on a threshold ---*;				
				%if %UPCASE(&auto) = %UPCASE(TRUE) %then %do;
					
					%if(%length(&droppedeff)) %then %do;
					
						data epe_del;
							set epe_less_sig;
						run;
					
						data &hkmimmodel;
							set &hkmimmodel;
							if trait = &trt and qtl1 = &epi_qtl1 and qtl2 = &epi_qtl2 
								and efftype1 = &droppedtype1 and efftype2 = &droppedtype2 then delete;
							else if neffect > 0 then do;
								neffect = neffect - 1;
							end;							
						run;

					%end; *--- end if droppedeff1 is true ---*;	
				%end; *--- end if auto is true ---*;
				
				proc datasets nolist;
					delete epe_less_sig;
				quit;
				
			%end; *--- end if NOBS_id_getepi >= 1 ---*;
			
			*--- delete temporary data sets ---*;
			proc datasets nolist;
				delete fitmodel_sum
					fitmodel_st
					interc_cov
					getepi;			
			quit;
						
		%end; *--- end if numbqtlin > 0 ---*;
	%end; *--- end if NOBS_hkmodel_tmp >=1 ---*;
	
	proc datasets nolist;
	        delete hkmodel_tmp
	        	hkdata3;
	quit;

	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);	

	ods exclude none;
	ods results;

	%exitdepi:;
	
%mend dropepieff;

/* backdropepieff() tests each epistatic effect. The argument AUTO gives the option to remove or not the lesser significant 
epistatic effect from the model if the test does not exceed a threshold. Its default value is TRUE. If AUTO= TRUE, this function 
keeps removing epistatic effects if they don't exceed a threshold*/
%macro backdropepieff(trt=, ng=, hkmimdata=, hkmimmodel=, criterion=, lod_threshold=, probf_stay=, auto=, mapgrid=, wsize=);

	%global flag_warn1_bde warn1_bde_message;
	%let flag_warn1_bde = 0;	
	%let warn1_bde_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local id_nqtlin_tmp;
	%local NOBS_nqtlin_tmp;
	%local numbqtlout;
	%local numbeffout; 
	%local InfinityLikeout;
	%local IsRefined;
	%let IsRefined= 0; *0 means not refined*;

	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitbdepi;
	%end;
				
	*--- get number of QTL and effects in the input model ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;

	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
		
	*--- check if trait = trt is in the input model --- *;
	%let id_nqtlin_tmp = %sysfunc(open(nqtlin_tmp));
	%let NOBS_nqtlin_tmp =%sysfunc(attrn(&id_nqtlin_tmp, NOBS));
	%let id_nqtlin_tmp = %sysfunc(close(&id_nqtlin_tmp)); *--- very important to close ---*;

	%if &NOBS_nqtlin_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;
		%let exiterror=1;
		%goto exitbdepi;
	%end;
	
	*--- call the procedure to drop a non-significant epistatic effect ---*; 
	%dropepieff(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
		criterion= &criterion, lod_threshold= &lod_threshold, probf_stay= &probf_stay, 
		auto= &auto);

	%if %sysfunc(exist(depi_epe_del)) %then %do;
		data epe_del_all;
			set depi_epe_del;
		run;
	%end;	
	
	*--- get number of QTL and effects after trying to remove epistatic effect ---*;
	proc means data=&hkmimmodel noprint;
		where trait=&trt;
		var nqtl neffect InfinityLike;
		output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
	run;
		
	data _null_;
		set nqtlout_tmp;
		call symputx("numbqtlout", numbqtl, 'L');
		call symputx("numbeffout", numbeff, 'L');
		call symputx("InfinityLikeout", InfinityLike, 'L');
	run;
	%*put numbqtlin = &numbqtlin;
	%*put numbqtlout = &numbqtlout;

	%if &numbeffout = &numbeffin %then %do;
		%let flag_warn1_bde = 1;
		%let warn1_bde_message= No QTL x QTL effects were removed from the model.;
		%put WARNING: &warn1_bde_message;
	%end;
	%else %do;

		%do %while(%sysevalf(&numbeffout < &numbeffin) and &InfinityLikeout=0);

			%let numbeffin = &numbeffout;

			*--- detele data set to avoid pilling results from previous runs ---*;
			proc datasets nolist;
				delete depi_epe_del;
			quit;

			*--- call the procedure to drop a non-significant epistatic effect ---*; 
			%dropepieff(trt= &trt, ng= &ng, hkmimdata= &hkmimdata, hkmimmodel= &hkmimmodel, 
				criterion= &criterion, lod_threshold= &lod_threshold, probf_stay= &probf_stay, 
				auto= &auto);

			%if %sysfunc(exist(depi_epe_del)) %then %do;
				data epe_del_all;
					set epe_del_all depi_epe_del;
				run;
			%end;		

			*--- get number of QTL and effects after trying to remove epistatic effect ---*;
			proc means data=&hkmimmodel noprint;
				where trait=&trt;
				var nqtl neffect InfinityLike;
				output out=nqtlout_tmp max=numbqtl numbeff InfinityLike;
			run;

			data _null_;
				set nqtlout_tmp;
				call symputx("numbqtlout", numbqtl, 'L');
				call symputx("numbeffout", numbeff, 'L');
				call symputx("InfinityLikeout", InfinityLike, 'L');
			run;
			
			*--- refine model before exiting the while loop ---*;
			%if &InfinityLikeout = 1 %then %do;
				%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
					wsize=&wsize, criterion=&criterion);	
				%let IsRefined= 1;
			%end;
			%else %if %sysevalf(&numbeffout = &numbeffin) and &IsRefined = 0 %then %do;
				%refinepos(trt=&trt, ng=&ng, hkmimdata=&hkmimdata, mapgrid=&mapgrid, hkmimmodel=&hkmimmodel, 
					wsize=&wsize, criterion=&criterion);
				*setting IsRefined to 1 to stop refining QTL position*;
				%let IsRefined= 1;
				*setting numbeffin to a lesser value to force the while loop to continue after refining QTL position*;
				%let numbeffin= %sysevalf(&numbeffin + 1);
			%end;
			%else %if %sysevalf(&numbeffout > &numbeffin) and &IsRefined = 1 %then %do;
				*setting IsRefined to 0 to allow refining QTL position*;
				%let IsRefined= 0;
			%end;

		%end; *--- end while loop ---*;
		
		%let flag_warn1_de = 0; /*comes from macro dropepieff()*/
		%let flag_warn2_de = 0; /*comes from macro dropepieff()*/
	%end;
	
	*--- delete temporary data sets ---*;
	proc datasets nolist;
		delete 	nqtlin_tmp
			nqtlout_tmp;
	quit;
	
	*--- delete temporary data sets ---*;
	proc datasets nolist;	
		delete depi_epe_del;
	quit;
	
	%exitbdepi:;

%mend backdropepieff;

/* refinepos() re-estimates the positions of QTL in the model */
%macro refinepos(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=, criterion=);

	ods exclude all;
	ods noresults;

	%global flag_warn1_rp warn1_rp_message;
	%global flag_warn2_rp warn2_rp_message;
	%let flag_warn1_rp=0;
	%let flag_warn2_rp=0;
	%let warn1_rp_message=;
	%let warn2_rp_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local continue_refine;
	%local rounds;
	%local maxrounds;
	%local posupdated_flag;
	%local geffect_main_red;
	%local geffect_epi_red;
	%local geffect_qxc_red;
	%local geffect_main_full;
	%local geffect_epi_full;
	%local geffect_qxc_full;	
	%local RefineWindow;
	%let RefineWindow= 15; /*--- region (in cM) on each side of a QTL to refine position ---*/
	
	%local teststmt;
	
	%local orderdirection;
	%local bool;
	%local neg2loglike0;
	%local AIC0;
	%local AICC0;
	%local BIC0;	
	
	%local bestInfinityLike;
						
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitref;
	%end;
	
	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/

	*--- created temporary data set that holds the model for the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*data OutLib.hkmodel_tmp; *set hkmodel_tmp; *run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitref;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);
	
	*--- get the number of QTL in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;

	%if &numbqtlin < 1 %then %do;
		%let flag_warn1_rp = 1;	
		%let warn1_rp_message= The current model has no QTL. Therefore, the refinepos procedure was aborted.;
		%put WARNING: &warn1_rp_message; 
	%end;
	%else %do;
		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		%*put # # # Number of QTL = &numbqtlin # # # # # #;
		%*put # # # Number of Effects = &numbeffin # # # #;
		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		
		%let orderdirection = ;
		%let bool = < ;
		%if "&criterion" = "LOD" %then %do; 
			%let orderdirection = descending;
			%let bool = > ;
		%end;
		
		*---begin loop over rounds of refinement ---*;
		%let continue_refine = 1;
		%let rounds = 1;
		%let maxrounds = 5;
		%do %while(&continue_refine = 1 and &rounds <= &maxrounds);

			%*put # # # round= &rounds;

			%let posupdated_flag= 0;

			*---begin loop over QTL in the input model ---*;
			%do i=1 %to &numbqtlin;
				*--- create a temporary data set that set holds QTL in the input model ---*;
				data qtl1_tmp(keep= trait qtl1 ch1 pos1 posid1) qtl2_tmp(keep= trait qtl2 ch2 pos2 posid2);
					set hkmodel_tmp;
					where trait=&trt and nqtl > 0;				
				run;

				proc sort data= qtl1_tmp;
					by qtl1;
				run;

				proc sort data= qtl2_tmp;
					by qtl2;
				run;

				data qtl_tmp1;
					merge qtl1_tmp qtl2_tmp(rename= (qtl2= qtl1 ch2= ch1 pos2= pos1 posid2= posid1));
					by qtl1;
				run;

				data allqtl_tmp;
					set qtl_tmp1;
					by qtl1;
					if first.qtl1 then output;
				run;

				proc sort data=allqtl_tmp;
					by ch1 DESCENDING posid1;
				run;

				data allqtl_tmp(drop= pos1right);
					set allqtl_tmp;
					by ch1;
					retain pos1right; 
					if first.ch1 then do;
						*rightbound=10000000;
						rightbound= pos1 + &RefineWindow;
						pos1right = pos1;
					end;
					else do;
						*rightbound = max(pos1 + (pos1right - pos1)/2, pos1right - &wsize);
						rightbound = min(pos1 + &RefineWindow, max(pos1, pos1right - &wsize));
						pos1right = pos1;
					end;
				run;

				proc sort data=allqtl_tmp;
					by ch1 posid1;
				run;

				data allqtl_tmp(drop= pos1left);
					set allqtl_tmp;
					by ch1;
					retain pos1left; 
					if first.ch1 then do;
						*leftbound= -10000000;
						leftbound= pos1 - &RefineWindow;
						pos1left = pos1;
					end;
					else do;
						*leftbound = min(pos1 - (pos1 - pos1left)/2, pos1left + &wsize);
						leftbound = max(pos1 - &RefineWindow, min(pos1, pos1left + &wsize));
						pos1left = pos1;
					end;
				run;

				proc sort data=allqtl_tmp(rename= (trait= curr_trait qtl1=curr_qtl ch1=curr_ch pos1=curr_pos posid1=curr_posid));
					by curr_ch curr_posid;
				run;

				proc datasets nolist;
					delete qtl1_tmp qtl2_tmp qtl_tmp1;
				quit;

				*--- create data sets for grid search ---*;
				data RefinedGrid_tmp;
					*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;	
					set &mapgrid(rename= (&CHROMNUM= chrom) 
						%if %length(&g_firstobs) %then firstobs= &g_firstobs;
						%if %length(&g_obs) %then obs=&g_obs;
						);
				run;

				data RefinedGrid_tmp;
					merge RefinedGrid_tmp allqtl_tmp(drop= curr_trait 
									rename=(curr_qtl= QTL 
										curr_ch= chrom 
										curr_pos= cm 
										curr_posid= cmid)
									);
					by chrom cmid;
					if QTL ^=. then QTLIndicator= 1;
					else QTLIndicator= 0;
				run;
				
				data RefinedGrid_tmp;
					set RefinedGrid_tmp;
					by chrom;
					if first.chrom then do;
						retainrightbound =rightbound;
						retainrightqtl = QTL;		
					end;
					retain retainrightbound;
					retain retainrightqtl;
					if QTL ^= . then do;
						retainrightbound =rightbound;
						retainrightqtl = QTL;
					end;
					if QTL = . and cm <= retainrightbound then QTL = retainrightqtl;
					if first.chrom then do;
						if QTL = . then do; 
							retainrightqtl = .; 
							QTL= retainrightqtl;
						end;
						else do;
							retainrightqtl = QTL; 
							QTL= retainrightqtl;
						end;
					end;
				run;

				proc sort data=RefinedGrid_tmp; 
					by chrom DESCENDING cmid;
				run;

				data RefinedGrid_tmp;
					set RefinedGrid_tmp;
					by chrom;
					if first.chrom then do;
						retainleftbound =leftbound;
						retainleftqtl = QTL;		
					end;
					retain retainleftbound;
					retain retainleftqtl;
					if QTL ^= . then do;
						retainleftbound =leftbound;
						retainleftqtl = QTL;
					end;
					if QTL = . and cm > retainleftbound then QTL = retainleftqtl;
				run;

				proc sort data=RefinedGrid_tmp (drop= retainrightbound retainrightqtl retainleftbound retainleftqtl);
					by chrom cmid;
				run;

				*---begin loop over the neighborhood of the QTL that is being refined ---*;		
				data CurrQtlGrid_tmp;
					set RefinedGrid_tmp;
					where QTL = &i;
				run;
				
				%if &ng = 4 %then %do;
					data _null_;
						set CurrQtlGrid_tmp;
						if QTLIndicator = 1 then call symputx("case_", contrast_, "L");
					run;
					%*put case_=&case_;
					
					data CurrQtlGrid_tmp;
						set CurrQtlGrid_tmp;
						where contrast_ = input("&case_", $1.);
					run;
				%end;
					
				%*goto exitref;

				data hkmodel_red_tmp;
					set hkmodel_tmp;
					*--- main and interaction effects of the current QTL are removed from the complete model ---*; 
					if qtl1 ^= &i and qtl2 ^= &i then output hkmodel_red_tmp;
					if qtl1 = &i and qtl2 = &i then do;
						call symputx("qtlch", trim(left(ch1)) , "L");
						call symputx("qtlip", trim(left(posid1)), "L");
					end;
				run;			
				%*put c=&qtlch;
				%*put p=&qtlip;

				*--- get effects of QTL in the reduced model ---*;
				%let geffect_main_red=;
				%let geffect_epi_red=;
				%let geffect_qxc_red=;

				proc sql;
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
							'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;

					select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
						'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_red separated by ' '
					from hkmodel_red_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;					
					
				quit;
				%put geffect_main_red= &geffect_main_red;
				%put geffect_epi_red= &geffect_epi_red;
				%put geffect_qxc_red= &geffect_qxc_red;

				%prephkdata3(ng=&ng, qtlvars=&geffect_main_red, talldsn=&hkmimdata);			
				
				proc mixed data=hkdata3 method=ml;
					model y&trt = &xcovar &geffect_main_red &geffect_epi_red / solution;
					ods output fitstatistics=fs0 solutionf=sf0;
				run;

				proc transpose data=fs0 out=fs01(rename= (
									_2_Log_Likelihood=neg2loglike0
									AIC__smaller_is_better_ =AIC0
									AICC__smaller_is_better_= AICC0
									BIC__smaller_is_better_=BIC0
									));
					var value;
					id descr;
				run;
				
				*--- get effects of QTL that is being refined ---*;
				%let geffect_main_full=;
				%let geffect_epi_full=;
				%let geffect_qxc_full=;

				proc sql;
					*--- collect effects for the contrast statement ---*;
					create table tested_main_tmp as 
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' length=30 as TestedEffect
					from hkmodel_tmp
					where nqtl > 0 and qtl1 = qtl2 and qtl1= &i and inter_cofactor_id = .;
					
					create table tested_epi_tmp as 
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'!'||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.))))||'!' length= 30 as TestedEffect
					from hkmodel_tmp
					where nqtl > 0 and qtl1 ^= qtl2 and (qtl1= &i or qtl2= &i) and inter_cofactor_id = .;
					
					create table tested_qxc_tmp as 
					select ('x'||trim(left(put(inter_cofactor_id,7.)))||'!'||'*'||
						'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' length= 30 as TestedEffect
					from hkmodel_tmp
					where nqtl > 0 and qtl1 ^= qtl2 and (qtl1= &i or qtl2= &i) and inter_cofactor_id ^= .;					
					
					*--- collect effects for the model statement ---*;
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' into: geffect_main_full  separated by ' '
					from hkmodel_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
					
					select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'!'||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.))))||'!' into: geffect_epi_full separated by ' '
					from hkmodel_tmp
					where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
					
					select ('x'||trim(left(put(inter_cofactor_id,7.)))||'!'||'*'||
						'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' into: geffect_qxc_full separated by ' '
					from hkmodel_tmp
					where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;					
				quit;
				%put geffect_main_full= &geffect_main_full;
				%put geffect_epi_full= &geffect_epi_full;
				%put geffect_qxc_full= &geffect_qxc_full;

				data eff_tested_qtl;
					set tested_main_tmp tested_epi_tmp tested_qxc_tmp; 
				run;

				%let teststmt= ;
				proc sql;
					select b.TestedEffect into: teststmt separated by ' 1, '
					from eff_tested_qtl as b;
				quit;
				%put teststmt= &teststmt;

				%*put q=&i;
				%*put c=&qtlch;
				%*put p=&qtlip;

				%if &ng=2 %then %do;
					%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e1c&qtlch.p&qtlip.!, xa2_!));			
					%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e1c&qtlch.p&qtlip.!, xa2_!));
					%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e1c&qtlch.p&qtlip.!, xa2_!));
					%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e1c&qtlch.p&qtlip.!, xa2_!));
				%end;
				%else %if &ng = 3 %then %do;
					%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e1c&qtlch.p&qtlip.!, xa2_!));			
					%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e1c&qtlch.p&qtlip.!, xa2_!));
					%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e1c&qtlch.p&qtlip.!, xa2_!));
					%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e1c&qtlch.p&qtlip.!, xa2_!));

					%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e2c&qtlch.p&qtlip.!, xd_!));
					%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e2c&qtlch.p&qtlip.!, xd_!));
					%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e2c&qtlch.p&qtlip.!, xd_!));
					%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e2c&qtlch.p&qtlip.!, xd_!));
				%end;				
				%else %if &ng = 4 %then %do;
					%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e1c&qtlch.p&qtlip.!, xa2_!));			
					%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e1c&qtlch.p&qtlip.!, xa2_!));
					%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e1c&qtlch.p&qtlip.!, xa2_!));
					%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e1c&qtlch.p&qtlip.!, xa2_!));

					%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e2c&qtlch.p&qtlip.!, xd_!));
					%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e2c&qtlch.p&qtlip.!, xd_!));
					%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e2c&qtlch.p&qtlip.!, xd_!));
					%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e2c&qtlch.p&qtlip.!, xd_!));
				
					%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e3c&qtlch.p&qtlip.!, xa1_!));
					%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e3c&qtlch.p&qtlip.!, xa1_!));
					%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e3c&qtlch.p&qtlip.!, xa1_!));
					%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e3c&qtlch.p&qtlip.!, xa1_!));
				%end;				
				
				%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, !, ));
				%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, !, ));
				%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, !, ));
				%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), !, ));
				
				%put geffect_main_full= &geffect_main_full;
				%put geffect_epi_full= &geffect_epi_full;
				%put geffect_qxc_full= &geffect_qxc_full;

				%let teststmt = %str('contq_' &teststmt 1);
				%put teststmt= &teststmt;

				%prephkdata1(ng=&ng, qtlvars=&geffect_main_red, talldsn=&hkmimdata, mappedgr=CurrQtlGrid_tmp, chrom_test=, cmid_test=);			
						
				ods output contrasts= contrq CovParms=CovP;
				proc mixed data=hkdata1 method=ml;
					by chrom cm cmid;
					model y&trt = &xcovar &geffect_main_full &geffect_epi_full &geffect_qxc_full/ solution;
					ods output fitstatistics=fs solutionf= sf nobs=nobm;
					contrast &teststmt / chisq;				
				run;
				
				data CovP;
					set CovP;
					where  CovParm = 'Residual';
					if Estimate=0 then InfinityLike= 1;
					else InfinityLike= 0;
				run; 				

				proc transpose data=fs out=fs1(rename= (
									_2_Log_Likelihood=neg2loglike
									AIC__smaller_is_better_ =AIC
									AICC__smaller_is_better_= AICC
									BIC__smaller_is_better_=BIC
									));
					var value;
					id descr;
					by chrom cm cmid;
				run;

				data nobm(keep=chrom cm cmid NObsRead NObsUsed);
					set nobm;
					where label = "Number of Observations Read";
				run;	
				
				%*goto exitref;

				data sf(drop= Effect stderr df tvalue probt);
					%if &ng = 2 %then %do;
						set sf(rename = (Estimate = Additive2));
						where Effect = "xa2_";
					%end;
					%else %if &ng = 3 %then %do;
						set sf(rename = (Estimate = Additive2));
						where Effect = "xa2_";
						set sf(rename = (Estimate = Dominance));
						where Effect = "xd_";
					%end;
					%else %do;
						%if &case_ = 2 %then %do;
							set sf(rename = (Estimate = Additive1));
							where Effect = "xa1_";
						%end;
						%else %if &case_ = 3 %then %do;
							set sf(rename = (Estimate = Additive2));
							where Effect = "xa2_";
							set sf(rename = (Estimate = Dominance));
							where Effect = "xd_";
						%end;	
						%else %if &case_ = 4 %then %do;
							set sf(rename = (Estimate = Dominance));
							where Effect = "xd_";
							set sf(rename = (Estimate = Additive1));
							where Effect = "xa1_";
						%end;
						%else %if &case_ = 5 %then %do;
							set sf(rename = (Estimate = Additive2));
							where Effect = "xa2_";
						%end;
						%else %if &case_ = 6 %then %do;
							set sf(rename = (Estimate = Additive2));
							where Effect = "xa2_";
							set sf(rename = (Estimate = Additive1));
							where Effect = "xa1_";
						%end;
						%else %if &case_ = 7 %then %do;
							set sf(rename = (Estimate = Additive2));
							where Effect = "xa2_";
							set sf(rename = (Estimate = Dominance));
							where Effect = "xd_";
						%end;
						%else %if &case_ = 8 %then %do;
							set sf(rename = (Estimate = Additive2));
							where Effect = "xa2_";
							set sf(rename = (Estimate = Dominance));
							where Effect = "xd_";
							set sf(rename = (Estimate = Additive1));
							where Effect = "xa1_";
						%end;
						%else %do;
							%put ERROR: Case not valid.;
						%end;
					%end;					
				run;

				data refineprof;
					merge sf CovP fs1 nobm contrq;
					*by chrom cm cmid;
					by chrom cmid;
					LOD = (1/(2 * 2.30258509299404))*(&neg2loglike0 - neg2loglike);
					%if "&criterion" = "LOD" %then %do; 
						valuet = LOD;
					%end;
					%else %if "&criterion" = "ProbF" %then %do; 
						valuet = probf;
					%end;	
					%else %if "&criterion" = "AIC" %then %do;
						valuet = AIC;
					%end;
					%else %if "&criterion" = "AICC" %then %do;
						valuet = AICC;
					%end;
					%else %if "&criterion" = "BIC" %then %do;
						valuet = BIC;
					%end;
				run;
				
				data bestref;
					set refineprof;
					where valuet ^=. ;
				run;

				proc sort data=bestref;
					by &orderdirection valuet;
				run;

				data _null_;
					set bestref(obs=1);
					call symputx("bestchrom", trim(left(chrom)), 'L');
					call symputx("bestcmid", trim(left(cmid)), 'L');
					call symputx("bestcm", trim(left(cm)), 'L');
					call symputx("bestInfinityLike", InfinityLike, "L");
				run;
				%*put qtl=&i;
				%*put bestchrom= &bestchrom;
				%*put bestcmid= &bestcmid;
				%*put bestcm= &bestcm;
				
				%*put qtlch= &qtlch;
				%*put qtlip= &qtlip;

				*---selecting best position of QTL ---*;
				%if (%sysevalf(&bestchrom = &qtlch and &bestcmid ^= &qtlip)) %then %do;
					
					data hkmodel_tmp;
						set hkmodel_tmp;
						if qtl1= &i then do;
							pos1= &bestcm; 
							posid1 = &bestcmid; 
						end;
						if qtl2= &i then do;
							pos2= &bestcm;
							posid2= &bestcmid;
						end;
						InfinityLike= &bestInfinityLike;
					run;						

					%let posupdated_flag = 1;
					
				%end;	
				
				%*goto exitref;

			%end; *--- end do loop over QTL in the input model ---*;

			%let rounds = %sysevalf(&rounds + 1);
			
			%*put posupdated_flag = &posupdated_flag;
			%if &posupdated_flag = 1 %then %do;
				%let continue_refine= 1;
			%end;
			%else %do;
				%let continue_refine= 0;
			%end;
			
			%*goto exitref;

		%end; *--- end while loop over number of refinement rounds ---*; 
		
		%*goto exitref;
		
		*--- save best model ---*;
		data &hkmimmodel;
			merge &hkmimmodel hkmodel_tmp;
			by trait;
		run;
	
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete 	hkmodel_red_tmp
				CurrQtlGrid_tmp
				RefinedGrid_tmp
				allqtl_tmp;
		quit;
		
		%*put rounds=&rounds;
		%*put maxrounds=&maxrounds;
		%*put posupdated_flag= &posupdated_flag;
		
		%if(&rounds > &maxrounds and &posupdated_flag = 1) %then %do;
			%let flag_warn2_rp = 1;
			%let warn2_rp_message= Positions of QTL did not converge to stable values during refinement. Consider removing some QTL from the model.;
			%put WARNING: &warn2_rp_message;
		%end;
	
	%end; *--- end if numbqtlin > 0 ---*;
	
	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp
			hkdata1
			hkdata3;
	quit;

	%exitref:;
	
	ods exclude none;
	ods results;
				
%mend refinepos;

/* profile() creates significance level profiling in the genome grid. Main and interaction effects of each QTL are jointly tested
to produce the significance level profiling on their neighborhoods.*/
%macro profile(trt=, ng=, hkmimdata=, mapgrid=, hkmimmodel=, wsize=);

	*ods html;
	ods exclude all;
	ods noresults;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local geffect_main_red;
	%local geffect_epi_red;
	%local geffect_qxc_red;
	%local geffect_main_full;
	%local geffect_epi_full;
	%local geffect_qxc_full;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local teststmt;	
	%local id_ChromNoQtlGrid_tmp;
	%local NOBs_ChromNoQtlGrid_tmp;
			
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitprof;
	%end;
	
	/*
	*---set contrast to estimate effects of covariates---*;
	%let xcovar = ;
	%do i = 1 %to &nx;
		%let xcovar = &xcovar x&i;
	%end;
	*/
		
	*--- created temporary data set that holds the model for the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitprof;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);	
	
	*--- get the number of QTL in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%put # # # # # # # # # # # # # # # # # # # # # # ;
	%put # # # Number of QTL = &numbqtlin # # # # # #;
	%put # # # Number of Effects = &numbeffin # # # #;
	%put # # # # # # # # # # # # # # # # # # # # # # ;

	%if &numbqtlin > 0 %then %do;
		*--- profile when there is QTL in the model ---*;
		
		*data OutLib.q_hkmodel_tmp; *set hkmodel_tmp; *run;
	
		*--- create a temporary data set that set holds QTL in the input model ---*;
		data qtl1_tmp(keep= trait qtl1 ch1 pos1 posid1) qtl2_tmp(keep= trait qtl2 ch2 pos2 posid2);
			set hkmodel_tmp;
			where trait=&trt and nqtl > 0;				
		run;

		proc sort data= qtl1_tmp;
			by qtl1;
		run;

		proc sort data= qtl2_tmp;
			by qtl2;
		run;

		data qtl_tmp1;
			merge qtl1_tmp qtl2_tmp(rename= (qtl2= qtl1 ch2= ch1 pos2= pos1 posid2= posid1));
			by qtl1;
		run;

		data allqtl_tmp;
			set qtl_tmp1;
			by qtl1;
			if first.qtl1 then output;
		run;

		/*
		proc sort data=allqtl_tmp;
			by ch1 DESCENDING posid1;
		run;

		data allqtl_tmp(drop= pos1right);
			set allqtl_tmp;
			by ch1;
			retain pos1right; 
			if first.ch1 then do;
				rightbound=1000000000;
				pos1right = pos1;
			end;
			else do;
				*rightbound = max(pos1 + (pos1right - pos1)/2, pos1right - &wsize);
				rightbound = max(pos1, pos1right - &wsize);
				pos1right = pos1;
			end;
		run;

		proc sort data=allqtl_tmp;
			by ch1 posid1;
		run;

		data allqtl_tmp(drop= pos1left);
			set allqtl_tmp;
			by ch1;
			retain pos1left; 
			if first.ch1 then do;
				leftbound= -1000000000;
				pos1left = pos1;
			end;
			else do;
				leftbound = min(pos1, pos1left + &wsize);
				pos1left = pos1;
			end;
		run;
		*/
		
		proc sort data=allqtl_tmp;
			*by ch1 DESCENDING posid1;
		     	by ch1 posid1;
		run;

		data _null_;
			set allqtl_tmp end=_eof_;
			by ch1;
			retain c;
			if _n_ = 1 then c=0;
			if first.ch1 then do;
				c= c + 1;
				call symputx("ch"||trim(left(c)),ch1);
			end;
			if _eof_ then call symputx("nch",c);
		run;
		%put ch1= &ch1;
		%put nch= &nch;

		data allqtlbounds;
			delete;
		run;

		%do i=1 %to &nch;

			data currch_tmp;
				set allqtl_tmp end=_eof_;
				retain c;
				if _n_= 1 then c=0;
				if ch1= &&ch&i then do;
					c=c+1;
					call symputx("pos"||trim(left(c)), pos1);
					call symputx("posID"||trim(left(c)), posid1);
					call symputx("qtlID"||trim(left(c)), qtl1);	
				end;	
				if _eof_ then call symput("nq", trim(left(c)));
			run;
			%put nq= &nq;

			%do j=1 %to &nq;
				%let a= %sysevalf(&j + 1);
				%let b= %sysevalf(&j - 1);

				%if &j=1 %then %do;
					%if &nq > 1 %then %do;
						%let d1= %sysevalf(&&pos&a - &&pos&j);
					%end;
					%else %do;
						%let d1= &wsize;
					%end;
					%put j= &j d1= &d1;
					%let lbound&j = %sysevalf(&&pos&j - &wsize);
					%if %sysevalf(&wsize < &d1/2) %then %let rbound&j = %sysevalf(&&pos&j + &wsize);
					%else %let rbound&j= %sysevalf(&&pos&j + &d1/2);
				%end;
				%if &j > 1 and &j < &nq %then %do;
						%let d1= %sysevalf(&&pos&j - &&pos&b);
						%let d2= %sysevalf(&&pos&a - &&pos&j);
						%if %sysevalf(&wsize < &d1/2) %then %let lbound&j = %sysevalf(&&pos&j - &wsize);
						%else %let lbound&j= %sysevalf(&&pos&j - &d1/2);
						%if %sysevalf(&wsize < &d2/2) %then %let rbound&j = %sysevalf(&&pos&j + &wsize);
						%else %let rbound&j= %sysevalf(&&pos&j + &d2/2);
				%end;
				
				%if &j=&nq %then %do;
					%if &nq > 1 %then %do;
						%let d1= %sysevalf(&&pos&j - &&pos&b);
					%end;
					%else %do;
						%let d1= &wsize;
					%end;			
					%put j= &j d1= &d1;	
					%if %sysevalf(&wsize < &d1/2) %then %let lbound&j = %sysevalf(&&pos&j - &wsize);
					%else %let lbound&j= %sysevalf(&&pos&j - &d1/2);
					%let rbound&j = %sysevalf(&&pos&j + &wsize);
				%end;
				
				%put j= &j lb= &&lbound&j up= &&rbound&j;

				data qtlbounds;
					trait= &trt;
					qtl1= &&qtlID&j;
					ch1= &&ch&i;
					pos1= &&pos&j;
					posid1= &&posID&j;
					leftbound= &&lbound&j;
					rightbound= &&rbound&j;
				run;
				
				data allqtlbounds;
					set allqtlbounds qtlbounds;
				run;
			%end;	
		%end;

		proc datasets nolist noprint;
			delete allqtl_tmp;
			change allqtlbounds= allqtl_tmp;
		run;
		quit;		

		proc sort data=allqtl_tmp(rename= (trait= curr_trait qtl1=curr_qtl ch1=curr_ch pos1=curr_pos posid1=curr_posid));
			by curr_ch curr_posid;
		run;

		proc datasets nolist;
			delete qtl1_tmp qtl2_tmp qtl_tmp1;
		quit;

		*--- create data sets for profiling results ---*;
		*data OutLib.q_mapgrid; *set &mapgrid; *run;
		
		data profile;
			*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;	
			set &mapgrid(rename= (&CHROMNUM= chrom) 
						%if %length(&g_firstobs) %then firstobs= &g_firstobs;
						%if %length(&g_obs) %then obs=&g_obs;
						);
			retain Additive2 %if &ng >= 3 %then Dominance; %if &ng = 4 %then Additive1; .;
		run;

		data profile;
			merge profile allqtl_tmp(drop= curr_trait rename=(curr_qtl= QTL curr_ch= chrom curr_pos= cm curr_posid= cmid));
			by chrom cmid;
			if QTL ^=. then QTLIndicator= 1;
			else QTLIndicator= 0;
		run;

		data profile;
			set profile;
			by chrom;
			if first.chrom then do;
				retainrightbound =rightbound;
				retainrightqtl = QTL;		
			end;
			retain retainrightbound;
			retain retainrightqtl;
			if QTL ^= . then do;
				retainrightbound =rightbound;
				retainrightqtl = QTL;
			end;
			if QTL = . and cm <= retainrightbound then QTL = retainrightqtl;
			if first.chrom then do;
				if QTL = . then do; 
					retainrightqtl = .; 
					QTL= retainrightqtl;
				end;
				else do;
					retainrightqtl = QTL; 
					QTL= retainrightqtl;
				end;
			end;	
		run;

		proc sort data=profile; 
			by chrom DESCENDING cmid;
		run;

		data profile;
			set profile;
			by chrom;
			if first.chrom then do;
				retainleftbound =leftbound;
				retainleftqtl = QTL;		
			end;
			retain retainleftbound;
			retain retainleftqtl;
			if QTL ^= . then do;
				retainleftbound =leftbound;
				retainleftqtl = QTL;
			end;
			if QTL = . and cm > retainleftbound then QTL = retainleftqtl;
		run;
		
		*data OutLib.q_profile1; *set profile; *run;

		proc sort data=profile(drop= retainrightbound retainrightqtl retainleftbound retainleftqtl);
			by chrom cmid;
		run;
		
		data profile;
			set profile;
			if QTL =. then QTLIndicator = .;
		run;
		
		*data OutLib.q_profile2; *set profile; *run;
	
 		*---begin loop over QTL in the input model ---*;
		%do i=1 %to &numbqtlin;

			*---begin loop over the neighborhood of the QTL that is being profiled ---*;			
			data CurrQtlGrid_tmp;
				set profile;
				where QTL = &i;
			run;
			
			*data Outlib.q_CurrQtlGrid_tmp; *set CurrQtlGrid_tmp; *run;
			
			%if &ng = 4 %then %do;
				data _null_;
					set CurrQtlGrid_tmp;
					if QTLIndicator = 1 then call symputx("case_", contrast_, "L");
				run;
				%*put case_=&case_;
				
				data CurrQtlGrid_tmp;
					set CurrQtlGrid_tmp;
					where contrast_ = input("&case_", $1.);
				run;
			%end;
			
			%*goto exitprof;

			data hkmodel_red_tmp;
				set hkmodel_tmp;
				*--- main and interaction effects of the current QTL are removed from the complete model ---*; 
				if qtl1 ^= &i and qtl2 ^= &i then 
					output hkmodel_red_tmp;
				if qtl1 = &i and qtl2 = &i then do;
					call symputx("qtlch", trim(left(ch1)) , "L");
					call symputx("qtlip", trim(left(posid1)), "L");
				end;
			run;
			
			%put c=&qtlch;
			%put p=&qtlip;
			
			*--- get effects of QTL in the reduced model ---*;
			%let geffect_main_red=;
			%let geffect_epi_red =;
			%let geffect_qxc_red =;

			proc sql;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_red separated by ' '
				from hkmodel_red_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
				
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
					'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_red separated by ' '
				from hkmodel_red_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
				
				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_red separated by ' '
				from hkmodel_red_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;
				
			quit;
			%put geffect_main_red= &geffect_main_red;
			%put geffect_epi_red= &geffect_epi_red;
			%put geffect_qxc_red= &geffect_qxc_red;

			%prephkdata1(ng=&ng, qtlvars=&geffect_main_red, talldsn=&hkmimdata, mappedgr=CurrQtlGrid_tmp, chrom_test=, cmid_test=);			

			*data OutLib.q_hkdata1; *set hkdata1; *run;

			*--- get value of criterion for the reduced model ---*;		
			proc mixed data=hkdata1 method=ml;
				by chrom cm cmid;
				model y&trt = &xcovar &geffect_main_red &geffect_epi_red &geffect_qxc_red/ solution;
				ods output fitstatistics=fs0 solutionf=sf0;
			run;
			
			%*goto exitprof;

			proc transpose data=fs0 out=fs01(rename= (
								_2_Log_Likelihood=neg2loglike0
								AIC__smaller_is_better_ =AIC0
								AICC__smaller_is_better_= AICC0
								BIC__smaller_is_better_=BIC0
								));
				var value;
				id descr;
				by chrom cm cmid;
			run;

			*--- get effects of QTL that is being profiled ---*;
			%let geffect_main_full=;
			%let geffect_epi_full=;

			proc sql;
				*--- collect effects for the contrast statement ---*;
				create table tested_main_tmp as 
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' length=30 as TestedEffect
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and qtl1= &i and inter_cofactor_id = .;
				
				create table tested_epi_tmp as 
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'!'||'*'||
					'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.))))||'!' length= 30 as TestedEffect
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and (qtl1= &i or qtl2= &i) and inter_cofactor_id = .;
				
				create table tested_qxc_tmp as 
				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'!'||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' length= 30 as TestedEffect
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and (qtl1= &i or qtl2= &i) and inter_cofactor_id ^= .;					
				
				*--- collect effects for the model statement ---*;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' into: geffect_main_full  separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
				
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'!'||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.))))||'!' into: geffect_epi_full separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;
				
				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'!'||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.))))||'!' into: geffect_qxc_full separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;				
			quit;
			%put geffect_main_full= &geffect_main_full;
			%put geffect_epi_full= &geffect_epi_full;
			%put geffect_qxc_full= &geffect_qxc_full;
			
			data eff_tested_qtl;
				set tested_main_tmp tested_epi_tmp; 
			run;

			%let teststmt= ;
			proc sql;
				select b.TestedEffect into: teststmt separated by ' 1, '
				from eff_tested_qtl as b;
			quit;
			%*put teststmt= &teststmt;
			
			%*goto exitprof;
			
			%*put c=&qtlch;
			%*put p=&qtlip;
			
			%if &ng=2 %then %do;
				%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e1c&qtlch.p&qtlip.!, xa2_!));			
				%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e1c&qtlch.p&qtlip.!, xa2_!));
				%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e1c&qtlch.p&qtlip.!, xa2_!));
				%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e1c&qtlch.p&qtlip.!, xa2_!));
			%end;
			%else %if &ng = 3 %then %do;
				%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e1c&qtlch.p&qtlip.!, xa2_!));			
				%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e1c&qtlch.p&qtlip.!, xa2_!));
				%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e1c&qtlch.p&qtlip.!, xa2_!));
				%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e1c&qtlch.p&qtlip.!, xa2_!));
						
				%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e2c&qtlch.p&qtlip.!, xd_!));
				%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e2c&qtlch.p&qtlip.!, xd_!));
				%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e2c&qtlch.p&qtlip.!, xd_!));
				%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e2c&qtlch.p&qtlip.!, xd_!));
			%end;
			%else %if &ng = 4 %then %do;
				%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e1c&qtlch.p&qtlip.!, xa2_!));			
				%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e1c&qtlch.p&qtlip.!, xa2_!));
				%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e1c&qtlch.p&qtlip.!, xa2_!));
				%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e1c&qtlch.p&qtlip.!, xa2_!));
						
				%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e2c&qtlch.p&qtlip.!, xd_!));
				%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e2c&qtlch.p&qtlip.!, xd_!));
				%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e2c&qtlch.p&qtlip.!, xd_!));
				%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e2c&qtlch.p&qtlip.!, xd_!));
				
				%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, e3c&qtlch.p&qtlip.!, xa1_!));
				%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, e3c&qtlch.p&qtlip.!, xa1_!));
				%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, e3c&qtlch.p&qtlip.!, xa1_!));
				%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), e3c&qtlch.p&qtlip.!, xa1_!));				
			%end;
			
			%let geffect_main_full= %sysfunc(tranwrd(&geffect_main_full, !, ));
			%let geffect_epi_full= %sysfunc(tranwrd(&geffect_epi_full, !, ));
			%let geffect_qxc_full= %sysfunc(tranwrd(&geffect_qxc_full, !, ));
			%let teststmt= %sysfunc(tranwrd(%quote(&teststmt), !, ));
						
			%put geffect_main_full= &geffect_main_full;
			%put geffect_epi_full= &geffect_epi_full;
			%put geffect_qxc_full= &geffect_qxc_full;
			
			%let teststmt = %str('contq_' &teststmt 1);
			%put teststmt= &teststmt;

			*--- get value of criterion for the full model ---*;	
			ods output contrasts= contrq CovParms=CovP;
			proc mixed data=hkdata1 method=ml;
				by chrom cm cmid;
				model y&trt = &xcovar &geffect_main_full &geffect_epi_full &geffect_qxc_full/ solution;
				ods output fitstatistics=fs solutionf= sf nobs=nobm;
				contrast &teststmt / chisq;				
			run;
			
			data CovP;
				set CovP;
				where  CovParm = 'Residual';
				if Estimate=0 then InfinityLike= 1;
				else InfinityLike= 0;
			run;			
			
			proc transpose data=fs out=fs1(rename= (
								_2_Log_Likelihood=neg2loglike
								AIC__smaller_is_better_ =AIC
								AICC__smaller_is_better_= AICC
								BIC__smaller_is_better_=BIC
								));
				var value;
				id descr;
				by chrom cm cmid;
			run;
								
			data nobm(keep=chrom cm cmid NObsRead NObsUsed);
				set nobm;
				where label = "Number of Observations Read";
			run;	
			
			%*goto exitprof;
			
			data sf(drop= Effect stderr df tvalue probt);
				%if &ng = 2 %then %do;
					set sf(rename = (Estimate = Additive2));
					where Effect = "xa2_";
				%end;
				%else %if &ng = 3 %then %do;
					set sf(rename = (Estimate = Additive2));
					where Effect = "xa2_";
					
					set sf(rename = (Estimate = Dominance));
					where Effect = "xd_";
				%end;
				%else %do;
					%if &case_ = 2 %then %do;
						set sf(rename = (Estimate = Additive1));
						where Effect = "xa1_";
					%end;
					%else %if &case_ = 3 %then %do;
						set sf(rename = (Estimate = Additive2));
						where Effect = "xa2_";
						set sf(rename = (Estimate = Dominance));
						where Effect = "xd_";
					%end;	
					%else %if &case_ = 4 %then %do;
						set sf(rename = (Estimate = Dominance));
						where Effect = "xd_";
						set sf(rename = (Estimate = Additive1));
						where Effect = "xa1_";
					%end;
					%else %if &case_ = 5 %then %do;
						set sf(rename = (Estimate = Additive2));
						where Effect = "xa2_";
					%end;
					%else %if &case_ = 6 %then %do;
						set sf(rename = (Estimate = Additive2));
						where Effect = "xa2_";
						set sf(rename = (Estimate = Additive1));
						where Effect = "xa1_";
					%end;
					%else %if &case_ = 7 %then %do;
						set sf(rename = (Estimate = Additive2));
						where Effect = "xa2_";
						set sf(rename = (Estimate = Dominance));
						where Effect = "xd_";
					%end;
					%else %if &case_ = 8 %then %do;
						set sf(rename = (Estimate = Additive2));
						where Effect = "xa2_";
						set sf(rename = (Estimate = Dominance));
						where Effect = "xd_";
						set sf(rename = (Estimate = Additive1));
						where Effect = "xa1_";
					%end;
					%else %do;
						%put ERROR: Case not valid.;
					%end;
				%end;
			run;

			data profile;
				merge profile sf CovP fs1 fs01 nobm contrq;
				*by chrom cm cmid;
				by chrom cmid;
				LOD = (1/(2 * 2.30258509299404))*(neg2loglike0 - neg2loglike);
			run;

			%*goto exitprof;
						
		%end; 
		*--- end do loop over QTL in the input model ---*;

		%*goto exitprof;
		
		*data OutLib.q_allqtl_tmp; *set allqtl_tmp; *run;
		
		*--- profile chromosomes (if any) without QTL ---*; 			
		data ChromNoQtlGrid_tmp;
			set profile;
			where missing(QTL);
			drop QTL Additive: Dominance leftbound rightbound QTLIndicator;
		run;
		
		%let id_ChromNoQtlGrid_tmp=%sysfunc(open(ChromNoQtlGrid_tmp));
		%let NObs_ChromNoQtlGrid_tmp=%sysfunc(attrn(&id_ChromNoQtlGrid_tmp,NOBS));
		%let id_ChromNoQtlGrid_tmp=%sysfunc(close(&id_ChromNoQtlGrid_tmp)); *--- very important to close ---*;
		
		%if &Nobs_ChromNoQtlGrid_tmp > 0 %then %do;
			*--- NOBs_id_ChromNoQtlGrid_tmp > 0 ---*;

			*--- get effects of QTL in the input model ---*;
			%let geffect_main_in=;
			%let geffect_epi_in =;
			%let geffect_qxc_in =;
			
			proc sql;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;
				
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
						'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;

				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;				
			quit;
			%*put geffect_main_in= &geffect_main_in;
			%*put geffect_epi_in= &geffect_epi_in;
			%*put geffect_qxc_in= &geffect_qxc_in;

			%prephkdata1(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata, mappedgr=ChromNoQtlGrid_tmp, chrom_test=, cmid_test=);		

			%*goto exitprof;

			%if &ng= 2 %then %do;
				%let teststmt= 'contq_' xa2_ 1;
			%end;
			%else %if &ng = 3 %then %do;
				%let teststmt= 'contq_' xa2_ 1, xd_ 1; 
			%end;
			%else %if &ng = 4 %then %do;
				%*let teststmt= %str(&teststmt, xa1_ 1); 
			
				*case2*;
				%let teststmt2= '2' xa1_ 1;
				*case3;
				%let teststmt3= '3' xa2_ 1, xd_  1;
				*case4*;
				%let teststmt4= '4' xd_  1, xa1_ 1; 
				*case3;
				%let teststmt5= '5' xa2_ 1;
				*case6*;
				%let teststmt6= '6' xa2_ 1, xa1_ 1; 
				*case7*;
				%let teststmt7= '7' xa2_ 1, xd_  1; 
				*case8*;
				%let teststmt8= '8' xa2_ 1, xd_  1, xa1_ 1;				
			%end;

			%*goto exitprof;

			*---fitting chromosomes without QTL---*;
			ods output contrasts= contrq CovParms=CovP;
			proc mixed data=hkdata1 method=ml;
				by chrom cm cmid;
				model y&trt = &xcovar xa2_ %if &ng >= 3 %then xd_; 
							   %if &ng = 4 %then xa1_; 
							   &geffect_main_in &geffect_epi_in &geffect_qxc_in
							   / solution;
				ods output fitstatistics=fs solutionf=sf nobs=nobm;
				%if &ng ^= 4 %then %do;
					contrast &teststmt / chisq;
				%end;
				%else %do;
					contrast &teststmt2 / chisq;
					contrast &teststmt3 / chisq;
					contrast &teststmt4 / chisq;
					contrast &teststmt5 / chisq;
					contrast &teststmt6 / chisq;
					contrast &teststmt7 / chisq;
					contrast &teststmt8 / chisq;
				%end;
			run;

			data CovP;
				set CovP;
				where  CovParm = 'Residual';
				if Estimate=0 then InfinityLike= 1;
				else InfinityLike= 0;
			run; 
		
			*---fitting model with all QTLs ---*;
			proc mixed data=hkdata1 method=ml;
				by chrom CM cmid;
				model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in/ solution;
				ods output fitstatistics=fs0 solutionf=sf0;
			run;		

			proc transpose data=fs out=fs1(rename= (
								_2_Log_Likelihood=neg2loglike
								AIC__smaller_is_better_ =AIC
								AICC__smaller_is_better_= AICC
								BIC__smaller_is_better_=BIC
								));
				var value;
				id descr;
				by chrom cm cmid;
			run;

			proc transpose data=fs0 out=fs01(rename= (
								_2_Log_Likelihood=neg2loglike0
								AIC__smaller_is_better_ =AIC0
								AICC__smaller_is_better_= AICC0
								BIC__smaller_is_better_=BIC0
								));
				var value;
				id descr;
				by chrom cm cmid;
			run;
			
			data nobm(keep=chrom cm cmid NObsRead NObsUsed);
				set nobm;
				where label = "Number of Observations Read";
			run;	
			
			%*goto exitprof;

			data sf(drop= Effect stderr df tvalue probt);
				%if &ng = 2 %then %do;
					set sf(rename = (Estimate = Additive2));
					where Effect = "xa2_";
				%end;
				%else %if &ng = 3 %then %do;
					set sf(rename = (Estimate = Additive2));
					where Effect = "xa2_";
					set sf(rename = (Estimate = Dominance));
					where Effect = "xd_";
				%end;
				%else %do;
					set sf(rename = (Estimate = Additive2));
					where Effect = "xa2_";
					set sf(rename = (Estimate = Dominance));
					where Effect = "xd_";			
					set sf(rename = (Estimate = Additive1));
					where Effect = "xa1_";
				%end;
			run;
			
			%if &ng = 4 %then %do;
				data hkdata1_sub;
					set hkdata1;
					by chrom cmid;
					if first.cmid then output;
				run;
					
				proc sql;
					create table contrq1 as
					select a.*
					from contrq as a, hkdata1_sub as b
					where a.chrom= b.chrom and a.cmid= b.cmid and a.label= b.contrast_;
				quit;
				
				proc datasets nolist;
					delete contrq;
					change contrq1= contrq;
					modify contrq;
					rename label = contrast_;
					delete hkdata1_sub;
				quit;
			%end;
				
			data profile;
				merge profile sf CovP fs1 fs01 nobm contrq;
				*by chrom cm cmid;
				by chrom cmid;
				LOD = (1/(2 * 2.30258509299404))*(neg2loglike0 - neg2loglike);
			run;

			%*goto exitprof;

		*--- end NOBs_id_ChromNoQtlGrid_tmp > 0 ---*;
		%end; 
		*--- end profile chromosomes (if any) without QTL ---*;
		
		*--- CHROMNUM, MARKERLAB, MARKERPOS, CHROMLAB are global variable passed from SingleQTLModel.sas ---*;	
		data profile(keep = chrom cm cmid &MARKERPOS 
					AIC BIC AICC LOD NumDF DenDF ProbF
					Additive2 
					%if &ng>=3 %then Dominance; 
					%if &ng=4 %then Additive1 contrast_; 
					MarkerIndicator &MARKERLAB NObsRead NObsUsed
					QTL QTLIndicator rightbound leftbound
					%if %length(&CHROMLAB)^=0 %then &CHROMLAB;); 
			merge &mapgrid(rename=(&CHROMNUM= chrom)) profile;
			*by chrom cm cmid;
			by chrom cmid;
		run;
		%*goto exitprof;

		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp
				hkmodel_red_tmp
				CurrQtlGrid_tmp
				allqtl_tmp
				ChromNoQtlGrid_tmp;
		quit;	
		
	*--- end profile when there is QTL in the model ---*;			
	%end; /*--- end if &numbqtlin > 0 ---*/
	%else %do;
		*--- profile when there is no QTL in the model ---*;
		
		data MappedGrid_tmp;
			*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;	
			set &mapgrid(rename= (&CHROMNUM= chrom) 
					%if %length(&g_firstobs) %then firstobs= &g_firstobs;
					%if %length(&g_obs) %then obs=&g_obs;
					);
		run;
	
		%prephkdata1(ng=&ng, qtlvars=, talldsn=&hkmimdata, mappedgr=MappedGrid_tmp, chrom_test=, cmid_test=);		

		%if &ng = 2 %then %do;
			%let teststmt= 'contq_' xa2_ 1;
		%end;
		%else %if &ng = 3 %then %do;
			%let teststmt= 'contq_' xa2_ 1, xd_ 1;
		%end;
		%else %if &ng = 4 %then %do;
			%*let teststmt= %str(&teststmt, xa1_ 1); 
			
			*case2*;
			%let teststmt2= '2' xa1_ 1;
			*case3;
			%let teststmt3= '3' xa2_ 1;
			*case4*;
			%let teststmt4= '4' xd_ 1, xa1_ 1; 
			*case3;
			%let teststmt5= '5' xa2_ 1;
			*case6*;
			%let teststmt6= '6' xa2_ 1, xa1_ 1; 
			*case7*;
			%let teststmt7= '7' xa2_ 1, xd_ 1; 
			*case8*;
			%let teststmt8= '8' xa2_ 1, xd_ 1, xa1_ 1;
		%end;
		
		%*goto exitprof;
		
		ods output contrasts= contrq CovParms=CovP;
		proc mixed data=hkdata1 method=ml;
			by chrom cm cmid;
			model y&trt = &xcovar %if &ng= 2 %then xa2_;
						%else %if &ng = 3 %then xa2_ xd_; 
						%else xa2_ xd_ xa1_; 
						/ solution;
			ods output fitstatistics=fs solutionf=sf nobs=nobm;
			%if &ng ^= 4 %then %do;
				contrast &teststmt / chisq;
			%end;
			%else %do;
				contrast &teststmt2 / chisq;
				contrast &teststmt3 / chisq;
				contrast &teststmt4 / chisq;
				contrast &teststmt5 / chisq;
				contrast &teststmt6 / chisq;
				contrast &teststmt7 / chisq;
				contrast &teststmt8 / chisq;
			%end;
		run;
		
		data CovP;
			set CovP;
			where  CovParm = 'Residual';
			if Estimate=0 then InfinityLike= 1;
			else InfinityLike= 0;
		run; 		

		proc mixed data=hkdata1 method=ml;
			by chrom cm cmid;
			model y&trt = &xcovar / solution;
			ods output fitstatistics=fs0 solutionf=sf0;
		run;		
		
		proc transpose data=fs out=fs1(rename= (
							_2_Log_Likelihood=neg2loglike
							AIC__smaller_is_better_ =AIC
							AICC__smaller_is_better_= AICC
							BIC__smaller_is_better_=BIC
							));
			var value;
			id descr;
			by chrom cm cmid;
		run;

		proc transpose data=fs0 out=fs01(rename= (
							_2_Log_Likelihood=neg2loglike0
							AIC__smaller_is_better_ =AIC0
							AICC__smaller_is_better_= AICC0
							BIC__smaller_is_better_=BIC0
							));
			var value;
			id descr;
			by chrom cm cmid;
		run;
		
		data nobm(keep=chrom cmid NObsRead NObsUsed);
			set nobm;
			where label = "Number of Observations Read";
		run;		

		data sf(drop= Effect stderr df tvalue probt);
			%if &ng = 2 %then %do;
				set sf(rename = (Estimate = Additive2));
				where Effect = "xa2_";
			%end;
			%else %if &ng = 3 %then %do;
				set sf(rename = (Estimate = Additive2));
				where Effect = "xa2_";
				set sf(rename = (Estimate = Dominance));
				where Effect = "xd_";
			%end;
			%else %do;
				set sf(rename = (Estimate = Additive2));
				where Effect = "xa2_";
				set sf(rename = (Estimate = Dominance));
				where Effect = "xd_";			
				set sf(rename = (Estimate = Additive1));
				where Effect = "xa1_";
			%end;
		run;
		
		%if &ng = 4 %then %do;
			data hkdata1_sub;
				set hkdata1;
				by chrom cmid;
				if first.cmid then output;
			run;
				
			proc sql;
				create table contrq1 as
				select a.*
				from contrq as a, hkdata1_sub as b
				where a.chrom= b.chrom and a.cmid= b.cmid and a.label= b.contrast_;
			quit;
			
			proc datasets nolist;
				delete contrq;
				change contrq1= contrq;
				modify contrq;
				rename label = contrast_;
				delete hkdata1_sub;
			quit;
		%end;
		
		%*goto exitprof;
		
		data profile;
			merge sf CovP fs1 fs01 nobm contrq;
			*by chrom cm cmid;
			by chrom cmid;
			LOD = (1/(2 * 2.30258509299404))*(neg2loglike0 - neg2loglike);
		run;
		
		%*goto exitprof;
		
		*--- CHROMNUM, MARKERPOS, MARKERLAB, CHROMLAB are global variable passed from SingleQTLModel.sas ---*;	
		data profile(keep = chrom cm cmid &MARKERPOS
					AIC BIC AICC LOD NumDF DenDF ProbF
					Additive2 
					%if &ng>=3 %then Dominance; 
					%if &ng=4 %then Additive1 contrast_; 
					MarkerIndicator &MARKERLAB NObsRead NObsUsed
					QTL QTLIndicator rightbound leftbound
					%if %length(&CHROMLAB)^=0 %then &CHROMLAB;); 
			merge &mapgrid(rename=(&CHROMNUM= chrom)) profile;
			*by chrom cm cmid;
			by chrom cmid;
			QTL= .;
			QTLIndicator = .;
			rightbound = .;
			leftbound = .;
		run;
		
		proc datasets nolist;
			delete MappedGrid_tmp;
		quit;
		
		%*goto exitprof;
		
	*--- end profile when there is no QTL in the model ---*;	
	%end; 
		
	*--- format output data set ---*;
	*--- CHROMNUM, MARKERLAB, MARKERPOS, CHROMLAB, and TRAITVAR are global variables passed from SingleQTLModel.sas ---*;	
	data profile(drop = rightbound leftbound 
			/*%if &ng=4 %then contrast_;*/
		);
		retain &CHROMNUM 
			%if %length(&CHROMLAB)^=0 %then &CHROMLAB;
			&MARKERPOS
			Morgans CM CMID MarkerIndicator 
			&MARKERLAB
			LOD&trt AIC&trt AICC&trt 
			BIC&trt 
			NObsRead&trt
			NObsUsed&trt 
			NumDF&trt DenDF&trt ProbF&trt 
			QTL&trt QTLIndicator&trt 
			Additive2&trt %if &ng>=3 %then Dominance&trt;
			%if &ng=4 %then %do;
				Additive1&trt
				contrast_
			%end;
			;
		set profile(rename= (chrom= &CHROMNUM 
					QTL= QTL&trt QTLIndicator= QTLIndicator&trt
					LOD= LOD&trt AIC=AIC&trt AICC=AICC&trt 
					NObsRead= NObsRead&trt
					NObsUsed= NObsUsed&trt 
					BIC=BIC&trt NumDF= NumDF&trt DenDF= DenDF&trt 
					ProbF= ProbF&trt Additive2= Additive2&trt 
					%if &ng>=3 %then Dominance= Dominance&trt;
					%if &ng=4 %then Additive1= Additive1&trt;
					)
				);
		label &CHROMNUM= "&CHROMNUM";
		%if %length(&CHROMLAB)^=0 %then %do;
			label &CHROMLAB= "&CHROMLAB";
		%end;
		label CM= "CM";
		label CMID= "CMID";
		label MarkerIndicator= "Marker Indicator";
		label &MARKERLAB= "&MARKERLAB";
		label &MARKERPOS= "&MARKERPOS";
		label QTL&trt= "QTL Profile Region &TRAITVAR";
		label QTLIndicator&trt = "QTL Position Indicator &TRAITVAR";
		label LOD&trt= "LOD &TRAITVAR";
		label AIC&trt= "AIC &TRAITVAR";
		label AICC&trt= "AICC &TRAITVAR";
		label BIC&trt= "BIC &TRAITVAR";
		label NObsRead&trt = "Num Obs Read &TRAITVAR";
		label NObsUsed&trt = "Num Obs Used &TRAITVAR";
		label NumDF&trt= "Num DF &TRAITVAR";
		label DenDF&trt= "Den DF &TRAITVAR";
		label ProbF&trt= "ProbF &TRAITVAR";
		%if &ng < 4 %then %do;
			label Additive2&trt= "Additive &TRAITVAR";
		%end;
		%else %do;
			***assuming maximum ng = 4 ***;
			label Additive2&trt= "Additive2 &TRAITVAR";
			label Additive1&trt= "Additive1 &TRAITVAR";
			
			if Contrast_= 2 then do;
				Additive2&trt = 0;
				Dominance&trt = 0;
			end;
			else if contrast_= 3 then do;
				Additive1&trt = 0;
			end;
			else if contrast_= 4 then do;
				Additive2&trt = 0;
			end;
			else if contrast_= 5 then do;
				Dominance&trt = 0;
				Additive1&trt = 0;
			end;
			else if contrast_= 6 then do;
				Dominance&trt = 0;
			end;
			else if contrast_= 7 then do;
				Additive1&trt = 0;
			end;
			else if contrast_= 8 then do;
				*all effects are estimable*;
			end;
			label Contrast_= "Contrast Type";			
		%end;
		%if &ng >= 3 %then %do;
			label Dominance&trt= "Dominance &TRAITVAR";
		%end;
		Morgans = CM/100;
		label Morgans = "Morgans";
	run;	
	
	proc datasets nolist;
		delete hkmodel_tmp
			hkdata1;	
	quit;

	%exitprof:;
	
	ods exclude none;
	ods results;

%mend profile;

/*VarExit() checks if a variable exists in the dataset, and returns 1 if yes and 0 if no */
%macro VarExist(ds=,var=);
	%local rc dsid result;
	%let dsid=%sysfunc(open(&ds));
	%if %sysfunc(varnum(&dsid,&var)) > 0 %then %do;
		%let result=1;
		%*put NOTE: Var &var exists in &ds;
	%end;
	%else %do;
		%let result=0;
		%*put NOTE: Var &var not exists in &ds;
	%end;
	%let rc=%sysfunc(close(&dsid));
	&result
%mend VarExist;

/*PrepTablesToAdd() takes the output dataset "fitmodel_sum" from the macro fitmodel() and splits it into datasets that will be used
to display tables via JSL code. It also renames output "fitmodel_st" from the macro fitmodel() to "HKMIMFitTable_OutDataName", because 
"fitmodel_st" may get deleted within dropmainqtlonly() and dropepieff()*/
%macro PrepTablesToAdd(fittedstat=, fittedmodel=, modelnumb= );

	data &HKMIMFitTable_OutData.&modelnumb;
		set &fittedstat;
	run;
	
	%global modelwithQTL H0JointMain H0JointMainInter;
	
	%let modelwithQTL= %VarExist(ds=&fittedmodel,var=QTL1);
	%let H0JointMain= %VarExist(ds=&fittedmodel,var=H0_JointMain);
	%let H0JointMainInter= %VarExist(ds=&fittedmodel,var=H0_JointMainInter);
	
	%local isQTL2missing;
	%local i;
	%local fvar;	

	data &HKMIMCofTable_OutData.&modelnumb (keep= Predictors Estimate StdErr DF_IndTest tValue_IndTest Probt_IndTest)
		%if &modelwithQTL = 1 %then %do;
			&HKMIMQtlTable_OutData.&modelnumb (Keep= QTL1 Ch1 Pos1 
				%if &H0JointMain = 1 %then %do;
					H0_JointMain NumDF_JointMain LOD_JointMain
				%end;
				%if &H0JointMainInter = 1 %then %do;
					H0_JointMainInter NumDF_JointMainInter LOD_JointMainInter
				%end;
			)
			&HKMIMEffTable_OutData.&modelnumb (keep= QTL1 QTL2 Effect_type Effect Estimate StdErr H0_IndTest DF_IndTest tValue_IndTest Probt_IndTest)
		%end;
		;
		set &fittedmodel; /*output data set from fitmodel()*/
		%if %length(&FactorVarsNew) %then %do;
			if inter_cofactor_id ^= . then do;
				Effect = tranwrd(trim(left(Effect)), "x"||trim(left(inter_cofactor_id)),scan("&FactorVarsNew", inter_cofactor_id));
				QTL2 = .;
			end;
		%end;
		if trim(left(Predictors)) ^= "QTL Effect" then output &HKMIMCofTable_OutData.&modelnumb;
		%if &modelwithQTL = 1 %then %do;
			if trim(left(Predictors)) = "QTL Effect" and trim(left(Effect_type)) = "Main" then output &HKMIMQtlTable_OutData.&modelnumb;
			if trim(left(Predictors)) = "QTL Effect" then do;
				if trim(left(Effect_type)) = "Main" then QTL2 = .;
				output &HKMIMEffTable_OutData.&modelnumb;
			end;	
		%end;
	run;
	
	data &HKMIMCofTable_OutData.&modelnumb;
		set &HKMIMCofTable_OutData.&modelnumb;
		%if %length(&FactorVarsNew) %then %do;
			%let i = 1;
			%let fvar = %scan(&FactorVarsNew,&i);
			%do %while(%length(&fvar));
				if trim(left(Predictors)) = "x&i" then Predictors= "&fvar";									
				%let i = %eval(&i+1);
				%let fvar = %scan(&FactorVarsNew,&i);
			%end;		
		%end;		
		label Predictors = "Predictors";
		label Estimate = "Estimate";
		label StdErr = "Standard Error";
		label DF_IndTest = "DF";
		label tValue_IndTest="t Value";
		label Probt_IndTest="Pr > |t|";				
	run;
	
	%if &modelwithQTL = 1 %then %do;
		data &HKMIMQtlTable_OutData.&modelnumb;
			set &HKMIMQtlTable_OutData.&modelnumb;
			by QTL1;
			if first.QTL1 then output;
			label QTL1 = "QTL1";
			*--- CHROMNUM is a global variable passed from SingleQTLModel.sas ---*;	
			%if %length(&CHROMNUM) %then %do;
				label Ch1 = "&CHROMNUM";
			%end;
			%else %do;
				label Ch1 = "ChNum";
			%end;
			*--- GDMETRIC is a global variable passed from SingleQTLModel.sas ---*;	
			label Pos1 = "&GDMETRIC";
			%if &H0JointMain = 1 %then %do;
				label H0_JointMain = "H02";
				label NumDF_JointMain = "NumDF for H02";
				label LOD_JointMain = "LOD for H02";
			%end;
			%if &H0JointMainInter = 1 %then %do;
				label H0_JointMainInter = "H03";
				label NumDF_JointMainInter = "NumDF for H03";
				label LOD_JointMainInter = "LOD for H03";
			%end;
		run;

		proc means data=&HKMIMEffTable_OutData.&modelnumb noprint;
			var QTL2;
			output out=maxQTL2tmp max=maxQTL2;
		run;

		%local isQTL2missing;
		data _null_;
			set maxQTL2tmp;
			call symputx("isQTL2missing", maxQTL2, 'L');
		run;
		%*put isQTL2missing= &isQTL2missing;

		data &HKMIMEffTable_OutData.&modelnumb(drop= %if &isQTL2missing = . %then QTL2;);
			set &HKMIMEffTable_OutData.&modelnumb;
			%if &isQTL2missing = . %then %do;
				label QTL1 = "QTL1";
			%end;
			%else %do;
				label QTL1 = "QTL1";
				label QTL2 = "QTL2";
			%end;
			label Effect_type = "Effect Type";
			label Effect = "Effect";
			label Estimate = "Estimate";
			label StdErr = "Standard Error";
			label H0_IndTest = "H01";
			label DF_IndTest = "DenDF";
			label tValue_IndTest="t Value";
			label Probt_IndTest="Pr > |t|";
		
		run;
	%end;
	
	%exitpreptable:;

%mend PrepTablesToAdd;

%macro CodeToAddDataTableToJsl(iswithindatastep= FALSE, fname=);

	%local opensqbr clossqbr;
	
	%if %UPCASE(&iswithindatastep) = %UPCASE(FALSE) %then %do;
		data _null_;
			file "&fname" mod lrecl=32767;
	%end;
			%let opensqbr = \[;
			%let clossqbr = ]\;
			put "Here:AppendTableFn = Function({nobs, nvar, varlabs, vartyps, allobs, oboxtitle, appendtobox, modelnumb, tablenumb},{default local}," ;
			put "	Eval(Parse(EvalInsert( ";
			put "		""&opensqbr ";			
			put "		Here:hlb^modelnumb^_^tablenumb^ = hlistbox(" ;
			put "			Here:o^modelnumb^_^tablenumb^ = Outline Box(oboxtitle," ;
			put "			);" ;
			put "		); ";
			put "		&clossqbr"" "; 
			put "		); ";
			put "	);); ";
			put "	k = 1;" ;
			put "	nclm = Parse(EvalInsert( ";
			put "				""&opensqbr ";
			put "					if(trim(vartyps[^k^]) == ""C"", ";
			put "						"" String Col Box(varlabs[^k^], {allobs[^k^]})"" ";
			put "					, //else ";
			put "						"" Number Col Box(varlabs[^k^], {allobs[^k^]})"" ";
			put "					); ";
			put "				&clossqbr"" "; 
			put "				); ";
			put "		); ";
			put "	k = k + 1;" ;
			put "	for(j = 2, j <= nvar, j++," ;
        		put "		nclm = Parse(EvalInsert( ";
        		put "					""&opensqbr ";
			put "						if(trim(vartyps[^k^]) == ""C"", ";
        		put "							""^nclm^, String Col Box(varlabs[^k^], {allobs[^k^]})"" ";
			put "						, //else ";
        		put "							""^nclm^, Number Col Box(varlabs[^k^], {allobs[^k^]})"" ";
			put "						); ";
			put "					&clossqbr"" ";
			put "				); ";
			put "		); ";
			put "		k = k + 1;" ;
			put "	);" ;
			put "	Eval(Parse(EvalInsert( ";
			put "				""&opensqbr ";
			put "				//Here:tb = Table Box(^nclm^;);";
			put "				Here:tb^modelnumb^_^tablenumb^ = Table Box(^nclm^;);";
			put "				Here:o^modelnumb^_^tablenumb^ << append(Here:tb^modelnumb^_^tablenumb^);";
			put "				&clossqbr"" ";
			put "			); "; 
			put "		); "; 
			put "	); ";	
			put "	k= nvar + 1;" ;
			put "	for(i = 2, i <= nobs, i++," ;
			put "		nrow = Parse(EvalInsert( ";
			put "					""&opensqbr ";
			put "					""allobs[^k^]"" ";
			put "					&clossqbr"" ";
			put "				); ";
			put "	 		); ";
			put "  		k = k + 1;" ;	
			put "		for(j = 2, j <= nvar, j++," ;
			put "			nrow = Parse(EvalInsert( ";
			put "						""&opensqbr ";
			put "						""^nrow^, allobs[^k^]"" ";
			put "						&clossqbr"" ";
			put "					); ";
			put "				); ";
			put "  			k = k + 1;" ;
			put "		);" ;
			put "		Eval(Parse(EvalInsert( ";
			put "					""&opensqbr ";		
			put "					//Here:tb << add row({^nrow^}); ";
			put "					Here:tb^modelnumb^_^tablenumb^ << add row({^nrow^}); ";
			put "					&clossqbr"" ";
			put "				); ";
			put "			); ";
			put "		); ";
			put "	);" ;
			put "   ";
			put "	Eval(Parse(EvalInsert("; 
			put "			""&opensqbr ";
			put "			appendtobox << Append(Here:hlb^modelnumb^_^tablenumb^);" ;
			put "			//show(Here:tb^modelnumb^_^tablenumb^);";
			put "			//Print( Here:o^modelnumb^_^tablenumb^ << Get Title );";
			put "			//Print( Here:tb^modelnumb^_^tablenumb^ << Get Names() );";
			put "			&clossqbr"" ";
			put "			); ";
			put "		); ";
			put "	); ";
			put ");" ;
			put ;
	%if %UPCASE(&iswithindatastep) = %UPCASE(FALSE) %then %do;	
		run;
	%end;
	
	%exitaddtable:;
	
%mend CodeToAddDataTableToJsl;	

/*GetTableValuesToMacroVarAndJsl() write the observations of table into a jsl file.*/
%macro GetTableValuesToMacroVarAndJsl(dsn=, tablename=, iswithindatastep= FALSE, fname=);
	
	%local dsid rc NOBS_dsid NVAR_dsid varnam varlab vartyp ncounter obsval;

	%let dsid=%sysfunc(open(&dsn));
	%let NOBS_dsid =%sysfunc(attrn(&dsid, NOBS));
	%let NVAR_dsid =%sysfunc(attrn(&dsid, NVARS));
	%*put NVAR_dsid = &NVAR_dsid;
	
	%if %UPCASE(&iswithindatastep) = %UPCASE(FALSE) %then %do;
		data _null_;
		file "&fname" mod lrecl=32767;
	%end;
		put ;
		put "Here:GetTableValuesExpr = Expr(" ; 
		put "	oboxtitle= ""&tablename""; ";
		put "	nobs = &NOBS_dsid;" ;
		put "	nvar = &NVAR_dsid;" ;
		put "	varlabs= {" ;
		%do j = 1 %to %sysevalf(&NVAR_dsid - 1);
			%let varlab = %sysfunc(VARLABEL(&dsid,&j));
		put "	""&varlab"", " @;
		%end;
		%do j = &NVAR_dsid %to &NVAR_dsid;
			%let varlab = %sysfunc(VARLABEL(&dsid,&j));
		put "	""&varlab"" " ;
		%end;
		put "	};" ;
		put "	vartyps= {" ;	
		%do j = 1 %to %sysevalf(&NVAR_dsid - 1);
			%let vartyp = %sysfunc(VARTYPE(&dsid,&j));
		put "	""&vartyp"", " @;
		%end;
		%do j = &NVAR_dsid %to &NVAR_dsid;
			%let vartyp = %sysfunc(VARTYPE(&dsid,&j));
		put "	""&vartyp"" " ;
		%end;
		put "	};" ;
		%syscall set(dsid);
		%let ncounter = 1;
		put "	allobs = {" ;
		%do i = 1 %to %sysevalf(&NOBS_dsid);
			%let rc=%sysfunc(fetchobs(&dsid,&i));
			%do j = 1 %to &NVAR_dsid;
				%let varnam = %sysfunc(VARNAME(&dsid,&j));
				%let obsval = %str(&&&varnam);
				%let vartyp = %sysfunc(VARTYPE(&dsid,&j));
				%if &ncounter ^= (&NOBS_dsid * &NVAR_dsid) %then %do;
					%if &vartyp = N %then %do;
		put "	%trim(&obsval)," @;		
					%end;
					%else %do;
		put "	""%trim(&obsval)""," @;
					%end;
				%end;
				%else %do;
					%if &vartyp = N %then %do;
		put "	%trim(&obsval)" @;			
					%end;
					%else %do;
		put "	""%trim(&obsval)"" " @;
					%end;	
				%end;
				%let ncounter = %sysevalf(&ncounter + 1);
			%end;
			put ;
		%end;
		put "	};";
		put ");" ;		
		put "Here:GetTableValuesExpr; " ;
		*put ;
	%if %UPCASE(&iswithindatastep) = %UPCASE(FALSE) %then %do;
		run;	
	%end;
	
	%let dsid=%sysfunc(close(&dsid));
	
	%exitgetvalues:;
	
%mend GetTableValuesToMacroVarAndJsl;

/*PrepModelTables() writes into a jsl file a function to append hlistbox into a outlinebox.*/
%macro PrepModelTables(iswithindatastep= FALSE, fname=, modelnumb=);
	
	***ClientOutPath is defined in the jsl code that is created from IM analysis***;
	***add code to jsl file***;
	%if %UPCASE(&iswithindatastep) = %UPCASE(FALSE) %then %do;
		data _null_;
		file "&fname" mod lrecl=32767;
	%end;
		put "//append model table" ;
		put "Here:AppendHlbToOlbFn = Function({AppendToBox},{default local}, ";
		put "Here:hlbt = hlistbox(" ;
		
		put "Here:AppendTableAfterRunFn(workfolder=Here:ClientOutPath, dsname= ""&HKMIMFitTable_OutDataName""||char(Here:Nmodels), outboxtitle= ""Model Fitness Information"", OLBox= AppendToBox, modelnumb=Here:NModels, tablenumb=1);";
	
		%if &modelwithQTL = 1 %then %do; 
			put "Here:AppendTableAfterRunFn(workfolder=Here:ClientOutPath, dsname= ""&HKMIMQtlTable_OutDataName""||char(Here:Nmodels), outboxtitle= ""Table of Mapped QTL"", OLBox= AppendToBox, modelnumb=Here:NModels, tablenumb=2);";

			put "Here:AppendTableAfterRunFn(workfolder=Here:ClientOutPath, dsname= ""&HKMIMEffTable_OutDataName""||char(Here:Nmodels), outboxtitle= ""Table of QTL Effects"", OLBox= AppendToBox, modelnumb=Here:NModels, tablenumb=3);";

		%end;

		put "Here:AppendTableAfterRunFn(workfolder=Here:ClientOutPath, dsname= ""&HKMIMCofTable_OutDataName""||char(Here:Nmodels), outboxtitle= ""Table of Cofactor Effects"", OLBox= AppendToBox, modelnumb=Here:NModels, tablenumb=4);";

		put "//Here:AppendTableAfterRunFn(workfolder=Here:ClientOutPath, dsname= ""&HKMIMSetTable_OutDataName""||char(Here:Nmodels), outboxtitle= ""Table of Settings Used to Build the Model"", OLBox= AppendToBox, modelnumb=Here:NModels, tablenumb=5);";
		
		/*
		%GetTableValuesToMacroVarAndJsl(dsn=fitmodel_st_final&modelnumb, tablename= Model Fitness Information, iswithindatastep= TRUE, fname=);
		****fitmodel_st_final comes from PrepTablesToAdd()****;		
		
		put "Here:AppendTableFn(nobs=nobs, nvar=nvar, varlabs=varlabs, vartyps=vartyps, allobs=allobs, oboxtitle=oboxtitle, AppendToBox=AppendToBox, modelnumb=Here:NModels, tablenumb=1); ";
		
		%if &modelwithQTL = 1 %then %do; 
			****modelwithQTL is a global variable that comes from PrepTablesToAdd()****;
			
			%GetTableValuesToMacroVarAndJsl(dsn=qtlmapped&modelnumb, tablename= Table of Mapped QTL, iswithindatastep= TRUE, fname=);
			****qtlmapped comes from PrepTablesToAdd()****;
			
			put "Here:AppendTableFn(nobs=nobs, nvar=nvar, varlabs=varlabs, vartyps=vartyps, allobs=allobs, oboxtitle=oboxtitle, AppendToBox=AppendToBox, modelnumb=Here:NModels, tablenumb=2); ";
		
			%GetTableValuesToMacroVarAndJsl(dsn=qtleffects&modelnumb, tablename= Table of QTL Effects, iswithindatastep= TRUE, fname=);
			****qtleffects comes from PrepTablesToAdd()****;
	
			put "Here:AppendTableFn(nobs=nobs, nvar=nvar, varlabs=varlabs, vartyps=vartyps, allobs=allobs, oboxtitle=oboxtitle, AppendToBox=AppendToBox, modelnumb=Here:NModels, tablenumb=3); ";

		%end;
		
		%GetTableValuesToMacroVarAndJsl(dsn=nonqtleffects&modelnumb, tablename= Table of Cofactor Effects, iswithindatastep= TRUE, fname=);
		****nonqtleffects comes from PrepTablesToAdd()****;
	
		put "Here:AppendTableFn(nobs=nobs, nvar=nvar, varlabs=varlabs, vartyps=vartyps, allobs=allobs, oboxtitle=oboxtitle, AppendToBox=AppendToBox, modelnumb=Here:NModels, tablenumb=4); ";

		%GetTableValuesToMacroVarAndJsl(dsn=mimsettingsdata&modelnumb, tablename= Table of Settings Used to Build the Model, iswithindatastep= TRUE, fname=);
		****mimsettingsdata comes from SingleQTLModel.sas or MultiQTLModel_DrillDowns.jsl****;
	
		put "Here:AppendTableFn(nobs=nobs, nvar=nvar, varlabs=varlabs, vartyps=vartyps, allobs=allobs, oboxtitle=oboxtitle, AppendToBox=AppendToBox, modelnumb=Here:NModels, tablenumb=5); ";
		*/
		
		put ");" ;
		put "AppendToBox << Append(Here:hlbt);" ;
		put ");" ;
	
	%if %UPCASE(&iswithindatastep) = %UPCASE(FALSE) %then %do;
		run;	
	%end;
	
	%exitprepmodeltables:;
		
%mend PrepModelTables;

/* InteractivelyAddMain() adds one QTL with main effect to the model*/
%macro InteractivelyAddMain(trt=, ng=, hkmimmodel=, chrom=, pos=, posid=, contrastype=);

	*ods html;
	ods exclude all;
	ods noresults;

	%global flag_warn1_addm warn1_addm_message;
	%let flag_warn1_addm =0;
	%let warn1_addm_message=;

	*--- NEVER uses same prefix anywhere throughout the code ---*;	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;	
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	
	%local totalqtl;	
	%local totaleff;
	
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitaddm;
	%end;
	
	*--- created temporary data set that holds the model for the current trait ---*;
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;	

	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitaddm;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;	
	%put xcovar= &xcovar;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);	

	*--- get the number of QTL in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;	
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%if &InfinityLikein %then %do;
		%let flag_warn1_addm = 1;
		%let warn1_addm_message = No more QTL effects can be added into the model because the residual variance is zero.;
		%put WARNING: ;
		%goto exitaddm;
	%end;	
	
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
		

	*---beging Interactively adding main effect to the model---*;
	%let totalqtl= %sysevalf(&numbqtlin + 1);
	
	%if &ng=2 %then %do;
		%let totaleff = %sysevalf(&numbeffin + 1);
		%let case_= .;
	%end;
	%else %if &ng=3 %then %do;
		%let totaleff = %sysevalf(&numbeffin + 2);
		%let case_= .;
	%end;
	%else %if &ng=4 %then %do;
		%let case_= &contrastype;
		%if &case_ = 2 %then %let totaleff = %sysevalf(&numbeffin + 1);
		%else %if &case_ = 3 %then %let totaleff = %sysevalf(&numbeffin + 2);
		%else %if &case_ = 4 %then %let totaleff = %sysevalf(&numbeffin + 2);
		%else %if &case_ = 5 %then %let totaleff = %sysevalf(&numbeffin + 1);
		%else %if &case_ = 6 %then %let totaleff = %sysevalf(&numbeffin + 2);
		%else %if &case_ = 7 %then %let totaleff = %sysevalf(&numbeffin + 2);
		%else %if &case_ = 8 %then %let totaleff = %sysevalf(&numbeffin + 3);				
	%end;		
	
	data add2_tmp;
		retain 
		trait &trt 
		cofactors "&xcovar"
		ncofactors &nx
		inter_cofactor_id .
		nqtl &totalqtl 
		neffect &totaleff
		indic_epi_eff 0
		indic_qtl_without_inter .		
		qtl1 &totalqtl 
		qtl2 &totalqtl 
		ch1 &chrom
		ch2 &chrom
		pos1 &pos
		pos2 &pos
		posid1 &posid
		posid2 &posid
		efftype1 1 
		efftype2 1
		contrast &case_;	
	run;
	
	data dom_tmp;
		retain 
		trait &trt 
		cofactors "&xcovar"
		ncofactors &nx
		inter_cofactor_id .
		nqtl &totalqtl 
		neffect &totaleff
		indic_epi_eff 0
		indic_qtl_without_inter .		
		qtl1 &totalqtl 
		qtl2 &totalqtl 
		ch1 &chrom
		ch2 &chrom
		pos1 &pos
		pos2 &pos
		posid1 &posid
		posid2 &posid
		efftype1 2 
		efftype2 2
		contrast &case_;
	run;
	
	data add1_tmp;
		retain 
		trait &trt 
		cofactors "&xcovar"
		ncofactors &nx
		inter_cofactor_id .
		nqtl &totalqtl 
		neffect &totaleff
		indic_epi_eff 0
		indic_qtl_without_inter .	
		qtl1 &totalqtl 
		qtl2 &totalqtl 
		ch1 &chrom
		ch2 &chrom
		pos1 &pos
		pos2 &pos
		posid1 &posid
		posid2 &posid
		efftype1 3 
		efftype2 3
		contrast &case_;	
	run;	
	
	data &hkmimmodel;
		set &hkmimmodel 
			%if %sysevalf(&ng=2) %then add2_tmp; 
			%if %sysevalf(&ng=3) %then add2_tmp dom_tmp; 
			%if %sysevalf(&ng=4) %then %do;
				%if &case_=2 %then add1_tmp;
				%else %if &case_=3 %then add2_tmp dom_tmp;
				%else %if &case_=4 %then add1_tmp dom_tmp;
				%else %if &case_=5 %then add2_tmp;
				%else %if &case_=6 %then add2_tmp add1_tmp;
				%else %if &case_=7 %then add2_tmp dom_tmp;
				%else %if &case_=8 %then add2_tmp dom_tmp add1_tmp;
			%end;
		;
		if trait= &trt and qtl1 ^=. then do;
			nqtl = &totalqtl;
			neffect = &totaleff;
		end;
	run;		
	*---end Interactively adding main effect to the model---*;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp
			add2_tmp
			dom_tmp
			add1_tmp;
	quit;	

	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);
	
	%exitaddm:;

	ods exclude none;
	ods results;	
	
%mend InteractivelyAddMain;

/*InteractivelyAddEpi0D() adds one interaction effect between two QTL already in the model*/
%macro InteractivelyAddEpi0D(trt=, ng=, hkmimmodel=, qtl1=, qtl2=, effect1=, effect2=);

	ods exclude all;
	ods noresults;

	%global flag_warn1_add0D warn1_add0D_message;
	%let flag_warn1_add0D=0;
	%let warn1_add0D_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local totalqtl;
	%local totaleff;
			
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitadde0D;
	%end;
	
	*---crete temporary data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitadde0D;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;	
	%put xcovar= &xcovar;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);	
		
	*--- get number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%if &InfinityLikein %then %do;
		%let flag_warn1_add0D = 1;
		%let warn1_add0D_message = No more QTL effects can be added into the model because the residual variance is zero.;
		%put WARNING: &warn1_add0D_message;
		%goto exitadde0D;
	%end;	
	
	%if %sysevalf(&numbqtlin < 2) %then %do;
	
		%let flag_warn1_add0D = 1;		
		%put WARNING: The Interactively Add Interaction procedure needs a model with at least 2 QTL.;  
			
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitadde0D;	
	%end;

	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	*--- begin Interactively adding two-way interaction among QTL in the model ---*;
	%let efftype1= ;
	%let efftype2= ;
	%if &ng < 4 %then %do;
		%if %upcase(&effect1) = ADDITIVE %then
			%let efftype1= 1;
		%else %if %upcase(&effect1) = DOMINANCE %then
			%let efftype1= 2;
		
		%if %upcase(&effect2) = ADDITIVE %then
			%let efftype2= 1;
		%else %if %upcase(&effect2) = DOMINANCE %then
			%let efftype2= 2;
	%end;
	%else %do;
		***assuming maximum ng = 4 ***;
		%if %upcase(&effect1) = ADDITIVE2 %then
			%let efftype1= 1;
		%else %if %upcase(&effect1) = DOMINANCE %then
			%let efftype1= 2;
		%else %if %upcase(&effect1) = ADDITIVE1 %then
			%let efftype1= 3;
		
		%if %upcase(&effect2) = ADDITIVE2 %then
			%let efftype2= 1;
		%else %if %upcase(&effect2) = DOMINANCE %then
			%let efftype2= 2;
		%else %if %upcase(&effect2) = ADDITIVE1 %then
			%let efftype2= 3;	
	%end;
	
	data _null_;
		set hkmodel_tmp;
		if qtl1= &qtl1 and efftype1 = &efftype1 then do;
			call symputx("chrom1", trim(left(ch1)));
			call symputx("pos1", trim(left(pos1)));
			call symputx("posid1", trim(left(posid1)));
			call symputx("epitype1", trim(left(efftype1)));
		end;
		else if qtl1= &qtl2 and efftype2= &efftype2 then do;
			call symputx("chrom2", trim(left(ch2)));
			call symputx("pos2", trim(left(pos2)));
			call symputx("posid2", trim(left(posid2)));
			call symputx("epitype2", trim(left(efftype2)));
		end;
	run;
	
	%let totalqtl= &numbqtlin;
	%let totaleff= %sysevalf(&numbeffin + 1);
	
	data neweff_tmp;
		retain 
		trait &trt
		cofactors "&xcovar"
		ncofactors &nx
		inter_cofactor_id .
		nqtl &totalqtl
		neffect &totaleff
		indic_epi_eff 1
		indic_qtl_without_inter . 	
		qtl1 &qtl1 
		qtl2 &qtl2
		ch1 &chrom1
		ch2 &chrom2
		pos1 &pos1
		pos2 &pos2
		posid1 &posid1
		posid2 &posid2
		efftype1 &epitype1
		efftype2 &epitype2
		contrast .;
	run;

	data &hkmimmodel;
		set &hkmimmodel neweff_tmp;
		if trait= &trt and neffect > 0 then do;
			nqtl = &totalqtl;
			neffect = &totaleff;
		end;
	run;

	proc datasets nolist;
		delete neweff_tmp;
	quit;
	*--- end Interactively adding two-way interaction among QTL in the model ---*;	

	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp;
	quit;
	
	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);
				
	%exitadde0D:;
	
	ods exclude none;
	ods results;
		
%mend InteractivelyAddEpi0D;

/*InteractivelyDelQTLEff() deletes one QTL from the model. If the QTL has interaction, the interaction is also deleted*/
%macro InteractivelyDelQTLEff(trt=, ng=, hkmimmodel=, qtl1=, qtl2=, effect1=, effect2=, cofactorID=);

	ods exclude all;
	ods noresults;
	
	%put qtl1= &qtl1;
	%put qtl2= &qtl2;
	%put effect1= &effect1;
	%put effect2= &effect2;
	%put cofactorID= &cofactorID;

	%global flag_warn1_deleff warn1_deleff_message;
	%let flag_warn1_deleff=0;
	%let warn1_deleff_message=;
	
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp; 
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local case_;
	%local NumbDelMain_;
	%local NumbDelQXQ_;
	%local NumbDelQxC_;
		
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitdeleff;
	%end;

	*--- create data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;	
		%let exiterror=1;
	%end;
	%else %do;
		*--- get the number of QTL and effects in the input model ---*;
		proc means data=hkmodel_tmp noprint;
			var nqtl neffect InfinityLike;
			output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
		run;

		*data OutLib.nqtlin_tmp; *set nqtlin_tmp; *run;

		data _null_;
			set nqtlin_tmp;
			call symputx("numbqtlin", numbqtl, 'L');
			call symputx("numbeffin", numbeff, 'L');
			call symputx("InfinityLikein", InfinityLike, 'L');
		run;	

		proc datasets nolist;
			delete nqtlin_tmp;
		quit;

		%put # # # # # # # # # # # # # # # # # # # # # # ;
		%put # # # Number of QTL = &numbqtlin # # # # # #;
		%put # # # Number of Effects = &numbeffin # # # #;
		%put # # # # # # # # # # # # # # # # # # # # # # ;
		
		%if %sysevalf(&numbqtlin < 1) %then %do;
			%let flag_warn1_deleff = 1;	
			%let warn1_deleff_message= The deletion procedure of QTL effects needs a model with some QTL.;
			%put WARNING: &warn1_deleff_message;  				
		%end;
		%else %do;		
			%if(&qtl1 = &qtl2) and %length(&cofactorID) = 0 %then %do;
				*--- effect to be deleted is a main effect ---*;
		
 				%let case_ = ;
 				%let NumbDelMain_= 0;
 				%let NumbDelQxQ_= 0;
 				%let NumbDelQxC_= 0;
				data _null_;
					set &hkmimmodel end= _eof_;
					if _n_ = 1 then do;
						countDelMain_= 0;
						countDelQxQ_= 0;
						countDelQxC_= 0;	
					end;
					if trait = &trt and (qtl1 = qtl2 and (qtl1 = &qtl1 and qtl2 = &qtl1) and inter_cofactor_id = .) then do;
						*-- main --*;
						countDelMain_ = countDelMain_ + 1;
						*call symputx('case_', trim(left(contrast_)), 'L');
					end;
					else if trait = &trt and (qtl1 ^= qtl2 and (qtl1 = &qtl1 or qtl2 = &qtl1) and inter_cofactor_id = .) then do;
						*-- qxq interaction --*;
						countDelQxQ_ = countDelQxQ_ + 1;
					end;
					else if trait = &trt and (qtl1 = qtl2 and (qtl1 = &qtl1 and qtl2 = &qtl1) and inter_cofactor_id ^= .) then do;
						*-- qxc --*;
						countDelQxC_ = countDelQxC_ + 1;
					end;					
					retain countDelMain_ countDelQxQ_ countDelQxC_;
					if _eof_ then do;
						call symputx('NumbDelMain_', trim(left(countDelMain_)), 'L');
						call symputx('NumbDelQxQ_', trim(left(countDelQxQ_)), 'L');
						call symputx('NumbDelQxC_', trim(left(countDelQxC_)), 'L');
					end;					
					
				run;
				%put case_ = &case_;
				%put NumbDelMain_ = &NumbDelMain_;
				%put NumbDelQxQ_ = &NumbDelQxQ_;
				%put NumbDelQxC_ = &NumbDelQxC_;
							
				data &hkmimmodel;
					set &hkmimmodel;
					if trait = &trt and (qtl1 = qtl2 and (qtl1 = &qtl1 and qtl2 = &qtl1) and inter_cofactor_id = .) then do;
						*-- main --*;
						delete;
					end;
					else if trait = &trt and (qtl1 ^= qtl2 and (qtl1 = &qtl1 or qtl2 = &qtl1) and inter_cofactor_id = .) then do;
						*-- qxq interaction --*;
						delete;
					end;
					else if trait = &trt and (qtl1 = qtl2 and (qtl1 = &qtl1 and qtl2 = &qtl1) and inter_cofactor_id ^= .) then do;
						*-- qxc interaction --*;
						delete;
					end;					
					else if neffect > 0 then do;
						nqtl = nqtl - 1;
						neffect = neffect - &NumbDelMain_ - &NumbDelQxQ_ - &NumbDelQxC_;
					end;
				run;		
			%end;
			%else %if(&qtl1 ^= &qtl2) and %length(&cofactorID) = 0 %then %do;
				*--- effect to be deleted is qxq interaction effect ---*;
				
				%let efftype1= ;
				%let efftype2= ;
				%if &ng < 4 %then %do;
					%if %upcase(&effect1) = ADDITIVE %then
						%let efftype1= 1;
					%else %if %upcase(&effect1) = DOMINANCE %then
						%let efftype1= 2;
					
					%if %upcase(&effect2) = ADDITIVE %then
						%let efftype2= 1;
					%else %if %upcase(&effect2) = DOMINANCE %then
						%let efftype2= 2;	
				%end;
				%else %do;
					***assuming maximum ng = 4 ***;
					%if %upcase(&effect1) = ADDITIVE2 %then
						%let efftype1= 1;
					%else %if %upcase(&effect1) = DOMINANCE %then
						%let efftype1= 2;
					%else %if %upcase(&effect1) = ADDITIVE1 %then
						%let efftype1= 3;
	
					%if %upcase(&effect2) = ADDITIVE2 %then
						%let efftype2= 1;
					%else %if %upcase(&effect2) = DOMINANCE %then
						%let efftype2= 2;
					%else %if %upcase(&effect2) = ADDITIVE1 %then
						%let efftype2= 3;				
				%end;
						
				data &hkmimmodel;
					set &hkmimmodel;
					if (trait = &trt) and 
						(qtl1 ^= qtl2) and
						(qtl1 = &qtl1 and qtl2 = &qtl2) and 
						(efftype1 = &efftype1 and efftype2 = &efftype2) and
						(inter_cofactor_id = .) then do;
						*-- qxq interaction --*;
						delete;
					end;
					else if neffect > 0 then do;
						neffect = neffect - 1;
					end;
				run;		
			%end;
			%else %if(&qtl1 = &qtl2) and %length(&cofactorID) > 0 %then %do;
				*--- effect to be deleted is qxc interaction effect ---*;
				%let efftype1= ;
				%if &ng < 4 %then %do;
					%if %upcase(&effect1) = ADDITIVE %then
						%let efftype1= 1;
					%else %if %upcase(&effect1) = DOMINANCE %then
						%let efftype1= 2;
				%end;
				%else %do;
					***assuming maximum ng = 4 ***;
					%if %upcase(&effect1) = ADDITIVE2 %then
						%let efftype1= 1;
					%else %if %upcase(&effect1) = DOMINANCE %then
						%let efftype1= 2;
					%else %if %upcase(&effect1) = ADDITIVE1 %then
						%let efftype1= 3;				
				%end;
												
				data &hkmimmodel;
					set &hkmimmodel;
					if (trait = &trt) and 
						(qtl1 = qtl2) and
						(qtl1 = &qtl1 and qtl2 = &qtl1) and 
						(efftype1 = &efftype1 and efftype2 = &efftype1) and
						(inter_cofactor_id = &cofactorID) then do;
						*-- qxc interaction --*;
						delete;
					end;
					else if neffect > 0 then do;
						neffect = neffect - 1;
					end;
				run;		
			%end;			
		%end; *--- end if numbqtlin > 0 ---*;
	%end; *--- end if NOBS_hkmodel_tmp >=1 ---*;

	proc datasets nolist;
		delete hkmodel_tmp;
	quit;

	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel); 
	
	ods exclude none;
	ods results;	
	
	%exitdeleff:;
	
%mend InteractivelyDelQTLEff;

/*InteractivelyAddCofactor() adds cofactor in the model*/
%macro InteractivelyAddCofactor(trt=, ng=, hkmimdata=, hkmimmodel=, cofactorID=, buildhkdata3=);

	ods exclude all;
	ods noresults;

	%global flag_warn_addc warn_addc_message;
	%let flag_warn_addc=0;
	%let warn_addc_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local totalqtl;
	%local totaleff;
	%local xcovarout;
	%local newx;
	%local totalnx;
	%local geffect_main_in;
	%local geffect_epi_in;
	%local geffect_qxc_in;
	%local InfinityLikeOut;
			
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitaddc;
	%end;
	
	*---crete temporary data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitaddc;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;	
	%put xcovar= &xcovar;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);	
		
	*--- get number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%if &InfinityLikein %then %do;
		%let flag_warn_addc = 1;
		%let warn_addc_message= No more cofactors can be added into the model because the residual variance is zero.;
		%put WARNING: &warn_addc_message;
		%goto exitaddc;
	%end;	

	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	%if %sysevalf(&numbqtlin > 0) %then %do;
	
		%if &buildhkdata3 %then %do;
			*--- begin Individual effects tests ---*;
			%let geffect_main_in= ;
			%let geffect_epi_in= ;
			%let geffect_qxc_in= ;

			proc sql;
				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_main_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id = .;

				select ('e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))||'*'||
					'e'||trim(left(put(efftype2,7.)))||'c'||trim(left(put(ch2,7.)))||'p'||trim(left(put(posid2,7.)))) into: geffect_epi_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 ^= qtl2 and inter_cofactor_id = .;

				select ('x'||trim(left(put(inter_cofactor_id,7.)))||'*'||
					'e'||trim(left(put(efftype1,7.)))||'c'||trim(left(put(ch1,7.)))||'p'||trim(left(put(posid1,7.)))) into: geffect_qxc_in separated by ' '
				from hkmodel_tmp
				where nqtl > 0 and qtl1 = qtl2 and inter_cofactor_id ^= .;				
			quit;
			%put geffect_main_in= &geffect_main_in;
			%put geffect_epi_in= &geffect_epi_in;
			%put geffect_qxc_in= &geffect_qxc_in;

			%prephkdata3(ng=&ng, qtlvars=&geffect_main_in, talldsn=&hkmimdata);

			proc datasets nolist;
				delete hkdata3f;
				change hkdata3=hkdata3f;
			quit;

			*data OutLib.ft_hkdata3f; *set hkdata3f; *run;
		%end;

		*--- get estimates and ProbT for each effect in the model (individual effect tests - ProbT) ---;
		*--- fit input model ---*;
		ods output CovParms=CovP;
		proc mixed data=hkdata3f method=ml;
			model y&trt = &xcovar &geffect_main_in &geffect_epi_in &geffect_qxc_in / solution;
			ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
		run;
		
		data fitmodel_st;
			set st_tmp1;
			if upcase(descr) = "-2 LOG LIKELIHOOD" then 
				call symputx("currNeg2LogLike", trim(left(value)), 'L');
			label value = "Value";
			format value 8.2;
		run;
		
		data CovP;
			set CovP;
			where  CovParm = 'Residual';
			if Estimate=0 then InfinityLike= 1;
			else InfinityLike= 0;
			call symputx("InfinityLikeOut", InfinityLike, "L");
		run;			
		
	%end;
	%else %do;
		
		%if &buildhkdata3 %then %do;				
			%prephkdata3(ng=&ng, qtlvars=, talldsn=&hkmimdata);
					
			proc datasets nolist;
				delete hkdata3f;
				change hkdata3=hkdata3f;
			quit;
			
			*data OutLib.ft_hkdata3f; *set hkdata3f; *run;
		%end;

		*--- fit input model ---*;
		ods output CovParms=CovP;
		proc mixed data=hkdata3f method=ml;
			model y&trt = &xcovar / solution;
			ods output fitstatistics=st_tmp1 solutionf=es_tmp1;
		run;
		
		data CovP;
			set CovP;
			where  CovParm = 'Residual';
			if Estimate=0 then InfinityLike= 1;
			else InfinityLike= 0;
			call symputx("InfinityLikeOut", InfinityLike, "L");
		run;
	%end;	
	
	%let totalqtl= &numbqtlin;
	%let totaleff= &numbeffin;
	%let newx= %str(x)%trim(%left(&cofactorID));
	%let xcovarout= %trim(%left(&xcovar &newx));
	%let totalnx = %sysevalf(&nx + 1);
	
	data &hkmimmodel;
		set &hkmimmodel;
		if trait= &trt then do;
			cofactors= "&xcovarout";
			ncofactors= &totalnx;
			InfinityLike= &InfinityLikeOut;
		end;
	run;
	*--- end Interactively adding cofactors in the model ---*;	

	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp;
	quit;
	
	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);
				
	%exitaddc:;
	
	ods exclude none;
	ods results;
		
%mend InteractivelyAddCofactor;

/*InteractivelyAddQTLxCofactor() adds interaction effect between Cofactor and QTL already in the model*/
%macro InteractivelyAddQTLxCofactor(trt=, ng=, hkmimmodel=, qtl1=, effect1=, cofactorID=);

	ods exclude all;
	ods noresults;

	%global flag_warn_addqxc warn_addqxc_message;
	%let flag_warn_addqxc=0;
	%let warn_addqxc_message=;
	
	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp;
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local totalqtl;
	%local totaleff;
			
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitaddqxc;
	%end;
	
	*---crete temporary data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the input model.;
		%let exiterror=1;
		%goto exitaddqxc;
	%end;
	
	*---get covariates in the input model---*;
	%let xcovar = ;
	data _null_;
		set hkmodel_tmp(obs=1);
		call symputx("xcovar", trim(left(cofactors)), 'L');
	run;	
	%put xcovar= &xcovar;
	
	%let u = 1;
	%let cofvar= %scan(&xcovar,&u);
	%do %while(%length(&cofvar));		
		%let u= %sysevalf(&u + 1);		
		%let cofvar= %scan(&xcovar,&u);
	%end;
	%let nx= %sysevalf(&u - 1);	
		
	*--- get number of QTL and effects in the input model ---*;
	proc means data=hkmodel_tmp noprint;
		var nqtl neffect InfinityLike;
		output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
	run;
	
	data _null_;
		set nqtlin_tmp;
		call symputx("numbqtlin", numbqtl, 'L');
		call symputx("numbeffin", numbeff, 'L');
		call symputx("InfinityLikein", InfinityLike, 'L');
	run;
	
	proc datasets nolist;
		delete nqtlin_tmp;
	quit;
	
	%if &InfinityLikein %then %do;
		%let flag_warn_addqxc = 1;
		%put WARNING: No more QTL effects can be added into the model because the residual variance is zero.;
		%goto exitaddqxc;
	%end;	
	
	%if %sysevalf(&numbqtlin < 1 or &nx < 1) %then %do;
		%let flag_warn_addqxc = 1;	

		%let warn_addqxc_message= ;
		%if &numbqtlin < 1 %then %do;
			%let warn_addqxc_message = &warn_addqxc_message The Interactively Add Cofactor x QTL Interaction procedure needs a model with at least 1 QTL.;
		%end;
		%if &nx < 1 %then %do;
			%let warn_addqxc_message = &warn_addqxc_message The Interactively Add Cofactor x QTL Interaction procedure needs a model with at least 1 cofactor.;		
		%end;		
		%put WARNING: &warn_addqxc_message;  	
		
		*--- deleting temporary data set ---*;
		proc datasets nolist;
			delete hkmodel_tmp;
		quit;
		
		%goto exitaddqxc;	
	%end;

	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	%*put # # # Number of QTL = &numbqtlin # # # # # #;
	%*put # # # Number of Effects = &numbeffin # # # #;
	%*put # # # # # # # # # # # # # # # # # # # # # # ;
	
	*--- begin Interactively adding two-way interaction among QTL in the model ---*;
	%let efftype1=;
	%if &ng < 4 %then %do;
		%if %upcase(&effect1) = ADDITIVE %then
			%let efftype1= 1;
		%else %if %upcase(&effect1) = DOMINANCE %then
			%let efftype1= 2;
	%end;
	%else %do;
		***assuming maximum ng = 4 ***;
		%if %upcase(&effect1) = ADDITIVE2 %then
			%let efftype1= 1;
		%else %if %upcase(&effect1) = DOMINANCE %then
			%let efftype1= 2;
		%else %if %upcase(&effect1) = ADDITIVE1 %then
			%let efftype1= 3;	
	%end;
	
	data _null_;
		set hkmodel_tmp;
		if qtl1= &qtl1 and efftype1 = &efftype1 then do;
			call symputx("chrom1", trim(left(ch1)));
			call symputx("pos1", trim(left(pos1)));
			call symputx("posid1", trim(left(posid1)));
			call symputx("epitype1", trim(left(efftype1)));
		end;
	run;
	
	%let totalqtl= &numbqtlin;
	%let totaleff= %sysevalf(&numbeffin + 1);
	
	data neweff_tmp;
		retain 
		trait &trt
		cofactors "&xcovar"
		ncofactors &nx
		inter_cofactor_id &cofactorID
		nqtl &totalqtl
		neffect &totaleff
		indic_epi_eff 0
		indic_qtl_without_inter . 	
		qtl1 &qtl1 
		qtl2 &qtl1
		ch1 &chrom1
		ch2 &chrom1
		pos1 &pos1
		pos2 &pos1
		posid1 &posid1
		posid2 &posid1
		efftype1 &epitype1
		efftype2 &epitype1
		contrast .;
	run;

	data &hkmimmodel;
		set &hkmimmodel neweff_tmp;
		if trait= &trt and neffect > 0 then do;
			nqtl = &totalqtl;
			neffect = &totaleff;
		end;
	run;

	proc datasets nolist;
		delete neweff_tmp;
	quit;
	*--- end Interactively adding two-way interaction among QTL in the model ---*;	

	*--- sort data set by trait ---*;
	proc sort data=&hkmimmodel;
		by trait;
	run;

	*--- deleting temporary data set ---*;
	proc datasets nolist;
		delete hkmodel_tmp;
	quit;
	
	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel);
				
	%exitaddqxc:;
	
	ods exclude none;
	ods results;
		
%mend InteractivelyAddQTLxCofactor;

/*InteractivelyDelCofactorEff() deletes one Cofactor from the model. If the Cofactor has interaction, the interaction is also deleted*/
%macro InteractivelyDelCofactorEff(trt=, hkmimmodel=, cofactorID=);

	ods exclude all;
	ods noresults;
	
	%put cofactorID= &cofactorID;

	%global flag_warn_delcofactor warn_delcofactor_message;
	%let flag_warn_delcofactor=0;
	%let warn_delcofactor_message=;

	%local nx;
	%local xcovar;
	%local cofvar;
	%local u;	
	%local id_hkmodel_tmp;
	%local NOBS_hkmodel_tmp; 
	%local numbqtlin;
	%local numbeffin;
	%local InfinityLikein;
	%local NumbDelEff_;
		
	*--- check if data set with input model exists ---*;
	%if ^%sysfunc(exist(&hkmimmodel)) %then %do;
		%put ERROR: Data set &hkmimmodel does not exist.;
		%let exiterror=1;
		%goto exitdelcofactor;
	%end;

	*--- create data set that holds the model from the current trait ---*;	
	data hkmodel_tmp;
		set &hkmimmodel;
		where trait = &trt;
	run;
	
	*--- check if trait = trt is in the input model --- *;
	%let id_hkmodel_tmp = %sysfunc(open(hkmodel_tmp));
	%let NOBS_hkmodel_tmp =%sysfunc(attrn(&id_hkmodel_tmp, NOBS));
	%let id_hkmodel_tmp = %sysfunc(close(&id_hkmodel_tmp)); *--- very important to close ---*;

	%if &NOBS_hkmodel_tmp < 1 %then %do;
		%put ERROR: Trait &trt is not in the current model.;	
		%let exiterror=1;
	%end;
	%else %do;
	
		*---get covariates in the input model---*;
		%let xcovar = ;
		data _null_;
			set hkmodel_tmp(obs=1);
			call symputx("xcovar", trim(left(cofactors)), 'L');
		run;
		
		%let u = 1;
		%let cofvar= %scan(&xcovar,&u);
		%do %while(%length(&cofvar));		
			%let u= %sysevalf(&u + 1);		
			%let cofvar= %scan(&xcovar,&u);
		%end;
		%let nx= %sysevalf(&u - 1);
	
		*--- get the number of QTL and effects in the input model ---*;
		proc means data=hkmodel_tmp noprint;
			var nqtl neffect InfinityLike;
			output out=nqtlin_tmp max=numbqtl numbeff InfinityLike;
		run;

		data _null_;
			set nqtlin_tmp;
			call symputx("numbqtlin", numbqtl, 'L');
			call symputx("numbeffin", numbeff, 'L');
			call symputx("InfinityLikein", InfinityLike, 'L');
		run;	

		proc datasets nolist;
			delete nqtlin_tmp;
		quit;

		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		%*put # # # Number of QTL = &numbqtlin # # # # # #;
		%*put # # # Number of Effects = &numbeffin # # # #;
		%*put # # # # # # # # # # # # # # # # # # # # # # ;
		
		%if %sysevalf(/*&numbqtlin < 1 or*/ &nx < 1) %then %do;
			%let flag_warn_delcofactor = 1;	
			%let warn_delcofactor_message= The deletion procedure of Cofactor effects needs a model with at least 1 cofactor.;
			%put WARNING: &warn_delcofactor_message;  				
		%end;
		%else %do;		
			
 			%let NumbDelEff_= 0;
			data _null_;
				set &hkmimmodel end= _eof_;
				if _n_ = 1 then do;
					countDelEff_= 0;
				end;					
				if trait = &trt and (qtl1 = qtl2 and inter_cofactor_id = &cofactorID) then do;
					*-- all effects in which cofactorID is present --*;
					countDelEff_ = countDelEff_ + 1;
				end;
				retain countDelEff_;
				if _eof_ then do;
					call symputx('NumbDelEff_', trim(left(countDelEff_)), 'L');
				end;
				
				
			run;
			%put NumbDelEff_ = &NumbDelEff_;

			data &hkmimmodel;
				set &hkmimmodel;
				if trait = &trt and find(cofactors, "x"||trim(left(&cofactorID))) then do;
					*-- cofactors --*;
					cofactors = trim(left(tranwrd(cofactors, "x"||trim(left(&cofactorID)), "")));
				end;
				if trait = &trt and inter_cofactor_id = &cofactorID then do;
					*-- all effects in which cofactorID is present --*;
					delete;
				end;					
				ncofactors = ncofactors - 1;
				if neffect > 0 then neffect = neffect - &NumbDelEff_;
			run;			
			
		%end; *--- end if nx > 0 ---*;
	%end; *--- end if NOBS_hkmodel_tmp >=1 ---*;

	proc datasets nolist;
		delete hkmodel_tmp;
	quit;

	*--- re-number QTL after sorting by chromosome and position ---*;
	%sortqtl(trt=&trt, hkmimmodel= &hkmimmodel); 
	
	ods exclude none;
	ods results;	
	
	%exitdelcofactor:;
	
%mend InteractivelyDelCofactorEff;



