/*
JMP Genomics Analytical Process 

Name: Basic Expression Workflow
Author: Russ Wolfinger, SAS Institute Inc.
History:
	sasrdw 08Jan08 initial coding 

Description: 
		This process runs a basic workflow
		for expression data, including options
		for quality control and BasicExpressionWorkflow.
*/

/*
-----------------------------------------------------------
Copyright (C) 2008 SAS Institute, Inc. All rights reserved. 
 
Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/



*ProcessVariables;

%global StudyName WFOutName InData ByVars LabelVar BaseMergeVars ChrVar PosVar WhereEddmt
AnnoInData AnnoMergeVars AnnoChrVar AnnoPosVar AnnoKeepVars MaxLength 
AnnoOrganism AnnoAccession AnnoLabel AnnoGeneSymbol AnnoDescription AnnoGeneId AnnoSnpId Anno2WhereStmt
ExpInData ColorVars ExpLabelVar GroupVars LComponents  OutDataComp PCAFixedEffects PCARandomEffects
Dist CorrPCA CorrGroup StandardizationMethod QCTime AccProportion VCFirstPC VCLastPC
ClassVars FixedEffects RandomEffects LsmeansEffects EstInFile JournalChrom
ClusterLSM LSMeansDiffSet LSMeansControlValues DiffInData UDiffs PValues AdjPValues
MultipleTestingMethod Alpha NegLog10pCutoff UseXaxisFilter XaxisCutoff XaxisDirection
LogFlag LogQC ShiftFactor /*LogNormalization ShiftFactor2*/ InDataLogQC InDataLogNorm DataType DataDistribution
TrackPath TrackFilter TrackFiles TrackAdd
OutPath ServerOutPath MacroPath ExonInData;

*ProcessBody;

%macro BasicmiRNAseqWorkflow;

* initialization and argument checking;
%let ProcessName = BasicmiRNAseqWorkflow; 

%include "&MacroPath./UtilityMacros.sas" / nosource;

* check methods applied;
%if &Dist=No & &CorrPCA=No & &CorrGroup=No & &StandardizationMethod=NONE &
	^%length(&FixedEffects) & ^%length(&RandomEffects) & ^%length(&LSMeansEffects) %then %do;
	%put ERROR: There is no Analysis Process selected.  Please check your settings on the last two tabs.;
	%goto exit;
%end;


%let TrackAdd = 0;
* check if the ClassVars is defined when Fixed effects or Random effects is specified;
%if %length(&FixedEffects) or %length(&RandomEffects) or %length(&LSMeansEffects) %then %do;
	%if ^%length(&ClassVars) %then %do;
		%put ERROR: Differential Expression Effects, Additional Fixed effects or Random effects should be either numeric or specified in the Nominal Variables on Design and Model .;
		%goto exit;	
	%end;
%end;
%let PCARandomEffects =  %trim(%left(&LSMeansEffects &FixedEffects &RandomEffects));
/*kemicl make these into new macros to follow one-way anova processing logic*/
%let OrigFixedEffects = &FixedEffects;
%let FixedEffects =  %trim(%left(&FixedEffects &LSMeansEffects));

%let ColorVars = ;
%if %length(&LSMeansEffects) %then %do;
	%let nLSmeansvars = %sysfunc(countw(&LSMeansEffects, "* "));
		%do ii=1 %to &nLSmeansvars;
			%let _tmpLSvar_ = %scan(&LSMeansEffects,&ii,'* ');
			%if ^%sysfunc(findw(&ColorVars,&_tmpLSvar_)) %then %do; 
				%let ColorVars = &ColorVars &_tmpLSvar_;
			%end;
		%end;
	%let ColorVars = %nrbquote(&ColorVars);
%end;

%GetOrigFileName(InData,OrigInData);

%let QCTime=Before;
%if &StandardizationMethod ne NONE %then %let QCTime=Both;

%CheckInData(_ExpFlag=1);

%if ^%length(&StudyName) %then %do;
	%if %length(&WFOutName) %then %let StudyName=&WFOutName;
	%else %let StudyName=&InData;
%end;

%if (&InPath ^= &OutPath) %then %do;
	data OutLib.&InData;
		set InLib.&InData;
	run;
%end;

%put ExpInPath = &ExpInPath;
%put OutPath = &OutPath;

%if (&ExpInPath ^= &OutPath) & ^%length(&WhereEddmt) %then %do;
	data OutLib.&ExpInData;
		set ExpInLib.&ExpInData;
	run;
%end;
%if (&syserr>4 | &exiterror) %then %goto exit;

%if %length(&InData)>27 %then %let InDataTmp=%substr(&InData,1,27);
%else %let InDataTmp=&InData;
%CheckDataName(_name=InDataLogQC,_default=&InDataTmp._logQ);

/*%CheckDataName(_name=InDataLogNorm,_default=&InDataTmp._log);*/

%let mRNA_Seq=0;
%let ANOVA_log=0;
%let ANOVA_unlog=0;
%if (&LogQC = yes and (&Dist=Yes | &CorrPCA=Yes | &CorrGroup=Yes) and (%index(&QCTime,After) | %index(&QCTime,Both))) %then %do;
	%let mRNA_Seq=1;
%end;		
* check if data has been log transformed;
%let _nFeatures = &_nTallCol_;
%put _nFeatures = &_nFeatures;
%if %upcase(&LogFlag) = YES %then %do;
	%let _MaxLogValue_ = 40; 
	%let _TopRowsToCheck_ = 5000;
	%let LogErr = 0;
	data _null_;
		set OutLib.&InData;
		array _C_ {&_nFeatures} %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
		do i = 1 to &_nFeatures;
			if _C_(i) > &_MaxLogValue_ then do;
				call symput("LogErr",1);
				stop;
			end;
		end;
		if _n_ > &_TopRowsToCheck_ then stop;
	run; 

	*the input data is not log-transformed, but the user specify the data as lot transformation;
	%if &LogErr %then %do;
		%put ERROR: There are large values in the input data. This implies the data are not log-transformed. Please uncheck the option that the input data is log-transformed.;
		%let exiterror = 1;
		%goto exit;
	%end;
%end;
%else %do;
	*no log transformed input data, but user specified mixed, we automatically do log transformation;
	%if &LogQC = yes %then %do;
		%put The data type you specified for Basic miRNA-Seq Workflow is count data, so we transform your input data to log scale.;
		%put InDataLogQC=&InDataLogQC;
		data OutLib.&InDataLogQC;
			set OutLib.&InData;
			array _C_ {&_nFeatures} %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
			do i = 1 to &_nFeatures;
				%if &ShiftFactor=0 %then %do;
					if _C_(i) >0 then _C_(i)=log2(_C_(i));
					else _C_(i)=.;
				%end;
				%else %do;
					_C_(i)=log2(_C_(i)+&ShiftFactor);
				%end;
			end;
			drop i;
		run;
	%end;
%end;

/*Start section to remove, need to deal with mRNA_Seq*/
/*
%if %index(&LogNormalization,Normalization) %then %do;
		data OutLib.&InDataLogNorm;
			set OutLib.&InData;
			array _C_ {&_nFeatures} %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
			do i = 1 to &_nFeatures;
				%if &ShiftFactor=0 %then %do;
					if _C_(i) >0 then _C_(i)=log2(_C_(i));
					else _C_(i)=.;
				%end;
				%else %do;
					_C_(i)=log2(_C_(i)+&ShiftFactor2);
				%end;
			end;
			drop i;
		run;
%end;
%if %index(&LogNormalization,ANOVA) %then %do;
	%let mRNA_Seq=1;
	%if &LogQC = no %then %let mRNA_Seq=2;
%end;
*/
/*End section to remove*/

%if %length(&AnnoOrganism) and %length(&AnnoGeneSymbol) %then %let mir_flag=1;
%else %let mir_flag=0;

%if (&Dist=Yes) %then %let ezdist = 1;
%else %let ezdist = 0;

%if (&CorrPCA=Yes) %then %let ezpca = 1;
%else %let ezpca = 0;

%if (&CorrGroup=Yes) %then %let ezgroup = 1;
%else %let ezgroup = 0;

%if %index(&StandardizationMethod,Quantile) %then %let StandardizationMethod=QKD;
%else %if %index(&StandardizationMethod,Loess) %then %let StandardizationMethod=LKD;

/*There is no miss_flag*/

%if %length(&FixedEffects) or %length(&RandomEffects) %then %do;
	
	* Check ByVars for determining applying OnWay or ANOVA;
	%let OneWayFlag = 1;
	%if %length(&ByVars) %then %do;
		proc sort data=InLib.&InData(keep=&ByVars) out=tmp nodupkey;
			by &ByVars;
		run;
		%nObsVars(tmp);
		%let nTmp = &nObs;
		%nObsVars(InLib.&InData);
		%if &nTmp ^= &nObs %then %let OneWayFlag = 0; * Mutilple obs per &ByVars so go with ANOVA;
	%end;
	
	%let ezanova = 1;
	%let f=1;
	%do %while(%length(%scan(&FixedEffects,&f,' |')));
		%let f = %eval(&f+1);
	%end;
	%let nf = %eval(&f-1);
	%let r=1;
	%do %while(%length(%scan(&RandomEffects,&r,' |')));
		%let r = %eval(&r+1);
	%end;
	%let nr = %eval(&r-1);
	/*Adding OnewayFlag*/
	%if (&nf <= 1) and (&nr <= 1) and ^%index(&LSMeansDiffSet,Simple) & &OneWayFlag and %upcase(&DataType)=CONTINUOUS %then %do;
		%let oneway = 1;
		%if &oneway %then %do;
			%if ^%length(&FixedEffects) %then %do;
				%put ERROR: At least one fixed effect (Model these Fixed Effects) must be specified.;
				%goto exit;
			%end;
		%end;
		%if %length(&DataDistribution) %then %do;
			%let oneway = 0;
		%end;
		* make effects blank separated;
		%let bf = ;
		%if (&nf = 1) %then %do;
			%let f=1;
			%do %while(%length(%scan(&FixedEffects,&f,'*()')));
				%let bf = &bf %scan(&FixedEffects,&f,'*()');
				%let f = %eval(&f+1);
			%end;
		%end;
		%let br = ;
		%if (&nr = 1) %then %do;
			%let r=1;
			%do %while(%length(%scan(&RandomEffects,&r,'*()')));
				%let br = &br %scan(&RandomEffects,&r,'*()');
				%let r = %eval(&r+1);
			%end;
		%end;
	%end;
	%else %do;
		%let oneway = 0;
	%end;

/*Add to Check ANOVA input*/
	%put nf=&nf;
	%* check if the input data is log transform or not and type of data is continuous or count data;
	%if &DataType =Continuous & &LogQC = no & &LogFlag = no & (^%length(&DataDistribution) | %index(&DataDistribution,Gaussian)) %then %let ANOVA_log=1;
	%if &mRNA_Seq=1 & &ANOVA_log=1 %then %let ANOVA_log=2;
	%if &DataType =Continuous & &LogFlag =yes %then %let ANOVA_log=2;
	%if &DataType =Count & &LogFlag = no %then %let ANOVA_log=0;		
	%if &DataType =Count & (&LogFlag = yes or (&LogQC = yes & &StandardizationMethod ne NONE)) %then %let ANOVA_unlog=1;
%end;
%else %let ezanova = 0;

%let OutName = &StudyName;
%if ^%length(&Alpha) %then %let Alpha = 0.05;
%if ^%length(&NegLog10pCutoff) %then %let NegLog10pCutoff = 0;

%let _FETmp_ =;
%if %length(&FixedEffects) %then %do;
	%let f=1;
	%do %while(%length(%scan(&FixedEffects,&f,'*()|')));
		%if ^%index(%upcase(&ClassVars),%upcase(%scan(&FixedEffects,&f,'*()|@'))) %then %do;
			%let _FETmp_ = &_FETmp_ %scan(&FixedEffects,&f,'*()|@');
		%end;
		%let f = %eval(&f+1);
	%end;
%end;
%*if %length(&_FETmp_) %then %let ClassVars =&ClassVars &_FETmp_;

%let _RETmp_ =;
%if %length(&RandomEffects) %then %do;
	%let r=1;
	%do %while(%length(%scan(&RandomEffects,&r,'*()|')));
		%if ^%index(%upcase(&ClassVars),%upcase(%scan(&RandomEffects,&r,'*()|@'))) %then %do;
			%let _RETmp_ = &_RETmp_ %scan(&RandomEffects,&r,'*()|@');
		%end;
		%let r = %eval(&r+1);
	%end;
%end;
%*if %length(&_RETmp_) %then %let ClassVars =&ClassVars &_RETmp_;

%if (&ezpca = 1) %then %do;

	%if %index(%nrbquote(&PCAFixedEffects),%quote(|)) %then %let FixedTmp = %sysfunc(tranwrd(%nrbquote(&PCAFixedEffects),%quote(|), ));
	%else %if %index(%nrbquote(&PCAFixedEffects),%quote(*)) %then %let FixedTmp = %sysfunc(tranwrd(%nrbquote(&PCAFixedEffects),%quote(*), ));
	%else %let FixedTmp = &PCAFixedEffects;
	%if %index(%nrbquote(&PCARandomEffects),%quote(|)) %then %let RandomTmp = %sysfunc(tranwrd(%nrbquote(&PCARandomEffects),%quote(|), ));
	%else %if %index(%nrbquote(&PCARandomEffects),%quote(*)) %then %let RandomTmp = %sysfunc(tranwrd(%nrbquote(&PCARandomEffects),%quote(*), ));
	%else %let RandomTmp=&PCARandomEffects;
	%if %index(%nrbquote(&FixedTmp),%quote(@)) %then %let FixedTmp2 = %substr(&FixedTmp, 1, %index(%nrbquote(&FixedTmp),%quote(@))-1);
	%else %let FixedTmp2 = &FixedTmp;
	%if %index(%nrbquote(&RandomTmp),%quote(@)) %then %let RandomTmp2 = %substr(&RandomTmp, 1, %index(%nrbquote(&RandomTmp),%quote(@))-1);
	%else %let RandomTmp2 = &RandomTmp;
	%let ClassVars2 = &FixedTmp2 &RandomTmp2 ;
%end;

/*kemicl reset Fixed Effects */
%let FixedEffects = &OrigFixedEffects;

* create JSL file;
%global BEW_Script;
%let BEW_Script=;
%PrepOutFile(BEW_Script,BEW_ScriptName,&ProcessName Script for &InData..jsl);

%let allc = ;
%do i = 1 %to &_nAllColChunk_; 
	%let allc = &allc &&_AllCol_&i._;
%end;
%put allc = &allc;


* put JSL statements;
data _null_;
   file "&BEW_Script";
	put "//!";
	put "Names Default to Here(1);";	
	put "if(hostis(windows), _slash_ = ""\"", _slash_ = ""/"");";
	put "if(Host Is(Windows),";
	put "   jmpGenomicsDir = ConvertFilePath(""$LS_HOME"", absolute, windows),";
	put "   jmpGenomicsDir = ConvertFilePath(""$LS_HOME"", absolute, POSIX)";
	put ");";
	put "include(jmpGenomicsDir||""Generator""||_slash_||""WizardHelperFunctions.jsl"");";
	put "include(jmpGenomicsDir||""Generator""||_slash_||""WorkflowHelperFunctions.jsl"");";
	put "jg:_BEWF_ = 1;";
	put "here:EZ_ProcessName = ""&ProcessName"";";
	put "here:JMPG_WF_Flag =1;";
	put "here:EZ_AE_StudyID = ""&StudyName"";";
	put "EZ_AE_CHPIE = 0;";
	put "EZ_AE_CHPFormat = "" "";";
	put "CDFRadio = 0;";
	put "LabelVar = ""&LabelVar"";";
	put "BaseMergeVars = ""&BaseMergeVars"";";
	put "ChrVar = ""&ChrVar"";";
	put "PosVar = ""&PosVar"";";
	put "LComponents = ""&LComponents"";";
	put "OutDataComp=""&OutDataComp"";";
	put "jg:AnnoRadio = 3;";
	put "jg:AnnoInData = ""&AnnoInData"";";
	put "AnnoMergeVars = ""&AnnoMergeVars"";";
	put "AnnoAccession = ""&AnnoAccession"";";
	put "jg:AnnoOrganism = ""&AnnoOrganism"";";
	put "AnnoLabel = ""&AnnoLabel"";";
	put "jg:AnnoChrVar = ""&AnnoChrVar"";";
	put "jg:AnnoPosVar = ""&AnnoPosVar"";";
	put "jg:AnnoGeneSymbol = ""&AnnoGeneSymbol"";";
	put "jg:AnnoDescription = ""&AnnoDescription"";";
	put "AnnoGeneId = ""&AnnoGeneId"";";
	put "Anno2WhereStmt = ""\[%nrbquote(&Anno2WhereStmt)]\"";";
	put "TrackPath = ""&TrackPath"";";
	put "jg:TrackFiles = ""&TrackFiles"";";
	put "TrackAdd = &TrackAdd;";
	put "ExonInData = ""&ExonInData"";";
	put "jg:EZ_AE_OutPath = ""&OutPath"";";
	put "EZ_AE_ExpInPath =""&ExpInPath"";";
	put "jg:EZ_AE_ClientOutPath = ""&ClientOutPath"";";
	put "jg:EZ_AE_ServerOutPath = ""&ServerOutPath"";";
	put "jg:EZ_AE_InData = ""&InData"";";
	put "jg:DataType=""&DataType"";";
	put "here:ANOVA_log=&ANOVA_log;";
	put "here:ANOVA_unlog=&ANOVA_unlog;";
	%if &LogQC = yes %then %do;
		put "jg:EZ_InData_LogQC = ""&InDataLogQC"";";
	%end;
	/*Start remove*/
	/*
	%if %index(&LogNormalization,Normalization) %then %do;
		put "jg:EZ_InData_LogNorm = ""&InDataLogNorm"";";
		put "jg:logNorm=1;";
	%end;
	*/
	/*End remove*/
	put "jg:ShiftFactor = ""&ShiftFactor"";";
	/*put "jg:ShiftFactor2 = ""&ShiftFactor2"";";*/
	put "EZ_AE_ExpInData = ""&ExpInData"";";
	put "ColorVars = ""&ColorVars"";";
	put "ExpLabelVar = ""&ExpLabelVar"";";
	put "GroupVars = ""&GroupVars"";";
	put "jg:MIRNA = ""yes"";";
	put "jg:mRNA_Seq=&mRNA_Seq;";
	put "jg:mir_flag = &mir_flag;";
	put "jg:LogFlag=""&LogFlag"";";	
	put "jg:AllCol=""&allc"";";
	%if %index(&QCTime,Before) %then %do;	
		put "jg:EZ_AE_Dist = &ezdist;";
		put "jg:EZ_AE_CorrPCA = &ezpca;";
		put "jg:EZ_AE_CorrGroup = &ezgroup;";
		put "jg:EZ_AE_QC2 = 0;";
		put "jg:EZ_AE_Dist2 = 0;";
		put "jg:EZ_AE_CorrPCA2 = 0;";
		put "jg:EZ_AE_CorrGroup2 = 0;";
	%end;
	%else %if %index(&QCTime,After) %then %do;;
		put "jg:EZ_AE_Dist = 0;";
		put "jg:EZ_AE_CorrPCA = 0;";
		put "jg:EZ_AE_CorrGroup = 0;";		
		put "jg:EZ_AE_QC2 = 1;";
		put "jg:EZ_AE_Dist2 = &ezdist;";
		put "jg:EZ_AE_CorrPCA2 = &ezpca;";
		put "jg:EZ_AE_CorrGroup2 = &ezgroup;";
	%end;
	%else %do;
		put "jg:EZ_AE_Dist = &ezdist;";
		put "jg:EZ_AE_CorrPCA = &ezpca;";
		put "jg:EZ_AE_CorrGroup = &ezgroup;";
		put "jg:EZ_AE_QC2 = 1;";
		put "jg:EZ_AE_Dist2 = &ezdist;";
		put "jg:EZ_AE_CorrPCA2 = &ezpca;";
		put "jg:EZ_AE_CorrGroup2 = &ezgroup;";	
	%end;
	put "jg:NormMethod = ""&StandardizationMethod"";";
	put "jg:EZ_AE_ANOVA = &ezanova;";
	put "ClassVars = ""&ClassVars"";";
	put "FixedEffects = ""&FixedEffects"";";
	put "RandomEffects = ""&RandomEffects"";";
	%if (&ezpca = 1) %then %do;
	put "AccProportion = ""&AccProportion"";";
	put "PCARandomEffects = ""&PCARandomEffects"";";
	put "PCAFixedEffects = ""&PCAFixedEffects"";";
	put "ClassVars2 = ""&ClassVars2"";";
	put "VCFirstPC = ""&VCFirstPC"";";
	put "VCLastPC = ""&VCLastPC"";";
	%end;
	%if (&ezanova = 1) %then %do;
		%if (&oneway = 1) %then %do;
			put "jg:EZ_AE_OneWay = 1;";
			put "FixedEffects = ""&bf"";";
			put "BlockVars = ""&br"";";
		%end;
		%else %do;
			put "jg:EZ_AE_OneWay = 0;";
			put "ByVars = ""&ByVars"";";
			put "DiffInData= ""&DiffInData"";";
			put "jg:DataDistribution=""&DataDistribution"";";
		%end;
		put "LSMeansEffects = ""&LSMeansEffects"";";
		put "EZ_AE_MeanDiff = ""&LSMeansDiffSet"";";
		put "ClusterLSM = ""yes"";";
		put "EZ_AE_ControlLevel1 = {};";
		put "Here:UDiffs = ""&UDiffs"";";
		put "Here:PValues = ""&PValues"";";
		put "Here:AdjPValues = ""&AdjPValues"";";
		%let len = %length(&LSMeansControlValues);
		%put LSMeansControlValues = &LSMeansControlValues;
		%put len = &len;
		%if &len %then %do;
			%let LSM = %quote(&LSMeansControlValues);
			%put LSM = &LSM;
			%let LSMCV = ;
			%do i=1 %to &len;
				%let c1 = %qsubstr(&LSM,&i,1);
				%put i=&i c1=&c1;
				%if ^%length(&c1.&c1) %then %do;
					%let LSMCV = &LSMCV.%str( );
				%end;
				%else %if &c1.&c1 = %str(%"%") %then %do;
					%let LSMCV = &LSMCV.\!&c1;
				%end;
				%else %do;
					%let LSMCV = &LSMCV.&c1;
				%end;
			%end;
			%put LSMCV = &LSMCV;
			put "c1 = ""&LSMCV"";";
			put "lc1 = length(c1);";
			put "c2 = substr(c1,2,lc1-2);";
			put "InsertInto(EZ_AE_ControlLevel1,c2);";
		%end;
		put "EZ_AE_ControlLevel2 = {};";
		put "EstInFile = ""&EstInFile"";";
		put "JournalChrom = ""&JournalChrom"";";
		put "MultiTest = ""&MultipleTestingMethod"";";
		put "UseXaxisFilter = ""&UseXaxisFilter"";";
		put "XaxisCutoff = ""&XaxisCutoff"";";
		put "XaxisDirection = ""&XaxisDirection"";";
		put "EZ_AE_Alpha = &Alpha;";
		put "Cutoff = &NegLog10pCutoff;";
	%end;
	put "EZ_AE_uploadGene = 0;";
	put "_CreateWorkflow;";
	put 'jg:_BEWF_=.;';
	put "here:JMPG_WF_Flag =0;";
run;

%PackageInsertFile(jsl code from &ProcessName, &OutPath, &BEW_ScriptName, , text/plain, yes);

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend BasicmiRNAseqWorkflow;


* call the macro;
%BasicmiRNAseqWorkflow;


