/*
JMP Genomics Analytical Process

Name: Build Consensus Linkage Map
Author: Luciano Silva
History:
	sbrlsi 03/11/15 Initial Coding
Description:
		Build Consensus Linkage Map takes as input a series of linkage maps that will be used to estimate a single consensus linkage map 
		via linear programming optimization subject to marker order constraints imposed by each input linkage map as well as
		optional weights for each input linkage map, as in Endelman and Plomion (2014, Bioinformatics 30(11):16231624). 
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InPath FileFilter InMapFiles LGWhereStmt MaxWindowList MapWeightList  
	ServerOutPath OutPath OutName MacroPath BuildConsLinkMap_Script 
;

%macro BuildConsensusLinkageMap;

	%let FlagDebug= 0;

	ods html close;
	ods listing close;
	*ods html;
	*ods listing;

   	data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
   		or164_enabled= sysprod('PRODNUM164');
    		if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,1);
   	run;

	%global LGVar MarkerVar PositionVar;
	%let LGVar = LinkageGroup; /*must be in each map file*/
	%let MarkerVar = MarkerName; /*must be in each map file*/
	%let PositionVar = MarkerPosition; /*must be in each map file*/

	**ProcessName must be declared before including UtilityMacros*;
	%global FullAPName ProcessName;
	%let FullAPName = %str(Build Consensus Linkage Map);	
	%let ProcessName = BuildConsensusLinkageMap;

	***begin: Including macros*;
	**The macro variable ClientOutpath is created upon including the UtilityMacros*;
	%include "&MacroPath./UtilityMacros.sas" / nosource;
	%CheckEndSlash(InPath);
	***end: Including macros*;
	
	**defining libnames**;
	libname InLib "&InPath";
	libname OutLib "&OutPath";

	**begin: extract file names**;
	%let f = 1;
	%let fLen = 0;
	%let FileList = %trim(%left(&InMapFiles)); 
	%put FileList = &FileList;
	%do %while(%length(&FileList));
		* extract one file name;
		%let flag = 0;
		%let name = ;
		%do %While(^&flag);
			%let e = %index(&FileList,%str(|));
			%if ^&e %then %do;
				%let name = %left(&name &FileList);
				%let flag = 1;
				%let FileList = ;
			%end;
			%else %do;
				%let name = %left(&name %substr(&FileList,1,&e-1));
				%if %index(&name,.) %then %let flag = 1;
				%let FileList = %substr(&FileList,&e+1);
			%end;
		%end;
		%let InFile&f = &name;
		%put InFile&f = &&InFile&f;
		%let ext = %scan(&name,-1);
		%let InMapName&f = %substr(&name,1,%length(&name)-%length(&ext)-1);
		%put InMapName&f = &&InMapName&f;
		%CheckDataName(_name=InMapName&f);
		%put InMapName&f = &&InMapName&f;

		%let f = %eval(&f+1);
	%end;
	%*let nFiles = %eval(&f-1);
	%*put nFiles = &nFiles;
	%let nMaps= %sysevalf(&f - 1);
	%put nMaps= &nMaps;
	**end: extract file names**;
	
	**Checking First File Name and creaging OutName in case it is blank*;
	%CheckDataName(_name=OutName,_default=&InMapName1);
	%put OutName = &OutName;
			
	%*goto exit;

	**begin: checking if number of elements in MapWeightList is the same as the number of map files**;
	%put MapWeightList = &MapWeightList;
	%if %length(&MapWeightList) %then %do;
		%let WgtNoNumber = 0;
		%let minWgt = 1000000000;
		%let k = 1;
		%do %while(%length(%scan(&MapWeightList, &k, %str( ))) and &WgtNoNumber = 0);
			%if %sysfunc(inputn(%scan(&MapWeightList, &k, %str( )), Best20)) = %str(.) %then %do;			
				%let WgtNoNumber= 1;
			%end;		
			%else %do;
				%put k= &k %scan(&MapWeightList, &k, %str( ));
				%let minWgt = %sysfunc(min(&minWgt, %scan(&MapWeightList, &k, %str( ))));
			%end;
			%let k= %sysevalf(&k + 1);
		%end;
		%let nWgt= %sysevalf(&k - 1);
		%put nWgt= &nWgt;
		%put minWgt= &minWgt;

		%if &WgtNoNumber %then %do;
			%put ERROR: All elements in the Linkage Map Weight List must be NUMERIC values greater than zero.;
			%let exiterror = 1;
			%goto exit;
		%end;

		%if &minWgt <= 0 %then %do;
			%put ERROR: All elements in the Linkage Map Weight List must be numeric values GREATER THAN ZERO.;
			%let exiterror = 1;
			%goto exit;
		%end;

		%if &nMaps ^= &nWgt %then %do;
			%if &nMaps > &nWgt %then %do;
				%put ERROR: Number of elements in the Linkage Map Weight List is smaller than the number of map files listed in the Linkage Map Files field.;
			%end;
			%else %do;
				%put ERROR: Number of elements in the Linkage Map Weight List is greater than the number of map files listed in the Linkage Map Files field.;
			%end;
			%let exiterror = 1;
			%goto exit;
		%end;
		**end: checking if number of elements in MapWeightList is the same as the number of map files**;
	%end;
	%else %do;
		%let MapWeightList = ;
		%do i=1 %to &nMaps;
			%let MapWeightList = &MapWeightList 1;
		%end;
	%end;
	%put MapWeightList = &MapWeightList;

	**begin: creating data set of map weights**;
	data MapWeights;
		length _MapName $32;
		%do m=1 %to &nMaps;
			_MapIndex= &m;
			_MapName= "&&InMapName&m";
			_Weight= %scan(&MapWeightList, &m, %str( ));
			output;
		%end;
	run;
	
	%if &FlagDebug %then %do;
		data OutLib.MapWeights;	set MapWeights;	run;
	%end;
	**end: creating data set of map weights**;
	
	**begin: creating macro variables for each maximum interval window**;
	%let MaxNoNumber = 0;
	%let MaxNoInteger= 0;
	%let j=1;
	%let minmaxwd = 1000000;
	%let maxmaxwd = -1000000;
	%do %while(%length(%scan(&MaxWindowList, &j, %str( ))) and &MaxNoNumber = 0 and &MaxNoInteger = 0);
		%let maxwd&j = %scan(&MaxWindowList, &j, %str( ));
		%put maxwd&j = &&maxwd&j;
		%if %sysfunc(inputn(&&maxwd&j, Best20)) = %str(.) %then %do;			
			%let MaxNoNumber= 1;
		%end;
		%else %do;
			%if %sysfunc(INT(&&maxwd&j)) ^= &&maxwd&j %then %do;
				%let MaxNoInteger= 1;	
			%end;
			%else %do;
				%let minmaxwd = %sysfunc(min(&minmaxwd, &&maxwd&j));
				%let maxmaxwd = %sysfunc(max(&maxmaxwd, &&maxwd&j));			
			%end;
		%end;
		%let j= %sysevalf(&j + 1);
	%end;
	%let nmaxwd= %sysevalf(&j - 1);
	%put minmaxwd = &minmaxwd;
	%put maxmaxwd = &maxmaxwd;
	
	%if &MaxNoNumber %then %do;
		%put ERROR: All elements in the Maximum Interval Window List must be NUMERIC interger values greater than zero.;
		%let exiterror= 1;
		%goto exit;	
	%end;
		
	%if &MaxNoInteger %then %do;
		%put ERROR: All elements in the Maximum Interval Window List must be numeric INTEGER values greater than zero.;
		%let exiterror= 1;
		%goto exit;	
	%end;
	
	%if &minmaxwd <= 0 %then %do;
		%put ERROR: All elements in the Maximum Interval Window List must be numeric integer values GREATER THAN ZERO.;
		%let exiterror= 1;
		%goto exit;
	%end;

	**S1178649: fixing problem with unsorted input list*;
	data MaxWinTmp;	
		%do w=1 %to &nmaxwd;
			maxwind= &&maxwd&w;
			output;
		%end;	
	run;
	
	proc sort data=MaxWinTmp overwrite;
		by maxwind;
	run;
	
	%if &FlagDebug %then %do;
		data OutLib.MaxWinTmp; set MaxWinTmp; run;
	%end;
	
	data _null_;
		set MaxWinTmp;
		call symputx("maxwd"||trim(left(_n_)), trim(left(maxwind)));
	run;
	**end: creating macro variables for each maximum interval window**;

	**begin: collecting markers from all maps into a single data set**;
	%let k= 1;
	data &&InMapName&k;
		retain &MarkerVar &LGVar &PositionVar;
		set InLib.&&InMapName&k;
		keep &MarkerVar &LGVar &PositionVar;
	run;
	
	%let _dsd_ = %sysfunc(open(&&InMapName&k));
	%let _lgvarnumb = %sysfunc(varnum(&_dsd_, &LGVar));
	%let _lgvartype&k = %sysfunc(vartype(&_dsd_, &_lgvarnumb));
	%let _mkrvarnumb = %sysfunc(varnum(&_dsd_, &MarkerVar));
	%let _mkrvartype&k = %sysfunc(vartype(&_dsd_, &_mkrvarnumb));
	%let _posvarnumb = %sysfunc(varnum(&_dsd_, &PositionVar));	
	%let _posvartype&k = %sysfunc(vartype(&_dsd_, &_posvarnumb));	
	%let _rc_ = %sysfunc(close(&_dsd_));
	
	%if &_lgvarnumb = 0 %then %do;
		%put ERROR: Variable &LGVar is not present in file &&InMapName&k..;
		%let exiterror = 1;
		%goto exit;
	%end;
	
	%if &_mkrvarnumb = 0 %then %do;
		%put ERROR: Variable &MarkerVar is not present in file &&InMapName&k..;
		%let exiterror = 1;
		%goto exit;
	%end;
	
	%if &_posvarnumb = 0 %then %do;
		%put ERROR: Variable &PositionVar is not present in file &&InMapName&k..;
		%let exiterror = 1;
		%goto exit;
	%end;
	%put _lgvartype&k _mkrvartype&k _posvartype&k = &&_lgvartype&k &&_mkrvartype&k &&_posvartype&k;
	
	%let LGlen= 8;
	%if &&_lgvartype&k = C %then %let LGlen= $32;
	%let MKlen= 8;
	%if &&_mkrvartype&k = C %then %let MKlen= $32;

	proc sort data= &&InMapName&k overwrite;
		by &LGVar &PositionVar &MarkerVar;
	run;

	data &&InMapName&k;
		set &&InMapName&k;
		by &LGVar &PositionVar;
		_lagpos= lag(&PositionVar);
		if first.&LGVar then do;
			_StepPosMap&k = .;
			_BinMap&k = 1;
		end;
		else do;
			_StepPosMap&k = &PositionVar - _lagpos;
			if first.&PositionVar then do;
				_BinMap&k = _BinMap&k + 1;
			end;
		end;
		retain _BinMap&k;
		drop _lagpos;
	run;

	data MapAll_LGAll;
		length _ConsensusMarker &MKlen _ConsensusLG &LGlen;
		set &&InMapName&k;
		_ConsensusMarker = &MarkerVar;
		_ConsensusLG = &LGVar;
		rename 
			&MarkerVar = _MarkerMap&k 
			&LGVar = _LGMap&k 
			&PositionVar = _PosMap&k
		;
	run;
	
	%do k= 2 %to &nMaps;

		data &&InMapName&k;
			retain &MarkerVar &LGVar &PositionVar;
			set InLib.&&InMapName&k;
			keep &MarkerVar &LGVar &PositionVar;
		run;
		
		%let _dsd_ = %sysfunc(open(&&InMapName&k));
		%let _lgvarnumb = %sysfunc(varnum(&_dsd_, &LGVar));
		%let _lgvartype&k = %sysfunc(vartype(&_dsd_, &_lgvarnumb));
		%let _mkrvarnumb = %sysfunc(varnum(&_dsd_, &MarkerVar));
		%let _mkrvartype&k = %sysfunc(vartype(&_dsd_, &_mkrvarnumb));
		%let _posvarnumb = %sysfunc(varnum(&_dsd_, &PositionVar));	
		%let _posvartype&k = %sysfunc(vartype(&_dsd_, &_posvarnumb));	
		%let _rc_ = %sysfunc(close(&_dsd_));
	
		%if &_lgvarnumb = 0 %then %do;
			%put ERROR: Variable &LGVar is not present in file &&InMapName&k..;
			%let exiterror = 1;
			%goto exit;
		%end;
	
		%if &_mkrvarnumb = 0 %then %do;
			%put ERROR: Variable &MarkerVar is not present in file &&InMapName&k..;
			%let exiterror = 1;
			%goto exit;
		%end;
	
		%if &_posvarnumb = 0 %then %do;
			%put ERROR: Variable &PositionVar is not present in file &&InMapName&k..;
			%let exiterror = 1;
			%goto exit;
		%end;
		%put _lgvartype&k _mkrvartype&k _posvartype&k = &&_lgvartype&k &&_mkrvartype&k &&_posvartype&k;
		
		%if &_lgvartype1 ^= &&_lgvartype&k %then %do;
			%put ERROR: The variable &LGVar from Linkage Map File &&InMapName&k is not of same type as that of other Linkage Map Files. Variable &LGVar must be of same type for all Linkage Map Files.;
			%let exiterror = 1;
			%goto exit;
		%end;
		
		%if &_mkrvartype1 ^= &&_mkrvartype&k %then %do;
			%put ERROR: The variable &MarkerVar from Linkage Map File &&InMapName&k is not of same type as that of other Linkage Map Files. Variable &MarkerVar must be of same type for all Linkage Map Files.;
			%let exiterror = 1;
			%goto exit;
		%end;

		%if &_posvartype1 ^= &&_posvartype&k %then %do;
			%put ERROR: The variable &PositionVar from Linkage Map File &&InMapName&k is not of same type as that of other Linkage Map Files. Variable &PositionVar must be of same type for all Linkage Map Files.;
			%let exiterror = 1;
			%goto exit;
		%end;
		
		proc sort data= &&InMapName&k overwrite;
			by &LGVar &PositionVar &MarkerVar;
		run;

		data &&InMapName&k;
			retain &MarkerVar &LGVar &PositionVar;
			set &&InMapName&k;
			by &LGVar &PositionVar;
			_lagpos= lag(&PositionVar);
			if first.&LGVar then do;
				_StepPosMap&k = .;
				_BinMap&k = 1;
			end;
			else do;
				_StepPosMap&k = &PositionVar - _lagpos;
				if first.&PositionVar then do;
					_BinMap&k = _BinMap&k + 1;
				end;
			end;
			retain _BinMap&k;
			drop _lagpos;
		run;

		proc sql UNDO_POLICY= NONE;
			create table MapAll_LGAll as
			select *
			from
			MapAll_LGAll as a1 
			full join
			&&InMapName&k(rename=(
					&MarkerVar = _MarkerMap&k 
					&LGVar = _LGMap&k  
					&PositionVar = _PosMap&k
				)) as a2
			on 
			a1._ConsensusLG = a2._LGMap&k
			and
			a1._ConsensusMarker = a2._MarkerMap&k 
			order by _ConsensusLG, _ConsensusMarker
			;
		quit; 
		
		data MapAll_LGAll;
			set MapAll_LGAll;
			if missing(_ConsensusLG) then do;
				_ConsensusMarker= _MarkerMap&k;
				_ConsensusLG= _LGMap&k;
			end;
		run;

		proc sort data=MapAll_LGAll overwrite;
			by _ConsensusLG _ConsensusMarker;
		run;		
	%end;
	
	%if &FlagDebug %then %do;
		data OutLib.MapAll_LGAll0; set MapAll_LGAll; run;
	%end;

	%NObsVars(MapAll_LGAll);
	%if &nobs = 0 %then %do;
		%put ERROR: Merging input maps was not possible.;
		%let exiterror = 1;
		%goto exit;
	%end;
	**end: collecting markers from all maps into a single data set**;
	
	%*goto exit;

	**begin: applying where statement to filter out linkage groups**;
	%if %length(&LGWhereStmt) %then %do;
		%put LGWhereStmt= &LGWhereStmt;
		
		%let LGWhereStmt= %upcase(&LGWhereStmt);
		
		%if ^%index(&LGWhereStmt, LINKAGEGROUP) %then %do;
			%put ERROR: The where statement provided in the Filter to Include Linkage Groups field is invalid. This where statement must always reffer to the LinkageGroup variable.;
			/*This is causing error*/
			%*put For instance, LinkageGroup = "H1", if LinkageGroup is a character variable. Or, LinkageGroup = 1, if LinkageGroup is a numeric variable.;
			
			%let exiterror = 1;
			%goto exit;
		%end;
	
		%let LGWhereStmt= %sysfunc(tranwrd(&LGWhereStmt, LINKAGEGROUP, _CONSENSUSLG));
		%put LGWhereStmt= &LGWhereStmt;
	
		data MapAll_LGAll;
			set MapAll_LGAll;
			where &LGWhereStmt;
		run;
		
		%if %index(%upcase("&SYSERRORTEXT"), IS NOT ON FILE WORK.MAPALL_LGALL) %then %do;
			%put ERROR: The where statement provided in the Filter to Include Linkage Groups field is invalid. This where statement must always reffer to the LinkageGroup variable.;
			/*This is causing error*/
			%*put For instance, LinkageGroup = "H1", if LinkageGroup is a character variable. Or, LinkageGroup = 1, if LinkageGroup is a numeric variable.;
			
			%let exiterror = 1;
			%goto exit;		
		%end;

		%NObsVars(MapAll_LGAll);
		%if &nobs = 0 %then %do;
			%put ERROR: Applying the where statement &LGWhereStmt to filter linkage groups produced a data set with zero observations.;
			%let exiterror = 1;
			%goto exit;
		%end;
	%end;
	
	%if &FlagDebug %then %do;
		data OutLib.MapAll_LGAll1; set MapAll_LGAll; run;
	%end;
	**end: applying where statement to filter out linkage groups**;
	
	%*goto exit;
	
	**begin: checking for the presence of same marker on different linkage groups from distinct input map files**;
	data MapAll_LGAll_Long;
		*length &LGVar &LGlen _MapName $32 _MapIndex 8 &MarkerVar &MKlen &PositionVar 8;
		length &MarkerVar &MKlen &LGVar &LGlen _MapName $32 _MapIndex 8 &PositionVar 8;
		set MapAll_LGAll;
		%do i=1 %to &nMaps;  
			if not missing(_MarkerMap&i) then do;
				&LGVar= _LGMap&i;
				_MapName= "&&InMapName&i";
				_MapIndex= &i;
				&MarkerVar= _MarkerMap&i;
				&PositionVar= _PosMap&i;
				StepPosition= _StepPosMap&i;
				Bin= _BinMap&i;
				output;
			end;
		%end;
		keep &LGVar _MapName _MapIndex &MarkerVar &PositionVar StepPosition Bin;
	run;
	
	proc sort data= MapAll_LGAll_Long overwrite;
		by &MarkerVar &LGVar _MapIndex;
	run;
	
	%if &FlagDebug %then %do;
		data OutLib.MapAll_LGAll_Long1; set MapAll_LGAll_Long; run;
	%end;
	
	data LGsByMarker;
		set MapAll_LGAll_Long;
		by &MarkerVar &LGVar;
		if first.&LGVar then do;
			output;
		end;
	run;
	
	%if &FlagDebug %then %do;
		data OutLib.LGsByMarker1; set LGsByMarker; run;
	%end;	
	
	proc freq data= LGsByMarker noprint;
		tables &MarkerVar / out= LGsByMarkerFreq;
	run;
		
	%if &FlagDebug %then %do;
		data OutLib.LGsByMarkerFreq1; set LGsByMarkerFreq; run;
	%end;	
	
	data DupLGsByMarker;
		set LGsByMarkerFreq;
		if COUNT > 1 then output;
	run;
	
	%NObsVars(DupLGsByMarker);
	%let ndup= &nobs;
	
	%let LGLab = Linkage Group;
	%let MarkerLab = Marker Name;
	%let PositionLab = Marker Position;
	%let MapIndexLab = Map File Index;
	%let MapNameLab = Map File Name;
	
	%if &ndup %then %do;
		proc sql UNDO_POLICY= NONE;
			create table DupLGsByMarker(drop= StepPosition Bin _Marker_ COUNT PERCENT) as 
			select *
			from 
			MapAll_LGAll_Long as a1 
			right join 
			DupLGsByMarker(rename= (&MarkerVar= _Marker_)) as a2
			on a1.&MarkerVar = a2._Marker_;
		quit;
		
		%if &FlagDebug %then %do;
			data OutLib.DupLGsByMarker1; set DupLGsByMarker; run;
		%end;	
		
		%global BuildConsLinkMapDUP_OutData;
		%PrepOutDS(BuildConsLinkMapDUP_OutData,BuildConsLinkMapDUP_OutDataName,&OutName,_DUP);
		
		data OutLib.&BuildConsLinkMapDUP_OutDataName;
			set DupLGsByMarker;
			label &LGVar= &LGLab;
			label _MapName= &MapNameLab;
			label _MapIndex= &MapIndexLab;
			label &MarkerVar= &MarkerLab; 
			label &PositionVar= &PositionLab; 
		run;
		%PackageInsertDataset(Distinct Linkage Groups with Markers of Identical Names Data from &ProcessName, outlib, &BuildConsLinkMapDUP_OutDataName, no, no, no);						
		
		%global BuildConsLinkMap_ScriptErr;
		%PrepOutFile(BuildConsLinkMap_ScriptErr,BuildConsLinkMap_ScriptErrName,saserror.jsl);
		
		data _null_;
			file "&BuildConsLinkMap_ScriptErr" lrecl=32767;
			put '//!';
			put "Names Default to Here(1);";
			put "Here:dupfilename = ""&ClientOutPath.&BuildConsLinkMapDUP_OutDataName..sas7bdat""; ";
			put "New Window(""WARNING running &FullAPName"",showmenu(0),showtoolbars(0),";
			/*put "	tb1 = TextBox(""Process stopped because distinct linkage groups have instances of markers with identical names. A summary SAS Data Set was created and saved as:"", <<Set Wrap(1000)),";*/
			put "	tb1 = TextBox(""Distinct linkage groups have instances of markers with identical names. A summary SAS Data Set was created and saved as:"", <<Set Wrap(1000)),";
			put "	tb1 << Font Color(""Red""),";
			put "	tb2 = TextBox("" ""), ";
			put "	teb = TextEditBox(Here:dupfilename, <<Set Wrap(1000)),";
			put "	teb << Font Color(""Red""), ";
			put "	tb3 = TextBox("" ""),";
			put "	ButtonBox(""Click here to open the data set"",";
			put "		open(Here:dupfilename);";
			put "	);";
			put "); ";		
		run;
		
	  	%PackageInsertFile(jsl code from &ProcessName, &OutPath, &BuildConsLinkMap_ScriptErrName, no, no, yes);

		%put WARNING: Process stopped because different linkage groups have instances of markers with identical names.;
		%*let exiterror = 1;
		%*goto exit;
	%end;
	**end: checking for the presence of same marker on different linkage groups from distinct input map files**;

	**begin: collecting input linkage group names and number**;
	data LGNames;
		set MapAll_LGAll(keep= _ConsensusLG);
	run;

	proc sort data=LGNames nodupkey;
		by _ConsensusLG;
	run;

	data LGNames;
		set LGNames end=_eof_;
		call symputx("_lg"||left(_n_), _ConsensusLG);
		if _eof_ then call symputx("nlg", left(_n_));
	run;
	%put nlg= &nlg;
	**end: collecting input linkage group names and number**;

	data ConstrDroppedAll;
		length _ConstraintID 8 _MapIndex 8 _MapName $32 _LinkageGroup &LGlen 
			_LeftMarker &MKlen _LeftPosition 8 _LeftConsensusBin 8 
			_ConstraintSense $10 _RightMarker &MKlen _RightPosition 8 
			_RightConsensusBin 8;  
		delete;
	run;

	data ConsensusMapAllWide;
		length _ConsensusMarker &MKlen _ConsensusLG &LGlen _MaxWindow 8 
			_ConsensusPosition 8 _ConsensusStepPos 8 _ConsensusBin 8;
		delete;
	run;

	data ConsensusRMSEDataAll;
		length _ConsensusLG &LGlen _MaxWindow 8 _MapIndex 8 _RMSE 8;
		delete;
	run;

	data ConsensusRMSESummAll;
		length _ConsensusLG &LGlen _MaxWindow 8;
		delete;
	run;

	%let FlagExistAtLeast2Bins= 0;
	%let FlagExistAtLeast2Maps= 0;
	%let FlagExistAtLeast1FinalMap= 0;
	%let MaxWindFinal= &maxmaxwd;
	/*begin loop for linkage groups*/
	%do c= 1 %to &nlg; 
		%put Processing Linkage Group &&_lg&c.;

		**begin: selecting the current linkage group**;
		data MapAll_LG&c;
			set MapAll_LGAll;
			where 
			%if &_lgvartype1 = C %then %do;
				_ConsensusLG= "&&_lg&c";
			%end; 
			%else %do;
				_ConsensusLG= &&_lg&c;
			%end;	
		run;
		
		%if &FlagDebug %then %do;
			data OutLib.MapAll_LG&c._0; set MapAll_LG&c; run;
		%end;
		
		proc sort data= MapAll_LG&c overwrite;
			by _ConsensusMarker;
		run;
		**end: selecting the current linkage group**;

		**begin: creating list of bins for the consensus map**;
		data MapAll_LG&c;
			set MapAll_LG&c;
			array p{*} _BinMap:;
 			do i=1 to dim(p);	
				if missing(p[i]) then p[i]= 0;
			end;
			_conbin0 = trim(left(p[1])) %do j=2 %to &nMaps; ||trim(left(p[&j])) %end; ;
 			do i=1 to dim(p);	
				if p[i]= 0 then p[i]= .;
			end;
			drop i;
		run;

		proc sort data=MapAll_LG&c overwrite;
			by _conbin0;
		run;

		%let nconbins= 0;
		data MapAll_LG&c;
			set MapAll_LG&c end=_eof_;
			by _conbin0;
			if _n_ = 1 then _conbin1 = 0; 
			if first._conbin0 then do;
				_conbin1 = _conbin1 + 1;
			end;
			retain _conbin1;
			**total of bins in the consensus map**;
			if _eof_ then call symputx("nconbins", trim(_conbin1)); 
		run;
		%put nconbins= &nconbins;
		
		%if &FlagDebug %then %do;
			data OutLib.MapAll_LG&c._1; set MapAll_LG&c; run;
		%end;
		
		%if ^&nconbins %then %do;
			%put WARNING: Estimating a consensus map for linkage group &&_lg&c was not possible because this linkage group does not have a minimum of two markers on distinct positions.;
			%goto endLGloop;
		%end;
		%let FlagExistAtLeast2Bins= 1;

		**begin: might be unnecessary**;
		proc sort data=MapAll_LG&c overwrite;
			by _ConsensusMarker;
		run;

		data MapAll_LG&c;
			set MapAll_LG&c;
			by _ConsensusMarker;
			if _n_= 1 then do;
				sub_ = _conbin1;
				_conbin2 = 1;
			end;
			else do;
				if _conbin1 >= sub_ then _conbin2 = _conbin1 - sub_ + 1;
				else if _conbin1 < sub_ then _conbin2 = &nconbins - _conbin1 + 1;
			end;
			retain sub_ ;
			drop sub_;
		run;
		**end: might be unnecessary**;
		
		%if &FlagDebug %then %do;
			data OutLib.MapAll_LG&c._2; set MapAll_LG&c; run;
		%end;

		/*****************************************************************
		begin: finding constraints for the consensus map based on each map. 
		Also, collecting distance between markers on each map.
		******************************************************************/
		data ConstrLG&c._MapAll;
			length _MapIndex 8 _MapName $32 
				_ConsensusMarkerX &MKlen _ConsensusLGX &LGlen _MarkerMapX &MKlen _LGMapX &LGlen 
				PosMapX 8 _StepPosMapX 8 _BinMapX 8 _conbin0X $32767 _conbin1X 8 _conbin2X 8
				_ConsensusMarkerY &MKlen _ConsensusLGY &LGlen _MarkerMapY &MKlen _LGMapY &LGlen 
				PosMapY 8 _StepPosMapY 8 _BinMapY 8 _conbin0Y $32767 _conbin1Y 8 _conbin2Y 8
				;		
			delete;
		run;

		data DistPosLG&c._MapAll;
			length _Wd 8 _MapIndex 8 _MapName $32 _ConsensusMarker &MKlen 
				_ConsensusLG &LGlen _MarkerMap &MKlen _LGMap &LGlen _PosMap 8
				_StepPosMap 8 _BinMap 8 _conbin0 $32767 _conbin1 8 _conbin2 8 
				_conLBin 8 _conRBin 8 _MapLBin 8 _MapRBin 8 _DistPos 8 _Indic 8; 
			delete;
		run;

		%let TotConstLG&c= 0;
		%let NumbMapsLG&c= 0;
		%let MapIndexLG&c= ;
		%let MapNamesLG&c= ;
		%let MaxWindLG&c= &maxmaxwd;
		/*begin loop for maps*/
		%do i=1 %to &nMaps; 
			%put Processing Map &&InMapName&i for Linkage Group &&_lg&c.;
		
			data ConstrLG&c._Map&i;
				delete;
			run;
		
			**removes all maps that do not contain the current linkage group**;
			data MapAll_LG&c._Map&i;
				set MapAll_LG&c;
				where not missing(_MarkerMap&i); 
				keep _ConsensusLG _ConsensusMarker _MarkerMap&i _LGMap&i _PosMap&i _StepPosMap&i _BinMap&i _conbin0 _conbin1 _conbin2;
			run;
			
			%if &FlagDebug %then %do;
				data OutLib.MapAll_LG&c._Map&i._o0; set MapAll_LG&c._Map&i; run;
			%end;

			%NObsVars(MapAll_LG&c._Map&i);
			%if &nobs %then %do;
				**collecting constraints within only for those maps that contain the current linkage group**;
				%put Collecting Constraints from Map &&InMapName&i for Linkage Group &&_lg&c.;
				
				%let NumbMapsLG&c= %sysevalf(&&NumbMapsLG&c + 1);
				%let MapIndexLG&c = &&MapIndexLG&c &i;
				%let MapNamesLG&c = &&MapNamesLG&c &&InMapName&i;

				proc sort data=MapAll_LG&c._Map&i overwrite;
					by _BinMap&i _conbin2;
				run;
				
				%if &FlagDebug %then %do;
					data OutLib.MapAll_LG&c._Map&i._o1; set MapAll_LG&c._Map&i; run;
				%end;
				
				data BinsByMap;
					set MapAll_LG&c._Map&i end=_eof_;
					by _BinMap&i _conbin2;
					if first._conbin2 then do;
						call symputx("_binByMap"||trim(left(_BinMap&i)), trim(left(_BinMap&i)));
						output;
					end;
					if _eof_ then call symputx("_nbinByMap", trim(left(_BinMap&i)));
				run;
				%put _nbinByMap= &_nbinByMap;
				
				%if &FlagDebug %then %do;
					data OutLib.BinsByMap; set BinsByMap; run;
				%end;
				
				%do n1= 2 %to &_nbinByMap;
					%let n0 = %sysevalf(&n1 - 1);
					proc sql UNDO_POLICY= NONE;
						create table ConstraintData as
						select *
						from 
						MapAll_LG&c._Map&i 
							(rename=(_ConsensusMarker= _ConsensusMarkerX 
							_ConsensusLG= _ConsensusLGX _MarkerMap&i= _MarkerMapX _LGMap&i= _LGMapX
							_PosMap&i= PosMapX _StepPosMap&i= _StepPosMapX _BinMap&i=_BinMapX 
							_conbin0= _conbin0X _conbin1= _conbin1X _conbin2= _conbin2X) 
							where = (_BinMapX= &&_binByMap&n0))
						,
						MapAll_LG&c._Map&i 
							(rename=(_ConsensusMarker= _ConsensusMarkerY 
							_ConsensusLG= _ConsensusLGY _MarkerMap&i= _MarkerMapY _LGMap&i= _LGMapY
							_PosMap&i= PosMapY _StepPosMap&i= _StepPosMapY _BinMap&i=_BinMapY
							_conbin0= _conbin0Y _conbin1= _conbin1Y _conbin2= _conbin2Y) 
							where = (_BinMapY= &&_binByMap&n1))
						;
					quit;
					
					data ConstrLG&c._Map&i;
						set ConstrLG&c._Map&i ConstraintData;
					run;
				%end;
				
				proc sort data= ConstrLG&c._Map&i overwrite nodupkey;
					by _conbin2X _conbin2Y;
				run;
				
				proc sort data= ConstrLG&c._Map&i overwrite; /**testing purpose**/
					by _BinMapX _BinMapY;
				run;
					
				%if &FlagDebug %then %do;
					data OutLib.ConstrLG&c._Map&i; set ConstrLG&c._Map&i; run;
				%end;
				
				%NObsVars(ConstrLG&c._Map&i);
				%let ncontrmap&i = &nobs;
				%put ncontrmap&i = &&ncontrmap&i;
				%let TotConstLG&c= %sysevalf(&&TotConstLG&c + &&ncontrmap&i);				
									
				**begin: collecting constraints from all maps into a single data set**;
				data ConstrLG&c._MapAll;
					set 
					ConstrLG&c._MapAll
					ConstrLG&c._Map&i(in=_b_) end = _eof_;
					if _b_ then do;
						_MapIndex= &i;
						_MapName= "&&InMapName&i";
					end;
				run;
				**end: collecting constraints from all maps into a single data set**;				
				
				%*goto exit;

				/*******************************************************************************
				begin: collecting distance between markers on each map.
				Each map will be circularized and distances will be collected by wraping-
				around each map in order to make the consensus map length corresponding  
				with the average length of all single maps, as in Endelman and Plomion (2014).
				*********************************************************************************/
				proc sort data= MapAll_LG&c._Map&i out= DistPosLG&c._Map&i nodupkey;
					by _conbin2;
				run;
				
				proc sort data= DistPosLG&c._Map&i overwrite;
					by _BinMap&i;
				run;
				
				**begin: need to stack data set twice to wrap-around each linkage group*;
				data DistPosLG&c._Map&i;
					set DistPosLG&c._Map&i(in= _a_) DistPosLG&c._Map&i end=_eof_;
					if _a_ then _rep_ = 1;
					else _rep_ = 2; 
					if _eof_ then call symputx("nconbinbymap", left(_n_/2));
				run; 
				%put LG Map nconbinbymap MaxWindLG&c= &c &i &nconbinbymap;	
				
				proc sort data=DistPosLG&c._Map&i out=DistPosLG&c._Map&i;
					by _rep_ _BinMap&i;
				run;

				data DistPosLG&c._Map&i;
					set DistPosLG&c._Map&i;
					by _rep_ _BinMap&i;
					**begin: loop for interval window**;
					%do w = 1 %to %sysfunc(min(&maxmaxwd, &nconbinbymap - 1));
						if ((&w < &nconbinbymap) and (_n_ >= &w + 1) and (_n_ <= &nconbinbymap + &w))
							then _IndicLag&w = 1;
						_mapbinlag&w= lag&w(_BinMap&i);
						_conbinlag&w= lag&w(_conbin2);
						_poslag&w= lag&w(_PosMap&i);
						_WdLag&w = &w;
						if _n_ = 1 then do;
							_mapLBinLag&w = .;
							_mapRBinLag&w = .;
							_conLBinLag&w = .;
							_conRBinLag&w = .;
							_DistPosLag&w = .;
						end;
						else do;
							if _mapbinlag&w <= _BinMap&i then do;
								_mapLBinLag&w = _mapbinlag&w;
								_mapRBinLag&w = _BinMap&i;
								_conLBinLag&w = _conbinlag&w;
								_conRBinLag&w = _conbin2;
							end;
							else do;
								_mapLBinLag&w = _BinMap&i;
								_mapRBinLag&w = _mapbinlag&w;
								_conLBinLag&w = _conbin2;
								_conRBinLag&w = _conbinlag&w;							
							end;
							_DistPosLag&w = abs(_PosMap&i - _poslag&w);
						end;
						
					%end; 
					**end: loop interval window**;
				run;
				
				%if &FlagDebug %then %do;
					data OutLib.DistPosLG&c._Map&i._b; set DistPosLG&c._Map&i; run;
				%end;
				**end: need to stack data set twice to wrap-around each linkage group*;
								
				%*goto exit;

				**begin: loop for interval window**;
				%do w = 1 %to %sysfunc(min(&maxmaxwd, &nconbinbymap - 1));
					data DistPosLG&c._MapAll;
						set DistPosLG&c._MapAll
							DistPosLG&c._Map&i(
							keep = _ConsensusMarker _ConsensusLG _MarkerMap&i _LGMap&i _PosMap&i _StepPosMap&i _BinMap&i _conbin0 _conbin1 _conbin2 
								_WdLag&w
								_IndicLag&w
								_mapLBinLag&w
								_mapRBinLag&w
								_conLBinLag&w
								_conRBinLag&w
								_DistPosLag&w
							rename = (_MarkerMap&i= _MarkerMap
									_LGMap&i= _LGMap
									_PosMap&i= _PosMap
									_StepPosMap&i= _StepPosMap
									_BinMap&i= _BinMap
									_WdLag&w=_Wd 
									_IndicLag&w=_Indic 
									_mapLBinLag&w= _mapLBin
									_mapRBinLag&w= _mapRBin
									_conLBinLag&w= _conLBin
									_conRBinLag&w= _conRBin
									_DistPosLag&w= _DistPos
								)
							in = _b_ 
							);
						if _b_ then do;
							_MapIndex= &i;
							_MapName= "&&InMapName&i";
						end;
					run;
				%end;
				**end: loop for interval window**;
				
				**collect maximum interval window on each linkage group across input maps**;
				%let MaxWindLG&c= %sysfunc(min(&&MaxWindLG&c, &nconbinbymap - 1));
				
				/********************************************************************************
				end: collecting distance between markers on each map.
				Each map will be circularized and distances will be collected by wraping-
				around each map in order to make the consensus map length corresponding  
				with the average length of all single maps, as in Endelman and Plomion (2014).
				**********************************************************************************/
			%end;
		%end; /*end loop for maps*/
		%put TotConstLG&c= &&TotConstLG&c;
		%put NumbMapsLG&c= &&NumbMapsLG&c;
		%put MapIndexLG= &&MapIndexLG&c;
		%put MapNamesLG= &&MapNamesLG&c;
		%put MaxWindLG= &&MaxWindLG&c;
		
		/****************************************************************
		end: finding constraints for the consensus map based each map. 
		Also, collecting distance between markers on each map.
		*****************************************************************/
		
		**Estimate consensus maps only if linkage group appears on at least two maps**;
		%if &&NumbMapsLG&c < 2 %then %do;
			%put WARNING: Estimating a consensus map for linkage group &&_lg&c was not possible because this linkage group appears on one input map only.;
			%goto endLGloop;
		%end;
		%let FlagExistAtLeast2Maps= 1;		
				
		**begin: creating constraints coefficients based on bins from the consensus map **;
		%if &FlagDebug %then %do;
			data OutLib.ConstrLG&c._MapAll0; set ConstrLG&c._MapAll; run;
		%end;

		data ConstrLG&c._MapAll;
			set ConstrLG&c._MapAll;
			array _bincol{&nconbins};
			_bincol[_conbin2X] = -1;
			_bincol[_conbin2Y] = 1;
			_ConstraintID= _n_;
		run;
		
		%if &FlagDebug %then %do;
			data OutLib.ConstrLG&c._MapAll1; set ConstrLG&c._MapAll; run;
		%end;
		**end: creating constraints coefficients based on bins from the consensus map **;
		
		**begin: creating matrix of constraints coefficients for the IML code**;
		data ConstrLG&c._Matrix;
			set ConstrLG&c._MapAll;
			array _bincol{*} _bincol:;
			do _i=1 to dim(_bincol);
				if missing (_bincol[_i]) then _bincol[_i] = 0;
			end;
			*keep _bincol:;
			drop _i;
		run;
		
		%if &FlagDebug %then %do;
			data OutLib.ConstrLG&c._Matrix; set ConstrLG&c._Matrix; run;
		%end;
		**end creating matrix of constraints coefficients for the IML code**;
				
		**begin: creating distance coefficients based on bins from the consensus map**;
		%if &FlagDebug %then %do;
			data OutLib.DistPosLG&c._MapAll0; set DistPosLG&c._MapAll; run;
		%end;
		
		data DistPosLG&c._MapAll;
			set DistPosLG&c._MapAll;
			array _bincol{&nconbins};
			if _Indic = 1 then do;
				*if _mapLBin <= _mapRBin then do; /*has to be based on bins from each map file*/
					_bincol[_conLBin] = -1;
					_bincol[_conRBin] = 1;
				*end;
				*else do;
				*	_bincol[_conLBin] = 1;
				*	_bincol[_conRBin] = -1;
				*end;
			end;
		run;
			
		%if &FlagDebug %then %do;
			data OutLib.DistPosLG&c._MapAll1; set DistPosLG&c._MapAll; run;
		%end;
		**end: creating distance coefficients based on bins from the consensus map**;

		**begin: creating matrix of distance coefficients for the consensus map and distances between marker on each map to be used in the IML code**;
		data DistPosLG&c._Matrix;
			set DistPosLG&c._MapAll;
			array _bincol{*} _bincol:;
			do _i=1 to dim(_bincol);
				if missing (_bincol[_i]) then _bincol[_i] = 0;
			end;
			where _Indic = 1 and _Wd <= &&MaxWindLG&c; /*selecting maximum interval window that are smaller than or equal to the number of unique consensus bins for each linkage group across all input maps*/
			drop _i;
		run;
		
		%if &FlagDebug %then %do;
			data OutLib.DistPosLG&c._Matrix; set DistPosLG&c._Matrix; run;
		%end;
		**end: creating matrix of distance coefficients for the consensus map and distances between marker on each map to be used in the IML code**;
		
		%*goto exit;

		**begin: selecting maximum feasible subsystem (Chinneck 2001)**;
		%let FindSubSysStatus= 0;
		%let NConstrToDrop= 0;
		proc iml;
			/* minimize object*x, subject to coef*x <= b */
			use ConstrLG&c._Matrix;
			read all Var ("_ConstraintID") into AllRows; /* matrix of order constraint */
			*nrk= nrow(AllRows);
			*nck= ncol(AllRows);
			*print 'nrk nck from AllRows' nrk nck;
			read all Var ("_bincol1":"_bincol&nconbins") into A; /* matrix of order constraint */
			nra = nrow(A);
			nca = ncol(A);
			*create OutLib.Amat from A;
			*append from A; 
			*print A;
			coef = A || Diag(J(nra, 1, 1)); /* nra x (nca + nra) */						
			nrc = nrow(coef); /* nrc= nra */
			ncc = ncol(coef); /* ncd= nca + nra */
			*print coef;
			object = J(nca, 1, 0)//J((ncc - nca), 1, 1); /* ncc x 1 */
			*print object;
			b = J(nrc, 1, 1); /* nrc x 1 */
			*print b;
			do i=1 to nrc;	
				** nrc x 1 *;
				rowsense= rowsense // 'G'; 
			end;
			*print rowsense (nrow(rowsense));

			/*
			do i=1 to ncc;
				** ncc x 1 *;
				coltype = coltype // 'C'; 
			end;
			*print coltype (nrow(coltype));
			cntl= {1, ., ., ., ., ., ., ., ., ., ., ., ., .};
			CALL MILPSOLVE(rc, objvalue, x, relgap, object, coef, b, cntl, coltype, rowsense, , , );
			print 'seleting maximum feasible' rc, objvalue, xtoprint, relgap;
			*/
			
			cntl= {1, ., ., ., ., ., ., .};
			*cntl= {1, ., ., ., ., 1, ., .};
			CALL LPSOLVE(rc, objvalue, x, dual, reducost, object ,coef, b, cntl, rowsense, , , );
			*xtoprint=x[(nca+1):ncc];
			*print 'seleting maximum feasible' rc, objvalue, xtoprint;

			if rc ^= 0 then do; /**checking status**/
				/**error solving the system**/
				call symputx("FindSubSysStatus", left(char(rc)));
				abort;
			end;
			else do; 
				if objvalue < 1e-6 then do;
					/**system is feasible**/
					call symputx("NConstrToDrop", left(char(0)));
					*abort;
				end;
				else do;
					DropRows= {};
					ElasticVars = x[(nca+1):ncc];
					*print 'ElasticVars' ElasticVars;
					SelectedIndex = loc(ElasticVars > 1e-4);
					*print 'SelectedIndex' SelectedIndex;
					if ncol(SelectedIndex) = 1 then do;
						DropRows = DropRows // SelectedIndex[1];							
					end;
					else do;
						SelectedValue = ElasticVars[SelectedIndex];
						*print 'SelectedValue' SelectedValue;
						SelectedValueAndIndex = SelectedValue || SelectedIndex`;
						*print 'SelectedValueAndIndex' SelectedValueAndIndex;
						call SORTNDX(SortOrder, SelectedValueAndIndex, {1}, {1});
						*print 'SortedIndex' SortOrder;	
						SelectedIndexSorted = SelectedIndex[SortOrder];
						*print 'SelectedIndexSorted' SelectedIndexSorted;						
					
						nloop = 1;
						ContinueWhileLoop = 1;
						do while(ContinueWhileLoop);
						*do while(ContinueWhileLoop & nloop <= 1); /** using nloop for testing purpose **/
							MinValue = 1e10; 
							TestIndexSorted = SelectedIndexSorted;
							*print 'nloop TestIndexSorted' nloop TestIndexSorted;
							nTest= nrow(TestIndexSorted);
							do z=1 to nTest;
								flagEndLoops= 0;
								free coltype rowsense; /**needs to be reset here**/
								
								if nrow(DropRows) = 0 then do;
									DropRowsTmp= TestIndexSorted[z];
								end;
								else do;
									DropRowsTmp= DropRows // TestIndexSorted[z];
								end;			
								SubAllRows = remove(AllRows, DropRowsTmp);		
								
								subA = A[SubAllRows,]; 
								nra = nrow(subA);
								nca = ncol(subA);
								*print subA;
								coef = subA || Diag(J(nra, 1, 1));
								nrc = nrow(coef); 
								ncc = ncol(coef); 
								*print coef;
								object = J(nca, 1, 0)//J((ncc - nca), 1, 1); 
								*print object;
								b = J(nrc, 1, 1); 
								*print b;
								do i=1 to nrc;	
									rowsense= rowsense // 'G'; 
								end;
								*print rowsense (nrow(rowsense));
															
								/*
								do i=1 to ncc;
									coltype = coltype // 'C'; 
								end;
								*print coltype (nrow(coltype));
								cntl= {1, ., ., ., ., ., ., ., ., ., ., ., ., .};
								CALL MILPSOLVE(rc, objvalue, x, relgap, object, coef, b, cntl, coltype, rowsense, , , );
								print 'nloop seleting maximum feasible' nloop rc, objvalue, xtoprint, relgap;						
								*/
								
								cntl= {1, ., ., ., ., ., ., .};
								*cntl= {1, ., ., ., ., 1, ., .};
								CALL LPSOLVE(rc, objvalue, x, dual, reducost, object ,coef, b, cntl, rowsense, , , );
								*xtoprint=x[(nca+1):ncc];
								*print 'nloop z DropRowsTmp seleting maximum feasible' nloop z DropRowsTmp rc, objvalue, xtoprint;

								if rc ^= 0 then do; /**checking status**/
									/**error solving the system**/
									call symputx("FindSubSysStatus", left(char(rc)));
									abort;
								end;
								else do;
									if objvalue < 1e-6 then do;
										/**system is feasible**/
										DropRows= DropRowsTmp;
										z= nTest + 1; /*stops the do loop*/
										flagEndLoops= 1;
									end;
									else do;
										if objvalue < MinValue then do;
											MinValue = objvalue;
											DropRowBest = TestIndexSorted[z];
											
											/*get row indexes from SUB matrix A*/
											subElasticVars = x[(nca+1):ncc];
											subSelectedIndex = loc(subElasticVars > 1e-4);											
																						
											subSelectedValue = subElasticVars[subSelectedIndex];
											*print 'nloop z subSelectedValue' nloop z subSelectedValue;
											subSelectedValueAndIndex = subSelectedValue || subSelectedIndex`;
											*print 'nloop z subSelectedValueAndIndex' nloop z subSelectedValueAndIndex;
											call SORTNDX(subSortOrder, subSelectedValueAndIndex, {1}, {1});
											*print 'nloop z subSortedIndex' nloop z subSortOrder;	
											subSelectedIndexSorted = subSelectedIndex[subSortOrder];
											*print 'nloop z subSelectedIndexSorted' nloop z subSelectedIndexSorted;
											
											/*get row indexes from FULL matrix A*/
											SelectedIndexSorted= SubAllRows[subSelectedIndexSorted];
											*print 'nloop z SelectedIndexSorted' nloop z SelectedIndexSorted;
											
											if nrow(SelectedIndexSorted) = 1 then do;
												DropRowFinal= SelectedIndexSorted[1];
											end;
											else do;
												DropRowFinal= {.};										
											end;
										end;
									end;/*objvalue >= 1e-6*/
								end;/*rc = 0*/
							end; /*end do loop*/
							
							if flagEndLoops then do;
								ContinueWhileLoop= 0; /*stops the while loop*/
								*print 'nloop ContinueWhileLoop' nloop ContinueWhileLoop;
							end;
							else do;
								DropRows= DropRows // DropRowBest;
								
								*print 'nloop DropRowFinal' nloop DropRowFinal;
								if DropRowFinal ^= . then do;
									DropRows= DropRows // DropRowFinal;
									ContinueWhileLoop= 0; /*stops the while loop*/
								end;
								else do;
									*DropRows= DropRows // DropRowBest;
									ContinueWhileLoop= 1; /*continue the while loop*/	
								end;
							end;
							
							nloop = nloop + 1;
						end; /*end while loop*/
					end;/*else statement for ncol(SelectedIndex) > 1*/	
					
					nDropped= nrow(DropRows);
					call symput("NConstrToDrop", left(char(nDropped)));
					if nDropped > 0 then do;
						*print 'DropRows' DropRows;
						create ConstrToDrop from DropRows[colname="_DropConstr"];
						append from DropRows; 
						KeepRows= remove(AllRows, DropRows)`;
						create ConstrToKeep from KeepRows[colname="_KeepConstr"];
						append from KeepRows; 
					end;
				end; /*end else statement for objvalue >= 1e-6*/
			end; /*end if statment for status = 0*/
		quit;
		%put FindSubSysStatus= &FindSubSysStatus;
		%put NConstrToDrop= &NConstrToDrop;
		**end: selecting maximum feasible subsystem (Chinneck 2001)**;
		
		**begin: checking if the selection of feasible subsystem worked properly**;
		%if &FindSubSysStatus %then %do;
			%if &rc = 1 %then %let StatusString= the time limit was exceeded;
			%else %if &rc = 2 %then %let StatusString= the number of node limit was exceeded;
			%else %if &rc = 3 %then %let StatusString= the solution is infeasible;
			%else %if &rc = 4 %then %let StatusString= the solution is unbounded or infeasible;
			%else %if &rc = 5 %then %let StatusString= the subroutine could not obtain enough memory;
			%else %if &rc = 6 %then %let StatusString= the subroutine failed to solve the problem;
			%put ERROR: Estimating a consensus map for linkage group &&_lg&c failed because &StatusString..;
			%let exiterror = 1;
			%goto exit;
		%end; 
		**end: checking if the selection of feasible subsystem worked properly**;

		%*goto exit;

		**begin: collecting constraints to keep and to drop**;
		%if &NConstrToDrop > 0 %then %do;
		
			%if &FlagDebug %then %do;
				data OutLib.ConstrToKeep0; set ConstrToKeep; run;
			%end;
				
			**begin: creating a data set of all subset constraints that are feasible**;
			proc sql UNDO_POLICY= NONE;
				create table Constr_Matrix as
				select *
				from ConstrToKeep as a1 left join ConstrLG&c._Matrix a2
				on a1._KeepConstr= a2._ConstraintID
				order by _ConstraintID;
			quit;
			**end: creating a matrix of subset constraints that are feasible**; 
			
			**begin: creating a data set with dropout constraints**;
			%if &FlagDebug %then %do;
				data OutLib.ConstrToDrop0; set ConstrToDrop; run;
			%end;

			proc sql UNDO_POLICY= NONE;
				create table ConstrToDrop(drop= _DropConstr)as
				select *
				from ConstrToDrop as a1 left join ConstrLG&c._Matrix a2
				on a1._DropConstr= a2._ConstraintID
				order by _ConstraintID;
			quit;	
			
			%if &FlagDebug %then %do;
				data OutLib.ConstrToDrop1; set ConstrToDrop; run;
				data OutLib.ConstrLG&c._Matrix; set ConstrLG&c._Matrix; run;
			%end;

			data _null_;
				set ConstrToDrop;
				call symputx("ConstrID"||trim(left(_n_)), trim(left(_ConstraintID)));
				call symputx("MapIndexForConstr"||trim(left(_n_)), trim(left(_MapIndex)));
				call symputx("MapNameForConstr"||trim(left(_n_)), trim(left(_MapName)));
				call symputx("conLBinForConstr"||trim(left(_n_)), trim(left(_conbin2X)));
				call symputx("conRBinForConstr"||trim(left(_n_)), trim(left(_conbin2Y)));
			run;

			%do nc= 1 %to &NConstrToDrop;
				%put ConstrID Map LBin RBin = &&ConstrID&nc &&MapIndexForConstr&nc &&conLBinForConstr&nc &&conRBinForConstr&nc;
			
				%if &FlagDebug %then %do;
					data Outlib.MapAll_LG&c; set MapAll_LG&c; run;
				%end;
			
				data ConstrDropL ConstrDropR;
					set MapAll_LG&c;
					if _conbin2 = &&conLBinForConstr&nc then output ConstrDropL;
					if _conbin2 = &&conRBinForConstr&nc then output ConstrDropR;
					keep _MarkerMap&&MapIndexForConstr&nc _LGMap&&MapIndexForConstr&nc 
						_PosMap&&MapIndexForConstr&nc _BinMap&&MapIndexForConstr&nc 
						_conbin2
					;
				run;
				
				%if &FlagDebug %then %do;
					data OutLib.ConstrDropL; set ConstrDropL; run;
					data OutLib.ConstrDropR; set ConstrDropR; run;
				%end;

				proc sql UNDO_POLICY= NONE;
					create table ConstrDropped as
					select *
					from 
					ConstrDropL(rename=(_MarkerMap&&MapIndexForConstr&nc=_LeftMarker 
								_LGMap&&MapIndexForConstr&nc=_LeftLG 
								_PosMap&&MapIndexForConstr&nc= _LeftPosition 
								_BinMap&&MapIndexForConstr&nc= _LeftBin
								_conbin2= _LeftConsensusBin)), 
					ConstrDropR(rename=(_MarkerMap&&MapIndexForConstr&nc=_RightMarker 
								_LGMap&&MapIndexForConstr&nc=_RightLG 
								_PosMap&&MapIndexForConstr&nc= _RightPosition 
								_BinMap&&MapIndexForConstr&nc= _RightBin
								_conbin2= _RightConsensusBin))
					;
				quit;
				
				%if &FlagDebug %then %do;
					data OutLib.ConstrDropped; set ConstrDropped; run;
				%end;
			
				data ConstrDroppedAll;
					set 
						ConstrDroppedAll 
						ConstrDropped(
							in= _a_ 
							rename=(_LeftLG = _LinkageGroup)
							drop= _RightLG _LeftBin _RightBin
							);
					if _a_ then do;
						_ConstraintSense = "<";
						_ConstraintID = &&ConstrID&nc;
						_MapIndex = &&MapIndexForConstr&nc;
						_MapName = "&&MapNameForConstr&nc";
					end;
				run;
			%end;
			
			%if &FlagDebug %then %do;
				data OutLib.ConstrDroppedAll; set ConstrDroppedAll; run;
			%end;
			**end: creating a data set with dropout constraints**;
		%end;
		%else %do;
			data Constr_Matrix;
				set ConstrLG&c._Matrix;
			run;
		%end;
		
		%if &FlagDebug %then %do;
			data OutLib.Constr_Matrix; set Constr_Matrix; run;
		%end;
		**end: collecting constraints to keep and to drop**;
		
		/*begin loop for interval window*/
		%do w = 1 %to &nmaxwd;
		
			%if %sysevalf(&&maxwd&w > &&MaxWindLG&c) %then %do;
				%goto endmaxwind3;
			%end;
		
			%let SolveSubSysStatus = 0;

			data DistPos_Matrix;
				set DistPosLG&c._Matrix;
				where _Wd <= &&maxwd&w;
				_ord_ = _n_;
			run;			

			proc means data=DistPos_Matrix N noprint;
				var _indic;
				by _MapIndex;
				output out=NError n=_NError;
			run;
			
			%if &FlagDebug %then %do;
				data OutLib.NError; set NError; run;
			%end;

			proc sql UNDO_POLICY= NONE;
				create table DistPos_Matrix(drop= _MapTmp) as
				select *
				from DistPos_Matrix as a1 left join NError(keep= _MapIndex _NError rename=(_MapIndex= _MapTmp)) as a2
				on a1._MapIndex = a2._MapTmp;

				create table DistPos_Matrix(drop= _ord_ _MapTmp) as
				select *
				from DistPos_Matrix as a1 left join MapWeights(keep= _MapIndex _Weight rename=(_MapIndex= _MapTmp)) as a2
				on a1._MapIndex = a2._MapTmp
				order by _ord_;
			quit;
			
			%if &FlagDebug %then %do;
				data OutLib.DistPos_Matrix; set DistPos_Matrix; run;
			%end;

			**begin: solving the selected feasible subsystem **;
			proc iml;
				/* minimize object*x, subject to coef*x <= b */
				use Constr_Matrix;
				read all var ("_bincol1":"_bincol&nconbins") into A; /* matrix of order constraints */
				nra = nrow(A);
				nca = ncol(A);
				*print A;
				use DistPos_Matrix;
				read all var ("_bincol1":"_bincol&nconbins") into H; /* matrix of distance coefficients */
				nrh = nrow(H);
				nch = ncol(H);
				coef = (A || J(nra, nrh, 0)) // (H || Diag(J(nrh, 1, 1))) // (-H || Diag(J(nrh, 1, 1)));
				nrc = nrow(coef); /* nrc= nra + 2*nrh */
				ncc = ncol(coef); /* ncc= nca + nrh. Notice that nca= nch */
				*print coef;
				use DistPos_Matrix;
				read all var ("_Weight") into W; /* vector of map weights */
				read all var ("_NError") into N; /* vector of number of distances */
				object = J(nca, 1, 0) // (W # (1/N)); /*(nca + nrh) x 1 */ 
				*print object (nrow(object));
				use DistPos_Matrix;
				read all var ("_DistPos") into d; /* vector of distances */
				nrd = nrow(d);
				ncd = ncol(d);
				b = J(nra, 1, 0) // d // -d; /* (nra + 2*nrd) x 1. Notice that nrd= nrh */
				*print b (nrow(b));
				do i=1 to nrc;	
					** nrc x 1 *;
					rowsense= rowsense // 'G'; 
				end;
				*print rowsense (nrow(rowsense));
				
				/*
				do i=1 to ncc;
					** ncc x 1 *;
					coltype = coltype // 'C'; 
				end;
				*print coltype (nrow(coltype));
				cntl= {1, ., ., ., ., ., ., ., ., ., ., ., ., .};
				CALL MILPSOLVE(rc, objvalue, x, relgap, object, coef, b, cntl, coltype, rowsense, , , );
				*print rc, objvalue, x, relgap;
				*/
				
				cntl= {1, ., ., ., ., ., ., .};
				*cntl= {1, ., ., ., ., 1, ., .};
				CALL LPSOLVE(rc, objvalue, x, dual, reducost, object ,coef, b, cntl, rowsense, , , );
				*print rc, objvalue, x;
				
				if rc ^= 0 then do; /*checking status*/
					/**error solving the system**/
					call symputx("SolveSubSysStatus", left(char(rc)));
					abort;
				end;
				else do;
					create SubSysSolution from x[colname="_PrimalSolution_"];
					append from x; 
				end;

			quit;
			**end: solving the selected feasible subsystem **;

			**begin: checking if system was solved properly**;
			%if &SolveSubSysStatus %then %do;
				%if &rc = 1 %then %let StatusString= the time limit was exceeded;
				%else %if &rc = 2 %then %let StatusString= the number of node limit was exceeded;
				%else %if &rc = 3 %then %let StatusString= the solution is infeasible;
				%else %if &rc = 4 %then %let StatusString= the solution is unbounded or infeasible;
				%else %if &rc = 5 %then %let StatusString= the subroutine could not obtain enough memory;
				%else %if &rc = 6 %then %let StatusString= the subroutine failed to solve the problem;
				%put ERROR: Estimating a consensus map for linkage group &&_lg&c failed because &StatusString..;
				%let exiterror = 1;
				%goto exit;
			%end;
			**end: checking if system was solved properly**;
			
			**begin: merging consensus marker list to estimated solution for marker position**;
			data SubSysSolution;
				set SubSysSolution;
				if _n_ <= &nconbins then do;
					_count = 0;
					_var= 'x'||left(_n_);
					_conbinsol= _n_;
				end;
				else do;
					_count = _count + 1;
					_var= 'z'||left(_count);
					_conbinsol= .;
				end;
				retain _count;
				drop _count;
			run;

			proc sql UNDO_POLICY= NONE;
				create table ConsensusMap as
				select *
				from MapAll_LG&c as a1 left join SubSysSolution(drop=_var rename=(_PrimalSolution_= _ConsensusPosition)) as a2
				on a1._conbin2= a2._conbinsol
				order by _ConsensusPosition, _ConsensusMarker;
			quit;
			
			data ConsensusMap;
				set ConsensusMap;
				by _ConsensusPosition;
				_lagpos= lag(_ConsensusPosition);
				if _n_ = 1 then do;
					_ConsensusStepPos = .;
				end;
				else do;
					_ConsensusStepPos = _ConsensusPosition - _lagpos;
				end;
				drop _lagpos;
			run;

			data ConsensusMapAllWide;
				*retain _ConsensusMarker _ConsensusLG _MaxWindow _ConsensusPosition _ConsensusBin;
				set ConsensusMapAllWide ConsensusMap(in= _b_ rename= (_conbinsol= _ConsensusBin));
				if _b_ then _MaxWindow= &&maxwd&w;
				drop _conbin0 _conbin1 _conbin2;
			run;
			**end: merging consensus marker list to estimated solution for marker position**;

			**begin: estimating RMSE**;
			%if &FlagDebug %then %do;
				data OutLib.ConsensusMap; set ConsensusMap; run;
			%end;
			
			data ConsensusRMSEData ConsensusRMSEDataTmp;
				set ConsensusMap end=_eof_;
				if _n_ = 1 then do;
					%do j= 1 %to &nMaps;
						_sumSqDevMap&j= 0;
						_nobsMap&j = 0;
					%end;
				end;
				%do j= 1 %to &nMaps;
					_SqDevMap&j= (_ConsensusPosition - _PosMap&j)**2;
					if not missing(_SqDevMap&j) then do;
						_sumSqDevMap&j= _sumSqDevMap&j + _SqDevMap&j;
						_nobsMap&j = _nobsMap&j + 1;
					end;
				%end;
				if _eof_ then do;
					%do j= 1 %to &nMaps;
						RMSEMap&j = sqrt(_sumSqDevMap&j/_nobsMap&j);
					%end;
					output ConsensusRMSEData;
				end;
				retain 
					%do j= 1 %to &nMaps;
						_sumSqDevMap&j _nobsMap&j
					%end;
				;
				output ConsensusRMSEDataTmp;
			run;
			
			%if &FlagDebug %then %do;
				data OutLib.ConsensusRMSEDataTmp; set ConsensusRMSEDataTmp; run;
				data OutLib.ConsensusRMSEData0; set ConsensusRMSEData; run;
			%end;

			proc transpose data= ConsensusRMSEData out= ConsensusRMSEData(rename=(col1=_RMSE));
				var 
				%do j= 1 %to &nMaps;
					RMSEMap&j
				%end;
				;
			run;
			
			%if &FlagDebug %then %do;
				data OutLib.ConsensusRMSEData1; set ConsensusRMSEData; run;
			%end;

			data ConsensusRMSEDataAll;
				set ConsensusRMSEDataAll ConsensusRMSEData(in=_b_);
				if _b_ then do;
					_MaxWindow= &&maxwd&w;
					%if &_lgvartype1 = C %then %do;
						_ConsensusLG= "&&_lg&c";
					%end;
					%else %do;
						_ConsensusLG= &&_lg&c;
					%end;
					_MapIndex= input(tranwrd(_name_, 'RMSEMap', ''), best12.);
				end;
				drop _name_;
			run;

			%*goto exit;

			proc means data=ConsensusRMSEData mean std noprint;
				var _RMSE;
				output out=ConsensusRMSESumm(drop=_type_ _freq_) mean=_RMSE_Mean std=_RMSE_SD;
			run;						

			proc means data= ConsensusMap max noprint;
				var _ConsensusPosition;
				output out=ConsensusMapLength(drop=_type_ _freq_) max=_ConsensusMapLength;
			run;

			data ConsensusRMSESumm;
				merge ConsensusRMSESumm ConsensusMapLength;
			run;

			data ConsensusRMSESummAll;
				*retain _ConsensusLG _MaxWindow;
				set ConsensusRMSESummAll ConsensusRMSESumm(in=_b_);
				if _b_ then do;
					_MaxWindow= &&maxwd&w;
					%if &_lgvartype1 = C %then %do;
						_ConsensusLG= "&&_lg&c";
					%end;
					%else %do;
						_ConsensusLG= &&_lg&c;
					%end;
				end;
			run;						
			**end: estimating RMSE**;
			
			%let FlagExistAtLeast1FinalMap= 1;
			
			%endmaxwind3:;
		%end; /*end loop for interval window*/
		
		%let MaxWindFinal= %sysfunc(min(&MaxWindFinal, &&MaxWindLG&c));
		%endLGloop:;
	%end; /*end loop for linkage groups*/

	**begin: checking if a consensus map was estimated properly. Must pass all three flags**;	
	%if ^&FlagExistAtLeast2Bins %then %do;
		%put ERROR: Estimating consensus map was not possible because each linkage group does not have a minimum of two markers on distinct positions.;
		%let exiterror= 1;
		%goto exit;	
	%end;
	
	%if ^&FlagExistAtLeast2Maps %then %do;
		%put ERROR: Estimating consensus map was not possible because each linkage group appears on one input map only.;
		%let exiterror= 1;
		%goto exit;
	%end;
	
	%if ^&FlagExistAtLeast1FinalMap %then %do;
		%put ERROR: Estimating consensus map was not possible because solving the linear equations failed.;
		%let exiterror= 1;
		%goto exit;			
	%end;
	**end: checking if a consensus map was estimated properly. Must pass all three flags**;
	
	**begin: preparing output in long format**;
	data ConsensusMapAllLong;
		length &LGVar &LGlen _MaxWindow 8 _MapName $32 _MapIndex 8 &MarkerVar &MKlen &PositionVar 8;
		set ConsensusMapAllWide;
		if not missing(_ConsensusMarker) then do;
			_MapIndex= 0;
			_MapName= "Consensus";
			&MarkerVar= _ConsensusMarker;
			&LGVar= _ConsensusLG;
			&PositionVar= _ConsensusPosition;
			StepPosition= _ConsensusStepPos;
			Bin= _ConsensusBin;
			LG_Map_MaxWindow= trim(left(&LGVar))||"."||trim(left(_MapName))||".mw"||trim(left(_MaxWindow));
			LG_Con_MaxWindow= trim(left(&LGVar))||"."||trim(left(_MapName))||".mw"||trim(left(_MaxWindow));
			output;
		end;
		%do i=1 %to &nMaps;  
			if not missing(_MarkerMap&i) then do;
				_MapIndex= &i;
				_MapName= "&&InMapName&i";
				&MarkerVar= _MarkerMap&i;
				&LGVar= _LGMap&i;
				&PositionVar= _PosMap&i;
				StepPosition= _StepPosMap&i;
				Bin= _BinMap&i;
				_MaxWindow= 0;
				LG_Map_MaxWindow= trim(left(&LGVar))||"."||trim(left(_MapName));
				LG_Con_MaxWindow= trim(left(&LGVar));
				output;
			end;
		%end;
		*rename _MaxWindow= MaxWindow;
	run;
	
	proc sort data= ConsensusMapAllLong overwrite nodupkey;
		by &LGVar _MaxWindow _MapIndex &PositionVar &MarkerVar;
	run;
	
	%if &FlagDebug %then %do;
		data OutLib.ConsensusMapAllLong; set ConsensusMapAllLong; run;
	%end;
	**end: preparing output in long format**;
	
	**begin: Estimating map lengths**;	
	proc means data= ConsensusMapAllLong min max noprint;
		var &PositionVar;
		by &LGVar _MaxWindow _MapIndex _MapName;
		output out=MinMaxPos(drop = _type_ _freq_) min=_FirstPos max=_LastPos;
	run;
	
	data MapLengthDataAll;
		set MinMaxPos;
		if trim(left(_MapName)) ^= "Consensus" then _MaxWindow= 0;
		_MapLength= _LastPos - _FirstPos;
		if _MaxWindow ^= 0 then do;
			LG_Map_MaxWindow= trim(left(&LGVar))||"."||trim(left(_MapName))||".mw"||trim(left(_MaxWindow));
			LG_Con_MaxWindow= trim(left(&LGVar))||"."||trim(left(_MapName))||".mw"||trim(left(_MaxWindow));	
		end;
		else do;
			LG_Map_MaxWindow= trim(left(&LGVar))||"."||trim(left(_MapName));
			LG_Con_MaxWindow= trim(left(&LGVar));			
		end;
	run;
	
	%if &FlagDebug %then %do;
		data OutLib.MapLengthDataAll_1; set MapLengthDataAll; run;
	%end;
	
	proc sort data=MapLengthDataAll overwrite nodupkey;
		by &LGVar _MaxWindow _MapIndex;
	run;
	
	%if &FlagDebug %then %do;
		data OutLib.MapLengthDataAll_2; set MapLengthDataAll; run;
	%end;
	
	data _null_;
		set MapLengthDataAll end=_eof_;
		by &LGVar;
		if _n_ = 1 then nlgout= 0;
		if first.&LGVar then do;
			nlgout = nlgout + 1;
			call symputx("lgout"||trim(left(nlgout)),trim(left(&LGVar)));
		end;
		if _eof_ then call symputx("nlgout", trim(left(nlgout)));
		retain nlgout;
	run;
	**end: Estimating map lengths**;
	
	**begin: setting RMSE macro variables for the jsl code**;
	proc sql UNDO_POLICY= NONE;
		create table ConsensusRMSEDataAll(drop=_MapIndexTmp) as
		select * from
		ConsensusRMSEDataAll as a1 left join MapWeights(keep= _MapIndex _MapName rename=(_MapIndex= _MapIndexTmp)) as a2
		on a1._MapIndex = a2._MapIndexTmp
		order by _ConsensusLG, _MaxWindow, _MapIndex
		;
	quit;
	
	data ConsensusRMSEDataAll;
		set ConsensusRMSEDataAll;
		*LG_Map_MaxWindow= trim(left(_ConsensusLG))||"."||trim(left(_MapName))||".mw"||trim(left(_MaxWindow));
		*LG_Con_MaxWindow= trim(left(_ConsensusLG))||".mw"||trim(left(_MaxWindow));
	run;
	
	proc means data= ConsensusRMSEDataAll min max noprint;
		var _RMSE;
		by _ConsensusLG;
		output out=MinMaxRMSE min=_minrmse max=_maxrmse;
	run;
	
	data _null_;
		set MinMaxRMSE end=_eof_;
		call symputx("LGrmse"||trim(left(_n_)), trim(left(_ConsensusLG)));
		call symputx("MinRMSE"||trim(left(_n_)), trim(left(_minrmse)));
		call symputx("MaxRMSE"||trim(left(_n_)), trim(left(_maxrmse)));
		if _eof_ then call symputx("nLGrmse", trim(left(_n_)));
	run;
	%do i= 1 %to &nLGrmse;
		%let MinRMSE&i= %sysevalf(&&MinRMSE&i - 1);
		%let MaxRMSE&i= %sysevalf(&&MaxRMSE&i + 1);
	%end;
	**end: setting RMSE macro variables for the jsl code**;	
	
	**begin: setting consensus map macro variables for the jsl code**;
	proc means data= ConsensusMapAllLong min max noprint;
		var &PositionVar;
		output out=MinMaxPos min=_minpos max=_maxpos;
	run;
	
	data _null_;
		set MinMaxPos;
		call symputx("maxpos", trim(left(_maxpos + 11)));
		call symputx("minpos", trim(left(_minpos - 11)));
	run;		
	**end: setting consensus map macro variables for the jsl code**;
	
	**begin: create output SAS Data Sets**;
	%*let LGLab = Linkage Group;
	%*let MarkerLab = Marker Name;
	%*let PositionLab = Marker Position;
	%let ConsensusLGLab = &LGLab;
	%let MaxWindowLab = Maximum Interval Window;
	%*let MapIndexLab = Map File Index;
	%*let MapNameLab = Map File Name;
	%let RMSELab = Root Mean Square Error;	
	%let MapLengthLab = Map Length;
	%let LGMapMaxWindowLab= Linkage Group and Map File and Max Interval Window;
	%let LGConMaxWindowLab= Linkage Group and Consensus and Max Interval Window;
	%let StepPositionLab= Step Position;
	%let FirstPosLab= Starting Position;
	%let LastPosLab= Ending Position;
	%let BinLab= Marker Bin;
	
	%global BuildConsLinkMapCDW_OutData BuildConsLinkMapCDL_OutData 
		BuildConsLinkMapCRM_OutData BuildConsLinkMapCST_OutData
		BuildConsLinkMapCML_OutData;
	%PrepOutDS(BuildConsLinkMapCDW_OutData,BuildConsLinkMapCDW_OutDataName,&OutName,_cdw);
	%PrepOutDS(BuildConsLinkMapCDL_OutData,BuildConsLinkMapCDL_OutDataName,&OutName,_cdl);
	%PrepOutDS(BuildConsLinkMapCRM_OutData,BuildConsLinkMapCRM_OutDataName,&OutName,_crm);
	%PrepOutDS(BuildConsLinkMapCST_OutData,BuildConsLinkMapCST_OutDataName,&OutName,_cst);
	%PrepOutDS(BuildConsLinkMapCML_OutData,BuildConsLinkMapCML_OutDataName,&OutName,_cml);
	
	data OutLib.&BuildConsLinkMapCDW_OutDataName;
		set ConsensusMapAllWide(drop= _ConsensusBin);
		label _ConsensusMarker = "&MarkerLab for Consensus Map";
		label _ConsensusLG = "&LGLab for Consensus Map";
		label _MaxWindow = "&MaxWindowLab";
		label _ConsensusPosition = "&PositionLab for Consensus Map";
		label _ConsensusStepPos = "&StepPositionLab for Consensus Map";
		%do i=1 %to &nMaps;  
			label _MarkerMap&i = "&MarkerLab for Map &&InMapName&i";
			label _LGMap&i = "&LGLab for Map &&InMapName&i";
			label _PosMap&i = "&PositionLab for Map &&InMapName&i";
			label _BinMap&i = "&BinLab for Map &&InMapName&i";
			label _StepPosMap&i= "&StepPositionLab for Map &&InMapName&i";
		%end;
	run;
	%PackageInsertDataset(Consensus Map Data in Wide Format from &ProcessName, outlib, &BuildConsLinkMapCDW_OutDataName, _type=other);					

	data OutLib.&BuildConsLinkMapCDL_OutDataName;
		set ConsensusMapAllLong;
		drop 
			Bin
			_ConsensusMarker
			_ConsensusLG
			_ConsensusPosition
			_ConsensusStepPos
			_ConsensusBin
			%do i=1 %to &nMaps; 
				_MarkerMap&i
				_LGMap&i
				_PosMap&i
				_StepPosMap&i
				_BinMap&i
			%end;
		;
		label &LGVar= "&LGLab";
		label _MaxWindow = "&MaxWindowLab";
		label _MapName = "&MapNameLab";
		label _MapIndex = "&MapIndexLab";
		label &MarkerVar= "&MarkerLab";
		label &PositionVar= "&PositionLab";
		label StepPosition= "&StepPositionLab";
		label LG_Map_MaxWindow= "&LGMapMaxWindowLab";
		label LG_Con_MaxWindow= "&LGConMaxWindowLab";
	run;
	%PackageInsertDataset(Consensus Map Data in Long Format from &ProcessName, outlib, &BuildConsLinkMapCDL_OutDataName, no, no, no);						

	data OutLib.&BuildConsLinkMapCML_OutDataName;
		set MapLengthDataAll;
		label &LGVar= "&LGLab";
		label _MaxWindow = "&MaxWindowLab";
		label _MapIndex = "&MapIndexLab";
		label _MapName = "&MapNameLab";
		label _FirstPos= "&FirstPosLab";
		label _LastPos= "&LastPosLab";
		label _MapLength = "&MapLengthLab";
		label LG_Map_MaxWindow= "&LGMapMaxWindowLab";
		label LG_Con_MaxWindow= "&LGConMaxWindowLab";
	run;
	%PackageInsertDataset(Consensus Map Data in Long Format from &ProcessName, outlib, &BuildConsLinkMapCML_OutDataName, no, no, no);						

	data OutLib.&BuildConsLinkMapCRM_OutDataName;
		retain _ConsensusLG _MaxWindow _MapIndex _MapName _RMSE;
		set ConsensusRMSEDataAll;
		label _ConsensusLG = "&ConsensusLGLab";
		label _MaxWindow = "&MaxWindowLab";
		label _MapIndex = "&MapIndexLab";
		label _MapName = "&MapNameLab";
		label _RMSE = "&RMSELab";
	run;
	%PackageInsertDataset(RMSE Data from &ProcessName, outlib, &BuildConsLinkMapCRM_OutDataName, no, no, no);					
	
	data OutLib.&BuildConsLinkMapCST_OutDataName;
		set ConsensusRMSESummAll;
		label _ConsensusLG= "&ConsensusLGLab";
		label _MaxWindow= "&MaxWindowLab";
		label _RMSE_Mean= "&RMSELab Mean";
		label _RMSE_SD= "&RMSELab SD";
		label _ConsensusMapLength= "Consensus &MapLengthLab";
	run;
	%PackageInsertDataset(Summary Statistics Data from &ProcessName, outlib, &BuildConsLinkMapCST_OutDataName, _type=other);					
	
	%if &NConstrToDrop > 0 %then %do;
		%global BuildConsLinkMapCCD_OutData;
		%PrepOutDS(BuildConsLinkMapCCD_OutData,BuildConsLinkMapCCD_OutDataName,&OutName,_ccd);
		
		%if &FlagDebug %then %do;
			data OutLib.ConstrDroppedAll; set ConstrDroppedAll; run;
		%end;
		
		data OutLib.&BuildConsLinkMapCCD_OutDataName;
			set ConstrDroppedAll;
			label _MapIndex= "&MapIndexLab";
			label _MapName= "&MapNameLab";
			label _LinkageGroup= "&LGLab";
			label _LeftMarker= "Left Marker";
			*label _LeftPosition= "Left Position";
			label _LeftConsensusBin= "Left Consensus Bin";
			label _ConstraintSense= "Constraint Sense";
			label _RightMarker= "Right Marker";
			*label _RightPosition= "";
			label _RightConsensusBin= "Right Consensus Bin";
			drop _ConstraintID _LeftPosition _RightPosition _LeftConsensusBin _RightConsensusBin;
		run;
		%PackageInsertDataset(Deleted Constraints Data from &ProcessName, outlib, &BuildConsLinkMapCCD_OutDataName, _type=other);					
	%end;
	**end: create output SAS Data Sets**;
	
	%let ColorVar= &LGVar;
	%let ColorLab= &LGLab;
	%let lsq= %str(\[);
	%let rsq= %str(]\);
	%let LGCompVar= LGCompID;

	**begin: create JSL file**;
	%PrepOutFile(BuildConsLinkMap_Script,BuildConsLinkMap_ScriptName,&ProcessName Script for &OutName..jsl);
	%global _mainjslfile_;
	%let _mainjslfile_ = %quote(&BuildConsLinkMap_Script);
	
	data _null_;
		file "&BuildConsLinkMap_Script" lrecl=32767;
		put '//!';
		put "Names Default to Here(1);";
		put "Here:DataList = {};";
		put "Here:TabButtonList = {};";
		put "Here:TabCodeList = {};";
		put "Here:TabAfterCodeList = {};";
		put "Here:TabOpenList = {};";	
		put "wait(0.5);";
		put "	";
		/* Plot Linkage Groups*/
		/*
		put "Here:GB_results = Expr(";
		put " Here:gb = Graph Builder(";
		put "		Size(625, 525),";
		put "		Show Control Panel(0),";
		put "		Auto Stretching( 0 ),";
		put "		Variables(";
		put "			X(:Name(""&LGMapMaxWindowLab"")),";
		put "			Y(:Name(""&PositionLab"")),";
		put " 		Color(:Name(""&LGConMaxWindowLab""))";
		put "		),";
		put "		Elements(Points(X,Y,Color,Legend(4), Jitter( 0 ))),";
		put "		SendToReport(";
		put "			Dispatch({}, ""graph title"", TextEditBox,";
		put "				{Set Text(""Genetic Distances for Linkage Groups"")}";
		put "			),";
		put "			Dispatch({},""&LGMapMaxWindowLab"", ScaleBox,	{Label Row( Label Orientation( ""Angled"" ) )}),";		
		put "			Dispatch({}, ""&PositionLab"", ScaleBox, {Min( &maxpos), Max( &minpos), Rotated Labels(""Automatic"")}),";
		put "			Dispatch({}, ""Graph Builder"", FrameBox, {Marker Size(3)}),";
		put "			Dispatch( {}, ""400"", LegendBox, {Set Title( ""[Linkage Group].[Map].[Max Window]"" )} ),";
		put "			Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""[Linkage Group].[Map].[Max Window]"" )} ),";
		put "			Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Marker Position"" )} )";		
		put "		)";
		put "	);";
		put ");";
		put "Here:DF_results = Expr(";
		put "	Here:df = odat << Data Filter(";
		put "		Mode( Show( 1 ), Include( 1 ) ),";
		put "		Conditional,";
		put "		Add Filter(";
		put "			Columns(";
		put "				:Name(""&LGLab""),";
		put "				:Name(""&MaxWindowLab""),";
		put "				:Name(""&MapNameLab""),";
		put "				:Name(""&StepPositionLab""),";
		put "				:Name(""&MarkerLab""),";
		put "			),";
		put "			Where(:&MapNameLab == ""Consensus""),";
		put "		),";
		put " 	);";
		put ");";
		put "Here:GBfunc = Function({},{default local},";
		put "	If(IsEmpty(Here:odat),";
		put "		Here:odat = open(""&ClientOutPath.&BuildConsLinkMapCDL_OutDataName..sas7bdat"",invisible);";
		put "			Column(""&LGLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MapNameLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MarkerLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&LGMapMaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&LGConMaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MarkerLab"")<<Label(1);";
		put " ,"; 
		put "			Current Data Table(Here:odat);";
		put "	);";
		put "	VListBox(";
		put " 		Here:LO_HB = HListBox(";
		put "			Here:GB_results, Here:DF_results;";
		put " 		);"; 
		put "	);"; 
		put ");"; 
		put "InsertInto(DataList,{{""Here:odat""}});";
		put "InsertInto(TabButtonList, ""Linkage Map"");";
		put "InsertInto(TabCodeList, expr(Here:GBfunc()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
		put "	";
		*/
		
		put "Here:GB_results = Function({},{default local},";
		put " Here:vlb0= VListBox(";
		put " Here:olb0 = OutlineBox(""Consensus Maps"",";
		put " Here:hlb0=HListBox(";
		put " Here:gb = Graph Builder(";
		put "		Size(625, 525),";
		put "		Show Control Panel(0),";
		put "		Show Footer(0),";
		put "		Auto Stretching( 0 ),";
		put "		Variables(";
		/*put "			X(:Name(""&LGLab"")),";*/
		put "			X(:Name(""&LGMapMaxWindowLab"")),";
		put "			Y(:Name(""&PositionLab"")),";
		/*put " 		Color(:Name(""&ColorLab""))";*/
		put " 		Color(:Name(""&LGConMaxWindowLab""))";
		put "		),";
		put "		Elements(Points(X,Y,Color,Legend(4), Jitter( 0 ))),";
		put "		//Where(:&MapNameLab == ""Consensus""),";
		put "		SendToReport(";
		put "			Dispatch({}, ""graph title"", TextEditBox,";
		put "				{Set Text(""Genetic Distances for Linkage Groups"")}";
		put "			),";
		put "			Dispatch({},""&LGMapMaxWindowLab"", ScaleBox,	{Label Row( Label Orientation( ""Angled"" ) )}),";		
		put "			Dispatch({}, ""&PositionLab"", ScaleBox, {Min( &maxpos), Max( &minpos), Rotated Labels(""Automatic"")}),";
		put "			Dispatch({}, ""Graph Builder"", FrameBox, {Marker Size(3)}),";
		put "			Dispatch( {}, ""400"", LegendBox, {Set Title( ""[Linkage Group].[Map].[Max Window]"" )} ),";
		put "			Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""[Linkage Group].[Map].[Max Window]"" )} ),";
		put "			Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Marker Position"" )} ),";
		/*Using Show Footer (0) instead of Hide(1)*/
		/*put "			Dispatch( {}, """", TextBox, {Hide( 1 )} )"; */
		put "		)";
		put "		),";
		put "		),";
		put "		),";
		put "		//try(olb0[TextBox(1)]<<Delete;);";
		put "	);";
		put ");";
		put "Here:DF_results = Function({},{default local},";
		put "	Here:df = odat << Data Filter(";
		put "		Mode( Show( 1 ), Include( 1 ) ),";
		put "		Conditional,";
		put "		Add Filter(";
		put "			Columns(";
		put "				:Name(""&LGLab""),";
		put "				:Name(""&MaxWindowLab""),";
		put "				:Name(""&MapNameLab""),";
		put "				:Name(""&StepPositionLab""),";
		put "				:Name(""&MarkerLab""),";
		put "			),";
		put "			//Where(:&MapNameLab == ""Consensus""),";
		put "		),";
		put " 	);";
		put ");";
		put "Here:GBfunc = Function({},{default local},";
		put "	If(IsEmpty(Here:odat),";
		put "		Here:odat = open(""&ClientOutPath.&BuildConsLinkMapCDL_OutDataName..sas7bdat"",invisible);";
		put "			Column(""&LGLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MapNameLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MarkerLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&LGMapMaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&LGConMaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "			Column(""&MarkerLab"")<<Label(1);";
		put " ,"; 
		put "			Current Data Table(Here:odat);";
		put "	);";
		put "	//VListBox(";
		put " 	//	Here:LO_HB = HListBox(";
		put "			Here:GB_results;//, Here:DF_results;";
		put " 	//	);"; 
		put "	//);"; 
		put ");"; 
		put "Here:AppendFun = Function({},{default local},";
		put "	Here:hb= HListBox(";
		put "	Here:df = odat << Data Filter(";
		put "		Mode( Show( 1 ), Include( 1 ) ),";
		put "		Conditional,";
		put "		Add Filter(";
		put "			Columns(";
		put "				:Name(""&LGLab""),";
		put "				:Name(""&MaxWindowLab""),";
		put "				:Name(""&MapNameLab""),";
		put "				:Name(""&StepPositionLab""),";
		put "				:Name(""&MarkerLab""),";
		put "			),";
		put "			Where(:&MapNameLab == ""Consensus""),";
		put "		),";
		put " 	),";
		put " 	);";
		put " 	Here:hlb0 << Append(Here:hb);";
		put " );";		
		put "InsertInto(DataList,{{""Here:odat""}});";
		put "InsertInto(TabButtonList, ""Linkage Map"");";
		put "InsertInto(TabCodeList, expr(Here:GBfunc()));";
		put "InsertInto(TabAfterCodeList, expr(Here:AppendFun()));";
		put "InsertInto(TabOpenList,1);";
		put "	";		
		
		/* Plot Map Length*/
		put "Here:LengthPlotFunc= Function({},{default local},";
		put "	If(IsEmpty(Here:lengthdat),";
		put " 	Here:lengthdat = open(""&ClientOutPath.&BuildConsLinkMapCML_OutDataName..sas7bdat"",invisible);";
		put "		Column(""&LGLab"")<<Set Modeling Type(Nominal);";
		put "		Column(""&MaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "		Column(""&MapIndexLab"")<<Set Modeling Type(Nominal);";
		put "		Column(""&MapNameLab"")<<Set Modeling Type(Nominal);";
		put "		Column(""&LGMapMaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "		Column(""&LGConMaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "		Here:lengthdat << Color Or Mark by Column(:Name(""&MaxWindowLab""));";
		put "		,//else"; 
		put "		Current Data Table(Here:lengthdat);";
		put "	);";

		put "	HListBox(";
		put "		HListBox(";
		%do i= 1 %to &nlgout;
			put "			HListBox(";
			put "				ob&i = OutlineBox(""Linkage Group &&lgout&i"",";
			put "				Graph Builder(";
			put "					Show Control Panel( 0 ),";
			put "					Show Legend( 0 ),";
			put "					Variables( X( :&LGConMaxWindowLab ), Y( :&MapLengthLab ) ),";
			put "					Elements(";
			put "						Points( X, Y, Legend( 1 ) ),";
			put "						Points( X, Y, Legend( 4 ) ),";
			put "						Box Plot( X, Y, Legend( 5 ) )";
			put "					),";
								%if &_lgvartype1 = C %then %do;
			put "						Where( :&LGLab == ""&&lgout&i"" ),";
								%end;
								%else %do;
			put "						Where( :&LGLab == &&lgout&i ),";
								%end;
			put "						SendToReport( ";
			put "						Dispatch(";
			put "							{},";
			put "							""&MapLengthLab"",";
			put "							ScaleBox,";
			put "							{Format( ""Fixed Dec"", 9, 2 ), Inc( 5 ),";
			put "							Minor Ticks( 0 ), Label Row Nesting( 1 ),Rotated Labels( ""Automatic"" )}";
			put "						),";
			put "						Dispatch( {}, ""graph title"", TextEditBox, {Set Text( ""Map Length for Linkage Group &&lgout&i"" )} ),";
			put "						Dispatch( {}, ""&LGConMaxWindowLab"", ScaleBox, {Label Row( Label Orientation( ""Angled"" ) )}),";
			put "						Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""[Linkage Group].[Map].[Max Window]"" )} ),";
			put "						Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Map Length"" )} ),";
			put "						Dispatch( {}, ""Graph Builder"", FrameBox, {Marker Size( 2 )} )";
			put "					)";
			put "				),";
			put "				),";
			put "				try(ob&i[TextBox(1)]<<Delete;);";
			put "			),";
		%end;
		put "		),";
		put "	),";
		put ");";
		put "InsertInto(DataList,{{""Here:lengthdat""}});";
		put "InsertInto(TabButtonList, ""Map Length"");";
		put "InsertInto(TabCodeList, expr(LengthPlotFunc()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";	
		put "	";
		/* Plot RMSE */
		put "Here:RMSEPlotFunc= Function({},{default local},";
		put "	If(IsEmpty(Here:rmsedat),";
		put " 	Here:rmsedat = open(""&ClientOutPath.&BuildConsLinkMapCRM_OutDataName..sas7bdat"",invisible);";
		put "		Column(""&ConsensusLGLab"")<<Set Modeling Type(Nominal);";
		put "		Column(""&MaxWindowLab"")<<Set Modeling Type(Nominal);";
		put "		Here:rmsedat << Color Or Mark by Column(:Name(""&MaxWindowLab""));";
		put "		, // else";
		put "		Current Data Table(Here:rmsedat);";
		put "	);";
		put "  	HListBox(";
		put "		Here:rmseplot=Oneway(";
		put "			Y( :&RMSELab ),";
		put "			X( :&MaxWindowLab ),";
		put "			By( :&ConsensusLGLab ),";
		put "			With Best( 1, Difference Matrix( 1 ) ),";
		put "			Means( 1 ),";
		put "			Box Plots( 1 ),";
		put "			X Axis Proportional( 0 ),";
		put "			Mean Lines( 1 ),";
		put "			Mean Diamonds( 0 ),";
		put "			Comparison Circles( 1 ),"; 
		put "			Points Jittered( 1 ),";
		put "			Grand Mean( 0 ),";
		put "			SendToReport(";
		put "				Dispatch(";
		put "					{},";
		put "					""Oneway Plot"",";
		put "					FrameBox,";
		put "					Frame Size( 424, 296 )";
		put "				),";
		put "				Dispatch( {}, """", NomAxisBox, Rotated Tick Labels( 0 ) ),";
		put "			),";
		put "		);";
		put "	);";
		put ");";
		put "InsertInto(DataList,{{""Here:rmsedat""}});";
		put "InsertInto(TabButtonList, ""RMSE"");";
		put "InsertInto(TabCodeList, expr(RMSEPlotFunc()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";	
		put "	";
		/* Drill Downs*/
		put "//Drill Downs";
		put "	ProcessName= ""&ProcessName"";";
		put "	BuildConsLinkMap_Script= ""&BuildConsLinkMap_Script"";";				
		put "	BuildConsLinkMapCDL_OutDataName= ""&BuildConsLinkMapCDL_OutDataName"";";		
		put "	OutPath= ""&OutPath"";";
		put "	MacroPath= ""&MacroPath"";";
		put "	OutName= ""&OutName"";";
		put "	ClientOutPath = ""&ClientOutPath"";";
		put "	LGCompVar= ""&LGCompVar"";";  
		put "	LGVar= ""&LGVar"";";
		put "	LGLab= ""&LGLab"";";
		put "	MarkerVar= ""&MarkerVar"";"; 
		put "	MarkerLab= ""&MarkerLab"";";
		put "	PositionVar= ""&PositionVar"";";
		put "	PositionLab= ""&PositionLab"";";
		put "	MaxWindowVar= ""_MaxWindow"";";
		put "	MaxWindowLab= ""&MaxWindowLab"";";
		put "	MapNameVar = ""_MapName"";";
		put "	MapNameLab= ""&MapNameLab"";";
		put "	MaxWindowLab= ""&MaxWindowLab"";";
		put "	LGMapMaxWindowVar = ""LG_Map_MaxWindow"";";
		put "	LGMapMaxWindowLab= ""&LGMapMaxWindowLab"";";
		put "	MapIndexVar = ""_MapIndex"";";
		put "	LGConMaxWindowVar= ""LG_Con_MaxWindow"";"; 
		put "	FlagDebug= &FlagDebug;";
		put "	";		
		/*Create subset data for a user-chosen MaxWindow */
		put "	Here:CreateSubsetLMData = Function({subfilename=""""},{default local},";
		put "		If(IsEmpty(Here:odat),";
		put "			Here:odat = open(""&ClientOutPath.&BuildConsLinkMapCDL_OutDataName..sas7bdat"",invisible);"; 
		put "			Current Data Table(Here:odat);";
		put "		);";
		put "		mwlist = {";
		%do w=1 %to &nmaxwd;
			%if %sysevalf(&&maxwd&w <= &MaxWindFinal) %then %do;
				%if %sysevalf(&&maxwd&w < &MaxWindFinal) %then %do;
					put "		""&&maxwd&w"", ";
				%end;
				%else %do;
					put "		""&&maxwd&w""  ";
				%end;
			%end;
		%end;
		put "		};";
		put "		smw = New Window(""Create Subset Consensus Linkage Map Data"", <<ReturnResult, <<modal,";
		put "			TextBox(""Select a Maximum Interval Window Value"", <<Set Wrap(1000)),";
		put "				vlistbox(";
		put "				mwlistb=listBox(mwlist,maxselected(1)),";
		put "				buttonbox(""OK"")";
		put "				)";
		put "		);";
		put " 		if (smw[""button""] == -1, throw());";
		put "		smwlist=mwlistb<<Get Selected;";
		put "		if(Nitems(smwlist)==0,";
		put "			New Window(""JMP Message"",showmenu(0),showtoolbars(0),";
		put "				tb = TextBox(""You must select one Maximum Interval Window value.""),";
		put "				tb << Font Color(3)";
		put "			);";
		put "			stop();";
		put "		);";
		put "		whichmw = Loc(mwlist,smwlist[1]);";
		put "		show(whichmw, smwlist, mwlist);";
		put "		Here:odat << Clear Select();";
		put "		Here:odat << select where(:&MaxWindowLab==Num(smwlist[1]));";
		put "		Here:SubDat = Data Table( Here:odat ) << Subset(";
		put "			Selected Rows( 1 ),";
		put "			Selected columns only( 0 )";
		put "		);";
		put "		Here:odat << Clear Select();";
		put "		Here:subfilename= ClientOutpath||OutName||""_sub""||smwlist[1]||"".sas7bdat""; ";
		put "		Close( Here:SubDat, Save(Here:subfilename) );";
		put "		//open( Here:subfilename) ;";
		put "		New Window(""JMP Message"",showmenu(0),showtoolbars(0),";
		put "			tb1 = TextBox(""A subset consensus linkage map data for Maximum Interval Window of "" || smwlist[1] || "" was created and saved as:"", <<Set Wrap(1000)),";
		put "			tb1 << Font Color(""Blue""),";
		put "			tb2 = TextBox("" ""), ";
		put "			teb = TextEditBox(Here:subfilename, <<Set Wrap(1000)),";
		put "			teb << Font Color(""Blue""), ";
		put "			tb3 = TextBox("" ""),";
		put "			ButtonBox(""Click here to open the data set"",";
		put "				open(Here:subfilename);";
		put "			);";
		put "		); ";
		put "	);";	
		put "Here:ActionButtonDesc = {};";
		put "Here:ActionButtonList = {};";
		put "Here:ActionCodeList = {};";
		put "InsertInto(Here:ActionButtonDesc ,{""Click the button below to create a subset consensus linkage map data set""});";	
		put "InsertInto(Here:ActionButtonList,{{""Create Subset Linkage Map Data""}});";
		put "InsertInto(Here:ActionCodeList,{{Here:CreateSubsetLMData}});";
		/* Follow-Up Processes*/	
		put "//Follow-up processes";
		put "if(host is(Windows),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, windows) ),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, POSIX) ));";
		put "	";
		put "ColExists = Function({colname},";
		put "   cnames = CurrentDataTable() << GetColumnNames(String);";
		put "   // show(cnames,colname);";
		put "   if(contains(cnames,colname),";
		put "      rc = 1,";
		put "      rc = 0;";
		put "      NewWindow(""Column Not Found"",<<modal,";
		put "         TextBox(""The column ""||colname||"" does not exist in the currently active JMP table.  ""||";
		put "            ""Try selecting rows on a different tab with results that use this column."" "; 
		put "         )";
		put "      );";
		put "   );";
		put "   rc";
		put ");";
		put "	";
		put "fupbuttonswindow=Function({},{default local},";
		put "		VListBox(";
		/* Linkage Map Viewer */
		put "			TextBox(""Click the button below to launch the Linkage Map Viewer process to draw the full linkage map.""),";
		put "			ButtonBox(""Linkage Map Viewer"",";		
		put "				Here:CurrDat = Current Data Table();";
		put "				//show(Here:CurrDat);";
		put "				if(ColExists(""&PositionLab"") & ColExists(""&LGLab"") & ColExists(""&MaxWindowLab"") & ColExists(""&MapNameLab"") & ColExists(""&LGMapMaxWindowLab""),";
		put "					nsel = 0;";
		put "					For Each Row( ";
		put "						If (Selected(), ";
		put "							nsel++;";
		put "						);";
		put "					);";
		put "					if (nsel==0,";
		put "						New Window(""JMP Message"",showmenu(0),showtoolbars(0),";
		put "							tb = TextBox(""You must first select points in the graphs or rows in the tables.""),";
		put "							tb << Font Color(3)";
		put "						);";
		put "						stop();";
		put "					);";
		put "					Here:SubDat = Data Table( Here:CurrDat ) << Subset(";
		put "						Selected Rows( 1 ),";
		put "						Selected columns only( 0 )";
		put "					);";
		put "					Here:NoDupDat = Data Table( Here:SubDat ) << Join(";
		put "						With( Data Table( Here:SubDat ) ),";
		put "						By Matching Columns(";
		put "							:&LGLab = :&LGLab,";
		put "							:&MaxWindowLab = :&MaxWindowLab,";
		put "							:&MapNameLab = :&MapNameLab";
		put "						),";
		put "						Drop multiples( 1, 1 ),";
		put "						Include Nonmatches( 0, 0 ),";
		put "						Preserve main table order( 0 )";
		put "					);";
		put "					Close( Here:SubDat, NoSave );";
		put "					nviewlgs= N Row(Here:NoDupDat);";
		put "					Here:ViewLG= "" "";";
		put "					for(i=1, i<=nviewlgs,i++,";
		put "						if(i <= nviewlgs - 1, ";
                put "                                        	   Here:ViewLG= Here:ViewLG || Eval Insert(  ";
                put "                                        	   ""\!""^:&LGMapMaxWindowLab[i]^\!"""" ";
                put "                                        	   ) || "","" ";
		put "						  ,";
                put "                                        	   Here:ViewLG= Here:ViewLG || Eval Insert(  ";
                put "                                        	   ""\!""^:&LGMapMaxWindowLab[i]^\!"""" ";
                put "                                        	   ) ";
		/*
		put '                                           	Here:ViewLG= Here:ViewLG || Eval Insert( "\!"^:&LGMapMaxWindowLab[i]^\!"" ) || ","';		
		put "						,";
		put '                                           	Here:ViewLG= Here:ViewLG || Eval Insert( "\!"^:&LGMapMaxWindowLab[i]^\!"" )';
		*/
		put "						);";
		put "					);";
		put "					Here:ViewLG= ""LG_Map_MaxWindow in ("" || Here:ViewLG || "")""; ";
		put "					Close( Here:NoDupDat, NoSave );";
		put "	";
		put "					if(host is(Windows),";
		put "						Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/LinkageMapViewer.jsl"", absolute, windows)),";
		put "						Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/LinkageMapViewer.jsl"", absolute, POSIX) ));";
		put "	";
		/* Set input data set and Output Folder */
		put "					jg:setDefaultParmValues(2, ""LinkageMapViewer"");";				
		put "					jg:LinkageMapViewer_updateDeps;";
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""InData""),""&ClientOutPath.&BuildConsLinkMapCDL_OutDataName..sas7bdat"");";		
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""WhereStmt""),Here:ViewLG);";	
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""OutPath""),""&ClientOutPath"");";										
		put "					jg:LinkageMapViewer_updateDeps;";				
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""ChrVar""),""LG_Map_MaxWindow"");";	
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""Label""),""&MarkerVar"");";
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""PosVar""),""&PositionVar"");";
		/*
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""ColorVar""),""&ColorVar"");";
		put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""ColorType""),""Continuous"");";
		*/
		put "					jg:LinkageMapViewer_updateDeps;";
		/*put "					jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""Theme""),""Blue to Gray to Red"");"; */
		put "					Wait(.5);";
		/* put "				jg:runProcess( 2, ""LinkageMapViewer"" );"; */
		put "					,";
		put "					stop();";
		put "				),";	
		put "			),";
		put "	"; 
		/* Compare Linkage Maps */
		put "			Here:CountComp=0;";
		put "			TextBox("" ""),";
		put "			TextBox(""Click the button below to launch the Compare Linkage Maps process.""),";
		put "			ButtonBox(""Compare Linkage Maps"",";
		put "	";		
		put "					if(host is(Windows),";
		put "   					Include( ConvertFilePath(""$GENOMICS_HOME/JSLFiles/BuildConsensusLinkageMap_DrillDowns.jsl"", absolute, windows));";
		put "						,";
		put "   					Include( ConvertFilePath(""$GENOMICS_HOME/JSLFiles/BuildConsensusLinkageMap_DrillDowns.jsl"", absolute, POSIX));";	
		put "					);";
		put "	";
		put "					Here:CompMaps;";
		put "	";
		put "			),";
		put "		),";
		put ");";
	
		put "Here:LaunchCode = fupbuttonswindow();";
	run;
	**end: create JSL file**;
	
	%*PackageInsertFile(JSL Code from &ProcessName, &OutPath, &BuildConsLinkMap_ScriptName, "&BuildConsLinkMapCDW_OutDataName..sas7bdat", yes);
	%PackageInsertFile(JSL Code from &ProcessName, &OutPath, &BuildConsLinkMap_ScriptName, , text/plain, yes);
	
	%exit:;
	%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);
	
	ods html close;
	ods listing;

%mend;
%BuildConsensusLinkageMap;
	