/*
JMP Genomics Analytical Process 

Name: DataCorrelation
Authors: John Bander and Russ Wolfinger, SAS Institute Inc.
Support: Tzu-Ming Chu, SAS Institute Inc.
History:
   sasjab 01Aug03 initial coding
   sasrdw 10Mar04 refinement
   sasrdw 21May05 added PCA

Description: 
This process computes correlations between 
numeric variables.  The default type is
standard Pearson product-moment correlations, 
but you may optionally specify Hoeffding D, 
Kendall tau-b, or Spearman rank-order correlations.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData Variables WithVars ListVars ListWithVars  WhereStmt OutType CovMat ExpInData 
ColorVars ExpLabelVar FactorVars ClassVars FixedEffects RandomEffects AccProportion NFactors
VCFirstPC VCOutFile VCTitle1 WAPRanking VCListAll ScaleRows CenterRows
DataCorrelation_OutData DataCorrelation_OutDataTest DataCorrelation_OutDataEig DataCorrelation_OutDataVC
DataCorrelation_OutDataPVC OutPath ServerOutPath MacroPath DataCorrelationScript

AnnoInData   
BaseMergeVars AnnoMergeVars AnnoWhereStmt
;
* removed VCLastPC from input parameters (TC 05/20/11);

*ProcessBody;

%macro DataCorrelation;

* initialization and argument checking;
%let ProcessName = DataCorrelation;

* Path for SAS macros.;
%include "&MacroPath./UtilityMacros.sas" / nosource;

%GetOrigFileName(InData,OrigInData);

%if %length(&ExpInData) %then %let ed = 1;
%else %let ed = 0;
%CheckInData(_ExpFlag=&ed);
%if &exiterror %then %goto exit;

* re-asign variable info;
%if %length(&Variables) | %length(&ListVars) %then %do;
	%CheckInData2(InLib.&InData,_Vars=&Variables,_VarListStyle=&ListVars);
%end;

%if ^&ed & ^%length(&Variables) & ^%length(&ListVars) %then %do;
	%put ERROR: Variables, List-Style Variables, or Experimental Design Data Set must be specified.;
	%let exiterror = 1;
	%goto exit;
%end;
%if &ed %then %do;
	%NObsVars(OutLib.&ExpInData);
	%if &nObs=1 %then %do;
		%put ERROR:There is only one observation in your experimental design file.  You need to have more than one sample to run this process.;
		%let exiterror = 1;
		%goto exit;
	%end;
%end;

%if (&AccProportion=0) & (&nFactors=0) %then %do;
	%let exiterror = 1;
	%put ERROR: Both parameters Cumulative Proportion of Variation to Explain and Maximum Number of Principal Components to Apply were set to 0 and disabled.;
	%put ERROR: At least one of them shall be actived to perform PCA.;
	%goto exit;
%end;

%let VCLastPC = &nFactors; * fixed the last PC to the maximum nuber of PCs;

libname OutLib "&OutPath"; 

%if %upcase(&CovMat)=YES and %upcase(&OutType)=PEARSON %then %do;
	%let ProcFactorOptions = COV;
	%let ctype = COV;
	%let csuff = _cov;
	%let clabel = Covariance;
%end;
%else %do;
	%let ProcFactorOptions = ;
	%let ctype = CORR;
	%let csuff = _cor;
	%let clabel = Correlation;
%end;

%if %length(&AnnoInData) %then %do;
	* treat the "FilterInData" as a special AnnoIndata;
	%CheckAnno(_CheckAccession=0,_CheckLabel=0);
	%if &exiterror %then %goto exit;
	
	%if %length(&AnnoWhereStmt) %then %do;
		data Anno;
			set AnnoLib.&AnnoData;
			where &AnnoWhereStmt;
		run;
		%let FilterInData = Anno;
	%end;
	%else %do;
		%let FilterInData = AnnoLib.&AnnoData;
	%end;
	%Merge2sets(InLib.&InData,&FilterInData,indat,&BaseMergeVars,&AnnoMergeVars,
		_SubSet2=1,_KeepVars2=&AnnoMergeVars);
	%let indat = indat;
%end;
%else %do;
	%let indat = InLib.&InData;
%end;

%if %length(&WhereStmt) %then %do;
	data indat;
		set &indat;
		where &WhereStmt;
	run;
	%let indat = indat;
%end;

* center and/or scale rows;
data ind;
	set &indat;
	array vr{*} %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
	call symput('nVariables',trim(left(dim(vr))));
		
	%if &CenterRows=Yes or &ScaleRows=Yes %then %do;
		%if &CenterRows=Yes %then %do;
			_rowmean_ = mean(of %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;);
		%end;
		%else %do;
			_rowmean_ = 0;
		%end;

		%if &ScaleRows=Yes %then %do;
			_rowstd_ = std(of %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;);
		%end;
		%else %do;
			_rowstd_ = 1;
		%end;
		
		do _i_=1 to dim(vr);
			if _rowstd_ > 0 then vr[_i_] = (vr[_i_] - _rowmean_)/_rowstd_;
			else vr[_i_] = .;
		end;
		drop _i_;
		* center and scale with variables separately;
		%if %length(&WithVars) or %length(&ListWithVars) %then %do;
			array wvr{*} &WithVars &ListWithVars;

			%if &CenterRows=Yes %then %do;
				_wrowmean_ = mean(of &WithVars &ListWithVars);
			%end;
			%else %do;
				_wrowmean_ = 0;
			%end;

			%if &ScaleRows=Yes %then %do;
				_wrowstd_ = std(of &WithVars &ListWithVars);
			%end;
			%else %do;
				_wrowstd_ = 1;
			%end;

			do _i_=1 to dim(wvr);
				if _wrowstd_ > 0 then wvr[_i_] = (wvr[_i_] - _wrowmean_)/_wrowstd_;
				else wvr[_i_] = .;
			end;
			drop _i_;
		%end;
	%end;
run;		
%let indat = ind;

%if &nVariables=1 %then %do;
	%put ERROR: You need to have more than 1 variable selected to run this process.;
	%let exiterror = 1;
	%goto exit;
%end;

* check name for OutData;
%if %length(&OrigInData) < 29 %then %let OID = &OrigInData;
%else %let OID = %substr(&OrigInData,1,28);
%PrepOutDS(DataCorrelation_OutData,DataCorrelation_OutDataName,&OID,&csuff);
%PrepOutDS(DataCorrelation_OutDataTest,DataCorrelation_OutDataTestName,&OID,_tst);

%if ^%length(&WithVars) and ^%length(&ListWithVars) %then %do;
	%PrepOutDS(DataCorrelation_OutDataEig,DataCorrelation_OutDataEigName,&OID,_eig);
	%if %length(&RandomEffects) %then %do;
		%PrepOutDS(DataCorrelation_OutDataVC,DataCorrelation_OutDataVCName,&OID,_vc);
		%PrepOutDS(DataCorrelation_OutDataPVC,DataCorrelation_OutDataPVCName,&OID,_pvc);
	%end;
	%if (&DataCorrelation_OutData = &DataCorrelation_OutDataEig) or
		(&DataCorrelation_OutData = &DataCorrelation_OutDataVC) or
		(&DataCorrelation_OutData = &DataCorrelation_OutDataPVC) or
		(&DataCorrelation_OutDataEig = &DataCorrelation_OutDataVC) %then %do;
		%put ERROR:  Output data set names must not be the same.;
		%let exiterror = 1;
		%goto exit;
	%end;
%end;

%let fl=%substr(&OutType,1,1);

* compute correlations;
ods exclude all;
ods noresults;
proc corr data=&indat out&fl=thecorrs &OutType 
	%if %upcase(&CovMat)=YES and %upcase(&OutType)=PEARSON %then %do; COV  %end; ;
  	var %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
  	%if %length(&WithVars) or %length(&ListWithVars) %then %do;
      with &WithVars &ListWithVars;
  	%end;
  	ods output &OutType.Corr=morecorrs;
run;
ods exclude none;
ods results;

%if &syserr>4 %then %do;
	%put Check the specification of Variables, With Variables, and List-Style variables.;
	%let exiterror = 1;
	%goto exit;
%end;

/*
title "corrs";
proc print data=thecorrs(obs=10);
run;

title "more corrs";
proc print data=morecorrs(obs=10);
run;
title;
*/

* compute principal components using output matrix from proc corr;
%if ^%length(&WithVars) and ^%length(&ListWithVars) %then %do;
	/*
	%if %length(&RandomEffects) and (&AccProportion=0) and (&VCLastPC > &NFactors) %then %do;
		%let NFactors = &VCLastPC;
	%end;
	*/
	/*
	proc factor data=thecorrs outstat=f1 nfactors=&NFactors &ProcFactorOptions
		scree;
		var &Variables;
	run;
	%let prefix = Factor;
	*/

	proc princomp data=thecorrs outstat=f1 %if &AccProportion=0 %then n=&NFactors; &ProcFactorOptions;
		var %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
		ods output eigenvalues=&DataCorrelation_OutDataEig;
	run;
	
	* 3 cases here:
	%put AccProportion= &AccProportion nFactors=&nFactors;
	1. &AccProportion>0 & &nFactors=0
	2. &AccProportion>0 & &nFactors>0
	3. &AccProportion=0 & &nFactors>0;
	
	data _null_;
		set &DataCorrelation_OutDataEig end=_e_; 
		%if &AccProportion>0 %then %do;
			* cases 1 & 2;
			if round(Cumulative,0.0001) >= &AccProportion then do;
				call symput('NFactors2',trim(left(_n_)));
				call symput('Cumulative',Cumulative);
				stop;
			end; 
		%end;
		%else %do;
			* case 3;
			if _n_ = &NFactors then call symput('Cumulative',Cumulative);
			if _e_ then call symput('MaxFactors',_n_);
		%end;
	run; 

	%if &AccProportion>0 %then %do;
		* further checking for cases 1 and 2;
		%if (&nFactors>0) %then %do;
			* case 2;
			%if &nFactors < &nFactors2 %then %do;
				* applying &nFactors as PCA selection cutoff, so reset &Cumulative;
				data _null_;
					set &DataCorrelation_OutDataEig; 
					if _n_ = &nFactors then call symput('Cumulative',Cumulative);
				run;
			%end;
			%else %do;
				* applying &AccProportion as PCA selectioncutoff, so reset &nFactors;
				%let nFactors = &nfactors2;
			%end;
		%end;
		%else %do;
			* case 1;
			* reset &nFactors;
			%let nFactors = &nfactors2;
		%end;
	%end;
	%else %do;
		* further checking for case 3;
		%if &MaxFactors < &nFactors %then %do;
			%let nFactors = &MaxFactors;
			%let Cumulative = 1;
			%put WARNING: The specified Maximum Number of Principal Components to Apply excesses the total number of prinicipal components calculated from the data.;
			%put WARNING: The Maximum Number of Principal Components to Apply was reset to %trim(%left(&nFactors)).;
		%end;
	%end;
	%let VCLastPC = &nFactors; * fixed &VCLastPC;

	%put Number Of Components = &NFactors;
	%put Cumulative Proportion of Variation = &Cumulative;
	
	%let prefix = Prin;
	
	data _null_;
		set &DataCorrelation_OutDataEig;
		if proportion >= 0.0995 then do;
			call symput("prinlab"||trim(left(_n_)),"&prefix" || trim(left(_n_)) || ' (' || put(100*proportion,4.1) || '%)');
		end;
		else do;
			call symput("prinlab"||trim(left(_n_)),"&prefix" || trim(left(_n_)) || ' (' || put(100*proportion,3.1) || '%)');		
		end;
		%if &AccProportion>0 %then %do;
			if _n_ > &NFactors then stop;
		%end;
	run;
	%put ctype=&ctype;
	
	%if &AccProportion>0 %then %do;
		data f1;
			set f1;
			if _TYPE_="SCORE" & (substr(_NAME_,5)+0) > &NFactors then stop;
		run;
	%end;
	
	/*
	proc princomp data=InLib.&InData outstat=f1 n=&NFactors &ProcFactorOptions;
		var &Variables &ListVars;
	run;
	%let prefix = Prin;
	*/
	
	/*
	proc print data=f1;
	run;
	*/
	
	data f1;
		set f1;
		if _type_ ne "&ctype";
		if length(_name_) then idvar = _name_;
		else idvar = _type_; 
	run;
	
	proc transpose data=f1 out=f2;
		var %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
		id idvar;
	run;
	
	data f2;
		set f2;
		keep _name_ &prefix.1-&prefix.&NFactors;
		label 
			%do i=1 %to &NFactors;
				&prefix.&i = "&&prinlab&i" 
			%end;
		;	
	run;
	
	* set up _sindex_ to save variable order for the heat map;
	data thecorrs;
		set thecorrs;
		_sindex_ = _n_;
	run;
	
	proc sort data=thecorrs;
		by _name_;
	run;
	
	proc sort data=f2;
		by _name_;
	run;
	
	data thecorrs;
		merge thecorrs f2;
		by _name_;
	run;
%end;

* main output data set in rectangular format;
data &DataCorrelation_OutData;
	set thecorrs;
	if _type_="&ctype";
	label _name_ = '_NAME_';
run;

* stacked output data set;
* proc corr stores p values under variable names with a P prefix;
data tp;
	set morecorrs;
	drop %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
run;

data tp;
	set tp(obs=1);
	keep P:;
run;

%let _tid_  = %sysfunc(open(tp));
%let nv = %sysfunc(attrn(&_tid_, NVARS));
%put nv = &nv;


data &DataCorrelation_OutDataTest;
	set morecorrs;
	length With $ 32;
	With = Variable;
	array corrs{*} %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
	array pvals{*} %do i=1 %to &nv; %sysfunc(varname(&_tid_,&i)) %end; ;
	%do i=1 %to &nv;
		%let vn = %sysfunc(varname(&_tid_,&i));
		%let len = %length(&vn);
		Variable = "%substr(&vn,2,%eval(&len-1))";
		Corr = corrs[&i];
		Raw_p = pvals[&i];
		if (Raw_p <= 10**(-312)) then NegLog10_p = 312; /* 1E-312 didn't work on client server. so change it to 10**(-312);*/
		else NegLog10_p = -log10(Raw_p);
		%if %length(&WithVars) or %length(&ListWithVars) %then %do;
			if not missing(Raw_p) then output;
		%end;
		%else %do;
			if &i < _n_ and not missing(Raw_p) then output;
		%end;
	%end;
	keep Variable With Corr Raw_p NegLog10_p;
run;

%let _tid_ 	 = %sysfunc(close(&_tid_));

%NObsVars(&DataCorrelation_OutDataTest);
%if ^&nobs %then %do;
	%put ERROR: No observations in the correlation test data set.;
	%put Check the specification of Variables, With Variables, and List-Style variables.;
	%let exiterror = 1;
	%goto exit;
%end;

proc sort data=&DataCorrelation_OutDataTest;
	by descending Corr;
run;


* merge experimental design metadata;
%if %length(&ExpInData) %then %do;

	%if ^%length(&ColorVars) %then %let ColorVars = ColumnName;
	%if %length(&ExpLabelVar) %then %let labvar = &ExpLabelVar;
	%else %let labvar = ColumnName;
	
	%let i = 1;
	%do %while(%length(%scan(&ColorVars,&i)));
	   %let ColorVar&i = %scan(&ColorVars,&i);
		%let i = %eval(&i+1);
	%end;
	%let nColorVar = %eval(&i-1);
	
	data &DataCorrelation_OutData;
		set &DataCorrelation_OutData;
		ColumnName = _name_;
		ColumnNameUpcase = upcase(ColumnName);
		drop _name_;
	run;

	proc sort data=&DataCorrelation_OutData;
		by ColumnNameUpcase;
	run;
	
	data ed;
		set ed;
		ColumnNameUpcase = upcase(ColumnName);
	run;

	proc sort data=ed;
		by ColumnNameUpcase;
	run;
	
	%CheckLengthForMerging(ed,&DataCorrelation_OutData,ColumnNameUpcase);
		
	data &DataCorrelation_OutData;
		merge ed &DataCorrelation_OutData;
		by ColumnNameUpcase;
		if _type_="&ctype";
		%if &nColorVar > 1 %then %do;
			ColorIndex = trim(left(&ColorVar1))
				%do i = 2 %to &nColorVar;
					   ||"_"|| trim(left(&&ColorVar&i)) 
				%end; 
			;
		%end;
	run;
	
	%if %length(&ExpLabelVar) %then %let labvar = &ExpLabelVar;
	%else %let labvar = ColumnName;
	%CheckLabel(&DataCorrelation_OutData,labvar);
	
	%if &nColorVar = 1 %then %CheckLabel(&DataCorrelation_OutData,ColorVars);
%end;
%else %let labvar = _NAME_;

%if ^%length(&WithVars) and ^%length(&ListWithVars) %then %do;
	proc sort data=&DataCorrelation_OutData;
		by _sindex_;
	run;

	data &DataCorrelation_OutData;
		set &DataCorrelation_OutData;
		drop _sindex_;
	run;

	%if %length(&RandomEffects) %then %do;
/*
		%if &AccProportion>0 %then %do;
			%let pcfirst = 1;
			%let pclast = &VCLastPC;
		%end;
		%else %do;
*/
			%let pcfirst = &VCFirstPC;
			%let pclast = &VCLastPC;
			%if (&pclast > &NFactors) %then %let pclast = &NFactors;
			%if (&pcfirst < 1) %then %let pcfirst = 1;
			%if (&pcfirst > &pclast) %then %let pcfirst = &pclast;
		%*end;
		%put pcfirst = &pcfirst;
		%put pclast = &pclast;
		%put NFactors = &NFactors;
		
		%let Cumulative1 = 0;
		%if &pcfirst > 1 %then %do;
			data _null_;
				set &DataCorrelation_OutDataEig; 
				if _n_ = %eval(&pcfirst-1) then do;
					call symput('Cumulative1',Cumulative);
					stop;
				end;
			run; 
		%end;
		data _null_;
			cum = &Cumulative-&Cumulative1;
			call symput('cum',put(cum,5.2));
		run;

		%if %length(&VCOutFile) %then %do;
			%if %length(&VCOutFile) > 4 %then %do;
				%if %upcase(%substr(&VCOutFile,%length(&VCOutFile)-3)) ne .RTF %then %let VCOutFile = &VCOutFile..rtf;
			%end;
			%else %do;
				%let VCOutFile = &VCOutFile..rtf;
			%end;
		%end;	

		%if %length(&VCTitle1) %then %do;
			title1 "&VCTitle1";
		%end;

		%include "&MacroPath./PCA_VCA.sas" / nosource;

		%PCA_VCA(
			pdataset = &DataCorrelation_OutData,
			edataset = &DataCorrelation_OutDataEig,
			class_vars = &ClassVars,
			fixed_effects = &FixedEffects,
			random_effects = &RandomEffects,
			pcfirst = &pcfirst,
			pclast = &pclast,
			listall = &VCListAll,
			vcdataset = &DataCorrelation_OutDataVC,
			pvcdataset = &DataCorrelation_OutDataPVC,
			rtffile = &VCOutFile,
			matrixlabel = &clabel
		);
		run;

		* set label for PVC dataset;
		data &DataCorrelation_OutDataPVC;
			set &DataCorrelation_OutDataPVC;
			label %do i = &pcfirst %to &pclast;
				Prin&i = "&&prinlab&i"
			%end;;
		run;
		* get variance component list;
		%if %upcase(&WAPRanking) = YES %then %do;		
			proc sort data=&DataCorrelation_OutDataVC;
				by Variance_Component;
			run;
			
			proc sort data=&DataCorrelation_OutDataPVC;
				by VarianceComponent;
			run;
			
			data &DataCorrelation_OutDataPVC;
				merge &DataCorrelation_OutDataPVC &DataCorrelation_OutDataVC;
				drop Variance_Component;
			run;
			
			proc sort data=&DataCorrelation_OutDataPVC;
				by descending Wt_Avg_Proportion;
			run;			
		%end;
		
		data _null_;
			set &DataCorrelation_OutDataPVC end=_e_;
			retain _c_ 0;
			if upcase(VarianceComponent) ne "RESIDUAL" then do;
				_c_+1;
				name = "_vc_" || trim(left(_c_));
				call symput(name,VarianceComponent);
			end;
			if _e_ then call symput("_nVC_",_n_);
		run;
		%let _vcList_ = ""%trim(&_vc_1)"";
		%do i = 2 %to %eval(&_nVC_-1);
			%let _vcList_ = &_vcList_ , ""%trim(&&_vc_&i)"";
		%end;
		%let _vcList_ = &_vcList_ , ""Residual"";
	%end;
%end;


* create macro variables for JSL;
data a;
   set InLib.&InData(obs=1);
   if _n_>1 then stop;
run;

proc transpose data=a out=at;
   var %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
run;
%if &syserr %then %goto exit;

data _null_;
   set at end=_last;
   length mnm$ 10;
   mnm = "var" || trim(left(_n_));
   call symput(mnm,trim(left(_name_)));
   if (_last) then call symput("nv",trim(left(_n_)));
run;

* convert names to labels for JMP;
%do i = 1 %to %eval(&nv);
   %let mvar = &&var&i;
 	%CheckLabel(&DataCorrelation_OutData,mvar);
	%let var&i = &mvar;  
%end;

* check number of observations for &DataCorrelation_OutDataPVC	to correct inproper x-axis ticks in 
pvca plot;
%nObsVars(&DataCorrelation_OutDataPVC);
%let nx_pvca = &nObs;

%PrepOutFile(DataCorrelationScript,DataCorrelationScriptName,&ProcessName Script for &InData..jsl);


%global _mainjslfile_ FullAPName;
%let _mainjslfile_ = %quote(&DataCorrelationScript);
%let FullAPName = %str(Correlation and Principal Variance Component Analysis);

* create JSL code;
data _null_;
   file "&DataCorrelationScript" lrecl=32767;
	put '//!';
	put "Names Default to Here(1);";
	put "Here:DataList = {};";
	put "Here:TabButtonList = {};";
	put "Here:TabCodeList = {};";
	put "Here:TabAfterCodeList = {};";
	put "Here:TabOpenList = {};";
	
   %if ^%length(&WithVars) and ^%length(&ListWithVars) %then %do;
		put "Here:scree = Function({},{default local}," ;
		put " If(IsEmpty(Here:eigdata),";
		put "   Here:eigdata=open(""&ClientOutpath.&DataCorrelation_OutDataEigName..sas7bdat"", invisible);";
		put "   Column(""Number"") << label(1);";
		put "   Column(""Cumulative"") << label(1);";
		put " ,";
		put "   Current Data Table( Here:eigdata );";
		put " );";
		put "   vListBox( ";
		
		put "Graph Builder(	Size( 450, 400 ), Show Control Panel( 0 ), Show Footer( 0 ), Show Title(0),";
		put "  Variables( X( :Number ), Y( :Proportion ), Y( :Cumulative, Position( 1 ) ) ),";
		put "  Elements(Line( X, Y( 1 ), Y( 2 ), Legend( 7 ) ),Points( X, Y( 1 ), Y( 2 ), Legend( 8 ) )),";
		put '  SendToReport(Dispatch( {}, "Graph Builder", OutlineBox, {Set Title( "Scree Plot" )} ),';
		put '  Dispatch( {}, "X title", TextEditBox, {Set Text( "Component Number" )}),';
		put '  Dispatch({},"Y title",TextEditBox,{Set Text( "Protion of Explained Variance" )})';
		put "  )";
		put ");";
		/*put "      Overlay Plot(X( :Number), Y( :Proportion, :Cumulative), Connect Thru Missing(1)),";*/
		put "   )";
		put ");";
		put " ";
		%if %length(&RandomEffects) %then %do;
		/* v4.0
			put "vcdata=open("".\&DataCorrelation_OutDataVCName..sas7bdat"");";
			put "vcdata<<Minimize Window;";
			put 'VC = New Window("Variance Components Chart",';
			put "   Chart(";
			put "      X(:Variance Component),";
			put "      Y(:Weighted Average Proportion),";
			put "      Label by Percent of Total Values,";
			put "      Show Labels,";
			put "      Bar Chart(1),";
			put "      Y[1] << Label Format( 5, 1 )";
			put "   );";		
			put '   << Set Window Icon("LS_HOME/Documentation/Icons/DistributionAnalysis.bmp");';
			put ");";
		*/
			* add one more plot in v4.1;
			put "Here:VC1= Function({},{default local}," ;
			put " If(IsEmpty(Here:vcdata),";
			put "	  Here:vcdata=open(""&ClientOutpath.&DataCorrelation_OutDataVCName..sas7bdat"",invisible));";
			put "   col1 = Column(1);";
			put "   col1 << Set Property(""Value Ordering"",{&_vcList_});";
			put " If(IsEmpty(Here:pvcdata),";
			put "   Here:pvcdata=open(""&ClientOutpath.&DataCorrelation_OutDataPVCName..sas7bdat"",invisible));";
			put "   col2 = Column(1);";
			put "   col2 << Set Property(""Value Ordering"",{&_vcList_});";
			put "   V List Box(";
			%if &pcfirst < &pclast %then %do;
				put "      TextBox(""Principal Components Applied: PC&pcfirst to PC&pclast""),";
			%end; 
			%else %do;
				put "      TextBox(""Principal Components Applied: PC&pcfirst""),";
			%end;
			put "      TextBox(""The Cumulative Proportion of Variation: &Cum""),";
			put '      TextBox(" "),';
			put " 	  Current Data Table( Here:vcdata );";
			/*Start of Chart*/
			/*put "      Chart(";
			put "         X( :Variance Component ),";
			put "         Y( :Weighted Average Proportion ),";
			put "         Label by Percent of Total Values,";
			put "         Show Labels,";
			put "         Bar Chart( 1 ),";
			put '         Y[1] << Label Format( "Percent", 5, 1 ),';
			put "         SendToReport(";
			put "            Dispatch(";
			put "               {},";
			put '               "Chart",';
			put "               OutlineBox,";
			put "               Set Title(";
			put '                  "Weighted Average Proportion across Principal Components"';
			put "               )";
			put "            ),";
			put '            Dispatch( {}, "Chart Graph", FrameBox, {Frame Size( 300, 200 )} )';
			put "         )";
			put "      ),";*//*End of Chart*/
			
			put "Graph Builder(";
			put "  Size( 450, 320 ),Show Control Panel( 0 ),Show Legend( 0 ),Show Title( 0 ),";
			put "  Variables( X( :Variance Component ), Y( :Weighted Average Proportion ) ),";
			put '  Elements(Bar( X, Y, Legend( 1 ), Label( "Label by Percent of Total Values" ) )),';
			put "  SendToReport(";
			put '    Dispatch({},"Graph Builder",OutlineBox,';
			put '    {Set Title( "Weighted Average Proportion across Principal Components" ),';
			put "    Image Export Display( Normal )}))";
			put "),";/*End of Graph Builder*/
			
			put 'textbox(" "),';
			put "      //Current Data Table( Here:pvcdata );";
			put 'Here:stckdt=Data Table( Here:pvcdata )<<stack(invisible, output table("PCProportion"),';
			put '	columns(:name("' "&&prinlab&pcfirst" '"),';
			put %do i = %eval(&pcfirst+1) %to %eval(&pclast-1); ':name("' "&&prinlab&i" '"),' %end; ':name("' "&&prinlab&i" '")),';
			put 'Source Label Column( "Components" ),';
			put 'Stacked Data Column( "Variance Proportion" ));'; /*End of stack*/
			put "Current Data Table( Here:stckdt);";
			put "Graph Builder(";
			put "      Size( 550, 400 ),Show Control Panel( 0 ),Show Title( 0 ),Show Footer( 0 ),";
			put "      Variables(X( :Variance Component ),";
			put "      Y( :Variance Proportion ),";
			put "      //Group Y( :PCs, Show Title(0) ),";
			put "      Overlay(:Components)),";
			put "Elements( Points( X, Y ), Line( X, Y ) ),";
			put "Local Data Filter(";
			put "	Add Filter(";
			put "   columns(:Components))),";
			put '		SendToReport(Dispatch({},"Graph Builder",OutlineBox,';
			put '{Set Title( "Variance Proportion by Principal Component" )}))';
			put ")));";
			
			**put "      Overlay Plot(";
			
			/*put "      Graph Builder(";
			put "      Size( 450, 320 ),Show Control Panel( 0 ),Show Title( 0 ),Show Footer( 0 ),";
			put "      Variables(X( :Variance Component ),";
			put '      Y(:name("' "&&prinlab&pcfirst" '")),'; 
			put        %do i = %eval(&pcfirst+1) %to %eval(&pclast-1); 'Y(:name("' "&&prinlab&i" '"),position(1)),' %end; 'Y(:name("' "&&prinlab&i" '"),position(1))),';
			put '   Elements(';
			put "Points(X, ";
			put %do i=&pcfirst %to %eval(&pclast-1); 'Y('"&i"'),' %end; 'Y('"&i"'),Legend(13)),';
			put "Line(X,";
			put %do i=&pcfirst %to %eval(&pclast-1); 'Y('"&i"'),' %end; 'Y('"&i"'),Legend(14))';
			put '),';
			put "  SendToReport(";
			put '  	Dispatch({},"Graph Builder",OutlineBox,{Set Title( "Variance Proportion by Principal Component" ), Image Export Display(normal)}';
			put '),';
			put '  Dispatch( {}, "Y title", TextEditBox, {Set Text( "Variance Proportion" )} )';
			put '  ))));';  */ /*End of  graph builder*/
			
			/*
			put "         X( :Variance Component ),";
			put "         Y( " %do i = &pcfirst %to %eval(&pclast-1); 'name("' "&&prinlab&i" '"),' %end; 'name("' "&&prinlab&i" '") ),';
			put "         Overlay Axis << {{}},";
			put "         Separate Axes( 1 ),";
			put "         Connect Points( 1 ),";
			put "         SendToReport(";
			put "            Dispatch(";
			put "               {},";
			put '               "Overlay Plot",';
			put "               OutlineBox,";
			put '               Set Title( "Variance Proportion by Principal Component" )';
			put "            ),";
			put "            Dispatch(";
			put "               {},";
			put '               "101",';
			put "               ScaleBox,";
			put "               {min(0.5),Max(%sysevalf(&nx_pvca+0.5)),Inc( 1 ), Minor Ticks( 0 ), Rotated Labels( 1 )}";
			put "            ),";
			put '            Dispatch( {}, "", AxisBox( 2 ), Add Axis Label( "Proportion" ) ),';
			put '            Dispatch( {}, "Overlay Plot Graph", FrameBox, {Frame Size( 300, 200 )} ),';
			put '            Dispatch( {}, "", LegendBox, {Set Wrap(3)} )';
			put "         )";
			put "      )";/*End of Overlay plot*/
			*/
			put " ";
			put "//close(Here:pvcdata,nosave);";
			put "InsertInto(DataList,{{""Here:stckdt"",""Here:vcdata"",""Here:pvcdata""}});";
			put "InsertInto(TabButtonList, ""Variance Components Charts"");";
			put "InsertInto(TabCodeList,expr(Here:VC1()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);";
		%end;	
	%end;
   
   * filter outliers;
	put "expindata = ""&ExpInData"";";
	put "OutPath = ""&OutPath"";";
	put "droplist = {" ;
	put "   column(""_TYPE_""),";
   %do i = 1 %to %eval(&nv-1);
      put "   column(""&&var&i""),";
   %end;
   put "   column(""&&var&i"")";
	put "};";
   %if ^%length(&WithVars) and ^%length(&ListWithVars) %then %do;
		put "pclist = {" ;
		%do i = 1 %to %eval(&nFactors-1);
			put "   column(""&&prinlab&i""),";
		%end;
		put "   column(""&&prinlab&i"")";
		put "};";
   	put "InsertInto(droplist,pclist);";
   %end;
	put "include(""$GENOMICS_HOME\JSLFiles\Filter Outliers.JSL"");";

   %if ^%length(&WithVars) and ^%length(&ListWithVars) %then %do;
		%if &NFactors > 10 %then %let NFactor2 = 10;
		%else %let nFactor2 = &nFactors;
		put "Here:pca = Function({},{default local}," ;
		put "If(IsEmpty(Here:pcadata),";	
		put "  Here:pcadata=open(""&ClientOutPath.&DataCorrelation_OutDataName..sas7bdat"",invisible);";
		%if %length(&ColorVars) %then %do;
			put "  Here:pcadata<<Color By Column(" @@;
			%if &nColorVar = 1 %then put "  Name(""&ColorVars""));";
			%else put "  :ColorIndex);"; ;
			%let colorstring = ;
		%end;
		%else %do;
			%let colorstring = %str(ColorClusters(1),);
		%end;
		put "  %str(column(%"&labvar%") << label(1);)" / " ";
		put "  Here:pcadata << Select Where(1);";
		put "  Here:pcadata << Markers(8);";
		put "  Here:pcadata << Clear Select;";
		put ",";  /*is empty else stmnt */
		put "Current Data Table(Here:pcadata);";
		put ");";
		put "HListBox( ";
		put "   VListBox( ";
		put "      ButtonBox(""Create Subset Experimental Design Data Set, Excluding Selected Points"",";
		put "         SaveInvSubset;";
		put "      ),";
    	put "      Multivariate(Y( ";
		%do i=1 %to %eval(&nFactor2-1); 
			put "         Column(""&&prinlab&i""),";
		%end;
		put "         Column(""&&prinlab&nFactor2"")),";
		put "         Scatterplot Matrix(Density Ellipses(1), Ellipse Color(1), Horizontal(1)),";
		put "         Parallel Coord Plot(1),Outlier Analysis,"; 
		put '         title("Correlation Analysis of Principal Components"),'; 
		put "         SendToReport(";
		%do i=1 %to %eval(&nFactor2 * &nFactor2);
			put '            Dispatch({"Scatterplot Matrix"},"Multiv Plot",' @;
			put " FrameBox(&i)," @;
         put ' {Marker Drawing Mode(Fast)}),';
		%end;
		put '            Dispatch({"Outlier Analysis","Mahalanobis Distances"},"Multiv Outlier",';
		put "               FrameBox, {Marker Drawing Mode(Fast)}),";
		put '            Dispatch({}, "Parallel Coordinate Plot", OutlineBox,'; 
      put '               Set Title("Parallel Plot of Principal Components"))';
		put "         )";
		put "      )";
      put "   )";
     
      put "));";
   	put " ";
		%if &nFactor2 > 2 %then %do;
			put "Here:PCA3D = Function({},{default local}," ;
			put "If(IsEmpty(Here:pcadata),";	
			put "  Here:pcadata=open(""&ClientOutPath.&DataCorrelation_OutDataName..sas7bdat"",invisible);";
			%if %length(&ColorVars) %then %do;
				put "  Here:pcadata<<Color By Column(" @@;
				%if &nColorVar = 1 %then put "  Name(""&ColorVars""));";
				%else put "  :ColorIndex);"; ;
				%let colorstring = ;
			%end;
			%else %do;
				%let colorstring = %str(ColorClusters(1),);
			%end;
			put "  %str(column(%"&labvar%") << label(1);)" / " ";
			put "  Here:pcadata << Select Where(1);";
			put "  Here:pcadata << Markers(8);";
			put "  Here:pcadata << Clear Select;";
			put ",";  /*is empty else stmnt */
			put "  Current Data Table(Here:pcadata);";
			put ");";
			put "vListBox( ";
			put "   ButtonBox(""Create Subset Experimental Design Data Set, Excluding Selected Points"",";
			put "      SaveInvSubset;";
			put "   ),";
			put "   Scatterplot 3D(";
			put "      Y(Column(""&prinlab1""),Column(""&prinlab2""),Column(""&prinlab3"")),";
			put "      Frame3D(";
			put "	      Set Graph Size( 450, 400 ),";
			put "	      Set Rotation( -79.1304781643282, 0.39728042891252, 44.9772463556906 ),";
			put "         Set Marker Transparency( 0.6 ),";
			put "         Set Marker Quality( 1 ),";
			put "         Set Marker Scale( 2.5 )";
			put "      )";
			put "   ),";
      	
			put "));";
			put "InsertInto(DataList,{{""Here:pcadata""}});";
			put "InsertInto(TabButtonList, ""3D PCA Plot"");";
			put "InsertInto(TabCodeList,expr(Here:PCA3D()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);";
		%end;
		put "InsertInto(DataList,{{""Here:pcadata""}});";
		put "InsertInto(TabButtonList, ""2D PCA Plots"");";
		put "InsertInto(TabCodeList,expr(Here:pca()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
		put " ";

		%if %length(&ExpInData) & %length(&FactorVars)%then %do;
			%let i = 1;
			%do %while(%length(%scan(&FactorVars,&i))); 
				%let _Var&i = %scan(&FactorVars,&i);
				%let i = %eval(&i+1);
			%end;
			%let _nV = %eval(&i-1); 
			put "Here:ftrDensity = Function({},{default local}," ;
			put "If(IsEmpty(Here:pcadata),";	
			put "  Here:pcadata=open(""&ClientOutPath.&DataCorrelation_OutDataName..sas7bdat"",invisible);";
			%if %length(&ColorVars) %then %do;
				put "  Here:pcadata<<Color By Column(" @@;
				%if &nColorVar = 1 %then put "  Name(""&ColorVars""));";
				%else put "  :ColorIndex);"; ;
				%let colorstring = ;
			%end;
			%else %do;
				%let colorstring = %str(ColorClusters(1),);
			%end;
			put "  %str(column(%"&labvar%") << label(1);)" / " ";
			put "  Here:pcadata << Select Where(1);";
			put "  Here:pcadata << Markers(8);";
			put "  Here:pcadata << Clear Select;";
			put ",";  /*is empty else stmnt */
			put "  Current Data Table(Here:pcadata);";
			put ");";
			put "   vListBox(";
			put "      Distribution(";
			%do i = 1 %to %eval(&_nV-1); 
				put "         Nominal Distribution( Column(Name(""&&_Var&i"")) ),";
			%end;
			put "         Nominal Distribution( Column(Name(""&&_Var&i"")) )";
			put "      )";
		
			put "   )";
			put ");";
			put " ";
			put 'InsertInto(DataList,{{"Here:pcadata"}});';
			put 'InsertInto(TabButtonList, "Factor Exploration");';
			put 'InsertInto(TabCodeList,expr(Here:ftrDensity()));';
			put 'InsertInto(TabAfterCodeList, " ");';
			put 'InsertInto(TabOpenList,1);';
			put " ";
		%end;
   %end;
   
   %if ^%length(&WithVars) and ^%length(&ListWithVars) %then %do;
		put "InsertInto(DataList,{{""Here:eigdata""}});";
		put "InsertInto(TabButtonList, ""Scree Plot"");";
		put "InsertInto(TabCodeList,expr(Here:scree()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
	%end;

   put "Here:hm = Function({},{default local}," ;
	put "If(IsEmpty(Here:pcadata),";	
	put "  Here:pcadata=open(""&ClientOutPath.&DataCorrelation_OutDataName..sas7bdat"",invisible);";
	%if %length(&ColorVars) %then %do;
		put "  Here:pcadata<<Color By Column(" @@;
		%if &nColorVar = 1 %then put "  Name(""&ColorVars""));";
		%else put "  :ColorIndex);"; ;
		%let colorstring = ;
	%end;
	%else %do;
		%let colorstring = %str(ColorClusters(1),);
	%end;
	put "  %str(column(%"&labvar%") << label(1);)" / " ";
	put "  Here:pcadata << Select Where(1);";
	put "  Here:pcadata << Markers(8);";
	put "  Here:pcadata << Clear Select;";
	put ",";  /*is empty else stmnt */
	put "  Current Data Table(Here:pcadata);";
	put ");";
   put "   HListBox(, ";	
	put "   VListBox(";
	put "      ButtonBox(""Create Subset Experimental Design Data Set, Excluding Selected Rows"",";
	put "				Here:pcadata << Clear Column Selection;";
	put "         SaveInvSubset;";
	put "      ),";
   put "      Hierarchical Cluster(Y(";
   %do i = 1 %to %eval(&nv-1);
      put "         Name(""&&var&i""), ";
   %end;
   put "         Name(""&&var&i"")), " /
       "         Method(Ward), Standardize(0), Geometric X Scale(1), &colorstring " /
       "         Two way clustering, Number of Clusters(10), Color Map(Blue to Gray to Red), Legend(1)," /
		 '         title("Hierarchical Clustering Analysis on Correlations"), Dendrogram Scale( Distance Scale ),';
   put '         SendToReport(Dispatch({}, "Dendrogram", OutlineBox, Set Title("Heat Map and Dendrogram")),'/
       '            Dispatch({"Dendrogram"},"", CellPlotBox(2), Frame Size(350,350))' /
       '         )';
	put '      )';
	put "   ),";
	
   put ")); "; 
   put " ";

	put "InsertInto(TabButtonList, ""Correlation Heat Map"");";
	put "InsertInto(TabCodeList,expr(Here:hm()));";
	put "InsertInto(TabAfterCodeList, "" "");";
	put "InsertInto(TabOpenList,1);";
	put "InsertInto(DataList,{{""Here:pcadata""}});";
	put " ";
	
	put "Here:dddd=Function({},{default local},";
	put "  If(IsEmpty(Here:pdata),";
	put "   Here:pdata=open(""&ClientOutPath.&DataCorrelation_OutDataTestName..sas7bdat"",invisible),";
	put "   Current Data Table(Here:pdata));";
	put "   vlistbox(";	
	put "      Distribution(";
	put "         Continuous Distribution( Column( :Corr ), Outlier Box Plot( 0 ) ),";
	put "         Continuous Distribution( Column( :Raw_p ), Outlier Box Plot( 0 ) ),";
	put "         Continuous Distribution( Column( :NegLog10_p ), Outlier Box Plot( 0 ) )";
	put "      )";
	put "   )";
	put ");";
	put " ";
	put "InsertInto(DataList,{{""Here:pdata""}});";
	put "InsertInto(TabButtonList, ""Correlation Distribution"");";
	put "InsertInto(TabCodeList,expr(Here:dddd()));";
	put "InsertInto(TabAfterCodeList, "" "");";
	put "InsertInto(TabOpenList,1);";
	put " ";
run; 

%let incds = &DataCorrelation_OutDataName..sas7bdat,&DataCorrelation_OutDataTestName..sas7bdat,&DataCorrelation_OutDataEigName..sas7bdat; 
%if %length(&RandomEffects) %then %do;
	%let incds = &incds,&DataCorrelation_OutDataVCName..sas7bdat,&DataCorrelation_OutDataPVCName..sas7bdat;
%end;

%PackageInsertFile(jsl code from &ProcessName, &OutPath, &DataCorrelationScriptName, "&incds", text/plain, yes);

/*%if (&exiterror=0) and ^%length(&WithVars) and %length(&RandomEffects) %then %do;
	%PackageInsertFile(rtf file from &ProcessName, &OutPath, &VCOutFile, , rtf, yes);
%end;  */

/* Button to launch Array Group Corr AP */

* create JSL file;

filename jslbtns "&DataCorrelationScript";
%global _iewindow_;
%let _iewindow_=dcbuttonswindow;

data _null_;
	file jslbtns lrecl=32767 mod;
	put "//!";
	put "Names Default to Here(1);";
	put "wait(0.5);";
	put "if(host is(Windows),";
	put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, windows) ),";
	put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, POSIX) ));";
	put "dcbuttonswindow = Function({},{default local},";
	put "		VListBox(";
	put "			TextBox(""Click on the button below to launch the Correlation and Grouped Scatterplots process using the same data sets."",<<Set Wrap(210)),";
	put "			TextBox("" ""),";
	put "			ButtonBox(""Correlation and Grouped Scatterplots"",";
	put "				if(host is(Windows),";
	put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/ArrayGroupCorrelation.jsl"", absolute, windows)),";
	put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/ArrayGroupCorrelation.jsl"", absolute, POSIX) ));";
	put "	";
	put "				jg:setDefaultParmValues(2, ""ArrayGroupCorrelation"");";
	put "				jg:setCurrentParmValue(2, ""ArrayGroupCorrelation"", jg:getParmNum(""ArrayGroupCorrelation"",""InData""),""&InPath.&InData..sas7bdat"");";
	put "				jg:ArrayGroupCorrelation_updateDeps;";
	%if %length(&ExpInData) %then %do;
		put "				jg:setCurrentParmValue(2, ""ArrayGroupCorrelation"", jg:getParmNum(""ArrayGroupCorrelation"",""ExpInData""),""&ExpInPath.&ExpInData..sas7bdat"");";
		put "				jg:ArrayGroupCorrelation_updateDeps;";
	%end;
	put "				jg:setCurrentParmValue(2, ""ArrayGroupCorrelation"", jg:getParmNum(""ArrayGroupCorrelation"",""OutPath""),""&ClientOutPath"");";
	put "			)";;
	put "		),";
	put ");";
	put "Here:LaunchCode = dcbuttonswindow();";
run;
%AddDatatoStudy(&DataCorrelation_OutDataName,_type=tall);

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend DataCorrelation;

* execute macro;
%DataCorrelation;
