/*
JMP Genomics Analytical Process

Name: DataDistribution
Author: Russ Wolfinger, SAS Institute Inc.
Support: Tzu-Ming Chu, SAS Institute Inc.
History:
   sasrdw 28Apr05 initial coding                          

Description: 
DataDistribution displays univariate distribution
results for variables in a SAS data set.  It also
optionally computes and overlays density estimates
for the variables.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData Vars IDVar ListVars  WhereStmt DetailPlots Uniform Overlay 
SmoothingBandwidthMethod SmoothingBandwidthMultiplier 
NGrid GFormat ExpInData GroupVars ColorVars ExpLabelVar BoxPlots StdMeanPlots 
CumSumPlots OutPath ServerOutPath ServerOutPath  ServerOutPath MacroPath 
DataDistribution_OutData DataDistribution_OutDen DataDistribution_OutBox DataDistribution_OutGrpExp
DataDistribution_OutCum DataDistributionScript;

 
*ProcessBody;

%macro DataDistribution;

* initialization and argument checking;
%let ProcessName = DataDistribution; 

* default values for parameters;
%if ^%length(&ngrid) %then %let ngrid = 100;
%if ^%length(&GFormat) %then %let GFormat = 5.2;
%if ^%length(&SmoothingBandwidthMultiplier) %then %let SmoothingBandwidthMultiplier = 1;
	
%include "&MacroPath./UtilityMacros.sas" / nosource;

%if %upcase(&DetailPlots)=NO & %upcase(&Overlay)=NO & %upcase(&BoxPlots)=NO %then %do;
   	%put ERROR: One of the following options must be checked: Overlay Density Estimates, Display Side-by-Side Box Plots, or Display Detailed Distribution Statistics and Histograms.;
	%let exiterror = 1;
	%goto exit;
%end;

%GetOrigFileName(InData,OrigInData);

%PathName(&InData);
%if &exiterror %then %do;
   %put ERROR: Input dataset specification is invalid.;
   %let exiterror = 1;
   %goto exit;
%end;
%global InPath;
%let InPath = &TmpPath;
%put InPath = &InPath;

%let eid = %nrbquote(&ExpInData);
%put eid = &eid;
%if %length(&eid) %then %let ed = 1;
%else %let ed = 0;
%CheckInData(_ExpFlag=&ed);
%if &ed %then %let nV = &_nTallCol_;
%else %let nV = 0;
%if &exiterror %then %goto exit;

* add listvars to vars;
%if %length(&ListVars) | %length(&Vars) %then %do;
	%CheckInData2(InLib.&InData,_Vars=&Vars,_VarListStyle=&ListVars);
	%let nV = &_nTallCol_;
%end;

* vars from edf by default;
%*if &ed & ^%length(&Vars) %then %let Vars = &_AllCol_;

* prepare output data sets;
%PrepOutDS(DataDistribution_OutDen,DataDistribution_OutDenName,&OrigInData,_den);
%PrepOutDS(DataDistribution_OutBox,DataDistribution_OutBoxName,&OrigInData,_box);
%if (&ed) and %length(&GroupVars) %then %do;
	%PrepOutDS(DataDistribution_OutData,DataDistribution_OutDataName,&OrigInData,_dgm);
	%PrepOutDS(DataDistribution_OutGrpExp,DataDistribution_OutGrpExpName,&OrigInData,_dge);
%end;
%else %do;
	%PrepOutDS(DataDistribution_OutData,DataDistribution_OutDataName,&OrigInData);
%end;
%if %upcase(&CumSumPlots)=YES %then %do;
	%PrepOutDS(DataDistribution_OutCum,DataDistribution_OutCumName,&OrigInData,_cum);
%end;

%if %length(&WhereStmt) %then %do;
	data indat;
		set InLib.&InData;
		where &WhereStmt;
	run;
	%let indat = indat;
	%let diff = 1;
%end;
%else %do;
	%let indat = InLib.&InData;
	* check if indata equals outdata;
	data _null_;
		inpath = "&InPath";
		OutPath = "&OutPath";
		indata = "&InData";
		outdata = "&DataDistribution_OutData";
		if (inpath = outpath) and (indata = outdata) then diff = 0;
		else diff = 1;
		call symput('diff',trim(left(diff)));
	run;
%end;

%if (&diff = 1) and (%upcase(&DetailPlots)=YES or (&ed and %length(&GroupVars))) %then %do;
	data &DataDistribution_OutData;
		set &indat;
	run;
	%let usedat = &DataDistribution_OutData;
%end;
%else %do;
	%let usedat = &indat;
%end;
%if &syserr %then %goto exit;

* make sure all variables are numeric;
%let allnum = 1;
%do j = 1 %to &nV;
	%let vName = &&_ColName_&j._;
	%let d1 = %sysfunc(open(&usedat));
	%let vNum = %sysfunc(varnum(&d1,&vName));
	%let vType = %sysfunc(vartype(&d1,&vNum));
	%if (&vType = C) %then %do;
		%put ERROR: This process only handles numeric variables, but variable &vName is not numeric.;
		%put Please remove this variable from the selection list.;
		%let allnum = 0;
	%end;
	%let d1 = %sysfunc(close(&d1));
%end;

%if (&allnum = 0) %then %do;
	%let exiterror = 1;
	%goto exit;
%end;

%put nv = &nv;
%put;

%if (&nv <= 0) %then %do;
	%put ERROR: You must select one or more variables for which to display distributions.;
	%let exiterror = 1;
	%goto exit;
%end;

* create new data sets for groups;
%if (&ed) and %length(&GroupVars) %then %do;

	* create group macro variables;
	%GroupMacroVars(data=ed,vars=&GroupVars,prefix=g,maxnamelen=24);
	
	* replace design and input data sets with ones corresponding to group means;
	proc sort data=ed nodupkey;
		by &GroupVars;
	run;
	
	* build new column names from values of grouping variables;
	data cn;
		length ColumnName $ 32;
		%do i=1 %to &gNum;
			ColumnName = "Mean_" || "&&gID&i";
			
			* convert to valid SAS variable name;
			if substr(ColumnName,1,1) in ('0','1','2','3','4','5','6','7','8','9') then do;
				ColumnName = "_" || ColumnName;
			end;
			lencn = length(trim(left(ColumnName)));
			do j=1 to lencn;
				charj = lowcase(substr(ColumnName,j,1));
				if charj ^in ('_','0','1','2','3','4','5','6','7','8','9',
					'a','b','c','d','e','f','g','h','i','j',
					'k','l','m','n','o','p','q','r','s','t',
					'u','v','w','x','y','z') then do;
					substr(ColumnName,j,1) = '_';
				end;
			end;
			output;
			call symput("mean&i",ColumnName);
			call symput("std&i","Std_"||substr(ColumnName,6,lencn-5));
		%end;
		keep ColumnName;
	run;
	
	data ed;
		set ed;
		drop ColumnName;
	run;
	
	data ed;
		merge ed cn;
	run;

	data &DataDistribution_OutGrpExp;
		set ed;
	run;
	
	data &DataDistribution_OutData;
		set &DataDistribution_OutData;
		%do i=1 %to &gNum;
			&&mean&i = mean(of %do k=1 %to &&gNum&i; &&G&i.R&k %end;);
		%end;
		%do i=1 %to &gNum;
			%if (&&gNum&i < 2) %then %do;
				&&std&i = .;
			%end;
			%else %do;
				&&std&i = std(of %do k=1 %to &&gNum&i; &&G&i.R&k %end;);
			%end;
		%end;
		* labels for JMP;
		label 	
			%do i=1 %to &gNum;
				&&mean&i="Mean &&gID&i" 
				&&std&i="StdDev &&gID&i" 
			%end;
		;
		drop %do i = 1 %to &nv; &&_ColName_&i._ %end;;
	run;
	
	%*let Vars = ;
	%do i=1 %to &gNum;
		%*let Vars = &Vars &&mean&i;
		%let _ColName_&i._ = &&mean&i;
	%end;	
	%let nv = &gNum;
%end;

proc means data=&usedat min max noprint;
	var %do i = 1 %to &nv; &&_ColName_&i._ %end;;
	output out=stats(drop=_TYPE_);
run;
/*
title "Summary Statistics";
proc print data=stats;
run;
*/

* kernel density estimates;
%if %upcase(&Overlay)=YES | %upcase(&BoxPlots) = YES %then %do;

	%if &SmoothingBandwidthMultiplier <= 0 %then %do;
   	%put ERROR: The smoothing bandwidth multiplier must be greater than 0.;
   	%let exiterror = 1;
   	%goto exit;
   %end;
	
   %if &ngrid<=1 %then %do;
   	%put ERROR: The number of density grid points must be greater than 1.;
   	%let exiterror = 1;
   	%goto exit;
   %end;
   
   /*data OutLib.dd_stats; set stats; run;*/

	data mm;
      set stats; 
      if _STAT_ in ("MIN","MAX");
      array Chp{*} %do i = 1 %to &nv; &&_ColName_&i._ %end;;
      med = median(of %do i = 1 %to &nv; &&_ColName_&i._ %end;);
      mad = mad(of %do i = 1 %to &nv; &&_ColName_&i._ %end;);
	   max = med;
	   min = med;
		flag = 0;
		fCount = 0;
      do i = 1 to dim(Chp);
			name = "BadVFlag"||trim(left(i));
		   if Chp{i} > med+10*mad | Chp{i} < med-10*mad then do;
			   flag = 1;
				fCount = fCount + 1;
				call symput(name,1);
			end;
			else do;
            if max < Chp{i} then max = Chp{i};
            if min > Chp{i} then min = Chp{i};
				call symput(name,0);
			end;
      end;
	   if _STAT_ = "MAX" then do;
			call symputx("maxv",max);
			if max < 50 then call symputx("LogFlag",1);
			else call symputx("LogFlag",0);
		end;
	   if _STAT_ = "MIN" then call symputx("minv",min);
		call symput("nBadVar",fCount);
   run;
      
   * updated &minv and &maxv (S0979644);     
   %let x = %sysevalf((&maxv-&minv)/10**7);
   %let minv = %sysevalf(&minv-&x);
   %let maxv = %sysevalf(&maxv+&x);
      
   data _null_;
   	range = &maxv - &minv;
   	call symputx('range',range);
   	call symputx('incr',range/&ngrid);
   run;

	%put minv = &minv;
	%put maxv = &maxv;
	%put range = &range;
	%put ngrid = &ngrid;
	%put incr = &incr;

	data 
		%do i = 1 %to &nV;
			%let sv&i = &&_ColName_&i._;
			k&i(keep= &&_ColName_&i._)
		%end;;
		set &usedat;
	run;

	%if %upcase(&CumSumPlots)=YES %then %do;
		* extract total features;
		data _null_;
			set stats;
      	if _STAT_ = "N";
      	array Chp{&nv} %do i = 1 %to &nv; &&_ColName_&i._ %end;;
			call symput("TotalFeatures",_FREQ_);
			do i = 1 to &nv;
				name = "N_"||trim(left(i));
				call symput(name,Chp(i)); * number of non-missing features;
			end;
		run;		

		* extract summary;
		data _null_;
			set stats;
      	if _STAT_ = "MEAN";
      	array Chp{&nv} %do i = 1 %to &nv; &&_ColName_&i._ %end;;
			array N_{&nv} _TEMPORARY_ (%do i=1 %to &nv; &&N_&i %end;);
			do i = 1 to dim(Chp);
				name = "Sum_"||trim(left(i));
				call symput(name,Chp(i)*N_(i));
			end;
		run;

		* preparing data for cumulative summary plots;
		%do i = 1 %to &nv;
			proc sort data=k&i out=cs&i;
				by descending &&sv&i;
			run;

			data cs&i;
				set cs&i;
				if &logFlag then do;
					if &&sv&i=. then Var2_&i=0;
					else Var2_&i = 2**&&sv&i;
				end;
				else Var2_&i = log2(&&sv&i);
			run;

			proc summary data=cs&i noprint;
				var Var2_&i;
				output out=stats2 sum=;
			run;

			data _null_;
				set stats2;
				call symput("Sum2_&i",Var2_&i);
			run;
		
			* keep ~5,000 point for drawing plots;
			%if &TotalFeatures > 10000 %then %do;
				%let nKeep = %sysfunc(round(&TotalFeatures/5000));
			%end;
			%if &LogFlag %then %do;
				%let string = Count;
				%*let string2 = Logarithm;
			%end;
			%else %do;
				%let string = Logarithm;
				%*let string2 = Count;
			%end;
			data cs&i;
				set cs&i;
				retain CumPtg1_&i CumPtg2_&i GenPtg 0;
				GenPtg = _n_ / &TotalFeatures * 100;
				CumPtg1_&i = CumPtg1_&i + (&&sv&i/&&Sum_&i) * 100;
				CumPtg2_&i = CumPtg2_&i + (Var2_&i/&&Sum2_&i) * 100;
				%if &TotalFeatures > 10000 %then %do;
					if mod(_n_,&nKeep)=1;
				%end;
				label GenPtg = "Cumulative % of Features"
					CumPtg1_&i = "Cumulative % Summary of &&sv&i"
					CumPtg2_&i = "Cumulative % Summary of &&sv&i in &string Scale"
					Var2_&i = "&&sv&i in &string Scale";
			run;
		%end;
		data &DataDistribution_OutCum;
			merge %do i = 1 %to &nv; cs&i %end;;
		run;
		
		%RemoveData(%do i = 1 %to &nv; cs&i %end;);
	%end;

	* loop over variables and compute density estimates;	
	%*let Vars2 = ;
	%do j=1 %to &nv; 
		%put;
		%put Computing kernel density for &&sv&j, variable &j of &nv;
		%put;
		/* this may not be a good way for flaging outlier.;
	   	%if (&&BadVFlag&j) %then %do;
			%put ERROR: Column &sv has a very large range difference from the other columns.;
			%put The overlayed kernel density estimated plot will not include this column.;
         		%goto jexit;
		%end;
		*/
		%*let Vars2 = &Vars2 &&sv&j;
      
      		/* subsetting might help performance for really large data sets;
      		data k;
      			set &usedat;
      			keep &sv;
      		run;
      		*/
      		* do binning in a data step for speed;
      		data a&j;
      			set k&j;
      			idx = ceil((&&sv&j - &minv)/&incr); 
      		run;
      
      		data k&j;
      			set k&j end=the_end;
      			array g{*} g1-g&ngrid;
      			retain g1-g&ngrid 0;
      	
      			* tally counts as weights;
      			idx = ceil((&&sv&j - &minv)/&incr);
      			if not missing(idx) and idx >= 1 and idx <= &ngrid then g[idx] = g[idx] + 1;
      	
      			* output midpoints and weights;
      			if (the_end) then do;
      				&&sv&j = &minv - &incr/2;
      				do i=1 to &ngrid;
      					&&sv&j = &&sv&j + &incr;
      					w = g[i];
      					output;
      				end;
      			end;
      			keep &&sv&j w;
      		run;
      
      		/*
      		proc print;
      		run;
      		*/
      
      		/*data OutLib.dd_k&j; set k&j; run;*/
      
		ods exclude all;
		ods noresults;
		
		proc kde data=k&j;
			univar &&sv&j / out=kde bwm=&SmoothingBandwidthMultiplier
				gridl=&minv gridu=&maxv ngrid=&ngrid method=&SmoothingBandwidthMethod;
			weight w;
		run;
				
		ods exclude none;
		ods results;
         		
		%if &syserr %then %do;
			%put The data may be extremely skewed.  If so, it will help to transform the data or specify a different Bandwidth Computation Method.;
			%let exiterror = 1;
			%goto exit;
		%end;
		
		%*RemoveData(k&j);

		data kde2 kde3;;
			set kde;
			format _Index &GFormat;
			_Index = Value;
			keep _Index Density;
		run;
		
		data kde2;
			set kde2;
			&&sv&j = Density;
			keep _Index &&sv&j;
		run;
	
		%if (&j=1) %then %do;
			data den;
				set kde2;
			run;
			data density;
				length ColumnName $ 32;
				set kde3;
				ColumnName = "&&sv&j";
			run;
      
      			%if %upcase(&BoxPlots) = YES %then %do;
			         data &DataDistribution_OutBox;
					set kde(rename=(Var=ColumnName Value=Response));
					if count > 0;
					label ColumnName="ColumnName" Response="Response";
				run;
			%end;
		%end;
		%else %do;
			data den;
				merge den kde2;
				by _Index;
			run;
      		   	%if %upcase(&BoxPlots) = YES %then %do;
			         data &DataDistribution_OutBox;
					set &DataDistribution_OutBox kde(rename=(Var=ColumnName Value=Response));
					if count > 0;
				run;
			%end;
			data density;
				set density kde3;
				if ColumnName = "" then ColumnName ="&&sv&j";
			run;
		%end;
	   	%jexit:;
	%end;

	*Get the max _Index;
	proc sort data = density out = den_out;
		by descending _Index;
	run;
	data _null_;
		set den_out end=last;
		if _n_=1 then call symput("maxInd",trim(left(_Index)));
		if last then call symput("minInd",trim(left(_Index)));
	run;
	*Get the max density;
	proc sort data = density out = den_out;
		by descending Density;
	run;
	data _null_;
		set den_out end=last;
		if _n_=1 then call symput("maxDen",trim(left(Density)));
		if last then call symput("minDen",trim(left(Density)));
	run;
	
	* transpose for parallel plots;
	data den;
		set den;
		did = "d_" || trim(left(put(_index,&GFormat)));
	run;
	option label;
	proc transpose data=den out=&DataDistribution_OutDen;
		var %do i = 1 %to &nV; &&sv&i %end;;
		id did;
		idlabel _Index;
	run;
	
	data &DataDistribution_OutDen;
		set density;
		/*ColumnName = _name_;
		drop _name_;*/
	run;
	
	%let ColorFlag =0;
	%let ColorVarGB =;
	* merge experimental design metadata;
	%if %length(&ExpInData) %then %do;
	   %if ^%length(&ColorVars) %then %let ColorVars = ColumnName;
	   %else %let ColorFlag = 1;
	   %if ^%length(&ExpLabelVar) %then %let LabelVar = ColumnName;
	   %else %let LabelVar = &ExpLabelVar;
   
   	%let ColorVarGB =&ColorVars;
		%let i = 1;
		%do %while(%length(%scan(&ColorVars,&i)));
		   %let ColorVar&i = %scan(&ColorVars,&i);
		   %let ColorVarGB = &&ColorVar&i ;
			%let i = %eval(&i+1);
		%end;
		%let nColorVar = %eval(&i-1);
		proc sort data=&DataDistribution_OutDen;
			by ColumnName;
		run;
		
		proc sort data=ed;
			by ColumnName;
		run;

		/*
		libname c "C:\temp";
		data c.den;
			set &DataDistribution_OutDen;
		run;
		
		data c.ed;
			set ed;
		run;
		*/
		
		/*data OutLib.d_ed; set ed; run;*/
		/*data OutLib.d_OutDen; set &DataDistribution_OutDen; run;*/
		
		data &DataDistribution_OutDen;
			merge ed &DataDistribution_OutDen;
			by ColumnName;
			%if &nColorVar > 1 %then %do;
				ColorIndex = trim(left(&ColorVar1))
					%do i = 2 %to &nColorVar;
						||"_"|| trim(left(&&ColorVar&i)) 
					%end; ;
			%end;
		run;
	%end;
	%else %do;
		%let nColorVar = 0;
		%let LabelVar = ;
	%end;
	
	* create macro variables for JMP;
	%MakeVarList(_data=&DataDistribution_OutDen,_nvar=ng,_prefix=d_);
	%if &nColorVar = 1 %then %CheckLabel(&DataDistribution_OutDen,ColorVars);
	%if %length(&LabelVar) %then %CheckLabel(&DataDistribution_OutDen,LabelVar);		

%end;
	
%if &exiterror %then %goto exit;

* create list of labels for JMP;
%if %upcase(&Overlay)=YES %then %do;
	%do j=1 %to &ng;
		%CheckLabel(&DataDistribution_OutDen,d_&j);
	%end;
%end;
%if (%upcase(&DetailPlots)=YES) or (&ed and %length(&GroupVars)) %then %do;
	%let lv = ;
	%do j=1 %to &nv; 
		%let sv = &&_ColName_&j._ ;
		/* %put Checking label for &sv, variable &j of &nv; */
		%CheckLabel(&DataDistribution_OutData,sv,_New=lv);
		%let Label&j = &lv;
		%if (&ed and %length(&GroupVars) and %upcase(&StdMeanPlots) = YES) %then %do;
			%let sv = &&std&j;
			%CheckLabel(&DataDistribution_OutData,sv,_New=lv);
			%let StdLabel&j = &lv;
		%end;
	%end;
%end;
%if %upcase(&CumSumPlots)=YES %then %do;
	%let lv = ;
	%do j=1 %to &nv; 
		%let sv = CumPtg1_&j;
		%CheckLabel(&DataDistribution_OutCum,sv,_New=lv);
		%let CumLabel1_&j = &lv;
		%let sv = CumPtg2_&j;
		%CheckLabel(&DataDistribution_OutCum,sv,_New=lv);
		%let CumLabel2_&j = &lv;
	%end;
	%let sv = GenPtg; 
	%CheckLabel(&DataDistribution_OutCum,sv,_New=lv);
	%let GenPtgLabel = &lv;
%end;

* create JSL file;
%PrepOutFile(DataDistributionScript,DataDistributionScriptName,&ProcessName Script for &InData..jsl);
%global _mainjslfile_ FullAPName;
%let _mainjslfile_ = &DataDistributionScript;
%let FullAPName=%str(Distribution Analysis);


* put JSL statements;
data _null_;
   file "&DataDistributionScript" lrecl=32767;
   put "//!";
   put "Names Default to Here(1);";
	put "Here:DataList = {};";
	put "Here:TabButtonList = {};";
	put "Here:TabCodeList = {};";
	put "Here:TabAfterCodeList = {};";
	put "Here:TabOpenList = {};";
	/*
	put "show(Here:DataList,Here:TabButtonList,Here:TabCodeList,Here:TabAfterCodeList,Here:TabOpenList);";
	*/
   %if %upcase(&Overlay)=YES %then %do;
   	put " ";
		* set up for filter outliers drill down;
		put "expindata = ""&ExpInData"";";
		put "OutPath = ""&OutPath"";";
		put "droplist = {" @@;
	/*	%do j=1 %to %eval(&ng-1);
			put "column(""&&d_&j""),";
		%end;
		put "column(""&&d_&ng"")";*/
		put "column(""Density""),";
		PUT "column(""_Index"")";
		put "};";
		put "include(""$GENOMICS_HOME\JSLFiles\Filter Outliers.JSL"");";
   	put "Here:ddkde=Function({},{default local},";
   	put "   If(IsEmpty(Here:data_den),";
		put "      Here:data_den=open(""&ClientOutPath.&DataDistribution_OutDenName..sas7bdat"",invisible);";
   	%if %length(&ColorVars) %then %do;

   	%let i = 1;
		%do %while(%length(%scan(&ColorVars,&i)));
		   %let ColorVar&i = %scan(&ColorVars,&i);
		   %let ColorVarGB = &&ColorVar&i ;
		   	put "	   Column(""&&ColorVar&i"")<<Data Type( Character );";
			%let i = %eval(&i+1);
		%end;
   			
		put "      Here:data_den<<Color By Column(";
         		%if &nColorVar = 1 %then put "Name(""&ColorVars""));";
			%else put ":ColorIndex);"; ;
		
		
   	%end;
	%else %do;
	   put "      Here:data_den<<Color By Column(:ColumnName);";
	%end;
   	%if %length(&LabelVar) %then %do;
		   put "      %str(Column(%"&LabelVar%") << label(1);)";
   	%end; 
		%else %do;
		   put '      Column("ColumnName") << label(1);';
		%end;
   	put "   );";
		put "   Current Data Table(Here:data_den);";
		*put "   ScrollBox(Size(800,450), Vlistbox(";	
		put "   Vlistbox(";		
		put "      ButtonBox(""Create Subset Experimental Design Data Set, Excluding Selected Curves"",";
		put "         SaveInvSubset;";
		put "      ),";
	   put "      TextBox(""Bandwidth Multiplier = &SmoothingBandwidthMultiplier""),";
/*   	put "      Parallel Plot( Y(";
		%do j=1 %to &ng;
			%if (&j > 1) %then %do;
				put ", ";
			%end;
			put "         %str(:Name(%"&&d_&j%"))" @@;
		%end;
   	put "      ),";
   	put '         SendToReport(';
		put '            Dispatch( {}, "", AxisBox, {Add Axis Label( "Density" )} ),';
   	put '            Dispatch({}, "Parallel Coord", FrameBox, {Frame Size(950, 300)})';
   	put '         ),';
   	put "         Scale Uniformly(1)";
   	put "      )"; */
   	*put "   ))";
   	put "   Hlistbox(";
   		put "maxDen=&maxDen+0.05;";
  		put "	gb=Graph Builder(";
  		put "	Size( 775,475 ),";
  		put "		Show Control Panel( 0 ),";
  		%if ^%length(&GroupVars) %then %do;
  		put "		Show Legend( 0 ),";
  		%end;
  		%if %length(&ColorVars) %then %do;
  			put "		Variables( X( :_Index ), Y( :Density ), Overlay( :ColumnName ) , Color(";
  				%if &nColorVar = 1 %then put " ""&ColorVarGB"" )),";
  				%else put ":ColorIndex)),"; ;
  		%end;
  				/*put "      Here:data_den<<Color By Column(";
		         		%if &nColorVar = 1 %then put "Name(""&ColorVars""));";
			%else put ":ColorIndex);"; ;*/
  		
  		%else %do;
  				put "		Variables( X( :_Index ), Y( :Density ), Overlay( :ColumnName ) ),";
  		%end;
  		put "		Elements( Line( X, Y, Legend( 7 ) ) ),";
  		put "		SendToReport(";
  		put "			Dispatch(";
  		put "				{},";
  		put "				""Graph Builder"",";
  		put "				OutlineBox,";
  		put "				{Set Title( ""Kernel Density Distributions"" )}";
  		put "			),";
  		put "			Dispatch(";
  		put "				{},";
  		put "				""_Index"",";
  		put "				ScaleBox,";
  		put "				{Format( ""Best"", 12 ), Min( &minInd ), Max( &maxInd ), ";
  		put "				Label Row( {Show Major Grid( 1 ), Show Minor Grid( 1 )} )}";
  		put "			),";
		put "			Dispatch(";
		put "				{},";
		put "				""Density"",";
		put "				ScaleBox,";
		put "				{Min( &minDen ), Max( maxDen ) }";
		put "			),";
  		put "			Dispatch( {}, ""graph title"", TextEditBox, {Set Text( "" "" )} ),";
  		put "			Dispatch( {}, ""X title"", TextEditBox, {Set Text( "" "" )} )";  		
  		put "      )"; 

  		put "		),";/* end of graph builder;*/
  		%if %length(&ExpInData) %then %do;
  			* adding a local data filter of etd;	
  			put " dkeylab = ""&ColorVars"";";
 			put " dlab = Words(dkeylab,"" "");";
  			put " For( i = 1, i <= N Items( dlab ), i++,";
  			put "	//	mergetmp = "":Name(\!""""||dlab[i]||""\!"")"";";
  			put "		mergetmp = "":""||dlab[i];";
  			put " 	If(i==1,";
  			put "			mergetxt = mergetmp;";
  			put "		,//else";
  			put "			mergetxt = mergetxt||"",""||mergetmp;";
  			put "		);";
  			put "		);"; 	
  			
  			put "	eval(parse(evalinsert(""\[";  			
			put "      gb << Local Data Filter(""on left"" ";
			%if &ColorFlag %then %do;
			put "         ,Add Filter( Columns(^mergetxt^)),";
			put "		    Mode( Show(1), Include(1) )";
			%end;
			put "      )";
			put "	]\"")));";
		%end;
   	put "   )";
   	put "   )";
	   put ");";
		put "InsertInto(DataList,{{""Here:data_den""}});";
		put "InsertInto(TabButtonList, ""Kernel Density Estimates"");";
		put "InsertInto(TabCodeList,expr(Here:ddkde()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
	%end;
   %if %upcase(&BoxPlots) = YES %then %do;
   	put " ";
   	put "Here:ddbp=Function({},{default local},";
   	put "   If(IsEmpty(Here:data_box),";
		put "      Here:data_box=open(""&ClientOutPath.&DataDistribution_OutBoxName..sas7bdat"",invisible);";
		put "   );";
		put "   Current Data Table(Here:data_box);";
		*put "   ScrollBox(Size(800,450), Vlistbox(";	
		put "   Vlistbox(";		
		put "      Oneway(Y( :Response), X( :ColumnName), Freq(:Count),";
		put "         Points(0), Box Plots(1), Mean Diamonds(1), Grand Mean(0)";
		put "      )";
		*put "   ))";
		put "   )";
		put ");";
		put "InsertInto(DataList,{{""Here:data_box""}});";
		put "InsertInto(TabButtonList, ""Box Plots"");";
		put "InsertInto(TabCodeList,expr(Here:ddbp()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
	%end;	
   %if (%upcase(&DetailPlots)=YES) or (&ed and %length(&GroupVars) and %upcase(&StdMeanPlots)=YES) %then %do;
   	put " ";
		%if (&ed and %length(&GroupVars) and %upcase(&StdMeanPlots)=YES) %then %do;
   		put "Here:ddsdm=Function({},{default local},";
 		  	put "   If(IsEmpty(Here:data_dp),";
			put "      Here:data_dp=open(""&ClientOutPath.&DataDistribution_OutDataName..sas7bdat"",invisible);";
			%if %length(&IDVar) %then %do;
				%CheckLabel(&DataDistribution_OutData,IDVar);
				put "      %str(Column(%"&IDVar%") << label(1);)";
			%end; 
			put "   );";
		   put "   Current Data Table(Here:data_dp);";
			put "  Vlistbox(";		
			%do j=1 %to &nv;
				put "      Bivariate(Y(:Name(""&&StdLabel&j"")),X(:Name(""&&Label&j"")),Fit Spline(1));";
			%end;
			put "   )";
			put ");";
			put "InsertInto(DataList,{{""Here:data_dp""}});";
			put "InsertInto(TabButtonList, ""Std Dev vs Mean Plots"");";
			put "InsertInto(TabCodeList,expr(Here:ddsdm()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);";		
		%end;
		%if %upcase(&DetailPlots) = YES %then %do;
   		put "Here:dddd=Function({},{default local},";
 		  	put "   If(IsEmpty(Here:data_dp),";
			put "      Here:data_dp=open(""&ClientOutPath.&DataDistribution_OutDataName..sas7bdat"",invisible);";
			%if %length(&IDVar) %then %do;
				%CheckLabel(&DataDistribution_OutData,IDVar);
				put "      %str(Column(%"&IDVar%") << label(1);)";
			%end; 
			put "   );";
		   put "   Current Data Table(Here:data_dp);";
			put "   ScrollBox(Size(950,650), Vlistbox(";		
			put "      Distribution(";
			%if %upcase(&Uniform) = YES %then %do;
				put "         Uniform Scaling(1),";
			%end;
			%do j=1 %to &nv;
				%if (&j > 1) %then %do;
					put ", ";
				%end;
				put "         Continuous Distribution(Column(Name(""&&Label&j"")),More Moments(1))" @@;
			%end;
			put "      ),";
			put "   ))";
			put ");";
			put "InsertInto(DataList,{{""Here:data_dp""}});";
			put "InsertInto(TabButtonList, ""Distribution Details"");";
			put "InsertInto(TabCodeList,expr(Here:dddd()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);";		
		%end;
	%end;
	%if %upcase(&CumSumPlots)=YES %then %do;
		put "Here:cump=Function({},{default local},";
 		put "   If(IsEmpty(Here:data_cump),";
		put "      Here:data_cump=open(""&ClientOutPath.&DataDistribution_OutCumName..sas7bdat"",invisible);";
		put "   );";
		put "   Current Data Table(Here:data_cump);";

		*put "   ScrollBox(Size(800,600), hlistbox(";	
		put "   hlistbox(";		
		put "      Graph Builder(";
		put "         Size( 450, 450 ),";
		put "         Show Control Panel( 0 ),";
		put "         Show Legend( 0 ), Auto Stretching(0),";
		put "         Variables(";
		put "            X( :NAME(""&GenPtgLabel"") ),";
		put "            Y( :NAME(""&CumLabel1_1"") ),";
		%do i = 2 %to %eval(&nv-1);
			put "            Y( :NAME(""&&CumLabel1_&i"") , Position( 1 ) ),";
		%end;
		put "            Y( :NAME(""&&CumLabel1_&i"") , Position( 1 ) )";
		put "         ),";
		put "         Elements(";
		put "            Line(";
		put "               X,";
		%do i = 1 %to %eval(&nv-1);
			put "               Y( &i ),";
		%end;
		put "               Y( &i )";
		put "            )";
		put "         ),";
		put "         SendToReport(";
		put "            Dispatch( {}, ""graph title"", TextEditBox, {Set Text( ""Original Scale"" )} ),";
		put "            Dispatch(";
		put "               {},";
		put '               "Y title",';
		put "               TextEditBox,";
		put '               {Set Text( "Cumulative % of Summary" )}';
		put "            )";
		put "         )";
		put "      ),";
	
		put "      Graph Builder(";
		put "         Size( 450, 450 ),";
		put "         Show Control Panel( 0 ),";
		put "         Show Legend( 0 ), Auto Stretching(0),";
		put "         Variables(";
		put "            X( :NAME(""&GenPtgLabel"") ),";
		put "            Y( :NAME(""&CumLabel2_1"") ),";
		%do i = 2 %to %eval(&nv-1);
			put "            Y( :NAME(""&&CumLabel2_&i"") , Position( 1 ) ),";
		%end;
		put "            Y( :NAME(""&&CumLabel2_&i"") , Position( 1 ) )";
		put "         ),";
		put "         Elements(";
		put "            Line(";
		put "               X,";
		%do i = 1 %to %eval(&nv-1);
			put "               Y( &i ),";
		%end;
		put "               Y( &i )";
		put "            )";
		put "         ),";
		put "         SendToReport(";
		put "            Dispatch( {}, ""graph title"", TextEditBox, {Set Text( ""&string Scale"" )} ),";
		put "            Dispatch(";
		put "               {},";
		put '               "Y title",';
		put "               TextEditBox,";
		put "               {Set Text( ""Cumulative % of Summary in &string Scale"" )}";
		put "            )";
		put "         )";
		put "      ),";
		*put "   ))";
		put "    )";
		put ");";
		put "InsertInto(DataList,{{""Here:data_cump""}});";
		put "InsertInto(TabButtonList, ""Cumulative Plots"");";
		put "InsertInto(TabCodeList,expr(Here:cump()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
	%end;

	* buttons to launch other aps;
	put "ddlaunchbuttons=Function({},{default local},";
	put "   VListBox(";
	put "			ButtonBox(""Data Standardize"",";
	put "				if(host is(Windows),";
	put "					Include(ConvertFilePath(""$HOME/JMPLS/ProcessLibrary/DataStandardize.jsl"", absolute, windows)),";
	put "					Include( ConvertFilePath(""$HOME/JMPLS/ProcessLibrary/DataStandardize.jsl"", absolute, POSIX) ));";
	put "	";
	put "				jg:setDefaultParmValues(1, ""DataStandardize"");";
	put "				jg:setCurrentParmValue(1,""DataStandardize"", jg:getParmNum(""DataStandardize"",""InData""),""&InPath.&InData..sas7bdat"");";
	%if %length(&ExpInData) %then %do;
		put "				jg:setCurrentParmValue(1,""DataStandardize"", jg:getParmNum(""DataStandardize"",""ExpInData""),""&ExpInPath.&ExpInData..sas7bdat"");";		
	%end;
	put "				jg:DataStandardize_updateDeps;";	
	put "				jg:setCurrentParmValue(1,""DataStandardize"", jg:getParmNum(""DataStandardize"",""OutPath""),""&ClientOutPath"");";

	put "			) ,";		
	put "	";
	
	put "			ButtonBox(""Correlation and Principal Components"",";
	put "				if(host is(Windows),";
	put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/DataCorrelation.jsl"", absolute, windows)),";
	put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/DataCorrelation.jsl"", absolute, POSIX) ));";
	put "	";		
	put "				jg:setDefaultParmValues(2, ""DataCorrelation"");";

	put "				jg:setCurrentParmValue(2,""DataCorrelation"", jg:getParmNum(""DataCorrelation"",""InData""),""&InPath.&InData..sas7bdat"");";
	put "				jg:DataCorrelation_updateDeps;";		
	%if %length(&ExpInData) %then %do;
		put "				jg:setCurrentParmValue(2,""DataCorrelation"", jg:getParmNum(""DataCorrelation"",""ExpInData""),""&ExpInPath.&ExpInData..sas7bdat"");";		
		put "				jg:DataCorrelation_updateDeps;";		
		put "				jg:setCurrentParmValue(2,""DataCorrelation"", jg:getParmNum(""DataCorrelation"",""ColorVars""),""&ColorVars"");";	
		put "				jg:setCurrentParmValue(2,""DataCorrelation"", jg:getParmNum(""DataCorrelation"",""ExpLabelVar""),""&ExpLabelVar"");";
	%end;
	put "				jg:setCurrentParmValue(2,""DataCorrelation"", jg:getParmNum(""DataCorrelation"",""OutPath""),""&ClientOutPath"");";	
	put "			),";
	put "	";
	
	put "			ButtonBox(""Filter Intensities"",";
	put "				if(host is(Windows),";
	put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/FilterIntensities.jsl"", absolute, windows)),";
	put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/FilterIntensities.jsl"", absolute, POSIX) ));";
	put "	";		
	put "				jg:setDefaultParmValues(2, ""FilterIntensities"");";	
	put "				jg:setCurrentParmValue(2,""FilterIntensities"", jg:getParmNum(""FilterIntensities"",""InData""),""&InPath.&InData..sas7bdat"");";
	put "				jg:FilterIntensities_updateDeps;";
	put "				jg:setCurrentParmValue(2,""FilterIntensities"", jg:getParmNum(""FilterIntensities"",""Variables""),";
	/*%if %length(&Vars) < 128 %then %do;
		put "             ""&Vars"" ";
	%end;
	%else %do;
		%let cutlength = 120;
		%let Vars1 = &Vars;
		%do %while(%length(&Vars1));
			%if %length(&Vars1) > &cutlength %then %do;
				%let Vars2 = %qsubstr(&Vars1,1,&cutlength);
				put "             ""&Vars2"" || ";
				%let Vars1 = %qsubstr(&Vars1,%eval(&cutlength+1));
			%end;
			%else %do;
				put "             ""&Vars1"" ";
				%let Vars1 = ;
			%end;
		%end;
	%end;*/
	
	%let _nVarInARow_ = %sysfunc(ceil(&nv/10));
	%let _Remainder_ = %eval(&nV - (&_nVarInARow_-1)*10);
	%do i = 1 %to &_nVarInARow_;
		%if &i < &_nVarInARow_ %then %do;
			%let Vars2 = ;
			%do j = 1 %to 10;
				%let k = %eval((&i-1)*10 + &j);
				%let Vars2 = &Vars2 &&_ColName_&k._;
			%end;
			put "             ""&Vars2"" || ";
		%end;
		%else %do;
			%let Vars2 = ;
			%do j = 1 %to &_Remainder_;
				%let k = %eval((&_nVarInARow_-1)*10 + &j);
				%let Vars2 = &Vars2 &&_ColName_&k._;
			%end;
			put "             ""&Vars2"" ";
		%end;
	%end;
	
	
	
	
	put "				);";	
	put "				jg:setCurrentParmValue(2,""FilterIntensities"", jg:getParmNum(""FilterIntensities"",""ListVars""),""&ListVars"");";		
	%if %length(&ExpInData) %then %do;
		put "				jg:setCurrentParmValue(2,""FilterIntensities"", jg:getParmNum(""FilterIntensities"",""ExpInData""),""&ExpInPath.&ExpInData..sas7bdat"");";		
		put "				jg:FilterIntensities_updateDeps;";	
	%end;
	put "				jg:setCurrentParmValue(2,""FilterIntensities"", jg:getParmNum(""FilterIntensities"",""OutPath""),""&ClientOutPath"");";	
	put "			)";
	put "   )";
	put ");";	
	put "Here:LaunchCode = ddlaunchbuttons();";
	/*
	put "show(DataList,TabButtonList,TabCodeList,TabAfterCodeList,TabOpenList);";
	*/
run;

* create string of data sets to include in package;
%let ds_string = ;
%if (%upcase(&DetailPlots)=YES) or (&ed and %length(&GroupVars)) %then %do;
	%let ds_string = &DataDistribution_OutDataName..sas7bdat;
	%if (&ed) and %length(&GroupVars) %then %do;
		%let ds_string = &ds_string,&DataDistribution_OutGrpExpName..sas7bdat;
	%end;	
%end;
%if %upcase(&Overlay)=YES %then %do;
	%if ^%length(&ds_string) %then %let ds_string = &DataDistribution_OutDenName..sas7bdat;
	%else %let ds_string = &ds_string,&DataDistribution_OutDenName..sas7bdat;
%end;
%if %upcase(&BoxPlots)=YES %then %do;
	%if ^%length(&ds_string) %then %let ds_string = &DataDistribution_OutBoxName..sas7bdat;
	%else %let ds_string = &ds_string,&DataDistribution_OutBoxName..sas7bdat;
%end;
%put ds_string = &ds_string;

%PackageInsertFile(jsl code from &ProcessName, &OutPath, &DataDistributionScriptName, "&ds_string", text/plain, yes);
%AddDatatoStudy(&DataDistribution_OutDenName,_type=tall);

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend DataDistribution;

* run the macro;
%DataDistribution;


