/*
JMP Genomics Analytical Process 

Name: Basic Expression Workflow
Author: Russ Wolfinger, SAS Institute Inc.
History:
	sasrdw 08Jan08 initial coding 

Description: 
		This process runs a basic workflow
		for expression data, including options
		for quality control and BasicExpressionWorkflow.
*/

/*
-----------------------------------------------------------
Copyright (C) 2008 SAS Institute, Inc. All rights reserved. 
 
Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/



*ProcessVariables;

%global StudyName WFOutName InData ByVars LabelVar BaseMergeVars ChrVar PosVar ListVars Variables
AnnoInData AnnoMergeVars AnnoChrVar AnnoPosVar AnnoKeepVars MaxLength 
AnnoOrganism AnnoAccession AnnoLabel AnnoGeneSymbol AnnoDescription AnnoGeneId AnnoSnpId 
ExpInData ColorVars ExpLabelVar GroupVars  LComponents OutDataComp WhereEddmt
HierarchicalClustering TwoWayClustering CenterRows kMeansClustering Method NumberOfClusters CenterRows2 
NumberOfComponents ReplaceMeans PCA CenterCols ScaleCols Include3D CenterRows3 ScaleRows
Dist CorrPCA CorrGroup StandardizationMethod QCTime UDiffs PValues AdjPValues
ClassVars FixedEffects RandomEffects LsmeansEffects EstInFile JournalChrom
ClusterLSM LSMeansDiffSet LSMeansControlValues DiffInData
MultipleTestingMethod Alpha NegLog10pCutoff UseXaxisFilter XaxisCutoff XaxisDirection
TrackPath TrackFilter TrackFiles TrackAdd
OutPath ServerOutPath MacroPath ExonInData;

*ProcessBody;

%macro ExpressionStatisticsWorkflow;

* initialization and argument checking;
%let ProcessName = ExpressionStatisticsWorkflow; 
%global FullAPName;
%let FullAPName = %str(Expression Statistics Workflow);

%include "&MacroPath./UtilityMacros.sas" / nosource;

%let TrackAdd = 0;
/*kemicl make these into new macros to follow one-way anova processing logic*/
%let OrigFixedEffects = &FixedEffects;
%let FixedEffects = %trim(%left(&FixedEffects &LSMeansEffects));

* check methods applied;
%if &Dist=No & &CorrPCA=No & &CorrGroup=No & &StandardizationMethod=NONE &
	^%length(&FixedEffects) & ^%length(&RandomEffects) %then %do;
	%put ERROR: There is no Analysis Process selected.  Please check your settings on the last two tabs.;
	%goto exit;
%end;


%GetOrigFileName(InData,OrigInData);

%CheckInData(_ExpFlag=1);

%if ^%length(&StudyName) %then %do;
	%if %length(&WFOutName) %then %let StudyName=&WFOutName;
	%else %let StudyName=&InData;
%end;

* re-asign variable info;
%if %length(&Variables) | %length(&ListVars) %then %do;
	%CheckInData2(InLib.&InData,_Vars=&Variables,_VarListStyle=&ListVars);
%end;



%if (&InPath ^= &OutPath) %then %do;
	data OutLib.&InData;
		set InLib.&InData;
	run;
%end;

* use a data set to save variable names and labels;
data dsix;
	set OutLib.&InData;
	keep %do i = 1 %to &_nAllColChunk_; &&_AllCol_&i._ %end;;
	if (0);
run;
%let _dsix_  = %sysfunc(open(dsix));
%let nix  = %sysfunc(attrn(&_dsix_, NVARS));


/* For Experimental Design */
%put ExpInPath = &ExpInPath;
%put OutPath = &OutPath;

%if (&ExpInPath ^= &OutPath) & ^%length(&WhereEddmt) %then %do;
	data OutLib.&ExpInData;
		set ExpInLib.&ExpInData;
	run;
%end;
%if (&syserr>4 | &exiterror) %then %goto exit;

%let ezdist = 0;

%let ezpca = 0;

%let ezgroup = 0;
%let StandardizationMethod = NONE;

/*For Pattern Discovery */
/* For Hierarchical Clustering setting*/
%if %upcase(&HierarchicalClustering) = YES %then %let HClustering =1;
%else %let HClustering=0;


/* For k-Means Clustering setting*/
%if %upcase(&kMeansClustering) = YES %then %let KMClustering=1;
%else %let KMClustering=0;

/* For PCA Clustering setting*/
%if %upcase(&PCA) = YES %then %let PCAClustering =1;
%else %let PCAClustering=0;

/* For ANOVA settings*/
%if %length(&FixedEffects) or %length(&RandomEffects) %then %do;
	%let ezanova = 1;
	%let f=1;
	%do %while(%length(%scan(&FixedEffects,&f,' |')));
		%let f = %eval(&f+1);
	%end;
	%let nf = %eval(&f-1);
	%let r=1;
	%do %while(%length(%scan(&RandomEffects,&r,' |')));
		%let r = %eval(&r+1);
	%end;
	%let nr = %eval(&r-1);
	%if (&nf <= 1) and (&nr <= 1) and ^%index(&LSMeansDiffSet,Simple) %then %do;
		%let oneway = 1;
		%if &oneway %then %do;
			%if ^%length(&FixedEffects) %then %do;
				%put ERROR: At least one fixed effect (Model these Fixed Effects) must be specified.;
				%goto exit;
			%end;
		%end;
		* make effects blank separated;
		%let bf = ;
		%if (&nf = 1) %then %do;
			%let f=1;
			%do %while(%length(%scan(&FixedEffects,&f,'*()')));
				%let bf = &bf %scan(&FixedEffects,&f,'*()');
				%let f = %eval(&f+1);
			%end;
		%end;
		%let br = ;
		%if (&nr = 1) %then %do;
			%let r=1;
			%do %while(%length(%scan(&RandomEffects,&r,'*()')));
				%let br = &br %scan(&RandomEffects,&r,'*()');
				%let r = %eval(&r+1);
			%end;
		%end;
	%end;
	%else %do;
		%let oneway = 0;
	%end;
%end;
%else %let ezanova = 0;
%let OutName = &StudyName;
%if ^%length(&Alpha) %then %let Alpha = 0.05;
%if ^%length(&NegLog10pCutoff) %then %let NegLog10pCutoff = 0;

%let _FETmp_ =;
%if %length(&FixedEffects) %then %do;
	%let f=1;
	%do %while(%length(%scan(&FixedEffects,&f,'*()|')));
		%if ^%index(%upcase(&ClassVars),%upcase(%scan(&FixedEffects,&f,'*()|@'))) %then %do;
			%let _FETmp_ = &_FETmp_ %scan(&FixedEffects,&f,'*()|@');
		%end;
		%let f = %eval(&f+1);
	%end;
%end;
%if %length(&_FETmp_) %then %let ClassVars =&ClassVars &_FETmp_;

%let _RETmp_ =;
%if %length(&RandomEffects) %then %do;
	%let r=1;
	%do %while(%length(%scan(&RandomEffects,&r,'*()|')));
		%if ^%index(%upcase(&ClassVars),%upcase(%scan(&RandomEffects,&r,'*()|@'))) %then %do;
			%let _RETmp_ = &_RETmp_ %scan(&RandomEffects,&r,'*()|@');
		%end;
		%let r = %eval(&r+1);
	%end;
%end;
%if %length(&_RETmp_) %then %let ClassVars =&ClassVars &_RETmp_;
/*kemicl reset Fixed Effects */
%let FixedEffects = &OrigFixedEffects;

* create JSL file;
%global BEW_Script;
%PrepOutFile(BEW_Script,BEW_ScriptName,&ProcessName Script for &InData..jsl);

* put JSL statements;
data _null_;
   file "&BEW_Script";
	put "//!";
	put "Names Default to Here(1);";	
	put "if(hostis(windows), _slash_ = ""\"", _slash_ = ""/"");";
	put "if(Host Is(Windows),";
	put "   jmpGenomicsDir = ConvertFilePath(""$LS_HOME"", absolute, windows),";
	put "   jmpGenomicsDir = ConvertFilePath(""$LS_HOME"", absolute, POSIX)";
	put ");";
	put "include(jmpGenomicsDir||""Generator""||_slash_||""WizardHelperFunctions.jsl"");";
	put "include(jmpGenomicsDir||""Generator""||_slash_||""WorkflowHelperFunctions.jsl"");";
	put "jg:_BEWF_ = 1;";
	put "here:EZ_ProcessName = ""&ProcessName"";";
	put "here:EZ_AE_StudyID = ""&StudyName"";";
	put "here:JMPG_WF_Flag =1;";
	put "EZ_AE_CHPIE = 0;";
	put "EZ_AE_CHPFormat = "" "";";
	put "CDFRadio = 0;";
	put "LabelVar = ""&LabelVar"";";
	put "BaseMergeVars = ""&BaseMergeVars"";";
	put "ChrVar = ""&ChrVar"";";
	put "PosVar = ""&PosVar"";";
	put "jg:AnnoRadio = 3;";
	put "jg:AnnoInData = ""&AnnoInData"";";
	put "AnnoMergeVars = ""&AnnoMergeVars"";";
	put "AnnoAccession = ""&AnnoAccession"";";
	put "AnnoLabel = ""&AnnoLabel"";";
	put "jg:AnnoGeneSymbol = ""&AnnoGeneSymbol"";";
	put "jg:AnnoDescription = ""&AnnoDescription"";";
	put "AnnoGeneId = ""&AnnoGeneId"";";
	put "jg:AnnoChrVar = ""&AnnoChrVar"";";
	put "jg:AnnoPosVar = ""&AnnoPosVar"";";
	put "jg:Anno2InData = """";";
	put "LComponents = ""&LComponents"";";
	put "OutDataComp=""&OutDataComp"";";
	put "TrackPath = ""&TrackPath"";";
	put "jg:TrackFiles = ""&TrackFiles"";";
	put "TrackAdd = &TrackAdd;";
	put "ExonInData = ""&ExonInData"";";
	put "jg:EZ_AE_OutPath = ""&OutPath"";";
	put "jg:EZ_AE_ClientOutPath = ""&ClientOutPath"";";
	put "jg:EZ_AE_ServerOutPath = ""&ServerOutPath"";";
	put "jg:EZ_AE_InData = ""&InData"";";
	put "EZ_AE_ExpInData = ""&ExpInData"";";
	put "ColorVars = ""&ColorVars"";";
	put "ExpLabelVar = ""&ExpLabelVar"";";
	put "GroupVars = ""&GroupVars"";";
	put "ListVars = ""&ListVars"";";

	put "jg:EZ_AE_Dist = &ezdist;";
	put "jg:EZ_AE_CorrPCA = &ezpca;";
	put "jg:EZ_AE_CorrGroup = &ezgroup;";
	put "jg:EZ_AE_QC2 = 0;";
	put "jg:EZ_AE_Dist2 = &ezdist;";
	put "jg:EZ_AE_CorrPCA2 = &ezpca;";
	put "jg:EZ_AE_CorrGroup2 = &ezgroup;";	
	
	put "jg:NormMethod = ""&StandardizationMethod"";";
	put "jg:EZ_AE_ANOVA = &ezanova;";
	put "ClassVars = ""&ClassVars"";";
	put "FixedEffects = ""&FixedEffects"";";
	put "RandomEffects = ""&RandomEffects"";";
	put "Variables = ";
		%do i=1 %to &nix;
			%let v=%sysfunc(varlabel(&_dsix_,&i));
			%if ^%length(&v) %then %let v = %sysfunc(varname(&_dsix_,&i));
			%if (&i=1) %then  %do;
				put " ""&v"" "@@;
			%end;
			
			%else %do;
				put " || "" ""  ||"@@;
				put " ""&v"" "@@;
			%end;
							
			%if (&i>&nix) %then %do;
				put "\!N";
			%end;
		%end; 
	put ";";
		
	put "EZ_AE_AP = """";";
	put "jg:EZ_AE_statistics = 0;";
	put "jg:EZ_AE_PCA = &PCAClustering;";
	put "jg:EZ_AE_kmcluster = &KMClustering;";
	put "jg:EZ_AE_hcluster = &HClustering;";
		
	%if (&ezanova = 1) %then %do;
		%if (&oneway = 1) %then %do;
			put "jg:EZ_AE_OneWay = 1;";
			put "FixedEffects = ""&bf"";";
			put "BlockVars = ""&br"";";
		%end;
		%else %do;
			put "jg:EZ_AE_OneWay = 0;";
			put "ByVars = ""&ByVars"";";
			put "DiffInData= ""&DiffInData"";";
		%end;
		put "LSMeansEffects = ""&LSMeansEffects"";";
		put "EZ_AE_MeanDiff = ""&LSMeansDiffSet"";";
		put "ClusterLSM = ""&ClusterLSM"";";
		put "EZ_AE_ControlLevel1 = {};";
		put "Here:UDiffs = ""&UDiffs"";";
		put "Here:PValues = ""&PValues"";";
		put "Here:AdjPValues = ""&AdjPValues"";";
		%let len = %length(&LSMeansControlValues);
		%put LSMeansControlValues = &LSMeansControlValues;
		%put len = &len;
		%if &len %then %do;
			%let LSM = %quote(&LSMeansControlValues);
			%put LSM = &LSM;
			%let LSMCV = ;
			%do i=1 %to &len;
				%let c1 = %qsubstr(&LSM,&i,1);
				%put i=&i c1=&c1;
				%if ^%length(&c1.&c1) %then %do;
					%let LSMCV = &LSMCV.%str( );
				%end;
				%else %if &c1.&c1 = %str(%"%") %then %do;
					%let LSMCV = &LSMCV.\!&c1;
				%end;
				%else %do;
					%let LSMCV = &LSMCV.&c1;
				%end;
			%end;
			%put LSMCV = &LSMCV;
			put "c1 = ""&LSMCV"";";
			put "lc1 = length(c1);";
			put "c2 = substr(c1,2,lc1-2);";
			put "InsertInto(EZ_AE_ControlLevel1,c2);";
		%end;
		put "EZ_AE_ControlLevel2 = {};";
		put "EstInFile = ""&EstInFile"";";
		put "MultiTest = ""&MultipleTestingMethod"";";
		put "UseXaxisFilter = ""&UseXaxisFilter"";";
		put "XaxisCutoff = ""&XaxisCutoff"";";
		put "XaxisDirection = ""&XaxisDirection"";";
		put "EZ_AE_Alpha = ""&Alpha"";";
		put "Cutoff = &NegLog10pCutoff;";
		put "JournalChrom = ""&JournalChrom"";";
	%end;
	
	%if (&HClustering =1) %then %do;
		put "EZ_AE_crows = ""&CenterRows"";";
		put "EZ_ZE_twoway =""&TwoWayClustering"";";
	%end;
	
	%if (&KMClustering =1) %then %do;
		put "EZ_AE_kmethod = ""&Method"";";
		put "EZ_AE_cluster = ""&NumberOfClusters"";";
		put "EZ_AE_crows2 = ""&CenterRows2"";";
		put "EZ_AE_RMean = ""&ReplaceMeans"";";
	%end;
	
	%if (&PCAClustering =1) %then %do;
		put "EZ_AE_ccols = ""&CenterCols"";";
		put "EZ_AE_crows3 = ""&CenterRows3"";";
		put "EZ_AE_scols = ""&ScaleCols"";";
		put "EZ_AE_srows = ""&ScaleRows"";";
		put "EZ_AE_inc3D = ""&Include3D"";";
		put "EZ_AE_NoComp= ""&NumberOfComponents"";";
	%end;
	
	put "EZ_AE_uploadGene = 0;";
	put "_CreateWorkflow;";
	put 'jg:_BEWF_=.;';
	put "here:JMPG_WF_Flag =0;";
run;

%PackageInsertFile(jsl code from &ProcessName, &OutPath, &BEW_ScriptName, , text/plain, yes);

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend ExpressionStatisticsWorkflow;


* call the macro;
%ExpressionStatisticsWorkflow;


