/*
JMP Genomics Analytical Process

Name: Haplotype Estimation
Author: Wendy Czika

Description:
Estimates haplotype frequencies via EM algorithm
using PROC HAPLOTYPE.  Sliding windows 
(optionally overlapping) of markers can be 
examined. If a trait variable is specified, 
a LRT is performed to test for association 
of the trait with the haplotypes across the 
markers, and the p-value from this test is 
plotted across windows of markers.  See the 
HAPLOTYPE procedure in the SAS/Genetics 
User's Guide for more information.  
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData AlleleVars ListMarkerVars MarkerFmt GenoDelim ByVars WhereStmt AnnoInData AnnoLabel AnnoMName AnnoGroup AnnoWindow AnnoLoc AnnoWhereStmt
EstMethod StepTrim InitMethod NStart MaxIter Converge SEMethod CutOff Seed TestAllelicAssociation 
PhaseData OutCut TraitVar IDVars TestIndividualHaplotype Perm OutName WSize WUnit Overlap Conv 
MTC PValAdj Tau Bandwidth Alpha OnlySig CreateDSforHTSNP MostProbHap CreateHTML OutPath ServerOutPath MacroPath HaplotypeEstimation_OutData HaplotypeEstimationScript HaplotypeEstimation_Frame 
HaplotypeEstimation_Body HaplotypeEstimation_Contents;

*ProcessBody;

%macro HaplotypeEstimation;

%include "&MacroPath./GenMacros.sas" / nosource;

%global AnnoAccession;
%let AnnoAccession=;

%global FullAPName;
%let FullAPName=%str(Haplotype Estimation);

%GenInit(HaplotypeEstimation,23);
%if &exiterror=1 %then %goto exit;

%if %length(&ListMarkerVars &AlleleVars)=0 %then %do;
 %put ERROR: Marker variables must be specified in either the Marker Variables or List-Style Specification of Marker Variables parameters.;
 %let exiterror=1;
 %goto exit;
%end;

%let AlleleVars = &AlleleVars &ListMarkerVars;

%if %length(&TraitVar)=0 and %upcase(&TestAllelicAssociation) ne YES and %upcase(&CreateHTML) ne YES and %upcase(&CreateDSforHTSNP) ne YES and %upcase(&PhaseData) ne YES %then %do;
	%put ERROR: You must select results to be created via options on the Association Tests or Output tab.;
	%let exiterror=1;
	%goto exit;
%end;

%PrepOutDS(HaplotypeEstimation_OutData,HaplotypeEstimation_OutDataName,&OutName._hest);
%PrepOutFile(HaplotypeEstimationScript,HaplotypeEstimationScriptName,&ProcessName Script for &OutName..jsl);
%if %upcase(&CreateHTML)=YES %then %do;
	%PrepHTML(HaplotypeEstimation_Body, HaplotypeEstimation_Frame, HaplotypeEstimation_Contents,&OutName._hest_body.html,&OutName._hest_frame.html,&OutName._hest_contents.html);
	ods html close;
%end;

%global ReorderingVars;
%let ReorderingVars = &AnnoGroup &AnnoWindow &AnnoLoc;
%if %length(&AnnoWindow) %then %do;
	%if %length(&AnnoWhereStmt) %then %let AnnoWhereStmt = %quote(&AnnoWhereStmt and missing(&AnnoWindow)=0);
	%else %let AnnoWhereStmt = %quote(missing(&AnnoWindow)=0);
%end;

/* For launching APs */
%let OrigAnnoLabel = &AnnoLabel;
%let OrigAnnoLoc = &AnnoLoc;
%let OrigAnnoWindow = &AnnoWindow;
%let OrigAnnoGroup = &AnnoGroup;

%PrepMapandBy;
run;
%if &exiterror=1 %then %goto exit;

%if &nummark<2 %then %do;
	%put ERROR: Allele or genotype variables for 2 or more markers must be specified.;
	%let exiterror = 1;
	%goto exit;
%end;

data at;
	set &ds;
	keep &AlleleVars;
run;
	
proc transpose data=at out=at;
	var &AlleleVars;
run;

%if &syserr>4 %then %goto exit;

%NObsVars(at);
%let Navars = &nobs;

%do i=1 %to &Navars; 
	%global _tv&i;
%end; 

data _null_;
	set at;
	length is$ 16;
	is = "_av" || strip(_n_);
	call symput(is,strip(_name_));
run;

data _null_; 
	set mapdata;
	by &AnnoGroup;
	if first.&AnnoGroup then do;
		_gp+1;
		_count=1;
		call symput("firstingroup"||strip(_gp),strip(_n_));
	end;
	else _count+1;
	if last.&AnnoGroup then do;
		call symput("HGNum"||strip(_gp),strip(_count));
	end;
	call symput('Locus'||strip(_n_),strip(&AnnoLabel));
	m=vlength(&AnnoLabel);
	call symput('Mlen',m);
run;

* setup for html output;
%if %upcase(&CreateHTML)=YES %then %do;
ods html 
   body = "&HaplotypeEstimation_Body" (url="&bodyfile")
   frame = "&HaplotypeEstimation_Frame"  (url="&framefile")
   contents = "&HaplotypeEstimation_Contents" (url="&contentsfile")
;
%end;
ods listing close;

*Proc haplotype option;
%let HaploOptions = init=&InitMethod cutoff= &CutOff maxiter=&MaxIter se=&SEMethod conv=&Converge nstart=&NStart est=&EstMethod ;
%if (&InitMethod = Random | (&Perm>0) | (&NStart>2)) and %length(&Seed)>0 %then %do;
 %let HaploOptions = &HaploOptions seed=&Seed;
%end;
%if (%upcase(&TestAllelicAssociation)=YES) %then %let HaploOptions = &HaploOptions LD;
%if (%upcase(&EstMethod)=STEPEM) %then %let HaploOptions = %str(&HaploOptions steptrim=&StepTrim);
%if %upcase(&PhaseData)=YES %then %do;
	%let PhaseData=&OutName._hph;
	%let HaploOptions = &HaploOptions out=TmpPhaseData outcut=0;
%end;
%else %let PhaseData=;

%let TraitStmt = ;	
%if %length(&TraitVar) %then %do; 

	/* Check that trait is binary */
	proc sort data=&ds out=isbin nodupkey;
		 by &TraitVar;
		 where missing(&TraitVar)=0 %if %length(&WhereStmt) %then and (&WhereStmt);;
	run;
		
	%NObsVars(isbin);
	%if &nobs ^= 2 %then %do;
		%put ERROR: The trait variable is non-binary.;
		%goto exit;
	%end;
	
	%let dsid=%sysfunc(open(&ds,i));
  	%if (%sysfunc(vartype(&dsid,%sysfunc(varnum(&dsid,&TraitVar)))) = N) %then %do;
   		%let TraitStmt = trait &TraitVar._char;
	 	%let NumTrait=1;
	%end;
	%else %do;
		%let TraitStmt = trait &TraitVar;
		%let NumTrait = 0;
	%end;
	%let rc=%sysfunc(close(&dsid));	

	**start S1575273: removing by groups for which the trait variable has only one level to avoid error in proc haplotype.**;
	%if &NumTrait and %length(&ByVars) %then %do;
		data tmpds;
			set &ds;
			_tmpord_ = _n_;
			%if %length(&WhereStmt) %then %do;
				where &WhereStmt;
			%end;
			keep &TraitVar &ByVars _tmpord_;
		run;
		
		proc sort data= tmpds;
			by 
			%let j=1;
			%do %while(%length(%scan(&ByVars,&j))); 
					%scan(&ByVars,&j)
					%let j=%eval(&j + 1);
			%end;
			;
		run;
		
		proc freq data= tmpds noprint;
			tables &TraitVar * 
			%let j=1;
			%do %while(%length(%scan(&ByVars,&j)));
					%if &j>1 %then %do;
						%str(*)
					%end; 
					%scan(&ByVars,&j)
					%let j=%eval(&j + 1);
			%end;
			/out=fby1
			;		
		run;
		
		data fby1_2;
			set fby1;
			where not missing(&TraitVar);
		run;
		
		proc freq data=fby1_2 noprint;
			tables 
			%let j=1;
			%do %while(%length(%scan(&ByVars,&j)));
				%if &j>1 %then %do;
					%str(*)
				%end; 
				%scan(&ByVars,&j)
				%let j=%eval(&j + 1);
			%end;
			/out=fby2
			;		
		run;
		
		proc sort data=fby2;
			by COUNT;
		run;
		
		proc sql UNDO_POLICY= NONE;
			create table tmpds2 as
			select a.*, b.*
			from 
			tmpds as a left join fby2(drop=PERCENT rename=(count=_count_)) as b
			on 
			%let j=1;
			%do %while(%length(%scan(&ByVars,&j)));
				%if &j>1 %then %do;
					%str(and)
				%end; 
				a.%scan(&ByVars,&j)=b.%scan(&ByVars,&j)
				%let j=%eval(&j + 1);
			%end;
			;
		quit;
		
		data tmpdskeep tmpdsdrop;
			set tmpds2;
			if _count_ > 1 then output tmpdskeep;
			else output tmpdsdrop;
			drop _count_;
		run;
		
		%NObsVars(tmpdsdrop);
		%if &nobs > 0 %then %do;
		
			proc sort data= tmpdskeep nodupkey;
				by 
				%let j=1;
				%do %while(%length(%scan(&ByVars,&j))); 
						%scan(&ByVars,&j)
						%let j=%eval(&j + 1);
				%end;
				;
			run;

			proc sort data= tmpdsdrop nodupkey;
				by 
				%let j=1;
				%do %while(%length(%scan(&ByVars,&j))); 
						%scan(&ByVars,&j)
						%let j=%eval(&j + 1);
				%end;
				;
			run;

			data null;
				set tmpdsdrop end=_eof_;
				%let j=1;
				%do %while(%length(%scan(&ByVars,&j))); 
						_tmp_= "byv"||"&j"||"l";
						call symputx(_tmp_||trim(left(_n_)), trim(left(%scan(&ByVars,&j))));
						%let j=%eval(&j + 1);
				%end;
				if _eof_ then call symputx("nbydrop", trim(left(_n_)));
			run;
			%put nbydrop= &nbydrop;

			%do k=1 %to &nbydrop;
				%let WhereStmtByDrop&k= ;
				%let j=1;
				%do %while(%length(%scan(&ByVars,&j))); 
					/*%put byv&j.l&k= &&byv&j.l&k;*/
					%if &j > 1 %then %do;
					%let WhereStmtByDrop&k= %str(&&WhereStmtByDrop&k and);
					%end;
					%let WhereStmtByDrop&k= &&WhereStmtByDrop&k %scan(&ByVars,&j)=&&byv&j.l&k;
					%let j=%eval(&j + 1);
				%end;
				%put WARNING: Dropping levels of By Variables for which there is only one value of the Binary Trait Variable.;
				%put WhereStmtByDrop&k= &&WhereStmtByDrop&k;	
			%end;

			%if %Index(&ds, &InData) %then %do;

				data tmpds;
					set &ds;
					%do k=1 %to &nbydrop;
						if (&&WhereStmtByDrop&k) then delete; 
					%end;
				run;

				%let ds= tmpds;
			%end;
			%else %do;
				data &ds;
					set &ds;
					%do k=1 %to &nbydrop;
						if (&&WhereStmtByDrop&k) then delete; 
					%end;
				run;
			%end;									

			%NObsVars(&ds);
			%if &nobs = 0 %then %do;
				%put ERROR: Trait &TraitVar cannot be analyzed because it has only one level.;
				%let exiterror=1;
				%goto exit;
			%end;
		%end;
	%end;
	**end S1575273: removing by groups for which the trait variable has only one level to avoid error in proc haplotype.**;
			
	data cc_all;
		delete;
 	run;
 	
 	data htt_all;
 		delete;
 	run;
 	
	%if (&Perm > 0) %then %do;
		%let TraitStmt = &TraitStmt/Perm=&Perm;
		%if %upcase(&TestIndividualHaplotype)=YES %then %let TraitStmt = &TraitStmt testall;
	%end;
   	%else %if %upcase(&TestIndividualHaplotype)=YES %then %let TraitStmt = &TraitStmt / testall;	
%end;
%else %let NumTrait=0;

%if (%upcase(&TestAllelicAssociation)=YES) %then %do;
	data ld_all;
		delete;
	run;
	%let FreqVar = H1Freq;
%end;
%else %let FreqVar=Freq;
  
%let winno=1;   
%let maxhaplen=3;
%let mhaplen=2;

%if %length(&PhaseData) %then %do;
	data OutLib.&PhaseData;
		WindowID=.;
		delete;
  	run;
%end;

%if %upcase(&CreateDSforHTSNP)=YES %then %do;
	data OutLib.&OutName._hfr;
		WindowID=.;
		FirstMarkerLocation=.;
		delete;
	run;
%end;

options label;

%let anygood=0;
%do i = 1 %to &NumGps;
	    
	data map_sub; 
		set mapdata;
		%if %quote(&WUnit)=%str(Location Variable unit) %then %do;
			by &AnnoGroup;
			retain _firstloc;
			if first.&AnnoGroup then do;
				_Window=1;
				_firstloc=&AnnoLoc;
			end;
			else if (&AnnoLoc-_firstloc > &WSize)  then do;
			 _window+1;              
			 _firstloc=&AnnoLoc;
			end;
			%let AnnoWindow=_Window;
			drop _firstloc;
		%end;
		%if (&GpType=C) %then %do;
			where &AnnoGroup="&&Groupn&i";
		%end;
    		%else %do;
     			where &AnnoGroup=&&Groupn&i;
    		%end;
	run;	
	
	%if &NumGps>1 %then %ClockStart(&AnnoGroup &&Groupn&i);
	
	%put HGNum&i= &&HGNum&i; 
	%if &&HGNum&i ge 2 %then %do;
	
		data haps; 
			set &ds;
			%if %length(&WhereStmt) %then %do;
				where &WhereStmt;
			%end;
			%if %length(&GenoCol) %then %do;
				array __g__{%eval(&&HGNum&i)} %if %upcase(&MarkerFmt) ne %str(NUMERIC GENOTYPES) %then $33;;
				array g_orig{*}  &AlleleVars;
				do i=1 to %eval(&&HGNum&i);
					mindex=&&firstingroup&i + i - 1;
					%if %upcase(&MarkerFmt) ne %str(NUMERIC GENOTYPES) %then %do;
						if missing(g_orig{mindex}) then __g__{i}='';
						else __g__{i}=strip(g_orig{mindex});
					%end;
					%else %do;
						if missing(g_orig{mindex}) then __g__{i}=.;
						else __g__{i}=g_orig{mindex};
					%end;					
				end;
				%if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) %then %do;
					format __g__: numtochar.;
				%end;	   
			%end;
			%else %do;
				array __a__{%eval(2*&&HGNum&i)} $;
				array a_orig{*} &AlleleVars;
				do i=1 to %eval(2*&&HGNum&i);
					mindex= 2*&&firstingroup&i + i - 2;
					if missing(a_orig{mindex}) then __a__{i}='';
					else __a__{i}=strip(a_orig{mindex});
				end;		
			%end;			
			%if &NumTrait=1 %then %do;
				if missing(&TraitVar) then &TraitVar._char="";
				else &TraitVar._char=compress(&TraitVar);
			%end;
			keep __g__: __a__: &ByVars &IDVars %if %length(&TraitVar) %then &TraitVar:;;
		run;
		
		/*data OutLib.hap_haps; set haps; run;*/
		/*data OutLib.hap_map_sub; set map_sub; run;*/

		%if %length(&AnnoWindow)=0 %then %do;
			%if (&WSize=0 or %length(&WSize)=0) %then %let nWsize=%eval(&&HGNum&i);
			%else %let nWsize=%sysfunc(min(&WSize,&&HGNum&i));			
			%let inc=%sysfunc(max(%eval(2*&nWSize-2*%sysfunc(min(&Overlap,&nWsize-1))),1));
		%end;
		%else %do;
			%let inc=1;
			%let wingroup = 1;
			data map_sub_win;
				set map_sub;
				by &AnnoWindow notsorted;
				retain start;
				if first.&AnnoWindow then do;
					start=_n_;
					firstgroup+1;
				end;
				if last.&AnnoWindow then do; 
					end=_n_;
					ws=end-start+1;
					if firstgroup=&wingroup then do;
						if ws>1 then call symput("nWSize",strip(ws));
						else call symput("nWSize",strip(1));
					end; 
				end;
			run;
		%end;
		%let la=%eval(2*&&HGNum&i-1);	
		
		%do firsta=1 %to &la %by &inc; 
			%let lasta=%sysfunc(Min(%eval(&firsta+2*&nWSize-1),2*&&HGNum&i));
			%let firstm=%eval((&firsta+1)/2);
			%let lastm=%sysfunc(Min(%eval(&lasta/2),&&HGNum&i));
			%if &firstm=&lastm %then %goto nextwindow;
			
			%let _fm=%sysevalf(&&firstingroup&i+&firstm-1);
			%let _lm=%sysevalf(&&firstingroup&i+&lastm-1);
		
			title "Markers &&Locus&_fm through &&Locus&_lm";
				 
    			data _null_;
    				set map_sub(firstobs=&firstm obs=&firstm);
    				call symput("winloc",&AnnoLoc);
    			run;
    		
    			%if %length(&TraitStmt) %then %do; 
				ods output CCTest=cct %if %upcase(&TestIndividualHaplotype)=YES %then haptraittest=htt;;
				%if %upcase(&CreateHTML)=YES %then ods html select all;;
   			%end;
	
      			%if %upcase(&CreateDSforHTSNP)=YES %then %do;
      	 			ods output HaplotypeFreq=hapfreqs AnalysisInfo=ai;
    			%end;
    		
    			%if (%upcase(&TestAllelicAssociation)=YES) %then %do;
    				ods output LDTest=ldtest;
    			%end;
    	
			proc haplotype data=haps ndata=map_sub(firstobs=&firstm obs=&lastm rename=(&AnnoLabel=name)) 
	   	           	&HaploOptions &GenoCol outid;
				%if %length(&ByVars) %then %do; 
					by &ByVars;
				%end;
				%if %length(&GenoCol) %then %do;
					var __g__&firstm - __g__&lastm;
				%end;
				%else %do;
					var __a__&firsta - __a__&lasta;
				%end;
				%if %length(&TraitStmt) %then %do; 
					&TraitStmt;
				%end;
				%if %length(&IDVars) %then %do; 
					ID &IDVars;
				%end;
  			run;
  			
 			%if &syserr <= 4 and (%length(&TraitStmt)=0 or %sysfunc(exist(cct))) %then %let anygood=1;
  			/* attempt at S0557976:
  			%else %if &syserr=4 and %index(%quote(&syswarningtext),%str(Output 'CCTest' was not created))>0 %then %do;
  				%let TraitStmt=;
  				%put The Trait Variable will be ignored.  See the log for details.;
  				%let anygood=1;
  			%end;
  			*/
  			%else %goto nextwindow;
  			
  			/*data OutLib.hap_cct; set cct; run;*/
  			/*data OutLib.hap_ldt; set ldtest; run;*/
  			/*data OutLib.hap_htt; set htt; run;*/
			
			/* Number of loci in haplotype (for current WindowID) */
			%let haplen = %eval(&lastm - &firstm +1);
      			%if &haplen > &mhaplen %then %let mhaplen = &haplen;
			
			%if %upcase(&CreateDSforHTSNP)=YES %then %do;
			    
				data _null_;
				 	set ai;
					if label1="Loci Used";
					call symput('names',cValue1);
				run;
				
				%if &syserr > 4 %then %goto nextwindow;
				
				%Nobsvars(hapfreqs);
				%if &nobs>0 %then %do;
				/* in case freq cutoff left none */
				
					data _null_;
						set hapfreqs;
						call symput("newmaxhaplen",vlength(haplotype));
					run;

					%if &newmaxhaplen > &maxhaplen %then %let maxhaplen = &newmaxhaplen;

					data _null_; 
						set hapfreqs;
						retain maxl 1;
						i=1;
						array a{&haplen} $ &maxhaplen;
						do i = 1 to &haplen; 
							a{i} = scan(haplotype, i, '-'); 
							maxl=max(maxl,lengthn(a{i}));
						end; 
						call symput('alen',strip(maxl));
					run;    

					data OutLib.&OutName._hfr; 
						length Markers $ %eval(2*&Mlen+2) WindowID 8 FirstMarkerLocation 8 Number 8 Haplotype $ &maxhaplen;
						set OutLib.&OutName._hfr hapfreqs(in=new); 
						array _Allele{&mhaplen} $ &alen; 
						if new then do;
							WindowID=&winno;
							Markers = "&&Locus&_fm--&&Locus&_lm";
							FirstMarkerLocation=&winloc;
							do i = 1 to &haplen; 
								_Allele{i} = scan(haplotype, i, '-'); 
							end;
						end; 
						drop i; 
					run;
				%end;
			%end;

      			%if %length(&PhaseData) %then %do; 
      
				%if %length(&GenoCol) %then %do;
					%let f=&firstm;
					%let l=&lastm;
					%let var=g;
					%let uvar=G;
				%end;
				%else %do;
					%let f=&firsta;
					%let l=&lasta;
					%let var=a;
					%let uvar=A;
				%end;
				
				%if %upcase(&CreateDSforHTSNP) ne YES %then %do;
				/*maxhaplen not defined */
					data _null_;
						set TmpPhaseData;
						_len1=vlength(haplotype1);
						_len2=vlength(haplotype2);
						_maxlen=max(_len1,_len2);
						call symput("newmaxhaplen",_maxlen);
					run;

					%if &newmaxhaplen > &maxhaplen %then %let maxhaplen = &newmaxhaplen;	
				%end;
				
				/*data OutLib.hap_TmpPhaseData&i; set TmpPhaseData; run;*/
				/*data OutLib.hap_PhaseDataTmp&i; set OutLib.&PhaseData; run;*/
					
				data OutLib.&PhaseData;
        				length Markers $ %eval(2*&Mlen+2) WindowID IndivID 8 Haplotype1 Haplotype2 $&maxhaplen Prob 8;
          				set OutLib.&PhaseData TmpPhaseData(in=new rename=(_id_= IndivID %do a=&f %to &l; __&var.__&a =_&uvar._%eval(&a-&f+1) %end;));
          				if new then do;
   	    					WindowID=&winno;
   	    					Markers = "&&Locus&_fm--&&Locus&_lm";
          				end;
					%if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) %then %do; 
						format _g_: ;
					%end;
					%if %length(&OutCut) %then %do;
						if Prob >= &OutCut;
					%end;
        			run;
        		           
			%end;
			
			%if %length(&TraitStmt) %then %do; 
				data cc_all; 
					set cc_all(in=old) cct(in=new);
					length Markers $ %eval(2*&Mlen+2) %if (&GpType=C) %then &AnnoGroup $&GpLen;;
          				if new then do;
						Markers = "&&Locus&_fm--&&Locus&_lm";
						WindowID=&winno;   
						FirstMarkerLocation=&winloc;
            					%if (&GpType=C) %then %do;
					 		&AnnoGroup="&&Groupn&i";
            					%end;
            					%else %do;
              						&AnnoGroup=&&Groupn&i;
            					%end;						
            					dfl=lag(df);
          					dfll=lag2(df);
          					dfHapCC = dfl + dfll - df;
					end;					
					if old or TraitVal='Combined';
					drop df:;
				run;
				
				%if %upcase(&TestIndividualHaplotype)=YES %then %do;
					data htt_all; 
						length %if (&GpType=C) %then &AnnoGroup $&GpLen; %else &AnnoGroup 8; Markers $ %eval(2*&Mlen+2) WindowID 8 ;
						set htt_all(in=old) htt(in=new);
						if new then do;
							Markers = "&&Locus&_fm--&&Locus&_lm";
							WindowID=&winno;   
							%if (&GpType=C) %then %do;
								&AnnoGroup="&&Groupn&i";
							%end;
							%else %do;
								&AnnoGroup=&&Groupn&i;
							%end;						
						end;					
						if old or CombinedFreq > &Cutoff;
					run;				
				
				%end;
				
				%if %upcase(&CreateHTML)=YES %then ods html exclude all;;
				ods output clear;
				proc datasets nolist;
					delete cct %if %upcase(&TestIndividualHaplotype)=YES %then htt;;
				run;
				quit;				
				
			%end;
			
			%if (%upcase(&TestAllelicAssociation)=YES) %then %do;
				data ld_all;
					length Markers $ %eval(2*&Mlen+2) %if (&GpType=C) %then &AnnoGroup $&GpLen;;
					set ld_all(in=old) ldtest(in=new);
          				if new then do;
						Markers = "&&Locus&_fm--&&Locus&_lm";
						WindowID=&winno;
						FirstMarkerLocation=&winloc;
            					%if (&GpType=C) %then %do;
					 		&AnnoGroup="&&Groupn&i";
            					%end;
            					%else %do;
              						&AnnoGroup=&&Groupn&i;
            					%end;          	
            					dfl=lag(df);
          					dfLD = df - dfl;
					end;					
					if old or Hypothesis='H1: Allelic Associations';
					drop df:;
				run;					
			%end;

			%let winno=%eval(&winno+1);
			%nextwindow:;			
			%if %length(&AnnoWindow)>0 %then %do;
				/* 1 gets added to firsta in do loop with inc */
				%let firsta = %eval(&lasta);
			
				%let wingroup = %eval(&wingroup+1);
				data _null_;
					set map_sub_win;
					by &AnnoWindow;
					if last.&AnnoWindow and firstgroup=&wingroup then do;
						if ws>1 then call symput("nWSize",strip(ws));
						else call symput("nWSize",strip(1));
					end; 
				run;
			%end;
			
		%end;
	%end; 
	%endloop:;
	%if &NumGps>1 %then %ClockStop(&AnnoGroup &&Groupn&i);		
%end;

%if %upcase(&MostProbHap)=YES %then %do;
	%if %length(&Seed)=0 %then %let Seed=0;
	
	data OutLib.&PhaseData;
		set OutLib.&PhaseData;
		/* _rand_ to break ties randomly*/
		_rand_ = ranuni(&seed);
	run;
	
	proc sort data=OutLib.&PhaseData;
		by WindowID IndivID descending Prob _rand_;
	run;
	
	data OutLib.&PhaseData;
		set OutLib.&PhaseData;
		by WindowID IndivID;
		if first.IndivID;
		drop _rand_;
	run;
%end;
	

%if &anygood=0 %then %do;
 %put ERROR: There were no successful runs of PROC HAPLOTYPE.  The process will exit without creating any results.  See the SAS log for more details.;
 %goto exit;
%end;

/* Need to define this for when psmooth isn't called; want to output everything but the p-value stuff if error in psmooth */
%let psmsyserr = 0;

%if %length(&TraitStmt) %then %do; 
	%NObsVars(cc_all);
	%if &nobs>0 %then %do;
		options nolabel;
		
		%if &Perm > 0 %then %let PStat=Exact;
		%else %let PStat=ChiSq;
		
		proc sort data=cc_all 
			out=/*&HaplotypeEstimation_OutData*/cc_all(
				drop=Trait: 
				rename=(Prob&PStat=ProbHapCC ChiSq=ChiSqHapCC)
			);
			by &ByVars &AnnoGroup WindowID;
		run; 
	
		%let AnnoLabel=Markers;
		%let AnnoLoc=FirstMarkerLocation;
		
		%if (%upcase(&TestAllelicAssociation)=YES) %then %do;
			proc sort data=ld_all out=ld_all(drop=Hypothesis rename=(ProbChiSq=ProbLD ChiSq=ChiSqLD));
				by &ByVars &AnnoGroup WindowID;
			run; 			
			
			/*
			data &HaplotypeEstimation_OutData;
				merge &HaplotypeEstimation_OutData ld_all;
				by &ByVars &AnnoGroup WindowID;
			run;
			
			%let Test=HapCC LD;
			*/
		%end;
		/*
		%else %do;
			%let Test=HapCC;
		%end;
		*/
		
		/*
		%Psmooth(&Test, &HaplotypeEstimation_OutData, &HaplotypeEstimation_OutDataName);	
		%let psmsyserr = &syserr;
		%if (&psmsyserr>4) %then %goto exit;
		*/

		%if %upcase(&TestIndividualHaplotype)=YES %then %do;
			proc sort data=htt_all out=/*OutLib.&OutName._hta*/htt_all(drop=number rename=(ProbChiSq=ProbIHapCC ChiSq=ChiSqIHapCC));
				by &ByVars &AnnoGroup WindowID;
			run; 		
		%end;
	%end;
	%else %goto exit;
%end;
%else %do;
	%if (%upcase(&TestAllelicAssociation)=YES) %then %do;
		options nolabel;
		proc sort data=ld_all out=/*&HaplotypeEstimation_OutData*/ld_all(drop=Hypothesis rename=(ProbChiSq=ProbLD ChiSq=ChiSqLD));
			by &ByVars &AnnoGroup WindowID;
		run; 			
		%let AnnoLabel=Markers;
		%let AnnoLoc=FirstMarkerLocation;
		/*
		%let Test=LD;
		%Psmooth(&Test, &HaplotypeEstimation_OutData, &HaplotypeEstimation_OutDataName);
		%let psmsyserr = &syserr;
		%if (&psmsyserr>4) %then %goto exit;
		*/
	%end;
%end;

%if %upcase(&CreateHTML)=YES %then ods html close;;
ods listing;

%if %upcase(&CreateHTML)=YES %then %PackageInsertHtml(html from &ProcessName, &OutPath, &bodyfile, &framefile, &contentsfile);
		
%if %length(&PhaseData) %then %do;
	%PackageInsertDataset(Phase assignment data from &ProcessName, OutLib, &PhaseData);
%end;
%if %upcase(&CreateDSforHTSNP)=YES %then %do;
	 %PackageInsertDataset(Haplotype frequency estimates data from &ProcessName, OutLib, &OutName._hfr,_type=other);
%end;
		
%if %length(&TraitStmt) or (%upcase(&TestAllelicAssociation) eq YES) %then %do; 
	%let OrigConv= &Conv; 
	%let OrigMTC= &MTC;
	%let OrigPValAdj= &PValAdj;
	%let OrigTau= &Tau;
	%let OrigBandwidth = &Bandwidth;

	%if %length(&TraitStmt) %then %do;
		%let Test= HapCC;
		/*%Psmooth(&Test, &HaplotypeEstimation_OutData, &HaplotypeEstimation_OutDataName);*/
		%Psmooth(HapCC, cc_all, ,_lsp=0);
		%let psmsyserr = &syserr;
		%if (&psmsyserr>4) %then %goto exit;
		%let HapCCAxisMax= &AxisMax;
		
		proc datasets nolist;
		   	change cc_all=&HaplotypeEstimation_OutDataName;
			run;
			copy in=work out=OutLib memtype=data move;
		   	select &HaplotypeEstimation_OutDataName;
			run;
		quit;			
		
		%*let AxisMax= &HapCCAxisMax;
		%JSLPvalMap(Haplotype, &HaplotypeEstimationScript, &HaplotypeEstimationScriptName, &HaplotypeEstimation_OutDataName);	
		
		%if %upcase(&TestIndividualHaplotype)=YES %then %do;
			%let Conv= &OrigConv; 
			%let MTC= &OrigMTC;
			%let PValAdj= &OrigPValAdj;
			%let Tau= &OrigTau;
			%let Bandwidth = &OrigBandwidth;		
		
			%let Test= IHapCC;
			%Psmooth(&Test, htt_all, );	
			%let psmsyserr = &syserr;
			%if (&psmsyserr>4) %then %goto exit;
			
			data OutLib.&OutName._hta;
				set htt_all;
			run;
			
			proc datasets nolist;
			   	change htt_all=&OutName._hta;
			run;
				copy in=work out=OutLib memtype=data move;
			   	select &OutName._hta;
			run;
			quit;		
			
		 	%PackageInsertDataset(Haplotype-trait association data from &ProcessName, OutLib, &OutName._hta,_type=other);

		%end;
	%end;

	%if (%upcase(&TestAllelicAssociation) eq YES) %then %do;
		/*These parameters may be changed in previous Psmooth calls*/
		%let Conv= &OrigConv; 
		%let MTC= &OrigMTC;
		%let PValAdj= &OrigPValAdj;
		%let Tau= &OrigTau;
		%let Bandwidth = &OrigBandwidth;
		
		/*%Psmooth(LD, &HaplotypeEstimation_OutData, &HaplotypeEstimation_OutDataName);	*/
		%let Test= LD;
		%Psmooth(&Test, ld_all, ,_lsp=0);
		%let psmsyserr = &syserr;
		%if (&psmsyserr>4) %then %goto exit;
		%let LDAxisMax= &AxisMax;
		%let PvalAdj= &OrigPvalAdj; /*PvalAdj is changed in the Psmooth macro*/
		
		**variables needed for plotting LD results**;
		%let dsid=%sysfunc(open(ld_all,I));
		%do p=1 %to &NPvars;
			%let CurrPVar=%scan(&PVars,&p);
			%let vn=%sysfunc(varnum(&dsid,&CurrPVar));
			%let PVarLab=%sysfunc(varlabel(&dsid,&vn));
			%if %length(&PVarLab)=0 %then %let PVarLab=&CurrPVar;
			%let PVar&p=%quote(&PVarLab);
		%end;
		%let rc=%sysfunc(close(&dsid));	
			
		%put AxisMax= &AxisMax;
		%let AxisMax = %sysevalf(&AxisMax + &AxisMax*0.005); 
		%put AxisMax= &AxisMax;		
		%let AxisMin= -0.05;
		%let _pvalplots=1;

		/*MTC correction*/
		%let currSubPVar = &PVar1;
		%let mtcmethod= None;
		%if %sysfunc(find(&PVar1, _Bon)) %then %do;
			%let mtcmethod= %str(Bonferroni);
			%let currSubPVar = %sysfunc(tranwrd(&PVar1, _Bon, ));
		%end;
		%else %if %sysfunc(find(&PVar1, _FDR)) %then %do;
			%let mtcmethod= %str(FDR);
			%let currSubPVar = %sysfunc(tranwrd(&PVar1, _FDR, ));
		%end;
		%else %if %sysfunc(find(&PVar1, _Sid)) %then %do;
			%let mtcmethod= %str(Sidak);
			%let currSubPVar = %sysfunc(tranwrd(&PVar1, _Sid, ));
		%end;	

		/*p-value adjustment*/					
		%let adjmethod= None;
		%let adjbandth= ;	
		%let adjtau= ;
		%if %sysfunc(find(&currSubPVar, _F)) %then %do;

			%let bg= %sysfunc(find(&currSubPVar, _F));
			%let ed= %sysfunc(find(&currSubPVar, %str(_), &bg+1));
			%if &ed %then %do;
				%let bwth= %substr(&currSubPVar, &bg + 2, &ed - &bg - 2);
			%end;
			%else %do;
				%let bwth= %substr(&currSubPVar, &bg + 2);
			%end;
			%let adjbandth= %str(Bandwith= &bwth);
			%let adjmethod= %str(Fisher / &adjbandth);
		%end;
		%else %if %sysfunc(find(&currSubPVar, _S)) %then %do;

			%let bg= %sysfunc(find(&currSubPVar, _S));
			%let ed= %sysfunc(find(&currSubPVar, %str(_), &bg+1));
			%if &ed %then %do;
				%let bwth= %substr(&currSubPVar, &bg + 2, &ed - &bg - 2);
			%end;
			%else %do;
				%let bwth= %substr(&currSubPVar, &bg + 2);
			%end;
			%let adjbandth= %str(Bandwith= &bwth);					
			%let adjmethod= %str(Simes / &adjbandth);
		%end;
		%else %if %sysfunc(find(&currSubPVar, _T)) %then %do;

			%let bg= %sysfunc(find(&currSubPVar, _T));
			%let ed= %sysfunc(find(&currSubPVar, %str(_), &bg+1));
			%if &ed %then %do;
				%let bwth= %substr(&currSubPVar, &bg + 2, &ed - &bg - 2);
			%end;
			%else %do;
				%let bwth= %substr(&currSubPVar, &bg + 2);
			%end;
			%let adjbandth= %str(Bandwith= &bwth);
			%if %symexist(Tau) %then %do;
				%let adjmethod= %str(TPM / &adjbandth / Tau= &Tau);	
			%end;
			%else %do;
				%let adjmethod= %str(TPM / &adjbandth);	
			%end;
		%end;

	
		%if %length(&TraitStmt) %then %do;
			data &HaplotypeEstimation_OutData;
				merge &HaplotypeEstimation_OutData ld_all;
				by &ByVars &AnnoGroup WindowID;
			run;
		%end;
		%else %do;
			proc datasets nolist;
			   	change ld_all=&HaplotypeEstimation_OutDataName;
				run;
				copy in=work out=OutLib memtype=data move;
			   	select &HaplotypeEstimation_OutDataName;
				run;
			quit;		
		%end;
	%end;
%end;

/* Buttons to launch other Haplotype APs */
%if %length(&PhaseData) or %upcase(&CreateDSforHTSNP)=YES or (%upcase(&TestAllelicAssociation) eq YES) %then %do;

	* create JSL file;
	filename jslbtns "&HaplotypeEstimationScript";

	data _null_;
		%if %length(&TraitStmt)=0 %then %do;
			/* main JSL script DOES NOT exist*/
			%global _mainjslfile_;
			%let _mainjslfile_ = %quote(&HaplotypeEstimationScript);
			file jslbtns lrecl=32767;
			put "//!";
			put "Names Default to Here(1);";
			put "wait(0.5);";
			
			%if (%upcase(&TestAllelicAssociation) eq YES) %then %do;
				put "Here:DataList = {};";
				put "Here:TabButtonList = {};";
				put "Here:TabCodeList = {};";
				put "Here:TabAfterCodeList = {};";
				put "Here:TabOpenList = {};";
			%end;
		%end;
		%else %do;
			/* main JSL script DOES ALREADY exist */
			file jslbtns lrecl=32767 mod;				
		%end;
			
		%if (%upcase(&TestAllelicAssociation) eq YES) %then %do;

			/*ploting p-values for the test allelic association for LD*/
			put " ";
			put "////////// Haplotype association LD plot";
			put "InsertInto(DataList,{{""Here:lddata""}});";
			put "InsertInto(Here:TabButtonList,""LD P-Value Plot"");";
			put "jg:drill_down = 0;";
			put "Here:LDpValue= function({},{default local},";
			put "if(isempty(Here:lddata),";
			put "if(jg:drill_down == 0,";
			put " Here:lddata=open(""&ClientOutPath.&HaplotypeEstimation_OutDataName..sas7bdat"", invisible);";
			put ");";
			%if %upcase(&OnlySig)=YES %then %do;
				%if %length(&Conv)=0 %then %let gtlt=%str(>);
				%else %let gtlt=%str(<);
				put "Here:lddata << select where(" @;
				/*
				put ":Name(""&PVar1"") &gtlt &r" @;
				%do p=2 %to &NPVars;
					put "& :Name(""&&PVar&p"") &gtlt &r" @;
				%end;
				*/
				put ")<<hide;";
			%end;  
			put "colnames = Here:lddata << GetColumnNames(String);";
			put 'if(Contains(colnames,"_excluded")>0,';
			put "Here:lddata << select where(:_excluded==1)<<hide);";
			put "Here:lddata << exclude;";
			put "nhidden=0;";
			put "for each row(";
			put "if(hidden(), nhidden=nhidden+1));";
			put "if(nhidden==nrow(Here:lddata),";
			put " throw(""All p-values non-significant in this data set.""));";
			put "Here:lddata << Clear Select;";
			%if %length(&AnnoAccession) %then %do;
				put "column(""&AnnoAccession"")<<data type(character);";
			%end;
			%if %length(&AnnoGroup) %then %do;
				put "column(""&AnnoGroup"")<<data type(character);";
			%end;
			put "column(""&AnnoLoc"")<<data type(numeric);";
			put "column(""&AnnoLabel"") << label(1);";			
			%if &_pvalplots %then %do;
				put "column(""CumOrder"") << modeling type(nominal);";
			%end;			
			%if &ProcessName=%str(HaplotypeEstimation) %then put "column(""WindowID"")<<label(1);";;	
			put ", //else";
			put "  currentdatatable(Here:lddata);";
			%if &_pvalplots %then %do;
				put "column(""CumOrder"") << modeling type(nominal);";
			%end;
			put "); //end isempty(lddata)";


			put "Data Filter Context Box(Hlistbox(";
			put "Here:ldvlb=Vlistbox(";
			put "Here:ld= Graph Builder(";
			put "	Show Control Panel( 0 ),";
			put "	Show Legend( 0 ),";
			put "	Graph Spacing( 0 ),";
			put "	Spacing Borders( 1 ),";
			put "	Variables(";
			%if %length(&AnnoGroup) and &NumGps>1 and &AnnoGroup ^= "_group_" %then %do;
				put "		X( ""&AnnoGroup"" ),";
				put "		X( :CumOrder, Position( 1 ) ),";
				put "		Color( ""&AnnoGroup"" ),";
			%end;
			%else %do;
				put "		X( :CumOrder ),";
			%end;
			%do p=1 %to &NPVars;
				put "		Y(Column(""&&PVar&p"")),";
			%end;
			/*
			%if %length(&ByVars)>0 %then %do;
				put "             Page(&JBy),";									
			%end;
			*/
			put "	),";
			%if %length(&ByVars)>0 %then %do;
				put "            By(&JBy),";						
			%end;					
			%do p=1 %to &NPVars;
				put "      Elements(Position(1,&p), Points( X( 1 )" @;
				%if %length(&AnnoGroup) and &NumGps>1 and &AnnoGroup ^= "_group_" %then %do;
					put ", X( 2 )" @;
				%end;
				put ", Y, Legend( &p + 6 ) ) ),";			
			%end;
			/*
			%if &NumGps > 1 %then %do;
				put "      Local Data Filter(""on right"",";
				put "         Add Filter( Columns(:Name(""&AnnoGroup"")), Display(:Name(""&AnnoGroup""), Size(210, min(400,&NumGps*20)), List Display)),";
				put "		    Mode( Show(1), Include(1) )";
				put "      ),";	
			%end;			
			*/
			put "	SendToReport(";			
			%if %length(&AnnoGroup) and &NumGps>1 and &AnnoGroup ^= "_group_" %then %do;
				put "			Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""&AnnoGroup"" )} ),";
			%end;
			%else %do;
				put "			Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""CumOrder"" )} ),";
			%end;
			put "			Dispatch( {}, ""graph title"", TextEditBox, {Set Text( """" )} ),";
			put "			Dispatch( {}, ""400"", LegendBox, {Sides( ""Left"" ), Set Wrap( 10 )} ),";			
			%do p=1 %to &NPVars;
				%if &p=1 %then %let k=;
				%else %let k= %sysevalf(&p - 1); 

				%if %sysfunc(find(&&PVar&p, LD)) %then %do;
					%let ytest= %str(Allelic LD Test);
					%let testflag= 1;
				%end;				
				%else %do;
					%let ytest= &&PVar&p;
					%let testflag= 0;
				%end;			

				/*MTC Correction*/
				%let ymtc=;
				/*
				%let currSubPVar = &&PVar&p;
				%if %sysfunc(find(&&PVar&p, _Bon)) %then %do;
					%let ymtc= %str(Bonferroni Corrected);
					%let currSubPVar = %sysfunc(tranwrd(&&PVar&p, _Bon, ));
				%end;
				%else %if %sysfunc(find(&&PVar&p, _FDR)) %then %do;
					%let ymtc= %str(FDR Corrected);
					%let currSubPVar = %sysfunc(tranwrd(&&PVar&p, _FDR, ));
				%end;
				%else %if %sysfunc(find(&&PVar&p, _Sid)) %then %do;
					%let ymtc= %str(Sidak Corrected);
					%let currSubPVar = %sysfunc(tranwrd(&&PVar&p, _Sid, ));
				%end;
				*/
				
				/*p-value adjustment*/						
				%let yadjs= ;
				%let ybwth= ;
				/*			
				%if %sysfunc(find(&currSubPVar, _F)) %then %do;

					%let bg= %sysfunc(find(&currSubPVar, _F));
					%let ed= %sysfunc(find(&currSubPVar, %str(_), &bg+1));
					%if &ed %then %do;
						%let bwth= %substr(&currSubPVar, &bg + 2, &ed - &bg - 2);
					%end;
					%else %do;
						%let bwth= %substr(&currSubPVar, &bg + 2);
					%end;
					%let ybwth= %str(Bandwith &bwth);
					%let yadjs= %str(Fisher &ybwth Adj);
				%end;
				%else %if %sysfunc(find(&currSubPVar, _S)) %then %do;

					%let bg= %sysfunc(find(&currSubPVar, _S));
					%let ed= %sysfunc(find(&currSubPVar, %str(_), &bg+1));
					%if &ed %then %do;
						%let bwth= %substr(&currSubPVar, &bg + 2, &ed - &bg - 2);
					%end;
					%else %do;
						%let bwth= %substr(&currSubPVar, &bg + 2);
					%end;
					%let ybwth= %str(Bandwith &bwth);					
					%let yadjs= %str(Simes &ybwth Adj);
				%end;
				%else %if %sysfunc(find(&currSubPVar, _T)) %then %do;

					%let bg= %sysfunc(find(&currSubPVar, _T));
					%let ed= %sysfunc(find(&currSubPVar, %str(_), &bg+1));
					%if &ed %then %do;
						%let bwth= %substr(&currSubPVar, &bg + 2, &ed - &bg - 2);
					%end;
					%else %do;
						%let bwth= %substr(&currSubPVar, &bg + 2);
					%end;
					%let ybwth= %str(Bandwith &bwth);
					%let yadjs= %str(TPM &ybwth Adj);					
				%end;

				/*Conversion*/
				%let yconv=;
				%if %sysfunc(find(%upcase(&&PVar&p), NEGLOG10)) %then %do;
					%let yconv= %str(Negative Log10 P-Value );
					%let ymin= -.1;
				%end;
				%else %if %sysfunc(find(%upcase(&&PVar&p), NEGLOG)) %then %do; 
					%let yconv= %str(Negative Log P-Value );
					%let ymin= -.1;
				%end;
				%else %do;
					%let yconv= %str(P-Value );
					%let ymin= -.01;
				%end;				

				put "			Dispatch(";
				put "			{},";
				put "			""Y &k title"",";
				put "			TextEditBox,";
				put "			{Set Text("" ";
				%if &testflag %then %do;
					put "			&ytest ";
					%if %length(&ymtc) %then %do;
						put "			&ymtc ";
					%end;
					%if %length(&yadjs) %then %do;
						put "			&yadjs ";
					%end;
					%if %length(&yconv) %then %do;
						put "			&yconv ";
					%end;					
				%end;
				%else %do;
					put "			&ytest ";
				%end;
				put "			 "")}";
				put "			),";					
			%end;	
			put "			Dispatch( {}, ""Graph Builder"", FrameBox, {Marker Size( 3 )} ),";
			put "	)";
			put "),";
			%if (&GpType = N) %then %do;
				put "column(""&AnnoGroup"")<<data type(character);";
			%end;			
			put "),";
			%if %length(&AnnoGroup) and &NumGps>1 and &AnnoGroup ^= "_group_" %then %do;
				put "Here:lddata<< Data Filter(";
				put " Conditional(1),";
				%if %length(&ByVars)=0 %then %do;
					put " Local,";			
				%end;
				put " Add Filter(Columns(""&AnnoGroup"")), Display(:Name(""&AnnoGroup""), Size(210, min(400,&NumGps*20)), List Display), Mode(Show(1), Include(1))";
				put ");";
			%end;
			put "),";
			put "), //end Data Filter Context Box";	
			/*S1149704 sbrlsi*/ 
			/*%if &tNumGps>1 %then %do;	
				put "Here:lddata << Color by Column(";
				put "&AnnoGroup, Color( 1 ),	Marker( 0 ),Continuous Scale( 0 ))";
			%end;*/

			put "); //end LDpValue function";

			put "////////LDpValue_after ";
			put "Here:LDpValue_after= Function({},{default local},";
			put "Here:dt = Today();";
			put "Eval(Parse(EvalInsert(""\[ldr^dt^ = Here:ld << Report;]\"")));";			
			put "agsum = Here:lddata << Summary( Group(:&AnnoGroup), Max(:CumOrder), invisible);"; 
			put "Wait( 0.5 );";
			put "Current Data Table( agsum );";
			put "agsum << Sort(By(:Name(""Max(CumOrder)"")),Replace Table);";
			put "Here:_chrvalues = {};";
			put "Here:_maxvalues = {};";
			put "Here:_minvalues = {1};";
			put "For Each Row(";
			put "  Insert Into( Here:_chrvalues, Column(""&AnnoGroup"")[] );";
			put "  Insert Into( Here:_maxvalues, Column(""Max(CumOrder)"")[] );";
			put "  if(row()>1, Insert Into(Here:_minvalues, lag(Column(""Max(CumOrder)"")[],1)+1));";
			put ");";
			put "pvnchr = N Items( Here:_chrvalues );";
			put "Here:_agmax = max(Here:_maxvalues);";		
			put "Here:_lowery = -(.0736* &Axismax);";
			%if &NbyVal > 1 %then %do;

				put "Here:ppfsize = 800;";
				put "Eval(Parse(EvalInsert(""\[";
				put "  for(_b=1, _b<=&NbyVal, _b++,";
				put "     for(_i=1, _i<=&NPVars, _i++,";
				put '      try(';
				put "        ldr^dt^[_b][framebox(_i)] << FrameSize(ppfsize,0.35*ppfsize);";
				put '  )));';
				put "  mpvlb << Prepend(";
				put "   VListBox(";
				put "   HListBox(";
				put "   	Text Box( "" Multiple testing correction: &mtcmethod"", << set wrap(800) ),";
				put "	),";
				put "   HListBox(";
				put "   	Text Box( "" P-value adjustment method: &adjmethod"", << set wrap(800) ),";
				put "	),";
				put "   HListBox(";
				put "   	Text Box( "" "" ),";
				put "	),";				
				put "  HListBox(TextBox("" Frame Size for Graphs Below""),";
				put "  frameSizeSlider^dt^ = SliderBox(num(200),num(800),Here:ppfsize,";
				put "  for(_b=1, _b<=&NbyVal, _b++,";
				put "     for(_i=1, _i<=&NPVars, _i++,";
				put "	  try(";
				put "        ldr^dt^[_b][framebox(_i)] << FrameSize(ppfsize,0.35*ppfsize);";
				put "  )));";				
				put ")))";
				put ")";
				put ']\")));';			

				put "for(_b=1, _b<=&NbyVal, _b++,";
				put "   for(_i=1, _i<=&NPVars, _i++,";
				put '    try(';
				put "Eval(Parse(EvalInsert(""\[";
				/*put "      ldr^dt^[_b][framebox(_i)] << FrameSize(650,175);";*/
				put "      ldr^dt^[_b][framebox(_i)] << add graphics script(pencolor(""red"");linestyle(""dashed"");hline(&r));";
				put "      ldr^dt^[_b][framebox(_i)]  << Marker Size(3);"; 
				put "      ldr^dt^[_b][framebox(_i)]  << Marker Drawing Mode(Normal);"; 
				%if %length(&AnnoGroup) and &NumGps>1 and &AnnoGroup ^= "_group_" %then %do;
					put "      ldr^dt^[_b][AxisBox(1)] << Label Row( 1, {Automatic Tick Marks( 0 ), Show Major Labels( 0 )});";
					put "      ldr^dt^[_b][AxisBox(1)] << Label Row( 2, {Automatic Tick Marks( 1 ), Show Major Labels( 1 )});";
				%end;
				%else %do;
					put "      ldr^dt^[_b][AxisBox(1)] << Label Row( 1, {Automatic Tick Marks( 1 ), Show Major Labels( 0 )});";
				%end;		
				%do p=1 %to &NPVars;
					/*put "      ldr^dt^[_b][Axisbox(&p+1)] << {Format( ""Fixed Dec"", 9, 2 ),Min(_lowery-4*((_lowery+&AxisMax)/175)), Max(&AxisMax)};";*/	
					put "      ldr^dt^[_b][Axisbox(&p+1)] << {Format( ""Fixed Dec"", 9, 2 ),Min(&ymin), Max(&AxisMax)};";
					/*put "      ldr^dt^[_b][Axisbox(&p+1)] << {Inc( %sysevalf((&AxisMax-&AxisMin)/6) ), Show Major Ticks( 0 ), Show Minor Ticks( 0 ), Remove Axis Label, Min(-0.5)};";*/
					put "      ldr^dt^[_b][Axisbox(&p+1)] << {Inc( %sysevalf((&AxisMax-&AxisMin)/6) ), Show Major Ticks( 0 ), Show Minor Ticks( 0 ), Remove Axis Label};";
				%end;
				/*put "      ldr^dt^[_b][framebox(_i)] << add graphics script(";*/
				/*put "        TextSize(7);";*/
				/*S1149704 sbrlsi %do j=1 %to &tNumGps %by &byinc;*/
				/*
				%if &NumGps > 1 %then %do;
					%let byinc=1;
					%do j=1 %to &NumGps %by &byinc; 
						put "         Text(Center Justified, {(Here:_minvalues[&j]+Here:_maxvalues[&j])/2 ,_lowery}, ""&&SGroupn&j"");";
					%end;
				%end;
				*/
				/*put "      ); // end graphics script "; */
				put "if(_i==1,";
			      	put "      //show(ldr^dt^[_b][OutLineBox(1)]);";
				put "      gtitle= ldr^dt^[_b][OutLineBox(1)]<< Get Text;";
      				put "      //show(gtitle);";
      				put "      gtitle = Trim(Munger(Trim(gtitle), 1, ""Graph Builder"", """"));";
      				put "      //show(gtitle);";
				put "      pos= Munger(gtitle, 1, ""Where("");";
				put "      //show(pos);";
				put "      if(pos, ";
				put "        	del= Trim(Substr(gtitle, pos - 6));";
				put "         	ldr^dt^[_b][OutLineBox(1)] << Dispatch( {}, del, TextBox, {Hide( 1 )} );";       		
				put "          	gtitle= Trim(Substr(gtitle, 1, pos - 6));";				
				put "      );";
				put "      //show(gtitle);";
      				put "      ldr^dt^[_b][OutLineBox(1)] << {Set Title(""Significance Test for ""||gtitle), Image Export Display( Normal )};";				
				put ");";
				put ']\")));';	
				put ")));";	
				put "Close( agsum, no save );"; 

			%end;
			%else %do;
				put "Here:ppfsize = 800;";
				put "Eval(Parse(EvalInsert(""\[";
				put "     for(_i=1, _i<=&NPVars, _i++,";
				put '      try(';
				put "        ldr^dt^[framebox(_i)] << FrameSize(ppfsize,0.35*ppfsize);";
				put "  ));";
				put "   ldvlb << Prepend(";
				put "   VListBox(";
				put "   HListBox(";
				put "   	Text Box( "" Multiple testing correction: &mtcmethod"", << set wrap(800) ),";
				put "	),";
				put "   HListBox(";
				put "   	Text Box( "" P-value adjustment method: &adjmethod"", << set wrap(800) ),";
				put "	),";
				put "   HListBox(";
				put "   	Text Box( "" "" ),";
				put "	),";				
				put "	HListBox(TextBox("" Frame Size for Graphs Below""),";
				put "     frameSizeSlider^dt^ = SliderBox(num(200),num(800),Here:ppfsize,";				
				put "     for(_i=1, _i<=&NPVars, _i++,";
				put "      try(";
				put "        ldr^dt^[framebox(_i)] << FrameSize(ppfsize,0.35*ppfsize);";
				put "  ));";				
				put ");";
				put ")";
				put "))";
				put ']\")));';				

				put "   for(_i=1, _i<=&NPVars, _i++,";
				put '    try(';
				put "Eval(Parse(EvalInsert(""\[";
				/*put "      ldr^dt^[framebox(_i)] << FrameSize(650,175);";*/
				put "      ldr^dt^[framebox(_i)] << add graphics script(pencolor(""red"");linestyle(""dashed"");hline(&r));";
				/*%if %upcase(&ProcessName) = QTLSINGLEMARKERANALYSIS %then %do;
					put "   	ldr^dt^[framebox(_i)]  << Marker Size(3);";
					put "   	ldr^dt^[framebox(_i)]  << Marker Drawing Mode(Normal);";  
				%end;
				%else %do;*/
					put "      ldr^dt^[framebox(_i)]  << Marker Size(3);";
					put "      ldr^dt^[framebox(_i)]  << Marker Drawing Mode(Normal);";
				/*%end;*/
				%if %length(&AnnoGroup) and &NumGps>1 and &AnnoGroup ^= "_group_" %then %do;
					put "      ldr^dt^[AxisBox(1)] << Label Row( 1, {Automatic Tick Marks( 0 ), Show Major Labels( 0 )});";
					put "      ldr^dt^[AxisBox(1)] << Label Row( 2, {Automatic Tick Marks( 1 ), Show Major Labels( 1 )});";
				%end;
				%else %do;
					put "      ldr^dt^[AxisBox(1)] << Label Row( 1, {Automatic Tick Marks( 1 ), Show Major Labels( 0 )});";
				%end;	
				%do p=1 %to &NPVars;
					/*put "      ldr^dt^[Axisbox(&p+1)] << {Format( ""Fixed Dec"", 9, 2 ),Min(_lowery-4*((_lowery+&AxisMax)/175)), Max(&AxisMax)};";*/
					put "      ldr^dt^[Axisbox(&p+1)] << {Format( ""Fixed Dec"", 9, 2 ),Min(&ymin), Max(&AxisMax)};";					
					/*put "      ldr^dt^[Axisbox(&p+1)] << {Inc( %sysevalf((&AxisMax-&AxisMin)/6) ), Show Major Ticks( 0 ), Show Minor Ticks( 0 ), Remove Axis Label, Min(-0.5)};";*/
					put "      ldr^dt^[Axisbox(&p+1)] << {Inc( %sysevalf((&AxisMax-&AxisMin)/6) ), Show Major Ticks( 0 ), Show Minor Ticks( 0 ), Remove Axis Label};";
				%end;
				/*put "      ldr^dt^[framebox(_i)] << add graphics script(";*/
				/*put "        TextSize(7);";*/
				/*S1149704 sbrlsi %do j=1 %to &tNumGps %by &byinc;*/
				/*
				%if &NumGps > 1 %then %do;
					%let byinc=1;
					%do j=1 %to &NumGps %by &byinc; 
						put "         Text(Center Justified, {(Here:_minvalues[&j]+Here:_maxvalues[&j])/2 ,_lowery}, ""&&SGroupn&j"");";
					%end;
				%end;
				*/
				/*put "      ); // end graphics script "; */

      				put "      ldr^dt^[OutLineBox(1)] << {Set Title(""Significance Test""), Image Export Display( Normal )};";

				put ']\")));';	
				put "));";	
				put "Close( agsum, no save );"; 
			%end;
			put "); //end LDpValue_after function";

			put "InsertInto(Here:TabCodeList,expr(LDpValue()));";
			put "InsertInto(TabAfterCodeList,expr(Here:LDpValue_after));";
			put "//InsertInto(TabAfterCodeList,"" "");";
			put "InsertInto(Here:TabOpenList,1);";				
				
		%end;

		put "if(host is(Windows),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, windows) ),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, POSIX) ));";
		put "	";
		put "hebuttonswindow=Function({},{default local},";
		put "		VListBox(";
		%if %length(&PhaseData) and %upcase(&CreateDSforHTSNP)=YES %then %do;
			put "			TextBox(""Click on a button below to launch a process using the appropriate output data set.""),";
		%end;
		%else %do;
			/*
			%if %length(&PhaseData) %then %let NextProcess = %str(Haplotype Trend Regression);
			%else %let NextProcess = %str(HTSNP Selection); 
			put "			TextBox(""Click on the button below to launch the &NextProcess process using the output data set.""),";
			*/
			%if %length(&PhaseData) %then %do;
				%if %length(&ByVars)=0 %then %do;
					put "			TextBox(""Click on a button below to launch a process using the appropriate output data set.""),";			
				%end;
				%else %do;
					%let NextProcess = %str(Haplotype Trend Regression);
					put "			TextBox(""Click on the button below to launch the &NextProcess process using the output data set.""),";			
				%end;
			%end;
			%else %do;
				%let NextProcess = %str(HTSNP Selection); 
				put "			TextBox(""Click on the button below to launch the &NextProcess process using the output data set.""),";			
			%end;
		%end;
		%if %length(&PhaseData) %then %do;
			put "			TextBox("" ""),";
			put "			ButtonBox(""Haplotype Trend Regression"",";
			put "				if(host is(Windows),";
			put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/HaplotypeTrendRegression.jsl"", absolute, windows)),";
			put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/HaplotypeTrendRegression.jsl"", absolute, POSIX) ));";
			put "	";

			/* Set input data set and Output Folder */
			put "				jg:setDefaultParmValues(2, ""HaplotypeTrendRegression"");";							
			put "				jg:setCurrentParmValue(2, ""HaplotypeTrendRegression"", jg:getParmNum(""HaplotypeTrendRegression"",""InData""),""&OutPath.&PhaseData..sas7bdat"");";
			put "				jg:HaplotypeTrendRegression_updateDeps;";
			put "				jg:setCurrentParmValue(2, ""HaplotypeTrendRegression"", jg:getParmNum(""HaplotypeTrendRegression"",""ByVars""),""&ByVars"");";			
			put "				jg:setCurrentParmValue(2, ""HaplotypeTrendRegression"", jg:getParmNum(""HaplotypeTrendRegression"",""OutPath""),""&ClientOutPath"");";
			%if %upcase(&CreateDSforHTSNP)=YES %then %do;
				put "				jg:setCurrentParmValue(2, ""HaplotypeTrendRegression"", jg:getParmNum(""HaplotypeTrendRegression"",""CombineBelow""),""0.01"");";			
				put "				jg:setCurrentParmValue(2, ""HaplotypeTrendRegression"", jg:getParmNum(""HaplotypeTrendRegression"",""HapFreqData""),""&OutPath.&OutName._hfr.sas7bdat"");";			
				put "				jg:HaplotypeTrendRegression_updateDeps;";
				put "				jg:setCurrentParmValue(2, ""HaplotypeTrendRegression"", jg:getParmNum(""HaplotypeTrendRegression"",""FreqVar""),""&FreqVar"");";			
			%end;			
			%if %upcase(&CreateDSforHTSNP)=YES %then %do;
				put "			) ,";		*end ButtonBox;
			%end;
			%else put "			),";;

			%if %length(&ByVars)=0 %then %do;
				put "			ButtonBox(""Haplotype Q-K Mixed Model"",";
				put "				if(host is(Windows),";
				put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/HaplotypeQKMixedModel.jsl"", absolute, windows)),";
				put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/HaplotypeQKMixedModel.jsl"", absolute, POSIX) ));";
				put "	";

				/* Set input data set and Output Folder */
				put "				jg:setDefaultParmValues(2, ""HaplotypeQKMixedModel"");";							
				put "				jg:setCurrentParmValue(2, ""HaplotypeQKMixedModel"", jg:getParmNum(""HaplotypeQKMixedModel"",""InData""),""&OutPath.&PhaseData..sas7bdat"");";
				put "				jg:HaplotypeQKMixedModel_updateDeps;";
				/*put "				jg:setCurrentParmValue(2, ""HaplotypeQKMixedModel"", jg:getParmNum(""HaplotypeQKMixedModel"",""ByVars""),""&ByVars"");";*/			
				put "				jg:setCurrentParmValue(2, ""HaplotypeQKMixedModel"", jg:getParmNum(""HaplotypeQKMixedModel"",""OutPath""),""&ClientOutPath"");";
				%if %upcase(&CreateDSforHTSNP)=YES %then %do;
					put "				jg:setCurrentParmValue(2, ""HaplotypeQKMixedModel"", jg:getParmNum(""HaplotypeQKMixedModel"",""CombineBelow""),""0.01"");";			
					put "				jg:setCurrentParmValue(2, ""HaplotypeQKMixedModel"", jg:getParmNum(""HaplotypeQKMixedModel"",""HapFreqData""),""&OutPath.&OutName._hfr.sas7bdat"");";			
					put "				jg:HaplotypeQKMixedModel_updateDeps;";
					put "				jg:setCurrentParmValue(2, ""HaplotypeQKMixedModel"", jg:getParmNum(""HaplotypeQKMixedModel"",""FreqVar""),""&FreqVar"");";			
				%end;			
				%if %upcase(&CreateDSforHTSNP)=YES %then %do;
					put "			) ,";		*end ButtonBox;
				%end;
				%else put "			),";;
			%end;
		%end;
		%if %upcase(&CreateDSforHTSNP)=YES %then %do;
			put "			ButtonBox(""htSNP Selection"",";
			put "				if(host is(Windows),";
			put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/HTSNPSelection.jsl"", absolute, windows)),";
			put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/HTSNPSelection.jsl"", absolute, POSIX) ));";
			put "	";		
			/* Set input data set and Output Folder */
			put "				jg:setDefaultParmValues(2, ""HTSNPSelection"");";					
			put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""InData""),""&OutPath.&OutName._hfr.sas7bdat"");";
			put "				jg:HTSNPSelection_updateDeps;";
			put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""OutPath""),""&ClientOutPath"");";	
			%let _allelevars=;
			%do i=1 %to &mhaplen; 
				%let _allelevars=&_allelevars _Allele&i;
			%end;
			put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""AlleleVars""),""&_allelevars"");";
			put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""FreqVar""),""&FreqVar"");";
			put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""MarkerID""),""Markers"");";
			put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""ByVars""),""&ByVars"");";			

			%if %length(&AnnoInData) %then %do;
				put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""AnnoInData""),""&AnnoInData"");";		
				put "				jg:HTSNPSelection_updateDeps;";			
				put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""AnnoGroup""),""&OrigAnnoGroup"");";
				put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""AnnoLoc""),""&OrigAnnoLoc"");";
				put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""AnnoLabel""),""&OrigAnnoLabel"");";
				put "				jg:HTSNPSelection_updateDeps;";							
				%if %length(&OrigAnnoWindow) %then %do;
					put "				jg:setCurrentParmValue(2, ""HTSNPSelection"", jg:getParmNum(""HTSNPSelection"",""AnnoWindow""),""&OrigAnnoWindow"");";
				%end;
			%end;
			put "			)";;
		%end;
		put "		)";
		put ");";
		put "Here:LaunchCode = hebuttonswindow();";
	run;

	%if %length(&TraitStmt)=0 /*and (%upcase(&TestAllelicAssociation) ne YES)*/ %then %do;
		%PackageInsertFile(jsl code from &ProcessName, &OutPath, &HaplotypeEstimationScriptName, , text/plain, yes);
	%end;

%end;

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend;

* run the macro;
%HaplotypeEstimation;




