/*
JMP Genomics Analytical Process

Name: Haplotype Trend Regression
Author: Wendy Czika

Description:
Performs "haplotype trend regression" using output 
data set from PROC HAPLOTYPE containing 
individuals' probabilities for haplotype pairs to 
test for association with a quantitative trait 
using PROC REG.  See the HAPLOTYPE procedure 
in the SAS/Genetics User's Guide and the REG 
procedure in the SAS/STAT User's Guide for more
information.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData TraitVars IDVars ByVars WhereStmt 
CombineBelow HapFreqData FreqVar DropBelow RegMethod 
CensorVar CensorValues Covar ClassVars Perms Seed
EventTrait OutName OutPath ServerOutPath MacroPath 
HaplotypeTrendReg_OutData ModelOut SaveEach BuildCombined 
HaplotypeTrendReg_Frame HaplotypeTrendReg_Body 
HaplotypeTrendReg_Contents HaplotypeTrendReg_Script

OutputResid OutputPred OutputFitStat Conv MTC PValAdj Tau Bandwidth Alpha OnlySig

AnnoInData AnnoLabel AnnoMName AnnoGroup AnnoWindow AnnoLoc AnnoWhereStmt
;

*ProcessBody;

%macro HapTrendReg;

%let pe= 0;
%let pest= 1;
%let OutputRSquare= No;
%let Odds= No;
%let DDFM= Residual;
%let GMethod= RSPL;

%let RandEff= ;
%let ie=0;
%let IntEffBuild= ;
%let IntEff= ;

/*S1199692: clear title to avoid showing titles from other processes*/
title;

*initialization and argument checking;
%global ProcessName;
%let ProcessName = HaplotypeTrendRegression; 

%global FullAPName;
%let FullAPName=%str(Haplotype Trend Regression);

%include "&MacroPath./UtilityMacros.sas" / nosource;
%include "&MacroPath./GenMacros.sas" / nosource; /*needed to produce plots*/

%GetOrigFileName(InData,OrigInData);

%PathName(&InData);
%let InData = &TmpName;
libname InLib "&TmpPath";
%if &exiterror=1 %then %goto exit;

%let HFData=;
%if %length(&HapFreqData) %then %do;
	%PathName(&HapFreqData);
	%if &exiterror=1 %then %goto exit;
	libname HFLib "&TmpPath";
	%let HFData = HFLib.&TmpName;

	data FreqReqVars; 
		set &HFData;
		where 0;
		keep WindowID Haplotype &FreqVar;
	run;

	%let _tid_ = %sysfunc(open(FreqReqVars));
	%let _nFreqReqVars_ = %sysfunc(attrn(&_tid_,NVARS));
	%let rc = %sysfunc(close(&_tid_));

	%if &_nFreqReqVars_ < 3 %then %do;
		%put ERROR: The Haplotype Frequency data set does not have the required variables: WindowID, &FreqVar, and Haplotype.;
		%let exiterror = 1;
		%goto exit;
	%end;

	/*S1203673: fixing error due to WhereStmt*/
	%if %length(&WhereStmt) %then %do;
		%let hfds= hfdat_sub;
		data &hfds;
			set &HFData;
			haplotype=tranwrd(haplotype,'-','_');
			where &WhereStmt; 
		run;
		
		%Nobsvars(&hfds);
		%if &nobs=0 %then %do;
			%put ERROR: There are no observations in the Haplotype Frequency Data Set to be analyzed.  Check that you have entered a valid Where Clause.;
			%let exiterror = 1;
			%goto exit;
		%end;
	
	%end;
	%else %do;
		%let hfds= &HFData;
	%end;
%end;

%if &RegMethod=Survival and %length(&CensorVar)=0 %then %do;
	%put ERROR: You must select a Censor Variable to use when Type of Trait is Survival.;
	%goto exit;
%end;

* check name for OutName;
%CheckDataName(_name=OutName,_default=&OrigInData);
%if %length(&OutName)>26 %then %let OutName=%substr(&OutName,1,26);

%if %length(&ByVars) %then %do;
   proc sort data=InLib.&InData out=a;
      by &ByVars;
   run;
   %let ds = a;
%end;
%else %let ds = InLib.&InData;

%if %length(&WhereStmt) %then %do;
	data ds_sub;
		set &ds;
		where &WhereStmt;
	run;

	%Nobsvars(ds_sub);
	%if &nobs=0 %then %do;
		%put ERROR: There are no observations in the data set to be analyzed.  Check that you have entered a valid Where Clause.;
		%goto exit;
	%end;
	%let ds= ds_sub;
%end;

%global Nbyval NByVars ByVarsOnly Jby OrigNByVal NonTraitByVars JNonTraitByVars lastby;
%if %length(&ByVars) %then %do;
	%let temp=%sysfunc(tranwrd(&ByVars,DESCENDING, ));
	%let ByVarsOnly=%sysfunc(compbl(&temp)); 
	%let JBy=%sysfunc(translate(&ByVarsOnly,","," "));
	%let i=1;
	%do %while (%length(%scan(&ByVarsOnly, &i))>0);
		%let i=%eval(&i+1);
	%end;
	%let NByVars = %eval(&i-1);
	data countby;
		set &ds;
		by &ByVars;
		%let lastby=%scan(&ByVarsOnly, &NByVars);
		if first.&lastby;
	run;
	%let dsid=%sysfunc(open(countby));
	%let Nbyval=%sysfunc(attrn(&dsid,NOBS));
	%let ret=%sysfunc(close(&dsid));	
	%if &Nbyval=1 and &ProcessName ne %str(ImputedSNPTraitAssociation)%then %do;
		%let ByVars=;
		%let ByVarsOnly=;
		%let JBy=;
	%end;
%end;
%else %do;
 %let Nbyval=1;
 %let ByVarsOnly=;
 %let NByVars = 0;
%end;

%let OrigNByVal = &NByVal;
%put OrigNByVal= &OrigNByVal;
%let NonTraitByVars = &ByVars;

%let b=1;
%do %while(%length(%scan(&NonTraitByVars,&b)));
	%let JNonTraitByVars = &JNonTraitByVars :%scan(&NonTraitByVars,&b),;
	%let b=%eval(&b+1);
%end;

data HapReqVars; 
	set &ds;
	where 0;
	keep WindowID IndivID Haplotype1 Haplotype2 Prob;
run;

%let _tid_ = %sysfunc(open(HapReqVars));
%let _nHapReqVars_ = %sysfunc(attrn(&_tid_,NVARS));
%let rc = %sysfunc(close(&_tid_));

%if &_nHapReqVars_ < 5 %then %do;
	%put ERROR: The Phased Haplotype data set does not have the required variables: WindowID, IndivID, Haplotype1, Haplotype2, Prob.;
	%let exiterror = 1;
	%goto exit;
%end;

%if &ModelOut = HTML %then %do;
	%PrepHTML(HaplotypeTrendReg_body, HaplotypeTrendReg_frame, HaplotypeTrendReg_contents,
		&OutName._htr_body.html,&OutName._htr_frame.html,&OutName._htr_contents.html);

	* setup for html output;
	ods html 
	   body = "&HaplotypeTrendReg_Body" (url="&bodyfile")
	   frame = "&HaplotypeTrendReg_Frame"  (url="&framefile")
	   contents = "&HaplotypeTrendReg_Contents" (url="&contentsfile")
	;
	ods listing close;
	%let OutputType = HTML;
%end;
%else %if &ModelOut = Text %then %do;
	%let OutputType = Listing;
%end;
%else %do;
	%let OutputType = None;
	ods exclude all;
	ods noresults;
%end;

data out1;
  set &ds;
  Haplotype=tranwrd(haplotype1,'-','_');
  drop haplotype1;
  /*
  %if %length(&WhereStmt) %then %do;
	where &WhereStmt;
  %end;
  */
run;

%if &syserr>4 %then %do;
	%let exiterror=1;
	%goto exit;
%end;

data out2;
  set &ds;
  haplotype=tranwrd(haplotype2,'-','_');
  drop haplotype2;
	/*
	%if %length(&WhereStmt) %then %do;
		where &WhereStmt;
	%end; 
	*/
run;

/*data OutLib.out1; set out1; run;*/
/*data OutLib.out2; set out2; run;*/

data outnew;
  set out1 out2;
run;
%if &syserr>4 %then %goto exit;

/*data OutLib.outnew; set outnew; run;*/ 

%if &Perms>0 %then %do;
	%if &RegMethod = Binary %then %do;
		/*S1218151: adding loop for TraitVars to convert all binary character variables to 0/1*/
		/* If trait var is character & binary, convert it to 0/1 */
		%let TraitVarsNum=;
				
		%let ti=1;
		%do %while(%length(%scan(&TraitVars,&ti))>0);			
			%let TraitVar = %scan(&TraitVars,&ti);
				
			%let dsid=%sysfunc(open(outnew,i));
			%let trnum=%sysfunc(varnum(&dsid,&TraitVar));
			%let tlen=%sysfunc(varlen(&dsid,&trnum));
			%let ttype=%sysfunc(vartype(&dsid,&trnum));
			%let rc=%sysfunc(close(&dsid));

			%if &ttype=C %then %do;
				%CheckBinaryChar(outnew,&TraitVar,_len=&tlen);

				%if &nvalue=2 %then %do;
					%let tnum=%substr(%trim(&TraitVar),1,28)_num;
					%put Converting &TraitVar to numeric 0/1 variable &tnum.;
					data outnew;
						set outnew;
						if &TraitVar="&cval1" then &tnum=0;
						else if &TraitVar="&cval2" then &tnum=1;
						else &tnum=.;
					run;
					
					%let TraitVarsNum= &TraitVarsNum &tnum;
				%end;
				%else %do;
					%put ERROR: Trait Variable &TraitVar selected is not binary.;
					%goto exit;
				%end;
			%end;
			%else %do;
				%let TraitVarsNum= &TraitVarsNum &TraitVar;
			%end;
			%let ti=%eval(&ti+1);
		%end;
		%let TraitVars= &TraitVarsNum;
	%end;
%end;

proc sort data=outnew;
 by WindowID haplotype;
run;

proc sort data=outnew out=winvals nodupkey;
	by WindowID;
run;

data _null_;
	set winvals end=last;
	call symput('winval'||strip(_n_),strip(WindowID));
	if last then call symput('Nwinvals',strip(_n_));
	call symput("wintype",vtype(WindowID));
run;

%let IDVars = IndivID &IDVars;
%let IndIDMerg= &IDVars;

* some windows may not contain the full data set, so assemble it here;
proc sort data=outnew nodupkey
	out=fulldat(keep=&ByVars &IDVars &TraitVars &Covar &CensorVar &ClassVars);
	by &ByVars &IDVars;
run;
%NObsVars(fulldat);
%let nobsfull = &nobs;

/*data OutLib.outnew; set outnew; run;*/
	
/* Just so don't have to add WindowID everywhere */
%let ByVars0 = &ByVars;
%let ByVars = WindowID &ByVars;

%global HaplotypeTrendReg_HRG HaplotypeTrendReg_HRS;
%PrepOutDS(HaplotypeTrendReg_HRG,HaplotypeTrendReg_HRGName,&OutName._hrg);
%PrepOutDS(HaplotypeTrendReg_HRS,HaplotypeTrendReg_HRSName,&OutName._hrs);

data OutLib.&HaplotypeTrendReg_HRGName;
	delete;
run;

data OutLib.&HaplotypeTrendReg_HRSName;
	delete;
run;

%global HaplotypeTrendReg_HRGE HaplotypeTrendReg_HRGR HaplotypeTrendReg_HRGP HaplotypeTrendReg_HRSE;
%let HaplotypeTrendReg_HRGE=;
%let HaplotypeTrendReg_HRGR=;
%let HaplotypeTrendReg_HRGP=;
%let HaplotypeTrendReg_HRSE=;

%PrepOutDS(HaplotypeTrendReg_HRGE,HaplotypeTrendReg_HRGEName,&OutName._HRGE);
%if %upcase(&OutputResid)=YES %then %PrepOutDS(HaplotypeTrendReg_HRGR,HaplotypeTrendReg_HRGRName,&OutName._HRGR);
%if %upcase(&OutputPred)=YES %then %PrepOutDS(HaplotypeTrendReg_HRGP,HaplotypeTrendReg_HRGPName,&OutName._HRGP);
%PrepOutDS(HaplotypeTrendReg_HRSE,HaplotypeTrendReg_HRSEName,&OutName._HRSE);

%if %upcase(&BuildCombined)=YES %then %do;
	%global HaplotypeTrendReg_HRW;
	%PrepOutDS(HaplotypeTrendReg_HRW,HaplotypeTrendReg_HRWName,&OutName._hrw);
	%let first = 1;
%end;

/* Table name from all but MIXED */
%global ParmEst EffName;
%let ParmEst=ParameterEstimates;
%let EffName=Effect;

%let ModOption=;
%let TrtOpt = ;
%if &RegMethod = Binary %then	%do;
	%let WhichProc = LOGISTIC;
	%let CTable = ContrastTest;
	%let StatC = WaldChisq;
	%let ProbC = ProbChiSq;
	%let StatCLabel = ChiSq-Statistic;
	%let ProbCLabel = Prob > ChiSq;
	%let ProbC_empLabel = Empirical Prob > ChiSq;	
	%let ConVarC=Contrast;
	%if %length(&EventTrait)>0 %then %let TrtOpt=%str((event="&EventTrait"));
	%else %let TrtOpt=%str((event=LAST));
	%let Gtable=ModelANOVA; /*S1146207: The logistic output formerly named Type3 is now named ModelANOVA*/
	%let Stat=ChiSq;
	%if %upcase(&OutputResid)=YES and %upcase(&OutputPred)=YES %then %let ModOption = %quote(&ModOption) %str(; output out=outres resdev=Resid predicted=pred);
	%else %if %upcase(&OutputResid)=YES and %upcase(&OutputPred)^=YES %then %let ModOption = %quote(&ModOption) %str(; output out=outres resdev=Resid);
	%else %if %upcase(&OutputResid)^=YES and %upcase(&OutputPred)=YES %then %let ModOption = %quote(&ModOption) %str(; output out=outres predicted=pred);
	%let EffName=Variable;
%end;
%else %if &RegMethod = Continuous %then %do;
	%*let WhichProc = GLM;
	%let WhichProc = MIXED;
	%let CTable =  Contrasts;
	%let StatC = FValue;
	%let ProbC = ProbF;
	%let StatCLabel = F-Statistic;
	%let ProbCLabel = Prob > F;
	%let ProbC_empLabel = Empirical Prob > F;
	%*let ConVarC=Source;
	%let ConVarC=Label;	
	%let Gtable=Tests3;
	%let Stat=F;
	%let ModOption=%quote(/ htype=3 ddfm=&DDFM s); 
	%if %upcase(&OutputResid)=YES or %upcase(&OutputPred)=YES %then %let ModOption = %quote(&ModOption residual outp=outres);
	%let ParmEst = SolutionF;
%end;
%else %if &RegMethod = Count %then %do;
	%let WhichProc = GLIMMIX;
	%let CTable = Contrasts;
	%let StatC = FValue;
	%let ProbC = ProbF;
	%let StatCLabel = F-Statistic;
	%let ProbCLabel = Prob > F;
	%let ProbC_empLabel = Empirical Prob > F;
	%let ConVarC=Label;
	%*let ModOption = %str(/link=log dist=p);
	%let ModOption=%quote(/ link=log dist=p htype=3 ddfm=&DDFM s);
	%if %upcase(&OutputResid)=YES and %upcase(&OutputPred)=YES %then %let ModOption = %quote(&ModOption) %str(; output out=outres resid=Resid predicted=Pred);
	%else %if %upcase(&OutputResid)=YES and %upcase(&OutputPred)^=YES %then %let ModOption = %quote(&ModOption) %str(; output out=outres resid=Resid);
	%else %if %upcase(&OutputResid)^=YES and %upcase(&OutputPred)=YES %then %let ModOption = %quote(&ModOption) %str(; output out=outres predicted=Pred);
	%let Gtable=Tests3;
	%let Stat=F;
%end;
%else %if &RegMethod=Survival %then %do;
	%let WhichProc=PHREG;
	%let CTable=TestStmts;
	%let StatC=WaldChiSq;
	%let ProbC = ProbChiSq;
	%let StatCLabel = ChiSq-Statistic;
	%let ProbCLabel = Prob > ChiSq;
	%let ProbC_empLabel = Empirical Prob > ChiSq;	
	%let ConVarC=Label;
	%let TrtOpt=%str(*&CensorVar(&CensorValues));
	%let Gtable=ModelANOVA;
	%let Stat=ChiSq;
	%let EffName = Parameter;
%end;

%if %upcase(&OutputResid)=YES or %upcase(&OutputPred)=YES %then %do;
	data outrest_o_all;
		length _Trait_ $32;
		delete;
	run;
%end;

data ta_o_all;
	length Effect $32;
	delete;
run;

%if %Upcase(&Odds)=YES  %then %do; 
	data or_o_all;
		delete;
	run;
%end;

data pat_o_all;
	length Effect $32;
	delete;
run;

data pat_s_all;
	length Effect $32;
	delete;
run;

%global global_flag single_flag;
%let global_flag= 0;
%let single_flag= 0;

%let firstwind= 1;
%do w=1 %to &Nwinvals;

	%put;
	%put Processing window &w of &Nwinvals....;
	
	%let flag_globalstatus= 0;

	%if %upcase(&SaveEach)=YES %then %do;
		%let HaplotypeTrendReg_OutData=;
		%PrepOutDS(HaplotypeTrendReg_OutData,HaplotypeTrendReg_OutDataName,&OutName._htr&w);
	%end;
	
	/*data OutLib.outnew&w; set outnew; run;*/
	
	data currwin;
		set outnew;
		where WindowID=&&winval&w;
	run;
	
	/*data OutLib.currwin&w; set currwin; run;*/

	%let Rare=;
	%if %length(&HFData) %then %do;
		
		/*S1203673: fixing error due to WhereStmt*/
		data hf_;
			set &hfds;
			haplotype=tranwrd(haplotype,'-','_');
			%if &wintype=C %then %do;
				where WindowID="&&winval&w";
			%end;
			%else %do;
				where WindowID=&&winval&w;
			%end;
		run;

		proc sort data=hf_;
		 by haplotype;
		run;
		
		/*data outlib.currwin; set currwin; run;*/
		/*data outlib.hf_; set hf_; run;*/

		data outnew_cut;
		 merge currwin hf_(in=keep);
		 by haplotype;
		 if keep;
		 if missing(IndivID) then delete;
		run;
		
		proc sort data=outnew_cut;
			by descending &FreqVar haplotype;
		run;
		
		/*data outlib.outnew_cut0_&w; set outnew_cut; run;*/

		data outnew2;
			set outnew_cut end=last;
			Orig_Haplotype= Haplotype;
			lagh=lag(haplotype);
			if &FreqVar>=&CombineBelow then do;
			 if haplotype ne lagh then num+1;
			 Hapname=compress("H_&w._"||num,' ');
			end;
			else do;
			 Hapname="H_&w._Rare";
			 call symput("Rare","Rare");
			 Haplotype='Rare';
			end;
			if last then do;
			 call symput('numhaps', strip(num));
			end;
		run;	
		
		/*data OutLib.outnew2; set outnew2; run;*/
		
		%if %length(&Rare) %then %do;
			%let numhaps=%eval(&numhaps+1);
		%end;

		%if &numhaps=0 %then %do;
			%put WARNING: There are no haplotypes with a frequency greater than the cutoff specified for window &&winval&w...;
			%goto nextwindow;
		%end;	
	%end;
	%else %do;

		data outnew2;
			set currwin end=last;
			lagh=lag(haplotype);
			if haplotype ne lagh then num+1;
			Hapname=compress("H_&w._"||num,' ');        
			if last then do;
				call symput('numhaps', strip(num));
			end;
		run;
	%end;

	%put numhaps = &numhaps;
	%if &numhaps=1 %then %do;
		%put WARNING: Haplotypes are constant for window &&winval&w...;
		%goto nextwindow;
	%end;	

	proc sort data=outnew2;
		by &ByVars &IDVars hapname;
	run;
	
	/*data outlib.outnew2; set outnew2; run;*/

	proc sort data=outnew2 out=hapnames(keep=Haplotype Hapname) nodupkey;
	 by hapname;
	run;
	
	%if %length(&HFData) %then %do;
		%put HereHFDATA;
		
		proc sort data=outnew2 out=hapnames_ho(keep=Haplotype Hapname &FreqVar Orig_Haplotype) nodupkey;
			by hapname Orig_Haplotype;
		run;	
		
		data hapnames_ho;
			set hapnames_ho;
			by hapname Orig_Haplotype;
			if first.hapname then HapNameFreq=&FreqVar;	
			else if (index(hapname, 'Rare')) then HapNameFreq+&FreqVar;
			if last.hapname;
			drop Orig_Haplotype;
		run;
		
		proc sort data=hapnames_ho out=hapnames(keep=Haplotype Hapname &FreqVar HapNameFreq) nodupkey;
			by hapname;
		run;
	%end;
	%else %do;
		proc sort data=outnew2 out=hapnames(keep=Haplotype Hapname) nodupkey;
			by hapname;
		run;
	%end;
	
	/*data OutLib.hapnames; set hapnames; run;*/

	data outt;
		set outnew2;
		by &ByVars &IDVars hapname;
		if first.hapname then totprob=prob/2;
		else totprob+prob/2;
		if last.hapname;
	run;
	
	/*data outlib.outt; set outt; run;*/
	
	options label;
	proc transpose data=outt out=outreg(drop=_NAME_) ;
		id hapname;
		idlabel haplotype;
		var totprob;
		by notsorted &ByVars &IDVars &TraitVars &Covar &CensorVar &ClassVars;
	run;
	
	* make sure data set is complete;
	%NObsVars(outreg);
	%if &nobs < &nobsfull %then %do;
		data t1;
			set fulldat;
			WindowID = &&winval&w;
		run;
		
		/*data OutLib.outreg0_&w; set outreg; run;*/
		
		data outreg;
			merge t1 outreg;
			by &ByVars &IDVars;
		run;
		
		/*data OutLib.outreg1_&w; set outreg; run;*/
	%end;
	
	%if %length(&HFData) %then %do;
		proc sort data=hapnames out=hapnamesInitial;
			by descending &FreqVar hapname;
		run; 
	
		data _null_; 
			set hapnamesInitial end=last;
			call symputx("hapInitial"||compress(_n_),compress(hapname));
			if last then call symputx("nhapsInitial",compress(_n_));
		run;
	%end;
	
	data outreg;
		%if %length(&HFData) %then %do;
			retain 
			%do i=1 %to &nhapsInitial;
				&&hapInitial&i
			%end;
			;
		%end;
		set outreg;
		array h{&numhaps} h_:;
		do i=1 to &numhaps;
		 if h{i}=. then h{i}=0; 
		end;
		keep &ByVars &IDVars &TraitVars h_: &Covar &CensorVar &ClassVars;
	run;

	/*data outlib.outreg2_&w; set outreg; run;*/

	%if &DropBelow > 0 %then %do;
		data outreg1;
			set outreg;
			drop h_:;
		run;
		data outreg2;
			set outreg;
			keep h_:;
		run;
		proc transpose data=outreg2 out=outreg2t;
			var h_:;
		run;
		data outreg2t;
			set outreg2t;
			array col{*} col:;
			nnz = 0;
			do i=1 to dim(col);
				if col[i] > 0 then nnz = nnz+1;
			end;
			freq = nnz/&nobsfull;
			if freq < &DropBelow then delete;
		run;
		data _null_;
			set outreg2t end=last;
			call symputx("hap"||compress(_n_),compress(_name_));
			if last then call symputx("nhaps",compress(_n_));
		run;
		%if &nhaps < &numhaps %then %do;
			%let numhaps = &nhaps;
			%put numhaps after dropping hapolotypes with frequency below &DropBelow = &numhaps;
			
			%if &numhaps = 0 %then %do;
				%put WARNING: There are no haplotypes with a frequency greater than the cutoff specified for window &&winval&w...;
				%goto nextwindow;
			%end;
			%else %if &numhaps = 1 %then %do;
				%put WARNING: Haplotypes are constant for window &&winval&w...;
				%goto nextwindow;
			%end;	

			proc transpose data=outreg2t out=outreg2tt(drop=_name_);
				var col:;
			run;
			data outreg;
				merge outreg1 outreg2tt;
			run;
			
			/*data OutLib.outreg3_&w; set outreg; run;*/
		%end;
	%end;
	%else %do;
		data outreg2;
			set outreg;
			keep h_:;
			if (0);
		run;
		%let tid = %sysfunc(open(outreg2));	
		%do i=1 %to &numhaps;
			%let hap&i = %sysfunc(varname(&tid,&i)); 
		%end;
		%let tid = %sysfunc(close(&tid));
	%end;

	%if %upcase(&SaveEach)=YES %then %do;
		data &HaplotypeTrendReg_OutData;
			set outreg;
		run;
	%end;
	
	%if %upcase(&BuildCombined)=YES %then %do;
		%if &first %then %do;
			/*data OutLib.outwide&w; set OutLib.&HaplotypeTrendReg_HRWName; run;*/
			/*data OutLib.outreg4_&w; set outreg; run;*/
			
			data OutLib.&HaplotypeTrendReg_HRWName;
				set outreg;
				drop WindowID;
			run;
			%let first = 0;
		%end;
		%else %do;
			/*data OutLib.outwide&w; set OutLib.&HaplotypeTrendReg_HRWName; run;*/
			/*data OutLib.outreg5_&w; set outreg; run; */
		
			data OutLib.&HaplotypeTrendReg_HRWName;
				merge OutLib.&HaplotypeTrendReg_HRWName outreg(keep=&ByVars0 &IDVars h_:);
				by &ByVars0 &IDVars;
			run;
		%end;
	%end;
	
	%if &Perms>0 %then %do;
		%if %length(&Seed)=0 %then %let Seed=0;
		
		%let DropStmt = %quote(missing(%scan(&TraitVars,1)));
		%let ti=2;
		%do %while(%length(%scan(&TraitVars,&ti))>0);
			%let DropStmt = %quote(&DropStmt and missing(%scan(&TraitVars,&ti)));
			%let ti=%eval(&ti+1);
		%end;		

		data p(keep=constant &ByVars &TraitVars &Covar &CensorVar)
			%if %length(&ClassVars) %then pclass(keep=&ClassVars);
		;
		 set outreg;
		 if &DropStmt then delete;	
		 constant=1;
		run;		
		
		data o;
		 set outreg;
		 if &DropStmt then delete;	
		 *keep h_&w._1-h_&w._&numhaps;
		  keep %do i=1 %to &numhaps; &&hap&i %end;; /*S1147948: making sure that H_1_Rare is kept in the data set if it is present*/
		run;

		options nonotes;
		data operms;
		 set %do i=1 %to &Perms; o %end;;
		run;
		options notes;
		
		%if &OutputType ^= None %then %do;
			ods &OutputType exclude all;
		%end;
		proc multtest data=p outsamp=perms perm n=&perms seed=&Seed;
		 class constant;
		 test mean(&TraitVars &Covar &CensorVar);
		 by &ByVars;
		run;

		proc sort data=perms;
			by _sample_ &ByVars;
		run;
		
		%if %length(&ClassVars) %then %do;
			data perms;
				set perms;
				set pclass point=_obs_;
				obs=_obs_;
			run;
		%end;

		data allp;
		 merge perms operms;
		run;
		
	%end;
	
	%let nhaps = &numhaps;
	
	/*
	%let nhaps=%eval(&numhaps-1);
	*/
	
	/* Need to do this for Proc Logistic so it will produce Type3 table when no Class var */
	%let NeedConstant=0;
	%if %length(&ClassVars) = 0 %then %do;
		/* Need to do this for Proc Logistic/Surveylogistic/PHREG so it will produce Type3 
		table for Trend test when no Class var is present*/
		%if (%index(&WhichProc,%str(LOGISTIC))>0 or %upcase(&WhichPROC)=%str(PHREG)) %then %do;
			%let NeedConstant=1;
		%end;
		%else %do;
			%let NeedConstant=0;
		%end;	
	
		%if &NeedConstant %then %do;
			data outreg;
				retain constant 0;
				set outreg;
			run;
			
			%if &Perms>0 %then %do;
				data allp;
					retain constant 0;
					set allp;
				run;
			%end;
			
			%*let ClassVars= constant;
		%end;
	%end;

	data OverallTest&&winval&w;
		delete;
	run;

	data SingleTests&&winval&w;
		delete;
	run;
	
	%if &w=1 %then %do;
		%global NTrait;
		%let NTrait= 0;
	%end;
	
	/* Loop over traits */
	%let ti=1;
	%do %while(%length(%scan(&TraitVars,&ti))>0);
		%let TraitVar = %scan(&TraitVars,&ti);

		%if &RegMethod = Binary %then	%do;
			proc sort data=outreg out=isbin nodupkey;
				by &TraitVar;
				where missing(&TraitVar)=0; 
			run;

			%NObsVars(isbin);
			%if &nobs ^= 2 %then %do;
				%put ERROR: The trait variable &TraitVar is non-binary.  This variable will not be used.;
				%goto nexttrait;
			%end;
		%end;
		
		/*data OutLib.outreg6_&w; set outreg; run;*/

		data ta; /* kemicl - added to ensure data is not repeated if PROC does not produce estimates */
			delete;
		run;		
		
		data ct;
			delete;
		run;
		
		/** Overall test **/
		title "Overall Test for Window= &&winval&w";
		*title "Overall Test";	
		*title2 "(Trait=&TraitVar)";

		%if &OutputType ^= None %then %do;
			ods &OutputType select all;
		%end;
		
		ods output 
			&CTable=ct
			&Gtable=ta
			%if &pe=0 %then %do;
				&ParmEst= pat
				%if &WhichProc ne SURVEYREG %then %do;
					nobs = tnobs %if &WhichProc ne PHREG %then (where=(Label ? "Observations Used"));
				%end;
				%if %upcase(&RegMethod)=BINARY %then responseprofile=rp; 
				/*begin: Luciano*/
				%if %upcase(&OutputRSquare)=YES and %index(%upcase(&WhichProc),LOGISTIC) %then
					rsquare= trsq;
				/*end: Luciano*/	
				%if %upcase(&OutputFitStat)=YES %then 
					FitStatistics= fst;
			%end;		
		;
		proc &WhichProc data=outreg;
			%if %length(&ClassVars) %then %do;
				class &ClassVars;
			%end;
			%if &NeedConstant %then class constant;;
			/*S1149775: removing Rare haplotyple from the model, test and contrast statements, otherwise the contrast is not estimable 
			because all haplotypes sum to 1 (linear combination)*/
			/*model &TraitVar &TrtOpt = &ClassVars &Covar %do i=1 %to &nhaps; &&hap&i %end; &ModOption;*/
			model &TraitVar &TrtOpt = &ClassVars %if &NeedConstant %then constant;
				&Covar %do i=1 %to &nhaps; %if ^%index(%upcase(&&hap&i), RARE) %then &&hap&i; %end; &ModOption; 
			%if &RegMethod=Survival %then %do;
				%if ^%index(%upcase(&hap1), RARE) %then %do;
					Overall: test &hap1 %do i=2 %to &nhaps; %if ^%index(%upcase(&&hap&i), RARE) %then , &&hap&i; %end;;
				%end;
				%else %do;
					Overall: test &hap2 %do i=3 %to &nhaps; , &&hap&i %end;;
				%end;
			%end;
			%else %do;
				%if ^%index(%upcase(&hap1), RARE) %then %do;
					contrast 'Overall' &hap1 1 %do i=2 %to &nhaps; %if ^%index(%upcase(&&hap&i), RARE) %then , &&hap&i 1; %end;; 
				%end;
				%else %do;
					contrast 'Overall' &hap2 1 %do i=3 %to &nhaps;, &&hap&i 1 %end;;
				%end;
			%end;
			%if (%upcase(&OutputResid)=YES or %upcase(&OutputPred)=YES) %then %do;
				id &IndIDMerg;
			%end;			
			%if %length(&ByVars) %then %do; 
				by &ByVars;
			%end;
		run; 
		quit;
		%if &syserr>4 %then %goto exit;
				
		/*
		%if %sysfunc(exist(ct))=0 %then %do;
			%put WARNING: Overall test could not be performed for Trait &TraitVar in Window &&winval&w...;
			%let haperr=1;
		%end;
		%else %do;
			%Nobsvars(ct);
			%if &nobs=0 %then %do;
				%put WARNING: Overall test could not be performed for Trait &TraitVar in Window &&winval&w...;
				%let haperr=1;
			%end;
			%else %do;
				%let haperr=0;
				data ma;
				 length Dependent $32;
				 set ct;
				 Dependent=propcase("&TraitVar");
				 Stat=&StatC;
				 ProbGlobal=&ProbC;
				 keep &ByVars Dependent Stat ProbGlobal;
				 label Stat="&StatCLabel" ProbGlobal="&ProbCLabel";
				run;	

				data ct;
					delete;
				run;
			%end;	
		%end;
		*/
		
		%NObsVars(ct);
		%if &nobs %then %do;
			data ct;
				set ct;
				if not missing(&ProbC);
			run;
			%NObsVars(ct);
		%end;
		
		%if &nobs=0 %then %do;
			%put WARNING: PROC &WhichProc generated no estimates for trait &TraitVar in window &&winval&w...  There may be excessive missing data or the contrast may be nonestimable. Please see the log for more details.;
			%put WARNING: Overall test could not be performed for Trait &TraitVar in Window &&winval&w...;
			%let haperr=1;
		%end;	
		%else %do;
			%let haperr=0;
			data ta_o;
				*length Dependent $32;
				length Trait $32;
				set ta;
				Trait=propcase("&TraitVar");
				*Dependent=propcase("&TraitVar");
				*Stat=&StatC;
				*ProbGlobalHap=&ProbC;
				*keep &ByVars Dependent Stat ProbGlobalHap;
				*label Stat="&StatCLabel" ProbGlobalHap="&ProbCLabel";
			run;	

			/*data OutLib.ta_0o&w; set ta_o; run;*/

			data ta;
				delete;
			run;

			data pat_o;
				*length Dependent $32;
				length Trait $32;
				set pat;
				Trait=propcase("&TraitVar");
				*Dependent=propcase("&TraitVar");
				*Stat=&StatC;
				*ProbGlobalHap=&ProbC;
				*keep &ByVars Dependent Stat ProbGlobalHap;
				*label Stat="&StatCLabel" ProbGlobalHap="&ProbCLabel";
			run;	

			/*data OutLib.pat_0o&w; set pat_o; run;*/

			data ma_o;
				*length Dependent $32;
				length Trait $32;					
				set ct;
				Trait=propcase("&TraitVar");
				*Dependent=propcase("&TraitVar");
				/*Stat=&StatC;*/
				&StatC.Global=&StatC;
				/*ProbGlobalHap=&ProbC;*/
				&ProbC.Global=&ProbC;
				*keep &ByVars Dependent Stat ProbGlobalHap;
				drop &StatC &ProbC;
				/*label Stat="&StatCLabel" ProbGlobalHap="&ProbCLabel";*/
			run;	

			/*data OutLib.ma_o0_&w; set ma_o; run;*/

			data ct;
				delete;
			run;

			%if %upcase(&RegMethod)=BINARY %then %do;
				proc transpose data=rp out=rp(drop=_name_) prefix=NumTrait;
					by &ByVars;
					var count;
					id %if %index(%upcase(&WhichProc),LOGISTIC) %then Outcome; %else &TraitVar;;
				run;

				%let datatomerge=rp;
			%end;
			%else %let datatomerge=;	

			%if &WhichProc ne SURVEYREG %then %do;
				data ta_o;
					merge tnobs(keep=&ByVars nobsused) &datatomerge ta_o;
					by &ByVars;
					rename nobsused=SampleSize;
				run;

				data pat_o;
					merge tnobs(keep=&ByVars nobsused) &datatomerge pat_o;
					by &ByVars;
					rename nobsused=SampleSize;
				run;
			%end;

			/*begin: including RSquare for logistic regression. Luciano*/
			%if %upcase(&OutputRSquare)=YES and %index(%upcase(&WhichProc),LOGISTIC) %then %do;
				data ta_o;
					merge ta_o trsq(keep= nValue1 nValue2 &ByVars);
					by &ByVars;
					rename nValue1= RSquare_Global nValue2=Rescaled_RSquare_Global;
				run;
			%end;
			/*end: Luciano*/

			/*data OutLib.ta_1o&w; set ta_o; run;*/

			%if %upcase(&OutputFitStat)=YES %then %do;
				%put WhichProc= &WhichProc;
				/*data OutLib.fst0; set fst; run;*/
				/*data OutLib.ta; set ta; run;*/

				proc transpose data=fst(
					%if %index(%upcase(&WhichProc),LOGISTIC) %then %do;
						rename=(
							criterion= descr 
							interceptandcovariates= value
						)
					%end;
					%else %if %index(%upcase(&WhichProc),PHREG) %then %do;
						rename=(
							criterion= descr  
							withcovariates= value
						)
					%end;
					)
					out=fst(drop= _name_); 
					var value;
					id descr;
					by &ByVars;
				run;

				/*data OutLib.fst1; set fst; run;*/

				data ta_o;
					merge ta_o fst;
					by &ByVars;
					rename
					%if %index(%upcase(&WhichProc),MIXED) %then %do;
						_2_Res_Log_Likelihood= _2_Res_Log_Likelihood_Global
						AIC__Smaller_is_Better_= AIC_Smaller_is_Better_Global
						AICC__Smaller_is_Better_= AICC_Smaller_is_Better_Global
						BIC__Smaller_is_Better_= BIC_Smaller_is_Better_Global
					%end;
					%else %if %index(%upcase(&WhichProc),GLIMMIX) %then %do;
						%if (%upcase(&RegMethod)=COUNT and %length(&RandEff)>0) 
							or (%upcase(&RegMethod)=BINARY and %length(&RandEff)>0) %then %do;
							%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
								_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global
							%end;
							%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
								_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global
							%end;
							%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
								_2_Log_Likelihood= _2_Log_Likelihood_Global
								AIC___smaller_is_better_= AIC_Smaller_is_Better_Global
								AICC__smaller_is_better_= AICC_Smaller_is_Better_Global
								BIC___smaller_is_better_= BIC_Smaller_is_Better_Global
								CAIC__smaller_is_better_= CAIC_smaller_is_better_Global
								HQIC__smaller_is_better_= HQIC_smaller_is_better_Global											
							%end;
						%end;
						%else %if %upcase(&RegMethod)=COUNT and %length(&RandEff)=0 %then %do; 		
							_2_Log_Likelihood= _2_Log_Likelihood_Global
							AIC___smaller_is_better_= AIC_Smaller_is_Better_Global
							AICC__smaller_is_better_= AICC_Smaller_is_Better_Global
							BIC___smaller_is_better_= BIC_Smaller_is_Better_Global
							CAIC__smaller_is_better_= CAIC_smaller_is_better_Global
							HQIC__smaller_is_better_= HQIC_smaller_is_better_Global
							Pearson_Chi_Square= Pearson_ChiSquare_Global
							Pearson_Chi_Square___DF= Pearson_ChiSq_DF_Ratio_Global											
						%end;
						%else %if %upcase(&RegMethod)=NOMINAL or %upcase(&RegMethod)=ORDINAL %then %do;
							%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
								_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Global
							%end;
							%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
								_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Global
							%end;
							%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
								_2_Log_Likelihood= _2_Log_Likelihood_Global
								AIC___smaller_is_better_= AIC_Smaller_is_Better_Global
								AICC__smaller_is_better_= AICC_Smaller_is_Better_Global
								BIC___smaller_is_better_= BIC_Smaller_is_Better_Global
								CAIC__smaller_is_better_= CAIC_smaller_is_better_Global
								HQIC__smaller_is_better_= HQIC_smaller_is_better_Global
							%end;	
						%end;
						%else %if (%upcase(&RegMethod)=BINARY and &kmodel=1) %then %do;
							%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
								_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global
							%end;
							%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
								_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global											
							%end;
						%end;										
					%end;
					%else %if %index(%upcase(&WhichProc),PHREG) %then %do;
						_2_LOG_L= _2_LOG_L_Global
						AIC= AIC_Global
						SBC= SBC_Global
					%end;
					%else %if %index(%upcase(&WhichProc),LOGISTIC) %then %do;
						AIC= AIC_Global
						SC= SC_Global
						_2_Log_L= _2_Log_L_Global
					%end;
					;
				run;

				/*data OutLib.ta_2o&w; set ta_o; run;*/

				data ma_o;
					merge ma_o fst;
					by &ByVars;
					rename
					%if %index(%upcase(&WhichProc),MIXED) %then %do;
						_2_Res_Log_Likelihood= _2_Res_Log_Likelihood_Global
						AIC__Smaller_is_Better_= AIC_Smaller_is_Better_Global
						AICC__Smaller_is_Better_= AICC_Smaller_is_Better_Global
						BIC__Smaller_is_Better_= BIC_Smaller_is_Better_Global
					%end;
					%else %if %index(%upcase(&WhichProc),GLIMMIX) %then %do;
						%if (%upcase(&RegMethod)=COUNT and %length(&RandEff)>0) 
							or (%upcase(&RegMethod)=BINARY and %length(&RandEff)>0) %then %do;
							%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
								_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global
							%end;
							%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
								_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global
							%end;
							%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
								_2_Log_Likelihood= _2_Log_Likelihood_Global
								AIC___smaller_is_better_= AIC_Smaller_is_Better_Global
								AICC__smaller_is_better_= AICC_Smaller_is_Better_Global
								BIC___smaller_is_better_= BIC_Smaller_is_Better_Global
								CAIC__smaller_is_better_= CAIC_smaller_is_better_Global
								HQIC__smaller_is_better_= HQIC_smaller_is_better_Global											
							%end;
						%end;
						%else %if %upcase(&RegMethod)=COUNT and %length(&RandEff)=0 %then %do; 		
							_2_Log_Likelihood= _2_Log_Likelihood_Global
							AIC___smaller_is_better_= AIC_Smaller_is_Better_Global
							AICC__smaller_is_better_= AICC_Smaller_is_Better_Global
							BIC___smaller_is_better_= BIC_Smaller_is_Better_Global
							CAIC__smaller_is_better_= CAIC_smaller_is_better_Global
							HQIC__smaller_is_better_= HQIC_smaller_is_better_Global
							Pearson_Chi_Square= Pearson_ChiSquare_Global
							Pearson_Chi_Square___DF= Pearson_ChiSq_DF_Ratio_Global											
						%end;
						%else %if %upcase(&RegMethod)=NOMINAL or %upcase(&RegMethod)=ORDINAL %then %do;
							%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
								_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Global
							%end;
							%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
								_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Global
							%end;
							%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
								_2_Log_Likelihood= _2_Log_Likelihood_Global
								AIC___smaller_is_better_= AIC_Smaller_is_Better_Global
								AICC__smaller_is_better_= AICC_Smaller_is_Better_Global
								BIC___smaller_is_better_= BIC_Smaller_is_Better_Global
								CAIC__smaller_is_better_= CAIC_smaller_is_better_Global
								HQIC__smaller_is_better_= HQIC_smaller_is_better_Global
							%end;	
						%end;
						%else %if (%upcase(&RegMethod)=BINARY and &kmodel=1) %then %do;
							%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
								_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global
							%end;
							%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
								_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Global
								Generalized_Chi_Square= Generalized_ChiSquare_Global
								Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Global											
							%end;
						%end;										
					%end;
					%else %if %index(%upcase(&WhichProc),PHREG) %then %do;
						_2_LOG_L= _2_LOG_L_Global
						AIC= AIC_Global
						SBC= SBC_Global
					%end;
					%else %if %index(%upcase(&WhichProc),LOGISTIC) %then %do;
						AIC= AIC_Global
						SC= SC_Global
						_2_Log_L= _2_Log_L_Global
					%end;
					;
				run;

				/*data OutLib.ma_o1_&w; set ma_o; run;*/					
			%end;

			%if (&ie=0 or %sysfunc(indexw(%upcase(&ClassVars),%upcase(%scan(&IntEff,&ie))))=0) and %upcase(&RegMethod) ne NOMINAL %then %do;
				data ta_o;
					%if &ie=0 %then %do;
						merge 
							pat_o(where=(
								%let first= 1;
								%do i=1 %to &nhaps; 
									%if &first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										&EffName="&&hap&i"
										%let first=0;
									%end;
									%else %if ^&first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										or &EffName="&&hap&i"
									%end;
								%end;
								) 
								keep=&EffName Estimate Stderr
							)
							ta_o(where=(
								%let first= 1;
								%do i=1 %to &nhaps;
									%if &first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										effect ? "&&hap&i"
										%let first= 0;
									%end;
									%else %if ^&first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										or effect ? "&&hap&i"
									%end;
								%end;
								)
							);
					%end;
					%else %do;
						merge 
							pat_o(where=(
								%let first= 1;
								%do i=1 %to &nhaps;
									%if &first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										&EffName ? "&&hap&i*" or &EffName ? "*&&hap&i"
										%let first= 0;
									%end;
									%else %if ^&first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										or &EffName ? "&&hap&i*" or &EffName ? "*&&hap&i"
									%end;
								%end;
								) 
								keep=&EffName Estimate Stderr
							)
							ta_o(where=(
								%let first= 1;
								%do i=1 %to &nhaps;
									%if &first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										Effect ? "&&hap&i*" or Effect ? "*&&hap&i"
										%let first= 0;
									%end;
									%else %if ^&first and ^%index(%upcase(&&hap&i), RARE) %then %do;
										or Effect ? "&&hap&i*" or Effect ? "*&&hap&i" 
									%end;
								%end;
								)
							);									
					%end;
					%if %quote(&EffName)=%str(Variable) %then drop &EffName;;
					rename Estimate = Estimate_Global StdErr= StdErr_Global;
				run;

				%let volcanoplot = 1;

				/*data OutLib.ta_3o&w; set ta_o; run;*/
			%end;

			data ta_o;  
				length Trait $32;
				set ta_o;
				%if &ie>0 %then %do;
					if orig=1
						%do i=1 %to &nhaps;
							 or index(Effect,"&&hap&i*")>0 or index(Effect,"*&&hap&i")>0
						%end;
					;
					/* To make sure interactions in same order for Global and Single tests */
					%do i=1 %to &nhaps; 
						if index(Effect,"&&hap&i*")>1 then do;
							Effect=cats("&&hap&i*",tranwrd(Effect,"&&hap&i*",""));
						end;
						else if index(Effect,"&&hap&i")>1 then do;
							Effect=cats("&&hap&i*",tranwrd(Effect,"*&&hap&i",""));
						end;
					%end;					
				%end;
				%else %do;
					%let first= 1;
					where
					%do i=1 %to &nhaps;
						%if &first and ^%index(%upcase(&&hap&i), RARE) %then %do;
							 index(Effect,"&&hap&i")>0
							 %let first= 0;
						%end;
						%else %if ^&first and ^%index(%upcase(&&hap&i), RARE) %then %do;
							 or index(Effect,"&&hap&i")>0
						%end;
					%end;
					;
				%end;				
			run;

			data ta_o_all;  
				length Trait $32;
				set ta_o_all(in=orig) ta_o;				
				if not orig then Trait= propcase("&TraitVar");
				label Prob&Stat=' '; 
			run;

			/*data OutLib.ta_o_all0_&w; set ta_o_all; run; */ 

			%if &ie=0 %then %do;
				%if %Upcase(&Odds)=YES  %then %do; 

					data or_o_all;
						set or_o_all or;
					run;

					data ta_o_all;
						length Trait $32;
						set ta_o_all; 
						set or_o_all;
						%if ^%length(&RandEff &KVars) %then %do;	
							drop unit;
						%end;
						%else %do;
							drop alpha lower upper;
						%end;							
					run;

					/*data OutLib.ta_o_all1_&w; set ta_o_all; run;*/  
				 %end;

				 %if %upcase(&OutputResid)=YES or %upcase(&OutputPred)=YES %then %do;
					%if %sysfunc(exist(outres)) %then %do;
						/*data OutLib.outres_&w; set outres; run;*/
						
						data outrest_o_all;
							set outrest_o_all(in=orig) outres;
							if orig=0 then do;
								_Trait_=propcase("&TraitVar");
							end;
						run;		
					%end;

					/*data OutLib.outrest_o_all; set outrest_o_all; run;*/
				%end;
			%end;

			%if &pest=1 %then %do;
				data pat_o_all;	
					%if %length(&IntEffBuild) %then %do;
						length Interaction $32;
					%end;
					set pat_o_all(in=orig) pat_o;
					if orig=0 then do;
						Trait=propcase("&TraitVar");
						%if %length(&IntEffBuild) %then %do;
							Interaction="&Interaction";
						%end;
					end;
				run;

				/*data OutLib.pat_o_all; set pat_o_all; run;*/
			%end;			
		%end;
			

		/** Single haplotype tests **/
		%let flag_sh= 0;
		%do i=1 %to &numhaps;
		
			title "Test for Haplotype &&hap&i";

			data ta;
				delete;
			run;
			
			data ct;
				delete;
			run;		

			%if &OutputType ^= None %then %do;
				ods &OutputType select all;
			%end;
			ods output
				&CTable=ct
				&Gtable=ta
				%if &pe=0 %then %do;
					&ParmEst= pat
					%if &WhichProc ne SURVEYREG %then %do;
						nobs = tnobs %if &WhichProc ne PHREG %then (where=(Label ? "Observations Used"));
					%end;
					%if %upcase(&RegMethod)=BINARY %then responseprofile=rp; 
					/*begin: Luciano*/
					%if %upcase(&OutputRSquare)=YES and %index(%upcase(&WhichProc),LOGISTIC) %then
						rsquare= trsq;
					/*end: Luciano*/	
					%if %upcase(&OutputFitStat)=YES %then 
						FitStatistics= fst;
				%end;			
			;
			proc &WhichProc data=outreg;
				%if %length(&ClassVars) %then %do;
					class &ClassVars;
				%end;	
				%if &NeedConstant %then class constant;;
				model &TraitVar &TrtOpt = &ClassVars %if &NeedConstant %then constant; &Covar &&hap&i &ModOption;
				%if &RegMethod=Survival %then %do;
					&&hap&i: test &&hap&i;
				%end;
				%else %do;
					contrast "&&hap&i" &&hap&i 1;
				%end;
				%if (%upcase(&OutputResid)=YES or %upcase(&OutputPred)=YES) %then %do;
					id &IndIDMerg;
				%end;				
				%if %length(&ByVars) %then %do; 
					by &ByVars;
				%end;
			run; 
			quit;
			
			/*
			%if %sysfunc(exist(ma_s))=0 %then %do;
				%put WARNING: Haplotype test could not be performed for Haplotype &&hap&i and Trait &TraitVar in Window &&winval&w...;
				%let sherr&i=1;
			%end;
			%else %do;
				%Nobsvars(ma_s);
				%if &nobs=0 %then %do;
					%put WARNING: Haplotype test could not be performed for Haplotype &&hap&i and Trait &TraitVar in Window &&winval&w...;
					%let sherr&i=1;
				%end;
				%else %do;				
					%let sherr&i=0;
					data ma_s&i;
					 length Dependent $32;
					 set ma_s;
					 Stat=&StatC;
					 ProbGlobal=&ProbC;
					 Dependent=propcase("&TraitVar");
					 keep &ByVars Dependent Stat ProbGlobal &ConVarC;
					 label Stat="&StatCLabel" ProbGlobal="&ProbCLabel";
					run;		

					data ma_s;
						delete;
					run;
				%end;	
			%end;
			*/
			
			%NObsVars(ct);
			%if &nobs %then %do;
				data ct;
					set ct;
					if not missing(&ProbC);
				run;
				%NObsVars(ct);
			%end;
			
			%if &nobs=0 %then %do;
				%put WARNING: PROC &WhichProc generated no estimates for Haplotype &&hap&i and Trait &TraitVar in Window &&winval&w...  There may be excessive missing data or the contrast may be nonestimable. Please see the log for more details.;
				%put WARNING: Haplotype test could not be performed for Haplotype &&hap&i and Trait &TraitVar in Window &&winval&w...;
				%let sherr&i=1;
			%end;
			%else %do;				
				%let sherr&i=0;
				%let flag_sh= 1;
				data ta_s&i;
					*length Dependent $32;
					length Trait $32;
					set ta;
					Trait=propcase("&TraitVar");
					*Dependent=propcase("&TraitVar");
					*Stat=&Stat;
					*ProbGlobalHap=&Prob;
					*keep &ByVars Dependent Stat ProbGlobalHap &ConVarC;
					*label Stat="&StatLabel" ProbGlobalHap="&ProbLabel";
				run;	

				/*data OutLib.ta_s&i; set ta_s&i;run;*/

				data ta;
					delete;
				run;
				
				data pat_s&i;
					*length Dependent $32;
					length Trait $32;
					set pat;
					Trait=propcase("&TraitVar");
					*Dependent=propcase("&TraitVar");
					*Stat=&Stat;	
					*ProbGlobalHap=&Prob;
					*keep &ByVars Dependent Stat ProbGlobalHap;
					*label Stat="&StatLabel" ProbGlobalHap="&ProbLabel";
					rename Estimate= Estimate_Single; /*GenMacro.sas volcano plot needs this variable*/
				run;	

				/*data OutLib.pat_s&i; set pat_s&i; run;*/
				
				data ma_s&i;
					*length Dependent $32;
					length Trait $32;
					set ct;
					Trait=propcase("&TraitVar");
					*Dependent=propcase("&TraitVar");
					/*Stat=&StatC;*/
					&StatC.Single=&StatC;
					/*ProbTrendHap=&ProbC;*/
					&ProbC.Single=&ProbC;
					*ProbGlobalHap=&ProbC;
					*keep &ByVars Dependent Stat ProbGlobalHap &ConVarC;
					drop &StatC &ProbC;
					*label Stat="&StatCLabel" ProbGlobalHap="&ProbCLabel";
				run;
				
				data ma_s&i;
					merge ma_s&i pat_s&i(where= (&EffName= "&&hap&i"));
				run;
				
				data ct;
					delete;
				run;

				/*data OutLib.ma_s&i.&w; set ma_s&i; run;*/
				
				%if %upcase(&OutputFitStat)=YES %then %do;
					%put WhichProc= &WhichProc;

					proc transpose data=fst(
						%if %index(%upcase(&WhichProc),LOGISTIC) %then %do;
							rename=(
								criterion= descr 
								interceptandcovariates= value
							)
						%end;
						%else %if %index(%upcase(&WhichProc),PHREG) %then %do;
							rename=(
								criterion= descr  
								withcovariates= value
							)
						%end;
						)
						out=fst(drop= _name_); 
						var value;
						id descr;
						by &ByVars;
					run;

					/*data OutLib.fst_s1; set fst; run;*/

					data ma_s&i;
						merge ma_s&i fst;
						by &ByVars;
						rename
						%if %index(%upcase(&WhichProc),MIXED) %then %do;
							_2_Res_Log_Likelihood= _2_Res_Log_Likelihood_Single
							AIC__Smaller_is_Better_= AIC_Smaller_is_Better_Single
							AICC__Smaller_is_Better_= AICC_Smaller_is_Better_Single
							BIC__Smaller_is_Better_= BIC_Smaller_is_Better_Single
						%end;
						%else %if %index(%upcase(&WhichProc),GLIMMIX) %then %do;
							%if (%upcase(&RegMethod)=COUNT and %length(&RandEff)>0) 
								or (%upcase(&RegMethod)=BINARY and %length(&RandEff)>0) %then %do;
								%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
									_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Single
									Generalized_Chi_Square= Generalized_ChiSquare_Single
									Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Single
								%end;
								%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
									_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Single
									Generalized_Chi_Square= Generalized_ChiSquare_Single
									Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Single
								%end;
								%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
									_2_Log_Likelihood= _2_Log_Likelihood_Single
									AIC___smaller_is_better_= AIC_Smaller_is_Better_Single
									AICC__smaller_is_better_= AICC_Smaller_is_Better_Single
									BIC___smaller_is_better_= BIC_Smaller_is_Better_Single
									CAIC__smaller_is_better_= CAIC_smaller_is_better_Single
									HQIC__smaller_is_better_= HQIC_smaller_is_better_Single											
								%end;
							%end;
							%else %if %upcase(&RegMethod)=COUNT and %length(&RandEff)=0 %then %do; 		
								_2_Log_Likelihood= _2_Log_Likelihood_Single
								AIC___smaller_is_better_= AIC_Smaller_is_Better_Single
								AICC__smaller_is_better_= AICC_Smaller_is_Better_Single
								BIC___smaller_is_better_= BIC_Smaller_is_Better_Single
								CAIC__smaller_is_better_= CAIC_smaller_is_better_Single
								HQIC__smaller_is_better_= HQIC_smaller_is_better_Single
								Pearson_Chi_Square= Pearson_ChiSquare_Single
								Pearson_Chi_Square___DF= Pearson_ChiSq_DF_Ratio_Single											
							%end;
							%else %if %upcase(&RegMethod)=NOMINAL or %upcase(&RegMethod)=ORDINAL %then %do;
								%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
									_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Single
								%end;
								%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
									_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Single
								%end;
								%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
									_2_Log_Likelihood= _2_Log_Likelihood_Single
									AIC___smaller_is_better_= AIC_Smaller_is_Better_Single
									AICC__smaller_is_better_= AICC_Smaller_is_Better_Single
									BIC___smaller_is_better_= BIC_Smaller_is_Better_Single
									CAIC__smaller_is_better_= CAIC_smaller_is_better_Single
									HQIC__smaller_is_better_= HQIC_smaller_is_better_Single
								%end;	
							%end;
							%else %if (%upcase(&RegMethod)=BINARY and &kmodel=1) %then %do;
								%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
									_2_Res_Log_Pseudo_Likelihood= _2_Res_Log_Pseud_Likel_Single
									Generalized_Chi_Square= Generalized_ChiSquare_Single
									Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Single
								%end;
								%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
									_2_Log_Pseudo_Likelihood= _2_Log_Pseud_Likel_Single
									Generalized_Chi_Square= Generalized_ChiSquare_Single
									Gener__Chi_Square___DF= Gener_ChiSq_DF_Ratio_Single											
								%end;
							%end;										
						%end;
						%else %if %index(%upcase(&WhichProc),PHREG) %then %do;
							_2_LOG_L= _2_LOG_L_Single
							AIC= AIC_Single
							SBC= SBC_Single
						%end;
						%else %if %index(%upcase(&WhichProc),LOGISTIC) %then %do;
							AIC= AIC_Single
							SC= SC_Single
							_2_Log_L= _2_Log_L_Single
						%end;
						;
					run;

					/*data OutLib.ma_s_fit&i; set ma_s&i;  run;*/				

				%end;
								
				%if &pest=1 %then %do;
					data pat_s_all;	
						%if %length(&IntEffBuild) %then %do;
							length Interaction $32;
						%end;
						set pat_s_all(in=orig) pat_s&i;
						if orig=0 then do;
							Trait=propcase("&TraitVar");
							%if %length(&IntEffBuild) %then %do;
								Interaction="&Interaction";
							%end;
						end;
					run;
	
					/*data OutLib.pat_s_all; set pat_s_all; run;*/
				%end;
			%end;	
		%end;

		%if &Perms>0 %then %do;		
			%if %length(&ByVars) %then %do;
				%let i=1;
				%do %while(%length(%scan(&ByVars,&i))>0);
				 %let i=%eval(&i+1);
				%end;
				%let lastby=%scan(&ByVars,%eval(&i-1));
			%end;

			data ma_all;
				delete;
			run;

			%if &haperr=1 %then %goto skipoverall;

			title "Permutations for Overall Test";
			%if &OutputType ^= None %then %do;
				ods &OutputType exclude all;
			%end;
			ods output &CTable=c_perm;		
			proc &WhichProc data=allp;
				%if %length(&ClassVars) %then %do;
					class &ClassVars;
				%end;	
				%if &NeedConstant %then class constant;;
				/*S1149775: removing Rare haplotyple from the model, test and contrast statements, otherwise the contrast is not estimable 
				because all haplotypes sum to 1 (linear combination)*/
				/*model &TraitVar &TrtOpt = &ClassVars &Covar %do i=1 %to &nhaps; &&hap&i %end; &ModOption;*/
				model &TraitVar &TrtOpt = &ClassVars %if &NeedConstant %then constant; &Covar %do i=1 %to &nhaps; %if ^%index(%upcase(&&hap&i), RARE) %then &&hap&i; %end; &ModOption;
				%if &RegMethod=Survival %then %do;
					%if ^%index(%upcase(&hap1), RARE) %then %do;
						Overall: test &hap1 %do i=2 %to &nhaps; %if ^%index(%upcase(&&hap&i), RARE) %then , &&hap&i; %end;;
					%end;
					%else %do;
						Overall: test &hap2 %do i=3 %to &nhaps;, &&hap&i %end;;
					%end;
				%end;
				%else %do;
					%if ^%index(%upcase(&hap1), RARE) %then %do;
						contrast 'Overall' &hap1 1 %do i=2 %to &nhaps; %if ^%index(%upcase(&&hap&i), RARE) %then , &&hap&i 1; %end;; 
					%end;
					%else %do;
						contrast 'Overall' &hap2 1 %do i=3 %to &nhaps;, &&hap&i 1 %end;; 
					%end;
				%end;
				by _sample_ &ByVars;
			run; 
			quit;
			
			%NobsVars(c_perm);
			%if &nobs %then %do;

				%if %length(&ByVars) %then %do;
					proc sort data=c_perm;
					 by &ByVars _sample_;
					run;
				%end;

				data c_perm;
					*length Dependent $32;
					length Trait $32;
					set c_perm;
					*Dependent=propcase("&TraitVar");
					Trait=propcase("&TraitVar");
				run;

				data c_perm;
					merge c_perm ma_o;
					*by Dependent &ByVars;
					by Trait &ByVars;
					drop df ss ms probf;
				run;

				data c_perm;
					set c_perm end=last;
					%if %length(&ByVars) %then %do;
						by &ByVars;
						if first.&lastby then Prob=0;
				 	%end;
					*if &StatC > Stat then Prob+1;
					if &StatC > &StatC.Global then Prob+1;
					%if %length(&ByVars) %then %do;
						if last.&lastby then do;
							Prob_emp = Prob / &Perms;
							output;
						end;			 	
					%end;
					%else %do;
						if last then do;
							Prob_emp = Prob / &Perms;
							output;
						end;
					%end;
					*keep &ByVars Dependent Stat ProbGlobal Prob_emp;
					drop &StatC &ProbC Prob _sample_;
					label Prob_emp="&ProbC_empLabel";
				run;	

				data OverallTest&&winval&w;
					set OverallTest&&winval&w c_perm(in=b);
					if b then GlobalTest_Status = 0;
				run;
				
				%let flag_globalstatus= 1;
			%end;

			%skipoverall:

			%do i=1 %to &numhaps;
				%if &&sherr&i=1 %then %goto skipsh;
				
				data c_perm;
					delete;
				run;				

				title "Permutations for Haplotype &&hap&i (Trait=&TraitVar)";
				%if &OutputType ^= None %then %do;
					ods &OutputType exclude all;
				%end;
				*ods output &Ctable=ma_s;
				ods output &Ctable=c_perm;
				proc &WhichProc data=allp;
					%if %length(&ClassVars) %then %do;
						class &ClassVars;
					%end;	
					%if &NeedConstant %then class constant;;
					model &TraitVar &TrtOpt= &ClassVars %if &NeedConstant %then constant; &Covar &&hap&i &ModOption;
					%if &RegMethod=Survival %then %do;
						&&hap&i: test &&hap&i;
					%end;
					%else %do;
						contrast "&&hap&i" &&hap&i 1;
					%end;
					by _sample_ &ByVars;				  
				run; 
				quit;
				
				%NobsVars(c_perm);
				%if &nobs %then %do;

					%if %length(&ByVars) %then %do;
						*proc sort data=ma_s;
						proc sort data=c_perm;
							by &ByVars _sample_;
						run;
					%end;

					data c_perm;
						*length Dependent $32;
						length Trait $32;
						set c_perm;
						*Dependent=propcase("&TraitVar");
						Trait=propcase("&TraitVar");
					run;

					/*data OutLib.c_c_perm; set c_perm; run;*/

					options label;
					data ma_s&i;
					 merge ma_s&i c_perm end=last;
					 *by Dependent &ByVars;
					 by Trait &ByVars;
					 %if %length(&ByVars) %then %do;
						 if first.&lastby then Prob=0;
					 %end;
					 *if &StatC > Stat then Prob+1;
					 if &StatC > &StatC.Single then Prob+1;
					 %if %length(&ByVars) %then %do;
						if last.&lastby then do;
							Prob_emp = Prob / &Perms;
							output;
						end;			 	
					 %end;
					 %else %do;
						if last then do;
							Prob_emp = Prob / &Perms;
							output;
						end;
					 %end;
					 *keep &ByVars Dependent Stat ProbGlobal Prob_emp &ConVarC;
					 drop &StatC &ProbC Prob _sample_;
					 label Prob_emp="&ProbC_empLabel";
					run;	

					/*data OutLib.c1_ma_s&i; set ma_s&i; run;*/

					data c_perm;
						delete;
					run;

					data ma_all;
					 length Hapname &ConVarC $16;
					 set ma_all ma_s&i(in=new);
					 if new then Hapname=strip(&ConVarC);
					run;	
				%end;

				%skipsh:;
			%end;

			%if &flag_sh %then %do;
				proc sort data=ma_all;
				 by hapname;
				run;

				data ma_all;
				 *length Dependent $32;
				 length Trait $32;
				 merge hapnames ma_all;
				 by hapname;
				 if Prob_Emp=. then Prob_Emp=0;
				 if missing(Trait) then Trait= propcase("&TraitVar");
				 drop &ConVarC;
				run;

				proc sort data=ma_all; 
				 *by &ByVars haplotype ;
				 by &ByVars %if %length(&HFData) %then descending &FreqVar; haplotype;
				run; 
			%end;
		%end;
		%else %do;
			%let flag_globalstatus= 1;
			
			%if &haperr=0 %then %do;
				data OverallTest&&winval&w;
					set OverallTest&&winval&w ma_o;
					GlobalTest_Status= 0;
				run;	
			%end;
			%else %do;	
				data OverallTest&&winval&w;
					length Trait %if &wintype=C %then WindowID; $32;
					Trait= propcase("&TraitVar");
					%if &wintype=C %then %do;
						WindowID= "&&winval&w";
					%end;
					%else %do;
						WindowID= &&winval&w;
					%end;
					GlobalTest_Status= 1;
				run;
			%end;
			
			/*data OutLib.OverallTest1_&&winval&w; set OverallTest&&winval&w; run;*/

			data ma_all; 
			 length Hapname &ConVarC $16;
			 set %do i=1 %to &numhaps; %if &&sherr&i=0 %then ma_s&i; %end;;
			 Hapname=strip(&ConVarC);
			 drop &ConVarC;
			run;

			proc sort data=ma_all;
			 by hapname;
			run;
			
			/*data OutLib.ma_all0; set ma_all; run;*/

			data ma_all;
			 *length Dependent $32;
			 length Trait $32;
			 merge hapnames ma_all;
			 by hapname;
			 *if missing(dependent) then delete;
			 *if missing(Trait) then delete;
			 if missing(Trait) then Trait= propcase("&TraitVar");			 
			run;

			proc sort data=ma_all; 
			 *by &ByVars haplotype ;
			 by &ByVars %if %length(&HFData) %then descending &FreqVar; haplotype;
			run;

		%end;
		
		/*data OutLib.ma_all; set ma_all; run;*/

		data SingleTests&&winval&w;
			set SingleTests&&winval&w ma_all;
		run;
		
		%if &w=1 %then %do;
			%let NTrait= %eval(&NTrait + 1);
			%global _tv&NTrait;
			%let _tv&NTrait = &TraitVar;
		%end;

		%nexttrait:
		%let ti=%eval(&ti+1);

	%end;
	
	%if &w=1 %then %do;
		%put NTrait= &NTrait;
	%end;	
	/* End loop over traits */

	/*data OutLib.OverallTest&&winval&w; set OverallTest&&winval&w; run;*/
	
	%if &flag_globalstatus %then %do;
		data OverallTest_Temp;
			set OverallTest&&winval&w;
			where GlobalTest_Status = 0;
		run;
		%Nobsvars(OverallTest_Temp);
	%end;
	%else %do;
		%Nobsvars(OverallTest&&winval&w);
	%end;
	%let nobs1=&nobs;
	%put nobs1= &nobs1;
	%Nobsvars(SingleTests&&winval&w);
	%let nobs2=&nobs;
	%put nobs2= &nobs2;
	
	proc datasets nolist;
		delete OverallTest_Temp;
	quit;

	%if %eval(&nobs1+&nobs2)>0 %then %do;
		%if &nobs1>0 %then %do;
			%let global_flag= 1;
		
			/*
			title "Global Test Statistic - Adjusted for Any Covariates (WindowID=&&winval&w)";
			proc print data=OverallTest&&winval&w(drop=WindowID) label; run;
			*/
			
			data OutLib.&HaplotypeTrendReg_HRGName;
				set OutLib.&HaplotypeTrendReg_HRGName Overalltest&&winval&w(in=new);
				%if &wintype=C %then %do;
					if new then WindowID = "&&winval&w";				
				%end;
				%else %do;
					if new then WindowID = &&winval&w;
				%end;	
				if new then _TempVar_= 1;
				label WindowID= "WindowID";
			run;
		%end;
		
		%if &nobs2>0 %then %do;
			%let single_flag= 1;
			
			/*
			title "Single Haplotype Test Statistics - Adjusted for Any Covariates (WindowID=&&winval&w)";
			proc print data=SingleTests&&winval&w(drop=WindowID) label; run;
			*/	
			
			%if &firstwind %then %do;
				%let nobs2_cum= 0;
				%let firstwind= 0;
			%end;
				
			data OutLib.&HaplotypeTrendReg_HRSName;
				set OutLib.&HaplotypeTrendReg_HRSName Singletests&&winval&w(in=new);
				%if &wintype=C %then %do;
					if new then WindowID = "&&winval&w";				
				%end;
				%else %do;
					if new then WindowID = &&winval&w;
				%end;
				if new then _TempVar_= 1;
				label WindowID= "WindowID";
			run;	
			
			%let nobs2_cum= %eval(&nobs2_cum + &nobs2);
		%end;
		
		proc datasets nolist;
			delete OverallTest&&winval&w SingleTests&&winval&w;
		quit;
	%end;
	
	%nextwindow:
%end;	

%if &ModelOut = None %then %do;
	ods exclude none;
	ods results;
%end;
%else %if &ModelOut = HTML %then %do;
	ods html close;
	%PackageInsertHtml(html from &ProcessName, &OutPath, &bodyfile, &framefile, &contentsfile);
	ods listing;
%end;
/*%PackageInsertDataset(Global tests data from &ProcessName, outlib, &HaplotypeTrendReg_HRGName,_type=other);*/
/*%PackageInsertDataset(Single tests data from &ProcessName, outlib, &HaplotypeTrendReg_HRSName,_type=other);*/
%if %upcase(&BuildCombined)=YES %then %do;
	%PackageInsertDataset(Wide data from &ProcessName, outlib, &HaplotypeTrendReg_HRWName,_type=other);
%end;

**begin: Creating an empty jsl file to trick ExitSDSProcess to output the jsl file named after the ProcessName*;
%PrepOutFile(HaplotypeTrendReg_Script,HaplotypeTrendReg_ScriptName,&ProcessName Script for &OutName..jsl);
%global _mainjslfile_;
%let _mainjslfile_ = %quote(&HaplotypeTrendReg_Script);

/*
data _null_;
	file "&HaplotypeTrendReg_Script" lrecl=32767;
	put "//!";
	put "Names Default to Here(1);";
	put "Here:DataList = {};";
run;
%PackageInsertFile(JSL Code from &ProcessName, &OutPath, &HaplotypeTrendReg_ScriptName, , text/plain, yes);
**end: Creating an empty jsl file to trick ExitSDSProcess to output the jsl file named after the ProcessName*;
*/

	%global maxlenAG NumGps AnnoOrganism AnnoAccession_All AnnoData 
		ANNOACCESSION GpType /*JBy NonTraitByVars
		Nbyval ByVarsOnly NByVars OrigNbyval JNONTRAITBYVARS*/
		SGroupn1 HRDepData nCallstoJSPV HaploTest;
		
	%if &NTrait > 1 %then %do;
		%let ByVars= Trait &ByVars0;
	%end;
	%else %do;
		%let ByVars= &ByVars0;
	%end;
 		
 	%if &NTrait>1 %then %do;
 		/* This stuff for Venns */
 		%if %length(&ByVars)>0 %then %do;
 			%let ByVars = Trait &ByVars;
 			%let ByVarsOnly = Trait &ByVarsOnly;
 			%let JBy = %str(Trait,&JBy);
 			%let JByOdds= &JBy; 
 		%end;
 		%else %do;
 			%let ByVars = Trait;
 			%let ByVarsOnly = Trait;
 			%let JBy=Trait;
 			%let JByOdds= &JBy; 
 		%end;
 		%let Nbyval = %eval(&Nbyval*&Ntrait);
	%end;		
				
	%let MTCOrig= &MTC;
	%let ConvOrig= &Conv;
			
	**begin: output statistics from global haplotype tests**;
	%put global_flag= &global_flag;
	%if &global_flag %then %do;
		%if &pest=1 %then %do;
			data &HaplotypeTrendReg_HRGE;
				set pat_o_all;
			run;

			%PackageInsertDataset(Parameter estimates data for the global tests from &ProcessName, OutLib, &HaplotypeTrendReg_HRGEName);
		%end;

		%if %upcase(&OutputResid)=YES or %upcase(&OutputPred)=YES %then %do;
			proc sort data=outrest_o_all;
				by &IndIDMerg WindowID _Trait_;
			run;	
		%end;

		%if %upcase(&OutputResid)=YES %then %do;
			proc transpose data=outrest_o_all out =outrestt(drop= _name_ _LABEL_) let
				%if &NTrait>1 %then %do;
					prefix=Res_ delimiter=_Trt_
				%end;
				%else %do;
					prefix=Resid_
				%end;
				;
				var resid;
				id WindowID %if &NTrait>1 %then _Trait_;;
				by &IndIDMerg;
			run;

			data &HaplotypeTrendReg_HRGR;
				set outrestt;
			*	 merge othervars outrestt;
			*	 by IndIDMerg;
			run;

			%PackageInsertDataset(Residual data for the global tests from &ProcessName, OutLib, &HaplotypeTrendReg_HRGRName);
		%end;

		%if %upcase(&OutputPred)=YES %then %do;
			proc transpose data=outrest_o_all out=outpredtt(drop= _name_ _LABEL_) let
				%if &NTrait>1 %then %do;
					prefix=Pred_ delimiter=_Trt_
				%end;
				%else %do;
					prefix=Pred_
				%end;
				;
				var pred;
				id WindowID %if &NTrait>1 %then _Trait_;;
				by &IndIDMerg;
			run;

			data &HaplotypeTrendReg_HRGP;
				set outpredtt;
				drop _LABEL_;
			*	 merge othervars outpredtt;
			*	 by IndIDMerg;
			run;

			%PackageInsertDataset(Predicted data for the global tests from &ProcessName, OutLib, &HaplotypeTrendReg_HRGPName);			
		%end;
		/*end new qk*/	

		%if %upcase(&OutputFitStat)= NO %then %do;
			%global PlotFitVars;
			%let PlotFitVars=0;
		%end;
		%else %if %upcase(&OutputFitStat)=YES %then %do;	
			%global PlotFitVars Group1OverlayVars Group2OverlayVars Group3OverlayVars;
			%let PlotFitVars=1;
			%let Group1OverlayVars=;
			%let Group2OverlayVars=;
			%let Group3OverlayVars=;

			** Global test **;
			%if %index(%upcase(&WhichProc),MIXED) %then %do;
				%let Group1OverlayVars= _2_Res_Log_Likelihood_Global AIC_Smaller_is_Better_Global AICC_Smaller_is_Better_Global BIC_Smaller_is_Better_Global;
			%end; 
			%else %if %index(%upcase(&WhichProc),GLIMMIX) %then %do;
				%if (%upcase(&RegMethod)=COUNT and %length(&RandEff)>0) 
					or (%upcase(&RegMethod)=BINARY and %length(&RandEff)>0) %then %do;
					%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
						%let Group1OverlayVars= _2_Res_Log_Pseud_Likel_Global;
						%let Group2OverlayVars= Generalized_ChiSquare_Global;
						%let Group3OverlayVars= Gener_ChiSq_DF_Ratio_Global;					
					%end;
					%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
						%let Group1OverlayVars= _2_Log_Pseud_Likel_Global;
						%let Group2OverlayVars= Generalized_ChiSquare_Global;
						%let Group3OverlayVars= Gener_ChiSq_DF_Ratio_Global;
					%end;
					%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
						%let Group1OverlayVars= _2_Log_Likelihood_Global AIC_Smaller_is_Better_Global AICC_Smaller_is_Better_Global BIC_Smaller_is_Better_Global CAIC_smaller_is_better_Global HQIC_smaller_is_better_Global;
					%end;
				%end;
				%else %if %upcase(&RegMethod)=COUNT and %length(&RandEff)=0 %then %do; 
					%let Group1OverlayVars= _2_Log_Likelihood_Global AIC_Smaller_is_Better_Global AICC_Smaller_is_Better_Global BIC_Smaller_is_Better_Global CAIC_smaller_is_better_Global HQIC_smaller_is_better_Global;	
					%let Group2OverlayVars= Pearson_ChiSquare_Global;	
					%let Group3OverlayVars= Pearson_ChiSq_DF_Ratio_Global;	
				%end;
				%else %if %upcase(&RegMethod)=NOMINAL or %upcase(&RegMethod)=ORDINAL %then %do;
					%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
						%let Group1OverlayVars= _2_Res_Log_Pseud_Likel_Global;
					%end;
					%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
						%let Group1OverlayVars= _2_Log_Pseud_Likel_Global;
					%end;
					%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
						%let Group1OverlayVars= _2_Log_Likelihood_Global AIC_Smaller_is_Better_Global AICC_Smaller_is_Better_Global BIC_Smaller_is_Better_Global CAIC_smaller_is_better_Global HQIC_smaller_is_better_Global;
					%end;
				%end;
				%else %if (%upcase(&RegMethod)=BINARY and &kmodel=1) %then %do;
					%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
						%let Group1OverlayVars= _2_Res_Log_Pseud_Likel_Global;
						%let Group2OverlayVars= Generalized_ChiSquare_Global;
						%let Group3OverlayVars= Gener_ChiSq_DF_Ratio_Global;
					%end;
					%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
						%let Group1OverlayVars= _2_Log_Pseud_Likel_Global;
						%let Group2OverlayVars= Generalized_ChiSquare_Global;
						%let Group3OverlayVars= Gener_ChiSq_DF_Ratio_Global;										
					%end;
				%end;
			%end;
			%else %if %index(%upcase(&WhichProc),PHREG) %then %do;
				%let Group1OverlayVars= _2_LOG_L_Global AIC_Global SBC_Global;
			%end;
			%else %if %index(%upcase(&WhichProc),LOGISTIC) %then %do;
				%let Group1OverlayVars= AIC_Global SC_Global _2_Log_L_Global;
			%end;

			%put PlotFitVars= &PlotFitVars; 
			%put Group1OverlayVars= &Group1OverlayVars; 
			%put Group2OverlayVars= &Group2OverlayVars; 
			%put Group3OverlayVars= &Group3OverlayVars; 
		%end;	

		/*data OutLib.HaplotypeTrendReg_HRG0; set &HaplotypeTrendReg_HRG; run;*/
		
		%if %length(&ByVars) or &NTrait > 1 %then %do;
			proc sort data= OutLib.&HaplotypeTrendReg_HRGName;
				by Trait &ByVars0 WindowID;
			run;
		%end;

		/*
		%*let ByVars= Trait;
		%if &NTrait > 1 %then %do;
			%let ByVars= Trait &ByVars0;
		%end;
		%else %do;
			%let ByVars= &ByVars0;
		%end;
		*/
		%let MTC= &MTCOrig;
		%let Conv= &ConvOrig;
		%let HTests= %sysfunc(substr(&ProbC, 5))%str(Global); /*GlobalHap*/
		%put HTests= &HTests;
		%Psmooth(&HTests, &HaplotypeTrendReg_HRG, &HaplotypeTrendReg_HRGName);
		%if (&syserr>4) %then %goto exit;
		%let PlotVarsGlobal= &PlotVars;
		%put PlotVarsGlobal= &PlotVarsGlobal;
		%let NPVars= &NPVars;

		/*data OutLib.HaplotypeTrendReg_HRG1; set &HaplotypeTrendReg_HRG; run;*/
		
		%let maxlenAG= 0;
		%let NumGps= 1;
		%let AnnoOrganism= ;
		%let AnnoAccession_All= ;
		%let AnnoGroup= _TempVar_;
		%let AnnoLoc= WindowID;
		%let AnnoLabel= &ConVarC;
		/*
		%*let ByVars= Trait &ByVars0;
		%let JBy= &ByVars;
		%let NbyVal= &NTrait;
		%let ByVarsOnly=;
		%let NByVars= 0;
		%let OrigNbyval= 1;
		%let JNONTRAITBYVARS=;
		*/
		/*
 		%if &NTrait>1 %then %do;
 			**This stuff for Venns*;
 			%if %length(&ByVars)>0 %then %do;
 				%let ByVars = Trait &ByVars;
 				%let ByVarsOnly = Trait &ByVarsOnly;
 				%let JBy = %str(Trait,&JBy);
 				%let JByOdds= &JBy; 
 			%end;
 			%else %do;
 				%let ByVars = Trait;
 				%let ByVarsOnly = Trait;
 				%let JBy=Trait;
 				%let JByOdds= &JBy; 
 			%end;
 			%let Nbyval = %eval(&Nbyval*&Ntrait);
		%end;		
		*/
		%let nCallstoJSPV= 0;
		%let HRDepData=;
		%let HaploTest= Global;
		%JSLPvalMap(Haplotype Q-K, &HaplotypeTrendReg_Script, &HaplotypeTrendReg_ScriptName, &HaplotypeTrendReg_HRGName, mod=0, _summchart=1, _volplot=0); 
	%end;
	%else %do;
		proc datasets library=outlib nolist;
			delete &HaplotypeTrendReg_HRGName;
		quit;
	%end;
	%let HRDepData_all= &HRDepData;
	
	**begin: output statistics from single haplotype tests**;
	/*data OutLib.HaplotypeTrendReg_HRS_01; set &HaplotypeTrendReg_HRS; run;*/

	%let modsingle= 0;	
	%if &global_flag %then %do;
		data _null_;
			file "&HaplotypeTrendReg_Script" mod lrecl=32767;
			put "////// JSL Single Starts Here //////";
		run;
		%let modsingle= 1;
	%end;
	
	/*data OutLib.HaplotypeTrendReg_HRS_0; set &HaplotypeTrendReg_HRS; run;*/
	
	proc sort data=&HaplotypeTrendReg_HRS;
		by WindowID %if %length(&HFData) %then descending &FreqVar; Hapname Trait;
	run;
	
	options label; /*must use options label here, otherwise labels won't appear in the output file*/
	data &HaplotypeTrendReg_HRS;
		set &HaplotypeTrendReg_HRS;
		by WindowID %if %length(&HFData) %then descending &FreqVar; Hapname Trait;
		if _n_ = 1 then do;	
			_HaplotypeID_= 0;
		end;
		if first.Hapname then _HaplotypeID_= _HaplotypeID_ + 1;
		retain _HaplotypeID_;
		%if %length(&ClassVars) %then %do;
			drop &ClassVars;
		%end;
		/*%if %length(&HFData) %then drop &FreqVar;;*/
		%if &Perms>0 %then %do;
			label Prob_emp="&ProbC_empLabel"; /*must put label here, otherwise it won't appear in the output file*/
		%end;
	run;
	
	/*data OutLib.HaplotypeTrendReg_HRS_1; set &HaplotypeTrendReg_HRS; run;*/
	
	%if %upcase(&OutputFitStat)= NO %then %do;
		%global PlotFitVars;
		%let PlotFitVars=0;
	%end;
	%else %if %upcase(&OutputFitStat)=YES %then %do;	
		
		%global PlotFitVars Group1OverlayVars Group2OverlayVars Group3OverlayVars;
		%let PlotFitVars=1;
		%let Group1OverlayVars=;
		%let Group2OverlayVars=;
		%let Group3OverlayVars=;
				
		** Single test **;
		
		%if %index(%upcase(&WhichProc),MIXED) %then %do;
			%let Group1OverlayVars= &Group1OverlayVars _2_Res_Log_Likelihood_Single AIC_Smaller_is_Better_Single AICC_Smaller_is_Better_Single BIC_Smaller_is_Better_Single;
		%end; 
		%else %if %index(%upcase(&WhichProc),GLIMMIX) %then %do;
			%if (%upcase(&RegMethod)=COUNT and %length(&RandEff)>0) 
				or (%upcase(&RegMethod)=BINARY and %length(&RandEff)>0) %then %do;
				%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Res_Log_Pseud_Likel_Single;
					%let Group2OverlayVars= &Group2OverlayVars Generalized_ChiSquare_Single;
					%let Group3OverlayVars= &Group3OverlayVars Gener_ChiSq_DF_Ratio_Single;					
				%end;
				%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Log_Pseud_Likel_Single;
					%let Group2OverlayVars= &Group2OverlayVars Generalized_ChiSquare_Single;
					%let Group3OverlayVars= &Group3OverlayVars Gener_ChiSq_DF_Ratio_Single;
				%end;
				%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Log_Likelihood_Single AIC_Smaller_is_Better_Single AICC_Smaller_is_Better_Single BIC_Smaller_is_Better_Single CAIC_smaller_is_better_Single HQIC_smaller_is_better_Single;
				%end;
			%end;
			%else %if %upcase(&RegMethod)=COUNT and %length(&RandEff)=0 %then %do; 
				%let Group1OverlayVars= &Group1OverlayVars _2_Log_Likelihood_Single AIC_Smaller_is_Better_Single AICC_Smaller_is_Better_Single BIC_Smaller_is_Better_Single CAIC_smaller_is_better_Single HQIC_smaller_is_better_Single;	
				%let Group2OverlayVars= &Group2OverlayVars Pearson_ChiSquare_Single;	
				%let Group3OverlayVars= &Group3OverlayVars Pearson_ChiSq_DF_Ratio_Single;	
			%end;
			%else %if %upcase(&RegMethod)=NOMINAL or %upcase(&RegMethod)=ORDINAL %then %do;
				%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Res_Log_Pseud_Likel_Single;
				%end;
				%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Log_Pseud_Likel_Single;
				%end;
				%else %if %upcase(&GMethod)=QUAD or %upcase(&GMethod)=LAPLACE %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Log_Likelihood_Single AIC_Smaller_is_Better_Single AICC_Smaller_is_Better_Single BIC_Smaller_is_Better_Single CAIC_smaller_is_better_Single HQIC_smaller_is_better_Single;
				%end;
			%end;
			%else %if (%upcase(&RegMethod)=BINARY and &kmodel=1) %then %do;
				%if %upcase(&GMethod)=RSPL or %upcase(&GMethod)=RMPL %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Res_Log_Pseud_Likel_Single;
					%let Group2OverlayVars= &Group2OverlayVars Generalized_ChiSquare_Single;
					%let Group3OverlayVars= &Group3OverlayVars Gener_ChiSq_DF_Ratio_Single;
				%end;
				%else %if %upcase(&GMethod)=MSPL or %upcase(&GMethod)=MMPL %then %do;
					%let Group1OverlayVars= &Group1OverlayVars _2_Log_Pseud_Likel_Single;
					%let Group2OverlayVars= &Group2OverlayVars Generalized_ChiSquare_Single;
					%let Group3OverlayVars= &Group3OverlayVars Gener_ChiSq_DF_Ratio_Single;										
				%end;
			%end;
		%end;
		%else %if %index(%upcase(&WhichProc),PHREG) %then %do;
			%let Group1OverlayVars= &Group1OverlayVars _2_LOG_L_Single AIC_Single SBC_Single;
		%end;
		%else %if %index(%upcase(&WhichProc),LOGISTIC) %then %do;
			%let Group1OverlayVars= &Group1OverlayVars AIC_Single SC_Single _2_Log_L_Single;
		%end;
		
		%put PlotFitVars= &PlotFitVars; 
		%put Group1OverlayVars= &Group1OverlayVars; 
		%put Group2OverlayVars= &Group2OverlayVars; 
		%put Group3OverlayVars= &Group3OverlayVars; 
	%end;
	%*let PlotFitVars=0; /*when 0, prevents from plotting fit stats for single tests */
	
	%if %length(&ByVars0) or &NTrait > 1 %then %do;
		proc sort data= OutLib.&HaplotypeTrendReg_HRSName;
			by Trait &ByVars0 WindowID %if %length(&HFData) %then descending &FreqVar; haplotype;
		run;
	%end;	
	
	/*
	%*let ByVars= Trait;
	%if &NTrait > 1 %then %do;
		%let ByVars= Trait &ByVars0;
	%end;
	%else %do;
		%let ByVars= &ByVars0;
	%end;
	*/
	%let MTC= &MTCOrig;
	%let Conv= &ConvOrig;
	%let HTests= %sysfunc(substr(&ProbC, 5))%str(Single); /*TrendHap*/;
	%put HTests= &HTests;
	%Psmooth(&HTests, &HaplotypeTrendReg_HRS, &HaplotypeTrendReg_HRSName);
	%if (&syserr>4) %then %goto exit;
	%let PlotVarsSingle= &PlotVars;
	%put PlotVarsSingle= &PlotVarsSingle;
	%let NPVars= &NPVars;
	
	%let maxlenAG= 0;
	%let NumGps= 1;
	%let AnnoOrganism= ;
	%let AnnoAccession_All= ;
	%let AnnoGroup= _TempVar_;
	%let AnnoLoc= _HaplotypeID_;
	%let AnnoLabel= Haplotype;
	/*
	%*let ByVars= Trait &ByVars0;
	%let JBy= &ByVars;
	%let NbyVal= &NTrait;
	%let ByVarsOnly=;
 	%let NByVars= 0;
 	%let OrigNbyval= 1;
 	%let JNONTRAITBYVARS=;
 	*/
 	/*	
 	%if &NTrait>1 %then %do;
 		** This stuff for Venns *;
 		%if %length(&ByVars)>0 %then %do;
 			%let ByVars = Trait &ByVars;
 			%let ByVarsOnly = Trait &ByVarsOnly;
 			%let JBy = %str(Trait,&JBy);
 			%let JByOdds= &JBy; 
 		%end;
 		%else %do;
 			%let ByVars = Trait;
 			%let ByVarsOnly = Trait;
 			%let JBy=Trait;
 			%let JByOdds= &JBy; 
 		%end;
 		%let Nbyval = %eval(&Nbyval*&Ntrait);
	%end;
 	*/
 	%let nCallstoJSPV= 1; 
 	%let HaploTest= Single;
	%JSLPvalMap(Haplotype Q-K, &HaplotypeTrendReg_Script, &HaplotypeTrendReg_ScriptName, &HaplotypeTrendReg_HRSName, mod=&modsingle, _summchart=1, _volplot=1);
	
	%if &global_flag %then %do; 
		%let HRDepData_all= &HRDepData_all, &HRDepData;
	%end;
	%else %do;
		%let HRDepData_all= &HRDepData;
	%end;
	
	%if &pest=1 %then %do;
		data &HaplotypeTrendReg_HRSE;
			set pat_s_all;
		run;
		
		%PackageInsertDataset(Parameter estimates data for the single tests from &ProcessName, OutLib, &HaplotypeTrendReg_HRSEName);
	%end;
	
	%put HRDepData_all= &HRDepData_all;
	%PackageInsertFile(jsl code from &ProcessName, &OutPath, &HaplotypeTrendReg_ScriptName, "&HRDepData_all", text/plain, yes);

**S1199692: clear title to avoid showing titles on other processes*;
title;

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend ;

* run the macro;
%HapTrendReg;
