/*
JMP Genomics Analytical Process

Name: Impute Missing Genotypes
Author: Luciano Silva

Description:
		The Impute Missing Genotypes process imputes numeric missging marker genotypes (0, 1, or 2) for diploid organisms using 
		the k-nearest neighbor imputation (kNNi) or the linkage disequilibrium k-nearest neighbor imputation (LD-kNNi) methods.  
		LD between markers is computed using the the ALLELE procedure in the SAS/Genetics, distances between samples are computed
		using the DISTANCE procedure in the SAS/STAT, and k-nearest neighbor samples is computed using the MODECLUS procedure in 
		the SAS/STAT.		
*/

/*
-----------------------------------------------------------
Copyright (C) 2017 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

/*References: 
Daniel Money, Kyle Gardner, Zo Migicovsky, Heidi Schwaninger, Gan-Yuan Zhong, and Sean Myles. 
(2015) LinkImpute: Fast and Accurate Genotype Imputation for Nonmodel Organisms. G3. 5: 2383-2390.
*/

*ProcessVariables;

%global StudyName InData AlleleVars IDVars ToKeepVars ListMarkerVars WhereStmt  
AnnoInData AnnoLabel AnnoMName AnnoGroup AnnoLoc AnnoAccession AnnoWhereStmt
PloidyNum NumbMrkNBHD NumbIndNBHD WUnit LDNeighbors WMarkerDist WLocDist 
KMeansClusterMethod CorrRadius NumberOfClusters
OutPath ServerOutPath OutName MacroPath ImpMissGeno_OutData ImpMissGeno_Script 
;

*ProcessBody;

%macro ImputeMissingGenotypes;

%if %symexist(missgID) %then %do;
	%let missgID = %sysfunc(close(&missgID));
%end;
%if %symexist(sampID) %then %do;
	%let sampID = %sysfunc(close(&sampID));
%end;

%include "&MacroPath./GenMacros.sas" / nosource;

* Set to blank vars needed for GenMacros *;
%global MarkerFmt ByVars AnnoMergeVars FullInPath;
%let MarkerFmt= %str(Numeric Genotypes);
%let ByVars = ;
%let FullInPath = &InData;

%GenInit(ImputeMissingGenotypes,28);
%if &exiterror=1 %then %goto exit;
%global FullAPName;
%let FullAPName=%str(Impute Missing Genotypes);

%if %length(&ListMarkerVars &AlleleVars)=0 %then %do;
	%let ErrorMessage= Marker variables must be specified in either the Marker Variables or List-Style Specification of Marker Variables parameters.; 
	%put ERROR: &ErrorMessage;
	%let exiterror=1;
	%goto exit;
%end;

%if %length(&PloidyNum) = 0 %then %let PloidyNum= 2;

%if &PloidyNum > 2 and %upcase(&LDNeighbors) = YES %then %do;
	%let ErrorMessage= Ploidy Number greater than 2 cannot be analyzed with the option Select Neighbor Markers with strongest LD within Baseline Unit specified in the Options tab.; 
	%put ERROR: &ErrorMessage; 
	%let exiterror=1;
	%goto exit;
%end;

%let OrigWunit= &Wunit;
%let MaxDist=0;
%if %upcase(&Wunit) = MARKERS %then %do;
	%let Wunit= Markers;
	%let MaxDist= &WMarkerDist;
%end;
%else %if %upcase(&Wunit) = %upcase(Annotation Location Variable) %then %do;
	%let Wunit= Location Variable unit;
	%let MaxDist= &WLocDist;
%end;
%else %do;
	%let ErrorMessage= The Baseline Unit &Wunit is incorrect.;
	%put ERROR: &ErrorMessage; 
	%let exiterror = 1;
	%goto exit;
%end;
%put MaxDist= &MaxDist;

%if %upcase(&Wunit) = MARKERS and %upcase(&LDNeighbors) = YES and &MaxDist > 0 %then %do;
	%if &NumbMrkNBHD > &WMarkerDist %then %do;
		%let ErrorMessage= Number of Neighbor Markers cannot be larger than Marker Units. Please, redefine one or both parameters in the Options tab.;
		%put ERROR: &ErrorMessage;
		%let exiterror= 1;
		%goto exit;
	%end;
%end;

%let OrigNumbMrkNBHD= &NumbMrkNBHD;
%if &NumbMrkNBHD = 0 %then %do;
	%let NumbMrkNBHD= 1000000000;
%end;
%put NumbMrkNBHD= &NumbMrkNBHD;

%if %quote(&Wunit)=%str(Location Variable unit) and %length(&AnnoLoc) = 0 %then %do;
	%*if &MaxDist > 0 %then %do;
		%let ErrorMessage= Annotation Location Variable Baseline Unit cannot be used without specifying an Annotation data set with a valid Annotation Location Variable.;
		%put ERROR: &ErrorMessage;
		%let exiterror = 1;
		%goto exit;
	%*end;
%end;

%let OrigMaxDist= &MaxDist;
%if &MaxDist = 0 %then %do;
	%if %upcase(&LDNeighbors) = YES %then %do;
		%put WARNING: Linkage disequilibrium will be computed between each imputing target marker and all other markers.;
		%if %quote(&Wunit)=%str(Location Variable unit) %then %do;
			%let ErrorString1= %str(The option to Select Neighbor Markers with strongest LD within Baseline Unit cannot be checked when Annotation Location Variable Units is &MaxDist..);
			%let ErrorString2= %str(To include all markers listed in the Marker Variables and List-Style Specification of Marker Variables when computing distance between samples, set Annotation Location Variable Units to zero and uncheck the option to select neighbor markers with strongest LD within Baseline Unit.);
		%end;
		%else %do;
			%let ErrorString1= %str(The option to Select Neighbor Markers with strongest LD within Baseline Unit cannot be checked when Marker Units is &MaxDist..);
			%let ErrorString2= %str(To include all markers listed in the Marker Variables and List-Style Specification of Marker Variables when computing distance between samples, set Marker Units to zero and uncheck the option to select neighbor markers with strongest LD within Baseline Unit.);
		%end;
		%*put ERROR: &ErrorString1 &ErrorString2;
		%*let exiterror=1;
		%*goto exit;
	%end;
	%else %do;
		%put WARNING: All markers in the input data set will be used to compute distance between samples.;
	%end;
	%if %length(&WLocDist) > 0 %then %do;
		%put WARNING: The option Annotation Location Variable Units will be ignored.;
	%end;
	%else %if %length(&WMarkerDist) > 0 %then %do;
		%put WARNING: The option Marker Units will be ignored.;
	%end;
	%else %if %length(&KMeansClusterMethod) > 0 %then %do;
		%put WARNING: The option K-Means Cluster Units.;
	%end;	
	
	%*let MaxDist= 1000000000;
%end;
%put OrigMaxDist= &OrigMaxDist;
%put MaxDist= &MaxDist;

* data with imputated genotypes. Cannot put this before calling %GenInit *;
%PrepOutDS(ImpMissGeno_OutData, ImpMissGeno_OutDataName, &OutName,_img);

* need to combine AlleleVars and ListMarkerVars into AlleleVars for %PrepMapandBy *;
%let OrigAlleleVars = &AlleleVars;
%let AlleleVars = &AlleleVars &ListMarkerVars;
%let XLVARSTMT= ;

%global ReorderingVars;
* set ReorderingVars to empty to avoid certain conflict when and AnnoLoc is empty and call %GenSubset within %PrepMapandBy;
%let ReorderingVars = ; 
%if %upcase(&LDNeighbors)= YES %then %do;
	%if %length(&AnnoLoc)>0 %then %let ReorderingVars = &AnnoGroup &AnnoLoc &AnnoLabel;
%end;
 
* AnnoWhereStmt statement applied in the %PrepMapandBy *;
* If AnnoWhereStmt is not empty, then AlleleVars is updated in the %PrepMapandBy *;
* ds is created within %PrepMapandBy *;
%PrepMapandBy;
run;
%if &exiterror=1 %then %goto exit;

* create index for observations if there are no IDVars *;
%let NewIDVar= ;
%if %length(&WhereStmt) %then %do;
	data indatn;
		set &ds;
		id_old= _n_;
	run;
	%let NewIDVar= id_old;
	
	data indatn;
		set indatn;
		%*if %length(&WhereStmt) %then %do;
			where &WhereStmt;
		%*end;
		keep &NewIDVar &IDVars &AlleleVars;
	run;
	
	%Nobsvars(indatn);
	%if &nobs=0 %then %do;
		%let ErrorMessage= There are no observations in the input data set that meet the criteria in the Where Clause.;
		%put ERROR: &ErrorMessage; 
		%let exiterror=1;
		%goto exit;
	%end;		
	
	%let dsnew= indatn;
%end;
%else %do;
	%let dsnew= &ds;
%end;

%Nobsvars(&dsnew);
%let NInd= &nobs;
%put NInd= &Nind;

%let MaxNumbIndNBHD= &NumbIndNBHD;
%if &MaxNumbIndNBHD > &NInd %then %do;
 	* Number of neighbor samples must be smaller or equal to the sample size after filtering observations *;
	%put WARNING: Number of neighbor samples is greater than number of individuals.;
	%put Number of neighbor samples= &MaxIndNBHD;
	%put Number of individuals= &NInd;
	%let MaxNumbIndNBHD = &NInd;
%end;
%else %if (&MaxNumbIndNBHD < &NInd) %then %do;
	* Number of neighbor samples must be large enough to account for possible discarted neighbor samples with *;
	* missing genotypes at the target imputing marker. I am adding 10% extra samples as a safe marging *;
	%let MaxNumbIndNBHD = %sysfunc(min(&NInd, %sysevalf(1.10 * &NumbIndNBHD)));
%end;
%put MaxNumbIndNBHD= &MaxNumbIndNBHD;

* collect markers that have missing genotypes *;
proc means data=&dsnew nmiss max noprint;
	var &AlleleVars;
	output out=mstat;
run;

/*data OutLib.mstat; set mstat; run;*/

data nstat maxst;
	set mstat;
	if _stat_="N" then output nstat;
	if _stat_="MAX" then output maxst;
	drop _Type_ _Freq_ /*_Stat_*/;
run;

/*data OutLib.maxst; set maxst; run;*/
/*data OutLib.nstat; set nstat; run;*/

data _null_;
	set maxst;
	call symputx("ObsPloidyNum", max(of &AlleleVars));
run;
%put ObsPloidyNum= &ObsPloidyNum;

%if &ObsPloidyNum > &PloidyNum %then %do;
	%let ErrorMessage= The specified Ploidy Number (&PloidyNum) is smaller than the observed maximum allele count (&ObsPloidyNum) found in the Input SAS Data Set.; 
	%put ERROR: &ErrorMessage;
	%let exiterror=1;
	%goto exit;
%end;

proc transpose data=nstat(drop=_stat_) out=nstat(rename=(_name_= _marker_ col1= NIndiv));
	var _all_;
run; 

/*data OutLib.nstatt; set nstat; run;*/

data mrkmissg;
	set nstat;
	where NIndiv < &NInd;
	*where 0;
run;
/*data OutLib.mrkmissg; set mrkmissg; run;*/

data mnames; 
	set &dsnew(keep= &AlleleVars);
	where 0;
run;
/*data OutLib.mnames0; set mnames; run;*/

%let AnnoMNameTmp= &AnnoMName; 
%if ^%length(&AnnoMName) %then %do;
	%let AnnoMNameTmp= _marker_;
%end;	
%put AnnoMNameTmp= &AnnoMNameTmp;
	
proc transpose data=mnames out=mnamest(rename=(_name_= &AnnoMNameTmp));
run;

data mnamest;
	set mnamest;
	_tmpname_= "_a_"||strip(_n_); 
run;

* Luciano, assigning _marker_ holding input marker names to AnnoMNameTmp, and use it in place of AnnoLabel*;
* mapdata is created by %PrepMapandBy *;
data mapdata;
	merge mapdata %if ^%length(&AnnoMName) %then mnamest;;
	%if &OrigMaxDist = 0 %then %do;
		* assigning all markers to linkage group 1 to garantee that all markers will be used when OrigMaxDist = 0*;
		_LGroupTmp_= 1;		
	%end;
	_ord_= _n_;
run;
%let OrigAnnoGroup= &AnnoGroup;
%if &OrigMaxDist = 0 %then %do;
	%let AnnoGroup= _LGroupTmp_;
%end;

/*data OutLib.mapdata0; set mapdata; run; */

%if %upcase(&Wunit)= %upcase(Location Variable unit) and &OrigMaxDist > 0 %then %do;
	**S1370720: checking if Annotation Location Variable Units is too small**;
	
	proc sort data=mapdata out=mapdatas;
		by %if %length(&AnnoGroup) %then &AnnoGroup; &AnnoLoc;
	run;
	
	%let TooBigMaxDist= 0; 
	data checkmaxdist;
	set mapdatas;
	_lagAnnoLoc_= lag(&AnnoLoc);
	%if %length(&AnnoGroup) %then %do;
		*by &AnnoGroup &AnnoLoc;
		by &AnnoGroup;
		if first.&AnnoGroup then do;
			_diff1_= 0;
			_lagAnnoLoc_= 0;
		end;
		else do;
			_diff1_= &AnnoLoc - _lagAnnoLoc_;
		end;
		
	%end;
	%else %do;
		if _n_=1 then _diff1_= 0;
		else _diff1_= &AnnoLoc - _lagAnnoLoc_;
	%end;
	if _diff1_ > &MaxDist then do;
		_flag_= 1;
		call symputx("TooBigMaxDist", 1);
		stop;
	end; 
	run;
	%put TooBigMaxDist= &TooBigMaxDist;
	
	/*data OutLib.ip_checkmaxdist; set checkmaxdist; run;*/

	%if &TooBigMaxDist %then %do;
		%let ErrorMessage= The value specified in the Annotation Location Variable Units field is too small, making it impossible to choose neighbor markers. Please, specify a larger value for the Annotation Location Variable Units.;
		%put ERROR: &ErrorMessage;
		%let exiterror=1;
		%goto exit;
	%end;
%end;

* If there is no genetic map (phisical or linkage), we may cluster markers before computing LD to reduce computational burden *;
%if %upcase(&WUnit) = CLUSTER %then %do;
	
	%let ErrorMessage= Baseline unit not yet available.;
	%put ERROR: &ErrorMessage;
	%let exiterror = 1;
	%goto exit;

	******** Luciano, I have to check this entire section *********;

	* get first and last marker name and create new liststyle that will not *;
	%let tid = %sysfunc(open(mnames));
	%let nobs = %sysfunc(attrn(&tid,NOBS));
	%let nvars = %sysfunc(attrn(&tid,NVARS));
	%put nObs =&nObs nVars=&nVars;
	%let fmvar = %sysfunc(varname(&tid,1));
	%let lmvar = %sysfunc(varname(&tid,&nvars));
	%let tid = %sysfunc(close(&tid));

	%if &nvars<2 %then %do;
		%if &nvars = 0 %then %do;
			%let ErrorMessage= The data set contains no markers to be analyzed.  At least two markers are needed to impute missing genotypes.;
			%put ERROR: &ErrorMessage;
		%end;
		%else %do;
			%let ErrorMessage= The data set only contains one marker to be analyzed.  At least two markers are needed to impute missing genotypes.;
			%put ERROR: &ErrorMessage;
		%end;
		%let exiterror = 1;
		%goto exit;
	%end;

	/* add here and make marker list data set.  if no annotation is used, build temp annotation from markerlist*/
	%let _tid_ = %sysfunc(open(mnames));
	%let _nVars_ = %sysfunc(attrn(&_tid_,NVARS));
	
	data markerlist;
		length _MarkerName_ $100; 
		%if %index(%Upcase(&MarkerFmt), ALLELES) %then %do;
			%let _nm_ = %sysfunc(floor(%eval(&_nVars_/2)));
			%put There are &_nm_ Markers;
			%if %length(&AnnoInData) %then %do;
				%do i = 1 %to &_nm_;
					_MarkerName_ = "&&AMName&i";
					output;
				%end;			
			%end;
			%else %do;
				%do i = 1 %to &_nm_;
					_MarkerName_ = "Marker_&i";
					output;
				%end;
			%end;
		%end;
		%else %do;
			%do i = 1 %to &_nVars_;
				_MarkerName_ = "%sysfunc(varname(&_tid_,&i))";
				output;
			%end;
		%end;
	run;
	
	*close opened file*;
	%let _rc_ = %sysfunc(close(&_tid_)); 
	
	%let Annodat = ;
	/*kemicl added to make data set for anno to avoid errors with allele format */
	%if %length(&AnnoInData) %then %do;
		%let Annodat = AnnoLib.&AnnoData;
		%if %length(&AnnoWhereStmt) %then %do;
			data adat; 
				set AnnoLib.&AnnoData;
				where &AnnoWhereStmt;
			run;
			%let Annodat = adat;

			%Nobsvars(&Annodat);
			%if &syserr>4 or &nobs=0 %then %do;
				%let ErrorMessage= You have entered an invalid Filter to Include Markers or no markers remained after filtering.;
				%put ERROR: &ErrorMessage;
				%let exiterror=1;
				%goto exerr;
			%end;			
			
		%end;
		/*data outlib.Annodat; set &Annodat; run;*/
	%end;
	
	/* Use k-means clustering to create groups of markers. Linkage Disequilibrium will then be computed only within each group*/
	%if %upcase(&Wunit) = CLUSTERS %then %do;
	
		%if %length(&AnnoInData) = 0 %then %do;
			%let Annodat = markerlist;
			%let AnnoMName = _MarkerName_;
		%end;
	
		%put Annodat= &Annodat;
		data AnnoTmp;
			set &Annodat;
		run;
		/*data OutLib.AnnoTmp; set AnnoTmp; run;*/
		/*data OutLib.markerlist; set markerList; run;*/
		
		proc sort data=AnnoTmp;
			by &AnnoMName;
		run;
		
		/*ds is a data set with input genoytpe. transpose input data set to conform with KMeansClustering format*/
		%let fclusdat = gen_trans;
		%let fclusvar_prefix= Ind;
		proc transpose data= &dsnew out=&fclusdat prefix= &fclusvar_prefix;
			var &AlleleVars;
		run;
	
		data &fclusdat;
			set &fclusdat;
			_ordg_=_n_;
		run;
				
		proc sort data=&fclusdat;
			by _name_;
		run;
	
		*Merge fclusdat with Annodat if SkeletonVar is present in either Annotation or Framework Map data sets*;
		data &fclusdat;
			merge &fclusdat(in=a) AnnoTmp(rename=(&AnnoMName=_name_) keep=&AnnoMName);
			by _name_;
			if a;
		run;
		
		proc sort data=&fclusdat out=orgfclusdat;
			by _ordg_;
		run;
		
		/*create variables to be used by KMeansClustering in what follows*/
		%global ListVars;
		%let ListVars= &fclusvar_prefix:;
		%put ListVars= &ListVars;
		
		/*columns become rows after transpositon. Centering and/or scaling markers. It is a good 
		practice to center and scale rows to avoid differences in clustering outcomes. If, for 
		instance, markers in a backcross data were coded as 1 and 2, or 0 and 1, these would
		produce different results if they were not centered and scaled.*/
		%let CenterRows = Yes; 
		%let ScaleRows = Yes; 
		%let CenterRows = No; 
		%let ScaleRows = No; 
		
		* center and/or scale rows;
		%if &CenterRows=Yes or &ScaleRows=Yes %then %do;
			data ind_dat;
				set &fclusdat;
				%if &CenterRows=Yes %then %do;
					_rowmean_ = mean(of &ListVars);
				%end;
				%else %do;
					_rowmean_ = 0;
				%end;
				%if &ScaleRows=Yes %then %do;
					_rowstd_ = std(of &ListVars);
				%end;
				%else %do;
					_rowstd_ = 1;
				%end;
				array v{*} &ListVars;
				do i=1 to dim(v);
					if _rowstd_ > 0 then v[i] = (v[i] - _rowmean_)/_rowstd_;
					else v[i] = .;
				end;
				drop i;
			run;	
			%let fclusdat = ind_dat;		
		%end;	
		
		%global ImpMissGeno_KMD_OutData ImpMissGeno_KMC_OutData;
		%PrepOutDS(ImpMissGeno_KMD_OutData, ImpMissGeno_KMD_OutDataName, &OutName,_kmdi);
		%PrepOutDS(ImpMissGeno_KMC_OutData, ImpMissGeno_KMC_OutDataName, &OutName,_kmci);
		
		%let FullInPath = &ClientOutPath.&ImpMissGeno_KMC_OutDataName..sas7bdat;
			
		* determine radius as a function of a correlation;
		%if %index(&KMeansClusterMethod,Radius) %then %do;
			data std_dat;
				set &fclusdat;
				_rowuss_ = uss(of &ListVars);
				keep _rowuss_;
			run;
			
			proc means data=std_dat noprint;
				var _rowuss_;
				output out=stats(drop=_TYPE_);
			run;
			
			data stats;
				set stats;
				if _STAT_ in ("MEAN");
				call symputx('mean_uss',_rowuss_);
			run;
			
			data _null_;
				call symputx('Radius',sqrt(2 * &mean_uss * (1 - &CorrRadius)));
			run;
			
			%put mean_uss = &mean_uss;
		%end;
		/*data outlib.fclusdat; set &fclusdat; run;*/
	
		ods exclude all;
		ods noresults;
		proc fastclus data=&fclusdat 
				out=kmd(rename=(cluster=_KMCluster_ distance=_KMClusterDistance_)) 
				outseed=kmc(rename=(cluster=_KMCluster_))
				%if %index(&KMeansClusterMethod,Radius) %then %do; 
					%put radius=&Radius;
					radius=&Radius 
					maxclusters=100000
				%end;
				%else %do; 
					%put maxclusters=&NumberOfClusters;
					maxclusters=&NumberOfClusters 
				%end; 
			;
		   	var &ListVars; 
		run;
		ods exclude none;
		ods results;
		
		%if &syserr %then %do;
			%let exiterror = 1;
			%goto exit;
		%end;	
		
		/*data outlib.kmd0; set kmd; run; */
		/*data outlib.kmc0; set kmc; run;  */
			
		/*This is a more general code than need for the marker reduction within the linkage groups AP.
		In the linkage groups AP it is mandatory to replace centroids with the original marker values,
		otherwise the remaining code will produce errors. Nevertheless, this following code has options
		to either replace the means of the centroids with values of distances from the input data, or to 
		use the centroid mean values them selves.*/
		%let UseOrginalValues = Yes;
		%let ReplaceMeans = No;
		%if %upcase(&UseOrginalValues)=YES %then %do;
			/*data outlib.orgfclusdat; set orgfclusdat; run;*/
			
			proc sql UNDO_POLICY=NONE;
				create table kmd(drop=_name2_) as
				select *
				from orgfclusdat(rename=(_name_= _name2_)) as a 
				right join 
				kmd(keep = %if %length(&SkeletonVar) %then &SkeletonVar; _name_ _KMCluster_ _KMClusterDistance_) as b
				on a._name2_= b._name_
				;
			quit;
			
			/*data outlib.kmd1;set kmd;run;*/
			
			*proc sort data=kmd out=os;
			proc sort data=kmd;
				by %if %length(&SkeletonVar) %then &SkeletonVar; _KMCluster_ _KMClusterDistance_ _name_;
			run;
						
			data os kmd;
				*set os;
				set kmd;			
				by %if %length(&SkeletonVar) %then &SkeletonVar; _KMCluster_ _KMClusterDistance_ _name_;
				_KMClusterCenter_= 0;
				if first._KMCluster_ then do;
					_KMClusterCenter_= 1;
					output os;
				end;
				output kmd;
			run;
			
			/*data outlib.os; set os; run;*/
			
			*merging by row number*;
			%let misscluster= 0;
			data kmc;
				merge kmc(drop= &ListVars) os; 
				by %if %length(&SkeletonVar) %then &SkeletonVar; _KMCluster_;
				if missing(_KMCluster_) then do;
					call symputx("misscluster", 1);
					delete;
					*delete missing(_KMCluster_) is necessary to eliminate clusters of monomorphic markers*;
				end;
				if _freq_ = 0 then delete;
			run;
			%if &misscluster %then %do;
				%put WARNING: Monomorphic markers were found in the Input Data Set and they were removed from the analysis.;
			%end;
		%end;
		
		/*get representative markers from each cluster and save them to a new gendat.*/
		proc transpose data=kmc out=gendat(drop=_name_);
			var &ListVars;
			id _name_;
		run;
		
		*create annotation data set with markers, clusters, and distance*;
		%if %length(&AnnoIndata) %then %do;		
			proc sort data=kmd out=kmd(rename=(_name_= &AnnoMName));
				by _name_;
			run;
			
			data kmdgen OutLib.&ImpMissGeno_KMD_OutDataName(drop= &ListVars _ordg_);
				merge AnnoTmp kmd(in=a);
				by &AnnoMName;
				if a;
			run;
			
			proc sort data=OutLib.&ImpMissGeno_KMD_OutDataName;
				by %if %length(&SkeletonVar) %then &SkeletonVar; _KMCluster_ _KMClusterDistance_ &AnnoMName;
			run;
		%end;
		%else %do;
			data kmdgen(rename=(_name_= _MarkerName_)) OutLib.&ImpMissGeno_KMD_OutDataName(keep=_MarkerName_ _KMCluster_ _KMClusterDistance_ _KMClusterCenter_ %if %length(&SkeletonVar) %then &SkeletonVar;);
				set kmd(rename=(_name_= _MarkerName_));
			run;
		%end;
	*end: k-means clustering*;
	%end;	
%end;

%put GenoCol= &GenoCol;
*compute LD between markers to select strongest LD markers for imputation*;
%if %upcase(&LDNeighbors)= YES %then %do;

	* nummark is created by %PrepMapandBy *;
	%put nummark = &nummark;
	%if &nummark < 2 %then %do;
		%let ErrorMessage= Two or more markers must be selected.;
		%put ERROR: &ErrorMessage;
		%goto exit;
	%end;

	* nummark is created by %PrepMapandBy *;
	%*if &nummark > 1000 %then %do;
		* If there are too many markers, then computing the LD between all pairwise markers up front instead *; 
		* of computing the LD between each marker with missing genotype and all other markers in the marker loop *;
		* can be prohibitive due to very large output data set. However, computing LD within the marker loop *;
		* can increase computational time *;
		%let LDByTargetMarker = 1;
	%*end;
	%*else %do;
		%*let LDByTargetMarker = 0;
	%*end;
	
	%let LDMeasures= CorrCoeff;
	%let Haplo = NONE;
	%if %length(&Haplo)=0 %then %let Haplo=est;
	%let CalcRho= No;
	%if %upcase(&Haplo) = NONE %then %let CalcRho= No; 
	
	%if ^&LDByTargetMarker %then %do;

		/*%global UseGrid OnlyDS CreateHTML ChartHaploFreqs ContPlotMeas PlotLDDecay LD_Measures LD_Stats LD_ContourPlot;*/
		%let UseGrid= No;
		%let OnlyDS= Yes;
		%let LD_Measures= out_ldm;
		%let LD_Stats= out_lds;
		%let LD_ContourPlot= out_ldc;

		%if %upcase(&OnlyDS)=YES %then %do;
			%let CreateHTML= No;
			%let ContPlotMeas= ;
			%let PlotLDDecay= ;
			%let ChartHaploFreqs= ; 
		%end;
		%else %do;
			%let CreateHTML= No;
			%let ContPlotMeas = ;
			%let PlotLDDecay = CorrCoeff;
			%let ChartHaploFreqs= No;
		%end; 

		* remove dataset "ldmeasures" to avoid error when running the 2nd time;
		proc datasets library=work nolist; 
			delete ldmeasures ldstats; 
		quit;

		*Perform LD calculations for all pairs within annotation groups*;
		%let AllPairs= Yes;

		data _null_;
			* mapdata is created by %PrepMapandBy *;
			set mapdata;
			call symput('HapGp'||left(_n_),strip(&AnnoGroup));
		run;

		* GenoCol is created by %GenInit *;
		%if %length(&GenoCol) %then %do;
			%let maxavlen=0;
			%let dsid=%sysfunc(open(count));
			* nummark is created by %PrepMapandBy *;
			%do m=1 %to &nummark;
				%let templen=%sysfunc(varlen(&dsid,&m));
				%if &templen > &maxavlen %then %let maxavlen = &templen;
			%end;
			%let rc = %sysfunc(close(&dsid));

			data indat;
				set &dsnew;
				%*if %length(&WhereStmt) %then %do;
				*	where &WhereStmt;
				%*end;
				array av{&nummark} &AlleleVars;
				array _a_{&nummark} %if %upcase(&MarkerFmt) ne %str(NUMERIC GENOTYPES) %then $&maxavlen;;
				do i=1 to &nummark;
					%if %upcase(&MarkerFmt) ne %str(NUMERIC GENOTYPES) %then %do;
						if missing(av{i}) then _a_{i}='';
						else _a_{i}=strip(av{i});
					%end;
					%else %do;
						if missing(av{i}) then _a_{i}=.;
						else _a_{i}=av{i};
					%end;    
				end;
				%if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) %then %do;
					* format is created by %GenInit *;
					format _a_: numtochar.;
				%end;
				* Luciano, included NewIDVar and IDVars*;
				keep &NewIDVar &IDVars &ByVars _a_1-_a_&nummark; 
			run;
			%*let nmv=&nummark;
		%end;
		%else %do;
			%let maxavlen=0;
			%let dsid=%sysfunc(open(count));
			* n2snps is created by %PrepMapandBy *;
			%do m=1 %to &n2snps;
				%let templen=%sysfunc(varlen(&dsid,&m));
				%if &templen > &maxavlen %then %let maxavlen = &templen;
			%end;
			%let rc = %sysfunc(close(&dsid));

			data indat;
				set &dsnew;
				%*if %length(&WhereStmt) %then %do;
				*	where &WhereStmt;
				%*end;
				array av{&n2snps} &AlleleVars;
				array _a_{&n2snps} $&maxavlen;
				do i=1 to &n2snps;
					if missing(av{i}) then _a_{i}='';
					else _a_{i}=strip(av{i});
				end;
				* Luciano, included NewIDVar and IDVars *;
				keep &NewIDVar &IDVars &ByVars _a_1-_a_&n2snps; 
			run;
			%*let nmv=&n2snps;
		%end;

		/*data OutLib.indat0; set indat; run;*/

		* When NeedLDM=0 the output data set with LD measure will not be created, and only a output data set with chisq and probchisq will be created *;
		* The output data set with LD measure will be large when number of markers is large, which can take a lot of disk space *;
		* For the pourpose of finding strongst LD markers, the probchisq statistic is enough *;
		%if %upcase(&CalcRho)=NO and (%upcase(&OnlyDS)=YES or (%length(&ContPlotMeas)=0 and %length(&PlotLDDecay)=0 and %upcase(&ChartHaploFreqs)=NO)) 
			%then %let NeedLDM = 0;
		%else %let NeedLDM=1;
	
		%put NeedLDM = &NeedLDM;
	
		%let firstloop=1;

		* Grid stuff *;
		%let bp=1;
		%let max_tries = 100;
		%let waitlist = ;
		%if %upcase(&UseGrid)=YES %then %do;
			%let jmpgindex = %index(&USER_HOME,JMPG);
			%let LSUserHome = %substr(&USER_HOME,1,%eval(&jmpgindex-1))%str(JMPLS/);
			filename gofile "&LSUserHome.GridOptions.sas";
			%if %sysfunc(fexist(gofile)) %then %do;
				%include gofile;

				/* The grdsvc_enable call will go out to the SAS Metadata Server and   */
				/* find the SAS Grid Server definition. A return code of 0 means that  */
				/* all signons will use the grid. A non-0 return code means that there */
				/* is a problem that should be investigated.                           */
				/* %global rc; */
				%let rc=%sysfunc(grdsvc_enable(_all_,resource="SASApp"));
				/* %rc_check(routine=grdsvc_enable,rc=&rc); */
				%if &rc %then %do;
					%let exiterror = 1;
					%goto exit;
				%end;
			%end;
			%else %do;
				%let ErrorMessage= &LSUserHome.GridOptions.sas not found. Please fill out the Grid tab in the Configure Life Sciences Settings process under the File menu to create this file.;
				%put ERROR: &ErrorMessage;
				%let exiterror = 1;
				%goto exit;
			%end;

			/* The grdsvc_nnodes call provides the number of grid nodes        */
			/* available in the grid.                                          */
			%let nnodes=%sysfunc(grdsvc_nnodes("SASApp"));
			%put number of available nodes = &nnodes;

			/*
			%let maxnodes = 10; %* ^WorkflowBuilder_ParmsVal[20]^;
			%if (&maxnodes > 0) and (&nnodes > &maxnodes) %then %do;
				%put restricting maximum number of nodes to &maxnodes.;
				%let nnodes = &maxnodes;
			%end;
			*/

			/* You can view the progress of the signons using the xlsbatch tool or Grid Manager. */
			/* Watch for job names such as SASGrid:xxxx where xxxx is the value of */
			/* the sysjobid of this SAS session.                                   */
			%put Job Name=SASGrid:&sysjobid;

			%let libn=;
		%end;
		%else %do;
			* use libname for grid in proc allele statement*;
			%let libn=;
		%end;

		%do i=1 %to &NumGps;
			%put i=&i of &NumGps;
			%*goto loopend;

			data indat_sub;
				set indat;
				%if &NumGps=1 %then %do;
					keep &ByVars _a_:; 
				%end;
				%else %do;
					keep &ByVars 
					%do j=1 %to &nummark;
						%if %quote(&&HapGp&j)=%quote(&&Groupn&i) %then %do;
							%if %length(&GenoCol) %then %do;
								_a_&j
							%end;
							%else %do;
								%let j1=%eval(2*&j-1);
								%let j2=%eval(2*&j);
								_a_&j1 _a_&j2
							%end;
						%end;
					%end;
					;
				%end;
			run;

			data map_sub&i;
				set mapdata end=_end_;
				%if (&GpType=C) %then %do;
					where &AnnoGroup="&&Groupn&i";
				%end;
				%else %do;
					where &AnnoGroup=&&Groupn&i;
				%end;
				%if %quote(&Wunit)=%str(Location Variable unit) %then %do;	
					/* Workaround for bug in Proc Allele - problems when ByVars, Location var used 
					and the last marker is not within MaxDist of 2nd to last */
					%if %length(&ByVars) %then %do;
						lagloc = lag(&AnnoLoc);
						if _end_ and abs(&AnnoLoc - lagloc) > &MaxDist then &AnnoLoc = lagloc + &MaxDist;
					%end;
					rename &AnnoLoc = Location;
				%end;
				%else %if %upcase(&AnnoLoc)=LOCATION %then %do;
					rename &AnnoLoc=_Location_;
				%end;
				keep &AnnoLabel &AnnoLoc &AnnoMNameTmp;
			run;		 

			%Nobsvars(map_sub&i);
			%if &nobs<2 %then %goto loopend;

			***************************************************************************;
			%let BlockSize = %sysfunc(min(%sysevalf(3000000/&nind,floor),&nobs));
			%*let BlockSize = %sysfunc(min(%sysevalf(10000/&nind,floor),&nobs)); /*Luciano, testing pourpose only*/
			%put BlockSize = &BlockSize;

			%let first = 1;
			%let last = &BlockSize;

			* Luciano, added AllPairs in the if statement to avoid redundant estimation of LD *;
			%if %length(&MaxDist)=0 or %quote(&Wunit)=%str(Location Variable unit) or %upcase(&AllPairs) = YES %then %let MD = 0;
			%else %if &MaxDist > &BlockSize %then %let MD = &BlockSize;
			%else %let MD = &MaxDist;
			%put MD= &MD;

			data map_sub&i;
				set map_sub&i;
				%let first1 = 1;
				%let last1 = &BlockSize;
				%let Block=1;

				%do %until (&&first&Block > &nobs - &MD);
					if _n_ >=&&first&Block and _n_ <= &&last&Block then _Block_&Block = 1;
					%if %quote(&Wunit)=%str(Location Variable unit) %then %do;
						if _n_ = &&first&Block then call symput("firstlocation&Block",strip(location));
					%end;
					%let first%eval(&Block + 1) = %eval(&&last&Block + 1 - &MD);
					%let last%eval(&Block + 1) = %eval(&&last&Block + &BlockSize);
					%let Block = %eval(&Block + 1);
					%put first&Block &&first&Block lastb &&last&Block;
					%put first&Block &&first&Block %eval(&nobs - &MD);
					%if &&last&Block > &nobs %then %let last&Block = &nobs;
					%put first&Block &&first&Block lastb &&last&Block;
				%end;
			run;
			/*data OutLib.map_sub0_&i; set map_sub&i; run;*/

			%let NBlocks = %eval(&Block-1);
			%put NBlocks = &NBlocks;

			%if %quote(&Wunit)=%str(Location Variable unit) %then %do;
				* Need to set again to figure out how far backwards need to set first marker of block to be sure to cover all markers *;
				data map_sub&i;
					set map_sub&i;
					array foundfirst{&NBlocks} _TEMPORARY_ (&NBlocks*0);
					%do b=2 %to &NBlocks;
						if &&firstlocation&b - location le &MaxDist then do;
							_Block_&b = 1;
							if foundfirst{&b} = 0 then do;
								call symput("first&b",strip(_n_));
								foundfirst{&b}=1;
							end;
						end;
					%end;
				run;
			%end;	

			data %if %length(&ByVars)>0 %then byvars(keep=&ByVars); avars&i(keep=_a_:);
				set indat_sub;
				array getname{*} _a_:;
				%do Block=1 %to &NBlocks;
					%if %length(&GenoCol) %then %do;
						_f = vname(getname(&&first&block));
						_l = vname(getname(&&last&Block));
					%end;
					%else %do;
						_f = vname(getname(%eval(2*&&first&block-1)));
						_l = vname(getname(%eval(2*&&last&Block)));
					%end;

					call symput("firstvar&Block",_f);
					call symput("lastvar&Block",_l);			

				%end;
			run;	

			%if %length(&ByVars)>0 %then %do;
				/* Need to make sure AlleleVars are in consecutive variables */
				data avars&i;
					set byvars; set avars&i;
				run;
			%end;

			/*data OutLib.avars&i; set avars&i; run;*/
			/*data OutLib.map_sub1_&i; set map_sub&i; run;*/

			%do blocki=1 %to &NBlocks;
				%if %upcase(&AllPairs)=YES %then %do;
					%let LastBlock=&NBlocks;
				%end;
				%else %let LastBlock=&blocki;

				%do blockj=&blocki %to &LastBlock;

					%put Calculating LD for block &blocki with block &blockj.;

					%if %upcase(&AllPairs)=YES %then %do;
						* Luciano, use AnnoMNameTmp instead of AnnoLabel *;
						%let pastmt = %str(data=&libn.avars&i ndata=&libn.map_sub&i(rename=(&AnnoMNameTmp=name) where=(_Block_&Blocki=1 or _Block_&Blockj=1)));
						%*let pastmt = %str(data=&libn.avars&i ndata=&libn.map_sub&i(rename=(&AnnoLabel=name) where=(_Block_&Blocki=1 or _Block_&Blockj=1)));

						%let varstmt = %str(var &&firstvar&Blocki -- &&lastvar&Blocki;);  
						%if &Blocki ne &Blockj %then %let varstmt = &varstmt %str(with &&firstvar&Blockj -- &&lastvar&Blockj;);
					%end;
					%else %do;
						* Luciano, use AnnoMNameTmp instead of AnnoLabel *;
						%let pastmt = %str(data=&libn.avars&i ndata=&libn.map_sub&i(rename=(&AnnoMNameTmp=name) where=(_Block_&Blocki=1)));
						%*let pastmt = %str(data=&libn.avars&i ndata=&libn.map_sub&i(rename=(&AnnoLabel=name) where=(_Block_&Blocki=1)));

						%let varstmt = %str(var &&firstvar&Blocki -- &&lastvar&Blocki;);
					%end;

					%if %length(&MaxDist)=0 %then %let MaxDist2=&BlockSize;
					%else %let MaxDist2=&MaxDist;	

					%if &NumGps>1 %then %ClockStart(%quote(&AnnoGroup &&Groupn&i, Block &blocki with &blockj));

					%if %upcase(&UseGrid)=YES %then %do;
						/************* GRID *********************/

						%let node=%eval(%sysfunc(mod(%eval(&bp-1),&nnodes))+1);
						%let waitlist=&waitlist gnode&node;
						%let tries = 1;
						%let finished = 0;
						%do %while(^&finished);
							%put Sign on attempt &tries of &max_tries to grid node &node....;
							signon gnode&node;
							%let node_name = ;
							%let node_name = %sysfunc(grdsvc_getname(gnode&node));
							%* %put node_name = &node_name; 
							%if %length(&node_name) %then %do;
								%put Successful sign on to GNODE&node = &node_name;
								%let finished = 1;
							%end;
							%else %do;
								%let tries = %eval(&tries+1);
								%if &tries > &max_tries %then %do;
									%let ErrorMessage= Grid sign-on to GNODE&node failed after &max_tries tries.;
									%put ERROR: &ErrorMessage;
									%let exiterror = 1;
									%let finished = 1;
									signoff _all_;
									%*goto cv_exit;
								%end;
							%end;
						%end;

						%if %sysfunc(sysprod(prodnum007)) ne 0 %then %do; 
							*S1313431: SAS/CONNECT be licensed*;
							%macro ConnectVars01;
								%syslput i = &i;
								%syslput NeedLDM = &NeedLDM;
								%syslput pastmt = &pastmt;
								%syslput AllPairs = &AllPairs;
								%syslput LDMeasures = &LDMeasures;
								%syslput CalcRho = &CalcRho;
								%syslput Haplo = &Haplo;
								%syslput MaxDist2 = &MaxDist2;
								%syslput GenoCol = &GenoCol;
								%syslput ByVars = &ByVars;
								%syslput varstmt = &varstmt;
								%syslput ProcessName = LinkageDisequilbrium&node; 
								%syslput BP = &bp;		
								%syslput Blocki = &Blocki;
								%syslput Blockj = &Blockj;
							%mend;
						%end;
						%else %do;
							%macro ConnectVars01;
								%put EMPTY.;	
							%mend;
						%end;
						%ConnectVars01;

						/* rsubmit block of work to be submitted to grid node */
						rsubmit cwait=no persist=no;

						%put rsubmit for &ProcessName, block pair &bp;
						%put node number = &node;
						%put node name = &node_name;
						%put pastmt = &pastmt;

						/* copy data sets to work */
						proc upload data=avars&i out=avars&i;
						run;
						proc upload data=map_sub&i out=map_sub&i;
						run;

						%if &NeedLDM=1 %then %do;
							data ldm&bp;
								delete;
							run;
						%end;

						ods output %if &NeedLDM=1 %then LDMeasures=ldm&bp;;
						ods listing close;
						proc allele &pastmt gmin=1 &LDMeasures %if %upcase(&CalcRho)=YES %then rho; outstat=out&bp 
							%if &Blocki ne &Blockj %then (where=(Locus1 ne Locus2)); haplo=&Haplo maxdist=&MaxDist2 &GenoCol lognote; 
							%if %length(&ByVars) %then %do; 
								by &ByVars;
							%end;
							&varstmt
						run;	

						%if %sysfunc(sysprod(prodnum007)) ne 0 %then %do; 
							*S1313431: SAS/CONNECT be licensed*;
							%macro ConnectVars02;
								%nrstr(%sysrput syserr&BP = &syserr);
							%mend;
						%end;
						%else %do;
							%macro ConnectVar02;
								%put EMPTY.;
							%mend;
						%end;
						%ConnectVars02;

						proc download data=out&bp out=out&bp;
						run;
						%if &NeedLDM=1 %then %do;
							proc download data=ldm&bp out=ldm&bp;
							run;
						%end;

						endrsubmit;
						/*****************************************/
					%end;
					%else %do;
						/* Same PROC ALLELE code as above */
						%if &NeedLDM=1 %then %do;
							data ldm&bp;
								delete;
							run;
						%end;

						ods output %if &NeedLDM=1 %then LDMeasures=ldm&bp;;
						ods listing close;
						proc allele &pastmt gmin=1 &LDMeasures %if %upcase(&CalcRho)=YES %then rho; outstat=out&bp
							%if &Blocki ne &Blockj %then (where=(Locus1 ne Locus2)); haplo=&Haplo maxdist=&MaxDist2 &GenoCol lognote; 
							%if %length(&ByVars) %then %do; 
								by &ByVars;
							%end;
							&varstmt
						run;	
					%end;

					%if (&syserr>4) %then %do;
						%if &NumGps>1 %then %goto blockend;
						%else %goto exit;
					%end;

					%if %upcase(&UseGrid)=NO %then %do;
						%if &NumGps>1 %then %ClockStop(%quote(&AnnoGroup &&Groupn&i, Block &blocki with &blockj));
						%if %upcase(&OnlyDS)=NO %then ods listing;;

						%if &NeedLDM=1 %then %do;
							%NObsVars(ldm&bp);
							%if &nobs=0 %then %do;
								%put WARNING: LD calculations were unable to be performed for current block.;
								%goto blockend;
							%end;

							proc datasets nolist;
								%if &firstloop=0 %then %do;
									append base=ldmeasures data=ldm&bp force;
								%end;
								%else %do;
									change ldm&bp=ldmeasures;
								%end;
							run;

							/*data OutLib.ldmeasures; set ldmeasures; run;*/
						%end;

						proc datasets nolist; 
							%if &firstloop=0 %then %do;
								append base=ldstats data=out&bp force;
								delete out&bp;
							%end;
							%else %do;
								change out&bp=ldstats;
							%end;
						quit;

						/*data OutLib.ldstats&blocki.&blockj; set ldstats; run;*/
					%end;

					%let firstloop=0;  

					%blockend:;
					%let bp=%eval(&bp+1);
				 %end;
			%end;

			%loopend:;  
		%end;

		%let NumBp = %eval(&bp-1);
		%put NumBp = &NumBp;

		%if %upcase(&UseGrid)=YES %then %do;
			* WAIT FOR ALL RSUBMITS TO FINISH;
			%put waitlist = &waitlist;
			waitfor _ALL_ &waitlist;

			%do n=1 %to &NumBP;
				%put syserr&n = &&syserr&n;
				%if &&syserr&n > 4 %then %do;
					%let ErrorMessage= Block pair &n did not run successfully.;
					%put ERROR: &ErrorMessage;
					%let exiterror=1;
					%goto exit;
				%end;
			%end;

			data ldstats;
				set out1-out&NumBp;
			run;

			%if &NeedLDM=1 %then %do;	
				data ldmeasures;
					set ldm1-ldm&NumBP;
				run;
			%end;

		%end;

		%if &NeedLDM=1 %then %do;
			%if %sysfunc(exist(ldmeasures))=0 %then %do;	
				%let ErrorMessage= LD calculations were unable to be performed.;
				%put ERROR: &ErrorMessage;
				%goto exit;
			%end;
		%end;	

		%if %upcase(&CreateHTML)=YES %then %do;	
			ods html 
				body = "&LD_Body" (url="&bodyfile")
				frame = "&LD_Frame"  (url="&framefile")
				contents = "&LD_Contents" (url="&contentsfile")
			;
			ods listing close;

			%if &NeedLDM=1 %then %do;
				proc print data=ldmeasures(drop=ControlVar) label;
				 title 'LD Measures';
				run;
			%end;

			proc print data=ldstats label;
				title 'Test for LD';
			run;
			ods html close; 
		%end;

		ods listing;
		options nolabel;

		proc sort data=ldstats;
			by Locus1;
		run;

		%if %upcase(&ChartHaploFreqs)=YES %then %do;
			data ldmeasures;
				set ldmeasures;
				Count=round(2*NIndiv*Freq);
			run;
		%end;

		proc datasets library=work nolist;
			modify ldstats;
				rename Locus1=Marker1 Locus2=Marker2;
			modify mapdata;
				* Luciano, use AnnoMNameTmp instead of AnnoLabel *;
				rename &AnnoMNameTmp= Marker1 _tmpname_= _tmpname1_;
				*rename &AnnoLabel=Marker1;
		quit;

		proc sort data=mapdata;
			by Marker1;
		run;

		%if &NeedLDM = 1 %then %do;	
			proc sort data=ldmeasures out=&LD_Measures(rename=(Locus1-Locus2=Marker1-Marker2) drop=ControlVar) noduprecs;
				by Locus1;
			run;

			/*data OutLib.&LD_Measures; set &LD_Measures; run;*/

			data temp1m; 
				merge &LD_Measures(in=inc) mapdata;
				by Marker1;
				Locus1=&AnnoLoc+0;
				label Locus1 = ' ';
				drop &AnnoLoc;
				l1=lag(Marker1);
				l2=lag(Marker2);
				if Marker1 ne l1 or Marker2 ne l2 or missing(Marker1) or missing(Marker2);
				if inc;
				%if %length(&AnnoAccession) %then %do;
					%if %quote(&AnnoAccession) = %quote(&AnnoLabel) %then %do;
					/* Since renamed AnnoLabel var above, need to re-set AnnoAccession if it is not the same var */
					_annoacc1 = Marker1;
					%end;
					%else %do;
						_annoacc1 = &AnnoAccession;
					%end;
				%end;
			run;
		%end;

		/*data OutLib.ldstats; set ldstats; run;*/
		/*data OutLib.mapdata; set mapdata; run;*/

		data temp1;
			merge ldstats(in=inc) mapdata;
			by Marker1;
			Locus1=&AnnoLoc+0;
			label Locus1 = ' ';
			drop &AnnoLoc;
			if inc;
		run;

		%*let exiterror=1;
		%*goto exit;

		proc datasets nolist;
			delete ldstats;
		run; quit;

		proc sort data=temp1;
			by Marker2;
		run;

		data mapdata(keep=Locus2 Marker2 _tmpname2_) mapdatam(keep=Locus2 Marker2 _tmpname2_ %if %length(&AnnoAccession) %then _annoacc2;);
			set mapdata;
			Locus2=&AnnoLoc+0;
			%if %length(&AnnoAccession)>0 %then %do;
				%if %quote(&AnnoAccession) = %quote(&AnnoLabel) %then %do;
				/* Since renamed AnnoLabel var above, need to re-set AnnoAccession if it is not the same var */
					_annoacc2 = Marker1;
				%end;
				%else %do;
					_annoacc2 = &AnnoAccession;	
				%end;
			%end;
			rename Marker1=Marker2 _tmpname1_=_tmpname2_;
		run;
		/*data OutLib.mapdata1; set mapdata; run; */

		%if &NeedLDM = 1 %then %do;	
			proc sort data=temp1m;
				by Marker2;
			run;	

			data temp2m;
				merge temp1m(in=inc) mapdatam;	
				by Marker2;
				if inc;
				keep &ByVarsOnly Locus1 Locus2 Marker1 Marker2 &ContPlotMeas &PlotLDDecay &AnnoGroup D &LDMeasures 
					%if %length(&AnnoAccession)>0 %then _annoacc1 _annoacc2; %if %upcase(&CalcRho)=YES %then rho k_rho;;
			run;	

			proc sort data=temp2m;
				by &ByVars &AnnoGroup Locus1 Marker1 Locus2 Marker2;
			run;	
		%end;	

		data temp2;
			%if %quote(&AnnoAccession) = %quote(&AnnoLabel) %then %do;
				length Marker1 Marker2 &AnnoAccession $&AnnoLen;
			%end;
			%if %quote(&AnnoAccession) = %quote(&AnnoLoc) %then %do;
				length Marker1 Marker2 $&AnnoLen &AnnoAccession $32;
			%end;	
			merge temp1(in=inc) mapdata;
			by Marker2;
			label ProbChi=' ';
			if inc;
			%if %length(&AnnoAccession)>0 %then %do;
				if Test ne "HWE" then &AnnoAccession=' ';
				%if %quote(&AnnoAccession) = %quote(&AnnoLabel) %then %do;
				/* Since renamed AnnoLabel var above, need to re-set AnnoAccession if it is not the same var */
					else &AnnoAccession = Marker1;
				%end;
				%else %if %quote(&AnnoAccession) = %quote(&AnnoLoc) %then %do;
				/* Since renamed AnnoLoc var above, need to re-set AnnoAccession if it is not the same var */
					else &AnnoAccession = strip(Locus1);
				%end;		
			%end;
			keep &ByVarsOnly Locus1 Locus2 Marker1 Marker2 Distance Test NIndiv ChiSq ProbChi &AnnoGroup &AnnoAccession _tmpname1_ _tmpname2_;
		run;

		proc datasets nolist;
			delete temp1;
		run; quit;

		proc sort data=temp2 noduprecs;
			by &ByVars &AnnoGroup Locus1 Marker1 Locus2 Marker2;
		run;

		options label;
		data &LD_Stats;
			merge temp2 
				%if %upcase(&CalcRho)=YES %then temp2m; 
				%if %length(&AnnoAccession) %then(drop=_annoacc1 _annoacc2);
				;
			by &ByVars &AnnoGroup Locus1 Marker1 Locus2 Marker2;
			temp=Locus1;
			tl=Marker1;
			if Locus1 gt Locus2 then do;
				Locus1=Locus2+0;
				Locus2=temp;
				Marker1=Marker2;
				Marker2=tl;
			end;
			Distance=abs(Locus2-locus1);
			if Test="LD" and NIndiv>0 then CorrCoeff2 = chisq/NIndiv;
			else CorrCoeff2 = .;
			length ColorVar $11;
			if ProbChi=. then ColorVar="Missing";
			else if ProbChi<0.01 then ColorVar="[0.00,0.01)";
			else if ProbChi<0.05 then ColorVar="[0.01,0.05)";
			else ColorVar="[0.05,1.00]";
			label ColorVar="P-Value Range";
			drop temp tl;
		run;
		/*data OutLib.&LD_Stats; set &LD_Stats; run;*/

		proc datasets nolist;
			delete temp2;
		run; quit;

		/* Need this extra sort to get everything in right order after switching Locus 1<->2 - in case not in right order in anno data set */
		proc sort data=&LD_Stats;
			by &ByVars &AnnoGroup Locus1 Marker1 Locus2 Marker2;
		run;

		%if (%length(&ContPlotMeas)>0 or %length(&PlotLDDecay)>0) %then %do;
			%let ContPlotMeas=%upcase(&ContPlotMeas);
			%let LDMeas=%upcase(&PlotLDDecay);

			data &LD_ContourPlot;
				set temp2m temp2m(in=switch);
				temp=Locus1;
				tl=Marker1;
				if switch then do;
					Locus1=Locus2+0;
					Locus2=temp;
					Marker1=Marker2;
					Marker2=tl;
					%if %length(&AnnoAccession) %then &AnnoAccession = _annoacc2;;
				end;
				%if %length(&AnnoAccession) %then else &AnnoAccession = _annoacc1;;
				Distance=abs(Locus2-locus1);
				%if (&ContPlotMeas=DPRIME or &LDMeas=DPRIME) %then %do;
					AbsDprime=abs(dprime);
				%end;
				%if (&ContPlotMeas=CORRCOEFF or &LDMeas=CORRCOEFF) %then %do;
					CorrCoeff2=corrcoeff*corrcoeff;
				%end;
				%let i=1;
				%if %length(&ContPlotMeas) %then %do;
					%if &ContPlotMeas=DPRIME %then %let ContPlotMeas = AbsDprime;
					%else %let ContPlotMeas = CorrCoeff2;
				%end;
				%if %length(&LDMeas) %then %do;
					%if &LDMeas=DPRIME %then %let LDMeas = AbsDprime;
					%else %let LDMeas = CorrCoeff2;
				%end;
				drop temp tl %if %length(&AnnoAccession) %then _annoacc:;;
			run;
			/*data OutLib.&LD_ContourPlot; set &LD_ContourPlot; run;*/

			%if %length(&ContPlotMeas)>0 %then %do;
				proc means data=&LD_ContourPlot noprint;
					var &ContPlotMeas;
					output out=minmax min=min max=max;
				run;

				data _null_;
					set minmax;
					call symput('contmax',strip(max));
					call symput('contmin',strip(min));
				run;
			%end;
		%end; 

		%let mnameID= %sysfunc(open(work.mnamest));
		%let nmrk_mnameID= %sysfunc(attrn(&mnameID,NOBS));
		%syscall set(mnameID);
		proc datasets lib=work nolist;
			modify mapdata;
			rename Marker2= &AnnoMNameTmp;
			modify indat;
			rename 
				%do i= 1 %to &nmrk_mnameID;
					%let b= %sysfunc(fetchobs(&mnameID,&i));
					_a_&i = &&&AnnoMNameTmp
				%end;
			;
			* Need to remove format introduced into the indat for the Proc Allele *;
			format _a_:;
		quit;
		%let mnameID= %sysfunc(close(&mnameID));
		%let dsnew= indat;
	%end;
	%else %do;
	
		* GenoCol is created by %PrepMapandBy *;
		%if %length(&GenoCol) %then %do;

			data indat;
				set &dsnew;
				%*if %length(&WhereStmt) %then %do;
				*	where &WhereStmt;
				%*end;
				* format is created by %GenInit *;
				format &AlleleVars numtochar.;
   				* Luciano, included NewIDVar and IDVars*;
   				keep &NewIDVar &IDVars &ByVars &AlleleVars; 
  			run;
			%*let nmv=&nummark;
		%end;
		%else %do;
			%let ErrorMessage= Not GenoCol.;
			%put ERROR: &ErrorMessage;
			%let exiterror=1;
			%goto exit;
		%end;
	
	%end;
	%endLD:;	
* ending LD computation *;
%end;

%*let exiterror=1;
%*goto exit;

* begin imputation of missing genotypes *;
%*let mrkname= &AnnoMNameTmp;
%let missgID= %sysfunc(open(work.mrkmissg));
%let numbmissmrk= %sysfunc(attrn(&missgID,NOBS));
%put numbmissmrk= &numbmissmrk;
%syscall set(missgID);

%if %upcase(&Wunit) = %upcase(Location Variable Unit) or %upcase(&Wunit) = MARKERS %then %do;
	%if %upcase(&Wunit) = %upcase(Location Variable Unit) %then %do;
		%let PositionVar= &AnnoLoc;
	%end;
	%else %do;
		%let PositionVar= _ord_;
	%end;
%end;

/*data OutLib.mapdata_final; set mapdata; run; */

%*put ERROR: &Wunit..;
%*let exiterror=1;
%*goto exit;

%let CopyVars= &NewIDVar;
%let DistPrefix= CIT;

%put OrigMaxDist= &OrigMaxDist;
%put LDNeighbors= &LDNeighbors;

%do mk= 1 %to &numbmissmrk;
	%let b= %sysfunc(fetchobs(&missgID,&mk));
	%put ### Processing marker= &_marker_ nindiv= &nindiv ###;

	%*goto endmrkloop;	

	%if &OrigMaxDist > 0 or %upcase(&LDNeighbors)= YES %then %do;

		%if &OrigMaxDist > 0 %then %do;
			data _null_;
				set mapdata;
				where &AnnoMNameTmp = "&_marker_";
				call symputx("mkpos", trim(left(&PositionVar))); 
				%if %length(&AnnoGroup) %then call symputx("mkgrp", &AnnoGroup);;
			run;

			data _null_;
				set mapdata end=_eof_ ;
				if _n_ = 1 then _count_= 1;
				* Use AnnoMNameTmp ne "&_marker_" to exclude the target marker among neighbors *;
				%if (&GpType=C) %then %do;
					if %if %length(&AnnoGroup) %then &AnnoGroup="&mkgrp" and; abs(&PositionVar - &mkpos) <= &MaxDist and &AnnoMNameTmp ne "&_marker_" then do;
						call symputx("ldmrk"||strip(_count_), trim(left(&AnnoMNameTmp)));
						*call symputx("ldtmp"||strip(_count_), trim(left(_nbortmp_)));
						output;					
						_count_ = _count_ + 1;
					end;
				%end;
				%else %do;
					if %if %length(&AnnoGroup) %then &AnnoGroup=&mkgrp and; abs(&PositionVar - &mkpos) <= &MaxDist and &AnnoMNameTmp ne "&_marker_" then do;
						call symputx("ldmrk"||strip(_count_), trim(left(&AnnoMNameTmp)));
						*call symputx("ldtmp"||strip(_count_), trim(left(_nbortmp_)));						
						output;
						_count_ = _count_ + 1;
					end;
				%end;
				if _eof_ then call symputx("nmrknbor", trim(left(_count_ - 1)));
				retain _count_;
			run;

			%if %upcase(&LDNeighbors)= YES %then %do;		
				%if &LDByTargetMarker %then %do;
					******** begining: using L-strongst LD markers to select neighbor samples ********;

					/*ods output %if &NeedLDM=1 %then LDMeasures=ldm&bp;;*/
					ods listing close;
					proc allele data=indat
						gmin=1 &LDMeasures %if %upcase(&CalcRho)=YES %then rho; 
						outstat=ldnbors(rename= (Locus1=Marker1 Locus2=Marker2)) 
						haplo=&Haplo &GenoCol lognote; 
					   	%if %length(&ByVars) %then %do; 
							by &ByVars;
					   	%end;
					   	var &_marker_;
					   	with 
							%do i= 1 %to &nmrknbor;
								&&ldmrk&i
							%end;
						;
					run;
					/*data OutLib.ldnbors&mk; set ldnbors; run;*/

					data ldnbors;
						length _nbormarker_ _nbortmp_ $100;
						set ldnbors;
						where (Test ne "HWE" and Marker1 ne Marker2);
						if Marker1 = "&_marker_" then do;
							_nbormarker_= Marker2;
							_nbortmp_= _tmpname2_;
						end;
						else do;
							_nbormarker_= Marker1;
							_nbortmp_= _tmpname1_;
						end;
						if Test="LD" and NIndiv>0 then CorrCoeff2 = chisq/NIndiv;
						else CorrCoeff2 = .;
					run;

					%*goto endmrkloop;
				%end;
				%else %do;
					******** begining: using L-strongst LD markers to select neighbor samples ********;
					data ldnbors;
						length _nbormarker_ _nbortmp_ $100;
						set &LD_Stats;
						where (Test ne "HWE" and Marker1 ne Marker2);
						if Marker1 = "&_marker_" then do;
							_nbormarker_= Marker2;
							_nbortmp_= _tmpname2_;
						end;
						else do;
							_nbormarker_= Marker1;
							_nbortmp_= _tmpname1_;
						end;
					run;
				%end;

				proc sort data=ldnbors out=ldnbors;
					by descending CorrCoeff2;
				run;
				/*data OutLib.ldnbors&mk; set ldnbors; run;*/

				data _null_/*nbmarkers*/;
					set ldnbors end=_eof_ ;
					if _n_ = 1 then _count_= 1;
					* Do not include the target marker among neighbors *;
					if (_count_ <= &NumbMrkNBHD) then do;
						* l-strongst (l= NumbMrkNBHD) LD markers *;
						call symputx("ldmrk"||strip(_count_), trim(left(_nbormarker_)));
						*call symputx("ldtmp"||strip(_count_), trim(left(_nbortmp_)));
						output;
						_count_ = _count_ + 1;
					end;
					if _eof_ then call symputx("nmrknbor", trim(left(_count_ - 1)));
					retain _count_;
				run;

				/*data OutLib.nbmarkers&mk; set nbmarkers; run;*/
			%end;
			%put nmrknbor= &nmrknbor;

			* 1. compute taxicab distance (cityblock) between samples *;
			proc distance data= &dsnew out=d13out method=CITYBLOCK nostd prefix= &DistPrefix shape=SQR;
				*var interval(&AlleleVars);
				var interval(
						%do i= 1 %to &nmrknbor;
							&&ldmrk&i
						%end;	
					);
			run;
			/*data OutLib.d13out&mk; set d13out; run;*/	
		%end;
		%else %if &OrigMaxDist = 0 and %upcase(&LDNeighbors)= YES %then %do;		
			%if &LDByTargetMarker %then %do;
				/*ods output %if &NeedLDM=1 %then LDMeasures=ldm&bp;;*/
				ods listing close;
				proc allele data=indat 
					gmin=1 &LDMeasures %if %upcase(&CalcRho)=YES %then rho; 
					outstat=ldnbors(rename= (Locus1=Marker1 Locus2=Marker2)) 
					haplo=&Haplo &GenoCol lognote; 
				   	%if %length(&ByVars) %then %do; 
						by &ByVars;
				   	%end;
				   	var &_marker_;
				   	with &AlleleVars;
				run;

				/*data OutLib.ldnbors&mk; set ldnbors; run;*/

				data ldnbors;
					length _nbormarker_ _nbortmp_ $100;
					set ldnbors;
					where (Test ne "HWE" and Marker1 ne Marker2);
					if Marker1 = "&_marker_" then do;
						_nbormarker_= Marker2;
						_nbortmp_= _tmpname2_;
					end;
					else do;
						_nbormarker_= Marker1;
						_nbortmp_= _tmpname1_;
					end;
					if Test="LD" and NIndiv>0 then CorrCoeff2 = chisq/NIndiv;
					else CorrCoeff2 = .;
				run;

				%*goto endmrkloop;
			%end;
			%else %do;
				******** begining: using L-strongst LD markers to select neighbor samples ********;
				data ldnbors;
					length _nbormarker_ _nbortmp_ $100;
					set &LD_Stats;
					where (Test ne "HWE" and Marker1 ne Marker2);
					if Marker1 = "&_marker_" then do;
						_nbormarker_= Marker2;
						_nbortmp_= _tmpname2_;
					end;
					else do;
						_nbormarker_= Marker1;
						_nbortmp_= _tmpname1_;
					end;
				run;
			%end;

			proc sort data=ldnbors out=ldnbors;
				by descending CorrCoeff2;
			run;

			/*data OutLib.ldnbors_f&mk; set ldnbors; run;*/

			data _null_;
				set ldnbors end=_eof_ ;
				if _n_ = 1 then _count_= 1;
				* Do not include the target marker among neighbors *;
				if (_count_ <= &NumbMrkNBHD) then do;
					* l-strongst (l= NumbMrkNBHD) LD markers *;
					call symputx("ldmrk"||strip(_count_), trim(left(_nbormarker_)));
					*call symputx("ldtmp"||strip(_count_), trim(left(_nbortmp_)));
					output;
					_count_ = _count_ + 1;
				end;
				if _eof_ then call symputx("nmrknbor", trim(left(_count_ - 1)));
				retain _count_;
			run;
			%put nmrknbor= &nmrknbor;
			/*data OutLib.nbmarkers&mk; set nbmarkers; run;*/

			* 2. compute taxicab distance (cityblock) between samples *;
			proc distance data= &dsnew out=d13out method=CITYBLOCK nostd prefix= &DistPrefix shape=SQR;
				*var interval(&AlleleVars);
				var interval(
						%do i= 1 %to &nmrknbor;
							&&ldmrk&i
						%end;	
					);
			run;
			/*data OutLib.d13out&mk; set d13out; run;*/							
		%end;
		
		* Use K=p to find nearest p-1 neighbors for each sample*;
		ods listing close;
		ods html close;
		ods output neighbor=IndNbors;
		proc modeclus data=d13out method=1 k=&MaxNumbIndNBHD Neighbor; 
			var &DistPrefix:;
			*id &NewIDVar;
		run;
		/*data OutLib.IndNbors0_&mk; set IndNbors; run;*/
	
		data IndNbors; 
			set IndNbors; 
			if not missing(id) then do;
				id_new= input(id, 10.);
				retain id_new;
			end;
			nbor_num= input(nbor, 10.);
		run;
		/*data OutLib.IndNbors1_&mk; set IndNbors; run;*/
		
	%end;
	%else %if &OrigMaxDist = 0 and &mk = 1 %then %do;
		******** begining: using all markers to select neighbor samples ***********;
		
		* 3. compute taxicab distance (cityblock) between samples *;
		proc distance data= &dsnew out=d13out method=CITYBLOCK nostd prefix= &DistPrefix shape=SQR;
			var interval(&AlleleVars);
		run;
		/*data OutLib.d13out&mk; set d13out; run;*/
		
		* Use K=p to find nearest p-1 neighbors for each sample*;
		ods listing close;
		ods html close;
		ods output neighbor=IndNbors;
		proc modeclus data=d13out method=1 k=&MaxNumbIndNBHD Neighbor; 
			var &DistPrefix:;
			*id &NewIDVar;
		run;
		/*data OutLib.IndNbors0_&mk; set IndNbors; run;*/
	
		data IndNbors; 
			set IndNbors; 
			if not missing(id) then do;
				id_new= input(id, 10.);
				retain id_new;
			end;
			nbor_num= input(nbor, 10.);
		run;
		/*data OutLib.IndNbors1_&mk; set IndNbors; run;*/	
				
	%end;
		
	%*goto endmrkloop;
	
	/*data OutLib.dsnew; set &dsnew; run;*/
	
	%put ### Processing marker= &_marker_ nindiv= &nindiv ###;
	* keep only observations with missing genotype *;
	data missobsdat allobsdat;
		set &dsnew(keep= &_marker_ %if %length(&WhereStmt) %then &NewIDVar;);
		if missing(&_marker_) then do;
			id_new= _n_;
			output missobsdat;
		end;
		output allobsdat;
	run;
	/*data OutLib.missobsdat&mk; set missobsdat; run;*/
	/*data OutLib.allobsdat; set allobsdat; run;*/
	
	%NObsVars(missobsdat);
	%let nmissobs= &nobs;
	
	/*data OutLib.IndNbors_&mk; set IndNbors; run;*/
	
	data Mrk_IndNbors; 
		merge IndNbors(in=a) missobsdat(in=b); 
		by id_new;
		if b and a;
	run;	
	/*data OutLib.Mrk_IndNbors_&mk; set Mrk_IndNbors; run;*/
	
	%*goto exit;

	%let ObsNotImputableFlag= 0;
		
	*ods listing;
	*ods html;
	proc iml;
		use Mrk_IndNbors;
		read all var {id_new nbor_num distance} into NBO;
		close Mrk_IndNbors;
		*NBO= num(NBO);
		*print NBO;
			
		use allobsdat;
		*read all var (MN[m]) into MS;
		read all var {&_marker_} into MS;
		*no= nrow(MS);
		close allobsdat;
		
		use missobsdat;
		read all var {id_new %if %length(&WhereStmt) %then &NewIDVar;} into loc_missobs;
		close missobsdat;
		*loc_missobs= loc(MS=.);
		*no= ncol(loc_missobs);
		
		NotImp= 0;
		* samples loop *;
		do o=1 to &nmissobs;
			*goto endo;
			loc_o= loc(NBO[,1]=loc_missobs[o,1]);
			nloc_o= ncol(loc_o);
			if (nloc_o > 0) then do; 
				*k-nearest neighbor samples loop*;
				s= 1;
				k= 1;
				*g= {0 0 0};
				g= J(&PloidyNum+1,1,0);
				flag_error = 0;
				do while(s <= nloc_o & k <= &NumbIndNBHD);
					*i= NBO[loc_o[s],2];
					z= MS[NBO[loc_o[s],2]];
					*print o s k i z;
					*print m o s k i z;
					if z ^= . then do;
						*if NBO[loc_o[s],3] > 0 then do;
							%do i=1 %to %sysevalf(&PloidyNum+1);
								if z = %sysevalf(&i-1) then do;
									g[&i]= g[&i] + 1/(1 + NBO[loc_o[s],3]);
									*g[&i]= g[&i] + 1/DIS[loc_o[s]];
									*k= k + 1;
								end;
								else flag_error = 1;
							%end;							
						*end;
						k= k + 1;
					end;
					s= s + 1;
				end;
				
				* retriving location of maximum *;
				loc_maxg= g[<:>]; 
				*print o g loc_maxg;
				* imputing missing marker *;
				MS[loc_missobs[o,1]]= loc_maxg - 1; 
			end;
			else do;
				call symputx("ObsNotImputableNewID", loc_missobs[o,1]);
				%if %length(&WhereStmt) %then %do;
					call symputx("ObsNotImputableOldID", loc_missobs[o,2]);	
				%end;
				call symputx("ObsNotImputableFlag", 1);
				NotImp= 1;
				o = &nmissobs + 1;
			end;
			endo:;
		end;
		*print MS;
		
		if NotImp=0 then do;
			create impdat from MS[colname="&_marker_"];
			append from MS;
		end;
	quit;
	
	%*goto exit;
	
	%if ^&ObsNotImputableFlag %then %do;
		/*data OutLib.impdat&mk; set impdat; run;*/
	
		%if &mk = 1 %then %do;
			proc datasets nolist;
				change impdat= imputed;
			quit;
		%end;
		%else %do;
			data imputed;
				merge imputed impdat;
			run;	
		%end;
	%end;
	%else %do;
		%if %length(&WhereStmt) = 0 %then %do;
			%let ErrorMessage= Observation number &ObsNotImputableNewID from marker &_marker_ could not be imputed. Please, increase the Number of Neighbor Markers in the Options tab or remove this observation and/or marker from the analysis.;
			%put ERROR: &ErrorMessage;
		%end;
		%else %do;
			%let ErrorMessage= Observation number &ObsNotImputableOldID from marker &_marker_ could not be imputed. Please, increase the Number of Neighbor Markers in the Options tab or remove this observation and/or marker from the analysis.;	
			%put ERROR: &ErrorMessage;	
		%end;
		%let missgID= %sysfunc(close(&missgID));
		%let exiterror= 1;
		%goto exit;
	%end;
	
	%endmrkloop:;
*end marker loop*;
%end;
%let missgID= %sysfunc(close(&missgID));

%*let exiterror=1;
%*goto exit;

data imputed;
	merge &dsnew(in=ina) imputed(in=inb);
	*by var1;
	*if inb;
	keep &AlleleVars;
run;

%if %length(&IDVars &ToKeepVars) %then %do;
	data keepdat;
		set &ds;
		%if %length(&WhereStmt) %then %do;
			where &WhereStmt;
		%end;
		keep &IDVars &ToKeepVars;
	run;
	
	data imputed;
		merge keepdat imputed(in=inb);
	run;
%end;

data OutLib.&ImpMissGeno_OutDataName;
	set imputed;
run;

%PackageInsertDataset(Imputed Missing Genotypes data from &ProcessName, OutLib, &ImpMissGeno_OutDataName, no);

%PrepOutFile(ImpMissGeno_Script,ImpMissGeno_ScriptName,&ProcessName Script for &OutName..jsl);

%global _mainjslfile_;
%let _mainjslfile_ = %quote(&ImpMissGeno_Script);

**creating a JSL file without code just to force the final jsl file to contain the process name on it instead of sasmessage**;
* create JSL code;
data _null_;
	file "&ImpMissGeno_Script" lrecl=32767;
	put '//!';
	put "Names Default to Here(1);";
	put "Here:DataList = {};";
	put "ClientOutPath = ""&ClientOutPath"";";
	put "OutDataName = ""&ImpMissGeno_OutDataName"";";
	put "lobuttonwindow=Function({},{default local},";
	put "		VListBox(";
	put "			TextBox(""Click the button below to launch the Marker Properties process.""),";
	put "			ButtonBox(""Marker Properties"",";
	put "				if(host is(Windows),";
	put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/MarkerProperties.jsl"", absolute, windows)),";
	put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/MarkerProperties.jsl"", absolute, POSIX) ));";
	put "	";
	** Set input data set and Output Folder *;
	put "				jg:setDefaultParmValues(2, ""MarkerProperties"");";				
	put "				jg:MarkerProperties_updateDeps;";
	put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""InData""),""&ClientOutPath.&&ImpMissGeno_OutDataName..sas7bdat"");";		
	put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""OutPath""),""&ClientOutPath"");";										
	put "				jg:MarkerProperties_updateDeps;";				
	put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AlleleVars""),""&OrigAlleleVars"");";
	put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""ListMarkerVars""),""&ListMarkerVars"");";
	%if %length(&AnnoInData) %then %do;
		put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoInData""),""&AnnoInData"");";
		put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoLabel""),""&AnnoLabel"");";
		put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoMName""),""&AnnoMName"");";
		put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoGroup""),""&OrigAnnoGroup"");";
		put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoLoc""),""&AnnoLoc"");";
		put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""AnnoAccession""),""&AnnoAccession"");";
	%end;
	put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""PloidyNum""),""&PloidyNum"");";	
	put "				jg:MarkerProperties_updateDeps;";	
	put "				jg:setCurrentParmValue(2, ""MarkerProperties"", jg:getParmNum(""MarkerProperties"",""MarkerFmt""),""Numeric Genotypes"");";
	put "				jg:MarkerProperties_updateDeps;";			
	put "			)";
	put "		)";
	put ");";
	put "Here:LaunchCode = lobuttonwindow();";	
run;

%PackageInsertFile(JSL Code from &ProcessName, &OutPath, &ImpMissGeno_ScriptName, ,text/plain, yes);

%exit:;

%if &exiterror %then %do;

	%PrepOutFile(ImpMissGeno_Script,ImpMissGeno_ScriptName,&ProcessName Script for &OutName..jsl);

	%global _mainjslfile_;
	%let _mainjslfile_ = %quote(&ImpMissGeno_Script);

	* create JSL code;
	data _null_;
		file "&ImpMissGeno_Script" lrecl=32767;
		put '//!';
		put "Names Default to Here(1);";
		put "Here:DataList = {};";
		put "Here:TabButtonList = {};";
		put "Here:TabCodeList = {};";
		put "Here:TabAfterCodeList = {};";
		put "Here:TabOpenList = {};";	
		put "nwfn= Function({},{default local},";
		put "// New Window(""&ProcessName Error Message"",showmenu(0),showtoolbars(0),";
		put " TextBox("" &ErrorMessage \!N\!N"",";
		put " <<Font Color(5),<<Set Font Size(10),<<Set Wrap(400)";
		put " );";
		put "//);";					
		put ");";
		put "InsertInto(DataList,{{""""}});";
		put "InsertInto(TabButtonList, ""Error Messages"");";	
		put "InsertInto(TabCodelist,expr(Here:nwfn()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";	
	run;

	%PackageInsertFile(JSL Code from &ProcessName, &OutPath, &ImpMissGeno_ScriptName, ,text/plain, yes);

%end;

%if %symexist(mnameID) %then %do;
	%let mnameID= %sysfunc(close(&mnameID));
%end;

%if %symexist(missgID) %then %do;
	%let missgID = %sysfunc(close(&missgID));
%end;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%exitcheck:;

%mend ImputeMissingGenotypes;
%ImputeMissingGenotypes;	