/*
JMP Genomics Analytical Process

Name: Linkage Map Order
Author: Kelci Miclaus
History:
	kemicl 03/03/11 Initial Coding
	sbrlsi 11/14/19 Adding LOD Constraint
Description:
		Linkage Map Order determines the probable order of genetic markers within linkage groups based on recombination frequencies
		and calculates the genetic distances between markers to produce a linkage map.  The input to this process should be the output data 
		set from the Recombination and Linkage Groups AP containing a symmetric matrix of pairwise recombination rates between
		markers and a variable whose values determine the linkage group to which each marker belongs.  Marker ordering is done only within
		linkage groups.  Multidimensional Scaling is used to find a one dimensional ordering that best represents the structure found in the 
		matrix of recombination frequencies.  A residual plot of the MDS fit can help to determine and flag outlier markers that do not fit well in the 
		ordering. 
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData DistVars IDVar LGroupVar SkOrderVar ColorVar KeepVars ListDistVars WhereStmt OutPath ServerOutPath
LODInData LODVars LODIDVar ListLODVars OrderInData AnnoMName OrderGrpVar OSkOrderVar OrderVar
SNPIndata AlleleVars LabelVar ListMarkerVars MarkerFmt GenoDelim GenoMCode DoCellPlots
BreakGrps BreakGrpsVar BreakGrpsGD BreakGrpsR BreakGrpsLOD UseStoppingRules OptimalityGap MaxTimeStep
MapFunc Algorithm Threshold LODThreshold MDSSqRoot UseWeights OutDataPrefix LinkageMapOrderScript MacroPath AddDroppedMarkers
ClusterVar ClusterCenterVar OldNumCode CrossType GenNum RanMaNum RILBiasCorr;

*ProcessBody;
%macro LinkageOrder;
%global exiterror flag_rf_error; 
%let flag_rf_error=0;

%let ProcessName = LinkageMapOrder;
/*kemicl development check, change to 1 to output interim data sets*/
%let debug = 0;
%include "&MacroPath./UtilityMacros.sas" / nosource;
%include "&MacroPath./qtlUtilityMacros.sas" / nosource; /*Luciano: add dropped out markers back*/

%GetOrigFileName(InData,OrigInData);
%PathName(&InData);
%if &exiterror %then %do;
   %put ERROR: The Input Data Set specification is invalid.;
   %goto exit;
%end;
%let InData = &TmpName;
libname InLib "&TmpPath";
libname OutLib "&OutPath";

%let LODDataFlag= 0;
%if %length(&LODInData) %then %do;
	%PathName(&LODInData);
	%if &exiterror %then %do;
		%put ERROR: The LOD Data Set specification is invalid.;
		%let exiterror=1;
		%goto exit;
	%end;
	%let LODDataFlag = 1;
	libname LODLib "&TmpPath";
	%let LODData = LODLib.&TmpName;
%end;
%put LODDataFlag= &LODDataFlag;

%if %length(&OrderInData) %then %do;
	%PathName(&OrderInData);
	%if &exiterror %then %do;
		%put ERROR: The Order Data Set specification is invalid.;
		%let exiterror=1;
		%goto exit;
	%end;
	%let OrderData = &TmpName;
	libname OrderLib "&TmpPath";
%end;

%if %length(&SNPInData) %then %do;
	%PathName(&SNPInData);
	%if &exiterror %then %do;
		%put ERROR: The Genotype Data Set specification is invalid.;
		%let exiterror=1;
		%goto exit;
	%end;
	%let SNPData = &TmpName;
	libname SNPLib "&TmpPath";		
%end;

/*begin: Luciano updated this code to when incorporating the optnet procedure*/
/* error checking for skeleton order specifications from dialog */
/*S1056137: (Luciano) Adding a check for &OSkOrderVar. Otherwise, when specifying a Frame Work Order 
variable in the Order Data Set and using an algorithm other than MAP ORDER OPTIMIZATION 
no error was issued.*/
%put Algorithm= &Algorithm;
%if (%length(&SkOrderVar) or %length(&OSkOrderVar)) and %nrbquote(%upcase(&Algorithm)) ^= MAP ORDER OPTIMIZATION %then %do;
	%put ERROR: The Map Order Optimization Algorithm must be used when a Framework Order Variable is specified.;
	%put ERROR: Please choose this marker ordering algorithm on the Analysis Tab.;
	%let exiterror =1;
	%goto exit;
%end;

%if (%length(&SkOrderVar) or %length(&OSkOrderVar)) and %index(%nrbquote(%upcase(&Algorithm)),ACCELERATED)=1 %then %do;
	%put ERROR: The Map Order Optimization Algorithm must be used when a Framework Order Variable is specified.;
	%put ERROR: Please choose this marker ordering algorithm on the Analysis Tab.;
	%let exiterror =1;
	%goto exit;
%end;

%if %upcase(&AddDroppedMarkers)= YES and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0 %then %do;
	%put ERROR: The Multidimensional Scaling Algorithm cannot be used when the option to "Add markers dropped out from marker reduction analysis back to the linkage groups" is checked in the Analysis tab.;
	%put ERROR: Please choose another marker ordering algorithm on the Analysis Tab.;
	%let exiterror =1;
	%goto exit;
%end;

%if %upcase(&AddDroppedMarkers)= YES and %length(&SNPIndata)=0 %then %do;
	%put ERROR: The Genotype Data Set must be specified when the option to "Add markers dropped out from marker reduction analysis back to the linkage groups" is checked in the Analysis tab. Please specify a Genotype Data Set in the Genotype Data Set Tab.;
	%let exiterror =1;
	%goto exit;
%end;
/*end: Luciano*/

%if %length(&OSkOrderVar) and ^%length(&OrderGrpVar) %then %do;
	%put ERROR:  The Linkage Group Variable and Framework Order Variable must be specified from the same data set.;
	%put ERROR:  Please specify both of these variables either from the Input Data Set, or the Order Data set.;
	%let exiterror = 1;
	%goto exit;
%end;

%if ^%length(&OutDataPrefix) %then %let OutDataPrefix = &OrigInData;
%if %length(&OutDataPrefix)>27 %then %let OutDataPrefix = %substr(&OutDataPrefix,1,27);
%let OutName = &OutDataPrefix;

*begin: Luciano, add dropped out markers from k-means clustering*;
%if &debug %then %do;
	data OutLib.InData;set InLib.&InData;run;
%end;

*checking if the _KMCluster_ variable exists in the Input Data Set*;
%let _KMClusterCenter_VarExist= 0;
%*let vtypeClus=;
%*let vtypeLG=;
data _null_;
	_dsid_= open("InLib.&InData");
	_check_= varnum(_dsid_,'_KMClusterCenter_');
	if _check_> 0 then call symputx("_KMClusterCenter_VarExist", 1);
	*_checkClus_= vartype(_dsid_, varnum(_dsid_,"&ClusterVar"));
	*call symputx("vtypeClus", _checkClus_);
	*_checkLg_= vartype(_dsid_, varnum(_dsid_,"&LGroupVar"));
	*call symputx("vtypeLG", _checkLG_);
run;
%put _KMClusterCenter_VarExist= &_KMClusterCenter_VarExist;

%put AddDroppedMarkers= &AddDroppedMarkers;
%if %length(&AddDroppedMarkers)= 0 %then 
	%let AddDroppedMarkers= No;

%if %upcase(&AddDroppedMarkers)= NO and &_KMClusterCenter_VarExist= 1 %then %do;
	%if %index(%upcase(&WhereStmt), %upcase(&ClusterCenterVar)) = 0 %then %do;
		%let string1= A variable named _KMClusterCenter_ was found in Input Data Set, this is an indicative that this data set was produced from the Recombination and Linkage Groups process with the option to "Append markers dropped out from marker reduction analysis to the output data sets" checked.;
		%let string2= %str(If this is really the case, then do the following: 1- Specify a Cluster Variable in the General tab (if the input SAS data set was produced from the Recombination and Linkage Groups process, this variable will be named _KMCluster_); 2- Specify a valid Where Statement to select only the marker in the center of each cluster from the Input Data Set (for instance, "_KMClusterCenter_=1", without the quotes).);
		%let string3= If you want to add dropped out markers back to linkage groups, then check the option to "Add markers dropped out from marker reduction analysis back to the linkage groups" in the Analysis tab. If the Input Data Set was not produced from the Recombination and Linkage Groups process with the option to "Append markers dropped out from marker reduction analysis to the output data sets" checked, please rename the _KMClusterCenter_ variable and rerun the process.;
		%put ERROR: &string1 &string2 &string3;  
		%let exiterror =1;
		%goto exit;
	%end;
%end;
	
%if %upcase(&AddDroppedMarkers)= YES %then %do;

	%if %length(&LGroupVar)=0 %then %do;
		%let string1= A Linkage Group variable must be specified on the General tab.;
		%let string2= Please, return to the General tab and specify a Linkage Group Variable.;
		%let string3= If the input SAS data set was produced from the Recombination and Linkage Groups process, this variable may be named LGroup.;
		%put ERROR: &string1 &string2 &string3;
		%let exiterror =1;
		%goto exit;
	%end;
	
	%if %length(&ClusterVar)=0 %then %do;
		%let string1= A Cluster Variable was not specified in the General tab.;
		%let string2= Please, return to the General tab and specify a Cluster Variable.;
		%let string3= If the input SAS data set was produced from the Recombination and Linkage Groups process, this variable will be named _KMCluster_.;
		%put ERROR: &string1 &string2 &string3; 
		%let exiterror =1;
		%goto exit;
	%end;
	
	%if %length(&ClusterCenterVar)=0 %then %do;
		%let string1= A Cluster Center Variable was not specified in the General tab.;
		%let string2= Please, return to the General tab and specify a Cluster Center Variable.;
		%let string3= Cluster Center Variable must be a binary variable with a value of 1 indicating which marker is the center of each cluster and 0 otherwise.;
		%let string4= If the input SAS data set was produced from the Recombination and Linkage Groups process, this variable will be named _KMClusterCenter_.;
		%put ERROR: &string1 &string2 &string3 &string4; 
		%let exiterror =1;
		%goto exit;
	%end;
	
	%let string2= Please, return to the General tab and specify a valid Where Statement to exclude missing values of the &LGroupVar variable from the Input Data Set.;
	%let string3= If the input SAS data set was produced from the Recombination and Linkage Groups process the Where Statement may be "&ClusterCenterVar=1", without the quotes.;
	%if %length(&WhereStmt)=0 %then %do;
		%let string1= A Where Statement to select only the marker in the center of each cluster from the Input Data Set was not specified in the field Filter to Include Observations from the General tab.;
		%put ERROR: &string1 &string2 &string3;
		%let exiterror =1;
		%goto exit;
	%end;
	%else %if %index(%upcase(&WhereStmt), %upcase(&ClusterCenterVar)) = 0 %then %do;
		%let string1= The Where Statement specified in the field Filter to Include Observations from the General tab is incorrect.;
		%put ERROR: &string1 &string2 &string3;
		%let exiterror =1;
		%goto exit;
	%end;
	
	*checking if ClusterVar is different from the LGroupVar*;
	%if %upcase(&ClusterVar) = %upcase(&LGroupVar) %then %do;
		%put ERROR: The Cluster Variable and Linkage Group Variable from the General tab cannot be the same variable.;
		%let exiterror =1;
		%goto exit;		
	%end;	
	
	%let nClusCenter= ;
	%let ClusCenterType=;
	%GetLevels(_Ds =InLib.&InData,_Var=&ClusterCenterVar,_LPrefix= ClusCen ,_nLvls=nClusCenter,_vType= ClusCenterType);
	%if &nClusCenter > 2 or %upcase(&ClusCenterType) = C %then %do;		
		%put ERROR: Cluster Center Variable must be a numeric binary variable with a value of 1 indicating which marker is in the center of each cluster and 0 for all markers.;
		%let exiterror =1;
		%goto exit;
	%end;
	%else %do;
		%if &nClusCenter = 1 %then %do;	
			%if &ClusCen1 ^= 1 %then %do;
				%put ERROR: Cluster Center Variable must be a numeric binary variable with a value of 1 indicating which marker is in the center of each cluster and 0 for all markers.;
				%let exiterror =1;
				%goto exit;
			%end;
		%end;
		%else %if &nClusCenter = 2 %then %do;
			%if &ClusCen1 ^= 0 and &ClusCen2 ^= 1 %then %do;
				%put ERROR: Cluster Center Variable must be a numeric binary variable with a value of 1 indicating which marker is in the center of each cluster and 0 for all markers.;
				%let exiterror =1;
				%goto exit;
			%end;
		%end;
	%end;
	
	%let nClus=;
	%let ClusType=;
	%GetLevels(_Ds =InLib.&InData,_Var=&ClusterVar,_LPrefix = Clus,_nLvls=nClus,_vType= ClusType);
	%put nClus= &nClus;
	
	%do i=1 %to &nClus;
		%if %upcase(&ClusType)= C %then %do;
			%let Clus&i= "&&Clus&i";
			%put Clus&i= &&Clus&i;
		%end;	
		
		*checking if there are missing values for ClusterVar*;
		%if &&Clus&i= . or %length(&&Clus&i)= 0 %then %do;
			%put ERROR: The Cluster Variable &ClusterVar cannot have missing values.;
			%let exiterror =1;
			%goto exit;
		%end;
		
		proc sql UNDO_POLICY=NONE;
			create table currCluster as
			select &LGroupVar, &ClusterVar, &ClusterCenterVar
			from InLib.&InData as d
			where d.&ClusterVar= &&Clus&i and not missing(d.&ClusterVar)
			order by &LGroupVar;
		run;
		
		*checking if each cluster has exactly one center marker with no missing value*;
		proc sort data=currCluster out=nCenterMk nodupkey;
			by &ClusterCenterVar;
		run;
		
		%NObsVars(nCenterMk);
		%if &nobs= 1 %then %do;
			%let missCenter= 0;
			data _null_;
				set nCenterMk;
				if _n_=1 and &ClusterCenterVar ^= 1 then call symputx("missCenter", 1);
			run;
		
			%if &missCenter %then %do;
				%put ERROR: Cluster &&Clus&i do not have a center marker. Please check the Cluster Center Variable &ClusterCenterVar..;
				%let exiterror =1;
				%goto exit;
			%end;
		%end;
		%else %if &nobs= 2 %then %do;
			%let missCenter= 0;
			data _null_;
				set nCenterMk;
				if _n_=2 and &ClusterCenterVar ^= 1 then call symputx("missCenter", 1);
			run;
		
			%if &missCenter %then %do;
				%put ERROR: Cluster &&Clus&i do not have a center marker. Please check the Cluster Center Variable &ClusterCenterVar..;
				%let exiterror =1;
				%goto exit;
			%end;
		%end;
		%else %if &nobs= 3 %then %do;		
			%put ERROR: Cluster Center Variable be a binary variable with a value of 1 indicating which marker is the center of each cluster and 0 otherwise.;
			%let exiterror =1;
			%goto exit;
		%end;	
	%end;
	
	*checking if number of cluster is equal to the number of markers*;
	%NObsVars(InLib.&InData);	
	%if &nobs = &nClus %then %do;
		%put WARNING: The number of cluster is equal to the number of markers. Therefore, there are no markers to be added back to the linkage groups.;
		%let AddDroppedMarkers= No;
		%let WhereStmt= ;
	%end;
%end;
*end: Luciano, add dropped markers from k-means clustering*;

%if %length(&LGroupVar &OrderGrpVar)=0 %then %do;
	 %put ERROR: A Linkage Group variable must be specified either on the General tab or on the Order Data Set Tab for this process to run.;
	 %let exiterror=1;
	 %goto exit;
%end;

%if %length(&DistVars &ListDistVars)=0 %then %do;
	%put ERROR: Recombination Rate variables must be specified in either the Recombination Rate Variables or List-Style Specification of Recombination Rate Variables parameters.;
	%let exiterror=1;
	%goto exit;
%end;

%if %length(&SkOrderVar) or %length(&OSkOrderVar) %then %do;
	%let Threshold = 1; /*recombination threshold*/
	%let LODThreshold = 1000;
%end;
%global RECThreshold;
%let RECThreshold = &Threshold;

%if (^%length(&LODInData) and %index(%upcase(&BreakGrpsVar),LOD)) %then %do;
	%put ERROR: Breaking linkage groups based LOD Score is not allowed when a LOD Score Data Set is not specified.;
	%put ERROR: Please choose either Genetic Distance or Recombination Fraction in the Analysis Tab.;
	%let exiterror =1;
	%goto exit;
%end;

/*set up vars for breaking further linkage groups */
%if %index(%upcase(&BreakGrpsVar),GENETIC) %then %do;
	%let BreakGrpsVar = cM_Step;
	%let BreakGrpsValue = &BreakGrpsGD;
	%let BGimlprefix = cm;
	%let BreakFunc = max;
	%let BreakOperator = %str(>);
%end;
%else %if %index(%upcase(&BreakGrpsVar),LOD) %then %do;
	%let BreakGrpsVar = LOD;
	%let BreakGrpsValue = &BreakGrpsLOD;
	%let BGimlprefix = lod;
	%let BreakFunc = min;
	%let BreakOperator = %str(<);
%end;
%else %do;
	%let BreakGrpsVar = RecombinationF;
	%let BreakGrpsValue = &BreakGrpsR;
	%let BGimlprefix = rf;
	%let BreakFunc = max;
	%let BreakOperator = %str(>);
%end;

%GLOBAL LinkageMapOrder_OutData LinkageMapOrder_OutDataRes LinkageMapOrder_OutDataEx;
%let LinkageMapOrder_OutData = ;
%let LinkageMapOrder_OutDataRes = ;
%let LinkageMapOrder_OutDataEx = ;
/* Prep output data linkage map data set and residuals dataset */
%PrepOutDS(LinkageMapOrder_OutData,LinkageMapOrder_OutDataName,&OutDataPrefix,_lmo);
%PrepOutDS(LinkageMapOrder_OutDataRes,LinkageMapOrder_OutDataResName,&OutDataPrefix,_lmr);
%PrepOutDS(LinkageMapOrder_OutDataEx,LinkageMapOrder_OutDataExName,&OutDataPrefix,_eml);

/*Check if the SAS/OR component is installed if Algorithm is MOO
%if %index(%upcase(&Algorithm),MULTIDIMENSIONAL)=0 and  %length(&OrderVar)=0 %then %do; 
	data _null_;
		x = sysprod('or');
		call symput('CheckOR',trim(left(x)));
	run;
	%if &CheckOR %then %do;
		%put SAS/OR is licensed, Map Order Optimization will be used.;
	%end;
	%else %do;
		%put ERROR: The SAS/OR component is not licensed and must be purchased separately from JMP Genomics.  Map Order Optimization Cannot Be Used.; 
		%let exiterror =1;
		%goto exit;
	%end;
%end;
 */
/* check that the matrix is square */
data sqrchk; set InLib.&InData;
	keep &DistVars &ListDistVars;
run;
%nobsvars(sqrchk);
%if &nobs ne &nvars %then %do;
	%put ERROR: The matrix of recombination fractions is not a square matrix of values.;
	%put ERROR: There are &nobs marker rows and &nvars marker columns, which should be equal values. Please check the input data.;
	%let exiterror =1;
	%goto exit;
%end;

/* Check first 100 recombination rate ID variable names to match to Column names */
/* Get RF variable names to put in a list to check column with values matching variable names */
data mnames; set InLib.&InData;
	where 0;
	keep &DistVars &ListDistVars;
run;

%let _tid_ = %sysfunc(open(mnames));
%let _nVars_ = %sysfunc(attrn(&_tid_,NVARS));

%let allmatch = 1;

data markerlist;
	length _IDVAR_ $32; 	
	%do i = 1 %to %sysfunc(min(100,&_nVars_));
		_IDVAR_ = "%sysfunc(varname(&_tid_,&i))";
		output;
	%end;
run;
data _null_; 
	set InLib.&InData(keep = &IDVar); set markerlist;
	if _n_ <= 100;
	if upcase(strip(&IDVar)) ne upcase(strip(_IDVAR_)) then do;
		call symputx("allmatch",0);
	end;
run;

%if &allmatch = 0 %then %do;
	%put ERROR: The values of the Marker ID Variable specified does not match the values of the Recombination Fraction column names.;
	%put ERROR: An ID Variable called _IDVAR_ will be created and used that assumes the rows are in the same order as the columns.;
	/*Create IDvar from recomb freq variable names to merge into recombination data */
	data markerlist;
		length _IDVAR_ $32; 	
		%do i = 1 %to &_nVars_;
			_IDVAR_ = "%sysfunc(varname(&_tid_,&i))";
			output;
		%end;
	run;
	%let IDVar = _IDVAR_;
%end;
/* Close data set after getting variable names */
%let rc = %sysfunc(close(&_tid_));

%if &LODDataFlag %then %do;
	/* Check first 100 LOD ID variable names to match to Column names */
	/* Get LOD variable names to put in a list to check column with values matching variable names */
	data LODmnames; 
		set &LODData;
		where 0;
		keep &LODVars &ListLODVars;
	run;
	
	%let _tid_ = %sysfunc(open(LODmnames));
	%let _nVars_ = %sysfunc(attrn(&_tid_,NVARS));
	
	%let allmatch = 1;
	
	data LODmarkerlist;
		length _LODIDVAR_ $32; 	
		%do i = 1 %to %sysfunc(min(100,&_nVars_));
			_LODIDVAR_ = "%sysfunc(varname(&_tid_,&i))";
			output;
		%end;
	run;
	
	data _null_; 
		set &LODData(keep = &LODIDVar); 
		set LODmarkerlist;
		if _n_ <= 100;
		if upcase(strip(&LODIDVar)) ne upcase(strip(_LODIDVAR_)) then do;
			call symputx("allmatch",0);
		end;
	run;
	
	%if &allmatch = 0 %then %do;
		%put ERROR: The values of the LOD ID Variable specified does not match the values of the LOD score column names.;
		%put ERROR: An ID Variable called _LODIDVAR_ will be created and used that assumes the rows are in the same order as the columns.;
		*Create LODIDvar from LOD variable names to merge into LOD data *;
		data LODmarkerlist;
			length _LODIDVAR_ $32; 	
			%do i = 1 %to &_nVars_;
				_LODIDVAR_ = "%sysfunc(varname(&_tid_,&i))";
				output;
			%end;
		run;
		%let LODIDVar = _LODIDVAR_;
	%end;
	* Close data set after getting variable names *;
	%let rc = %sysfunc(close(&_tid_));
	
	/*data OutLib.lmo_lodmarkerlist; set lodmarkerlist; run;*/
	/*data OutLib.lmo_recmarkerlist; set markerlist; run;*/
	
	/* Check first 1000 LOD ID variable names to match to Recombination Rate variable names */
	data _null_; 
		set markerlist(keep= _IDVAR_ ); 
		set LODmarkerlist(keep= _LODIDVAR_);
		if _n_ <= 1000;
		if upcase(strip(_LODIDVAR_)) ne upcase(strip(_IDVAR_)) then do;
			call symputx("allmatch",0);
		end;
	run;
	
	%if &allmatch = 0 %then %do;
		%put ERROR: The values of the LOD ID Variable specified does not match the values of the Marker ID Variable.;
		%let exiterror=1;
		%goto exit;
	%end;	
%end;

*begin: Luciano, add dropped out markers*;
**S1190637: Check marker format when SNPInData is not empty**;
%put SNPIndata= SNPIndata;
%if %upcase(&AddDroppedMarkers)= YES or %length(&SNPIndata) %then %do;
			
	*Get names of markers in the IDVar from InData*;
	proc transpose data=InLib.&InData out=indatamnames(drop=_NAME_);
		var &IDVar;
		id &IDVar;
	run;
	
	%let idvarsmnames= ;
	%let _tid_ = %sysfunc(open(indatamnames));
	%let _n_IDVars_ = %sysfunc(attrn(&_tid_,NVARS));
	/*
	%do i=1 %to &_n_IDVars_;
		%let idvarsmnames= &idvarsmnames %sysfunc(varname(&_tid_, &i)); 
	%end;
	*/
	
	*S1098339: The text expression length (65539) exceeds maximum length (655534).*;
	%let nvarblock= 0;
	%let idvarsmnames= ;
	%do i=1 %to &_n_IDVars_;
		%if %length(&idvarsmnames) le 32700 %then %do;
			%let idvarsmnames = &idvarsmnames %sysfunc(varname(&_tid_, &i)); 
		%end;
		%else %do;
			%let nvarblock= %sysevalf(&nvarblock + 1);
			%let idvarsmnames&nvarblock = &idvarsmnames %sysfunc(varname(&_tid_, &i)); 
			%let idvarsmnames = ;
		%end;
	%end;
	%if &nvarblock = 0 %then %do;
		%let nvarblock= %sysevalf(&nvarblock + 1);
		%let idvarsmnames&nvarblock = &idvarsmnames;
		%let idvarsmnames = ;
	%end;
	%else %if &nvarblock > 0 and %length(&idvarsmnames) > 0 %then %do;
		%let nvarblock= %sysevalf(&nvarblock + 1);
		%let idvarsmnames&nvarblock = &idvarsmnames;
		%let idvarsmnames = ;
	%end;
	
	%let rc = %sysfunc(close(&_tid_));
		
	*Get number of matching marker names between SNPData and InData sets*;
	data snpmnames gendat;
		set SNPLib.&SNPData;
		if _n_= 1 then output snpmnames;
		output gendat;
		*keep &idvarsmnames;
		keep 	%do z=1 %to &nvarblock;
				&&idvarsmnames&z
			%end;
		;
	run;

	%let _tid_= %sysfunc(open(snpmnames));
	%let _n_SNPVars_= %sysfunc(attrn(&_tid_, nvars));
	%let rc = %sysfunc(close(&_tid_));	
	
	%if %sysevalf(&_n_SNPVars_ ~= &_n_IDVars_) %then %do;
		%put ERROR: The values of the Marker ID variable specified does not match marker variable names in the Genotype Data Set;
		%put Please check that the marker variable names in the Genotype Data Set exactly matches the name of markers specified by the Marker ID variable from the Input Data Set columns.;		
		
		%exiterror=1;
		%goto exit;
	%end;
	
	*begin: variables needed to run rfest()*;
	%global Parent Cross; *Luciano: will be used in rfest() macro*;
	%let OrigAlleleVars = &AlleleVars;
	%let AlleleVars = &AlleleVars &ListMarkerVars;
	
	%if %length(&AlleleVars) = 0 %then %do;
		%put ERROR: Marker variables must be specified in either the Marker Variables or List-Style Specification of Marker Variables parameters in the Genotype Data Set tab.;
		%let exiterror=1;
		%goto exit;
	%end;
	
	%let Parent = 0;
	%if %index(%upcase(&CrossType),BC) %then %do;
		%let cross = 1;
		%if %upcase(&CrossType) = BC1 %then %let Parent = 1;
		%else %let Parent = 2;
	%end;
	%else %if %upcase(&CrossType) = SFN  %then %let cross = 22;
	%else %if %upcase(&CrossType) = RI1 %then %let cross = 3; /*RILself*/
	%else %if %upcase(&CrossTYpe) = RI2 %then %let cross = 4; /*RILsib*/
	%else %if %upcase(&CrossType) = RI0 %then %let cross = 5; /*DH*/
	%else %if %upcase(&CrossType) = RFN %then %let cross = 6; /* goes with SFn, also known as AIL*/
	%else %if %upcase(&CrossType) = IRILSELF %then %let cross = 7; /* goes with RILSELF */
	%else %if %upcase(&CrossType) = IRILSIB %then %let cross = 8; /* goes with RILSELF */

	%let OrigGenNum= &GenNum;
	%if (%length(&GenNum)=0 or &GenNum < 2) and (%upcase(&CrossType) = SFN or %upcase(&CrossType) = RFN) %then %do;
		%let string1= Genotyping Generation (n) must be specified when crosstype is &CrossType.;
		%let string2= Please go to the Genotyping Generation (n) on the Recombination tab and specify a value for this parameter.;
		%put ERROR: &string1.. &string2;
		%let exiterror=1;
		%goto exit;
	%end;
	%else %if ^%length(&GenNum) %then 
		%let GenNum = 2;

	%let OrigRanMaNum= &RanMaNum;
	%if (%length(&RanMaNum)=0 or &RanMaNum < 2) and (%upcase(&CrossType) = IRILSELF or %upcase(&CrossType) = IRILSIB) %then %do;
		%let string1= Random Mating Generation (t) Prior to Inbreeding must be specified when crosstype is &CrossType.;
		%let string2= Please go to the Random Mating Generation (t) Prior to Inbreeding on the Recombination tab and specify a value for this parameter.;
		%put ERROR: &string1.. &string2;
		%let exiterror=1;
		%goto exit;
	%end;
	%else %if ^%length(&RanMaNum) %then 
		%let RanMaNum = 2;

	/*
	For both SFn and RFn, RanMaNum is empty and GenNum is NOT empty.
	For both IRILself and IRILsib, RanMaNum is NOT empty and GenNum is empty.
	In the equations, RanMaNum= GenNum, therefore we optioned to use GenNum in the
	entire code.
	*/
	%if (%upcase(&CrossType) = IRILSELF or %upcase(&CrossType) = IRILSIB) 
		%then %let GenNum = &RanMaNum;

	*to run the analysis of f2 in the old code written solely for it;
	%if &cross = 22 and &GenNum = 2 %then %let cross = 2; 
	%put cross= &cross;
	
	%if %Upcase(&RILBiasCorr) = YES %then %let RILBiasCorr =1;
	%else %let RILBiasCorr = 0;

	/* if AIL and no spec'd Seg ratios, create an indicator to not perform test right now */ 
	%global no_st; *Luciano: will be used in rfest() macro*;
	%let no_st = 1;
	
	%global MaxR; *Luciano: will be used in rfest() macro*;
	%let MaxR= 0.55;
	
	%global BB_ef AB_ef AA_ef MultipleTestingMethod SEGRATIOS AnnoInData;
	%let BB_ef= 0;
	%let AB_ef= 0;
	%let AA_ef= 0;
	*end: variables needed to run rfest()*;	
	
	/*data outlib.gendat; set gendat;run;*/
	
	/* marker recoding macro to convert to numeric for IML code */
	%RecodeGenos;

	/*S1190637: RecodeGenos creates code0, code11, code12, and code2 only when marker format
	is not numeric. Therefore, need to create them here.*/
	%if %index(%Upcase(&MarkerFmt),NUMERIC) ^= 0  %then %do;
		%if %index(%Upcase(&MarkerFmt), NONDELIMITED)>0%then %do;
			%let code0 = %scan(&GenoMCode,1,' ')%scan(&GenoMCode,1,' ');
			%let code11 = %scan(&GenoMCode,1,' ')%scan(&GenoMCode,2,' ');
			%let code12 = %scan(&GenoMCode,2,' ')%scan(&GenoMCode,1,' ');
			%let code2 =  %scan(&GenoMCode,2,' ')%scan(&GenoMCode,2,' ');			
		%end;
		%else %if %index(%Upcase(&MarkerFmt), HOMOZYGOUS)>0 %then %do;
			%let code0 = %scan(&GenoMCode,1,' ');
			%let code11 = %scan(&GenoMCode,1,' ')%scan(&GenoMCode,2,' ');  /* AB coded */
			%let code12 = H;  /* H coded*/
			%let code2 =  %scan(&GenoMCode,2,' ');
		%end;
		%else %if %Upcase(&MarkerFmt) = GENOTYPES %then %do; /* delimited */
			%let code0 = %scan(&GenoMCode,1,' ')&GenoDelim%scan(&GenoMCode,1,' ');
			%let code11 = %scan(&GenoMCode,1,' ')&GenoDelim%scan(&GenoMCode,2,' ');
			%let code12 =  %scan(&GenoMCode,2,' ')&GenoDelim%scan(&GenoMCode,1,' ');
			%let code2 =  %scan(&GenoMCode,2,' ')&GenoDelim%scan(&GenoMCode,2,' ');
		%end;
		%else %do;
			%let code0 = 0; 
			%let code11 = 1;
			%let code12 = 1;
			%let code2 = 2;
		%end;
		%put Genotype Codes: &code0, &code11, &code12, &code2;

	%end;

	/* Add a check for numeric format for negative values and throw error if found */
	%OldGenoCheck;
	%if &exiterror %then %goto exit;

	%if &syserr>4 %then %do;
		%put ERROR: The genotype data could not be converted to numeric format for recombination computation.;
		%put Please check the Marker Format coding of your data for possible errors.;
	 	%let exiterror=1;
		%goto exit;
	%end;
%end;
	
%if %upcase(&AddDroppedMarkers)= YES %then %do;

	/*S1051771. Luciano: When markers are in a format other than numeric, %recodegenos will return 
	marker names as _geno_1 _geno_2 and so on. Therefore, I need to rename these markers to 
	their orginal names. It is important to notice that Linkage Groups AP does not accept
	allele marker format, and for that reason I can use the code to rename markers as follows:*/
	%if %index(%Upcase(&MarkerFmt),NUMERIC) = 0  %then %do;
		%let dsid = %sysfunc(open(mnames));
		%let _nVars_= %sysfunc(attrn(&dsid, nvars));
		
		/*data OutLib.gendat0; set gendat; run;*/
		
		proc datasets library=work nolist;
			modify gendat;
			rename 
			%do i=1 %to &_nVars_;
				_geno_&i = %sysfunc(varname(&_tid_,&i))
			%end;
			;
		quit;
		%let rc = %sysfunc(close(&dsid));
		
		/*data OutLib.gendat1; set gendat; run;*/
	%end;
	
	%let OrigLGroupVar= &LGroupVar;
	%let OrigClusterVar= &ClusterVar;
	%let OrigClusterCenterVar= &ClusterCenterVar;
	%put OrigLGroupVar= &OrigLGroupVar;
	%put OrigClusterVar= &OrigClusterVar;
	
	%let KeepVars= &KeepVars _LGroupVar_ _ClusterVar_ _ClusterCenterVar_;
%end;
*end: Luciano, add dropped out markers*;

%if %length(&OrderVar &OrderGrpVar &OSkOrderVar)>0 %then %do;
	/* Use Order var and or Order group var from Order data set */
	/* Also use Skeleton order varo and Order Group var if both specified on Order data set*/

	data indat; 
		set InLib.&InData;
		*_Ord_ = _n_;
		orig_order= _n_;
		%if %upcase(&AddDroppedMarkers)= YES %then %do;
			%if %length(&OrigLGroupVar) %then
				_LGroupVar_= &OrigLGroupVar;;
			%if %length(&OrigClusterVar) %then
				_ClusterVar_= &OrigClusterVar;;
			%if %length(&OrigClusterCenterVar) %then
				_ClusterCenterVar_= &OrigClusterCenterVar;;
			%end;
		%if %length(&OrderVar) %then drop &OrderVar;;
		%if %length(&OrderGrpVar) %then drop &OrderGrpVar;;
		%if %length(&OSkOrdervar) %then drop &OSkOrderVar;;
	run;
	
	proc sql UNDO_POLICY=NONE; 
		create table indat as  
			select *
			from indat as d1 join OrderLib.&OrderData(keep = &OrderVar &OrderGrpVar &OSkOrderVar &AnnoMName) as d2
			on d1.&IDVar = d2.&AnnoMName;
	quit;
	
	%if &debug %then %do;
		data outlib.indat1;set indat;run;
	%end;
	
	/* kemicl - get number of markers in indata and  after merge in Order data */
	%nobsvars(InLib.&InData);
	%let nm_in = &nobs;
	%nobsvars(indat);
	%let nm_out = &nobs;
	/* If nobs doesn't match throw error and exit */
	%if &nm_in ne &nm_out %then %do;
		%put ERROR: Merging in the Order Data Set resulted in only &nm_out markers when the Input Data Set had &nm_in markers.;
		%put ERROR: Please check that the values of the Marker ID Merge Variable in the Order Data Set exactly matches the values of the Marker ID Variable in the Input Data Set.;
		%let exiterror = 1;
		%goto exit;
	%end;	
	
	proc sort data=indat out=indat;
		by orig_order;
	run;

	%if &syserr>4 %then %do;
		%put ERROR: Merging in the Order Data Set produced errors.  Please check that the Marker Merge Variable matches the Input Data Set ID Variable.;
		%let exiterror = 1;
		%goto exit;
	%end;
	%let indat = indat;
	
	/* use Order Group Var instead of Linkage Group var put on input data set */
	%if %length(&OrderGrpVar) %then %let LGroupVar = &OrderGrpVar;
	/* use Skeleton Order Var from Order data set instead of any SkOrderVar specified on input */
	%if %length(&OSkOrderVar) %then %let SkOrderVar = &OSkOrderVar;
%end;
%else %do;
	data indat;
		set InLib.&InData;
		%if %upcase(&AddDroppedMarkers)= YES %then %do;
			%if %length(&OrigLGroupVar) %then
				_LGroupVar_= &OrigLGroupVar;;
			%if %length(&OrigClusterVar) %then
				_ClusterVar_= &OrigClusterVar;;
			%if %length(&OrigClusterCenterVar) %then
				_ClusterCenterVar_= &OrigClusterCenterVar;;
		%end;
		orig_order=_n_;
	run;
	%let indat = indat;
%end;

%let lod_indat= ;
%let lod_data= ;
%if &LODDataFlag %then %do;
	%let lod_indat= &LODData;
	%let lod_data= lod_data;
	%put lod_data= &lod_data;	
%end;

%if &debug %then %do;
	data outlib.lmo_indat_orig1; set &indat; run;
%end;

%*goto exit;

*begin: Luciano, added a general where statement to filter out observations from the Input Data Set*;
%put WhereStmt= &WhereStmt;
%if %length(&WhereStmt) %then %do;
	%if %upcase(&AddDroppedMarkers)= YES %then %do;
		/*
		%if %length(&OrigLGroupVar) %then 
			%let WhereStmt= %sysfunc(tranwrd(&WhereStmt, &OrigLGroupVar, _LGroupVar_));
			
		*/
		
		data Orig_Indat;
			set &indat;
			*_order_= _n_;
		run;
		
	%end;
	%put WhereStmt= &WhereStmt;

	*subseting indat*;
	data KeepIndat;
		retain _countKeep_ _countDrop_ 0;
		set &indat end=_eof_;
		if &WhereStmt then do;
			output KeepIndat;
			_countDrop_ + 0;
			_countKeep_ + 1;
			call symputx("KeepMK"||trim(left(_countKeep_)), trim(left(&IDVar)));
		end;
		else do;
			_countKeep_ + 0;
			_countDrop_ + 1;
			call symputx("DropMK"||trim(left(_countDrop_)), trim(left(&IDVar)));		
		end;
		if _eof_ then do;
			call symputx("TotKeepMK", trim(left(_countKeep_)));
			call symputx("TotDropMK", trim(left(_countDrop_)));
		end;
		drop _countKeep_ _countDrop_;
	run;
	%put DropMK1= &DropMK1;
	%put TotDropMK= &TotDropMK;
	%if &debug %then %do;
		data outlib.lmo_Keepindat;set Keepindat;run;
	%end;
	
	%NObsVars(&indat);
	%let nobs_in= &nobs;
	%NObsVars(KeepIndat);
	%let nobs_out= &nobs;
	
	%if &nobs_in = &nobs_out %then %do;
		%put ERROR: No observation was removed from the Input Data Set when the Where Statement specified in the General Tab was applyied to this data set. Please return to the General Tab and change the Where Statement that is being specified.; 
		%let exiterror=1;
		%goto exit;
	%end;
	
	%if &nobs_out < 2 %then %do;
		%put ERROR: &nobs observations left in the Input Data Set after applying the Where Statement specified in the General Tab. At least 2 observations must remain in the Input Data Set after applying the Where Statement.; 
		%let exiterror=1;
		%goto exit;
	%end;

	*dropping unwanted columns*;
	data KeepIndat;
		set KeepIndat;
		drop 
		%do i=1 %to &TotDropMK;
			&&DropMK&i
		%end;
		;
	run;
	%let indat= KeepIndat;
	
	%let OrigListDistVars= &ListDistVars;
	%let OrigDistVars= &DistVars;
	
	%let ListDistVars= ;
	%let DistVars= ;
	%do i=1 %to &TotKeepMK;
		%let ListDistVars= &ListDistVars &&KeepMK&i;
	%end;
	%*let ListDistVars= &KeepMK1--&&KeepMK&TotKeepMK;
	%*let DistVars= ;
	
	%if %upcase(&AddDroppedMarkers)= YES %then %do;
		*checking if clusters are nested withing linkage groups*;
		proc sort data=&indat out=LGByCluster nodupkey;
			by &LGroupVar &ClusterVar;
		run;
		
		%let notnestedlg= 0;
		%let notnestedcl= ;
		data _null_;
			set LGByCluster;
			retain clus_;
			if &ClusterVar= clus_ then do;
				call symputx("notnestedlg", trim(left(1)));
				call symputx("notnestedcl", trim(left(&ClusterVar)));
			end;
			clus_ = &ClusterVar;
		run;
		
		%if &notnestedlg %then %do;
			%put ERROR: Cluster &notnestedcl appears in more than one linkage group.;
			%let exiterror=1;
			%goto exit;
		%end;
	%end;		
%end;
*end: Luciano, added a general where statement to filter out observations from the Input Data Set*;

/* if a skeleton order var exists, create a data set for subsetting for each linkage group to get any anchored marker positions*/
/* in Linkage Group loop, will subset and check if any markers in curr lgroup have skeleton positions to use ordinally in TSP */
%if %length(&SkOrderVar) %then %do;
	data skeleton; 
		set &indat;
		where not missing(&SkOrderVar);
		keep &IDVar &LGroupVar &SkOrderVar;
	run;
%end;

%if %length(&SkOrderVar) or %length(&OrderVar) %then %do;
	data tmp; 
		set &indat;
		where 0;
	run;
	
	%let dsid = %sysfunc(open(tmp));
	%if %length(&SkOrderVar) %then %do;
		%let SkPos=%sysfunc(varnum(&dsid,&SkOrderVar));
		%let SkType = %sysfunc(vartype(&dsid, &SkPos));
		%if &SkType = C %then %do;
			%put ERROR: The Framework Order Variable must be numeric.;
			%let exiterror = 1;
			%goto exit;
		%end;
	%end;
	%if %length(&OrderVar) %then %do;
		%let OPos = %sysfunc(varnum(&dsid,&OrderVar));
		%Let OType = %sysfunc(vartype(&dsid, &OPos));
		%if &OType = C %then %do;
			%put ERROR: The Marker Order Variable must be numeric.;
			%let exiterror = 1;
			%goto exit;
		%end;		
	%end;
	%let rc = %sysfunc(close(&dsid));
%end;

/* check type and Length of Linkage Group Variable and replace missing with -99 if numeric and MISSING if character */
data tmp; 
	set &indat;
	where 0;
	keep &LGroupVar;
run;

%let dsid = %sysfunc(open(tmp));
%let LGType = %sysfunc(vartype(&dsid, 1));
%let LGLen = %sysfunc(varlen(&dsid,1));
%let rc = %sysfunc(close(&dsid));

%if &LGType = C %then %do;
	%let LGLen = %sysfunc(max(&LGLen,7));
	%let LGLen = %eval(&LGLen +3);
%end;

%if &debug %then %do;
	data outlib.indat;set &indat;run;
%end;

data &indat; 
	 %if &LGType = C %then %do;
 		length &LGroupVar $&LGLen;
 	%end;
	%if &allmatch = 0 %then %do; merge MarkerList &indat; %end; %else %do; set &indat; %end;
 	%if &LGType = C %then %do;
		if missing(&LGroupVar) then &LGroupVar = "MISSING";
	%end;
	%else %do;
		if missing(&LGroupVar) then &LGroupVar = -99;
	%end;
run;

%if &debug %then %do;
	data outlib.indatbe;set &indat;run;
%end;

data &indat(keep =&LGroupVar &IDVar &DistVars &ListDistVars)  
	copydat(keep = &LGroupVar &IDVar &KeepVars &SkOrderVar &ColorVar orig_order &OrderVar &AnnoMName);
	set &indat; 
	*orig_order = _n_;
run;

%if &debug %then %do;
	data outlib.indataf;set &indat;run;
	data outlib.copydat;set copydat;run;
%end;

/*Get Levels of the Linkage Group Var (in case it's been recoded) */
/* Loop over linkage groups and create data sets for each */
/*
%let nGrps = ;
%*GetLevels(_Ds =indat,_Var=&LGroupVar,_LPrefix = LGrp,_nLvls=nGrps,_vType= LGType);
%GetLevels(_Ds =&indat,_Var=&LGroupVar,_LPrefix = LGrp,_nLvls=nGrps,_vType= LGType);
%put There are &nGrps Linkage Groups for ordering;
%put The linkage group variable is of type &LGType;
%if &syserr>4 %then %goto exit;
*/

/* Set up stored macro library if using Map Order Optimization*/ 
%put Algorithm =&Algorithm;
%if %index(%nrbquote(%upcase(&Algorithm)),OPTIMIZATION)>0 %then %do;
	options mstored sasmstore=orlib;
	%if %length(%sysfunc(pathname(orlib))) = 0 %then %do;
		libname orlib "&MacroPath.";
	%end;
 	/* define macros for solving for optimal order in PROC OPTMODEL macros */
	%Global exiterror relobjgap1 relobjgap2 maxtime;
	%let relobjgap1 = ;
	%let relobjgap2 = ;
	%let maxtime = ;
	%if %upcase(&UseStoppingRules) = YES %then %do;
		/* convert time to seconds */
		%let MaxTimeStep = %sysevalf(&MaxTimeStep*60);
		%let OptGap1 = %sysevalf(&OptimalityGap*2);
		%let relobjgap1 = relobjgap=&OptGap1;
		%let relobjgap2 = relobjgap=&OptimalityGap;
		%let maxtime = maxtime=&MaxTimeStep;
		%put The Algorithm will solve the marker order within &OptimalityGap of optimality with max time of &MaxTimeStep seconds per MILP solve.;
	%end;
%end;

%let UseMapDistance = NO;
/* use this macrovar to see if a group had an order created, if 1 then dim data set exists otherwise it was a singleton group*/
%global Singletons sflag allsflag OrderCreated BreakMaxCheck;
%let BreakMaxCheck = 0;
%let EverBroke = 0;  /*update this if a group was ever broken so I know to resort lmo data */
/* create a macrovariable  storing the linkage groups and markers that have only one marker */
%let Singletons = ; 
%let sflag = 0;
%let allsflag = 1;

%macro OrderMarkers(_Group,iter,OrigGrp=);

	/*data OutLib.lmo_lg0_&iter; set _lg; run;*/

	data lgdata;
		retain &LGroupVar &IDVar %do j=1 %to &LGtot; &&MVar&j %end;;
		set _lg;
		where &LGroupVar = &_Group;
		keep &LGroupVar &IDVar %do j=1 %to &LGtot; &&MVar&j %end;;
	run;
	%put Number of markers in Group &_Group = &LGTot;
	
	/*Luciano moved this data step to this location to fix defect S1039528*/
	data namesformerge; set lgdata(keep = &IDVar);
		row = _n_;
	run;
	
	/*data outlib.lmo_lgdata0_&iter; set lgdata; run;*/

	%nobsvars(lgdata);
	%put nobs = &nobs;
	%let nvars = %eval(&nvars - 2);
	%put nvars = &nvars;
	
	/* check that the matrix is symmetric for the current linkage group */
	%if &nobs ne &nvars %then %do;
		%put ERROR: The matrix of recombination fractions for &LGroupVar = &_Group is not a square matrix of values.;
		%put ERROR: There are &nobs marker rows and &nvars marker columns, which should be equal values. Please check the input data.;
		%let exiterror =1;
		%goto exito;
	%end;
	
	%if &LODDataFlag %then %do;
	
		/*data OutLib.lmo_lglod0_&iter; set _lglod; run;*/
	
		data &lod_data;
			length &IDVar $50;
			if 0 then set _lg;
			if _N_ = 1 then do;
				declare hash h(dataset:"_lg");
				h.defineKey("&IDVar");
				h.defineData("&IDVar");
				h.defineDone();
			end;
			set _lglod;
			if h.find() = 0;
			keep &IDVar %do j=1 %to &LGtot; &&MVar&j %end;;
		run;	
		
		/*data outlib.lmo_lod_data0_&iter; set &lod_data; run;*/	
	%end;	
	
	%if &nobs <2 %then %do;
		%put Note: A linkage group must contain at least 2 markers to compute order and distances ;
		%put Note: There is only one marker in the linkage group, &LGroupVar = &_Group... Please add this marker to another group if you want it included in analysis. ;
		data lgdata;
			set currcopydat;
			keep &LGroupVar &IDVar &ColorVar &KeepVars Original_Group;
			%if %length(&OrigGrp) %then %do;
				%if %UPCASE(&LGType) = C %then %do;
					length Original_Group $&LGLen;
				%end;
				Original_Group = &OrigGrp;
			%end;
		run;

		%if &sflag=0 %then %do;
			%let Singletons = "&_Group";
			*data Outlib.&LinkageMapOrder_OutDataExName; 
			data singletons_all; /*Luciano*/
				set lgdata;
			run;
		%end;
		%else %do;
			%let Singletons = &Singletons "&_Group";
			/*proc append base=Outlib.&LinkageMapOrder_OutDataExName data=lgdata; run;*/
			proc append base=singletons_all data=lgdata; run;
		%end;
		%let sflag = 1;
		%put Singleton List = &Singletons;
		%let OrderCreated = 0;
	%end; /*end: &nobs <2, where nobs is number of markers on a linkage group*/
	%else %do;
		%let OrderCreated = 1;
		%let allsflag = 0;
		%let nobscheck = &nobs;

		%if %length(&OrderVar) or %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)=0 %then %do;
			/* kemicl change to PROC SQL to avoid data OOM 
			create a data set of names for output when MDS isn't run to get stacked data 
			*/
			proc sql;
				create table stackedmlist as
				select f1.&IDVar, f1.row1, f2._COL_, f2.row2
				from namesformerge(rename = (row=row1)) as f1 
				cross join namesformerge(rename = (&IDVar=_COL_ row=row2)) as f2
				where f1.row1 <= f2.row2
				order by f2.row2, f1.row1
			;
			quit;
			
			/*data outlib.stackedmlist;set stackedmlist;run;*/
			
			data stackedmlist; set stackedmlist;
				rename &IDVar = _ROW_;
				drop row1 row2;
				%if %upcase(&AddDroppedMarkers)= YES %then
					&LGroupVar = &_Group;; /*Luciano: needed if adding markers back */
			run;	
		%end;

		%if ^%length(&OrderVar) %then %do;
			%if %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0 %then %do;
				/* no order is specified, calculate from MDS */

				%if %Upcase(&UseMapDistance) = YES %then %do;
					data lgdata;	 set lgdata;
						array d{*} %do j=1 %to &LGtot; &&MVar&j %end; ;
						do i = 1 to &LGtot;
							if not missing(d{i}) then do;
								%if %Upcase(&MapFunc) = HALDANE %then %do;
									d{i} = min(200, 100*(-(1/2)*log(1-2*min(d{i},0.49))));
								%end;
								%else %do;
									d{i} = min(200, 100*((1/4)*log((1+2*min(d{i},0.49))/(1-2*min(d{i},0.49)))));
								%end;
							end;
						end;
						drop i;
					run;
				%end;
				
				/*Never enters the if condition because UseWeights=NO in the xml file*/	
				%put UseWeights= &UseWeights;
				%if %Upcase(&UseWeights) = YES %then %do;
					data lgdata;	 set lgdata;
						array d{*}  %do j=1 %to &LGtot; &&MVar&j %end;;
						array weights{&LGtot};
						do i = 1 to &LGtot;
							weights{i} = 1-d{i};
						end;
						drop i;
					run;
				%end;			

				%put MDS order processing for Linkage Group = &_Group;
				options DKROCOND=nowarn;

				proc mds data=lgdata dim=1 formula = 2 %if %upcase(&MDSSqRoot)= YES %then fit = 0.5;
					OCONFIG pfinal pfit 
					out=Moutdim(drop=_name_ _matrix_ _type_)
					outres=Moutres outfit=mf;
					var %do j=1 %to &LGtot; &&MVar&j %end; ;
					%if %Upcase(&UseWeights) = YES %then %do;
						weight weights:;
					%end;
					id &IDVar;
				run;

				%if &syserr>4 %then %goto exito;

				data Moutres; set Moutres;
					*&LGroupVar = &_Group;
					%if %upcase(&AddDroppedMarkers)= YES %then 
						&LGroupVar = &_Group;; /*Luciano: needed if adding markers back */
					rename Data = Recombination;
					label FitDist = "MDS Distance" _ROW_ = ' ' _COL_= ' ' Data = "Recombination";
					drop _MATRIX_ _DIMENS_ TranData Distance ;
				run;

				data mf; set mf;
					%if &LGType = C %then %do;
						length &LGroupVar $&LGLen;
					%end;
					&LGroupVar = &_Group;
					if missing(_MATRIX_) then delete;
				run;

				/* merge back in orig_order and copyvars */
				proc sql; 
					create table dims as 
						select *
						from currcopydat d1, Moutdim d2
						where d1.&IDVar = d2.&IDVar;
				quit;

				proc sort data = dims out=dims(drop = Dim1 _DIMENS_);
					by Dim1;
				run;

				data dims; set dims;
					Init_Map_Order= _n_;
					%if %UPCASE(&LGType) = C %then %do;
						NumLinkageGroup = &iter;
					%end;
				run;

				/* data set with a column of marker orders, ordered by the orignal RF matrix, ready for IML manipulation */
				proc sort data = dims out=_ord_1(keep = Init_Map_Order);
					by orig_order;
				run;

				%put &nobscheck;
				%if &nobscheck > 2 %then %do;
					proc iml;
						start CostFunction(r) global(recomb,s,iter); 
							nc = ncol(r);
							cost_r =0; 
							do i = 1 to nc-1; 
							  cost_r = cost_r + recomb[loc(r=i),loc(r=i+1)]; 
							end; 
							s = r;
							if recomb[loc(s=2),loc(s=3)]>recomb[loc(s=1),loc(s=3)] then do;
								xi = loc(s=1);
								xj = loc(s=2);
								sxi =s[loc(s=1)];
								sxj = s[loc(s=2)];
								s[xi] = sxj;
								s[xj] =  sxi;
							end;
							/* local optimization  */ 
							do i = iter+1 to nc-(iter+1);
									tmp12 = recomb[loc(s=i-iter),loc(s=i)];
									tmp34 = recomb[loc(s=i+iter),loc(s=i+(iter+1))];
									tmp13 = recomb[loc(s=i-iter),loc(s=i+iter)];
									tmp24 = recomb[loc(s=i),loc(s=i+(iter+1))];
									if tmp12 +tmp34 > tmp13+tmp24 then do;
										xi = loc(s=i);
										xj = loc(s=i+iter);
										sxi =s[loc(s=i)];
										sxj = s[loc(s=i+iter)];
										s[xi] = sxj;
										s[xj] =  sxi;

									end;
							end;

							if recomb[loc(s=nc-1),loc(s=nc)]>recomb[loc(s=nc-2),loc(s=nc)] then do;
								xi = loc(s=nc-1);
								xj = loc(s=nc);
								sxi =s[loc(s=nc-1)];
								sxj = s[loc(s=nc)];
								s[xi] = sxj;
								s[xj] =  sxi;
							end;

							/* compute objective function */ 
							cost_s =0; 
							do i = 1 to nc-1; 
							  cost_s = cost_s + recomb[loc(s=i),loc(s=i+1)]; 
							end; 
							if cost_s < cost_r then
								cost = cost_s;
							else do;
								cost = cost_r;
								s = r;
							end;
							/* 
							if uniform(1234)<=p then 
								r = s; 
							*/
							return (cost); 
						finish; 

						use lgdata;
						read all var{ %do j=1 %to &LGtot; &&MVar&j %end;} into recomb;
						use _ord_1;
						read all into ord;
						nloci = ncol(recomb);

						/* run cost function with MDS order */
						init_o = T(ord);
						*print init_o;
						initcost =0; 
						do i = 1 to nloci-1; 
						  initcost = initcost + recomb[loc(init_o=i),loc(init_o=i+1)]; 
						end; 
						iter = 1;
						newcost = CostFunction(init_o);
						print "Local Optimization of MDS Solution";
						print "Initial Cost"  initcost  "		Local Optimized Cost" newcost;
						newo = s;
						maxiter = min(10, floor(nloci/4));
						do iter = 2 to maxiter; 
							oldcost = newcost;
							newcost = CostFunction(newo);
							if oldcost <= newcost then i=maxiter;
							else do;
								newo = s;
								print "Previous Cost" oldcost "New Cost" newcost;
							end;
						end; 				

						bestmembert = T(newo);
						create ordermat from bestMembert;
						append from bestMembert;
					quit;

					data _ord_; set ordermat(rename= (col1 = Map_Order)); run;

					data tord; merge _ord_1 ordermat(rename= (col1 = Map_Order)); run;	

					proc sort data=tord;
						by Init_Map_Order;
					run;

					data dims; merge dims tord; by Init_Map_Order;run;

					proc sort data=dims; by Map_Order; run;
				%end; /*end: &nobscheck > 2, where nobscheck is the number of markers on a linkage group*/
				%else %do;
					data dims; set dims;
						Map_Order = Init_Map_Order;
					run;

					proc sort data = dims out=_ord_(keep = Map_Order);
						by Map_Order;
					run;
				%end;
			%end; /*end: %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0 */
			%else %do;
				%*put MacroPath= &MacroPath;
				%let foundcatmacro=%sysfunc(fileexist(&MacroPath.\sasmacr.sas7bcat));
				%put foundcatmacro= &foundcatmacro;

				%if &foundcatmacro = 0 %then %do;
					%put ERROR: The stored macro catalog "sasmacr.sas7bcat" was not found in "&MacroPath.".;
					%put This file should be copied from;
					%put "&MacroPath.\Stored32" or;
					%put "&MacroPath.\Stored64" depending on your OS.;
					%let exiterror=1;
					%goto exito;
				%end;
				
				/*data OutLib.lmo_lgdata; set lgdata; run;*/
				/*
				%if &LODDataFlag %then %do;
					data OutLib.lmo_lod_data; set &lod_data; run;
				%end;
				*/
				
				/* use PROC OPTmodel to perform Map Order Optimization using TSP algorithms*/
				/* constraint on maximum recomb rate between markers placed next to each other */
				%put Recombination constraint threshold = &Threshold;
				%put LOD constraint threshold = &LODThreshold;
				
				%if %length(&SkOrderVar) %then %do;
					%nobsvars(skdata_lg);

					%if &nobs > 1 %then %do;
						%put At least two markers in linkage group &_Group have skeleton positions.  Constrained TSP will be used.;
						/* constrained TSP macro to honor ordinality of skeleton markers */
						%ClockStart(Linkage Group OPTMODEL PROCESSING);
						%SkeletonLMO(indata=lgdata, skeletondata=skdata_lg, SPosVar=&SkOrderVar);
						%ClockStop(Linkage Group OPTMODEL PROCESSING);
					%end;
					%else %do;
						%put Markers in linkage group &_Group have no skeleton positions.  General TSP will be used.;
						/* no skeleton marker constraints for current linkage group- general TSP solver */
						/* iterative solution using the subtour formulation from SAS PROC OPTMODEL*/
						%ClockStart(Linkage Group OPTMODEL PROCESSING);
						%if &LODThreshold > 0 %then %do;
							%LMOOptModel(rf_indata=lgdata, lod_indata= &lod_data);
						%end;
						%else %do;
							%LMOOptModel(rf_indata=lgdata, lod_indata= );
						%end;
						%ClockStop(Linkage Group OPTMODEL PROCESSING);
					%end;
				%end;
				%else %do;
					/* no skeleton marker constraints - general TSP solver */
					/* iterative solution using the subtour formulation from SAS PROC OPTMODEL*/
					%if %nrbquote(%upcase(&Algorithm)) = MAP ORDER OPTIMIZATION %then %do;
						%put Algorithm= OPTMODEL;
						%ClockStart(Linkage Group OPTMODEL PROCESSING);
						%if &LODThreshold > 0 %then %do;
							%LMOOptModel(rf_indata=lgdata, lod_indata= &lod_data);
						%end;
						%else %do;						
							%LMOOptModel(rf_indata=lgdata, lod_indata= );
						%end;
						%ClockStop(Linkage Group OPTMODEL PROCESSING);
						
					%end;
					%else %if %index(%nrbquote(%upcase(&Algorithm)),ACCELERATED)=1 %then %do;
						%put Algorithm= OPTNET;
						%ClockStart(Linkage Group OPTNET PROCESSING);
						%if &LODThreshold > 0 %then %do;
							%LMOOptNet(rf_indata=lgdata, lod_indata= &lod_data);
						%end;
						%else %do;
							%LMOOptNet(rf_indata=lgdata, lod_indata= );
						%end;
						%ClockStop(Linkage Group OPTNET PROCESSING);
					%end;
				%end;

				%if &exiterror %then %do;
					%let LODConstraintMsg= ;
					%if &LODDataFlag %then %do;
						%let LODConstraintMsg = %str(and/or decrease the Nearby LOD Constraint);					
					%end;
						
					%if %nrbquote(%upcase(&Algorithm)) = MAP ORDER OPTIMIZATION %then %do;
						%put ERROR: Map Order Optimization is infeasible for Linkage Group &_Group..;
						%put You may need to increase the Nearby Marker Recombination Constraint &LODConstraintMsg on;
						%put the Optimization Options Tab or further split this linkage group.;
					%end;
					%else %if %index(%nrbquote(%upcase(&Algorithm)),ACCELERATED)=1 %then %do;
						%put ERROR: Accelerated Map Order Optimization is infeasible for Linkage Group &_Group..;
						%put You may need to increase the Nearby Marker Recombination Constraint &LODConstraintMsg on;
						%put the Optimization Options Tab or further split this linkage group.;	
					%end;
					%goto exito;
				%end;
				
				data dims; 
					set currcopydat;
				run;
				
				data ordermat;
					set solData2; 
					keep Order;
				run;
				
				data dims; merge dims ordermat(rename= (Order = Opt_Order)); run;
				
				proc sort data=dims;
					by Opt_order;
				run;
				
				data dims; set dims;
					Map_Order= _n_;
					%if %UPCASE(&LGType) = C %then %do;
						NumLinkageGroup = &iter;
					%end;
					drop Opt_Order; /*Luciano*/
				run;
				/* data set with a column of marker orders, ordered by the orignal RF matrix, ready for IML manipulation */
				proc sort data = dims out=_ord_(keep = Map_Order);
					by orig_order;
				run;
			%end; /*end: Map Order Optimization or Accelerated Map Order Optimization*/
		%end; /*end: ^%length(&OrderVar)*/
		%else %do;
			/* Use Order from Order data set */
			data dims;
				set currcopydat;
			run;
			proc sort data = dims;
				by &OrderVar;
			run;
			data dims; set dims;
				Map_Order= _n_;
				%if %UPCASE(&LGType) = C %then %do;
					NumLinkageGroup = &iter;
				%end;
			run;
			/* data set with a column of marker orders, ordered by the orignal RF matrix, ready for IML manipulation */
			proc sort data = dims out=_ord_(keep = Map_Order);
				by orig_order;
			run;				
		%end;
		
		/*data OutLib.lmo_lgdata&iter; set lgdata; run;*/
		/*
		%if &LODDataFlag %then %do;
			data OutLib.lmo_lod_data&iter; set &lod_data; run;
		%end;
		*/

		proc iml;
			use lgdata;
			read all var{ %do j=1 %to &LGtot; &&MVar&j %end;} into r;
			close lgdata;
			use _ord_;
			read all into ord;
			close _ord_;
			nm= nrow(ord);
			%if &LODDataFlag %then %do;
				use &lod_data;
				read all var{ %do j=1 %to &LGtot; &&MVar&j %end;} into lod;
				close &lod_data;
				lodstep = J(nm,1,.);
			%end;
			cmstep = J(nm, 1, .);
			cmstep[1,1] = 0;
			rfstep = J(nm,1,.);
			cmtot = J(nm, 1, .);
			cmmax = J(1,1,.);
			rfmax = J(1,1,.);
			do i = 1 to nm-1;
				do j = i to nm;
					if j-i = 1 then do;
						/* add in option to output recombination frequency between ordered markers */
						rfstep[i+1] = r[loc(ord=i),loc(ord=j)];
						%if &LODDataFlag %then %do;
							lodstep[i+1] = lod[loc(ord=i),loc(ord=j)];
						%end;
						%if %Upcase(&UseMapDistance) = YES %then %do;
							cmstep[i+1] = r[loc(ord=i),loc(ord=j)];
						%end;
						%else %do;
							if r[loc(ord=i),loc(ord=j)] = . then do;
								/* the recomb freq is missing, something is wrong */
								cmstep[i+1] = .;
							end;
							else if r[loc(ord=i),loc(ord=j)] >= 0.5 then do;
								cmstep[i+1] = 200;
							end;
							else do;
								%if %Upcase(&MapFunc) = HALDANE %then %do;
									/* Haldane's Function */
									cmstep[i+1] = 100*(-(1/2)*log(1-2*r[loc(ord=i),loc(ord=j)]));
								%end;
								%else %do; /* Kosambi */
									cmstep[i+1] = 100*((1/4)*log((1+2*r[loc(ord=i),loc(ord=j)])/(1-2*r[loc(ord=i),loc(ord=j)])));
								%end;
							end;
							/* cap at 200 */
							if cmstep[i+1] > 200 then cmstep[i+1] = 200;
						%end;
					end;
				end;
				cmtot[i,1] = sum(cmstep[1:i]);
			end;
			cmtot[nm,1] = sum(cmstep[1:nm]);
			%if %upcase(&BreakGrps) = YES %then %do;
				&BGimlprefix.&BreakFunc = &BreakFunc(&BGimlprefix.step[1:nm]);
				if &BGimlprefix.&BreakFunc &BreakOperator &BreakGrpsValue then call symput("BreakMaxCheck",trim(left(char(1))));
			%end;
			%if %length(&OrderVar) or %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)=0 %then %do;
				rstack = symsqr(r);

				create rstack from rstack;
				append from rstack;
			%end;

			create cmmat from cmtot;
			append from cmtot;
			create cmstep from cmstep;
			append from cmstep;
			create rfstep from rfstep;
			append from rfstep;
			%if &LODDataFlag %then %do;
				create lodstep from lodstep;
				append from lodstep;			
			%end;
		quit;
		
		%put BreakMaxCheck = &BreakMaxCheck..  A value of 1 means Linkage Group &_Group will be broken.;
		%if &syserr>4 %then %do;
			%put ERROR: The genetic distances could not be calculated.;
			%let exiterror=1;
			%goto exito;
		%end;		

		data dims; 
			merge dims 
				cmmat(rename =(col1 = cM_&MapFunc)) 
				cmstep(rename =(col1 =cM_Step)) 
				rfstep(rename=(col1=RecombinationF))
				%if &LODDataFlag %then %do;
					lodstep(rename=(col1=LOD))
				%end;
				;
			cM_&MapFunc = round(cM_&MapFunc,0.01);
			cM_Step = round(cM_Step,0.01);
			RecombinationF = round(RecombinationF, 0.001);
			%if &LODDataFlag %then %do;
				LOD = round(LOD, 0.001);
			%end;
			%if %length(&OrigGrp) %then %do;
				%if %UPCASE(&LGType) = C %then %do;
					length Original_Group $&LGLen;
				%end;
				Original_Group = &OrigGrp;
			%end;
		run;
		
		%if &debug %then %do;
			data outlib.lmo_dims&iter; set dims; run;
		%end;
	%end; /*end: &nobs >= 2, where nobs is the number of marker on a linkage group*/
	%exito:;
%mend OrderMarkers;

%let AppendStarted = 0;
/* macro for appending the dims data out of ordering into a complete data set for lmo and lmr;*/
%macro BuildLMOData(first=);
		/* need descending for computing reverse order */
		proc sort data=dims;	
			by &LGroupVar descending Map_Order;
		run;
		data dims; set dims;
			retain _RMO_ _cMR_ ;
			by &LGroupVar;
			_cMStepR_ = lag1(CM_step);
			_rfR_ = lag1(RecombinationF);
			%if &LODDataFlag %then %do;
				_LODR_ = lag1(LOD);
			%end;
			if first.&LGroupVar then do;
				_cMR_ = 0;
				_cMStepR_ = 0;
				_rfR_ = .;
				_RMO_ = 1;
				%if &LODDataFlag %then %do;
					_LODR_ = .;
				%end;				
			end;
			else do;
				_RMO_ = _RMO_ + 1;
				_cMR_ = _cMR_ + _cMStepR_;
			end;
			if CM_step >= 200 then 
				call symputx("flag_rf_error", 1);
		run;

		proc sort data=dims;	
			by &LGroupVar Map_Order;
		run;

		/* create stacked recombination matrix output for interactive plots */
		%if %length(&OrderVar) or %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)=0 %then %do;
			data Moutres;
				merge stackedmlist rstack(rename = (col1 = Recombination));
				label Recombination = "Recombination";
				if _ROW_ = _COL_ then delete;
			run;
		%end;

		%if &first %then %do;
			data outdims; set dims; run;
			data outres; set Moutres; run;
			%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0 %then %do;   /* fit data doesn't exist when order var spec'd */
				data outfit; set mf; run; 
			%end;
		%end;
		%else %do;

			proc append base=outdims data=dims;run;

			proc append base=Outres data=Moutres;run;

			%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0  %then %do; 
				proc append base=outfit data=mf;run; 
			%end;
		%end;

		%if ^%length(&OrderVar) and %nrbquote(%Upcase(&Algorithm))=MDS  %then %do;
			proc datasets library=work nolist;
				delete Moutdim Moutres mf;
			run;
		%end;
		%else %do;
			proc datasets library=work nolist;
				delete Moutdim Moutres ;
			run;
		%end;
		/* this macr has been run*/
		%let AppendStarted = 1;
%mend BuildLMOData;

%macro LoopOverLGs(nGrps_=, LGrpPrefix_=);

	/*data OutLib.lmo_indat; set &indat; run;*/

	/* loop over groups to do marker ordering and calculate genetic distance */
	%do i=1 %to &nGrps_;
		%put Processing Linkage Group &&&LGrpPrefix_&i;

		%global OLGrp&i;
		%let OLGrp&i = &&&LGrpPrefix_&i;
		%if %UPCASE(&LGType) = C %then %do;
			%let &LGrpPrefix_&i = "&&&LGrpPrefix_&i";
		%end;

		/* make data set of linkage group and list of the variables to keep for this linkage group for MDS processing data set 	*/
		data _lg; 
			*set indat end=_e_;
			set &indat end=_e_;
			where &LGroupVar = &&&LGrpPrefix_&i;  
			call symput("MVar"||trim(left(_n_)), trim(left(&IDVar)));
			if _e_ then call symput("LGtot",trim(left(_n_)));
		run;
		%put LGtot= &LGtot;
		
		%if &LODDataFlag %then %do;
			data _lglod;
				length &IDVar $50;
				if 0 then set _lg;
				if _N_ = 1 then do;
					declare hash h(dataset:"_lg");
					h.defineKey("&IDVar");
					h.defineData("&IDVar");
					h.defineDone();
				end;
				set &lod_indat(rename=(&LODIDVar=&IDVar));
				if h.find() = 0;
			run;	
		%end;

		/* Luciano moved this part to the begining of OrderMarkers() macro to fix defect S1039528;
		data namesformerge; set _lg(keep = &IDVar);
			row = _n_;
		run;
		*/

		/* get current copy data for linkage group merging.  need to break up so that can handle and change it for new groups*/
		data currcopydat;
			set copydat;
			where &LGroupVar = &&&LGrpPrefix_&i; 
		run;
		/* changed copydat reference in macro to currcopydat, break this up for new groups before recall of macro*/
		%if &debug %then %do;
			data outlib.copydatL;set copydat;run;
			data outlib.currcopydatL;set currcopydat;run;
		%end;
		
		%if %length(&SkOrderVar) %then %do;
			data skdata_lg; set skeleton;
				where &LGroupVar = &&&LGrpPrefix_&i;
			run;
		%end;

		 /*call macro defined above to order markers or detect singletons; */
		%OrderMarkers(&&&LGrpPrefix_&i, &i, OrigGrp = &&&LGrpPrefix_&i);
		%if &exiterror = 1 %then %goto exitL;

		%if &OrderCreated = 1 %then %do;	
			/* Kemicl check if too large of distance, if so, split groups and rerun order */
			%if &BreakMaxCheck %then %do;
				%let EverBroke = 1;
				%let OrigGrpValue = &&&LGrpPrefix_&i;
				data newgps;
					set dims(Keep = &IDVar &LGroupVar &BreakGrpsVar) end=last;
					Original_Group = &LGroupVar;
					if _n_ = 1 then do;
						%if %UPCASE(&LGType) = C %then %do;
							newLGroup = 1;
							NewNumLGrp = &bngrps;
						%end;
						%else %do;
							newLGroup = &bngrps;
						%end;
					end;
					retain NewLGroup;
					%if %UPCASE(&LGType) = C %then retain NewNumLGrp;;
					if &BreakGrpsVar &BreakOperator &BreakGrpsValue then do;
						NewLGroup+1;
						%if %UPCASE(&LGType) = C %then NewNumLGrp+1;;
					end;
					%if %UPCASE(&LGType) = C %then %do;
						&LGroupVar = trim(left(&LGroupVar))||"_"||trim(left(NewLGroup));
					%end;
					%else %do;
						if newLGroup ne &bngrps then &LGroupVar = newLGroup;
					%end;
					%if %UPCASE(&LGType) = C %then %do;
						if last then  call symput("bnGrps",trim(left(NewNumLGrp)));
					%end;
					%else %do;
						if last then  call symput("bnGrps",trim(left(NewLGroup)));
					%end;					
					keep &IDVar &LGroupVar %if %UPCASE(&LGType) = C %then NewNumLGrp; NewLGroup;
				run;

				%if &debug %then %do;
					data outlib.newgps; set newgps; run;
				%end;
				/* create macro variables for new groups to loop through for ordering from current original group*/
				%let NnewGrps = ;
				%GetLevels(_Ds =newgps,_Var=&LGroupVar,_LPrefix = NLGrp,_nLvls=NnewGrps,_vType= nLGType);
				%put New Number of Groups = &BnGrps;
				%put There are &NnewGrps new Linkage Groups for ordering broken from &&&LGrpPrefix_&i ;

				/* get values of the new linkage group numbers for the iteration count */
				%if %UPCASE(&LGType) = C %then %do;
					%let newitervar = NewNumLGrp;
					%let tmpdropvars = NewNumLGrp NewLGroup;
				%end;
				%else %do;
					%let newitervar = NewLGroup;
					%let tmpdropvars = NewLGroup;
				%end;

				proc sort data=newgps out=getiter nodupkey;
					by &newitervar;
				run;

				data _null_; 
					set getiter;
					call symput("niter"||trim(left(_n_)),trim(left(&newitervar)));
				run;

				data newlgdata;
					if 0 then set newgps(drop = &tmpdropvars);
					if _N_ = 1 then do;
						declare hash h(dataset:"newgps");
						h.defineKey("&IDVar");
						h.defineData("&IDVar","&LGroupVar");
						h.defineDone();
					end;
					set _lg(drop = &LGroupVar);
					if h.find() = 0;
				run;

				data newcopydat;
					if 0 then set newgps(drop = &tmpdropvars);
					if _N_ = 1 then do;
						declare hash h(dataset:"newgps");
						h.defineKey("&IDVar");
						h.defineData("&IDVar","&LGroupVar");
						h.defineDone();
					end;
					set currcopydat(drop = &LGroupVar);
					if h.find() = 0;
				run;

				%if %length(&SkOrderVar) %then %do;
					data newskdat;
						if 0 then set newgps(drop = &tmpdropvars);
						if _N_ = 1 then do;
							declare hash h(dataset:"newgps");
							h.defineKey("&IDVar");
							h.defineData("&IDVar","&LGroupVar");
							h.defineDone();
						end;
						set skdata_lg(drop = &LGroupVar);
						if h.find() = 0;
					run;
				%end;

				/* reset breakmax back to 0 so it doesn't keep looping through */
				%let BreakMaxCheck = 0;

				/*reloop through OrderMacro with new smaller groups here*/
				%do ng = 1 %to &NnewGrps;
					%put Loop through new linkage group &ng..;
					%if %UPCASE(&LGTYPE) = C %then %let NLGrp&ng = "&&NLGrp&ng";
					
					data _lg; 
						set newlgdata end=_e_;
						where &LGroupVar = &&NLGrp&ng;
						call symput("MVar"||trim(left(_n_)), trim(left(&IDVar)));
						if _e_ then call symput("LGtot",trim(left(_n_)));
					run;
						
					/*
					*dont't need to recreate _lglod because there is no recursion in this code*;
					%if &LODDataFlag %then %do;
						data _lglod;
							length &IDVar $50;
							if 0 then set _lg;
							if _N_ = 1 then do;
								declare hash h(dataset:"_lg");
								h.defineKey("&IDVar");
								h.defineData("&IDVar");
								h.defineDone();
							end;
							set &lod_indat(rename=(&LODIDVar=&IDVar));
							if h.find() = 0;
						run;	
					%end;
					*/

					%if &debug %then %do;
						data outlib._newlgdata&i; set newlgdata; run;
						data outlib._lg&i._&ng; set _lg; run;
					%end;

					data currcopydat; 
						set newcopydat;
						where &LGroupVar = &&NLGrp&ng;
					run;

					%if %length(&SkOrderVar) %then %do;
						data skdata_lg; 
							set newskdat;
							where &LGroupVar = &&NLGrp&ng;
						run;
					%end;

					 /*call macro defined above to order markers or detect singletons; */
					%if &ng = 1 %then %let newi = &i; 
					%else %let newi = &&niter&ng;
					%OrderMarkers(&&NLGrp&ng, &newi, OrigGrp = &&&LGrpPrefix_&i);
					%if &exiterror = 1 %then %goto exitL;
					%if &OrderCreated = 1 %then %do;	
						%if &AppendStarted = 0 %then %let fcheck = 1;
						%else %let fcheck = 0;
						%BuildLMOData(first=&fcheck);
					%end;
				%end;
				/*Once that done, at end after dim data appends, recreate macros for each lgroup value and final number of groups for jsl */
			%end; /* end for breaking groups and recreating order and dims data */
			%else %do;
				%if &AppendStarted = 0 %then %let fcheck = 1;
				%else %let fcheck = 0;
				%BuildLMOData(first = &fcheck);
			%end;
		%end;  /* end loop for if a marker belongs in a linkage group */
	%end;   /* end loop over linkage groups */
				
	%exitL:;
%mend LoopOverLGs;

%macro ErrorCheckAfterOrdering;
	%if &allsflag %then %do;
		%put ERROR: No linkage groups contain more than 1 marker.  All markers have been excluded from analysis.;
		%put ERROR: This process cannot be run without linkage groups containing two or more markers.;
		%let exiterror =1;
		%goto exitE;
	%end;
	
	%if %sysfunc(exist(outdims))=0 %then %do;
		%put ERROR: No linkage groups contain more than 1 marker after breaking groups using the specified threshold.  All markers have been excluded from analysis.;
		%put ERROR: Please increase the value for which to break linkage groups on the Analysis Tab.;
		%let exiterror =1;
		%goto exitE;
	%end;
	
	%exitE:;
%mend ErrorCheckAfterOrdering;

/*Begin: Luciano*/
/*Get Levels of the Linkage Group Var (in case it's been recoded) */
/* Loop over linkage groups and create data sets for each */
%let nGrps = ;
%*GetLevels(_Ds =indat,_Var=&LGroupVar,_LPrefix = LGrp,_nLvls=nGrps,_vType= LGType);
%GetLevels(_Ds =&indat,_Var=&LGroupVar,_LPrefix = LGrp,_nLvls=nGrps,_vType= LGType);
%put There are &nGrps Linkage Groups for ordering;
%put The linkage group variable is of type &LGType;

%if &syserr>4 %then %goto exit;

*make LoopOverLGs macro initialized to original number of groups, but if they get broke update to have new number of groups *;
%global bngrps;
%if &LGType = N %then %do;
	/**S1548417: to fix problem when linkage groups need to be broken down and are not 
	numbered with increments of one unit, for instance 1 and 3, instead of 1 and 2**/
	proc means data=tmp noprint;
		var &LGroupVar;
		output out=maxlg max=_maxlg_;
	run;
	
	data _null_;
		set maxlg;
		call symputx("bnGrps", _maxlg_);
	run;
%end;
%else %do;
	%let bnGrps = &nGrps;
%end;
%put bnGrps = &bnGrps;

%put AddDroppedMarkers= &AddDroppedMarkers;
%if %upcase(&AddDroppedMarkers) = NO %then %do;	
	%LoopOverLGs(nGrps_=&nGrps, LGrpPrefix_= LGrp);
	%if &exiterror= 1 %then 
		%goto exit;
	%if &sflag %then %do;
		data Outlib.&LinkageMapOrder_OutDataExName;
			set singletons_all;
		run;
	%end;
	%ErrorCheckAfterOrdering;
	%if &exiterror= 1 %then 
		%goto exit;
%end;
%else %do;
	%*let debug=1;

	*ordering cluster center markers*;
	%put BreakGrps= &BreakGrps;
	%let OrigBreakGrps= &BreakGrps;
	%let BreakGrps= NO;
	%put BreakGrps= &BreakGrps;
	%LoopOverLGs(nGrps_=&nGrps, LGrpPrefix_= LGrp);
	%if &exiterror= 1 %then 
		%goto exit;
	%ErrorCheckAfterOrdering;
	%if &exiterror= 1 %then 
		%goto exit;
		
	*clearing result data sets and storeing them on new data sets*;
	proc datasets nolist;
		delete oldoutdims oldoutres;
		change outdims= oldoutdims;
		change outres= oldoutres;
		%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0 %then %do; 
			delete oldoutfit;
			change outfit= oldoutfit;
		%end;
		%if &sflag %then %do;
			delete oldsingletons_all;
			change singletons_all= oldsingletons_all;
		%end;
	quit;
	
	%if &debug %then %do;
		data outlib.oldoutdims;set oldoutdims;run;
		data outlib.oldoutres;set oldoutres;run;
		%if &sflag %then %do;
			data outlib.oldsingletons_all; set oldsingletons_all;run;
		%end;
	%end;
		
	*collecting number of linkage groups after ordering cluster center markers*;
	%let nGrpsOrd=;
	%let LGTypeOrd=;
	%GetLevels(_Ds =oldoutdims,_Var=&LGroupVar,_LPrefix = LGrpOrd,_nLvls=nGrpsOrd,_vType= LGTypeOrd);
	%put There are &nGrpsOrd ordered Linkage Groups;
	%put nGrpsOrd= &nGrpsOrd;
			
	*collecting number of observations in the Orig_Indat data set*;
	%NObsVars(Orig_Indat);
	%let nobs_in= &nobs;
	
	%let ApeendSingleMk= 0; *indicates if appending started*;
	%let AppendNewODims= 0; *indicates if appending started*;
	
	*reseting marcro variables*;
	%let EverBroke_old= &EverBroke;
	%let EverBroke_out= 0;
	%let EverBroke_sin= 0;
	%let EverBroke= 0;
	%let sflag_old= &sflag;
	%let sflag_out= 0;
	%let sflag_sin= 0;
	%let sflag= 0;
	%let allsflag_old = &allsflag;
	%let allsflag_out= 1;
	%let allsflag_sin= 1;
	%let allsflag = 1;
		
	*loop over ordered LGs adding back dropped out markers*;
	%do ii=1 %to &nGrpsOrd;
		%put Processing ordered Linkage Group &&LGrpOrd&ii;
		
		*EverAddedMarkers indicates if current LG has markers to be added back*;
		%let EverAddedMarkers= 0;
			
		%*global OLGrp&i;
		%*let OLGrp&i = &&LGrpOrd&i;
		%if %upcase(&LGTypeOrd)= C %then %do;
			%let LGrpOrd&ii= "&&LGrpOrd&ii";
			%put LGrpOrd&ii= &&LGrpOrd&ii;
		%end;
	
		*collecting cluster center markers for the current ordered LG, and clusters within the current ordered LG*;
		data currLGrp;
			retain _count_ 0;
			set oldoutdims end=_eof_;
			*if &LGroupVar= &&LGrpOrd&ii and _ClusterCenterVar_ = 1 then do;
			if &LGroupVar= &&LGrpOrd&ii then do;
				 _count_ + 1;
				 call symputx("Clus"||trim(left(_count_)),trim(left(_ClusterVar_)));
				 call symputx("ClusCen"||trim(left(_count_)),trim(left(_ClusterCenterVar_)));
				 call symputx("MrkName"||trim(left(_count_)),trim(left(&IDVar)));
				 output;
			end;
			else  _count_ + 0;
			if _eof_ then call symputx("nMkByLG", trim(left(_count_)));
			drop _count_;
		run;
		%put nMkByLG= &nMkByLG;
		%put Clus&nMkByLG= &&Clus&nMkByLG;
		%if &debug %then %do;
			data outlib.currLGrp&ii;set currLGrp;run;
		%end;
		
		data updatedLGrp;
			set currLGrp;
			_OldMapOrder_= Map_Order;
			drop cM_&MapFunc cM_Step RecombinationF _RMO_ _cMR_ _cMStepR_ _rfR_ %if &LODDataFlag %then LOD _LODR_;;
		run;
		%if &debug %then %do;
			data outlib.updatedLGrpin&ii;set updatedLGrp;run;
		%end;

		*reseting parameter to split linkage groups*;
		%let BreakMaxCheck= 0;
							
		*loop over markers the current ordered LG*;
		%do jj=1 %to &nMkByLG;
			%put Processing cluster &&Clus&jj associated with marker center &&MrkName&jj of ordered Linkage Group &&LGrpOrd&ii .;
			%let sufix= lg&&LGrpOrd&ii..clu&&Clus&jj;
			
			%put ClusCen&jj= &&ClusCen&jj;
			%if &&ClusCen&jj = 1 %then %do;
				*if current marker is a cluster center, then proceed*;
				
				%let ListDistVars= ;
				%let DistVars= ;

				%if %upcase(&ClusType)= C %then %do;
					%let Clus&ii= "&&Clus&jj";
					%put Clus&ii= &&Clus&jj;
				%end;

				*collecting markers that belong to the current cluster*;
				data currClusIndat;
					retain _countKeep_ _countDrop_ 0;
					*set Orig_Indat(drop=_order_) end=_eof_;
					set Orig_Indat end=_eof_;
					if &ClusterVar= &&Clus&jj then do;
						_countDrop_ + 0;
						_countKeep_ + 1;
						call symputx("KeepMK"||trim(left(_countKeep_)), trim(left(&IDVar)));
						output currClusIndat;
					end;
					else do;
						_countDrop_ + 1;
						_countKeep_ + 0;
						call symputx("DropMK"||trim(left(_countDrop_)), trim(left(&IDVar)));	
					end;
					if _eof_ then do;
						call symputx("TotDropMK", trim(left(_countDrop_)));
						call symputx("TotKeepMK", trim(left(_countKeep_)));
					end;
					drop _countKeep_ _countDrop_;
				run;
				%put TotDropMK= &TotDropMK;

				%let ListKeepMK= ;
				%do kk= 1 %to &TotKeepMK;
					%let ListKeepMK= &ListKeepMK &&KeepMK&kk;
				%end;

				/*	
				*dropping marker columns that do not belong to the current cluster*;
				data currClusIndat;
					set currClusIndat;
					drop 
					%do u=1 %to &TotDropMK;
						&&DropMK&u
					%end;
					;
				run;
				*/	
				/*data outlib.currClusIndat&sufix;set currClusIndat;run;*/

				*error checking*;
				%NObsVars(currClusIndat);
				%let nobs_out= &nobs;
				%if &nobs_in = &nobs_out %then %do;
					%put ERROR: No observation was removed from the Input Data Set when the filter &ClusterVar= &&Clus&jj was applyied to this data set.; 
					%let exiterror=1;
					%goto exit;
				%end;

				*adding dropped out markers back to linkage group if current cluster have more than one marker*;
				%if &nobs_out > 1 %then %do;
					%put Adding markers from cluster &&Clus&jj back to ordered Linkage Group &&LGrpOrd&ii;
					%let EverAddedMarkers= 1;

					*create list of markers that belong to the current cluster plus additional markers*;
					*collecting markers that belong to the current cluster plus all markers up to the next right-cluster center marker*;
					%if &jj < &nMkByLG %then %do;
						%let jjl= %sysevalf(&jj);
						%let jjr= %sysevalf(&jj + 1);
					%end;
					%else %do;
						%let jjl= %sysevalf(&jj - 1);
						%let jjr= %sysevalf(&jj);						
					%end;
					
					/*data outlib.lmo_updatedLGrp&sufix;set updatedLGrp;run;*/
					
					data _null_ 
						%if &jj < &nMkByLG %then tailcurrLGrp;
						%if &jj > 1 %then headcurrLGrp;
						;
						retain _count_ 0 _lcenter_ 10000000 _rcenter_ 10000000;
						set updatedLGrp end=_eof_;
						if &IDVar="&&MrkName&jjl" then do;
							 _lcenter_= _n_;
						end;
						if &IDVar="&&MrkName&jjr" then do;
							 _rcenter_= _n_;
						end;
						if _n_ >= _lcenter_ and _n_ <= _rcenter_ then do;
							_count_ + 1;
							call symputx("PlusMK"||trim(left(_count_)), trim(left(&IDVar))); 
						end;
						%if &jj > 1 %then %do;
							if _n_ < _lcenter_ then output headcurrLGrp;
						%end;
						%if &jj < &nMkByLG %then %do;
							if _n_ > _rcenter_ then output tailcurrLGrp;
						%end;
						if _eof_ then call symputx("TotPlusMK", trim(left(_count_)));
						drop _count_ _lcenter_ _rcenter_;
					run;
					
					%let ListPlusMK=;
					%do kk= 1 %to &TotPlusMK;
						%let ListPlusMK= &ListPlusMK &&PlusMK&kk;
					%end;
		
					%let ListDistVars= &ListKeepMK &ListPlusMK;
					%put DistVars&sufix= &DistVars;
					%put ListDistVars&sufix= &ListDistVars;
					
					*creating genotypic data for recombination estimation with markers belonging the current cluster plus markers up to the next right-cluster center*;			
					data currgendat;
						set gendat;
						keep &ListDistVars;
					run;
					%if &debug %then %do;
						data outlib.currgendat&sufix;set currgendat;run;
					%end;

					*estimating recombination for the current cluster plus additional markers*;
					%rfest(_gendata_=currgendat, _annodata_=, _RandLG_AP_=0, _runchisq_=0, _swindow_=0, _wsize_=);
					%if &exiterror %then %goto exit;
					
					%if &debug %then %do;
						data outlib.rfmat0&sufix;set rfmat;run;
					%end;

					*adding a column of marker names to the rfmat data set for merging purpose*;
					data rfnames; 
						set rfmat;
						where 0;
					run;

					proc transpose data=rfnames out=rfnames;
					run;

					data rfmat;
						merge rfnames rfmat;
						_ord_=_n_;
					run;
					
					%if &debug %then %do;
						data outlib.rfmat1&sufix;set rfmat;run;
					%end;

					%*goto exit;

					*creating an indat data set for the current cluster plus additional markers*;
					proc sql UNDO_POLICY=NONE;
						create table currClusPlusIndat(drop= _name_ _ord_) as 
						select a.*, b.*
						from Orig_Indat(keep = &LGroupVar &IDVar &KeepVars &SkOrderVar &ColorVar orig_order &OrderVar &AnnoMName) as a join rfmat as b
						on a.&IDVar= b._name_
						order by _ord_;
					quit;
					
					%if &debug %then %do;
						data outlib.currClusPlusIndat&sufix;set currClusPlusIndat;run;
					%end;

					data currClusPlusIndat;
						set currClusPlusIndat(rename=(&LGroupVar= _OriginalGroup_));
						&LGroupVar= &&LGrpOrd&ii;
					run;
					%let indat= currClusPlusIndat;
					
					%if &debug %then %do;
						data outlib.indatout&sufix;set &indat;run;
					%end;

					*rebuild skeleton data set for the current cluster plus additional markers*;
					%if %length(&SkOrderVar) %then %do;
						data skeleton; 
							set &indat;
							where not missing(&SkOrderVar);
							keep &IDVar &LGroupVar &SkOrderVar;
						run;
					%end;

					*preparing final indat and copydat that will be used in the ordering process*;
					data &indat(keep =&LGroupVar &IDVar &DistVars &ListDistVars)  
						copydat(keep = &LGroupVar &IDVar &KeepVars &SkOrderVar &ColorVar orig_order &OrderVar &AnnoMName _OriginalGroup_);
						set &indat; 
						*orig_order = _n_;
					run;
					%if &debug %then %do;
						data outlib.finalindat&sufix;set &indat;run;
						data outlib.finalcopydat&sufix;set copydat;run;
					%end;

					*begin: preparing data sets for the OrderMarkers macro*;
					* make data set of linkage group and list of the variables to keep for this linkage group for MDS processing data set 	*;
					data _lg; 
						set &indat end=_e_;
						where &LGroupVar = &&LGrpOrd&ii;  
						call symput("MVar"||trim(left(_n_)), trim(left(&IDVar)));
						if _e_ then call symput("LGtot",trim(left(_n_)));
					run;
					
					%if &LODDataFlag %then %do;
						*adding a column of marker names to the lodmat data set for merging purpose*;
						data lodnames; 
							set lodmat;
							where 0;
						run;
	
						proc transpose data=lodnames out=lodnames(rename= (_name_= &IDVar));
						run;
	
						*data lodmat;
						data _lglod;
							merge lodnames lodmat;
							_ord_=_n_;
						run;
						
						%if &debug %then %do;
							data outlib.lmo_lg_lod1&sufix; set lodmat; run;
						%end;
						
						%*let lod_indat= lodmat;
						%let lod_indat= _lglod;
					%end;					
					
					%if &debug %then %do;
						data outlib._lg&sufix; set _lg; run;
					%end;

					* get current copy data for linkage group merging.  need to break up so that can handle and change it for new groups*;
					data currcopydat;
						set copydat;
						where &LGroupVar = &&LGrpOrd&ii; 
					run;
					* changed copydat reference in macro to currcopydat, break this up for new groups before recall of macro*;

					%if %length(&SkOrderVar) %then %do;
						data skdata_lg; set skeleton;
							where &LGroupVar = &&LGrpOrd&ii;
						run;
					%end;
					*end: preparing data sets for the OrderMarkers macro*;

					*call macro defined above to order markers or detect singletons; *;
					%OrderMarkers(&&LGrpOrd&ii, &ii, OrigGrp = &&LGrpOrd&ii);
					%if &exiterror = 1 %then %goto exit;
					
					%if &debug %then %do;
						data outlib.lmo_dims0&sufix;set dims;run;
					%end;
				
					*checking if it is needed to flip the order of new order after adding markers back*;
					proc sql UNDO_POLICY= NONE;
						create table dims(rename=(_OriginalGroup_=Original_Group) 
							drop=_MK_ cM_&MapFunc cM_Step RecombinationF %if &LODDataFlag %then LOD;) as
						select a.*, b._MK_, b._OldMapOrder_
						/*from dims as a left join currLGrp(rename=(&IDVar= _MK_ Map_Order=_MapOrder_)) as b*/
						from dims(drop=Original_Group) as a left join updatedLGrp(rename=(&IDVar= _MK_)) as b
						on a.&IDVar= b._MK_
						order by Map_Order;
					quit;
					
					%if &debug %then %do;
						data outlib.lmo_dims1&sufix;set dims;run;
					%end;

					data dimscheckorder;
						set dims;
						where not missing(_OldMapOrder_);
					run;

					%NObsVars(dimscheckorder);
					%if &nobs < 2 %then %do;
						%put ERROR: Cluster &&Clus&jj on Linkage Group &&LGrpOrd&ii must have at least two anchoring ordered markers.;
						%let exiterror=1;
						%goto exit;
					%end;

					data _null_;
						set dimscheckorder end=_eof_;
						if _n_= 1 then call symputx("lmkord", trim(left(_OldMapOrder_)));
						if _eof_ then call symputx("rmkord", trim(left(_OldMapOrder_)));
					run;

					%put lmkord= &lmkord;
					%put rmkord= &rmkord;
					%if &lmkord > &rmkord %then %do;
						proc sort data=dims;
							by descending Map_Order ;
						run;

						data dims;
							set dims;
							Map_Order= _n_;
						run;
					%end;
					
					%if &debug %then %do;
						data outlib.lmo_dims2&sufix;set dims;run;
					%end;
						
					*updating updatedLGrp is necessary before starting the loop again*;
					%if &jj > 1 %then %do;
						proc append base=headcurrLGrp data=dims;
						run;
						
						proc datasets nolist;
							delete dims;
							change headcurrLGrp= dims;
						run;
					%end;
					%if &jj < &nMkByLG %then %do;
						/*data outlib.lmo_tailcurrLGrp&sufix;set tailcurrLGrp;run;*/
					
						proc append base=dims data=tailcurrLGrp;
						run;
					%end;
										
					data updatedLGrp;
						set dims;
						Map_Order=_n_;
						_OldMapOrder_= Map_Order;
					run;
					 
					%*goto exit;
					
				%end; /*end: if &nobs_out >= 2, cluster has more than one marker */
				%else %do;
					%put Cluster &&Clus&jj associated with marker cluster center &&MrkName&JJ has only one marker, so there is no marker to be added back near this marker center.; 
				%end; /*end: if &nobs_out = 1, cluster has one marker */
			%end; /*end if current marker is a cluster center*/
			%else %do;
				%put Marker &&MrkName&jj is not a cluster center.; 
			%end;  /*end if a marker is NOT a cluster center*/
			
			%if &debug %then %do;
				data outlib.updatedLGrp&sufix;set updatedLGrp;run;
			%end;
			
			%*goto exit;
		%end; /*end: loop over clusters within current linkage group*/
		
		%if &debug %then %do;
			data outlib.updatedLGrpO&ii; set updatedLGrp;run;
		%end;
		
		%if &EverAddedMarkers = 1 %then %do;
			
			proc transpose data= updatedLGrp(keep=&IDVar) out=mnamesUpLGrp(drop=_name_);
				ID &IDVar;
			run;
			
			%if &debug %then %do;
				data outlib.mnamesUpLGrp;set mnamesUpLGrp;run;
			%end;
			
			%let _uplgmk_= ;
			%let _idlg_= %sysfunc(open(mnamesUpLGrp));
			%let _nvlg_= %sysfunc(attrn(&_idlg_, nvars));
			%do uu= 1 %to &_nvlg_;
				%let _uplgmk_= &_uplgmk_ %sysfunc(varname(&_idlg_, &uu));
			%end;
			%let _rclg_= %sysfunc(close(&_idlg_));
			
			data upgendat;
				retain &_uplgmk_;
				set gendat(keep=&_uplgmk_);
			run;
			
			%if &debug %then %do;
				data outlib.upgendat;set upgendat;run;
			%end;
			
			*estimating recombination for added markers*;
			%rfest(_gendata_=upgendat, _annodata_=, _RandLG_AP_=0, _runchisq_=0, _swindow_=1, _wsize_=1);
			%if &exiterror %then %goto exit;
			
			%if &debug %then %do;
				data outlib.rfmatup&ii;set rfmat;run;
				data outlib.rfadjup&ii;set rfadj;run;
				data outlib.lodadjup&ii;set lodadj;run;
			%end;
			
			%put BreakGrpsVarL= &BreakGrpsVar;
			%put BreakGrpsValueL= &BreakGrpsValue;
			data updatedLGrp;
				merge updatedLGrp rfadj(rename=(col1=RecombinationF)) %if &LODDataFlag %then %do; lodadj(rename=(col1=LOD)) %end;;
				retain _cumcM_ 0;
				%if %Upcase(&MapFunc) = HALDANE %then %do;
					/* Haldane's Function */
					if RecombinationF < 0.50 then do;
						cM_Step = round(100*(-(1/2)*log(1-2*RecombinationF)),0.01);
					end;
					else do;
						cM_Step = 200;
					end;
				%end;
				%else %do; /* Kosambi */
					if RecombinationF < 0.50 then do;
						cM_Step = round(100*((1/4)*log((1+2*RecombinationF)/(1-2*RecombinationF))),0.01);
					end;
					else do;
						cM_Step = 200;
					end;
				%end;
				if cM_Step >= 200 then 
					call symputx("flag_rf_error", 1);
				_cumcM_ + cM_Step;
				cM_&MapFunc = round(_cumcM_, 0.01);
				RecombinationF = round(RecombinationF, 0.001);
				%if &LODDataFlag %then %do;
					LOD = round(LOD, 0.001);
					if _n_ = 1 then LOD = .;
				%end;
				%if %upcase(&OrigBreakGrps) = YES %then %do;
					if &BreakGrpsVar &BreakOperator &BreakGrpsValue then do;
						call symputx("BreakMaxCheck",trim(left(1)));
					end;
				%end;
				if _n_ = 1 then RecombinationF = .;
				drop _cumcM_;
			run;
			
			*begin: code similar to that found in the BuildLMOData macro*;
			* need descending for computing reverse order *;
			proc sort data=updatedLGrp;	
				by &LGroupVar descending Map_Order;
			run;
			
			data updatedLGrp; set updatedLGrp;
				retain _RMO_ _cMR_ ;
				by &LGroupVar;
				_cMStepR_ = lag1(CM_step);
				_rfR_ = lag1(RecombinationF);
				%if &LODDataFlag %then %do;
					_LODR_ = lag1(LOD);
				%end;				
				if first.&LGroupVar then do;
					_cMR_ = 0;
					_cMStepR_ = 0;
					_rfR_ = .;
					_RMO_ = 1;
					%if &LODDataFlag %then %do;
						_LODR_ = .;
					%end;					
				end;
				else do;
					_RMO_ = _RMO_ +1;
					_cMR_ = _cMR_ + _cMStepR_;
				end;
			run;

			proc sort data=updatedLGrp;	
				by &LGroupVar Map_Order;
			run;
			*end: code similar to that found in the BuildLMOData macro*;
		%end;
		%else %do;
			data updatedLGrp;
				set currLGrp;
				_OldMapOrder_= Map_Order;
				%if %upcase(&OrigBreakGrps) = YES %then %do;
					if &BreakGrpsVar &BreakOperator &BreakGrpsValue then do;
						call symputx("BreakMaxCheck",trim(left(1)));
					end;
				%end;
			run;
		%end;
		
		%if &debug %then %do;
			data outlib.updatedLGrpout&ii;set updatedLGrp;run;
		%end;
		
		%if %upcase(&OrigBreakGrps) = YES %then %do;
			*Similar to "Kemicl check if too large of distance, if so, split groups and rerun order", however, DO NOT rerun order*;
			%put BreakMaxCheck= &BreakMaxCheck;
			%if &BreakMaxCheck %then %do;
				%let EverBroke = 1;
				%let OrigGrpValue = &&LGrpOrd&ii;
				data updatedLGrp;
					retain _cumcM_ 0 _cumMap_ 0;
					set updatedLGrp end=last;
					*set updatedLGrp(Keep = &IDVar &LGroupVar &BreakGrpsVar) end=last;					
					if _n_ = 1 then do;
						%if %UPCASE(&LGTypeOrd) = C %then %do;
							newLGroup = 1;
							NewNumLGrp = &bngrps;
						%end;
						%else %do;
							newLGroup = &bngrps;
						%end;
					end;
					retain NewLGroup;
					%if %UPCASE(&LGTypeOrd) = C %then retain NewNumLGrp;;
					if &BreakGrpsVar &BreakOperator &BreakGrpsValue then do;
						NewLGroup+1;
						%if %UPCASE(&LGTypeOrd) = C %then NewNumLGrp+1;;
						Map_Order= 1;
						_cumMap_ = 0;
						cM_Step= 0;
						RecombinationF= .;
						cM_&MapFunc= 0;
						_cumcM_= 1;
						%if &LODDataFlag %then %do;
							LOD= .;
						%end;
					end;
					_cumMap_ + 1;
					Map_Order= _cumMap_;
					_cumcM_ + cM_Step;
					cM_&MapFunc = round(_cumcM_, 0.01);
					%if %UPCASE(&LGTypeOrd) = C %then %do;
						&LGroupVar = trim(left(&LGroupVar))||"_"||trim(left(NewLGroup));
					%end;
					%else %do;
						if newLGroup ne &bngrps then &LGroupVar = newLGroup;
					%end;
					%if %UPCASE(&LGTypeOrd) = C %then %do;
						if last then  call symput("bnGrps",trim(left(NewNumLGrp)));
					%end;
					%else %do;
						if last then  call symput("bnGrps",trim(left(NewLGroup)));
					%end;	
					*keep &IDVar &LGroupVar NewNumLGrp NewLGroup;
					drop _cumcM_ _cumMap_ _cMR_ _cMStepR_ _rfR_ _RMO_ NewLGroup %if &LODDataFlag %then _LODR_;;
				run;
				
				%if &debug %then %do;
					data outlib.updatedLGrpNew&ii; set updatedLGrp; run;
				%end;
								
				*begin: code similar to that found in the BuildLMOData macro*;
				* need descending for computing reverse order *;
				proc sort data=updatedLGrp;	
					by &LGroupVar descending Map_Order;
				run;
				data updatedLGrp; set updatedLGrp;
					retain _RMO_ _cMR_ ;
					by &LGroupVar;
					_cMStepR_ = lag1(CM_step);
					_rfR_ = lag1(RecombinationF);
					%if &LODDataFlag %then %do;
						_LODR_ = lag1(LOD);
					%end;					
					if first.&LGroupVar then do;
						_cMR_ = 0;
						_cMStepR_ = 0;
						_rfR_ = .;
						_RMO_ = 1;
						%if &LODDataFlag %then %do;
							_LODR_ = .;
						%end;						
					end;
					else do;
						_RMO_ = _RMO_ +1;
						_cMR_ = _cMR_ + _cMStepR_;
					end;
				run;

				proc sort data=updatedLGrp;	
					by &LGroupVar Map_Order;
				run;
				*end: code similar to that found in the BuildLMOData macro*;
				
				*begin: collecing linkage groups with one marker only*;
				proc sql UNDO_POLICY= NONE;
					create table singlemk as
					select *
					from updatedLGrp
					group by &LGroupVar
					having n(Map_Order) < 2
					order by &LGroupVar, Map_Order;
				quit;
				*end: collecing linkage groups with one marker only*;
				
				%NObsVars(singlemk);
				%if &nobs > 0 %then %do;
					%let sflag= 1;
					
					%if &ApeendSingleMk= 0 %then %do;
						data singlemk_all;
							set singlemk(drop= Map_Order _OldMapOrder_ RecombinationF cM_Step 
							cM_&MapFunc _cMR_ _cMStepR_ _rfR_ _RMO_ %if &LODDataFlag %then LOD _LODR_;)							;
							*set singlemk(drop= Map_Order RecombinationF cM_Step cM_&MapFunc _cMR_ _cMStepR_ _rfR_ _RMO_ %if &LODDataFlag %then LOD _LODR_;);
						run;
						%let ApeendSingleMk= 1;
					%end;
					%else %do;
						proc append base= singlemk_all data= singlemk(drop= Map_Order _OldMapOrder_ RecombinationF cM_Step 
						cM_&MapFunc _cMR_ _cMStepR_ _rfR_ _RMO_ %if &LODDataFlag %then LOD _LODR_;);
						*proc append base= singlemk_all data= singlemk(drop= Map_Order RecombinationF cM_Step cM_&MapFunc _cMR_ _cMStepR_ _rfR_ _RMO_ %if &LODDataFlag %then LOD _LODR_;);
						run;
					%end;
					
					%if &debug %then %do;
						data outlib.singlemk_all; set singlemk_all;run;
					%end;
					
					*removing singletons from the updatedLGrp data set*;
					proc sql UNDO_POLICY= NONE;
						create table updatedLGrp as
						select *
						from updatedLGrp
						group by &LGroupVar
						having n(Map_Order) > 1
						order by &LGroupVar, Map_Order;
					quit;
											
				%end; 
			%end; /* end for breaking groups and recreating order and dims data */	
		%end; /*if OrigBreakGrps= YES*/
		
		%if &debug %then %do;
			data outlib.updatedLGrpfinal&ii; set updatedLGrp;run;
		%end;
		
		%NObsVars(updatedLGrp);
		%if &nobs > 1 %then %do; 
			%let allsflag= 0;
			%if &AppendNewODims %then %do;
				proc append base=newoutdims data=updatedLGrp(drop=_OldMapOrder_);
				*proc append base=newoutdims data=updatedLGrp;
				run;
			%end;
			%else %do;
				data newoutdims;
					set updatedLGrp(drop=_OldMapOrder_ );
					*set updatedLGrp;
				run;
				%let AppendNewODims= 1;
			%end;
			
			%if &debug %then %do;
				data outlib.lmo_newoutdims&ii; set newoutdims; run;
			%end;
		%end;
		%*goto exit;
			
	%end; /*end: loop over linkage groups in the outdims build with cluster center markers*/

	*reseting macro variables used in the LoopOverLGs;	
	%let EverBroke_out = &EverBroke;
	%let EverBroke = 0;
	%let sflag_out= &sflag;
	%let sflag = 0;
	%let allsflag_out= &allsflag;
	%let allsflag = 1;
	%let BreakMaxCheck = 0;
	%let AppendStarted = 0;

	%*goto exit;
	
	*begin: checking if any singleton from the original ordering is a cluster center, if add back markers*;
	%if &sflag %then %do;
		*This part is not working*;		
		data singletons_center singletons_noncenter;
			retain _count_ _countno_ 0;
			set oldsingletons_all end=_eof_;
			if _ClusterCenterVar_= 1 then do;
				 _count_ + 1;
				 _countno_ + 0;
				 call symputx("Clus"||trim(left(_count_)),trim(left(_ClusterVar_)));
				 call symputx("ClusCen"||trim(left(_count_)),trim(left(_ClusterCenterVar_)));
				 call symputx("MrkName"||trim(left(_count_)),trim(left(&IDVar)));
				 output singletons_center;
			end;
			else do;
				_count_ + 0;
				_countno_ + 1;
				output singletons_noncenter;
			end;
			if _eof_ then do;
				call symputx("nSingCenterMk", trim(left(_count_)));
				call symputx("nSingNoCenterMk", trim(left(_countno_)));
			end;
			drop _count_;
		run;
		%put nSingCenterMk= &nSingCenterMk;
		
		%if &debug %then %do;
			data outlib.singletons_center;set singletons_center;run;
		%end;
		
		%if &nSingCenterMk > 0 %then %do;
	
			*creating sing_indat data set with markers and recombination fractions;
			proc sql;
				/*create table sing_indat(drop=_order_ _cvar_) as*/
				create table sing_indat(drop= _cvar_) as
				select a._cvar_, a.&LGroupVar, b.*
				from singletons_center(rename=(_ClusterVar_=_cvar_)) as a left join orig_indat(drop= &LGroupVar) as b
				on a._cvar_= b._ClusterVar_
				/*order by _order_;*/
				order by orig_order;
			quit;
			%let indat= sing_indat;
			
			%if &debug %then %do;
				data outlib.sing_indatb;set &indat;run;
			%end;
			
			data &indat(keep =&LGroupVar &IDVar &OrigDistVars &OrigListDistVars)  
				copydat(keep = &LGroupVar &IDVar &KeepVars &SkOrderVar &ColorVar orig_order &OrderVar &AnnoMName);
				set &indat; 
				*orig_order = _n_;
			run;
			
			%if &debug %then %do;
				data outlib.sing_indatf;set &indat;run;
				data outlib.sing_copydat;set copydat;run;
			%end;
			%*goto exit;
			
			*collecting number of singletons that are cluster centers*;
			%let nGrpsSin=0;
			%let LGTypeSin=;
			%GetLevels(_Ds =singletons_center,_Var=&LGroupVar,_LPrefix = LGrpSin,_nLvls=nGrpsSin,_vType= LGTypeSin);
			%put There are &nGrpsSin singletons that are cluster centers.;
			%put nGrpsSin= &nGrpsSin;
			%put LGrpSin1= &LGrpSin1;
								
			%let nGrpsOS= 0;
			%if &sflag_out %then %do;
				%let nGrpsOS=;
				%let LGTypeOS=;
				%GetLevels(_Ds =singlemk_all,_Var=&LGroupVar,_LPrefix = LGrpOS,_nLvls=nGrpsOS,_vType= LGTypeOS);
			%end;
			%let nGrpsOO= 0;
			%let LGTypeOO=;
			%GetLevels(_Ds =newoutdims,_Var=&LGroupVar,_LPrefix = LGrpOO,_nLvls=nGrpsOO,_vType= LGTypeOO);

			%let bnGrps = %sysevalf(&nGrpsOO + &nGrpsOS + &nGrpsSin);
			*end: reseting macro variables used in the LoopOverLGs;
				
			proc datasets;
				delete dims outdims;
			quit;
				
			%LoopOverLGs(nGrps_=&nGrpsSin, LGrpPrefix_= LGrpSin);
			%if &exiterror= 1 %then 
				%goto exit;
				
			%if %sysfunc(exist(outdims)) %then %do;
			
				%if &debug %then %do;
					data outlib.newoutdims_s;set newoutdims;run;
					data outlib.outdims_s;set outdims;run;
					data outlib.dims_s;set dims;run;
				%end;
			
				data outdims;
					set outdims;
					Original_Group = _LGroupVar_;
				run;
			
				*appending outdims from the singletons to the newoutdims from linked markers*;
				proc append base= newoutdims data= outdims;
				run;
			%end;
			
			%*goto exit;

			%put sflag= &sflag;
			%put nSingNoCenterMk= &nSingNoCenterMk;
			%put sflag_out= &sflag_out; 
			*appeding singletons, if any is found*;	
			%if &sflag %then %do;
				%if &nSingNoCenterMk %then %do;
					proc append base=singletons_all data=singletons_noncenter;
					run;
				%end;
				
				%if &sflag_out %then %do;
					proc append base=singletons_all data=singlemk_all;
					run;
				%end;
			%end;
			%else %if &nSingNoCenterMk %then %do;
				data singletons_all;
					set singletons_noncenter;
				run;
				
				%if &sflag_out %then %do;
					proc append base=singletons_all data=singlemk_all;
					run;
				%end;
			%end;
			%else %if &sflag_out %then %do;
				data singletons_all;
					set singlemk_all;
				run;
			%end;
			
			%if &sflag or &nSingNoCenterMk or &sflag_out %then %do;
				data Outlib.&LinkageMapOrder_OutDataExName;
					set singletons_all(drop= _LGroupVar_ _ClusterVar_ _ClusterCenterVar_);
				run;
			%end;
			
			%*goto exit;
		*end: adding back markers from singletons of original ordering*;					
		%end; 
		%else %do;
			%if &sflag_out %then %do;
				proc append base=oldsingletons_all data=singlemk_all;
				run;
				data Outlib.&LinkageMapOrder_OutDataExName;
					set oldsingletons_all(drop= _LGroupVar_ _ClusterVar_ _ClusterCenterVar_);
				run;
			%end;
		%end;	
	
		%let EverBroke_sin = &EverBroke;
		%if &nSingNoCenterMk or &sflag %then
			%let sflag_sin= 1;
		%let allsflag_sin= &allsflag;
	
	*end: checking if any singleton from the original ordering is a cluster center*;
	%end;
	%else %do;	
		%if &sflag_out %then %do;
			data Outlib.&LinkageMapOrder_OutDataExName;
				set singlemk_all(drop= _LGroupVar_ _ClusterVar_ _ClusterCenterVar_);
			run;
		%end;
	%end;
		
	*creating new outdims from newoutdims after adding markers back to linkage groups*;
	proc datasets nolist;
		delete outdims;
	quit;
	%NObsVars(newoutdims);
	%if &nobs > 1 %then %do;
		
		proc datasets nolist;
			change newoutdims= outdims;
		run;
		
		proc sort data=outdims out=outdims(drop= _LGroupVar_ _ClusterVar_ _ClusterCenterVar_);
			by &LGroupVar Map_Order;
		run;
	
		%if &debug %then %do;
			data outlib.outdims_final;set outdims;run;
		%end;
	%end;
		
	%let EverBroke= 0;
	%let sflag= 0;
	%let allsflag= 1;
	%if &EverBroke_out or &EverBroke_sin %then
		%let EverBroke= 1;
	%if &sflag_out or &sflag_sin %then
		%let sflag= 1;
	%if &allsflag_out=0 or &allsflag_sin=0 %then
		%let allsflag= 0;
		
	%ErrorCheckAfterOrdering;
	%if &exiterror= 1 %then 
		%goto exit;
		
	%if &sflag %then %do;
		%let nGrpsS=;
		%let LGTypeS=;
		%GetLevels(_Ds =Outlib.&LinkageMapOrder_OutDataExName,_Var=&LGroupVar,_LPrefix = LGrpS,_nLvls=nGrpsS,_vType= LGTypeS);
		%let Singletons= ;
		%do i=1 %to &nGrpsS;
			%let Singletons = &Singletons "&&LGrpS&i";
		%end;
		%put Singletons= &Singletons;
	%end;
	
	*creating new LGrp and OldGrp macro vars, and building the rstack data set*;
	proc sort data=outdims;
		by &LGroupVar Map_Order;
	run;
	
	%GetLevels(_Ds =outdims,_Var=&LGroupVar,_LPrefix = LGrp,_nLvls=nGrps,_vType= LGType);
	%put nGrpsFinal= &nGrps;
	%do ii=1 %to &nGrps;
		%put Processing Linkage Group &&LGrp&ii;
		%let OLGrp&ii = &&LGrp&ii;
		%if %UPCASE(&LGType) = C %then %do;
			%let LGrp&ii = "&&LGrp&ii";
		%end;
		
		/*data OutLib.lmo_outdims; set outdims; run;*/
		
		data namesformerge;
			set outdims;
			where &LGroupVar= &&LGrp&ii;
			row = _n_;
			keep &IDVar row;
		run;

		proc sql;
			create table stackedmlist as
			select f1.&IDVar, f1.row1, f2._COL_, f2.row2
			from namesformerge(rename = (row=row1)) as f1 
			cross join namesformerge(rename = (&IDVar=_COL_ row=row2)) as f2
			where f1.row1 <= f2.row2
			order by f2.row2, f1.row1
			;
		quit;
		
		data stackedmlist; set stackedmlist;
			rename &IDVar = _ROW_;
			drop row1 row2;
		run;
		
		%if &debug %then %do;
			data outlib.stackedmlist&ii;set stackedmlist;run;
			data outlib.namesformerge&ii;set namesformerge;run;
		%end;
		
		proc transpose data=namesformerge(keep=&IDVar) out=namesformerge(drop=_name_);
			ID &IDVar;
		run;
		
		%let Markers_lg= ;
		%let _tidlg_ = %sysfunc(open(namesformerge));
		%let _nvarslg_ = %sysfunc(attrn(&_tid_,NVARS));
		%do j=1 %to &_nvarslg_;
			%let Markers_lg= &Markers_lg %sysfunc(varname(&_tidlg_, &j)); 
		%end;
		%let rclg = %sysfunc(close(&_tidlg_));
		%put Markers_lg= &Markers_lg;
		
		data gendat_lg;
			retain &Markers_lg;
			set gendat(keep=&Markers_lg);
		run;
		
		%if &debug %then %do;
			data outlib.gendat_lg&ii;set gendat_lg;run;
		%end;
			
		*estimating recombination for final linkage map*;
		%rfest(_gendata_=gendat_lg, _annodata_=, _RandLG_AP_=0, _runchisq_=0, _swindow_=1, _wsize_=10);
		%if &exiterror %then %goto exit;
		
		%if &debug %then %do;
			data outlib.lmo_rfmatlg&ii;set rfmat;run;
			%if &LODDataFlag %then %do;
				data outlib.lmo_lodmatlg&ii;set lodmat;run;
			%end;
		%end;
		
		proc iml;
			use rfmat;
			read all var{&Markers_lg} into r;
			rstack = symsqr(r);

			create rstack from rstack;
			append from rstack;			
		quit;
		
		%if &debug %then %do;
			data outlib.lmo_rstack&ii; set rstack;run;
		%end;
		
		data Moutres;
			merge stackedmlist rstack(rename = (col1 = Recombination));
			label Recombination = "Recombination";
			if _ROW_ = _COL_ then delete;
		run;
		
		%if &ii=1 %then %do;
			data outres; set Moutres; run;
		%end;
		%else %do;
			proc append base=Outres data=Moutres;run;
		%end;
	%end;
	
	%if &debug %then %do;
		data outlib.outresnew;set outres;run;
	%end;
	
%end;
/*end: Luciano*/

/* kemicl added so we don't have label issues for main data */
options nolabel;

%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)=0 %then %let OrderVar = Map_Order;

%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0 %then %do;
	/* sum up absolute value of residuals to color by as indication of poorly fitted markers */
	data absres; set outres;
		keep _ROW_ _COL_ Residual;
		Residual = abs(Residual);
	run;

	proc sort data=absres;
		by _ROW_;
	run;

	proc means data = absres sum noprint;
		by _ROW_;
		var Residual;
		output out = rowsums(keep = _ROW_ srow) sum=srow;
	run;

	proc sort data=absres;
		by _COL_;
	run;

	proc means data = absres sum noprint;
		by _COL_;
		var Residual;
		output out = colsums(keep = _COL_ scol) sum=scol;
	run;

	proc sql UNDO_POLICY=NONE;
		create table rsums as 
			select *
			from rowsums as d1 full join colsums as d2
			on d1._ROW_ = d2._COL_;
	quit;

	data rsums;
		set rsums;
		if missing(srow) then srow = 0;
		if missing(scol) then scol = 0;
		Sum_AbsResidual = srow + scol;
		if missing(_ROW_) then _ROW_ = _COL_;
		drop _COL_ scol srow;
	run;

	proc sql UNDO_POLICY=NONE;
		create table Outlib.&LinkageMapOrder_OutDataName(drop=_ROW_ Init_Map_Order) as 
			select *
			from outdims d1, rsums d2
			where d1.&IDVar = d2._ROW_;
	quit;
	
	/* for plots in JMP to color by if no colorvar */
	%if ^%length(&ColorVar) %then %do;
		%let ColorVar = Sum_AbsResidual;
	%end;
%end;
%else %do;
	data Outlib.&LinkageMapOrder_OutDataName; 
		set outdims; 
		*drop Opt_Order; /*Luciano*/
	run;
	
	%if ^%length(&ColorVar) %then %do;
		%if &EverBroke %then %let ColorVar = Original_Group;
		%else %let ColorVar = &LGroupVar;
	%end;
	/* %CheckLabel(outdims, ColorLabel); */
%end;

%*goto exit;

%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0  %then %do;
/* columns to keep when order was specified and output is from MDS */
	%let keepcols1 = _ROW_, _COL_, Recombination, FitData, FitDist, Weight, Residual,Locus1, O1;
	%let keepcols2 = Marker1, _COL_, Recombination, FitData, FitDist, Weight, Residual, &LGroupVar, Locus1, Locus2, O1, O2;
%end;
%else %do;
	%let keepcols1 = _ROW_, _COL_, Recombination, Locus1, O1;
	%let keepcols2 = Marker1, _COL_, Recombination, &LGroupVar, Locus1, Locus2, O1, O2;
%end;
proc sql UNDO_POLICY=NONE;
	create table res1(rename = (_ROW_ = Marker1)) as 
		select &keepcols1
		from outres as d1, outdims (rename = (cm_&MapFunc = Locus1 Map_Order = O1)) as d2 
		where d1._Row_ = d2.&IDVar;
		
	create table res(rename = (_COL_ = Marker2)) as 
		select &keepcols2
		from res1 as d1, outdims (rename = (cm_&MapFunc = Locus2 Map_Order = O2)) as d2 
		where d1._COL_ = d2.&IDVar;
quit;

data matchmark; set outdims(rename = (&IDVar = Marker1 cm_&MapFunc = Locus1 Map_Order = O1)); set outdims(rename = (&IDVar = Marker2 cm_&MapFunc = Locus2 Map_Order = O2));
	Recombination = 0;
	keep Marker1 Marker2 Locus1 Locus2 Recombination &LGroupVar O1 O2;
run;

data res;
	set res matchmark;
	if O1 > O2 then do;
		tmp = O1;
		O1 = O2;
		O2 = tmp;
		tmp = Locus1;
		Locus1 = Locus2;
		Locus2 = tmp;
		tmp1 = Marker1;
		Marker1 = Marker2;
		Marker2 = tmp1;
	end;
	drop tmp tmp1;
run;

proc sort data=res out=Outlib.&LinkageMapOrder_OutDataResName;
	by &LGroupVar O1 O2;
run;

/* resort data if breakgroups was possible */
%if &EverBroke %then %do;
	proc sort data=Outlib.&LinkageMapOrder_OutDataName;
		by &LGroupVar Map_Order;
	run;
	
	%GetLevels(_Ds =Outlib.&LinkageMapOrder_OutDataName,_Var=&LGroupVar,_LPrefix = LGrp,_nLvls=nGrps,_vType= LGType);
	%do i=1 %to &nGrps;
		%put Processing Linkage Group &&LGrp&i;
		%let OLGrp&i = &&LGrp&i;
		%if %UPCASE(&LGType) = C %then %do;
			%let LGrp&i = "&&LGrp&i";
		%end;
	%end;
%end;

%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0 %then %do;
	/* Get the badness of fit and fit correlation into macro variables */
	data _null_; set outfit;
		CRITER = Round(CRITER,0.001);
		FitCorr = Round(FitCorr,0.001);
		call symput("BoF"||trim(left(_n_)),trim(left(CRITER)));
		call symput("Corr"||trim(left(_n_)), trim(left(FitCORR)));
	run;
%end;

/* get maxpos and minpos to switch axes with */
proc means data = outdims noprint;
	var cM_&MapFunc;
	output out = mp max = maxpos;
run;

data _null_;
	set mp;
	call symput("maxpos",trim(left(maxpos+11)));
run;
%let minpos = -11;

%if &sflag %then %do;
	%PackageInsertDataset(Excluded Markers from &ProcessName, OutLib, &LinkageMapOrder_OutDataExName,no);
%end;

/* create macro vars for the maxposition of each linkage group*/
/* create new lgrp macros and number of lgrps if option to break linkage groups enabled*/
data checkmax;
	set Outlib.&LinkageMapOrder_OutDataName(keep=&LGroupVar cm_&MapFunc);
	by &LGroupVar;
	if last.&LGroupVar;
run;

data _null_;
	set checkmax;
	call symput("LGMax"||trim(left(_n_)),trim(left(cm_&MapFunc)));
run;

/* Create all the macro variables of the markers to use in cell plots */
/* also use lgtot for reversal of maps */
%do i=1 %to &nGrps;
	%if %index(&Singletons,"&&LGrp&i")= 0 %then %do;
		data _null_;
			set Outlib.&LinkageMapOrder_OutDataName(keep=&IDVar &LGroupVar) end=_e_;
			where &LGroupVar = &&LGrp&i;
			%if %length(&SNPInData) and %upcase(&DoCellPlots) = YES %then %do;
				call symput("MVar&i"||trim(left(_n_)), trim(left(&IDVar)));
			%end;
			if _e_ then call symput("LGtot&i",trim(left(_n_)));
		run;
		/* check list 
			%if %index(&Singletons,"&&LGrp&i")= 0 %then %do;
				%do m=1 %to &&LGtot&i;
				%put &&&MVar&i&m;
				%end;
			%end;
		*/
	%end;
%end;

%if %length(&SNPInData) %then %do;
	/*Kemicl set marker format to numeric for launching subset reorder process */
	%if %index(&MarkerFmt,Homozygous) %then %let MarkerFmt = Numeric Genotypes;
%end;

%*goto exit;

* create JSL file;
%PrepOutFile(LinkageMapOrderScript,LinkageMapOrderScriptName,&ProcessName Script for &OutDataPrefix..jsl);
%global _mainjslfile_ FullAPName;
%let _mainjslfile_ = &LinkageMapOrderScript;
%let FullAPName=%str(Linkage Map Order);
data _null_;
   file "&LinkageMapOrderScript" lrecl=32767;
	put "//!";
	put "Names Default to Here(1);";
	put "Here:DataList = {};";
	put "Here:TabButtonList = {};";
	put "Here:TabCodeList = {};";
	put "Here:TabAfterCodeList = {};";
	put "Here:TabOpenList = {};";
	put "vfsize = 650;";
	put "hfsize = 300;";
	put "OutPath = ""&OutPath"";";
	put "ClientOutPath = ""&ClientOutPath"";";
	put "OutDataName = ""&LinkageMapOrder_OutDataName"";";
	put "OutDataResName = ""&LinkageMapOrder_OutDataResName"";";
	put "GroupVar = ""&LGroupVar"";";
	put "LGType = ""&LGtype"";";
	put " if(host is(Windows),";
	put "   include( ConvertFilePath(""$GENOMICS_HOME/JSLFiles/Linkage Group Expressions.jsl"", absolute, windows));";
	put " ,";
	put "   include( ConvertFilePath(""$GENOMICS_HOME/JSLFiles/Linkage Group Expressions.jsl"", absolute, POSIX));";	
	put " );";	
	put "Here:GB_results = Expr(";
	put " Here:gb = Graph Builder(";
	put "		Size(625, 525),";
	put "		Show Control Panel(0),";
	put "		Auto Stretching( 0 ),";
	put "		Variables(";
	put "			X(:Name(""&LGroupVar"")),";
	put "			Y(:Name(""cm_&MapFunc"")),";
	put " 		Color(:Name(""&ColorVar""))";
	put "		),";
	put "		Elements(Points(X,Y,Color,Legend(4), Jitter( 0 ))),";
	put "		SendToReport(";
	put "			Dispatch({}, ""graph title"", TextEditBox,";
	put "				{Set Text(""Genetic Distances For Linkage Groups"")}";
	put "			),";
	put "			Dispatch({}, ""cm_&MapFunc"", ScaleBox, {Min( &maxpos), Max( &minpos), Rotated Labels(""Automatic"")}),";
	put "			Dispatch({}, ""Graph Builder"", FrameBox, {Marker Size(3)})";
	put "		)";
	put "	);";
	put ");";
	
	put "Here:DF_results = Expr(";
	put "	Here:df = odat << Data Filter(";
	put "		Add Filter(";
	put "			Columns(";
	%if &EverBroke %then %do;
		put "				:Name(""Original_Group""), ";
	%end;
	put "				:Name(""cM_Step""),"; 
	put "				:Name(""RecombinationF"")";
	%if &LODDataFlag %then %do;
		put "				,:Name(""LOD"")";
	%end;
	put "			)";
	put "		)";
	put " );";
	put ");";
	
	put "Here:GBfunc = Function({},{default local},";
	put "	If(IsEmpty(Here:odat),";
	put " 	Here:odat = open(""&ClientOutPath.&LinkageMapOrder_OutDataName..sas7bdat"",invisible);";
	put "		Column(""&LGroupVar"")<<Set Modeling Type(Nominal);";
	%if &EverBroke %then %do;
		put "		Column(""Original_Group"")<<Set Modeling Type(Nominal);";
	%end;
	%if %length(&SkOrderVar) %then %do;
		put "		Here:odat << Select Where(!Is Missing(:Name(""&SkOrderVar"")[]));";
		put "		Here:odat << Markers(11);";
	%end;
	put "		Column(""&IDVar"")<<Label(1);";
	put " ,"; /* else */
	put "		Current Data Table(Here:odat);";
	put "	);";
	put "	VListBox(";
	%if &sflag %then %do;
		put "		TextBox(""  NOTE: Some Linkage Groups contained only one marker. These markers are excluded from the analysis and output map.    A data set of the excluded markers can be found under the Output Data tab."", <<Set Wrap(715)),";
	%end;
	%if %length(&SkOrderVar) %then %do;
			put "		TextBox(""  Framework Markers are denoted with a star."", <<Set Wrap(415)),";
	%end;
	put " Here:LO_HB = HListBox(";
	put "		Here:GB_results, Here:DF_results;";
	put " ));"; /* Hlistbox and VlistBox */
	
	put ");"; /* Close Function */
	
	put "InsertInto(DataList,{{""Here:odat""}});";
	put "InsertInto(TabButtonList, ""Linkage Order Results"");";
	put "InsertInto(TabCodeList, expr(Here:GBfunc()));";
	put "InsertInto(TabAfterCodeList, "" "");";
	put "InsertInto(TabOpenList,1);";
	%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0  %then %do;
		put "Here:Rfunc = Function({},{default local},";
		put "	If(IsEmpty(Here:rdat),";
		put " 	Here:rdat = open(""&ClientOutPath.&LinkageMapOrder_OutDataResName..sas7bdat"",invisible);";
		put "		Column(""&LGroupVar"")<<Set Modeling Type(Nominal);";
		put "		Column(""Marker1"")<<Label(1);";
		put "		Column(""Marker2"")<<Label(1);";
		put " ,"; 
		put "		Current Data Table(Here:rdat);";
		put "	);";
		put "	VListBox(";
		%do i = 1 %to &nGrps;
			%if %index(&Singletons,"&&LGrp&i")= 0 %then %do; 
				put "  OutlineBox(""Linkage Group &&OLGrp&i"",";
				put "  	TextBox(""Recombination Fraction and Fitted Distance Correlation = &&Corr&i"", <<Set Wrap(600)),";
				put "   	TextBox(""MDS Model Badness of Fit (lower is better) = &&BoF&i"", <<Set Wrap(600)),";
				put "  	biv&i = Bivariate(";
				put "			Y(:Name(""Residual"")), X(:Name(""FITDIST"")),";
				%if %upcase(&LGType) = C %then %do;
					put "			Where (:Name(""&LGroupVar"") == ""&&OLGrp&i""),";
				%end;
				%else %do;
					put "			Where (:Name(""&LGroupVar"") == &&LGrp&i),";
				%end;				
				put "			Fit Spline(1), Histogram Borders(1),";
				put "			SendToReport(";
				put "				Dispatch({}, ""Bivar Plot"", FrameBox, {Frame Size(375,225)}),";
				put "				Dispatch( {}, ""Smoothing Spline Fit"", OutlineBox, {Close( 1 )} )";
				put "			)";
				put "	 	)";
				put "  ),";
			%end;
		%end;
		put "));";

		put "InsertInto(DataList,{{""Here:rdat""}});";
		put "InsertInto(TabButtonList, ""MDS Residual Plots"");";
		put "InsertInto(TabCodeList, expr(Here:Rfunc()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,0);";	
		put "Here:TabsSectionFlag = 1;";
	%end;
	/* Kemicl added tab to show cell plots Jan192012*/
	%if %upcase(&DoCellPlots) = YES %then %do;
		put "Here:CPfunc = Function({},{default local},";
		put "	If(IsEmpty(Here:sdat),";
		put " 	Here:sdat = open(""&SNPInData"",invisible);";
		/*Force to use SAS names so markers match*/
		put "		cols = Here:sdat  << GetColumnNames();";
		put "		for(i=1,i<=NItems(cols),i++,";
		put "			ci = Column(cols[i]);";
		put "			sasnm = char(ci << GetProperty(""SAS Name""));";
		put "			if(length(sasnm),ci << SetName(sasnm))";
		put "		);";
		%do i=1 %to &nGrps;
			%if %index(&Singletons,"&&LGrp&i")= 0 %then %do;
				%do m=1 %to &&LGtot&i;
					put "  Column(""&&&MVar&i&m"")<<Set Modeling Type(""Ordinal"");";
					%if %upcase(&AddDroppedMarkers)= YES or %length(&SNPIndata) %then %do;
						put "  Column(""&&&MVar&i&m"")<<Set Property(""Value Colors"",{""&code2""=21,""&code11""=1, ""&code12"" = 1, ""&code0""=19});"; 
					%end;
				%end;
			%end;
		%end;
		%if %length(&LabelVar) %then %do;
			put "		Column(""&LabelVar"")<<Label(1);";
		%end;
		put " ,"; 
		put "		Current Data Table(Here:sdat);";
		put "	);";
		put " HListBox(";
		put "VListBox(";
		put "            LineUpBox(NCol(2),";
		put "               Text Box( "" Vertical Frame Size "" ),";
		put "         		  vframeSizeSlider = SliderBox(num(200),num(1500),vfsize,";
		%do i=1 %to &nGrps;
			%if %index(&Singletons,"&&LGrp&i")= 0 %then %do;
			put "	try((Here:cp&i<<report)[CellPlotBox(1)] << FrameSize(hfsize,vfsize);); ";
			%end;
		%end;
		put "					  ),";
		put "               Text Box( "" Horizontal Frame Size "" ),";
		put "         		  hframeSizeSlider = SliderBox(num(100),num(800),hfsize,";
		%do i=1 %to &nGrps;
			%if %index(&Singletons,"&&LGrp&i")= 0 %then %do;
			put "								try((Here:cp&i<<report)[CellPlotBox(1)] << FrameSize(hfsize,vfsize); );";
			%end;
		%end;
		put "					  ),";
		put "            ),";		
		 put '  panelbox("Legend",';
		 put '   t1=Textbox("Homozygous for Parental Line 1"),';
		 put '   t1<<Font Color(3),';
		 put '   t1<<Bullet Point(1);';
		 put '   t2=Textbox("Heterozygous"),';
		 put '   t2<<Font Color(1),';
		 put '   t2<<Bullet Point(1);';
		 put '   t3=Textbox("Homozygous for Parental Line 2"),';
		 put '   t3<<Font Color(5),';
		 put '   t3<<Bullet Point(1);';
		 put '  )';
		put "),";
		%do i=1 %to &nGrps;
			%if %index(&Singletons,"&&LGrp&i")= 0 %then %do;
				put "		Here:cp&i =  Cell Plot( ScaleUniformly(1), Center at Zero(0),";
				put "			Y(";
				%do m=1 %to &&LGtot&i;
					put "			:Name(""&&&MVar&i&m""),";
				%end;
				put "			),";
				%if %length(&LabelVar) %then %do;
					put "			Label(:Name(""&LabelVar"")),";
				%end;
				put "			SendToReport(";
				put "				Dispatch({},""Cell Plot"",OutlineBox,";
				put "					{Set Title(""Genotype Color Plot for Linkage Group &&OLGrp&i"")}";
				put "				),";
				%if &&LGtot&i > 50 %then %do;
					put "				Dispatch( {}, """", CellPlotBox, {Frame Size(300, 650), No separator lines(1)})";
				%end;
				%else %do;
					put "				Dispatch( {}, """", CellPlotBox, {Frame Size(300, 650)})";
				%end;
				put "			)";
				put "		);";
			%end;
		%end;
		put " ));";
		
		put "InsertInto(DataList,{{""Here:sdat""}});";
		put "InsertInto(TabButtonList, ""Genotype Color Plots"");";
		put "InsertInto(TabCodeList, expr(Here:CPfunc()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";
	%end;
	put "InsertInto(DataList,{{"""" }});";
	%let tmpcount = 1;
	%do i = 1 %to &nGrps;
		%if %index(&Singletons,"&&LGrp&i")= 0 %then %do;
			put "ldhvp&i = Function({},{default local},";
			put "	If(IsEmpty(Here:rdat),";
			put " 	Here:rdat = open(""&ClientOutPath.&LinkageMapOrder_OutDataResName..sas7bdat"",invisible);";
			put "		Column(""&LGroupVar"")<<Set Modeling Type(Nominal);";
			put "		Column(""Marker1"")<<Label(1);";
			put "		Column(""Marker2"")<<Label(1);";
			put " ,"; 
			put "		Current Data Table(Here:rdat);";
			put "	);";
			put "	Here:rdat << Clear Column Selection();";
			put "	Here:rdat << Clear Select;";
			put " If(IsEmpty(Here:data&i),";
				%if %upcase(&LGType) = C %then %do;
					put "	Here:data&i = Subset(Where(:Name(""&LGroupVar"") == ""&&OLGrp&i""), Output Table Name(OutDataResName||""_&&OLGrp&i""));";
				%end;
				%else %do;
						put "	Here:data&i = Subset(Where(:Name(""&LGroupVar"") == &&LGrp&i), Output Table Name(OutDataResName||""_&&LGrp&i""));";
				%end;
			put "	,";
			put "		Current Data Table(Here:data&i);";
			put ");";
			put "Here:dt = Today();";
			put "Here:_ag_=""for Linkage Group &&OLGrp&i"";";
			put "maxrows = 100000;";
			put "if(nrows(Here:data&i)<=maxrows,";
			put " Here:ld_i=1;";
			put " jg:theme_index = 8;";
			put " Here:MeasureVar = ""Recombination"";";
			put " Here:BlocksDefined = 0;";		
			put " Here:currchr = ""&i"";";	
			put "	Here:currby = ""1"";";
			put " Here:ExtraTitle = """";";
			put " Eval(Parse(Eval(ldexprexpr)));";

			put ", //else too big";
			put " errwin=New Window(""JMP Genomics Message"",";
			put "  TextBox(""You cannot view the Linkage Order and Recombination plot for more than ""||char(maxrows)||"" pairs.""));";
			put ' errwin << SetWindowIcon("$LS_HOME/Documentation/Icons/JMPGenomicsApp.ico");';				
			put ");";
			put "); //end func";
			%if &i = 1 %then %do;
				put "InsertInto(DataList,{{""Here:rdat""}});";
			%end;
			put "InsertInto(DataList,{{""Here:data&i"" }}) ;";
			put "_ag_buttonname=""Linkage Group &&OLGrp&i"";";
			%if &tmpcount < 2 %then %do;
				put "if(isEmpty(Here:ActionButtonDesc) | NItems(Here:ActionButtonDesc)==0,";
				put "   Here:ActionButtonList = {{""""}};";
				put "   Here:ActionCodeList = {{""""}};";
				put "   Here:ActionButtonDesc = {0};";	
				put "   LArryIndex=1;";
				put ", //else";
				put "   LArryIndex=NItems(Here:ActionButtonList)+1;";
				put "   InsertInto(Here:ActionButtonList,{{""""}});";
				put "   InsertInto(Here:ActionCodeList,{{""""}});";
				put "   InsertInto(Here:ActionButtonDesc,0);";					
				put ");";
				put "Here:ActionButtonDesc[LArryIndex]=""Launch Interactive Order and Recombination Plots"";";
			%end;
			%if &tmpcount>1 %then %do;
				%if &tmpcount>2 %then %do;
					put "InsertInto(Here:ActionButtonList[LArryIndex],0);";
					put "InsertInto(Here:ActionCodeList[LArryIndex],{{""""}});";				
				%end;
				put "Here:ActionButtonList[LArryIndex][&tmpcount]= _ag_buttonname;";
				put "Here:ActionCodeList[LArryIndex][&tmpcount]=expr(ldhvp&i());";
			%end;
			%else %do;
				put "Here:ActionButtonList[LArryIndex][1]= _ag_buttonname;";
				put "Here:ActionCodeList[LArryIndex][1]=expr(ldhvp&i());";				
			%end;
			%let tmpcount = %eval(&tmpcount+1);
		%end;
	%end;
	
	/* kemicl Jan242012 added work to make interactive triangle plots reorder for reversal*/
	put "Here:RevGroups = Function({},{default local},";
		%if &nGrps > 1 %then %do;
			put " aglist = { ""_All_"" " @;
		%end;
		%else %do;
			put "	aglist = { " @;
		%end;
			%do jj=1 %to &nGrps;
				%if %index(&Singletons,"&&LGrp&jj")= 0 %then %do; 
					put ", ""&&OLGrp&jj"" " @;
				%end;
			%end;
			put "};";
			/* create a list of grps and max gds to pass to sas*/
			put " maxdlist = {};";
			put "	nmarklist = {};";
			put " grplist = {};";
			%let maxi = 1;
			%do jj=1 %to &nGrps;
				%if %index(&Singletons,"&&LGrp&jj")= 0 %then %do;
					put " InsertInto(maxdlist, &&&&LgMax&&maxi);";
					put "	InsertInto(nmarklist, &&LGtot&jj);";
					%if &LGType = C %then %do;
						put "	InsertInto(grplist, ""&&OLGrp&jj"");";
					%end;
					%else %do;
						put "	InsertInto(grplist, &&OLGrp&jj);";
					%end;	
					%let maxi = %eval(&maxi + 1);
				%end;	
			%end;
			/* put "Show(maxdlist, grplist);"; */
			put " dlgg = New Window(""Select the Linkage Groups for Order Reversal"",<<modal,";
			put "   vlistbox(";
			put "    cglistb=listBox(aglist),";
			put "    buttonbox(""OK"", cglist=cglistb<<Get Selected; selindex = cglistb<<Get Selected Indices;)));";
			put "	try(Current Data Table(Here:odat), Throw(""The Linkage Map Order results data set is not open.""));";
			put "	//show(cglist); show(current data table());";
			put " if(nitems(cglist)>0, ";
			put " 	try(close(Here:rdat,nosave););";
			/* change values in open map data */
			put "		if(Contains(cglist,""_All_""),";
			/* create selected groups and maxes to ready for SAS to flip orders for interactive plots */
			put "			selgrps = grplist;";
			put "			selmax = maxdlist;";
			put "			selnmark = nmarklist;";
			put "			numsel = nitems(selgrps);";
			put "			for each row(";
			put "				cms_tmp = :cM_Step; cm_tmp = :cM_&MapFunc; ord_tmp = :Map_Order; rf_tmp = :RecombinationF;";
			%if &LODDataFlag %then %do;
				put "				LOD_tmp = :LOD;";
			%end;				
			put "				:cM_Step = :_cMStepR_; :cM_&MapFunc = :_cMR_; :Map_Order = :_RMO_; :RecombinationF = :_rfR_;";
			put "				:_cMStepR_ = cms_tmp; :_cMR_ = cm_tmp; :_RMO_ = ord_tmp; :_rfR_ = rf_tmp;";
			%if &LODDataFlag %then %do;
				put "				:_LODR_ = lod_tmp;";
			%end;
			put "			);";
			put "		,";  /*else for if contains all */
			%if &nGrps > 1 %then %do;
				put "			selindex = selindex -1;";
			%end;
			put "			selgrps = grplist[selindex];";
			put "			selmax = maxdlist[selindex];";
			put "			selnmark = nmarklist[selindex];";
			put "			numsel = nitems(cglist);";
			put "			for each row(";
			put "				if(Contains(cglist,Char(:&LGroupVar)),";
			put "					cms_tmp = :cM_Step; cm_tmp = :cM_&MapFunc; ord_tmp = :Map_Order; rf_tmp = :RecombinationF;";
			%if &LODDataFlag %then %do;
				put "				LOD_tmp = :LOD;";
			%end;			
			put "					:cM_Step = :_cMStepR_; :cM_&MapFunc = :_cMR_; :Map_Order = :_RMO_; :RecombinationF = :_rfR_;";
			put "					:_cMStepR_ = cms_tmp; :_cMR_ = cm_tmp; :_RMO_ = ord_tmp; :_rfR_ = rf_tmp;";
			%if &LODDataFlag %then %do;
				put "				:_LODR_ = lod_tmp;";
			%end;			
			put "				);";
			put "			);";			
			put "		); "; /* close All list contain */
			/* remove gb so I can resort the data table and save it with new orders*/
			put "		try(Here:gb<<Delete);";
			put " 		try(Here:df<<Delete);";
			put " 		wait(0.5);";
			put " 		Here:odats = Here:odat << Sort(By( :&LGroupVar, :Map_Order ),Order( Ascending, Ascending ), Invisible);";	
			put " 		Here:odats << save(""&ClientOutPath.&LinkageMapOrder_OutDataName..sas7bdat"", Replace(1));";
			put " 		Here:FlipOrderExpr;";
			put " 		Close(Here:odats, nosave);";
			put " 		Close(Here:odat, nosave);";
			put " 		Here:odat = open(""&ClientOutPath.&LinkageMapOrder_OutDataName..sas7bdat"",invisible);";
			put " 		Column(""&LGroupVar"")<<Set Modeling Type(Nominal);";
			put " 		Column(""&IDVar"")<<Label(1);";	
			put " 		Current Data Table(Here:odat);";
			put " 		try(Here:LO_HB<<Prepend(HListBox(Eval(Here:GB_results)), Horizontal));";					
			%if ^%length(&OrderVar) and %index(%nrbquote(%upcase(&Algorithm)),MULTIDIMENSIONAL)>0  %then %do;
				put "		gettabsel = trtb << Get Selected();";
				put " 		mdstab = trtb<<Index of Tab(""MDS Residual Plots"");";
				put "		if(mdstab>0,";
				put "			trtb<<Delete(mdstab);";
				put "			trtb<<Append(""MDS Residual Plots"", Eval(expr(Here:Rfunc())));";
				put "			trtb << Set Selected(gettabsel);";
				put "		);";
			%end;
			put "	);"; /* Close If loop for if groups were selected */
	put ");";		
	
	put "InsertInto(Here:ActionButtonDesc ,""Click to Select Linkage Groups for Marker Order Reversal"");";	
	put "InsertInto(Here:ActionButtonList,{{""Reverse Linkage Group Marker Orders""}});";
	put "InsertInto(Here:ActionCodeList,{{Here:RevGroups}});";
	
	put "if(host is(Windows),";
	put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, windows) ),";
	put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, POSIX) ));";
	put "	";
	
	put "srbuttonswindow=Function({},{default local},";
	put "		VListBox(";
	put "			TextBox(""Click the button below to launch the Linkage Map Viewer process to draw the full linkage map.""),";
	put "			ButtonBox(""Linkage Map Viewer"",";
	put "				if(host is(Windows),";
	put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/LinkageMapViewer.jsl"", absolute, windows)),";
	put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/LinkageMapViewer.jsl"", absolute, POSIX) ));";
	put "	";
	/* Set input data set and Output Folder */
	put "				jg:setDefaultParmValues(2, ""LinkageMapViewer"");";				
	put "				jg:LinkageMapViewer_updateDeps;";
	put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""InData""),""&ClientOutPath.&LinkageMapOrder_OutDataName..sas7bdat"");";		
	put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""OutPath""),""&ClientOutPath"");";										
	put "				jg:LinkageMapViewer_updateDeps;";				
	put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""ChrVar""),""&LGroupVar"");";	
	put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""Label""),""&IDVar"");";
	put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""PosVar""),""cM_&MapFunc"");";
	/*
	put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""ColorVar""),""&ColorVar"");";
	put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""ColorType""),""Continuous"");";
	*/
	put "				jg:LinkageMapViewer_updateDeps;";
	/*put "				jg:setCurrentParmValue(2, ""LinkageMapViewer"", jg:getParmNum(""LinkageMapViewer"",""Theme""),""Blue to Gray to Red"");"; */
	put "				Wait(.5);";
	/* put "				jg:runProcess( 2, ""LinkageMapViewer"" );"; */
	put "			),";
	put "			TextBox("" ""),";	
	put "			TextBox(""Click the button below to launch the Subset and Reorder Genetic Data process.""),";
	put "			TextBox(""The genotype marker data must be reordered to match the linkage map order for QTL analysis.""),";
	put "			ButtonBox(""Subset and Reorder Genetic Data"",";
	put "				if(host is(Windows),";
	put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/SubsetandReorderGeneticData.jsl"", absolute, windows)),";
	put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/SubsetandReorderGeneticData.jsl"", absolute, POSIX) ));";
	put "	";
	/* Set input data set and Output Folder */
	put "				jg:setDefaultParmValues(2, ""SubsetandReorderGeneticData"");";	
	%if %length(&SNPInData) %then %do;
		put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""InData""),""&SNPInData"");";
		put "				jg:SubsetandReorderGeneticData_updateDeps;";
		%if %length(&AlleleVars) le 32767 %then %do;		
			put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""AlleleVars""),""&AlleleVars"");";
		%end;
		put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""ListMarkerVars""),""&ListMarkerVars"");";
		put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""MarkerFmt""),""&MarkerFmt"");";
		put "				jg:SubsetandReorderGeneticData_updateDeps;";
		put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""GenoDelim""),""&GenoDelim"");";
	%end;
	put "				jg:SubsetandReorderGeneticData_updateDeps;";
	put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""MinPercent""),""0"");";		
	put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""OutPath""),""&ClientOutPath"");";					
	put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""AnnoInData""),""&ClientOutPath.&LinkageMapOrder_OutDataName..sas7bdat"");";					
	put "				jg:SubsetandReorderGeneticData_updateDeps;";				
	put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""AnnoMName""),""&IDVar"");";	
	put "				jg:setCurrentParmValue(2, ""SubsetandReorderGeneticData"", jg:getParmNum(""SubsetandReorderGeneticData"",""ReorderVars""),""&LGroupVar Map_Order"");";	
	
	put "				jg:SubsetandReorderGeneticData_updateDeps;";	
	put "			)";	
	put "		)";
	put ");";
	
	put "Here:LaunchCode = srbuttonswindow();";
run;

%PackageInsertFile(JSL Code from &ProcessName, &OutPath, &LinkageMapOrderScriptName, "&LinkageMapOrder_OutDataName..sas7bdat,&LinkageMapOrder_OutDataResName..sas7bdat", yes);


/* for now just launch file 
%PackageInsertDataset(Output data from &ProcessName, OutLib, &LinkageMapOrder_OutDataName);
*/

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend;

%LinkageOrder;
