/*
JMP Genomics Analytical Process

Name: Marker Properties
Author: Wendy Czika

Description:
Marker Properties calculates various single marker 
measures, such as marker informativeness, 
test for HWE, and allele and genotype frequencies.
A cell plot representing the data set containing
individuals' genotypes is created, with red
representing the genotype homozygous for the major 
allele, gray representing a heterozygous genotype,
and blue representing the genotype homozygous for
the minor allele.

See the ALLELE procedure in the SAS/Genetics User's
Guide for more information.
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData AlleleVars ListMarkerVars PloidyNum MarkerFmt GenoDelim SexVar ByVars WhereStmt HWEWhereStmt AnnoWhereStmt
AnnoInData AnnoLabel AnnoMName AnnoLoc AnnoGroup AnnoAccession XLVarStmt HWETest Perms Boot BAlpha Seed Conv MTC PValAdj Tau Bandwidth Alpha OnlySig
ChartFreqs CreateHTML CreateCellPlot TransposeCellPlotData Recode OutPath ServerOutPath OutName MacroPath
MarkerPropertiesScript MarkerProperties_OutData MarkerProperties_Frame MarkerProperties_Body MarkerProperties_Contents
NumGenoData AlleleFreqData GenotypeFreqData PropMissData;

*ProcessBody;

%macro MarkerProperties;

%include "&MacroPath./GenMacros.sas" / nosource;

%GenInit(MarkerProperties,28);
%if &exiterror=1 %then %goto exit;
%global FullAPName;
%let FullAPName=%str(Marker Properties);

%if %length(&ListMarkerVars &AlleleVars)=0 %then %do;
 %put ERROR: Marker variables must be specified in either the Marker Variables or List-Style Specification of Marker Variables parameters.;
 %let exiterror=1;
 %goto exit;
%end;

%global OldAlleleVars;
%let OldAlleleVars = &AlleleVars;
%let AlleleVars = &AlleleVars &ListMarkerVars;
%global XLVars;
%let XLVars =;

%if %length(&PloidyNum) = 0 %then %let PloidyNum= 2;

%if &PloidyNum > 2 %then %do;
	%if (^%Index(&MarkerFmt, %str(Numeric)) and ^%Index(&MarkerFmt, %str(Nondelimited))) %then %do;
		%put ERROR: Ploidy Number greater than 2 can only be analyzed for either Numeric Genotypes or Nondelimited Genotypes marker format options in the Options tab.; 
		%let exiterror=1;
	%end;
	
	%if &Boot > 0 %then %do;
		%put ERROR: Ploidy Number greater than 2 cannot be analyzed with the Bootstrap Sample option in the Options tab. The Boostrap Sample has be set to zero.; 
		%let exiterror=1;
	%end;

	%if %upcase(&HWETest)=EXACT or %upcase(&HWETest)=PERMUTATION %then %do;
		%put ERROR: Ploidy Number greater than 2 cannot be analyzed with either Exact or Permutation Test for HWE in the Options tab.; 
		%let exiterror=1;
	%end;
	
	%if %length(&HWEWhereStmt) %then %do;
		%put ERROR: Ploidy Number greater than 2 cannot be analyzed with Filter to Include Observations for HWE Test in the General tab.; 	
		%let exiterror=1;
	%end;
	
	%if %length(&SexVar) %then %do;
		%put ERROR: Ploidy Number greater than 2 cannot be analyzed with Sex Variable in the General tab.; 	
		%let exiterror=1;	
	%end;
	
	%if %length(&XLVarStmt) %then %do;
		%put ERROR: Ploidy Number greater than 2 cannot be analyzed with X-linked Markers Clause in the Annotation tab.; 	
		%let exiterror=1;	
	%end;

	%if &exiterror=1 %then %do;
		%goto exit;
	%end;
%end;

%PrepOutDS(MarkerProperties_OutData, MarkerProperties_OutDataName, &OutName._hwe);

%PrepOutDS(NumGenoData, NumGenoDataName, &OutName._num);

%PrepOutDS(GenotypeFreqData, GenotypeFreqDataName, &OutName._gfr);

%PrepOutDS(AlleleFreqData, AlleleFreqDataName, &OutName._afr);

%PrepOutDS(PropMissData, PropMissDataName, &OutName._ipm);

%PrepOutFile(MarkerPropertiesScript,MarkerPropertiesScriptName,&ProcessName Script for &OutName..jsl);

%PrepMapandBy;
run;
%if &exiterror=1 %then %goto exit;

/*
S1574488: Creating a list of variables that might be in the AnnoInData and are calculated in this process. 
If these variables are not deleted from the AnnoInData, they will appear as a results of this process, 
therefore, the output data table will have wrong results. 
*/ 
%let DropAnnoVars=  NAllele PIC Het Div ProbChiSq ChiSq df NIndiv _Total MajorAllele MinorAllele MAF PropMissing;
data mapdata;
	set mapdata;
	drop &DropAnnoVars;
run;

%if %sysfunc(exist(work.mapdata_orig)) %then %do;
	data mapdata_orig;
		set mapdata_orig;
		drop &DropAnnoVars;
	run;
%end;

%PrepHTML(markerproperties_body, markerproperties_frame, markerproperties_contents, &OutName._&ProcessName._body.html, &OutName._&ProcessName._frame.html, &OutName._&ProcessName._contents.html);

%if %upcase(&CreateHTML)=YES %then %do;
* setup for html output;
ods html 
   body = "&MarkerProperties_Body" (url="&bodyfile")
   frame = "&MarkerProperties_Frame"  (url="&framefile")
   contents = "&MarkerProperties_Contents" (url="&contentsfile")
;
%end;

ods listing close;
%if %length(&WhereStmt) %then %do;
 data countobs; 
 	set &ds;
  where &WhereStmt;
 run;
 %Nobsvars(countobs);
 %let ds=countobs;
%end;
%else %Nobsvars(&ds);

%if %length(&Perms)=0 %then %let Perms=0;

%let nm=&nummark;

%if &nobs=0 %then %do;
	%put ERROR: There are no observations in the data set to be analyzed.  Check that you have entered a valid Where Clause.;
	%goto exit;
%end;

%if &Boot>0 %then %let BootDiv = 10;
%else %let BootDiv = 1;

%let BlockSize = %sysfunc(min(100000,%sysevalf((20000000/&nobs)/&BootDiv,floor)));

%if &BlockSize > &nm %then %let BlockSize = &nm;
%put BLOCKSIZE = &BlockSize;

%let sexclause =;
%if %length(&XLVars) %then %do;
	%let numnxmark = %eval(&nm - &numxmark);
	/*S1151346: Luciano, making sure that SexVar is not empty and XLVars is being used*/
	%*if %length(&SexVar)=0 %then %let sexvar=sex;
	%if %length(&SexVar)=0 %then %do;
		%put ERROR: A Sex Variable must be specified in the General Tab when using the X-Link Markers Clause from the Annotation Tab.;
		%let exiterror = 1;
		%goto exit;
	%end;
	%let dsid=%sysfunc(open(&ds,i));
	%let sextype=%sysfunc(vartype(&dsid,%sysfunc(varnum(&dsid,&sexvar))));
	%let rc=%sysfunc(close(&dsid));
	%if %quote(&sextype) = %str(N) %then %do;
		%let sexclause = %nrbquote(&sexvar = 2);
	%end;
	%else %do;
		%let sexclause = %nrbquote(substr(&sexvar,1,1) in ("F","f"));
	%end;
%end;
%else %do;
	%let numnxmark=&nm;
	%let sexclause = 1;
%end;
%put sexclause= &sexclause;

%let first = 1;
%let last=%sysfunc(min(&BlockSize,&numnxmark)); /* last marker in block */

data _null_;
	set count;
	%if %length(&AlleleVars) %then %do;
		array getname{*} &AlleleVars;
	%end;
	%if %length(&XLVars) %then %do;
		array getnamex{*} &XLVars;
	%end;		
	%let Block=1;
	%if %length(&AlleleVars) %then %do;
		%do %until (&first > &numnxmark);	
			%if %length(&GenoCol) %then %do;
				_f = vname(getname(&first));
				_l = vname(getname(&last));
			%end;
			%else %do;
				_f = vname(getname(%eval(2*&first-1)));
				_l = vname(getname(%eval(2*&last)));
			%end;

			call symput("firstvar&Block",_f);
			call symput("lastvar&Block",_l);
			%let Size&Block = %eval(&last - &first + 1);

			%let first = %eval(&last + 1);
			%let last = %eval(&first + &BlockSize - 1);
			%let Block = %eval(&Block + 1);
			%if &last > &numnxmark %then %let last = &numnxmark;
		%end;
	%end;
	/* Xlinked Markers - don't want a mix of xlinked & nonxlinked markers in same block*/
	%if %length(&XLVars) %then %do;
		%let first=1;
		%let last = %sysfunc(min(&BlockSize,&numxmark));
		%do %until (&first > &numxmark);	
			%if %length(&GenoCol) %then %do;
				_f = vname(getnamex(&first));
				_l = vname(getnamex(&last));
			%end;
			%else %do;
				_f = vname(getnamex(%eval(2*&first-1)));
				_l = vname(getnamex(%eval(2*&last)));
			%end;

			call symput("firstvar&Block",_f);
			call symput("lastvar&Block",_l);
			%let Size&Block = %eval(&last - &first + 1);

			%let first = %eval(&last + 1);
			%let last = %eval(&first + &BlockSize - 1);
			%let Block = %eval(&Block + 1);
			%if &last > &numxmark %then %let last = &numxmark;
		%end;		
	%end;	
run;	
		
%let HWEWhereStmtFlag=0;		
%if %length(&HWEWhereStmt) %then %let HWEWhereStmtFlag=1;

%let NBlocks = %eval(&Block-1);
%let NChunks=%eval((&NBlocks-1)/100 + 1);
%let ChunkSize = %sysfunc(min(100,&NBlocks)); /* Number of blocks in chunk */
%put NBlocks = &NBlocks;
%put NChunks = &NChunks;
%put ChunkSize = &ChunkSize;

%let NIndivVar=NIndiv;
%let validhwefilter=0;

%let first = 1; /* first marker in block */
%let last = &Size1;
%let Block = 1;	
%RemoveData(ms allelefreqs genotypefreqs);

**S1439803: check if mapdata has a variable named name**;
%let _ds_=%sysfunc(open(mapdata));
%let FlagName= %sysfunc(varnum(&_ds_,name));
%let _rc_=%sysfunc(close(&_ds_));		
%if &FlagName and %upcase(&AnnoLabel) ^= NAME %then %do;
	proc datasets nolist;
		modify mapdata;
		rename name= name_orig;
	quit;
%end;

%put NChunks= &NChunks;
%do c=1 %to &NChunks;

	/* Last block in current chunk */
	%let BinC = %sysfunc(min(%eval(&Block + &ChunkSize -1),&NBlocks));
	%put c = &c;
	%put BinC = &BinC;
	
	data %do b=&Block %to &BinC; avars&b(keep=&ByVars &&firstvar&b--&&lastvar&b usehwe &sexvar) %end; %if &c=1 %then other(drop=&AlleleVars &XLVars usehwe);;
	/* Need to make sure AlleleVars are in consecutive variables */
		set &ds(drop=&AlleleVars &XLVars);
		set &ds(keep=&AlleleVars &XLVars);
		
		%if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) and &PloidyNum = 2 %then %do;
			format &firstvar1--&&lastvar&NBlocks numtochar.;
		%end;
		
		%if &HWEWhereStmtFlag eq 1 %then %do;
			if %unquote(&HWEWhereStmt) then do;
				usehwe=1;
				%if &Block=1 %then call symputx("validhwefilter",1);;
			end;
			else do;
				usehwe=0;
			end;
		%end;
	run;

	%if &HWEWhereStmtFlag eq 1 and &Block=1 and &validhwefilter=0 %then %do;
		%put ERROR: No observations meet the criteria specified in the Filter to Include Observations for the HWE Test.  This filter is ignored.;
		%let HWEWhereStmtFlag=0;
	%end;
		
	/*data OutLib.mp_mapdata; set mapdata; run;*/	

	%do %while (&Block le &BinC);
	
		%put Block = &Block;
	
		%if &first gt &numnxmark %then %do;
		/* For X-linked markers, need to get stats for males and females to get overall MAF - must be corrected for males*/
			%let NewByVars = &ByVars &sexvar;
			
			proc sort data=avars&Block out=avarssort;
				by &NewByVars;
			run;
			
			%let mindata=avarssort;
		%end;
		%else %do;
			%let NewByVars = &ByVars;
			%let mindata=avars&Block;
		%end;

		/*data OutLib.mp_mindata&Block; set &mindata; run;*/	
		%if &PloidyNum = 2 %then %do;
		
			%put HWEWhereStmtFlag= &HWEWhereStmtFlag;
			%if &HWEWhereStmtFlag eq 1 %then title "Results Without HWE Test Filter";;
			ods output MarkerSumm=ms_;
			ods output AlleleFreq=allelefreqs_;
			ods output GenotypeFreq=genotypefreqs_;
			proc allele data=&mindata ndata=mapdata(rename=(&AnnoLabel=name) firstobs=&first obs=&last) %if %length(&BAlpha) %then alpha=&BAlpha; exact=&Perms boot=&Boot &GenoCol %if %length(&Seed) %then seed=&Seed; ;  
				 %if %length(&NewByVars) %then %do; 
						by &NewByVars;
				 %end;
				 var &&firstvar&Block -- &&lastvar&Block;  
			run;
			
			/*data OutLib.mp_ms_d; set ms_; run;*/
			/*data OutLib.mp_allelefreqs_d; set allelefreqs_; run;*/
			/*data OutLib.mp_genotypefreqs_d; set genotypefreqs_; run;*/				
		
		%end;
		%else %do;			
			
			/*data OutLib.mp_mindata; set &mindata; run;*/
		
			***allele frequency***;
			proc transpose data=&mindata out=trans/*(drop=_name_)*/ prefix=indiv;
				var &&firstvar&Block -- &&lastvar&Block;  
			run;			
		
			/*data OutLib.mp_trans1; set trans; run;*/
		
			data af_poly;
				set trans;
				retain NIndiv NAllele MajorAllele MinorAllele MAF;
				array samp{*} indiv:;
				Total= dim(samp);
				*put Total;
				
				NIndiv=0; *total of nonmissing samples*;
				CountA=0; *total of minorallele A*;
				CountB=0; *total of majorallele B*;
				do i=1 to dim(samp);
					if not missing(samp[i]) and samp[i]^="." then do;
						NIndiv= NIndiv + 1;
						%if %upcase(&MarkerFmt)=%str(NONDELIMITED GENOTYPES) %then %do;
							CountA= CountA + countc(samp[i], "A"); 
							CountB= CountB + countc(samp[i], "B"); 	
						%end;
						%else %if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) %then %do;
							CountA= CountA + samp[i]; 	
							CountB= CountB + (&PloidyNum - samp[i]); 
						%end;							
					end;
				end;
				*CountB= NIndiv * &PloidyNum - CountA;
				*total of other alleles*;
				CountO= NIndiv * &PloidyNum - (CountA + CountB); 
				FreqA= 0;
				FreqB= 0;
				if NIndiv > 0 then do;
					FreqA= CountA/(NIndiv * &PloidyNum);
					FreqB= CountB/(NIndiv * &PloidyNum);
				end;
				MAF= 0;
				if CountA <= CountB then do;  
					*minor allele*;
					MinorAllele= "A"; 
					MajorAllele= "B";
					if NIndiv > 0 then do;	
						*minor allele frequency*;
						MAF= CountA/(NIndiv * &PloidyNum); 
					end;
				end;
				else do;	     
					*minor allele*;
					MinorAllele= "B";
					MajorAllele= "A";
					if NIndiv > 0 then do;		 
						*minor allele frequency*;
						MAF= CountB/(NIndiv * &PloidyNum); 
					end;
				end;
				if CountA > 0 and CountB > 0 and CountO > 0 then do;
					NAllele=.;
				end;
				else if CountA > 0 and CountB > 0 and CountO = 0 then do;
					NAllele=2;
				end;
				else if (CountA = 0 or CountB = 0) and CountO = 0 then do;
					NAllele=1;
					MinorAllele = "";
					MajorAllele = "";
				end;
				drop i indiv:;
			run;
			
			/*data OutLib.mp_af_poly0; set af_poly; run;*/
			/*data OutLib.mp_mapdata; set mapdata; run;*/
		
			data allelefreqs_;
				merge mapdata(rename=(&AnnoLabel= Locus) /*keep=&AnnoLabel*/) af_poly;
				Allele= "A";
				Count= CountA;
				Freq= FreqA;
				output;
				Allele= "B";
				Count= CountB;
				Freq= FreqB;
				output;
				keep Locus Allele Count Freq;
				*drop CountA CountB FreqA FreqB;	
			run;
			
			/*data OutLib.mp_allelefreqs_; set allelefreqs_; run;*/
			
			*** genotype counts ***;
			data &mindata;
				set &mindata;
				_rowid_= _n_;
			run;
		
			proc transpose data=&mindata out=trans;
				var &&firstvar&Block -- &&lastvar&Block;  
				by _rowid_;
			run;
			
			data trans;
				set trans;
				by _rowid_;
				if first._rowid_ then _mkorder_=1;
				else _mkorder_ = _mkorder_ + 1;
				retain _mkorder_;
			run;
			
			proc sort data=trans;
				by _mkorder_;
			run;
			
			/*data OutLib.mp_trans2; set trans; run;*/
			
			proc freq data=trans;
				tables col1;
				by _mkorder_ notsorted;
				ods output OneWayFreqs=genotypefreqs_;
			run;
			
			data genotypefreqs_;
				set genotypefreqs_;
				Freq= Percent/100;
				rename col1= Genotype Frequency=Count;
				drop Table F_Col1 Percent CumFrequency CumPercent;
			run;
			
			proc sql UNDO_POLICY=NONE;
				create table genotypefreqs_ as
				select a1.*, a2.*
				from genotypefreqs_ as a1 left join mapdata(rename=(&AnnoLabel= Locus) keep= &AnnoLabel _order_) as a2
				on a1._mkorder_=a2._order_;
			quit;			
			
			/*data OutLib.mp_genotypefreqs_; set genotypefreqs_; run;*/	

			**compute chisquare for hwe**;
			data dummy;
				%do k=0 %to &PloidyNum;
					_order_= -99;
					%if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) %then %do;
						Genotype= %sysevalf(&k);
					%end;
					%else %do;
						*nondelimited*;
						Genotype= substr(repeat('A', &k)||repeat("B",%sysevalf(&PloidyNum - &k)),2,&PloidyNum);
					%end;
					output;
				%end;
			run;
			
			data geno_hwe;
				set dummy genotypefreqs_;
				%if %upcase(&MarkerFmt)=%str(NONDELIMITED GENOTYPES) %then %do;
					if Genotype ^= "." and Genotype ^= " " then _Genotype_ = countc(Genotype, "A");
					else _Genotype_ = ".";
				%end;
			run;
			
			/*data OutLib.mp_geno_hwe; set geno_hwe; run;*/
			
			proc transpose data=geno_hwe out=geno_hwet prefix=GenoClass;
				var Count;
					%if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) %then %do;
						id Genotype;
					%end;
					%else %do;
						*nondelimited*;
						id _Genotype_;
					%end;				
				by _order_;
			run;
			
			/*data OutLib.mp_geno_hwet0; set geno_hwet; run;*/
			
			data af_poly;
				merge mapdata(rename=(&AnnoLabel= Locus) ) af_poly(keep= _name_ NIndiv NAllele FreqA FreqB);
			run;
			
			/*data OutLib.mp_af_poly1; set af_poly; run; */

			proc sql UNDO_POLICY=NONE;
				create table geno_hwet as
				select a1.*, a2.*
				from geno_hwet(drop= _name_) as a1 left join af_poly(keep= _order_ _name_ Locus NIndiv NAllele FreqA FreqB) as a2 
				on a1._order_=a2._order_
				;
			quit;
			
			/*data OutLib.mp_geno_hwet1; set geno_hwet; run;*/
			data ms_;
				retain Locus NIndiv NAllele ChiSq DF ProbChiSq %do k=0 %to &PloidyNum; GenoClass&k %end;;
				set geno_hwet;
				ChiSq = 0;
				*DF= 0;
				%do k=0 %to &PloidyNum; 
					if GenoClass&k = . then GenoClass&k=0;
					*if GenoClass&k > 0 then df= df + 1;
					EGenoClass&k = NIndiv * COMB(&PloidyNum, &k) * (FreqA**&k) * (FreqB**(&PloidyNum - &k));
					ChiSq = ChiSq + ((GenoClass&k - EGenoClass&k)**2)/EGenoClass&k; 	
				%end; 
				*if DF > 0 then DF = DF - 1;
				DF = NAllele - 1;
				ProbChiSq= 1 - PROBCHI(ChiSq, DF); 
				if _order_ = -99 then delete;
				Keep Locus NIndiv NAllele ChiSq DF ProbChiSq %do k=0 %to &PloidyNum; GenoClass&k %end;;
			run;

			/*data OutLib.mp_ms_; set ms_; run;*/			
		%end;
		
		%if %length(&XLVars) and &first le &numnxmark and %upcase(&HWETest)=EXACT and &HWEWhereStmtFlag eq 0 %then %do;
			/* If XL vars exist, for non-XLinked, must create "_ALL" vars for HWE exact test */
			data ms_;
				set ms_;
				NIndiv_All=NIndiv;
				NAllele_All=NAllele; 
				Het_All=Het;
			run;
		%end;
		
		%if &HWEWhereStmtFlag eq 1 or &first gt &numnxmark %then %do;
		
			%if &first gt &numnxmark %then %do;
				/* Need to get combined allele freqs, females only for all other stats */
				data allelefreqs_;
					set allelefreqs_;
					_oldorder_= _n_;
				run;
				
				proc sort data=allelefreqs_;
					by locus allele;
				run;
				
				data aff afm;
					set allelefreqs_;
					if %unquote(&sexclause) then output aff;
					else output afm;
				run;
				
				/* Adjust allele counts for males */
				data allelefreqs_;
					merge aff(rename=(count=countf freq=freqf)) afm(rename=(count=countm freq=freqm));
					by Locus allele;
					if countf=. then countf=0;
					if countm=. then countm=0;
					if freqf=. then freqf=0;
					if freqm=. then freqm=0;
					Count = Countf + Countm/2;
					Freq = (Countf*Freqf + Countm*Freqm/2) / Count;
					drop countf countm freqf freqm stderr &sexvar;
				run;
				
				proc sort data=allelefreqs_ out=allelefreqs_(drop=_order_);
					by _oldorder_;
				run;				
				
				/* Need to get combined genotype freqs*/
				data genotypefreqs_;
					set genotypefreqs_;
					_oldorder_= _n_;
				run;
				
				proc sort data=genotypefreqs_;
					by locus genotype;
				run;
				
				data gff gfm;
					set genotypefreqs_;
					if %unquote(&sexclause) then output gff;
					else output gfm;
				run;
				
				data genotypefreqs_;
					merge gff(rename=(count=countf freq=freqf)) gfm(rename=(count=countm freq=freqm));
					by Locus genotype;
					if countf=. then countf=0;
					if countm=. then countm=0;
					if freqf=. then freqf=0;
					if freqm=. then freqm=0;					
					Count = Countf + Countm;
					Freq = (Countf*Freqf + Countm*Freqm) / Count;
					drop countf countm freqf freqm stderr d &sexvar;
				run;		
				
				proc sort data=genotypefreqs_ out=genotypefreqs_(drop=_order_);
					by _oldorder_;
				run;							
				
				/* Females only for other stats: PIC, Div, Het */
				data ms_;
					set ms_;
					_oldorder_= _n_;
				run;
				
				proc sort data=ms_;
					by locus;
				run;
				
				data ms_f ms_m;
					set ms_;
					if %unquote(&sexclause) then output ms_f;
					else output ms_m;
				run;
				
				/* set males first so females overwrites, combine N_Indiv */
				data ms_;
					merge  ms_m(rename=(NIndiv=NIndiv_m )) ms_f(rename=(NIndiv=NIndiv_f));
					by Locus ;
					NIndiv = NIndiv_m + NIndiv_f;
					drop NIndiv_m  NIndiv_f &sexvar;
				run;			
				
				proc sort data=ms_ out=ms_(drop=_order_);
					by _oldorder_;
				run;				
			
			%end;
			
			/* Run on HWE filter only for HWE test */
			
			/*data OutLib.mp_avars&block; set avars&block; run;*/
			
			%if &HWEWhereStmtFlag eq 1 %then title "Results Using HWE Test Filter";;
			ods output MarkerSumm=hwe 
			%if %upcase(&HWETest)=EXACT %then allelefreq=afhwe_;;
			ods exclude %if %upcase(&HWETest) ne EXACT %then allelefreq; genotypefreq;
			proc allele data=avars&Block ndata=mapdata(rename=(&AnnoLabel=name) firstobs=&first obs=&last) exact=&Perms &GenoCol %if %length(&Seed) %then seed=&Seed; ;  
				 %if %length(&ByVars) %then %do; 
						by &ByVars;
				 %end;
				 var &&firstvar&Block -- &&lastvar&Block;  
				 where 
				 %if &HWEWhereStmtFlag eq 1 and &first gt &numnxmark %then %do;
				 	usehwe=1 and %unquote(&sexclause)
				 %end;
				 %else %if &HWEWhereStmtFlag eq 1 %then %do;
				 	usehwe=1
				 %end;
				 %else %do;
				 	%unquote(&sexclause)
				 %end;
				 ;
			run;		
			
			data ms_;
				/* For HWE exact, need to keep both sets of NIndiv, NAllele, Het;
					ChiSq stats only needed from HWE filter run */
				set ms_%if %upcase(&HWETest)=EXACT %then(rename=(NIndiv=NIndiv_All NAllele=NAllele_All Het=Het_All));;
				set hwe(keep=&ByVars Locus ChiSq Df ProbChiSq
					%if %upcase(&HWETest)=PERMUTATION %then ProbExact;
					%if %upcase(&HWETest)=EXACT %then NIndiv NAllele Het;)
				;
			run;			
			
			%if %upcase(&HWETest)=EXACT %then %let NIndivVar=NIndiv_All;
			
		%end;

		%if %sysfunc(exist(allelefreqs_))=0 %then %do;
			%put ERROR: PROC ALLELE computations could not be completed.  See the SAS log for more details.;
			%let exiterror=1;
			%goto exit;
		%end;
		
		proc datasets nolist;
		%if &Block=1 %then %do;
			change ms_=ms allelefreqs_=allelefreqs genotypefreqs_=genotypefreqs;
			%if (&HWEWhereStmtFlag eq 1 or &first gt &numnxmark) and %upcase(&HWETest)=EXACT %then change afhwe_=afhwe;;
		%end;
		%else %do;
			append base=ms data=ms_ force;
			append base=allelefreqs data=allelefreqs_ force;
			append base=genotypefreqs data=genotypefreqs_ force;
			%if (&HWEWhereStmtFlag eq 1 or &first gt &numnxmark) and %upcase(&HWETest)=EXACT %then append base=afhwe data=afhwe_ force;;			
		%end;
			%if %upcase(&CreateCellPlot) ne YES and %upcase(&Recode) ne YES and (&Block>1 or %length(&ByVars)=0) 
			%then delete avars&Block;;
		quit;

		%let first = %eval(&last + 1);
		%let Block = %eval(&Block + 1);		
		%if &Block le &NBlocks %then %let last = %eval(&first + &&Size&Block - 1);
	%end;

%end; /* end chunk */

title " ";

%if %upcase(&CreateHTML)=YES %then %do;
	ods html close; 
%end;

%if &syserr>4 %then %goto exit;

proc sort data=ms;
 by &ByVars Locus;
run;

%if %length(&ByVars) %then %do;
	data bycts;
		set avars1(keep=&ByVars);
		const=1;
	run;

	proc means data=bycts noprint;
		by &ByVars;
		var const;
		output out=bycounts(keep=&ByVars _Total) n=_Total;
	run;	
	
	data ms;
		merge ms bycounts;
		by &ByVars;
	run;
%end;

data mapdata;
	set mapdata;
	_neworder_ = _n_;
run;

proc sort data=mapdata out=mapsort;
 by &AnnoLabel;
run;

proc sort data=allelefreqs out=maf;
	by &ByVars Locus descending freq;
run;

data maf;
	set maf;
	by &ByVars Locus;
	retain MajorAllele MinorAllele MAF;
	if first.locus then do;
		MajorAllele=Allele;
		MAF = 1-Freq;
		allct = 1;
	end;
	else allct + 1;
	if last.locus then do;
		if allct =2 then MinorAllele = Allele;
		else MinorAllele = "";
		output;
	end;
	keep &ByVars Locus MAF MajorAllele MinorAllele;
run;	

%if (&HWEWhereStmtFlag eq 1 or %length(&XLVars)) and %upcase(&HWETest)=EXACT %then %do;

	proc sort data=afhwe out=mafhwe nodupkey;
		by &ByVars Locus descending freq;
	run;

	data mafhwe;
		set mafhwe;
		by &ByVars Locus;
		if first.locus;
		keep &ByVars Locus MAF;
		MAF = 1-Freq;
	run;	
	
	data maf;
		merge maf(rename=(MAF=MAF_All)) mafhwe(in=sub);
		by &ByVars Locus;
		if sub=0 then MAF = MAF_all;
	run;
%end;

data msmaf;
	merge ms maf;
	by &ByVars Locus;
run;

proc sort data=msmaf; 
	by Locus;
run;

%let NDigits = %sysevalf(%sysfunc(floor(%sysfunc(log10(&nm))))+1);
%if %length(&AnnoLabel)>29 %then %let AnnoSubL = %substr(&AnnoLabel,29);
%else %let AnnoSubL = &AnnoLabel;

/*data OutLib.mp_mapsort; set mapsort; run;*/
/*data OutLib.mp_msmaf; set msmaf; run;*/

data temp;
	%if (%upcase(&CreateCellPlot)=YES and %upcase(&MarkerFmt) ne %str(NUMERIC GENOTYPES)) or %upcase(&Recode)=YES %then %do;
		length &AnnoSubL._RG $%eval(7+&NDigits);
	%end;
	merge mapsort msmaf(in=k rename=(Locus=&AnnoLabel));
	by &AnnoLabel;
	label &AnnoLoc= ' ';
	%if %length(&ByVars) %then PropMissing=1 - &NIndivVar/_Total;
	%else PropMissing = 1 - &NIndivVar/&nobs;;
	if k;
	%if %upcase(&HWETest)=PERMUTATION %then rename ProbExact = ProbHWEPerm;;
	rename ProbChiSq = ProbHWE ChiSq=ChiSqHWE df=dfHWE;;
	%if (%upcase(&CreateCellPlot)=YES and %upcase(&MarkerFmt) ne %str(NUMERIC GENOTYPES)) or %upcase(&Recode)=YES %then %do;
		&AnnoSubL._rg = "numgeno"||strip(_order_);
		%if %length(&AnnoMName) %then %let AnnoMName = &AnnoSubL._RG;
	%end;
	drop _neworder_;
run;

%if %upcase(&HWETest) = EXACT %then %do;
	data temp;
		set temp;
		if NAllele=2 then do;
			_X = round(NIndiv * Het);
			_Na = round(2*NIndiv*MAF);
			_Nb = round(2*NIndiv - _Na);
			_Const = lgamma(NIndiv+1) + lgamma(_Na+1) + lgamma(_Nb+1) - lgamma(2*Nindiv+1);
			_NonconstX = lgamma((_Na-_X)/2+1) + lgamma(_X+1) + lgamma((_Nb-_X)/2+1) - _X*log(2);
			ProbExact_obs = exp(_Const - _NonconstX);
			ProbExact = ProbExact_obs;
			ProbExact_last = ProbExact_obs;
			if mod(_X,2)=0 then _start=2;
			else _start=3;
			/* Recursive calculation given by Wiggington et al. 2005 */
			do _het = _X to _start by -2;	
				ProbExact_curr = ProbExact_last *  _het * (_het-1) / (4 * ((_Na-_Het)/2 + 1) * ((_Nb-_Het)/2 + 1));
				if ProbExact_curr <= ProbExact_obs then ProbExact + ProbExact_curr;
				ProbExact_last = ProbExact_curr;
			end;
			ProbExact_last = ProbExact_obs;
			do _het=_X to min(_Na,_Nb)-2 by 2;
				ProbExact_curr = ProbExact_last * (_Na-_Het) * (_Nb-_Het) / ((_het+2)*(_het+1));
				if ProbExact_curr <= ProbExact_obs then ProbExact + ProbExact_curr;
				ProbExact_last = ProbExact_curr;
			end;		
			ProbExact = Min(ProbExact,1.0);
		end;
		else ProbExact = .;
		rename ProbExact = ProbHWEExact;
		drop _X _Na _Nb _Const _NonconstX _het _start ProbExact_:;
		%if &HWEWhereStmtFlag eq 1 or %length(&XLVars) %then %do;
			drop MAF NIndiv NAllele Het;
			rename MAF_All = MAF NIndiv_All=NIndiv NAllele_All=NAllele Het_All=Het;
		%end;
	run;
%end;

proc sort data=temp out=&MarkerProperties_OutData;
 by &ByVars &AnnoGroup &AnnoLoc;
run;

%if %upcase(&HWETest)=ASYMPTOTIC %then %let Test=HWE;
%else %if %upcase(&HWETest)=EXACT %then %let Test=HWEExact;
%else %let Test=HWEPerm;

%let nsubs=&nobs;

/*data OutLib.mp_MarkerProperties_OutData0; set &MarkerProperties_OutData; run;*/

%if %upcase(&HWETest) ne NONE %then %do;
	%Psmooth(&Test, &MarkerProperties_OutData, &MarkerProperties_OutDataName);
	%if (&syserr>4) %then %let HWETest = NONE;
	
	/*data OutLib.mp_MarkerProperties_OutData1; set &MarkerProperties_OutData; run;*/
%end;
%else %do;
/* Code from Psmooth to order data correctly */

	%if %length(&AnnoWhereStmt) %then %do;
		proc sort data=&MarkerProperties_OutData out=&MarkerProperties_OutData;
		 by _oldorder_;
		run;	

		/* Need to do this so don't have missing values for BY variables for excluded markers */
		%if %length(&ByVars) %then %do;
			proc sort data=&MarkerProperties_OutData out=byvars(keep=&ByVars) nodupkey;
				by &ByVars;
			run;

			%NObsvars(mapdata_orig);
			%let mapobs=&nobs;
			%NObsvars(byvars);

			data mapdata_orig_by;
				set byvars;
				do _i=1 to &mapobs;
					output;
				end;
				drop _i;
			run;

			data mapdata_orig_by;
				set mapdata_orig_by;
				set %do i=1 %to &nobs; mapdata_orig %end;;
			run;

			%if &syserr>4 %then %goto endps;

			proc sort data=mapdata_orig_by out=mapdata_orig;
				by _oldorder_ &ByVars;
			run;

			proc datasets nolist;
				delete mapdata_orig_by;
			quit;

		%end;

		data &MarkerProperties_OutData;
			merge mapdata_orig(drop=keepthis _xlvar) &MarkerProperties_OutData end=_closeds;
			by _oldorder_;
			%if %symexist(MarkerFmt) %then %do;
				%if &MarkerFmt = %str(Numeric Genotypes) %then %do;
					retain _ads_;
					if _n_=1 then do;
						if exist("count_old") then  _ads_=open("count_old","i"); 
						else _ads_=open("count","i");
					end;
					_Column_Name_ = varname(_ads_,_oldorder_);
					if _closeds then _rc_=close(_ads_);
				%end;
			%end;
			drop _oldorder_ _order_ _ads_ _rc_;
		run;
	%end;
	%else %do;
		proc sort data=&MarkerProperties_OutData out=&MarkerProperties_OutData(drop=_order_);
			by _order_;
		run;	
	%end;
%end;

proc sort data=allelefreqs ;
	by  Locus &ByVars;
run;

data &AlleleFreqData;
  merge mapsort allelefreqs(rename=(locus=&AnnoLabel));
  by &AnnoLabel;
  drop ControlVar _neworder_ _oldorder_;
run;

proc sort data=&AlleleFreqData out=&AlleleFreqData(drop=_order_);
	by _order_ &ByVars ;
run;
 
proc sort data=genotypefreqs ;
	by  Locus &ByVars;
run;

data &GenotypeFreqData;
  merge mapsort Genotypefreqs(rename=(locus=&AnnoLabel));
  by &AnnoLabel;
  drop ControlVar _neworder_ _oldorder_;
run;

proc sort data=&GenotypeFreqData out=&GenotypeFreqData(drop=_order_);
	by  _order_ &ByVars;
run;

%if (&syserr>4) %then %goto exit;

/* Fix for S0944012  - using _: for list-style spec of marker vars */
%let countid=%sysfunc(open(count));
%let ncv = %sysfunc(attrn(&countid,nvars));

%put nummark = &nummark;
  
data &PropMissData(drop=&AlleleVars &XLVars _i);
	length _w1 _w2 %do k= 3 %to &PloidyNum; _w&k %end; $16;
	set &ds;
	array __a{*} %do _i_=1 %to &ncv; %sysfunc(varname(&countid,&_i_)) %end; &XLVars;
	array __g__{&nummark} _temporary_;

	%if %index(%upcase(&MarkerFmt),%str(GENOTYPES)) %then %do;
		do _i=1 to &nummark;
			__g__{_i}=1;
			if missing(__a{_i}) then __g__{_i}=.;
			%if %upcase(&MarkerFmt) = %str(GENOTYPES) %then %do;
				else do;
					_w1=strip(scan(__a{_i},1,"&GenoDelim"));
					_w2=strip((scan(__a{_i},2,"&GenoDelim")));
					if missing(_w1) or missing(_w2) then __g__{_i}=.;
				end;
			%end;
			%else %if %upcase(&MarkerFmt)=%str(NONDELIMITED GENOTYPES) %then %do;
				else do;
					_w1=substr(__a{_i},1,1);
					_w2=substr(__a{_i},2,1);
					%do k= 3 %to &PloidyNum;
						_w&k=substr(__a{_i},&k,1);
					%end;					
					if missing(_w1) or missing(_w2) %do k= 3 %to &PloidyNum; or missing(_w&k) %end; then __g__{_i}=.;
				end;	
			%end;
		end;
	%end;
	%else %do;
		do _i=1 to &nummark;
			if missing(__a{2*_i-1}) or missing(__a{2*_i}) then __g__{_i}=.;
			else __g__{_i}=1;
		end;
	%end;
	PropMissingGenotypes = nmiss(of __g__{*})/&nummark;
	drop _w1 _w2 %do k= 3 %to &PloidyNum; _w&k %end; ;
run;	

%let rc = %sysfunc(close(&countid));
	
%if %upcase(&CreateCellPlot)=YES or %upcase(&Recode)=YES %then %do;
	
	%if %upcase(&MarkerFmt)=%str(NUMERIC GENOTYPES) %then %do;
		data &NumGenoData;		
			set &ds;
			array _av{*} &AlleleVars &XLVars;
			do i=1 to dim(_av);
				call symput("Label"||strip(i),vlabel(_av{i}));
			end;
			drop i;
		run;

		%if (%upcase(&TransposeCellPlotData) = YES) %then %do;
			options nolabel;
			proc transpose data=&NumGenoData out=&NumGenoData prefix=Subject name=Marker;
				var &AlleleVars &XLVars;
			run;

			proc datasets library=OutLib nolist;
				modify &NumGenoDataName;
				label Marker=' ';
			quit;

		%end;			
	%end;	
	%else %do;

		%if %length(&ByVars)>0 %then %do;
			proc sort data=allelefreqs;
				by Locus Allele;
			run;

			data af;
				set allelefreqs;
				by locus allele;
				if first.allele then count_sum=count;
				else count_sum+count;
				if last.allele then output;
				drop &ByVars count;
				rename count_sum=count;
			run;

		%end;
		%else %do;
			%RemoveData(af);
			proc datasets nolist;
				change allelefreqs=af;
			quit;
		%end;

		proc sort data=af;
			by locus descending count;
		run;

		data afsub;
			merge af mapsort(rename=(&AnnoLabel=locus));
			by locus;
			if first.locus;
		run;		

		%if (%upcase(&TransposeCellPlotData) ne YES) %then %do;
			data _null_;
				set afsub;
				call symput('Label'||left(_order_), compress(Locus));
			run;
		%end;		

		%let first = 1; /* first marker in block */
		%let last=&Size1; /* last marker in block */
		%let Block = 1;	
		%removedata(trans);
		%do %while (&Block le &NBlocks);

			%Nobsvars(avars&Block);

			proc transpose data=avars&Block out=trans_ prefix=indiv;
				var &&firstvar&Block -- &&lastvar&Block;
			run;	

			%if &syserr>4 %then %let exiterror=1;

			proc datasets nolist;
				%if &block>1 %then %do;
					append base=trans data=trans_ force;
					delete trans_;
				%end;
				%else %do;
					change trans_=trans;
				%end;
				delete avars&Block;
			run;		

			%let first = %eval(&last + 1);
			%let Block = %eval(&Block + 1);		
			%if &Block le &NBlocks %then %let last = %eval(&first + &&Size&Block - 1);	

		%end;

		proc sort data=afsub;
			by _neworder_;
		run;

		%if %length(&GenoCol)=0 %then %do;
			data trans1(rename=(indiv1-indiv&nobs=findiv1-findiv&nobs)) trans2(rename=(indiv1-indiv&nobs=sindiv1-sindiv&nobs));
				set trans;
				retain maxlen 0;
				%do i=1 %to &nobs;
					if length(strip(indiv&i)) > maxlen then maxlen = length(strip(indiv&i));
				%end;
				call symputx("maxavlen",maxlen);				
				if mod(_n_,2)=1 then output trans1;
				else output trans2;
			run;

			data trans;
				set trans1; set trans2;
				array indiv{&nobs} $%eval(2*&maxavlen + 1);
				array findiv{&nobs} findiv1-findiv&nobs;
				array sindiv{&nobs} sindiv1-sindiv&nobs;
				do _i=1 to &nobs;
					if missing(findiv{_i}) or missing(sindiv{_i}) then indiv{_i}="";
					else indiv{_i} = cats(findiv{_i},"/",sindiv{_i});
				end;
				keep indiv:;
			run;

			proc datasets nolist;
				delete trans1 trans2;
			quit;

			%let GenoDelim=%str(/);

		%end;
		%else %do;
			data _null_;
				set trans;
				retain maxlen 0;
				array iarray{*} indiv1-indiv&nobs;
				do i=1 to dim(iarray);
					if length(strip(iarray{i})) > maxlen then maxlen = length(strip(iarray{i}));
				end;
				call symputx("maxavlen",maxlen);
			run;
			%let maxavlen = %sysevalf(&maxavlen/2,ceil);
		%end;			

		%let Block = 1;	
		%let fobs=1;
		%let lobs=&Size1;

		options compress=binary;

		%do c=1 %to &NChunks;	

			/* Last block in current chunk */
			%let BinC = %sysfunc(min(%eval(&Block + &ChunkSize -1),&NBlocks));

			data transafsub;
				set trans; set afsub;
			run;

			proc sort data=transafsub;
				by _order_;
			run;

			data %do b=&Block %to &BinC; mrge&b %end;;
				length _w1 _w2 %do k= 3 %to &PloidyNum; _w&k %end; _a_ $&maxavlen;
				set transafsub;
				array samp{&nobs} indiv:;
				array _ng{&nobs} 8 %if (%upcase(&TransposeCellPlotData)=YES) %then Subject1-Subject&nobs;;;
				n__=_n_;
				do _i=1 to &nobs;
					%if %length(&GenoDelim) %then %do;
						_w1=compress(scan(samp{_i},1,"&GenoDelim"));
						_w2=strip(trim(scan(samp{_i},2,"&GenoDelim")));
					%end;
					%else %do;
						/* nondelimited genotype */
						%do k= 1 %to &PloidyNum;
							_w&k=substr(samp{_i},&k,1);
						%end;						
					%end;					
					_a_=strip(Allele);
					if missing(_w1) or missing(_w2) %do k= 3 %to &PloidyNum; or missing(_w&k) %end; then _ng{_i}=.;
					else _ng{_i}= &PloidyNum - ((_w1=_a_) + (_w2=_a_) %do k= 3 %to &PloidyNum; + (_w&k=_a_) %end; );
				end;
				drop indiv: _i _w1 _w2 %do k= 3 %to &PloidyNum; _w&k %end; n__ _a_;
				if &fobs <= n__ <= &lobs then output mrge&Block;
				%let Blockt=%eval(&Block+1);
				%let fobst=&lobs;
				%if &Blockt le &NBlocks %then %let lobst=%eval(&lobs+&&Size&Blockt);
				%do %while (&Blockt le &BinC);
					else if &fobst <= n__ <= &lobst then output mrge&Blockt;
					%let Blockt=%eval(&Blockt+1);
					%let fobst=%eval(&lobst+1);
					%if &Blockt le &NBlocks %then %let lobst = %eval(&fobst+&&Size&Blockt - 1);
				%end;
			run;

			%if (%upcase(&TransposeCellPlotData) ne YES) %then %do;
				options label;

				%let FirstBlock=&Block;

				%do %while (&Block le &BinC);

					proc transpose data=mrge&Block out=retrans&Block(drop=_name_) prefix=numgeno;
						var _ng1 - _ng&nobs;
						id _order_;
						idlabel locus;
					run;
					%if &syserr>4 %then %let exiterror=1;

					proc datasets nolist;
					delete mrge&Block;
					quit;

					%if &Block=&BinC %then %do;

						data &NumGenoData;
							%if &c=1 %then %do;
								set other;
							%end;
							%else %do;
								set &NumGenoData;
							%end;
							%do b=&FirstBlock %to &Block;/* Separate SET statements instead of a single MERGE are processed much faster! */
								set retrans&b;
							%end;
						run;	
						%if &syserr>4 %then %do;
							%put ERROR: The data set with numerically coded genotypes was not created.;
						%end;

						proc datasets nolist;
							delete %do b=&FirstBlock %to &Block; retrans&b %end;;
						quit;
					%end;

					%let fobs = %eval(&lobs + 1);
					%let Block = %eval(&Block + 1);		
					%if &Block le &NBlocks %then %let lobst = %eval(&fobst+&&Size&Block - 1);

				%end;	
				options nolabel;
			%end;

			%else %do;

				data &NumGenoData;
					set %if &Block>1 %then &NumGenoData; 
					%do %while (&Block le &BinC); 
						mrge&Block 
						%let fobs = %eval(&lobs + 1);
						%let Block = %eval(&Block + 1);
						%if &Block le &NBlocks %then %let lobst = %eval(&fobst+&&Size&Block - 1);
					%end;
					;
					rename locus=Marker;
				run;

			%end;		
		%end; /* end chunk */			

		proc datasets nolist;
			delete trans afsub;
		quit;

	%end; /* end not numeric genotypes */

%end; /* end of loop for cell plot */

%if (%upcase(&CreateCellPlot)=NO and %upcase(&ChartFreqs)=YES) %then %do;

 data count; 
  set &ds;
  keep &AlleleVars &XLVars;
 run;

 %let dsid=%sysfunc(open(count));
 %let n2snps=%sysfunc(attrn(&dsid,NVARS));
 %if %length(&GenoCol) %then %do;
  %let n2snps=%eval(2*&n2snps);
 %end;
 %let nobs=%sysfunc(attrn(&dsid,NOBS));
 %let ret=%sysfunc(close(&dsid));
 %let nm=%eval(&n2snps/2);
%end;
%else %let nobs=&nsubs;

%if &NumGps>1 %then %let JByAnno = %quote(&AnnoGroup,&JBy);
%else %let JByAnno = &JBy;

%let SummaryDescription = %quote(Type of HWE Test:);
%if &Perms > 0 %then %let SummaryDescription = %quote(&SummaryDescription &HWETest (&Perms));
%else %let SummaryDescription = %quote(&SummaryDescription &HWETest);

%if &NByVal > 1 %then %let ssr=0;
%else %let ssr=1;

%global jsl_after_tabbedreport;
%let jsl_after_tabbedreport=;
%if &ssr %then %do;
	%let jsl_after_tabbedreport=%str($GENOMICS_HOME\JSLFiles\MPSubset.jsl);
%end;

%JSLPvalMap(&Test, &MarkerPropertiesScript, &MarkerPropertiesScriptName, &MarkerProperties_OutDataName, _summchart = %eval(%upcase(&HWETest) ne NONE),  _pvalplots = %eval(%upcase(&HWETest) ne NONE), _subsetdd=&ssr);
%PackageInsertDataset(Marker statistics data from &ProcessName, OutLib, &MarkerProperties_OutDataName,_type=);/* already inserted in macro above*/	

%if %upcase(&CreateHTML)=YES %then %do;
	%PackageInsertHtml(html from &ProcessName, &OutPath, &bodyfile, &framefile, &contentsfile);
%end;
%if %upcase(&ChartFreqs) ne YES %then %do;
	%PackageInsertDataset(Allele frequency data from &ProcessName, OutLib, &AlleleFreqDataName,_type=other);	
	%PackageInsertDataset(Genotype frequency data from &ProcessName, OutLib, &GenotypeFreqDataName,_type=other);	
%end;
%if %upcase(&Recode)=YES %then %PackageInsertDataset(Numerically coded genotypes data from &ProcessName, OutLib, &NumGenoDataName, no);

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend;

* run the macro;
%MarkerProperties;
