/*
JMP Genomics Analytical Process

Name: Relationship Matrix
Author: Kelci Miclaus, SAS Institute Inc.

History:
   kemicl 08Jun09 initial coding 
   kemicl 09Jun10 add pedigree ID options
   kemicl 09Jun10 add option to report IBS pairs


Description: 
This Analytical Process computes a symmetric matrix of pairwise relatedness measures for rows of the input
data set across all SNP loci. The input dataset should have the NxM wide format with N samples/individuals
as rows and M markers in numeric format as columns, and output data will contain the NxN matrix which can 
optionally be used as input to the Q-K Mixed Model AP based on Yu et al. (2006); a paper that proposed a 
mixed model using a random effect for genome-wide relatedness that has a covariance structure proportional
to an estimated IBD matrix. 

 Three genome-wide relatedness estimates can be calculated from this AP: Identity By Descent (based on a 
 simple and logical method that follows previously published algorithms as reviewed in Oliehoek et al. 2006),
 Identity By State, and Allele Sharing Similarity (the latter two calculated using the Distance Procedure in 
 SAS/STAT based on Gower's similarity metric).
 
*/

/*
-----------------------------------------------------------
Copyright (C) 2004 SAS Institute, Inc. All rights reserved. 

Notice: 
The above copyright notice and this notice must appear in
any whole or partial copy of this code and any related
documentation. Permission to use, copy, and modify the
source code contained in this file is hereby granted, but
is limited to customers of SAS with a valid license for the
SAS software product with which this file was provided. 
Except as expressly set forth herein, the terms of such
license apply.
-----------------------------------------------------------
*/

*ProcessVariables;

%global StudyName InData Variables IDVars PedID CopyVars  ColorVar LabelVar ListVars WhereStmt 
AnnoInData AnnoWhereStmt
MatrixType Prefix Constrain ChunkSize SVDRoot IBSThresh IBDThresh DoHeat HeatTheme DoPCA Include3D NFactors OutName
RelationMatrix_OutData RelationMatrix_OutDataEig RelationMatrix_OutDataSVD RelationMatrix_OutStacked
RelationshipMatrixScript OutPath ServerOutPath MacroPath;

*ProcessBody;

%macro RelationshipMatrix;

*initialization and argument checking;
%let ProcessName = RelationshipMatrix; 

data _null_;
   infile "&MacroPath./UtilityMacros.sas";
run;
%if &syserr %then %do;
   %put ERROR: Path for SAS macros is invalid.;
   %goto exit;
%end;
%include "&MacroPath./UtilityMacros.sas" / nosource;
%include "&MacroPath./GenMacros.sas" / nosource;

%if ^%length(&Variables) and ^%length(&ListVars) %then %do;
	%put ERROR: You must specify Marker Variables Within Which to Compute Distances.;
	%let exiterror = 1;
	%goto exit;
%end;

%RemoveData(forchol);

%GetOrigFileName(InData,OrigInData);
run;

%if &syserr %then %goto exit;
%let clockrunning = 1;
%PathName(&InData);
%if &exiterror %then %do;
   %put ERROR: Input dataset specification is invalid.;
   %goto exit;
%end;
%let InData = &TmpName;
libname InLib "&TmpPath";
%let InPath = &TmpPath;
libname OutLib "&OutPath"; 

%if %length(&IBDThresh) and ^%length(&IBSThresh) %then %do;
	%let IBSThresh = &IBDThresh;
%end;

* check name for DistanceMatrix_OutData;
%CheckDataName(_name=OutName,_default=&OrigInData);
%if %length(&OutName)>28 %then %let OutName=%substr(&OutName,1,28);

%if %UPCASE(&MatrixType)=%nrbquote(IDENTITY BY DESCENT) %then %do;
	%let DistMetric=IBD;
	%PrepOutDS(RelationMatrix_OutData,RelationMatrix_OutDataName,&OutName._ibd);
	%PrepOutFile(RelationshipMatrixScript,RelationshipMatrixScriptName,&ProcessName Script for &OutName._ibd.jsl);
%end;
%else %if %UPCASE(&MatrixType)=%nrbquote(IDENTITY BY STATE) %then %do;
	%let DistMetric=IBS;
	%PrepOutDS(RelationMatrix_OutData,RelationMatrix_OutDataName,&OutName._ibs);
	%PrepOutFile(RelationshipMatrixScript,RelationshipMatrixScriptName,&ProcessName Script for &OutName._ibs.jsl);	
%end;
%else %do;	
	%let DistMetric=AS;
	%PrepOutDS(RelationMatrix_OutData,RelationMatrix_OutDataName,&OutName._as);
	%PrepOutFile(RelationshipMatrixScript,RelationshipMatrixScriptName,&ProcessName Script for &OutName._as.jsl);	
%end;

%if %UPCASE(&DoPCA)=YES %then %do;
	%PrepOutDS(RelationMatrix_OutDataEig,RelationMatrix_OutDataEigName,&OutName._eig);
%end;

%if %UPCASE(&SVDRoot)=YES %then %do;
	%PrepOutDS(RelationMatrix_OutDataSVD, RelationMatrix_OutDataSVDName,&OutName._rm);
%end;

%if %sysevalf(&IBSThresh ^=0) and ^%length(&PedId)%then %do;
	%PrepOutDS(RelationMatrix_OutStacked, RelationMatrix_OutStackedName,&OutName._prs);
%end;
/* Creates &nobs - number of obs and &nvars - number of variables */
%NObsVars(Inlib.&InData);
%let NobLen=%length(&nobs);

%if &nobs=0 %then %do;
	%put ERROR: There are no observations in the Input Data Set.  Please check the data set for accuracy.;
	%let exiterror=1;
	%goto exit;	
%end;

%if %Upcase(&DistMetric) = AS %then %let IBSTHRESH = 0;

/* Shorten Prefix to avoid errors with too long of prefix */
%if %length(&Prefix) %then %do;
	%CheckDataName(_name=Prefix, _default=&Prefix);
	%let PreIDTot = %eval(32-&NobLen);
	%if %length(&Prefix) le %eval(32-(&NobLen)) %then %do;
		%let NewPLen=%length(&Prefix);
	%end; 
	%else %if %eval(32-(&NobLen))>0 %then %do;
		%let NewPLen=%eval(32-(&NobLen));
	%end;
	%else %do;
		%let NewPlen=4;
	%end;
		%let Prefix=%substr(&Prefix,1,&NewPlen);
%end;

%if &syserr>4 %then %goto exit;

/* Add ColorVar and LabelVar to CopyVars macro */
%let CopyVars= &CopyVars &ColorVar &LabelVar;

* create combined id variable and create subset for Where Statement;
%if %length(&IDVars) %then %do;
	%let IDCon = trim(left(%scan(%quote(&IDVars),1)));
	%let i = 2;
	%do %while(%length(%scan(%quote(&IDVars),&i)));
		%let IDCon = &IDCon || "_" || trim(left(%scan(%quote(&IDVars),&i)));
		%let i = %eval(&i+1);
	%end; 
	data indat;
		set InLib.&InData%if %length(&WhereStmt) %then(where=(&WhereStmt));;
		retain maxIDLen 0;
		ID_Row = &IDCon;
		if lengthn(ID_ROW) > maxIDlen then maxIDlen = lengthn(ID_ROW);
		call symputx("maxIDLen",maxIDlen);
		drop maxidlen;
	run;
	%let PrefLen=%length(&Prefix);
	%if &maxIDlen > %eval(32-&PrefLen) %then %do;
		%put ERROR: The length of the combined ID Variables is too long to append with the Prefix. The maximum length Prefix + ID can be is &PreIDTot.;
		%put Please specify less ID Variables or a shorter Prefix. ;
		%let exiterror = 1;
		%goto exit;
	%end;	
	%let CopyVars = &IDVars &CopyVars;
	%let IDVars = ID_Row;
	%let _ds = indat;
	
%end;
%else %do;
	%if %length(&WhereStmt) %then %do;
		data indat;	
			set InLib.&InData;
			where &WhereStmt;
		run;
		%let _ds = indat;	
	%end;
	%else %do;
		%let _ds = Inlib.&InData;		
	%end;
%end;	

%let _tid_ = %sysfunc(open(&_ds));
%let _nObs_ = %sysfunc(attrn(&_tid_,NOBS));
%let _tid_ = %sysfunc(close(&_tid_));

%if &_nObs_<=1 %then %do;
	%put ERROR: There must be at least two observations in the Input Data Set that meet the Filter for Including Observations.;
	%put ERROR: Currently &_nObs_ observations that meet the current filter.;
	%let exiterror=1;
	%goto exit;
%end;
%if &_nObs_ < &NFactors %then %do;
	%let NFactors = &_nObs_;
%end;

/*Make default prefix match matrix type */
%if %UPCASE(&DistMetric)=IBS %then %do;
	%if ^%length(&Prefix) %then %let Prefix = IBS_;
	%else %let Prefix = %unquote(&Prefix);
%end;
%else %if %UPCASE(&DistMetric)=IBD %then %do;	
	%if ^%length(&Prefix) %then %let Prefix = IBD_;
	%else %let Prefix = %unquote(&Prefix);
%end;
%else %do;
	%if ^%length(&Prefix) %then %let Prefix = AS_;
	%else %let Prefix = %unquote(&Prefix);
%end;

/* Create &nix for distance chunking if IBS and for transpose chunking if IBD */
data vars count;
	set &_ds;
	keep &Variables &ListVars;
	if (0);
run;
%let _dsix_ = %sysfunc(open(vars));
%let nix = %sysfunc(attrn(&_dsix_,NVARS));
%let nummark = &nix;

%if %length(&PedID) %then %do;
	proc sort data=&_ds out=ds_sort;
		by &PedId;
	run;
	%let _ds = ds_sort;
%end;

%let AlleleVars = &Variables &ListVars;
%if %length(&AnnoWhereStmt) %then %do;

	%GetOrigFileName(AnnoInData,OrigAnnoInData);
	%PathName(&AnnoInData);
	%let AnnoData = &TmpName;
	libname AnnoLib "&TmpPath";
	
	data mapdata;
		set AnnoLib.&AnnoData;
		_order_=_n_;
		if _order_ <= &nummark;
	run;

	%let dsid=%sysfunc(open(mapdata));
	%let nmapobs=%sysfunc(attrn(&dsid,NOBS));
	%let ret=%sysfunc(close(&dsid));

	%if (&nmapobs<&nummark) %then %do;
		data map2;
			do _order_=%eval(&nmapobs+1) to &nummark;
					output;
			end;
		run;

		data mapdata;
			set mapdata map2;
		run;
	%end;

	%let ReorderingVars=;
	%let ds = &_ds;
	%let GenoCol = 1;

	%GenSubset(%quote(&AnnoWhereStmt));
	%let _dsold = &_ds;
	%let _ds = &ds;
	
	%if &exiterror=1 %then %goto exit;
	
	%let _rc_ = %sysfunc(close(&_dsix_));
	/* Need to use new count from GenSubset*/
	%let _dsix_ = %sysfunc(open(count));
	%let nix = %sysfunc(attrn(&_dsix_,NVARS));
%end;

%if %UPCASE(&DistMetric)=IBD %then %do;
	%if %length(&PedId) %then %do;	
		ods exclude all;
		ods noresults;
		proc means data=&_ds mean;
			var &AlleleVars;
			output out=ds_mean(drop=_type_ _freq_) mean=;
		run;
		ods exclude none;
		ods results;
		proc transpose data=ds_mean out=meant;
		run;

		data meant; set meant;
			_std_=sqrt(col1*(1-col1/2));
			rename col1=_xbar_;
		run;
	%end;
	
	%if %length(&CopyVars) %then %do;
		data cv;
			set &_ds(keep=&CopyVars %if %length(&IDVars) %then ID_Row;);
		run;
	%end;
	
	/* Transpose data for IBD calcs. */
	/* Have to do in blocks to prevent Out of Memory in Proc Transpose */
	%let BlockSize=%sysfunc(min(&nix,100000));
	%let first = 1;
	%let last = &BlockSize;

	data _null_;
		/* Make sure &Variables &ListVars are in consecutive variables */
		set &_ds(drop=&AlleleVars);
		set &_ds(keep=&AlleleVars);
		array getname{*} &AlleleVars;
		%let Block=1;
		%do %until (&first > &nix);	

			_f = vname(getname(&first));
			_l = vname(getname(&last));

			call symput("firstv&Block",_f);
			call symput("lastv&Block",_l);

			%let first = %eval(&last + 1);
			%let last = %eval(&first + &BlockSize - 1);
			%let Block = %eval(&Block + 1);
			%if &last > &nix %then %let last = &nix;
		%end;
	run;

	%let NBlocks = %eval(&Block-1);	

	data %do b=1 %to &NBlocks; sub_trans&b(keep=%if %length(&IDVars) %then ID_Row; &PedId &&firstv&b--&&lastv&b) %end;;
		/* Make sure &Variables &ListVars are in consecutive variables */
		set &_ds(drop=%if %length(&IDVars)  %then ID_Row; &PedId  &AlleleVars);
		set &_ds(keep= %if %length(&IDVars) %then ID_Row; &PedId &AlleleVars);
	run;

	%do b=1 %to &NBlocks;
		proc transpose data=sub_trans&b out=trans_ prefix=&Prefix;
			%if %length(&PedId) %then %do;
				by &PedId;
			%end;
			var &&firstv&b--&&lastv&b;
			%if %length(&IDVars) and ^%length(&PedId) %then %do;
				id ID_Row;
			%end;
		run;
		
		/*data OutLib.trans_; set trans_; run;*/

		%if &syserr>4 %then %do;
			%put ERROR: Unable to calculate IBD Matrix, please check that the ID Variables specified are unique to each row;
			%let exiterror=1;
			%goto exit;
		%end;

		proc datasets nolist;
			 %if &b>1 %then %do; 
				append base=tdat data=trans_ force;
			%end;
			%else %do;
				delete tdat;
				change trans_=tdat;
			%end;
			delete sub_trans&b;
		quit;
	%end;
	%let _rc_ = %sysfunc(close(&_dsix_));
	
	/*data OutLib.tdat0; set tdat; run;*/
	
	/* Scale individuals x to be (x-2p)/2pq */
	%if %length(&PedId) %then %do;
		proc sql UNDO_POLICY=NONE;
			create table tdat(drop=tmp_name) as 
			select * from tdat as in1 join meant(rename=(_name_=tmp_name)) as in2
			on in1._name_ = in2.tmp_name;
		quit;
		
		data tdat; set tdat;
			array av{*} &Prefix:;
			do i=1 to dim(av);
				if _std_>0 then av{i}=(av{i}-_xbar_)/_std_;
				else av{i}=.;
			end;
			drop i;
		run;
	%end;
	%else %do;
		data tdat; set tdat;
			array av{*} &Prefix:;
			_xbar_=mean(of &Prefix:);
			_std_=sqrt(_xbar_*(1-_xbar_/2));
			do i=1 to dim(av);
				if _std_>0 then av{i}=(av{i}-_xbar_)/_std_;
				else av{i}=.;
			end;
		run;
	%end;
	
	/*data OutLib.tdat1; set tdat; run;*/
	
	ods exclude all;
	ods noresults;
	proc corr data=tdat  outp=corr(where=(_Name_ ne ' ') drop=_Type_) vardef=n;
		%if %length(&PedId) %then %do;
			by &PedId;
		%end;
		var &Prefix:;
	run;
	ods exclude none;
	ods results;
	
	%if %length(&PedId) %then %do;
		data corr; set corr;
			if sum(of &Prefix:)=. then delete;
			array d{*} &Prefix:;
			do i=1 to dim(d);
				if d{i}=. then d{i}=0;
			end;
			drop i;
		run;
	%end;
	
	%if &syserr>4 %then %do;
		%put ERROR: Unable to calculate IBD Matrix. ;
		%let exiterror = 1;
		%goto exit;
	%end;
	
	data &RelationMatrix_OutData; 
		length Sample_Index $32;
		%if %length(&CopyVars) %then %do;
			merge cv corr;
		%end;
		%else %do;	
			set corr;
		%end;
		%if %length(&IDVars) %then %do; 
			Sample_Index="&Prefix"||trimn(left(ID_Row));
		%end;
		%else %do;
			Sample_Index="&Prefix"||trimn(left(_n_));
		%end;
		%if %UPCASE(&Constrain)=YES %then %do;
			array d{*} &Prefix:;
				do j=1 to dim(d);
					if d{j}<0 then d{j}=0;
				end;
				drop j;
		%end;
		drop _name_;
	run;	
%end;
%else %do;
	/*Luciano: 
	Up until ChangeList 169733 the IBS measure was defined as ibs= 1 - abs(xstar - ystar), where xstar= x/2 and ystar= y/2,
	with x and y coded as 0(aa), 1(Aa), and 2(AA). This produced, for instance, ibs= 1 for two heterozygous individuals (Aa and Aa),
	instead of ibs= 0.5 according to the majority of literature (Endelman, J.B. and Jannink, J.-L. 2012; Oliehoek et al. 2006).
	The IBS outcome values will coincide with those obtained by the Distance Matrix menu option in Tassel 5.2.13 software. 
	Also, the ASS measure was defined as ass= 1 - abs(x - y), where x and y were coded as 0(aa), 1(Aa), and 2(AA). 
	This produced, for instance, ass= 0 for two individuals AA and Aa, instead of ass= 0.5 according to the literature 
	(Gao, X. and Martin, E.R. 2009; Gao, X. and Starmer, J. 2007. The ASS measure in these two articles were not scaled for the 
	number of alleles (2) in a diploid individual, however, in the code here implemented we scaled it by 2 in order for the ASS 
	measure to fall between 0 and 1).
	The only difference between IBS and ASS (according to the above references) is that for two heterozygous individuals (Aa and Aa), 
	ASS= 1 and IBS= 0.5. A trick was used to get the IBS matrix: first compute the GOWER matrix with recoded values xstar= x/2 
	and ystar= y/2 (the GOWER matrix will be exactly the ASS matrix), then compute the DOT distance with recoded values 
	xstar= x if x=1 (heterozygous), and xstar=0 otherwise. Likewise for ystar. Then, set diagnonal elements from the DOD matrix 
	to zero, divide all elements by 2, and finally subtract this resulting matrix from the GOWER (ASS) matrix to correct for the 
	inflated values of heterozygous pairs.
	*/

	* check for chunking to enable calculations on really wide data sets;
	* applies only for AS calculations;
	* when possible, chunk the data for I/O performance;
	%let startv = 1;
	%let endv = %eval(&startv + &ChunkSize - 1);
	%if (&endv > &nix) %then %let endv = &nix;
	%let morevars = 1;
	%let firstone = 1;
	%let firstone2 = 1;
	
	**creating data set for the GOWER measure**;
	data dtmp; set &_ds;
		array d{*} &AlleleVars;
		do i=1 to dim(d);
			d{i}=d{i}/2;
		end;
		drop i;
	run;
	
	%if %UPCASE(&DistMetric)=IBS %then %do;
		**creating data set for the DOT measure**;
		data dtmpibs; set &_ds;
			array d{*} &AlleleVars;
			do i=1 to dim(d);
				if d{i} ^= 1 and d{i} ^= . then d{i}=0;
			end;
			drop i;
		run;
		
		/*data OutLib.dtmpibs; set dtmpibs; run;*/
	%end;
	
	%if %symexist(_dsold)=0 %then %let _dsold = &_ds;
	%let _ds = dtmp;
	
	/*data OutLib._ds; set &_ds; run;*/
	
	%let kk= 1;
	%do %while(&morevars=1);

		%if %UPCASE(&DistMetric)=AS %then %do;
			%put Computing AS distance matrix for vars &startv to &endv out of &nix..;
		%end;
		%else %if %UPCASE(&DistMetric)=IBS %then %do;
			%put Computing IBS distance matrix for vars &startv to &endv out of &nix..;
		%end;

		*****begin AS computations and first step for the IBS computations*******;
		* delete output data set as a way to check that new one is created;
		%if %sysfunc(exist(d1out)) %then %do;
			proc datasets lib=work nolist;
				delete d1out;
			run;
		%end;

		%if (&firstone) %then %let pref1 = &Prefix;
		%else %let pref1 = _TD_;

		proc distance data=&_ds out=d1out method=GOWER nostd prefix=&pref1 shape=SQR  ;
			var interval(%do i=&startv %to &endv; %sysfunc(varname(&_dsix_,&i)) %end;);
			%if %length(&IDVars) %then %do;
				id &IDVars;
			%end;
			%if %length(&PedId) %then %do;
				by &PedId;
			%end;
			%if %length(&CopyVars) %then %do;
				copy &CopyVars;
			%end;
		run;

		%if &syserr>4 %then %do;
			%put ERROR: Unable to calculate distance matrix.;
			%let exiterror = 1;
			%goto exit;
		%end;

		/*data OutLib.d1out; set d1out; run;*/

		%if ^%sysfunc(exist(d1out)) %then %do;
			%put ERROR: Unable to calculate distance matrix as per warnings in the SAS Log.;
			%if %trim(%scan(%nrbquote(&syswarningtext), 7))=%trim(standardized) %then %do;
				%if %UPCASE(&DistMetric)=IBS %then %do;
					%put ERROR: For IBS, all SNPs should be polymorphic.  Please filter out SNPs;
					%put with minor allele frequency MAF=0 by running Marker Properties, and Subset Reorder APs.;
				%end;
			%end;
			%let exiterror = 1;
			%goto exit;
		%end;	

		%if (&firstone) %then %do;
			data GOWERMatrix;
				length Sample_Index $32;
				set d1out;
				%if %length(&IDVars) %then %do; 
					Sample_Index="&Prefix"||trimn(left(ID_Row));
				%end;
				%else %do;
					Sample_Index="&Prefix"||trimn(left(_n_));
				%end;				
			run;
			%let firstone = 0;
		%end;
		%else %do;
			data d1out;
				set d1out;
				keep _TD_:;
			run;
			data GOWERMatrix;
				merge GOWERMatrix d1out;
				array _d1{*} &Prefix.:;
				array _d2{*} _TD_:;
				do i=1 to dim(_d1);
					if _d1[i] ne . and _d2[i] ne . then do;
						_d1[i] = ((&startv - 1)*_d1[i] + (&endv - &startv + 1)*_d2[i])/&endv;
					end;
					else if _d1[i]=. then do;
						_d1[i] = _d2[i];
					end; 
				end;
				drop i _TD_:;
			run;
		%end;
		
		/*data OutLib.GOWERMatrix&kk; set GOWERMatrix; run;*/
		*****end AS computations and first step for the IBS computations*******;
	
		*****begin second step for the IBS computations*******;
		%if %UPCASE(&DistMetric)=IBS %then %do;
			* delete output data set as a way to check that new one is created;
			%if %sysfunc(exist(d1outdot)) %then %do;
				proc datasets lib=work nolist;
					delete d1outdot;
				run;
			%end;

			%if (&firstone2) %then %let pref1ibs = D&Prefix;
			%else %let pref1ibs = _TD_;

			proc distance data=dtmpibs out=d1outdot method=DOT nostd prefix=&pref1ibs shape=SQR;
				var ratio(%do i=&startv %to &endv; %sysfunc(varname(&_dsix_,&i)) %end;);
				%if %length(&IDVars) %then %do;
					id &IDVars;
				%end;
				%if %length(&PedId) %then %do;
					by &PedId;
				%end;
				%if %length(&CopyVars) %then %do;
					copy &CopyVars;
				%end;
			run;

			%if &syserr>4 %then %do;
				%put ERROR: Unable to calculate distance matrix.;
				%let exiterror = 1;
				%goto exit;
			%end;

			%if ^%sysfunc(exist(d1outdot)) %then %do;
				%put ERROR: Unable to calculate distance matrix as per warnings in the SAS Log.;
				%if %trim(%scan(%nrbquote(&syswarningtext), 7))=%trim(standardized) %then %do;
					%if %UPCASE(&DistMetric)=IBS %then %do;
						%put ERROR: For IBS, all SNPs should be polymorphic.  Please filter out SNPs;
						%put with minor allele frequency MAF=0 by running Marker Properties, and Subset Reorder APs.;
					%end;
				%end;
				%let exiterror = 1;
				%goto exit;
			%end;	
			
			/*data OutLib.d1outdot; set d1outdot; run;*/

			%if (&firstone2) %then %do;
				data DotProdMatrix;
					length Sample_Index $32;
					set d1outdot;
					%if %length(&IDVars) %then %do; 
						Sample_Index="D&Prefix"||trimn(left(ID_Row));
					%end;
					%else %do;
						Sample_Index="D&Prefix"||trimn(left(_n_));
					%end;				
				run;
				%let firstone2 = 0;
			%end;
			%else %do;
				data d1outdot;
					set d1outdot;
					keep _TD_:;
				run;
				data DotProdMatrix;
					merge DotProdMatrix d1outdot;
					array _d1{*} D&Prefix.:;
					array _d2{*} _TD_:;
					do i=1 to dim(_d1);
						if _d1[i] ne . and _d2[i] ne . then do;
							_d1[i] = ((&startv - 1)*_d1[i] + (&endv - &startv + 1)*_d2[i])/&endv;
						end;
						else if _d1[i]=. then do;
							_d1[i] = _d2[i];
						end; 
					end;
					drop i _TD_:;
				run;
			%end;

			/*data OutLib.DotProdMatrix&kk; set DotProdMatrix; run;*/
		%end;
		*****end second step for the IBS computations*******;

		%if &endv = &nix %then %do;
			%let morevars = 0;
		%end;
		%else %do;
			%let startv = %eval(&endv+1);
			%let endv = %eval(&startv + &chunksize - 1);
			%if (&endv > &nix) %then %let endv = &nix;
		%end;
		%let kk= %sysevalf(&kk + 1);
	%end;
	%let _rc_ = %sysfunc(close(&_dsix_));	

	%if %UPCASE(&DistMetric)=AS %then %do;
		data &RelationMatrix_OutData; 
			set GOWERMatrix;
		run;
	%end;
	%else %if %UPCASE(&DistMetric)=IBS %then %do;
		*****begin third step for the IBS computations*******;
		%if %length(&PedId) %then %do;
			data DotProdMatrix;
				set DotProdMatrix;
				by Ped_ID;
				array d{*} D&Prefix.:;
				if first.&PedId then _i= 1;
				else _i= _i + 1;
				if not missing(d[_i]) then d[_i] = 0;
				retain _i;
				drop _i;
			run;
		%end;
		%else %do;
			data DotProdMatrix;
				set DotProdMatrix;
				array d{*} D&Prefix.:;
				if not missing(d[_n_]) then d[_n_] = 0;
			run;			
		%end;
		
		data &RelationMatrix_OutData;
			merge GOWERMatrix DotProdMatrix;
			array _d1{*} &Prefix.:;
			array _d2{*} D&Prefix.:;
			do i=1 to dim(_d1);
				if _d1[i] ne . and _d2[i] ne . then do;
					_d1[i] = _d1[i] - _d2[i]*0.5;
				end;
			end;
			drop i D&Prefix.:;
		run;
		
		/*data OutLib.GOWERMatrixDot; set &RelationMatrix_OutData; run;*/
		*****end third step for the IBS computations*******;
	%end;
	
	%if %length(&PedId) %then %do;
		data &RelationMatrix_OutData; 
			set &RelationMatrix_OutData;
			if sum(of &Prefix:)=. then delete;
			array d{*} &Prefix:;
			do i=1 to dim(d);
				if d{i}=. then d{i}=0;
			end;
			drop i;
		run;
	%end;
%end; /* End for IBS and AS */ /*Luciano*/

%if %UPCASE(&SVDRoot)=YES %then %do;
	data forchol(keep=&Prefix:) %if %length(&PedId) %then outped(keep=&PedId);; 
		set &RelationMatrix_OutData;
		array col{*} &Prefix:;
		call symput('maxcol',trim(left(dim(col))));
		do i=1 to dim(col);
			if col{i}=. then col{i}=0;
		end;
	run;
				
	%if %sysfunc(exist(Outlib.&RelationMatrix_OutDataSVD)) %then %do;
		proc datasets lib=Outlib nolist;
			delete &RelationMatrix_OutDataSVD;
		run;
	%end;		
		
	proc iml;
		use forchol;
		read all var _num_ into x;
			%if %length(&PedID) %then %do;
				use outped;
				read all var {&PedID} into ped;
				ur=uniqueby(ped,{1},1:nrow(ped));
				start=1;
				do i=1 to nrow(ur)-1;
				 size=ur[i+1]-ur[i]-1;
				 subx=x[start:start+size,1:1+size];
				 call svd(u,d,v,subx);
				 crsub=u*diag(sqrt(d));
				 if ncol(crsub) < &maxcol then do ;
					 empty = J(nrow(crsub),(&maxcol-ncol(crsub)),.) ;
					 crsub = crsub||empty ;
				 end ;
				 cr= crsub;
				 if i=1 then create svdrt from cr;
				 append from cr ;
				 start=start+size+1;
				end;
				size=nrow(x)-ur[nrow(ur)];
				subx=x[start:start+size,1:1+size];
				call svd(u,d,v,subx);
				crsub=u*diag(sqrt(d));
				if ncol(crsub) < &maxcol then do ;
					empty = J(nrow(crsub),(&maxcol-ncol(crsub)),.) ;
					 crsub = crsub||empty ;
				end ;
				cr= crsub;
				if i=1 then create svdrt from cr;	
			%end;
			%else %do;
				call svd(u,d,v,x);
				cr=u*diag(sqrt(d));
				create svdrt from cr;
			%end;
			append from cr;
	quit;
	
	%NobsVars(svdrt);
	
	data svdrt; set svdrt;
		array d{*} Col:;
		do i=1 to dim(d);
			if d{i}=. then d{i}=0;
		end;
		drop i;
	run;

	proc datasets library=work nolist;
		modify svdrt;
		rename %do i=1 %to &nVars; Col&i = &DistMetric.&i %end;;
	quit;
	
	data &RelationMatrix_OutDataSVD;
		%if %symexist(_dsold) %then %do;
			set &_dsold; 
		%end;
		%else %do;
			set &_ds;
		%end;
		set svdrt;
	run;

	%if &syserr>4 %then %do;
		%let exiterror=1;
		%goto exit;
	%end;	
		
%end; 

%if %sysevalf(&IBSThresh ^= 0) and ^%length(&PedId) %then %do;
	%if ^%sysfunc(exist(forchol)) %then %do;
		data forchol(keep=&Prefix:) %if %length(&PedId) %then outped(keep=&PedId);; 
			set &RelationMatrix_OutData;
			array col{*} &Prefix:;
			call symput('maxcol',trim(left(dim(col))));
			do i=1 to dim(col);
				if col{i}=. then col{i}=0;
			end;
		run;
	%end;
	%if %length(&LabelVar) %then %do;
		data labelds(keep=&Labelvar);
			set &RelationMatrix_OutData;
		run;
		
		%let forcholid = %sysfunc(open(labelds,i));
		%let labeltype = %sysfunc(vartype(&forcholid,1));
		%let rc= %sysfunc(close(&forcholid));
		
		%let SubLabel = %substr(&LabelVar,1,31);
	%end;
	
	data row_refs(keep= Sample1 Sample2 %if %length(&LabelVar) %then &SubLabel.1 &SubLabel.2;);
		length %if %length(&LabelVar) %then &SubLabel.1; Sample1 %if %length(&LabelVar) %then &SubLabel.2; Sample2 $32;
		dsid = open("forchol");
		%if %length(&LabelVar) %then %do;
			labelid = open("labelds");
		%end;
		do i=1 to &maxcol;
			Sample1 = varname(dsid,i);
			%if %length(&LabelVar) %then %do;
				rc=fetchobs(labelid,i);
				%if &labeltype=C %then %do;
					&SubLabel.1 = getvarc(labelid,1);
				%end;
				%else %do;
					&SubLabel.1 = strip(getvarn(labelid,1));
				%end;
			%end;
			do j=1 to i;
				Sample2 = varname(dsid,j);
				%if %length(&LabelVar) %then %do;
					rc=fetchobs(labelid,j);
					%if &labeltype=C %then %do;
						&SubLabel.2 = getvarc(labelid,1);
					%end;
					%else %do;
						&SubLabel.2 = strip(getvarn(labelid,1));
					%end;
				%end;				
				output;
			end;
		end;
		rc=close(dsid);
		%if %length(&LabelVar) %then %do;
			rc=close(labelid);
		%end;
	run;
	
	proc iml;
		start main;
			use forchol;
			read all into d2;
			nrows = nrow(d2);
			/* size of output stacked matrix of measures */
			d1size = nrows*(nrows+1)/2;
			/* matrix for output stacked values in 1 dimension */
			d1 = J(d1size,1,.);
			/* insert values into new matrix */
			do i=1 to nrows;
				do j=1 to i;
					_index = i*(i-1)/2+j;
					d1[_index] = d2[i,j];
				end;
			end;

			create stackout from d1;
			append from d1;
		finish;
		run main;
	quit;
		
	data &RelationMatrix_OutStacked;
		merge row_refs stackout;
		if Sample1 = Sample2 then delete;
		if Col1 < &IBSThresh then delete;
		rename Col1=&DistMetric._Value;
	run;	
	/* create a flag for whether any pairs passed the threshold */
	%NobsVars(&RelationMatrix_OutStacked);
	%if &nobs=0 %then %let NoPairs=1;
	%else %let NoPairs=0;
%end;
	

%if %UPCASE(&DoPCA)=YES %then %do;

	ods exclude all;
	proc princomp data=&RelationMatrix_OutData outstat=f1 n=&NFactors ; 
		var &Prefix:;
		ods output eigenvalues=&RelationMatrix_OutDataEig;
	run;
	ods exclude none;

	%if &syserr>4 %then %do;
		%let exiterror = 1;
		%goto exit;
	%end;

	data _null_;
		set &RelationMatrix_OutDataEig; 
		if _n_ = &NFactors then call symput('Cumulative',Cumulative);
	run; 
	%put Number Of Components = &NFactors;
	%put Cumulative Proportion of Variation = &Cumulative;

	%let PPref = Prin;

	data _null_;
		set &RelationMatrix_OutDataEig;
		if proportion >= 0.0995 then do;
			call symput("prinlab"||trim(left(_n_)),"&PPref" || trim(left(_n_)) || ' (' || put(100*proportion,4.1) || '%)');
		end;
		else do;
			call symput("prinlab"||trim(left(_n_)),"&PPref" || trim(left(_n_)) || ' (' || put(100*proportion,3.1) || '%)');		
		end;
	run;
	
	data f1;
		set f1;
		if _type_ ne "CORR";
		if length(_name_) then idvar = _name_;
		else idvar = _type_; 
	run;

	proc transpose data=f1 out=f2;
		var &Prefix: ;
		id idvar;
	run;

	*S1278486: need to add options label because the GenSubSet macro sets options nolabel*;
	options label; 
	data f2;
		set f2;
		keep _name_ &PPref.1-&PPref.&NFactors;
		label 
			%do i=1 %to &NFactors;
				&PPref.&i = "&&prinlab&i" 
			%end;
		;	
	run;
	
	data &RelationMatrix_OutData;
		set &RelationMatrix_OutData;
		set f2;
		drop _Name_;
	run;
%end;

%MakeVarList(_data=&RelationMatrix_OutData,_nvar=nv,_prefix=&Prefix,_typeflag=0);

/*Remove labels of distance variables in case there are repeated ID values */
proc datasets nolist library=OutLib;
	modify &RelationMatrix_OutDataName;
	label %do i=1 %to &nv; &&&Prefix&i=  %end;;
quit;

%let LabelVarPlot= &LabelVar;
%if %length(&LabelVar) %then %do;
	%CheckLabel(OutLib.&RelationMatrix_OutDataName,LabelVarPlot);
%end;

%let ColorVarPlot= &ColorVar;		
%if %length(&ColorVar) %then %do;
	%CheckLabel(OutLib.&RelationMatrix_OutDataName,ColorVarPlot);
%end;	
			

%global _mainjslfile_ FullAPName;
%let _mainjslfile_ = &RelationshipMatrixScript;
%let FullAPName=%str(Relationship Matrix);

data _null_;
   file "&RelationshipMatrixScript" lrecl=32767;
	put '//!';
	put "Names Default to Here(1);";
	
	put "Here:DataList = {};";
	put "Here:TabButtonList = {};";
	put "Here:TabCodeList = {};";
	put "Here:TabAfterCodeList = {};";
	put "Here:TabOpenList = {};";
	
	%if %UPCASE(&DoHeat)=YES %then %do;
	   /* Make Heatmap of matrix */ 
		put "Here:Htmp = Function({},{default local},";
		put '   << Set Window Icon("HierCluster");';		
		put " If(IsEmpty(Here:maindata), ";
		put "   Here:maindata = open(""&ClientOutpath.&RelationMatrix_OutDataName..sas7bdat"",invisible);";
		put "   Here:maindata << Select Where(1);";
		put "   Here:maindata << Markers(8);";
		put "   Here:maindata << Clear Select;";
		%if %length(&ColorVar) %then %do;
			put " Here:maindata<<Color By Column(Name(""&ColorVarPlot""));";
			%let colorstring = ;
		%end;
		%else %do;
		 %let colorstring = %str(ColorClusters(1),);
		%end;	
		%if %length(&LabelVar) %then %do;
			put "  %str(column(%"&LabelVarPlot%") << label(1);)" ;
		%end;
		%else %if %length(&IDVars) %then %do;
			put "  %str(column(%"ID_Row%") << label(1);)" ;
  	 	%end;
  	 	put ",";  /* end of IsEmpty If stmnt */
      		put "  Current Data Table(Here:maindata));"; /* IsEmpty else stmnt */
		put " Vlistbox(";
		put "   Hierarchical Cluster(Y(" /
			%do j = 1 %to %eval(&nv-1);
				":&&&Prefix&j, " /
			%end;
		 ":&&&Prefix&j), " /
		"      Method(Fast Ward), Standardize(0), Geometric X Scale(1),  " /
		"      Two way clustering, Number of Clusters(10), &colorstring, Dendrogram Scale(""Distance Scale""), Color Map(&HeatTheme), Legend(1)," /
		/*  '      title("' "Hierarchical Clustering Analysis of &DistMetric Estimates" '"),' */
		;
		put '      SendToReport(Dispatch({}, "Dendrogram", OutlineBox, Set Title("Heat Map and Dendrogram")),'/
			 '         Dispatch({"Dendrogram"},"", CellPlotBox, Frame Size(500,500)),'/
			 '         Dispatch({"Dendrogram"},"Clust Dendro", FrameBox(2), Frame Size(125,500)),'/  
			 '         Dispatch({"Dendrogram"},"Clust Dendro", FrameBox(3), Frame Size(500,125)),'/
			 '       )';
		put "   );";
		put "); "; 
	   put ");";
	   put " "; 
	   
	   /*set up first tab as heatmap if requested */
		put "InsertInto(DataList,{{""Here:maindata""}});";
		put "InsertInto(TabButtonList, "" Heat Map"");";
		put "InsertInto(TabCodeList, expr(Here:Htmp()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";  
	 %end; 
	 
	 %if %sysevalf(&IBSThresh ^= 0) and ^%length(&PedId) %then %do;
	 	put "Here:ibsprs = Function({}, {default local},";
	 	%if &NoPairs=0 %then %do;
	 		put " If(IsEmpty(Here:ibsdata), ";
	 		put " 	 Here:ibsdata = open(""&ClientOutpath.&RelationMatrix_OutStackedName..sas7bdat"",invisible));";
			%if %length(&LabelVar) %then %do;
		 		put " column(""&SubLabel.1"") << label(1);";
		 		put " column(""&SubLabel.2"") << label(1);";
			%end;
			%else %do;
		 		put " column(""Sample1"") << label(1);";
		 		put " column(""Sample2"") << label(1);";
			%end;			
	 		put " Current Data Table(Here:ibsdata);";
	 		put "  	Vlistbox(";
	 		put "   		Distribution(Continuous Distribution(Column(:&DistMetric._Value),";
	 		put "    		Moments(1), Horizontal Layout(1), Vertical(0)";
	 		put "       ));";
	 		put "    )";
	 		put ");";  /* close Function for this condition */
			/*set up tab for IBS pairs*/
			put "InsertInto(DataList,{{""Here:ibsdata""}});";
			put "InsertInto(TabButtonList, "" &DistMetric Pairs Results"");";
			put "InsertInto(TabCodeList, expr(Here:ibsprs()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);"; 
		%end;
		%else %do;
			put " VListBox(";
			put "	 TextBox(""No sample pairs have a relatedness measure greater than or equal to &IBSThresh.."",";
			put "    <<Font Color(5), <<Set Wrap(500))";
			put " );";
			put ");"; /* close function*/
			put "InsertInto(DataList,{{"" ""}});";
			put "InsertInto(TabButtonList, "" &DistMetric Pairs Results"");";
			put "InsertInto(TabCodeList, expr(Here:ibsprs()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);"; 
		%end;
	%end;
			
   %if %UPCASE(&DoPCA)=YES %then %do;
   	%if &NFactors > 10 %then %let NFactor2 = 10;
		%else %let nFactor2 = &nFactors;
		put "Here:PCAplots = Function({},{default local},";
		put " If(IsEmpty(Here:maindata), ";
		put "   Here:maindata = open(""&ClientOutpath.&RelationMatrix_OutDataName..sas7bdat"",invisible);";
		put "   Here:maindata << Select Where(1);";
		put "   Here:maindata << Markers(8);";
		put "   Here:maindata << Clear Select;";
		%if %length(&ColorVar) %then %do;
			put " Here:maindata<<Color By Column(Name(""&ColorVarPlot""));";
			%let colorstring = ;
		%end;
		%else %do;
		 %let colorstring = %str(ColorClusters(1),);
		%end;	
		%if %length(&LabelVar) %then %do;
			put "%str(column(%"&LabelVarPlot%") << label(1);)";
		%end;
		%else %if %length(&IDVars) %then %do;
			put "%str(column(%"ID_Row%") << label(1);)";
		%end;
		put ",";  /* end of IsEmpty If stmnt */
      		put "  Current Data Table(Here:maindata));";  /*IsEmpty else stmnt */
		put " Vlistbox(";
		put "   Multivariate(Y( ";
		%do i=1 %to %eval(&nFactor2-1); 
			put "      Column(""&&prinlab&i""),";
		%end;
		put "      Column(""&&prinlab&nFactor2"")),";
		put "      Scatterplot Matrix(Density Ellipses(1), Ellipse Color(1), Horizontal(1)),";
		put "      Parallel Coord Plot(1),Outlier Analysis,"; 
		put '      title("Correlation Analysis of Principal Components"),'; 
		put "      SendToReport(";
		%do i=1 %to %eval(&nFactor2 * &nFactor2);
			put '         Dispatch({"Scatterplot Matrix"},"Multiv Plot",' @;
			put " FrameBox(&i)," @;
			put ' {Marker Drawing Mode(Fast)}),';
		%end;
		put '         Dispatch({"Outlier Analysis","Mahalanobis Distances"},"Multiv Outlier",';
		put "            FrameBox, {Marker Drawing Mode(Fast)}),";
		put '         Dispatch({}, "Parallel Coordinate Plot", OutlineBox,'; 
		put '            Set Title("Parallel Plot of Principal Components"))';
		put "      )";
		put "   )";
		put ");";
		put ");";
		put " ";
		
		put "InsertInto(DataList,{{""Here:maindata""}});";
		put "InsertInto(TabButtonList, "" PCA 2D Row Scores"");";
		put "InsertInto(TabCodeList, expr(Here:PCAplots()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);";  		
		
		
		%if &Include3D = Yes and &nFactor2 > 2 %then %do;
			put "Here:d3pca = Function({},{default local},";
     	   put " If(IsEmpty(Here:maindata), ";
			put "   Here:maindata = open(""&ClientOutpath.&RelationMatrix_OutDataName..sas7bdat"",invisible);";
			put "   Here:maindata << Select Where(1);";
			put "   Here:maindata << Markers(8);";
			put "   Here:maindata << Clear Select;";
			%if %length(&ColorVar) %then %do;
				put "    Here:maindata<<Color By Column(Name(""&ColorVarPlot""));";
				%let colorstring = ;
			%end;
			%else %do;
				%let colorstring = %str(ColorClusters(1),);
			%end;	
			%if %length(&LabelVar) %then %do;
				put "%str(column(%"&LabelVarPlot%") << label(1);)";
			%end;
			%else %if %length(&IDVars) %then %do;
				put "%str(column(%"ID_Row%") << label(1);)";
			%end;
			put ",";  /* end of IsEmpty If stmnt */
			put "  Current Data Table(Here:maindata));";  /*IsEmpty else stmnt */			
			put " Vlistbox(";
			put "   Scatterplot 3D(";
			put "      Y(Column(""&prinlab1""),Column(""&prinlab2""),Column(""&prinlab3"")),";
			put "      Frame3D(";
			put "         Set Marker Transparency( 0.6 ),";
			put "         Set Marker Quality( 1 ),";
			put "         Set Marker Scale( 2.5 )";
			put "      )";
			put "   ),";
			put "));";
			put " ";
			
			put "InsertInto(DataList,{{""Here:maindata""}});";
			put "InsertInto(TabButtonList, "" PCA 3D Row Scores"");";
			put "InsertInto(TabCodeList, expr(Here:d3pca()));";
			put "InsertInto(TabAfterCodeList, "" "");";
			put "InsertInto(TabOpenList,1);";  				
		%end;
		
		/* Scree Plot */
		put "Here:screeplt = Function({},{default local},";
		put " If(IsEmpty(Here:eigdata), ";
		put "		Here:eigdata=open(""&ClientOutpath.&RelationMatrix_OutDataEigName..sas7bdat"", invisible);";
	   put "		Column(""Number"")<<Modeling Type(""Ordinal"");";	
	   put "	,";
		put " 	Current Data Table(Here:eigdata)";
		put " );";
		put " Vlistbox(";
		put "		Graph Builder(";
		put "			Size( 575, 450 ),Show Control Panel( 0 ),Show Legend( 1 ),Legend Position( ""Right"" ),";
		put "			Variables( X( :Number ), Y( :Proportion ), Y( :Cumulative, Position( 1 ) ) ),";
		put "			Elements( Line( X, Y( 1 ), Y( 2 ), Legend( 4 ) ), Points( X, Y( 1 ), Y( 2 ), Legend( 5 ) ) ),";
		put "			SendToReport(";
		put "				Dispatch( {}, ""Graph Builder"", OutlineBox, {Set Title( ""Scree Plot"" )} ),";
		put "				Dispatch( {}, ""graph title"", TextEditBox, {Set Text( """" )} ),";
		put "				Dispatch( {}, ""X title"", TextEditBox, {Set Text( ""Component Number"" )} ),";
		put "				Dispatch( {}, ""Y title"", TextEditBox, {Set Text( ""Portion of Explained Variance"" )} )";
		put "			)";
		put "		);";
		put "));";
		put " ";
		
		put "InsertInto(DataList,{{""Here:eigdata""}});";
		put "InsertInto(TabButtonList, "" Scree Plot"");";
		put "InsertInto(TabCodeList, expr(Here:screeplt()));";
		put "InsertInto(TabAfterCodeList, "" "");";
		put "InsertInto(TabOpenList,1);"; 		
	%end;
run;

%PackageInsertFile(jsl code from &ProcessName, &OutPath, &RelationshipMatrixScriptName, &RelationMatrix_OutDataName..sas7bdat, 
   text/plain, yes);
 
 /* if in C-S mode, copy over eigenvalue data */
%if &Server^=Local %then %do;
	%if %UPCASE(&DoPCA)=YES %then %do;
		%PackageInsertDataset(Eigenvalue data from &ProcessName, outlib, &RelationMatrix_OutDataEigName, yes, no);
	%end;
	 %if %sysevalf(&IBSThresh ^= 0) and ^%length(&PedId) %then %do;
		 %PackageInsertDataset(IBS Pairs data from &ProcessName, outlib, &RelationMatrix_OutStackedName, yes, no);
	 %end;	
%end;
   
%if %UPCASE(&DoHeat) ne YES and %UPCASE(&DoPCA) ne YES %then %do;
    %PackageInsertDataset(&DistMetric Relationship Matrix from &ProcessName, OutLib, &RelationMatrix_OutDataName,no);
 %end;
%if %UPCASE(&SVDRoot)=YES %then %do;
	%PackageInsertDataset(Input Data with &DistMetric SVD Root from &ProcessName, OutLib, &RelationMatrix_OutDataSVDName,no);
%end;
/* %PackagePublish(&OutPath,to_requester);
   */
   
/* Button to launch MDS AP */
%if %UPCASE(&SVDROOT)=YES or ^%length(&PedId) %then %do;
	* add to JSL file;
	filename jslbtns "&RelationshipMatrixScript";

	data _null_;
		file jslbtns mod lrecl=32767;
		put "if(host is(Windows),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, windows) ),";
		put "	include( ConvertFilePath(""$LS_HOME/Generator/DialogHelperFunctions.jsl"", absolute, POSIX) ));";
		put "	";
		put "rmbuttonswindow=Function({},{default local},";
		put "		VListBox(";
		put "			TextBox(""Click on one of the buttons below to launch a follow-up process using the output data set.""),";
		put "			TextBox("" ""),";
		%if %UPCASE(&SVDRoot)=YES %then %do;
			put "			ButtonBox(""Q-K Mixed Model"",";
			put "				if(host is(Windows),";
			put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/QKMixedModel.jsl"", absolute, windows)),";
			put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/QKMixedModel.jsl"", absolute, POSIX) ));";
			put "	";
			put "				jg:setDefaultParmValues(2, ""QKMixedModel"");";
			put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""InData""),""&OutPath.&RelationMatrix_OutDataSVDName..sas7bdat"");";
			put "				jg:QKMixedModel_updateDeps;";		
			%if %length(&Variables) le 32767 %then %do;
				put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""AlleleVars""),""&Variables"");";
			%end;
			put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""PedID""),""&PedID"");";
			put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""AnnoInData""),""&AnnoInData"");";			
			put "				jg:QKMixedModel_updateDeps;";				
			put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""ListMarkerVars""),""&ListVars"");";
			put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""OutPath""),""&ClientOutPath"");";
			put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""MarkerFmt""),""Numeric Genotypes"");";
			put "				jg:setCurrentParmValue(2, ""QKMixedModel"", jg:getParmNum(""QKMixedModel"",""ListKVars""),""&DistMetric.1-&Distmetric.&maxcol."");";	
			put "			),";;
		%end;
		%if ^%length(&PedId) %then %do;
			put "			ButtonBox(""Multidimensional Scaling"",";
			put "				if(host is(Windows),";
			put "					Include(ConvertFilePath(""$HOME/JMPLS/ProcessLibrary/MultidimensionalScaling.jsl"", absolute, windows)),";
			put "					Include( ConvertFilePath(""$HOME/JMPLS/ProcessLibrary/MultidimensionalScaling.jsl"", absolute, POSIX) ));";
			put "	";
			put "				jg:setDefaultParmValues(2, ""MultidimensionalScaling"");";
			put "				jg:setCurrentParmValue(2, ""MultidimensionalScaling"", jg:getParmNum(""MultidimensionalScaling"",""InData""),""&OutPath.&RelationMatrix_OutDataName..sas7bdat"");";
			put "				jg:MultidimensionalScaling_updateDeps;";
			put "				jg:setCurrentParmValue(2, ""MultidimensionalScaling"", jg:getParmNum(""MultidimensionalScaling"",""IDVar""),""&IDVars"");";
			put "				jg:setCurrentParmValue(2, ""MultidimensionalScaling"", jg:getParmNum(""MultidimensionalScaling"",""ListDistVars""),""&Prefix:"");";			
			put "				jg:setCurrentParmValue(2, ""MultidimensionalScaling"", jg:getParmNum(""MultidimensionalScaling"",""OutPath""),""&ClientOutPath"");";
			put "			),";
			/* add autolaunch of new K matrix compression AP */
				put "			ButtonBox(""K Matrix Compression"",";
				put "				if(host is(Windows),";
				put "					Include(ConvertFilePath(""$HOME/JMPG/ProcessLibrary/KMatrixCompression.jsl"", absolute, windows)),";
				put "					Include( ConvertFilePath(""$HOME/JMPG/ProcessLibrary/KMatrixCompression.jsl"", absolute, POSIX) ));";
				put "	";
				put "				jg:setDefaultParmValues(2, ""KMatrixCompression"");";
				put "				jg:setCurrentParmValue(2, ""KMatrixCompression"", jg:getParmNum(""KMatrixCompression"",""InData""),""&OutPath.&RelationMatrix_OutDataName..sas7bdat"");";
				put "				jg:KMatrixCompression_updateDeps;";		
				put "				jg:setCurrentParmValue(2, ""KMatrixCompression"", jg:getParmNum(""KMatrixCompression"",""ListMatrixVars""),""&Prefix:"");";
				put "				jg:setCurrentParmValue(2, ""KMatrixCompression"", jg:getParmNum(""KMatrixCompression"",""CopyVars""),""&CopyVars"");";
				put "				jg:setCurrentParmValue(2, ""KMatrixCompression"", jg:getParmNum(""KMatrixCompression"",""OutPath""),""&ClientOutPath"");";
				put "				jg:setCurrentParmValue(2, ""KMatrixCompression"", jg:getParmNum(""KMatrixCompression"",""SNPInData""),""&InPath.&Indata..sas7bdat"");";
				put "				jg:KMatrixCompression_updateDeps;";	
				%if %length(&Variables) le 32767 %then %do;
					put "				jg:setCurrentParmValue(2, ""KMatrixCompression"", jg:getParmNum(""KMatrixCompression"",""AlleleVars""),""&Variables"");";
				%end;
				put "				jg:setCurrentParmValue(2, ""KMatrixCompression"", jg:getParmNum(""KMatrixCompression"",""ListMarkerVars""),""&ListVars"");";
				put "			)";
		%end;
		put "		),";	
		put ");";

		put "Here:LaunchCode = rmbuttonswindow();";
		
	run;
%end;

%exit:;
%ExitSDSProcess(&OutPath, &ProcessName, &syserr, &exiterror, &clockrunning);

%mend RelationshipMatrix;

* run the macro;
%RelationshipMatrix;

