/* SAS Macros to be used in Recombination and Linkage Groups, 
   Linkage Map Order, Linkage Map Order for Outcross, Cross Evaluation 
*/
/* Support: Kelci Miclaus (kemicl)
	History:
		kemicl Initial Coding
		sbrlsi 11/14/19 Adding LOD Constraint and Moving Needed Macros from LinkageMapOrder.sas to here.
*/

/* This file must be run in SAS to create sasmacr.sas7bcat, which will then be placed 
   in Stored32 and Stored64 folders in /LifeScience/MacroLib
*/

/* initiate options to store macros in the MacroLib */
options mstored sasmstore=orlib;
libname orlib 'C:\PerforceWorkspaces\JMP\LifeSciences\LifeSciences\4101\JMP\MacroLib\Stored64';

/* define macros for solving for optimal order in PROC OPTMODEL */
%macro postSolve /store secure; /* This macro is encrypted */
	if _solution_status_ = 'INFEASIBLE' then do;
		call symputx('exiterror', 1);
		put "Problem is infeasible.  Exiting...";
		leave;
	end;
%mend postSolve;

%macro findConnectedComponents /store secure; /* This macro is encrypted */
	/* find connected components of support graph */
	EDGES1 = {<i,j> in EDGES: round(x[i,j].sol) = 1};
	EDGES1 = EDGES1 union {setof {<i,j> in EDGES1} <j,i>};
	VERTICES1 = INITVERTICES;
	CLOSURE = EDGES1;
	for {i in INITVERTICES} component[i] = 0;
	for {i in VERTICES1} do;
		NEIGHBORS = slice(<i,*>,CLOSURE);
		CLOSURE = CLOSURE union (NEIGHBORS cross NEIGHBORS);
	end;
	numcomp = 0;
	do while (card(VERTICES1) > 0);
		numcomp = numcomp + 1;
		for {i in VERTICES1} do;
			NEIGHBORS = slice(<i,*>,CLOSURE);
			for {j in NEIGHBORS} component[j] = numcomp;
			VERTICES1 = VERTICES1 diff NEIGHBORS;
			leave;
		end;
	end;
	put numcomp=;
	put;
	if numcomp > 1 then do;
		numsubtour = numsubtour + numcomp;
		for {comp in 1..numcomp} do;
			SUBTOUR[numsubtour-numcomp+comp]
			= {i in VERTICES: component[i] = comp};
		end;
	end;
%mend findConnectedComponents;

/* rowGenerationLoop is used when no skeleton marker exists */
%macro rowGenerationLoop(simple=) /store secure; /* This macro is encrypted */
	 %if &simple = 1 %then %do;
		  /* generate constraints dynamically as needed: simple version */
		  do until (numcomp = 1);
			iter = iter + 1;
			put iter=;
			/* solve with current set of constraints */
			put "Solving to optimality...";
			solve with milp %if %upcase(&UseStoppingRules) = YES %then  / &relobjgap2 &maxtime;;
			%postSolve;
			%findConnectedComponents;
			/* can use optimal objective as global lower bound on all subsequent solves 
			lower_bound = obj.sol;
			put lower_bound=;
			*/
		  end;
	 %end;
	 %else %do;
		  /* generate constraints dynamically as needed: fancy version */
		  do until (numcomp = 1);
			iter = iter + 1;
			put iter=;
			/* solve with current set of constraints */
			solve with milp %if %upcase(&UseStoppingRules) = YES %then / &relobjgap1 &maxtime;;
			%postSolve;
			%findConnectedComponents;
			/* put lower_bound=; */

			if (numcomp = 1) then do;
				iter = iter + 1;
				put iter=;
				 /* if solution has no violations, then solve to optimality (starting with feasible solution) */
				 put "Solving to optimality...";
				 solve with milp / primalin %if %upcase(&UseStoppingRules) = YES %then &relobjgap2 &maxtime;;
				 %postSolve;
				 %findConnectedComponents;
				 /* can use optimal objective as global lower bound on all subsequent solves 
				 lower_bound = obj.sol;
			  	 put lower_bound=;
				 */
			end;
		  end;
	 %end;
%mend rowGenerationLoop;

/* Macros used when skeleton marker orders exist for constrained TSP */
%macro computeOrder /store secure; /* This macro is encrypted */
	/* traverse arcs to determine order */
	SUPPORT = {<i,j> in EDGES: round(x[i,j].sol) = 1};
	OPTIMAL_SUPPORT = SUPPORT;
	curr_vert = 'DUMMY';
	for {curr_order in 1..card(VERTICES)-1} do;
		for {<(curr_vert),j> in SUPPORT} do;
			order[j] = curr_order;
			unrank[curr_order] = j;
			curr_vert = j;
			SUPPORT = SUPPORT diff {<curr_vert,j>};
		end;
	end;
%mend computeOrder;

%macro findBadPaths /store secure; /* This macro is encrypted */
	/* find paths that violate precedence constraints */
	violates_precedence = 0;
	put (card(VIOLATED_PAIRS))=;
	for {<i,j> in VIOLATED_PAIRS} do;
		violates_precedence = 1;
		PATH = {};
		order_this = order[j];
		do while (order_this < order[i]);
			PATH = PATH union {<unrank[order_this],unrank[order_this+1]>};
			order_this = order_this + 1;
		end;
		numbadpath = numbadpath + 1;
		BADPATH[numbadpath] = PATH;
	end;
%mend findBadPaths;

%macro SrowGenerationLoop(simple=) /store secure; /* This macro is encrypted */
	 %if &simple = 1 %then %do;
		  /* generate constraints dynamically as needed: simple version */
		  do until (violates_precedence = 0);
			  do until (numcomp = 1);
				feasible = 1;
				iter = iter + 1;
				put iter=;
					/* solve with current set of constraints */
					put "Solving to optimality...";
					solve with milp %if %upcase(&UseStoppingRules) = YES %then  / &relobjgap2 &maxtime;;
					%postSolve;
					%findConnectedComponents;
					/* can use optimal objective as global lower bound on all subsequent solves 
					lower_bound = obj.sol;
					put lower_bound=;
					*/
			  end;
			if feasible = 1 then do;
				%computeOrder;
				%findBadPaths;
			end;
			else leave;
		  end;
	 %end;
	 %else %do;
		  /* generate constraints dynamically as needed: fancy version */
		  do until (violates_precedence = 0);
			do until (numcomp = 1);
				feasible = 1;
				iter = iter + 1;
				put iter=;
				/* solve with current set of constraints */
				solve with milp %if %upcase(&UseStoppingRules) = YES %then / &relobjgap1 &maxtime;;
				%postSolve;
				%findConnectedComponents;
				/* put lower_bound=; */
			end;
			if feasible = 1 then do;
				%computeOrder;
				%findBadPaths;
			end;
			else leave;

			if (numcomp = 1 and violates_precedence = 0) then do;
				feasible = 1;
				iter = iter + 1;
				put iter=;
				 /* if solution has no violations, then solve to optimality (starting with feasible solution) */
				 put "Solving to optimality...";
				 solve with milp / primalin %if %upcase(&UseStoppingRules) = YES %then &relobjgap2 &maxtime;;
				 %postSolve;
				 %findConnectedComponents;
				 /* can use optimal objective as global lower bound on all subsequent solves 
				 lower_bound = obj.sol;
				 put lower_bound=;
				 */
				%computeOrder;
				%findBadPaths;
			end;
		end;
	 %end;
%mend SrowGenerationLoop;

/* Macro called in Linkage Map Order to perform Map Order Optimization */
%macro LMOOptModel(rf_indata=, lod_indata=)/store secure; /* This macro is encrypted */
    /*--- check if OR is licensed; if not, conditionally setinit ---*/
    data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
   		or164_enabled= sysprod('PRODNUM164');
    		if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,1);
    run;
    
	 proc optmodel presolver=none;
		set <str> VERTICES;
		set <str,str> EDGES = {i in VERTICES, j in VERTICES: i > j};
		num numsubtour init 0;
		set <str> SUBTOUR {1..numsubtour};
		read data &rf_indata into VERTICES=[&IdVar];
		VERTICES = VERTICES union {'DUMMY'};
		num c {VERTICES, VERTICES} init 0;
		read data &rf_indata into [&IdVar] {j in VERTICES diff {'DUMMY'}} <c[&IdVar,j]=col(j)>;

		/* x[i,j] = 1 if edge (i,j) is traversed, 0 otherwise */
		var x {EDGES} binary;
		/* minimize the total cost */
		min obj =
			sum {<i,j> in EDGES} c[i,j] * x[i,j];
		/* each vertex has exactly one in-edge and one out-edge */
		con two_match {i in VERTICES}:
			sum {j in VERTICES: i > j} x[i,j]
			+ sum {j in VERTICES: i < j} x[j,i] = 2;
		/* no subtours (these constraints are generated dynamically) */
		con subtour_elim {s in 1..numsubtour}:
			sum {<i,j> in EDGES: (i in SUBTOUR[s] and j not in SUBTOUR[s])
			or (i not in SUBTOUR[s] and j in SUBTOUR[s])} x[i,j] >= 2;
		 num lower_bound init 0;
		 con lower_bound_def:
			  obj >= lower_bound;

		%if %length(&lod_indata) %then %do;
			num d {VERTICES, VERTICES} init 1000;
			read data &lod_indata into [&IdVar] {j in VERTICES diff {'DUMMY'}} <d[&IdVar,j]=col(j)>;	
			*print d;		
			for {<i,j> in EDGES: c[i,j] > &recthreshold or d[i,j] < &lodthreshold} fix x[i,j] = 0;
		%end;
		%else %do;
			for {<i,j> in EDGES: c[i,j] > &recthreshold} fix x[i,j] = 0;
		%end;

		/* this starts the algorithm to find violated subtours */
		set <str,str> EDGES1;
		set <str> INITVERTICES = setof{<i,j> in EDGES1} i;
		set <str> VERTICES1;
		set <str> NEIGHBORS;
		set <str,str> CLOSURE;
		num component {INITVERTICES};
		num numcomp;
		num iter init 0;
		set ITERS = 1..iter;

		%rowGenerationLoop(simple=0);

		set <str,str> SUPPORT init {<i,j> in EDGES: round(x[i,j].sol) = 1};
		create data solData from [i j]=SUPPORT c;

		/* traverse arcs to determine Order */
		num Order {VERTICES};
		str curr_vert init 'DUMMY';
		Order['DUMMY'] = 0;
		for {curr_order in 1..card(VERTICES)-1} do;
			for {<i,j> in SUPPORT: curr_vert in {i,j}} do;
				for {other_end in {i,j} diff {curr_vert}} do;
					Order[other_end] = curr_order;
					curr_vert = other_end;
				end;
				SUPPORT = SUPPORT diff {<i,j>};
				leave;
			end;
		end;
		create data solData2 from [&IdVar]=(VERTICES diff {'DUMMY'}) Order;
	quit;
	/* Turn off conditional setinit for OR and OptModel */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;
%mend;

/*Constrained TSP algorithm macro to honor fixed marker orders as specified by a skeleton marker map */
%macro SkeletonLMO(indata=, skeletondata=, SPosVar=)/store secure;
	 /*--- check if OR is licensed; if not, conditionally setinit ---*/
	 data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
		or164_enabled= sysprod('PRODNUM164');
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
		if or164_enabled ne 1 then rc=zzqbxwf(164,1);
	run;
	
	/* iterative contstrained TSP solution using the subtour formulation */
	proc optmodel presolver=none;
		set <str> VERTICES;
		set <str,str> EDGES = {i in VERTICES, j in VERTICES: i ne j};
		num numsubtour init 0;
		set <str> SUBTOUR {1..numsubtour};
		read data &indata into VERTICES=[&IdVar];
	
		VERTICES = VERTICES union {'DUMMY'};
		num c {VERTICES, VERTICES} init 0;
		read data &indata into [&IdVar] {j in VERTICES diff {'DUMMY'}} <c[&IdVar,j]=col(j)>;
	
		set <str,str> PATH;
		num numbadpath init 0;
		set <str,str> BADPATH {1..numbadpath};
		set <str> SKELETON;
		num position {SKELETON};
		read data &skeletondata into SKELETON=[&IdVar] position=&SPosVar;
		/* print position; */
		num violates_precedence;
		num order_this;
		set PRECEDENCE_PAIRS = {i in SKELETON, j in SKELETON: position[i] < position[j]};
		put (card(PRECEDENCE_PAIRS))=;
	
		/* x[i,j] = 1 if edge (i,j) is traversed, 0 otherwise */
		var x {EDGES} binary;
		/* minimize the total cost */
		min obj =
			sum {<i,j> in EDGES} c[i,j] * x[i,j];
		/* each vertex has exactly one in-edge and one out-edge */
		con assign_i {i in VERTICES}:
			sum {j in VERTICES: i ne j} x[i,j] = 1;
		con assign_j {j in VERTICES}:
			sum {i in VERTICES: i ne j} x[i,j] = 1;
		/* no subtours (these constraints are generated dynamically) */
		con subtour_elim {s in 1..numsubtour}:
			sum {<i,j> in EDGES: i in SUBTOUR[s] and j not in SUBTOUR[s]} x[i,j] >= 1;
		/* no bad paths (these constraints are generated dynamically) */
		con badpath_elim {p in 1..numbadpath}:
			sum {<i,j> in BADPATH[p]} x[i,j] <= card(BADPATH[p]) - 1;
	    	num lower_bound init 0;
	    	con lower_bound_def:
	        obj >= lower_bound;
	
		for {<i,j> in EDGES: c[i,j] > &recthreshold} fix x[i,j] = 0;
		for {<i,j> in PRECEDENCE_PAIRS} fix x[j,i] = 0;
	
		/* this starts the algorithm to find violated subtours */
		set <str,str> EDGES1;
		set <str> INITVERTICES = setof{<i,j> in EDGES1} i;
		set <str> VERTICES1;
		set <str> NEIGHBORS;
		set <str,str> CLOSURE;
		num component {INITVERTICES};
		num numcomp;
		num iter init 0;
		set ITERS = 1..iter;
	
		set <str,str> SUPPORT init {};
		set <str,str> OPTIMAL_SUPPORT init {};
		num order {VERTICES};
		str unrank {0..card(VERTICES)-1};
		order['DUMMY'] = 0;
		unrank[0] = 'DUMMY';
		str curr_vert;
		set VIOLATED_PAIRS = {<i,j> in PRECEDENCE_PAIRS: order[i] > order[j]};
		num feasible;
	
		%SrowGenerationLoop(simple=0);
	
		if feasible = 1 then do;
			/* create data solData from [i j]=OPTIMAL_SUPPORT c; */
			create data solData2 from [&IdVar]=(VERTICES diff {'DUMMY'}) order;
		end;
	quit;
	
	/* Turn off conditional setinit for OR and OptModel */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;
%mend;

/* Find minimum spanning tree of undirected graph */
/* Author: Rob Pratt, SAS/OR, Kelci Miclaus, JMP Genomics    */
/* input data set contains stacked genotype pairs (i,j) and recombination fractions[i,j] */
/* outdata data set outgroups contains edges in tree which defines to linkage groups */
%macro MinimumSpanningTree(rf_indata=, lod_indata=, marki=, markj=, cost=, rfthresh=, lodcost=, lodthresh=, lgthresh=)/store secure;
    /*--- check if OR is licensed; if not, conditionally setinit ---*/
    data _check_;
	or5_enabled  = sysprod('PRODNUM5'  );
   	or164_enabled= sysprod('PRODNUM164');
    	if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
    	if or164_enabled ne 1 then rc=zzqbxwf(164,1);
    run;
	
	proc sort data=&rf_indata;
		by &cost;
	run;
	
	proc optmodel;
		set <str>     NODES;                                                                                                                   
		set <str,str> EDGES;
		num cost {EDGES};
		set <str,str> TREE_EDGES init {};
		num num_components;                                                                                                          
		num component_id {NODES};
		set <str> COMPONENT {k in 1..num_components};

		/* read edges from recombination data set */
		read data &rf_indata into EDGES=[&marki &markj] cost=&cost;
		*print cost;
		
		NODES = union {<i,j> in EDGES} {i, j};
		/* initially, each node is its own component */
		num c init 0;
		for {i in NODES} do;
			c = c + 1;
			num_components = c;
			component_id[i] = c;
			COMPONENT[c] = {i};
		end;

		/* if i and j are in different components, merge the two components and add edge to tree */
		num ci;
		num cj;
		set TEMP_EDGES init EDGES;
/*		num min_cost = min {<i,j> in TEMP_EDGES} cost[i,j];*/
		num min_cost;
		%if %length(&lod_indata) %then %do;	
			set <str,str> LODEDGES;			
			num lodcost {LODEDGES};			
			read data &lod_indata into LODEDGES=[&marki &markj] lodcost=&lodcost;
			*print lodcost;		
			num max_lodcost;
			*print min_cost;
			*print max_lodcost;
/*			do until ((card(setof {g in NODES} component_id[g]) <= &lgthresh) or (min_cost >= &rfthresh));*/
			do until ((card(NODES)-card(TREE_EDGES) <= &lgthresh) or (min_cost >= &rfthresh) or (max_lodcost <= &lodthresh));
/*				for {<i,j> in TEMP_EDGES: cost[i,j] = min_cost} do;*/
				*put TEMP_EDGES=;
				for {<i,j> in TEMP_EDGES} do;
					min_cost = cost[i,j];
					max_lodcost = lodcost[i,j];
					ci = component_id[i]; 
					cj = component_id[j];
					*print min_cost max_lodcost ci cj;

					if ((min_cost <= &rfthresh) and (max_lodcost >= &lodthresh)) then do; 
						if ci ne cj then do;
							/* update smaller component */
							if card(COMPONENT[ci]) < card(COMPONENT[cj]) then do;
								for {k in COMPONENT[ci]} component_id[k] = cj;
								COMPONENT[cj] = COMPONENT[cj] union COMPONENT[ci];
								COMPONENT[ci] = {};
							end;
							else do;
								for {k in COMPONENT[cj]} component_id[k] = ci;
								COMPONENT[ci] = COMPONENT[ci] union COMPONENT[cj];
								COMPONENT[cj] = {};
							end;

							/* add edge to tree */
							*print component_id;
							TREE_EDGES = TREE_EDGES union {<i,j>};

						end;
						TEMP_EDGES = TEMP_EDGES diff {<i,j>};
						leave;
					end;
					TEMP_EDGES = TEMP_EDGES diff {<i,j>};
				end;
/*				put (card(setof {g in NODES} component_id[g]))=;*/
/*				put (card(TREE_EDGES))=;*/
/*				put (card(NODES)-card(TREE_EDGES))=;*/
			end;
		%end;
		%else %do;
			*print min_cost;
/*			do until ((card(setof {g in NODES} component_id[g]) <= &lgthresh) or (min_cost >= &rfthresh));*/
			do until ((card(NODES)-card(TREE_EDGES) <= &lgthresh) or (min_cost >= &rfthresh));
/*				for {<i,j> in TEMP_EDGES: cost[i,j] = min_cost} do;*/
				*put TEMP_EDGES=;
				for {<i,j> in TEMP_EDGES} do;
					min_cost = cost[i,j];
					ci = component_id[i]; 
					cj = component_id[j];
					*print min_cost ci cj;
					if ci ne cj then do;
						/* update smaller component */
						if card(COMPONENT[ci]) < card(COMPONENT[cj]) then do;
							for {k in COMPONENT[ci]} component_id[k] = cj;
							COMPONENT[cj] = COMPONENT[cj] union COMPONENT[ci];
							COMPONENT[ci] = {};
						end;
						else do;
							for {k in COMPONENT[cj]} component_id[k] = ci;
							COMPONENT[ci] = COMPONENT[ci] union COMPONENT[cj];
							COMPONENT[cj] = {};
						end;

						/* add edge to tree */
						*print component_id;
						TREE_EDGES = TREE_EDGES union {<i,j>};

					end;
					TEMP_EDGES = TEMP_EDGES diff {<i,j>};
					leave;
				end;
/*				put (card(setof {g in NODES} component_id[g]))=;*/
/*				put (card(TREE_EDGES))=;*/
/*				put (card(NODES)-card(TREE_EDGES))=;*/
			end;		
		%end;
/*		put "minimum spanning tree has cost " (sum {<i,j> in TREE_EDGES} cost[i,j]); */
		/* write tree edges to data set */
		/*create data &outdata from [&marki &markj]=TREE_EDGES &cost=cost; */
		create data outgroups from [&marki] = NODES LGroupnum = component_id;
	quit;
	
	/* Turn off conditional setinit for OR and OptModel */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;
%mend MinimumSpanningTree;

%macro SkeletonMST(rf_indata=, lod_indata=, sdata=, marki=, markj=, cost=, rfthresh=, lodcost=, lodthresh=, lgthresh=)/store secure;
    /*--- check if OR is licensed; if not, conditionally setinit ---*/
    data _check_;
	or5_enabled  = sysprod('PRODNUM5'  );
   	or164_enabled= sysprod('PRODNUM164');
    	if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
    	if or164_enabled ne 1 then rc=zzqbxwf(164,1);
    run;
	
	proc sort data=&rf_indata;
		by &cost;
	run;
	
	proc optmodel;
		set <str>     NODES;                                                                                                                   
		set <str,str> EDGES;
		num cost {EDGES};
		set <str,str> TREE_EDGES init {};

		/* read edges from recombination data set */
		read data &rf_indata into EDGES=[&marki &markj] cost=&cost;
		*print cost;
		
		NODES = union {<i,j> in EDGES} {i, j};
		num component_id {NODES};

		/* read component_id in from skeleton data */
		read data &sdata into [&marki] component_id=_skeleton_;
		num num_components init max {i in NODES} component_id[i];                                                                                                          
		set <str> COMPONENT {1..num_components} init {};

		num c;
		for {i in NODES} do;
			c = component_id[i];
			COMPONENT[c] = COMPONENT[c] union {i};
		end;
	*	put component_id[*]=;
	*	put COMPONENT[*]=;
	*	put NODES=;

		/* if i and j are in different components, merge the two components and add edge to tree */
		num ci;
		num cj;
	/*	set TEMP_EDGES init EDGES;
		set TEMP_EDGES init {<i,j> in EDGES: component_id[i] > &lgthresh and component_id[j] > &lgthresh};
	*/
		set SKELETON = {i in NODES: component_id[i] <= &lgthresh};
		set TEMP_EDGES init {<i,j> in EDGES: i not in SKELETON or j not in SKELETON};
	*	put TEMP_EDGES=;
	/*	num min_cost = min {<i,j> in TEMP_EDGES} cost[i,j];*/
		num min_cost;
		%if %length(&lod_indata) %then %do;	
			set <str,str> LODEDGES;
			num lodcost {LODEDGES};		
			read data &lod_indata into LODEDGES=[&marki &markj] lodcost=&lodcost;
			*print lodcost;		
			num max_lodcost;
			do until ((card(setof {n in NODES} component_id[n]) <= &lgthresh) or (min_cost >= &rfthresh) or (max_lodcost <= &lodthresh));
		/*		for {<i,j> in TEMP_EDGES: cost[i,j] = min_cost} do; */
				for {<i,j> in TEMP_EDGES} do;
					min_cost = cost[i,j];	
					max_lodcost = lodcost[i,j];
		*			put i= j=;
		*			put min_cost=;
					ci = component_id[i]; 
					cj = component_id[j];	
		*			put "ci cj =" ci cj;
					if ((min_cost <= &rfthresh) and (max_lodcost >= &lodthresh)) then do; 

						if (ci ne cj) and (ci > &lgthresh or cj >&lgthresh) then do;
			*				put "Change ci cj =" ci cj;
							/* update the smaller numbered component to add to skeleton groups*/
							if ci > cj then do;
								for {k in COMPONENT[ci]} component_id[k] = cj;
								COMPONENT[cj] = COMPONENT[cj] union COMPONENT[ci];
								COMPONENT[ci] = {};
							end;
							else do;
								for {k in COMPONENT[cj]} component_id[k] = ci;
								COMPONENT[ci] = COMPONENT[ci] union COMPONENT[cj];
								COMPONENT[cj] = {};
							end;

							/* add edge to tree */
							/* print component_id */
							TREE_EDGES = TREE_EDGES union {<i,j>};

						end;
						TEMP_EDGES = TEMP_EDGES diff {<i,j>};
						leave;
					end;
					TEMP_EDGES = TEMP_EDGES diff {<i,j>};
				end;
			end;
		%end;
		%else %do;
			do until ((card(setof {n in NODES} component_id[n]) <= &lgthresh) or (min_cost >= &rfthresh));
		/*		for {<i,j> in TEMP_EDGES: cost[i,j] = min_cost} do; */
				for {<i,j> in TEMP_EDGES} do;
					min_cost = cost[i,j];	
		*			put i= j=;
		*			put min_cost=;
					ci = component_id[i]; 
					cj = component_id[j];	
		*			put "ci cj =" ci cj;
					if (ci ne cj) and (ci > &lgthresh or cj >&lgthresh) then do;
			*			put "Change ci cj =" ci cj;
						/* update the smaller numbered component to add to skeleton groups*/
						if ci > cj then do;
							for {k in COMPONENT[ci]} component_id[k] = cj;
							COMPONENT[cj] = COMPONENT[cj] union COMPONENT[ci];
							COMPONENT[ci] = {};
						end;
						else do;
							for {k in COMPONENT[cj]} component_id[k] = ci;
							COMPONENT[ci] = COMPONENT[ci] union COMPONENT[cj];
							COMPONENT[cj] = {};
						end;

						/* add edge to tree */
						/* print component_id */
						TREE_EDGES = TREE_EDGES union {<i,j>};

					end;
					TEMP_EDGES = TEMP_EDGES diff {<i,j>};
					leave;
				end;
			end;		
		%end;
		
		/*
		put "minimum spanning tree has cost " (sum {<i,j> in TREE_EDGES} cost[i,j]);
		*/
		/* write tree edges to data set */
		/*create data &outdata from [&marki &markj]=TREE_EDGES &cost=cost; */
		create data outgroups from [&marki] = NODES LGroupnum = component_id;
	quit;
	/* Turn off conditional setinit for OR and OptModel */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;
%mend SkeletonMST;

%macro FeatureOptimization /store secure;

	data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
		or164_enabled= sysprod('PRODNUM164');
		 if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
		 if or164_enabled ne 1 then rc=zzqbxwf(164,1);
	run;

	proc optmodel;
		set FEATURES;
		str feature_name {FEATURES};
		read data trans into FEATURES=[_N_] feature_name=_name_;

		num a {FEATURES};
		set PAIRS = {i in FEATURES, j in FEATURES: i < j};
		num b {PAIRS};
		read data upper_triangular(firstobs=2) nomiss into [i=_N_] a=&UseDep {j in FEATURES} <b[i,j]=col(feature_name[j])>;
		
		%if %index(&Opt_Algorithm,Greedy) %then %do;

			put 'Running greedy forward selection...';
			set SELECTED;
			set SELECTED_PAIRS = {i in SELECTED, j in SELECTED: i < j};

			%if %index(&Opt_Criterion,Minimum) %then %do;
				* mRMR;
				max CRITERION = (sum {i in SELECTED} a[i])/card(SELECTED) - 
					(2*sum {<i,j> in SELECTED_PAIRS} b[i,j] + card(SELECTED))/(2*card(SELECTED_PAIRS) + card(SELECTED));
			%end;
			%else %do;
				* CFS;
				max CRITERION = (sum {i in SELECTED} a[i])^2 /
					(card(SELECTED) + 4 * sum {<i,j> in SELECTED_PAIRS} b[i,j]);
			%end;

			num max_CRITERION;
			num argmax_i;
			num iter;
			set INCUMBENT;

			INCUMBENT = {};
			SELECTED = {};
			max_CRITERION = -1;
			iter = 1;
			do while (1);
				argmax_i = -1;
				SELECTED = INCUMBENT;
				for {i in FEATURES diff INCUMBENT} do;
					SELECTED = SELECTED union {i};
					if max_CRITERION < CRITERION then do;
						max_CRITERION = CRITERION;
						argmax_i = i;
					end;
					SELECTED = SELECTED diff {i};
				end;
				if argmax_i = -1 then leave;
				else INCUMBENT = INCUMBENT union {argmax_i};
				put iter= max_CRITERION= INCUMBENT=;
				iter = iter+1;
			end;

			*print {i in INCUMBENT} feature_name[i];
			create data optc from [i]=INCUMBENT feature_name;
		%end;
		%else %do;

			put 'Running active set NLP...';
			/* X[i] = 1 if feature i is used, 0 otherwise */
			var X {FEATURES} binary;

			%if %index(&Opt_Criterion,Minimum) %then %do;
				* mRMR;
				/*
				max CRITERION = (sum {i in SELECTED} a[i])/card(SELECTED) - 
					(2*sum {<i,j> in SELECTED_PAIRS} b[i,j] + card(SELECTED))/(2*card(SELECTED_PAIRS) + card(SELECTED));
				*/
				max CRITERION = (sum {i in FEATURES} a[i] * X[i])/(sum {i in FEATURES} X[i]) -
					(2*sum {<i,j> in PAIRS} b[i,j] * X[i] * X[j] + sum {i in FEATURES} X[i])/(2*sum {<i,j> in PAIRS} X[i] * X[j] + sum {i in FEATURES} X[i]);
			%end;
			%else %do;
				* CFS;
				max CRITERION = (sum {i in FEATURES} a[i] * X[i])^2 
					/ (sum {i in FEATURES} X[i] + 4 * sum {<i,j> in PAIRS} b[i,j] * X[i] * X[j]);
			%end;

		   solve with NLP relaxint / algorithm=as hesstype=product;
		   put CRITERION=;
		   print feature_name X;
		   print {i in FEATURES: X[i].sol > 0.5} feature_name[i];
		   create data opta from [i] feature_name X;
		%end;		
	quit;
	
	/* Turn off conditional setinit for OR and OptModel */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;	
%mend;


/*Luciano: implementation of sas proc optnet for ordering markers.
Initial coding of proc optnet was written by Kelci Miclaus */
%macro PrepDataForOptNet(rf_indata=,lod_indata=,rfoutdata=, lodoutdata=) /store secure;
	/*--- check if OR is licensed; if not, conditionally setinit ---*/
    	data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
   		or164_enabled= sysprod('PRODNUM164');
    		if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,1);
	run;
	
	proc optmodel presolver=none;
	        set <str> VERTICES;
	        *set <str,str> EDGES = {i in VERTICES, j in VERTICES: i >= j}; * check for bug with self-link *;
	        set <str,str> EDGES = {i in VERTICES, j in VERTICES: i > j}; * using '>' rather than '>=' avoids bug with self-link *;
	        read data &rf_indata into VERTICES=[&IdVar];
	        VERTICES = VERTICES union {'DUMMY'};
	        num c {VERTICES, VERTICES} init 0;
	        read data &rf_indata into [&IdVar] {j in VERTICES diff {'DUMMY'}} <c[&IdVar,j]=col(j)>;

		%if %length(&lod_indata) %then %do;
			num d {VERTICES, VERTICES} init 1000;
			read data &lod_indata into [&IdVar] {j in VERTICES diff {'DUMMY'}} <d[&IdVar,j]=col(j)>;	
			*print d;
			create data &rfoutdata from [i j]={<i,j> in EDGES : c[i,j]<=&recthreshold and d[i,j]>=&lodthreshold} c=c[i,j];
			***create lodlinks for debug only;
			*create data &lodoutdata from [i j]={<i,j> in EDGES : c[i,j]<=&recthreshold and d[i,j]>=&lodthreshold} d=d[i,j];
		%end;
		%else %do;
	        	create data &rfoutdata from [i j]={<i,j> in EDGES : c[i,j]<=&recthreshold} c=c[i,j];
	        	*create data &rfoutdata from [i j]={<i,j> in EDGES : c[i,j]<=1} c=c[i,j];
	        %end;
	quit;
	
	/*S1056123: out_nodes from PROC OPTNET truncates names of nodes if DUMMY 
	has smaller length than marker names. Therefore, if length of marker names is less 
	than the number of characters in DUMMY (which is 5), the out_nodes results will be 
	truncated and it causes problems.*/
	data &rfoutdata;
		length i j $32;
		set &rfoutdata;
	run;
	
	
	/*
	***create lodlinks for debug only***;
	%if %length(&lod_indata) %then %do;
		data &lodoutdata;
			length i j $32;
			set &lodoutdata;
		run;	
	%end;
	*/
	
	/* Turn off conditional setinit for OR and OptNet */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;
	
%mend PrepDataForOptNet;

%macro SolveWithOptNet(indata=, milp=) /store secure;
	/*--- check if OR is licensed; if not, conditionally setinit ---*/
    	data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
   		or164_enabled= sysprod('PRODNUM164');
    		if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,1);
	run;
	
	proc optnet
		loglevel   = 2
		out_nodes  = TSPSolutionNodes   
		data_links = &indata;
	   	data_links_var
	      	from    = i
	      	to      = j
	      	weight  = c;
	   	tsp	
	      	out     = TSPSolutionLinks
	      	MILP	= &milp
	      	%if %upcase(&UseStoppingRules) = YES %then &relobjgap2 &maxtime;
	      	;
	run;
	%*put &_OROPTNET_;
	%*put &_OROPTNET_TSP_;
	
	%global _OROPTNET_macrovar _OROPTNET_TSP_macrovar ;
	%let _OROPTNET_macrovar= &_OROPTNET_;
	%put _OROPTNET_macrovar= &_OROPTNET_macrovar;
	%let _OROPTNET_TSP_macrovar= &_OROPTNET_TSP_;
	%put _OROPTNET_TSP_macrovar= &_OROPTNET_TSP_macrovar;
						
	/*
	%let bstatus= %index(&_OROPTNET_macrovar, STATUS=);
	%let btsp= %index(&_OROPTNET_macrovar, TSP=);
	%let status1= %substr(&_OROPTNET_macrovar, &bstatus + 7, &btsp - &bstatus - 7);
	%put status1= &status1;
	
	%let bstatus= %index(&_OROPTNET_TSP_macrovar, STATUS=);
	%let btsp= %index(&_OROPTNET_TSP_macrovar, TSP=);
	%let status1= %substr(&_OROPTNET_macrovar, &bstatus + 7, &btsp - &bstatus - 7);
	%put status1= &status1;	
		
	%global exiterror;
	%if %upcase(&status1)=OUT_OF_MEMORY or %upcase(&status1)=ERROR or %upcase(&status2)=INFEASIBLE %then %do;
		%let exiterror=1;
	%end;
	*/
	
	%global exiterror;
	%let exiterror=0;
	%if %sysfunc(find(%upcase(&_OROPTNET_), OUT_OF_MEMORY)) or 
		%sysfunc(find(%upcase(&_OROPTNET_), ERROR)) or
		%sysfunc(find(%upcase(&_OROPTNET_), INFEASIBLE)) 
		%then %do;
		%let exiterror=1;
	%end;	
	
	/* Turn off conditional setinit for OR and OptNet */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;
	
%mend SolveWithOptNet;

%macro GetNetOrd(indata=) / store secure;
	proc iml;
		use &indata;
		read all var {i j} into ordin;
		nrn= nrow(ordin);
		close &indata;
		ordout= ordin[1,1] // ordin[1,2];
		nrv= nrow(ordout);
		ordin= ordin[2:nrn,];
		nrn= nrow(ordin);
		do while (nrn > 0);
			l1= loc(ordin[,1]=ordout[1]);
			l2= loc(ordin[,2]=ordout[1]);
			r1= loc(ordin[,1]=ordout[nrv]);
			r2= loc(ordin[,2]=ordout[nrv]);
			if nrow(l1) > 0 then do;
				ordout = ordin[l1[1],2] // ordout;
				delrow= l1[1];
			end;
			else if nrow(l2) > 0 then do;
				ordout = ordin[l2[1],1] // ordout;
				delrow= l2[1];
			end;
			else if nrow(r1) > 0 then do;
				ordout = ordout // ordin[r1[1],2];
				delrow= r1[1];
			end;
			else if nrow(r2) > 0 then do;
				ordout = ordout // ordin[r2[1],1];
				delrow= r2[1];
				end;
		
			if nrn > 1 then do;
			nrv= nrow(ordout);
			ordin= ordin[setdif(1:nrn, delrow),];
			nrn= nrow(ordin);
			end;
			else do;
				nrn = 0;
			end;
		end;
		
		create solordnet from ordout[colname="&IdVar"];
		append from ordout;
		close solordnet;
	quit;
	
	data solordnet;
		set solordnet;
		order= _n_;
	run;
	
%mend GetNetOrd;

%macro LMOOptNet(rf_indata=,lod_indata=,milp=on) /store secure;
	/*--- check if OR is licensed; if not, conditionally setinit ---*/
    	data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
		or164_enabled= sysprod('PRODNUM164');
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
		if or164_enabled ne 1 then rc=zzqbxwf(164,1);
	run;
	
	%*let start=%sysfunc(time());	

	/*data OutLib.lmo_rf_indata; set &rf_indata; run;*/
	/*data OutLib.lmo_lod_indata; set &lod_indata; run;*/
	
	%PrepDataForOptNet(rf_indata=&rf_indata,lod_indata=&lod_indata,rfoutdata=rfdata_links, lodoutdata=loddata_links);	
	
	/*data OutLib.lmo_rfdata_links; set rfdata_links; run;*/
	
	/*
	***lodlinks for debug only***;
	%if %length(&lod_indata) %then %do;
		data OutLib.lmo_loddata_links; set loddata_links; run;
	%end;
	*/
	
	%SolveWithOptNet(indata=rfdata_links, milp=&milp);
	%*let stop=%sysfunc(time());
	%*put Total Time =  %sysevalf(&stop-&start);
	
	%if %sysfunc(exist(solData2)) %then %do;
		proc datasets nolist;
			delete solData2;
		quit;
	%end;
	
	%if ^&exiterror %then %do; 
	
		/*Using TSPSolutionLinks to assemble linear order of markers. IML code. This strategy
		does need a dummy marker*/
		/*
		data solData;
			set TSPSolutionLinks;
			where i ne 'DUMMY' and j ne 'DUMMY';
		run; 

		%GetNetOrd(indata=solData);  
		*/

		/*Using TSPSolutionNodes to assemble linear order of markers. This strategy works 
		only when there is a DUMMY marker*/

		/*data OutLib.lmo_TSPSolutionNodes; set TSPSolutionNodes; run;*/

		proc sort data=TSPSolutionNodes out=solordnet;
			by tsp_order;
		run;

		data solordnet;
			set solordnet end=last;
			if node eq 'DUMMY' then call symputx("_dummy_", trim(left(tsp_order)));
			if last then do;
				_nobs=_n_;
				call symputx("_nobs_", trim(left(_nobs)));
			end;
		run;

		data solordnet;
			set solordnet;
			retain count;
			if tsp_order <= &_dummy_ then count= 1;
			if tsp_order < &_dummy_ then order= &_dummy_ - _n_;
			else if tsp_order > &_dummy_ then do;
				order= &_nobs_ - count;
				count= count + 1;
			end;
			if node eq "DUMMY" then delete;
			rename node = &IdVar;
		run;

		proc sort data=solordnet(keep= &IdVar order);
			by order;
		run;	

		data orgmarkerorder;
			set &rf_indata;
			_orgord_=_n_;
			keep &IdVar _orgord_ ;
			rename &IdVar= node;
		run;

		proc sql;
			create table solData2(drop= node _orgord_) as 
			select * 
			from solordnet as b, orgmarkerorder as a 
			where a.node= b.&IdVar
			order a._orgord_;
		quit;

		proc datasets nolist;
			delete TSPSolutionLinks
				TSPSolutionNodes
				solordnet
				rfdata_links
				/*
				%if %length(&lod_indata) %then %do;
					loddata_links
				%end;
				*/
				orgmarkerorder;
		quit;
	
	%end;

	/* Turn off conditional setinit for OR and OptNet */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;
	  		
%mend LMOOptNet;

/* Given finite subset of R^d, find the subset of Pareto-optimal points */
/* (assume minimization for all components)                             */
/* Author: Rob Pratt, SAS/OR                                            */
/* Updated: Luciano Silva, JMPLS                                        */
/* indata data set contains n observations with variables x1-xd  */
/* outdata data set contains subset consisting of Pareto points  */
%macro FindParetoPoints(indata=, outdata=, outorder=) /store secure; /* This macro is encrypted */

	/*--- check if OR is licensed; if not, conditionally setinit ---*/
	data _check_;
		or5_enabled  = sysprod('PRODNUM5'  );
		or164_enabled= sysprod('PRODNUM164');
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,1);
		if or164_enabled ne 1 then rc=zzqbxwf(164,1);
	run;

	data sum;
		set &indata;
		sum = sum(of x:);
		_orig_order_= _n_;
	run;

	proc sort data=sum out=sorted;
		by sum;
	run;

	data sorted(keep= x:) &outorder(keep= _orig_order_ _sort_order_);
		set sorted;
		_sort_order_= _n_;
	run;

	proc optmodel;
		/* declare parameters and read data */
		set ISET;
		*num d = attrn(open("&indata"), "nvars");
		num d = attrn(open("sorted"), "nvars");
		set JSET = 1..d;
		num x {ISET, JSET};
		read data sorted into ISET=[_N_] {j in JSET} <x[_N_,j]=col('x'||j)>;
		set SUBSET init ISET;
		num dom;

		/* compare each i in ISET to each s in SUBSET to determine if dominated */
		/* since input data is sorted by sum, s cannot dominate i */
		for {i in ISET} do;
			if i not in SUBSET then continue;
			for {s in SUBSET: i < s} do;
				/* if i dominates s, remove s from SUBSET */
				dom = 1;
				for {j in JSET} do;
					if x[i,j] > x[s,j] then do;
						dom = 0;
						leave;
					end;
				end;
				if dom = 1 then SUBSET = SUBSET diff {s};
			end;
		end;

		/* write output to data set */
		create data &outdata(/*drop=i*/ rename= (i= pareto_points)) 
		from [i]=SUBSET {j in JSET} <col('x'||j)=x[i,j]>;
	quit;

	proc datasets nolist;
		delete sum sorted;
	quit;
	
	/* Turn off conditional setinit for OR and OptModel */
	data _null_;
		set _check_;
		if or5_enabled   ne 1 then rc=zzqbxwf(  5,0);
    		if or164_enabled ne 1 then rc=zzqbxwf(164,0);
	run;	

%mend FindParetoPoints;
